#define GE_USE_THREADS

/*
	description:

		"C declarations for the Gobo Eiffel runtime."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2005-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_EIFFEL_H
#define GE_EIFFEL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/* Class name mapping as defined in the FreeELKS library. */
#ifndef EIF_INTEGER
#define EIF_INTEGER EIF_INTEGER_32
#endif
#ifndef EIF_CHARACTER
#define EIF_CHARACTER EIF_CHARACTER_8
#endif
#ifndef EIF_REAL
#define EIF_REAL EIF_REAL_32
#endif
#ifndef EIF_DOUBLE
#define EIF_DOUBLE EIF_REAL_64
#endif
#ifndef GE_ms
#define GE_ms(s,c) GE_ms8((s),(c))
#endif

#if defined(__USE_POSIX) || defined(__unix__) || defined(_POSIX_C_SOURCE)
#include <unistd.h>
#endif
#if !defined(WIN32) && \
	(defined(WINVER) || defined(_WIN32_WINNT) || defined(_WIN32) || \
	defined(__WIN32__) || defined(__TOS_WIN__) || defined(_MSC_VER) || \
	defined(__MINGW32__))
#define WIN32 1
#endif
#ifdef WIN32
#define EIF_WINDOWS 1
#include <windows.h>
#endif

#define BYTEORDER 0x1234

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stddef.h>

/* Platform definition */
/* Unix definition */
#define EIF_IS_UNIX EIF_TRUE
/* Windows definition */
#ifdef EIF_WINDOWS
#define EIF_IS_WINDOWS EIF_TRUE
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_WINDOWS EIF_FALSE
#endif
/* VMS definition */
#ifdef EIF_VMS
#define EIF_IS_VMS EIF_TRUE
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_VMS EIF_FALSE
#endif
/* MAC definition */
#ifdef EIF_MAC
#define EIF_IS_MAC EIF_TRUE
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_MAC EIF_FALSE
#endif
/* VxWorks definition */
#ifdef EIF_VXWORKS
#define EIF_IS_VXWORKS EIF_TRUE
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_VXWORKS EIF_FALSE
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _MSC_VER /* MSVC */
typedef signed char int8_t;
typedef signed short int16_t;
typedef signed int int32_t;
typedef signed __int64 int64_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned __int64 uint64_t;
#else
#if defined (__BORLANDC__) && (__BORLANDC__ < 0x600) /* Borland before 6.0 */
typedef signed char int8_t;
typedef signed short int16_t;
typedef signed long int int32_t;
typedef signed __int64 int64_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long int uint32_t;
typedef unsigned __int64 uint64_t;
#else
#include <inttypes.h>
#endif
#endif

/* Portable integer pointers */
#ifdef EIF_WINDOWS
#ifndef _INTPTR_T_DEFINED
#define _INTPTR_T_DEFINED
#ifdef _WIN64
typedef __int64 intptr_t;
#else
typedef int intptr_t;
#endif
#endif
#ifndef _UINTPTR_T_DEFINED
#define _UINTPTR_T_DEFINED
#ifdef _WIN64
  typedef unsigned __int64 uintptr_t;
#else
  typedef unsigned int uintptr_t;
#endif
#endif
#endif

/* C type for underlying integer type identifying object's dynamic type. */
typedef uint16_t EIF_TYPE_INDEX;

/*
 * Abstraction representing an Eiffel type.
 * It is made of a compiler type-id,
 * and of some annotations (attached/detachable/separate/variant/frozen).
 */
typedef struct eif_type {
	EIF_TYPE_INDEX id;
	EIF_TYPE_INDEX annotations;
} EIF_TYPE;

/*
 * Since EIF_TYPE and EIF_ENCODED_TYPE have the same size, the encoded version
 * is basically a memcpy version of the EIF_TYPE representation.
 * It is used to provide backward compatibility to most Eiffel and
 * C APIs manipulating types as an INTEGER.
 */
typedef int32_t EIF_ENCODED_TYPE;
typedef EIF_ENCODED_TYPE EIF_TYPE_ID;
#define EIF_NO_TYPE (EIF_TYPE_ID)(-1)

/* Basic Eiffel types */
typedef char EIF_BOOLEAN;
typedef unsigned char EIF_CHARACTER_8;
typedef uint32_t EIF_CHARACTER_32;
typedef int8_t EIF_INTEGER_8;
typedef int16_t EIF_INTEGER_16;
typedef int32_t EIF_INTEGER_32;
typedef int64_t EIF_INTEGER_64;
typedef uint8_t EIF_NATURAL_8;
typedef uint16_t EIF_NATURAL_16;
typedef uint32_t EIF_NATURAL_32;
typedef uint64_t EIF_NATURAL_64;
typedef void* EIF_POINTER;
typedef float EIF_REAL_32;
typedef double EIF_REAL_64;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags;} EIF_ANY;
typedef EIF_ANY* EIF_REFERENCE;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags; EIF_REFERENCE area; EIF_INTEGER count;} EIF_STRING;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags; uint32_t offset; EIF_INTEGER count; EIF_INTEGER capacity;} EIF_SPECIAL;

#ifdef EIF_WINDOWS
typedef wchar_t EIF_NATIVE_CHAR;
#else
typedef char EIF_NATIVE_CHAR;
#endif
typedef EIF_NATIVE_CHAR* EIF_FILENAME;

#define EIF_VOID ((EIF_REFERENCE)0)
#define EIF_FALSE ((EIF_BOOLEAN)'\0')
#define EIF_TRUE ((EIF_BOOLEAN)'\1')
#define EIF_TEST(x) ((x) ? EIF_TRUE : EIF_FALSE)

#define EIF_IS_WORKBENCH EIF_FALSE
#define EIF_POINTER_DISPLAY "lX"

/* For INTEGER and NATURAL manifest constants */
#define GE_int8(x) x
#define GE_nat8(x) x
#define GE_int16(x) x
#define GE_nat16(x) x
#define GE_int32(x) x##L
#define GE_nat32(x) x##U
#if defined (_MSC_VER) && (_MSC_VER < 1400) /* MSC older than v8 */
#define GE_int64(x) x##i64
#define GE_nat64(x) x##ui64
#elif defined(__BORLANDC__) && (__BORLANDC__ < 0x600) /* Borland before 6.0 */
#define GE_int64(x) x##i64
#define GE_nat64(x) x##ui64
#else /* ISO C 99 */
#define GE_int64(x) x##LL
#define GE_nat64(x) x##ULL
#endif
#ifdef __LCC__
/* lcc-win32 reports a constant overflow for -21474836478. */
#define GE_min_int32 (-GE_int32(2147483647)-GE_int32(1))
#else
#define GE_min_int32 GE_int32(-2147483648)
#endif
#define GE_max_int32 GE_int32(2147483647)
#if defined(__LCC__) || defined(__GNUC__) || defined(__MINGW32__)
/* lcc-win32 reports a constant overflow for -9223372036854775808. */
/* gcc and mingw-win64 warn that integer constant is so large that it is unsigned. */
#define GE_min_int64 (-GE_int64(9223372036854775807)-GE_int64(1))
#else
#define GE_min_int64 GE_int64(-9223372036854775808)
#endif
#if defined(__LCC__) && !defined(_WIN64)
/* lcc-win32 does not consider 64 bit constants as constants in case statement. */
#define GE_case_int64(x) ((int32_t)(x))
#define GE_case_nat64(x) ((uint32_t)(x))
#else
#define GE_case_int64(x) (x)
#define GE_case_nat64(x) (x)
#endif

#ifdef _WIN64
#define GE_IS_64_BITS EIF_TRUE
#else
#define GE_IS_64_BITS EIF_TEST(sizeof(void*)==64)
#endif

/* Posix threads */
#if !defined(EIF_WINDOWS)
#define GE_USE_POSIX_THREADS
#endif

#ifdef _MSC_VER /* MSVC */
/* MSVC does not support ISO C 99's 'snprintf' from stdio.h */
#define snprintf(a,b,c,d) sprintf(a,c,d)
#endif

/*
 * Gobo compiler version.
 * Starts with 6080 (looks like GOBO) followed by 5 digits.
 */
#define GE_compiler_version() 608000001

/*
	Interoperability with ISE.
*/
#define RTI64C(x) GE_int64(x)
#define EIF_OBJECT EIF_REFERENCE
#define EIF_OBJ EIF_OBJECT
#define OVERHEAD sizeof(EIF_ANY)
/* Function pointer call to make sure all arguments are correctly pushed onto stack. */
/* FUNCTION_CAST is for standard C calls. */
/* FUNCTION_CAST_TYPE is for non-standard C calls. */
#define FUNCTION_CAST(r_type,arg_types) (r_type (*) arg_types)
#define FUNCTION_CAST_TYPE(r_type,call_type,arg_types) (r_type (call_type *) arg_types)
#define SIGBLOCK
#define SIGRESUME
#define rt_public				/* default C scope */
#define rt_private static		/* static outside a block means private */
#define rt_shared				/* data shared between modules, but not public */
typedef intptr_t rt_int_ptr;
typedef uintptr_t rt_uint_ptr;
#define RTMS(s) GE_str8(s)
#define RTMS_EX(s,c) GE_ms8((s),(c))

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C types used to implement class THREAD and related threading facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_THREAD_TYPES_H
#define GE_THREAD_TYPES_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef GE_USE_THREADS

#define EIF_THREADS
#ifdef GE_USE_POSIX_THREADS
#define EIF_POSIX_THREADS
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef GE_USE_POSIX_THREADS
#include <pthread.h>
#include <semaphore.h>
#elif defined EIF_WINDOWS
#include <windows.h>
#include <process.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_POSIX_THREADS

#define EIF_THR_TYPE            pthread_t
#define EIF_CS_TYPE				pthread_mutex_t
#define EIF_MUTEX_TYPE          pthread_mutex_t
#define EIF_COND_TYPE			pthread_cond_t
#define EIF_SEM_TYPE			sem_t
#define EIF_RWL_TYPE			pthread_rwlock_t

#elif defined EIF_WINDOWS

#define EIF_THR_TYPE            HANDLE
#define EIF_CS_TYPE				CRITICAL_SECTION
#define EIF_MUTEX_TYPE          CRITICAL_SECTION
#define EIF_SEM_TYPE            HANDLE

typedef struct {
		/* Semaphore used to queue up threads waiting for the condition to become signaled. */
	EIF_SEM_TYPE* semaphore;
		/* Serialize access to fields of Current. */
	EIF_CS_TYPE* csection;
		/* Number of waiters. */
	unsigned long num_waiting;
		/* Number of already awoken. */
	unsigned long num_wake;
		/* Number of time we signaled/broadcasted for improving fairness.
		 * This ensures one thread won't steal wakeups from other threads in queue. */
	unsigned long generation;
} EIF_COND_TYPE;

typedef struct {
	EIF_MUTEX_TYPE* m; /* Internal monitor lock. */
	int rwlock; /* >0 = # readers, <0 = writer, 0 = none */
	EIF_COND_TYPE* readers_ok; /* Start waiting readers. */
	unsigned int waiting_writers; /* Number of waiting writers. */
	EIF_COND_TYPE* writers_ok; /* Start a waiting writer. */
} EIF_RWL_TYPE;

#endif

typedef struct {
	unsigned int priority;
	unsigned int stack_size;
} EIF_THR_ATTR_TYPE;

/* Struct for thread context. */
typedef struct GE_thread_context_struct GE_thread_context;
struct GE_thread_context_struct {
	EIF_THR_TYPE thread_id; /* Thread identifier for associated thread. */
	EIF_REFERENCE current; /* Eiffel root object. */
	void (*routine)(EIF_REFERENCE); /* Eiffel routine. */
	void (*set_terminated)(EIF_REFERENCE,EIF_BOOLEAN); /* Eiffel routine to set {THREAD}.terminated. */
	unsigned int initial_priority; /* Initial priority. */
	EIF_THR_TYPE last_thread_id; /* Last thread created from current thread. */
	volatile int n_children; /* Number of direct thread children. */
	EIF_MUTEX_TYPE* children_mutex; /* Mutex to wait for thread children. */
	EIF_COND_TYPE* children_cond; /* Condition variable to wait for thread children. */
	GE_thread_context* parent_context;	/* Context of parent thread, NULL if main thread. */
	int thread_exiting; /* Has current thread already called GE_thread_exit? */
	volatile int is_alive; /* Is Current thread still alive? */
};

#ifdef __cplusplus
}
#endif

#endif
#endif

/*
	description:

		"C functions used to implement once features"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_ONCE_H
#define GE_ONCE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Struct to keep track of the call status
 * and results of once features.
 */
typedef struct {
	EIF_BOOLEAN* boolean_value;
	EIF_REFERENCE* boolean_exception;
	unsigned char* boolean_status;
	unsigned int boolean_count;
	EIF_CHARACTER_8* character_8_value;
	EIF_REFERENCE* character_8_exception;
	unsigned char* character_8_status;
	unsigned int character_8_count;
	EIF_CHARACTER_32* character_32_value;
	EIF_REFERENCE* character_32_exception;
	unsigned char* character_32_status;
	unsigned int character_32_count;
	EIF_INTEGER_8* integer_8_value;
	EIF_REFERENCE* integer_8_exception;
	unsigned char* integer_8_status;
	unsigned int integer_8_count;
	EIF_INTEGER_16* integer_16_value;
	EIF_REFERENCE* integer_16_exception;
	unsigned char* integer_16_status;
	unsigned int integer_16_count;
	EIF_INTEGER_32* integer_32_value;
	EIF_REFERENCE* integer_32_exception;
	unsigned char* integer_32_status;
	unsigned int integer_32_count;
	EIF_INTEGER_64* integer_64_value;
	EIF_REFERENCE* integer_64_exception;
	unsigned char* integer_64_status;
	unsigned int integer_64_count;
	EIF_NATURAL_8* natural_8_value;
	EIF_REFERENCE* natural_8_exception;
	unsigned char* natural_8_status;
	unsigned int natural_8_count;
	EIF_NATURAL_16* natural_16_value;
	EIF_REFERENCE* natural_16_exception;
	unsigned char* natural_16_status;
	unsigned int natural_16_count;
	EIF_NATURAL_32* natural_32_value;
	EIF_REFERENCE* natural_32_exception;
	unsigned char* natural_32_status;
	unsigned int natural_32_count;
	EIF_NATURAL_64* natural_64_value;
	EIF_REFERENCE* natural_64_exception;
	unsigned char* natural_64_status;
	unsigned int natural_64_count;
	EIF_POINTER* pointer_value;
	EIF_REFERENCE* pointer_exception;
	unsigned char* pointer_status;
	unsigned int pointer_count;
	EIF_REAL_32* real_32_value;
	EIF_REFERENCE* real_32_exception;
	unsigned char* real_32_status;
	unsigned int real_32_count;
	EIF_REAL_64* real_64_value;
	EIF_REFERENCE* real_64_exception;
	unsigned char* real_64_status;
	unsigned int real_64_count;
	EIF_REFERENCE* reference_value;
	EIF_REFERENCE* reference_exception;
	unsigned char* reference_status;
	unsigned int reference_count;
	EIF_REFERENCE* procedure_exception;
	unsigned char* procedure_status;
	unsigned int procedure_count;
} GE_onces;

/*
 * Variable to keep track of the call status
 * and results of once-per-process features.
 */
extern GE_onces* GE_process_onces;

/*
 * Initialize `GE_process_onces'.
 */
extern void GE_init_onces(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Create a new 'GE_onces' struct which can deal with the
 * numbers of once features passed as argument.
 */
extern GE_onces* GE_new_onces(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Free memory allocated by `a_onces'.
 */
extern void GE_free_onces(GE_onces* a_onces);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class EXCEPTION"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_EXCEPTION_H
#define GE_EXCEPTION_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_ONCE_H
#include "ge_once.h"
#endif
#ifdef GE_USE_THREADS
#ifndef GE_THREAD_TYPES_H
#include "ge_thread_types.h"
#endif
#endif

#include <setjmp.h>

/*
 * On Linux glibc systems, we need to use sig* versions of jmp_buf,
 * setjmp and longjmp to preserve the signal handling context.
 * One way to detect this is if _SIGSET_H_types has
 * been defined in /usr/include/setjmp.h.
 * NOTE: ANSI only recognizes the non-sig versions.
 */
#if (defined(_SIGSET_H_types) && !defined(__STRICT_ANSI__))
#define GE_jmp_buf sigjmp_buf
#define GE_setjmp(x) sigsetjmp((x),1)
#define GE_longjmp(x,y) siglongjmp((x),(y))
#else
#define GE_jmp_buf jmp_buf
#define GE_setjmp(x) setjmp((x))
#define GE_longjmp(x,y) longjmp((x),(y))
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Predefined exception codes.
 */
#define GE_EX_VOID		1			/* Feature applied to void reference */
#define GE_EX_MEM		2			/* No more memory */
#define GE_EX_PRE		3			/* Pre-condition violated */
#define GE_EX_POST		4			/* Post-condition violated */
#define GE_EX_FLOAT		5			/* Floating point exception (signal SIGFPE) */
#define GE_EX_CINV		6			/* Class invariant violated */
#define GE_EX_CHECK		7			/* Check instruction violated */
#define GE_EX_FAIL		8			/* Routine failure */
#define GE_EX_WHEN		9			/* Unmatched inspect value */
#define GE_EX_VAR		10			/* Non-decreasing loop variant */
#define GE_EX_LINV		11			/* Loop invariant violated */
#define GE_EX_SIG		12			/* Operating system signal */
#define GE_EX_BYE		13			/* Eiffel run-time panic */
#define GE_EX_RESC		14			/* Exception in rescue clause */
#define GE_EX_OMEM		15			/* Out of memory (cannot be ignored) */
#define GE_EX_RES		16			/* Resumption failed (retry did not succeed) */
#define GE_EX_CDEF		17			/* Create on deferred */
#define GE_EX_EXT		18			/* External event */
#define GE_EX_VEXP		19			/* Void assigned to expanded */
#define GE_EX_HDLR		20			/* Exception in signal handler */
#define GE_EX_IO		21			/* I/O error */
#define GE_EX_SYS		22			/* Operating system error */
#define GE_EX_RETR		23			/* Retrieval error */
#define GE_EX_PROG		24			/* Developer exception */
#define GE_EX_FATAL		25			/* Eiffel run-time fatal error */
#define GE_EX_DOL		26			/* $ applied to melted feature */
#define GE_EX_ISE_IO	27			/* I/O error raised by the ISE Eiffel runtime */
#define GE_EX_COM		28			/* COM error raised by EiffelCOM runtime */
#define GE_EX_RT_CHECK	29			/* Runtime check error such as out-of-bound array access */
#define GE_EX_OLD		30			/* Old violation */
#define GE_EX_SEL		31			/* Serialization failure */
#define GE_EX_DIRTY		32			/* SCOOP processor dirty exception. */
#define GE_EX_NEX		32			/* Number of internal exceptions */

/*
 * String buffer used to build the exception trace.
 */
typedef struct GE_exception_trace_buffer_struct GE_exception_trace_buffer;
struct GE_exception_trace_buffer_struct {
	char* area;
	uint32_t count;
	uint32_t capacity;
};

/*
 * Information about the feature being executed.
 */
typedef struct GE_call_struct GE_call;
struct GE_call_struct {
#ifdef GE_USE_CURRENT_IN_EXCEPTION_TRACE
	void* object; /* Current object */
#endif
	const char* class_name;
	const char* feature_name;
	GE_call* caller; /* previous feature in the call chain */
};

/*
 * Context of features containing a rescue clause.
 */
typedef struct GE_rescue_struct GE_rescue;
struct GE_rescue_struct {
	GE_jmp_buf jb;
	GE_rescue* previous; /* previous context in the call chain */
};

/*
 * Information about the execution context.
 * One such struct per thread.
 */
typedef struct GE_context_struct GE_context;
struct GE_context_struct {
	GE_call* call; /* Call stack */
	uint32_t in_assertion; /* Is an assertion evaluated? */
	GE_rescue* last_rescue; /* Context of last feature entered containing a rescue clause */
	uint32_t in_rescue; /* Number of rescue clauses currently being executed */
	EIF_REFERENCE exception_manager; /* Exception manager */
	char raising_exception; /* Is an exception currently being raised? */
	char exception_trace_enabled; /* Should exception trace be displayed? */
	long exception_code; /* Code of the exception currently being raised, 0 otherwise */
	const char* exception_tag; /* Tag of the exception currently being raised, NULL otherwise */
	GE_exception_trace_buffer exception_trace_buffer; /* String buffer used to build the exception trace */
	GE_exception_trace_buffer last_exception_trace; /* Last non-routine-failure exception trace */
	int pre_ecma_mapping_status; /* Do we map old names to new name? (i.e. STRING to STRING_8, INTEGER to INTEGER_32, ...). */
#ifdef GE_USE_THREADS
	GE_thread_context* thread; /* Thread context */
	GE_onces* process_onces; /* Cache for status and results of onces-per-process */
	GE_onces* thread_onces; /* Status and results of onces-per-thread */
	void* wel_per_thread_data; /* WEL private data */
#endif
};

/*
 * Default initialization for `GE_context'.
 */
extern GE_context GE_default_context;

/*
 * Execution context of main thread.
 */
extern GE_context* GE_main_context;

/*
 * Execution context of current thread.
 */
extern GE_context* GE_current_context(void);

/*
 * Initialization of exception handling.
 */
extern void GE_init_exception(GE_context* context);

/*
 * Free memory allocated in `a_context' for exception handling.
 */
extern void GE_free_exception(GE_context* a_context);

/*
 * Pointer to function to create a new exception manager object
 * (of type ISE_EXCEPTION_MANAGER).
 */
extern EIF_REFERENCE (*GE_new_exception_manager)(EIF_BOOLEAN);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.init_exception_manager.
 */
extern void (*GE_init_exception_manager)(GE_context*, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.last_exception.
 */
extern EIF_REFERENCE (*GE_last_exception)(GE_context*, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.once_raise.
 */
extern void (*GE_once_raise)(GE_context*, EIF_REFERENCE, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.set_exception_data.
 */
extern void (*GE_set_exception_data)(GE_context*, EIF_REFERENCE, EIF_INTEGER_32, EIF_BOOLEAN, EIF_INTEGER_32, EIF_INTEGER_32, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_INTEGER_32, EIF_BOOLEAN);

/*
 * Exception tag associated with `a_code'.
 */
extern char* GE_exception_tag(long a_code);

/*
 * Raise an exception with code `a_code'.
 */
extern void GE_raise(long a_code);

/*
 * Raise an exception with code `a_code' and message `msg'.
 */
extern void GE_raise_with_message(long a_code, const char* msg);

/*
 * Raise an exception from EXCEPTION_MANAGER.
 */
extern void GE_developer_raise(long a_code, char* a_meaning, char* a_message);

/*
 * Raise exception which was raised the first time a once routine
 * was executed when executing it again.
 */
extern void GE_raise_once_exception(GE_context* a_context, EIF_REFERENCE a_exception);

/*
 * Exception, if any, which was last raised in `a_context'.
 */
extern EIF_REFERENCE GE_last_exception_raised(GE_context* a_context);

/*
 * Jump to execute the rescue of the last routine with a rescue
 * in the call stack.
 */
extern void GE_jump_to_last_rescue(GE_context* a_context);

/*
 * Set `in_assertion' to 'not b'.
 * Return the opposite of previous value.
 */
extern EIF_BOOLEAN GE_check_assert(EIF_BOOLEAN b);

/*
 * Check whether the type id of `obj' is not in `type_ids'.
 * If it is, then raise a CAT-call exception. Don't do anything if `obj' is Void.
 * `nb' is the number of ids in `type_ids' and is expected to be >0.
 * `type_ids' is sorted in increasing order.
 * Return `obj'.
 */
#define GE_catcall(obj,type_ids,nb) GE_check_catcall((obj),(type_ids),(nb))
extern EIF_REFERENCE GE_check_catcall(EIF_REFERENCE obj, EIF_TYPE_INDEX type_ids[], int nb);

/*
 * Check whether `obj' is Void.
 * If it is, then raise a call-on-void-target exception.
 * Return `obj'.
 */
#define GE_void(obj) (!(obj)?GE_check_void(obj):(obj))
extern EIF_REFERENCE GE_check_void(EIF_REFERENCE obj);

/*
 * Check whether `ptr' is a null pointer.
 * If it is, then raise a no-more-memory exception.
 * Return `ptr'.
 */
#define GE_null(ptr) GE_check_null(ptr)
extern void* GE_check_null(void* ptr);

#ifdef EIF_WINDOWS
/*
 * Set default exception handler.
 */
extern void GE_set_windows_exception_filter(void);
#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class THREAD and related threading facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_THREAD_H
#define GE_THREAD_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef GE_USE_THREADS

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_THREAD_TYPES_H
#include "ge_thread_types.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_POSIX_THREADS
#include <unistd.h>
#endif

#ifdef GE_USE_POSIX_THREADS
#	define EIF_TSD_VAL_TYPE        void*
#	define EIF_TSD_TYPE            pthread_key_t
#	define EIF_TSD_CREATE(key,msg) \
		if (pthread_key_create(&(key),NULL)) \
	        GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_SET(key,val,msg) \
		if (pthread_setspecific((key), (EIF_TSD_VAL_TYPE)(val))) \
			GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_GET0(val_type,key,val) (val = pthread_getspecific(key))
#	define EIF_TSD_GET(val_type,key,val,msg) \
		if (EIF_TSD_GET0(val_type,key,val) == (void*) 0) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_DESTROY(key,msg) if (pthread_key_delete(key)) GE_raise_with_message(GE_EX_EXT, msg)
#elif defined EIF_WINDOWS
#	define EIF_TSD_VAL_TYPE        LPVOID
#	define EIF_TSD_TYPE            DWORD
#	define EIF_TSD_CREATE(key,msg) \
		if ((key=TlsAlloc())==TLS_OUT_OF_INDEXES) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_SET(key,val,msg) \
		if (!TlsSetValue((key),(EIF_TSD_VAL_TYPE)(val))) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_GET0(val_type,key,val) \
		val=(val_type) TlsGetValue(key)
#	define EIF_TSD_GET(val_type,key,val,msg) \
		EIF_TSD_GET0(val_type,key,val); \
		if (GetLastError() != NO_ERROR) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_DESTROY(key,msg) \
		if (!TlsFree(key)) GE_raise_with_message(GE_EX_EXT, msg)
#endif

/* Thread priority levels. */
#define EIF_MIN_THR_PRIORITY			0L
#define EIF_BELOW_NORMAL_THR_PRIORITY	100L
#define EIF_DEFAULT_THR_PRIORITY		127L
#define EIF_ABOVE_NORMAL_THR_PRIORITY	154L
#define EIF_MAX_THR_PRIORITY			255L

/*
 * Mutexes used to protect the calls to once-per-process features.
 */
typedef struct {
	EIF_POINTER* boolean_mutex;
	EIF_POINTER* character_8_mutex;
	EIF_POINTER* character_32_mutex;
	EIF_POINTER* integer_8_mutex;
	EIF_POINTER* integer_16_mutex;
	EIF_POINTER* integer_32_mutex;
	EIF_POINTER* integer_64_mutex;
	EIF_POINTER* natural_8_mutex;
	EIF_POINTER* natural_16_mutex;
	EIF_POINTER* natural_32_mutex;
	EIF_POINTER* natural_64_mutex;
	EIF_POINTER* pointer_mutex;
	EIF_POINTER* real_32_mutex;
	EIF_POINTER* real_64_mutex;
	EIF_POINTER* reference_mutex;
	EIF_POINTER* procedure_mutex;
} GE_once_mutexes;

/*
 * Mutexes used to protect the calls to once-per-process features.
 */
extern GE_once_mutexes* GE_process_once_mutexes;

/*
 * Create a new 'GE_once_mutexes' struct which can deal with the
 * numbers of once features passed as argument.
 */
extern GE_once_mutexes* GE_new_once_mutexes(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Keep track of the numbers of once-per-thread features.
 */
extern void GE_thread_onces_set_counts(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Initialize data to handle threads.
 * To be called at the beginning of the main function
 * on the main thread.
 */
extern void GE_init_thread(GE_context* a_context);

/*
 * Create a new thread with attributes `attr' and execute
 * Eiffel routine `routine' on object `current'.
 */
extern void GE_thread_create_with_attr(EIF_REFERENCE current, void (*routine)(EIF_REFERENCE), void (*set_terminated)(EIF_REFERENCE,EIF_BOOLEAN), EIF_THR_ATTR_TYPE* attr);

/*
 * Execution context of current thread.
 */
extern GE_context* GE_thread_current_context(void);

/*
 * Thead ID of current thread.
 */
extern EIF_POINTER GE_thread_id(void);

/*
 * Thread ID of last thread created from current thread.
 */
extern EIF_POINTER GE_last_thread_created(void);

#ifdef EIF_WINDOWS
/*
 * Support for Windows GUI that requires that all GUI operations are performed in the same thread.
 * Allocate new structure of the given size `a_size', assign it to `wel_per_thread_data'.
 * Return newly allocated memory block. It will be freed automatically on thread termination.
 */
extern void* GE_thread_create_wel_per_thread_data(size_t a_size);
#endif

/*
 * Waits until a child thread sets `terminated' from `obj' to True,
 * which means it is terminated. The calling thread must be the
 * direct parent of the thread, or the function might loop indefinitely.
 */
extern void GE_thread_wait(EIF_REFERENCE obj, EIF_BOOLEAN (*get_terminated)(EIF_REFERENCE));

/*
 * Waits until a child thread sets `terminated' from `obj' to True,
 * which means it is terminated, or reaching `a_timeout_ms'.
 * The calling thread must be the direct parent of the thread,
 * or the function might loop indefinitely.
 */
extern EIF_BOOLEAN GE_thread_wait_with_timeout(EIF_REFERENCE obj, EIF_BOOLEAN (*get_terminated)(EIF_REFERENCE), EIF_NATURAL_64 a_timeout_ms);

/*
 * Yields execution to other threads.
 */
extern void GE_thread_yield(void);

/*
 * The calling thread waits for all other children threads to terminate.
 */
extern void GE_thread_join_all(void);

/*
 * Function called to terminate a thread launched by Eiffel with `GE_thread_create_with_attr'.
 * This function must be called from the thread itself (not the parent).
 */
extern void GE_thread_exit(void);

/*
 * Default thread priority level.
 */
extern EIF_INTEGER GE_thread_default_priority(void);

/*
 * Minimum thread priority level.
 */
extern EIF_INTEGER GE_thread_min_priority(void);

/*
 * Maximum thread priority level.
 */
extern EIF_INTEGER GE_thread_max_priority(void);

/*
 * Create a new mutex.
 */
extern EIF_POINTER GE_mutex_create(void);

/*
 * Lock mutex.
 */
extern void GE_mutex_lock(EIF_POINTER a_mutex);

/*
 * Try to lock mutex. Return True on success.
 */
extern EIF_BOOLEAN GE_mutex_try_lock(EIF_POINTER a_mutex);

/*
 * Unlock mutex.
 */
extern void GE_mutex_unlock(EIF_POINTER a_mutex);

/*
 * Destroy and free all resources used by mutex.
 */
extern void GE_mutex_destroy(EIF_POINTER a_mutex);

/*
 * Create a new semaphore allowing `a_count' threads
 * to go into a critical section.
 */
extern EIF_POINTER GE_semaphore_create(EIF_INTEGER a_count);

/*
 * Decrement semaphore count, waiting if necessary
 * until that becomes possible.
 */
extern void GE_semaphore_wait(EIF_POINTER a_semaphore);

/*
 * Has client been successful in decrementing semaphore
 * count without waiting?
 */
extern EIF_BOOLEAN GE_semaphore_try_wait(EIF_POINTER a_semaphore);

/*
 * Increment semaphore count.
 */
extern void GE_semaphore_post(EIF_POINTER a_semaphore);

/*
 * Destroy and free all resources used by semaphore.
 */
extern void GE_semaphore_destroy(EIF_POINTER a_semaphore);

/*
 * Create a new condition variable.
 */
extern EIF_POINTER GE_condition_variable_create(void);

/*
 * Unblock all threads blocked on condition variable.
 */
extern void GE_condition_variable_broadcast(EIF_POINTER a_condition_variable);

/*
 * Unblock one thread blocked on condition variable.
 */
extern void GE_condition_variable_signal(EIF_POINTER a_condition_variable);

/*
 * Block calling thread on condition variable.
 */
extern void GE_condition_variable_wait(EIF_POINTER a_condition_variable, EIF_POINTER a_mutex);

/*
 * Block calling thread on condition variable for at most `a_timeout' milliseconds.
 * Return 1 is we got the condition variable on time, otherwise return 0.
 */
extern EIF_INTEGER GE_condition_variable_wait_with_timeout(EIF_POINTER a_condition_variable, EIF_POINTER a_mutex, EIF_INTEGER a_timeout);

/*
 * Destroy and free all resources used by condition variable.
 */
extern void GE_condition_variable_destroy(EIF_POINTER a_condition_variable);

/*
 * Create a new read-write lock.
 */
extern EIF_POINTER GE_read_write_lock_create(void);

/*
 * Acquire a read lock. Multiple readers can go if there are no writer.
 */
extern void GE_read_write_lock_read_lock(EIF_POINTER a_read_write_lock);

/*
 * Acquire a write lock. Only a single write can proceed.
 */
extern void GE_read_write_lock_write_lock(EIF_POINTER a_read_write_lock);

/*
 * Unlock a read or write lock.
 */
extern void GE_read_write_lock_unlock(EIF_POINTER a_read_write_lock);

/*
 * Destroy and free all resources used by read-write lock.
 */
extern void GE_read_write_lock_destroy(EIF_POINTER a_read_write_lock);

#ifdef __cplusplus
}
#endif

#endif
#endif

/*
	description:

		"C functions used to manipulate native strings"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2013-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_NATIVE_STRING_H
#define GE_NATIVE_STRING_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <string.h>
#ifdef EIF_WINDOWS
#ifdef __LCC__
/* With lcc-win32, stat.h should be included before wchar.h. */
#include <sys/stat.h>
#endif
#include <wchar.h>
#else
#include <sys/types.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef EIF_WINDOWS

/* Macro used to manipulate native strings, i.e: (wchar_t*) */
#define GE_nstrlen wcslen /* size of string */
#define GE_nstrncpy wcsncpy /* Copy n characters of one string to another */
#define GE_nstrcpy wcscpy /* Copy one string to another */
#define GE_nstrncat wcsncat /* Append characters of a string */
#define GE_nstrcat wcscat /* Append a string */
#define GE_nstrstr wcsstr /* Return a pointer to the first occurrence of a search string in a string. */
#define GE_nmakestr(quote) L##quote /* Manifest Native string declaration */
#define GE_nstr_fopen _wfopen /* Open file using native string name */
#define GE_nstrcmp wcscmp /* Compare two strings. */
#define GE_nstrdup _wcsdup /* Duplicate string. */
#define GE_nstr_cat_ascii(dest, src) {							\
		int i;													\
		size_t dest_len, src_len;								\
		dest_len = rt_nstrlen (dest);							\
		src_len = strlen (src);									\
		for (i = 0; i < src_len; i++) {							\
			dest[dest_len + i] = (EIF_NATIVE_CHAR)src[i];		\
		}														\
		dest[dest_len + src_len] = (EIF_NATIVE_CHAR)0;			\
	}

#else /* not EIF_WINDOWS */

/* Macro used to manipulate native strings, i.e: (char*) */
#define GE_nstrlen strlen /* size of string */
#define GE_nstrncpy strncpy /* Copy n characters of one string to another */
#define GE_nstrcpy strcpy /* Copy one string to another */
#define GE_nstrncat strncat /* Append characters of a string */
#define GE_nstrcat strcat /* Append a string */
#define GE_nstrstr strstr /* Return a pointer to the first occurrence of a search string in a string. */
#define GE_nmakestr(quote) quote /* Manifest Native string declaration */
#define GE_nstr_fopen fopen /* Open file using native string name */
#define GE_nstrcmp strcmp /* Compare two strings. */
#define GE_nstrdup strdup /* Duplicate string. */
#define GE_nstr_cat_ascii strcat

#endif


#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class ARGUMENTS"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_ARGUMENTS_H
#define GE_ARGUMENTS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int GE_argc;
extern EIF_NATIVE_CHAR** GE_argv;

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement type information"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_TYPES_H
#define GE_TYPES_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Type annotations.
 * When a type has no annotation, it means a detachable, non-separate, variant type.
 * In all other cases, there will be an annotation.
 */
#define ANNOTATION_MASK			0x007F	/* All possible annotations. */
#define ATTACHED_FLAG			0x0001
#define DETACHABLE_FLAG			0x0002	/* Only present when overriding an attached type. */
#define SEPARATE_FLAG			0x0004
#define VARIANT_FLAG			0x0008	/* Only present when overriding a frozen/poly type. */
#define UNUSABLE_FLAG			0x0010	/* Reserved for backward compatibility for storables. */
#define FROZEN_FLAG				0x0020
#define POLY_FLAG				0x0040

/*
 * Type flags.
 */
#define GE_TYPE_FLAG_SPECIAL		0x0010
#define GE_TYPE_FLAG_TUPLE			0x0020
#define GE_TYPE_FLAG_EXPANDED		0x0040
#define GE_TYPE_FLAG_DEFERRED		0x0080
#define GE_TYPE_FLAG_NONE			0x0100
#define GE_TYPE_FLAG_BASIC_MASK		0x000F /* One of "BOOLEAN", "CHARACTER_8", "CHARACTER_32", "INTEGER_8", "INTEGER_16", "INTEGER_32", "INTEGER_64", "NATURAL_8", "NATURAL_16", "NATURAL_32", "NATURAL_64", "POINTER", "REAL_32", "REAL_64" */
#define GE_TYPE_FLAG_BOOLEAN		0x0001
#define GE_TYPE_FLAG_CHARACTER_8	0x0002
#define GE_TYPE_FLAG_CHARACTER_32	0x0003
#define GE_TYPE_FLAG_INTEGER_8		0x0004
#define GE_TYPE_FLAG_INTEGER_16		0x0005
#define GE_TYPE_FLAG_INTEGER_32		0x0006
#define GE_TYPE_FLAG_INTEGER_64		0x0007
#define GE_TYPE_FLAG_NATURAL_8		0x0008
#define GE_TYPE_FLAG_NATURAL_16		0x0009
#define GE_TYPE_FLAG_NATURAL_32		0x000A
#define GE_TYPE_FLAG_NATURAL_64		0x000B
#define GE_TYPE_FLAG_POINTER		0x000C
#define GE_TYPE_FLAG_REAL_32		0x000D
#define GE_TYPE_FLAG_REAL_64		0x000E

/*
 * Convention for attribute types.
 * The values are in sync with REFLECTOR_CONSTANTS.
 */
#define GE_TYPE_KIND_INVALID		-1
#define GE_TYPE_KIND_POINTER		0
#define GE_TYPE_KIND_REFERENCE		1
#define GE_TYPE_KIND_CHARACTER_8	2
#define GE_TYPE_KIND_BOOLEAN		3
#define GE_TYPE_KIND_INTEGER_32		4
#define GE_TYPE_KIND_REAL_32		5
#define GE_TYPE_KIND_REAL_64		6
#define GE_TYPE_KIND_EXPANDED		7
#define GE_TYPE_KIND_INTEGER_8		9
#define GE_TYPE_KIND_INTEGER_16		10
#define GE_TYPE_KIND_INTEGER_64 	11
#define GE_TYPE_KIND_CHARACTER_32	12
#define GE_TYPE_KIND_NATURAL_8		13
#define GE_TYPE_KIND_NATURAL_16		14
#define GE_TYPE_KIND_NATURAL_32 	15
#define GE_TYPE_KIND_NATURAL_64 	16

/*
 * Object flags.
 */
#define GE_OBJECT_FLAG_MARKED		0x0001

/*
 * Ancestor relationship between two types X and Y.
 */
#ifdef GE_USE_ANCESTORS
typedef struct {
	EIF_TYPE_INDEX type_id; /* Type id of Y */
	EIF_BOOLEAN conforms; /* Does X conform to Y? */
	void (**qualified_calls)(); /* Function pointers, indexed by call id, when the static type of the target is Y and the dynamic type is X */
} GE_ancestor;
#endif

/*
 * Attribute.
 */
#ifdef GE_USE_ATTRIBUTES
typedef struct {
#ifdef GE_USE_ATTRIBUTE_NAME
	const char* name; /* Attribute name */
#endif
#ifdef GE_USE_ATTRIBUTE_TYPE_ID
	EIF_ENCODED_TYPE type_id; /* Static type id */
#endif
#ifdef GE_USE_ATTRIBUTE_OFFSET
	uint32_t offset; /* Address offset in object */
#endif
} GE_attribute;
#endif

/*
 * Type information.
 */
typedef struct {
	EIF_TYPE_INDEX type_id;
	uint16_t flags;
#ifdef GE_USE_TYPE_GENERATOR
	const char* generator; /* Generator class name */
#endif
#ifdef GE_USE_TYPE_NAME
	const char* name; /* Full type name */
#endif
#ifdef GE_USE_TYPE_GENERIC_PARAMETERS
	EIF_ENCODED_TYPE* generic_parameters;
	uint32_t generic_parameter_count;
#endif
#ifdef GE_USE_ANCESTORS
	GE_ancestor** ancestors;
	uint32_t ancestor_count;
#endif
#ifdef GE_USE_ATTRIBUTES
	GE_attribute** attributes;
	uint32_t attribute_count;
#endif
#ifdef GE_USE_TYPE_OBJECT_SIZE
	uint64_t object_size;
#endif
	EIF_REFERENCE (*new_instance)();
	void (*dispose)(GE_context*, EIF_REFERENCE);
} GE_type_info;

typedef struct {
	EIF_TYPE_INDEX id; /* Type id of the "TYPE [X]" object */
	EIF_INTEGER type_id; /* Type id of the type "X" */
	EIF_BOOLEAN is_special;
	void (*dispose)(GE_context*, EIF_REFERENCE);
	EIF_REFERENCE a1; /* internal_name */
	EIF_REFERENCE a2; /* internal_name_32 */
} EIF_TYPE_OBJ;

/*
 * Types indexed by type id.
 * Generated by the compiler.
 */
extern EIF_TYPE_OBJ GE_types[][2];
extern GE_type_info GE_type_infos[];

/*
 * Number of type infos in `GE_type_infos'.
 * Do not take into account the fake item at index 0.
 */
extern int GE_type_info_count;

/*
 * Encode a EIF_TYPE into a EIF_ENCODED_TYPE.
 * The lower part of EIF_ENCODED_TYPE contains the .id field,
 * and the upper part the .annotations.
 */
extern EIF_ENCODED_TYPE GE_encoded_type(EIF_TYPE a_type);

/*
 * Decode a EIF_ENCODED_TYPE into a EIF_TYPE.
 * The lower part of EIF_ENCODED_TYPE contains the .id field,
 * and the upper part the .annotations.
 */
extern EIF_TYPE GE_decoded_type(EIF_ENCODED_TYPE a_type);

/*
 * Type with `a_id' and `a_annotations'.
 */
extern EIF_TYPE GE_new_type(EIF_TYPE_INDEX a_id, EIF_TYPE_INDEX a_annotations);

/*
 * Type of object `obj'.
 */
#define GE_object_type(obj)	GE_new_type(((EIF_REFERENCE)(obj))->id, 0x0)
#define GE_object_encoded_type(obj) GE_encoded_type(GE_object_type(obj))

/*
 * Attachment status of `a_type'.
 */
#define GE_is_attached_type(a_type) EIF_TEST(((a_type).annotations & ATTACHED_FLAG) || GE_is_expanded_type_index((a_type).id))
#define GE_is_attached_encoded_type(a_type) GE_is_attached_type(GE_decoded_type(a_type))

/*
 * Associated detachable type of `a_type' if any,
 * otherwise `a_type'.
 */
extern EIF_TYPE GE_non_attached_type(EIF_TYPE a_type);
#define GE_non_attached_encoded_type(a_type) GE_encoded_type(GE_non_attached_type(GE_decoded_type(a_type)))

/*
 * Associated attached type of `a_type' if any,
 * otherwise `a_type'.
 */
extern EIF_TYPE GE_attached_type(EIF_TYPE a_type);
#define GE_attached_encoded_type(t) GE_encoded_type(GE_attached_type(GE_decoded_type(t)))

/*
 * Is `a_type' a SPECIAL type?
 */
#define GE_is_special_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_SPECIAL)
#define GE_is_special_encoded_type(a_type) GE_is_special_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_object(obj) GE_is_special_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of user-defined expanded type?
 */
extern EIF_BOOLEAN GE_is_special_of_expanded_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_expanded_encoded_type(a_type) GE_is_special_of_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_expanded_object(obj) GE_is_special_of_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of reference type?
 */
extern EIF_BOOLEAN GE_is_special_of_reference_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_reference_encoded_type(a_type) GE_is_special_of_reference_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_reference_object(obj) GE_is_special_of_reference_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of reference type or basic expanded type?
 * (Note that user-defined expanded types are excluded.)
 */
extern EIF_BOOLEAN GE_is_special_of_reference_or_basic_expanded_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_reference_or_basic_expanded_encoded_type(a_type) GE_is_special_of_reference_or_basic_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_reference_or_basic_expanded_object(obj) GE_is_special_of_reference_or_basic_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a TUPLE type?
 */
#define GE_is_tuple_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_TUPLE)
#define GE_is_tuple_encoded_type(a_type) GE_is_tuple_type_index(GE_decoded_type(a_type).id)
#define GE_is_tuple_object(obj) GE_is_tuple_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' an expanded type?
 */
#define GE_is_expanded_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_EXPANDED)
#define GE_is_expanded_encoded_type(a_type) GE_is_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_expanded_object(obj) GE_is_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a type whose base class is deferred?
 */
#define GE_is_deferred_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_DEFERRED)
#define GE_is_deferred_encoded_type(a_type) GE_is_deferred_type_index(GE_decoded_type(a_type).id)

/*
 * Does `i'-th field of `a_object + a_physical_offset' (which is expected to be reference)
 * denote a reference with copy semantics?
 */
extern EIF_BOOLEAN GE_is_copy_semantics_field(EIF_INTEGER i, EIF_POINTER a_object, EIF_INTEGER a_physical_offset);

/*
 * Does `i'-th item of special `a_object' (which is expected to be reference)
 * denote a reference with copy semantics?
 */
extern EIF_BOOLEAN GE_is_special_copy_semantics_item(EIF_INTEGER i, EIF_POINTER a_object);

/*
 * Generator class name of `a_type'.
 */
extern EIF_REFERENCE GE_generator_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_generator_of_encoded_type(a_type) GE_generator_of_type_index(GE_decoded_type(a_type).id)

/*
 * Full name of `a_type'.
 */
extern EIF_REFERENCE GE_generating_type_of_encoded_type(EIF_ENCODED_TYPE a_type);

/*
 * Encoded type whose name is `a_name'.
 * -1 if no such type.
 */
extern EIF_ENCODED_TYPE GE_encoded_type_from_name(EIF_POINTER a_name);

/*
 * Does `a_type_1' conform to `a_type_2'?
 */
extern EIF_BOOLEAN GE_encoded_type_conforms_to(EIF_ENCODED_TYPE a_type_1, EIF_ENCODED_TYPE a_type_2);

/*
 * Number of generic parameters.
 */
extern EIF_INTEGER GE_generic_parameter_count_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_generic_parameter_count_of_encoded_type(a_type) GE_generic_parameter_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Type of `i'-th generic parameter of `a_type'.
 */
extern EIF_INTEGER GE_generic_parameter_of_type_index(EIF_TYPE_INDEX a_type, EIF_INTEGER i);
#define GE_generic_parameter_of_encoded_type(a_type,i) GE_generic_parameter_of_type_index(GE_decoded_type(a_type).id, (i))

/*
 * Number of fields of an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_count_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_field_count_of_encoded_type(a_type) GE_field_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Physical offset of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_offset_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_offset_of_encoded_type(i, a_type) GE_field_offset_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Name of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_POINTER GE_field_name_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_name_of_encoded_type(i, a_type) GE_field_name_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Static type of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_static_type_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_static_type_of_encoded_type(i, a_type) GE_field_static_type_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Kind of type of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_type_kind_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_type_kind_of_encoded_type(i, a_type) GE_field_type_kind_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Physical size of `a_object'.
 */
extern EIF_NATURAL_64 GE_object_size(EIF_POINTER a_object);

/*
 * Is `i'-th field of objects of type `a_type' a user-defined expanded attribute?
 */
extern EIF_BOOLEAN GE_is_field_expanded_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_is_field_expanded_of_encoded_type(i, a_type) GE_is_field_expanded_of_type_index((i), GE_decoded_type(a_type).id)

#define GE_field_address_at(a_field_offset, a_object, a_physical_offset) ((char*)(a_object) + (a_physical_offset) + (a_field_offset))
#define GE_object_at_offset(a_enclosing, a_physical_offset) (EIF_REFERENCE)(GE_field_address_at(0, (a_enclosing), (a_physical_offset)))
#define GE_raw_object_at_offset(a_enclosing, a_physical_offset) (EIF_POINTER)(GE_field_address_at(0, (a_enclosing), (a_physical_offset)))
#define GE_object_encoded_type_at_offset(a_enclosing, a_physical_offset) GE_object_encoded_type(GE_raw_object_at_offset((a_enclosing), (a_physical_offset)))
#define GE_boolean_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_BOOLEAN*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_character_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_CHARACTER_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_character_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_CHARACTER_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_16_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_16*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_16_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_16*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_pointer_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_POINTER*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_real_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REAL_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_real_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REAL_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_raw_reference_field_at(a_field_offset, a_object, a_physical_offset) (EIF_POINTER)*(EIF_REFERENCE*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_reference_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REFERENCE*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_set_boolean_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_boolean_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_character_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_character_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_character_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_character_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_16_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_16_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_16_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_16_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_pointer_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_pointer_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_real_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_real_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_real_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_real_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_reference_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_reference_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value

#if defined(GE_USE_ATTRIBUTES) && defined(GE_USE_ATTRIBUTE_OFFSET)
#define GE_field_address(i, a_object, a_physical_offset) GE_field_address_at(GE_type_infos[((EIF_REFERENCE)(a_object))->id].attributes[(i) - 1]->offset, (a_object), (a_physical_offset))
#define GE_boolean_field(i, a_object, a_physical_offset) *(EIF_BOOLEAN*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_character_8_field(i, a_object, a_physical_offset) *(EIF_CHARACTER_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_character_32_field(i, a_object, a_physical_offset) *(EIF_CHARACTER_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_8_field(i, a_object, a_physical_offset) *(EIF_INTEGER_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_16_field(i, a_object, a_physical_offset) *(EIF_INTEGER_16*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_32_field(i, a_object, a_physical_offset) *(EIF_INTEGER_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_64_field(i, a_object, a_physical_offset) *(EIF_INTEGER_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_8_field(i, a_object, a_physical_offset) *(EIF_NATURAL_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_16_field(i, a_object, a_physical_offset) *(EIF_NATURAL_16*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_32_field(i, a_object, a_physical_offset) *(EIF_NATURAL_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_64_field(i, a_object, a_physical_offset) *(EIF_NATURAL_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_pointer_field(i, a_object, a_physical_offset) *(EIF_POINTER*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_real_32_field(i, a_object, a_physical_offset) *(EIF_REAL_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_real_64_field(i, a_object, a_physical_offset) *(EIF_REAL_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_reference_field(i, a_object, a_physical_offset) *(EIF_REFERENCE*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_set_boolean_field(i, a_object, a_physical_offset, a_value) GE_boolean_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_character_8_field(i, a_object, a_physical_offset, a_value) GE_character_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_character_32_field(i, a_object, a_physical_offset, a_value) GE_character_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_8_field(i, a_object, a_physical_offset, a_value) GE_integer_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_16_field(i, a_object, a_physical_offset, a_value) GE_integer_16_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_32_field(i, a_object, a_physical_offset, a_value) GE_integer_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_64_field(i, a_object, a_physical_offset, a_value) GE_integer_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_8_field(i, a_object, a_physical_offset, a_value) GE_natural_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_16_field(i, a_object, a_physical_offset, a_value) GE_natural_16_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_32_field(i, a_object, a_physical_offset, a_value) GE_natural_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_64_field(i, a_object, a_physical_offset, a_value) GE_natural_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_pointer_field(i, a_object, a_physical_offset, a_value) GE_pointer_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_real_32_field(i, a_object, a_physical_offset, a_value) GE_real_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_real_64_field(i, a_object, a_physical_offset, a_value) GE_real_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_reference_field(i, a_object, a_physical_offset, a_value) GE_reference_field((i), (a_object), (a_physical_offset)) = (a_value)
#else
#define GE_boolean_field(i, a_object, a_physical_offset) (EIF_BOOLEAN)0
#define GE_character_8_field(i, a_object, a_physical_offset) (EIF_CHARACTER_8)0
#define GE_character_32_field(i, a_object, a_physical_offset) (EIF_CHARACTER_32)0
#define GE_integer_8_field(i, a_object, a_physical_offset) (EIF_INTEGER_8)0
#define GE_integer_16_field(i, a_object, a_physical_offset) (EIF_INTEGER_16)0
#define GE_integer_32_field(i, a_object, a_physical_offset) (EIF_INTEGER_32)0
#define GE_integer_64_field(i, a_object, a_physical_offset) (EIF_INTEGER_64)0
#define GE_natural_8_field(i, a_object, a_physical_offset) (EIF_NATURAL_8)0
#define GE_natural_16_field(i, a_object, a_physical_offset) (EIF_NATURAL_16)0
#define GE_natural_32_field(i, a_object, a_physical_offset) (EIF_NATURAL_32)0
#define GE_natural_64_field(i, a_object, a_physical_offset) (EIF_NATURAL_64)0
#define GE_pointer_field(i, a_object, a_physical_offset) (EIF_POINTER)0
#define GE_real_32_field(i, a_object, a_physical_offset) (EIF_REAL_32)0
#define GE_real_64_field(i, a_object, a_physical_offset) (EIF_REAL_64)0
#define GE_reference_field(i, a_object, a_physical_offset) (EIF_REFERENCE)0
#define GE_set_boolean_field(i, a_object, a_physical_offset, a_value)
#define GE_set_character_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_character_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_16_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_16_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_pointer_field(i, a_object, a_physical_offset, a_value)
#define GE_set_real_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_real_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_reference_field(i, a_object, a_physical_offset, a_value)
#endif

/*
 * Number of non-transient fields of an object of dynamic type `a_type'.
 * TODO: storable not implemented yet.
 */
#define GE_persistent_field_count_of_type_index(a_type) GE_field_count_of_type_index(a_type)
#define GE_persistent_field_count_of_encoded_type(a_type) GE_persistent_field_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Is `i'-th field of objects of type `a_type' a transient field?
 * TODO: storable not implemented yet.
 */
#define GE_is_field_transient_of_type_index(i, a_type) EIF_FALSE
#define GE_is_field_transient_of_encoded_type(i, a_type) GE_is_field_transient_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Storable version of objects of type `a_type'.
 * TODO: storable not implemented yet.
 */
#define GE_storable_version_of_type_index(a_type) EIF_VOID
#define GE_storable_version_of_encoded_type(a_type) GE_storable_version_of_type_index(GE_decoded_type(a_type).id)

/*
 * Get a lock on `GE_mark_object' and `GE_unmark_object' routines so that
 * 2 threads cannot `GE_mark_object' and `GE_unmark_object' at the same time.
 */
extern void GE_lock_marking(void);

/*
 * Release a lock on `GE_mark_object' and `GE_unmark_object', so that another
 * thread can use `GE_mark_object' and `GE_unmark_object'.
 */
extern void GE_unlock_marking(void);

/*
 * Is `obj' marked?
 */
extern EIF_BOOLEAN GE_is_object_marked(EIF_POINTER obj);

/*
 * Mark `obj'.
 */
extern void GE_mark_object(EIF_POINTER obj);

/*
 * Unmark `obj'.
 */
extern void GE_unmark_object(EIF_POINTER obj);

/*
 * New instance of dynamic `a_type'.
 * Note: returned object is not initialized and may
 * hence violate its invariant.
 * `a_type' cannot represent a SPECIAL type, use
 * `GE_new_special_of_reference_instance_of_type_index' instead.
 */
extern EIF_REFERENCE GE_new_instance_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_new_instance_of_encoded_type(a_type) GE_new_instance_of_type_index(GE_decoded_type(a_type).id)

/*
 * New instance of dynamic `a_type' that represents
 * a SPECIAL with can contain `a_capacity' elements of reference type.
 * To create a SPECIAL of basic type, use class SPECIAL directly.
 */
extern EIF_REFERENCE GE_new_special_of_reference_instance_of_type_index(EIF_TYPE_INDEX a_type, EIF_INTEGER a_capacity);
#define GE_new_special_of_reference_instance_of_encoded_type(a_type, a_capacity) GE_new_special_of_reference_instance_of_type_index(GE_decoded_type(a_type).id, (a_capacity))

/*
 * New instance of tuple of type `a_type'.
 * Note: returned object is not initialized and may
 * hence violate its invariant.
 */
#define GE_new_tuple_instance_of_type_index(a_type) GE_new_instance_of_type_index(a_type)
#define GE_new_tuple_instance_of_encoded_type(a_type) GE_new_tuple_instance_of_type_index(GE_decoded_type(a_type).id)

/*
 * New instance of TYPE for object of type `a_type'.
 */
extern EIF_REFERENCE GE_new_type_instance_of_encoded_type(EIF_ENCODED_TYPE a_type);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to manipulate strings"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_STRING_H
#define GE_STRING_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * New Eiffel empty string of type "STRING_8" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_str8(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "IMMUTABLE_STRING_8" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_istr8(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "STRING_32" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_str32(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "IMMUTABLE_STRING_32" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_istr32(EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_8" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ms8(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_8" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_str8(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_8" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ims8(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ms32(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing the
 * first `c' 32-bit characters built from `s' by reading
 * groups of four bytes with little-endian byte order.
 */
extern EIF_REFERENCE GE_ms32_from_utf32le(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_str32(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing
 * the first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ims32(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing the
 * first `c' 32-bit characters built from `s' by reading
 * groups of four bytes with little-endian byte order.
 */
extern EIF_REFERENCE GE_ms32_from_utf32le(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_istr32(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing the
 * first `n' native characters found in native string `s'.
 * Invalid native characters are escaped.
 */
extern EIF_REFERENCE GE_ims32_from_nstr(EIF_NATIVE_CHAR* s, EIF_INTEGER n);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing all
 * characters found in the null-terminated native string `s'.
 * Invalid native characters are escaped.
 */
extern EIF_REFERENCE GE_istr32_from_nstr(EIF_NATIVE_CHAR* s);

/*
 * New Eiffel string of type "STRING" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'
 */
extern EIF_REFERENCE GE_str(const char* s);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class CONSOLE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_CONSOLE_H
#define GE_CONSOLE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Initialize mutex to determine whether a new
 * console needs to be created.
 */
#ifdef EIF_WINDOWS
extern void GE_init_console(void);
#else
#define GE_init_console()
#endif

/*
 * Create a new DOS console if needed (i.e. in case of a Windows application).
 */
#ifdef EIF_WINDOWS
extern void GE_show_console(void);
#else
#define GE_show_console()
#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement the program initialization"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_MAIN_H
#define GE_MAIN_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int GE_main(int argc, EIF_NATIVE_CHAR** argv);

/*
 * System name.
 */
extern char* GE_system_name;

/*
 * Root class name.
 */
extern char* GE_root_class_name;

#ifdef EIF_WINDOWS

#include <windows.h>

/*
 * Used in WEL.
 */
extern HINSTANCE eif_hInstance;
extern HINSTANCE eif_hPrevInstance;
extern LPWSTR eif_lpCmdLine;
extern int eif_nCmdShow;

/*
 * Main entry point when compiling a Windows application.
 * See:
 *    http://en.wikipedia.org/wiki/WinMain
 *    http://msdn2.microsoft.com/en-us/library/ms633559.aspx
 */
extern int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow);

#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to access garbage collector facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_GC_H
#define GE_GC_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef GE_USE_BOEHM_GC
/*
 *	Use the Boehm garbage collector.
 *	See:
 *		http://en.wikipedia.org/wiki/Boehm_GC
 *		http://www.hpl.hp.com/personal/Hans_Boehm/gc/
 */
#define GC_IGNORE_WARN
#ifdef GE_USE_THREADS
#define GC_THREADS
#undef GC_NO_THREAD_REDIRECTS
#endif
#include "gc.h"
#endif

/*
 * GC initialization.
 */

#ifdef GE_USE_BOEHM_GC
#ifdef GE_USE_THREADS
#define GE_init_gc() \
	GC_INIT(); \
	GC_allow_register_threads(); \
	GC_enable_incremental()
#else
#define GE_init_gc() \
	GC_INIT(); \
	GC_enable_incremental()
#endif
#else /* No GC */
#define GE_init_gc() /* do nothing */
#endif

/*
 * Memory allocation.
 */

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc(size) GE_null(GC_MALLOC(size))
#else /* No GC */
#define GE_malloc(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc(size) GC_MALLOC(size)
#else /* No GC */
#define GE_unprotected_malloc(size) malloc(size)
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_atomic(size) GE_null(GC_MALLOC_ATOMIC(size))
#else /* No GC */
#define GE_malloc_atomic(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_atomic(size) GC_MALLOC_ATOMIC(size)
#else /* No GC */
#define GE_unprotected_malloc_atomic(size) malloc(size)
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc(nelem, elsize) GE_null(GC_MALLOC((nelem) * (elsize)))
#else /* No GC */
#define GE_calloc(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_calloc(nelem, elsize) GC_MALLOC((nelem) * (elsize))
#else /* No GC */
#define GE_unprotected_calloc(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_atomic(nelem, elsize) memset(GE_null(GC_MALLOC_ATOMIC((nelem) * (elsize))), 0, (nelem) * (elsize))
#else /* No GC */
#define GE_calloc(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_uncollectable(size) GE_null(GC_MALLOC_UNCOLLECTABLE(size))
#else /* No GC */
#define GE_malloc_uncollectable(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_uncollectable(size) GC_MALLOC_UNCOLLECTABLE(size)
#else /* No GC */
#define GE_unprotected_malloc_uncollectable(size) malloc(size)
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_atomic_uncollectable(size) GE_null(GC_malloc_atomic_uncollectable(size))
#else /* No GC */
#define GE_malloc_atomic_uncollectable(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_atomic_uncollectable(size) GC_malloc_atomic_uncollectable(size)
#else /* No GC */
#define GE_unprotected_malloc_atomic_uncollectable(size) malloc(size)
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_uncollectable(nelem, elsize) GE_null(GC_MALLOC_UNCOLLECTABLE((nelem) * (elsize)))
#else /* No GC */
#define GE_calloc_uncollectable(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_calloc_uncollectable(nelem, elsize) GC_MALLOC_UNCOLLECTABLE((nelem) * (elsize))
#else /* No GC */
#define GE_unprotected_calloc_uncollectable(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_atomic_uncollectable(nelem, elsize) memset(GE_null(GC_malloc_atomic_uncollectable((nelem) * (elsize))), 0, (nelem) * (elsize))
#else /* No GC */
#define GE_calloc_atomic_uncollectable(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
extern void* GE_unprotected_calloc_atomic_uncollectable(size_t nelem, size_t elsize);
#else /* No GC */
#define GE_unprotected_calloc_atomic_uncollectable(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is not necessarily zeroed.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_realloc(p, size) GE_null(GC_REALLOC((p), (size)))
#else /* No GC */
#define GE_realloc(p, size) GE_null(realloc((p), (size)))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is not necessarily zeroed.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_realloc(p, size) GC_REALLOC((p), (size))
#else /* No GC */
#define GE_unprotected_realloc(p, size) realloc((p), (size))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is zeroed.
 * Raise an exception when no-more-memory.
 */
extern void* GE_recalloc(void* p, size_t old_nelem, size_t new_nelem, size_t elsize);

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is zeroed.
 * Do not raise an exception when no-more-memory.
 */
extern void* GE_unprotected_recalloc(void* p, size_t old_nelem, size_t new_nelem, size_t elsize);

/*
 * Explicitly deallocate an object.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_free(p) GC_FREE(p)
#else /* No GC */
#define GE_free(p) free(p)
#endif

/*
 * When defined, `GE_malloc_cleared' means that `GE_malloc' and
 * `GE_malloc_uncollectable' make sure that the allocated memory
 * is zeroed.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_cleared
#else /* No GC */
#endif

/*
 * When defined, `GE_malloc_atomic_cleared' means that `GE_malloc_atomic'
 * and `GE_malloc_atomic_uncollectable' make sure that the allocated memory
 * is zeroed.
 */
#ifdef GE_USE_BOEHM_GC
/* #define GE_malloc_atomic_cleared */
#else /* No GC */
#endif

/*
 * Dispose
 */

/*
 * Register dispose routine `disp' to be called on object `obj' when it will be collected.
 */
#ifdef GE_USE_BOEHM_GC
extern void GE_boehm_dispose(void* C, void* disp); /* Call dispose routine `disp' on object `C'. */
#define GE_register_dispose(obj, disp) GC_REGISTER_FINALIZER_NO_ORDER((void*)(obj), (void (*) (void*, void*)) &GE_boehm_dispose, NULL, NULL, NULL)
#else /* No GC */
#define GE_register_dispose(obj, disp) /* do nothing */
#endif

/*
 * Access to objects, useful with GCs which move objects in memory.
 * This is not the case here, since the Boehm GC is not a moving GC.
 */

/* Access object through hector. */
#define eif_access(obj) (obj)
/* Freeze memory address. */
#define eif_freeze(obj) (obj)
/* The C side adopts an object. */
#define eif_adopt(obj) (obj)
/* The C side protects an object. */
#define eif_protect(obj) (obj)
/* The C side weans adopted object. */
#define eif_wean(obj) (obj)
/* Forget a frozen memory address. */
#define eif_unfreeze(obj)
/* Always frozen since they do not move. */
#define eif_frozen(obj) 1
/* Always frozen since they do not move. */
#define spfrozen(obj) 1

#endif

/*
	description:

		"C functions used to implement class IDENTIFIED"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_IDENTIFIED_H
#define GE_IDENTIFIED_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Initialize data to keep track of object ids.
 */
extern void GE_init_identified(void);

/*
 * Get a new id for `object', assuming it is NOT in the stack.
 */
extern EIF_INTEGER_32 GE_object_id(EIF_OBJECT object);

/*
 * Return the object associated with `id'.
 */
extern EIF_REFERENCE GE_id_object(EIF_INTEGER_32 id);

/*
 * Remove the object associated with `id' from the stack.
 */
extern void GE_object_id_free(EIF_INTEGER_32 id);

#ifdef __cplusplus
}
#endif

#endif

#ifdef __cplusplus
extern "C" {
#endif

#define T0 EIF_ANY

/* CHARACTER */
#define EIF_CHARACTER EIF_CHARACTER_8

/* WIDE_CHARACTER */
#define EIF_WIDE_CHAR EIF_CHARACTER_32

/* INTEGER */
#define EIF_INTEGER EIF_INTEGER_32

/* NATURAL */
#define EIF_NATURAL EIF_NATURAL_32

/* REAL */
#define EIF_REAL EIF_REAL_32

/* DOUBLE */
#define EIF_DOUBLE EIF_REAL_64

/* BOOLEAN */
#define T1 EIF_BOOLEAN
extern T0* GE_boxed1(T1 a1);
typedef struct Sb1 Tb1;

/* CHARACTER_8 */
#define T2 EIF_CHARACTER_8
extern T0* GE_boxed2(T2 a1);
typedef struct Sb2 Tb2;

/* CHARACTER_32 */
#define T3 EIF_CHARACTER_32
extern T0* GE_boxed3(T3 a1);
typedef struct Sb3 Tb3;

/* INTEGER_8 */
#define T4 EIF_INTEGER_8
extern T0* GE_boxed4(T4 a1);
typedef struct Sb4 Tb4;

/* INTEGER_16 */
#define T5 EIF_INTEGER_16
extern T0* GE_boxed5(T5 a1);
typedef struct Sb5 Tb5;

/* INTEGER_32 */
#define T6 EIF_INTEGER_32
extern T0* GE_boxed6(T6 a1);
typedef struct Sb6 Tb6;

/* INTEGER_64 */
#define T7 EIF_INTEGER_64
extern T0* GE_boxed7(T7 a1);
typedef struct Sb7 Tb7;

/* NATURAL_8 */
#define T8 EIF_NATURAL_8
extern T0* GE_boxed8(T8 a1);
typedef struct Sb8 Tb8;

/* NATURAL_16 */
#define T9 EIF_NATURAL_16
extern T0* GE_boxed9(T9 a1);
typedef struct Sb9 Tb9;

/* NATURAL_32 */
#define T10 EIF_NATURAL_32
extern T0* GE_boxed10(T10 a1);
typedef struct Sb10 Tb10;

/* NATURAL_64 */
#define T11 EIF_NATURAL_64
extern T0* GE_boxed11(T11 a1);
typedef struct Sb11 Tb11;

/* REAL_32 */
#define T12 EIF_REAL_32
extern T0* GE_boxed12(T12 a1);
typedef struct Sb12 Tb12;

/* REAL_64 */
#define T13 EIF_REAL_64
extern T0* GE_boxed13(T13 a1);
typedef struct Sb13 Tb13;

/* POINTER */
#define T14 EIF_POINTER
extern T0* GE_boxed14(T14 a1);
typedef struct Sb14 Tb14;

/* [detachable] SPECIAL [CHARACTER_8] */
typedef struct S15 T15;

/* [detachable] SPECIAL [CHARACTER_32] */
typedef struct S16 T16;

/* [detachable] STRING_8 */
typedef struct S17 T17;

/* [detachable] STRING_32 */
typedef struct S18 T18;

/* [detachable] IMMUTABLE_STRING_32 */
typedef struct S20 T20;

/* [detachable] ISE_EXCEPTION_MANAGER */
typedef struct S21 T21;

/* [detachable] GEC */
typedef struct S25 T25;

/* [detachable] CELL [detachable EXCEPTION] */
typedef struct S26 T26;

/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
typedef struct S27 T27;

/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
typedef struct S28 T28;

/* [detachable] CELL [[attached] NO_MORE_MEMORY] */
typedef struct S29 T29;

/* detachable C_STRING */
typedef struct S30 T30;

/* [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
typedef struct S31 T31;

/* [detachable] KL_TEXT_INPUT_FILE */
typedef struct S32 T32;

/* [detachable] KL_ARGUMENTS */
typedef struct S33 T33;

/* [detachable] ET_ISE_VARIABLES */
typedef struct S34 T34;

/* [detachable] ET_ERROR_HANDLER */
typedef struct S35 T35;

/* detachable ET_SYSTEM */
typedef struct S36 T36;

/* [detachable] KL_EXCEPTIONS */
typedef struct S38 T38;

/* [detachable] AP_PARSER */
typedef struct S39 T39;

/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
typedef struct S40 T40;

/* [detachable] AP_STRING_OPTION */
typedef struct S41 T41;

/* [detachable] AP_FLAG */
typedef struct S43 T43;

/* [detachable] UT_VERSION */
typedef struct S44 T44;

/* [detachable] AP_ENUMERATION_OPTION */
typedef struct S45 T45;

/* [detachable] AP_BOOLEAN_OPTION */
typedef struct S46 T46;

/* [detachable] AP_INTEGER_OPTION */
typedef struct S47 T47;

/* [detachable] ET_NULL_ERROR_HANDLER */
typedef struct S51 T51;

/* [detachable] ET_ECF_SYSTEM_PARSER */
typedef struct S53 T53;

/* [detachable] ET_ECF_ERROR_HANDLER */
typedef struct S54 T54;

/* detachable ET_ECF_SETTINGS */
typedef struct S55 T55;

/* [detachable] ET_ECF_SYSTEM */
typedef struct S56 T56;

/* [detachable] ET_ECF_TARGET */
typedef struct S57 T57;

/* detachable ET_ECF_CAPABILITIES */
typedef struct S59 T59;

/* detachable ET_ECF_VARIABLES */
typedef struct S60 T60;

/* [detachable] ET_DYNAMIC_SYSTEM */
typedef struct S61 T61;

/* [detachable] ET_SYSTEM_PROCESSOR */
typedef struct S62 T62;

/* [detachable] ET_C_GENERATOR */
typedef struct S65 T65;

/* [detachable] KL_SHELL_COMMAND */
typedef struct S66 T66;

/* detachable DT_DATE_TIME */
typedef struct S67 T67;

/* detachable DS_HASH_SET [[attached] STRING_8] */
typedef struct S69 T69;

/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
typedef struct S70 T70;

/* [detachable] ET_CLASS_TYPE */
typedef struct S71 T71;

/* [detachable] ET_TOKEN_CONSTANTS */
typedef struct S72 T72;

/* [detachable] ET_CLASS */
typedef struct S73 T73;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S75 T75;

/* [detachable] KL_OPERATING_SYSTEM */
typedef struct S76 T76;

/* [detachable] UT_CANNOT_READ_FILE_ERROR */
typedef struct S79 T79;

/* [detachable] UT_VERSION_NUMBER */
typedef struct S81 T81;

/* [detachable] UT_MESSAGE */
typedef struct S82 T82;

/* [detachable] RX_PCRE_REGULAR_EXPRESSION */
typedef struct S83 T83;

/* [detachable] KL_STRING_ROUTINES */
typedef struct S84 T84;

/* [detachable] KL_EXECUTION_ENVIRONMENT */
typedef struct S87 T87;

/* [detachable] AP_ERROR */
typedef struct S88 T88;

/* [detachable] VOID_TARGET */
typedef struct S91 T91;

/* [detachable] TYPE [[attached] VOID_TARGET] */
#define T92 EIF_TYPE_OBJ

/* [detachable] ROUTINE_FAILURE */
typedef struct S93 T93;

/* [detachable] TYPE [[attached] ROUTINE_FAILURE] */
#define T94 EIF_TYPE_OBJ

/* [detachable] OLD_VIOLATION */
typedef struct S95 T95;

/* [detachable] TYPE [[attached] OLD_VIOLATION] */
#define T96 EIF_TYPE_OBJ

/* [detachable] NO_MORE_MEMORY */
typedef struct S97 T97;

/* [detachable] INVARIANT_VIOLATION */
typedef struct S98 T98;

/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
typedef struct S99 T99;

/* [detachable] IO_FAILURE */
typedef struct S100 T100;

/* [detachable] OPERATING_SYSTEM_FAILURE */
typedef struct S101 T101;

/* [detachable] COM_FAILURE */
typedef struct S102 T102;

/* [detachable] EIFFEL_RUNTIME_PANIC */
typedef struct S103 T103;

/* [detachable] PRECONDITION_VIOLATION */
typedef struct S105 T105;

/* [detachable] POSTCONDITION_VIOLATION */
typedef struct S106 T106;

/* [detachable] FLOATING_POINT_FAILURE */
typedef struct S107 T107;

/* [detachable] CHECK_VIOLATION */
typedef struct S108 T108;

/* [detachable] BAD_INSPECT_VALUE */
typedef struct S109 T109;

/* [detachable] VARIANT_VIOLATION */
typedef struct S110 T110;

/* [detachable] LOOP_INVARIANT_VIOLATION */
typedef struct S111 T111;

/* [detachable] RESCUE_FAILURE */
typedef struct S112 T112;

/* [detachable] RESUMPTION_FAILURE */
typedef struct S113 T113;

/* [detachable] CREATE_ON_DEFERRED */
typedef struct S114 T114;

/* [detachable] EXTERNAL_FAILURE */
typedef struct S115 T115;

/* [detachable] VOID_ASSIGNED_TO_EXPANDED */
typedef struct S116 T116;

/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
typedef struct S117 T117;

/* [detachable] MISMATCH_FAILURE */
typedef struct S118 T118;

/* [detachable] DEVELOPER_EXCEPTION */
typedef struct S119 T119;

/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
typedef struct S120 T120;

/* [detachable] SERIALIZATION_FAILURE */
typedef struct S121 T121;

/* [detachable] EXECUTION_ENVIRONMENT */
typedef struct S122 T122;

/* [detachable] KL_WINDOWS_FILE_SYSTEM */
typedef struct S123 T123;

/* [detachable] KL_UNIX_FILE_SYSTEM */
typedef struct S124 T124;

/* [detachable] PRIMES */
typedef struct S125 T125;

/* [detachable] SPECIAL [[attached] INTEGER_32] */
typedef struct S126 T126;

/* [detachable] SPECIAL [[attached] BOOLEAN] */
typedef struct S127 T127;

/* detachable KL_LINKABLE [[attached] CHARACTER_8] */
typedef struct S129 T129;

/* [detachable] MANAGED_POINTER */
typedef struct S131 T131;

/* [detachable] FILE_INFO */
typedef struct S132 T132;

/* [detachable] MUTEX */
typedef struct S133 T133;

/* [detachable] UT_ERROR_HANDLER */
typedef struct S134 T134;

/* [detachable] KL_STANDARD_FILES */
typedef struct S135 T135;

/* [detachable] KL_STDERR_FILE */
typedef struct S136 T136;

/* [detachable] KL_STDOUT_FILE */
typedef struct S137 T137;

/* [detachable] ST_WORD_WRAPPER */
typedef struct S138 T138;

/* [detachable] AP_DISPLAY_HELP_FLAG */
typedef struct S140 T140;

/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
typedef struct S141 T141;

/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S142 T142;

/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
typedef struct S143 T143;

/* [detachable] AP_ERROR_HANDLER */
typedef struct S144 T144;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
typedef struct S145 T145;

/* [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
typedef struct S146 T146;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
typedef struct S147 T147;

/* [detachable] DS_LINKED_LIST [[attached] STRING_8] */
typedef struct S148 T148;

/* [detachable] KL_STRING_EQUALITY_TESTER */
typedef struct S149 T149;

/* detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
typedef struct S150 T150;

/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
typedef struct S153 T153;

/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
typedef struct S156 T156;

/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
typedef struct S157 T157;

/* [detachable] ET_ECF_AST_FACTORY */
typedef struct S158 T158;

/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
typedef struct S159 T159;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET] */
typedef struct S160 T160;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
typedef struct S161 T161;

/* detachable TUPLE */
typedef struct S162 T162;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
typedef struct S163 T163;

/* [detachable] XM_ELEMENT */
typedef struct S164 T164;

/* detachable XM_POSITION_TABLE */
typedef struct S165 T165;

/* [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
typedef struct S166 T166;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
typedef struct S167 T167;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
typedef struct S168 T168;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
typedef struct S170 T170;

/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
typedef struct S171 T171;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
typedef struct S172 T172;

/* [detachable] XM_EIFFEL_PARSER */
typedef struct S174 T174;

/* [detachable] XM_TREE_CALLBACKS_PIPE */
typedef struct S175 T175;

/* [detachable] XM_CALLBACKS_TO_TREE_FILTER */
typedef struct S178 T178;

/* [detachable] ET_ECF_STATE */
typedef struct S179 T179;

/* [detachable] ET_IDENTIFIER */
typedef struct S180 T180;

/* [detachable] ET_ECF_SYSTEM_CONFIG */
typedef struct S181 T181;

/* [detachable] ET_ECF_LIBRARY */
typedef struct S183 T183;

/* [detachable] ET_ADAPTED_LIBRARIES */
typedef struct S184 T184;

/* [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S186 T186;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
typedef struct S187 T187;

/* [detachable] XM_DOCUMENT */
typedef struct S188 T188;

/* [detachable] ET_COMPRESSED_POSITION */
typedef struct S189 T189;

/* [detachable] XM_STOP_ON_ERROR_FILTER */
typedef struct S191 T191;

/* detachable XM_ATTRIBUTE */
typedef struct S192 T192;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
typedef struct S194 T194;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S195 T195;

/* detachable ET_ECF_TARGET_PARENT */
typedef struct S196 T196;

/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
typedef struct S197 T197;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
typedef struct S198 T198;

/* detachable ET_ECF_TARGETS */
typedef struct S199 T199;

/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S200 T200;

/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S201 T201;

/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
typedef struct S202 T202;

/* [detachable] ET_ECF_ADAPTED_LIBRARY */
typedef struct S203 T203;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
typedef struct S204 T204;

/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
typedef struct S206 T206;

/* [detachable] ET_ECF_DOTNET_ASSEMBLY */
typedef struct S207 T207;

/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
typedef struct S209 T209;

/* [detachable] ET_ECF_ERROR */
typedef struct S210 T210;

/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
typedef struct S211 T211;

/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
typedef struct S212 T212;

/* [detachable] TUPLE [[attached] STRING_8] */
typedef struct S213 T213;

/* [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
typedef struct S214 T214;

/* [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
typedef struct S215 T215;

/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
typedef struct S217 T217;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE */
typedef struct S219 T219;

/* [detachable] ET_DYNAMIC_FEATURE_LIST */
typedef struct S220 T220;

/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
typedef struct S221 T221;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
typedef struct S222 T222;

/* [detachable] TUPLE [[attached] ET_CLASS] */
typedef struct S223 T223;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
typedef struct S224 T224;

/* [detachable] TUPLE [INTEGER_32] */
typedef struct S225 T225;

/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
typedef struct S226 T226;

/* [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
typedef struct S227 T227;

/* detachable ET_DYNAMIC_FEATURE */
typedef struct S232 T232;

/* [detachable] ET_ACTUAL_PARAMETER_LIST */
typedef struct S236 T236;

/* [detachable] ET_NESTED_TYPE_CONTEXT */
typedef struct S237 T237;

/* [detachable] ET_FORMAL_PARAMETER_TYPE */
typedef struct S246 T246;

/* [detachable] ET_EXTERNAL_FUNCTION */
typedef struct S247 T247;

/* [detachable] ET_TUPLE_TYPE */
typedef struct S248 T248;

/* detachable ET_FORMAL_ARGUMENT_LIST */
typedef struct S249 T249;

/* [detachable] ARRAY [[attached] ET_TYPE] */
typedef struct S250 T250;

/* [detachable] SPECIAL [[attached] ET_TYPE] */
typedef struct S251 T251;

/* [detachable] ET_FORMAL_ARGUMENT */
typedef struct S252 T252;

/* [detachable] ET_QUERY_LIST */
typedef struct S253 T253;

/* [detachable] ET_PROCEDURE_LIST */
typedef struct S254 T254;

/* [detachable] ET_DYNAMIC_TYPE_SET_LIST */
typedef struct S255 T255;

/* [detachable] KL_STRING_OUTPUT_STREAM */
typedef struct S256 T256;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
typedef struct S257 T257;

/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S258 T258;

/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
typedef struct S259 T259;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
typedef struct S260 T260;

/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
typedef struct S261 T261;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
typedef struct S262 T262;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
typedef struct S263 T263;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
typedef struct S264 T264;

/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
typedef struct S265 T265;

/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S266 T266;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S267 T267;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
typedef struct S268 T268;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S269 T269;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S270 T270;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
typedef struct S271 T271;

/* [detachable] ET_CURRENT */
typedef struct S273 T273;

/* [detachable] ET_ACTUAL_ARGUMENT_LIST */
typedef struct S274 T274;

/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
typedef struct S275 T275;

/* [detachable] ET_QUALIFIED_CALL_EXPRESSION */
typedef struct S278 T278;

/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
typedef struct S279 T279;

/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S280 T280;

/* [detachable] ET_MANIFEST_TUPLE */
typedef struct S281 T281;

/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S282 T282;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
typedef struct S283 T283;

/* [detachable] ARRAY [[attached] INTEGER_32] */
typedef struct S284 T284;

/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
typedef struct S285 T285;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
typedef struct S286 T286;

/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S287 T287;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S288 T288;

/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
typedef struct S289 T289;

/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
typedef struct S290 T290;

/* [detachable] ET_IDENTIFIER_TESTER */
typedef struct S291 T291;

/* [detachable] KL_TEXT_OUTPUT_FILE */
typedef struct S293 T293;

/* [detachable] ET_DYNAMIC_TUPLE_TYPE */
typedef struct S294 T294;

/* [detachable] DS_STRING_HASH_TABLE */
typedef struct S295 T295;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
typedef struct S296 T296;

/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
typedef struct S297 T297;

/* detachable ET_DYNAMIC_PRECURSOR */
typedef struct S298 T298;

/* detachable ET_DYNAMIC_PRECURSOR_LIST */
typedef struct S299 T299;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S300 T300;

/* [detachable] ET_RESULT */
typedef struct S301 T301;

/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
typedef struct S302 T302;

/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
typedef struct S306 T306;

/* [detachable] ET_OBJECT_TEST */
typedef struct S308 T308;

/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
typedef struct S309 T309;

/* [detachable] ET_EQUALITY_EXPRESSION */
typedef struct S310 T310;

/* [detachable] ET_DYNAMIC_SPECIAL_TYPE */
typedef struct S311 T311;

/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S314 T314;

/* detachable ET_DYNAMIC_SECONDARY_TYPE */
typedef struct S315 T315;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
typedef struct S316 T316;

/* [detachable] ET_CREATE_EXPRESSION */
typedef struct S320 T320;

/* [detachable] ET_QUALIFIED_CALL */
typedef struct S321 T321;

/* [detachable] ARRAY [[attached] STRING_8] */
typedef struct S324 T324;

/* [detachable] SPECIAL [[attached] STRING_8] */
typedef struct S325 T325;

/* [detachable] KL_DIRECTORY */
typedef struct S327 T327;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S328 T328;

/* [detachable] ET_EXTENDED_ATTRIBUTE */
typedef struct S329 T329;

/* detachable ET_COMPOUND */
typedef struct S330 T330;

/* detachable ET_LOCAL_VARIABLE_LIST */
typedef struct S331 T331;

/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
typedef struct S335 T335;

/* [detachable] ET_IMPLICIT_TYPE_MARK */
typedef struct S342 T342;

/* [detachable] ET_SYMBOL */
typedef struct S343 T343;

/* [detachable] PROCEDURE [[attached] TUPLE] */
typedef struct S344 T344;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S345 T345;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S346 T346;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S347 T347;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S348 T348;

/* [detachable] UT_INTEGER_FORMATTER */
typedef struct S349 T349;

/* [detachable] STD_FILES */
typedef struct S352 T352;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S354 T354;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S356 T356;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S357 T357;

/* detachable ET_ACROSS_COMPONENT_LIST */
typedef struct S361 T361;

/* [detachable] ET_CONSTANT_ATTRIBUTE */
typedef struct S363 T363;

/* [detachable] ET_UNIQUE_ATTRIBUTE */
typedef struct S365 T365;

/* [detachable] ET_REGULAR_INTEGER_CONSTANT */
typedef struct S366 T366;

/* [detachable] ET_ATTRIBUTE */
typedef struct S367 T367;

/* [detachable] NATIVE_STRING */
typedef struct S370 T370;

/* [detachable] ET_SYSTEM_MULTIPROCESSOR */
typedef struct S371 T371;

/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
typedef struct S372 T372;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
typedef struct S373 T373;

/* [detachable] TYPE [[attached] STRING_8] */
#define T374 EIF_TYPE_OBJ

/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
typedef struct S375 T375;

/* [detachable] ET_FEATURE_CHECKER */
typedef struct S376 T376;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
typedef struct S377 T377;

/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
typedef struct S378 T378;

/* [detachable] ET_TYPE_CHECKER */
typedef struct S380 T380;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
typedef struct S382 T382;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S383 T383;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S384 T384;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S385 T385;

/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S386 T386;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S387 T387;

/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
typedef struct S388 T388;

/* [detachable] ET_OBJECT_TEST_SCOPE */
typedef struct S389 T389;

/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
typedef struct S390 T390;

/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
typedef struct S391 T391;

/* [detachable] ET_ACROSS_CURSOR_SCOPE */
typedef struct S392 T392;

/* [detachable] ET_ATTACHMENT_SCOPE */
typedef struct S393 T393;

/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
typedef struct S394 T394;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S395 T395;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
typedef struct S396 T396;

/* [detachable] ET_CLIENT_LIST */
typedef struct S397 T397;

/* [detachable] ET_CLIENT */
typedef struct S398 T398;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S399 T399;

/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
typedef struct S400 T400;

/* detachable ET_PRECONDITIONS */
typedef struct S401 T401;

/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
typedef struct S402 T402;

/* detachable ET_FEATURE_LIST */
typedef struct S404 T404;

/* [detachable] ET_UNKNOWN_GROUP */
typedef struct S406 T406;

/* [detachable] ET_BASE_TYPE_LIST */
typedef struct S408 T408;

/* [detachable] ET_KEYWORD */
typedef struct S409 T409;

/* [detachable] ET_CLASS_CODES */
typedef struct S410 T410;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S411 T411;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S412 T412;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S413 T413;

/* [detachable] UC_UTF8_ROUTINES */
typedef struct S416 T416;

/* [detachable] PATH */
typedef struct S417 T417;

/* [detachable] RX_BYTE_CODE */
typedef struct S419 T419;

/* [detachable] RX_CHARACTER_SET */
typedef struct S420 T420;

/* [detachable] RX_CASE_MAPPING */
typedef struct S421 T421;

/* [detachable] UC_UNICODE_ROUTINES */
typedef struct S423 T423;

/* [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
typedef struct S424 T424;

/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
typedef struct S425 T425;

/* [detachable] UC_STRING */
typedef struct S426 T426;

/* [detachable] KL_ANY_ROUTINES */
typedef struct S427 T427;

/* [detachable] STRING_TO_INTEGER_CONVERTOR */
typedef struct S428 T428;

/* [detachable] ARGUMENTS_32 */
typedef struct S431 T431;

/* [detachable] ET_ECF_OPTIONS */
typedef struct S437 T437;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S438 T438;

/* [detachable] XM_NAMESPACE */
typedef struct S439 T439;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
typedef struct S440 T440;

/* [detachable] ET_LIKE_CURRENT */
typedef struct S442 T442;

/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
typedef struct S443 T443;

/* [detachable] ET_DYNAMIC_FUNCTION_TYPE */
typedef struct S444 T444;

/* [detachable] DT_SHARED_SYSTEM_CLOCK */
typedef struct S445 T445;

/* [detachable] DT_SYSTEM_CLOCK */
typedef struct S446 T446;

/* [detachable] ET_EIFFEL_PREPARSER */
typedef struct S447 T447;

/* [detachable] ET_MASTER_CLASS_CHECKER */
typedef struct S448 T448;

/* [detachable] ET_EIFFEL_PARSER */
typedef struct S449 T449;

/* [detachable] ET_PROVIDER_CHECKER */
typedef struct S450 T450;

/* [detachable] ET_ANCESTOR_BUILDER */
typedef struct S451 T451;

/* [detachable] ET_FEATURE_FLATTENER */
typedef struct S452 T452;

/* [detachable] ET_INTERFACE_CHECKER */
typedef struct S453 T453;

/* [detachable] ET_IMPLEMENTATION_CHECKER */
typedef struct S454 T454;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
typedef struct S455 T455;

/* [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
typedef struct S456 T456;

/* [detachable] ET_AST_NULL_PROCESSOR */
typedef struct S457 T457;

/* [detachable] ET_AST_FACTORY */
typedef struct S458 T458;

/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
typedef struct S460 T460;

/* [detachable] DT_DATE_TIME_DURATION */
typedef struct S461 T461;

/* [detachable] UT_CONFIG_PARSER */
typedef struct S462 T462;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
typedef struct S463 T463;

/* [detachable] ET_DO_PROCEDURE */
typedef struct S466 T466;

/* [detachable] ET_DEFERRED_PROCEDURE */
typedef struct S468 T468;

/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
typedef struct S469 T469;

/* [detachable] KL_PLATFORM */
typedef struct S470 T470;

/* UTF_CONVERTER */
typedef struct S471 T471;
extern T0* GE_boxed471(T471 a1);
typedef struct Sb471 Tb471;

/* [detachable] CELL [[attached] INTEGER_32] */
typedef struct S472 T472;

/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
typedef struct S473 T473;

/* [detachable] KL_PATHNAME */
typedef struct S475 T475;

/* [detachable] EXCEPTIONS */
typedef struct S476 T476;

/* [detachable] SPECIAL [[attached] NATURAL_8] */
typedef struct S477 T477;

/* TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
typedef struct S478 T478;
extern T0* GE_boxed478(T478 a1);
typedef struct Sb478 Tb478;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
typedef struct S479 T479;

/* [detachable] SPECIAL [[attached] AP_OPTION] */
typedef struct S481 T481;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
typedef struct S482 T482;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S483 T483;

/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S484 T484;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S485 T485;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S486 T486;

/* [detachable] SPECIAL [detachable STRING_8] */
typedef struct S487 T487;

/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
typedef struct S488 T488;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S489 T489;

/* detachable DS_LINKABLE [[attached] STRING_8] */
typedef struct S490 T490;

/* detachable DS_LINKABLE [[attached] BOOLEAN] */
typedef struct S491 T491;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
typedef struct S492 T492;

/* detachable DS_LINKABLE [[attached] INTEGER_32] */
typedef struct S493 T493;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
typedef struct S494 T494;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
typedef struct S496 T496;

/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
typedef struct S497 T497;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
typedef struct S500 T500;

/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
typedef struct S501 T501;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
typedef struct S504 T504;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
typedef struct S505 T505;

/* [detachable] XM_EIFFEL_SCANNER */
typedef struct S506 T506;

/* [detachable] XM_DEFAULT_POSITION */
typedef struct S507 T507;

/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
typedef struct S509 T509;

/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
typedef struct S510 T510;

/* [detachable] XM_CALLBACKS_NULL */
typedef struct S511 T511;

/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
typedef struct S512 T512;

/* [detachable] XM_NULL_EXTERNAL_RESOLVER */
typedef struct S514 T514;

/* [detachable] XM_DTD_CALLBACKS_NULL */
typedef struct S516 T516;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
typedef struct S517 T517;

/* [detachable] SPECIAL [detachable ANY] */
typedef struct S518 T518;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S519 T519;

/* [detachable] XM_EIFFEL_PARSER_NAME */
typedef struct S520 T520;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S521 T521;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
typedef struct S522 T522;

/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S523 T523;

/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
typedef struct S524 T524;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
typedef struct S525 T525;

/* [detachable] XM_DTD_EXTERNAL_ID */
typedef struct S526 T526;

/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
typedef struct S527 T527;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S528 T528;

/* [detachable] XM_DTD_ELEMENT_CONTENT */
typedef struct S529 T529;

/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S530 T530;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
typedef struct S531 T531;

/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S532 T532;

/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
typedef struct S533 T533;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S534 T534;

/* [detachable] XM_DTD_ATTRIBUTE_CONTENT */
typedef struct S535 T535;

/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S536 T536;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
typedef struct S537 T537;

/* [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
typedef struct S538 T538;

/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
typedef struct S539 T539;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
typedef struct S540 T540;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
typedef struct S541 T541;

/* [detachable] XM_EIFFEL_DECLARATION */
typedef struct S542 T542;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
typedef struct S543 T543;

/* [detachable] XM_EIFFEL_ENTITY_DEF */
typedef struct S546 T546;

/* [detachable] XM_EIFFEL_SCANNER_DTD */
typedef struct S547 T547;

/* [detachable] XM_EIFFEL_PE_ENTITY_DEF */
typedef struct S549 T549;

/* [detachable] XM_NAMESPACE_RESOLVER */
typedef struct S550 T550;

/* [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
typedef struct S551 T551;

/* [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
typedef struct S552 T552;

/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
typedef struct S553 T553;

/* [detachable] ET_CLUSTERS */
typedef struct S554 T554;

/* [detachable] ET_CLASS_NAME_TESTER */
typedef struct S555 T555;

/* [detachable] ET_PARENT */
typedef struct S557 T557;

/* [detachable] ET_PARENT_LIST */
typedef struct S558 T558;

/* [detachable] ET_BUILTIN_CONVERT_FEATURE */
typedef struct S559 T559;

/* [detachable] ET_MASTER_CLASS */
typedef struct S560 T560;

/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
typedef struct S561 T561;

/* detachable ET_RENAME_LIST */
typedef struct S562 T562;

/* detachable ET_EXPORT_LIST */
typedef struct S563 T563;

/* detachable ET_KEYWORD_FEATURE_NAME_LIST */
typedef struct S564 T564;

/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
typedef struct S567 T567;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S568 T568;

/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
typedef struct S569 T569;

/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
typedef struct S572 T572;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
typedef struct S574 T574;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
typedef struct S575 T575;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S576 T576;

/* [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
typedef struct S577 T577;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
typedef struct S579 T579;

/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S581 T581;

/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
typedef struct S582 T582;

/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
typedef struct S583 T583;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S584 T584;

/* [detachable] ET_DOTNET_ASSEMBLIES */
typedef struct S585 T585;

/* [detachable] ET_DOTNET_ASSEMBLY */
typedef struct S586 T586;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
typedef struct S589 T589;

/* [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
typedef struct S590 T590;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
typedef struct S591 T591;

/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
typedef struct S592 T592;

/* detachable ET_FORMAL_PARAMETER_LIST */
typedef struct S593 T593;

/* detachable ET_FEATURE_IDS */
typedef struct S594 T594;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S595 T595;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S596 T596;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
typedef struct S599 T599;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
typedef struct S600 T600;

/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
typedef struct S601 T601;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
typedef struct S602 T602;

/* [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
typedef struct S603 T603;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
typedef struct S605 T605;

/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S609 T609;

/* [detachable] ET_BRACKET_SYMBOL */
typedef struct S610 T610;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S612 T612;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
typedef struct S613 T613;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
typedef struct S614 T614;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
typedef struct S615 T615;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
typedef struct S616 T616;

/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S617 T617;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S618 T618;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S619 T619;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
typedef struct S620 T620;

/* [detachable] SPECIAL [[attached] ET_EXPRESSION] */
typedef struct S622 T622;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
typedef struct S623 T623;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
typedef struct S624 T624;

/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
typedef struct S625 T625;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
typedef struct S626 T626;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S627 T627;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S628 T628;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S629 T629;

/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
typedef struct S631 T631;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
typedef struct S632 T632;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
typedef struct S633 T633;

/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S634 T634;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S635 T635;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S636 T636;

/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S637 T637;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S638 T638;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S639 T639;

/* [detachable] SPECIAL [[attached] ET_AGENT] */
typedef struct S640 T640;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
typedef struct S641 T641;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
typedef struct S642 T642;

/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
typedef struct S643 T643;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
typedef struct S645 T645;

/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S646 T646;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S647 T647;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S648 T648;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S649 T649;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S650 T650;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S651 T651;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
typedef struct S653 T653;

/* [detachable] SPECIAL [[attached] ET_FEATURE] */
typedef struct S654 T654;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
typedef struct S655 T655;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
typedef struct S656 T656;

/* [detachable] SPECIAL [[attached] ET_CONSTANT] */
typedef struct S657 T657;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
typedef struct S659 T659;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
typedef struct S660 T660;

/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
typedef struct S661 T661;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
typedef struct S663 T663;

/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
typedef struct S664 T664;

/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S666 T666;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
typedef struct S667 T667;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S668 T668;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
typedef struct S669 T669;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
typedef struct S672 T672;

/* [detachable] TYPE [[attached] ET_IDENTIFIER] */
#define T673 EIF_TYPE_OBJ

/* TYPED_POINTER [[attached] ANY] */
typedef struct S674 T674;
extern T0* GE_boxed674(T674 a1);
typedef struct Sb674 Tb674;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S675 T675;

/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S676 T676;

/* [detachable] ARRAY [[attached] BOOLEAN] */
typedef struct S677 T677;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S680 T680;

/* [detachable] WORKER_THREAD */
typedef struct S681 T681;

/* [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
typedef struct S682 T682;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
typedef struct S686 T686;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
typedef struct S687 T687;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
typedef struct S688 T688;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
typedef struct S689 T689;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
typedef struct S690 T690;

/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
typedef struct S691 T691;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
typedef struct S692 T692;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S693 T693;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
typedef struct S694 T694;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S695 T695;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S696 T696;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S697 T697;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
typedef struct S698 T698;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S699 T699;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S700 T700;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S701 T701;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
typedef struct S702 T702;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S703 T703;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S704 T704;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S705 T705;

/* [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S706 T706;

/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S707 T707;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S708 T708;

/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S709 T709;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S710 T710;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S711 T711;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
typedef struct S712 T712;

/* [detachable] ET_NAMED_OBJECT_TEST */
typedef struct S715 T715;

/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S716 T716;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S717 T717;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S718 T718;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
typedef struct S719 T719;

/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT] */
typedef struct S721 T721;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT] */
typedef struct S722 T722;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT] */
typedef struct S723 T723;

/* [detachable] DS_HASH_SET [[attached] INTEGER_32] */
typedef struct S724 T724;

/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S726 T726;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S727 T727;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S728 T728;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
typedef struct S729 T729;

/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
typedef struct S731 T731;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
typedef struct S732 T732;

/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
typedef struct S733 T733;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
typedef struct S734 T734;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S735 T735;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S736 T736;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S737 T737;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
typedef struct S738 T738;

/* [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
typedef struct S739 T739;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
typedef struct S741 T741;

/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
typedef struct S742 T742;

/* [detachable] UC_UTF8_STRING */
typedef struct S743 T743;

/* [detachable] SPECIAL [[attached] NATURAL_32] */
typedef struct S745 T745;

/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
typedef struct S746 T746;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
typedef struct S747 T747;

/* [detachable] SPECIAL [[attached] NATURAL_64] */
typedef struct S748 T748;

/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
typedef struct S749 T749;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
typedef struct S750 T750;

/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
typedef struct S751 T751;

/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
typedef struct S752 T752;

/* [detachable] KL_INTEGER_ROUTINES */
typedef struct S753 T753;

/* [detachable] STRING_8_SEARCHER */
typedef struct S755 T755;

/* detachable ARRAYED_LIST [[attached] INTEGER_32] */
typedef struct S756 T756;

/* [detachable] ET_SYSTEM_ERROR */
typedef struct S757 T757;

/* [detachable] ET_INTERNAL_ERROR */
typedef struct S758 T758;

/* [detachable] ET_VALIDITY_ERROR */
typedef struct S759 T759;

/* detachable ET_ECF_NOTE_ELEMENT */
typedef struct S760 T760;

/* detachable ET_ECF_CLUSTERS */
typedef struct S761 T761;

/* detachable ET_ECF_ADAPTED_LIBRARIES */
typedef struct S762 T762;

/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
typedef struct S763 T763;

/* detachable ET_ECF_FILE_RULES */
typedef struct S764 T764;

/* detachable ET_ECF_EXTERNAL_CFLAGS */
typedef struct S765 T765;

/* detachable ET_ECF_EXTERNAL_INCLUDES */
typedef struct S766 T766;

/* detachable ET_ECF_EXTERNAL_LIBRARIES */
typedef struct S767 T767;

/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
typedef struct S768 T768;

/* detachable ET_ECF_EXTERNAL_MAKES */
typedef struct S769 T769;

/* detachable ET_ECF_EXTERNAL_OBJECTS */
typedef struct S770 T770;

/* detachable ET_ECF_EXTERNAL_RESOURCES */
typedef struct S771 T771;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
typedef struct S772 T772;

/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
typedef struct S773 T773;

/* detachable ET_ECF_VERSION */
typedef struct S775 T775;

/* detachable ET_ECF_CLUSTER */
typedef struct S776 T776;

/* detachable ET_ECF_EXTERNAL_CFLAG */
typedef struct S777 T777;

/* detachable ET_ECF_EXTERNAL_INCLUDE */
typedef struct S778 T778;

/* detachable ET_ECF_EXTERNAL_LIBRARY */
typedef struct S779 T779;

/* detachable ET_ECF_EXTERNAL_LINKER_FLAG */
typedef struct S780 T780;

/* detachable ET_ECF_EXTERNAL_MAKE */
typedef struct S781 T781;

/* detachable ET_ECF_EXTERNAL_OBJECT */
typedef struct S782 T782;

/* detachable ET_ECF_EXTERNAL_RESOURCE */
typedef struct S783 T783;

/* detachable ET_ECF_FILE_RULE */
typedef struct S784 T784;

/* detachable ET_ECF_ACTION */
typedef struct S785 T785;

/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S786 T786;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S787 T787;

/* [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
typedef struct S788 T788;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
typedef struct S789 T789;

/* detachable ET_ECF_ROOT_CLASS */
typedef struct S790 T790;

/* [detachable] ET_ECF_ROOT_ALL_CLASSES */
typedef struct S791 T791;

/* [detachable] ET_ECF_ORED_CONDITIONS */
typedef struct S792 T792;

/* detachable ET_ECF_ANDED_CONDITIONS */
typedef struct S793 T793;

/* detachable ET_ECF_VISIBLE_CLASS */
typedef struct S794 T794;

/* detachable ET_ECF_BUILD_CONDITION */
typedef struct S796 T796;

/* detachable ET_ECF_CONCURRENCY_CONDITION */
typedef struct S797 T797;

/* detachable ET_ECF_VOID_SAFETY_CONDITION */
typedef struct S798 T798;

/* detachable ET_ECF_CUSTOM_CONDITION */
typedef struct S799 T799;

/* detachable ET_ECF_DOTNET_CONDITION */
typedef struct S800 T800;

/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
typedef struct S801 T801;

/* detachable ET_ECF_PLATFORM_CONDITION */
typedef struct S802 T802;

/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
typedef struct S804 T804;

/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
typedef struct S805 T805;

/* [detachable] UT_COUNTER */
typedef struct S807 T807;

/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
typedef struct S808 T808;

/* [detachable] TUPLE [[attached] UT_COUNTER] */
typedef struct S809 T809;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
typedef struct S810 T810;

/* [detachable] TUPLE [[attached] ET_UNIVERSE] */
typedef struct S811 T811;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
typedef struct S812 T812;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
typedef struct S813 T813;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S814 T814;

/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S816 T816;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S817 T817;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S818 T818;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S819 T819;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S820 T820;

/* [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S821 T821;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
typedef struct S822 T822;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
typedef struct S823 T823;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
typedef struct S824 T824;

/* [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S825 T825;

/* [detachable] ET_LIBRARY */
typedef struct S826 T826;

/* [detachable] TUPLE [[attached] ET_LIBRARY] */
typedef struct S827 T827;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
typedef struct S828 T828;

/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
typedef struct S829 T829;

/* [detachable] TUPLE [[attached] ET_CLUSTER] */
typedef struct S831 T831;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
typedef struct S832 T832;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
typedef struct S833 T833;

/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S834 T834;

/* [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
typedef struct S835 T835;

/* [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
typedef struct S836 T836;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
typedef struct S837 T837;

/* [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S838 T838;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
typedef struct S839 T839;

/* [detachable] ET_ANCESTORS_STATUS_CHECKER */
typedef struct S840 T840;

/* [detachable] ET_FLATTENING_STATUS_CHECKER */
typedef struct S841 T841;

/* [detachable] ET_INTERFACE_STATUS_CHECKER */
typedef struct S842 T842;

/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
typedef struct S843 T843;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
typedef struct S844 T844;

/* [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
typedef struct S846 T846;

/* [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
typedef struct S847 T847;

/* [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
typedef struct S848 T848;

/* [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
typedef struct S849 T849;

/* [detachable] TUPLE [BOOLEAN] */
typedef struct S850 T850;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
typedef struct S851 T851;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
typedef struct S854 T854;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
typedef struct S858 T858;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
typedef struct S859 T859;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
typedef struct S860 T860;

/* [detachable] ET_NONE_GROUP */
typedef struct S861 T861;

/* detachable ET_PARENTHESIS_EXPRESSION */
typedef struct S862 T862;

/* [detachable] ET_REGULAR_MANIFEST_STRING */
typedef struct S863 T863;

/* [detachable] ET_INFIX_FREE_NAME */
typedef struct S864 T864;

/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S866 T866;

/* [detachable] ET_FORMAL_PARAMETER */
typedef struct S867 T867;

/* TYPED_POINTER [[attached] NATURAL_8] */
typedef struct S868 T868;
extern T0* GE_boxed868(T868 a1);
typedef struct Sb868 Tb868;

/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
typedef struct S869 T869;

/* [detachable] UC_STRING_EQUALITY_TESTER */
typedef struct S871 T871;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S873 T873;

/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S874 T874;

/* [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
typedef struct S875 T875;

/* [detachable] XM_EIFFEL_INPUT_STREAM */
typedef struct S876 T876;

/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
typedef struct S877 T877;

/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
typedef struct S878 T878;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
typedef struct S879 T879;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
typedef struct S880 T880;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
typedef struct S881 T881;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S882 T882;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S884 T884;

/* [detachable] SPECIAL [[attached] ET_QUERY] */
typedef struct S885 T885;

/* detachable ET_ALIAS_NAME */
typedef struct S886 T886;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
typedef struct S887 T887;

/* [detachable] SPECIAL [[attached] ET_PROCEDURE] */
typedef struct S888 T888;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
typedef struct S889 T889;

/* [detachable] TYPE [[attached] INTEGER_32] */
#define T894 EIF_TYPE_OBJ

/* [detachable] CONSOLE */
typedef struct S896 T896;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
typedef struct S897 T897;

/* [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
typedef struct S899 T899;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
typedef struct S900 T900;

/* [detachable] INTEGER_OVERFLOW_CHECKER */
typedef struct S901 T901;

/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
typedef struct S902 T902;

/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
typedef struct S903 T903;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S904 T904;

/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S905 T905;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S906 T906;

/* detachable C_DATE */
typedef struct S907 T907;

/* [detachable] YY_FILE_BUFFER */
typedef struct S908 T908;

/* [detachable] KL_STDIN_FILE */
typedef struct S909 T909;

/* [detachable] YY_BUFFER */
typedef struct S910 T910;

/* [detachable] YY_UNICODE_FILE_BUFFER */
typedef struct S914 T914;

/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
typedef struct S915 T915;

/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S916 T916;

/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S917 T917;

/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
typedef struct S918 T918;

/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
typedef struct S919 T919;

/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
typedef struct S920 T920;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S921 T921;

/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST] */
typedef struct S922 T922;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST] */
typedef struct S923 T923;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
typedef struct S924 T924;

/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
typedef struct S925 T925;

/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
typedef struct S926 T926;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
typedef struct S927 T927;

/* [detachable] SPECIAL [detachable ET_KEYWORD] */
typedef struct S928 T928;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
typedef struct S929 T929;

/* detachable ET_AGENT_KEYWORD */
typedef struct S930 T930;

/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
typedef struct S931 T931;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
typedef struct S932 T932;

/* detachable ET_PRECURSOR_KEYWORD */
typedef struct S933 T933;

/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
typedef struct S934 T934;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
typedef struct S935 T935;

/* [detachable] SPECIAL [detachable ET_SYMBOL] */
typedef struct S936 T936;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
typedef struct S937 T937;

/* [detachable] SPECIAL [detachable ET_POSITION] */
typedef struct S938 T938;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
typedef struct S939 T939;

/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
typedef struct S941 T941;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
typedef struct S942 T942;

/* [detachable] SPECIAL [detachable ET_BREAK] */
typedef struct S944 T944;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
typedef struct S945 T945;

/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
typedef struct S947 T947;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
typedef struct S948 T948;

/* [detachable] SPECIAL [detachable ET_CURRENT] */
typedef struct S949 T949;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
typedef struct S950 T950;

/* detachable ET_FREE_OPERATOR */
typedef struct S951 T951;

/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
typedef struct S952 T952;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
typedef struct S953 T953;

/* [detachable] SPECIAL [detachable ET_IDENTIFIER] */
typedef struct S954 T954;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
typedef struct S955 T955;

/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
typedef struct S957 T957;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
typedef struct S958 T958;

/* detachable ET_KEYWORD_OPERATOR */
typedef struct S959 T959;

/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
typedef struct S960 T960;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
typedef struct S961 T961;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
typedef struct S962 T962;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
typedef struct S963 T963;

/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
typedef struct S965 T965;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
typedef struct S966 T966;

/* [detachable] SPECIAL [detachable ET_RESULT] */
typedef struct S967 T967;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
typedef struct S968 T968;

/* detachable ET_RETRY_INSTRUCTION */
typedef struct S969 T969;

/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
typedef struct S970 T970;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
typedef struct S971 T971;

/* detachable ET_SYMBOL_OPERATOR */
typedef struct S972 T972;

/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
typedef struct S973 T973;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
typedef struct S974 T974;

/* detachable ET_VOID */
typedef struct S975 T975;

/* [detachable] SPECIAL [detachable ET_VOID] */
typedef struct S976 T976;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
typedef struct S977 T977;

/* detachable ET_SEMICOLON_SYMBOL */
typedef struct S978 T978;

/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
typedef struct S979 T979;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
typedef struct S980 T980;

/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
typedef struct S981 T981;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
typedef struct S982 T982;

/* detachable ET_QUESTION_MARK_SYMBOL */
typedef struct S983 T983;

/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
typedef struct S984 T984;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
typedef struct S985 T985;

/* detachable ET_ACROSS_EXPRESSION */
typedef struct S986 T986;

/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
typedef struct S987 T987;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
typedef struct S988 T988;

/* detachable ET_ACROSS_INSTRUCTION */
typedef struct S989 T989;

/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
typedef struct S990 T990;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S991 T991;

/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S992 T992;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S993 T993;

/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S994 T994;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
typedef struct S995 T995;

/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
typedef struct S996 T996;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
typedef struct S997 T997;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
typedef struct S998 T998;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S999 T999;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1001 T1001;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
typedef struct S1002 T1002;

/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
typedef struct S1003 T1003;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
typedef struct S1004 T1004;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
typedef struct S1005 T1005;

/* [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
typedef struct S1006 T1006;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
typedef struct S1007 T1007;

/* [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
typedef struct S1008 T1008;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
typedef struct S1009 T1009;

/* [detachable] SPECIAL [detachable ET_ASSIGNER] */
typedef struct S1011 T1011;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
typedef struct S1012 T1012;

/* detachable ET_BRACKET_EXPRESSION */
typedef struct S1013 T1013;

/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
typedef struct S1014 T1014;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
typedef struct S1015 T1015;

/* detachable ET_CALL_AGENT */
typedef struct S1016 T1016;

/* [detachable] SPECIAL [detachable ET_CALL_AGENT] */
typedef struct S1017 T1017;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
typedef struct S1018 T1018;

/* detachable ET_CHECK_INSTRUCTION */
typedef struct S1019 T1019;

/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
typedef struct S1020 T1020;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
typedef struct S1021 T1021;

/* [detachable] SPECIAL [detachable ET_CHOICE] */
typedef struct S1023 T1023;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
typedef struct S1024 T1024;

/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
typedef struct S1026 T1026;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
typedef struct S1027 T1027;

/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
typedef struct S1029 T1029;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
typedef struct S1030 T1030;

/* detachable ET_CHOICE_LIST */
typedef struct S1031 T1031;

/* [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
typedef struct S1032 T1032;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
typedef struct S1033 T1033;

/* [detachable] SPECIAL [detachable ET_CLASS] */
typedef struct S1034 T1034;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
typedef struct S1035 T1035;

/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
typedef struct S1036 T1036;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
typedef struct S1037 T1037;

/* detachable ET_CLIENTS */
typedef struct S1038 T1038;

/* [detachable] SPECIAL [detachable ET_CLIENTS] */
typedef struct S1039 T1039;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
typedef struct S1040 T1040;

/* [detachable] SPECIAL [detachable ET_COMPOUND] */
typedef struct S1041 T1041;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
typedef struct S1042 T1042;

/* [detachable] SPECIAL [detachable ET_CONDITIONAL] */
typedef struct S1044 T1044;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
typedef struct S1045 T1045;

/* [detachable] SPECIAL [detachable ET_CONSTANT] */
typedef struct S1046 T1046;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1047 T1047;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1049 T1049;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
typedef struct S1050 T1050;

/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
typedef struct S1051 T1051;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
typedef struct S1052 T1052;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
typedef struct S1053 T1053;

/* detachable ET_CONSTRAINT_CREATOR */
typedef struct S1054 T1054;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
typedef struct S1055 T1055;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
typedef struct S1056 T1056;

/* detachable ET_CONSTRAINT_RENAME_LIST */
typedef struct S1057 T1057;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
typedef struct S1058 T1058;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1059 T1059;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1061 T1061;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
typedef struct S1062 T1062;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
typedef struct S1064 T1064;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
typedef struct S1065 T1065;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
typedef struct S1067 T1067;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
typedef struct S1068 T1068;

/* detachable ET_CONVERT_FEATURE_LIST */
typedef struct S1069 T1069;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
typedef struct S1070 T1070;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
typedef struct S1071 T1071;

/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
typedef struct S1072 T1072;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
typedef struct S1073 T1073;

/* detachable ET_CREATION_REGION */
typedef struct S1074 T1074;

/* [detachable] SPECIAL [detachable ET_CREATION_REGION] */
typedef struct S1075 T1075;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
typedef struct S1076 T1076;

/* detachable ET_CREATOR */
typedef struct S1077 T1077;

/* [detachable] SPECIAL [detachable ET_CREATOR] */
typedef struct S1078 T1078;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
typedef struct S1079 T1079;

/* detachable ET_CREATOR_LIST */
typedef struct S1080 T1080;

/* [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
typedef struct S1081 T1081;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
typedef struct S1082 T1082;

/* detachable ET_DEBUG_INSTRUCTION */
typedef struct S1083 T1083;

/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
typedef struct S1084 T1084;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
typedef struct S1085 T1085;

/* detachable ET_ELSEIF_EXPRESSION */
typedef struct S1086 T1086;

/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
typedef struct S1087 T1087;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
typedef struct S1088 T1088;

/* detachable ET_ELSEIF_EXPRESSION_LIST */
typedef struct S1089 T1089;

/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
typedef struct S1090 T1090;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
typedef struct S1091 T1091;

/* detachable ET_ELSEIF_PART */
typedef struct S1092 T1092;

/* [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
typedef struct S1093 T1093;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
typedef struct S1094 T1094;

/* detachable ET_ELSEIF_PART_LIST */
typedef struct S1095 T1095;

/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
typedef struct S1096 T1096;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
typedef struct S1097 T1097;

/* [detachable] SPECIAL [detachable ET_EXPORT] */
typedef struct S1099 T1099;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
typedef struct S1100 T1100;

/* [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
typedef struct S1101 T1101;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
typedef struct S1102 T1102;

/* [detachable] SPECIAL [detachable ET_EXPRESSION] */
typedef struct S1103 T1103;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
typedef struct S1104 T1104;

/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
typedef struct S1105 T1105;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
typedef struct S1106 T1106;

/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
typedef struct S1107 T1107;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
typedef struct S1108 T1108;

/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
typedef struct S1110 T1110;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
typedef struct S1111 T1111;

/* detachable ET_FEATURE_CLAUSE */
typedef struct S1112 T1112;

/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
typedef struct S1113 T1113;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
typedef struct S1114 T1114;

/* detachable ET_FEATURE_CLAUSE_LIST */
typedef struct S1115 T1115;

/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
typedef struct S1116 T1116;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
typedef struct S1117 T1117;

/* detachable ET_FEATURE_EXPORT */
typedef struct S1118 T1118;

/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
typedef struct S1119 T1119;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
typedef struct S1120 T1120;

/* [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
typedef struct S1121 T1121;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
typedef struct S1122 T1122;

/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
typedef struct S1124 T1124;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
typedef struct S1125 T1125;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
typedef struct S1126 T1126;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1127 T1127;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1129 T1129;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S1130 T1130;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S1131 T1131;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
typedef struct S1132 T1132;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
typedef struct S1133 T1133;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1134 T1134;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1136 T1136;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
typedef struct S1137 T1137;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
typedef struct S1138 T1138;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
typedef struct S1139 T1139;

/* detachable ET_IF_EXPRESSION */
typedef struct S1140 T1140;

/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
typedef struct S1141 T1141;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
typedef struct S1142 T1142;

/* detachable ET_IF_INSTRUCTION */
typedef struct S1143 T1143;

/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
typedef struct S1144 T1144;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
typedef struct S1145 T1145;

/* detachable ET_INDEXING_LIST */
typedef struct S1146 T1146;

/* [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
typedef struct S1147 T1147;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
typedef struct S1148 T1148;

/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
typedef struct S1150 T1150;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
typedef struct S1151 T1151;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
typedef struct S1152 T1152;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
typedef struct S1153 T1153;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
typedef struct S1155 T1155;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
typedef struct S1156 T1156;

/* detachable ET_INDEXING_TERM_LIST */
typedef struct S1157 T1157;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
typedef struct S1158 T1158;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
typedef struct S1159 T1159;

/* [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
typedef struct S1160 T1160;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
typedef struct S1161 T1161;

/* detachable ET_INSPECT_INSTRUCTION */
typedef struct S1162 T1162;

/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
typedef struct S1163 T1163;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
typedef struct S1164 T1164;

/* [detachable] SPECIAL [detachable ET_INSTRUCTION] */
typedef struct S1166 T1166;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
typedef struct S1167 T1167;

/* detachable ET_INVARIANTS */
typedef struct S1168 T1168;

/* [detachable] SPECIAL [detachable ET_INVARIANTS] */
typedef struct S1169 T1169;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
typedef struct S1170 T1170;

/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
typedef struct S1171 T1171;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
typedef struct S1172 T1172;

/* [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
typedef struct S1174 T1174;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
typedef struct S1175 T1175;

/* detachable ET_LOCAL_VARIABLE */
typedef struct S1176 T1176;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
typedef struct S1177 T1177;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1178 T1178;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1180 T1180;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S1181 T1181;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S1182 T1182;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
typedef struct S1183 T1183;

/* detachable ET_LOOP_INVARIANTS */
typedef struct S1184 T1184;

/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
typedef struct S1185 T1185;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
typedef struct S1186 T1186;

/* detachable ET_MANIFEST_ARRAY */
typedef struct S1187 T1187;

/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
typedef struct S1188 T1188;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
typedef struct S1189 T1189;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
typedef struct S1191 T1191;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
typedef struct S1192 T1192;

/* detachable ET_MANIFEST_STRING_LIST */
typedef struct S1193 T1193;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
typedef struct S1194 T1194;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
typedef struct S1195 T1195;

/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
typedef struct S1196 T1196;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
typedef struct S1197 T1197;

/* [detachable] SPECIAL [detachable ET_OBSOLETE] */
typedef struct S1199 T1199;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
typedef struct S1200 T1200;

/* detachable ET_PARENTHESIZED_EXPRESSION */
typedef struct S1201 T1201;

/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
typedef struct S1202 T1202;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
typedef struct S1203 T1203;

/* [detachable] SPECIAL [detachable ET_PARENT] */
typedef struct S1204 T1204;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
typedef struct S1205 T1205;

/* detachable ET_PARENT_CLAUSE_LIST */
typedef struct S1206 T1206;

/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
typedef struct S1207 T1207;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
typedef struct S1208 T1208;

/* [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
typedef struct S1209 T1209;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
typedef struct S1210 T1210;

/* [detachable] SPECIAL [detachable ET_PARENT_LIST] */
typedef struct S1211 T1211;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
typedef struct S1212 T1212;

/* detachable ET_POSTCONDITIONS */
typedef struct S1213 T1213;

/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
typedef struct S1214 T1214;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
typedef struct S1215 T1215;

/* [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
typedef struct S1216 T1216;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
typedef struct S1217 T1217;

/* [detachable] SPECIAL [detachable ET_PROCEDURE] */
typedef struct S1218 T1218;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
typedef struct S1219 T1219;

/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
typedef struct S1221 T1221;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
typedef struct S1222 T1222;

/* [detachable] SPECIAL [detachable ET_QUERY] */
typedef struct S1223 T1223;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
typedef struct S1224 T1224;

/* [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
typedef struct S1226 T1226;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
typedef struct S1227 T1227;

/* [detachable] SPECIAL [detachable ET_RENAME_LIST] */
typedef struct S1228 T1228;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
typedef struct S1229 T1229;

/* detachable ET_STATIC_CALL_EXPRESSION */
typedef struct S1230 T1230;

/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
typedef struct S1231 T1231;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
typedef struct S1232 T1232;

/* detachable ET_STRIP_EXPRESSION */
typedef struct S1233 T1233;

/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
typedef struct S1234 T1234;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
typedef struct S1235 T1235;

/* [detachable] SPECIAL [detachable ET_TYPE] */
typedef struct S1236 T1236;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
typedef struct S1237 T1237;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
typedef struct S1239 T1239;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1240 T1240;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1242 T1242;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
typedef struct S1243 T1243;

/* detachable ET_TYPE_CONSTRAINT_LIST */
typedef struct S1244 T1244;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
typedef struct S1245 T1245;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
typedef struct S1246 T1246;

/* [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
typedef struct S1248 T1248;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
typedef struct S1249 T1249;

/* [detachable] SPECIAL [detachable ET_TYPE_LIST] */
typedef struct S1251 T1251;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
typedef struct S1252 T1252;

/* detachable ET_VARIANT */
typedef struct S1253 T1253;

/* [detachable] SPECIAL [detachable ET_VARIANT] */
typedef struct S1254 T1254;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
typedef struct S1255 T1255;

/* detachable ET_WHEN_PART */
typedef struct S1256 T1256;

/* [detachable] SPECIAL [detachable ET_WHEN_PART] */
typedef struct S1257 T1257;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
typedef struct S1258 T1258;

/* detachable ET_WHEN_PART_LIST */
typedef struct S1259 T1259;

/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
typedef struct S1260 T1260;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
typedef struct S1261 T1261;

/* [detachable] SPECIAL [detachable ET_WRITABLE] */
typedef struct S1262 T1262;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
typedef struct S1263 T1263;

/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
typedef struct S1264 T1264;

/* [detachable] ET_PARENT_CHECKER1 */
typedef struct S1265 T1265;

/* [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
typedef struct S1266 T1266;

/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
typedef struct S1267 T1267;

/* [detachable] ET_FEATURE_NAME_TESTER */
typedef struct S1268 T1268;

/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
typedef struct S1270 T1270;

/* [detachable] ET_ALIAS_NAME_TESTER */
typedef struct S1271 T1271;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
typedef struct S1273 T1273;

/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1274 T1274;

/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
typedef struct S1275 T1275;

/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
typedef struct S1276 T1276;

/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
typedef struct S1277 T1277;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
typedef struct S1278 T1278;

/* [detachable] ET_ANCHORED_TYPE_CHECKER */
typedef struct S1279 T1279;

/* [detachable] ET_SIGNATURE_CHECKER */
typedef struct S1280 T1280;

/* [detachable] ET_PARENT_CHECKER2 */
typedef struct S1281 T1281;

/* [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
typedef struct S1282 T1282;

/* [detachable] ET_BUILTIN_FEATURE_CHECKER */
typedef struct S1283 T1283;

/* [detachable] ET_PRECURSOR_CHECKER */
typedef struct S1284 T1284;

/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
typedef struct S1285 T1285;

/* [detachable] DS_HASH_SET [[attached] ET_CLASS] */
typedef struct S1286 T1286;

/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
typedef struct S1287 T1287;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
typedef struct S1288 T1288;

/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
typedef struct S1289 T1289;

/* [detachable] ET_CALL_NAME_TESTER */
typedef struct S1290 T1290;

/* [detachable] ET_RENAME */
typedef struct S1295 T1295;

/* [detachable] ET_PARENT_CHECKER3 */
typedef struct S1296 T1296;

/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
typedef struct S1297 T1297;

/* [detachable] DS_HASH_SET [[attached] ET_QUERY] */
typedef struct S1298 T1298;

/* [detachable] ET_SUPPLIER_BUILDER */
typedef struct S1299 T1299;

/* detachable ET_PARENT_FEATURE */
typedef struct S1302 T1302;

/* [detachable] ET_INHERITED_FEATURE */
typedef struct S1303 T1303;

/* [detachable] SPECIAL [[attached] ET_CLASS] */
typedef struct S1304 T1304;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
typedef struct S1305 T1305;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
typedef struct S1306 T1306;

/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
typedef struct S1307 T1307;

/* [detachable] UT_SYNTAX_ERROR */
typedef struct S1308 T1308;

/* [detachable] UT_TOO_MANY_INCLUDES_ERROR */
typedef struct S1309 T1309;

/* detachable ET_DYNAMIC_TARGET_LIST */
typedef struct S1311 T1311;

/* [detachable] SPECIAL [[attached] NATIVE_STRING] */
typedef struct S1312 T1312;

/* [detachable] ARRAY [detachable STRING_8] */
typedef struct S1313 T1313;

/* [detachable] XM_EIFFEL_CHARACTER_ENTITY */
typedef struct S1314 T1314;

/* detachable DS_BILINKABLE [[attached] XM_POSITION] */
typedef struct S1316 T1316;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
typedef struct S1317 T1317;

/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
typedef struct S1318 T1318;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
typedef struct S1319 T1319;

/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
typedef struct S1322 T1322;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
typedef struct S1323 T1323;

/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S1326 T1326;

/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S1327 T1327;

/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1328 T1328;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S1329 T1329;

/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S1330 T1330;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S1331 T1331;

/* detachable DS_BILINKABLE [[attached] STRING_8] */
typedef struct S1332 T1332;

/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
typedef struct S1333 T1333;

/* [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
typedef struct S1336 T1336;

/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
typedef struct S1337 T1337;

/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
typedef struct S1338 T1338;

/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
typedef struct S1339 T1339;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
typedef struct S1340 T1340;

/* [detachable] SPECIAL [[attached] XM_NAMESPACE] */
typedef struct S1341 T1341;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
typedef struct S1342 T1342;

/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
typedef struct S1344 T1344;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
typedef struct S1345 T1345;

/* [detachable] ET_BRACED_TYPE_LIST */
typedef struct S1346 T1346;

/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
typedef struct S1347 T1347;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1348 T1348;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1349 T1349;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1350 T1350;

/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
typedef struct S1351 T1351;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1353 T1353;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1354 T1354;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1355 T1355;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1356 T1356;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
typedef struct S1358 T1358;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S1359 T1359;

/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1360 T1360;

/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1361 T1361;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1362 T1362;

/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1364 T1364;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1365 T1365;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1366 T1366;

/* [detachable] THREAD_ATTRIBUTES */
typedef struct S1367 T1367;

/* [detachable] CELL [[attached] BOOLEAN] */
typedef struct S1368 T1368;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
typedef struct S1370 T1370;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
typedef struct S1372 T1372;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
typedef struct S1373 T1373;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
typedef struct S1374 T1374;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
typedef struct S1375 T1375;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S1376 T1376;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT] */
typedef struct S1377 T1377;

/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
typedef struct S1378 T1378;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
typedef struct S1382 T1382;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
typedef struct S1383 T1383;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
typedef struct S1385 T1385;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
typedef struct S1386 T1386;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
typedef struct S1393 T1393;

/* [detachable] TYPE [[attached] NATURAL_64] */
#define T1394 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] NATURAL_32] */
#define T1395 EIF_TYPE_OBJ

/* [detachable] KL_CHARACTER_BUFFER */
typedef struct S1396 T1396;

/* [detachable] UT_TRISTATE */
typedef struct S1398 T1398;

/* [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
typedef struct S1399 T1399;

/* [detachable] TUPLE [[attached] UT_TRISTATE] */
typedef struct S1400 T1400;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
typedef struct S1401 T1401;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
typedef struct S1402 T1402;

/* [detachable] PREDICATE [[attached] TUPLE] */
typedef struct S1403 T1403;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
typedef struct S1405 T1405;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
typedef struct S1406 T1406;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S1407 T1407;

/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
typedef struct S1408 T1408;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
typedef struct S1409 T1409;

/* [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
typedef struct S1410 T1410;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
typedef struct S1411 T1411;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S1413 T1413;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
typedef struct S1414 T1414;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S1415 T1415;

/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
typedef struct S1416 T1416;

/* TYPED_POINTER [[attached] NATURAL_16] */
typedef struct S1418 T1418;
extern T0* GE_boxed1418(T1418 a1);
typedef struct Sb1418 Tb1418;

/* [detachable] XM_COMMENT */
typedef struct S1419 T1419;

/* [detachable] XM_PROCESSING_INSTRUCTION */
typedef struct S1420 T1420;

/* [detachable] XM_CHARACTER_DATA */
typedef struct S1421 T1421;

/* [detachable] ET_LIKE_N */
typedef struct S1424 T1424;

/* [detachable] KL_STRING_INPUT_STREAM */
typedef struct S1427 T1427;

/* [detachable] KL_UNICODE_CHARACTER_BUFFER */
typedef struct S1429 T1429;

/* detachable ET_OBJECT_TEST_LIST */
typedef struct S1431 T1431;

/* detachable ET_C3_CHARACTER_CONSTANT */
typedef struct S1432 T1432;

/* detachable ET_SPECIAL_MANIFEST_STRING */
typedef struct S1433 T1433;

/* detachable ET_VERBATIM_STRING */
typedef struct S1434 T1434;

/* detachable ET_UNDERSCORED_INTEGER_CONSTANT */
typedef struct S1435 T1435;

/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
typedef struct S1436 T1436;

/* detachable ET_OCTAL_INTEGER_CONSTANT */
typedef struct S1437 T1437;

/* detachable ET_BINARY_INTEGER_CONSTANT */
typedef struct S1438 T1438;

/* detachable ET_REGULAR_REAL_CONSTANT */
typedef struct S1439 T1439;

/* detachable ET_UNDERSCORED_REAL_CONSTANT */
typedef struct S1440 T1440;

/* detachable ET_TRUE_CONSTANT */
typedef struct S1442 T1442;

/* detachable ET_FALSE_CONSTANT */
typedef struct S1443 T1443;

/* detachable ET_C1_CHARACTER_CONSTANT */
typedef struct S1444 T1444;

/* detachable ET_C2_CHARACTER_CONSTANT */
typedef struct S1445 T1445;

/* detachable ET_TAGGED_INDEXING */
typedef struct S1447 T1447;

/* detachable ET_INDEXING */
typedef struct S1448 T1448;

/* detachable ET_CUSTOM_ATTRIBUTE */
typedef struct S1449 T1449;

/* detachable ET_CONSTRAINED_FORMAL_PARAMETER */
typedef struct S1450 T1450;

/* detachable ET_TYPE_RENAME_CONSTRAINT */
typedef struct S1451 T1451;

/* detachable ET_CONSTRAINT_NAMED_TYPE */
typedef struct S1452 T1452;

/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
typedef struct S1453 T1453;

/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
typedef struct S1454 T1454;

/* detachable ET_ALL_EXPORT */
typedef struct S1455 T1455;

/* detachable ET_CONVERT_FUNCTION */
typedef struct S1457 T1457;

/* detachable ET_CONVERT_PROCEDURE */
typedef struct S1458 T1458;

/* detachable ET_DO_FUNCTION */
typedef struct S1460 T1460;

/* detachable ET_ONCE_FUNCTION */
typedef struct S1461 T1461;

/* detachable ET_DEFERRED_FUNCTION */
typedef struct S1462 T1462;

/* detachable ET_ONCE_PROCEDURE */
typedef struct S1464 T1464;

/* detachable ET_EXTERNAL_PROCEDURE */
typedef struct S1465 T1465;

/* detachable ET_PREFIX_NAME */
typedef struct S1466 T1466;

/* detachable ET_PREFIX_FREE_NAME */
typedef struct S1467 T1467;

/* detachable ET_INFIX_NAME */
typedef struct S1468 T1468;

/* detachable ET_ALIASED_FEATURE_NAME */
typedef struct S1469 T1469;

/* detachable ET_ALIAS_FREE_NAME */
typedef struct S1470 T1470;

/* detachable ET_CLASS_ASSERTION */
typedef struct S1473 T1473;

/* detachable ET_LABELED_ACTUAL_PARAMETER */
typedef struct S1474 T1474;

/* detachable ET_LIKE_FEATURE */
typedef struct S1476 T1476;

/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
typedef struct S1477 T1477;

/* detachable ET_QUALIFIED_LIKE_TYPE */
typedef struct S1478 T1478;

/* detachable ET_ASSIGNER_INSTRUCTION */
typedef struct S1479 T1479;

/* detachable ET_ASSIGNMENT */
typedef struct S1480 T1480;

/* detachable ET_ASSIGNMENT_ATTEMPT */
typedef struct S1481 T1481;

/* detachable ET_LOOP_INSTRUCTION */
typedef struct S1482 T1482;

/* detachable ET_BANG_INSTRUCTION */
typedef struct S1484 T1484;

/* detachable ET_CREATE_INSTRUCTION */
typedef struct S1485 T1485;

/* detachable ET_CHOICE_RANGE */
typedef struct S1486 T1486;

/* detachable ET_PRECURSOR_INSTRUCTION */
typedef struct S1487 T1487;

/* detachable ET_STATIC_CALL_INSTRUCTION */
typedef struct S1489 T1489;

/* detachable ET_PRECURSOR_EXPRESSION */
typedef struct S1490 T1490;

/* detachable ET_FEATURE_ADDRESS */
typedef struct S1491 T1491;

/* detachable ET_CURRENT_ADDRESS */
typedef struct S1492 T1492;

/* detachable ET_RESULT_ADDRESS */
typedef struct S1493 T1493;

/* detachable ET_EXPRESSION_ADDRESS */
typedef struct S1494 T1494;

/* detachable ET_INFIX_EXPRESSION */
typedef struct S1495 T1495;

/* detachable ET_INFIX_AND_THEN_OPERATOR */
typedef struct S1496 T1496;

/* detachable ET_INFIX_OR_ELSE_OPERATOR */
typedef struct S1497 T1497;

/* detachable ET_MANIFEST_TYPE */
typedef struct S1498 T1498;

/* detachable ET_PREFIX_EXPRESSION */
typedef struct S1499 T1499;

/* detachable ET_OLD_EXPRESSION */
typedef struct S1500 T1500;

/* detachable ET_OLD_OBJECT_TEST */
typedef struct S1501 T1501;

/* detachable ET_ONCE_MANIFEST_STRING */
typedef struct S1502 T1502;

/* detachable ET_DO_FUNCTION_INLINE_AGENT */
typedef struct S1503 T1503;

/* detachable ET_ONCE_FUNCTION_INLINE_AGENT */
typedef struct S1504 T1504;

/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
typedef struct S1505 T1505;

/* detachable ET_DO_PROCEDURE_INLINE_AGENT */
typedef struct S1506 T1506;

/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
typedef struct S1507 T1507;

/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
typedef struct S1508 T1508;

/* detachable ET_AGENT_OPEN_TARGET */
typedef struct S1509 T1509;

/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
typedef struct S1510 T1510;

/* detachable ET_TAGGED_ASSERTION */
typedef struct S1513 T1513;

/* [detachable] ET_FILE_POSITION */
typedef struct S1515 T1515;

/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
typedef struct S1517 T1517;

/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
typedef struct S1518 T1518;

/* [detachable] ET_REDECLARED_FEATURE */
typedef struct S1520 T1520;

/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
typedef struct S1522 T1522;

/* [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1524 T1524;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
typedef struct S1527 T1527;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1528 T1528;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1529 T1529;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
typedef struct S1530 T1530;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1531 T1531;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1532 T1532;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1533 T1533;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1534 T1534;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1535 T1535;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1536 T1536;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1537 T1537;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
typedef struct S1538 T1538;

/* [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
typedef struct S1539 T1539;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
typedef struct S1540 T1540;

/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1541 T1541;

/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
typedef struct S1542 T1542;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
typedef struct S1545 T1545;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S1546 T1546;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1547 T1547;

/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1548 T1548;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1549 T1549;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1550 T1550;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1551 T1551;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1552 T1552;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1553 T1553;

/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1554 T1554;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
typedef struct S1556 T1556;

/* [detachable] SPECIAL [[attached] ET_UNIVERSE] */
typedef struct S1557 T1557;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
typedef struct S1559 T1559;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1560 T1560;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1562 T1562;

/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1563 T1563;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
typedef struct S1564 T1564;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
typedef struct S1565 T1565;

/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
typedef struct S1566 T1566;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
typedef struct S1567 T1567;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
typedef struct S1568 T1568;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1569 T1569;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1570 T1570;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1571 T1571;

/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
typedef struct S1573 T1573;

/* [detachable] UC_UTF16_ROUTINES */
typedef struct S1574 T1574;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
typedef struct S1575 T1575;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
typedef struct S1576 T1576;

/* [detachable] SPECIAL [[attached] ET_LIBRARY] */
typedef struct S1577 T1577;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
typedef struct S1578 T1578;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1579 T1579;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1580 T1580;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1581 T1581;

/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1582 T1582;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1583 T1583;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1584 T1584;

/* [detachable] YY_UNICODE_BUFFER */
typedef struct S1585 T1585;

/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
typedef struct S1586 T1586;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
typedef struct S1587 T1587;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S1588 T1588;

/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S1589 T1589;

/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST] */
typedef struct S1590 T1590;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST] */
typedef struct S1591 T1591;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST] */
typedef struct S1592 T1592;

/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST] */
typedef struct S1593 T1593;

/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
typedef struct S1594 T1594;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
typedef struct S1595 T1595;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
typedef struct S1596 T1596;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1597 T1597;

/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
typedef struct S1598 T1598;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
typedef struct S1600 T1600;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
typedef struct S1602 T1602;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
typedef struct S1603 T1603;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1606 T1606;

/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
typedef struct S1607 T1607;

/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
typedef struct S1608 T1608;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
typedef struct S1609 T1609;

/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
typedef struct S1610 T1610;

/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
typedef struct S1613 T1613;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
typedef struct S1614 T1614;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
typedef struct S1615 T1615;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
typedef struct S1617 T1617;

/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
typedef struct S1618 T1618;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
typedef struct S1619 T1619;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
typedef struct S1620 T1620;

/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
typedef struct S1621 T1621;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
typedef struct S1622 T1622;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1624 T1624;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
typedef struct S1625 T1625;

/* [detachable] SPECIAL [[attached] ET_CLIENT] */
typedef struct S1626 T1626;

/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME] */
typedef struct S1627 T1627;

/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
typedef struct S1628 T1628;

/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
typedef struct S1629 T1629;

/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
typedef struct S1630 T1630;

/* [detachable] ET_REPLICATED_FEATURE */
typedef struct S1632 T1632;

/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1633 T1633;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
typedef struct S1634 T1634;

/* [detachable] ET_DOTNET_SIGNATURE_TESTER */
typedef struct S1635 T1635;

/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
typedef struct S1636 T1636;

/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
typedef struct S1638 T1638;

/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
typedef struct S1640 T1640;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
typedef struct S1641 T1641;

/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
typedef struct S1642 T1642;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
typedef struct S1643 T1643;

/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
typedef struct S1644 T1644;

/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
typedef struct S1645 T1645;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
typedef struct S1646 T1646;

/* [detachable] TYPE [[attached] ET_FEATURE] */
#define T1647 EIF_TYPE_OBJ

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
typedef struct S1649 T1649;

/* [detachable] TYPE [[attached] ET_CLASS] */
#define T1650 EIF_TYPE_OBJ

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
typedef struct S1651 T1651;

/* [detachable] SPECIAL [[attached] ET_RENAME] */
typedef struct S1652 T1652;

/* [detachable] SPECIAL [[attached] ET_CALL_NAME] */
typedef struct S1653 T1653;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
typedef struct S1656 T1656;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
typedef struct S1657 T1657;

/* [detachable] TYPE [[attached] ET_RENAME] */
#define T1658 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] ET_CALL_NAME] */
#define T1659 EIF_TYPE_OBJ

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
typedef struct S1660 T1660;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
typedef struct S1661 T1661;

/* [detachable] SPECIAL [[attached] YY_BUFFER] */
typedef struct S1662 T1662;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
typedef struct S1663 T1663;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
typedef struct S1664 T1664;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
typedef struct S1665 T1665;

/* [detachable] TYPE [detachable STRING_8] */
#define T1666 EIF_TYPE_OBJ

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1667 T1667;

/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1668 T1668;

/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1670 T1670;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1671 T1671;

/* detachable DS_LINKABLE [detachable STRING_8] */
typedef struct S1672 T1672;

/* [detachable] SPECIAL [[attached] ET_CLUSTER] */
typedef struct S1673 T1673;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
typedef struct S1674 T1674;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
typedef struct S1675 T1675;

/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
typedef struct S1676 T1676;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
typedef struct S1677 T1677;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
typedef struct S1678 T1678;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1679 T1679;

/* [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
typedef struct S1680 T1680;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
typedef struct S1681 T1681;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S1682 T1682;

/* [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1684 T1684;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
typedef struct S1685 T1685;

/* [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1686 T1686;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
typedef struct S1687 T1687;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
typedef struct S1690 T1690;

/* [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S1691 T1691;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
typedef struct S1692 T1692;

/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
typedef struct S1693 T1693;

/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
typedef struct S1698 T1698;

/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
typedef struct S1703 T1703;

/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
typedef struct S1704 T1704;

/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
typedef struct S1706 T1706;

/* [detachable] ET_SEEDED_QUERY_COMPARATOR */
typedef struct S1707 T1707;

/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
typedef struct S1710 T1710;

/* [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
typedef struct S1714 T1714;

/* [detachable] ET_FORMAL_COMMA_ARGUMENT */
typedef struct S1715 T1715;

/* [detachable] ET_LOCAL_COMMA_VARIABLE */
typedef struct S1716 T1716;

/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
typedef struct S1717 T1717;

/* [detachable] ET_KEYWORD_EXPRESSION */
typedef struct S1719 T1719;

/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
typedef struct S1724 T1724;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
typedef struct S1725 T1725;

/* [detachable] SPECIAL [[attached] ET_EXPORT] */
typedef struct S1727 T1727;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
typedef struct S1728 T1728;

/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
typedef struct S1730 T1730;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
typedef struct S1731 T1731;

/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1732 T1732;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1733 T1733;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1736 T1736;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
typedef struct S1737 T1737;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
typedef struct S1738 T1738;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1739 T1739;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
typedef struct S1740 T1740;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1741 T1741;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
typedef struct S1742 T1742;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1743 T1743;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
typedef struct S1744 T1744;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1745 T1745;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
typedef struct S1746 T1746;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1747 T1747;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
typedef struct S1748 T1748;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1749 T1749;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
typedef struct S1750 T1750;

/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S1755 T1755;

/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1758 T1758;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1759 T1759;

/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
typedef struct S1760 T1760;

/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
typedef struct S1762 T1762;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
typedef struct S1763 T1763;

/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1764 T1764;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1766 T1766;

/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
typedef struct S1768 T1768;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
typedef struct S1769 T1769;

/* [detachable] SPECIAL [[attached] ET_CREATOR] */
typedef struct S1770 T1770;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
typedef struct S1771 T1771;

/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
typedef struct S1772 T1772;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
typedef struct S1773 T1773;

/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
typedef struct S1774 T1774;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
typedef struct S1775 T1775;

/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
typedef struct S1776 T1776;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
typedef struct S1777 T1777;

/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
typedef struct S1779 T1779;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
typedef struct S1780 T1780;

/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
typedef struct S1781 T1781;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
typedef struct S1782 T1782;

/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
typedef struct S1784 T1784;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
typedef struct S1785 T1785;

/* [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
typedef struct S1786 T1786;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
typedef struct S1787 T1787;

/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1789 T1789;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1790 T1790;

/* [detachable] SPECIAL [[attached] ET_WHEN_PART] */
typedef struct S1791 T1791;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
typedef struct S1792 T1792;

/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
typedef struct S1794 T1794;

/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1797 T1797;

/* [detachable] ET_STANDARD_ONCE_KEYS */
typedef struct S1802 T1802;

/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
typedef struct S1804 T1804;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
typedef struct S1805 T1805;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
typedef struct S1806 T1806;

/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1807 T1807;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1808 T1808;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1809 T1809;

/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1810 T1810;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1811 T1811;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1812 T1812;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
typedef struct S1813 T1813;

/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
typedef struct S1814 T1814;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
typedef struct S1815 T1815;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1816 T1816;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1817 T1817;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1818 T1818;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1819 T1819;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1820 T1820;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1821 T1821;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1822 T1822;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1823 T1823;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1824 T1824;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1825 T1825;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1826 T1826;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1827 T1827;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1828 T1828;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1829 T1829;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1830 T1830;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1831 T1831;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1832 T1832;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1833 T1833;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1834 T1834;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1835 T1835;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1836 T1836;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1837 T1837;

/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1838 T1838;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1839 T1839;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1840 T1840;

/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1841 T1841;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1842 T1842;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1843 T1843;

/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1844 T1844;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1845 T1845;

/* detachable DS_LINKABLE [[attached] CHARACTER_8] */
typedef struct S1847 T1847;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
typedef struct S1848 T1848;

/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1849 T1849;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1850 T1850;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1851 T1851;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
typedef struct S1852 T1852;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME] */
typedef struct S1853 T1853;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
typedef struct S1854 T1854;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
typedef struct S1857 T1857;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
typedef struct S1858 T1858;

/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
typedef struct S1859 T1859;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
typedef struct S1861 T1861;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1862 T1862;

/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1863 T1863;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
typedef struct S1864 T1864;

/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
typedef struct S1865 T1865;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
typedef struct S1866 T1866;

/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
typedef struct S1867 T1867;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
typedef struct S1868 T1868;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
typedef struct S1869 T1869;

/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
typedef struct S1870 T1870;

/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
typedef struct S1873 T1873;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
typedef struct S1874 T1874;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
typedef struct S1875 T1875;

/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
typedef struct S1876 T1876;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
typedef struct S1877 T1877;

/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
typedef struct S1878 T1878;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
typedef struct S1879 T1879;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
typedef struct S1880 T1880;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
typedef struct S1882 T1882;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
typedef struct S1883 T1883;

/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
typedef struct S1884 T1884;

/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
typedef struct S1885 T1885;

/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1886 T1886;

/* [detachable] ET_CLUSTER_ERROR */
typedef struct S1887 T1887;

/* [detachable] ET_SYNTAX_ERROR */
typedef struct S1888 T1888;

/* [detachable] ET_DOTNET_ASSEMBLY_ERROR */
typedef struct S1889 T1889;

/* [detachable] AP_OPTION_COMPARATOR */
typedef struct S1890 T1890;

/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
typedef struct S1891 T1891;

/* [detachable] ST_SPLITTER */
typedef struct S1898 T1898;

/* [detachable] KL_CHARACTER_ROUTINES */
typedef struct S1899 T1899;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
typedef struct S1900 T1900;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
typedef struct S1901 T1901;

/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
typedef struct S1902 T1902;

/* [detachable] TYPE [[detachable] DEVELOPER_EXCEPTION] */
#define T1903 EIF_TYPE_OBJ

/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1905 T1905;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1907 T1907;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
typedef struct S1909 T1909;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
typedef struct S1910 T1910;

/* [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
typedef struct S1911 T1911;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
typedef struct S1913 T1913;

/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1914 T1914;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1916 T1916;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
typedef struct S1924 T1924;

/* [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
typedef struct S1925 T1925;

/* [detachable] ET_INFIX_CAST_EXPRESSION */
typedef struct S1926 T1926;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
typedef struct S1930 T1930;

/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
typedef struct S1931 T1931;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
typedef struct S1934 T1934;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
typedef struct S1935 T1935;

/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1936 T1936;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
typedef struct S1939 T1939;

/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
typedef struct S1940 T1940;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
typedef struct S1943 T1943;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
typedef struct S1944 T1944;

/* [detachable] ET_UNIVERSE_ERROR */
typedef struct S1945 T1945;

/* [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S1946 T1946;

/* detachable ET_PARENTHESIS_INSTRUCTION */
typedef struct S1947 T1947;

/* [detachable] TYPE [detachable RX_REGULAR_EXPRESSION] */
#define T1952 EIF_TYPE_OBJ

/* [detachable] ET_PARENTHESIS_SYMBOL */
typedef struct S1967 T1967;

/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
typedef struct S1973 T1973;

/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
typedef struct S1974 T1974;

/* [detachable] TYPE [[attached] ET_NESTED_TYPE_CONTEXT] */
#define T1977 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] ET_ACROSS_COMPONENT] */
#define T1978 EIF_TYPE_OBJ

/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S1981 T1981;

/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
typedef struct S1986 T1986;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
typedef struct S1987 T1987;

/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
typedef struct S1988 T1988;

/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
typedef struct S1989 T1989;

/* [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
typedef struct S1990 T1990;

/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
typedef struct S1992 T1992;

/* [detachable] ET_CONVERT_FROM_EXPRESSION */
typedef struct S1993 T1993;

/* [detachable] ET_CONVERT_TO_EXPRESSION */
typedef struct S1994 T1994;

/* detachable DS_CELL [detachable ET_CLASS] */
typedef struct S1997 T1997;

/* [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
typedef struct S1998 T1998;

/* [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
typedef struct S2000 T2000;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
typedef struct S2002 T2002;

/* [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
typedef struct S2003 T2003;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
typedef struct S2004 T2004;

/* [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
typedef struct S2005 T2005;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
typedef struct S2006 T2006;

/* [detachable] TUPLE [[attached] ET_ECF_STATE] */
typedef struct S2007 T2007;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
typedef struct S2009 T2009;

/* [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
typedef struct S2010 T2010;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
typedef struct S2012 T2012;

/* [detachable] LX_DFA_WILDCARD */
typedef struct S2014 T2014;

/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
typedef struct S2017 T2017;

/* [detachable] TUPLE [[attached] LX_WILDCARD] */
typedef struct S2019 T2019;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
typedef struct S2020 T2020;

/* [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
typedef struct S2021 T2021;

/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
typedef struct S2023 T2023;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
typedef struct S2024 T2024;

/* [detachable] SPECIAL [[attached] LX_WILDCARD] */
typedef struct S2025 T2025;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
typedef struct S2028 T2028;

/* [detachable] CHARACTER_PROPERTY */
typedef struct S2029 T2029;

/* [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2034 T2034;

/* [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2035 T2035;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2036 T2036;

/* [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2037 T2037;

/* [detachable] ET_AST_NULL_LEAF */
typedef struct S2039 T2039;

/* [detachable] LX_WILDCARD_PARSER */
typedef struct S2040 T2040;

/* [detachable] LX_DESCRIPTION */
typedef struct S2041 T2041;

/* [detachable] LX_FULL_DFA */
typedef struct S2042 T2042;

/* [detachable] TUPLE [[attached] ET_GROUP] */
typedef struct S2044 T2044;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
typedef struct S2045 T2045;

/* [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S2047 T2047;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
typedef struct S2048 T2048;

/* [detachable] SPECIAL [[attached] NATURAL_16] */
typedef struct S2051 T2051;

/* [detachable] ARRAY [[attached] NATURAL_16] */
typedef struct S2052 T2052;

/* [detachable] ARRAY [[attached] NATURAL_32] */
typedef struct S2053 T2053;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
typedef struct S2054 T2054;

/* [detachable] LX_START_CONDITIONS */
typedef struct S2055 T2055;

/* [detachable] LX_ACTION_FACTORY */
typedef struct S2056 T2056;

/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
typedef struct S2057 T2057;

/* [detachable] LX_SYMBOL_CLASS */
typedef struct S2058 T2058;

/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
typedef struct S2059 T2059;

/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
typedef struct S2060 T2060;

/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
typedef struct S2061 T2061;

/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
typedef struct S2062 T2062;

/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
typedef struct S2063 T2063;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
typedef struct S2064 T2064;

/* [detachable] SPECIAL [[attached] STRING_32] */
typedef struct S2065 T2065;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
typedef struct S2066 T2066;

/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
typedef struct S2067 T2067;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
typedef struct S2068 T2068;

/* [detachable] LX_NFA */
typedef struct S2069 T2069;

/* [detachable] SPECIAL [[attached] LX_NFA] */
typedef struct S2070 T2070;

/* detachable LX_RULE */
typedef struct S2071 T2071;

/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
typedef struct S2072 T2072;

/* [detachable] LX_EQUIVALENCE_CLASSES */
typedef struct S2073 T2073;

/* [detachable] LX_UNRECOGNIZED_RULE_ERROR */
typedef struct S2074 T2074;

/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
typedef struct S2075 T2075;

/* [detachable] LX_MISSING_QUOTE_ERROR */
typedef struct S2076 T2076;

/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
typedef struct S2077 T2077;

/* [detachable] LX_BAD_CHARACTER_ERROR */
typedef struct S2078 T2078;

/* [detachable] LX_FULL_AND_META_ERROR */
typedef struct S2079 T2079;

/* [detachable] LX_FULL_AND_REJECT_ERROR */
typedef struct S2080 T2080;

/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
typedef struct S2081 T2081;

/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
typedef struct S2082 T2082;

/* [detachable] ARRAY [[attached] LX_RULE] */
typedef struct S2083 T2083;

/* [detachable] SPECIAL [[attached] LX_RULE] */
typedef struct S2084 T2084;

/* [detachable] LX_DFA_STATE */
typedef struct S2085 T2085;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
typedef struct S2086 T2086;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
typedef struct S2087 T2087;

/* detachable LX_SYMBOL_PARTITIONS */
typedef struct S2088 T2088;

/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
typedef struct S2089 T2089;

/* [detachable] ARRAY [detachable LX_RULE] */
typedef struct S2090 T2090;

/* [detachable] SPECIAL [detachable LX_RULE] */
typedef struct S2091 T2091;

/* [detachable] LX_START_CONDITION */
typedef struct S2092 T2092;

/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
typedef struct S2093 T2093;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
typedef struct S2094 T2094;

/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
typedef struct S2095 T2095;

/* [detachable] LX_NFA_STATE */
typedef struct S2096 T2096;

/* detachable ARRAY [detachable LX_SYMBOL_CLASS] */
typedef struct S2097 T2097;

/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
typedef struct S2098 T2098;

/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
typedef struct S2099 T2099;

/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
typedef struct S2100 T2100;

/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2101 T2101;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
typedef struct S2102 T2102;

/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
typedef struct S2103 T2103;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
typedef struct S2104 T2104;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
typedef struct S2105 T2105;

/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
typedef struct S2106 T2106;

/* [detachable] SPECIAL [[attached] LX_START_CONDITION] */
typedef struct S2107 T2107;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
typedef struct S2108 T2108;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
typedef struct S2109 T2109;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
typedef struct S2110 T2110;

/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
typedef struct S2111 T2111;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
typedef struct S2113 T2113;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
typedef struct S2114 T2114;

/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
typedef struct S2115 T2115;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
typedef struct S2117 T2117;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
typedef struct S2119 T2119;

/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2120 T2120;

/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2123 T2123;

/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2124 T2124;

/* [detachable] LX_ACTION */
typedef struct S2126 T2126;

/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
typedef struct S2127 T2127;

/* [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
typedef struct S2128 T2128;

/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
typedef struct S2129 T2129;

/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
typedef struct S2130 T2130;

/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
typedef struct S2132 T2132;

/* [detachable] SPECIAL [[attached] LX_NFA_STATE] */
typedef struct S2134 T2134;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
typedef struct S2136 T2136;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
typedef struct S2137 T2137;

/* [detachable] SPECIAL [[attached] LX_DFA_STATE] */
typedef struct S2139 T2139;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
typedef struct S2140 T2140;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
typedef struct S2141 T2141;

/* [detachable] KL_ARRAY [[attached] LX_RULE] */
typedef struct S2142 T2142;

/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
typedef struct S2143 T2143;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
typedef struct S2144 T2144;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
typedef struct S2146 T2146;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
typedef struct S2150 T2150;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
typedef struct S2153 T2153;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2156 T2156;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2157 T2157;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2158 T2158;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
typedef struct S2159 T2159;

/* [detachable] KL_ARRAY [[attached] STRING_8] */
typedef struct S2160 T2160;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
typedef struct S2162 T2162;

/* Struct for boxed version of type BOOLEAN */
struct Sb1 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 z1; /* item */
};

/* Struct for boxed version of type CHARACTER_8 */
struct Sb2 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 z1; /* item */
};

/* Struct for boxed version of type CHARACTER_32 */
struct Sb3 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T3 z1; /* item */
};

/* Struct for boxed version of type INTEGER_8 */
struct Sb4 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T4 z1; /* item */
};

/* Struct for boxed version of type INTEGER_16 */
struct Sb5 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T5 z1; /* item */
};

/* Struct for boxed version of type INTEGER_32 */
struct Sb6 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 z1; /* item */
};

/* Struct for boxed version of type INTEGER_64 */
struct Sb7 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T7 z1; /* item */
};

/* Struct for boxed version of type NATURAL_8 */
struct Sb8 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T8 z1; /* item */
};

/* Struct for boxed version of type NATURAL_16 */
struct Sb9 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T9 z1; /* item */
};

/* Struct for boxed version of type NATURAL_32 */
struct Sb10 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T10 z1; /* item */
};

/* Struct for boxed version of type NATURAL_64 */
struct Sb11 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T11 z1; /* item */
};

/* Struct for boxed version of type REAL_32 */
struct Sb12 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T12 z1; /* item */
};

/* Struct for boxed version of type REAL_64 */
struct Sb13 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T13 z1; /* item */
};

/* Struct for boxed version of type POINTER */
struct Sb14 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 z1; /* item */
};

/* Struct for type UTF_CONVERTER */
struct S471 {
	char dummy;
};

/* Struct for boxed version of type UTF_CONVERTER */
struct Sb471 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T471 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] NATURAL_16] */
struct S1418 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] NATURAL_16] */
struct Sb1418 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1418 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] NATURAL_8] */
struct S868 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] NATURAL_8] */
struct Sb868 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T868 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] ANY] */
struct S674 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] ANY] */
struct Sb674 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T674 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
struct S478 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
struct Sb478 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T478 z1; /* item */
};

/* Struct for type [detachable] SPECIAL [CHARACTER_8] */
struct S15 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T2 z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [CHARACTER_32] */
struct S16 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T3 z2[1]; /* item */
};

/* Struct for type [detachable] STRING_8 */
struct S17 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] STRING_32 */
struct S18 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] IMMUTABLE_STRING_32 */
struct S20 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* area_lower */
};

/* Struct for type [detachable] ISE_EXCEPTION_MANAGER */
struct S21 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] GEC */
struct S25 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* error_handler */
	T0* a2; /* ecf_filename */
	T0* a3; /* last_system */
	T0* a4; /* target_option */
	T0* a5; /* setting_option */
	T0* a6; /* capability_option */
	T0* a7; /* variable_option */
	T0* a8; /* finalize_flag */
	T0* a9; /* gelint_flag */
	T0* a10; /* ise_option */
	T0* a11; /* catcall_option */
	T0* a12; /* c_compile_option */
	T0* a13; /* split_option */
	T0* a14; /* split_size_option */
	T0* a15; /* new_instance_types_option */
	T0* a16; /* gc_option */
	T0* a17; /* thread_option */
	T0* a18; /* silent_flag */
	T0* a19; /* no_benchmark_flag */
	T0* a20; /* nested_benchmark_flag */
	T0* a21; /* metrics_flag */
	T0* a22; /* verbose_flag */
	T0* a23; /* version_flag */
	T0* a24; /* override_settings */
	T0* a25; /* override_capabilities */
	T0* a26; /* override_variables */
	T1 a27; /* ecf_gelint_option */
	T0* a28; /* ise_version */
	T0* a29; /* new_instance_types */
	T6 a30; /* split_size */
};

/* Struct for type [detachable] CELL [detachable EXCEPTION] */
struct S26 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
struct S27 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T0* a2; /* content */
	T0* a3; /* keys */
	T0* a4; /* deleted_marks */
	T0* a5; /* indexes_map */
	T6 a6; /* iteration_position */
	T6 a7; /* count */
	T6 a8; /* deleted_item_position */
	T6 a9; /* control */
	T6 a10; /* found_item */
	T1 a11; /* has_default */
	T6 a12; /* item_position */
	T6 a13; /* ht_lowest_deleted_position */
	T6 a14; /* ht_deleted_item */
	T6 a15; /* ht_deleted_key */
	T1 a16; /* object_comparison */
};

/* Struct for type [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
struct S28 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] CELL [[attached] NO_MORE_MEMORY] */
struct S29 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type detachable C_STRING */
struct S30 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* managed_data */
	T6 a2; /* count */
};

/* Struct for type [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
struct S31 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 z1;
	T6 z2;
	T6 z3;
	T0* z4;
	T0* z5;
	T0* z6;
	T0* z7;
	T0* z8;
	T0* z9;
	T6 z10;
	T1 z11;
};

/* Struct for type [detachable] KL_TEXT_INPUT_FILE */
struct S32 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_file */
	T0* a2; /* last_string */
	T6 a3; /* mode */
	T0* a4; /* name */
	T0* a5; /* character_buffer */
	T2 a6; /* last_character */
	T14 a7; /* file_pointer */
	T1 a8; /* descriptor_available */
	T0* a9; /* internal_name */
	T0* a10; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] KL_ARGUMENTS */
struct S33 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* program_name */
};

/* Struct for type [detachable] ET_ISE_VARIABLES */
struct S34 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_ERROR_HANDLER */
struct S35 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* info_file */
	T0* a3; /* mutex */
	T0* a4; /* error_file */
	T0* a5; /* warning_file */
	T1 a6; /* is_ise */
	T1 a7; /* is_verbose */
	T1 a8; /* has_eiffel_error */
	T1 a9; /* has_internal_error */
	T1 a10; /* is_ge */
};

/* Struct for type detachable ET_SYSTEM */
struct S36 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implicit_attachment_type_mark */
	T0* a2; /* integer_type */
	T0* a3; /* current_system */
	T0* a4; /* tuple_type */
	T0* a5; /* boolean_type */
	T0* a6; /* pointer_type */
	T0* a7; /* system_object_parents */
	T0* a8; /* any_parents */
	T0* a9; /* system_object_type */
	T0* a10; /* any_type */
	T0* a11; /* register_class_mutex */
	T0* a12; /* external_include_pathnames */
	T0* a13; /* external_object_pathnames */
	T0* a14; /* external_library_pathnames */
	T0* a15; /* external_make_pathnames */
	T0* a16; /* external_resource_pathnames */
	T0* a17; /* external_cflags */
	T0* a18; /* external_linker_flags */
	T1 a19; /* console_application_mode */
	T1 a20; /* alias_transition_mode */
	T1 a21; /* total_order_on_reals_mode */
	T1 a22; /* attachment_type_conformance_mode */
	T1 a23; /* scoop_mode */
	T6 a24; /* default_create_seed */
	T0* a25; /* universe */
	T0* a26; /* name */
	T0* a27; /* libraries */
	T0* a28; /* clusters */
	T0* a29; /* dotnet_assemblies */
	T0* a30; /* master_classes */
	T0* a31; /* master_class_mutex */
	T0* a32; /* system_name */
	T0* a33; /* detachable_any_type */
	T0* a34; /* detachable_separate_any_type */
	T0* a35; /* any_parent */
	T0* a36; /* any_clients */
	T0* a37; /* detachable_tuple_type */
	T0* a38; /* tuple_identity_type */
	T0* a39; /* unfolded_empty_tuple_actual_parameters */
	T0* a40; /* array_any_type */
	T0* a41; /* array_detachable_any_type */
	T0* a42; /* array_none_type */
	T0* a43; /* array_identity_type */
	T0* a44; /* character_type */
	T0* a45; /* character_8_type */
	T0* a46; /* character_8_convert_feature */
	T0* a47; /* character_32_type */
	T0* a48; /* character_32_convert_feature */
	T0* a49; /* double_type */
	T0* a50; /* exception_type */
	T0* a51; /* detachable_exception_type */
	T0* a52; /* exception_manager_type */
	T0* a53; /* function_identity_any_type */
	T0* a54; /* immutable_string_8_type */
	T0* a55; /* immutable_string_32_type */
	T0* a56; /* integer_8_type */
	T0* a57; /* integer_8_convert_feature */
	T0* a58; /* integer_16_type */
	T0* a59; /* integer_16_convert_feature */
	T0* a60; /* integer_32_type */
	T0* a61; /* integer_32_convert_feature */
	T0* a62; /* integer_64_type */
	T0* a63; /* integer_64_convert_feature */
	T0* a64; /* ise_exception_manager_type */
	T0* a65; /* iterable_detachable_separate_any_type */
	T0* a66; /* natural_type */
	T0* a67; /* natural_8_type */
	T0* a68; /* natural_8_convert_feature */
	T0* a69; /* natural_16_type */
	T0* a70; /* natural_16_convert_feature */
	T0* a71; /* natural_32_type */
	T0* a72; /* natural_32_convert_feature */
	T0* a73; /* natural_64_type */
	T0* a74; /* natural_64_convert_feature */
	T0* a75; /* none_type */
	T0* a76; /* detachable_none_type */
	T0* a77; /* predicate_identity_type */
	T0* a78; /* procedure_identity_type */
	T0* a79; /* real_type */
	T0* a80; /* real_32_type */
	T0* a81; /* real_32_convert_feature */
	T0* a82; /* real_64_type */
	T0* a83; /* real_64_convert_feature */
	T0* a84; /* routine_identity_type */
	T0* a85; /* special_any_type */
	T0* a86; /* special_detachable_any_type */
	T0* a87; /* special_identity_type */
	T0* a88; /* string_type */
	T0* a89; /* detachable_string_type */
	T0* a90; /* string_8_type */
	T0* a91; /* string_8_convert_feature */
	T0* a92; /* string_32_type */
	T0* a93; /* string_32_convert_feature */
	T0* a94; /* system_string_type */
	T0* a95; /* type_detachable_any_type */
	T0* a96; /* detachable_type_exception_type */
	T0* a97; /* type_detachable_exception_type */
	T0* a98; /* type_detachable_like_current_type */
	T0* a99; /* type_identity_type */
	T0* a100; /* typed_pointer_identity_type */
	T0* a101; /* wide_character_type */
	T1 a102; /* is_read_only */
	T6 a103; /* registered_class_count */
	T6 a104; /* copy_seed */
	T6 a105; /* is_equal_seed */
	T6 a106; /* routine_call_seed */
	T6 a107; /* function_item_seed */
	T6 a108; /* dispose_seed */
	T6 a109; /* iterable_new_cursor_seed */
	T6 a110; /* iteration_cursor_item_seed */
	T6 a111; /* iteration_cursor_after_seed */
	T6 a112; /* iteration_cursor_forth_seed */
	T1 a113; /* target_type_attachment_mode */
	T1 a114; /* obsolete_routine_type_mode */
	T1 a115; /* is_dotnet */
};

/* Struct for type [detachable] KL_EXCEPTIONS */
struct S38 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] AP_PARSER */
struct S39 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* options */
	T0* a2; /* alternative_options_lists */
	T0* a3; /* parameters */
	T0* a4; /* parameters_description */
	T0* a5; /* help_option */
	T0* a6; /* error_handler */
	T0* a7; /* application_description */
	T0* a8; /* current_options */
	T1 a9; /* is_first_option */
	T0* a10; /* last_option_parameter */
};

/* Struct for type [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
struct S40 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* introduction_option */
	T0* a2; /* parameters_description */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] AP_STRING_OPTION */
struct S41 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* has_default_parameter */
	T1 a6; /* needs_parameter */
	T0* a7; /* long_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
	T2 a12; /* short_form */
};

/* Struct for type [detachable] AP_FLAG */
struct S43 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* occurrences */
	T0* a2; /* description */
	T0* a3; /* long_form */
	T2 a4; /* short_form */
	T1 a5; /* has_short_form */
	T1 a6; /* is_hidden */
	T1 a7; /* is_mandatory */
	T6 a8; /* maximum_occurrences */
};

/* Struct for type [detachable] UT_VERSION */
struct S44 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_major */
	T6 a2; /* internal_minor */
	T6 a3; /* internal_revision */
	T6 a4; /* internal_build */
};

/* Struct for type [detachable] AP_ENUMERATION_OPTION */
struct S45 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* possible_values */
	T0* a5; /* parameter_description */
	T1 a6; /* needs_parameter */
	T0* a7; /* long_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
	T2 a12; /* short_form */
};

/* Struct for type [detachable] AP_BOOLEAN_OPTION */
struct S46 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T1 a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* needs_parameter */
	T0* a6; /* long_form */
	T1 a7; /* is_hidden */
	T1 a8; /* is_mandatory */
	T6 a9; /* maximum_occurrences */
	T1 a10; /* has_short_form */
	T2 a11; /* short_form */
};

/* Struct for type [detachable] AP_INTEGER_OPTION */
struct S47 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T6 a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* needs_parameter */
	T0* a6; /* long_form */
	T2 a7; /* short_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
};

/* Struct for type [detachable] ET_NULL_ERROR_HANDLER */
struct S51 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* info_file */
	T0* a3; /* mutex */
	T0* a4; /* error_file */
	T0* a5; /* warning_file */
	T1 a6; /* is_ise */
	T1 a7; /* is_verbose */
	T1 a8; /* has_eiffel_error */
	T1 a9; /* has_internal_error */
	T1 a10; /* is_ge */
};

/* Struct for type [detachable] ET_ECF_SYSTEM_PARSER */
struct S53 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_system */
	T0* a2; /* ast_factory */
	T0* a3; /* override_settings */
	T0* a4; /* override_capabilities */
	T0* a5; /* override_variables */
	T0* a6; /* parsed_libraries */
	T0* a7; /* parsed_dotnet_assemblies */
	T0* a8; /* redirected_locations */
	T0* a9; /* xml_parser */
	T0* a10; /* tree_pipe */
	T0* a11; /* error_handler */
	T0* a12; /* ise_version */
};

/* Struct for type [detachable] ET_ECF_ERROR_HANDLER */
struct S54 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* error_count */
	T0* a2; /* error_file */
	T0* a3; /* warning_file */
	T0* a4; /* info_file */
};

/* Struct for type detachable ET_ECF_SETTINGS */
struct S55 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_settings */
	T0* a2; /* secondary_settings */
};

/* Struct for type [detachable] ET_ECF_SYSTEM */
struct S56 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* filename */
	T0* a2; /* targets */
	T0* a3; /* library_target */
	T0* a4; /* libraries */
	T0* a5; /* selected_target */
	T0* a6; /* dotnet_assemblies */
	T0* a7; /* master_classes */
	T0* a8; /* root_type */
	T0* a9; /* none_type */
	T0* a10; /* any_type */
	T0* a11; /* root_creation */
	T6 a12; /* default_create_seed */
	T0* a13; /* boolean_type */
	T0* a14; /* character_8_type */
	T0* a15; /* character_32_type */
	T0* a16; /* integer_8_type */
	T0* a17; /* integer_16_type */
	T0* a18; /* integer_32_type */
	T0* a19; /* integer_64_type */
	T0* a20; /* natural_8_type */
	T0* a21; /* natural_16_type */
	T0* a22; /* natural_32_type */
	T0* a23; /* natural_64_type */
	T0* a24; /* real_32_type */
	T0* a25; /* real_64_type */
	T0* a26; /* pointer_type */
	T0* a27; /* special_any_type */
	T0* a28; /* string_8_type */
	T0* a29; /* string_32_type */
	T0* a30; /* immutable_string_8_type */
	T0* a31; /* immutable_string_32_type */
	T0* a32; /* array_any_type */
	T0* a33; /* special_identity_type */
	T0* a34; /* typed_pointer_identity_type */
	T0* a35; /* routine_identity_type */
	T0* a36; /* detachable_tuple_type */
	T0* a37; /* universe */
	T0* a38; /* ise_exception_manager_type */
	T0* a39; /* detachable_none_type */
	T1 a40; /* attachment_type_conformance_mode */
	T1 a41; /* use_boehm_gc */
	T0* a42; /* external_include_pathnames */
	T0* a43; /* external_cflags */
	T0* a44; /* external_linker_flags */
	T0* a45; /* external_library_pathnames */
	T0* a46; /* external_object_pathnames */
	T0* a47; /* character_type */
	T0* a48; /* wide_character_type */
	T0* a49; /* integer_type */
	T0* a50; /* natural_type */
	T0* a51; /* real_type */
	T0* a52; /* double_type */
	T0* a53; /* external_make_pathnames */
	T0* a54; /* external_resource_pathnames */
	T6 a55; /* routine_call_seed */
	T6 a56; /* function_item_seed */
	T6 a57; /* is_equal_seed */
	T6 a58; /* dispose_seed */
	T1 a59; /* trace_mode */
	T1 a60; /* exception_trace_mode */
	T1 a61; /* multithreaded_mode */
	T1 a62; /* total_order_on_reals_mode */
	T1 a63; /* console_application_mode */
	T0* a64; /* system_name */
	T0* a65; /* name */
	T0* a66; /* ecf_namespace */
	T0* a67; /* ecf_version */
	T0* a68; /* description */
	T0* a69; /* notes */
	T1 a70; /* is_read_only */
	T0* a71; /* uuid */
	T0* a72; /* register_class_mutex */
	T1 a73; /* alias_transition_mode */
	T0* a74; /* current_system */
	T0* a75; /* tuple_type */
	T0* a76; /* detachable_exception_type */
	T0* a77; /* exception_type */
	T0* a78; /* string_type */
	T1 a79; /* is_preparsed */
	T0* a80; /* clusters */
	T0* a81; /* implicit_attachment_type_mark */
	T0* a82; /* master_class_mutex */
	T0* a83; /* detachable_any_type */
	T0* a84; /* detachable_separate_any_type */
	T0* a85; /* any_parent */
	T0* a86; /* any_parents */
	T0* a87; /* any_clients */
	T0* a88; /* tuple_identity_type */
	T0* a89; /* unfolded_empty_tuple_actual_parameters */
	T0* a90; /* array_detachable_any_type */
	T0* a91; /* array_none_type */
	T0* a92; /* array_identity_type */
	T0* a93; /* character_8_convert_feature */
	T0* a94; /* character_32_convert_feature */
	T0* a95; /* exception_manager_type */
	T0* a96; /* function_identity_any_type */
	T0* a97; /* integer_8_convert_feature */
	T0* a98; /* integer_16_convert_feature */
	T0* a99; /* integer_32_convert_feature */
	T0* a100; /* integer_64_convert_feature */
	T0* a101; /* iterable_detachable_separate_any_type */
	T0* a102; /* natural_8_convert_feature */
	T0* a103; /* natural_16_convert_feature */
	T0* a104; /* natural_32_convert_feature */
	T0* a105; /* natural_64_convert_feature */
	T0* a106; /* predicate_identity_type */
	T0* a107; /* procedure_identity_type */
	T0* a108; /* real_32_convert_feature */
	T0* a109; /* real_64_convert_feature */
	T0* a110; /* special_detachable_any_type */
	T0* a111; /* detachable_string_type */
	T0* a112; /* string_8_convert_feature */
	T0* a113; /* string_32_convert_feature */
	T0* a114; /* system_object_type */
	T0* a115; /* system_object_parents */
	T0* a116; /* system_string_type */
	T0* a117; /* type_detachable_any_type */
	T0* a118; /* detachable_type_exception_type */
	T0* a119; /* type_detachable_exception_type */
	T0* a120; /* type_detachable_like_current_type */
	T0* a121; /* type_identity_type */
	T0* a122; /* scm_read_mapping_builder */
	T0* a123; /* scm_write_mapping_builder */
	T6 a124; /* registered_class_count */
	T1 a125; /* scoop_mode */
	T1 a126; /* is_dotnet */
	T1 a127; /* obsolete_routine_type_mode */
	T1 a128; /* target_type_attachment_mode */
	T6 a129; /* copy_seed */
	T6 a130; /* iterable_new_cursor_seed */
	T6 a131; /* iteration_cursor_item_seed */
	T6 a132; /* iteration_cursor_after_seed */
	T6 a133; /* iteration_cursor_forth_seed */
};

/* Struct for type [detachable] ET_ECF_TARGET */
struct S57 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* settings */
	T0* a2; /* capabilities */
	T0* a3; /* options */
	T0* a4; /* system_config */
	T0* a5; /* root */
	T0* a6; /* variables */
	T0* a7; /* name */
	T0* a8; /* parent */
	T0* a9; /* description */
	T0* a10; /* precompiled_library */
	T0* a11; /* version */
	T1 a12; /* is_abstract */
	T0* a13; /* clusters */
	T0* a14; /* libraries */
	T0* a15; /* dotnet_assemblies */
	T0* a16; /* file_rules */
	T0* a17; /* class_mappings */
	T0* a18; /* external_cflags */
	T0* a19; /* external_includes */
	T0* a20; /* external_libraries */
	T0* a21; /* external_linker_flags */
	T0* a22; /* external_makes */
	T0* a23; /* external_objects */
	T0* a24; /* external_resources */
	T0* a25; /* pre_compile_actions */
	T0* a26; /* post_compile_actions */
	T0* a27; /* notes */
};

/* Struct for type detachable ET_ECF_CAPABILITIES */
struct S59 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_use_capabilities */
	T0* a2; /* primary_support_capabilities */
	T0* a3; /* secondary_capabilities */
};

/* Struct for type detachable ET_ECF_VARIABLES */
struct S60 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* secondary_variables */
	T0* a2; /* primary_variables */
};

/* Struct for type [detachable] ET_DYNAMIC_SYSTEM */
struct S61 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* dynamic_types */
	T1 a3; /* catcall_error_mode */
	T0* a4; /* current_system */
	T0* a5; /* system_processor */
	T0* a6; /* dynamic_generic_types_by_name */
	T0* a7; /* null_dynamic_type_set_builder */
	T1 a8; /* full_class_checking */
	T1 a9; /* catcall_warning_mode */
	T0* a10; /* new_instance_types */
	T0* a11; /* dynamic_type_set_builder */
	T0* a12; /* unknown_type */
	T0* a13; /* boolean_type */
	T0* a14; /* character_8_type */
	T0* a15; /* character_32_type */
	T0* a16; /* immutable_string_8_type */
	T0* a17; /* immutable_string_32_type */
	T0* a18; /* integer_8_type */
	T0* a19; /* integer_16_type */
	T0* a20; /* integer_32_type */
	T0* a21; /* integer_64_type */
	T0* a22; /* natural_8_type */
	T0* a23; /* natural_16_type */
	T0* a24; /* natural_32_type */
	T0* a25; /* natural_64_type */
	T0* a26; /* real_32_type */
	T0* a27; /* real_64_type */
	T0* a28; /* pointer_type */
	T0* a29; /* string_8_type */
	T0* a30; /* string_32_type */
	T0* a31; /* special_character_8_type */
	T0* a32; /* special_character_32_type */
	T0* a33; /* ise_exception_manager_type */
	T0* a34; /* any_type */
	T0* a35; /* none_type */
	T0* a36; /* root_type */
	T0* a37; /* root_creation_procedure */
	T0* a38; /* special_count_feature */
	T0* a39; /* special_capacity_feature */
	T0* a40; /* array_area_feature */
	T0* a41; /* array_lower_feature */
	T0* a42; /* array_upper_feature */
	T0* a43; /* typed_pointer_to_pointer_feature */
	T0* a44; /* routine_closed_operands_feature */
	T0* a45; /* routine_rout_disp_feature */
	T0* a46; /* routine_set_rout_disp_final_feature */
	T0* a47; /* ise_exception_manager_init_exception_manager_feature */
	T0* a48; /* ise_exception_manager_last_exception_feature */
	T0* a49; /* ise_exception_manager_once_raise_feature */
	T0* a50; /* ise_exception_manager_set_exception_data_feature */
	T0* a51; /* ise_runtime_type_conforms_to_feature */
	T1 a52; /* in_dynamic_primary_type */
	T0* a53; /* ise_runtime_new_type_instance_of_feature */
	T1 a54; /* in_create_meta_type */
	T1 a55; /* all_attributes_used */
	T0* a56; /* ise_runtime_reference_field_feature */
	T0* a57; /* ise_runtime_set_reference_field_feature */
};

/* Struct for type [detachable] ET_SYSTEM_PROCESSOR */
struct S62 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* interface_checker */
	T0* a2; /* stop_request */
	T0* a3; /* eiffel_parser */
	T0* a4; /* implementation_checker */
	T0* a5; /* error_handler */
	T1 a6; /* benchmark_shown */
	T0* a7; /* ise_version */
	T0* a8; /* eiffel_preparser */
	T0* a9; /* master_class_checker */
	T0* a10; /* provider_checker */
	T0* a11; /* ancestor_builder */
	T0* a12; /* feature_flattener */
	T0* a13; /* processed_class_count_stack */
	T0* a14; /* ast_factory */
	T0* a15; /* dotnet_assembly_consumer */
	T0* a16; /* ecma_version */
	T1 a17; /* nested_benchmark_shown */
	T1 a18; /* metrics_shown */
	T1 a19; /* flat_mode */
	T1 a20; /* flat_dbc_mode */
	T1 a21; /* preparse_shallow_mode */
	T1 a22; /* preparse_single_mode */
	T1 a23; /* preparse_multiple_mode */
	T1 a24; /* unknown_builtin_reported */
	T1 a25; /* qualified_anchored_types_cycle_detection_enabled */
	T1 a26; /* use_attribute_keyword */
	T1 a27; /* use_note_keyword */
	T1 a28; /* use_reference_keyword */
	T1 a29; /* use_attached_keyword */
	T1 a30; /* use_detachable_keyword */
	T1 a31; /* cluster_dependence_enabled */
	T6 a32; /* postponed_class_count */
	T6 a33; /* processed_class_count */
	T1 a34; /* suppliers_enabled */
	T1 a35; /* preparse_override_mode */
	T1 a36; /* preparse_readonly_mode */
	T1 a37; /* providers_enabled */
	T1 a38; /* use_cluster_dependence_pathnames */
};

/* Struct for type [detachable] ET_C_GENERATOR */
struct S65 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* never_void_target_count */
	T6 a2; /* can_be_void_target_count */
	T1 a3; /* has_fatal_error */
	T0* a4; /* system_processor */
	T0* a5; /* current_dynamic_system */
	T0* a6; /* dynamic_types */
	T1 a7; /* short_names */
	T1 a8; /* split_mode */
	T6 a9; /* split_threshold */
	T1 a10; /* use_boehm_gc */
	T0* a11; /* system_name */
	T0* a12; /* current_file */
	T0* a13; /* header_file */
	T0* a14; /* current_type */
	T0* a15; /* current_feature */
	T0* a16; /* extra_dynamic_type_sets */
	T0* a17; /* current_dynamic_type_sets */
	T0* a18; /* current_function_header_buffer */
	T0* a19; /* current_function_body_buffer */
	T0* a20; /* temp_variables */
	T0* a21; /* used_temp_variables */
	T0* a22; /* free_temp_variables */
	T0* a23; /* frozen_temp_variables */
	T0* a24; /* conforming_types */
	T0* a25; /* conforming_type_set */
	T0* a26; /* non_conforming_types */
	T0* a27; /* equality_type_set */
	T0* a28; /* equality_common_types */
	T0* a29; /* operand_stack */
	T0* a30; /* call_operands */
	T0* a31; /* attachment_dynamic_type_ids */
	T0* a32; /* target_dynamic_type_ids */
	T0* a33; /* target_dynamic_types */
	T0* a34; /* standalone_type_sets */
	T0* a35; /* deep_twin_types */
	T0* a36; /* deep_equal_types */
	T0* a37; /* deep_feature_target_type_sets */
	T0* a38; /* current_object_tests */
	T0* a39; /* current_object_equalities */
	T0* a40; /* current_equalities */
	T0* a41; /* current_agents */
	T0* a42; /* agent_open_operands */
	T0* a43; /* agent_closed_operands */
	T0* a44; /* agent_target */
	T0* a45; /* agent_arguments */
	T0* a46; /* agent_instruction */
	T0* a47; /* agent_expression */
	T0* a48; /* agent_closed_operands_type */
	T0* a49; /* wrapper_expression */
	T0* a50; /* agent_tuple_item_expressions */
	T0* a51; /* agent_manifest_tuple */
	T0* a52; /* formal_arguments */
	T0* a53; /* manifest_array_types */
	T0* a54; /* big_manifest_array_types */
	T0* a55; /* manifest_tuple_types */
	T0* a56; /* once_features */
	T0* a57; /* once_per_process_counts */
	T0* a58; /* once_per_thread_counts */
	T0* a59; /* constant_features */
	T0* a60; /* inline_constants */
	T0* a61; /* dispose_procedures */
	T0* a62; /* dynamic_type_id_set_names */
	T0* a63; /* called_features */
	T0* a64; /* called_static_features */
	T0* a65; /* included_header_filenames */
	T0* a66; /* included_runtime_header_files */
	T0* a67; /* included_runtime_c_files */
	T0* a68; /* c_filenames */
	T1 a69; /* finalize_mode */
	T0* a70; /* locals_written_in_body */
	T0* a71; /* locals_written_in_rescue */
	T0* a72; /* locals_written */
	T0* a73; /* locals_read_in_body */
	T0* a74; /* locals_read_in_rescue */
	T0* a75; /* locals_read */
	T0* a76; /* external_c_regexp */
	T0* a77; /* external_c_macro_regexp */
	T0* a78; /* external_c_struct_regexp */
	T0* a79; /* external_c_inline_regexp */
	T0* a80; /* old_external_c_regexp */
	T0* a81; /* old_external_c_macro_regexp */
	T0* a82; /* old_external_c_struct_regexp */
	T0* a83; /* external_cpp_regexp */
	T0* a84; /* external_cpp_struct_regexp */
	T0* a85; /* external_cpp_inline_regexp */
	T0* a86; /* external_dllwin_regexp */
	T0* a87; /* old_external_dllwin32_regexp */
	T1 a88; /* in_static_feature */
	T1 a89; /* type_info_ancestors_used */
	T1 a90; /* type_info_attributes_used */
	T1 a91; /* type_info_attribute_name_used */
	T1 a92; /* type_info_attribute_type_id_used */
	T1 a93; /* type_info_attribute_offset_used */
	T1 a94; /* type_info_generator_used */
	T1 a95; /* type_info_name_used */
	T1 a96; /* type_info_generic_parameters_used */
	T1 a97; /* type_info_object_size_used */
	T0* a98; /* c_file */
	T6 a99; /* c_file_size */
	T0* a100; /* cpp_file */
	T6 a101; /* cpp_file_size */
	T6 a102; /* indentation */
	T0* a103; /* assignment_target */
	T1 a104; /* in_operand */
	T0* a105; /* call_target_type */
	T0* a106; /* current_agent */
	T1 a107; /* call_target_check_void */
	T1 a108; /* has_rescue */
	T1 a109; /* result_written */
	T1 a110; /* result_read */
	T1 a111; /* has_retry */
};

/* Struct for type [detachable] KL_SHELL_COMMAND */
struct S66 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* exit_code */
	T0* a2; /* string_command */
	T0* a3; /* command */
	T6 a4; /* return_code */
	T1 a5; /* is_system_code */
};

/* Struct for type detachable DT_DATE_TIME */
struct S67 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* time_storage */
	T6 a2; /* date_storage */
};

/* Struct for type detachable DS_HASH_SET [[attached] STRING_8] */
struct S69 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* internal_cursor */
	T0* a3; /* item_storage */
	T0* a4; /* equality_tester */
	T6 a5; /* count */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* free_slot */
	T6 a9; /* last_position */
	T6 a10; /* modulus */
	T6 a11; /* clashes_previous_position */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
struct S70 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* dynamic_type_set_count */
	T1 a2; /* has_fatal_error */
	T0* a3; /* current_dynamic_system */
	T0* a4; /* feature_checker */
	T0* a5; /* current_dynamic_type */
	T0* a6; /* current_dynamic_feature */
	T0* a7; /* dynamic_type_sets */
	T0* a8; /* object_id_dynamic_type_set */
	T0* a9; /* constant_indexes */
	T1 a10; /* catcall_error_mode */
	T1 a11; /* catcall_warning_mode */
	T1 a12; /* no_debug */
	T1 a13; /* no_assertion */
	T1 a14; /* is_built */
	T0* a15; /* alive_conforming_descendants_per_type */
	T0* a16; /* type_checker */
	T0* a17; /* current_type */
	T0* a18; /* current_class */
	T0* a19; /* current_feature */
	T0* a20; /* current_feature_impl */
	T0* a21; /* current_class_impl */
	T0* a22; /* enclosing_inline_agents */
	T0* a23; /* unused_overloaded_procedures_list */
	T0* a24; /* unused_overloaded_queries_list */
	T0* a25; /* unused_overloaded_features_list */
	T0* a26; /* unused_call_infos */
	T0* a27; /* unused_contexts */
	T0* a28; /* current_context */
	T0* a29; /* current_target_type */
	T0* a30; /* current_object_test_types */
	T0* a31; /* current_object_test_scope */
	T0* a32; /* object_test_scope_builder */
	T0* a33; /* current_across_cursor_types */
	T0* a34; /* current_across_cursor_scope */
	T0* a35; /* current_initialization_scope */
	T0* a36; /* current_attachment_scope */
	T0* a37; /* attachment_scope_builder */
	T0* a38; /* unused_attachment_scopes */
	T0* a39; /* common_ancestor_type_list */
	T0* a40; /* indexing_term_list */
	T0* a41; /* default_creation_call_name */
	T0* a42; /* default_creation_call */
	T0* a43; /* vape_non_descendant_clients */
	T0* a44; /* vape_creation_clients */
	T0* a45; /* vape_client */
	T0* a46; /* unused_adapted_base_classes */
	T0* a47; /* adapted_base_class_checker */
	T0* a48; /* system_processor */
	T6 a49; /* current_index */
	T6 a50; /* result_index */
	T6 a51; /* attached_result_index */
	T6 a52; /* void_index */
	T1 a53; /* in_precursor */
	T1 a54; /* in_static_feature */
	T1 a55; /* in_rescue */
	T0* a56; /* current_inline_agent */
	T0* a57; /* supplier_handler */
	T1 a58; /* in_invariant */
	T1 a59; /* in_precondition */
	T1 a60; /* in_postcondition */
	T1 a61; /* in_check_instruction */
	T0* a62; /* precursor_queries */
	T0* a63; /* precursor_procedures */
};

/* Struct for type [detachable] ET_CLASS_TYPE */
struct S71 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T0* a3; /* actual_parameters */
	T0* a4; /* named_base_class */
	T1 a5; /* tuple_actual_parameters_unfolded_1 */
	T1 a6; /* tuple_actual_parameters_unfolded_2 */
};

/* Struct for type [detachable] ET_TOKEN_CONSTANTS */
struct S72 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_CLASS */
struct S73 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* group */
	T0* a2; /* status_mutex */
	T1 a3; /* unprotected_is_parsed */
	T1 a4; /* unprotected_has_syntax_error */
	T1 a5; /* unprotected_ancestors_built */
	T1 a6; /* unprotected_has_ancestors_error */
	T1 a7; /* unprotected_features_flattened */
	T1 a8; /* unprotected_has_flattening_error */
	T1 a9; /* unprotected_interface_checked */
	T1 a10; /* unprotected_has_interface_error */
	T1 a11; /* unprotected_implementation_checked */
	T1 a12; /* unprotected_has_implementation_error */
	T0* a13; /* name */
	T6 a14; /* id */
	T0* a15; /* ancestors */
	T0* a16; /* conforming_ancestors */
	T0* a17; /* queries */
	T0* a18; /* procedures */
	T0* a19; /* class_keyword */
	T0* a20; /* end_keyword */
	T0* a21; /* processing_mutex */
	T0* a22; /* formal_parameter_types_mutex */
	T0* a23; /* named_base_class */
	T6 a24; /* time_stamp */
	T8 a25; /* class_code */
	T1 a26; /* tuple_actual_parameters_unfolded_1 */
	T1 a27; /* tuple_actual_parameters_unfolded_2 */
	T1 a28; /* is_checking_implementation */
	T0* a29; /* formal_parameters */
	T6 a30; /* index */
	T0* a31; /* class_mark */
	T1 a32; /* has_deferred_features */
	T0* a33; /* formal_parameter_types */
	T0* a34; /* parent_clauses */
	T0* a35; /* suppliers */
	T1 a36; /* redeclared_signatures_checked */
	T0* a37; /* invariants */
	T0* a38; /* filename */
	T0* a39; /* creators */
	T0* a40; /* convert_features */
	T1 a41; /* is_ignored */
	T0* a42; /* providers */
	T1 a43; /* is_interface */
	T1 a44; /* unprotected_is_marked */
	T1 a45; /* has_utf8_bom */
	T6 a46; /* registered_feature_count */
	T0* a47; /* external_keyword */
	T0* a48; /* frozen_keyword */
	T0* a49; /* feature_clauses */
	T0* a50; /* first_indexing */
	T0* a51; /* second_indexing */
	T6 a52; /* tuple_constraint_position */
	T0* a53; /* obsolete_message */
	T6 a54; /* registered_inline_constant_count */
	T0* a55; /* leading_break */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S75 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] KL_OPERATING_SYSTEM */
struct S76 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] UT_CANNOT_READ_FILE_ERROR */
struct S79 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_VERSION_NUMBER */
struct S81 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_MESSAGE */
struct S82 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] RX_PCRE_REGULAR_EXPRESSION */
struct S83 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* match_count */
	T0* a2; /* subject */
	T6 a3; /* subject_end */
	T6 a4; /* subject_start */
	T0* a5; /* offset_vector */
	T6 a6; /* offset_vector_count */
	T6 a7; /* brastart_capacity */
	T0* a8; /* brastart_vector */
	T6 a9; /* brastart_lower */
	T6 a10; /* brastart_count */
	T6 a11; /* eptr_capacity */
	T0* a12; /* eptr_vector */
	T6 a13; /* eptr_lower */
	T6 a14; /* eptr_upper */
	T0* a15; /* byte_code */
	T0* a16; /* internal_start_bits */
	T0* a17; /* pattern_buffer */
	T6 a18; /* subexpression_count */
	T0* a19; /* pattern */
	T6 a20; /* pattern_count */
	T6 a21; /* pattern_position */
	T6 a22; /* code_index */
	T6 a23; /* maxbackrefs */
	T10 a24; /* optchanged */
	T7 a25; /* first_character */
	T7 a26; /* required_character */
	T6 a27; /* regexp_countlits */
	T0* a28; /* start_bits */
	T0* a29; /* error_message */
	T1 a30; /* is_anchored */
	T0* a31; /* character_case_mapping */
	T0* a32; /* word_set */
	T6 a33; /* subject_next_start */
	T6 a34; /* error_code */
	T6 a35; /* error_position */
	T1 a36; /* is_startline */
	T1 a37; /* is_caseless */
	T1 a38; /* is_bol */
	T1 a39; /* is_eol */
	T1 a40; /* is_multiline */
	T1 a41; /* is_dotall */
	T1 a42; /* is_extended */
	T1 a43; /* is_empty_allowed */
	T1 a44; /* is_dollar_endonly */
	T1 a45; /* is_greedy */
	T1 a46; /* is_strict */
	T1 a47; /* is_ichanged */
	T6 a48; /* first_matched_index */
	T6 a49; /* eptr */
	T6 a50; /* offset_top */
	T1 a51; /* is_matching_caseless */
	T1 a52; /* is_matching_multiline */
	T1 a53; /* is_matching_dotall */
};

/* Struct for type [detachable] KL_STRING_ROUTINES */
struct S84 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_EXECUTION_ENVIRONMENT */
struct S87 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] AP_ERROR */
struct S88 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_template */
	T0* a3; /* code */
};

/* Struct for type [detachable] VOID_TARGET */
struct S91 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] ROUTINE_FAILURE */
struct S93 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* routine_name */
	T0* a2; /* class_name */
	T0* a3; /* throwing_exception */
	T0* a4; /* internal_trace */
	T0* a5; /* c_description */
	T0* a6; /* recipient_name */
	T0* a7; /* type_name */
};

/* Struct for type [detachable] OLD_VIOLATION */
struct S95 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] NO_MORE_MEMORY */
struct S97 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_trace */
	T6 a2; /* internal_code */
	T0* a3; /* throwing_exception */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] INVARIANT_VIOLATION */
struct S98 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T1 a2; /* is_entry */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
struct S99 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* signal_code */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] IO_FAILURE */
struct S100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_code */
	T0* a2; /* throwing_exception */
	T6 a3; /* error_code */
	T0* a4; /* internal_trace */
	T0* a5; /* c_description */
	T0* a6; /* recipient_name */
	T0* a7; /* type_name */
};

/* Struct for type [detachable] OPERATING_SYSTEM_FAILURE */
struct S101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* error_code */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] COM_FAILURE */
struct S102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* hresult_code */
	T0* a3; /* exception_information */
	T6 a4; /* hresult */
	T0* a5; /* internal_trace */
	T0* a6; /* c_description */
	T0* a7; /* recipient_name */
	T0* a8; /* type_name */
};

/* Struct for type [detachable] EIFFEL_RUNTIME_PANIC */
struct S103 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_code */
	T0* a2; /* throwing_exception */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] PRECONDITION_VIOLATION */
struct S105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] POSTCONDITION_VIOLATION */
struct S106 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] FLOATING_POINT_FAILURE */
struct S107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] CHECK_VIOLATION */
struct S108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] BAD_INSPECT_VALUE */
struct S109 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] VARIANT_VIOLATION */
struct S110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] LOOP_INVARIANT_VIOLATION */
struct S111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] RESCUE_FAILURE */
struct S112 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] RESUMPTION_FAILURE */
struct S113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] CREATE_ON_DEFERRED */
struct S114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXTERNAL_FAILURE */
struct S115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] VOID_ASSIGNED_TO_EXPANDED */
struct S116 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
struct S117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] MISMATCH_FAILURE */
struct S118 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] DEVELOPER_EXCEPTION */
struct S119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
struct S120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] SERIALIZATION_FAILURE */
struct S121 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXECUTION_ENVIRONMENT */
struct S122 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* return_code */
};

/* Struct for type [detachable] KL_WINDOWS_FILE_SYSTEM */
struct S123 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_UNIX_FILE_SYSTEM */
struct S124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PRIMES */
struct S125 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] INTEGER_32] */
struct S126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T6 z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] BOOLEAN] */
struct S127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T1 z2[1]; /* item */
};

/* Struct for type detachable KL_LINKABLE [[attached] CHARACTER_8] */
struct S129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] MANAGED_POINTER */
struct S131 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* item */
	T1 a2; /* is_shared */
	T6 a3; /* count */
};

/* Struct for type [detachable] FILE_INFO */
struct S132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* buffered_file_info */
	T1 a2; /* is_following_symlinks */
	T1 a3; /* exists */
	T0* a4; /* internal_file_name */
	T0* a5; /* internal_name_pointer */
};

/* Struct for type [detachable] MUTEX */
struct S133 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* owner_thread_id */
	T14 a2; /* mutex_pointer */
};

/* Struct for type [detachable] UT_ERROR_HANDLER */
struct S134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* error_file */
	T0* a2; /* warning_file */
	T0* a3; /* info_file */
};

/* Struct for type [detachable] KL_STANDARD_FILES */
struct S135 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_STDERR_FILE */
struct S136 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_STDOUT_FILE */
struct S137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ST_WORD_WRAPPER */
struct S138 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* new_line_indentation */
	T6 a2; /* broken_words */
	T6 a3; /* maximum_text_width */
};

/* Struct for type [detachable] AP_DISPLAY_HELP_FLAG */
struct S140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* description */
	T2 a2; /* short_form */
	T1 a3; /* has_short_form */
	T0* a4; /* long_form */
	T6 a5; /* occurrences */
	T1 a6; /* is_hidden */
	T1 a7; /* is_mandatory */
	T6 a8; /* maximum_occurrences */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
struct S141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* count */
	T0* a3; /* equality_tester */
	T0* a4; /* storage */
	T0* a5; /* special_routines */
	T6 a6; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
struct S143 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* internal_cursor */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] AP_ERROR_HANDLER */
struct S144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* error_file */
	T0* a3; /* warning_file */
	T0* a4; /* info_file */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
struct S145 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
struct S146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
struct S147 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] STRING_8] */
struct S148 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
	T0* a5; /* equality_tester */
};

/* Struct for type [detachable] KL_STRING_EQUALITY_TESTER */
struct S149 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
struct S150 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
struct S153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
struct S156 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
struct S157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
};

/* Struct for type [detachable] ET_ECF_AST_FACTORY */
struct S158 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
struct S159 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET] */
struct S160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
struct S161 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable TUPLE */
struct S162 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
struct S163 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] XM_ELEMENT */
struct S164 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* namespace */
	T0* a3; /* parent */
	T0* a4; /* children */
};

/* Struct for type detachable XM_POSITION_TABLE */
struct S165 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
};

/* Struct for type [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
struct S166 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
struct S167 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
struct S168 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
struct S170 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T6 a6; /* count */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* position */
	T6 a10; /* capacity */
	T0* a11; /* equality_tester */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
struct S171 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
struct S172 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* internal_cursor */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] XM_EIFFEL_PARSER */
struct S174 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_last_error_description */
	T0* a2; /* scanner */
	T0* a3; /* error_positions */
	T0* a4; /* scanners */
	T1 a5; /* in_external_dtd */
	T0* a6; /* callbacks */
	T0* a7; /* entities */
	T0* a8; /* pe_entities */
	T0* a9; /* dtd_resolver */
	T0* a10; /* entity_resolver */
	T1 a11; /* use_namespaces */
	T6 a12; /* string_mode */
	T0* a13; /* dtd_callbacks */
	T0* a14; /* last_string_value */
	T0* a15; /* yyss */
	T0* a16; /* yyspecial_routines1 */
	T6 a17; /* yyvsc1 */
	T0* a18; /* yyvs1 */
	T0* a19; /* yyspecial_routines2 */
	T6 a20; /* yyvsc2 */
	T0* a21; /* yyvs2 */
	T0* a22; /* yyspecial_routines3 */
	T6 a23; /* yyvsc3 */
	T0* a24; /* yyvs3 */
	T0* a25; /* yyspecial_routines4 */
	T6 a26; /* yyvsc4 */
	T0* a27; /* yyvs4 */
	T0* a28; /* yyspecial_routines5 */
	T6 a29; /* yyvsc5 */
	T0* a30; /* yyvs5 */
	T0* a31; /* yyspecial_routines6 */
	T6 a32; /* yyvsc6 */
	T0* a33; /* yyvs6 */
	T0* a34; /* yyspecial_routines7 */
	T6 a35; /* yyvsc7 */
	T0* a36; /* yyvs7 */
	T0* a37; /* yyspecial_routines8 */
	T6 a38; /* yyvsc8 */
	T0* a39; /* yyvs8 */
	T0* a40; /* yyspecial_routines9 */
	T6 a41; /* yyvsc9 */
	T0* a42; /* yyvs9 */
	T0* a43; /* yyspecial_routines10 */
	T6 a44; /* yyvsc10 */
	T0* a45; /* yyvs10 */
	T0* a46; /* yyspecial_routines11 */
	T6 a47; /* yyvsc11 */
	T0* a48; /* yyvs11 */
	T0* a49; /* yytranslate */
	T0* a50; /* yyr1 */
	T0* a51; /* yytypes1 */
	T0* a52; /* yytypes2 */
	T0* a53; /* yydefact */
	T0* a54; /* yydefgoto */
	T0* a55; /* yypact */
	T0* a56; /* yypgoto */
	T0* a57; /* yytable */
	T0* a58; /* yycheck */
	T6 a59; /* yy_parsing_status */
	T6 a60; /* yy_suspended_yystacksize */
	T6 a61; /* yy_suspended_yystate */
	T6 a62; /* yy_suspended_yyn */
	T6 a63; /* yy_suspended_yychar1 */
	T6 a64; /* yy_suspended_index */
	T6 a65; /* yy_suspended_yyss_top */
	T6 a66; /* yy_suspended_yy_goto */
	T6 a67; /* yyssp */
	T6 a68; /* error_count */
	T1 a69; /* yy_lookahead_needed */
	T6 a70; /* yyerrstatus */
	T6 a71; /* last_token */
	T6 a72; /* yyvsp1 */
	T6 a73; /* yyvsp2 */
	T6 a74; /* yyvsp3 */
	T6 a75; /* yyvsp4 */
	T6 a76; /* yyvsp5 */
	T6 a77; /* yyvsp6 */
	T6 a78; /* yyvsp7 */
	T6 a79; /* yyvsp8 */
	T6 a80; /* yyvsp9 */
	T6 a81; /* yyvsp10 */
	T6 a82; /* yyvsp11 */
	T0* a83; /* last_detachable_any_value */
};

/* Struct for type [detachable] XM_TREE_CALLBACKS_PIPE */
struct S175 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* start */
	T0* a2; /* tree */
	T0* a3; /* error */
	T0* a4; /* last */
};

/* Struct for type [detachable] XM_CALLBACKS_TO_TREE_FILTER */
struct S178 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_position_table */
	T0* a2; /* document */
	T0* a3; /* source_parser */
	T0* a4; /* current_element */
	T0* a5; /* namespace_cache */
	T0* a6; /* next */
};

/* Struct for type [detachable] ET_ECF_STATE */
struct S179 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* ise_version */
};

/* Struct for type [detachable] ET_IDENTIFIER */
struct S180 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T6 a2; /* index */
	T0* a3; /* name */
	T2 a4; /* status_code */
	T6 a5; /* compressed_position */
	T6 a6; /* hash_code */
};

/* Struct for type [detachable] ET_ECF_SYSTEM_CONFIG */
struct S181 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* filename */
	T0* a3; /* universe */
	T0* a4; /* ecf_namespace */
	T0* a5; /* ecf_version */
	T0* a6; /* description */
	T0* a7; /* notes */
	T0* a8; /* targets */
	T1 a9; /* is_read_only */
	T0* a10; /* uuid */
	T0* a11; /* library_target */
};

/* Struct for type [detachable] ET_ECF_LIBRARY */
struct S183 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* filename */
	T0* a3; /* ecf_namespace */
	T0* a4; /* ecf_version */
	T0* a5; /* description */
	T0* a6; /* notes */
	T0* a7; /* targets */
	T1 a8; /* is_read_only */
	T0* a9; /* uuid */
	T0* a10; /* library_target */
	T0* a11; /* current_system */
	T0* a12; /* library */
	T0* a13; /* libraries */
	T0* a14; /* clusters */
	T0* a15; /* dotnet_assemblies */
	T0* a16; /* implicit_attachment_type_mark */
	T0* a17; /* master_classes */
	T0* a18; /* master_class_mutex */
	T0* a19; /* any_type */
	T0* a20; /* detachable_any_type */
	T0* a21; /* detachable_separate_any_type */
	T0* a22; /* any_parent */
	T0* a23; /* any_parents */
	T0* a24; /* any_clients */
	T0* a25; /* tuple_type */
	T0* a26; /* detachable_tuple_type */
	T0* a27; /* tuple_identity_type */
	T0* a28; /* unfolded_empty_tuple_actual_parameters */
	T0* a29; /* array_any_type */
	T0* a30; /* array_detachable_any_type */
	T0* a31; /* array_none_type */
	T0* a32; /* array_identity_type */
	T0* a33; /* boolean_type */
	T0* a34; /* character_type */
	T0* a35; /* character_8_type */
	T0* a36; /* character_8_convert_feature */
	T0* a37; /* character_32_type */
	T0* a38; /* character_32_convert_feature */
	T0* a39; /* double_type */
	T0* a40; /* exception_type */
	T0* a41; /* detachable_exception_type */
	T0* a42; /* exception_manager_type */
	T0* a43; /* function_identity_any_type */
	T0* a44; /* immutable_string_8_type */
	T0* a45; /* immutable_string_32_type */
	T0* a46; /* integer_type */
	T0* a47; /* integer_8_type */
	T0* a48; /* integer_8_convert_feature */
	T0* a49; /* integer_16_type */
	T0* a50; /* integer_16_convert_feature */
	T0* a51; /* integer_32_type */
	T0* a52; /* integer_32_convert_feature */
	T0* a53; /* integer_64_type */
	T0* a54; /* integer_64_convert_feature */
	T0* a55; /* ise_exception_manager_type */
	T0* a56; /* iterable_detachable_separate_any_type */
	T0* a57; /* natural_type */
	T0* a58; /* natural_8_type */
	T0* a59; /* natural_8_convert_feature */
	T0* a60; /* natural_16_type */
	T0* a61; /* natural_16_convert_feature */
	T0* a62; /* natural_32_type */
	T0* a63; /* natural_32_convert_feature */
	T0* a64; /* natural_64_type */
	T0* a65; /* natural_64_convert_feature */
	T0* a66; /* none_type */
	T0* a67; /* detachable_none_type */
	T0* a68; /* pointer_type */
	T0* a69; /* predicate_identity_type */
	T0* a70; /* procedure_identity_type */
	T0* a71; /* real_type */
	T0* a72; /* real_32_type */
	T0* a73; /* real_32_convert_feature */
	T0* a74; /* real_64_type */
	T0* a75; /* real_64_convert_feature */
	T0* a76; /* routine_identity_type */
	T0* a77; /* special_any_type */
	T0* a78; /* special_detachable_any_type */
	T0* a79; /* special_identity_type */
	T0* a80; /* string_type */
	T0* a81; /* detachable_string_type */
	T0* a82; /* string_8_type */
	T0* a83; /* string_8_convert_feature */
	T0* a84; /* string_32_type */
	T0* a85; /* string_32_convert_feature */
	T0* a86; /* system_object_type */
	T0* a87; /* system_object_parents */
	T0* a88; /* system_string_type */
	T0* a89; /* type_detachable_any_type */
	T0* a90; /* detachable_type_exception_type */
	T0* a91; /* type_detachable_exception_type */
	T0* a92; /* type_detachable_like_current_type */
	T0* a93; /* type_identity_type */
	T0* a94; /* typed_pointer_identity_type */
	T0* a95; /* wide_character_type */
	T0* a96; /* selected_target */
	T1 a97; /* obsolete_routine_type_mode */
	T1 a98; /* is_preparsed */
};

/* Struct for type [detachable] ET_ADAPTED_LIBRARIES */
struct S184 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* libraries */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
struct S186 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
struct S187 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] XM_DOCUMENT */
struct S188 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* root_element */
	T0* a2; /* children */
};

/* Struct for type [detachable] ET_COMPRESSED_POSITION */
struct S189 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
};

/* Struct for type [detachable] XM_STOP_ON_ERROR_FILTER */
struct S191 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* last_error */
	T0* a3; /* next */
};

/* Struct for type detachable XM_ATTRIBUTE */
struct S192 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* namespace */
	T0* a3; /* value */
	T0* a4; /* parent */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
struct S194 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T6 a4; /* last_position */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* key_equality_tester */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* special_key_routines */
	T0* a15; /* key_storage */
	T0* a16; /* equality_tester */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S195 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* position */
	T0* a3; /* internal_cursor */
	T0* a4; /* key_storage */
	T0* a5; /* item_storage */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* equality_tester */
	T6 a18; /* free_slot */
	T0* a19; /* hash_function */
};

/* Struct for type detachable ET_ECF_TARGET_PARENT */
struct S196 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* location_id */
	T0* a3; /* target */
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
struct S197 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
struct S198 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type detachable ET_ECF_TARGETS */
struct S199 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* targets */
};

/* Struct for type [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S200 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* hash_function */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S201 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
struct S202 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] ET_ECF_ADAPTED_LIBRARY */
struct S203 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* classname_prefix */
	T1 a7; /* is_read_only */
	T1 a8; /* use_application_options */
	T0* a9; /* options */
	T0* a10; /* class_options */
	T0* a11; /* class_renamings */
	T0* a12; /* visible_classes */
	T0* a13; /* conditions */
	T0* a14; /* notes */
	T0* a15; /* name */
	T0* a16; /* library */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
struct S204 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
struct S206 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type [detachable] ET_ECF_DOTNET_ASSEMBLY */
struct S207 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* referenced_assemblies */
	T0* a3; /* current_system */
	T0* a4; /* dotnet_assembly */
	T0* a5; /* name */
	T0* a6; /* implicit_attachment_type_mark */
	T0* a7; /* master_classes */
	T0* a8; /* master_class_mutex */
	T0* a9; /* any_type */
	T0* a10; /* detachable_any_type */
	T0* a11; /* detachable_separate_any_type */
	T0* a12; /* any_parent */
	T0* a13; /* any_parents */
	T0* a14; /* any_clients */
	T0* a15; /* tuple_type */
	T0* a16; /* detachable_tuple_type */
	T0* a17; /* tuple_identity_type */
	T0* a18; /* unfolded_empty_tuple_actual_parameters */
	T0* a19; /* array_any_type */
	T0* a20; /* array_detachable_any_type */
	T0* a21; /* array_none_type */
	T0* a22; /* array_identity_type */
	T0* a23; /* boolean_type */
	T0* a24; /* character_type */
	T0* a25; /* character_8_type */
	T0* a26; /* character_8_convert_feature */
	T0* a27; /* character_32_type */
	T0* a28; /* character_32_convert_feature */
	T0* a29; /* double_type */
	T0* a30; /* exception_type */
	T0* a31; /* detachable_exception_type */
	T0* a32; /* exception_manager_type */
	T0* a33; /* function_identity_any_type */
	T0* a34; /* immutable_string_8_type */
	T0* a35; /* immutable_string_32_type */
	T0* a36; /* integer_type */
	T0* a37; /* integer_8_type */
	T0* a38; /* integer_8_convert_feature */
	T0* a39; /* integer_16_type */
	T0* a40; /* integer_16_convert_feature */
	T0* a41; /* integer_32_type */
	T0* a42; /* integer_32_convert_feature */
	T0* a43; /* integer_64_type */
	T0* a44; /* integer_64_convert_feature */
	T0* a45; /* ise_exception_manager_type */
	T0* a46; /* iterable_detachable_separate_any_type */
	T0* a47; /* natural_type */
	T0* a48; /* natural_8_type */
	T0* a49; /* natural_8_convert_feature */
	T0* a50; /* natural_16_type */
	T0* a51; /* natural_16_convert_feature */
	T0* a52; /* natural_32_type */
	T0* a53; /* natural_32_convert_feature */
	T0* a54; /* natural_64_type */
	T0* a55; /* natural_64_convert_feature */
	T0* a56; /* none_type */
	T0* a57; /* detachable_none_type */
	T0* a58; /* pointer_type */
	T0* a59; /* predicate_identity_type */
	T0* a60; /* procedure_identity_type */
	T0* a61; /* real_type */
	T0* a62; /* real_32_type */
	T0* a63; /* real_32_convert_feature */
	T0* a64; /* real_64_type */
	T0* a65; /* real_64_convert_feature */
	T0* a66; /* routine_identity_type */
	T0* a67; /* special_any_type */
	T0* a68; /* special_detachable_any_type */
	T0* a69; /* special_identity_type */
	T0* a70; /* string_type */
	T0* a71; /* detachable_string_type */
	T0* a72; /* string_8_type */
	T0* a73; /* string_8_convert_feature */
	T0* a74; /* string_32_type */
	T0* a75; /* string_32_convert_feature */
	T0* a76; /* system_object_type */
	T0* a77; /* system_object_parents */
	T0* a78; /* system_string_type */
	T0* a79; /* type_detachable_any_type */
	T0* a80; /* detachable_type_exception_type */
	T0* a81; /* type_detachable_exception_type */
	T0* a82; /* type_detachable_like_current_type */
	T0* a83; /* type_identity_type */
	T0* a84; /* typed_pointer_identity_type */
	T0* a85; /* wide_character_type */
	T1 a86; /* is_read_only */
	T1 a87; /* is_preparsed */
};

/* Struct for type [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
struct S209 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* assembly_culture */
	T0* a7; /* assembly_key */
	T0* a8; /* assembly_name */
	T0* a9; /* assembly_version */
	T0* a10; /* classname_prefix */
	T1 a11; /* is_read_only */
	T0* a12; /* options */
	T0* a13; /* class_options */
	T0* a14; /* class_renamings */
	T0* a15; /* conditions */
	T0* a16; /* notes */
	T0* a17; /* name */
	T0* a18; /* dotnet_assembly */
};

/* Struct for type [detachable] ET_ECF_ERROR */
struct S210 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* system_config */
	T0* a4; /* position */
	T0* a5; /* code */
};

/* Struct for type [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
struct S211 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* hash_agent */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
struct S212 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T0* a5; /* key_storage */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_keys */
	T0* a8; /* hash_function */
	T6 a9; /* position */
	T6 a10; /* last_position */
	T6 a11; /* capacity */
	T6 a12; /* slots_position */
	T6 a13; /* free_slot */
	T0* a14; /* equality_tester */
	T6 a15; /* modulus */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* clashes */
	T0* a19; /* slots */
	T0* a20; /* special_key_routines */
};

/* Struct for type [detachable] TUPLE [[attached] STRING_8] */
struct S213 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
struct S214 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
struct S215 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
struct S217 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* hash_function */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE */
struct S219 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* attribute_count */
	T0* a2; /* queries */
	T1 a3; /* is_alive */
	T6 a4; /* id */
	T0* a5; /* base_type */
	T0* a6; /* attached_type */
	T0* a7; /* base_class */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* procedures_by_seed */
	T0* a11; /* procedures */
	T0* a12; /* queries_by_seed */
	T6 a13; /* hash_code */
	T0* a14; /* next_type */
	T1 a15; /* has_reference_attributes */
	T1 a16; /* has_generic_expanded_attributes */
	T0* a17; /* query_calls */
	T1 a18; /* has_static */
	T0* a19; /* procedure_calls */
	T0* a20; /* equality_expressions */
	T0* a21; /* object_equality_expressions */
};

/* Struct for type [detachable] ET_DYNAMIC_FEATURE_LIST */
struct S220 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
struct S221 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* special_key_routines */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T6 a15; /* found_position */
	T0* a16; /* internal_cursor */
	T6 a17; /* count */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
struct S222 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* modulus */
	T6 a11; /* clashes_previous_position */
	T0* a12; /* special_item_routines */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_key_routines */
	T0* a16; /* key_storage */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLASS] */
struct S223 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
struct S224 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [INTEGER_32] */
struct S225 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 z1;
};

/* Struct for type [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
struct S226 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_dynamic_system */
	T1 a3; /* catcall_error_mode */
	T1 a4; /* catcall_warning_mode */
	T1 a5; /* no_debug */
	T1 a6; /* no_assertion */
	T0* a7; /* alive_conforming_descendants_per_type */
};

/* Struct for type [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
struct S227 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type detachable ET_DYNAMIC_FEATURE */
struct S232 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_feature */
	T0* a2; /* dynamic_type_sets */
	T6 a3; /* id */
	T0* a4; /* target_type */
	T1 a5; /* is_generated */
	T0* a6; /* result_type_set */
	T8 a7; /* builtin_class_code */
	T8 a8; /* builtin_feature_code */
	T1 a9; /* is_regular */
	T0* a10; /* first_precursor */
	T1 a11; /* is_creation */
	T0* a12; /* other_precursors */
	T1 a13; /* is_built */
	T1 a14; /* is_address */
	T1 a15; /* is_static_generated */
};

/* Struct for type [detachable] ET_ACTUAL_PARAMETER_LIST */
struct S236 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type [detachable] ET_NESTED_TYPE_CONTEXT */
struct S237 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* root_context */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_TYPE */
struct S246 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T6 a3; /* index */
	T0* a4; /* implementation_class */
};

/* Struct for type [detachable] ET_EXTERNAL_FUNCTION */
struct S247 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T8 a10; /* builtin_class_code */
	T8 a11; /* builtin_feature_code */
	T0* a12; /* arguments */
	T0* a13; /* postconditions */
	T0* a14; /* assigner */
	T0* a15; /* declared_type */
	T0* a16; /* extended_name */
	T0* a17; /* frozen_keyword */
	T0* a18; /* language */
	T0* a19; /* obsolete_message */
	T0* a20; /* alias_clause */
	T0* a21; /* clients */
	T0* a22; /* is_keyword */
	T0* a23; /* end_keyword */
	T0* a24; /* semicolon */
	T0* a25; /* feature_clause */
	T0* a26; /* first_indexing */
	T0* a27; /* object_tests */
	T1 a28; /* validity_checked */
	T1 a29; /* has_validity_error */
	T6 a30; /* id */
	T6 a31; /* version */
	T0* a32; /* synonym */
};

/* Struct for type [detachable] ET_TUPLE_TYPE */
struct S248 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* actual_parameters */
	T0* a2; /* type_mark */
	T0* a3; /* named_base_class */
	T0* a4; /* tuple_keyword */
};

/* Struct for type detachable ET_FORMAL_ARGUMENT_LIST */
struct S249 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type [detachable] ARRAY [[attached] ET_TYPE] */
struct S250 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE] */
struct S251 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_FORMAL_ARGUMENT */
struct S252 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_QUERY_LIST */
struct S253 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* declared_count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_PROCEDURE_LIST */
struct S254 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* declared_count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_DYNAMIC_TYPE_SET_LIST */
struct S255 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] KL_STRING_OUTPUT_STREAM */
struct S256 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* string */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
struct S257 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S258 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
struct S259 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
struct S260 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* clashes */
	T0* a4; /* slots */
};

/* Struct for type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
struct S261 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T1 a3; /* is_never_void */
	T0* a4; /* dynamic_types */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
struct S262 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
struct S263 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* count */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
struct S264 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T6 a3; /* last_position */
	T6 a4; /* capacity */
	T6 a5; /* count */
	T6 a6; /* free_slot */
	T6 a7; /* slots_position */
	T6 a8; /* modulus */
	T6 a9; /* clashes_previous_position */
	T0* a10; /* key_equality_tester */
	T0* a11; /* equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
struct S265 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S266 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* count */
	T6 a3; /* position */
	T0* a4; /* item_storage */
	T6 a5; /* capacity */
	T6 a6; /* modulus */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S267 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
struct S268 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S269 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
struct S270 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
struct S271 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_CURRENT */
struct S273 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* compressed_position */
};

/* Struct for type [detachable] ET_ACTUAL_ARGUMENT_LIST */
struct S274 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_symbol */
	T0* a4; /* right_symbol */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
struct S275 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* qualified_name */
	T0* a3; /* arguments */
	T0* a4; /* parenthesis_call */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL_EXPRESSION */
struct S278 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T6 a2; /* index */
	T0* a3; /* qualified_name */
	T0* a4; /* arguments */
	T0* a5; /* parenthesis_call */
};

/* Struct for type [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
struct S279 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* arguments */
	T6 a3; /* index */
	T0* a4; /* parenthesis_call */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S280 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_MANIFEST_TUPLE */
struct S281 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_symbol */
	T0* a4; /* right_symbol */
	T6 a5; /* index */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S282 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
struct S283 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* count */
	T6 a3; /* last_position */
	T6 a4; /* free_slot */
	T6 a5; /* modulus */
	T6 a6; /* slots_position */
	T6 a7; /* clashes_previous_position */
	T0* a8; /* key_equality_tester */
	T0* a9; /* equality_tester */
	T0* a10; /* internal_cursor */
	T6 a11; /* found_position */
	T0* a12; /* item_storage */
	T0* a13; /* key_storage */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T6 a16; /* capacity */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ARRAY [[attached] INTEGER_32] */
struct S284 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
struct S285 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T0* a10; /* equality_tester */
	T0* a11; /* key_equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
struct S286 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* equality_tester */
	T0* a9; /* key_equality_tester */
	T6 a10; /* found_position */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S287 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T6 a7; /* capacity */
	T0* a8; /* equality_tester */
	T0* a9; /* key_equality_tester */
	T0* a10; /* internal_cursor */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
struct S288 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
struct S289 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T0* a4; /* key_storage */
	T0* a5; /* key_equality_tester */
	T0* a6; /* internal_keys */
	T6 a7; /* count */
	T6 a8; /* last_position */
	T6 a9; /* free_slot */
	T6 a10; /* capacity */
	T6 a11; /* slots_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* equality_tester */
	T6 a15; /* found_position */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
struct S290 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T0* a6; /* equality_tester */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T6 a9; /* clashes_previous_position */
	T0* a10; /* special_item_routines */
	T0* a11; /* item_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_IDENTIFIER_TESTER */
struct S291 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_TEXT_OUTPUT_FILE */
struct S293 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* mode */
	T14 a3; /* file_pointer */
	T1 a4; /* descriptor_available */
	T0* a5; /* last_string */
	T0* a6; /* internal_name */
	T0* a7; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] ET_DYNAMIC_TUPLE_TYPE */
struct S294 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* procedures_by_seed */
	T0* a11; /* procedures */
	T0* a12; /* queries_by_seed */
	T0* a13; /* item_type_sets */
	T1 a14; /* has_reference_attributes */
	T1 a15; /* has_generic_expanded_attributes */
	T6 a16; /* hash_code */
	T0* a17; /* next_type */
	T0* a18; /* query_calls */
	T1 a19; /* has_static */
	T0* a20; /* procedure_calls */
	T0* a21; /* equality_expressions */
	T0* a22; /* object_equality_expressions */
};

/* Struct for type [detachable] DS_STRING_HASH_TABLE */
struct S295 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T0* a5; /* key_equality_tester */
	T0* a6; /* slots */
	T0* a7; /* hash_function */
	T0* a8; /* key_storage */
	T0* a9; /* clashes */
	T6 a10; /* count */
	T6 a11; /* capacity */
	T6 a12; /* slots_position */
	T6 a13; /* free_slot */
	T6 a14; /* last_position */
	T0* a15; /* internal_keys */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* equality_tester */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
struct S296 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
struct S297 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type detachable ET_DYNAMIC_PRECURSOR */
struct S298 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_static_generated */
	T0* a2; /* target_type */
	T0* a3; /* static_feature */
	T0* a4; /* dynamic_type_sets */
	T0* a5; /* current_feature */
	T6 a6; /* id */
	T0* a7; /* result_type_set */
	T1 a8; /* is_regular */
	T1 a9; /* is_creation */
	T1 a10; /* is_address */
	T8 a11; /* builtin_class_code */
	T0* a12; /* first_precursor */
	T0* a13; /* other_precursors */
	T1 a14; /* is_built */
	T8 a15; /* builtin_feature_code */
	T1 a16; /* is_generated */
	T0* a17; /* parent_type */
};

/* Struct for type detachable ET_DYNAMIC_PRECURSOR_LIST */
struct S299 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S300 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sorted_items */
	T0* a2; /* cycle */
	T0* a3; /* indexes */
	T0* a4; /* items */
	T0* a5; /* counts */
	T0* a6; /* successors */
};

/* Struct for type [detachable] ET_RESULT */
struct S301 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* compressed_position */
};

/* Struct for type detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
struct S302 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_call */
	T0* a2; /* target_type_set */
	T0* a3; /* result_type_set */
	T0* a4; /* current_feature */
	T0* a5; /* current_type */
	T0* a6; /* next */
};

/* Struct for type detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
struct S306 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T0* a2; /* next */
	T0* a3; /* static_call */
	T0* a4; /* current_feature */
	T0* a5; /* current_type */
};

/* Struct for type [detachable] ET_OBJECT_TEST */
struct S308 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* attached_keyword */
	T0* a3; /* declared_type */
	T0* a4; /* expression */
};

/* Struct for type [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
struct S309 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* operator */
	T0* a4; /* right */
};

/* Struct for type [detachable] ET_EQUALITY_EXPRESSION */
struct S310 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* operator */
	T0* a4; /* right */
};

/* Struct for type [detachable] ET_DYNAMIC_SPECIAL_TYPE */
struct S311 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* procedures_by_seed */
	T0* a11; /* procedures */
	T0* a12; /* queries_by_seed */
	T0* a13; /* item_type_set */
	T1 a14; /* has_reference_attributes */
	T1 a15; /* has_generic_expanded_attributes */
	T6 a16; /* hash_code */
	T0* a17; /* next_type */
	T0* a18; /* query_calls */
	T1 a19; /* has_static */
	T0* a20; /* procedure_calls */
	T0* a21; /* equality_expressions */
	T0* a22; /* object_equality_expressions */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S314 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type detachable ET_DYNAMIC_SECONDARY_TYPE */
struct S315 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_type */
	T0* a2; /* meta_type */
	T0* a3; /* internal_base_type */
	T0* a4; /* type_mark */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
struct S316 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_CREATE_EXPRESSION */
struct S320 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* creation_call */
	T0* a3; /* creation_type */
	T0* a4; /* create_keyword */
	T0* a5; /* creation_region */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL */
struct S321 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* arguments */
	T0* a2; /* qualified_name */
};

/* Struct for type [detachable] ARRAY [[attached] STRING_8] */
struct S324 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] STRING_8] */
struct S325 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_DIRECTORY */
struct S327 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_input */
	T0* a2; /* last_entry */
	T6 a3; /* mode */
	T0* a4; /* name */
	T0* a5; /* entry_buffer */
	T1 a6; /* old_end_of_input */
	T14 a7; /* directory_pointer */
	T14 a8; /* last_entry_pointer */
	T0* a9; /* lastentry */
	T0* a10; /* internal_name */
	T0* a11; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S328 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_EXTENDED_ATTRIBUTE */
struct S329 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* compound */
	T0* a7; /* locals */
	T0* a8; /* rescue_clause */
	T0* a9; /* first_precursor */
	T0* a10; /* other_precursors */
	T0* a11; /* preconditions */
	T6 a12; /* hash_code */
	T0* a13; /* postconditions */
	T0* a14; /* assigner */
	T0* a15; /* extended_name */
	T0* a16; /* declared_type */
	T0* a17; /* frozen_keyword */
	T0* a18; /* obsolete_message */
	T0* a19; /* clients */
	T0* a20; /* end_keyword */
	T0* a21; /* semicolon */
	T0* a22; /* feature_clause */
	T0* a23; /* first_indexing */
	T0* a24; /* object_tests */
	T1 a25; /* validity_checked */
	T1 a26; /* has_validity_error */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type detachable ET_COMPOUND */
struct S330 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* keyword */
};

/* Struct for type detachable ET_LOCAL_VARIABLE_LIST */
struct S331 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* local_keyword */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
struct S335 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] ET_IMPLICIT_TYPE_MARK */
struct S342 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_expanded_mark */
	T1 a2; /* is_attached_mark */
	T1 a3; /* is_reference_mark */
	T1 a4; /* is_separate_mark */
	T1 a5; /* is_detachable_mark */
};

/* Struct for type [detachable] ET_SYMBOL */
struct S343 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE] */
struct S344 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S345 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S346 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S347 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S348 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] UT_INTEGER_FORMATTER */
struct S349 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] STD_FILES */
struct S352 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_output */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S354 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S356 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S357 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T1 z4;
};

/* Struct for type detachable ET_ACROSS_COMPONENT_LIST */
struct S361 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_CONSTANT_ATTRIBUTE */
struct S363 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* constant */
	T0* a5; /* implementation_feature */
	T0* a6; /* other_seeds */
	T0* a7; /* first_precursor */
	T0* a8; /* other_precursors */
	T6 a9; /* hash_code */
	T0* a10; /* assigner */
	T0* a11; /* extended_name */
	T0* a12; /* declared_type */
	T0* a13; /* frozen_keyword */
	T0* a14; /* clients */
	T0* a15; /* is_keyword */
	T0* a16; /* semicolon */
	T0* a17; /* feature_clause */
	T0* a18; /* first_indexing */
	T0* a19; /* object_tests */
	T1 a20; /* validity_checked */
	T1 a21; /* has_validity_error */
	T6 a22; /* id */
	T6 a23; /* version */
	T0* a24; /* synonym */
};

/* Struct for type [detachable] ET_UNIQUE_ATTRIBUTE */
struct S365 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constant */
	T6 a2; /* id */
	T6 a3; /* first_seed */
	T0* a4; /* across_components */
	T0* a5; /* implementation_class */
	T0* a6; /* other_seeds */
	T0* a7; /* implementation_feature */
	T0* a8; /* first_precursor */
	T0* a9; /* other_precursors */
	T6 a10; /* hash_code */
	T0* a11; /* assigner */
	T0* a12; /* extended_name */
	T0* a13; /* declared_type */
	T0* a14; /* frozen_keyword */
	T0* a15; /* clients */
	T0* a16; /* is_keyword */
	T0* a17; /* unique_keyword */
	T0* a18; /* semicolon */
	T0* a19; /* feature_clause */
	T0* a20; /* first_indexing */
	T0* a21; /* object_tests */
	T1 a22; /* validity_checked */
	T1 a23; /* has_validity_error */
	T6 a24; /* version */
	T0* a25; /* synonym */
};

/* Struct for type [detachable] ET_REGULAR_INTEGER_CONSTANT */
struct S366 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T11 a1; /* value */
	T0* a2; /* sign */
	T0* a3; /* literal */
	T1 a4; /* has_overflow */
	T6 a5; /* compressed_position */
	T6 a6; /* index */
	T0* a7; /* cast_type */
	T0* a8; /* type */
};

/* Struct for type [detachable] ET_ATTRIBUTE */
struct S367 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T6 a8; /* hash_code */
	T0* a9; /* assigner */
	T0* a10; /* extended_name */
	T0* a11; /* declared_type */
	T0* a12; /* frozen_keyword */
	T0* a13; /* clients */
	T0* a14; /* semicolon */
	T0* a15; /* feature_clause */
	T0* a16; /* first_indexing */
	T0* a17; /* object_tests */
	T1 a18; /* validity_checked */
	T1 a19; /* has_validity_error */
	T6 a20; /* id */
	T6 a21; /* version */
	T0* a22; /* synonym */
};

/* Struct for type [detachable] NATIVE_STRING */
struct S370 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* managed_data */
	T6 a2; /* unit_count */
};

/* Struct for type [detachable] ET_SYSTEM_MULTIPROCESSOR */
struct S371 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* interface_checker */
	T0* a2; /* stop_request */
	T0* a3; /* eiffel_parser */
	T0* a4; /* implementation_checker */
	T0* a5; /* error_handler */
	T1 a6; /* benchmark_shown */
	T0* a7; /* ise_version */
	T0* a8; /* other_processors */
	T0* a9; /* eiffel_preparser */
	T0* a10; /* master_class_checker */
	T0* a11; /* provider_checker */
	T0* a12; /* ancestor_builder */
	T0* a13; /* feature_flattener */
	T0* a14; /* ecma_version */
	T1 a15; /* nested_benchmark_shown */
	T1 a16; /* metrics_shown */
	T1 a17; /* flat_mode */
	T1 a18; /* flat_dbc_mode */
	T0* a19; /* processed_class_count_stack */
	T0* a20; /* ast_factory */
	T0* a21; /* dotnet_assembly_consumer */
	T1 a22; /* use_attribute_keyword */
	T1 a23; /* use_note_keyword */
	T1 a24; /* use_reference_keyword */
	T1 a25; /* use_attached_keyword */
	T1 a26; /* use_detachable_keyword */
	T1 a27; /* preparse_shallow_mode */
	T1 a28; /* preparse_single_mode */
	T1 a29; /* preparse_multiple_mode */
	T1 a30; /* unknown_builtin_reported */
	T1 a31; /* qualified_anchored_types_cycle_detection_enabled */
	T1 a32; /* cluster_dependence_enabled */
	T6 a33; /* postponed_class_count */
	T6 a34; /* processed_class_count */
	T1 a35; /* suppliers_enabled */
	T1 a36; /* preparse_override_mode */
	T1 a37; /* preparse_readonly_mode */
	T1 a38; /* providers_enabled */
	T1 a39; /* use_cluster_dependence_pathnames */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
struct S372 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
struct S373 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
struct S375 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_FEATURE_CHECKER */
struct S376 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_checker */
	T0* a2; /* current_type */
	T0* a3; /* current_class */
	T0* a4; /* current_feature */
	T0* a5; /* current_feature_impl */
	T0* a6; /* current_class_impl */
	T0* a7; /* enclosing_inline_agents */
	T0* a8; /* unused_overloaded_procedures_list */
	T0* a9; /* unused_overloaded_queries_list */
	T0* a10; /* unused_overloaded_features_list */
	T0* a11; /* unused_call_infos */
	T0* a12; /* unused_contexts */
	T0* a13; /* current_context */
	T0* a14; /* current_target_type */
	T0* a15; /* current_object_test_types */
	T0* a16; /* current_object_test_scope */
	T0* a17; /* object_test_scope_builder */
	T0* a18; /* current_across_cursor_types */
	T0* a19; /* current_across_cursor_scope */
	T0* a20; /* current_initialization_scope */
	T0* a21; /* current_attachment_scope */
	T0* a22; /* attachment_scope_builder */
	T0* a23; /* unused_attachment_scopes */
	T0* a24; /* common_ancestor_type_list */
	T0* a25; /* indexing_term_list */
	T0* a26; /* default_creation_call_name */
	T0* a27; /* default_creation_call */
	T0* a28; /* vape_non_descendant_clients */
	T0* a29; /* vape_creation_clients */
	T0* a30; /* vape_client */
	T0* a31; /* unused_adapted_base_classes */
	T0* a32; /* adapted_base_class_checker */
	T0* a33; /* system_processor */
	T1 a34; /* has_fatal_error */
	T1 a35; /* in_static_feature */
	T1 a36; /* in_precondition */
	T0* a37; /* supplier_handler */
	T1 a38; /* in_invariant */
	T0* a39; /* precursor_queries */
	T0* a40; /* precursor_procedures */
	T1 a41; /* in_postcondition */
	T1 a42; /* in_precursor */
	T1 a43; /* in_rescue */
	T0* a44; /* current_inline_agent */
	T1 a45; /* in_check_instruction */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
struct S377 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T6 a3; /* free_slot */
	T6 a4; /* position */
	T0* a5; /* equality_tester */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_cursor */
	T6 a8; /* found_position */
	T0* a9; /* item_storage */
	T0* a10; /* key_storage */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
struct S378 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* modulus */
	T6 a7; /* slots_position */
	T6 a8; /* clashes_previous_position */
	T0* a9; /* key_equality_tester */
	T0* a10; /* equality_tester */
	T6 a11; /* capacity */
	T6 a12; /* free_slot */
	T0* a13; /* special_item_routines */
	T0* a14; /* special_key_routines */
	T0* a15; /* key_storage */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* hash_function */
	T6 a19; /* count */
};

/* Struct for type [detachable] ET_TYPE_CHECKER */
struct S380 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_context */
	T0* a2; /* current_class_impl */
	T0* a3; /* current_feature_impl */
	T0* a4; /* constraint_context */
	T0* a5; /* current_class */
	T0* a6; /* target_context */
	T0* a7; /* other_context */
	T0* a8; /* adapted_base_class_checker */
	T0* a9; /* adapted_base_classes */
	T0* a10; /* system_processor */
	T0* a11; /* supplier_handler */
	T1 a12; /* has_fatal_error */
	T1 a13; /* class_interface_error_only */
	T1 a14; /* in_qualified_anchored_type */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
struct S382 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S383 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S384 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S385 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S386 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S387 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
struct S388 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T0* a7; /* equality_tester */
	T0* a8; /* key_equality_tester */
	T6 a9; /* found_position */
	T0* a10; /* key_storage */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_OBJECT_TEST_SCOPE */
struct S389 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* object_tests */
	T6 a2; /* hidden_count */
};

/* Struct for type [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
struct S390 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* scope */
	T1 a5; /* is_negated */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
struct S391 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T0* a7; /* equality_tester */
	T0* a8; /* key_equality_tester */
	T6 a9; /* found_position */
	T0* a10; /* key_storage */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* internal_keys */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_ACROSS_CURSOR_SCOPE */
struct S392 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* hidden_count */
	T0* a2; /* across_components */
};

/* Struct for type [detachable] ET_ATTACHMENT_SCOPE */
struct S393 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* locals_attached */
	T0* a2; /* arguments_attached */
	T0* a3; /* attributes_attached */
	T1 a4; /* result_attached */
	T1 a5; /* is_code_unreachable */
};

/* Struct for type [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
struct S394 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* scope */
	T1 a2; /* is_negated */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
struct S395 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
struct S396 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] ET_CLIENT_LIST */
struct S397 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_CLIENT */
struct S398 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* named_base_class */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S399 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
struct S400 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class_impl */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T1 a4; /* feature_flattening_error_only */
	T1 a5; /* has_fatal_error */
	T1 a6; /* class_interface_error_only */
};

/* Struct for type detachable ET_PRECONDITIONS */
struct S401 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* validity_checked */
	T1 a2; /* has_validity_error */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* require_keyword */
	T0* a6; /* else_keyword */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
struct S402 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T6 a11; /* found_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* hash_function */
};

/* Struct for type detachable ET_FEATURE_LIST */
struct S404 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
};

/* Struct for type [detachable] ET_UNKNOWN_GROUP */
struct S406 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T1 a2; /* use_obsolete_syntax */
	T0* a3; /* cached_absolute_pathname */
};

/* Struct for type [detachable] ET_BASE_TYPE_LIST */
struct S408 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_KEYWORD */
struct S409 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T0* a2; /* text */
	T6 a3; /* compressed_position */
};

/* Struct for type [detachable] ET_CLASS_CODES */
struct S410 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S411 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S412 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S413 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] UC_UTF8_ROUTINES */
struct S416 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PATH */
struct S417 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_name */
	T0* a2; /* storage */
	T1 a3; /* is_normalized */
};

/* Struct for type [detachable] RX_BYTE_CODE */
struct S419 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* byte_code */
	T0* a3; /* character_sets */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] RX_CHARACTER_SET */
struct S420 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_empty */
	T1 a2; /* is_negated */
	T11 a3; /* first_set */
	T11 a4; /* second_set */
	T11 a5; /* third_set */
	T11 a6; /* fourth_set */
	T0* a7; /* other_sets */
	T1 a8; /* is_reverted */
};

/* Struct for type [detachable] RX_CASE_MAPPING */
struct S421 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* lower_table */
	T0* a2; /* flip_table */
};

/* Struct for type [detachable] UC_UNICODE_ROUTINES */
struct S423 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
struct S424 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
struct S425 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] UC_STRING */
struct S426 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* byte_count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* internal_case_insensitive_hash_code */
	T0* a5; /* area */
	T6 a6; /* last_byte_index_input */
	T6 a7; /* last_byte_index_result */
};

/* Struct for type [detachable] KL_ANY_ROUTINES */
struct S427 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] STRING_TO_INTEGER_CONVERTOR */
struct S428 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* sign */
	T11 a2; /* part1 */
	T11 a3; /* part2 */
	T6 a4; /* last_state */
	T1 a5; /* internal_overflowed */
	T0* a6; /* leading_separators */
	T0* a7; /* trailing_separators */
	T1 a8; /* leading_separators_acceptable */
	T1 a9; /* trailing_separators_acceptable */
	T6 a10; /* conversion_type */
};

/* Struct for type [detachable] ARGUMENTS_32 */
struct S431 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_ECF_OPTIONS */
struct S437 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_assertions */
	T0* a2; /* primary_debugs */
	T0* a3; /* primary_warnings */
	T0* a4; /* primary_options */
	T0* a5; /* description */
	T0* a6; /* secondary_options */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
struct S438 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] XM_NAMESPACE */
struct S439 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* uri */
	T0* a2; /* ns_prefix */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
struct S440 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] ET_LIKE_CURRENT */
struct S442 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* current_keyword */
};

/* Struct for type [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
struct S443 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* next_type */
	T0* a11; /* procedures */
	T1 a12; /* has_static */
	T0* a13; /* query_calls */
	T0* a14; /* procedure_calls */
	T0* a15; /* open_operand_type_sets */
	T6 a16; /* hash_code */
	T0* a17; /* procedures_by_seed */
	T1 a18; /* has_reference_attributes */
	T0* a19; /* queries_by_seed */
	T0* a20; /* set_rout_disp_final_feature */
	T1 a21; /* has_generic_expanded_attributes */
	T0* a22; /* equality_expressions */
	T0* a23; /* object_equality_expressions */
};

/* Struct for type [detachable] ET_DYNAMIC_FUNCTION_TYPE */
struct S444 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* next_type */
	T0* a11; /* procedures */
	T1 a12; /* has_static */
	T0* a13; /* query_calls */
	T0* a14; /* procedure_calls */
	T0* a15; /* open_operand_type_sets */
	T0* a16; /* result_type_set */
	T6 a17; /* hash_code */
	T0* a18; /* procedures_by_seed */
	T1 a19; /* has_reference_attributes */
	T0* a20; /* queries_by_seed */
	T0* a21; /* set_rout_disp_final_feature */
	T1 a22; /* has_generic_expanded_attributes */
	T0* a23; /* equality_expressions */
	T0* a24; /* object_equality_expressions */
};

/* Struct for type [detachable] DT_SHARED_SYSTEM_CLOCK */
struct S445 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DT_SYSTEM_CLOCK */
struct S446 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* second */
	T6 a2; /* year */
	T6 a3; /* month */
	T6 a4; /* day */
	T6 a5; /* hour */
	T6 a6; /* minute */
	T6 a7; /* millisecond */
	T0* a8; /* local_clock */
};

/* Struct for type [detachable] ET_EIFFEL_PREPARSER */
struct S447 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* eiffel_buffer */
	T0* a2; /* system_processor */
	T6 a3; /* last_text_count */
	T6 a4; /* last_literal_start */
	T0* a5; /* filename */
	T0* a6; /* group */
	T0* a7; /* verbatim_marker */
	T6 a8; /* verbatim_marker_count */
	T0* a9; /* verbatim_open_white_characters */
	T0* a10; /* verbatim_close_white_characters */
	T0* a11; /* yy_pushed_start_conditions */
	T0* a12; /* input_buffer */
	T0* a13; /* yy_state_stack */
	T6 a14; /* yy_start_state */
	T6 a15; /* yy_line */
	T6 a16; /* yy_column */
	T6 a17; /* yy_position */
	T6 a18; /* yy_start */
	T6 a19; /* yy_end */
	T6 a20; /* line */
	T6 a21; /* column */
	T6 a22; /* position */
	T0* a23; /* yy_nxt */
	T0* a24; /* yy_chk */
	T0* a25; /* yy_base */
	T0* a26; /* yy_def */
	T0* a27; /* yy_ec */
	T0* a28; /* yy_meta */
	T0* a29; /* yy_accept */
	T0* a30; /* yy_acclist */
	T0* a31; /* yy_content */
	T0* a32; /* yy_content_area */
	T0* a33; /* yy_unicode_content_area */
	T0* a34; /* last_classname */
	T1 a35; /* class_keyword_found */
	T6 a36; /* last_token */
	T1 a37; /* yy_more_flag */
	T6 a38; /* yy_more_len */
	T6 a39; /* yy_state_count */
	T6 a40; /* yy_last_accepting_state */
	T6 a41; /* yy_last_accepting_cpos */
	T6 a42; /* yy_lp */
	T6 a43; /* yy_looking_for_trail_begin */
	T6 a44; /* yy_full_match */
	T6 a45; /* yy_full_state */
	T6 a46; /* yy_full_lp */
	T1 a47; /* yy_rejected */
	T6 a48; /* last_literal_end */
	T6 a49; /* pushed_start_condition_count */
	T6 a50; /* ms_column */
	T6 a51; /* ms_line */
	T6 a52; /* last_break_end */
	T6 a53; /* last_comment_end */
	T3 a54; /* verbatim_opening_character */
};

/* Struct for type [detachable] ET_MASTER_CLASS_CHECKER */
struct S448 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
	T1 a3; /* has_fatal_error */
};

/* Struct for type [detachable] ET_EIFFEL_PARSER */
struct S449 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* eiffel_buffer */
	T0* a2; /* counters */
	T0* a3; /* last_formal_arguments_stack */
	T0* a4; /* last_local_variables_stack */
	T0* a5; /* last_keywords */
	T0* a6; /* last_symbols */
	T0* a7; /* last_object_tests_stack */
	T0* a8; /* last_object_tests_pool */
	T0* a9; /* last_across_components_stack */
	T0* a10; /* last_across_components_pool */
	T0* a11; /* assertions */
	T0* a12; /* assertion_counters */
	T0* a13; /* assertion_kinds */
	T6 a14; /* assertion_kind */
	T0* a15; /* queries */
	T0* a16; /* procedures */
	T0* a17; /* constraints */
	T0* a18; /* providers */
	T0* a19; /* current_class */
	T0* a20; /* system_processor */
	T6 a21; /* last_text_count */
	T6 a22; /* last_literal_start */
	T0* a23; /* filename */
	T0* a24; /* group */
	T0* a25; /* verbatim_marker */
	T6 a26; /* verbatim_marker_count */
	T0* a27; /* verbatim_open_white_characters */
	T0* a28; /* verbatim_close_white_characters */
	T0* a29; /* yyss */
	T0* a30; /* yy_pushed_start_conditions */
	T0* a31; /* input_buffer */
	T0* a32; /* yyspecial_routines1 */
	T6 a33; /* yyvsc1 */
	T0* a34; /* yyvs1 */
	T0* a35; /* yyspecial_routines2 */
	T6 a36; /* yyvsc2 */
	T0* a37; /* yyvs2 */
	T0* a38; /* yyspecial_routines3 */
	T6 a39; /* yyvsc3 */
	T0* a40; /* yyvs3 */
	T0* a41; /* yyspecial_routines4 */
	T6 a42; /* yyvsc4 */
	T0* a43; /* yyvs4 */
	T0* a44; /* yyspecial_routines5 */
	T6 a45; /* yyvsc5 */
	T0* a46; /* yyvs5 */
	T0* a47; /* yyspecial_routines6 */
	T6 a48; /* yyvsc6 */
	T0* a49; /* yyvs6 */
	T0* a50; /* yyspecial_routines7 */
	T6 a51; /* yyvsc7 */
	T0* a52; /* yyvs7 */
	T0* a53; /* yyspecial_routines8 */
	T6 a54; /* yyvsc8 */
	T0* a55; /* yyvs8 */
	T0* a56; /* yyspecial_routines9 */
	T6 a57; /* yyvsc9 */
	T0* a58; /* yyvs9 */
	T0* a59; /* yyspecial_routines10 */
	T6 a60; /* yyvsc10 */
	T0* a61; /* yyvs10 */
	T0* a62; /* yyspecial_routines11 */
	T6 a63; /* yyvsc11 */
	T0* a64; /* yyvs11 */
	T0* a65; /* yyspecial_routines12 */
	T6 a66; /* yyvsc12 */
	T0* a67; /* yyvs12 */
	T0* a68; /* yyspecial_routines13 */
	T6 a69; /* yyvsc13 */
	T0* a70; /* yyvs13 */
	T0* a71; /* yyspecial_routines14 */
	T6 a72; /* yyvsc14 */
	T0* a73; /* yyvs14 */
	T0* a74; /* yyspecial_routines15 */
	T6 a75; /* yyvsc15 */
	T0* a76; /* yyvs15 */
	T0* a77; /* yyspecial_routines16 */
	T6 a78; /* yyvsc16 */
	T0* a79; /* yyvs16 */
	T0* a80; /* yyspecial_routines17 */
	T6 a81; /* yyvsc17 */
	T0* a82; /* yyvs17 */
	T0* a83; /* yyspecial_routines18 */
	T6 a84; /* yyvsc18 */
	T0* a85; /* yyvs18 */
	T0* a86; /* yyspecial_routines19 */
	T6 a87; /* yyvsc19 */
	T0* a88; /* yyvs19 */
	T0* a89; /* yyspecial_routines20 */
	T6 a90; /* yyvsc20 */
	T0* a91; /* yyvs20 */
	T0* a92; /* yyspecial_routines21 */
	T6 a93; /* yyvsc21 */
	T0* a94; /* yyvs21 */
	T0* a95; /* yyspecial_routines22 */
	T6 a96; /* yyvsc22 */
	T0* a97; /* yyvs22 */
	T0* a98; /* yyspecial_routines23 */
	T6 a99; /* yyvsc23 */
	T0* a100; /* yyvs23 */
	T0* a101; /* yyspecial_routines24 */
	T6 a102; /* yyvsc24 */
	T0* a103; /* yyvs24 */
	T0* a104; /* yyspecial_routines25 */
	T6 a105; /* yyvsc25 */
	T0* a106; /* yyvs25 */
	T0* a107; /* yyspecial_routines26 */
	T6 a108; /* yyvsc26 */
	T0* a109; /* yyvs26 */
	T0* a110; /* yyspecial_routines27 */
	T6 a111; /* yyvsc27 */
	T0* a112; /* yyvs27 */
	T0* a113; /* yyspecial_routines28 */
	T6 a114; /* yyvsc28 */
	T0* a115; /* yyvs28 */
	T0* a116; /* yyspecial_routines29 */
	T6 a117; /* yyvsc29 */
	T0* a118; /* yyvs29 */
	T0* a119; /* yyspecial_routines30 */
	T6 a120; /* yyvsc30 */
	T0* a121; /* yyvs30 */
	T0* a122; /* yyspecial_routines31 */
	T6 a123; /* yyvsc31 */
	T0* a124; /* yyvs31 */
	T0* a125; /* yyspecial_routines32 */
	T6 a126; /* yyvsc32 */
	T0* a127; /* yyvs32 */
	T0* a128; /* yyspecial_routines33 */
	T6 a129; /* yyvsc33 */
	T0* a130; /* yyvs33 */
	T0* a131; /* yyspecial_routines34 */
	T6 a132; /* yyvsc34 */
	T0* a133; /* yyvs34 */
	T0* a134; /* yyspecial_routines35 */
	T6 a135; /* yyvsc35 */
	T0* a136; /* yyvs35 */
	T0* a137; /* yyspecial_routines36 */
	T6 a138; /* yyvsc36 */
	T0* a139; /* yyvs36 */
	T0* a140; /* yyspecial_routines37 */
	T6 a141; /* yyvsc37 */
	T0* a142; /* yyvs37 */
	T0* a143; /* yyspecial_routines38 */
	T6 a144; /* yyvsc38 */
	T0* a145; /* yyvs38 */
	T0* a146; /* yyspecial_routines39 */
	T6 a147; /* yyvsc39 */
	T0* a148; /* yyvs39 */
	T0* a149; /* yyspecial_routines40 */
	T6 a150; /* yyvsc40 */
	T0* a151; /* yyvs40 */
	T0* a152; /* yyspecial_routines41 */
	T6 a153; /* yyvsc41 */
	T0* a154; /* yyvs41 */
	T0* a155; /* yyspecial_routines42 */
	T6 a156; /* yyvsc42 */
	T0* a157; /* yyvs42 */
	T0* a158; /* yyspecial_routines43 */
	T6 a159; /* yyvsc43 */
	T0* a160; /* yyvs43 */
	T0* a161; /* yyspecial_routines44 */
	T6 a162; /* yyvsc44 */
	T0* a163; /* yyvs44 */
	T0* a164; /* yyspecial_routines45 */
	T6 a165; /* yyvsc45 */
	T0* a166; /* yyvs45 */
	T0* a167; /* yyspecial_routines46 */
	T6 a168; /* yyvsc46 */
	T0* a169; /* yyvs46 */
	T0* a170; /* yyspecial_routines47 */
	T6 a171; /* yyvsc47 */
	T0* a172; /* yyvs47 */
	T0* a173; /* yyspecial_routines48 */
	T6 a174; /* yyvsc48 */
	T0* a175; /* yyvs48 */
	T0* a176; /* yyspecial_routines49 */
	T6 a177; /* yyvsc49 */
	T0* a178; /* yyvs49 */
	T0* a179; /* yyspecial_routines50 */
	T6 a180; /* yyvsc50 */
	T0* a181; /* yyvs50 */
	T0* a182; /* yyspecial_routines51 */
	T6 a183; /* yyvsc51 */
	T0* a184; /* yyvs51 */
	T0* a185; /* yyspecial_routines52 */
	T6 a186; /* yyvsc52 */
	T0* a187; /* yyvs52 */
	T0* a188; /* yyspecial_routines53 */
	T6 a189; /* yyvsc53 */
	T0* a190; /* yyvs53 */
	T0* a191; /* yyspecial_routines54 */
	T6 a192; /* yyvsc54 */
	T0* a193; /* yyvs54 */
	T0* a194; /* yyspecial_routines55 */
	T6 a195; /* yyvsc55 */
	T0* a196; /* yyvs55 */
	T0* a197; /* yyspecial_routines56 */
	T6 a198; /* yyvsc56 */
	T0* a199; /* yyvs56 */
	T0* a200; /* yyspecial_routines57 */
	T6 a201; /* yyvsc57 */
	T0* a202; /* yyvs57 */
	T0* a203; /* yyspecial_routines58 */
	T6 a204; /* yyvsc58 */
	T0* a205; /* yyvs58 */
	T0* a206; /* yyspecial_routines59 */
	T6 a207; /* yyvsc59 */
	T0* a208; /* yyvs59 */
	T0* a209; /* yyspecial_routines60 */
	T6 a210; /* yyvsc60 */
	T0* a211; /* yyvs60 */
	T0* a212; /* yyspecial_routines61 */
	T6 a213; /* yyvsc61 */
	T0* a214; /* yyvs61 */
	T0* a215; /* yyspecial_routines62 */
	T6 a216; /* yyvsc62 */
	T0* a217; /* yyvs62 */
	T0* a218; /* yyspecial_routines63 */
	T6 a219; /* yyvsc63 */
	T0* a220; /* yyvs63 */
	T0* a221; /* yyspecial_routines64 */
	T6 a222; /* yyvsc64 */
	T0* a223; /* yyvs64 */
	T0* a224; /* yyspecial_routines65 */
	T6 a225; /* yyvsc65 */
	T0* a226; /* yyvs65 */
	T0* a227; /* yyspecial_routines66 */
	T6 a228; /* yyvsc66 */
	T0* a229; /* yyvs66 */
	T0* a230; /* yyspecial_routines67 */
	T6 a231; /* yyvsc67 */
	T0* a232; /* yyvs67 */
	T0* a233; /* yyspecial_routines68 */
	T6 a234; /* yyvsc68 */
	T0* a235; /* yyvs68 */
	T0* a236; /* yyspecial_routines69 */
	T6 a237; /* yyvsc69 */
	T0* a238; /* yyvs69 */
	T0* a239; /* yyspecial_routines70 */
	T6 a240; /* yyvsc70 */
	T0* a241; /* yyvs70 */
	T0* a242; /* yyspecial_routines71 */
	T6 a243; /* yyvsc71 */
	T0* a244; /* yyvs71 */
	T0* a245; /* yyspecial_routines72 */
	T6 a246; /* yyvsc72 */
	T0* a247; /* yyvs72 */
	T0* a248; /* yyspecial_routines73 */
	T6 a249; /* yyvsc73 */
	T0* a250; /* yyvs73 */
	T0* a251; /* yyspecial_routines74 */
	T6 a252; /* yyvsc74 */
	T0* a253; /* yyvs74 */
	T0* a254; /* yyspecial_routines75 */
	T6 a255; /* yyvsc75 */
	T0* a256; /* yyvs75 */
	T0* a257; /* yyspecial_routines76 */
	T6 a258; /* yyvsc76 */
	T0* a259; /* yyvs76 */
	T0* a260; /* yyspecial_routines77 */
	T6 a261; /* yyvsc77 */
	T0* a262; /* yyvs77 */
	T0* a263; /* yyspecial_routines78 */
	T6 a264; /* yyvsc78 */
	T0* a265; /* yyvs78 */
	T0* a266; /* yyspecial_routines79 */
	T6 a267; /* yyvsc79 */
	T0* a268; /* yyvs79 */
	T0* a269; /* yyspecial_routines80 */
	T6 a270; /* yyvsc80 */
	T0* a271; /* yyvs80 */
	T0* a272; /* yyspecial_routines81 */
	T6 a273; /* yyvsc81 */
	T0* a274; /* yyvs81 */
	T0* a275; /* yyspecial_routines82 */
	T6 a276; /* yyvsc82 */
	T0* a277; /* yyvs82 */
	T0* a278; /* yyspecial_routines83 */
	T6 a279; /* yyvsc83 */
	T0* a280; /* yyvs83 */
	T0* a281; /* yyspecial_routines84 */
	T6 a282; /* yyvsc84 */
	T0* a283; /* yyvs84 */
	T0* a284; /* yyspecial_routines85 */
	T6 a285; /* yyvsc85 */
	T0* a286; /* yyvs85 */
	T0* a287; /* yyspecial_routines86 */
	T6 a288; /* yyvsc86 */
	T0* a289; /* yyvs86 */
	T0* a290; /* yyspecial_routines87 */
	T6 a291; /* yyvsc87 */
	T0* a292; /* yyvs87 */
	T0* a293; /* yyspecial_routines88 */
	T6 a294; /* yyvsc88 */
	T0* a295; /* yyvs88 */
	T0* a296; /* yyspecial_routines89 */
	T6 a297; /* yyvsc89 */
	T0* a298; /* yyvs89 */
	T0* a299; /* yyspecial_routines90 */
	T6 a300; /* yyvsc90 */
	T0* a301; /* yyvs90 */
	T0* a302; /* yyspecial_routines91 */
	T6 a303; /* yyvsc91 */
	T0* a304; /* yyvs91 */
	T0* a305; /* yyspecial_routines92 */
	T6 a306; /* yyvsc92 */
	T0* a307; /* yyvs92 */
	T0* a308; /* yyspecial_routines93 */
	T6 a309; /* yyvsc93 */
	T0* a310; /* yyvs93 */
	T0* a311; /* yyspecial_routines94 */
	T6 a312; /* yyvsc94 */
	T0* a313; /* yyvs94 */
	T0* a314; /* yyspecial_routines95 */
	T6 a315; /* yyvsc95 */
	T0* a316; /* yyvs95 */
	T0* a317; /* yyspecial_routines96 */
	T6 a318; /* yyvsc96 */
	T0* a319; /* yyvs96 */
	T0* a320; /* yyspecial_routines97 */
	T6 a321; /* yyvsc97 */
	T0* a322; /* yyvs97 */
	T0* a323; /* yyspecial_routines98 */
	T6 a324; /* yyvsc98 */
	T0* a325; /* yyvs98 */
	T0* a326; /* yyspecial_routines99 */
	T6 a327; /* yyvsc99 */
	T0* a328; /* yyvs99 */
	T0* a329; /* yyspecial_routines100 */
	T6 a330; /* yyvsc100 */
	T0* a331; /* yyvs100 */
	T0* a332; /* yyspecial_routines101 */
	T6 a333; /* yyvsc101 */
	T0* a334; /* yyvs101 */
	T0* a335; /* yyspecial_routines102 */
	T6 a336; /* yyvsc102 */
	T0* a337; /* yyvs102 */
	T0* a338; /* yyspecial_routines103 */
	T6 a339; /* yyvsc103 */
	T0* a340; /* yyvs103 */
	T0* a341; /* yyspecial_routines104 */
	T6 a342; /* yyvsc104 */
	T0* a343; /* yyvs104 */
	T0* a344; /* yyspecial_routines105 */
	T6 a345; /* yyvsc105 */
	T0* a346; /* yyvs105 */
	T0* a347; /* yyspecial_routines106 */
	T6 a348; /* yyvsc106 */
	T0* a349; /* yyvs106 */
	T0* a350; /* yyspecial_routines107 */
	T6 a351; /* yyvsc107 */
	T0* a352; /* yyvs107 */
	T0* a353; /* yyspecial_routines108 */
	T6 a354; /* yyvsc108 */
	T0* a355; /* yyvs108 */
	T0* a356; /* yyspecial_routines109 */
	T6 a357; /* yyvsc109 */
	T0* a358; /* yyvs109 */
	T0* a359; /* yyspecial_routines110 */
	T6 a360; /* yyvsc110 */
	T0* a361; /* yyvs110 */
	T0* a362; /* yyspecial_routines111 */
	T6 a363; /* yyvsc111 */
	T0* a364; /* yyvs111 */
	T0* a365; /* yyspecial_routines112 */
	T6 a366; /* yyvsc112 */
	T0* a367; /* yyvs112 */
	T0* a368; /* yyspecial_routines113 */
	T6 a369; /* yyvsc113 */
	T0* a370; /* yyvs113 */
	T0* a371; /* yyspecial_routines114 */
	T6 a372; /* yyvsc114 */
	T0* a373; /* yyvs114 */
	T0* a374; /* yyspecial_routines115 */
	T6 a375; /* yyvsc115 */
	T0* a376; /* yyvs115 */
	T0* a377; /* yyspecial_routines116 */
	T6 a378; /* yyvsc116 */
	T0* a379; /* yyvs116 */
	T0* a380; /* yyspecial_routines117 */
	T6 a381; /* yyvsc117 */
	T0* a382; /* yyvs117 */
	T0* a383; /* yyspecial_routines118 */
	T6 a384; /* yyvsc118 */
	T0* a385; /* yyvs118 */
	T0* a386; /* yyspecial_routines119 */
	T6 a387; /* yyvsc119 */
	T0* a388; /* yyvs119 */
	T0* a389; /* yyspecial_routines120 */
	T6 a390; /* yyvsc120 */
	T0* a391; /* yyvs120 */
	T0* a392; /* yyspecial_routines121 */
	T6 a393; /* yyvsc121 */
	T0* a394; /* yyvs121 */
	T0* a395; /* yyspecial_routines122 */
	T6 a396; /* yyvsc122 */
	T0* a397; /* yyvs122 */
	T0* a398; /* yyspecial_routines123 */
	T6 a399; /* yyvsc123 */
	T0* a400; /* yyvs123 */
	T0* a401; /* yyspecial_routines124 */
	T6 a402; /* yyvsc124 */
	T0* a403; /* yyvs124 */
	T0* a404; /* yyspecial_routines125 */
	T6 a405; /* yyvsc125 */
	T0* a406; /* yyvs125 */
	T0* a407; /* yyspecial_routines126 */
	T6 a408; /* yyvsc126 */
	T0* a409; /* yyvs126 */
	T0* a410; /* yyspecial_routines127 */
	T6 a411; /* yyvsc127 */
	T0* a412; /* yyvs127 */
	T0* a413; /* yyspecial_routines128 */
	T6 a414; /* yyvsc128 */
	T0* a415; /* yyvs128 */
	T0* a416; /* yytranslate */
	T0* a417; /* yyr1 */
	T0* a418; /* yytypes1 */
	T0* a419; /* yytypes2 */
	T0* a420; /* yydefact */
	T0* a421; /* yydefgoto */
	T0* a422; /* yypact */
	T0* a423; /* yypgoto */
	T0* a424; /* yytable */
	T0* a425; /* yycheck */
	T0* a426; /* yy_state_stack */
	T6 a427; /* yy_start_state */
	T6 a428; /* yy_line */
	T6 a429; /* yy_column */
	T6 a430; /* yy_position */
	T6 a431; /* yy_start */
	T6 a432; /* yy_end */
	T6 a433; /* line */
	T6 a434; /* column */
	T6 a435; /* position */
	T0* a436; /* yy_nxt */
	T0* a437; /* yy_chk */
	T0* a438; /* yy_base */
	T0* a439; /* yy_def */
	T0* a440; /* yy_ec */
	T0* a441; /* yy_meta */
	T0* a442; /* yy_accept */
	T0* a443; /* yy_acclist */
	T0* a444; /* yy_content */
	T0* a445; /* yy_content_area */
	T0* a446; /* yy_unicode_content_area */
	T6 a447; /* time_stamp */
	T0* a448; /* last_class */
	T6 a449; /* yy_parsing_status */
	T6 a450; /* yy_suspended_yystacksize */
	T6 a451; /* yy_suspended_yystate */
	T6 a452; /* yy_suspended_yyn */
	T6 a453; /* yy_suspended_yychar1 */
	T6 a454; /* yy_suspended_index */
	T6 a455; /* yy_suspended_yyss_top */
	T6 a456; /* yy_suspended_yy_goto */
	T6 a457; /* yyssp */
	T6 a458; /* error_count */
	T1 a459; /* yy_lookahead_needed */
	T6 a460; /* yyerrstatus */
	T6 a461; /* last_token */
	T0* a462; /* last_clients */
	T0* a463; /* last_export_clients */
	T0* a464; /* last_feature_clause */
	T6 a465; /* last_literal_end */
	T1 a466; /* yy_rejected */
	T6 a467; /* yy_state_count */
	T6 a468; /* yy_full_match */
	T6 a469; /* yy_lp */
	T6 a470; /* yy_looking_for_trail_begin */
	T6 a471; /* yy_full_lp */
	T6 a472; /* yy_full_state */
	T6 a473; /* pushed_start_condition_count */
	T1 a474; /* yy_more_flag */
	T6 a475; /* yy_more_len */
	T6 a476; /* yy_last_accepting_state */
	T6 a477; /* yy_last_accepting_cpos */
	T6 a478; /* yyvsp1 */
	T6 a479; /* yyvsp2 */
	T6 a480; /* yyvsp3 */
	T6 a481; /* yyvsp4 */
	T6 a482; /* yyvsp5 */
	T6 a483; /* yyvsp6 */
	T6 a484; /* yyvsp7 */
	T6 a485; /* yyvsp8 */
	T6 a486; /* yyvsp9 */
	T6 a487; /* yyvsp10 */
	T6 a488; /* yyvsp11 */
	T6 a489; /* yyvsp12 */
	T6 a490; /* yyvsp13 */
	T6 a491; /* yyvsp14 */
	T6 a492; /* yyvsp15 */
	T6 a493; /* yyvsp16 */
	T6 a494; /* yyvsp17 */
	T6 a495; /* yyvsp18 */
	T6 a496; /* yyvsp19 */
	T6 a497; /* yyvsp20 */
	T6 a498; /* yyvsp21 */
	T6 a499; /* yyvsp22 */
	T6 a500; /* yyvsp23 */
	T6 a501; /* yyvsp24 */
	T6 a502; /* yyvsp25 */
	T6 a503; /* yyvsp26 */
	T6 a504; /* yyvsp27 */
	T6 a505; /* yyvsp28 */
	T6 a506; /* yyvsp29 */
	T6 a507; /* yyvsp30 */
	T6 a508; /* yyvsp31 */
	T6 a509; /* yyvsp32 */
	T6 a510; /* yyvsp33 */
	T6 a511; /* yyvsp34 */
	T6 a512; /* yyvsp35 */
	T6 a513; /* yyvsp36 */
	T6 a514; /* yyvsp37 */
	T6 a515; /* yyvsp38 */
	T6 a516; /* yyvsp39 */
	T6 a517; /* yyvsp40 */
	T6 a518; /* yyvsp41 */
	T6 a519; /* yyvsp42 */
	T6 a520; /* yyvsp43 */
	T6 a521; /* yyvsp44 */
	T6 a522; /* yyvsp45 */
	T6 a523; /* yyvsp46 */
	T6 a524; /* yyvsp47 */
	T6 a525; /* yyvsp48 */
	T6 a526; /* yyvsp49 */
	T6 a527; /* yyvsp50 */
	T6 a528; /* yyvsp51 */
	T6 a529; /* yyvsp52 */
	T6 a530; /* yyvsp53 */
	T6 a531; /* yyvsp54 */
	T6 a532; /* yyvsp55 */
	T6 a533; /* yyvsp56 */
	T6 a534; /* yyvsp57 */
	T6 a535; /* yyvsp58 */
	T6 a536; /* yyvsp59 */
	T6 a537; /* yyvsp60 */
	T6 a538; /* yyvsp61 */
	T6 a539; /* yyvsp62 */
	T6 a540; /* yyvsp63 */
	T6 a541; /* yyvsp64 */
	T6 a542; /* yyvsp65 */
	T6 a543; /* yyvsp66 */
	T6 a544; /* yyvsp67 */
	T6 a545; /* yyvsp68 */
	T6 a546; /* yyvsp69 */
	T6 a547; /* yyvsp70 */
	T6 a548; /* yyvsp71 */
	T6 a549; /* yyvsp72 */
	T6 a550; /* yyvsp73 */
	T6 a551; /* yyvsp74 */
	T6 a552; /* yyvsp75 */
	T6 a553; /* yyvsp76 */
	T6 a554; /* yyvsp77 */
	T6 a555; /* yyvsp78 */
	T6 a556; /* yyvsp79 */
	T6 a557; /* yyvsp80 */
	T6 a558; /* yyvsp81 */
	T6 a559; /* yyvsp82 */
	T6 a560; /* yyvsp83 */
	T6 a561; /* yyvsp84 */
	T6 a562; /* yyvsp85 */
	T6 a563; /* yyvsp86 */
	T6 a564; /* yyvsp87 */
	T6 a565; /* yyvsp88 */
	T6 a566; /* yyvsp89 */
	T6 a567; /* yyvsp90 */
	T6 a568; /* yyvsp91 */
	T6 a569; /* yyvsp92 */
	T6 a570; /* yyvsp93 */
	T6 a571; /* yyvsp94 */
	T6 a572; /* yyvsp95 */
	T6 a573; /* yyvsp96 */
	T6 a574; /* yyvsp97 */
	T6 a575; /* yyvsp98 */
	T6 a576; /* yyvsp99 */
	T6 a577; /* yyvsp100 */
	T6 a578; /* yyvsp101 */
	T6 a579; /* yyvsp102 */
	T6 a580; /* yyvsp103 */
	T6 a581; /* yyvsp104 */
	T6 a582; /* yyvsp105 */
	T6 a583; /* yyvsp106 */
	T6 a584; /* yyvsp107 */
	T6 a585; /* yyvsp108 */
	T6 a586; /* yyvsp109 */
	T6 a587; /* yyvsp110 */
	T6 a588; /* yyvsp111 */
	T6 a589; /* yyvsp112 */
	T6 a590; /* yyvsp113 */
	T6 a591; /* yyvsp114 */
	T6 a592; /* yyvsp115 */
	T6 a593; /* yyvsp116 */
	T6 a594; /* yyvsp117 */
	T6 a595; /* yyvsp118 */
	T6 a596; /* yyvsp119 */
	T6 a597; /* yyvsp120 */
	T6 a598; /* yyvsp121 */
	T6 a599; /* yyvsp122 */
	T6 a600; /* yyvsp123 */
	T6 a601; /* yyvsp124 */
	T6 a602; /* yyvsp125 */
	T6 a603; /* yyvsp126 */
	T6 a604; /* yyvsp127 */
	T6 a605; /* yyvsp128 */
	T0* a606; /* last_detachable_any_value */
	T0* a607; /* last_detachable_et_keyword_value */
	T0* a608; /* last_detachable_et_agent_keyword_value */
	T0* a609; /* last_detachable_et_precursor_keyword_value */
	T0* a610; /* last_detachable_et_symbol_value */
	T0* a611; /* last_detachable_et_position_value */
	T0* a612; /* last_detachable_et_boolean_constant_value */
	T0* a613; /* last_detachable_et_break_value */
	T0* a614; /* last_detachable_et_character_constant_value */
	T0* a615; /* last_detachable_et_current_value */
	T0* a616; /* last_detachable_et_free_operator_value */
	T0* a617; /* last_detachable_et_identifier_value */
	T0* a618; /* last_detachable_et_integer_constant_value */
	T0* a619; /* last_detachable_et_keyword_operator_value */
	T0* a620; /* last_detachable_et_manifest_string_value */
	T0* a621; /* last_detachable_et_real_constant_value */
	T0* a622; /* last_detachable_et_result_value */
	T0* a623; /* last_detachable_et_retry_instruction_value */
	T0* a624; /* last_detachable_et_symbol_operator_value */
	T0* a625; /* last_detachable_et_void_value */
	T0* a626; /* last_detachable_et_semicolon_symbol_value */
	T0* a627; /* last_detachable_et_bracket_symbol_value */
	T0* a628; /* last_detachable_et_question_mark_symbol_value */
	T0* a629; /* last_formal_arguments */
	T0* a630; /* last_local_variables */
	T0* a631; /* last_object_tests */
	T0* a632; /* last_across_components */
	T6 a633; /* last_break_end */
	T6 a634; /* last_comment_end */
	T6 a635; /* ms_column */
	T6 a636; /* ms_line */
	T6 a637; /* break_kind */
	T3 a638; /* verbatim_opening_character */
	T1 a639; /* has_natural_64_overflow */
	T11 a640; /* last_natural_64 */
};

/* Struct for type [detachable] ET_PROVIDER_CHECKER */
struct S450 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
};

/* Struct for type [detachable] ET_ANCESTOR_BUILDER */
struct S451 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_sorter */
	T0* a2; /* ancestors */
	T0* a3; /* conforming_ancestors */
	T0* a4; /* parent_checker */
	T0* a5; /* formal_parameter_checker */
	T0* a6; /* parent_context */
	T0* a7; /* current_class */
	T0* a8; /* system_processor */
};

/* Struct for type [detachable] ET_FEATURE_FLATTENER */
struct S452 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* named_features */
	T0* a2; /* queries */
	T0* a3; /* procedures */
	T0* a4; /* aliased_features */
	T0* a5; /* clients_list */
	T0* a6; /* client_classes */
	T0* a7; /* feature_adaptation_resolver */
	T0* a8; /* dotnet_feature_adaptation_resolver */
	T0* a9; /* identifier_type_resolver */
	T0* a10; /* unfolded_tuple_actual_parameters_resolver */
	T0* a11; /* anchored_type_checker */
	T0* a12; /* signature_checker */
	T0* a13; /* parent_checker */
	T0* a14; /* formal_parameter_checker */
	T0* a15; /* builtin_feature_checker */
	T0* a16; /* precursor_checker */
	T0* a17; /* precursors */
	T0* a18; /* current_class */
	T0* a19; /* system_processor */
	T1 a20; /* has_signature_error */
};

/* Struct for type [detachable] ET_INTERFACE_CHECKER */
struct S453 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* classes_to_be_processed */
	T0* a2; /* qualified_anchored_type_checker */
	T0* a3; /* unfolded_tuple_actual_parameters_resolver */
	T0* a4; /* old_name_rename_table */
	T0* a5; /* new_name_rename_table */
	T0* a6; /* current_class */
	T0* a7; /* system_processor */
};

/* Struct for type [detachable] ET_IMPLEMENTATION_CHECKER */
struct S454 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
	T0* a3; /* feature_checker */
	T0* a4; /* named_features */
	T0* a5; /* feature_adaptation_resolver */
	T0* a6; /* dotnet_feature_adaptation_resolver */
	T0* a7; /* signature_checker */
	T0* a8; /* parent_checker3 */
	T0* a9; /* precursor_procedures */
	T0* a10; /* precursor_queries */
	T0* a11; /* supplier_builder */
	T0* a12; /* no_suppliers */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
struct S455 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
struct S456 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_AST_NULL_PROCESSOR */
struct S457 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_AST_FACTORY */
struct S458 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
struct S460 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* system_processor */
};

/* Struct for type [detachable] DT_DATE_TIME_DURATION */
struct S461 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* year */
	T6 a2; /* month */
	T6 a3; /* day */
	T6 a4; /* hour */
	T6 a5; /* minute */
	T6 a6; /* second */
	T6 a7; /* millisecond */
};

/* Struct for type [detachable] UT_CONFIG_PARSER */
struct S462 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* config_values */
	T0* a3; /* error_handler */
	T0* a4; /* defined_values */
	T0* a5; /* include_stack */
	T0* a6; /* line_nb_stack */
	T0* a7; /* last_string_value */
	T6 a8; /* line_nb */
	T0* a9; /* yyss */
	T0* a10; /* input_buffer */
	T6 a11; /* yy_end */
	T6 a12; /* yy_position */
	T6 a13; /* yy_line */
	T6 a14; /* yy_column */
	T6 a15; /* yy_parsing_status */
	T6 a16; /* yy_suspended_yystacksize */
	T6 a17; /* yy_suspended_yystate */
	T6 a18; /* yy_suspended_yyn */
	T6 a19; /* yy_suspended_yychar1 */
	T6 a20; /* yy_suspended_index */
	T6 a21; /* yy_suspended_yyss_top */
	T6 a22; /* yy_suspended_yy_goto */
	T0* a23; /* yyr1 */
	T6 a24; /* yyssp */
	T0* a25; /* yypgoto */
	T0* a26; /* yycheck */
	T0* a27; /* yytable */
	T0* a28; /* yydefgoto */
	T6 a29; /* error_count */
	T1 a30; /* yy_lookahead_needed */
	T6 a31; /* yyerrstatus */
	T0* a32; /* yypact */
	T6 a33; /* last_token */
	T0* a34; /* yytranslate */
	T0* a35; /* yydefact */
	T0* a36; /* yy_pushed_start_conditions */
	T0* a37; /* yyspecial_routines1 */
	T6 a38; /* yyvsc1 */
	T0* a39; /* yyvs1 */
	T0* a40; /* yyspecial_routines2 */
	T6 a41; /* yyvsc2 */
	T0* a42; /* yyvs2 */
	T0* a43; /* yyspecial_routines3 */
	T6 a44; /* yyvsc3 */
	T0* a45; /* yyvs3 */
	T0* a46; /* yytypes1 */
	T0* a47; /* yytypes2 */
	T6 a48; /* yy_start */
	T6 a49; /* yyvsp1 */
	T6 a50; /* yyvsp2 */
	T6 a51; /* yyvsp3 */
	T1 a52; /* yy_more_flag */
	T6 a53; /* yy_more_len */
	T6 a54; /* line */
	T6 a55; /* column */
	T6 a56; /* position */
	T6 a57; /* yy_start_state */
	T0* a58; /* yy_state_stack */
	T6 a59; /* yy_state_count */
	T0* a60; /* yy_content_area */
	T0* a61; /* yy_unicode_content_area */
	T0* a62; /* yy_accept */
	T6 a63; /* yy_last_accepting_state */
	T6 a64; /* yy_last_accepting_cpos */
	T0* a65; /* yy_content */
	T0* a66; /* yy_ec */
	T0* a67; /* yy_chk */
	T0* a68; /* yy_base */
	T0* a69; /* yy_def */
	T0* a70; /* yy_meta */
	T0* a71; /* yy_nxt */
	T6 a72; /* yy_lp */
	T0* a73; /* yy_acclist */
	T6 a74; /* yy_looking_for_trail_begin */
	T6 a75; /* yy_full_match */
	T6 a76; /* yy_full_state */
	T6 a77; /* yy_full_lp */
	T1 a78; /* yy_rejected */
	T0* a79; /* last_detachable_any_value */
	T6 a80; /* if_level */
	T6 a81; /* ignored_level */
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
struct S463 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_DO_PROCEDURE */
struct S466 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* across_components */
	T0* a2; /* implementation_class */
	T6 a3; /* first_seed */
	T0* a4; /* other_seeds */
	T0* a5; /* arguments */
	T0* a6; /* compound */
	T0* a7; /* implementation_feature */
	T0* a8; /* extended_name */
	T6 a9; /* hash_code */
	T0* a10; /* end_keyword */
	T0* a11; /* clients */
	T1 a12; /* validity_checked */
	T1 a13; /* has_validity_error */
	T0* a14; /* preconditions */
	T0* a15; /* first_precursor */
	T0* a16; /* other_precursors */
	T0* a17; /* postconditions */
	T0* a18; /* rescue_clause */
	T0* a19; /* locals */
	T0* a20; /* object_tests */
	T0* a21; /* frozen_keyword */
	T0* a22; /* obsolete_message */
	T0* a23; /* is_keyword */
	T0* a24; /* semicolon */
	T0* a25; /* feature_clause */
	T0* a26; /* first_indexing */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type [detachable] ET_DEFERRED_PROCEDURE */
struct S468 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implementation_feature */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T6 a4; /* first_seed */
	T0* a5; /* other_seeds */
	T0* a6; /* arguments */
	T0* a7; /* extended_name */
	T0* a8; /* deferred_keyword */
	T6 a9; /* hash_code */
	T0* a10; /* end_keyword */
	T0* a11; /* clients */
	T1 a12; /* validity_checked */
	T1 a13; /* has_validity_error */
	T0* a14; /* preconditions */
	T0* a15; /* first_precursor */
	T0* a16; /* other_precursors */
	T0* a17; /* postconditions */
	T0* a18; /* object_tests */
	T0* a19; /* frozen_keyword */
	T0* a20; /* obsolete_message */
	T0* a21; /* is_keyword */
	T0* a22; /* semicolon */
	T0* a23; /* feature_clause */
	T0* a24; /* first_indexing */
	T6 a25; /* id */
	T6 a26; /* version */
	T0* a27; /* synonym */
};

/* Struct for type [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
struct S469 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T1 a3; /* is_never_void */
	T0* a4; /* dynamic_types */
	T0* a5; /* targets */
	T1 a6; /* is_dynamic_types_readonly */
};

/* Struct for type [detachable] KL_PLATFORM */
struct S470 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] CELL [[attached] INTEGER_32] */
struct S472 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
};

/* Struct for type [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
struct S473 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T0* a2; /* content */
	T0* a3; /* keys */
	T0* a4; /* deleted_marks */
	T0* a5; /* indexes_map */
	T6 a6; /* iteration_position */
	T6 a7; /* count */
	T6 a8; /* deleted_item_position */
	T6 a9; /* control */
	T0* a10; /* found_item */
	T1 a11; /* has_default */
	T6 a12; /* item_position */
	T6 a13; /* ht_lowest_deleted_position */
	T0* a14; /* ht_deleted_item */
	T0* a15; /* ht_deleted_key */
	T1 a16; /* object_comparison */
};

/* Struct for type [detachable] KL_PATHNAME */
struct S475 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* drive */
	T0* a3; /* hostname */
	T0* a4; /* sharename */
	T1 a5; /* is_relative */
	T0* a6; /* components */
};

/* Struct for type [detachable] EXCEPTIONS */
struct S476 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_8] */
struct S477 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T8 z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
struct S479 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] AP_OPTION] */
struct S481 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
struct S482 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S483 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S484 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S485 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
struct S486 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [detachable STRING_8] */
struct S487 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
struct S488 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
struct S489 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* next_cursor */
	T0* a4; /* container */
	T1 a5; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] STRING_8] */
struct S490 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type detachable DS_LINKABLE [[attached] BOOLEAN] */
struct S491 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
struct S492 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] INTEGER_32] */
struct S493 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
struct S494 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
struct S496 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
struct S497 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
struct S500 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
struct S501 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
struct S504 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
struct S505 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_SCANNER */
struct S506 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_token */
	T0* a2; /* last_value */
	T0* a3; /* input_name */
	T6 a4; /* position */
	T6 a5; /* column */
	T6 a6; /* line */
	T0* a7; /* input_filter */
	T6 a8; /* yy_start_state */
	T0* a9; /* character_entity */
	T0* a10; /* input_stream */
	T0* a11; /* input_resolver */
	T1 a12; /* yy_more_flag */
	T6 a13; /* yy_more_len */
	T6 a14; /* yy_end */
	T6 a15; /* yy_start */
	T6 a16; /* yy_line */
	T6 a17; /* yy_column */
	T6 a18; /* yy_position */
	T0* a19; /* yy_state_stack */
	T6 a20; /* yy_state_count */
	T0* a21; /* yy_content_area */
	T0* a22; /* yy_unicode_content_area */
	T0* a23; /* yy_accept */
	T6 a24; /* yy_last_accepting_state */
	T6 a25; /* yy_last_accepting_cpos */
	T0* a26; /* yy_content */
	T0* a27; /* yy_ec */
	T0* a28; /* yy_chk */
	T0* a29; /* yy_base */
	T0* a30; /* yy_def */
	T0* a31; /* yy_meta */
	T0* a32; /* yy_nxt */
	T6 a33; /* yy_lp */
	T0* a34; /* yy_acclist */
	T6 a35; /* yy_looking_for_trail_begin */
	T6 a36; /* yy_full_match */
	T6 a37; /* yy_full_state */
	T6 a38; /* yy_full_lp */
	T0* a39; /* input_buffer */
	T1 a40; /* yy_rejected */
	T0* a41; /* yy_pushed_start_conditions */
	T0* a42; /* last_string_value */
	T0* a43; /* last_error */
	T6 a44; /* pushed_start_condition_count */
};

/* Struct for type [detachable] XM_DEFAULT_POSITION */
struct S507 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* row */
	T6 a2; /* column */
	T0* a3; /* source_name */
	T6 a4; /* byte_index */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
struct S509 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
struct S510 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* first_cell */
};

/* Struct for type [detachable] XM_CALLBACKS_NULL */
struct S511 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
struct S512 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* capacity */
	T0* a7; /* key_equality_tester */
	T0* a8; /* internal_keys */
	T6 a9; /* modulus */
	T6 a10; /* slots_position */
	T6 a11; /* clashes_previous_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* found_position */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* equality_tester */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] XM_NULL_EXTERNAL_RESOLVER */
struct S514 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_CALLBACKS_NULL */
struct S516 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
struct S517 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ANY] */
struct S518 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
struct S519 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_PARSER_NAME */
struct S520 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* use_namespaces */
	T6 a2; /* count */
	T0* a3; /* first */
	T0* a4; /* second */
	T0* a5; /* tail */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
struct S521 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
struct S522 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
struct S523 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* equality_tester */
	T6 a3; /* count */
	T6 a4; /* capacity */
	T6 a5; /* free_slot */
	T6 a6; /* last_position */
	T6 a7; /* modulus */
	T6 a8; /* slots_position */
	T6 a9; /* clashes_previous_position */
	T6 a10; /* found_position */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
struct S524 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
struct S525 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_EXTERNAL_ID */
struct S526 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* system_id */
	T0* a2; /* public_id */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
struct S527 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S528 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_ELEMENT_CONTENT */
struct S529 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* items */
	T2 a2; /* repetition */
	T0* a3; /* name */
	T2 a4; /* type */
	T1 a5; /* is_character_data_allowed */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S530 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
struct S531 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S532 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
struct S533 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S534 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_ATTRIBUTE_CONTENT */
struct S535 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* default_value */
	T2 a3; /* type */
	T1 a4; /* is_list_type */
	T0* a5; /* enumeration_list */
	T2 a6; /* value */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S536 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
struct S537 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
struct S538 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* equality_tester */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
struct S539 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
struct S540 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
struct S541 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_DECLARATION */
struct S542 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* encoding */
	T0* a2; /* version */
	T1 a3; /* stand_alone */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
struct S543 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] XM_EIFFEL_ENTITY_DEF */
struct S546 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* literal_name */
	T0* a2; /* value */
	T0* a3; /* resolver */
	T0* a4; /* external_id */
	T0* a5; /* character_entity */
	T0* a6; /* yy_pushed_start_conditions */
	T0* a7; /* input_buffer */
	T1 a8; /* in_use */
	T0* a9; /* input_name */
	T0* a10; /* last_value */
	T0* a11; /* last_string_value */
	T0* a12; /* last_error */
	T1 a13; /* yy_rejected */
	T6 a14; /* yy_state_count */
	T6 a15; /* yy_full_match */
	T6 a16; /* yy_lp */
	T6 a17; /* yy_looking_for_trail_begin */
	T6 a18; /* yy_full_lp */
	T6 a19; /* yy_full_state */
	T6 a20; /* pushed_start_condition_count */
	T6 a21; /* yy_start_state */
	T6 a22; /* yy_line */
	T6 a23; /* yy_column */
	T6 a24; /* yy_position */
	T6 a25; /* yy_start */
	T6 a26; /* yy_end */
	T6 a27; /* line */
	T6 a28; /* column */
	T6 a29; /* position */
	T1 a30; /* yy_more_flag */
	T6 a31; /* yy_more_len */
	T6 a32; /* yy_last_accepting_state */
	T6 a33; /* yy_last_accepting_cpos */
	T0* a34; /* yy_state_stack */
	T0* a35; /* yy_nxt */
	T0* a36; /* yy_chk */
	T0* a37; /* yy_base */
	T0* a38; /* yy_def */
	T0* a39; /* yy_ec */
	T0* a40; /* yy_meta */
	T0* a41; /* yy_accept */
	T0* a42; /* yy_content */
	T0* a43; /* yy_content_area */
	T0* a44; /* yy_unicode_content_area */
	T6 a45; /* last_token */
	T0* a46; /* input_filter */
	T0* a47; /* input_stream */
	T0* a48; /* input_resolver */
	T0* a49; /* yy_acclist */
};

/* Struct for type [detachable] XM_EIFFEL_SCANNER_DTD */
struct S547 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_token */
	T0* a2; /* last_value */
	T0* a3; /* input_name */
	T6 a4; /* position */
	T6 a5; /* column */
	T6 a6; /* line */
	T0* a7; /* input_filter */
	T6 a8; /* yy_start_state */
	T0* a9; /* character_entity */
	T0* a10; /* input_stream */
	T0* a11; /* input_resolver */
	T1 a12; /* decl_start_sent */
	T1 a13; /* decl_end_sent */
	T1 a14; /* yy_more_flag */
	T6 a15; /* yy_more_len */
	T6 a16; /* yy_end */
	T6 a17; /* yy_start */
	T6 a18; /* yy_line */
	T6 a19; /* yy_column */
	T6 a20; /* yy_position */
	T0* a21; /* yy_state_stack */
	T6 a22; /* yy_state_count */
	T0* a23; /* yy_content_area */
	T0* a24; /* yy_unicode_content_area */
	T0* a25; /* yy_accept */
	T6 a26; /* yy_last_accepting_state */
	T6 a27; /* yy_last_accepting_cpos */
	T0* a28; /* yy_content */
	T0* a29; /* yy_ec */
	T0* a30; /* yy_chk */
	T0* a31; /* yy_base */
	T0* a32; /* yy_def */
	T0* a33; /* yy_meta */
	T0* a34; /* yy_nxt */
	T6 a35; /* yy_lp */
	T0* a36; /* yy_acclist */
	T6 a37; /* yy_looking_for_trail_begin */
	T6 a38; /* yy_full_match */
	T6 a39; /* yy_full_state */
	T6 a40; /* yy_full_lp */
	T0* a41; /* input_buffer */
	T1 a42; /* yy_rejected */
	T0* a43; /* yy_pushed_start_conditions */
	T0* a44; /* last_string_value */
	T0* a45; /* last_error */
	T6 a46; /* pushed_start_condition_count */
};

/* Struct for type [detachable] XM_EIFFEL_PE_ENTITY_DEF */
struct S549 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* resolver */
	T0* a2; /* external_id */
	T0* a3; /* literal_name */
	T0* a4; /* value */
	T0* a5; /* character_entity */
	T0* a6; /* yy_pushed_start_conditions */
	T0* a7; /* input_buffer */
	T1 a8; /* in_use */
	T0* a9; /* input_name */
	T0* a10; /* last_value */
	T0* a11; /* last_string_value */
	T0* a12; /* last_error */
	T1 a13; /* yy_rejected */
	T6 a14; /* yy_state_count */
	T6 a15; /* yy_full_match */
	T6 a16; /* yy_lp */
	T6 a17; /* yy_looking_for_trail_begin */
	T6 a18; /* yy_full_lp */
	T6 a19; /* yy_full_state */
	T6 a20; /* pushed_start_condition_count */
	T6 a21; /* yy_start_state */
	T6 a22; /* yy_line */
	T6 a23; /* yy_column */
	T6 a24; /* yy_position */
	T6 a25; /* yy_start */
	T6 a26; /* yy_end */
	T6 a27; /* line */
	T6 a28; /* column */
	T6 a29; /* position */
	T1 a30; /* yy_more_flag */
	T6 a31; /* yy_more_len */
	T6 a32; /* yy_last_accepting_state */
	T6 a33; /* yy_last_accepting_cpos */
	T0* a34; /* yy_state_stack */
	T1 a35; /* pre_sent */
	T1 a36; /* post_sent */
	T0* a37; /* yy_nxt */
	T0* a38; /* yy_chk */
	T0* a39; /* yy_base */
	T0* a40; /* yy_def */
	T0* a41; /* yy_ec */
	T0* a42; /* yy_meta */
	T0* a43; /* yy_accept */
	T0* a44; /* yy_content */
	T0* a45; /* yy_content_area */
	T0* a46; /* yy_unicode_content_area */
	T6 a47; /* last_token */
	T0* a48; /* input_filter */
	T0* a49; /* input_stream */
	T0* a50; /* input_resolver */
	T0* a51; /* yy_acclist */
};

/* Struct for type [detachable] XM_NAMESPACE_RESOLVER */
struct S550 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* element_local_part */
	T0* a3; /* element_prefix */
	T0* a4; /* context */
	T1 a5; /* forward_xmlns */
	T0* a6; /* attributes_prefix */
	T0* a7; /* attributes_local_part */
	T0* a8; /* attributes_value */
};

/* Struct for type [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
struct S551 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
struct S552 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
struct S553 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T0* a7; /* special_item_routines */
	T0* a8; /* item_storage */
	T0* a9; /* clashes */
	T0* a10; /* slots */
	T6 a11; /* found_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* slots_position */
	T6 a14; /* count */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_CLUSTERS */
struct S554 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clusters */
};

/* Struct for type [detachable] ET_CLASS_NAME_TESTER */
struct S555 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_PARENT */
struct S557 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
	T0* a3; /* exports */
	T0* a4; /* undefines */
	T0* a5; /* redefines */
	T0* a6; /* selects */
	T0* a7; /* end_keyword */
};

/* Struct for type [detachable] ET_PARENT_LIST */
struct S558 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* inherit_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
	T0* a4; /* clients_clause */
};

/* Struct for type [detachable] ET_BUILTIN_CONVERT_FEATURE */
struct S559 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* type */
};

/* Struct for type [detachable] ET_MASTER_CLASS */
struct S560 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* mapped_class */
	T0* a2; /* first_overriding_class */
	T0* a3; /* intrinsic_class */
	T0* a4; /* name */
	T0* a5; /* universe */
	T0* a6; /* status_mutex */
	T0* a7; /* processing_mutex */
	T1 a8; /* unprotected_is_marked */
	T0* a9; /* first_imported_class */
	T0* a10; /* other_imported_classes */
	T1 a11; /* is_modified */
	T0* a12; /* first_local_override_class */
	T0* a13; /* first_local_non_override_class */
	T0* a14; /* first_local_ignored_class */
	T0* a15; /* first_local_hidden_class */
	T0* a16; /* other_local_override_classes */
	T0* a17; /* other_local_non_override_classes */
	T0* a18; /* other_local_ignored_classes */
	T0* a19; /* other_overriding_classes */
	T0* a20; /* other_local_hidden_classes */
};

/* Struct for type [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
struct S561 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* attachment_mark */
	T0* a2; /* separateness_keyword */
};

/* Struct for type detachable ET_RENAME_LIST */
struct S562 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* rename_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type detachable ET_EXPORT_LIST */
struct S563 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* export_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type detachable ET_KEYWORD_FEATURE_NAME_LIST */
struct S564 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
struct S567 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tuple_type */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
struct S568 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
struct S569 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
struct S572 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
struct S574 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
struct S575 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S576 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
struct S577 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
struct S579 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S581 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
	T0* a2; /* equality_tester */
	T0* a3; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
struct S582 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
struct S583 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S584 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLIES */
struct S585 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY */
struct S586 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* referenced_assemblies */
	T0* a3; /* current_system */
	T0* a4; /* dotnet_assembly */
	T0* a5; /* name */
	T0* a6; /* implicit_attachment_type_mark */
	T0* a7; /* master_classes */
	T0* a8; /* master_class_mutex */
	T0* a9; /* any_type */
	T0* a10; /* detachable_any_type */
	T0* a11; /* detachable_separate_any_type */
	T0* a12; /* any_parent */
	T0* a13; /* any_parents */
	T0* a14; /* any_clients */
	T0* a15; /* tuple_type */
	T0* a16; /* detachable_tuple_type */
	T0* a17; /* tuple_identity_type */
	T0* a18; /* unfolded_empty_tuple_actual_parameters */
	T0* a19; /* array_any_type */
	T0* a20; /* array_detachable_any_type */
	T0* a21; /* array_none_type */
	T0* a22; /* array_identity_type */
	T0* a23; /* boolean_type */
	T0* a24; /* character_type */
	T0* a25; /* character_8_type */
	T0* a26; /* character_8_convert_feature */
	T0* a27; /* character_32_type */
	T0* a28; /* character_32_convert_feature */
	T0* a29; /* double_type */
	T0* a30; /* exception_type */
	T0* a31; /* detachable_exception_type */
	T0* a32; /* exception_manager_type */
	T0* a33; /* function_identity_any_type */
	T0* a34; /* immutable_string_8_type */
	T0* a35; /* immutable_string_32_type */
	T0* a36; /* integer_type */
	T0* a37; /* integer_8_type */
	T0* a38; /* integer_8_convert_feature */
	T0* a39; /* integer_16_type */
	T0* a40; /* integer_16_convert_feature */
	T0* a41; /* integer_32_type */
	T0* a42; /* integer_32_convert_feature */
	T0* a43; /* integer_64_type */
	T0* a44; /* integer_64_convert_feature */
	T0* a45; /* ise_exception_manager_type */
	T0* a46; /* iterable_detachable_separate_any_type */
	T0* a47; /* natural_type */
	T0* a48; /* natural_8_type */
	T0* a49; /* natural_8_convert_feature */
	T0* a50; /* natural_16_type */
	T0* a51; /* natural_16_convert_feature */
	T0* a52; /* natural_32_type */
	T0* a53; /* natural_32_convert_feature */
	T0* a54; /* natural_64_type */
	T0* a55; /* natural_64_convert_feature */
	T0* a56; /* none_type */
	T0* a57; /* detachable_none_type */
	T0* a58; /* pointer_type */
	T0* a59; /* predicate_identity_type */
	T0* a60; /* procedure_identity_type */
	T0* a61; /* real_type */
	T0* a62; /* real_32_type */
	T0* a63; /* real_32_convert_feature */
	T0* a64; /* real_64_type */
	T0* a65; /* real_64_convert_feature */
	T0* a66; /* routine_identity_type */
	T0* a67; /* special_any_type */
	T0* a68; /* special_detachable_any_type */
	T0* a69; /* special_identity_type */
	T0* a70; /* string_type */
	T0* a71; /* detachable_string_type */
	T0* a72; /* string_8_type */
	T0* a73; /* string_8_convert_feature */
	T0* a74; /* string_32_type */
	T0* a75; /* string_32_convert_feature */
	T0* a76; /* system_object_type */
	T0* a77; /* system_object_parents */
	T0* a78; /* system_string_type */
	T0* a79; /* type_detachable_any_type */
	T0* a80; /* detachable_type_exception_type */
	T0* a81; /* type_detachable_exception_type */
	T0* a82; /* type_detachable_like_current_type */
	T0* a83; /* type_identity_type */
	T0* a84; /* typed_pointer_identity_type */
	T0* a85; /* wide_character_type */
	T1 a86; /* is_read_only */
	T1 a87; /* is_preparsed */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
struct S589 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
struct S590 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
struct S591 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
struct S592 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* modulus */
	T0* a2; /* key_equality_tester */
	T0* a3; /* item_storage */
	T0* a4; /* slots */
	T0* a5; /* hash_function */
	T0* a6; /* key_storage */
	T0* a7; /* clashes */
	T6 a8; /* position */
	T6 a9; /* last_position */
	T6 a10; /* capacity */
	T6 a11; /* slots_position */
	T6 a12; /* count */
	T0* a13; /* equality_tester */
	T6 a14; /* found_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* special_key_routines */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
};

/* Struct for type detachable ET_FORMAL_PARAMETER_LIST */
struct S593 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type detachable ET_FEATURE_IDS */
struct S594 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* feature_ids */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
struct S595 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
struct S596 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
struct S599 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
struct S600 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
struct S601 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
struct S602 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
struct S603 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
struct S605 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
struct S609 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_BRACKET_SYMBOL */
struct S610 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T2 a2; /* code */
	T6 a3; /* seed */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
struct S612 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
struct S613 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
struct S614 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
struct S615 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
struct S616 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S617 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S618 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S619 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
struct S620 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPRESSION] */
struct S622 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
struct S623 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
struct S624 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
struct S625 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
struct S626 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S627 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S628 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S629 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
struct S631 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
struct S632 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
struct S633 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S634 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S635 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S636 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
struct S637 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
struct S638 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
struct S639 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT] */
struct S640 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
struct S641 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
struct S642 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
struct S643 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
struct S645 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S646 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S647 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S648 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S649 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S650 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S651 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
struct S653 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE] */
struct S654 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
struct S655 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
struct S656 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONSTANT] */
struct S657 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
struct S659 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
struct S660 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
struct S661 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
struct S663 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
struct S664 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S666 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
struct S667 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
struct S668 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
struct S669 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
struct S672 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S675 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* free_slot */
	T6 a8; /* count */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* equality_tester */
	T0* a13; /* slots */
	T0* a14; /* clashes */
	T0* a15; /* special_item_routines */
	T0* a16; /* key_storage */
	T0* a17; /* special_key_routines */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S676 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T0* a2; /* special_routines */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ARRAY [[attached] BOOLEAN] */
struct S677 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
struct S680 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] WORKER_THREAD */
struct S681 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* thread_procedure */
	T0* a2; /* launch_mutex */
	T14 a3; /* thread_id */
	T1 a4; /* terminated */
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
struct S682 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
struct S686 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
struct S687 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
struct S688 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
struct S689 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
struct S690 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
struct S691 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
struct S692 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S693 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
struct S694 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S695 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S696 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S697 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
struct S698 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S699 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S700 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S701 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
struct S702 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* equality_tester */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S703 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S704 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S705 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
struct S706 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S707 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S708 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S709 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S710 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S711 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
struct S712 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_NAMED_OBJECT_TEST */
struct S715 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* attached_keyword */
	T0* a4; /* declared_type */
	T0* a5; /* expression */
	T0* a6; /* as_keyword */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
struct S716 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
struct S717 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
struct S718 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* special_routines */
	T0* a3; /* storage */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
struct S719 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT] */
struct S721 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT] */
struct S722 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT] */
struct S723 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] INTEGER_32] */
struct S724 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* count */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
struct S726 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
struct S727 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
struct S728 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
struct S729 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
struct S731 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
struct S732 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
struct S733 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
struct S734 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S735 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S736 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S737 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
struct S738 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
struct S739 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
struct S741 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
struct S742 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] UC_UTF8_STRING */
struct S743 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* area */
	T6 a3; /* byte_count */
	T6 a4; /* internal_hash_code */
	T6 a5; /* last_byte_index_input */
	T6 a6; /* last_byte_index_result */
	T6 a7; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_32] */
struct S745 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T10 z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
struct S746 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
struct S747 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_64] */
struct S748 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T11 z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
struct S749 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* position */
	T0* a3; /* internal_cursor */
	T6 a4; /* found_position */
	T0* a5; /* item_storage */
	T6 a6; /* modulus */
	T0* a7; /* key_equality_tester */
	T0* a8; /* key_storage */
	T0* a9; /* slots */
	T0* a10; /* hash_function */
	T0* a11; /* clashes */
	T0* a12; /* internal_keys */
	T6 a13; /* capacity */
	T6 a14; /* free_slot */
	T6 a15; /* last_position */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* equality_tester */
	T0* a19; /* special_item_routines */
	T0* a20; /* special_key_routines */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
struct S750 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
struct S751 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
struct S752 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_INTEGER_ROUTINES */
struct S753 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] STRING_8_SEARCHER */
struct S755 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* deltas */
};

/* Struct for type detachable ARRAYED_LIST [[attached] INTEGER_32] */
struct S756 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area_v2 */
	T6 a2; /* index */
};

/* Struct for type [detachable] ET_SYSTEM_ERROR */
struct S757 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
};

/* Struct for type [detachable] ET_INTERNAL_ERROR */
struct S758 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
};

/* Struct for type [detachable] ET_VALIDITY_ERROR */
struct S759 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* ise_reported */
	T1 a2; /* ge_reported */
	T0* a3; /* default_template */
	T0* a4; /* parameters */
	T0* a5; /* current_class */
	T0* a6; /* class_impl */
	T0* a7; /* position */
	T0* a8; /* code */
	T0* a9; /* etl_code */
	T1 a10; /* ise_fatal */
	T1 a11; /* ge_fatal */
};

/* Struct for type detachable ET_ECF_NOTE_ELEMENT */
struct S760 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* attributes */
	T0* a3; /* elements */
	T0* a4; /* content */
};

/* Struct for type detachable ET_ECF_CLUSTERS */
struct S761 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clusters */
};

/* Struct for type detachable ET_ECF_ADAPTED_LIBRARIES */
struct S762 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* libraries */
};

/* Struct for type detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
struct S763 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type detachable ET_ECF_FILE_RULES */
struct S764 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* file_rules */
};

/* Struct for type detachable ET_ECF_EXTERNAL_CFLAGS */
struct S765 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_INCLUDES */
struct S766 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LIBRARIES */
struct S767 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
struct S768 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_MAKES */
struct S769 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_OBJECTS */
struct S770 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_RESOURCES */
struct S771 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
struct S772 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
struct S773 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* eifgens_location */
	T0* a7; /* classname_prefix */
	T1 a8; /* is_read_only */
	T1 a9; /* use_application_options */
	T0* a10; /* options */
	T0* a11; /* class_options */
	T0* a12; /* class_renamings */
	T0* a13; /* visible_classes */
	T0* a14; /* conditions */
	T0* a15; /* name */
	T0* a16; /* library */
};

/* Struct for type detachable ET_ECF_VERSION */
struct S775 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_major */
	T6 a2; /* internal_minor */
	T6 a3; /* internal_release */
	T6 a4; /* internal_build */
	T0* a5; /* product */
	T0* a6; /* company */
	T0* a7; /* copyright */
};

/* Struct for type detachable ET_ECF_CLUSTER */
struct S776 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* pathname */
	T1 a3; /* is_relative */
	T0* a4; /* universe */
	T0* a5; /* target */
	T0* a6; /* description */
	T1 a7; /* is_hidden */
	T0* a8; /* classname_prefix */
	T1 a9; /* is_read_only */
	T1 a10; /* is_recursive */
	T0* a11; /* conditioned_subclusters */
	T0* a12; /* conditioned_file_rules */
	T0* a13; /* options */
	T0* a14; /* class_options */
	T0* a15; /* class_renamings */
	T0* a16; /* class_mappings */
	T0* a17; /* visible_classes */
	T0* a18; /* provider_groups */
	T0* a19; /* conditions */
	T0* a20; /* notes */
	T1 a21; /* is_override */
	T0* a22; /* overridden_group */
	T1 a23; /* overridden_constraint_enabled */
	T1 a24; /* scm_mapping_constraint_enabled */
	T0* a25; /* subclusters */
	T1 a26; /* use_obsolete_syntax */
	T0* a27; /* scm_read_mapping */
	T0* a28; /* scm_write_mapping */
	T1 a29; /* is_implicit */
	T0* a30; /* file_rules */
	T0* a31; /* parent */
	T0* a32; /* provider_constraint */
	T0* a33; /* dependant_constraint */
	T1 a34; /* is_preparsed */
	T1 a35; /* is_abstract */
	T0* a36; /* cached_absolute_pathname */
};

/* Struct for type detachable ET_ECF_EXTERNAL_CFLAG */
struct S777 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* flag */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_INCLUDE */
struct S778 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LIBRARY */
struct S779 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LINKER_FLAG */
struct S780 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* flag */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_MAKE */
struct S781 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_OBJECT */
struct S782 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_RESOURCE */
struct S783 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_FILE_RULE */
struct S784 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* exclude */
	T0* a2; /* include */
	T0* a3; /* exclude_regexp */
	T0* a4; /* include_regexp */
	T0* a5; /* conditions */
	T0* a6; /* description */
};

/* Struct for type detachable ET_ECF_ACTION */
struct S785 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* command_name */
	T0* a2; /* description */
	T1 a3; /* must_succeed */
	T0* a4; /* working_directory */
	T0* a5; /* conditions */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S786 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* last_position */
	T6 a4; /* position */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* key_equality_tester */
	T0* a9; /* internal_keys */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
struct S787 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
struct S788 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
struct S789 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable ET_ECF_ROOT_CLASS */
struct S790 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* creation_procedure_name */
	T0* a3; /* cluster_name */
};

/* Struct for type [detachable] ET_ECF_ROOT_ALL_CLASSES */
struct S791 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_ECF_ORED_CONDITIONS */
struct S792 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditions */
};

/* Struct for type detachable ET_ECF_ANDED_CONDITIONS */
struct S793 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditions */
};

/* Struct for type detachable ET_ECF_VISIBLE_CLASS */
struct S794 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* feature_name */
	T0* a3; /* new_class_name */
	T0* a4; /* new_feature_name */
};

/* Struct for type detachable ET_ECF_BUILD_CONDITION */
struct S796 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_CONCURRENCY_CONDITION */
struct S797 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_VOID_SAFETY_CONDITION */
struct S798 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_CUSTOM_CONDITION */
struct S799 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* value */
	T0* a3; /* match */
	T1 a4; /* is_excluded */
};

/* Struct for type detachable ET_ECF_DOTNET_CONDITION */
struct S800 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* value */
};

/* Struct for type detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
struct S801 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* value */
};

/* Struct for type detachable ET_ECF_PLATFORM_CONDITION */
struct S802 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
struct S804 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* min_value */
	T0* a2; /* max_value */
};

/* Struct for type [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
struct S805 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* min_value */
	T0* a2; /* max_value */
};

/* Struct for type [detachable] UT_COUNTER */
struct S807 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
struct S808 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] UT_COUNTER] */
struct S809 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
struct S810 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_UNIVERSE] */
struct S811 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
struct S812 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
struct S813 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S814 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
struct S816 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* modulus */
	T6 a3; /* last_position */
	T6 a4; /* free_slot */
	T6 a5; /* position */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T0* a13; /* internal_cursor */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
struct S817 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
struct S818 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
struct S819 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
struct S820 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
struct S821 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
struct S822 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
struct S823 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
struct S824 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
struct S825 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_LIBRARY */
struct S826 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_system */
	T0* a2; /* library */
	T0* a3; /* name */
	T0* a4; /* libraries */
	T0* a5; /* clusters */
	T0* a6; /* dotnet_assemblies */
	T0* a7; /* implicit_attachment_type_mark */
	T0* a8; /* master_classes */
	T0* a9; /* master_class_mutex */
	T0* a10; /* any_type */
	T0* a11; /* detachable_any_type */
	T0* a12; /* detachable_separate_any_type */
	T0* a13; /* any_parent */
	T0* a14; /* any_parents */
	T0* a15; /* any_clients */
	T0* a16; /* tuple_type */
	T0* a17; /* detachable_tuple_type */
	T0* a18; /* tuple_identity_type */
	T0* a19; /* unfolded_empty_tuple_actual_parameters */
	T0* a20; /* array_any_type */
	T0* a21; /* array_detachable_any_type */
	T0* a22; /* array_none_type */
	T0* a23; /* array_identity_type */
	T0* a24; /* boolean_type */
	T0* a25; /* character_type */
	T0* a26; /* character_8_type */
	T0* a27; /* character_8_convert_feature */
	T0* a28; /* character_32_type */
	T0* a29; /* character_32_convert_feature */
	T0* a30; /* double_type */
	T0* a31; /* exception_type */
	T0* a32; /* detachable_exception_type */
	T0* a33; /* exception_manager_type */
	T0* a34; /* function_identity_any_type */
	T0* a35; /* immutable_string_8_type */
	T0* a36; /* immutable_string_32_type */
	T0* a37; /* integer_type */
	T0* a38; /* integer_8_type */
	T0* a39; /* integer_8_convert_feature */
	T0* a40; /* integer_16_type */
	T0* a41; /* integer_16_convert_feature */
	T0* a42; /* integer_32_type */
	T0* a43; /* integer_32_convert_feature */
	T0* a44; /* integer_64_type */
	T0* a45; /* integer_64_convert_feature */
	T0* a46; /* ise_exception_manager_type */
	T0* a47; /* iterable_detachable_separate_any_type */
	T0* a48; /* natural_type */
	T0* a49; /* natural_8_type */
	T0* a50; /* natural_8_convert_feature */
	T0* a51; /* natural_16_type */
	T0* a52; /* natural_16_convert_feature */
	T0* a53; /* natural_32_type */
	T0* a54; /* natural_32_convert_feature */
	T0* a55; /* natural_64_type */
	T0* a56; /* natural_64_convert_feature */
	T0* a57; /* none_type */
	T0* a58; /* detachable_none_type */
	T0* a59; /* pointer_type */
	T0* a60; /* predicate_identity_type */
	T0* a61; /* procedure_identity_type */
	T0* a62; /* real_type */
	T0* a63; /* real_32_type */
	T0* a64; /* real_32_convert_feature */
	T0* a65; /* real_64_type */
	T0* a66; /* real_64_convert_feature */
	T0* a67; /* routine_identity_type */
	T0* a68; /* special_any_type */
	T0* a69; /* special_detachable_any_type */
	T0* a70; /* special_identity_type */
	T0* a71; /* string_type */
	T0* a72; /* detachable_string_type */
	T0* a73; /* string_8_type */
	T0* a74; /* string_8_convert_feature */
	T0* a75; /* string_32_type */
	T0* a76; /* string_32_convert_feature */
	T0* a77; /* system_object_type */
	T0* a78; /* system_object_parents */
	T0* a79; /* system_string_type */
	T0* a80; /* type_detachable_any_type */
	T0* a81; /* detachable_type_exception_type */
	T0* a82; /* type_detachable_exception_type */
	T0* a83; /* type_detachable_like_current_type */
	T0* a84; /* type_identity_type */
	T0* a85; /* typed_pointer_identity_type */
	T0* a86; /* wide_character_type */
	T1 a87; /* is_read_only */
	T1 a88; /* is_preparsed */
};

/* Struct for type [detachable] TUPLE [[attached] ET_LIBRARY] */
struct S827 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
struct S828 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
struct S829 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLUSTER] */
struct S831 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
struct S832 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
struct S833 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
struct S834 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T6 a8; /* clashes_previous_position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T6 a13; /* found_position */
	T0* a14; /* internal_cursor */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
struct S835 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
struct S836 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
struct S837 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S838 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
struct S839 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] ET_ANCESTORS_STATUS_CHECKER */
struct S840 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_type_checker */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_FLATTENING_STATUS_CHECKER */
struct S841 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_type_checker */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_INTERFACE_STATUS_CHECKER */
struct S842 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* qualified_anchored_type_checker */
	T0* a2; /* class_type_checker */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
struct S843 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
struct S844 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
struct S846 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
struct S847 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
struct S848 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
struct S849 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [BOOLEAN] */
struct S850 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 z1;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
struct S851 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
struct S854 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
struct S858 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
struct S859 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
struct S860 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] ET_NONE_GROUP */
struct S861 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* universe */
	T0* a2; /* name */
	T1 a3; /* use_obsolete_syntax */
	T0* a4; /* cached_absolute_pathname */
};

/* Struct for type detachable ET_PARENTHESIS_EXPRESSION */
struct S862 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* name */
	T0* a3; /* arguments */
	T6 a4; /* index */
};

/* Struct for type [detachable] ET_REGULAR_MANIFEST_STRING */
struct S863 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* compressed_position */
	T6 a3; /* index */
	T0* a4; /* cast_type */
	T0* a5; /* type */
};

/* Struct for type [detachable] ET_INFIX_FREE_NAME */
struct S864 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* infix_keyword */
	T0* a2; /* operator_string */
	T2 a3; /* code */
	T6 a4; /* hash_code */
	T6 a5; /* seed */
	T0* a6; /* convert_keyword */
};

/* Struct for type detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S866 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER */
struct S867 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* index */
	T0* a3; /* implementation_class */
	T0* a4; /* type_mark */
};

/* Struct for type [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
struct S869 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] UC_STRING_EQUALITY_TESTER */
struct S871 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S873 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S874 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
struct S875 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first */
	T0* a2; /* second */
};

/* Struct for type [detachable] XM_EIFFEL_INPUT_STREAM */
struct S876 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_string */
	T6 a2; /* encoding */
	T0* a3; /* impl */
	T0* a4; /* utf_queue */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
struct S877 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
struct S878 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* modulus */
	T6 a3; /* last_position */
	T6 a4; /* free_slot */
	T6 a5; /* position */
	T0* a6; /* special_item_routines */
	T0* a7; /* item_storage */
	T0* a8; /* clashes */
	T0* a9; /* slots */
	T6 a10; /* found_position */
	T0* a11; /* internal_cursor */
	T6 a12; /* slots_position */
	T6 a13; /* count */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
struct S879 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
struct S880 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
struct S881 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
struct S882 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
struct S884 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_QUERY] */
struct S885 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable ET_ALIAS_NAME */
struct S886 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* alias_keyword */
	T0* a2; /* alias_string */
	T2 a3; /* code */
	T0* a4; /* convert_keyword */
	T6 a5; /* seed */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
struct S887 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PROCEDURE] */
struct S888 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
struct S889 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] CONSOLE */
struct S896 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* file_pointer */
	T6 a2; /* mode */
	T0* a3; /* last_string */
	T0* a4; /* internal_name */
	T0* a5; /* internal_detachable_name_pointer */
	T2 a6; /* last_character */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
struct S897 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
struct S899 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
struct S900 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] INTEGER_OVERFLOW_CHECKER */
struct S901 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* integer_overflow_state1 */
	T0* a2; /* integer_overflow_state2 */
	T0* a3; /* natural_overflow_state1 */
	T0* a4; /* natural_overflow_state2 */
};

/* Struct for type [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
struct S902 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
	T1 a4; /* object_comparison */
};

/* Struct for type [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
struct S903 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
struct S904 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
struct S905 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
struct S906 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable C_DATE */
struct S907 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* millisecond_now */
	T0* a2; /* internal_item */
	T1 a3; /* is_utc */
};

/* Struct for type [detachable] YY_FILE_BUFFER */
struct S908 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T0* a8; /* file */
	T1 a9; /* beginning_of_line */
	T6 a10; /* capacity */
	T1 a11; /* end_of_file */
	T1 a12; /* interactive */
};

/* Struct for type [detachable] KL_STDIN_FILE */
struct S909 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_file */
	T2 a2; /* last_character */
	T0* a3; /* last_string */
	T0* a4; /* character_buffer */
};

/* Struct for type [detachable] YY_BUFFER */
struct S910 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T1 a8; /* beginning_of_line */
	T6 a9; /* capacity */
};

/* Struct for type [detachable] YY_UNICODE_FILE_BUFFER */
struct S914 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T1 a6; /* has_utf8_bom */
	T6 a7; /* count */
	T1 a8; /* filled */
	T6 a9; /* capacity */
	T1 a10; /* end_of_file */
	T1 a11; /* interactive */
	T6 a12; /* encoding */
	T0* a13; /* file */
	T6 a14; /* default_encoding */
	T1 a15; /* beginning_of_line */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
struct S915 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S916 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
struct S917 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
struct S918 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
struct S919 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
struct S920 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
struct S921 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST] */
struct S922 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST] */
struct S923 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
struct S924 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
struct S925 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
struct S926 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* slots_position */
	T0* a10; /* special_item_routines */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T6 a13; /* found_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
struct S927 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD] */
struct S928 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
struct S929 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_AGENT_KEYWORD */
struct S930 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* name */
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
struct S931 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
struct S932 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PRECURSOR_KEYWORD */
struct S933 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* name */
	T6 a4; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
struct S934 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
struct S935 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_SYMBOL] */
struct S936 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
struct S937 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_POSITION] */
struct S938 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
struct S939 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
struct S941 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
struct S942 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BREAK] */
struct S944 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
struct S945 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
struct S947 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
struct S948 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CURRENT] */
struct S949 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
struct S950 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FREE_OPERATOR */
struct S951 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* hash_code */
	T6 a3; /* compressed_position */
	T0* a4; /* operator_name */
	T6 a5; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
struct S952 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
struct S953 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_IDENTIFIER] */
struct S954 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
struct S955 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
struct S957 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
struct S958 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_KEYWORD_OPERATOR */
struct S959 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
struct S960 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
struct S961 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
struct S962 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
struct S963 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
struct S965 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
struct S966 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RESULT] */
struct S967 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
struct S968 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_RETRY_INSTRUCTION */
struct S969 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* text */
};

/* Struct for type [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
struct S970 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
struct S971 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_SYMBOL_OPERATOR */
struct S972 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T6 a3; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
struct S973 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
struct S974 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_VOID */
struct S975 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type [detachable] SPECIAL [detachable ET_VOID] */
struct S976 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
struct S977 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_SEMICOLON_SYMBOL */
struct S978 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
};

/* Struct for type [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
struct S979 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
struct S980 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
struct S981 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
struct S982 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_QUESTION_MARK_SYMBOL */
struct S983 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T6 a3; /* index */
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
struct S984 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
struct S985 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ACROSS_EXPRESSION */
struct S986 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cursor_name */
	T0* a3; /* unfolded_cursor_name */
	T0* a4; /* across_keyword */
	T0* a5; /* iterable_expression */
	T0* a6; /* as_keyword */
	T0* a7; /* until_conditional */
	T0* a8; /* iteration_conditional */
	T0* a9; /* end_keyword */
	T1 a10; /* is_all */
	T0* a11; /* invariant_part */
	T0* a12; /* variant_part */
	T0* a13; /* new_cursor_expression */
	T0* a14; /* cursor_item_expression */
	T0* a15; /* cursor_after_expression */
	T0* a16; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
struct S987 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
struct S988 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ACROSS_INSTRUCTION */
struct S989 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* cursor_name */
	T0* a2; /* unfolded_cursor_name */
	T0* a3; /* across_keyword */
	T0* a4; /* iterable_expression */
	T0* a5; /* as_keyword */
	T0* a6; /* from_compound */
	T0* a7; /* until_conditional */
	T0* a8; /* loop_compound */
	T0* a9; /* end_keyword */
	T0* a10; /* invariant_part */
	T0* a11; /* variant_part */
	T0* a12; /* new_cursor_expression */
	T0* a13; /* cursor_item_expression */
	T0* a14; /* cursor_after_expression */
	T0* a15; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
struct S990 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
struct S991 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
struct S992 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
struct S993 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
struct S994 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
struct S995 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
struct S996 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
struct S997 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
struct S998 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S999 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1001 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
struct S1002 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
struct S1003 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_parenthesis */
	T0* a2; /* right_parenthesis */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
struct S1004 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
struct S1005 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
struct S1006 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
struct S1007 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
struct S1008 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
struct S1009 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ASSIGNER] */
struct S1011 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
struct S1012 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_BRACKET_EXPRESSION */
struct S1013 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* target */
	T0* a3; /* name */
	T0* a4; /* arguments */
};

/* Struct for type [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
struct S1014 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
struct S1015 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CALL_AGENT */
struct S1016 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* target */
	T0* a4; /* qualified_name */
	T0* a5; /* arguments */
	T0* a6; /* implicit_result */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CALL_AGENT] */
struct S1017 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
struct S1018 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CHECK_INSTRUCTION */
struct S1019 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* check_keyword */
	T0* a2; /* end_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
struct S1020 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
struct S1021 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE] */
struct S1023 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
struct S1024 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
struct S1026 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
struct S1027 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
struct S1029 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
struct S1030 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CHOICE_LIST */
struct S1031 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* when_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
struct S1032 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
struct S1033 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLASS] */
struct S1034 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
struct S1035 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
struct S1036 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
struct S1037 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CLIENTS */
struct S1038 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLIENTS] */
struct S1039 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
struct S1040 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_COMPOUND] */
struct S1041 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
struct S1042 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONDITIONAL] */
struct S1044 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
struct S1045 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTANT] */
struct S1046 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1047 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1049 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
struct S1050 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
struct S1051 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
struct S1052 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
struct S1053 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_CREATOR */
struct S1054 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* create_keyword */
	T0* a2; /* end_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
struct S1055 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
struct S1056 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_RENAME_LIST */
struct S1057 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* end_keyword */
	T0* a2; /* rename_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
struct S1058 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
struct S1059 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
struct S1061 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
struct S1062 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
struct S1064 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
struct S1065 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
struct S1067 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
struct S1068 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONVERT_FEATURE_LIST */
struct S1069 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* convert_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
struct S1070 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
struct S1071 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
struct S1072 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
struct S1073 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATION_REGION */
struct S1074 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* left_symbol */
	T0* a3; /* right_symbol */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATION_REGION] */
struct S1075 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
struct S1076 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATOR */
struct S1077 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* create_keyword */
	T0* a2; /* clients */
	T0* a3; /* storage */
	T6 a4; /* count */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATOR] */
struct S1078 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
struct S1079 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATOR_LIST */
struct S1080 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
struct S1081 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
struct S1082 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_DEBUG_INSTRUCTION */
struct S1083 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* keys */
	T0* a2; /* compound */
	T0* a3; /* end_keyword */
};

/* Struct for type [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
struct S1084 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
struct S1085 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_EXPRESSION */
struct S1086 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_keyword */
	T0* a3; /* then_expression */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
struct S1087 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
struct S1088 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_EXPRESSION_LIST */
struct S1089 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
struct S1090 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
struct S1091 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_PART */
struct S1092 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
struct S1093 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
struct S1094 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_PART_LIST */
struct S1095 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
struct S1096 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
struct S1097 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPORT] */
struct S1099 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
struct S1100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
struct S1101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
struct S1102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPRESSION] */
struct S1103 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
struct S1104 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
struct S1105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
struct S1106 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
struct S1107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
struct S1108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
struct S1110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
struct S1111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_CLAUSE */
struct S1112 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* feature_keyword */
	T0* a2; /* clients */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
struct S1113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
struct S1114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_CLAUSE_LIST */
struct S1115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
struct S1116 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
struct S1117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_EXPORT */
struct S1118 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clients_clause */
	T0* a2; /* storage */
	T6 a3; /* count */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
struct S1119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
struct S1120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
struct S1121 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
struct S1122 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
struct S1124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
struct S1125 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
struct S1126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
struct S1127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
struct S1129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S1130 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S1131 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
struct S1132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
struct S1133 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
struct S1134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
struct S1136 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
struct S1137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
struct S1138 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
struct S1139 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_IF_EXPRESSION */
struct S1140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* conditional */
	T0* a3; /* then_keyword */
	T0* a4; /* then_expression */
	T0* a5; /* else_keyword */
	T0* a6; /* else_expression */
	T0* a7; /* end_keyword */
	T0* a8; /* elseif_parts */
};

/* Struct for type [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
struct S1141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
struct S1142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_IF_INSTRUCTION */
struct S1143 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_compound */
	T0* a3; /* end_keyword */
	T0* a4; /* elseif_parts */
	T0* a5; /* else_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
struct S1144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
struct S1145 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INDEXING_LIST */
struct S1146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* indexing_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
struct S1147 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
struct S1148 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
struct S1150 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
struct S1151 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
struct S1152 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
struct S1153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
struct S1155 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
struct S1156 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INDEXING_TERM_LIST */
struct S1157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
struct S1158 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
struct S1159 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
struct S1160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
struct S1161 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INSPECT_INSTRUCTION */
struct S1162 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* when_parts */
	T0* a3; /* end_keyword */
	T0* a4; /* else_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
struct S1163 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
struct S1164 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INSTRUCTION] */
struct S1166 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
struct S1167 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INVARIANTS */
struct S1168 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implementation_class */
	T0* a2; /* invariant_keyword */
	T0* a3; /* storage */
	T6 a4; /* count */
	T0* a5; /* object_tests */
	T0* a6; /* across_components */
	T1 a7; /* validity_checked */
	T1 a8; /* has_validity_error */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INVARIANTS] */
struct S1169 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
struct S1170 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
struct S1171 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
struct S1172 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
struct S1174 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
struct S1175 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_LOCAL_VARIABLE */
struct S1176 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
struct S1177 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
struct S1178 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
struct S1180 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
struct S1181 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
struct S1182 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
struct S1183 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_LOOP_INVARIANTS */
struct S1184 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* invariant_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
struct S1185 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
struct S1186 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_MANIFEST_ARRAY */
struct S1187 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T0* a3; /* left_symbol */
	T6 a4; /* count */
	T0* a5; /* storage */
	T0* a6; /* right_symbol */
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
struct S1188 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
struct S1189 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
struct S1191 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
struct S1192 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_MANIFEST_STRING_LIST */
struct S1193 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_parenthesis */
	T0* a2; /* right_parenthesis */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
struct S1194 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
struct S1195 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
struct S1196 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
struct S1197 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_OBSOLETE] */
struct S1199 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
struct S1200 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PARENTHESIZED_EXPRESSION */
struct S1201 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
struct S1202 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
struct S1203 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT] */
struct S1204 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
struct S1205 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PARENT_CLAUSE_LIST */
struct S1206 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
struct S1207 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
struct S1208 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
struct S1209 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
struct S1210 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_LIST] */
struct S1211 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
struct S1212 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_POSTCONDITIONS */
struct S1213 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* validity_checked */
	T1 a2; /* has_validity_error */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* ensure_keyword */
	T0* a6; /* then_keyword */
};

/* Struct for type [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
struct S1214 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
struct S1215 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
struct S1216 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
struct S1217 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PROCEDURE] */
struct S1218 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
struct S1219 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
struct S1221 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
struct S1222 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUERY] */
struct S1223 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
struct S1224 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
struct S1226 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
struct S1227 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RENAME_LIST] */
struct S1228 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
struct S1229 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_STATIC_CALL_EXPRESSION */
struct S1230 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* feature_keyword */
	T0* a3; /* static_type */
	T0* a4; /* qualified_name */
	T0* a5; /* arguments */
	T0* a6; /* parenthesis_call */
};

/* Struct for type [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
struct S1231 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
struct S1232 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_STRIP_EXPRESSION */
struct S1233 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* strip_keyword */
	T0* a3; /* left_parenthesis */
	T6 a4; /* count */
	T0* a5; /* storage */
	T0* a6; /* right_parenthesis */
};

/* Struct for type [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
struct S1234 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
struct S1235 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE] */
struct S1236 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
struct S1237 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
struct S1239 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
struct S1240 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
struct S1242 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
struct S1243 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_TYPE_CONSTRAINT_LIST */
struct S1244 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
struct S1245 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
struct S1246 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
struct S1248 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
struct S1249 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_LIST] */
struct S1251 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
struct S1252 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_VARIANT */
struct S1253 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* variant_keyword */
	T0* a2; /* tag */
	T0* a3; /* expression */
};

/* Struct for type [detachable] SPECIAL [detachable ET_VARIANT] */
struct S1254 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
struct S1255 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_PART */
struct S1256 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* choices */
	T0* a2; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_PART] */
struct S1257 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
struct S1258 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_PART_LIST */
struct S1259 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
struct S1260 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
struct S1261 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_WRITABLE] */
struct S1262 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
struct S1263 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sorted_items */
	T0* a2; /* cycle */
	T0* a3; /* indexes */
	T0* a4; /* counts */
	T0* a5; /* successors */
	T0* a6; /* items */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
struct S1264 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_PARENT_CHECKER1 */
struct S1265 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* current_parent */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
struct S1266 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* all_base_types */
	T0* a3; /* base_types */
	T0* a4; /* formal_dependencies */
	T0* a5; /* base_type_dependencies */
	T0* a6; /* recursive_formal_constraints */
	T0* a7; /* recursive_formal_constraints_to_be_processed */
	T0* a8; /* current_class */
	T0* a9; /* system_processor */
	T0* a10; /* current_formal */
	T0* a11; /* current_type_constraint */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
struct S1267 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T6 a4; /* position */
	T0* a5; /* item_storage */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_keys */
	T6 a8; /* last_position */
	T6 a9; /* free_slot */
	T6 a10; /* modulus */
	T6 a11; /* slots_position */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* equality_tester */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T6 a17; /* capacity */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_FEATURE_NAME_TESTER */
struct S1268 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
struct S1270 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* position */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* count */
	T6 a9; /* free_slot */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* slots_position */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_item_routines */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_ALIAS_NAME_TESTER */
struct S1271 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
struct S1273 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1274 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T6 a16; /* count */
	T0* a17; /* internal_keys */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
struct S1275 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* rename_table */
	T0* a3; /* export_table */
	T0* a4; /* undefine_table */
	T0* a5; /* redefine_table */
	T0* a6; /* select_table */
	T0* a7; /* alias_mapping */
	T0* a8; /* replicable_features */
	T0* a9; /* current_class */
	T0* a10; /* system_processor */
	T0* a11; /* free_parent_feature */
	T0* a12; /* parent_feature_list */
	T0* a13; /* free_inherited_feature */
	T0* a14; /* inherited_feature_list */
	T0* a15; /* free_redeclared_feature */
	T0* a16; /* redeclared_feature_list */
};

/* Struct for type [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
struct S1276 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* dotnet_features */
	T0* a3; /* other_dotnet_features */
	T0* a4; /* current_class */
	T0* a5; /* system_processor */
	T0* a6; /* free_parent_feature */
	T0* a7; /* parent_feature_list */
	T0* a8; /* free_inherited_feature */
	T0* a9; /* inherited_feature_list */
	T0* a10; /* free_redeclared_feature */
	T0* a11; /* redeclared_feature_list */
};

/* Struct for type [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
struct S1277 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* current_feature */
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
struct S1278 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constraint_context */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_ANCHORED_TYPE_CHECKER */
struct S1279 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* anchored_type_sorter */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
	T0* a5; /* current_anchored_type */
};

/* Struct for type [detachable] ET_SIGNATURE_CHECKER */
struct S1280 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* parent_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
	T6 a5; /* processing_mode */
};

/* Struct for type [detachable] ET_PARENT_CHECKER2 */
struct S1281 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* constraint_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
struct S1282 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* constraint_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_BUILTIN_FEATURE_CHECKER */
struct S1283 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* builtin_features */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_PRECURSOR_CHECKER */
struct S1284 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_feature */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
struct S1285 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* item_storage */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* free_slot */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T6 a11; /* found_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* internal_keys */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_CLASS] */
struct S1286 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T6 a5; /* capacity */
	T6 a6; /* modulus */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* slots */
	T6 a11; /* found_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
struct S1287 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class_impl */
	T0* a3; /* current_class */
	T0* a4; /* adapted_base_class_checker */
	T0* a5; /* adapted_base_classes */
	T0* a6; /* target_context */
	T0* a7; /* other_context */
	T0* a8; /* classes_to_be_processed */
	T0* a9; /* system_processor */
	T1 a10; /* in_qualified_anchored_type */
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
struct S1288 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constraint_context */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
struct S1289 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T6 a3; /* position */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
	T0* a6; /* item_storage */
	T0* a7; /* key_storage */
	T0* a8; /* key_equality_tester */
	T0* a9; /* internal_keys */
	T6 a10; /* last_position */
	T6 a11; /* free_slot */
	T6 a12; /* modulus */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* equality_tester */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_CALL_NAME_TESTER */
struct S1290 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_RENAME */
struct S1295 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* old_name */
	T0* a2; /* as_keyword */
	T0* a3; /* new_name */
};

/* Struct for type [detachable] ET_PARENT_CHECKER3 */
struct S1296 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
struct S1297 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_QUERY] */
struct S1298 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] ET_SUPPLIER_BUILDER */
struct S1299 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* supplier_classes */
	T0* a2; /* current_class */
};

/* Struct for type detachable ET_PARENT_FEATURE */
struct S1302 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* parent */
	T0* a3; /* precursor_feature */
	T0* a4; /* new_name */
	T0* a5; /* undefine_name */
	T0* a6; /* redefine_name */
	T0* a7; /* select_name */
	T0* a8; /* merged_feature */
	T1 a9; /* has_other_deferred */
	T1 a10; /* has_other_effective */
};

/* Struct for type [detachable] ET_INHERITED_FEATURE */
struct S1303 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* parent_feature */
	T0* a3; /* flattened_feature */
	T0* a4; /* flattened_parent */
	T6 a5; /* first_seed */
	T0* a6; /* other_seeds */
	T1 a7; /* is_selected */
	T0* a8; /* replicated_features */
	T0* a9; /* replicated_seeds */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLASS] */
struct S1304 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
struct S1305 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
struct S1306 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
struct S1307 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
};

/* Struct for type [detachable] UT_SYNTAX_ERROR */
struct S1308 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_TOO_MANY_INCLUDES_ERROR */
struct S1309 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type detachable ET_DYNAMIC_TARGET_LIST */
struct S1311 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [[attached] NATIVE_STRING] */
struct S1312 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ARRAY [detachable STRING_8] */
struct S1313 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] XM_EIFFEL_CHARACTER_ENTITY */
struct S1314 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* code */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_POSITION] */
struct S1316 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
struct S1317 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
struct S1318 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
struct S1319 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
struct S1322 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
struct S1323 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
struct S1326 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
struct S1327 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1328 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T6 a2; /* count */
	T0* a3; /* last_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S1329 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S1330 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
struct S1331 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_BILINKABLE [[attached] STRING_8] */
struct S1332 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
struct S1333 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* context */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
struct S1336 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T6 a2; /* count */
	T0* a3; /* last_cell */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
struct S1337 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T0* a2; /* last_cell */
	T6 a3; /* count */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
struct S1338 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
struct S1339 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
struct S1340 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] XM_NAMESPACE] */
struct S1341 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
struct S1342 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
struct S1344 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
struct S1345 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_BRACED_TYPE_LIST */
struct S1346 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
struct S1347 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* equality_tester */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
struct S1348 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
struct S1349 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
struct S1350 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
struct S1351 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1353 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1354 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1355 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
struct S1356 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
struct S1358 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S1359 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1360 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1361 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1362 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
struct S1364 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
struct S1365 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
struct S1366 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] THREAD_ATTRIBUTES */
struct S1367 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* shared */
	T14 a2; /* internal_item */
	T0* a3; /* managed_pointer */
};

/* Struct for type [detachable] CELL [[attached] BOOLEAN] */
struct S1368 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
struct S1370 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
struct S1372 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
struct S1373 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
struct S1374 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
struct S1375 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
struct S1376 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT] */
struct S1377 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
struct S1378 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
struct S1382 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
struct S1383 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
struct S1385 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
struct S1386 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
struct S1393 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_CHARACTER_BUFFER */
struct S1396 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* as_special */
	T0* a2; /* area */
};

/* Struct for type [detachable] UT_TRISTATE */
struct S1398 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* value */
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
struct S1399 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] UT_TRISTATE] */
struct S1400 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
struct S1401 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
struct S1402 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE] */
struct S1403 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
struct S1405 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* last_position */
	T6 a3; /* capacity */
	T6 a4; /* slots_position */
	T6 a5; /* count */
	T6 a6; /* modulus */
	T6 a7; /* clashes_previous_position */
	T0* a8; /* key_equality_tester */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* item_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_key_routines */
	T0* a16; /* key_storage */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
struct S1406 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S1407 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
struct S1408 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
struct S1409 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
struct S1410 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
struct S1411 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S1413 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
struct S1414 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S1415 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
struct S1416 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T1 a3; /* is_never_void */
	T0* a4; /* dynamic_types */
	T0* a5; /* agent_type */
	T1 a6; /* is_dynamic_types_readonly */
};

/* Struct for type [detachable] XM_COMMENT */
struct S1419 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* data */
	T0* a2; /* parent */
};

/* Struct for type [detachable] XM_PROCESSING_INSTRUCTION */
struct S1420 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* data */
	T0* a3; /* parent */
};

/* Struct for type [detachable] XM_CHARACTER_DATA */
struct S1421 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T0* a2; /* parent */
};

/* Struct for type [detachable] ET_LIKE_N */
struct S1424 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* previous */
	T0* a3; /* type_mark */
	T0* a4; /* like_keyword */
};

/* Struct for type [detachable] KL_STRING_INPUT_STREAM */
struct S1427 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_input */
	T2 a2; /* last_character */
	T0* a3; /* last_string */
	T0* a4; /* string */
	T6 a5; /* location */
};

/* Struct for type [detachable] KL_UNICODE_CHARACTER_BUFFER */
struct S1429 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* as_unicode_special */
	T2 a2; /* invalid_character */
	T0* a3; /* unicode_area */
	T0* a4; /* area */
	T3 a5; /* invalid_unicode_character */
};

/* Struct for type detachable ET_OBJECT_TEST_LIST */
struct S1431 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type detachable ET_C3_CHARACTER_CONSTANT */
struct S1432 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T1 a3; /* has_invalid_code */
	T3 a4; /* value */
	T6 a5; /* compressed_position */
	T0* a6; /* literal */
	T0* a7; /* type */
};

/* Struct for type detachable ET_SPECIAL_MANIFEST_STRING */
struct S1433 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T1 a4; /* has_invalid_code */
	T6 a5; /* compressed_position */
	T0* a6; /* literal */
	T0* a7; /* type */
};

/* Struct for type detachable ET_VERBATIM_STRING */
struct S1434 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* marker */
	T0* a7; /* open_white_characters */
	T0* a8; /* close_white_characters */
	T1 a9; /* is_left_aligned */
	T0* a10; /* type */
};

/* Struct for type detachable ET_UNDERSCORED_INTEGER_CONSTANT */
struct S1435 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
struct S1436 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_OCTAL_INTEGER_CONSTANT */
struct S1437 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_BINARY_INTEGER_CONSTANT */
struct S1438 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_REGULAR_REAL_CONSTANT */
struct S1439 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sign */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* type */
};

/* Struct for type detachable ET_UNDERSCORED_REAL_CONSTANT */
struct S1440 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sign */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* type */
};

/* Struct for type detachable ET_TRUE_CONSTANT */
struct S1442 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type detachable ET_FALSE_CONSTANT */
struct S1443 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type detachable ET_C1_CHARACTER_CONSTANT */
struct S1444 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T6 a3; /* compressed_position */
	T3 a4; /* value */
	T0* a5; /* type */
};

/* Struct for type detachable ET_C2_CHARACTER_CONSTANT */
struct S1445 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T6 a3; /* compressed_position */
	T3 a4; /* value */
	T0* a5; /* type */
};

/* Struct for type detachable ET_TAGGED_INDEXING */
struct S1447 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tag */
	T0* a2; /* terms */
};

/* Struct for type detachable ET_INDEXING */
struct S1448 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* terms */
};

/* Struct for type detachable ET_CUSTOM_ATTRIBUTE */
struct S1449 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* creation_expression */
	T0* a2; /* settings */
	T0* a3; /* end_keyword */
};

/* Struct for type detachable ET_CONSTRAINED_FORMAL_PARAMETER */
struct S1450 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* arrow_symbol */
	T0* a2; /* constraint */
	T0* a3; /* creation_procedures */
	T0* a4; /* constraint_base_types */
	T0* a5; /* implementation_class */
	T0* a6; /* type_mark */
	T6 a7; /* index */
	T0* a8; /* name */
	T0* a9; /* recursive_formal_constraints */
};

/* Struct for type detachable ET_TYPE_RENAME_CONSTRAINT */
struct S1451 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
};

/* Struct for type detachable ET_CONSTRAINT_NAMED_TYPE */
struct S1452 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
};

/* Struct for type detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
struct S1453 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* colon */
	T0* a3; /* type */
};

/* Struct for type detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
struct S1454 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* comma */
	T0* a3; /* type */
};

/* Struct for type detachable ET_ALL_EXPORT */
struct S1455 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clients_clause */
	T0* a2; /* all_keyword */
};

/* Struct for type detachable ET_CONVERT_FUNCTION */
struct S1457 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* colon */
};

/* Struct for type detachable ET_CONVERT_PROCEDURE */
struct S1458 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type detachable ET_DO_FUNCTION */
struct S1460 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* declared_type */
	T0* a14; /* frozen_keyword */
	T0* a15; /* extended_name */
	T0* a16; /* obsolete_message */
	T0* a17; /* locals */
	T0* a18; /* compound */
	T0* a19; /* rescue_clause */
	T0* a20; /* clients */
	T0* a21; /* is_keyword */
	T0* a22; /* end_keyword */
	T0* a23; /* semicolon */
	T0* a24; /* feature_clause */
	T0* a25; /* first_indexing */
	T0* a26; /* object_tests */
	T1 a27; /* validity_checked */
	T1 a28; /* has_validity_error */
	T6 a29; /* id */
	T6 a30; /* version */
	T0* a31; /* synonym */
};

/* Struct for type detachable ET_ONCE_FUNCTION */
struct S1461 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* declared_type */
	T0* a14; /* keys */
	T0* a15; /* first_indexing */
	T0* a16; /* frozen_keyword */
	T0* a17; /* extended_name */
	T0* a18; /* obsolete_message */
	T0* a19; /* locals */
	T0* a20; /* compound */
	T0* a21; /* rescue_clause */
	T0* a22; /* clients */
	T0* a23; /* is_keyword */
	T0* a24; /* end_keyword */
	T0* a25; /* semicolon */
	T0* a26; /* feature_clause */
	T0* a27; /* object_tests */
	T1 a28; /* validity_checked */
	T1 a29; /* has_validity_error */
	T6 a30; /* id */
	T6 a31; /* version */
	T0* a32; /* synonym */
};

/* Struct for type detachable ET_DEFERRED_FUNCTION */
struct S1462 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* declared_type */
	T0* a14; /* frozen_keyword */
	T0* a15; /* extended_name */
	T0* a16; /* obsolete_message */
	T0* a17; /* clients */
	T0* a18; /* is_keyword */
	T0* a19; /* deferred_keyword */
	T0* a20; /* end_keyword */
	T0* a21; /* semicolon */
	T0* a22; /* feature_clause */
	T0* a23; /* first_indexing */
	T0* a24; /* object_tests */
	T1 a25; /* validity_checked */
	T1 a26; /* has_validity_error */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type detachable ET_ONCE_PROCEDURE */
struct S1464 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* compound */
	T0* a6; /* implementation_feature */
	T0* a7; /* first_precursor */
	T0* a8; /* other_precursors */
	T0* a9; /* preconditions */
	T6 a10; /* hash_code */
	T0* a11; /* arguments */
	T0* a12; /* postconditions */
	T0* a13; /* keys */
	T0* a14; /* first_indexing */
	T0* a15; /* extended_name */
	T0* a16; /* frozen_keyword */
	T0* a17; /* obsolete_message */
	T0* a18; /* locals */
	T0* a19; /* rescue_clause */
	T0* a20; /* clients */
	T0* a21; /* is_keyword */
	T0* a22; /* end_keyword */
	T0* a23; /* semicolon */
	T0* a24; /* feature_clause */
	T0* a25; /* object_tests */
	T1 a26; /* validity_checked */
	T1 a27; /* has_validity_error */
	T6 a28; /* id */
	T6 a29; /* version */
	T0* a30; /* synonym */
};

/* Struct for type detachable ET_EXTERNAL_PROCEDURE */
struct S1465 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* across_components */
	T0* a3; /* implementation_class */
	T0* a4; /* other_seeds */
	T0* a5; /* implementation_feature */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T8 a10; /* builtin_class_code */
	T8 a11; /* builtin_feature_code */
	T0* a12; /* arguments */
	T0* a13; /* postconditions */
	T0* a14; /* extended_name */
	T0* a15; /* frozen_keyword */
	T0* a16; /* language */
	T0* a17; /* obsolete_message */
	T0* a18; /* alias_clause */
	T0* a19; /* clients */
	T0* a20; /* is_keyword */
	T0* a21; /* end_keyword */
	T0* a22; /* semicolon */
	T0* a23; /* feature_clause */
	T0* a24; /* first_indexing */
	T0* a25; /* object_tests */
	T1 a26; /* validity_checked */
	T1 a27; /* has_validity_error */
	T6 a28; /* id */
	T6 a29; /* version */
	T0* a30; /* synonym */
};

/* Struct for type detachable ET_PREFIX_NAME */
struct S1466 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T0* a2; /* convert_keyword */
	T2 a3; /* code */
	T0* a4; /* prefix_keyword */
	T0* a5; /* operator_string */
};

/* Struct for type detachable ET_PREFIX_FREE_NAME */
struct S1467 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* hash_code */
	T6 a2; /* seed */
	T0* a3; /* prefix_keyword */
	T0* a4; /* operator_string */
	T0* a5; /* convert_keyword */
	T2 a6; /* code */
};

/* Struct for type detachable ET_INFIX_NAME */
struct S1468 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T0* a2; /* convert_keyword */
	T2 a3; /* code */
	T0* a4; /* infix_keyword */
	T0* a5; /* operator_string */
};

/* Struct for type detachable ET_ALIASED_FEATURE_NAME */
struct S1469 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* feature_name */
	T0* a2; /* alias_name */
};

/* Struct for type detachable ET_ALIAS_FREE_NAME */
struct S1470 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* convert_keyword */
	T6 a2; /* hash_code */
	T2 a3; /* code */
	T0* a4; /* alias_string */
	T0* a5; /* alias_keyword */
	T6 a6; /* seed */
};

/* Struct for type detachable ET_CLASS_ASSERTION */
struct S1473 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_keyword */
};

/* Struct for type detachable ET_LABELED_ACTUAL_PARAMETER */
struct S1474 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* declared_type */
	T0* a2; /* label_item */
};

/* Struct for type detachable ET_LIKE_FEATURE */
struct S1476 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* name */
	T6 a4; /* seed */
	T1 a5; /* is_procedure */
};

/* Struct for type detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
struct S1477 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* target_type */
	T0* a4; /* implementation_class */
	T0* a5; /* qualified_name */
	T0* a6; /* left_brace */
	T0* a7; /* right_brace */
};

/* Struct for type detachable ET_QUALIFIED_LIKE_TYPE */
struct S1478 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* target_type */
	T0* a3; /* implementation_class */
	T0* a4; /* qualified_name */
};

/* Struct for type detachable ET_ASSIGNER_INSTRUCTION */
struct S1479 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* call */
	T0* a2; /* source */
	T0* a3; /* assign_symbol */
	T0* a4; /* name */
};

/* Struct for type detachable ET_ASSIGNMENT */
struct S1480 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* source */
	T0* a3; /* assign_symbol */
};

/* Struct for type detachable ET_ASSIGNMENT_ATTEMPT */
struct S1481 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* source */
	T0* a3; /* assign_attempt_symbol */
};

/* Struct for type detachable ET_LOOP_INSTRUCTION */
struct S1482 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* from_compound */
	T0* a2; /* until_conditional */
	T0* a3; /* loop_compound */
	T0* a4; /* end_keyword */
	T0* a5; /* invariant_part */
	T0* a6; /* variant_part */
	T1 a7; /* has_old_variant_syntax */
};

/* Struct for type detachable ET_BANG_INSTRUCTION */
struct S1484 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* target */
	T0* a3; /* creation_call */
	T0* a4; /* left_bang */
	T0* a5; /* right_bang */
};

/* Struct for type detachable ET_CREATE_INSTRUCTION */
struct S1485 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* creation_type */
	T0* a2; /* target */
	T0* a3; /* creation_call */
	T0* a4; /* create_keyword */
	T0* a5; /* creation_region */
};

/* Struct for type detachable ET_CHOICE_RANGE */
struct S1486 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* lower */
	T0* a2; /* upper */
	T0* a3; /* dotdot */
};

/* Struct for type detachable ET_PRECURSOR_INSTRUCTION */
struct S1487 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parent_name */
	T0* a2; /* arguments */
	T0* a3; /* precursor_keyword */
	T1 a4; /* is_parent_prefixed */
	T0* a5; /* parenthesis_call */
	T0* a6; /* parent_type */
};

/* Struct for type detachable ET_STATIC_CALL_INSTRUCTION */
struct S1489 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T0* a2; /* feature_keyword */
	T0* a3; /* qualified_name */
	T0* a4; /* arguments */
	T0* a5; /* parenthesis_call */
};

/* Struct for type detachable ET_PRECURSOR_EXPRESSION */
struct S1490 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T1 a2; /* is_parent_prefixed */
	T0* a3; /* parent_name */
	T0* a4; /* precursor_keyword */
	T0* a5; /* arguments */
	T0* a6; /* parenthesis_call */
	T0* a7; /* parent_type */
};

/* Struct for type detachable ET_FEATURE_ADDRESS */
struct S1491 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* name */
};

/* Struct for type detachable ET_CURRENT_ADDRESS */
struct S1492 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* current_keyword */
};

/* Struct for type detachable ET_RESULT_ADDRESS */
struct S1493 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* result_keyword */
};

/* Struct for type detachable ET_EXPRESSION_ADDRESS */
struct S1494 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* expression */
};

/* Struct for type detachable ET_INFIX_EXPRESSION */
struct S1495 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* name */
	T0* a4; /* right */
	T1 a5; /* is_boolean_operator */
};

/* Struct for type detachable ET_INFIX_AND_THEN_OPERATOR */
struct S1496 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* and_keyword */
	T0* a2; /* then_keyword */
	T6 a3; /* seed */
};

/* Struct for type detachable ET_INFIX_OR_ELSE_OPERATOR */
struct S1497 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* or_keyword */
	T0* a2; /* else_keyword */
	T6 a3; /* seed */
};

/* Struct for type detachable ET_MANIFEST_TYPE */
struct S1498 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left_brace */
	T0* a3; /* type */
	T0* a4; /* right_brace */
};

/* Struct for type detachable ET_PREFIX_EXPRESSION */
struct S1499 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* expression */
	T1 a4; /* is_boolean_operator */
};

/* Struct for type detachable ET_OLD_EXPRESSION */
struct S1500 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* old_keyword */
	T0* a3; /* expression */
};

/* Struct for type detachable ET_OLD_OBJECT_TEST */
struct S1501 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* left_brace */
	T0* a4; /* attached_keyword */
	T0* a5; /* as_keyword */
	T0* a6; /* colon */
	T0* a7; /* declared_type */
	T0* a8; /* right_brace */
	T0* a9; /* expression */
};

/* Struct for type detachable ET_ONCE_MANIFEST_STRING */
struct S1502 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* once_keyword */
	T0* a3; /* manifest_string */
	T6 a4; /* id */
};

/* Struct for type detachable ET_DO_FUNCTION_INLINE_AGENT */
struct S1503 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* declared_type */
	T0* a5; /* end_keyword */
	T0* a6; /* implicit_result */
	T0* a7; /* object_tests */
	T0* a8; /* across_components */
	T0* a9; /* actual_arguments */
	T0* a10; /* preconditions */
	T0* a11; /* locals */
	T0* a12; /* compound */
	T0* a13; /* postconditions */
	T0* a14; /* rescue_clause */
	T0* a15; /* target */
	T6 a16; /* attached_result_index */
	T6 a17; /* result_index */
};

/* Struct for type detachable ET_ONCE_FUNCTION_INLINE_AGENT */
struct S1504 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* declared_type */
	T0* a5; /* end_keyword */
	T0* a6; /* implicit_result */
	T0* a7; /* object_tests */
	T0* a8; /* across_components */
	T0* a9; /* actual_arguments */
	T0* a10; /* preconditions */
	T0* a11; /* locals */
	T0* a12; /* keys */
	T0* a13; /* compound */
	T0* a14; /* postconditions */
	T0* a15; /* rescue_clause */
	T0* a16; /* target */
	T6 a17; /* attached_result_index */
	T6 a18; /* result_index */
};

/* Struct for type detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
struct S1505 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* language */
	T0* a4; /* object_tests */
	T0* a5; /* across_components */
	T0* a6; /* actual_arguments */
	T0* a7; /* preconditions */
	T0* a8; /* alias_clause */
	T0* a9; /* postconditions */
	T0* a10; /* formal_arguments */
	T0* a11; /* declared_type */
	T0* a12; /* end_keyword */
	T0* a13; /* implicit_result */
	T0* a14; /* target */
	T6 a15; /* attached_result_index */
	T6 a16; /* result_index */
};

/* Struct for type detachable ET_DO_PROCEDURE_INLINE_AGENT */
struct S1506 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* end_keyword */
	T0* a5; /* object_tests */
	T0* a6; /* across_components */
	T0* a7; /* actual_arguments */
	T0* a8; /* preconditions */
	T0* a9; /* locals */
	T0* a10; /* compound */
	T0* a11; /* postconditions */
	T0* a12; /* rescue_clause */
	T0* a13; /* target */
	T6 a14; /* attached_result_index */
	T6 a15; /* result_index */
};

/* Struct for type detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
struct S1507 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* end_keyword */
	T0* a5; /* object_tests */
	T0* a6; /* across_components */
	T0* a7; /* actual_arguments */
	T0* a8; /* preconditions */
	T0* a9; /* locals */
	T0* a10; /* keys */
	T0* a11; /* compound */
	T0* a12; /* postconditions */
	T0* a13; /* rescue_clause */
	T0* a14; /* target */
	T6 a15; /* attached_result_index */
	T6 a16; /* result_index */
};

/* Struct for type detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
struct S1508 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* language */
	T0* a4; /* object_tests */
	T0* a5; /* across_components */
	T0* a6; /* actual_arguments */
	T0* a7; /* preconditions */
	T0* a8; /* alias_clause */
	T0* a9; /* postconditions */
	T0* a10; /* formal_arguments */
	T0* a11; /* end_keyword */
	T0* a12; /* target */
	T6 a13; /* attached_result_index */
	T6 a14; /* result_index */
};

/* Struct for type detachable ET_AGENT_OPEN_TARGET */
struct S1509 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* type */
	T0* a3; /* right_brace */
	T6 a4; /* index */
};

/* Struct for type detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
struct S1510 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* left_brace */
	T0* a3; /* right_brace */
	T0* a4; /* question_mark */
	T6 a5; /* index */
};

/* Struct for type detachable ET_TAGGED_ASSERTION */
struct S1513 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tag */
	T0* a2; /* untagged_assertion */
};

/* Struct for type [detachable] ET_FILE_POSITION */
struct S1515 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T0* a2; /* filename */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
struct S1517 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
struct S1518 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] ET_REDECLARED_FEATURE */
struct S1520 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* flattened_feature */
	T0* a3; /* parent_feature */
	T6 a4; /* first_seed */
	T0* a5; /* other_seeds */
	T1 a6; /* is_selected */
	T0* a7; /* replicated_features */
	T0* a8; /* replicated_seeds */
};

/* Struct for type [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
struct S1522 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_state */
	T1 a2; /* internal_overflowed */
	T11 a3; /* part1 */
	T11 a4; /* part2 */
	T0* a5; /* leading_separators */
	T0* a6; /* trailing_separators */
	T6 a7; /* sign */
	T6 a8; /* conversion_type */
	T2 a9; /* internal_lookahead */
	T1 a10; /* leading_separators_acceptable */
	T1 a11; /* trailing_separators_acceptable */
};

/* Struct for type [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
struct S1524 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
struct S1527 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1528 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1529 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
struct S1530 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1531 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1532 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1533 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1534 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1535 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1536 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1537 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
struct S1538 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
struct S1539 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
struct S1540 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1541 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
struct S1542 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
struct S1545 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S1546 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1547 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1548 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1549 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1550 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
struct S1551 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
struct S1552 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
struct S1553 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
struct S1554 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
struct S1556 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_UNIVERSE] */
struct S1557 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
struct S1559 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
struct S1560 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
struct S1562 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
struct S1563 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
struct S1564 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
struct S1565 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
struct S1566 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
struct S1567 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
struct S1568 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1569 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1570 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1571 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
struct S1573 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* first_cell */
	T0* a3; /* last_cell */
};

/* Struct for type [detachable] UC_UTF16_ROUTINES */
struct S1574 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
struct S1575 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
struct S1576 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LIBRARY] */
struct S1577 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
struct S1578 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
struct S1579 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
struct S1580 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
struct S1581 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
struct S1582 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
struct S1583 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
struct S1584 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] YY_UNICODE_BUFFER */
struct S1585 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T1 a8; /* beginning_of_line */
	T6 a9; /* capacity */
};

/* Struct for type [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
struct S1586 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
struct S1587 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
struct S1588 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
struct S1589 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST] */
struct S1590 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST] */
struct S1591 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST] */
struct S1592 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST] */
struct S1593 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
struct S1594 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
struct S1595 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
struct S1596 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
struct S1597 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
struct S1598 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
struct S1600 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
struct S1602 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* modulus */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* equality_tester */
	T0* a13; /* internal_cursor */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
struct S1603 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1606 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
struct S1607 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T0* a2; /* special_routines */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
struct S1608 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T0* a3; /* storage */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
struct S1609 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
struct S1610 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
struct S1613 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
struct S1614 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
struct S1615 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
struct S1617 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
struct S1618 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
struct S1619 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
struct S1620 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
struct S1621 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
struct S1622 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1624 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
struct S1625 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT] */
struct S1626 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_CALL_NAME] */
struct S1627 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T0* a3; /* internal_cursor */
	T6 a4; /* count */
	T0* a5; /* item_storage */
	T6 a6; /* modulus */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* position */
	T0* a10; /* equality_tester */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* slots_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
struct S1628 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T6 a3; /* found_position */
	T0* a4; /* internal_cursor */
	T0* a5; /* key_storage */
	T0* a6; /* item_storage */
	T0* a7; /* key_equality_tester */
	T0* a8; /* internal_keys */
	T6 a9; /* position */
	T6 a10; /* last_position */
	T6 a11; /* free_slot */
	T6 a12; /* modulus */
	T0* a13; /* equality_tester */
	T6 a14; /* slots_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* clashes */
	T0* a18; /* slots */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
struct S1629 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* key_storage */
	T0* a4; /* item_storage */
	T0* a5; /* key_equality_tester */
	T0* a6; /* internal_keys */
	T6 a7; /* count */
	T6 a8; /* last_position */
	T6 a9; /* free_slot */
	T6 a10; /* capacity */
	T6 a11; /* modulus */
	T6 a12; /* slots_position */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* equality_tester */
	T0* a15; /* internal_cursor */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
struct S1630 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T0* a4; /* key_storage */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* key_equality_tester */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_REPLICATED_FEATURE */
struct S1632 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* features */
	T6 a2; /* selected_count */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
struct S1633 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
struct S1634 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_DOTNET_SIGNATURE_TESTER */
struct S1635 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
struct S1636 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* equality_tester */
	T6 a9; /* count */
	T6 a10; /* slots_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* internal_cursor */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
struct S1638 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* internal_cursor */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
struct S1640 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
struct S1641 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* cycle */
	T0* a2; /* counts */
	T0* a3; /* successors */
	T0* a4; /* items */
	T0* a5; /* sorted_items */
	T0* a6; /* indexes */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
struct S1642 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
struct S1643 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
struct S1644 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
struct S1645 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T6 a16; /* slots_position */
	T6 a17; /* count */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
struct S1646 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
struct S1649 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
struct S1651 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_RENAME] */
struct S1652 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CALL_NAME] */
struct S1653 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
struct S1656 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
struct S1657 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
struct S1660 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
struct S1661 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] YY_BUFFER] */
struct S1662 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
struct S1663 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
struct S1664 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
struct S1665 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1667 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1668 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1670 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1671 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* before */
	T0* a2; /* current_cell */
	T0* a3; /* next_cursor */
	T0* a4; /* container */
	T1 a5; /* after */
};

/* Struct for type detachable DS_LINKABLE [detachable STRING_8] */
struct S1672 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLUSTER] */
struct S1673 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
struct S1674 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
struct S1675 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
struct S1676 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
struct S1677 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
struct S1678 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
struct S1679 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
struct S1680 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
struct S1681 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S1682 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1684 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
struct S1685 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1686 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
struct S1687 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
struct S1690 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S1691 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
struct S1692 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T0* a6; /* hash_function */
	T6 a7; /* position */
	T6 a8; /* last_position */
	T6 a9; /* capacity */
	T6 a10; /* slots_position */
	T6 a11; /* count */
	T6 a12; /* free_slot */
	T0* a13; /* equality_tester */
	T6 a14; /* modulus */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* clashes */
	T0* a18; /* slots */
	T0* a19; /* special_key_routines */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
struct S1693 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T0* a6; /* hash_function */
	T6 a7; /* last_position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T6 a10; /* count */
	T0* a11; /* equality_tester */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
struct S1698 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* right */
	T0* a2; /* item */
};

/* Struct for type detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
struct S1703 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T0* a3; /* actual_parameters */
};

/* Struct for type detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
struct S1704 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* arguments */
	T0* a3; /* parenthesis_call */
};

/* Struct for type detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
struct S1706 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_cluster */
	T0* a2; /* group_names */
	T0* a3; /* group_pathnames */
	T0* a4; /* pathname_buffer */
};

/* Struct for type [detachable] ET_SEEDED_QUERY_COMPARATOR */
struct S1707 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
struct S1710 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
struct S1714 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_FORMAL_COMMA_ARGUMENT */
struct S1715 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_LOCAL_COMMA_VARIABLE */
struct S1716 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
struct S1717 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* declared_type */
	T0* a2; /* label_item */
};

/* Struct for type [detachable] ET_KEYWORD_EXPRESSION */
struct S1719 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* expression */
	T0* a2; /* keyword */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
struct S1724 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
struct S1725 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPORT] */
struct S1727 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
struct S1728 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
struct S1730 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
struct S1731 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
struct S1732 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
struct S1733 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1736 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
struct S1737 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
struct S1738 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1739 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
struct S1740 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1741 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
struct S1742 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1743 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
struct S1744 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1745 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
struct S1746 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1747 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
struct S1748 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1749 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
struct S1750 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S1755 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1758 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1759 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
struct S1760 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_expression */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
struct S1762 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
struct S1763 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1764 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1766 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
struct S1768 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
struct S1769 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CREATOR] */
struct S1770 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
struct S1771 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
struct S1772 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
struct S1773 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
struct S1774 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
struct S1775 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
struct S1776 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
struct S1777 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
struct S1779 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
struct S1780 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
struct S1781 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
struct S1782 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
struct S1784 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
struct S1785 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
struct S1786 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
struct S1787 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
struct S1789 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
struct S1790 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_WHEN_PART] */
struct S1791 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
struct S1792 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
struct S1794 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
};

/* Struct for type detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1797 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
	T0* a2; /* equality_tester */
	T0* a3; /* internal_cursor */
};

/* Struct for type [detachable] ET_STANDARD_ONCE_KEYS */
struct S1802 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
struct S1804 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
struct S1805 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
struct S1806 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1807 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1808 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1809 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1810 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1811 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1812 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
struct S1813 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
struct S1814 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
struct S1815 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1816 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1817 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1818 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1819 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1820 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1821 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1822 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1823 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1824 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1825 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1826 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1827 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1828 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1829 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1830 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1831 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1832 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1833 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1834 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1835 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1836 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1837 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1838 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1839 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1840 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1841 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1842 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
struct S1843 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
struct S1844 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
struct S1845 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_LINKABLE [[attached] CHARACTER_8] */
struct S1847 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
struct S1848 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1849 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1850 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1851 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
struct S1852 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME] */
struct S1853 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
struct S1854 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
struct S1857 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
struct S1858 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
struct S1859 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
struct S1861 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
struct S1862 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
struct S1863 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
struct S1864 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
struct S1865 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
struct S1866 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
struct S1867 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
struct S1868 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
struct S1869 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
struct S1870 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
struct S1873 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
struct S1874 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
struct S1875 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
struct S1876 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
struct S1877 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T6 a3; /* free_slot */
	T6 a4; /* position */
	T6 a5; /* modulus */
	T6 a6; /* capacity */
	T0* a7; /* equality_tester */
	T0* a8; /* key_equality_tester */
	T0* a9; /* internal_cursor */
	T6 a10; /* found_position */
	T0* a11; /* item_storage */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* hash_function */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
struct S1878 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
struct S1879 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
struct S1880 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
struct S1882 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
struct S1883 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
struct S1884 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* position */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
struct S1885 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1886 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* left */
	T0* a3; /* right */
};

/* Struct for type [detachable] ET_CLUSTER_ERROR */
struct S1887 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* cluster */
};

/* Struct for type [detachable] ET_SYNTAX_ERROR */
struct S1888 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* filename */
	T0* a6; /* position */
	T1 a7; /* ise_reported */
	T1 a8; /* ise_fatal */
	T1 a9; /* ge_reported */
	T1 a10; /* ge_fatal */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY_ERROR */
struct S1889 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* assembly */
};

/* Struct for type [detachable] AP_OPTION_COMPARATOR */
struct S1890 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
struct S1891 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] ST_SPLITTER */
struct S1898 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_escape_character */
	T2 a2; /* escape_character */
	T0* a3; /* separator_codes */
	T0* a4; /* separators */
};

/* Struct for type [detachable] KL_CHARACTER_ROUTINES */
struct S1899 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
struct S1900 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
struct S1901 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* actual_parameters */
	T6 a3; /* tuple_position */
	T0* a4; /* tuple_type */
};

/* Struct for type [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
struct S1902 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* actual_parameters */
	T6 a2; /* upper */
	T6 a3; /* lower */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
struct S1905 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
struct S1907 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
struct S1909 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
struct S1910 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
struct S1911 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
struct S1913 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
struct S1914 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
struct S1916 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
struct S1924 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* actual_arguments */
	T0* a3; /* left_symbol */
	T0* a4; /* storage */
	T0* a5; /* right_symbol */
};

/* Struct for type [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
struct S1925 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T8 z3;
};

/* Struct for type [detachable] ET_INFIX_CAST_EXPRESSION */
struct S1926 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* type */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
struct S1930 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
struct S1931 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
struct S1934 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
struct S1935 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1936 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table_cursor */
	T0* a2; /* container */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
struct S1939 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
struct S1940 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
struct S1943 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
struct S1944 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] ET_UNIVERSE_ERROR */
struct S1945 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* universe */
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S1946 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type detachable ET_PARENTHESIS_INSTRUCTION */
struct S1947 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* name */
	T0* a3; /* arguments */
};

/* Struct for type [detachable] ET_PARENTHESIS_SYMBOL */
struct S1967 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
struct S1973 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
struct S1974 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* agent_expression */
	T6 a2; /* argument_index */
	T6 a3; /* index */
};

/* Struct for type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S1981 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table_cursor */
	T0* a2; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
struct S1986 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T0* a5; /* key_equality_tester */
	T0* a6; /* internal_keys */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T0* a10; /* equality_tester */
	T6 a11; /* found_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
struct S1987 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
struct S1988 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T0* a2; /* static_equality */
	T0* a3; /* current_feature */
	T0* a4; /* current_type */
	T0* a5; /* next */
};

/* Struct for type [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
struct S1989 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T0* a2; /* static_equality */
	T0* a3; /* current_feature */
	T0* a4; /* current_type */
	T0* a5; /* next */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
struct S1990 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
struct S1992 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* convert_feature */
	T0* a4; /* type */
};

/* Struct for type [detachable] ET_CONVERT_FROM_EXPRESSION */
struct S1993 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* index */
	T0* a3; /* expression */
	T0* a4; /* convert_feature */
	T0* a5; /* type */
};

/* Struct for type [detachable] ET_CONVERT_TO_EXPRESSION */
struct S1994 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* index */
	T0* a3; /* expression */
	T0* a4; /* convert_feature */
};

/* Struct for type detachable DS_CELL [detachable ET_CLASS] */
struct S1997 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
struct S1998 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
struct S2000 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
struct S2002 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
struct S2003 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
struct S2004 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
struct S2005 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
struct S2006 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_STATE] */
struct S2007 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
struct S2009 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
struct S2010 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
struct S2012 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] LX_DFA_WILDCARD */
struct S2014 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* yy_nxt */
	T0* a2; /* yy_accept */
	T6 a3; /* subject_start */
	T6 a4; /* subject_end */
	T6 a5; /* match_count */
	T6 a6; /* matched_start */
	T6 a7; /* matched_end */
	T0* a8; /* yy_ec */
	T6 a9; /* yyNull_equiv_class */
	T6 a10; /* yyMax_symbol_equiv_class */
	T6 a11; /* yyNb_rows */
	T0* a12; /* pattern */
	T1 a13; /* is_case_insensitive */
	T0* a14; /* subject */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
struct S2017 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] TUPLE [[attached] LX_WILDCARD] */
struct S2019 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
struct S2020 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
struct S2021 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
struct S2023 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
struct S2024 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_WILDCARD] */
struct S2025 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
struct S2028 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] CHARACTER_PROPERTY */
struct S2029 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2034 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2035 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2036 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2037 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] ET_AST_NULL_LEAF */
struct S2039 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
};

/* Struct for type [detachable] LX_WILDCARD_PARSER */
struct S2040 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* successful */
	T0* a2; /* pending_rules */
	T0* a3; /* start_condition_stack */
	T0* a4; /* action_factory */
	T0* a5; /* old_singleton_lines */
	T0* a6; /* old_singleton_columns */
	T0* a7; /* old_singleton_counts */
	T0* a8; /* old_regexp_lines */
	T0* a9; /* old_regexp_columns */
	T0* a10; /* old_regexp_counts */
	T0* a11; /* unions_of_concatenations_of_symbol_classes_by_utf8_character_class */
	T0* a12; /* buffer */
	T6 a13; /* maximum_used_symbol */
	T0* a14; /* description */
	T0* a15; /* last_string_value */
	T0* a16; /* last_string_32_value */
	T0* a17; /* last_lx_symbol_class_value */
	T0* a18; /* error_handler */
	T0* a19; /* name_definitions */
	T0* a20; /* character_classes */
	T0* a21; /* character_classes_by_name */
	T0* a22; /* equiv_character_classes */
	T0* a23; /* utf8_character_classes_by_name */
	T0* a24; /* utf8_mode */
	T6 a25; /* line_nb */
	T0* a26; /* eiffel_verbatim_marker */
	T0* a27; /* yyss */
	T0* a28; /* input_buffer */
	T6 a29; /* yy_end */
	T6 a30; /* yy_position */
	T6 a31; /* yy_line */
	T6 a32; /* yy_column */
	T6 a33; /* yy_parsing_status */
	T6 a34; /* yy_suspended_yystacksize */
	T6 a35; /* yy_suspended_yystate */
	T6 a36; /* yy_suspended_yyn */
	T6 a37; /* yy_suspended_yychar1 */
	T6 a38; /* yy_suspended_index */
	T6 a39; /* yy_suspended_yyss_top */
	T6 a40; /* yy_suspended_yy_goto */
	T0* a41; /* yyr1 */
	T6 a42; /* yyssp */
	T0* a43; /* yypgoto */
	T0* a44; /* yycheck */
	T0* a45; /* yytable */
	T0* a46; /* yydefgoto */
	T6 a47; /* error_count */
	T1 a48; /* yy_lookahead_needed */
	T6 a49; /* yyerrstatus */
	T0* a50; /* yypact */
	T6 a51; /* last_token */
	T0* a52; /* yytranslate */
	T0* a53; /* yydefact */
	T0* a54; /* yy_pushed_start_conditions */
	T0* a55; /* yyspecial_routines1 */
	T6 a56; /* yyvsc1 */
	T0* a57; /* yyvs1 */
	T0* a58; /* yyspecial_routines2 */
	T6 a59; /* yyvsc2 */
	T0* a60; /* yyvs2 */
	T0* a61; /* yyspecial_routines3 */
	T6 a62; /* yyvsc3 */
	T0* a63; /* yyvs3 */
	T0* a64; /* yyspecial_routines4 */
	T6 a65; /* yyvsc4 */
	T0* a66; /* yyvs4 */
	T0* a67; /* yyspecial_routines5 */
	T6 a68; /* yyvsc5 */
	T0* a69; /* yyvs5 */
	T0* a70; /* yytypes1 */
	T0* a71; /* yytypes2 */
	T6 a72; /* yy_start */
	T6 a73; /* yyvsp1 */
	T6 a74; /* yyvsp2 */
	T6 a75; /* yyvsp3 */
	T6 a76; /* yyvsp4 */
	T6 a77; /* yyvsp5 */
	T1 a78; /* yy_more_flag */
	T6 a79; /* yy_more_len */
	T6 a80; /* line */
	T6 a81; /* column */
	T6 a82; /* position */
	T6 a83; /* yy_start_state */
	T0* a84; /* yy_state_stack */
	T6 a85; /* yy_state_count */
	T0* a86; /* yy_content_area */
	T0* a87; /* yy_unicode_content_area */
	T0* a88; /* yy_accept */
	T6 a89; /* yy_last_accepting_state */
	T6 a90; /* yy_last_accepting_cpos */
	T0* a91; /* yy_content */
	T0* a92; /* yy_ec */
	T0* a93; /* yy_chk */
	T0* a94; /* yy_base */
	T0* a95; /* yy_def */
	T0* a96; /* yy_meta */
	T0* a97; /* yy_nxt */
	T6 a98; /* yy_lp */
	T0* a99; /* yy_acclist */
	T6 a100; /* yy_looking_for_trail_begin */
	T6 a101; /* yy_full_match */
	T6 a102; /* yy_full_state */
	T6 a103; /* yy_full_lp */
	T1 a104; /* yy_rejected */
	T0* a105; /* last_detachable_any_value */
	T6 a106; /* last_integer_value */
	T1 a107; /* in_trail_context */
	T0* a108; /* rule */
	T1 a109; /* has_trail_context */
	T6 a110; /* head_count */
	T6 a111; /* head_line */
	T6 a112; /* head_column */
	T6 a113; /* trail_count */
	T6 a114; /* i_ */
	T6 a115; /* rule_line_nb */
	T1 a116; /* has_bol_context */
};

/* Struct for type [detachable] LX_DESCRIPTION */
struct S2041 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* maximum_symbol */
	T1 a2; /* utf8_mode */
	T1 a3; /* equiv_classes_used */
	T1 a4; /* full_table */
	T1 a5; /* meta_equiv_classes_used */
	T1 a6; /* reject_used */
	T1 a7; /* variable_trail_context */
	T0* a8; /* rules */
	T0* a9; /* start_conditions */
	T6 a10; /* array_size */
	T1 a11; /* line_pragma */
	T1 a12; /* inspect_used */
	T0* a13; /* eof_rules */
	T0* a14; /* eiffel_header */
	T1 a15; /* case_insensitive */
	T0* a16; /* equiv_classes */
	T1 a17; /* bol_needed */
	T0* a18; /* input_filename */
	T1 a19; /* has_utf8_enconding */
	T1 a20; /* actions_separated */
	T0* a21; /* eiffel_code */
	T1 a22; /* pre_action_used */
	T1 a23; /* post_action_used */
	T1 a24; /* pre_eof_action_used */
	T1 a25; /* post_eof_action_used */
	T1 a26; /* line_used */
	T1 a27; /* position_used */
};

/* Struct for type [detachable] LX_FULL_DFA */
struct S2042 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* yy_nxt */
	T0* a2; /* yy_accept */
	T0* a3; /* yy_ec */
	T6 a4; /* yyNb_rows */
	T6 a5; /* yyNull_equiv_class */
	T6 a6; /* yyMax_symbol_equiv_class */
	T0* a7; /* input_filename */
	T1 a8; /* has_utf8_enconding */
	T6 a9; /* array_size */
	T1 a10; /* inspect_used */
	T1 a11; /* actions_separated */
	T0* a12; /* eiffel_code */
	T0* a13; /* eiffel_header */
	T1 a14; /* bol_needed */
	T1 a15; /* pre_action_used */
	T1 a16; /* post_action_used */
	T1 a17; /* pre_eof_action_used */
	T1 a18; /* post_eof_action_used */
	T1 a19; /* line_pragma */
	T0* a20; /* yy_start_conditions */
	T6 a21; /* yyNb_rules */
	T0* a22; /* yy_rules */
	T6 a23; /* yyEnd_of_buffer */
	T1 a24; /* yyLine_used */
	T1 a25; /* yyPosition_used */
	T6 a26; /* minimum_symbol */
	T6 a27; /* maximum_symbol */
	T0* a28; /* states */
	T6 a29; /* backing_up_count */
	T0* a30; /* partitions */
	T6 a31; /* start_states_count */
	T1 a32; /* yyBacking_up */
	T0* a33; /* yy_eof_rules */
};

/* Struct for type [detachable] TUPLE [[attached] ET_GROUP] */
struct S2044 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
struct S2045 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S2047 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
struct S2048 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_16] */
struct S2051 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T9 z2[1]; /* item */
};

/* Struct for type [detachable] ARRAY [[attached] NATURAL_16] */
struct S2052 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] ARRAY [[attached] NATURAL_32] */
struct S2053 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
struct S2054 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_START_CONDITIONS */
struct S2055 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_ACTION_FACTORY */
struct S2056 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
struct S2057 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* position */
	T6 a5; /* count */
	T6 a6; /* capacity */
	T6 a7; /* free_slot */
	T6 a8; /* last_position */
	T0* a9; /* equality_tester */
	T6 a10; /* modulus */
	T6 a11; /* slots_position */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] LX_SYMBOL_CLASS */
struct S2058 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* upper */
	T1 a2; /* is_empty */
	T1 a3; /* is_negated */
	T11 a4; /* first_set */
	T11 a5; /* second_set */
	T11 a6; /* third_set */
	T11 a7; /* fourth_set */
	T0* a8; /* other_sets */
	T6 a9; /* lower */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
struct S2059 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* special_key_routines */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T6 a15; /* found_position */
	T0* a16; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
struct S2060 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* count */
	T6 a10; /* slots_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* internal_cursor */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
struct S2061 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* position */
	T6 a6; /* count */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* free_slot */
	T6 a10; /* last_position */
	T0* a11; /* equality_tester */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
struct S2062 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
struct S2063 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
struct S2064 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] STRING_32] */
struct S2065 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
struct S2066 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
struct S2067 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
struct S2068 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] LX_NFA */
struct S2069 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* in_trail_context */
	T0* a2; /* states */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_NFA] */
struct S2070 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable LX_RULE */
struct S2071 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* pattern */
	T0* a3; /* action */
	T6 a4; /* head_count */
	T6 a5; /* trail_count */
	T6 a6; /* line_count */
	T6 a7; /* column_count */
	T6 a8; /* line_nb */
	T1 a9; /* has_trail_context */
	T1 a10; /* is_useful */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
struct S2072 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] LX_EQUIVALENCE_CLASSES */
struct S2073 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* new_upper */
	T0* a2; /* storage */
};

/* Struct for type [detachable] LX_UNRECOGNIZED_RULE_ERROR */
struct S2074 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
struct S2075 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_MISSING_QUOTE_ERROR */
struct S2076 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
struct S2077 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_BAD_CHARACTER_ERROR */
struct S2078 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_META_ERROR */
struct S2079 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_REJECT_ERROR */
struct S2080 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
struct S2081 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
struct S2082 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] ARRAY [[attached] LX_RULE] */
struct S2083 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_RULE] */
struct S2084 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] LX_DFA_STATE */
struct S2085 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* accepted_rules */
	T0* a2; /* states */
	T0* a3; /* transitions */
	T0* a4; /* accepted_head_rules */
	T6 a5; /* code */
	T6 a6; /* id */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
struct S2086 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* equality_tester */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
struct S2087 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type detachable LX_SYMBOL_PARTITIONS */
struct S2088 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* symbols */
	T0* a2; /* storage */
	T6 a3; /* new_upper */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
struct S2089 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ARRAY [detachable LX_RULE] */
struct S2090 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable LX_RULE] */
struct S2091 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] LX_START_CONDITION */
struct S2092 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* id */
	T1 a3; /* is_exclusive */
	T0* a4; /* patterns */
	T0* a5; /* bol_patterns */
};

/* Struct for type [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
struct S2093 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* lower */
	T6 a2; /* upper */
	T6 a3; /* minimum_label */
	T6 a4; /* second_minimum_label */
	T6 a5; /* maximum_label */
	T0* a6; /* transitions */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
struct S2094 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
struct S2095 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* count */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* free_slot */
	T6 a8; /* last_position */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] LX_NFA_STATE */
struct S2096 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* in_trail_context */
	T0* a2; /* transition */
	T0* a3; /* epsilon_transition */
	T6 a4; /* id */
	T0* a5; /* accepted_rule */
};

/* Struct for type detachable ARRAY [detachable LX_SYMBOL_CLASS] */
struct S2097 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
struct S2098 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
struct S2099 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
struct S2100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
struct S2102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
struct S2103 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
struct S2104 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
struct S2105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
struct S2106 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_START_CONDITION] */
struct S2107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
struct S2108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
struct S2109 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
struct S2110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
struct S2111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
struct S2113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
struct S2114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
struct S2115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
struct S2117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
struct S2119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
struct S2120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
};

/* Struct for type [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
struct S2123 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* target */
};

/* Struct for type [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
struct S2124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* label */
	T0* a2; /* target */
};

/* Struct for type [detachable] LX_ACTION */
struct S2126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* text */
};

/* Struct for type [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
struct S2127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
struct S2128 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
	T0* a2; /* left */
	T0* a3; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
struct S2129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
struct S2130 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
struct S2132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_NFA_STATE] */
struct S2134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
struct S2136 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
struct S2137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_DFA_STATE] */
struct S2139 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
struct S2140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
struct S2141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_ARRAY [[attached] LX_RULE] */
struct S2142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
struct S2143 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
	T6 a17; /* count */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
struct S2144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
struct S2146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
struct S2150 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
struct S2153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2156 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2158 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
struct S2159 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_ARRAY [[attached] STRING_8] */
struct S2160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
struct S2162 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};


extern T1 GE_default1;
extern T2 GE_default2;
extern T3 GE_default3;
extern T4 GE_default4;
extern T5 GE_default5;
extern T6 GE_default6;
extern T7 GE_default7;
extern T8 GE_default8;
extern T9 GE_default9;
extern T10 GE_default10;
extern T11 GE_default11;
extern T12 GE_default12;
extern T13 GE_default13;
extern T14 GE_default14;
extern T15 GE_default15;
extern T16 GE_default16;
extern T17 GE_default17;
extern T18 GE_default18;
extern T20 GE_default20;
extern T21 GE_default21;
extern T25 GE_default25;
extern T26 GE_default26;
extern T27 GE_default27;
extern T28 GE_default28;
extern T29 GE_default29;
extern T30 GE_default30;
extern T31 GE_default31;
extern T32 GE_default32;
extern T33 GE_default33;
extern T34 GE_default34;
extern T35 GE_default35;
extern T36 GE_default36;
extern T38 GE_default38;
extern T39 GE_default39;
extern T40 GE_default40;
extern T41 GE_default41;
extern T43 GE_default43;
extern T44 GE_default44;
extern T45 GE_default45;
extern T46 GE_default46;
extern T47 GE_default47;
extern T51 GE_default51;
extern T53 GE_default53;
extern T54 GE_default54;
extern T55 GE_default55;
extern T56 GE_default56;
extern T57 GE_default57;
extern T59 GE_default59;
extern T60 GE_default60;
extern T61 GE_default61;
extern T62 GE_default62;
extern T65 GE_default65;
extern T66 GE_default66;
extern T67 GE_default67;
extern T69 GE_default69;
extern T70 GE_default70;
extern T71 GE_default71;
extern T72 GE_default72;
extern T73 GE_default73;
extern T75 GE_default75;
extern T76 GE_default76;
extern T79 GE_default79;
extern T81 GE_default81;
extern T82 GE_default82;
extern T83 GE_default83;
extern T84 GE_default84;
extern T87 GE_default87;
extern T88 GE_default88;
extern T91 GE_default91;
extern T93 GE_default93;
extern T95 GE_default95;
extern T97 GE_default97;
extern T98 GE_default98;
extern T99 GE_default99;
extern T100 GE_default100;
extern T101 GE_default101;
extern T102 GE_default102;
extern T103 GE_default103;
extern T105 GE_default105;
extern T106 GE_default106;
extern T107 GE_default107;
extern T108 GE_default108;
extern T109 GE_default109;
extern T110 GE_default110;
extern T111 GE_default111;
extern T112 GE_default112;
extern T113 GE_default113;
extern T114 GE_default114;
extern T115 GE_default115;
extern T116 GE_default116;
extern T117 GE_default117;
extern T118 GE_default118;
extern T119 GE_default119;
extern T120 GE_default120;
extern T121 GE_default121;
extern T122 GE_default122;
extern T123 GE_default123;
extern T124 GE_default124;
extern T125 GE_default125;
extern T126 GE_default126;
extern T127 GE_default127;
extern T129 GE_default129;
extern T131 GE_default131;
extern T132 GE_default132;
extern T133 GE_default133;
extern T134 GE_default134;
extern T135 GE_default135;
extern T136 GE_default136;
extern T137 GE_default137;
extern T138 GE_default138;
extern T140 GE_default140;
extern T141 GE_default141;
extern T142 GE_default142;
extern T143 GE_default143;
extern T144 GE_default144;
extern T145 GE_default145;
extern T146 GE_default146;
extern T147 GE_default147;
extern T148 GE_default148;
extern T149 GE_default149;
extern T150 GE_default150;
extern T153 GE_default153;
extern T156 GE_default156;
extern T157 GE_default157;
extern T158 GE_default158;
extern T159 GE_default159;
extern T160 GE_default160;
extern T161 GE_default161;
extern T162 GE_default162;
extern T163 GE_default163;
extern T164 GE_default164;
extern T165 GE_default165;
extern T166 GE_default166;
extern T167 GE_default167;
extern T168 GE_default168;
extern T170 GE_default170;
extern T171 GE_default171;
extern T172 GE_default172;
extern T174 GE_default174;
extern T175 GE_default175;
extern T178 GE_default178;
extern T179 GE_default179;
extern T180 GE_default180;
extern T181 GE_default181;
extern T183 GE_default183;
extern T184 GE_default184;
extern T186 GE_default186;
extern T187 GE_default187;
extern T188 GE_default188;
extern T189 GE_default189;
extern T191 GE_default191;
extern T192 GE_default192;
extern T194 GE_default194;
extern T195 GE_default195;
extern T196 GE_default196;
extern T197 GE_default197;
extern T198 GE_default198;
extern T199 GE_default199;
extern T200 GE_default200;
extern T201 GE_default201;
extern T202 GE_default202;
extern T203 GE_default203;
extern T204 GE_default204;
extern T206 GE_default206;
extern T207 GE_default207;
extern T209 GE_default209;
extern T210 GE_default210;
extern T211 GE_default211;
extern T212 GE_default212;
extern T213 GE_default213;
extern T214 GE_default214;
extern T215 GE_default215;
extern T217 GE_default217;
extern T219 GE_default219;
extern T220 GE_default220;
extern T221 GE_default221;
extern T222 GE_default222;
extern T223 GE_default223;
extern T224 GE_default224;
extern T225 GE_default225;
extern T226 GE_default226;
extern T227 GE_default227;
extern T232 GE_default232;
extern T236 GE_default236;
extern T237 GE_default237;
extern T246 GE_default246;
extern T247 GE_default247;
extern T248 GE_default248;
extern T249 GE_default249;
extern T250 GE_default250;
extern T251 GE_default251;
extern T252 GE_default252;
extern T253 GE_default253;
extern T254 GE_default254;
extern T255 GE_default255;
extern T256 GE_default256;
extern T257 GE_default257;
extern T258 GE_default258;
extern T259 GE_default259;
extern T260 GE_default260;
extern T261 GE_default261;
extern T262 GE_default262;
extern T263 GE_default263;
extern T264 GE_default264;
extern T265 GE_default265;
extern T266 GE_default266;
extern T267 GE_default267;
extern T268 GE_default268;
extern T269 GE_default269;
extern T270 GE_default270;
extern T271 GE_default271;
extern T273 GE_default273;
extern T274 GE_default274;
extern T275 GE_default275;
extern T278 GE_default278;
extern T279 GE_default279;
extern T280 GE_default280;
extern T281 GE_default281;
extern T282 GE_default282;
extern T283 GE_default283;
extern T284 GE_default284;
extern T285 GE_default285;
extern T286 GE_default286;
extern T287 GE_default287;
extern T288 GE_default288;
extern T289 GE_default289;
extern T290 GE_default290;
extern T291 GE_default291;
extern T293 GE_default293;
extern T294 GE_default294;
extern T295 GE_default295;
extern T296 GE_default296;
extern T297 GE_default297;
extern T298 GE_default298;
extern T299 GE_default299;
extern T300 GE_default300;
extern T301 GE_default301;
extern T302 GE_default302;
extern T306 GE_default306;
extern T308 GE_default308;
extern T309 GE_default309;
extern T310 GE_default310;
extern T311 GE_default311;
extern T314 GE_default314;
extern T315 GE_default315;
extern T316 GE_default316;
extern T320 GE_default320;
extern T321 GE_default321;
extern T324 GE_default324;
extern T325 GE_default325;
extern T327 GE_default327;
extern T328 GE_default328;
extern T329 GE_default329;
extern T330 GE_default330;
extern T331 GE_default331;
extern T335 GE_default335;
extern T342 GE_default342;
extern T343 GE_default343;
extern T344 GE_default344;
extern T345 GE_default345;
extern T346 GE_default346;
extern T347 GE_default347;
extern T348 GE_default348;
extern T349 GE_default349;
extern T352 GE_default352;
extern T354 GE_default354;
extern T356 GE_default356;
extern T357 GE_default357;
extern T361 GE_default361;
extern T363 GE_default363;
extern T365 GE_default365;
extern T366 GE_default366;
extern T367 GE_default367;
extern T370 GE_default370;
extern T371 GE_default371;
extern T372 GE_default372;
extern T373 GE_default373;
extern T375 GE_default375;
extern T376 GE_default376;
extern T377 GE_default377;
extern T378 GE_default378;
extern T380 GE_default380;
extern T382 GE_default382;
extern T383 GE_default383;
extern T384 GE_default384;
extern T385 GE_default385;
extern T386 GE_default386;
extern T387 GE_default387;
extern T388 GE_default388;
extern T389 GE_default389;
extern T390 GE_default390;
extern T391 GE_default391;
extern T392 GE_default392;
extern T393 GE_default393;
extern T394 GE_default394;
extern T395 GE_default395;
extern T396 GE_default396;
extern T397 GE_default397;
extern T398 GE_default398;
extern T399 GE_default399;
extern T400 GE_default400;
extern T401 GE_default401;
extern T402 GE_default402;
extern T404 GE_default404;
extern T406 GE_default406;
extern T408 GE_default408;
extern T409 GE_default409;
extern T410 GE_default410;
extern T411 GE_default411;
extern T412 GE_default412;
extern T413 GE_default413;
extern T416 GE_default416;
extern T417 GE_default417;
extern T419 GE_default419;
extern T420 GE_default420;
extern T421 GE_default421;
extern T423 GE_default423;
extern T424 GE_default424;
extern T425 GE_default425;
extern T426 GE_default426;
extern T427 GE_default427;
extern T428 GE_default428;
extern T431 GE_default431;
extern T437 GE_default437;
extern T438 GE_default438;
extern T439 GE_default439;
extern T440 GE_default440;
extern T442 GE_default442;
extern T443 GE_default443;
extern T444 GE_default444;
extern T445 GE_default445;
extern T446 GE_default446;
extern T447 GE_default447;
extern T448 GE_default448;
extern T449 GE_default449;
extern T450 GE_default450;
extern T451 GE_default451;
extern T452 GE_default452;
extern T453 GE_default453;
extern T454 GE_default454;
extern T455 GE_default455;
extern T456 GE_default456;
extern T457 GE_default457;
extern T458 GE_default458;
extern T460 GE_default460;
extern T461 GE_default461;
extern T462 GE_default462;
extern T463 GE_default463;
extern T466 GE_default466;
extern T468 GE_default468;
extern T469 GE_default469;
extern T470 GE_default470;
extern T471 GE_default471;
extern T472 GE_default472;
extern T473 GE_default473;
extern T475 GE_default475;
extern T476 GE_default476;
extern T477 GE_default477;
extern T478 GE_default478;
extern T479 GE_default479;
extern T481 GE_default481;
extern T482 GE_default482;
extern T483 GE_default483;
extern T484 GE_default484;
extern T485 GE_default485;
extern T486 GE_default486;
extern T487 GE_default487;
extern T488 GE_default488;
extern T489 GE_default489;
extern T490 GE_default490;
extern T491 GE_default491;
extern T492 GE_default492;
extern T493 GE_default493;
extern T494 GE_default494;
extern T496 GE_default496;
extern T497 GE_default497;
extern T500 GE_default500;
extern T501 GE_default501;
extern T504 GE_default504;
extern T505 GE_default505;
extern T506 GE_default506;
extern T507 GE_default507;
extern T509 GE_default509;
extern T510 GE_default510;
extern T511 GE_default511;
extern T512 GE_default512;
extern T514 GE_default514;
extern T516 GE_default516;
extern T517 GE_default517;
extern T518 GE_default518;
extern T519 GE_default519;
extern T520 GE_default520;
extern T521 GE_default521;
extern T522 GE_default522;
extern T523 GE_default523;
extern T524 GE_default524;
extern T525 GE_default525;
extern T526 GE_default526;
extern T527 GE_default527;
extern T528 GE_default528;
extern T529 GE_default529;
extern T530 GE_default530;
extern T531 GE_default531;
extern T532 GE_default532;
extern T533 GE_default533;
extern T534 GE_default534;
extern T535 GE_default535;
extern T536 GE_default536;
extern T537 GE_default537;
extern T538 GE_default538;
extern T539 GE_default539;
extern T540 GE_default540;
extern T541 GE_default541;
extern T542 GE_default542;
extern T543 GE_default543;
extern T546 GE_default546;
extern T547 GE_default547;
extern T549 GE_default549;
extern T550 GE_default550;
extern T551 GE_default551;
extern T552 GE_default552;
extern T553 GE_default553;
extern T554 GE_default554;
extern T555 GE_default555;
extern T557 GE_default557;
extern T558 GE_default558;
extern T559 GE_default559;
extern T560 GE_default560;
extern T561 GE_default561;
extern T562 GE_default562;
extern T563 GE_default563;
extern T564 GE_default564;
extern T567 GE_default567;
extern T568 GE_default568;
extern T569 GE_default569;
extern T572 GE_default572;
extern T574 GE_default574;
extern T575 GE_default575;
extern T576 GE_default576;
extern T577 GE_default577;
extern T579 GE_default579;
extern T581 GE_default581;
extern T582 GE_default582;
extern T583 GE_default583;
extern T584 GE_default584;
extern T585 GE_default585;
extern T586 GE_default586;
extern T589 GE_default589;
extern T590 GE_default590;
extern T591 GE_default591;
extern T592 GE_default592;
extern T593 GE_default593;
extern T594 GE_default594;
extern T595 GE_default595;
extern T596 GE_default596;
extern T599 GE_default599;
extern T600 GE_default600;
extern T601 GE_default601;
extern T602 GE_default602;
extern T603 GE_default603;
extern T605 GE_default605;
extern T609 GE_default609;
extern T610 GE_default610;
extern T612 GE_default612;
extern T613 GE_default613;
extern T614 GE_default614;
extern T615 GE_default615;
extern T616 GE_default616;
extern T617 GE_default617;
extern T618 GE_default618;
extern T619 GE_default619;
extern T620 GE_default620;
extern T622 GE_default622;
extern T623 GE_default623;
extern T624 GE_default624;
extern T625 GE_default625;
extern T626 GE_default626;
extern T627 GE_default627;
extern T628 GE_default628;
extern T629 GE_default629;
extern T631 GE_default631;
extern T632 GE_default632;
extern T633 GE_default633;
extern T634 GE_default634;
extern T635 GE_default635;
extern T636 GE_default636;
extern T637 GE_default637;
extern T638 GE_default638;
extern T639 GE_default639;
extern T640 GE_default640;
extern T641 GE_default641;
extern T642 GE_default642;
extern T643 GE_default643;
extern T645 GE_default645;
extern T646 GE_default646;
extern T647 GE_default647;
extern T648 GE_default648;
extern T649 GE_default649;
extern T650 GE_default650;
extern T651 GE_default651;
extern T653 GE_default653;
extern T654 GE_default654;
extern T655 GE_default655;
extern T656 GE_default656;
extern T657 GE_default657;
extern T659 GE_default659;
extern T660 GE_default660;
extern T661 GE_default661;
extern T663 GE_default663;
extern T664 GE_default664;
extern T666 GE_default666;
extern T667 GE_default667;
extern T668 GE_default668;
extern T669 GE_default669;
extern T672 GE_default672;
extern T674 GE_default674;
extern T675 GE_default675;
extern T676 GE_default676;
extern T677 GE_default677;
extern T680 GE_default680;
extern T681 GE_default681;
extern T682 GE_default682;
extern T686 GE_default686;
extern T687 GE_default687;
extern T688 GE_default688;
extern T689 GE_default689;
extern T690 GE_default690;
extern T691 GE_default691;
extern T692 GE_default692;
extern T693 GE_default693;
extern T694 GE_default694;
extern T695 GE_default695;
extern T696 GE_default696;
extern T697 GE_default697;
extern T698 GE_default698;
extern T699 GE_default699;
extern T700 GE_default700;
extern T701 GE_default701;
extern T702 GE_default702;
extern T703 GE_default703;
extern T704 GE_default704;
extern T705 GE_default705;
extern T706 GE_default706;
extern T707 GE_default707;
extern T708 GE_default708;
extern T709 GE_default709;
extern T710 GE_default710;
extern T711 GE_default711;
extern T712 GE_default712;
extern T715 GE_default715;
extern T716 GE_default716;
extern T717 GE_default717;
extern T718 GE_default718;
extern T719 GE_default719;
extern T721 GE_default721;
extern T722 GE_default722;
extern T723 GE_default723;
extern T724 GE_default724;
extern T726 GE_default726;
extern T727 GE_default727;
extern T728 GE_default728;
extern T729 GE_default729;
extern T731 GE_default731;
extern T732 GE_default732;
extern T733 GE_default733;
extern T734 GE_default734;
extern T735 GE_default735;
extern T736 GE_default736;
extern T737 GE_default737;
extern T738 GE_default738;
extern T739 GE_default739;
extern T741 GE_default741;
extern T742 GE_default742;
extern T743 GE_default743;
extern T745 GE_default745;
extern T746 GE_default746;
extern T747 GE_default747;
extern T748 GE_default748;
extern T749 GE_default749;
extern T750 GE_default750;
extern T751 GE_default751;
extern T752 GE_default752;
extern T753 GE_default753;
extern T755 GE_default755;
extern T756 GE_default756;
extern T757 GE_default757;
extern T758 GE_default758;
extern T759 GE_default759;
extern T760 GE_default760;
extern T761 GE_default761;
extern T762 GE_default762;
extern T763 GE_default763;
extern T764 GE_default764;
extern T765 GE_default765;
extern T766 GE_default766;
extern T767 GE_default767;
extern T768 GE_default768;
extern T769 GE_default769;
extern T770 GE_default770;
extern T771 GE_default771;
extern T772 GE_default772;
extern T773 GE_default773;
extern T775 GE_default775;
extern T776 GE_default776;
extern T777 GE_default777;
extern T778 GE_default778;
extern T779 GE_default779;
extern T780 GE_default780;
extern T781 GE_default781;
extern T782 GE_default782;
extern T783 GE_default783;
extern T784 GE_default784;
extern T785 GE_default785;
extern T786 GE_default786;
extern T787 GE_default787;
extern T788 GE_default788;
extern T789 GE_default789;
extern T790 GE_default790;
extern T791 GE_default791;
extern T792 GE_default792;
extern T793 GE_default793;
extern T794 GE_default794;
extern T796 GE_default796;
extern T797 GE_default797;
extern T798 GE_default798;
extern T799 GE_default799;
extern T800 GE_default800;
extern T801 GE_default801;
extern T802 GE_default802;
extern T804 GE_default804;
extern T805 GE_default805;
extern T807 GE_default807;
extern T808 GE_default808;
extern T809 GE_default809;
extern T810 GE_default810;
extern T811 GE_default811;
extern T812 GE_default812;
extern T813 GE_default813;
extern T814 GE_default814;
extern T816 GE_default816;
extern T817 GE_default817;
extern T818 GE_default818;
extern T819 GE_default819;
extern T820 GE_default820;
extern T821 GE_default821;
extern T822 GE_default822;
extern T823 GE_default823;
extern T824 GE_default824;
extern T825 GE_default825;
extern T826 GE_default826;
extern T827 GE_default827;
extern T828 GE_default828;
extern T829 GE_default829;
extern T831 GE_default831;
extern T832 GE_default832;
extern T833 GE_default833;
extern T834 GE_default834;
extern T835 GE_default835;
extern T836 GE_default836;
extern T837 GE_default837;
extern T838 GE_default838;
extern T839 GE_default839;
extern T840 GE_default840;
extern T841 GE_default841;
extern T842 GE_default842;
extern T843 GE_default843;
extern T844 GE_default844;
extern T846 GE_default846;
extern T847 GE_default847;
extern T848 GE_default848;
extern T849 GE_default849;
extern T850 GE_default850;
extern T851 GE_default851;
extern T854 GE_default854;
extern T858 GE_default858;
extern T859 GE_default859;
extern T860 GE_default860;
extern T861 GE_default861;
extern T862 GE_default862;
extern T863 GE_default863;
extern T864 GE_default864;
extern T866 GE_default866;
extern T867 GE_default867;
extern T868 GE_default868;
extern T869 GE_default869;
extern T871 GE_default871;
extern T873 GE_default873;
extern T874 GE_default874;
extern T875 GE_default875;
extern T876 GE_default876;
extern T877 GE_default877;
extern T878 GE_default878;
extern T879 GE_default879;
extern T880 GE_default880;
extern T881 GE_default881;
extern T882 GE_default882;
extern T884 GE_default884;
extern T885 GE_default885;
extern T886 GE_default886;
extern T887 GE_default887;
extern T888 GE_default888;
extern T889 GE_default889;
extern T896 GE_default896;
extern T897 GE_default897;
extern T899 GE_default899;
extern T900 GE_default900;
extern T901 GE_default901;
extern T902 GE_default902;
extern T903 GE_default903;
extern T904 GE_default904;
extern T905 GE_default905;
extern T906 GE_default906;
extern T907 GE_default907;
extern T908 GE_default908;
extern T909 GE_default909;
extern T910 GE_default910;
extern T914 GE_default914;
extern T915 GE_default915;
extern T916 GE_default916;
extern T917 GE_default917;
extern T918 GE_default918;
extern T919 GE_default919;
extern T920 GE_default920;
extern T921 GE_default921;
extern T922 GE_default922;
extern T923 GE_default923;
extern T924 GE_default924;
extern T925 GE_default925;
extern T926 GE_default926;
extern T927 GE_default927;
extern T928 GE_default928;
extern T929 GE_default929;
extern T930 GE_default930;
extern T931 GE_default931;
extern T932 GE_default932;
extern T933 GE_default933;
extern T934 GE_default934;
extern T935 GE_default935;
extern T936 GE_default936;
extern T937 GE_default937;
extern T938 GE_default938;
extern T939 GE_default939;
extern T941 GE_default941;
extern T942 GE_default942;
extern T944 GE_default944;
extern T945 GE_default945;
extern T947 GE_default947;
extern T948 GE_default948;
extern T949 GE_default949;
extern T950 GE_default950;
extern T951 GE_default951;
extern T952 GE_default952;
extern T953 GE_default953;
extern T954 GE_default954;
extern T955 GE_default955;
extern T957 GE_default957;
extern T958 GE_default958;
extern T959 GE_default959;
extern T960 GE_default960;
extern T961 GE_default961;
extern T962 GE_default962;
extern T963 GE_default963;
extern T965 GE_default965;
extern T966 GE_default966;
extern T967 GE_default967;
extern T968 GE_default968;
extern T969 GE_default969;
extern T970 GE_default970;
extern T971 GE_default971;
extern T972 GE_default972;
extern T973 GE_default973;
extern T974 GE_default974;
extern T975 GE_default975;
extern T976 GE_default976;
extern T977 GE_default977;
extern T978 GE_default978;
extern T979 GE_default979;
extern T980 GE_default980;
extern T981 GE_default981;
extern T982 GE_default982;
extern T983 GE_default983;
extern T984 GE_default984;
extern T985 GE_default985;
extern T986 GE_default986;
extern T987 GE_default987;
extern T988 GE_default988;
extern T989 GE_default989;
extern T990 GE_default990;
extern T991 GE_default991;
extern T992 GE_default992;
extern T993 GE_default993;
extern T994 GE_default994;
extern T995 GE_default995;
extern T996 GE_default996;
extern T997 GE_default997;
extern T998 GE_default998;
extern T999 GE_default999;
extern T1001 GE_default1001;
extern T1002 GE_default1002;
extern T1003 GE_default1003;
extern T1004 GE_default1004;
extern T1005 GE_default1005;
extern T1006 GE_default1006;
extern T1007 GE_default1007;
extern T1008 GE_default1008;
extern T1009 GE_default1009;
extern T1011 GE_default1011;
extern T1012 GE_default1012;
extern T1013 GE_default1013;
extern T1014 GE_default1014;
extern T1015 GE_default1015;
extern T1016 GE_default1016;
extern T1017 GE_default1017;
extern T1018 GE_default1018;
extern T1019 GE_default1019;
extern T1020 GE_default1020;
extern T1021 GE_default1021;
extern T1023 GE_default1023;
extern T1024 GE_default1024;
extern T1026 GE_default1026;
extern T1027 GE_default1027;
extern T1029 GE_default1029;
extern T1030 GE_default1030;
extern T1031 GE_default1031;
extern T1032 GE_default1032;
extern T1033 GE_default1033;
extern T1034 GE_default1034;
extern T1035 GE_default1035;
extern T1036 GE_default1036;
extern T1037 GE_default1037;
extern T1038 GE_default1038;
extern T1039 GE_default1039;
extern T1040 GE_default1040;
extern T1041 GE_default1041;
extern T1042 GE_default1042;
extern T1044 GE_default1044;
extern T1045 GE_default1045;
extern T1046 GE_default1046;
extern T1047 GE_default1047;
extern T1049 GE_default1049;
extern T1050 GE_default1050;
extern T1051 GE_default1051;
extern T1052 GE_default1052;
extern T1053 GE_default1053;
extern T1054 GE_default1054;
extern T1055 GE_default1055;
extern T1056 GE_default1056;
extern T1057 GE_default1057;
extern T1058 GE_default1058;
extern T1059 GE_default1059;
extern T1061 GE_default1061;
extern T1062 GE_default1062;
extern T1064 GE_default1064;
extern T1065 GE_default1065;
extern T1067 GE_default1067;
extern T1068 GE_default1068;
extern T1069 GE_default1069;
extern T1070 GE_default1070;
extern T1071 GE_default1071;
extern T1072 GE_default1072;
extern T1073 GE_default1073;
extern T1074 GE_default1074;
extern T1075 GE_default1075;
extern T1076 GE_default1076;
extern T1077 GE_default1077;
extern T1078 GE_default1078;
extern T1079 GE_default1079;
extern T1080 GE_default1080;
extern T1081 GE_default1081;
extern T1082 GE_default1082;
extern T1083 GE_default1083;
extern T1084 GE_default1084;
extern T1085 GE_default1085;
extern T1086 GE_default1086;
extern T1087 GE_default1087;
extern T1088 GE_default1088;
extern T1089 GE_default1089;
extern T1090 GE_default1090;
extern T1091 GE_default1091;
extern T1092 GE_default1092;
extern T1093 GE_default1093;
extern T1094 GE_default1094;
extern T1095 GE_default1095;
extern T1096 GE_default1096;
extern T1097 GE_default1097;
extern T1099 GE_default1099;
extern T1100 GE_default1100;
extern T1101 GE_default1101;
extern T1102 GE_default1102;
extern T1103 GE_default1103;
extern T1104 GE_default1104;
extern T1105 GE_default1105;
extern T1106 GE_default1106;
extern T1107 GE_default1107;
extern T1108 GE_default1108;
extern T1110 GE_default1110;
extern T1111 GE_default1111;
extern T1112 GE_default1112;
extern T1113 GE_default1113;
extern T1114 GE_default1114;
extern T1115 GE_default1115;
extern T1116 GE_default1116;
extern T1117 GE_default1117;
extern T1118 GE_default1118;
extern T1119 GE_default1119;
extern T1120 GE_default1120;
extern T1121 GE_default1121;
extern T1122 GE_default1122;
extern T1124 GE_default1124;
extern T1125 GE_default1125;
extern T1126 GE_default1126;
extern T1127 GE_default1127;
extern T1129 GE_default1129;
extern T1130 GE_default1130;
extern T1131 GE_default1131;
extern T1132 GE_default1132;
extern T1133 GE_default1133;
extern T1134 GE_default1134;
extern T1136 GE_default1136;
extern T1137 GE_default1137;
extern T1138 GE_default1138;
extern T1139 GE_default1139;
extern T1140 GE_default1140;
extern T1141 GE_default1141;
extern T1142 GE_default1142;
extern T1143 GE_default1143;
extern T1144 GE_default1144;
extern T1145 GE_default1145;
extern T1146 GE_default1146;
extern T1147 GE_default1147;
extern T1148 GE_default1148;
extern T1150 GE_default1150;
extern T1151 GE_default1151;
extern T1152 GE_default1152;
extern T1153 GE_default1153;
extern T1155 GE_default1155;
extern T1156 GE_default1156;
extern T1157 GE_default1157;
extern T1158 GE_default1158;
extern T1159 GE_default1159;
extern T1160 GE_default1160;
extern T1161 GE_default1161;
extern T1162 GE_default1162;
extern T1163 GE_default1163;
extern T1164 GE_default1164;
extern T1166 GE_default1166;
extern T1167 GE_default1167;
extern T1168 GE_default1168;
extern T1169 GE_default1169;
extern T1170 GE_default1170;
extern T1171 GE_default1171;
extern T1172 GE_default1172;
extern T1174 GE_default1174;
extern T1175 GE_default1175;
extern T1176 GE_default1176;
extern T1177 GE_default1177;
extern T1178 GE_default1178;
extern T1180 GE_default1180;
extern T1181 GE_default1181;
extern T1182 GE_default1182;
extern T1183 GE_default1183;
extern T1184 GE_default1184;
extern T1185 GE_default1185;
extern T1186 GE_default1186;
extern T1187 GE_default1187;
extern T1188 GE_default1188;
extern T1189 GE_default1189;
extern T1191 GE_default1191;
extern T1192 GE_default1192;
extern T1193 GE_default1193;
extern T1194 GE_default1194;
extern T1195 GE_default1195;
extern T1196 GE_default1196;
extern T1197 GE_default1197;
extern T1199 GE_default1199;
extern T1200 GE_default1200;
extern T1201 GE_default1201;
extern T1202 GE_default1202;
extern T1203 GE_default1203;
extern T1204 GE_default1204;
extern T1205 GE_default1205;
extern T1206 GE_default1206;
extern T1207 GE_default1207;
extern T1208 GE_default1208;
extern T1209 GE_default1209;
extern T1210 GE_default1210;
extern T1211 GE_default1211;
extern T1212 GE_default1212;
extern T1213 GE_default1213;
extern T1214 GE_default1214;
extern T1215 GE_default1215;
extern T1216 GE_default1216;
extern T1217 GE_default1217;
extern T1218 GE_default1218;
extern T1219 GE_default1219;
extern T1221 GE_default1221;
extern T1222 GE_default1222;
extern T1223 GE_default1223;
extern T1224 GE_default1224;
extern T1226 GE_default1226;
extern T1227 GE_default1227;
extern T1228 GE_default1228;
extern T1229 GE_default1229;
extern T1230 GE_default1230;
extern T1231 GE_default1231;
extern T1232 GE_default1232;
extern T1233 GE_default1233;
extern T1234 GE_default1234;
extern T1235 GE_default1235;
extern T1236 GE_default1236;
extern T1237 GE_default1237;
extern T1239 GE_default1239;
extern T1240 GE_default1240;
extern T1242 GE_default1242;
extern T1243 GE_default1243;
extern T1244 GE_default1244;
extern T1245 GE_default1245;
extern T1246 GE_default1246;
extern T1248 GE_default1248;
extern T1249 GE_default1249;
extern T1251 GE_default1251;
extern T1252 GE_default1252;
extern T1253 GE_default1253;
extern T1254 GE_default1254;
extern T1255 GE_default1255;
extern T1256 GE_default1256;
extern T1257 GE_default1257;
extern T1258 GE_default1258;
extern T1259 GE_default1259;
extern T1260 GE_default1260;
extern T1261 GE_default1261;
extern T1262 GE_default1262;
extern T1263 GE_default1263;
extern T1264 GE_default1264;
extern T1265 GE_default1265;
extern T1266 GE_default1266;
extern T1267 GE_default1267;
extern T1268 GE_default1268;
extern T1270 GE_default1270;
extern T1271 GE_default1271;
extern T1273 GE_default1273;
extern T1274 GE_default1274;
extern T1275 GE_default1275;
extern T1276 GE_default1276;
extern T1277 GE_default1277;
extern T1278 GE_default1278;
extern T1279 GE_default1279;
extern T1280 GE_default1280;
extern T1281 GE_default1281;
extern T1282 GE_default1282;
extern T1283 GE_default1283;
extern T1284 GE_default1284;
extern T1285 GE_default1285;
extern T1286 GE_default1286;
extern T1287 GE_default1287;
extern T1288 GE_default1288;
extern T1289 GE_default1289;
extern T1290 GE_default1290;
extern T1295 GE_default1295;
extern T1296 GE_default1296;
extern T1297 GE_default1297;
extern T1298 GE_default1298;
extern T1299 GE_default1299;
extern T1302 GE_default1302;
extern T1303 GE_default1303;
extern T1304 GE_default1304;
extern T1305 GE_default1305;
extern T1306 GE_default1306;
extern T1307 GE_default1307;
extern T1308 GE_default1308;
extern T1309 GE_default1309;
extern T1311 GE_default1311;
extern T1312 GE_default1312;
extern T1313 GE_default1313;
extern T1314 GE_default1314;
extern T1316 GE_default1316;
extern T1317 GE_default1317;
extern T1318 GE_default1318;
extern T1319 GE_default1319;
extern T1322 GE_default1322;
extern T1323 GE_default1323;
extern T1326 GE_default1326;
extern T1327 GE_default1327;
extern T1328 GE_default1328;
extern T1329 GE_default1329;
extern T1330 GE_default1330;
extern T1331 GE_default1331;
extern T1332 GE_default1332;
extern T1333 GE_default1333;
extern T1336 GE_default1336;
extern T1337 GE_default1337;
extern T1338 GE_default1338;
extern T1339 GE_default1339;
extern T1340 GE_default1340;
extern T1341 GE_default1341;
extern T1342 GE_default1342;
extern T1344 GE_default1344;
extern T1345 GE_default1345;
extern T1346 GE_default1346;
extern T1347 GE_default1347;
extern T1348 GE_default1348;
extern T1349 GE_default1349;
extern T1350 GE_default1350;
extern T1351 GE_default1351;
extern T1353 GE_default1353;
extern T1354 GE_default1354;
extern T1355 GE_default1355;
extern T1356 GE_default1356;
extern T1358 GE_default1358;
extern T1359 GE_default1359;
extern T1360 GE_default1360;
extern T1361 GE_default1361;
extern T1362 GE_default1362;
extern T1364 GE_default1364;
extern T1365 GE_default1365;
extern T1366 GE_default1366;
extern T1367 GE_default1367;
extern T1368 GE_default1368;
extern T1370 GE_default1370;
extern T1372 GE_default1372;
extern T1373 GE_default1373;
extern T1374 GE_default1374;
extern T1375 GE_default1375;
extern T1376 GE_default1376;
extern T1377 GE_default1377;
extern T1378 GE_default1378;
extern T1382 GE_default1382;
extern T1383 GE_default1383;
extern T1385 GE_default1385;
extern T1386 GE_default1386;
extern T1393 GE_default1393;
extern T1396 GE_default1396;
extern T1398 GE_default1398;
extern T1399 GE_default1399;
extern T1400 GE_default1400;
extern T1401 GE_default1401;
extern T1402 GE_default1402;
extern T1403 GE_default1403;
extern T1405 GE_default1405;
extern T1406 GE_default1406;
extern T1407 GE_default1407;
extern T1408 GE_default1408;
extern T1409 GE_default1409;
extern T1410 GE_default1410;
extern T1411 GE_default1411;
extern T1413 GE_default1413;
extern T1414 GE_default1414;
extern T1415 GE_default1415;
extern T1416 GE_default1416;
extern T1418 GE_default1418;
extern T1419 GE_default1419;
extern T1420 GE_default1420;
extern T1421 GE_default1421;
extern T1424 GE_default1424;
extern T1427 GE_default1427;
extern T1429 GE_default1429;
extern T1431 GE_default1431;
extern T1432 GE_default1432;
extern T1433 GE_default1433;
extern T1434 GE_default1434;
extern T1435 GE_default1435;
extern T1436 GE_default1436;
extern T1437 GE_default1437;
extern T1438 GE_default1438;
extern T1439 GE_default1439;
extern T1440 GE_default1440;
extern T1442 GE_default1442;
extern T1443 GE_default1443;
extern T1444 GE_default1444;
extern T1445 GE_default1445;
extern T1447 GE_default1447;
extern T1448 GE_default1448;
extern T1449 GE_default1449;
extern T1450 GE_default1450;
extern T1451 GE_default1451;
extern T1452 GE_default1452;
extern T1453 GE_default1453;
extern T1454 GE_default1454;
extern T1455 GE_default1455;
extern T1457 GE_default1457;
extern T1458 GE_default1458;
extern T1460 GE_default1460;
extern T1461 GE_default1461;
extern T1462 GE_default1462;
extern T1464 GE_default1464;
extern T1465 GE_default1465;
extern T1466 GE_default1466;
extern T1467 GE_default1467;
extern T1468 GE_default1468;
extern T1469 GE_default1469;
extern T1470 GE_default1470;
extern T1473 GE_default1473;
extern T1474 GE_default1474;
extern T1476 GE_default1476;
extern T1477 GE_default1477;
extern T1478 GE_default1478;
extern T1479 GE_default1479;
extern T1480 GE_default1480;
extern T1481 GE_default1481;
extern T1482 GE_default1482;
extern T1484 GE_default1484;
extern T1485 GE_default1485;
extern T1486 GE_default1486;
extern T1487 GE_default1487;
extern T1489 GE_default1489;
extern T1490 GE_default1490;
extern T1491 GE_default1491;
extern T1492 GE_default1492;
extern T1493 GE_default1493;
extern T1494 GE_default1494;
extern T1495 GE_default1495;
extern T1496 GE_default1496;
extern T1497 GE_default1497;
extern T1498 GE_default1498;
extern T1499 GE_default1499;
extern T1500 GE_default1500;
extern T1501 GE_default1501;
extern T1502 GE_default1502;
extern T1503 GE_default1503;
extern T1504 GE_default1504;
extern T1505 GE_default1505;
extern T1506 GE_default1506;
extern T1507 GE_default1507;
extern T1508 GE_default1508;
extern T1509 GE_default1509;
extern T1510 GE_default1510;
extern T1513 GE_default1513;
extern T1515 GE_default1515;
extern T1517 GE_default1517;
extern T1518 GE_default1518;
extern T1520 GE_default1520;
extern T1522 GE_default1522;
extern T1524 GE_default1524;
extern T1527 GE_default1527;
extern T1528 GE_default1528;
extern T1529 GE_default1529;
extern T1530 GE_default1530;
extern T1531 GE_default1531;
extern T1532 GE_default1532;
extern T1533 GE_default1533;
extern T1534 GE_default1534;
extern T1535 GE_default1535;
extern T1536 GE_default1536;
extern T1537 GE_default1537;
extern T1538 GE_default1538;
extern T1539 GE_default1539;
extern T1540 GE_default1540;
extern T1541 GE_default1541;
extern T1542 GE_default1542;
extern T1545 GE_default1545;
extern T1546 GE_default1546;
extern T1547 GE_default1547;
extern T1548 GE_default1548;
extern T1549 GE_default1549;
extern T1550 GE_default1550;
extern T1551 GE_default1551;
extern T1552 GE_default1552;
extern T1553 GE_default1553;
extern T1554 GE_default1554;
extern T1556 GE_default1556;
extern T1557 GE_default1557;
extern T1559 GE_default1559;
extern T1560 GE_default1560;
extern T1562 GE_default1562;
extern T1563 GE_default1563;
extern T1564 GE_default1564;
extern T1565 GE_default1565;
extern T1566 GE_default1566;
extern T1567 GE_default1567;
extern T1568 GE_default1568;
extern T1569 GE_default1569;
extern T1570 GE_default1570;
extern T1571 GE_default1571;
extern T1573 GE_default1573;
extern T1574 GE_default1574;
extern T1575 GE_default1575;
extern T1576 GE_default1576;
extern T1577 GE_default1577;
extern T1578 GE_default1578;
extern T1579 GE_default1579;
extern T1580 GE_default1580;
extern T1581 GE_default1581;
extern T1582 GE_default1582;
extern T1583 GE_default1583;
extern T1584 GE_default1584;
extern T1585 GE_default1585;
extern T1586 GE_default1586;
extern T1587 GE_default1587;
extern T1588 GE_default1588;
extern T1589 GE_default1589;
extern T1590 GE_default1590;
extern T1591 GE_default1591;
extern T1592 GE_default1592;
extern T1593 GE_default1593;
extern T1594 GE_default1594;
extern T1595 GE_default1595;
extern T1596 GE_default1596;
extern T1597 GE_default1597;
extern T1598 GE_default1598;
extern T1600 GE_default1600;
extern T1602 GE_default1602;
extern T1603 GE_default1603;
extern T1606 GE_default1606;
extern T1607 GE_default1607;
extern T1608 GE_default1608;
extern T1609 GE_default1609;
extern T1610 GE_default1610;
extern T1613 GE_default1613;
extern T1614 GE_default1614;
extern T1615 GE_default1615;
extern T1617 GE_default1617;
extern T1618 GE_default1618;
extern T1619 GE_default1619;
extern T1620 GE_default1620;
extern T1621 GE_default1621;
extern T1622 GE_default1622;
extern T1624 GE_default1624;
extern T1625 GE_default1625;
extern T1626 GE_default1626;
extern T1627 GE_default1627;
extern T1628 GE_default1628;
extern T1629 GE_default1629;
extern T1630 GE_default1630;
extern T1632 GE_default1632;
extern T1633 GE_default1633;
extern T1634 GE_default1634;
extern T1635 GE_default1635;
extern T1636 GE_default1636;
extern T1638 GE_default1638;
extern T1640 GE_default1640;
extern T1641 GE_default1641;
extern T1642 GE_default1642;
extern T1643 GE_default1643;
extern T1644 GE_default1644;
extern T1645 GE_default1645;
extern T1646 GE_default1646;
extern T1649 GE_default1649;
extern T1651 GE_default1651;
extern T1652 GE_default1652;
extern T1653 GE_default1653;
extern T1656 GE_default1656;
extern T1657 GE_default1657;
extern T1660 GE_default1660;
extern T1661 GE_default1661;
extern T1662 GE_default1662;
extern T1663 GE_default1663;
extern T1664 GE_default1664;
extern T1665 GE_default1665;
extern T1667 GE_default1667;
extern T1668 GE_default1668;
extern T1670 GE_default1670;
extern T1671 GE_default1671;
extern T1672 GE_default1672;
extern T1673 GE_default1673;
extern T1674 GE_default1674;
extern T1675 GE_default1675;
extern T1676 GE_default1676;
extern T1677 GE_default1677;
extern T1678 GE_default1678;
extern T1679 GE_default1679;
extern T1680 GE_default1680;
extern T1681 GE_default1681;
extern T1682 GE_default1682;
extern T1684 GE_default1684;
extern T1685 GE_default1685;
extern T1686 GE_default1686;
extern T1687 GE_default1687;
extern T1690 GE_default1690;
extern T1691 GE_default1691;
extern T1692 GE_default1692;
extern T1693 GE_default1693;
extern T1698 GE_default1698;
extern T1703 GE_default1703;
extern T1704 GE_default1704;
extern T1706 GE_default1706;
extern T1707 GE_default1707;
extern T1710 GE_default1710;
extern T1714 GE_default1714;
extern T1715 GE_default1715;
extern T1716 GE_default1716;
extern T1717 GE_default1717;
extern T1719 GE_default1719;
extern T1724 GE_default1724;
extern T1725 GE_default1725;
extern T1727 GE_default1727;
extern T1728 GE_default1728;
extern T1730 GE_default1730;
extern T1731 GE_default1731;
extern T1732 GE_default1732;
extern T1733 GE_default1733;
extern T1736 GE_default1736;
extern T1737 GE_default1737;
extern T1738 GE_default1738;
extern T1739 GE_default1739;
extern T1740 GE_default1740;
extern T1741 GE_default1741;
extern T1742 GE_default1742;
extern T1743 GE_default1743;
extern T1744 GE_default1744;
extern T1745 GE_default1745;
extern T1746 GE_default1746;
extern T1747 GE_default1747;
extern T1748 GE_default1748;
extern T1749 GE_default1749;
extern T1750 GE_default1750;
extern T1755 GE_default1755;
extern T1758 GE_default1758;
extern T1759 GE_default1759;
extern T1760 GE_default1760;
extern T1762 GE_default1762;
extern T1763 GE_default1763;
extern T1764 GE_default1764;
extern T1766 GE_default1766;
extern T1768 GE_default1768;
extern T1769 GE_default1769;
extern T1770 GE_default1770;
extern T1771 GE_default1771;
extern T1772 GE_default1772;
extern T1773 GE_default1773;
extern T1774 GE_default1774;
extern T1775 GE_default1775;
extern T1776 GE_default1776;
extern T1777 GE_default1777;
extern T1779 GE_default1779;
extern T1780 GE_default1780;
extern T1781 GE_default1781;
extern T1782 GE_default1782;
extern T1784 GE_default1784;
extern T1785 GE_default1785;
extern T1786 GE_default1786;
extern T1787 GE_default1787;
extern T1789 GE_default1789;
extern T1790 GE_default1790;
extern T1791 GE_default1791;
extern T1792 GE_default1792;
extern T1794 GE_default1794;
extern T1797 GE_default1797;
extern T1802 GE_default1802;
extern T1804 GE_default1804;
extern T1805 GE_default1805;
extern T1806 GE_default1806;
extern T1807 GE_default1807;
extern T1808 GE_default1808;
extern T1809 GE_default1809;
extern T1810 GE_default1810;
extern T1811 GE_default1811;
extern T1812 GE_default1812;
extern T1813 GE_default1813;
extern T1814 GE_default1814;
extern T1815 GE_default1815;
extern T1816 GE_default1816;
extern T1817 GE_default1817;
extern T1818 GE_default1818;
extern T1819 GE_default1819;
extern T1820 GE_default1820;
extern T1821 GE_default1821;
extern T1822 GE_default1822;
extern T1823 GE_default1823;
extern T1824 GE_default1824;
extern T1825 GE_default1825;
extern T1826 GE_default1826;
extern T1827 GE_default1827;
extern T1828 GE_default1828;
extern T1829 GE_default1829;
extern T1830 GE_default1830;
extern T1831 GE_default1831;
extern T1832 GE_default1832;
extern T1833 GE_default1833;
extern T1834 GE_default1834;
extern T1835 GE_default1835;
extern T1836 GE_default1836;
extern T1837 GE_default1837;
extern T1838 GE_default1838;
extern T1839 GE_default1839;
extern T1840 GE_default1840;
extern T1841 GE_default1841;
extern T1842 GE_default1842;
extern T1843 GE_default1843;
extern T1844 GE_default1844;
extern T1845 GE_default1845;
extern T1847 GE_default1847;
extern T1848 GE_default1848;
extern T1849 GE_default1849;
extern T1850 GE_default1850;
extern T1851 GE_default1851;
extern T1852 GE_default1852;
extern T1853 GE_default1853;
extern T1854 GE_default1854;
extern T1857 GE_default1857;
extern T1858 GE_default1858;
extern T1859 GE_default1859;
extern T1861 GE_default1861;
extern T1862 GE_default1862;
extern T1863 GE_default1863;
extern T1864 GE_default1864;
extern T1865 GE_default1865;
extern T1866 GE_default1866;
extern T1867 GE_default1867;
extern T1868 GE_default1868;
extern T1869 GE_default1869;
extern T1870 GE_default1870;
extern T1873 GE_default1873;
extern T1874 GE_default1874;
extern T1875 GE_default1875;
extern T1876 GE_default1876;
extern T1877 GE_default1877;
extern T1878 GE_default1878;
extern T1879 GE_default1879;
extern T1880 GE_default1880;
extern T1882 GE_default1882;
extern T1883 GE_default1883;
extern T1884 GE_default1884;
extern T1885 GE_default1885;
extern T1886 GE_default1886;
extern T1887 GE_default1887;
extern T1888 GE_default1888;
extern T1889 GE_default1889;
extern T1890 GE_default1890;
extern T1891 GE_default1891;
extern T1898 GE_default1898;
extern T1899 GE_default1899;
extern T1900 GE_default1900;
extern T1901 GE_default1901;
extern T1902 GE_default1902;
extern T1905 GE_default1905;
extern T1907 GE_default1907;
extern T1909 GE_default1909;
extern T1910 GE_default1910;
extern T1911 GE_default1911;
extern T1913 GE_default1913;
extern T1914 GE_default1914;
extern T1916 GE_default1916;
extern T1924 GE_default1924;
extern T1925 GE_default1925;
extern T1926 GE_default1926;
extern T1930 GE_default1930;
extern T1931 GE_default1931;
extern T1934 GE_default1934;
extern T1935 GE_default1935;
extern T1936 GE_default1936;
extern T1939 GE_default1939;
extern T1940 GE_default1940;
extern T1943 GE_default1943;
extern T1944 GE_default1944;
extern T1945 GE_default1945;
extern T1946 GE_default1946;
extern T1947 GE_default1947;
extern T1967 GE_default1967;
extern T1973 GE_default1973;
extern T1974 GE_default1974;
extern T1981 GE_default1981;
extern T1986 GE_default1986;
extern T1987 GE_default1987;
extern T1988 GE_default1988;
extern T1989 GE_default1989;
extern T1990 GE_default1990;
extern T1992 GE_default1992;
extern T1993 GE_default1993;
extern T1994 GE_default1994;
extern T1997 GE_default1997;
extern T1998 GE_default1998;
extern T2000 GE_default2000;
extern T2002 GE_default2002;
extern T2003 GE_default2003;
extern T2004 GE_default2004;
extern T2005 GE_default2005;
extern T2006 GE_default2006;
extern T2007 GE_default2007;
extern T2009 GE_default2009;
extern T2010 GE_default2010;
extern T2012 GE_default2012;
extern T2014 GE_default2014;
extern T2017 GE_default2017;
extern T2019 GE_default2019;
extern T2020 GE_default2020;
extern T2021 GE_default2021;
extern T2023 GE_default2023;
extern T2024 GE_default2024;
extern T2025 GE_default2025;
extern T2028 GE_default2028;
extern T2029 GE_default2029;
extern T2034 GE_default2034;
extern T2035 GE_default2035;
extern T2036 GE_default2036;
extern T2037 GE_default2037;
extern T2039 GE_default2039;
extern T2040 GE_default2040;
extern T2041 GE_default2041;
extern T2042 GE_default2042;
extern T2044 GE_default2044;
extern T2045 GE_default2045;
extern T2047 GE_default2047;
extern T2048 GE_default2048;
extern T2051 GE_default2051;
extern T2052 GE_default2052;
extern T2053 GE_default2053;
extern T2054 GE_default2054;
extern T2055 GE_default2055;
extern T2056 GE_default2056;
extern T2057 GE_default2057;
extern T2058 GE_default2058;
extern T2059 GE_default2059;
extern T2060 GE_default2060;
extern T2061 GE_default2061;
extern T2062 GE_default2062;
extern T2063 GE_default2063;
extern T2064 GE_default2064;
extern T2065 GE_default2065;
extern T2066 GE_default2066;
extern T2067 GE_default2067;
extern T2068 GE_default2068;
extern T2069 GE_default2069;
extern T2070 GE_default2070;
extern T2071 GE_default2071;
extern T2072 GE_default2072;
extern T2073 GE_default2073;
extern T2074 GE_default2074;
extern T2075 GE_default2075;
extern T2076 GE_default2076;
extern T2077 GE_default2077;
extern T2078 GE_default2078;
extern T2079 GE_default2079;
extern T2080 GE_default2080;
extern T2081 GE_default2081;
extern T2082 GE_default2082;
extern T2083 GE_default2083;
extern T2084 GE_default2084;
extern T2085 GE_default2085;
extern T2086 GE_default2086;
extern T2087 GE_default2087;
extern T2088 GE_default2088;
extern T2089 GE_default2089;
extern T2090 GE_default2090;
extern T2091 GE_default2091;
extern T2092 GE_default2092;
extern T2093 GE_default2093;
extern T2094 GE_default2094;
extern T2095 GE_default2095;
extern T2096 GE_default2096;
extern T2097 GE_default2097;
extern T2098 GE_default2098;
extern T2099 GE_default2099;
extern T2100 GE_default2100;
extern T2101 GE_default2101;
extern T2102 GE_default2102;
extern T2103 GE_default2103;
extern T2104 GE_default2104;
extern T2105 GE_default2105;
extern T2106 GE_default2106;
extern T2107 GE_default2107;
extern T2108 GE_default2108;
extern T2109 GE_default2109;
extern T2110 GE_default2110;
extern T2111 GE_default2111;
extern T2113 GE_default2113;
extern T2114 GE_default2114;
extern T2115 GE_default2115;
extern T2117 GE_default2117;
extern T2119 GE_default2119;
extern T2120 GE_default2120;
extern T2123 GE_default2123;
extern T2124 GE_default2124;
extern T2126 GE_default2126;
extern T2127 GE_default2127;
extern T2128 GE_default2128;
extern T2129 GE_default2129;
extern T2130 GE_default2130;
extern T2132 GE_default2132;
extern T2134 GE_default2134;
extern T2136 GE_default2136;
extern T2137 GE_default2137;
extern T2139 GE_default2139;
extern T2140 GE_default2140;
extern T2141 GE_default2141;
extern T2142 GE_default2142;
extern T2143 GE_default2143;
extern T2144 GE_default2144;
extern T2146 GE_default2146;
extern T2150 GE_default2150;
extern T2153 GE_default2153;
extern T2156 GE_default2156;
extern T2157 GE_default2157;
extern T2158 GE_default2158;
extern T2159 GE_default2159;
extern T2160 GE_default2160;
extern T2162 GE_default2162;

extern T0* GE_new_str8(T6 c);
extern T0* GE_new_str32(T6 c);
extern T0* GE_new_istr8(T6 c);
extern T0* GE_new_istr32(T6 c);
/* Call to [detachable] ANY.same_type */
extern T1 T23x49156T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.read_to_string */
extern T6 T37x5767169T0T6T6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.end_of_input */
extern T1 T37x5881861(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.last_character */
extern T2 T37x5881864(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.name */
extern T0* T37x5881863(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.is_closable */
extern T1 T37x5881858(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.short_form */
extern T2 T49x5275656(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.long_form */
extern T0* T49x5275651(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.description */
extern T0* T49x5275649(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.names */
extern T0* T49x5275653(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.is_hidden */
extern T1 T49x5275659(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.name */
extern T0* T49x5275652(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.needs_parameter */
extern T1 T49x5275662(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.has_short_form */
extern T1 T49x5275658(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.allows_parameter */
extern T1 T49x5275661(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.maximum_occurrences */
extern T6 T49x5275655(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.occurrences */
extern T6 T49x5275654(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.was_found */
extern T1 T49x5275665(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.is_mandatory */
extern T1 T49x5275660(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.example */
extern T0* T49x5275650(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.conforms_to_type_with_type_marks */
extern T1 T64x14925830T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.base_class */
extern T0* T64x21856259(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.index_of_label */
extern T6 T64x21282825T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.base_type_index_of_label */
extern T6 T64x21823501T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.same_base_type_with_type_marks */
extern T1 T64x21823537T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.to_text */
extern T0* T64x21823562(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.type_with_type_mark */
extern T0* T64x21823509T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.actual_parameters */
extern T0* T64x21282820(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T64x21823494T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.adapted_base_class_with_named_feature */
extern T0* T64x21823493T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_attached_with_type_mark */
extern T1 T64x21823520T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.named_base_class */
extern T0* T64x21856260(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.base_type_actual */
extern T0* T64x21823499T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.actual_parameter_count */
extern T6 T64x21282828(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.shallow_named_type_with_type_mark */
extern T0* T64x21823505T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_reference_with_type_mark */
extern T1 T64x21823518T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T64x21856262T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_BASE_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T64x21856261T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.is_expanded */
extern T1 T64x21282832(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.is_attached */
extern T1 T64x21282833(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.is_type_detachable_with_type_mark */
extern T1 T64x21823522T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_expanded_with_type_mark */
extern T1 T64x21823516T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T64x21823554T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T64x21823553T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T64x21823552T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.base_type_with_type_mark */
extern T0* T64x21823496T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.shallow_base_type_with_type_mark */
extern T0* T64x21823498T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.same_named_type */
extern T1 T64x21823534T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_BASE_TYPE.unaliased_to_text */
extern T0* T64x21823563(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T64x21823545T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T64x21823546T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.same_named_class_type_with_type_marks */
extern T1 T64x21823544T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.base_type_actual_count */
extern T6 T64x21823511T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.named_type_with_type_mark */
extern T0* T64x21823503T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.same_named_type_with_type_marks */
extern T1 T64x21823535T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.upper_name */
extern T0* T64x21708802(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.name */
extern T0* T64x21708801(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.same_as_base_class */
extern T1 T64x21282839(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.runtime_name_to_text */
extern T0* T64x21823564(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.code */
extern T10 T78x2048001T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] READABLE_STRING_GENERAL.count */
extern T6 T78x2048039(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.substring */
extern T0* T78x2048080T6T6(GE_context* ac, T0* C, T6 a1, T6 a2);
/* Call to [detachable] READABLE_STRING_GENERAL.item */
extern T3 T78x2048002T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] UT_ERROR.default_message */
extern T0* T80x11517955(GE_context* ac, T0* C);
/* Call to detachable KL_EQUALITY_TESTER [[attached] STRING_8].test */
extern T1 T150x7307265T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.ecf_version */
extern T0* T181x13713414(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.filename */
extern T0* T181x13713411(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.universe */
extern T0* T181x13713420(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.targets */
extern T0* T181x13713415(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.library_target */
extern T0* T181x13713416(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.column */
extern T6 T193x20004866(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.line */
extern T6 T193x20004865(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.to_text */
extern T0* T193x20004871(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.is_null */
extern T1 T193x20004867(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_primary_type */
extern T1 T219x23199759T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_class */
extern T0* T219x23609351(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_type */
extern T0* T219x23609350(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_query */
extern T0* T219x23199769T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_basic */
extern T1 T219x23609348(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_expanded */
extern T1 T219x23642113(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_procedure */
extern T0* T219x23199768T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_agent_type */
extern T1 T219x23199746(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_query */
extern T0* T219x23199767T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_procedure */
extern T0* T219x23199770T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.id */
extern T6 T219x23199757(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.queries */
extern T0* T219x23199765(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.attribute_count */
extern T6 T219x23199761(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_special */
extern T1 T219x23199747(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_builtin_attribute */
extern T1 T219x23199771T0T8T8(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.primary_type */
extern T0* T219x23609352(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.hash_code */
extern T6 T219x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_generic */
extern T1 T219x23609347(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_attributes */
extern T1 T219x23199763(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_alive */
extern T1 T219x23199745(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.attached_type */
extern T0* T219x23199755(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.next_type */
extern T0* T219x23199778(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforming_ancestors */
extern T0* T219x23199760(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_mark */
extern T0* T219x23609354(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedure_calls */
extern T0* T219x23199775(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.query_calls */
extern T0* T219x23199774(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedures */
extern T0* T219x23199766(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.meta_type */
extern T0* T219x23609353(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_tuple */
extern T1 T219x23199748(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_static */
extern T1 T219x23199749(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.last_leaf */
extern T0* T229x19791877(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.name */
extern T0* T229x20365313(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.lower_name */
extern T0* T229x20365314(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.position */
extern T0* T229x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.seed */
extern T6 T229x20365315(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.same_feature_name */
extern T1 T229x20496387T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_FEATURE_NAME.is_tuple_label */
extern T1 T229x20365321(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.across_cursor_name */
extern T0* T229x20365357(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_across_cursor */
extern T1 T229x20365319(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.object_test_local_name */
extern T0* T229x20365356(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_object_test_local */
extern T1 T229x20365318(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.local_name */
extern T0* T229x20365354(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_local */
extern T1 T229x20365317(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.argument_name */
extern T0* T229x20365355(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_argument */
extern T1 T229x20365320(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.hash_code */
extern T6 T229x475137(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.first_leaf */
extern T0* T229x19791876(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_prefix */
extern T1 T229x20365343(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME.is_infix */
extern T1 T229x20365323(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.hash_code */
extern T6 T230x475137(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.first_seed */
extern T6 T230x17743883(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_static */
extern T1 T230x17661990(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_exported_to */
extern T1 T230x17662007T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_PROCEDURE.arguments_count */
extern T6 T230x17662006(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.arguments */
extern T0* T230x17743882(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.overloaded_alias_name */
extern T0* T230x17661959(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.alias_name */
extern T0* T230x17661955(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.overloaded_name */
extern T0* T230x17661958(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.name */
extern T0* T230x17743880(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.lower_name */
extern T0* T230x18366467(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_creation_exported_to */
extern T1 T230x17662009T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_PROCEDURE.new_synonym */
extern T0* T230x17662013T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.has_seed */
extern T1 T230x17743878T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_PROCEDURE.implementation_class */
extern T0* T230x18366469(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_parenthesisable */
extern T1 T230x17661997(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_frozen */
extern T1 T230x17661979(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_deferred */
extern T1 T230x17661980(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.postconditions */
extern T0* T230x17252362(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.preconditions */
extern T0* T230x17252361(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.lower_name */
extern T0* T231x18366467(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.type */
extern T0* T231x17743881(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.first_seed */
extern T6 T231x17743883(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_attribute */
extern T1 T231x17661983(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.assigner */
extern T0* T231x18169859(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.overloaded_alias_name */
extern T0* T231x17661959(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.alias_name */
extern T0* T231x17661955(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.overloaded_name */
extern T0* T231x17661958(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.name */
extern T0* T231x17743880(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.hash_code */
extern T6 T231x475137(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.arguments */
extern T0* T231x17743882(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_exported_to */
extern T1 T231x17662007T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_QUERY.is_static */
extern T1 T231x17661990(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.arguments_count */
extern T6 T231x17662006(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.has_seed */
extern T1 T231x17743878T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_QUERY.implementation_class */
extern T0* T231x18366469(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.implementation_feature */
extern T0* T231x17252365(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_unique_attribute */
extern T1 T231x17661985(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_constant_attribute */
extern T1 T231x17661984(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.new_synonym */
extern T0* T231x17662013T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.is_prefixable */
extern T1 T231x17661995(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_infixable */
extern T1 T231x17661994(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_once_per_object */
extern T1 T231x17252356(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_once */
extern T1 T231x17252353(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_parenthesisable */
extern T1 T231x17661997(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_bracketable */
extern T1 T231x17661996(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_frozen */
extern T1 T231x17661979(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_deferred */
extern T1 T231x17661980(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.postconditions */
extern T0* T231x17252362(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.preconditions */
extern T0* T231x17252361(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.position */
extern T0* T234x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_type_detachable */
extern T1 T234x21823521T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_named_type */
extern T0* T234x21823504T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.is_type_attached */
extern T1 T234x21823519T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.base_type_index_of_label */
extern T6 T234x21823501T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T234x21823494T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_TYPE.adapted_base_class_with_named_feature */
extern T0* T234x21823493T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.same_base_type_with_type_marks */
extern T1 T234x21823537T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_named_type_with_type_marks */
extern T1 T234x21823535T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.to_text */
extern T0* T234x21823562(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_base_type */
extern T1 T234x21823514(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T234x21823540T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T234x21823543T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T234x21823538T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.is_like_current */
extern T1 T234x21823512(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.named_type */
extern T0* T234x21823502T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.conforms_to_type */
extern T1 T234x14925829T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE.same_named_type */
extern T1 T234x21823534T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_TYPE.is_named_type */
extern T1 T234x14925827(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_type_expanded */
extern T1 T234x21823515T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.same_syntactical_type */
extern T1 T234x21823532T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_TYPE.resolved_formal_parameters */
extern T0* T234x21823555T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_identifier_anchored_type */
extern T1 T234x21823525(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.depends_on_qualified_anchored_type */
extern T1 T234x21823526T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_formal_types */
extern T1 T234x21823527T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_unqualified_anchored_type */
extern T1 T234x21823524(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.type_with_type_mark */
extern T0* T234x21823509T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_base_type_with_type_mark */
extern T0* T234x21823498T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T234x21823539T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T234x21823542T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T234x21823541T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_type_with_type_marks */
extern T1 T234x21823533T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.shallow_named_type_with_type_mark */
extern T0* T234x21823505T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_reference_with_type_mark */
extern T1 T234x21823518T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_attached_with_type_mark */
extern T1 T234x21823520T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_detachable_with_type_mark */
extern T1 T234x21823522T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_expanded_with_type_mark */
extern T1 T234x21823516T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T234x21823554T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T234x21823553T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T234x21823552T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.base_type_with_type_mark */
extern T0* T234x21823496T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.named_base_class */
extern T0* T234x21823492T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T234x21823545T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.conforms_to_type_with_type_marks */
extern T1 T234x14925830T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T234x21823546T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_named_class_type_with_type_marks */
extern T1 T234x21823544T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.base_type_actual */
extern T0* T234x21823499T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_TYPE.base_type_actual_count */
extern T6 T234x21823511T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.named_type_with_type_mark */
extern T0* T234x21823503T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.base_class */
extern T0* T234x21823491T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.conforms_to_constraint */
extern T1 T234x21823550T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE.is_type_reference */
extern T1 T234x21823517T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_base_type */
extern T0* T234x21823497T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.type */
extern T0* T234x21872641(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.to_nested_type_context */
extern T0* T235x21856307(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.as_nested_type_context */
extern T0* T235x21856306(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.root_context */
extern T0* T235x21856257(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.is_root_context */
extern T1 T235x21856272(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_actual */
extern T0* T235x21856265T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_actual_count */
extern T6 T235x21856270(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_class */
extern T0* T235x21856259(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.scoop_mode */
extern T1 T235x21856286(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.attachment_type_conformance_mode */
extern T1 T235x21856285(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.named_type */
extern T0* T235x21856268(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.same_base_type_with_type_marks */
extern T1 T235x21856292T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_attached_with_type_mark */
extern T1 T235x21856279T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_reference_with_type_mark */
extern T1 T235x21856277T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.adapted_base_class_with_seeded_feature */
extern T0* T235x21856262T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_TYPE_CONTEXT.adapted_base_class_with_named_feature */
extern T0* T235x21856261T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T235x21856304T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_formal_parameter_type_with_type_marks */
extern T1 T235x21856294T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.named_type_with_type_mark */
extern T0* T235x21856269T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_type_with_type_marks */
extern T1 T235x21856288T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_to_type_with_type_marks */
extern T1 T235x21856300T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_tuple_type_with_type_marks */
extern T1 T235x21856295T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_class_type_with_type_marks */
extern T1 T235x21856293T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_tuple_type_with_type_marks */
extern T1 T235x21856305T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_class_type_with_type_marks */
extern T1 T235x21856303T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_detachable_with_type_mark */
extern T1 T235x21856281T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_expanded_with_type_mark */
extern T1 T235x21856275T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.named_base_class */
extern T0* T235x21856260(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_with_type_mark */
extern T0* T235x21856264T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.adapted_universe */
extern T0* T238x18890770T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.boolean_type */
extern T0* T238x18890793(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_32_type */
extern T0* T238x18890796(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_8_type */
extern T0* T238x18890795(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_64_type */
extern T0* T238x18890815(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_32_type */
extern T0* T238x18890814(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_16_type */
extern T0* T238x18890813(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_8_type */
extern T0* T238x18890812(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_64_type */
extern T0* T238x18890808(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_32_type */
extern T0* T238x18890807(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_16_type */
extern T0* T238x18890806(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_8_type */
extern T0* T238x18890805(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_64_convert_feature */
extern T0* T238x18890859(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_64_type */
extern T0* T238x18890823(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_32_convert_feature */
extern T0* T238x18890858(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_32_type */
extern T0* T238x18890822(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_64_convert_feature */
extern T0* T238x18890857(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_32_convert_feature */
extern T0* T238x18890856(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_16_convert_feature */
extern T0* T238x18890855(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_8_convert_feature */
extern T0* T238x18890854(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_64_convert_feature */
extern T0* T238x18890853(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_32_convert_feature */
extern T0* T238x18890852(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_16_convert_feature */
extern T0* T238x18890851(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_8_convert_feature */
extern T0* T238x18890850(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.shortest_path */
extern T0* T238x18890771T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.full_lower_name */
extern T0* T238x18890775T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_UNIVERSE.kind_name */
extern T0* T238x18890777(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_clients */
extern T0* T238x18890788(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_32_convert_feature */
extern T0* T238x18890849(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_8_convert_feature */
extern T0* T238x18890848(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_32_convert_feature */
extern T0* T238x18890861(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_32_type */
extern T0* T238x18890831(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_8_convert_feature */
extern T0* T238x18890860(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_8_type */
extern T0* T238x18890830(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.master_class */
extern T0* T238x18890759T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.has_master_class */
extern T1 T238x18890753T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.full_name */
extern T0* T238x18890774T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_UNIVERSE.classes_in_group */
extern T0* T238x18890764T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.tuple_type */
extern T0* T238x18890835(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.function_identity_any_type */
extern T0* T238x18890801(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.predicate_identity_type */
extern T0* T238x18890819(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.tuple_identity_type */
extern T0* T238x18890837(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.procedure_identity_type */
extern T0* T238x18890820(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_type */
extern T0* T238x18890804(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.iterable_detachable_separate_any_type */
extern T0* T238x18890810(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.pointer_type */
extern T0* T238x18890818(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.typed_pointer_identity_type */
extern T0* T238x18890844(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_identity_type */
extern T0* T238x18890843(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_any_type */
extern T0* T238x18890784(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.implicit_attachment_type_mark */
extern T0* T238x18890846(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.unfolded_empty_tuple_actual_parameters */
extern T0* T238x18890838(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.obsolete_routine_type_mode */
extern T1 T238x18890847(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.immutable_string_32_type */
extern T0* T238x18890803(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.immutable_string_8_type */
extern T0* T238x18890802(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_string_type */
extern T0* T238x18890834(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_type */
extern T0* T238x18890828(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.hash_code */
extern T6 T238x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_separate_any_type */
extern T0* T238x18890785(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.double_type */
extern T0* T238x18890797(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_type */
extern T0* T238x18890821(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_string_type */
extern T0* T238x18890829(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_type */
extern T0* T238x18890783(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.special_any_type */
extern T0* T238x18890825(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_any_type */
extern T0* T238x18890839(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.special_detachable_any_type */
extern T0* T238x18890826(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.exception_manager_type */
extern T0* T238x18890800(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_exception_type */
extern T0* T238x18890841(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.exception_type */
extern T0* T238x18890798(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_type_exception_type */
extern T0* T238x18890840(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_exception_type */
extern T0* T238x18890799(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_like_current_type */
extern T0* T238x18890842(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_object_parents */
extern T0* T238x18890833(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_parents */
extern T0* T238x18890787(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_object_type */
extern T0* T238x18890832(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.current_system */
extern T0* T238x18890769(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.resolved_formal_parameters_with_type */
extern T0* T239x21233669T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.type */
extern T0* T239x21233666(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.actual_parameter */
extern T0* T239x21233665(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_detachable_mark */
extern T1 T241x20217860(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_attached_mark */
extern T1 T241x20217859(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.position */
extern T0* T241x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_expandedness_mark */
extern T1 T241x20217869(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_attachment_mark */
extern T1 T241x20217857(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_implicit_mark */
extern T1 T241x20217875(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_separateness_mark */
extern T1 T241x20217865(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_expanded_mark */
extern T1 T241x20217871(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.overridden_type_mark */
extern T0* T241x20217876T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_MARK.is_explicit_attachment_mark */
extern T1 T241x20217858(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_separate_mark */
extern T1 T241x20217867(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_reference_mark */
extern T1 T241x20217872(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.last_leaf */
extern T0* T242x19791877(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.type */
extern T0* T242x21200899T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.count */
extern T6 T242x21200897(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.is_empty */
extern T1 T242x21200903(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.same_named_types */
extern T1 T242x21200911T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ACTUAL_PARAMETERS.index_of_label */
extern T6 T242x21200901T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T242x21200898T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T242x21200910T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T242x21200915T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T242x21200906(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T242x21200907T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T242x21200908T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T242x21200905(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T242x21200912T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to detachable ET_ACTUAL_PARAMETERS.named_types */
extern T0* T242x21200900T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.static_type */
extern T0* T244x23642117(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_never_void */
extern T1 T244x23642114(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.can_be_void */
extern T1 T244x23642115(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.dynamic_type */
extern T0* T244x23216129T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.count */
extern T6 T244x23216132(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.has_expanded */
extern T1 T244x23216136(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_expanded */
extern T1 T244x23642113(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_empty */
extern T1 T244x23216133(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.special_type */
extern T0* T244x23216130(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.dynamic_types */
extern T0* T244x23642119(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.has_type */
extern T1 T244x23216134T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TYPE.primary_type */
extern T0* T245x23609352(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.type_mark */
extern T0* T245x23609354(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_basic */
extern T1 T245x23609348(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_expanded */
extern T1 T245x23642113(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.hash_code */
extern T6 T245x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.conforms_to_type */
extern T1 T245x23609360T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TYPE.base_class */
extern T0* T245x23609351(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.base_type */
extern T0* T245x23609350(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.meta_type */
extern T0* T245x23609353(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_generic */
extern T1 T245x23609347(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.type_id */
extern T6 T245x23609355(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.index */
extern T6 T246x21561347(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.implementation_class */
extern T0* T246x21561348(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable_with_type_mark */
extern T1 T246x21823522T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached_with_type_mark */
extern T1 T246x21823520T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.type_with_type_mark */
extern T0* T246x21823509T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.overridden_type_mark */
extern T0* T246x21823490T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.name */
extern T0* T246x21708801(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.type */
extern T0* T246x21872641(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.last_leaf */
extern T0* T272x19791877(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.position */
extern T0* T272x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_instance_free */
extern T1 T272x16236548(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_false */
extern T1 T272x16515078(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.manifest_constant_convert_feature */
extern T0* T272x16515082T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_EXPRESSION.index */
extern T6 T272x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_instance_free */
extern T1 T272x16891906(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.count */
extern T6 T272x16236550(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.unparenthesized_expression */
extern T0* T272x16515076(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_current */
extern T1 T272x16891907(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_never_void */
extern T1 T272x16515077(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.lower_name */
extern T0* T276x18366467(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.name */
extern T0* T276x17743880(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.implementation_feature */
extern T0* T276x17252365(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.implementation_class */
extern T0* T276x18366469(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.arguments */
extern T0* T276x17743882(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_seed */
extern T1 T276x17743878T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.first_seed */
extern T6 T276x17743883(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_deferred */
extern T1 T276x17661980(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_static */
extern T1 T276x17661990(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.clients */
extern T0* T276x18366471(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_exported_to */
extern T1 T276x17662007T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FEATURE.extended_name */
extern T0* T276x17661956(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.arguments_count */
extern T6 T276x17662006(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.version */
extern T6 T276x17661964(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.are_postconditions_instance_free_recursive */
extern T1 T276x17661999(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.are_preconditions_instance_free_recursive */
extern T1 T276x17661998(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.alias_name */
extern T0* T276x17661955(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_constant_attribute */
extern T1 T276x17661984(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_procedure */
extern T1 T276x17661987(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.other_seeds */
extern T0* T276x17743884(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.type */
extern T0* T276x17743881(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_class_postcondition_recursive */
extern T1 T276x17662000(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_unique_attribute */
extern T1 T276x17661985(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_attribute */
extern T1 T276x17661983(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once_per_process */
extern T1 T276x17252354(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_frozen */
extern T1 T276x17661979(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_infixable */
extern T1 T276x17661994(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_prefixable */
extern T1 T276x17661995(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_parenthesisable */
extern T1 T276x17661997(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_bracketable */
extern T1 T276x17661996(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.id */
extern T6 T276x17661963(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.renamed_feature */
extern T0* T276x17662014T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.undefined_feature */
extern T0* T276x17662015T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.postconditions */
extern T0* T276x17252362(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.preconditions */
extern T0* T276x17252361(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.same_version */
extern T1 T276x17662012T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.is_dotnet */
extern T1 T276x17743877(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_query */
extern T1 T276x17661986(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once */
extern T1 T276x17252353(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.other_precursors */
extern T0* T276x17661966(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.first_precursor */
extern T0* T276x18366468(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_validity_error */
extern T1 T276x18366473(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.validity_checked */
extern T1 T276x18366472(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.hash_code */
extern T6 T276x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_NAME.last_leaf */
extern T0* T277x19791877(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_NAME.feature_name */
extern T0* T277x20971521(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.name */
extern T0* T303x16285698(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.arguments */
extern T0* T303x16285699(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.position */
extern T0* T303x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.is_tuple_label */
extern T1 T303x16285701(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.is_call_agent */
extern T1 T303x16285702(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.position */
extern T0* T304x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.seed */
extern T6 T304x20365315(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_tuple_label */
extern T1 T304x20365321(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.lower_name */
extern T0* T304x20365314(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.name */
extern T0* T304x20365313(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or_else */
extern T1 T304x20365337(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_implies */
extern T1 T304x20365331(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and_then */
extern T1 T304x20365325(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.precursor_keyword */
extern T0* T304x20365353(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_precursor */
extern T1 T304x20365351(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.hash_code */
extern T6 T304x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_not */
extern T1 T304x20365347(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_plus */
extern T1 T304x20365345(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_minus */
extern T1 T304x20365344(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_dotdot */
extern T1 T304x20365342(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_xor */
extern T1 T304x20365341(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_times */
extern T1 T304x20365340(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_power */
extern T1 T304x20365339(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_plus */
extern T1 T304x20365338(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or */
extern T1 T304x20365336(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_mod */
extern T1 T304x20365335(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_minus */
extern T1 T304x20365334(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_lt */
extern T1 T304x20365333(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_le */
extern T1 T304x20365332(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_gt */
extern T1 T304x20365330(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_ge */
extern T1 T304x20365329(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_divide */
extern T1 T304x20365327(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_div */
extern T1 T304x20365326(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and */
extern T1 T304x20365324(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_parenthesis */
extern T1 T304x20365349(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_bracket */
extern T1 T304x20365348(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.same_call_name */
extern T1 T304x20365352T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_CALL_NAME.feature_name */
extern T0* T304x20365358(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix */
extern T1 T304x20365323(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix */
extern T1 T304x20365343(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.target */
extern T0* T307x14024706(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.arguments */
extern T0* T307x14024707(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.implicit_target_position */
extern T0* T307x14024714(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.position */
extern T0* T307x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.is_inline_agent */
extern T1 T307x14024712(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.index */
extern T6 T307x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.is_qualified_call */
extern T1 T307x14024709(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.implicit_result */
extern T0* T307x14024708(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.expression */
extern T0* T308x16842756(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.index */
extern T6 T308x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.type */
extern T0* T308x16842755(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.name */
extern T0* T308x16842757(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_integer_constant */
extern T1 T313x15859715(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_character_constant */
extern T1 T313x15859714(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_boolean_constant */
extern T1 T313x15859713(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.index */
extern T6 T313x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.type */
extern T0* T322x21790721(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.position */
extern T0* T322x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.last_leaf */
extern T0* T322x19791877(GE_context* ac, T0* C);
/* Call to detachable ET_ARGUMENT_OPERANDS.actual_argument */
extern T0* T332x16236549T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ARGUMENT_OPERANDS.count */
extern T6 T332x16236550(GE_context* ac, T0* C);
/* Call to detachable ET_ARGUMENT_OPERANDS.is_empty */
extern T1 T332x16236546(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION_ITEM.expression */
extern T0* T334x16580609(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION_ITEM.position */
extern T0* T334x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.is_open_operand */
extern T1 T337x16891905(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.index */
extern T6 T337x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.is_current */
extern T1 T337x16891907(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.index */
extern T6 T338x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.is_instance_free */
extern T1 T338x16891906(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.position */
extern T0* T338x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.is_open_operand */
extern T1 T338x16891905(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.is_instance_free */
extern T1 T340x16891906(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.is_open_operand */
extern T1 T340x16891905(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.index */
extern T6 T340x16891908(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_ARGUMENTS.actual_argument */
extern T0* T351x16236549T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_ARGUMENTS.count */
extern T6 T351x16236550(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_ARGUMENTS.is_empty */
extern T1 T351x16236546(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.dynamic_type */
extern T0* T353x23216129T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.count */
extern T6 T353x23216132(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.index_of */
extern T6 T353x23216131T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.has_type */
extern T1 T353x23216134T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_CLOSURE.type */
extern T0* T358x17252357(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.arguments */
extern T0* T358x17252358(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.locals */
extern T0* T358x17252360(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.across_components */
extern T0* T358x17252364(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.object_tests */
extern T0* T358x17252363(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.position */
extern T0* T360x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.implicit_result */
extern T0* T360x14024708(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.formal_arguments */
extern T0* T360x17252358(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.actual_arguments */
extern T0* T360x14024707(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.target */
extern T0* T360x14024706(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.locals */
extern T0* T360x17252360(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.attached_result_index */
extern T6 T360x14336002(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.result_index */
extern T6 T360x14336001(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.type */
extern T0* T360x17252357(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.parenthesis_call */
extern T0* T362x19333123(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.arguments */
extern T0* T362x16285699(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.name */
extern T0* T362x16285698(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.target */
extern T0* T362x16285697(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.arguments_count */
extern T6 T362x19038211(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.index */
extern T6 T362x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.is_call_agent */
extern T1 T362x16285702(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.is_tuple_label */
extern T1 T362x16285701(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.index */
extern T6 T364x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.value_position */
extern T0* T364x15941643(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.cast_type */
extern T0* T364x15941640(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.value */
extern T0* T364x15941638(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.position */
extern T0* T364x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.is_string_32 */
extern T1 T364x15941634(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.is_string_8 */
extern T1 T364x15941633(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.as_invariants */
extern T0* T381x18366475(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.is_invariants */
extern T1 T381x18366466(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.as_feature */
extern T0* T381x18366474(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.is_feature */
extern T1 T381x18366465(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.first_precursor */
extern T0* T381x18366468(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.has_validity_error */
extern T1 T381x18366473(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.implementation_feature */
extern T0* T381x17252365(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.postconditions */
extern T0* T381x17252362(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.preconditions */
extern T0* T381x17252361(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.clients */
extern T0* T381x18366471(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.locals */
extern T0* T381x17252360(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.arguments */
extern T0* T381x17252358(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.type */
extern T0* T381x17252357(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.implementation_class */
extern T0* T381x18366469(GE_context* ac, T0* C);
/* Call to [detachable] ET_ASSERTION.expression */
extern T0* T405x14499842(GE_context* ac, T0* C);
/* Call to [detachable] ET_ASSERTION.is_class_assertion */
extern T1 T405x14499841(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.relative_lower_name */
extern T0* T407x18743317T0T2(GE_context* ac, T0* C, T0* a1, T2 a2);
/* Call to [detachable] ET_PRIMARY_GROUP.kind_name */
extern T0* T407x18743323(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.full_lower_name */
extern T0* T407x18743319T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_PRIMARY_GROUP.is_hidden */
extern T1 T407x18743306(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_cluster */
extern T1 T407x18743299(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.current_system */
extern T0* T407x18743313(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.universe */
extern T0* T407x18743312(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.use_obsolete_syntax */
extern T1 T407x18743307(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_override */
extern T1 T407x18743304(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_unknown */
extern T1 T407x18743303(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.cluster */
extern T0* T407x18743336(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.dotnet_assembly */
extern T0* T407x18743338(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_dotnet_assembly */
extern T1 T407x18743301(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_none */
extern T1 T407x18743302(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.alias_name */
extern T0* T467x20463618(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.feature_name */
extern T0* T467x20463617(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.same_call_name */
extern T1 T467x20463620T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.position */
extern T0* T467x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.same_extended_feature_name */
extern T1 T467x20463619T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.start_condition */
extern T6 T506x12271626(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.is_applicable_encoding */
extern T1 T506x28377092T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.end_of_file */
extern T1 T506x12271632(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.last_value */
extern T0* T506x28377101(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.last_token */
extern T6 T506x12271617(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.error_position */
extern T0* T506x28377099(GE_context* ac, T0* C);
/* Call to [detachable] XM_DOCUMENT_NODE.parent */
extern T0* T571x29294593(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.last_leaf */
extern T0* T607x19791877(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.label */
extern T0* T607x21233667(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.position */
extern T0* T607x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T607x21184516T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ACTUAL_PARAMETER.type */
extern T0* T607x21233666(GE_context* ac, T0* C);
/* Call to detachable ET_NAMED_TYPE.to_text */
extern T0* T608x21823562(GE_context* ac, T0* C);
/* Call to detachable ET_NAMED_TYPE.type_with_type_mark */
extern T0* T608x21823509T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T608x21823540T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T608x21823543T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.has_formal_types */
extern T1 T608x21823527T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T608x21823539T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T608x21823538T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_base_type_with_type_marks */
extern T1 T608x21823537T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T608x21823542T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T608x21823541T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_type_with_type_marks */
extern T1 T608x21823533T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.is_type_reference_with_type_mark */
extern T1 T608x21823518T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_attached_with_type_mark */
extern T1 T608x21823520T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_detachable_with_type_mark */
extern T1 T608x21823522T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_expanded_with_type_mark */
extern T1 T608x21823516T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T608x21823554T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T608x21823553T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T608x21823552T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T608x21823545T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.conforms_to_type_with_type_marks */
extern T1 T608x14925830T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T608x21823546T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_named_class_type_with_type_marks */
extern T1 T608x21823544T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_named_type_with_type_marks */
extern T1 T608x21823535T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.base_class */
extern T0* T608x21823491T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INDEXING_TERM.indexing_term_value */
extern T0* T730x15384578(GE_context* ac, T0* C);
/* Call to [detachable] ET_INDEXING_TERM.position */
extern T0* T730x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_INDEXING_TERM.has_indexing_term_value */
extern T1 T730x15384579T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_enabled */
extern T1 T795x12910593T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_capability_supported */
extern T1 T795x12910594T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_capability_aware */
extern T1 T795x12959745T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.hash_code */
extern T6 T815x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_UNIVERSE.lower_name */
extern T0* T883x18530307(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_UNIVERSE.name */
extern T0* T883x18530306(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.same_call_name */
extern T1 T886x20365352T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ALIAS_NAME.same_alias_name */
extern T1 T886x20283436T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ALIAS_NAME.is_infix_minus */
extern T1 T886x20365334(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_infix_plus */
extern T1 T886x20365338(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_prefix_minus */
extern T1 T886x20365344(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_prefix_plus */
extern T1 T886x20365345(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.hash_code */
extern T6 T886x475137(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.alias_lower_name */
extern T0* T886x20283425(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.lower_name */
extern T0* T886x20365314(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.position */
extern T0* T886x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_prefixable */
extern T1 T886x20283420(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_infixable */
extern T1 T886x20283421(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_infix */
extern T1 T886x20365323(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.convert_keyword */
extern T0* T886x20283431(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_prefix */
extern T1 T886x20365343(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_parenthesis */
extern T1 T886x20365349(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.is_bracket */
extern T1 T886x20365348(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.cast_type */
extern T0* T946x15843335(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.index */
extern T6 T946x16891908(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.value */
extern T3 T946x15843334(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.value_position */
extern T0* T946x15843339(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.is_character_32 */
extern T1 T946x15843330(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.is_character_8 */
extern T1 T946x15843329(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.cast_type */
extern T0* T956x15925260(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.index */
extern T6 T956x16891908(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_negative */
extern T1 T956x15925267(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.value */
extern T11 T956x15925250(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.to_integer_32 */
extern T6 T956x15925253(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.literal */
extern T0* T956x15925249(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.sign */
extern T0* T956x15925259(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.value_position */
extern T0* T956x15925264(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_64 */
extern T1 T956x15925279(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_binary */
extern T1 T956x15925270(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T956x15925269(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_64 */
extern T1 T956x15925275(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_32 */
extern T1 T956x15925274(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_32 */
extern T1 T956x15925278(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_16 */
extern T1 T956x15925277(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_8 */
extern T1 T956x15925276(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_16 */
extern T1 T956x15925273(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_8 */
extern T1 T956x15925272(GE_context* ac, T0* C);
/* Call to detachable ET_AGENT_ARGUMENT_OPERAND_ITEM.agent_actual_argument */
extern T0* T1000x14090241(GE_context* ac, T0* C);
/* Call to detachable ET_ASSIGNER.feature_name */
extern T0* T1010x17203201(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.upper */
extern T0* T1022x19087363(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.is_range */
extern T1 T1022x19087361(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.lower */
extern T0* T1022x19087362(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_CONSTANT.position */
extern T0* T1025x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_ITEM.choice */
extern T0* T1028x19136513(GE_context* ac, T0* C);
/* Call to detachable ET_CONDITIONAL.expression */
extern T0* T1043x16318465(GE_context* ac, T0* C);
/* Call to detachable ET_CONDITIONAL.position */
extern T0* T1043x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.resolved_syntactical_constraint_with_type */
extern T0* T1048x21413892T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.type */
extern T0* T1048x21413890(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.position */
extern T0* T1048x21413891(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_TYPE.resolved_syntactical_constraint */
extern T0* T1060x21528578T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_CONVERT_FEATURE.name */
extern T0* T1063x14991363(GE_context* ac, T0* C);
/* Call to detachable ET_CONVERT_FEATURE.is_convert_to */
extern T1 T1063x14991362(GE_context* ac, T0* C);
/* Call to detachable ET_CONVERT_FEATURE.is_convert_from */
extern T1 T1063x14991361(GE_context* ac, T0* C);
/* Call to detachable ET_EXTERNAL_ALIAS.manifest_string */
extern T0* T1109x17563649(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME_ITEM.feature_name */
extern T0* T1123x20463617(GE_context* ac, T0* C);
/* Call to detachable ET_FEATURE_NAME_ITEM.position */
extern T0* T1123x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.indexing_term */
extern T0* T1154x15384577(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.position */
extern T0* T1154x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.has_indexing_term_value */
extern T1 T1154x15384579T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_LIKE_TYPE.to_text */
extern T0* T1173x21823562(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.position */
extern T0* T1173x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.resolved_formal_parameters */
extern T0* T1173x21823555T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_LIKE_TYPE.like_keyword */
extern T0* T1173x21692418(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.has_formal_types */
extern T1 T1173x21823527T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_LIKE_TYPE.has_unqualified_anchored_type */
extern T1 T1173x21823524(GE_context* ac, T0* C);
/* Call to detachable ET_MANIFEST_STRING_ITEM.manifest_string */
extern T0* T1190x16760833(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.type */
extern T0* T1238x15712258(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.renames */
extern T0* T1238x15712259(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT_ITEM.type_constraint */
extern T0* T1241x15712257(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT_ITEM.type */
extern T0* T1241x15712258(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_ITEM.type */
extern T0* T1247x21872641(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.last_leaf */
extern T0* T1292x19791877(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.type_constraint */
extern T0* T1292x14925825T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_CONSTRAINT.count */
extern T6 T1292x14925828(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.types_to_text */
extern T0* T1292x14925831(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.type_constraint */
extern T0* T1293x14925825T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.count */
extern T6 T1293x14925828(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.conforms_to_type_with_type_marks */
extern T1 T1293x14925830T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.types_to_text */
extern T0* T1293x14925831(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.renames */
extern T0* T1294x15712259(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.type */
extern T0* T1294x15712258(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.base_class */
extern T0* T1294x14696449(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.named_feature */
extern T0* T1294x14696453T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FLATTENED_FEATURE.adapted_feature */
extern T0* T1301x17743889(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_adapted */
extern T1 T1301x17743876(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.inherited_feature */
extern T0* T1301x17743887(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_inherited */
extern T1 T1301x17743874(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.redeclared_feature */
extern T0* T1301x17743888(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_redeclared */
extern T1 T1301x17743875(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_immediate */
extern T1 T1301x17743873(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.flattened_feature */
extern T0* T1301x17743885(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.immediate_feature */
extern T0* T1301x17743886(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_query */
extern T0* T1371x14696451T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_class */
extern T0* T1371x14696449(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_procedure */
extern T0* T1371x14696452T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_type */
extern T0* T1371x14696450(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_feature */
extern T0* T1371x14696453T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_type_index_of_label */
extern T6 T1371x14696454T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_ELEMENT_NODE.parent */
extern T0* T1422x29294593(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.last_leaf */
extern T0* T1459x19791877(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.type */
extern T0* T1459x21544961(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.position */
extern T0* T1459x19791873(GE_context* ac, T0* C);
/* Call to detachable ET_EXTERNAL_LANGUAGE.manifest_string */
extern T0* T1463x17596417(GE_context* ac, T0* C);
/* Call to detachable ET_UNTAGGED_ASSERTION.is_class_assertion */
extern T1 T1512x14499841(GE_context* ac, T0* C);
/* Call to detachable ET_ASSERTION_ITEM.assertion */
extern T0* T1514x14532609(GE_context* ac, T0* C);
/* Call to detachable ET_ASSERTION_ITEM.position */
extern T0* T1514x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.first_feature */
extern T0* T1631x18219011(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.replicated_feature */
extern T0* T1631x18219010(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.has_replication */
extern T1 T1631x18219009(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.selected_count */
extern T6 T1631x18219012(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.class_options */
extern T0* T1683x13418502(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.options */
extern T0* T1683x13418501(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.implementation_feature */
extern T0* T1694x17252365(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.is_once_per_process */
extern T1 T1694x17252354(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.is_once */
extern T1 T1694x17252353(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.lower_name */
extern T0* T1694x18366467(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.rescue_clause */
extern T0* T1694x17924099(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.compound */
extern T0* T1694x17924097(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE.arguments */
extern T0* T1694x17743882(GE_context* ac, T0* C);
/* Call to [detachable] ET_ROUTINE.implementation_feature */
extern T0* T1695x17252365(GE_context* ac, T0* C);
/* Call to [detachable] ET_ROUTINE.arguments */
extern T0* T1695x17743882(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_CLOSURE.locals */
extern T0* T1696x17252360(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_CLOSURE.compound */
extern T0* T1696x17924097(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_CLOSURE.is_once_per_process */
extern T1 T1696x17252354(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_CLOSURE.rescue_clause */
extern T0* T1696x17924099(GE_context* ac, T0* C);
/* Call to detachable ET_AST_LEAF.last_position */
extern T0* T1700x19791875(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_prefix_not */
extern T1 T1701x20365347(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and_then */
extern T1 T1701x20365325(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and */
extern T1 T1701x20365324(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_implies */
extern T1 T1701x20365331(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or_else */
extern T1 T1701x20365337(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or */
extern T1 T1701x20365336(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_tuple_label */
extern T1 T1701x20365321(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.position */
extern T0* T1701x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.is_cluster */
extern T1 T1705x18743299(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.absolute_pathname */
extern T0* T1705x18743322(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.universe */
extern T0* T1705x18743312(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.parent */
extern T0* T1705x18743339(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.full_name */
extern T0* T1705x18743318T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_GROUP.full_lower_name */
extern T0* T1705x18743319T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_GROUP.kind_lower_name */
extern T0* T1705x18743324(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.cluster */
extern T0* T1705x18743336(GE_context* ac, T0* C);
/* Call to detachable ET_AST_NODE.position */
extern T0* T1735x19791873(GE_context* ac, T0* C);
/* Call to [detachable] ET_FREE_NAME.operator_name */
extern T0* T1754x20611074(GE_context* ac, T0* C);
/* Call to [detachable] ET_FREE_NAME.hash_code */
extern T6 T1754x475137(GE_context* ac, T0* C);
/* Call to [detachable] ET_FREE_NAME.is_infix_freeop */
extern T1 T1754x20365328(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.target */
extern T0* T1894x16285697(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.name */
extern T0* T1894x16285698(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.is_call_agent */
extern T1 T1894x16285702(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.is_tuple_label */
extern T1 T1894x16285701(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.arguments */
extern T0* T1894x16285699(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_INLINE_AGENT.implicit_result */
extern T0* T1897x14024708(GE_context* ac, T0* C);
/* Call to [detachable] ET_LIKE_IDENTIFIER.name */
extern T0* T1918x21659649(GE_context* ac, T0* C);
/* Call to [detachable] ET_LIKE_IDENTIFIER.to_text */
extern T0* T1918x21823562(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.expression */
extern T0* T1922x16351233(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.index */
extern T6 T1922x16891908(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.convert_feature */
extern T0* T1922x16351234(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENTS.name */
extern T0* T1956x19038209(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENTS.arguments */
extern T0* T1956x19038210(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL.name */
extern T0* T1958x16285698(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUERY_INLINE_AGENT.type */
extern T0* T1963x17252357(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUERY_INLINE_AGENT.formal_arguments */
extern T0* T1963x17252358(GE_context* ac, T0* C);
/* Call to [detachable] ET_PROCEDURE_INLINE_AGENT.formal_arguments */
extern T0* T1966x17252358(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.parenthesis_call */
extern T0* T1968x19333123(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.name */
extern T0* T1968x16285698(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.target */
extern T0* T1968x16285697(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.arguments */
extern T0* T1968x16285699(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE_CALL_EXPRESSION.arguments */
extern T0* T1970x16285699(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENT_LIST.arguments */
extern T0* T1972x19038210(GE_context* ac, T0* C);
/* Call to [detachable] ET_CREATION_COMPONENT.position */
extern T0* T1976x19791873(GE_context* ac, T0* C);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].target */
extern T0* T2121x25919489(GE_context* ac, T0* C);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2121x25919490T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable EXCEPTION.set_type_name */
extern void T22x1572888T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_recipient_name */
extern void T22x1572891T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_description */
extern void T22x1572893T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_exception_trace */
extern void T22x1572887T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_throwing_exception */
extern void T22x1572892T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.read_character */
extern void T37x5881868(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.close */
extern void T37x5881866(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.record_occurrence */
extern void T49x5275667T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] AP_OPTION.reset */
extern void T49x5275666(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.append_to_string */
extern void T64x21823568T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.add_adapted_base_classes_to_list */
extern void T64x21823569T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.context_add_adapted_base_classes_to_list */
extern void T64x21856309T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.process */
extern void T64x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.append_runtime_name_to_string */
extern void T64x21823566T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_integer */
extern void T74x5783566T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_character */
extern void T74x5914630T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_new_line */
extern void T74x5996546(GE_context* ac, T0* C);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_string */
extern void T74x5783568T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.append */
extern void T74x5914629T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_line */
extern void T74x5996547T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS_FILTER.set_next */
extern void T176x27426818T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_xml_declaration */
extern void T177x27361293T0T0T1(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* Call to [detachable] XM_CALLBACKS.on_attribute */
extern void T177x27361288T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] XM_CALLBACKS.on_error */
extern void T177x27361292T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_end_tag */
extern void T177x27361286T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] XM_CALLBACKS.on_start_tag_finish */
extern void T177x27361287(GE_context* ac, T0* C);
/* Call to [detachable] XM_CALLBACKS.on_start_tag */
extern void T177x27361289T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] XM_CALLBACKS.on_content */
extern void T177x27361285T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_processing_instruction */
extern void T177x27361291T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_CALLBACKS.on_comment */
extern void T177x27361290T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_finish */
extern void T177x27361294(GE_context* ac, T0* C);
/* Call to [detachable] XM_CALLBACKS.on_start */
extern void T177x27361295(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_library_target */
extern void T181x13713426T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_uuid */
extern void T181x13713425T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_read_only */
extern void T181x13713429T1(GE_context* ac, T0* C, T1 a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_targets */
extern void T181x13713427T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_notes */
extern void T181x13713421T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_description */
extern void T181x13713422T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_version */
extern void T181x13713423T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_namespace */
extern void T181x13713424T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_static */
extern void T219x23199794T1(GE_context* ac, T0* C, T1 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_procedure_call */
extern void T219x23199785T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_query_call */
extern void T219x23199786T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_object_equality_expression */
extern void T219x23199783T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_equality_expression */
extern void T219x23199784T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.use_all_attributes */
extern void T219x23199789T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_alive */
extern void T219x23199795(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attribute_position */
extern void T219x23199790T0T6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_next_type */
extern void T219x23199782T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_hash_code */
extern void T219x23199791T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attached_type */
extern void T219x23199793T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_id */
extern void T219x23199792T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_FEATURE_NAME.set_seed */
extern void T229x20365359T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_FEATURE_NAME.reset */
extern void T229x20365360(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.add_creation_clients_to */
extern void T230x17662037T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_PROCEDURE.reset_after_interface_checked */
extern void T230x17662039(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.reset_after_features_flattened */
extern void T230x17662040(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.reset */
extern void T230x17662041(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.set_across_components */
extern void T230x17252366T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.set_object_tests */
extern void T230x17252367T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.set_frozen_keyword */
extern void T230x17662025T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.reset_validity_checked */
extern void T230x18366477(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset_after_interface_checked */
extern void T231x17662039(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset_after_features_flattened */
extern void T231x17662040(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset */
extern void T231x17662041(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.set_across_components */
extern void T231x17252366T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.set_object_tests */
extern void T231x17252367T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.set_frozen_keyword */
extern void T231x17662025T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.reset_validity_checked */
extern void T231x18366477(GE_context* ac, T0* C);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_to_expression */
extern void T233x24199332T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_from_expression */
extern void T233x24199334T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_builtin_expression */
extern void T233x24199338T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unqualified_call_instruction */
extern void T233x24199176T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_static_call_instruction */
extern void T233x24199195T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_precursor_instruction */
extern void T233x24199218T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_create_instruction */
extern void T233x24199330T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_bang_instruction */
extern void T233x24199367T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_loop_instruction */
extern void T233x24199247T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assignment_attempt */
extern void T233x24199371T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assignment */
extern void T233x24199372T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assigner_instruction */
extern void T233x24199373T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_inspect_instruction */
extern void T233x24199264T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_if_instruction */
extern void T233x24199277T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_debug_instruction */
extern void T233x24199323T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_check_instruction */
extern void T233x24199354T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_across_instruction */
extern void T233x24199391T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_retry_instruction */
extern void T233x24199199T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_infix_cast_expression */
extern void T233x24199270T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_constrained_formal_parameter */
extern void T233x24199341T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_formal_parameter */
extern void T233x24199287T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_integer_constant */
extern void T233x24199207T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_binary_integer_constant */
extern void T233x24199364T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_octal_integer_constant */
extern void T233x24199237T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_hexadecimal_integer_constant */
extern void T233x24199282T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_underscored_integer_constant */
extern void T233x24199183T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_verbatim_string */
extern void T233x24199174T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_special_manifest_string */
extern void T233x24199197T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c3_character_constant */
extern void T233x24199356T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_like_current */
extern void T233x24199254T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_attribute */
extern void T233x24199368T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unique_attribute */
extern void T233x24199178T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_constant_attribute */
extern void T233x24199342T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_extended_attribute */
extern void T233x24199303T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_create_expression */
extern void T233x24199331T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_equality_expression */
extern void T233x24199307T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_object_equality_expression */
extern void T233x24199239T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_object_test */
extern void T233x24199238T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unqualified_call_expression */
extern void T233x24199177T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_call_instruction */
extern void T233x24199211T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_tuple_type */
extern void T233x24199188T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_function */
extern void T233x24199301T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_formal_parameter_type */
extern void T233x24199284T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_class_type */
extern void T233x24199348T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_procedure_inline_agent */
extern void T233x24199298T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_procedure_inline_agent */
extern void T233x24199230T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_procedure_inline_agent */
extern void T233x24199317T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_function_inline_agent */
extern void T233x24199300T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_function_inline_agent */
extern void T233x24199233T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_function_inline_agent */
extern void T233x24199319T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_manifest_string */
extern void T233x24199232T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_old_object_test */
extern void T233x24199235T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_old_expression */
extern void T233x24199236T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_prefix_expression */
extern void T233x24199216T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_type */
extern void T233x24199241T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_infix_expression */
extern void T233x24199268T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_expression_address */
extern void T233x24199305T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_result_address */
extern void T233x24199200T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_current_address */
extern void T233x24199325T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_feature_address */
extern void T233x24199296T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_precursor_expression */
extern void T233x24199219T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_like_type */
extern void T233x24199209T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_like_braced_type */
extern void T233x24199210T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_like_feature */
extern void T233x24199253T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_procedure */
extern void T233x24199299T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_procedure */
extern void T233x24199231T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_deferred_function */
extern void T233x24199322T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_function */
extern void T233x24199234T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_function */
extern void T233x24199320T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c2_character_constant */
extern void T233x24199357T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c1_character_constant */
extern void T233x24199358T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_false_constant */
extern void T233x24199297T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_true_constant */
extern void T233x24199189T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_underscored_real_constant */
extern void T233x24199182T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_real_constant */
extern void T233x24199205T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_strip_expression */
extern void T233x24199194T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_static_call_expression */
extern void T233x24199196T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_parenthesized_expression */
extern void T233x24199222T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_array */
extern void T233x24199245T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_if_expression */
extern void T233x24199278T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_call_agent */
extern void T233x24199355T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_bracket_expression */
extern void T233x24199360T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_across_expression */
extern void T233x24199392T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_void */
extern void T233x24199173T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_manifest_string */
extern void T233x24199206T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_named_object_test */
extern void T233x24199240T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_deferred_procedure */
extern void T233x24199321T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_procedure */
extern void T233x24199318T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_tuple */
extern void T233x24199242T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_result */
extern void T233x24199201T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_call_expression */
extern void T233x24199212T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_current */
extern void T233x24199326T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_identifier */
extern void T233x24199281T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_class */
extern void T233x24199350T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.add_adapted_base_classes_to_list */
extern void T234x21823569T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.reset */
extern void T234x14925834(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.process */
extern void T234x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.reset_qualified_anchored_types */
extern void T234x21823570(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.append_to_string */
extern void T234x21823568T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.append_unaliased_to_string */
extern void T234x21823567T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.append_runtime_name_to_string */
extern void T234x21823566T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.universes_do_if */
extern void T238x18890883T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_if_until */
extern void T238x18890889T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_if */
extern void T238x18890890T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_until */
extern void T238x18890891T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_all */
extern void T238x18890892T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.classes_do_if */
extern void T238x18890900T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.classes_do_until */
extern void T238x18890901T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.classes_do_all */
extern void T238x18890902T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_MARK.append_to_string_with_space */
extern void T241x20217880T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T242x21200919(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.reset */
extern void T242x21200920(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_to_string */
extern void T242x21200918T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T242x21200917T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T242x21200916T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.put_target */
extern void T244x23642122T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_DYNAMIC_TYPE_SET.set_never_void */
extern void T244x23642123(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.propagate_can_be_void */
extern void T244x23527425T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.put_type_from_type_set */
extern void T244x23527427T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_DYNAMIC_TYPE.put_target */
extern void T245x23642122T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_TYPE.set_meta_type */
extern void T245x23609365T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_EXPRESSION.process */
extern void T272x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_EXPRESSION.reset */
extern void T272x14499844(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.set_index */
extern void T272x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_id */
extern void T276x17662035T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.process */
extern void T276x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_other_seeds */
extern void T276x17662028T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_first_seed */
extern void T276x17662029T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_clients */
extern void T276x17662036T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_version */
extern void T276x17662032T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_other_precursors */
extern void T276x17662026T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_first_precursor */
extern void T276x17662027T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.reset_postconditions */
extern void T276x17662021(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.reset_preconditions */
extern void T276x17662022(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.set_implementation_class */
extern void T276x17662031T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_implementation_feature */
extern void T276x17662030T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_validity_error */
extern void T276x18366478(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.set_validity_checked */
extern void T276x18366479(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.set_seed */
extern void T304x20365359T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_AGENT.set_arguments */
extern void T307x14024716T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT.process */
extern void T307x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_CONSTANT.set_index */
extern void T313x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_CONSTANT.process */
extern void T313x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_OPERAND.set_index */
extern void T337x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_AGENT_TARGET.process */
extern void T338x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT_TARGET.set_index */
extern void T338x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_AGENT_TARGET.reset */
extern void T338x16891910(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.process */
extern void T340x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.reset */
extern void T340x16891910(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.set_result_index */
extern void T360x14336012T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_INLINE_AGENT.set_attached_result_index */
extern void T360x14336011T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_INLINE_AGENT.set_across_components */
extern void T360x17252366T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INLINE_AGENT.set_object_tests */
extern void T360x17252367T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INLINE_AGENT.set_actual_arguments */
extern void T360x14024716T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.set_index */
extern void T362x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_index */
extern void T364x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_type */
extern void T364x15941648T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_cast_type */
extern void T364x15941649T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ASSERTION.process */
extern void T405x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ASSERTION.reset */
extern void T405x14499844(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXTENDED_FEATURE_NAME.reset */
extern void T467x20463621(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.close_input */
extern void T506x28377112(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_input_from_resolver */
extern void T506x28377113T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_encoding */
extern void T506x28377111T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.push_start_condition_dtd_ignore */
extern void T506x28377116(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.read_token */
extern void T506x12271674(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_input_stream */
extern void T506x28377114T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_DOCUMENT_NODE.node_set_parent */
extern void T571x29147138T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TARGET.propagate_can_be_void */
extern void T678x23527425T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TARGET.put_type_from_type_set */
extern void T678x23527427T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.clusters_do_explicit */
extern void T815x18776106T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.dotnet_assemblies_do_if */
extern void T815x18776099T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_ALIAS_NAME.reset */
extern void T886x20365360(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.set_prefix */
extern void T886x20283441(GE_context* ac, T0* C);
/* Call to detachable ET_ALIAS_NAME.set_infix */
extern void T886x20283440(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.set_index */
extern void T946x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_CHARACTER_CONSTANT.set_cast_type */
extern void T946x15843343T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_CHARACTER_CONSTANT.set_type */
extern void T946x15843342T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_index */
extern void T956x16891909T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_cast_type */
extern void T956x15925289T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_type */
extern void T956x15925288T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_sign */
extern void T956x15925290T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ASSIGNER.reset */
extern void T1010x17203202(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.reset */
extern void T1022x19087365(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_CONSTANT.reset */
extern void T1025x19087365(GE_context* ac, T0* C);
/* Call to detachable ET_INSTRUCTION.process */
extern void T1165x19791882T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INSTRUCTION.reset */
extern void T1165x19398658(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.reset_qualified_anchored_types */
extern void T1173x21823570(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.reset */
extern void T1173x14925834(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.append_to_string */
extern void T1173x21823568T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_CONSTRAINT.reset */
extern void T1238x14925834(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.reset_renames */
extern void T1238x14925833(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.reset */
extern void T1292x14925834(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.reset_renames */
extern void T1292x14925833(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.add_overloaded_procedures */
extern void T1371x14696455T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_ADAPTED_CLASS.add_overloaded_queries */
extern void T1371x14696456T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_ELEMENT_NODE.node_set_parent */
extern void T1422x29147138T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_UNTAGGED_ASSERTION.reset */
extern void T1512x14499844(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.set_options */
extern void T1683x13418512T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_OPERATOR.set_seed */
extern void T1701x20365359T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_OPERATOR.reset */
extern void T1701x20365360(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.set_parenthesis_call */
extern void T1968x19611650T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENT_LIST.set_arguments */
extern void T1972x19054594T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2121x25919492T0(GE_context* ac, T0* C, T0* a1);
/* New instance of type [detachable] SPECIAL [CHARACTER_8] */
extern T0* GE_new15(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [CHARACTER_32] */
extern T0* GE_new16(T6 a1, T1 initialize);
/* New instance of type [detachable] STRING_8 */
extern T0* GE_new17(T1 initialize);
/* New instance of type [detachable] STRING_32 */
extern T0* GE_new18(T1 initialize);
/* New instance of type [detachable] IMMUTABLE_STRING_32 */
extern T0* GE_new20(T1 initialize);
/* New instance of type [detachable] ISE_EXCEPTION_MANAGER */
extern T0* GE_new21(T1 initialize);
/* New instance of type [detachable] GEC */
extern T0* GE_new25(T1 initialize);
/* New instance of type [detachable] CELL [detachable EXCEPTION] */
extern T0* GE_new26(T1 initialize);
/* New instance of type [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
extern T0* GE_new27(T1 initialize);
/* New instance of type [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
extern T0* GE_new28(T1 initialize);
/* New instance of type [detachable] CELL [[attached] NO_MORE_MEMORY] */
extern T0* GE_new29(T1 initialize);
/* New instance of type detachable C_STRING */
extern T0* GE_new30(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
extern T0* GE_new31(T1 initialize);
/* New instance of type [detachable] KL_TEXT_INPUT_FILE */
extern T0* GE_new32(T1 initialize);
/* New instance of type [detachable] KL_ARGUMENTS */
extern T0* GE_new33(T1 initialize);
/* New instance of type [detachable] ET_ISE_VARIABLES */
extern T0* GE_new34(T1 initialize);
/* New instance of type [detachable] ET_ERROR_HANDLER */
extern T0* GE_new35(T1 initialize);
/* New instance of type detachable ET_SYSTEM */
extern T0* GE_new36(T1 initialize);
/* New instance of type [detachable] KL_EXCEPTIONS */
extern T0* GE_new38(T1 initialize);
/* New instance of type [detachable] AP_PARSER */
extern T0* GE_new39(T1 initialize);
/* New instance of type [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
extern T0* GE_new40(T1 initialize);
/* New instance of type [detachable] AP_STRING_OPTION */
extern T0* GE_new41(T1 initialize);
/* New instance of type [detachable] AP_FLAG */
extern T0* GE_new43(T1 initialize);
/* New instance of type [detachable] UT_VERSION */
extern T0* GE_new44(T1 initialize);
/* New instance of type [detachable] AP_ENUMERATION_OPTION */
extern T0* GE_new45(T1 initialize);
/* New instance of type [detachable] AP_BOOLEAN_OPTION */
extern T0* GE_new46(T1 initialize);
/* New instance of type [detachable] AP_INTEGER_OPTION */
extern T0* GE_new47(T1 initialize);
/* New instance of type [detachable] ET_NULL_ERROR_HANDLER */
extern T0* GE_new51(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM_PARSER */
extern T0* GE_new53(T1 initialize);
/* New instance of type [detachable] ET_ECF_ERROR_HANDLER */
extern T0* GE_new54(T1 initialize);
/* New instance of type detachable ET_ECF_SETTINGS */
extern T0* GE_new55(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM */
extern T0* GE_new56(T1 initialize);
/* New instance of type [detachable] ET_ECF_TARGET */
extern T0* GE_new57(T1 initialize);
/* New instance of type detachable ET_ECF_CAPABILITIES */
extern T0* GE_new59(T1 initialize);
/* New instance of type detachable ET_ECF_VARIABLES */
extern T0* GE_new60(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_SYSTEM */
extern T0* GE_new61(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_PROCESSOR */
extern T0* GE_new62(T1 initialize);
/* New instance of type [detachable] ET_C_GENERATOR */
extern T0* GE_new65(T1 initialize);
/* New instance of type [detachable] KL_SHELL_COMMAND */
extern T0* GE_new66(T1 initialize);
/* New instance of type detachable DT_DATE_TIME */
extern T0* GE_new67(T1 initialize);
/* New instance of type detachable DS_HASH_SET [[attached] STRING_8] */
extern T0* GE_new69(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
extern T0* GE_new70(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE */
extern T0* GE_new71(T1 initialize);
/* New instance of type [detachable] ET_TOKEN_CONSTANTS */
extern T0* GE_new72(T1 initialize);
/* New instance of type [detachable] ET_CLASS */
extern T0* GE_new73(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new75(T1 initialize);
/* New instance of type [detachable] KL_OPERATING_SYSTEM */
extern T0* GE_new76(T1 initialize);
/* New instance of type [detachable] UT_CANNOT_READ_FILE_ERROR */
extern T0* GE_new79(T1 initialize);
/* New instance of type [detachable] UT_VERSION_NUMBER */
extern T0* GE_new81(T1 initialize);
/* New instance of type [detachable] UT_MESSAGE */
extern T0* GE_new82(T1 initialize);
/* New instance of type [detachable] RX_PCRE_REGULAR_EXPRESSION */
extern T0* GE_new83(T1 initialize);
/* New instance of type [detachable] KL_STRING_ROUTINES */
extern T0* GE_new84(T1 initialize);
/* New instance of type [detachable] KL_EXECUTION_ENVIRONMENT */
extern T0* GE_new87(T1 initialize);
/* New instance of type [detachable] AP_ERROR */
extern T0* GE_new88(T1 initialize);
/* New instance of type [detachable] VOID_TARGET */
extern T0* GE_new91(T1 initialize);
/* New instance of type [detachable] ROUTINE_FAILURE */
extern T0* GE_new93(T1 initialize);
/* New instance of type [detachable] OLD_VIOLATION */
extern T0* GE_new95(T1 initialize);
/* New instance of type [detachable] NO_MORE_MEMORY */
extern T0* GE_new97(T1 initialize);
/* New instance of type [detachable] INVARIANT_VIOLATION */
extern T0* GE_new98(T1 initialize);
/* New instance of type [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
extern T0* GE_new99(T1 initialize);
/* New instance of type [detachable] IO_FAILURE */
extern T0* GE_new100(T1 initialize);
/* New instance of type [detachable] OPERATING_SYSTEM_FAILURE */
extern T0* GE_new101(T1 initialize);
/* New instance of type [detachable] COM_FAILURE */
extern T0* GE_new102(T1 initialize);
/* New instance of type [detachable] EIFFEL_RUNTIME_PANIC */
extern T0* GE_new103(T1 initialize);
/* New instance of type [detachable] PRECONDITION_VIOLATION */
extern T0* GE_new105(T1 initialize);
/* New instance of type [detachable] POSTCONDITION_VIOLATION */
extern T0* GE_new106(T1 initialize);
/* New instance of type [detachable] FLOATING_POINT_FAILURE */
extern T0* GE_new107(T1 initialize);
/* New instance of type [detachable] CHECK_VIOLATION */
extern T0* GE_new108(T1 initialize);
/* New instance of type [detachable] BAD_INSPECT_VALUE */
extern T0* GE_new109(T1 initialize);
/* New instance of type [detachable] VARIANT_VIOLATION */
extern T0* GE_new110(T1 initialize);
/* New instance of type [detachable] LOOP_INVARIANT_VIOLATION */
extern T0* GE_new111(T1 initialize);
/* New instance of type [detachable] RESCUE_FAILURE */
extern T0* GE_new112(T1 initialize);
/* New instance of type [detachable] RESUMPTION_FAILURE */
extern T0* GE_new113(T1 initialize);
/* New instance of type [detachable] CREATE_ON_DEFERRED */
extern T0* GE_new114(T1 initialize);
/* New instance of type [detachable] EXTERNAL_FAILURE */
extern T0* GE_new115(T1 initialize);
/* New instance of type [detachable] VOID_ASSIGNED_TO_EXPANDED */
extern T0* GE_new116(T1 initialize);
/* New instance of type [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
extern T0* GE_new117(T1 initialize);
/* New instance of type [detachable] MISMATCH_FAILURE */
extern T0* GE_new118(T1 initialize);
/* New instance of type [detachable] DEVELOPER_EXCEPTION */
extern T0* GE_new119(T1 initialize);
/* New instance of type [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
extern T0* GE_new120(T1 initialize);
/* New instance of type [detachable] SERIALIZATION_FAILURE */
extern T0* GE_new121(T1 initialize);
/* New instance of type [detachable] EXECUTION_ENVIRONMENT */
extern T0* GE_new122(T1 initialize);
/* New instance of type [detachable] KL_WINDOWS_FILE_SYSTEM */
extern T0* GE_new123(T1 initialize);
/* New instance of type [detachable] KL_UNIX_FILE_SYSTEM */
extern T0* GE_new124(T1 initialize);
/* New instance of type [detachable] PRIMES */
extern T0* GE_new125(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] INTEGER_32] */
extern T0* GE_new126(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] BOOLEAN] */
extern T0* GE_new127(T6 a1, T1 initialize);
/* New instance of type detachable KL_LINKABLE [[attached] CHARACTER_8] */
extern T0* GE_new129(T1 initialize);
/* New instance of type [detachable] MANAGED_POINTER */
extern T0* GE_new131(T1 initialize);
/* New instance of type [detachable] FILE_INFO */
extern T0* GE_new132(T1 initialize);
/* New instance of type [detachable] MUTEX */
extern T0* GE_new133(T1 initialize);
/* New instance of type [detachable] UT_ERROR_HANDLER */
extern T0* GE_new134(T1 initialize);
/* New instance of type [detachable] KL_STANDARD_FILES */
extern T0* GE_new135(T1 initialize);
/* New instance of type [detachable] KL_STDERR_FILE */
extern T0* GE_new136(T1 initialize);
/* New instance of type [detachable] KL_STDOUT_FILE */
extern T0* GE_new137(T1 initialize);
/* New instance of type [detachable] ST_WORD_WRAPPER */
extern T0* GE_new138(T1 initialize);
/* New instance of type [detachable] AP_DISPLAY_HELP_FLAG */
extern T0* GE_new140(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
extern T0* GE_new141(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new142(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
extern T0* GE_new143(T1 initialize);
/* New instance of type [detachable] AP_ERROR_HANDLER */
extern T0* GE_new144(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
extern T0* GE_new145(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
extern T0* GE_new146(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
extern T0* GE_new147(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] STRING_8] */
extern T0* GE_new148(T1 initialize);
/* New instance of type [detachable] KL_STRING_EQUALITY_TESTER */
extern T0* GE_new149(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
extern T0* GE_new150(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
extern T0* GE_new153(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
extern T0* GE_new156(T1 initialize);
/* New instance of type [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
extern T0* GE_new157(T1 initialize);
/* New instance of type [detachable] ET_ECF_AST_FACTORY */
extern T0* GE_new158(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
extern T0* GE_new159(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET] */
extern T0* GE_new160(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
extern T0* GE_new161(T1 initialize);
/* New instance of type detachable TUPLE */
extern T0* GE_new162(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
extern T0* GE_new163(T1 initialize);
/* New instance of type [detachable] XM_ELEMENT */
extern T0* GE_new164(T1 initialize);
/* New instance of type detachable XM_POSITION_TABLE */
extern T0* GE_new165(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
extern T0* GE_new166(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
extern T0* GE_new167(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
extern T0* GE_new168(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
extern T0* GE_new170(T1 initialize);
/* New instance of type [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
extern T0* GE_new171(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
extern T0* GE_new172(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PARSER */
extern T0* GE_new174(T1 initialize);
/* New instance of type [detachable] XM_TREE_CALLBACKS_PIPE */
extern T0* GE_new175(T1 initialize);
/* New instance of type [detachable] XM_CALLBACKS_TO_TREE_FILTER */
extern T0* GE_new178(T1 initialize);
/* New instance of type [detachable] ET_ECF_STATE */
extern T0* GE_new179(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER */
extern T0* GE_new180(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM_CONFIG */
extern T0* GE_new181(T1 initialize);
/* New instance of type [detachable] ET_ECF_LIBRARY */
extern T0* GE_new183(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_LIBRARIES */
extern T0* GE_new184(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new186(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
extern T0* GE_new187(T1 initialize);
/* New instance of type [detachable] XM_DOCUMENT */
extern T0* GE_new188(T1 initialize);
/* New instance of type [detachable] ET_COMPRESSED_POSITION */
extern T0* GE_new189(T1 initialize);
/* New instance of type [detachable] XM_STOP_ON_ERROR_FILTER */
extern T0* GE_new191(T1 initialize);
/* New instance of type detachable XM_ATTRIBUTE */
extern T0* GE_new192(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
extern T0* GE_new194(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new195(T1 initialize);
/* New instance of type detachable ET_ECF_TARGET_PARENT */
extern T0* GE_new196(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new197(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
extern T0* GE_new198(T1 initialize);
/* New instance of type detachable ET_ECF_TARGETS */
extern T0* GE_new199(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new200(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new201(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
extern T0* GE_new202(T1 initialize);
/* New instance of type [detachable] ET_ECF_ADAPTED_LIBRARY */
extern T0* GE_new203(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
extern T0* GE_new204(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
extern T0* GE_new206(T1 initialize);
/* New instance of type [detachable] ET_ECF_DOTNET_ASSEMBLY */
extern T0* GE_new207(T1 initialize);
/* New instance of type [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
extern T0* GE_new209(T1 initialize);
/* New instance of type [detachable] ET_ECF_ERROR */
extern T0* GE_new210(T1 initialize);
/* New instance of type [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
extern T0* GE_new211(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
extern T0* GE_new212(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] STRING_8] */
extern T0* GE_new213(T1 initialize);
/* New instance of type [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
extern T0* GE_new214(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
extern T0* GE_new215(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
extern T0* GE_new217(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE */
extern T0* GE_new219(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_FEATURE_LIST */
extern T0* GE_new220(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
extern T0* GE_new221(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
extern T0* GE_new222(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLASS] */
extern T0* GE_new223(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
extern T0* GE_new224(T1 initialize);
/* New instance of type [detachable] TUPLE [INTEGER_32] */
extern T0* GE_new225(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
extern T0* GE_new226(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
extern T0* GE_new227(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_FEATURE */
extern T0* GE_new232(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_PARAMETER_LIST */
extern T0* GE_new236(T1 initialize);
/* New instance of type [detachable] ET_NESTED_TYPE_CONTEXT */
extern T0* GE_new237(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_TYPE */
extern T0* GE_new246(T1 initialize);
/* New instance of type [detachable] ET_EXTERNAL_FUNCTION */
extern T0* GE_new247(T1 initialize);
/* New instance of type [detachable] ET_TUPLE_TYPE */
extern T0* GE_new248(T1 initialize);
/* New instance of type detachable ET_FORMAL_ARGUMENT_LIST */
extern T0* GE_new249(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] ET_TYPE] */
extern T0* GE_new250(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE] */
extern T0* GE_new251(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_FORMAL_ARGUMENT */
extern T0* GE_new252(T1 initialize);
/* New instance of type [detachable] ET_QUERY_LIST */
extern T0* GE_new253(T1 initialize);
/* New instance of type [detachable] ET_PROCEDURE_LIST */
extern T0* GE_new254(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_TYPE_SET_LIST */
extern T0* GE_new255(T1 initialize);
/* New instance of type [detachable] KL_STRING_OUTPUT_STREAM */
extern T0* GE_new256(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
extern T0* GE_new257(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new258(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
extern T0* GE_new259(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
extern T0* GE_new260(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
extern T0* GE_new261(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
extern T0* GE_new262(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
extern T0* GE_new263(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
extern T0* GE_new264(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
extern T0* GE_new265(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new266(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new267(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
extern T0* GE_new268(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new269(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new270(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
extern T0* GE_new271(T1 initialize);
/* New instance of type [detachable] ET_CURRENT */
extern T0* GE_new273(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_ARGUMENT_LIST */
extern T0* GE_new274(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
extern T0* GE_new275(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL_EXPRESSION */
extern T0* GE_new278(T1 initialize);
/* New instance of type [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
extern T0* GE_new279(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new280(T1 initialize);
/* New instance of type [detachable] ET_MANIFEST_TUPLE */
extern T0* GE_new281(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new282(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
extern T0* GE_new283(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] INTEGER_32] */
extern T0* GE_new284(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
extern T0* GE_new285(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
extern T0* GE_new286(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new287(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new288(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
extern T0* GE_new289(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
extern T0* GE_new290(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER_TESTER */
extern T0* GE_new291(T1 initialize);
/* New instance of type [detachable] KL_TEXT_OUTPUT_FILE */
extern T0* GE_new293(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_TUPLE_TYPE */
extern T0* GE_new294(T1 initialize);
/* New instance of type [detachable] DS_STRING_HASH_TABLE */
extern T0* GE_new295(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
extern T0* GE_new296(T1 initialize);
/* New instance of type [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
extern T0* GE_new297(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_PRECURSOR */
extern T0* GE_new298(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_PRECURSOR_LIST */
extern T0* GE_new299(T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new300(T1 initialize);
/* New instance of type [detachable] ET_RESULT */
extern T0* GE_new301(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
extern T0* GE_new302(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
extern T0* GE_new306(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST */
extern T0* GE_new308(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
extern T0* GE_new309(T1 initialize);
/* New instance of type [detachable] ET_EQUALITY_EXPRESSION */
extern T0* GE_new310(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_SPECIAL_TYPE */
extern T0* GE_new311(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new314(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_SECONDARY_TYPE */
extern T0* GE_new315(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
extern T0* GE_new316(T1 initialize);
/* New instance of type [detachable] ET_CREATE_EXPRESSION */
extern T0* GE_new320(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL */
extern T0* GE_new321(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] STRING_8] */
extern T0* GE_new324(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] STRING_8] */
extern T0* GE_new325(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_DIRECTORY */
extern T0* GE_new327(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new328(T1 initialize);
/* New instance of type [detachable] ET_EXTENDED_ATTRIBUTE */
extern T0* GE_new329(T1 initialize);
/* New instance of type detachable ET_COMPOUND */
extern T0* GE_new330(T1 initialize);
/* New instance of type detachable ET_LOCAL_VARIABLE_LIST */
extern T0* GE_new331(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
extern T0* GE_new335(T1 initialize);
/* New instance of type [detachable] ET_IMPLICIT_TYPE_MARK */
extern T0* GE_new342(T1 initialize);
/* New instance of type [detachable] ET_SYMBOL */
extern T0* GE_new343(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE] */
extern T0* GE_new344(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new345(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new346(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new347(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new348(T1 initialize);
/* New instance of type [detachable] UT_INTEGER_FORMATTER */
extern T0* GE_new349(T1 initialize);
/* New instance of type [detachable] STD_FILES */
extern T0* GE_new352(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new354(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new356(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new357(T1 initialize);
/* New instance of type detachable ET_ACROSS_COMPONENT_LIST */
extern T0* GE_new361(T1 initialize);
/* New instance of type [detachable] ET_CONSTANT_ATTRIBUTE */
extern T0* GE_new363(T1 initialize);
/* New instance of type [detachable] ET_UNIQUE_ATTRIBUTE */
extern T0* GE_new365(T1 initialize);
/* New instance of type [detachable] ET_REGULAR_INTEGER_CONSTANT */
extern T0* GE_new366(T1 initialize);
/* New instance of type [detachable] ET_ATTRIBUTE */
extern T0* GE_new367(T1 initialize);
/* New instance of type [detachable] NATIVE_STRING */
extern T0* GE_new370(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_MULTIPROCESSOR */
extern T0* GE_new371(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
extern T0* GE_new372(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
extern T0* GE_new373(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
extern T0* GE_new375(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_CHECKER */
extern T0* GE_new376(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new377(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new378(T1 initialize);
/* New instance of type [detachable] ET_TYPE_CHECKER */
extern T0* GE_new380(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
extern T0* GE_new382(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new383(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new384(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new385(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new386(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new387(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new388(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST_SCOPE */
extern T0* GE_new389(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
extern T0* GE_new390(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new391(T1 initialize);
/* New instance of type [detachable] ET_ACROSS_CURSOR_SCOPE */
extern T0* GE_new392(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_SCOPE */
extern T0* GE_new393(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
extern T0* GE_new394(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new395(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
extern T0* GE_new396(T1 initialize);
/* New instance of type [detachable] ET_CLIENT_LIST */
extern T0* GE_new397(T1 initialize);
/* New instance of type [detachable] ET_CLIENT */
extern T0* GE_new398(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new399(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
extern T0* GE_new400(T1 initialize);
/* New instance of type detachable ET_PRECONDITIONS */
extern T0* GE_new401(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
extern T0* GE_new402(T1 initialize);
/* New instance of type detachable ET_FEATURE_LIST */
extern T0* GE_new404(T1 initialize);
/* New instance of type [detachable] ET_UNKNOWN_GROUP */
extern T0* GE_new406(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_LIST */
extern T0* GE_new408(T1 initialize);
/* New instance of type [detachable] ET_KEYWORD */
extern T0* GE_new409(T1 initialize);
/* New instance of type [detachable] ET_CLASS_CODES */
extern T0* GE_new410(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new411(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new412(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new413(T1 initialize);
/* New instance of type [detachable] UC_UTF8_ROUTINES */
extern T0* GE_new416(T1 initialize);
/* New instance of type [detachable] PATH */
extern T0* GE_new417(T1 initialize);
/* New instance of type [detachable] RX_BYTE_CODE */
extern T0* GE_new419(T1 initialize);
/* New instance of type [detachable] RX_CHARACTER_SET */
extern T0* GE_new420(T1 initialize);
/* New instance of type [detachable] RX_CASE_MAPPING */
extern T0* GE_new421(T1 initialize);
/* New instance of type [detachable] UC_UNICODE_ROUTINES */
extern T0* GE_new423(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
extern T0* GE_new424(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
extern T0* GE_new425(T6 a1, T1 initialize);
/* New instance of type [detachable] UC_STRING */
extern T0* GE_new426(T1 initialize);
/* New instance of type [detachable] KL_ANY_ROUTINES */
extern T0* GE_new427(T1 initialize);
/* New instance of type [detachable] STRING_TO_INTEGER_CONVERTOR */
extern T0* GE_new428(T1 initialize);
/* New instance of type [detachable] ARGUMENTS_32 */
extern T0* GE_new431(T1 initialize);
/* New instance of type [detachable] ET_ECF_OPTIONS */
extern T0* GE_new437(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new438(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE */
extern T0* GE_new439(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new440(T1 initialize);
/* New instance of type [detachable] ET_LIKE_CURRENT */
extern T0* GE_new442(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
extern T0* GE_new443(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_FUNCTION_TYPE */
extern T0* GE_new444(T1 initialize);
/* New instance of type [detachable] DT_SHARED_SYSTEM_CLOCK */
extern T0* GE_new445(T1 initialize);
/* New instance of type [detachable] DT_SYSTEM_CLOCK */
extern T0* GE_new446(T1 initialize);
/* New instance of type [detachable] ET_EIFFEL_PREPARSER */
extern T0* GE_new447(T1 initialize);
/* New instance of type [detachable] ET_MASTER_CLASS_CHECKER */
extern T0* GE_new448(T1 initialize);
/* New instance of type [detachable] ET_EIFFEL_PARSER */
extern T0* GE_new449(T1 initialize);
/* New instance of type [detachable] ET_PROVIDER_CHECKER */
extern T0* GE_new450(T1 initialize);
/* New instance of type [detachable] ET_ANCESTOR_BUILDER */
extern T0* GE_new451(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_FLATTENER */
extern T0* GE_new452(T1 initialize);
/* New instance of type [detachable] ET_INTERFACE_CHECKER */
extern T0* GE_new453(T1 initialize);
/* New instance of type [detachable] ET_IMPLEMENTATION_CHECKER */
extern T0* GE_new454(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
extern T0* GE_new455(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
extern T0* GE_new456(T1 initialize);
/* New instance of type [detachable] ET_AST_NULL_PROCESSOR */
extern T0* GE_new457(T1 initialize);
/* New instance of type [detachable] ET_AST_FACTORY */
extern T0* GE_new458(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
extern T0* GE_new460(T1 initialize);
/* New instance of type [detachable] DT_DATE_TIME_DURATION */
extern T0* GE_new461(T1 initialize);
/* New instance of type [detachable] UT_CONFIG_PARSER */
extern T0* GE_new462(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
extern T0* GE_new463(T1 initialize);
/* New instance of type [detachable] ET_DO_PROCEDURE */
extern T0* GE_new466(T1 initialize);
/* New instance of type [detachable] ET_DEFERRED_PROCEDURE */
extern T0* GE_new468(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
extern T0* GE_new469(T1 initialize);
/* New instance of type [detachable] KL_PLATFORM */
extern T0* GE_new470(T1 initialize);
/* New instance of type [detachable] CELL [[attached] INTEGER_32] */
extern T0* GE_new472(T1 initialize);
/* New instance of type [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
extern T0* GE_new473(T1 initialize);
/* New instance of type [detachable] KL_PATHNAME */
extern T0* GE_new475(T1 initialize);
/* New instance of type [detachable] EXCEPTIONS */
extern T0* GE_new476(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_8] */
extern T0* GE_new477(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
extern T0* GE_new479(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] AP_OPTION] */
extern T0* GE_new481(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
extern T0* GE_new482(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new483(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new484(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new485(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new486(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable STRING_8] */
extern T0* GE_new487(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
extern T0* GE_new488(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new489(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] STRING_8] */
extern T0* GE_new490(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] BOOLEAN] */
extern T0* GE_new491(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
extern T0* GE_new492(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] INTEGER_32] */
extern T0* GE_new493(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new494(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
extern T0* GE_new496(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new497(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new500(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
extern T0* GE_new501(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
extern T0* GE_new504(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
extern T0* GE_new505(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_SCANNER */
extern T0* GE_new506(T1 initialize);
/* New instance of type [detachable] XM_DEFAULT_POSITION */
extern T0* GE_new507(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
extern T0* GE_new509(T1 initialize);
/* New instance of type [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
extern T0* GE_new510(T1 initialize);
/* New instance of type [detachable] XM_CALLBACKS_NULL */
extern T0* GE_new511(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
extern T0* GE_new512(T1 initialize);
/* New instance of type [detachable] XM_NULL_EXTERNAL_RESOLVER */
extern T0* GE_new514(T1 initialize);
/* New instance of type [detachable] XM_DTD_CALLBACKS_NULL */
extern T0* GE_new516(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
extern T0* GE_new517(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ANY] */
extern T0* GE_new518(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new519(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PARSER_NAME */
extern T0* GE_new520(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new521(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
extern T0* GE_new522(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new523(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
extern T0* GE_new524(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
extern T0* GE_new525(T1 initialize);
/* New instance of type [detachable] XM_DTD_EXTERNAL_ID */
extern T0* GE_new526(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
extern T0* GE_new527(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new528(T1 initialize);
/* New instance of type [detachable] XM_DTD_ELEMENT_CONTENT */
extern T0* GE_new529(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new530(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
extern T0* GE_new531(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new532(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
extern T0* GE_new533(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new534(T1 initialize);
/* New instance of type [detachable] XM_DTD_ATTRIBUTE_CONTENT */
extern T0* GE_new535(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new536(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
extern T0* GE_new537(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
extern T0* GE_new538(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
extern T0* GE_new539(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
extern T0* GE_new540(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
extern T0* GE_new541(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_DECLARATION */
extern T0* GE_new542(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
extern T0* GE_new543(T6 a1, T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_ENTITY_DEF */
extern T0* GE_new546(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_SCANNER_DTD */
extern T0* GE_new547(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PE_ENTITY_DEF */
extern T0* GE_new549(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE_RESOLVER */
extern T0* GE_new550(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
extern T0* GE_new551(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
extern T0* GE_new552(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
extern T0* GE_new553(T1 initialize);
/* New instance of type [detachable] ET_CLUSTERS */
extern T0* GE_new554(T1 initialize);
/* New instance of type [detachable] ET_CLASS_NAME_TESTER */
extern T0* GE_new555(T1 initialize);
/* New instance of type [detachable] ET_PARENT */
extern T0* GE_new557(T1 initialize);
/* New instance of type [detachable] ET_PARENT_LIST */
extern T0* GE_new558(T1 initialize);
/* New instance of type [detachable] ET_BUILTIN_CONVERT_FEATURE */
extern T0* GE_new559(T1 initialize);
/* New instance of type [detachable] ET_MASTER_CLASS */
extern T0* GE_new560(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
extern T0* GE_new561(T1 initialize);
/* New instance of type detachable ET_RENAME_LIST */
extern T0* GE_new562(T1 initialize);
/* New instance of type detachable ET_EXPORT_LIST */
extern T0* GE_new563(T1 initialize);
/* New instance of type detachable ET_KEYWORD_FEATURE_NAME_LIST */
extern T0* GE_new564(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
extern T0* GE_new567(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new568(T1 initialize);
/* New instance of type [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new569(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new572(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new574(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
extern T0* GE_new575(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new576(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
extern T0* GE_new577(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
extern T0* GE_new579(T1 initialize);
/* New instance of type detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new581(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
extern T0* GE_new582(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
extern T0* GE_new583(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new584(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLIES */
extern T0* GE_new585(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY */
extern T0* GE_new586(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
extern T0* GE_new589(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
extern T0* GE_new590(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
extern T0* GE_new591(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new592(T1 initialize);
/* New instance of type detachable ET_FORMAL_PARAMETER_LIST */
extern T0* GE_new593(T1 initialize);
/* New instance of type detachable ET_FEATURE_IDS */
extern T0* GE_new594(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new595(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new596(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
extern T0* GE_new599(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
extern T0* GE_new600(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
extern T0* GE_new601(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
extern T0* GE_new602(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
extern T0* GE_new603(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
extern T0* GE_new605(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new609(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_BRACKET_SYMBOL */
extern T0* GE_new610(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new612(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
extern T0* GE_new613(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
extern T0* GE_new614(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
extern T0* GE_new615(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
extern T0* GE_new616(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new617(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new618(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new619(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new620(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPRESSION] */
extern T0* GE_new622(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
extern T0* GE_new623(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
extern T0* GE_new624(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
extern T0* GE_new625(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
extern T0* GE_new626(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new627(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new628(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new629(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
extern T0* GE_new631(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
extern T0* GE_new632(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
extern T0* GE_new633(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new634(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new635(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new636(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new637(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new638(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new639(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT] */
extern T0* GE_new640(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
extern T0* GE_new641(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
extern T0* GE_new642(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
extern T0* GE_new643(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
extern T0* GE_new645(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new646(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new647(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new648(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new649(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new650(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new651(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
extern T0* GE_new653(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE] */
extern T0* GE_new654(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
extern T0* GE_new655(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
extern T0* GE_new656(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONSTANT] */
extern T0* GE_new657(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
extern T0* GE_new659(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
extern T0* GE_new660(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
extern T0* GE_new661(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
extern T0* GE_new663(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
extern T0* GE_new664(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new666(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
extern T0* GE_new667(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new668(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
extern T0* GE_new669(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
extern T0* GE_new672(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new675(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new676(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] BOOLEAN] */
extern T0* GE_new677(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new680(T1 initialize);
/* New instance of type [detachable] WORKER_THREAD */
extern T0* GE_new681(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
extern T0* GE_new682(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new686(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new687(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new688(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new689(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
extern T0* GE_new690(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
extern T0* GE_new691(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
extern T0* GE_new692(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new693(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
extern T0* GE_new694(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new695(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new696(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new697(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
extern T0* GE_new698(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new699(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new700(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new701(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
extern T0* GE_new702(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new703(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new704(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new705(T1 initialize);
/* New instance of type [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new706(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new707(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new708(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new709(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new710(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new711(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new712(T1 initialize);
/* New instance of type [detachable] ET_NAMED_OBJECT_TEST */
extern T0* GE_new715(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new716(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new717(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new718(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new719(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new721(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new722(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new723(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] INTEGER_32] */
extern T0* GE_new724(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new726(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new727(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new728(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
extern T0* GE_new729(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
extern T0* GE_new731(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
extern T0* GE_new732(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
extern T0* GE_new733(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
extern T0* GE_new734(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new735(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new736(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new737(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
extern T0* GE_new738(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
extern T0* GE_new739(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
extern T0* GE_new741(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
extern T0* GE_new742(T1 initialize);
/* New instance of type [detachable] UC_UTF8_STRING */
extern T0* GE_new743(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_32] */
extern T0* GE_new745(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
extern T0* GE_new746(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
extern T0* GE_new747(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_64] */
extern T0* GE_new748(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
extern T0* GE_new749(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
extern T0* GE_new750(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
extern T0* GE_new751(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
extern T0* GE_new752(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_INTEGER_ROUTINES */
extern T0* GE_new753(T1 initialize);
/* New instance of type [detachable] STRING_8_SEARCHER */
extern T0* GE_new755(T1 initialize);
/* New instance of type detachable ARRAYED_LIST [[attached] INTEGER_32] */
extern T0* GE_new756(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_ERROR */
extern T0* GE_new757(T1 initialize);
/* New instance of type [detachable] ET_INTERNAL_ERROR */
extern T0* GE_new758(T1 initialize);
/* New instance of type [detachable] ET_VALIDITY_ERROR */
extern T0* GE_new759(T1 initialize);
/* New instance of type detachable ET_ECF_NOTE_ELEMENT */
extern T0* GE_new760(T1 initialize);
/* New instance of type detachable ET_ECF_CLUSTERS */
extern T0* GE_new761(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_LIBRARIES */
extern T0* GE_new762(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
extern T0* GE_new763(T1 initialize);
/* New instance of type detachable ET_ECF_FILE_RULES */
extern T0* GE_new764(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_CFLAGS */
extern T0* GE_new765(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_INCLUDES */
extern T0* GE_new766(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LIBRARIES */
extern T0* GE_new767(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
extern T0* GE_new768(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_MAKES */
extern T0* GE_new769(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_OBJECTS */
extern T0* GE_new770(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_RESOURCES */
extern T0* GE_new771(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
extern T0* GE_new772(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
extern T0* GE_new773(T1 initialize);
/* New instance of type detachable ET_ECF_VERSION */
extern T0* GE_new775(T1 initialize);
/* New instance of type detachable ET_ECF_CLUSTER */
extern T0* GE_new776(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_CFLAG */
extern T0* GE_new777(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_INCLUDE */
extern T0* GE_new778(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LIBRARY */
extern T0* GE_new779(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LINKER_FLAG */
extern T0* GE_new780(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_MAKE */
extern T0* GE_new781(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_OBJECT */
extern T0* GE_new782(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_RESOURCE */
extern T0* GE_new783(T1 initialize);
/* New instance of type detachable ET_ECF_FILE_RULE */
extern T0* GE_new784(T1 initialize);
/* New instance of type detachable ET_ECF_ACTION */
extern T0* GE_new785(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new786(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new787(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new788(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
extern T0* GE_new789(T1 initialize);
/* New instance of type detachable ET_ECF_ROOT_CLASS */
extern T0* GE_new790(T1 initialize);
/* New instance of type [detachable] ET_ECF_ROOT_ALL_CLASSES */
extern T0* GE_new791(T1 initialize);
/* New instance of type [detachable] ET_ECF_ORED_CONDITIONS */
extern T0* GE_new792(T1 initialize);
/* New instance of type detachable ET_ECF_ANDED_CONDITIONS */
extern T0* GE_new793(T1 initialize);
/* New instance of type detachable ET_ECF_VISIBLE_CLASS */
extern T0* GE_new794(T1 initialize);
/* New instance of type detachable ET_ECF_BUILD_CONDITION */
extern T0* GE_new796(T1 initialize);
/* New instance of type detachable ET_ECF_CONCURRENCY_CONDITION */
extern T0* GE_new797(T1 initialize);
/* New instance of type detachable ET_ECF_VOID_SAFETY_CONDITION */
extern T0* GE_new798(T1 initialize);
/* New instance of type detachable ET_ECF_CUSTOM_CONDITION */
extern T0* GE_new799(T1 initialize);
/* New instance of type detachable ET_ECF_DOTNET_CONDITION */
extern T0* GE_new800(T1 initialize);
/* New instance of type detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
extern T0* GE_new801(T1 initialize);
/* New instance of type detachable ET_ECF_PLATFORM_CONDITION */
extern T0* GE_new802(T1 initialize);
/* New instance of type [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
extern T0* GE_new804(T1 initialize);
/* New instance of type [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
extern T0* GE_new805(T1 initialize);
/* New instance of type [detachable] UT_COUNTER */
extern T0* GE_new807(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
extern T0* GE_new808(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] UT_COUNTER] */
extern T0* GE_new809(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
extern T0* GE_new810(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_UNIVERSE] */
extern T0* GE_new811(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
extern T0* GE_new812(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
extern T0* GE_new813(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new814(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new816(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new817(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new818(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new819(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new820(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new821(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
extern T0* GE_new822(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
extern T0* GE_new823(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
extern T0* GE_new824(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new825(T1 initialize);
/* New instance of type [detachable] ET_LIBRARY */
extern T0* GE_new826(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_LIBRARY] */
extern T0* GE_new827(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
extern T0* GE_new828(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
extern T0* GE_new829(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLUSTER] */
extern T0* GE_new831(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
extern T0* GE_new832(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
extern T0* GE_new833(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new834(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
extern T0* GE_new835(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
extern T0* GE_new836(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
extern T0* GE_new837(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new838(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
extern T0* GE_new839(T1 initialize);
/* New instance of type [detachable] ET_ANCESTORS_STATUS_CHECKER */
extern T0* GE_new840(T1 initialize);
/* New instance of type [detachable] ET_FLATTENING_STATUS_CHECKER */
extern T0* GE_new841(T1 initialize);
/* New instance of type [detachable] ET_INTERFACE_STATUS_CHECKER */
extern T0* GE_new842(T1 initialize);
/* New instance of type [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
extern T0* GE_new843(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
extern T0* GE_new844(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
extern T0* GE_new846(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
extern T0* GE_new847(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
extern T0* GE_new848(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
extern T0* GE_new849(T1 initialize);
/* New instance of type [detachable] TUPLE [BOOLEAN] */
extern T0* GE_new850(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
extern T0* GE_new851(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
extern T0* GE_new854(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
extern T0* GE_new858(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
extern T0* GE_new859(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
extern T0* GE_new860(T1 initialize);
/* New instance of type [detachable] ET_NONE_GROUP */
extern T0* GE_new861(T1 initialize);
/* New instance of type detachable ET_PARENTHESIS_EXPRESSION */
extern T0* GE_new862(T1 initialize);
/* New instance of type [detachable] ET_REGULAR_MANIFEST_STRING */
extern T0* GE_new863(T1 initialize);
/* New instance of type [detachable] ET_INFIX_FREE_NAME */
extern T0* GE_new864(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new866(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER */
extern T0* GE_new867(T1 initialize);
/* New instance of type [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new869(T1 initialize);
/* New instance of type [detachable] UC_STRING_EQUALITY_TESTER */
extern T0* GE_new871(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new873(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new874(T1 initialize);
/* New instance of type [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
extern T0* GE_new875(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_INPUT_STREAM */
extern T0* GE_new876(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
extern T0* GE_new877(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
extern T0* GE_new878(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
extern T0* GE_new879(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new880(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
extern T0* GE_new881(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new882(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new884(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_QUERY] */
extern T0* GE_new885(T6 a1, T1 initialize);
/* New instance of type detachable ET_ALIAS_NAME */
extern T0* GE_new886(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
extern T0* GE_new887(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PROCEDURE] */
extern T0* GE_new888(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
extern T0* GE_new889(T1 initialize);
/* New instance of type [detachable] CONSOLE */
extern T0* GE_new896(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new897(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
extern T0* GE_new899(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
extern T0* GE_new900(T1 initialize);
/* New instance of type [detachable] INTEGER_OVERFLOW_CHECKER */
extern T0* GE_new901(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
extern T0* GE_new902(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
extern T0* GE_new903(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new904(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new905(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new906(T1 initialize);
/* New instance of type detachable C_DATE */
extern T0* GE_new907(T1 initialize);
/* New instance of type [detachable] YY_FILE_BUFFER */
extern T0* GE_new908(T1 initialize);
/* New instance of type [detachable] KL_STDIN_FILE */
extern T0* GE_new909(T1 initialize);
/* New instance of type [detachable] YY_BUFFER */
extern T0* GE_new910(T1 initialize);
/* New instance of type [detachable] YY_UNICODE_FILE_BUFFER */
extern T0* GE_new914(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
extern T0* GE_new915(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new916(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new917(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
extern T0* GE_new918(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
extern T0* GE_new919(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new920(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new921(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new922(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new923(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new924(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new925(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
extern T0* GE_new926(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
extern T0* GE_new927(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD] */
extern T0* GE_new928(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
extern T0* GE_new929(T1 initialize);
/* New instance of type detachable ET_AGENT_KEYWORD */
extern T0* GE_new930(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
extern T0* GE_new931(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
extern T0* GE_new932(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_KEYWORD */
extern T0* GE_new933(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
extern T0* GE_new934(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
extern T0* GE_new935(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SYMBOL] */
extern T0* GE_new936(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
extern T0* GE_new937(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_POSITION] */
extern T0* GE_new938(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
extern T0* GE_new939(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
extern T0* GE_new941(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
extern T0* GE_new942(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BREAK] */
extern T0* GE_new944(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
extern T0* GE_new945(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
extern T0* GE_new947(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
extern T0* GE_new948(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CURRENT] */
extern T0* GE_new949(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
extern T0* GE_new950(T1 initialize);
/* New instance of type detachable ET_FREE_OPERATOR */
extern T0* GE_new951(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
extern T0* GE_new952(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
extern T0* GE_new953(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IDENTIFIER] */
extern T0* GE_new954(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
extern T0* GE_new955(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
extern T0* GE_new957(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
extern T0* GE_new958(T1 initialize);
/* New instance of type detachable ET_KEYWORD_OPERATOR */
extern T0* GE_new959(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
extern T0* GE_new960(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
extern T0* GE_new961(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
extern T0* GE_new962(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
extern T0* GE_new963(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
extern T0* GE_new965(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
extern T0* GE_new966(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RESULT] */
extern T0* GE_new967(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
extern T0* GE_new968(T1 initialize);
/* New instance of type detachable ET_RETRY_INSTRUCTION */
extern T0* GE_new969(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
extern T0* GE_new970(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
extern T0* GE_new971(T1 initialize);
/* New instance of type detachable ET_SYMBOL_OPERATOR */
extern T0* GE_new972(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
extern T0* GE_new973(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
extern T0* GE_new974(T1 initialize);
/* New instance of type detachable ET_VOID */
extern T0* GE_new975(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_VOID] */
extern T0* GE_new976(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
extern T0* GE_new977(T1 initialize);
/* New instance of type detachable ET_SEMICOLON_SYMBOL */
extern T0* GE_new978(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
extern T0* GE_new979(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
extern T0* GE_new980(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
extern T0* GE_new981(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
extern T0* GE_new982(T1 initialize);
/* New instance of type detachable ET_QUESTION_MARK_SYMBOL */
extern T0* GE_new983(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
extern T0* GE_new984(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
extern T0* GE_new985(T1 initialize);
/* New instance of type detachable ET_ACROSS_EXPRESSION */
extern T0* GE_new986(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
extern T0* GE_new987(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
extern T0* GE_new988(T1 initialize);
/* New instance of type detachable ET_ACROSS_INSTRUCTION */
extern T0* GE_new989(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
extern T0* GE_new990(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new991(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new992(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new993(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new994(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new995(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new996(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
extern T0* GE_new997(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
extern T0* GE_new998(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new999(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1001(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
extern T0* GE_new1002(T1 initialize);
/* New instance of type detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
extern T0* GE_new1003(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
extern T0* GE_new1004(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
extern T0* GE_new1005(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
extern T0* GE_new1006(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
extern T0* GE_new1007(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
extern T0* GE_new1008(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
extern T0* GE_new1009(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ASSIGNER] */
extern T0* GE_new1011(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
extern T0* GE_new1012(T1 initialize);
/* New instance of type detachable ET_BRACKET_EXPRESSION */
extern T0* GE_new1013(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
extern T0* GE_new1014(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
extern T0* GE_new1015(T1 initialize);
/* New instance of type detachable ET_CALL_AGENT */
extern T0* GE_new1016(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CALL_AGENT] */
extern T0* GE_new1017(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
extern T0* GE_new1018(T1 initialize);
/* New instance of type detachable ET_CHECK_INSTRUCTION */
extern T0* GE_new1019(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
extern T0* GE_new1020(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
extern T0* GE_new1021(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE] */
extern T0* GE_new1023(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
extern T0* GE_new1024(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
extern T0* GE_new1026(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
extern T0* GE_new1027(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
extern T0* GE_new1029(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
extern T0* GE_new1030(T1 initialize);
/* New instance of type detachable ET_CHOICE_LIST */
extern T0* GE_new1031(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
extern T0* GE_new1032(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
extern T0* GE_new1033(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLASS] */
extern T0* GE_new1034(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
extern T0* GE_new1035(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
extern T0* GE_new1036(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
extern T0* GE_new1037(T1 initialize);
/* New instance of type detachable ET_CLIENTS */
extern T0* GE_new1038(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLIENTS] */
extern T0* GE_new1039(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
extern T0* GE_new1040(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_COMPOUND] */
extern T0* GE_new1041(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
extern T0* GE_new1042(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONDITIONAL] */
extern T0* GE_new1044(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
extern T0* GE_new1045(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTANT] */
extern T0* GE_new1046(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1047(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1049(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1050(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
extern T0* GE_new1051(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1052(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
extern T0* GE_new1053(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_CREATOR */
extern T0* GE_new1054(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
extern T0* GE_new1055(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
extern T0* GE_new1056(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_RENAME_LIST */
extern T0* GE_new1057(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
extern T0* GE_new1058(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1059(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1061(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
extern T0* GE_new1062(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
extern T0* GE_new1064(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1065(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1067(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
extern T0* GE_new1068(T1 initialize);
/* New instance of type detachable ET_CONVERT_FEATURE_LIST */
extern T0* GE_new1069(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
extern T0* GE_new1070(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
extern T0* GE_new1071(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
extern T0* GE_new1072(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
extern T0* GE_new1073(T1 initialize);
/* New instance of type detachable ET_CREATION_REGION */
extern T0* GE_new1074(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATION_REGION] */
extern T0* GE_new1075(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
extern T0* GE_new1076(T1 initialize);
/* New instance of type detachable ET_CREATOR */
extern T0* GE_new1077(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATOR] */
extern T0* GE_new1078(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
extern T0* GE_new1079(T1 initialize);
/* New instance of type detachable ET_CREATOR_LIST */
extern T0* GE_new1080(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
extern T0* GE_new1081(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
extern T0* GE_new1082(T1 initialize);
/* New instance of type detachable ET_DEBUG_INSTRUCTION */
extern T0* GE_new1083(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
extern T0* GE_new1084(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1085(T1 initialize);
/* New instance of type detachable ET_ELSEIF_EXPRESSION */
extern T0* GE_new1086(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1087(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
extern T0* GE_new1088(T1 initialize);
/* New instance of type detachable ET_ELSEIF_EXPRESSION_LIST */
extern T0* GE_new1089(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
extern T0* GE_new1090(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
extern T0* GE_new1091(T1 initialize);
/* New instance of type detachable ET_ELSEIF_PART */
extern T0* GE_new1092(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
extern T0* GE_new1093(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
extern T0* GE_new1094(T1 initialize);
/* New instance of type detachable ET_ELSEIF_PART_LIST */
extern T0* GE_new1095(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
extern T0* GE_new1096(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
extern T0* GE_new1097(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPORT] */
extern T0* GE_new1099(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
extern T0* GE_new1100(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
extern T0* GE_new1101(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
extern T0* GE_new1102(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPRESSION] */
extern T0* GE_new1103(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
extern T0* GE_new1104(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
extern T0* GE_new1105(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
extern T0* GE_new1106(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
extern T0* GE_new1107(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
extern T0* GE_new1108(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
extern T0* GE_new1110(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
extern T0* GE_new1111(T1 initialize);
/* New instance of type detachable ET_FEATURE_CLAUSE */
extern T0* GE_new1112(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
extern T0* GE_new1113(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
extern T0* GE_new1114(T1 initialize);
/* New instance of type detachable ET_FEATURE_CLAUSE_LIST */
extern T0* GE_new1115(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
extern T0* GE_new1116(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
extern T0* GE_new1117(T1 initialize);
/* New instance of type detachable ET_FEATURE_EXPORT */
extern T0* GE_new1118(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
extern T0* GE_new1119(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
extern T0* GE_new1120(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
extern T0* GE_new1121(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1122(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1124(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
extern T0* GE_new1125(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
extern T0* GE_new1126(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1127(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1129(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new1130(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new1131(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
extern T0* GE_new1132(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
extern T0* GE_new1133(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1134(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1136(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
extern T0* GE_new1137(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
extern T0* GE_new1138(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
extern T0* GE_new1139(T1 initialize);
/* New instance of type detachable ET_IF_EXPRESSION */
extern T0* GE_new1140(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
extern T0* GE_new1141(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
extern T0* GE_new1142(T1 initialize);
/* New instance of type detachable ET_IF_INSTRUCTION */
extern T0* GE_new1143(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
extern T0* GE_new1144(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
extern T0* GE_new1145(T1 initialize);
/* New instance of type detachable ET_INDEXING_LIST */
extern T0* GE_new1146(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
extern T0* GE_new1147(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
extern T0* GE_new1148(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
extern T0* GE_new1150(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
extern T0* GE_new1151(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
extern T0* GE_new1152(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1153(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1155(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
extern T0* GE_new1156(T1 initialize);
/* New instance of type detachable ET_INDEXING_TERM_LIST */
extern T0* GE_new1157(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
extern T0* GE_new1158(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
extern T0* GE_new1159(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
extern T0* GE_new1160(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
extern T0* GE_new1161(T1 initialize);
/* New instance of type detachable ET_INSPECT_INSTRUCTION */
extern T0* GE_new1162(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
extern T0* GE_new1163(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
extern T0* GE_new1164(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INSTRUCTION] */
extern T0* GE_new1166(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
extern T0* GE_new1167(T1 initialize);
/* New instance of type detachable ET_INVARIANTS */
extern T0* GE_new1168(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INVARIANTS] */
extern T0* GE_new1169(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
extern T0* GE_new1170(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
extern T0* GE_new1171(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
extern T0* GE_new1172(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
extern T0* GE_new1174(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
extern T0* GE_new1175(T1 initialize);
/* New instance of type detachable ET_LOCAL_VARIABLE */
extern T0* GE_new1176(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
extern T0* GE_new1177(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1178(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1180(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new1181(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new1182(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
extern T0* GE_new1183(T1 initialize);
/* New instance of type detachable ET_LOOP_INVARIANTS */
extern T0* GE_new1184(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
extern T0* GE_new1185(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
extern T0* GE_new1186(T1 initialize);
/* New instance of type detachable ET_MANIFEST_ARRAY */
extern T0* GE_new1187(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
extern T0* GE_new1188(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1189(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1191(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
extern T0* GE_new1192(T1 initialize);
/* New instance of type detachable ET_MANIFEST_STRING_LIST */
extern T0* GE_new1193(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
extern T0* GE_new1194(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
extern T0* GE_new1195(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
extern T0* GE_new1196(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
extern T0* GE_new1197(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_OBSOLETE] */
extern T0* GE_new1199(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
extern T0* GE_new1200(T1 initialize);
/* New instance of type detachable ET_PARENTHESIZED_EXPRESSION */
extern T0* GE_new1201(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
extern T0* GE_new1202(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
extern T0* GE_new1203(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT] */
extern T0* GE_new1204(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
extern T0* GE_new1205(T1 initialize);
/* New instance of type detachable ET_PARENT_CLAUSE_LIST */
extern T0* GE_new1206(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
extern T0* GE_new1207(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
extern T0* GE_new1208(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
extern T0* GE_new1209(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
extern T0* GE_new1210(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_LIST] */
extern T0* GE_new1211(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
extern T0* GE_new1212(T1 initialize);
/* New instance of type detachable ET_POSTCONDITIONS */
extern T0* GE_new1213(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
extern T0* GE_new1214(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
extern T0* GE_new1215(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
extern T0* GE_new1216(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
extern T0* GE_new1217(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PROCEDURE] */
extern T0* GE_new1218(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
extern T0* GE_new1219(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
extern T0* GE_new1221(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
extern T0* GE_new1222(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUERY] */
extern T0* GE_new1223(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
extern T0* GE_new1224(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
extern T0* GE_new1226(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
extern T0* GE_new1227(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RENAME_LIST] */
extern T0* GE_new1228(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
extern T0* GE_new1229(T1 initialize);
/* New instance of type detachable ET_STATIC_CALL_EXPRESSION */
extern T0* GE_new1230(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
extern T0* GE_new1231(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
extern T0* GE_new1232(T1 initialize);
/* New instance of type detachable ET_STRIP_EXPRESSION */
extern T0* GE_new1233(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
extern T0* GE_new1234(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
extern T0* GE_new1235(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE] */
extern T0* GE_new1236(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
extern T0* GE_new1237(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
extern T0* GE_new1239(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1240(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1242(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
extern T0* GE_new1243(T1 initialize);
/* New instance of type detachable ET_TYPE_CONSTRAINT_LIST */
extern T0* GE_new1244(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
extern T0* GE_new1245(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
extern T0* GE_new1246(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
extern T0* GE_new1248(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
extern T0* GE_new1249(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_LIST] */
extern T0* GE_new1251(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
extern T0* GE_new1252(T1 initialize);
/* New instance of type detachable ET_VARIANT */
extern T0* GE_new1253(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_VARIANT] */
extern T0* GE_new1254(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
extern T0* GE_new1255(T1 initialize);
/* New instance of type detachable ET_WHEN_PART */
extern T0* GE_new1256(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_PART] */
extern T0* GE_new1257(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
extern T0* GE_new1258(T1 initialize);
/* New instance of type detachable ET_WHEN_PART_LIST */
extern T0* GE_new1259(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
extern T0* GE_new1260(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
extern T0* GE_new1261(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WRITABLE] */
extern T0* GE_new1262(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
extern T0* GE_new1263(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
extern T0* GE_new1264(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER1 */
extern T0* GE_new1265(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
extern T0* GE_new1266(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1267(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_NAME_TESTER */
extern T0* GE_new1268(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1270(T1 initialize);
/* New instance of type [detachable] ET_ALIAS_NAME_TESTER */
extern T0* GE_new1271(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1273(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1274(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
extern T0* GE_new1275(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
extern T0* GE_new1276(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
extern T0* GE_new1277(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
extern T0* GE_new1278(T1 initialize);
/* New instance of type [detachable] ET_ANCHORED_TYPE_CHECKER */
extern T0* GE_new1279(T1 initialize);
/* New instance of type [detachable] ET_SIGNATURE_CHECKER */
extern T0* GE_new1280(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER2 */
extern T0* GE_new1281(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
extern T0* GE_new1282(T1 initialize);
/* New instance of type [detachable] ET_BUILTIN_FEATURE_CHECKER */
extern T0* GE_new1283(T1 initialize);
/* New instance of type [detachable] ET_PRECURSOR_CHECKER */
extern T0* GE_new1284(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1285(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_CLASS] */
extern T0* GE_new1286(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
extern T0* GE_new1287(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
extern T0* GE_new1288(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
extern T0* GE_new1289(T1 initialize);
/* New instance of type [detachable] ET_CALL_NAME_TESTER */
extern T0* GE_new1290(T1 initialize);
/* New instance of type [detachable] ET_RENAME */
extern T0* GE_new1295(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER3 */
extern T0* GE_new1296(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
extern T0* GE_new1297(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_QUERY] */
extern T0* GE_new1298(T1 initialize);
/* New instance of type [detachable] ET_SUPPLIER_BUILDER */
extern T0* GE_new1299(T1 initialize);
/* New instance of type detachable ET_PARENT_FEATURE */
extern T0* GE_new1302(T1 initialize);
/* New instance of type [detachable] ET_INHERITED_FEATURE */
extern T0* GE_new1303(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLASS] */
extern T0* GE_new1304(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
extern T0* GE_new1305(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
extern T0* GE_new1306(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
extern T0* GE_new1307(T1 initialize);
/* New instance of type [detachable] UT_SYNTAX_ERROR */
extern T0* GE_new1308(T1 initialize);
/* New instance of type [detachable] UT_TOO_MANY_INCLUDES_ERROR */
extern T0* GE_new1309(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_TARGET_LIST */
extern T0* GE_new1311(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATIVE_STRING] */
extern T0* GE_new1312(T6 a1, T1 initialize);
/* New instance of type [detachable] ARRAY [detachable STRING_8] */
extern T0* GE_new1313(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_CHARACTER_ENTITY */
extern T0* GE_new1314(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_POSITION] */
extern T0* GE_new1316(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
extern T0* GE_new1317(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
extern T0* GE_new1318(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
extern T0* GE_new1319(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
extern T0* GE_new1322(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
extern T0* GE_new1323(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new1326(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new1327(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1328(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new1329(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new1330(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new1331(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] STRING_8] */
extern T0* GE_new1332(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
extern T0* GE_new1333(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
extern T0* GE_new1336(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
extern T0* GE_new1337(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
extern T0* GE_new1338(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
extern T0* GE_new1339(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
extern T0* GE_new1340(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_NAMESPACE] */
extern T0* GE_new1341(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
extern T0* GE_new1342(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
extern T0* GE_new1344(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
extern T0* GE_new1345(T1 initialize);
/* New instance of type [detachable] ET_BRACED_TYPE_LIST */
extern T0* GE_new1346(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
extern T0* GE_new1347(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1348(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1349(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1350(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new1351(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1353(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1354(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1355(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1356(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1358(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new1359(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1360(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1361(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1362(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1364(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1365(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1366(T1 initialize);
/* New instance of type [detachable] THREAD_ATTRIBUTES */
extern T0* GE_new1367(T1 initialize);
/* New instance of type [detachable] CELL [[attached] BOOLEAN] */
extern T0* GE_new1368(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1370(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1372(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1373(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
extern T0* GE_new1374(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
extern T0* GE_new1375(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new1376(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT] */
extern T0* GE_new1377(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new1378(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
extern T0* GE_new1382(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
extern T0* GE_new1383(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
extern T0* GE_new1385(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
extern T0* GE_new1386(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
extern T0* GE_new1393(T1 initialize);
/* New instance of type [detachable] KL_CHARACTER_BUFFER */
extern T0* GE_new1396(T1 initialize);
/* New instance of type [detachable] UT_TRISTATE */
extern T0* GE_new1398(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
extern T0* GE_new1399(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] UT_TRISTATE] */
extern T0* GE_new1400(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
extern T0* GE_new1401(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
extern T0* GE_new1402(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE] */
extern T0* GE_new1403(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
extern T0* GE_new1405(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
extern T0* GE_new1406(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new1407(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
extern T0* GE_new1408(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
extern T0* GE_new1409(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
extern T0* GE_new1410(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
extern T0* GE_new1411(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new1413(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
extern T0* GE_new1414(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new1415(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
extern T0* GE_new1416(T1 initialize);
/* New instance of type [detachable] XM_COMMENT */
extern T0* GE_new1419(T1 initialize);
/* New instance of type [detachable] XM_PROCESSING_INSTRUCTION */
extern T0* GE_new1420(T1 initialize);
/* New instance of type [detachable] XM_CHARACTER_DATA */
extern T0* GE_new1421(T1 initialize);
/* New instance of type [detachable] ET_LIKE_N */
extern T0* GE_new1424(T1 initialize);
/* New instance of type [detachable] KL_STRING_INPUT_STREAM */
extern T0* GE_new1427(T1 initialize);
/* New instance of type [detachable] KL_UNICODE_CHARACTER_BUFFER */
extern T0* GE_new1429(T1 initialize);
/* New instance of type detachable ET_OBJECT_TEST_LIST */
extern T0* GE_new1431(T1 initialize);
/* New instance of type detachable ET_C3_CHARACTER_CONSTANT */
extern T0* GE_new1432(T1 initialize);
/* New instance of type detachable ET_SPECIAL_MANIFEST_STRING */
extern T0* GE_new1433(T1 initialize);
/* New instance of type detachable ET_VERBATIM_STRING */
extern T0* GE_new1434(T1 initialize);
/* New instance of type detachable ET_UNDERSCORED_INTEGER_CONSTANT */
extern T0* GE_new1435(T1 initialize);
/* New instance of type detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
extern T0* GE_new1436(T1 initialize);
/* New instance of type detachable ET_OCTAL_INTEGER_CONSTANT */
extern T0* GE_new1437(T1 initialize);
/* New instance of type detachable ET_BINARY_INTEGER_CONSTANT */
extern T0* GE_new1438(T1 initialize);
/* New instance of type detachable ET_REGULAR_REAL_CONSTANT */
extern T0* GE_new1439(T1 initialize);
/* New instance of type detachable ET_UNDERSCORED_REAL_CONSTANT */
extern T0* GE_new1440(T1 initialize);
/* New instance of type detachable ET_TRUE_CONSTANT */
extern T0* GE_new1442(T1 initialize);
/* New instance of type detachable ET_FALSE_CONSTANT */
extern T0* GE_new1443(T1 initialize);
/* New instance of type detachable ET_C1_CHARACTER_CONSTANT */
extern T0* GE_new1444(T1 initialize);
/* New instance of type detachable ET_C2_CHARACTER_CONSTANT */
extern T0* GE_new1445(T1 initialize);
/* New instance of type detachable ET_TAGGED_INDEXING */
extern T0* GE_new1447(T1 initialize);
/* New instance of type detachable ET_INDEXING */
extern T0* GE_new1448(T1 initialize);
/* New instance of type detachable ET_CUSTOM_ATTRIBUTE */
extern T0* GE_new1449(T1 initialize);
/* New instance of type detachable ET_CONSTRAINED_FORMAL_PARAMETER */
extern T0* GE_new1450(T1 initialize);
/* New instance of type detachable ET_TYPE_RENAME_CONSTRAINT */
extern T0* GE_new1451(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_NAMED_TYPE */
extern T0* GE_new1452(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
extern T0* GE_new1453(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
extern T0* GE_new1454(T1 initialize);
/* New instance of type detachable ET_ALL_EXPORT */
extern T0* GE_new1455(T1 initialize);
/* New instance of type detachable ET_CONVERT_FUNCTION */
extern T0* GE_new1457(T1 initialize);
/* New instance of type detachable ET_CONVERT_PROCEDURE */
extern T0* GE_new1458(T1 initialize);
/* New instance of type detachable ET_DO_FUNCTION */
extern T0* GE_new1460(T1 initialize);
/* New instance of type detachable ET_ONCE_FUNCTION */
extern T0* GE_new1461(T1 initialize);
/* New instance of type detachable ET_DEFERRED_FUNCTION */
extern T0* GE_new1462(T1 initialize);
/* New instance of type detachable ET_ONCE_PROCEDURE */
extern T0* GE_new1464(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_PROCEDURE */
extern T0* GE_new1465(T1 initialize);
/* New instance of type detachable ET_PREFIX_NAME */
extern T0* GE_new1466(T1 initialize);
/* New instance of type detachable ET_PREFIX_FREE_NAME */
extern T0* GE_new1467(T1 initialize);
/* New instance of type detachable ET_INFIX_NAME */
extern T0* GE_new1468(T1 initialize);
/* New instance of type detachable ET_ALIASED_FEATURE_NAME */
extern T0* GE_new1469(T1 initialize);
/* New instance of type detachable ET_ALIAS_FREE_NAME */
extern T0* GE_new1470(T1 initialize);
/* New instance of type detachable ET_CLASS_ASSERTION */
extern T0* GE_new1473(T1 initialize);
/* New instance of type detachable ET_LABELED_ACTUAL_PARAMETER */
extern T0* GE_new1474(T1 initialize);
/* New instance of type detachable ET_LIKE_FEATURE */
extern T0* GE_new1476(T1 initialize);
/* New instance of type detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
extern T0* GE_new1477(T1 initialize);
/* New instance of type detachable ET_QUALIFIED_LIKE_TYPE */
extern T0* GE_new1478(T1 initialize);
/* New instance of type detachable ET_ASSIGNER_INSTRUCTION */
extern T0* GE_new1479(T1 initialize);
/* New instance of type detachable ET_ASSIGNMENT */
extern T0* GE_new1480(T1 initialize);
/* New instance of type detachable ET_ASSIGNMENT_ATTEMPT */
extern T0* GE_new1481(T1 initialize);
/* New instance of type detachable ET_LOOP_INSTRUCTION */
extern T0* GE_new1482(T1 initialize);
/* New instance of type detachable ET_BANG_INSTRUCTION */
extern T0* GE_new1484(T1 initialize);
/* New instance of type detachable ET_CREATE_INSTRUCTION */
extern T0* GE_new1485(T1 initialize);
/* New instance of type detachable ET_CHOICE_RANGE */
extern T0* GE_new1486(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_INSTRUCTION */
extern T0* GE_new1487(T1 initialize);
/* New instance of type detachable ET_STATIC_CALL_INSTRUCTION */
extern T0* GE_new1489(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_EXPRESSION */
extern T0* GE_new1490(T1 initialize);
/* New instance of type detachable ET_FEATURE_ADDRESS */
extern T0* GE_new1491(T1 initialize);
/* New instance of type detachable ET_CURRENT_ADDRESS */
extern T0* GE_new1492(T1 initialize);
/* New instance of type detachable ET_RESULT_ADDRESS */
extern T0* GE_new1493(T1 initialize);
/* New instance of type detachable ET_EXPRESSION_ADDRESS */
extern T0* GE_new1494(T1 initialize);
/* New instance of type detachable ET_INFIX_EXPRESSION */
extern T0* GE_new1495(T1 initialize);
/* New instance of type detachable ET_INFIX_AND_THEN_OPERATOR */
extern T0* GE_new1496(T1 initialize);
/* New instance of type detachable ET_INFIX_OR_ELSE_OPERATOR */
extern T0* GE_new1497(T1 initialize);
/* New instance of type detachable ET_MANIFEST_TYPE */
extern T0* GE_new1498(T1 initialize);
/* New instance of type detachable ET_PREFIX_EXPRESSION */
extern T0* GE_new1499(T1 initialize);
/* New instance of type detachable ET_OLD_EXPRESSION */
extern T0* GE_new1500(T1 initialize);
/* New instance of type detachable ET_OLD_OBJECT_TEST */
extern T0* GE_new1501(T1 initialize);
/* New instance of type detachable ET_ONCE_MANIFEST_STRING */
extern T0* GE_new1502(T1 initialize);
/* New instance of type detachable ET_DO_FUNCTION_INLINE_AGENT */
extern T0* GE_new1503(T1 initialize);
/* New instance of type detachable ET_ONCE_FUNCTION_INLINE_AGENT */
extern T0* GE_new1504(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
extern T0* GE_new1505(T1 initialize);
/* New instance of type detachable ET_DO_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1506(T1 initialize);
/* New instance of type detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1507(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1508(T1 initialize);
/* New instance of type detachable ET_AGENT_OPEN_TARGET */
extern T0* GE_new1509(T1 initialize);
/* New instance of type detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
extern T0* GE_new1510(T1 initialize);
/* New instance of type detachable ET_TAGGED_ASSERTION */
extern T0* GE_new1513(T1 initialize);
/* New instance of type [detachable] ET_FILE_POSITION */
extern T0* GE_new1515(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
extern T0* GE_new1517(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
extern T0* GE_new1518(T1 initialize);
/* New instance of type [detachable] ET_REDECLARED_FEATURE */
extern T0* GE_new1520(T1 initialize);
/* New instance of type [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
extern T0* GE_new1522(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1524(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1527(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1528(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1529(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1530(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1531(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1532(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1533(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1534(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1535(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1536(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1537(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
extern T0* GE_new1538(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
extern T0* GE_new1539(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
extern T0* GE_new1540(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1541(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new1542(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new1545(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new1546(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1547(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1548(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1549(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1550(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1551(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1552(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1553(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1554(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
extern T0* GE_new1556(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_UNIVERSE] */
extern T0* GE_new1557(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
extern T0* GE_new1559(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1560(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1562(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1563(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
extern T0* GE_new1564(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
extern T0* GE_new1565(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
extern T0* GE_new1566(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
extern T0* GE_new1567(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1568(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1569(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1570(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1571(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
extern T0* GE_new1573(T1 initialize);
/* New instance of type [detachable] UC_UTF16_ROUTINES */
extern T0* GE_new1574(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
extern T0* GE_new1575(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
extern T0* GE_new1576(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LIBRARY] */
extern T0* GE_new1577(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
extern T0* GE_new1578(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1579(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1580(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1581(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1582(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1583(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1584(T1 initialize);
/* New instance of type [detachable] YY_UNICODE_BUFFER */
extern T0* GE_new1585(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new1586(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new1587(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new1588(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new1589(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new1590(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new1591(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new1592(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST] */
extern T0* GE_new1593(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1594(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1595(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1596(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1597(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1598(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1600(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
extern T0* GE_new1602(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
extern T0* GE_new1603(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1606(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
extern T0* GE_new1607(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
extern T0* GE_new1608(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1609(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
extern T0* GE_new1610(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1613(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
extern T0* GE_new1614(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1615(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
extern T0* GE_new1617(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
extern T0* GE_new1618(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1619(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1620(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1621(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1622(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1624(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
extern T0* GE_new1625(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT] */
extern T0* GE_new1626(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_CALL_NAME] */
extern T0* GE_new1627(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
extern T0* GE_new1628(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1629(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1630(T1 initialize);
/* New instance of type [detachable] ET_REPLICATED_FEATURE */
extern T0* GE_new1632(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1633(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1634(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_SIGNATURE_TESTER */
extern T0* GE_new1635(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1636(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1638(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1640(T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1641(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1642(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1643(T1 initialize);
/* New instance of type detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1644(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
extern T0* GE_new1645(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1646(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
extern T0* GE_new1649(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME] */
extern T0* GE_new1651(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_RENAME] */
extern T0* GE_new1652(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CALL_NAME] */
extern T0* GE_new1653(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
extern T0* GE_new1656(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
extern T0* GE_new1657(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
extern T0* GE_new1660(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
extern T0* GE_new1661(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] YY_BUFFER] */
extern T0* GE_new1662(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
extern T0* GE_new1663(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
extern T0* GE_new1664(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
extern T0* GE_new1665(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1667(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1668(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1670(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1671(T1 initialize);
/* New instance of type detachable DS_LINKABLE [detachable STRING_8] */
extern T0* GE_new1672(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLUSTER] */
extern T0* GE_new1673(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
extern T0* GE_new1674(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
extern T0* GE_new1675(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
extern T0* GE_new1676(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
extern T0* GE_new1677(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
extern T0* GE_new1678(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1679(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1680(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
extern T0* GE_new1681(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new1682(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1684(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
extern T0* GE_new1685(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1686(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
extern T0* GE_new1687(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
extern T0* GE_new1690(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new1691(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
extern T0* GE_new1692(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
extern T0* GE_new1693(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new1698(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
extern T0* GE_new1703(T1 initialize);
/* New instance of type detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
extern T0* GE_new1704(T1 initialize);
/* New instance of type detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
extern T0* GE_new1706(T1 initialize);
/* New instance of type [detachable] ET_SEEDED_QUERY_COMPARATOR */
extern T0* GE_new1707(T1 initialize);
/* New instance of type [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
extern T0* GE_new1710(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1714(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_COMMA_ARGUMENT */
extern T0* GE_new1715(T1 initialize);
/* New instance of type [detachable] ET_LOCAL_COMMA_VARIABLE */
extern T0* GE_new1716(T1 initialize);
/* New instance of type [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
extern T0* GE_new1717(T1 initialize);
/* New instance of type [detachable] ET_KEYWORD_EXPRESSION */
extern T0* GE_new1719(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
extern T0* GE_new1724(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
extern T0* GE_new1725(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPORT] */
extern T0* GE_new1727(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
extern T0* GE_new1728(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1730(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1731(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1732(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1733(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1736(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
extern T0* GE_new1737(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
extern T0* GE_new1738(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1739(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
extern T0* GE_new1740(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1741(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
extern T0* GE_new1742(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1743(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
extern T0* GE_new1744(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1745(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
extern T0* GE_new1746(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1747(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
extern T0* GE_new1748(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1749(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
extern T0* GE_new1750(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new1755(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1758(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1759(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
extern T0* GE_new1760(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
extern T0* GE_new1762(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
extern T0* GE_new1763(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1764(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1766(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1768(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1769(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CREATOR] */
extern T0* GE_new1770(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
extern T0* GE_new1771(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1772(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1773(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
extern T0* GE_new1774(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
extern T0* GE_new1775(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
extern T0* GE_new1776(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
extern T0* GE_new1777(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
extern T0* GE_new1779(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
extern T0* GE_new1780(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1781(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1782(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1784(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1785(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
extern T0* GE_new1786(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
extern T0* GE_new1787(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1789(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1790(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_WHEN_PART] */
extern T0* GE_new1791(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
extern T0* GE_new1792(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
extern T0* GE_new1794(T1 initialize);
/* New instance of type detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1797(T1 initialize);
/* New instance of type [detachable] ET_STANDARD_ONCE_KEYS */
extern T0* GE_new1802(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1804(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1805(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1806(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1807(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1808(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1809(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1810(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1811(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1812(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1813(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1814(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1815(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1816(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1817(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1818(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1819(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1820(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1821(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1822(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1823(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1824(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1825(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1826(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1827(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1828(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1829(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1830(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1831(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1832(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1833(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1834(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1835(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1836(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1837(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1838(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1839(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1840(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1841(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1842(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1843(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1844(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1845(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] CHARACTER_8] */
extern T0* GE_new1847(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
extern T0* GE_new1848(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1849(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1850(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1851(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
extern T0* GE_new1852(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME] */
extern T0* GE_new1853(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME] */
extern T0* GE_new1854(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1857(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1858(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
extern T0* GE_new1859(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
extern T0* GE_new1861(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1862(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1863(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1864(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1865(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1866(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1867(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1868(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1869(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
extern T0* GE_new1870(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1873(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
extern T0* GE_new1874(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1875(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1876(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
extern T0* GE_new1877(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1878(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1879(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1880(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
extern T0* GE_new1882(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
extern T0* GE_new1883(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
extern T0* GE_new1884(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
extern T0* GE_new1885(T6 a1, T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1886(T1 initialize);
/* New instance of type [detachable] ET_CLUSTER_ERROR */
extern T0* GE_new1887(T1 initialize);
/* New instance of type [detachable] ET_SYNTAX_ERROR */
extern T0* GE_new1888(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY_ERROR */
extern T0* GE_new1889(T1 initialize);
/* New instance of type [detachable] AP_OPTION_COMPARATOR */
extern T0* GE_new1890(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
extern T0* GE_new1891(T1 initialize);
/* New instance of type [detachable] ST_SPLITTER */
extern T0* GE_new1898(T1 initialize);
/* New instance of type [detachable] KL_CHARACTER_ROUTINES */
extern T0* GE_new1899(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
extern T0* GE_new1900(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
extern T0* GE_new1901(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
extern T0* GE_new1902(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1905(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1907(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
extern T0* GE_new1909(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
extern T0* GE_new1910(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
extern T0* GE_new1911(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
extern T0* GE_new1913(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1914(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1916(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
extern T0* GE_new1924(T1 initialize);
/* New instance of type [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
extern T0* GE_new1925(T1 initialize);
/* New instance of type [detachable] ET_INFIX_CAST_EXPRESSION */
extern T0* GE_new1926(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
extern T0* GE_new1930(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
extern T0* GE_new1931(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
extern T0* GE_new1934(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
extern T0* GE_new1935(T1 initialize);
/* New instance of type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1936(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
extern T0* GE_new1939(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
extern T0* GE_new1940(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
extern T0* GE_new1943(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
extern T0* GE_new1944(T1 initialize);
/* New instance of type [detachable] ET_UNIVERSE_ERROR */
extern T0* GE_new1945(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new1946(T1 initialize);
/* New instance of type detachable ET_PARENTHESIS_INSTRUCTION */
extern T0* GE_new1947(T1 initialize);
/* New instance of type [detachable] ET_PARENTHESIS_SYMBOL */
extern T0* GE_new1967(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
extern T0* GE_new1973(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
extern T0* GE_new1974(T1 initialize);
/* New instance of type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new1981(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
extern T0* GE_new1986(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
extern T0* GE_new1987(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
extern T0* GE_new1988(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
extern T0* GE_new1989(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new1990(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
extern T0* GE_new1992(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_FROM_EXPRESSION */
extern T0* GE_new1993(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_TO_EXPRESSION */
extern T0* GE_new1994(T1 initialize);
/* New instance of type detachable DS_CELL [detachable ET_CLASS] */
extern T0* GE_new1997(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
extern T0* GE_new1998(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
extern T0* GE_new2000(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
extern T0* GE_new2002(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new2003(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
extern T0* GE_new2004(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
extern T0* GE_new2005(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
extern T0* GE_new2006(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_STATE] */
extern T0* GE_new2007(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
extern T0* GE_new2009(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
extern T0* GE_new2010(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
extern T0* GE_new2012(T1 initialize);
/* New instance of type [detachable] LX_DFA_WILDCARD */
extern T0* GE_new2014(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new2017(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] LX_WILDCARD] */
extern T0* GE_new2019(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
extern T0* GE_new2020(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
extern T0* GE_new2021(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
extern T0* GE_new2023(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
extern T0* GE_new2024(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_WILDCARD] */
extern T0* GE_new2025(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
extern T0* GE_new2028(T1 initialize);
/* New instance of type [detachable] CHARACTER_PROPERTY */
extern T0* GE_new2029(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2034(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2035(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2036(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2037(T1 initialize);
/* New instance of type [detachable] ET_AST_NULL_LEAF */
extern T0* GE_new2039(T1 initialize);
/* New instance of type [detachable] LX_WILDCARD_PARSER */
extern T0* GE_new2040(T1 initialize);
/* New instance of type [detachable] LX_DESCRIPTION */
extern T0* GE_new2041(T1 initialize);
/* New instance of type [detachable] LX_FULL_DFA */
extern T0* GE_new2042(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_GROUP] */
extern T0* GE_new2044(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
extern T0* GE_new2045(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new2047(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
extern T0* GE_new2048(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_16] */
extern T0* GE_new2051(T6 a1, T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] NATURAL_16] */
extern T0* GE_new2052(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] NATURAL_32] */
extern T0* GE_new2053(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
extern T0* GE_new2054(T1 initialize);
/* New instance of type [detachable] LX_START_CONDITIONS */
extern T0* GE_new2055(T1 initialize);
/* New instance of type [detachable] LX_ACTION_FACTORY */
extern T0* GE_new2056(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2057(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_CLASS */
extern T0* GE_new2058(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
extern T0* GE_new2059(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2060(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
extern T0* GE_new2061(T1 initialize);
/* New instance of type [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
extern T0* GE_new2062(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
extern T0* GE_new2063(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
extern T0* GE_new2064(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] STRING_32] */
extern T0* GE_new2065(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2066(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2067(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
extern T0* GE_new2068(T1 initialize);
/* New instance of type [detachable] LX_NFA */
extern T0* GE_new2069(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_NFA] */
extern T0* GE_new2070(T6 a1, T1 initialize);
/* New instance of type detachable LX_RULE */
extern T0* GE_new2071(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2072(T1 initialize);
/* New instance of type [detachable] LX_EQUIVALENCE_CLASSES */
extern T0* GE_new2073(T1 initialize);
/* New instance of type [detachable] LX_UNRECOGNIZED_RULE_ERROR */
extern T0* GE_new2074(T1 initialize);
/* New instance of type [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
extern T0* GE_new2075(T1 initialize);
/* New instance of type [detachable] LX_MISSING_QUOTE_ERROR */
extern T0* GE_new2076(T1 initialize);
/* New instance of type [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
extern T0* GE_new2077(T1 initialize);
/* New instance of type [detachable] LX_BAD_CHARACTER_ERROR */
extern T0* GE_new2078(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_META_ERROR */
extern T0* GE_new2079(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_REJECT_ERROR */
extern T0* GE_new2080(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
extern T0* GE_new2081(T1 initialize);
/* New instance of type [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
extern T0* GE_new2082(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] LX_RULE] */
extern T0* GE_new2083(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_RULE] */
extern T0* GE_new2084(T6 a1, T1 initialize);
/* New instance of type [detachable] LX_DFA_STATE */
extern T0* GE_new2085(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
extern T0* GE_new2086(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
extern T0* GE_new2087(T1 initialize);
/* New instance of type detachable LX_SYMBOL_PARTITIONS */
extern T0* GE_new2088(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
extern T0* GE_new2089(T1 initialize);
/* New instance of type [detachable] ARRAY [detachable LX_RULE] */
extern T0* GE_new2090(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable LX_RULE] */
extern T0* GE_new2091(T6 a1, T1 initialize);
/* New instance of type [detachable] LX_START_CONDITION */
extern T0* GE_new2092(T1 initialize);
/* New instance of type [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
extern T0* GE_new2093(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
extern T0* GE_new2094(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
extern T0* GE_new2095(T1 initialize);
/* New instance of type [detachable] LX_NFA_STATE */
extern T0* GE_new2096(T1 initialize);
/* New instance of type detachable ARRAY [detachable LX_SYMBOL_CLASS] */
extern T0* GE_new2097(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
extern T0* GE_new2098(T6 a1, T1 initialize);
/* New instance of type detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
extern T0* GE_new2099(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
extern T0* GE_new2100(T6 a1, T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2101(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2102(T1 initialize);
/* New instance of type [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
extern T0* GE_new2103(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
extern T0* GE_new2104(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
extern T0* GE_new2105(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
extern T0* GE_new2106(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_START_CONDITION] */
extern T0* GE_new2107(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
extern T0* GE_new2108(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
extern T0* GE_new2109(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
extern T0* GE_new2110(T6 a1, T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2111(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
extern T0* GE_new2113(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2114(T1 initialize);
/* New instance of type detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
extern T0* GE_new2115(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
extern T0* GE_new2117(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
extern T0* GE_new2119(T1 initialize);
/* New instance of type [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2120(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2123(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2124(T1 initialize);
/* New instance of type [detachable] LX_ACTION */
extern T0* GE_new2126(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new2127(T1 initialize);
/* New instance of type [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
extern T0* GE_new2128(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new2129(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
extern T0* GE_new2130(T1 initialize);
/* New instance of type [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
extern T0* GE_new2132(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_NFA_STATE] */
extern T0* GE_new2134(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
extern T0* GE_new2136(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
extern T0* GE_new2137(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_DFA_STATE] */
extern T0* GE_new2139(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
extern T0* GE_new2140(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
extern T0* GE_new2141(T1 initialize);
/* New instance of type [detachable] KL_ARRAY [[attached] LX_RULE] */
extern T0* GE_new2142(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
extern T0* GE_new2143(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
extern T0* GE_new2144(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
extern T0* GE_new2146(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
extern T0* GE_new2150(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
extern T0* GE_new2153(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2156(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2157(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2158(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2159(T1 initialize);
/* New instance of type [detachable] KL_ARRAY [[attached] STRING_8] */
extern T0* GE_new2160(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
extern T0* GE_new2162(T1 initialize);
/* [detachable] GEC.execute */
extern T0* T25c59(GE_context* ac);
/* [detachable] GEC.report_cannot_read_error */
extern void T25f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error */
extern void T51f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error_message */
extern void T51f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] MUTEX.unlock */
extern void T133f12(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_unlock */
extern void T133f15(GE_context* ac, T0* C, T14 a1);
/* [detachable] MUTEX.default_pointer */
extern T14 T133f3(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_error */
extern void T51f184(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error_message */
extern void T51f182p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] MUTEX.lock */
extern void T133f11(GE_context* ac, T0* C);
/* [detachable] MUTEX.current_thread_id */
extern T14 T133f4(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_lock */
extern void T133f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.message */
extern T0* T51f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_error */
extern void T35f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_error_message */
extern void T35f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_error */
extern void T35f184(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.report_error_message */
extern void T35f182p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_line */
extern void T136f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_new_line */
extern void T136f8(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.put_string */
extern void T136f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.put_string */
extern void T896f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.console_ps */
extern void T896f26(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_STDERR_FILE.console */
extern T0* T136f2(GE_context* ac, T0* C);
/* [detachable] STD_FILES.error */
extern T0* T352f4(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stderr */
extern T0* T896c23(GE_context* ac, T0* a1);
/* [detachable] CONSOLE.set_write_mode */
extern void T896f25(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_def */
extern T14 T896f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] CONSOLE.make_with_name */
extern void T896f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.make_empty */
extern T0* T17c55(GE_context* ac);
/* [detachable] STRING_8.make */
extern void T17f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.make */
extern T0* T17c45(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_filled */
extern T0* T15c12(GE_context* ac, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].fill_with */
extern void T15f15(GE_context* ac, T0* C, T2 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].extend */
extern void T15f19(GE_context* ac, T0* C, T2 a1);
/* INTEGER_32.min */
extern T6 T6f22(GE_context* ac, T6* C, T6 a1);
/* INTEGER_32.is_less_equal */
extern T1 T6f13(GE_context* ac, T6* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_empty */
extern void T15f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_empty */
extern T0* T15c16(GE_context* ac, T6 a1);
/* [detachable] CONSOLE.default_pointer */
extern T14 T896f9(GE_context* ac, T0* C);
/* [detachable] CONSOLE.set_name */
extern void T896f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] FILE_INFO.file_name_to_pointer */
extern T0* T132f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable C_STRING.set_string */
extern void T30f11(GE_context* ac, T0* C, T0* a1);
/* detachable C_STRING.set_substring */
extern void T30f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] MANAGED_POINTER.put_natural_8 */
extern void T131f15(GE_context* ac, T0* C, T8 a1, T6 a2);
/* POINTER.memory_copy */
extern void T14f16(GE_context* ac, T14* C, T14 a1, T6 a2);
/* POINTER.c_memcpy */
extern void T14f17(GE_context* ac, T14* C, T14 a1, T14 a2, T6 a3);
/* NATURAL_32.to_natural_8 */
extern T8 T10f16(GE_context* ac, T10* C);
/* detachable C_STRING.make_shared_from_pointer_and_count */
extern void T30f10(GE_context* ac, T0* C, T14 a1, T6 a2);
/* detachable C_STRING.make_shared_from_pointer_and_count */
extern T0* T30c10(GE_context* ac, T14 a1, T6 a2);
/* [detachable] MANAGED_POINTER.share_from_pointer */
extern T0* T131c14(GE_context* ac, T14 a1, T6 a2);
/* [detachable] MANAGED_POINTER.increment_counter */
extern void T131f12(GE_context* ac, T0* C);
/* UTF_CONVERTER.escaped_utf_32_substring_into_utf_8_0_pointer */
extern void T471f14(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3, T0* a4, T6 a5, T0* a6);
/* [detachable] CELL [[attached] INTEGER_32].put */
extern void T472f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] CELL [[attached] INTEGER_32].put */
extern T0* T472c2(GE_context* ac, T6 a1);
/* INTEGER_32.is_greater */
extern T1 T6f6(GE_context* ac, T6* C, T6 a1);
/* NATURAL_32.is_less_equal */
extern T1 T10f15(GE_context* ac, T10* C, T10 a1);
/* UTF_CONVERTER.to_natural_32 */
extern T10 T471f10(GE_context* ac, T471* C, T0* a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parsed_natural_32 */
extern T10 T1522f12(GE_context* ac, T0* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parse_string_with_type */
extern void T1522f19(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T1522f19ot1(T0* a1, T0** a2);
extern T1 T1522f19ot2(T0* a1, T0** a2);
/* NATURAL_32.is_valid_character_8_code */
extern T1 T10f17(GE_context* ac, T10* C);
/* INTEGER_32.to_natural_32 */
extern T10 T6f18(GE_context* ac, T6* C);
/* CHARACTER_32.is_character_8 */
extern T1 T3f3(GE_context* ac, T3* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parse_character */
extern void T1522f22(GE_context* ac, T0* C, T2 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.overflowed */
extern T1 T1522f16(GE_context* ac, T0* C);
/* [detachable] INTEGER_OVERFLOW_CHECKER.will_overflow */
extern T1 T901f7(GE_context* ac, T0* C, T11 a1, T11 a2, T6 a3, T6 a4);
/* NATURAL_64.is_greater */
extern T1 T11f11(GE_context* ac, T11* C, T11 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.overflow_checker */
extern T0* T1522f15(GE_context* ac, T0* C);
/* [detachable] INTEGER_OVERFLOW_CHECKER.make */
extern T0* T901c13(GE_context* ac);
/* NATURAL_64.to_natural_64 */
extern T11 T11f14(GE_context* ac, T11* C);
/* NATURAL_32.to_natural_64 */
extern T11 T10f30(GE_context* ac, T10* C);
/* NATURAL_16.to_natural_64 */
extern T11 T9f15(GE_context* ac, T9* C);
/* NATURAL_8.to_natural_64 */
extern T11 T8f17(GE_context* ac, T8* C);
/* INTEGER_64.to_natural_64 */
extern T11 T7f15(GE_context* ac, T7* C);
/* INTEGER_16.to_natural_64 */
extern T11 T5f3(GE_context* ac, T5* C);
/* INTEGER_8.to_natural_64 */
extern T11 T4f3(GE_context* ac, T4* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].extend */
extern void T748f13(GE_context* ac, T0* C, T11 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_empty */
extern void T748f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_empty */
extern T0* T748c11(GE_context* ac, T6 a1);
/* [detachable] STRING_8.has */
extern T1 T17f41(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_8.area_lower */
extern T6 T17f9(GE_context* ac, T0* C);
/* INTEGER_32.to_natural_64 */
extern T11 T6f27(GE_context* ac, T6* C);
/* CHARACTER_8.is_less_equal */
extern T1 T2f8(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.is_less */
extern T1 T2f15(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.natural_32_code */
extern T10 T2f10(GE_context* ac, T2* C);
/* CHARACTER_8.is_greater_equal */
extern T1 T2f7(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.is_hexa_digit */
extern T1 T2f30(GE_context* ac, T2* C);
/* NATURAL_8.is_greater */
extern T1 T8f2(GE_context* ac, T8* C, T8 a1);
/* CHARACTER_8.character_types */
extern T8 T2f11(GE_context* ac, T2* C, T6 a1);
/* CHARACTER_8.internal_character_types */
extern T0* T2f16(GE_context* ac, T2* C);
/* [detachable] SPECIAL [[attached] NATURAL_8].fill_with */
extern void T477f8(GE_context* ac, T0* C, T8 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_8].extend */
extern void T477f7(GE_context* ac, T0* C, T8 a1);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_empty */
extern void T477f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_empty */
extern T0* T477c6(GE_context* ac, T6 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.reset */
extern void T1522f18(GE_context* ac, T0* C, T6 a1);
/* UTF_CONVERTER.ctoi_convertor */
extern T0* T471f11(GE_context* ac, T471* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.set_trailing_separators_acceptable */
extern void T1522f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.set_leading_separators_acceptable */
extern void T1522f20(GE_context* ac, T0* C, T1 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.make */
extern T0* T1522c17(GE_context* ac);
/* [detachable] STRING_8.make_from_string */
extern T0* T17c70(GE_context* ac, T0* a1);
/* [detachable] SPECIAL [CHARACTER_8].copy_data */
extern void T15f14(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [CHARACTER_8].force */
extern void T15f18(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].move_data */
extern void T15f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].overlapping_move */
extern void T15f21(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].non_overlapping_move */
extern void T15f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* UTF_CONVERTER.is_hexa_decimal */
extern T1 T471f9(GE_context* ac, T471* C, T0* a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.is_integral_integer */
extern T1 T1522f14(GE_context* ac, T0* C);
/* INTEGER_32.is_greater_equal */
extern T1 T6f12(GE_context* ac, T6* C, T6 a1);
/* UTF_CONVERTER.utf_8_bytes_count */
extern T6 T471f6(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3);
/* INTEGER_32.do_nothing */
extern void T6f34(GE_context* ac, T6* C);
/* [detachable] FILE_INFO.multi_byte_to_utf_16 */
extern T6 T132f10(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] FILE_INFO.default_pointer */
extern T14 T132f9(GE_context* ac, T0* C);
/* detachable C_STRING.item */
extern T14 T30f4(GE_context* ac, T0* C);
/* detachable C_STRING.make */
extern T0* T30c9(GE_context* ac, T0* a1);
/* detachable C_STRING.make_empty */
extern void T30f7(GE_context* ac, T0* C, T6 a1);
/* detachable C_STRING.make_empty */
extern T0* T30c7(GE_context* ac, T6 a1);
/* UTF_CONVERTER.escaped_utf_32_substring_into_utf_16_0_pointer */
extern void T471f13(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3, T0* a4, T6 a5, T0* a6);
/* [detachable] MANAGED_POINTER.put_natural_16 */
extern void T131f16(GE_context* ac, T0* C, T9 a1, T6 a2);
/* NATURAL_32.to_natural_16 */
extern T9 T10f22(GE_context* ac, T10* C);
/* NATURAL_32.is_greater */
extern T1 T10f4(GE_context* ac, T10* C, T10 a1);
/* UTF_CONVERTER.utf_16_bytes_count */
extern T6 T471f7(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3);
/* [detachable] MANAGED_POINTER.make */
extern T0* T131c10(GE_context* ac, T6 a1);
/* [detachable] EXCEPTIONS.raise */
extern void T476f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.raise */
extern void T119f16(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.raise */
extern void T21f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.code */
extern T6 T119f8(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.default_pointer */
extern T14 T21f9(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.set_last_exception */
extern void T21f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable EXCEPTION].put */
extern void T26f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable EXCEPTION].put */
extern T0* T26c2(GE_context* ac, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.last_exception_cell */
extern T0* T21f2(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.in_rescue */
extern T1 T21f8(GE_context* ac, T0* C);
/* [detachable] DEVELOPER_EXCEPTION.is_ignored */
extern T1 T119f7(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.is_ignored */
extern T1 T21f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].has */
extern T1 T27f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].same_keys */
extern T1 T27f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].hash_code_of */
extern T6 T27f25(GE_context* ac, T0* C, T6 a1);
/* INTEGER_32.hash_code */
extern T6 T6f23(GE_context* ac, T6* C);
/* [detachable] ISE_EXCEPTION_MANAGER.ignored_exceptions */
extern T0* T21f3(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].make */
extern T0* T27c35(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_filled */
extern T0* T126c9(GE_context* ac, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].fill_with */
extern void T126f12(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].extend */
extern void T126f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_filled */
extern T0* T127c5(GE_context* ac, T1 a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].fill_with */
extern void T127f8(GE_context* ac, T0* C, T1 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] BOOLEAN].extend */
extern void T127f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_empty */
extern void T127f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_empty */
extern T0* T127c7(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_empty */
extern void T126f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_empty */
extern T0* T126c8(GE_context* ac, T6 a1);
/* [detachable] PRIMES.higher_prime */
extern T6 T125f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] PRIMES.is_prime */
extern T1 T125f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] PRIMES.default_create */
extern T0* T125c5(GE_context* ac);
/* [detachable] DEVELOPER_EXCEPTION.exception_manager */
extern T0* T119f6(GE_context* ac, T0* C);
/* [detachable] DEVELOPER_EXCEPTION.default_create */
extern T0* T119c10(GE_context* ac);
/* [detachable] EXCEPTIONS.default_create */
extern T0* T476c1(GE_context* ac);
/* [detachable] MANAGED_POINTER.default_pointer */
extern T14 T131f4(GE_context* ac, T0* C);
/* POINTER.memory_calloc */
extern T14 T14f2(GE_context* ac, T14* C, T6 a1, T6 a2);
/* POINTER.c_calloc */
extern T14 T14f5(GE_context* ac, T14* C, T6 a1, T6 a2);
/* INTEGER_32.max */
extern T6 T6f14(GE_context* ac, T6* C, T6 a1);
/* [detachable] MANAGED_POINTER.resize */
extern void T131f11(GE_context* ac, T0* C, T6 a1);
/* POINTER.memory_set */
extern void T14f12(GE_context* ac, T14* C, T6 a1, T6 a2);
/* POINTER.c_memset */
extern void T14f15(GE_context* ac, T14* C, T14 a1, T6 a2, T6 a3);
/* POINTER.memory_realloc */
extern T14 T14f3(GE_context* ac, T14* C, T6 a1);
/* POINTER.c_realloc */
extern T14 T14f7(GE_context* ac, T14* C, T14 a1, T6 a2);
/* [detachable] CONSOLE.buffered_file_info */
extern T0* T896f10(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.make */
extern T0* T132c19(GE_context* ac);
/* [detachable] FILE_INFO.make_filled_area */
extern void T132f21(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_filled */
extern T0* T477c5(GE_context* ac, T8 a1, T6 a2);
/* [detachable] FILE_INFO.stat_size */
extern T6 T132f7(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.io */
extern T0* T136f4(GE_context* ac, T0* C);
/* [detachable] STD_FILES.set_output_default */
extern void T352f8(GE_context* ac, T0* C);
/* [detachable] STD_FILES.output */
extern T0* T352f3(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stdout */
extern T0* T896c20(GE_context* ac, T0* a1);
/* [detachable] STD_FILES.default_create */
extern T0* T352c6(GE_context* ac);
/* [detachable] KL_STRING_ROUTINES.as_string */
extern T0* T84f7(GE_context* ac, T0* C, T0* a1);
extern T1 T84f7ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.string */
extern T0* T743f20(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.next_byte_index */
extern T6 T743f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_byte_count */
extern T6 T416f18(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.utf8 */
extern T0* T743f32(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.default_create */
extern T0* T416c55(GE_context* ac);
/* [detachable] KL_INTEGER_ROUTINES.to_character */
extern T2 T753f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.integer_ */
extern T0* T743f44(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.default_create */
extern T0* T753c5(GE_context* ac);
/* [detachable] KL_PLATFORM.maximum_character_code */
extern T6 T470f2(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.platform */
extern T0* T743f37(GE_context* ac, T0* C);
/* [detachable] KL_PLATFORM.default_create */
extern T0* T470c3(GE_context* ac);
/* [detachable] UC_UTF8_STRING.item_code_at_byte_index */
extern T6 T743f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_next_value */
extern T6 T416f21(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_first_value */
extern T6 T416f19(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_8.append_character */
extern void T17f49(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_8.reset_hash_codes */
extern void T17f51(GE_context* ac, T0* C);
/* [detachable] STRING_8.resize */
extern void T17f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].aliased_resized_area_with_default */
extern T0* T15f4(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].aliased_resized_area */
extern T0* T15f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.additional_space */
extern T6 T17f8(GE_context* ac, T0* C);
/* [detachable] STRING_8.capacity */
extern T6 T17f7(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.byte_item */
extern T2 T743f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.old_item */
extern T2 T743f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.set_count */
extern void T743f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.reset_byte_index_cache */
extern void T743f92(GE_context* ac, T0* C);
/* [detachable] STRING_8.string */
extern T0* T17f27(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_string */
extern T0* T743f19(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_utf8 */
extern T0* T743f43(GE_context* ac, T0* C);
/* [detachable] KL_ANY_ROUTINES.same_types */
extern T1 T427f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STRING_ROUTINES.any_ */
extern T0* T84f3(GE_context* ac, T0* C);
/* [detachable] KL_ANY_ROUTINES.default_create */
extern T0* T427c3(GE_context* ac);
/* [detachable] KL_STDERR_FILE.string_ */
extern T0* T136f1(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.default_create */
extern T0* T84c19(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.message */
extern T0* T35f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.make */
extern T0* T79c8(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].put */
extern void T324f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] STRING_8].make_filled */
extern T0* T324c7(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] STRING_8].make_filled_area */
extern void T324f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].make_filled */
extern T0* T325c7(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].fill_with */
extern void T325f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].extend */
extern void T325f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] STRING_8].make_empty */
extern void T325f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].make_empty */
extern T0* T325c8(GE_context* ac, T6 a1);
/* [detachable] GEC.process_system */
extern void T25f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SHELL_COMMAND.execute */
extern void T66f13(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.operating_system */
extern T0* T66f6(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.default_create */
extern T0* T76c6(GE_context* ac);
/* [detachable] KL_SHELL_COMMAND.system */
extern void T66f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SHELL_COMMAND.system_call */
extern T6 T66f9(GE_context* ac, T0* C, T14 a1);
/* [detachable] NATIVE_STRING.item */
extern T14 T370f3(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.make */
extern T0* T370c9(GE_context* ac, T0* a1);
/* [detachable] NATIVE_STRING.set_string */
extern void T370f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] NATIVE_STRING.set_substring */
extern void T370f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] NATIVE_STRING.unit_size */
extern T6 T370f5(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.upper_cell */
extern T0* T370f7(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.make_empty */
extern void T370f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SHELL_COMMAND.default_shell */
extern T0* T66f8(GE_context* ac, T0* C);
/* [detachable] STRING_32.make_empty */
extern T0* T18c29(GE_context* ac);
/* [detachable] STRING_32.make */
extern void T18f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.make */
extern T0* T18c20(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_filled */
extern T0* T16c8(GE_context* ac, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].fill_with */
extern void T16f11(GE_context* ac, T0* C, T3 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].extend */
extern void T16f15(GE_context* ac, T0* C, T3 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_empty */
extern void T16f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_empty */
extern T0* T16c12(GE_context* ac, T6 a1);
/* [detachable] KL_SHELL_COMMAND.item */
extern T0* T66f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] NATIVE_STRING.string */
extern T0* T370f6(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_8_0_pointer_to_escaped_string_32 */
extern T0* T471f2(GE_context* ac, T471* C, T0* a1);
/* UTF_CONVERTER.utf_8_0_pointer_into_escaped_string_32 */
extern void T471f18(GE_context* ac, T471* C, T0* a1, T0* a2);
/* UTF_CONVERTER.utf_8_0_subpointer_into_escaped_string_32 */
extern void T471f23(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3, T1 a4, T0* a5);
/* UTF_CONVERTER.escape_code_into */
extern void T471f24(GE_context* ac, T471* C, T0* a1, T9 a2);
/* NATURAL_16.to_hex_string */
extern T0* T9f10(GE_context* ac, T9* C);
/* [detachable] STRING_8.put */
extern void T17f68(GE_context* ac, T0* C, T2 a1, T6 a2);
/* INTEGER_32.to_hex_character */
extern T2 T6f29(GE_context* ac, T6* C);
/* CHARACTER_8.plus */
extern T2 T2f28(GE_context* ac, T2* C, T6 a1);
/* CHARACTER_8.minus */
extern T2 T2f27(GE_context* ac, T2* C, T6 a1);
/* NATURAL_16.to_integer_32 */
extern T6 T9f2(GE_context* ac, T9* C);
/* [detachable] STRING_8.make_filled */
extern T0* T17c57(GE_context* ac, T2 a1, T6 a2);
/* [detachable] STRING_8.fill_character */
extern void T17f62(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_32.append_string_general */
extern void T18f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.append */
extern void T18f21p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.reset_hash_codes */
extern void T18f27(GE_context* ac, T0* C);
/* [detachable] STRING_32.set_count */
extern void T18f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.append_code */
extern void T18f22(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_32.put_code */
extern void T18f28(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] STRING_32.resize */
extern void T18f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].aliased_resized_area_with_default */
extern T0* T16f4(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].aliased_resized_area */
extern T0* T16f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.capacity */
extern T6 T18f5(GE_context* ac, T0* C);
/* [detachable] STRING_32.append */
extern void T18f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_32].copy_data */
extern void T16f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [CHARACTER_32].force */
extern void T16f14(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].move_data */
extern void T16f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].overlapping_move */
extern void T16f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].non_overlapping_move */
extern void T16f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] STRING_32.additional_space */
extern T6 T18f6(GE_context* ac, T0* C);
/* NATURAL_8.to_hex_string */
extern T0* T8f11(GE_context* ac, T8* C);
/* NATURAL_8.to_integer_32 */
extern T6 T8f14(GE_context* ac, T8* C);
/* NATURAL_16.is_less_equal */
extern T1 T9f8(GE_context* ac, T9* C, T9 a1);
/* [detachable] STRING_32.append_character */
extern void T18f30(GE_context* ac, T0* C, T3 a1);
/* NATURAL_8.to_natural_16 */
extern T9 T8f9(GE_context* ac, T8* C);
/* [detachable] STRING_32.extend */
extern void T18f34(GE_context* ac, T0* C, T3 a1);
/* NATURAL_8.is_less_equal */
extern T1 T8f6(GE_context* ac, T8* C, T8 a1);
/* [detachable] MANAGED_POINTER.read_natural_8 */
extern T8 T131f5(GE_context* ac, T0* C, T6 a1);
/* TYPED_POINTER [[attached] NATURAL_8].memory_copy */
extern void T868f2(GE_context* ac, T868* C, T14 a1, T6 a2);
/* TYPED_POINTER [[attached] NATURAL_8].c_memcpy */
extern void T868f3(GE_context* ac, T868* C, T14 a1, T14 a2, T6 a3);
/* [detachable] STRING_32.grow */
extern void T18f33(GE_context* ac, T0* C, T6 a1);
/* UTF_CONVERTER.utf_16_0_pointer_to_escaped_string_32 */
extern T0* T471f1(GE_context* ac, T471* C, T0* a1);
/* UTF_CONVERTER.utf_16_0_pointer_into_escaped_string_32 */
extern void T471f17(GE_context* ac, T471* C, T0* a1, T0* a2);
/* UTF_CONVERTER.utf_16_0_subpointer_into_escaped_string_32 */
extern void T471f22(GE_context* ac, T471* C, T0* a1, T6 a2, T6 a3, T1 a4, T0* a5);
/* NATURAL_32.is_greater_equal */
extern T1 T10f8(GE_context* ac, T10* C, T10 a1);
/* NATURAL_16.to_natural_32 */
extern T10 T9f7(GE_context* ac, T9* C);
/* [detachable] MANAGED_POINTER.read_natural_16 */
extern T9 T131f7(GE_context* ac, T0* C, T6 a1);
/* TYPED_POINTER [[attached] NATURAL_16].memory_copy */
extern void T1418f2(GE_context* ac, T1418* C, T14 a1, T6 a2);
/* TYPED_POINTER [[attached] NATURAL_16].c_memcpy */
extern void T1418f3(GE_context* ac, T1418* C, T14 a1, T14 a2, T6 a3);
/* [detachable] NATIVE_STRING.make_from_pointer */
extern T0* T370c10(GE_context* ac, T14 a1);
/* [detachable] MANAGED_POINTER.make_from_pointer */
extern T0* T131c13(GE_context* ac, T14 a1, T6 a2);
/* POINTER.memory_alloc */
extern T14 T14f9(GE_context* ac, T14* C, T6 a1);
/* POINTER.c_malloc */
extern T14 T14f10(GE_context* ac, T14* C, T6 a1);
/* [detachable] NATIVE_STRING.pointer_length_in_bytes */
extern T6 T370f4(GE_context* ac, T0* C, T14 a1);
/* NATURAL_64.to_integer_32 */
extern T6 T11f6(GE_context* ac, T11* C);
/* NATURAL_64.is_less_equal */
extern T1 T11f5(GE_context* ac, T11* C, T11 a1);
/* [detachable] NATIVE_STRING.c_pointer_length_in_bytes */
extern T11 T370f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_SHELL_COMMAND.eif_getenv */
extern T14 T66f11(GE_context* ac, T0* C, T14 a1);
/* [detachable] UC_UTF8_STRING.old_is_empty */
extern T1 T743f17(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_empty */
extern T1 T17f12(GE_context* ac, T0* C);
/* INTEGER_32.out */
extern T0* T6f2(GE_context* ac, T6* C);
/* [detachable] STRING_8.append_integer */
extern void T17f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] GEC.c_compile_using_make */
extern T1 T25f53(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.same_string */
extern T1 T84f1(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T84f1ot1(T0* a1, T0** a2);
extern T1 T84f1ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.item_code */
extern T6 T743f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.byte_index */
extern T6 T743f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.item_code */
extern T6 T17f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.elks_same_string */
extern T1 T84f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UC_UTF8_STRING.same_string */
extern T1 T743f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.substring_index */
extern T6 T743f23(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T743f23ot1(T0* a1, T0** a2);
extern T1 T743f23ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.item */
extern T2 T743f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.character_item_at_byte_index */
extern T2 T743f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.item */
extern T2 T17f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.same_string */
extern T1 T17f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.same_characters */
extern T1 T17f20(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [CHARACTER_8].same_items */
extern T1 T15f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] UC_UTF8_STRING.area_lower */
extern T6 T743f36(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.same_unicode_string */
extern T1 T743f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.unicode_substring_index */
extern T6 T743f47(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T743f47ot1(T0* a1, T0** a2);
extern T1 T743f47ot2(T0* a1, T0** a2);
/* NATURAL_32.to_integer_32 */
extern T6 T10f6(GE_context* ac, T10* C);
/* [detachable] GEC.string_ */
extern T0* T25f55(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.parameter */
extern T0* T45f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].last */
extern T0* T146f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].is_empty */
extern T1 T146f8(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.make */
extern T0* T66c12(GE_context* ac, T0* a1);
/* [detachable] KL_SHELL_COMMAND.string_ */
extern T0* T66f7(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.absolute_pathname */
extern T0* T124f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname */
extern T0* T124f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STRING_ROUTINES.appended_string */
extern T0* T84f12(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T84f12ot1(T0* a1, T0** a2);
extern T1 T84f12ot2(T0* a1, T0** a2);
/* [detachable] STRING_8.append_string */
extern void T17f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append */
extern void T17f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.concat */
extern T0* T84f15(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T84f15ot1(T0* a1, T0** a2);
extern T1 T84f15ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.prefixed_string */
extern T0* T743f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.make */
extern void T743f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.make */
extern T0* T743c73(GE_context* ac, T6 a1);
/* [detachable] UC_UTF8_STRING.old_set_count */
extern void T743f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.reset_hash_codes */
extern void T743f95(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.byte_capacity */
extern T6 T743f55(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.capacity */
extern T6 T743f55p1(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.make */
extern void T743f73p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.substring_byte_count */
extern T6 T416f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
extern T1 T416f3ot1(T0* a1, T0** a2);
extern T1 T416f3ot2(T0* a1, T0** a2);
extern T1 T416f3ot3(T0* a1, T0** a2);
/* [detachable] UC_UTF8_ROUTINES.character_32_byte_count */
extern T6 T416f8(GE_context* ac, T0* C, T3 a1);
/* [detachable] UC_UTF8_ROUTINES.natural_32_code_byte_count */
extern T6 T416f10(GE_context* ac, T0* C, T10 a1);
/* [detachable] UC_UTF8_STRING.shifted_byte_index */
extern T6 T743f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_ROUTINES.dummy_uc_string */
extern T0* T416f7(GE_context* ac, T0* C);
/* [detachable] UC_STRING.make_empty */
extern T0* T426c9(GE_context* ac);
/* [detachable] UC_STRING.make */
extern void T426f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.old_set_count */
extern void T426f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.reset_hash_codes */
extern void T426f14(GE_context* ac, T0* C);
/* [detachable] UC_STRING.set_count */
extern void T426f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.byte_capacity */
extern T6 T426f8(GE_context* ac, T0* C);
/* [detachable] UC_STRING.capacity */
extern T6 T426f8p1(GE_context* ac, T0* C);
/* [detachable] UC_STRING.make */
extern void T426f10p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.reset_byte_index_cache */
extern void T426f11(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.character_8_byte_count */
extern T6 T416f6(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.any_ */
extern T0* T416f4(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.append_string */
extern void T743f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.append */
extern void T743f86(GE_context* ac, T0* C, T0* a1);
extern T1 T743f86ot1(T0* a1, T0** a2);
extern T1 T743f86ot2(T0* a1);
/* [detachable] UC_UTF8_STRING.dummy_uc_string */
extern T0* T743f59(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.gobo_append_substring */
extern void T743f70(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.put_substring_at_byte_index */
extern void T743f90(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4, T6 a5);
extern T1 T743f90ot1(T0* a1, T0** a2);
extern T1 T743f90ot2(T0* a1, T0** a2);
extern T1 T743f90ot3(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.put_code_at_byte_index */
extern void T743f101(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.code_byte_count */
extern T6 T416f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.put_character_at_byte_index */
extern void T743f89(GE_context* ac, T0* C, T2 a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.character_byte_count */
extern T6 T416f36(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.put_byte */
extern void T743f102(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.old_put */
extern void T743f88(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.resize_byte_storage */
extern void T743f87(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.resize */
extern void T743f103(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.cloned_string */
extern T0* T743f40(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.twin */
extern T0* T743f14(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.copy */
extern void T743f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.copy */
extern void T743f80p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_8].resized_area */
extern T0* T15f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.append */
extern void T743f86p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.additional_space */
extern T6 T743f60(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.any_ */
extern T0* T743f39(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.append_character */
extern void T743f69(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_STRING_ROUTINES.cloned_string */
extern T0* T84f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.twin */
extern T0* T17f11(GE_context* ac, T0* C);
/* [detachable] STRING_8.copy */
extern void T17f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.string_ */
extern T0* T124f21(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.cwd */
extern T0* T124f13(GE_context* ac, T0* C);
/* [detachable] PATH.utf_8_name */
extern T0* T417f6(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_32_string_to_utf_8_string_8 */
extern T0* T471f3(GE_context* ac, T471* C, T0* a1);
/* UTF_CONVERTER.utf_32_string_into_utf_8_string_8 */
extern void T471f19(GE_context* ac, T471* C, T0* a1, T0* a2);
/* [detachable] STRING_8.extend */
extern void T17f78(GE_context* ac, T0* C, T2 a1);
/* [detachable] IMMUTABLE_STRING_32.code */
extern T10 T20f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.grow */
extern void T17f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.name */
extern T0* T417f4(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_8_string_8_to_escaped_string_32 */
extern T0* T471f5(GE_context* ac, T471* C, T0* a1);
/* UTF_CONVERTER.utf_8_string_8_into_escaped_string_32 */
extern void T471f21(GE_context* ac, T471* C, T0* a1, T0* a2);
/* [detachable] IMMUTABLE_STRING_32.make_from_string */
extern T0* T20c12(GE_context* ac, T0* a1);
/* [detachable] STRING_32.area_lower */
extern T6 T18f14(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_16le_string_8_to_escaped_string_32 */
extern T0* T471f4(GE_context* ac, T471* C, T0* a1);
/* UTF_CONVERTER.utf_16le_string_8_into_escaped_string_32 */
extern void T471f20(GE_context* ac, T471* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.current_working_path */
extern T0* T122f3(GE_context* ac, T0* C);
/* [detachable] PATH.make_from_pointer */
extern T0* T417c17(GE_context* ac, T14 a1);
/* [detachable] PATH.reset_internal_data */
extern void T417f20(GE_context* ac, T0* C);
/* [detachable] PATH.normalize */
extern void T417f19(GE_context* ac, T0* C);
/* [detachable] STRING_8.remove_tail */
extern void T17f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.directory_separator */
extern T2 T417f13(GE_context* ac, T0* C);
/* [detachable] STRING_8.remove_head */
extern void T17f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.keep_tail */
extern void T17f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.root_end_position */
extern T6 T417f12(GE_context* ac, T0* C);
/* [detachable] PATH.next_directory_separator */
extern T6 T417f15(GE_context* ac, T0* C, T6 a1);
/* CHARACTER_8.as_lower */
extern T2 T2f25(GE_context* ac, T2* C);
/* CHARACTER_8.is_upper */
extern T1 T2f13(GE_context* ac, T2* C);
/* [detachable] STRING_8.keep_head */
extern void T17f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.is_character */
extern T1 T417f8(GE_context* ac, T0* C, T0* a1, T6 a2, T2 a3);
/* detachable C_STRING.substring */
extern T0* T30f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable C_STRING.read_substring_into */
extern void T30f12(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_8.put_code */
extern void T17f54(GE_context* ac, T0* C, T10 a1, T6 a2);
/* NATURAL_8.to_natural_32 */
extern T10 T8f5(GE_context* ac, T8* C);
/* [detachable] STRING_8.set_count */
extern void T17f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.unit_size */
extern T6 T417f5(GE_context* ac, T0* C);
/* [detachable] PATH.pointer_length_in_bytes */
extern T6 T417f7(GE_context* ac, T0* C, T14 a1);
/* [detachable] PATH.c_pointer_length_in_bytes */
extern T11 T417f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] PATH.make_from_string */
extern T0* T417c16(GE_context* ac, T0* a1);
/* [detachable] PATH.internal_append_into */
extern void T417f18(GE_context* ac, T0* C, T0* a1, T0* a2, T2 a3);
/* UTF_CONVERTER.escaped_utf_32_string_into_utf_8_string_8 */
extern void T471f16(GE_context* ac, T471* C, T0* a1, T0* a2);
/* UTF_CONVERTER.escaped_utf_32_string_into_utf_16le_string_8 */
extern void T471f15(GE_context* ac, T471* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.eif_dir_current */
extern T6 T122f5(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.execution_environment */
extern T0* T124f23(GE_context* ac, T0* C);
/* [detachable] EXECUTION_ENVIRONMENT.default_create */
extern T0* T122c9(GE_context* ac);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_absolute_pathname */
extern T1 T124f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.absolute_pathname */
extern T0* T123f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.current_drive */
extern T0* T123f15(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_directory_separator */
extern T1 T123f18(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.secondary_directory_separator */
extern T2 T123f26(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.string_ */
extern T0* T123f14(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname */
extern T0* T123f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.cwd */
extern T0* T123f13(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.execution_environment */
extern T0* T123f25(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_relative_pathname */
extern T1 T123f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_absolute_pathname */
extern T1 T123f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.file_system */
extern T0* T25f52(GE_context* ac, T0* C);
/* [detachable] GEC.unix_file_system */
extern T0* T25f58(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.make */
extern T0* T124c31(GE_context* ac);
/* [detachable] KL_OPERATING_SYSTEM.is_unix */
extern T1 T76f2(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.current_working_directory */
extern T0* T76f4(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.execution_environment */
extern T0* T76f5(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_equal */
extern T1 T17f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_OPERATING_SYSTEM.variable_value */
extern T0* T76f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.to_string_8 */
extern T0* T18f11(GE_context* ac, T0* C);
/* [detachable] STRING_32.as_string_8 */
extern T0* T18f16(GE_context* ac, T0* C);
/* [detachable] STRING_8.valid_code */
extern T1 T17f37(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_32.is_valid_as_string_8 */
extern T1 T18f10(GE_context* ac, T0* C);
/* [detachable] EXECUTION_ENVIRONMENT.item */
extern T0* T122f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXECUTION_ENVIRONMENT.eif_getenv */
extern T14 T122f4(GE_context* ac, T0* C, T14 a1);
/* [detachable] GEC.windows_file_system */
extern T0* T25f57(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.make */
extern T0* T123c34(GE_context* ac);
/* [detachable] GEC.c_compile_using_script */
extern T1 T25f51(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.plus */
extern T0* T743f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.plus */
extern T0* T17f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append_string_general */
extern void T17f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append */
extern void T17f48p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append_code */
extern void T17f52(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_8.new_string */
extern T0* T17f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_OPERATING_SYSTEM.is_windows */
extern T1 T76f1(GE_context* ac, T0* C);
/* [detachable] GEC.operating_system */
extern T0* T25f50(GE_context* ac, T0* C);
/* [detachable] GEC.no_c_compile */
extern T1 T25f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_feature_count */
extern T6 T61f72(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T75f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.alive_dynamic_type_count */
extern T6 T61f73(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.put_new_line */
extern void T137f8(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.put_integer */
extern void T137f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STDOUT_FILE.put_integer_64 */
extern void T137f11(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_STDOUT_FILE.put_character */
extern void T137f10(GE_context* ac, T0* C, T2 a1);
/* [detachable] CONSOLE.put_character */
extern void T896f22(GE_context* ac, T0* C, T2 a1);
/* [detachable] CONSOLE.console_pc */
extern void T896f27(GE_context* ac, T0* C, T14 a1, T2 a2);
/* [detachable] KL_STDOUT_FILE.console */
extern T0* T137f2(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.io */
extern T0* T137f4(GE_context* ac, T0* C);
/* INTEGER_32.to_integer_64 */
extern T7 T6f19(GE_context* ac, T6* C);
/* [detachable] KL_STDOUT_FILE.put_string */
extern void T137f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDOUT_FILE.string_ */
extern T0* T137f1(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate */
extern void T65f527(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].wipe_out */
extern void T212f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_wipe_out */
extern void T212f69(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_wipe_out */
extern void T212f68(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_wipe_out */
extern void T212f67(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] STRING_8].keep_head */
extern void T325f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].set_count */
extern void T325f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_wipe_out */
extern void T212f66(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].unset_found_item */
extern void T212f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].move_all_cursors_after */
extern void T212f65(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_next_cursor */
extern void T296f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_after */
extern void T296f17(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_compilation_script */
extern void T65f531(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.generate_compilation_makefile */
extern void T65f567(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.report_cannot_write_error */
extern void T65f533(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.error_handler */
extern T0* T65f130(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.make */
extern T0* T297c8(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.set_fatal_error */
extern void T65f532(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.close */
extern void T293f34(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_close */
extern void T293f39(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.default_pointer */
extern T14 T293f18(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_close */
extern void T293f47(GE_context* ac, T0* C, T14 a1);
/* [detachable] DS_STRING_HASH_TABLE.expanded_string */
extern T0* T295f27(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_STRING_HASH_TABLE.value */
extern T0* T295f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_item */
extern T0* T295f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.position_of_key */
extern T6 T295f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.clashes_item */
extern T6 T295f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_item */
extern T0* T295f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.hash_position */
extern T6 T295f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.hash_code */
extern T6 T743f56(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.hash_code */
extern T6 T743f56p1(GE_context* ac, T0* C);
/* [detachable] STRING_8.hash_code */
extern T6 T17f25(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.slots_item */
extern T6 T295f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.append_substring_to_string */
extern void T84f20(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
extern T1 T84f20ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.append_code */
extern void T743f75(GE_context* ac, T0* C, T10 a1);
/* [detachable] UC_UTF8_STRING.append_item_code */
extern void T743f94(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.appended_substring */
extern T0* T84f11(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
extern T1 T84f11ot1(T0* a1, T0** a2);
extern T1 T84f11ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.new_empty_string */
extern T0* T743f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.string_ */
extern T0* T295f22(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.is_open_write */
extern T1 T293f23(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_is_open_write */
extern T1 T293f19(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_write */
extern void T293f29(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.is_closed */
extern T1 T293f17(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_is_closed */
extern T1 T293f22(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_open_write */
extern void T293f41(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_write */
extern void T293f41p1(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_open */
extern T14 T293f21(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.internal_name_pointer */
extern T0* T293f20(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.string_name */
extern T0* T293f14(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_string_8 */
extern T0* T743f16(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.put_code */
extern void T743f82(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.put_item_code */
extern void T743f100(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.move_bytes_right */
extern void T743f98(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.move_bytes_left */
extern void T743f97(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.valid_code */
extern T1 T743f42(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_8.as_string_8 */
extern T0* T17f24(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.make */
extern T0* T293c28(GE_context* ac, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_make */
extern void T293f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.make_with_name */
extern void T293f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.set_name */
extern void T293f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.buffered_file_info */
extern T0* T293f25(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.string_ */
extern T0* T293f12(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.operating_system */
extern T0* T65f129(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_compilation_shell_script */
extern void T65f566(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] KL_TEXT_OUTPUT_FILE.change_mode */
extern void T293f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_change_mode */
extern void T293f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_chmod */
extern void T293f51(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_exists */
extern T1 T293f16(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_exists */
extern T1 T293f26(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.file_exists */
extern T1 T124f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.exists */
extern T1 T32f32(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_plain */
extern T1 T32f34(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.is_plain */
extern T1 T132f13(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.file_info */
extern T6 T132f14(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.buffered_file_info */
extern T0* T32f29(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.set_buffer */
extern void T32f73(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.fast_update */
extern void T132f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] FILE_INFO.eif_file_stat */
extern T6 T132f12(GE_context* ac, T0* C, T14 a1, T14 a2, T1 a3);
/* [detachable] KL_TEXT_INPUT_FILE.internal_name_pointer */
extern T0* T32f23(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_exists */
extern T1 T32f15(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_exists */
extern T1 T32f30(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_closed */
extern T1 T32f27(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.string_name */
extern T0* T32f13(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.reset */
extern void T32f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.tmp_file */
extern T0* T124f22(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.file_exists */
extern T1 T123f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.tmp_file */
extern T0* T123f24(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.remove_tail */
extern void T743f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.keep_head */
extern void T743f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.wipe_out */
extern void T743f91(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.has_extension */
extern T1 T124f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.has_extension */
extern T1 T123f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.file_system */
extern T0* T65f116(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.unix_file_system */
extern T0* T65f455(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.windows_file_system */
extern T0* T65f454(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].item */
extern T0* T296f5(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_item */
extern T0* T295f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_item */
extern T0* T212f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_item */
extern T0* T212f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].forth */
extern void T296f16(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_forth */
extern void T295f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.add_traversing_cursor */
extern void T295f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_position */
extern void T296f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.remove_traversing_cursor */
extern void T295f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_forth */
extern void T212f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].add_traversing_cursor */
extern void T212f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove_traversing_cursor */
extern void T212f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_item */
extern T6 T212f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].after */
extern T1 T296f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].start */
extern void T296f15(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_start */
extern void T295f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.cursor_off */
extern T1 T295f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].off */
extern T1 T296f11(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.is_empty */
extern T1 T295f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_start */
extern void T212f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_off */
extern T1 T212f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].is_empty */
extern T1 T212f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].back */
extern void T296f14(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_back */
extern void T295f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_before */
extern void T296f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_back */
extern void T212f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].key */
extern T0* T296f9(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_key */
extern T0* T295f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_key */
extern T0* T212f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_item */
extern T0* T212f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].is_last */
extern T1 T296f7(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_is_last */
extern T1 T295f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_is_last */
extern T1 T212f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].before */
extern T1 T296f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].finish */
extern void T296f13(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_finish */
extern void T295f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_finish */
extern void T212f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_cursor */
extern T0* T212f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].make */
extern T0* T296c12(GE_context* ac, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.item */
extern T0* T295f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.search_position */
extern void T295f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.add_external_c_files */
extern void T65f565(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] KL_DIRECTORY.close */
extern void T327f34(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_close */
extern void T327f35(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.default_pointer */
extern T14 T327f17(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dir_close */
extern void T327f39(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_STRING_ROUTINES.replaced_all_substrings */
extern T0* T84f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] KL_STRING_ROUTINES.substring_index */
extern T6 T84f14(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] KL_STRING_ROUTINES.platform */
extern T0* T84f16(GE_context* ac, T0* C);
/* [detachable] STRING_8.substring_index */
extern T6 T17f29(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] STRING_8_SEARCHER.substring_index */
extern T6 T755f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] STRING_8_SEARCHER.substring_index_with_deltas */
extern T6 T755f2(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] STRING_8_SEARCHER.internal_initialize_deltas */
extern void T755f8(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] STRING_8.string_searcher */
extern T0* T17f39(GE_context* ac, T0* C);
/* [detachable] STRING_8_SEARCHER.make */
extern T0* T755c6(GE_context* ac);
/* [detachable] ET_C_GENERATOR.multithreaded_mode */
extern T1 T65f265(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_last */
extern void T212f51(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_put */
extern void T212f64(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].force */
extern void T373f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_put */
extern void T212f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_put */
extern void T212f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_item */
extern T6 T212f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].hash_position */
extern T6 T212f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8].hash_code */
extern T6 T211f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].resize */
extern void T212f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_resize */
extern void T212f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].aliased_resized_area_with_default */
extern T0* T375f1(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].aliased_resized_area_with_default */
extern T0* T126f4(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].aliased_resized_area */
extern T0* T126f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].special_integer_ */
extern T0* T212f41(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].default_create */
extern T0* T375c6(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_resize */
extern void T212f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].aliased_resized_area */
extern T0* T373f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].aliased_resized_area */
extern T0* T325f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_resize */
extern void T212f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_resize */
extern void T212f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_modulus */
extern T6 T212f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_capacity */
extern T6 T212f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_put */
extern void T212f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].search_position */
extern void T212f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.extension */
extern T0* T124f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.extension */
extern T0* T123f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.ends_with */
extern T1 T743f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.area_upper */
extern T6 T17f21(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.area_upper */
extern T6 T743f46(GE_context* ac, T0* C);
/* [detachable] STRING_8.ends_with */
extern T1 T17f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.read_entry */
extern void T327f33(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.last_entry_8 */
extern T0* T327f15(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.pointer_to_file_name_8 */
extern T0* T132f15(GE_context* ac, T0* C, T14 a1);
/* detachable C_STRING.string */
extern T0* T30f5(GE_context* ac, T0* C);
/* detachable C_STRING.c_strlen */
extern T6 T30f6(GE_context* ac, T0* C, T14 a1);
/* detachable C_STRING.make_shared_from_pointer */
extern T0* T30c14(GE_context* ac, T14 a1);
/* [detachable] FILE_INFO.utf_16_to_multi_byte */
extern T6 T132f18(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_DIRECTORY.file_info */
extern T0* T327f24(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.readentry */
extern void T327f38(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_next */
extern T14 T327f22(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_C_GENERATOR.report_cannot_read_error */
extern void T65f633(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.is_open_read */
extern T1 T327f26(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.is_closed */
extern T1 T327f16(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.open_read */
extern void T327f32(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_open_read */
extern void T327f37(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dir_open */
extern T14 T327f19(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.internal_name_pointer */
extern T0* T327f18(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_is_readable */
extern T1 T327f13(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_is_readable */
extern T1 T327f28(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.old_exists */
extern T1 T327f12(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_exists */
extern T1 T327f27(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.make */
extern void T327f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.make */
extern T0* T327c31(GE_context* ac, T0* a1);
/* [detachable] KL_DIRECTORY.old_make */
extern void T327f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.make_with_name */
extern void T327f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.set_name */
extern void T327f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.string_ */
extern T0* T327f23(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_substring */
extern T0* T83f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_end_position */
extern T6 T83f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_start_position */
extern T6 T83f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.recognizes */
extern T1 T83f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.has_matched */
extern T1 T83f236(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_key_equality_tester */
extern void T212f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.string_equality_tester */
extern T0* T65f284(GE_context* ac, T0* C);
/* [detachable] KL_STRING_EQUALITY_TESTER.default_create */
extern T0* T149c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map */
extern void T212f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map */
extern T0* T212c48(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_with_equality_testers */
extern void T212f57(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_sparse_container */
extern void T212f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_internal_cursor */
extern void T212f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_slots */
extern void T212f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_clashes */
extern void T212f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_key_storage */
extern void T212f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].make */
extern T0* T373f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].default_create */
extern T0* T373c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_item_storage */
extern void T212f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.found_item */
extern T0* T295f21(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.found */
extern T1 T295f30(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.search */
extern void T295f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.force */
extern void T295f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_put */
extern void T295f57(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.slots_put */
extern void T295f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.clashes_put */
extern void T295f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.resize */
extern void T295f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.clashes_resize */
extern void T295f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.special_integer_ */
extern T0* T295f38(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_resize */
extern void T295f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_resize */
extern void T295f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.slots_resize */
extern void T295f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.new_modulus */
extern T6 T295f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.new_capacity */
extern T6 T295f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_put */
extern void T295f53(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.unset_found_item */
extern void T295f52(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.starts_with */
extern T1 T743f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.starts_with */
extern T1 T17f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.interpreted_string */
extern T0* T87f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.expanded_string */
extern T0* T87f5(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] KL_EXECUTION_ENVIRONMENT.variable_value */
extern T0* T87f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.environment_impl */
extern T0* T87f1(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.string_ */
extern T0* T87f2(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.replace_all */
extern T0* T83f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_replace_all_to_string */
extern void T83f313(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_unbounded_substring */
extern void T83f327(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_substring */
extern void T83f325(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_it */
extern void T83f340(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_start */
extern T1 T83f151(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_internal */
extern T6 T83f237(GE_context* ac, T0* C, T6 a1, T1 a2, T1 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_type */
extern T6 T83f303(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_not_repeated_characters */
extern T6 T83f302(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.infinity */
extern T6 T83f59(GE_context* ac, T0* C);
/* [detachable] KL_PLATFORM.maximum_integer */
extern T6 T470f1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.platform */
extern T0* T83f194(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_characters */
extern T6 T83f301(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_BYTE_CODE.character_item */
extern T10 T419f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_classes */
extern T6 T83f300(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_BYTE_CODE.character_set_has */
extern T1 T419f5(GE_context* ac, T0* C, T6 a1, T10 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].item */
extern T0* T746f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_refs */
extern T6 T83f299(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_ref */
extern T6 T83f305(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.space_set */
extern T0* T83f152(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.make */
extern T0* T420c16(GE_context* ac, T0* a1);
/* [detachable] RX_CHARACTER_SET.add_string */
extern void T420f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_CHARACTER_SET.add_character */
extern void T420f14(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].force */
extern void T749f60(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_put */
extern void T749f78(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].force */
extern void T747f6(GE_context* ac, T0* C, T0* a1, T10 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].fill_with */
extern void T745f9(GE_context* ac, T0* C, T10 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].extend */
extern void T745f11(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_put */
extern void T749f77(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_put */
extern void T749f76(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_item */
extern T6 T749f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_item */
extern T6 T749f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].hash_position */
extern T6 T749f28(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.hash_code */
extern T6 T10f26(GE_context* ac, T10* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].resize */
extern void T749f75(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_resize */
extern void T749f93(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].special_integer_ */
extern T0* T749f49(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_resize */
extern void T749f92(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].aliased_resized_area */
extern T0* T747f3(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].aliased_resized_area */
extern T0* T745f5(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_resize */
extern void T749f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].aliased_resized_area */
extern T0* T1393f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_64].aliased_resized_area */
extern T0* T748f5(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_item */
extern T10 T749f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_resize */
extern void T749f90(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_modulus */
extern T6 T749f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_capacity */
extern T6 T749f37(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_put */
extern void T749f65(GE_context* ac, T0* C, T11 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].force */
extern void T1393f4(GE_context* ac, T0* C, T0* a1, T11 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].fill_with */
extern void T748f12(GE_context* ac, T0* C, T11 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].search_position */
extern void T749f61(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].unset_found_item */
extern void T749f64(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].value */
extern T11 T749f34(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_item */
extern T11 T749f38(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].position_of_key */
extern T6 T749f22(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].has */
extern T1 T749f33(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_map */
extern T0* T749c50(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_with_equality_testers */
extern void T749f62(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_sparse_container */
extern void T749f83(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].set_internal_cursor */
extern void T749f69(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_cursor */
extern T0* T749f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].make */
extern T0* T750c10(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_slots */
extern void T749f97(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_clashes */
extern void T749f96(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_key_storage */
extern void T749f95(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].make */
extern T0* T747f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_empty */
extern void T745f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_empty */
extern T0* T745c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].default_create */
extern T0* T747c5(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_item_storage */
extern void T749f94(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].make */
extern T0* T1393f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].default_create */
extern T0* T1393c3(GE_context* ac);
/* [detachable] RX_CHARACTER_SET.masks */
extern T0* T420f9(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_filled */
extern T0* T748c9(GE_context* ac, T11 a1, T6 a2);
/* [detachable] RX_CHARACTER_SET.make_empty */
extern void T420f11(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.make_empty */
extern T0* T420c11(GE_context* ac);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.digit_set */
extern T0* T83f167(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.has */
extern T1 T420f10(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item */
extern T11 T749f36(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_ims_options */
extern void T83f338(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_dotall */
extern T1 T83f86(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_multiline */
extern T1 T83f85(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_caseless */
extern T1 T83f84(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_match_count */
extern void T83f348(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_next_start */
extern void T83f347(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_recursive */
extern T6 T83f298(GE_context* ac, T0* C, T6 a1, T1 a2, T1 a3);
/* [detachable] RX_BYTE_CODE.integer_item */
extern T10 T419f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.next_matching_alternate */
extern T6 T83f296(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_additional_bracket */
extern T6 T83f277(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.opcode_item */
extern T10 T419f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.special_integer_ */
extern T0* T83f54(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.flip_case */
extern T10 T421f4(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.to_integer_64 */
extern T7 T10f9(GE_context* ac, T10* C);
/* [detachable] RX_CASE_MAPPING.to_lower */
extern T10 T421f3(GE_context* ac, T0* C, T10 a1);
/* INTEGER_64.to_natural_32 */
extern T10 T7f3(GE_context* ac, T7* C);
/* INTEGER_64.is_greater_equal */
extern T1 T7f2(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_subject */
extern void T83f345(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_replacement_to_string */
extern void T83f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_captured_substring_to_string */
extern void T83f341(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.string_ */
extern T0* T83f70(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.execution_environment */
extern T0* T65f128(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.default_create */
extern T0* T87c6(GE_context* ac);
/* [detachable] KL_STRING_ROUTINES.new_empty_string */
extern T0* T84f2(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T84f2ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.old_wipe_out */
extern void T743f76(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.wipe_out */
extern void T743f76p1(GE_context* ac, T0* C);
/* [detachable] STRING_8.wipe_out */
extern void T17f59(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.string_ */
extern T0* T65f127(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match */
extern void T83f312(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].item */
extern T0* T143f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.current_system */
extern T0* T65f279(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_case_insensitive */
extern void T83f311(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_caseless */
extern void T83f324(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile */
extern void T83f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile */
extern void T83f310p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_startline */
extern void T83f323(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.has_startline */
extern T1 T83f235(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.first_significant_code */
extern T10 T83f189(GE_context* ac, T0* C, T10 a1, T10 a2, T1 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.find_firstchar */
extern T10 T83f234(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_anchored */
extern void T83f322(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.can_anchored */
extern T1 T83f233(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.ims_options */
extern T10 T83f149(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_dotall */
extern T10 T83f132(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_multiline */
extern T10 T83f130(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_caseless */
extern T10 T83f128(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_regexp */
extern void T83f321(GE_context* ac, T0* C, T10 a1, T1 a2, T1 a3, T7 a4);
/* [detachable] RX_BYTE_CODE.set_count */
extern void T419f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.find_fixed_code_length */
extern T6 T83f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_BYTE_CODE.put_integer */
extern void T419f14(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_branch */
extern void T83f339(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.meta_set */
extern T0* T83f148(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.maxlit */
extern T6 T83f147(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.append_character */
extern void T419f16(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.put_character */
extern void T419f22(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.resize_byte_code */
extern void T419f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].aliased_resized_area_with_default */
extern T0* T747f2(GE_context* ac, T0* C, T0* a1, T10 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].aliased_resized_area_with_default */
extern T0* T745f4(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.special_natural_32_ */
extern T0* T419f9(GE_context* ac, T0* C);
/* INTEGER_64.is_less_equal */
extern T1 T7f7(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_escape */
extern T7 T83f141(GE_context* ac, T0* C, T6 a1, T1 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_hex_number */
extern T10 T83f224(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.xdigit_set */
extern T0* T83f249(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_octal_number */
extern T10 T83f223(GE_context* ac, T0* C, T6 a1);
/* INTEGER_64.to_integer_32 */
extern T6 T7f11(GE_context* ac, T7* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.escape_character */
extern T7 T83f218(GE_context* ac, T0* C, T10 a1);
/* INTEGER_64.is_greater */
extern T1 T7f4(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.to_option_ims */
extern T10 T83f135(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_ichanged */
extern void T83f337(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_dotall */
extern T10 T83f133(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_multiline */
extern T10 T83f131(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_caseless */
extern T10 T83f129(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_decimal_number */
extern T10 T83f110(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_repeats */
extern void T83f344(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T7 a4, T6 a5);
/* [detachable] RX_BYTE_CODE.append_subcopy */
extern void T419f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.put_opcode */
extern void T419f19(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.move_right */
extern void T419f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_single_repeat */
extern void T83f346(GE_context* ac, T0* C, T10 a1, T6 a2, T6 a3, T6 a4, T10 a5, T10 a6);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_character_class */
extern void T83f343(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.append_character_set */
extern void T419f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].force_last */
extern void T746f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].force */
extern void T1385f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].fill_with */
extern void T425f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].extend */
extern void T425f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].resize */
extern void T746f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].aliased_resized_area */
extern T0* T1385f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].aliased_resized_area */
extern T0* T425f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].new_capacity */
extern T6 T746f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].extendible */
extern T1 T746f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_CHARACTER_SET.set_negated */
extern void T420f15(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_CHARACTER_SET.add_set */
extern void T420f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].wipe_out */
extern void T749f59(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_wipe_out */
extern void T749f82(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_wipe_out */
extern void T749f81(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_wipe_out */
extern void T749f80(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_32].keep_head */
extern void T745f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].set_count */
extern void T745f14(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_wipe_out */
extern void T749f79(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].keep_head */
extern void T748f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].set_count */
extern void T748f16(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].move_all_cursors_after */
extern void T749f67(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_next_cursor */
extern void T750f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_after */
extern void T750f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].forth */
extern void T750f11(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_forth */
extern void T749f66(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].add_traversing_cursor */
extern void T749f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_position */
extern void T750f14(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove_traversing_cursor */
extern void T749f84(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].force_new */
extern void T749f58(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].replace_found_item */
extern void T749f57(GE_context* ac, T0* C, T11 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].item */
extern T11 T750f7(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_item */
extern T11 T749f44(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].found_item */
extern T11 T749f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].found */
extern T1 T749f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].search */
extern void T749f56(GE_context* ac, T0* C, T10 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].key */
extern T10 T750f6(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_key */
extern T10 T749f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].after */
extern T1 T750f4(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_iterator */
extern T0* T749f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].start */
extern void T750f12(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_start */
extern void T749f63(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_off */
extern T1 T749f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].off */
extern T1 T750f9(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].forth */
extern void T749f53(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove */
extern void T749f55(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove_position */
extern void T749f74(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_keep_head */
extern void T749f89(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_keep_head */
extern void T749f88(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].move_cursors_forth */
extern void T749f87(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].internal_set_key_equality_tester */
extern void T749f86(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_for_iteration */
extern T10 T749f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].after */
extern T1 T749f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_after */
extern T1 T749f42(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].start */
extern void T749f51(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].copy */
extern void T749f54(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].copy */
extern void T749f54p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_clashes */
extern void T749f73(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] INTEGER_32].resized_area */
extern T0* T126f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].copy_data */
extern void T126f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] INTEGER_32].force */
extern void T126f11(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].move_data */
extern void T126f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].overlapping_move */
extern void T126f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].non_overlapping_move */
extern void T126f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_slots */
extern void T749f72(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_key_storage */
extern void T749f71(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_32].resized_area */
extern T0* T745f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].copy_data */
extern void T745f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] NATURAL_32].force */
extern void T745f16(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].move_data */
extern void T745f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].overlapping_move */
extern void T745f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].non_overlapping_move */
extern void T745f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_item_storage */
extern void T749f70(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].resized_area */
extern T0* T748f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].copy_data */
extern void T748f15(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] NATURAL_64].force */
extern void T748f18(GE_context* ac, T0* C, T11 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_64].move_data */
extern void T748f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].overlapping_move */
extern void T748f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].non_overlapping_move */
extern void T748f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].valid_cursor */
extern T1 T749f35(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].twin */
extern T0* T749f48(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].is_empty */
extern T1 T749f41(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.add_negated_set */
extern void T420f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].replace */
extern void T749f52(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_for_iteration */
extern T11 T749f27(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] RX_CHARACTER_SET].item */
extern T0* T424f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.class_sets */
extern T0* T83f159(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.default_word_set */
extern T0* T83f76(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.punct_set */
extern T0* T83f248(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.print_set */
extern T0* T83f247(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.graph_set */
extern T0* T83f246(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.cntrl_set */
extern T0* T83f245(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.ascii_set */
extern T0* T83f244(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.alnum_set */
extern T0* T83f243(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.upper_set */
extern T0* T83f242(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.lower_set */
extern T0* T83f241(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.alpha_set */
extern T0* T83f239(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.check_posix_name */
extern T6 T83f157(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] STRING_8].item */
extern T0* T324f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].count */
extern T6 T324f6(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.class_names */
extern T0* T83f240(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.check_posix_syntax */
extern T6 T83f155(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_counted_repeats */
extern T1 T83f88(GE_context* ac, T0* C, T6 a1, T7 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_comment */
extern void T83f342(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_undef */
extern T1 T83f61(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.append_integer */
extern void T419f13(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.append_opcode */
extern void T419f12(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_error */
extern void T83f320(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_default_internal_options */
extern void T83f319(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.wipe_out */
extern void T419f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].wipe_out */
extern void T746f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].keep_head */
extern void T425f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].set_count */
extern void T425f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].move_all_cursors_after */
extern void T746f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].set_next_cursor */
extern void T1386f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].set_after */
extern void T1386f6(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.wipe_out */
extern void T83f318(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.wipe_out */
extern void T83f318p1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.make */
extern T0* T83c309(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].make_filled */
extern T0* T375f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.make */
extern void T83f309p1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_default_options */
extern void T83f317(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_strict */
extern void T83f336(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_greedy */
extern void T83f335(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_eol */
extern void T83f334(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_bol */
extern void T83f333(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_dollar_endonly */
extern void T83f332(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_empty_allowed */
extern void T83f331(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_extended */
extern void T83f330(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_dotall */
extern void T83f329(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_multiline */
extern void T83f328(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_word_set */
extern void T83f316(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_character_case_mapping */
extern void T83f315(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.default_character_case_mapping */
extern T0* T83f75(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.make */
extern T0* T421c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] RX_CASE_MAPPING.add */
extern void T421f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_CASE_MAPPING.make_default */
extern void T421f7(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.clear */
extern void T421f9(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].make_filled */
extern T0* T747f1(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_filled */
extern T0* T745c7(GE_context* ac, T10 a1, T6 a2);
/* [detachable] RX_CASE_MAPPING.special_natural_32_ */
extern T0* T421f5(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.reset */
extern void T83f314(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.reset */
extern void T83f314p1(GE_context* ac, T0* C);
/* [detachable] STRING_32.wipe_out */
extern void T18f23(GE_context* ac, T0* C);
/* [detachable] STRING_32.twin */
extern T0* T18f8(GE_context* ac, T0* C);
/* [detachable] STRING_32.copy */
extern void T18f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_32].resized_area */
extern T0* T16f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.empty_pattern_buffer */
extern T0* T83f74(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.make */
extern T0* T419c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].make */
extern T0* T746c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].set_internal_cursor */
extern void T746f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].new_cursor */
extern T0* T746f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].make */
extern T0* T1386c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].make */
extern T0* T1385f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].make_empty */
extern T0* T425c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].default_create */
extern T0* T1385c3(GE_context* ac);
/* [detachable] ET_C_GENERATOR.c_config */
extern T0* T65f126(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.parse_file */
extern void T462f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.parse */
extern void T462f189(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_pop_last_value */
extern void T462f202(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_push_error_value */
extern void T462f201(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].force */
extern void T517f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ANY].fill_with */
extern void T518f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ANY].extend */
extern void T518f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].aliased_resized_area */
extern T0* T517f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ANY].aliased_resized_area */
extern T0* T518f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_do_action */
extern void T462f200(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].force */
extern void T540f4(GE_context* ac, T0* C, T0* a1, T1 a2, T6 a3);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].aliased_resized_area */
extern T0* T540f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].aliased_resized_area */
extern T0* T127f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.is_defined */
extern T1 T462f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].has */
extern T1 T212f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force */
extern void T212f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_CONFIG_PARSER.process_include */
extern void T462f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.make */
extern T0* T1309c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].put */
extern void T1307f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].force */
extern void T1663f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] YY_BUFFER].fill_with */
extern void T1662f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] YY_BUFFER].extend */
extern void T1662f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].put */
extern void T915f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.execution_environment */
extern T0* T462f166(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].is_full */
extern T1 T1307f5(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.undefine_value */
extern void T462f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove */
extern void T212f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove_position */
extern void T212f88(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_keep_head */
extern void T212f92(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_keep_head */
extern void T212f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].move_cursors_forth */
extern void T212f90(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].internal_set_key_equality_tester */
extern void T212f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.ignored */
extern T1 T462f137(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_push_last_value */
extern void T462f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.accept */
extern void T462f198(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_do_error_action */
extern void T462f196(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.report_error */
extern void T462f212(GE_context* ac, T0* C, T0* a1);
extern T1 T462f212ot1(T0* a1, T0** a2);
/* [detachable] UT_SYNTAX_ERROR.make */
extern T0* T1308c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.report_eof_expected_error */
extern void T462f211(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.read_token */
extern void T462f195(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_execute_action */
extern void T462f210(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_set_beginning_of_line */
extern void T462f219(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.item */
extern T2 T1396f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_FILE_BUFFER.set_beginning_of_line */
extern void T908f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] YY_BUFFER.set_beginning_of_line */
extern void T910f16(GE_context* ac, T0* C, T1 a1);
/* [detachable] UT_CONFIG_PARSER.text_item */
extern T2 T462f165(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.text_substring */
extern T0* T462f159(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.substring */
extern T0* T1396f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.text_count */
extern T6 T462f158(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.string_ */
extern T0* T462f154(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.text */
extern T0* T462f153(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.set_start_condition */
extern void T462f217(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_execute_eof_action */
extern void T462f209(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.terminate */
extern void T462f218(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.wrap */
extern T1 T462f132(GE_context* ac, T0* C);
extern T1 T462f132ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].remove */
extern void T1307f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] YY_BUFFER].keep_head */
extern void T1662f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] YY_BUFFER].set_count */
extern void T1662f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].remove */
extern void T915f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] INTEGER_32].keep_head */
extern void T126f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].set_count */
extern void T126f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].item */
extern T6 T915f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].item */
extern T0* T1307f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].is_empty */
extern T1 T1307f6(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_refill_input_buffer */
extern void T462f208(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_set_content */
extern void T462f205(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_BUFFER.count */
extern T6 T1396f6(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.as_unicode_special */
extern T0* T1396f5(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.fill */
extern void T908f21(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_stream */
extern T6 T1396f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_CHARACTER_BUFFER.put */
extern void T1396f12(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.compact_left */
extern void T908f24(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.move_left */
extern void T1396f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] STRING_8.subcopy */
extern void T17f79(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] YY_FILE_BUFFER.resize */
extern void T908f26(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.resize */
extern void T1396f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_FILE_BUFFER.default_capacity */
extern T6 T908f13(GE_context* ac, T0* C);
/* [detachable] YY_BUFFER.fill */
extern void T910f14(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_index */
extern void T908f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_BUFFER.set_index */
extern void T910f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_null_trans_state */
extern T6 T462f131(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_previous_state */
extern T6 T462f130(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.fatal_error */
extern void T462f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_MESSAGE.make */
extern T0* T82c8(GE_context* ac, T0* a1);
/* [detachable] UT_CONFIG_PARSER.report_invalid_unicode_character_error */
extern void T462f206(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.to_hex_string */
extern T0* T10f18(GE_context* ac, T10* C);
/* NATURAL_32.to_hex_character */
extern T2 T10f28(GE_context* ac, T10* C);
/* [detachable] KL_CHARACTER_BUFFER.item_code */
extern T10 T1396f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_at_beginning_of_line */
extern T6 T462f99(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].force */
extern void T375f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].aliased_resized_area */
extern T0* T375f4(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.special_integer_ */
extern T0* T462f105(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_init_value_stacks */
extern void T462f194(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_clear_all */
extern void T462f203(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.clear_all */
extern void T462f215(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.clear_stacks */
extern void T462f220(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_clear_value_stacks */
extern void T462f221(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] BOOLEAN].keep_head */
extern void T127f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].set_count */
extern void T127f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].keep_head */
extern void T518f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].set_count */
extern void T518f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.abort */
extern void T462f197(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.set_input_buffer */
extern void T462f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_load_input_buffer */
extern void T462f193(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_position */
extern void T908f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] YY_BUFFER.set_position */
extern void T910f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] UT_CONFIG_PARSER.new_file_buffer */
extern T0* T462f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.make */
extern T0* T908c18(GE_context* ac, T0* a1);
/* [detachable] YY_FILE_BUFFER.make_with_size */
extern void T908f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.make_with_size */
extern T0* T908c17(GE_context* ac, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.set_file */
extern void T908f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.set_file_with_size */
extern void T908f25(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.flush */
extern void T908f27(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.new_default_buffer */
extern T0* T908f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_CHARACTER_BUFFER.make */
extern T0* T1396c11(GE_context* ac, T6 a1);
/* [detachable] ET_C_GENERATOR.console_application_mode */
extern T1 T65f458(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.define_value */
extern void T462f184(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_CONFIG_PARSER.make */
extern T0* T462c183(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].make */
extern T0* T915c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].make */
extern T0* T375f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].make */
extern T0* T1307c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].make */
extern T0* T1663f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] YY_BUFFER].make_empty */
extern T0* T1662c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].default_create */
extern T0* T1663c2(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.string_equality_tester */
extern T0* T462f97(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.make_parser_skeleton */
extern void T462f187(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_build_parser_tables */
extern void T462f192(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yycheck_template */
extern T0* T462f82(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yyfixed_array */
extern T0* T462f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].to_special */
extern T0* T375f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yytable_template */
extern T0* T462f139(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yypgoto_template */
extern T0* T462f138(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yypact_template */
extern T0* T462f136(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yydefgoto_template */
extern T0* T462f133(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yydefact_template */
extern T0* T462f127(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytypes2_template */
extern T0* T462f126(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytypes1_template */
extern T0* T462f123(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yyr1_template */
extern T0* T462f119(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template */
extern T0* T462f116(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template_2 */
extern void T462f223(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yyarray_subcopy */
extern void T462f225(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].subcopy */
extern void T877f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ARRAY [[attached] INTEGER_32].subcopy */
extern void T284f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] UT_CONFIG_PARSER.integer_array_ */
extern T0* T462f182(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].default_create */
extern T0* T877c2(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template_1 */
extern void T462f222(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] INTEGER_32].make_filled */
extern T0* T284c10(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] INTEGER_32].make_filled_area */
extern void T284f12(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.yy_create_value_stacks */
extern void T462f191(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].make */
extern T0* T540f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].default_create */
extern T0* T540c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].make */
extern T0* T517f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].make_empty */
extern T0* T518c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].default_create */
extern T0* T517c3(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.make_scanner */
extern void T462f186(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.make_with_buffer */
extern void T462f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_initialize */
extern void T462f204(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_initialize */
extern void T462f204p1(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_build_tables */
extern void T462f216(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_accept_template */
extern T0* T462f173(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_fixed_array */
extern T0* T462f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_meta_template */
extern T0* T462f172(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_ec_template */
extern T0* T462f171(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_ec_template_1 */
extern void T462f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_array_subcopy */
extern void T462f226(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] UT_CONFIG_PARSER.yy_def_template */
extern T0* T462f170(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_base_template */
extern T0* T462f169(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_chk_template */
extern T0* T462f168(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_nxt_template */
extern T0* T462f167(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.empty_buffer */
extern T0* T462f83(GE_context* ac, T0* C);
/* [detachable] YY_BUFFER.make */
extern T0* T910c12(GE_context* ac, T0* a1);
/* [detachable] YY_BUFFER.make_from_buffer */
extern void T910f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_string */
extern void T1396f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_substring */
extern void T1396f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] YY_BUFFER.new_default_buffer */
extern T0* T910f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.put_new */
extern void T295f49(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_STRING_HASH_TABLE.set_key_equality_tester */
extern void T295f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.make_map */
extern T0* T295c45(GE_context* ac, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_with_equality_testers */
extern void T295f51(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_STRING_HASH_TABLE.make_sparse_container */
extern void T295f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.set_internal_cursor */
extern void T295f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.new_cursor */
extern T0* T295f37(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.make_slots */
extern void T295f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_clashes */
extern void T295f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_key_storage */
extern void T295f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_item_storage */
extern void T295f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.right_adjust */
extern void T84f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.left_adjust */
extern void T84f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.remove_head */
extern void T743f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.keep_tail */
extern void T743f96(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.read_line */
extern void T32f63(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.unread_character */
extern void T32f68(GE_context* ac, T0* C, T2 a1);
/* detachable KL_LINKABLE [[attached] CHARACTER_8].put_right */
extern void T129f4(GE_context* ac, T0* C, T0* a1);
/* detachable KL_LINKABLE [[attached] CHARACTER_8].make */
extern T0* T129c3(GE_context* ac, T2 a1);
/* [detachable] KL_STRING_ROUTINES.wipe_out */
extern void T84f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.string_ */
extern T0* T32f11(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.nested_pathname */
extern T0* T124f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.nested_pathname */
extern T0* T123f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.generate_c_code */
extern void T65f530(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.dummy_feature */
extern T0* T65f283(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.make */
extern T0* T232c30(GE_context* ac, T0* a1, T0* a2, T0* a3);
extern T1 T232f30ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_set_reference_field_feature */
extern void T61f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.use_all_attributes */
extern void T61f108(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_set_reference_field_dynamic_types */
extern void T226f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_set_reference_field_dynamic_types */
extern void T70f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_giaaa_error */
extern void T51f190(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_internal_error */
extern void T51f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_internal_error */
extern void T51f202(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_INTERNAL_ERROR.make_giaaa */
extern T0* T758c13(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.report_giaaa_error */
extern void T35f190(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_internal_error */
extern void T35f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_internal_error */
extern void T35f202(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.error_handler */
extern T0* T70f89(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_fatal_error */
extern void T70f113(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_formal_argument_dynamic_types */
extern void T70f128(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.item */
extern T0* T220f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_reference_field_feature */
extern void T61f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_reference_field_dynamic_types */
extern void T226f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_reference_field_dynamic_types */
extern void T70f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_result_dynamic_types */
extern void T70f123(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_new_type_instance_of_feature */
extern void T61f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_type_of_type_result_type */
extern void T61f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_type_of_type_result_type */
extern void T226f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_type_of_type_result_type */
extern void T70f122(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SYSTEM.create_meta_type */
extern void T61f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.meta_type */
extern T0* T61f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_primary_type */
extern T0* T61f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_dynamic_types */
extern void T61f92(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_alive_conforming_descendants */
extern void T61f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_alive_conforming_descendants */
extern void T226f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].forth */
extern void T378f40(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_forth */
extern void T378f44(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].add_traversing_cursor */
extern void T378f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_next_cursor */
extern void T897f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_position */
extern void T897f11(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].remove_traversing_cursor */
extern void T378f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_after */
extern void T897f9(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].clashes_item */
extern T6 T378f35(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_for_iteration */
extern T0* T378f22(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_item */
extern T0* T378f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_storage_item */
extern T0* T378f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].after */
extern T1 T378f34(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_after */
extern T1 T378f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].after */
extern T1 T897f7(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].start */
extern void T378f39(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_start */
extern void T378f43(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_off */
extern T1 T378f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].off */
extern T1 T897f5(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].is_empty */
extern T1 T378f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_alive_conforming_descendants */
extern void T70f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_attribute_type_sets */
extern void T61f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_predicate_type */
extern void T61f100(GE_context* ac, T0* C, T0* a1);
extern T1 T61f100ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_set_rout_disp_final_feature */
extern void T444f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_set_rout_disp_final_feature */
extern void T443f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_regular */
extern void T298f37(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_regular */
extern void T232f31(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.item */
extern T0* T299f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.is_predicate_class */
extern T1 T73f64(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.class_codes */
extern T0* T73f109(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.default_create */
extern T0* T410c35(GE_context* ac);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_function_type */
extern void T61f99(GE_context* ac, T0* C, T0* a1);
extern T1 T61f99ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.is_function_class */
extern T1 T73f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_procedure_type */
extern void T61f98(GE_context* ac, T0* C, T0* a1);
extern T1 T61f98ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.is_procedure_class */
extern T1 T73f62(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_typed_pointer_type */
extern void T61f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_typed_pointer_class */
extern T1 T73f61(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_array_type */
extern void T61f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_array_class */
extern T1 T73f60(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_special_type */
extern void T61f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_special_class */
extern T1 T73f59(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_interface_error */
extern T1 T73f80(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.interface_checked */
extern T1 T73f77(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_fatal_error */
extern void T61f91(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_implementation_error */
extern T1 T73f105(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.implementation_checked */
extern T1 T73f94(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].force_last */
extern void T222f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_put */
extern void T222f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_put */
extern void T222f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_put */
extern void T222f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_item */
extern T6 T222f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].hash_position */
extern T6 T222f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].resize */
extern void T222f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_resize */
extern void T222f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].special_integer_ */
extern T0* T222f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_resize */
extern void T222f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_resize */
extern void T222f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T412f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T411f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_item */
extern T0* T222f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_item */
extern T6 T222f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_resize */
extern void T222f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_modulus */
extern T6 T222f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_capacity */
extern T6 T222f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_put */
extern void T222f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T412f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].fill_with */
extern void T411f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].extend */
extern void T411f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].search_position */
extern void T222f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].unset_found_item */
extern void T222f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].found_item */
extern T0* T222f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_item */
extern T0* T222f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].found */
extern T1 T222f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].search */
extern void T222f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.type_name_buffer */
extern T0* T61f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_dynamic_primary_type */
extern T0* T61f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.register_dynamic_type */
extern void T61f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_conforming_ancestors */
extern void T61f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.has_class */
extern T1 T408f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.item */
extern T0* T408f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.null_system_processor */
extern T0* T72f23(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.make_null */
extern void T62f59(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.make_null */
extern T0* T62c59(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.set_qualified_anchored_types_cycle_detection_enabled_only */
extern void T62f74(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_unknown_builtin_reported_only */
extern void T62f73(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_preparse_shallow_mode_only */
extern void T62f72(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_default_keyword_usage_only */
extern void T62f71(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_detachable_keyword_only */
extern void T62f79(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_attached_keyword_only */
extern void T62f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_reference_keyword_only */
extern void T62f77(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_note_keyword_only */
extern void T62f76(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_attribute_keyword_only */
extern void T62f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.make */
extern T0* T460c3(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.null_error_handler */
extern T0* T72f126(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.make_null */
extern T0* T51c177(GE_context* ac);
/* [detachable] ET_NULL_ERROR_HANDLER.make_null */
extern void T51f177p1(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.null_output_stream */
extern T0* T51f23(GE_context* ac, T0* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.make */
extern T0* T157c3(GE_context* ac, T0* a1);
/* [detachable] MUTEX.make */
extern T0* T133c10(GE_context* ac);
/* [detachable] MUTEX.eif_thr_mutex_create */
extern T14 T133f6(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.null_ast_processor */
extern T0* T72f125(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_PROCESSOR.make */
extern T0* T457c1(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.default_ast_factory */
extern T0* T72f124(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.make */
extern T0* T458c362(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.tokens */
extern T0* T62f45(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.default_create */
extern T0* T72c1525(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make_default */
extern T0* T259c23(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make */
extern void T259f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make */
extern T0* T259c16(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].set_internal_cursor */
extern void T259f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].new_cursor */
extern T0* T259f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].make */
extern T0* T620c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].default_capacity */
extern T6 T259f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T266f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T266f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T266f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T266f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T266f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T266f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T266f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T266f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T266f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T266f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T266f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T266f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T266f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T266f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T266f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T266f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T266f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T266f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_equality_tester */
extern T0* T266f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T266f47(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_none */
extern T1 T73f179(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_index */
extern void T73f262(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T75f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T75f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T75f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].extendible */
extern T1 T75f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.make */
extern T0* T219c54(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T266c37(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T266f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T266f18(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T328c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T266f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T266f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T266f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T266f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T412f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_empty */
extern T0* T411c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].default_create */
extern T0* T412c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.empty_features */
extern T0* T219f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.make */
extern T0* T220c14(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T596f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].make_empty */
extern T0* T595c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.fixed_array */
extern T0* T220f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].default_create */
extern T0* T596c3(GE_context* ac);
/* [detachable] ET_CLASS.hash_code */
extern T6 T73f184(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_predicate_type */
extern T0* T61f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.make */
extern T0* T444c57(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.make_type */
extern void T444f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.empty_features */
extern T0* T444f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.empty_dynamic_type_sets */
extern T0* T61f70(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.make */
extern T0* T255c14(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].make */
extern T0* T615f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].make_empty */
extern T0* T614c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.fixed_array */
extern T0* T255f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].default_create */
extern T0* T615c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_SYSTEM.new_function_type */
extern T0* T61f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.new_procedure_type */
extern T0* T61f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.make */
extern T0* T443c57(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.make_type */
extern void T443f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.empty_features */
extern T0* T443f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_type_type */
extern T0* T61f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_class */
extern T1 T73f187(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_tuple_type */
extern T0* T61f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.make */
extern T0* T294c55(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.item */
extern T0* T255f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.make_type */
extern void T294f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.empty_features */
extern T0* T294f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_special_type */
extern T0* T61f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.make */
extern T0* T311c55(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.make_type */
extern void T311f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.empty_features */
extern T0* T311f43(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_tuple_class */
extern T1 T73f84(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_generic */
extern T1 T73f183(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_type_mark */
extern T0* T72f21(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_detachable_mark */
extern void T342f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.make */
extern T0* T342c26(GE_context* ac);
/* [detachable] ET_DYNAMIC_SYSTEM.tokens */
extern T0* T61f75(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_index */
extern void T1715f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_index */
extern void T252f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.name */
extern T0* T1715f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.identifier */
extern T0* T180f70(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.name */
extern T0* T252f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.put_last */
extern void T255f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].force */
extern void T615f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].fill_with */
extern void T614f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].extend */
extern void T614f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.type */
extern T0* T1715f7(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.type */
extern T0* T252f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.argument_type_set */
extern T0* T298f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_FEATURE.argument_type_set */
extern T0* T232f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.formal_argument */
extern T0* T249f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_ARGUMENT.formal_argument */
extern T0* T252f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.formal_argument */
extern T0* T1715f6(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.item */
extern T0* T249f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.make_with_capacity */
extern T0* T255c9(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.new_dynamic_type_set */
extern T0* T226f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_dynamic_type_set */
extern T0* T70f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.make */
extern T0* T469c15(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type */
extern void T469f20(GE_context* ac, T0* C, T0* a1);
extern T1 T469f20ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.force_last */
extern void T260f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.fixed_array */
extern T0* T260f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.hash_position */
extern T6 T260f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.modulus */
extern T6 T260f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.resize */
extern void T260f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.special_integer_ */
extern T0* T260f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.new_modulus */
extern T6 T260f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.new_capacity */
extern T6 T260f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.capacity */
extern T6 T260f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.put_last */
extern void T260f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.append_last */
extern void T260f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.make_with_capacity */
extern T0* T260c17(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.alive_conforming_descendants */
extern T0* T226f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].found_item */
extern T0* T378f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].found */
extern T1 T378f27(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].search */
extern void T378f41(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].search_position */
extern void T378f45(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].hash_position */
extern T6 T378f25(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].key_storage_item */
extern T0* T378f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].slots_item */
extern T6 T378f20(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_map_default */
extern T0* T378c38(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_map */
extern void T378f42(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_with_equality_testers */
extern void T378f46(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_sparse_container */
extern void T378f49(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_internal_cursor */
extern void T378f55(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].new_cursor */
extern T0* T378f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make */
extern T0* T897c8(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].unset_found_item */
extern void T378f54(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_slots */
extern void T378f53(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].new_modulus */
extern T6 T378f28(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_clashes */
extern void T378f52(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_key_storage */
extern void T378f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].make */
extern T0* T688f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].make_empty */
extern T0* T687c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].default_create */
extern T0* T688c3(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_item_storage */
extern void T378f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].default_capacity */
extern T6 T378f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.alive_conforming_descendants */
extern T0* T70f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.object_id_dynamic_type_set */
extern T0* T226f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.mark_type_alive */
extern void T226f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.current_system */
extern T0* T226f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.mark_type_alive */
extern void T70f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_system */
extern T0* T70f81(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_type */
extern T0* T61f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.make */
extern T0* T315c24(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_type_mark */
extern T0* T72f22(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_attached_mark */
extern void T342f28(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.tokens */
extern T0* T232f26(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.empty_dynamic_type_sets */
extern T0* T232f24(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.make */
extern T0* T466c78(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.empty_clients */
extern T0* T72f162(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.make */
extern T0* T397c15(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].make */
extern T0* T734f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].make_empty */
extern T0* T733c5(GE_context* ac, T6 a1);
/* [detachable] ET_CLIENT_LIST.fixed_array */
extern T0* T397f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].default_create */
extern T0* T734c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.end_keyword */
extern T0* T72f88(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_end */
extern T0* T409c47(GE_context* ac);
/* [detachable] ET_KEYWORD.make_token */
extern void T409f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_leaf */
extern void T409f45(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.tokens */
extern T0* T409f4(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.tokens */
extern T0* T466f32(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.make */
extern T0* T180c109(GE_context* ac, T0* a1);
/* [detachable] ET_IDENTIFIER.make_with_hash_code */
extern void T180f119(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_IDENTIFIER.make_with_hash_code */
extern T0* T180c119(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_IDENTIFIER.make_leaf */
extern void T180f120(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.new_hash_code */
extern T6 T180f19(GE_context* ac, T0* C, T0* a1);
/* CHARACTER_8.upper */
extern T2 T2f6(GE_context* ac, T2* C);
/* CHARACTER_8.is_lower */
extern T1 T2f14(GE_context* ac, T2* C);
/* [detachable] ET_C_GENERATOR.close_cpp_file */
extern void T65f564(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.close_c_file */
extern void T65f563(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.delete_file */
extern void T124f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.delete */
extern void T32f76(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_delete */
extern void T32f79(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_unlink */
extern void T32f81(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.delete_file */
extern void T123f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_file */
extern void T65f562(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.rename_file */
extern void T124f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_TEXT_INPUT_FILE.change_name */
extern void T32f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_change_name */
extern void T32f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.rename_file */
extern void T32f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.set_name */
extern void T32f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.file_rename */
extern void T32f82(GE_context* ac, T0* C, T14 a1, T14 a2);
/* [detachable] KL_TEXT_INPUT_FILE.same_physical_file */
extern T1 T32f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.canonical_pathname */
extern T0* T124f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname_to_string */
extern T0* T124f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.item */
extern T0* T475f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [detachable STRING_8].item */
extern T0* T1313f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_PATHNAME.is_parent */
extern T1 T475f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_PATHNAME.is_current */
extern T1 T475f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.root_directory */
extern T0* T124f15(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.set_canonical */
extern void T475f21(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.append_current */
extern void T475f23(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.append_name */
extern void T475f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [detachable STRING_8].force */
extern void T1313f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].move_data */
extern void T487f10(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].overlapping_move */
extern void T487f12(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].force */
extern void T487f13(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].non_overlapping_move */
extern void T487f11(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable STRING_8].capacity */
extern T6 T1313f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].set_area */
extern void T1313f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable STRING_8].aliased_resized_area */
extern T0* T487f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable STRING_8].fill_with */
extern void T487f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].extend */
extern void T487f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [detachable STRING_8].make_empty_area */
extern void T1313f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable STRING_8].make_empty */
extern void T487f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable STRING_8].make_empty */
extern T0* T487c5(GE_context* ac, T6 a1);
/* [detachable] ARRAY [detachable STRING_8].additional_space */
extern T6 T1313f6(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].empty_area */
extern T1 T1313f5(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].put */
extern void T1313f10(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.string_to_pathname */
extern T0* T124f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.append_parent */
extern void T475f24(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.set_relative */
extern void T475f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] KL_PATHNAME.make */
extern T0* T475c20(GE_context* ac);
/* [detachable] ARRAY [detachable STRING_8].make_filled */
extern T0* T1313c9(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable STRING_8].make_filled_area */
extern void T1313f12(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].make_filled */
extern T0* T487c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.canonical_pathname */
extern T0* T123f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname_to_string */
extern T0* T123f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.root_directory */
extern T0* T123f17(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.string_to_pathname */
extern T0* T123f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_drive */
extern void T475f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_sharename */
extern void T475f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_hostname */
extern void T475f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.file_system */
extern T0* T32f43(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.unix_file_system */
extern T0* T32f53(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.windows_file_system */
extern T0* T32f52(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.operating_system */
extern T0* T32f51(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.time_stamp */
extern T6 T32f42(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.date */
extern T6 T32f50(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.eif_file_date */
extern T6 T32f56(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_INPUT_FILE.count */
extern T6 T32f41(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_count */
extern T6 T32f49(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_size */
extern T6 T32f55(GE_context* ac, T0* C, T14 a1);
/* [detachable] FILE_INFO.size */
extern T6 T132f17(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_open_write */
extern T1 T32f54(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.inode */
extern T6 T32f40(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.inode */
extern T6 T132f16(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.tmp_file1 */
extern T0* T32f39(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.rename_file */
extern void T123f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T287f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T287f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T287f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T287f41(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T411f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_count */
extern void T411f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T287f40(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].keep_head */
extern void T664f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].set_count */
extern void T664f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T287f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T287f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T666f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T666f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].wipe_out */
extern void T286f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_wipe_out */
extern void T286f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_wipe_out */
extern void T286f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_wipe_out */
extern void T286f47(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].keep_head */
extern void T661f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].set_count */
extern void T661f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_wipe_out */
extern void T286f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].unset_found_item */
extern void T286f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].move_all_cursors_after */
extern void T286f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_next_cursor */
extern void T660f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_after */
extern void T660f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].wipe_out */
extern void T285f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_wipe_out */
extern void T285f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_wipe_out */
extern void T285f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_wipe_out */
extern void T285f48(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FEATURE].keep_head */
extern void T654f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE].set_count */
extern void T654f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_wipe_out */
extern void T285f47(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].keep_head */
extern void T657f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].set_count */
extern void T657f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].unset_found_item */
extern void T285f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].move_all_cursors_after */
extern void T285f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_next_cursor */
extern void T656f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_after */
extern void T656f9(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].fill_with */
extern void T284f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].wipe_out */
extern void T283f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_wipe_out */
extern void T283f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_wipe_out */
extern void T283f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_wipe_out */
extern void T283f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_wipe_out */
extern void T283f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].unset_found_item */
extern void T283f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].move_all_cursors_after */
extern void T283f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].set_next_cursor */
extern void T653f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].set_after */
extern void T653f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].wipe_out */
extern void T289f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_wipe_out */
extern void T289f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_wipe_out */
extern void T289f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_wipe_out */
extern void T289f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_wipe_out */
extern void T289f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].unset_found_item */
extern void T289f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].move_all_cursors_after */
extern void T289f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_next_cursor */
extern void T669f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_after */
extern void T669f9(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].wipe_out */
extern void T69f45(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_wipe_out */
extern void T69f64(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_wipe_out */
extern void T69f63(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_wipe_out */
extern void T69f62(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_wipe_out */
extern void T69f61(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].unset_found_item */
extern void T69f48(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].move_all_cursors_after */
extern void T69f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T372f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_after */
extern void T372f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].forth */
extern void T289f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_forth */
extern void T289f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].add_traversing_cursor */
extern void T289f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_position */
extern void T669f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].remove_traversing_cursor */
extern void T289f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_item */
extern T6 T289f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_for_iteration */
extern T0* T289f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_key */
extern T0* T289f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_item */
extern T0* T289f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_for_iteration */
extern T1 T289f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_item */
extern T1 T289f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_item */
extern T1 T289f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].after */
extern T1 T289f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_after */
extern T1 T289f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].after */
extern T1 T669f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].start */
extern void T289f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_start */
extern void T289f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_off */
extern T1 T289f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].off */
extern T1 T669f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].is_empty */
extern T1 T289f31(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].forth */
extern void T69f46(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_forth */
extern void T69f65(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].add_traversing_cursor */
extern void T69f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_position */
extern void T372f12(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].remove_traversing_cursor */
extern void T69f70(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_item */
extern T6 T69f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.include_runtime_file */
extern void T65f561(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_SET [[attached] STRING_8].item_for_iteration */
extern T0* T69f30(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_item */
extern T0* T69f36(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_item */
extern T0* T69f18(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].after */
extern T1 T69f22(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_after */
extern T1 T69f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].after */
extern T1 T372f8(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].start */
extern void T69f42(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_start */
extern void T69f58(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_off */
extern T1 T69f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].off */
extern T1 T372f5(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].is_empty */
extern T1 T69f33(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_end_extern_c */
extern void T65f560(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_main_function */
extern void T65f559(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.dedent */
extern void T65f609(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.remove_last */
extern void T255f11(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].keep_head */
extern void T615f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].keep_head */
extern void T614f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].set_count */
extern void T614f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_creation_expression */
extern void T65f632(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].wipe_out */
extern void T263f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].keep_head */
extern void T622f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].set_count */
extern void T622f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].move_all_cursors_after */
extern void T263f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].set_next_cursor */
extern void T624f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].set_after */
extern void T624f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_attachment_expression */
extern void T65f667(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_item_access */
extern void T65f637(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_non_void_expression */
extern void T65f645(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.print_expression */
extern void T65f644(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_item_name */
extern void T65f627(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_expression */
extern void T65f689(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_cast */
extern void T65f682(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_name */
extern void T65f626(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].wipe_out */
extern void T259f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].move_all_cursors_after */
extern void T259f25(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].set_next_cursor */
extern void T620f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].set_after */
extern void T620f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_last_new */
extern void T212f56(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].found_item */
extern T0* T212f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].found */
extern T1 T212f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].search */
extern void T212f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].item */
extern T6 T259f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].sort */
extern void T259f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].sort */
extern void T335f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].sort_with_comparator */
extern void T335f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].subsort_with_comparator */
extern void T335f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] ARRAY [[attached] INTEGER_32].force */
extern void T284f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] INTEGER_32].capacity */
extern T6 T284f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].set_area */
extern void T284f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] INTEGER_32].make_empty_area */
extern void T284f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] INTEGER_32].additional_space */
extern T6 T284f6(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].empty_area */
extern T1 T284f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].swap */
extern void T259f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].replace */
extern void T259f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].less_than */
extern T1 T463f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].detachable_less_than */
extern T1 T463f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].attached_less_than */
extern T1 T463f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] INTEGER_32].put */
extern void T284f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_C_GENERATOR.dynamic_type_id_sorter */
extern T0* T65f233(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].make */
extern T0* T335c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].make */
extern T0* T463c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].is_empty */
extern T1 T259f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].put_last */
extern void T259f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].resize */
extern void T259f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.argument_type_set_in_feature */
extern T0* T65f258(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].item */
extern T0* T263f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_creation_procedure_name */
extern void T65f681(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_type_name */
extern void T65f628(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.new_temp_variable */
extern T0* T65f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].force_last */
extern void T259f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].new_capacity */
extern T6 T259f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].extendible */
extern T1 T259f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T258f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T618f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].fill_with */
extern void T617f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].extend */
extern void T617f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T258f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T618f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T617f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T258f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].extendible */
extern T1 T258f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].force_last */
extern void T257f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].force */
extern void T589f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].fill_with */
extern void T590f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].extend */
extern void T590f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].resize */
extern void T257f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].aliased_resized_area */
extern T0* T589f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].aliased_resized_area */
extern T0* T590f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].new_capacity */
extern T6 T257f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].extendible */
extern T1 T257f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_temporary */
extern void T180f116(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].item */
extern T0* T257f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].replace */
extern void T258f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_C_GENERATOR.same_declaration_types */
extern T1 T65f459(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T258f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_writable */
extern void T65f680(GE_context* ac, T0* C, T0* a1);
extern T1 T65f680ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_attribute_access */
extern void T65f607(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_attribute_name */
extern void T65f629(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_type_cast */
extern void T65f603(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.current_keyword */
extern T0* T72f24(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.make */
extern T0* T273c32(GE_context* ac);
/* [detachable] ET_CURRENT.make_token */
extern void T273f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.make_leaf */
extern void T273f35(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.tokens */
extern T0* T273f5(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.tokens */
extern T0* T65f270(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].remove */
extern void T290f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].remove_position */
extern void T290f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_put */
extern void T290f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_put */
extern void T290f42(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_keep_head */
extern void T290f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_keep_head */
extern void T290f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].keep_head */
extern void T590f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].set_count */
extern void T590f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_put */
extern void T290f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_put */
extern void T290f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_item */
extern T6 T290f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].move_cursors_forth */
extern void T290f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_position */
extern void T672f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_next_cursor */
extern void T672f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_after */
extern void T672f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].internal_set_key_equality_tester */
extern void T290f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_equality_tester */
extern T0* T290f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_item */
extern T6 T290f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].hash_position */
extern T6 T290f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_item */
extern T0* T290f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_item */
extern T0* T290f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].search_position */
extern void T290f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TESTER.test */
extern T1 T291f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.same_identifier */
extern T1 T180f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.same_case_insensitive */
extern T1 T84f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* CHARACTER_8.lower */
extern T2 T2f5(GE_context* ac, T2* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code */
extern T6 T423f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.lower_codes */
extern T0* T423f7(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_lower_code_plane */
extern T0* T423f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].make_filled */
extern T0* T751c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].fill_with */
extern void T751f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].extend */
extern void T751f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].make_empty */
extern void T751f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.empty_lower_code_segment */
extern T0* T423f17(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.to_integer */
extern T6 T753f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.integer_ */
extern T0* T423f28(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_1 */
extern T0* T423f9(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_1_segment_4 */
extern T0* T423f27(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].make_filled */
extern T0* T752c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].fill_with */
extern void T752f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].extend */
extern void T752f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].make_empty */
extern void T752f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0 */
extern T0* T423f8(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_255 */
extern T0* T423f26(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_167 */
extern T0* T423f25(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_166 */
extern T0* T423f24(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_44 */
extern T0* T423f23(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_36 */
extern T0* T423f22(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_33 */
extern T0* T423f21(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_31 */
extern T0* T423f20(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_30 */
extern T0* T423f19(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_16 */
extern T0* T423f18(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_5 */
extern T0* T423f16(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_4 */
extern T0* T423f15(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_3 */
extern T0* T423f14(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_2 */
extern T0* T423f13(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_1 */
extern T0* T423f12(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_0 */
extern T0* T423f11(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.unicode */
extern T0* T84f13(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.default_create */
extern T0* T423c57(GE_context* ac);
/* [detachable] ET_IDENTIFIER.string_ */
extern T0* T180f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].unset_found_item */
extern void T290f40(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_local_variable */
extern void T65f696(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_local_name */
extern void T65f702(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].has */
extern T1 T290f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].force_last */
extern void T290f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].resize */
extern void T290f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_resize */
extern void T290f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].special_integer_ */
extern T0* T290f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_resize */
extern void T290f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_resize */
extern void T290f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_resize */
extern void T290f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_modulus */
extern T6 T290f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_capacity */
extern T6 T290f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_across_cursor */
extern void T65f695(GE_context* ac, T0* C, T0* a1);
extern T1 T65f695ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_across_cursor_name */
extern void T65f701(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_qualified_call_expression */
extern void T65f700(GE_context* ac, T0* C, T0* a1);
extern T1 T65f700ot1(T0* a1, T0** a2);
extern T1 T65f700ot2(T0* a1, T0** a2);
extern T1 T65f700ot3(T0* a1, T0** a2);
/* [detachable] ET_MANIFEST_TUPLE.wipe_out */
extern void T281f32(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].keep_head */
extern void T645f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].keep_head */
extern void T643f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].set_count */
extern void T643f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.fixed_array */
extern T0* T281f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].default_create */
extern T0* T645c3(GE_context* ac);
/* [detachable] ET_C_GENERATOR.mark_expressions_unfrozen */
extern void T65f723(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_unfrozen */
extern void T65f796(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.is_temp_variable_frozen */
extern T1 T65f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_call_name */
extern void T65f650(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T65f650ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_eif_any_type_name */
extern void T65f570(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.dynamic_type_set_in_feature */
extern T0* T65f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_PRECURSOR.dynamic_type_set */
extern T0* T298f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.dynamic_type_set */
extern T0* T232f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_attribute_type_id_access */
extern void T65f653(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_type_id_name */
extern void T65f639(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.seeded_query */
extern T0* T73f182(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUERY_LIST.seeded_feature */
extern T0* T253f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_adapted_named_query_call */
extern void T65f655(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_named_query_call */
extern T0* T65f655ac1(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T65f655ot1(T0* a1, T0** a2);
/* [detachable] PROCEDURE [[attached] TUPLE].set_rout_disp_final */
extern void T344f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_C_GENERATOR.print_attribute_tuple_item_access */
extern void T65f618(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_attribute_tuple_item_name */
extern void T65f642(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_adapted_query_call */
extern void T65f685(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_query_call */
extern T0* T65f685ac1(GE_context* ac, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_query_call */
extern void T65f697(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f697ot1(T0* a1, T0** a2);
extern T1 T65f697ot2(T0* a1, T0** a2);
extern T1 T65f697ot3(T0* a1, T0** a2);
extern T1 T65f697ot4(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_non_inlined_query_call */
extern void T65f705(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_query_call */
extern void T65f704(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_query_call */
extern void T65f745(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_runtime_name_call */
extern void T65f998(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_type_id_call */
extern void T65f997(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_expanded_call */
extern void T65f996(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_deferred_call */
extern void T65f995(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_attached_call */
extern void T65f994(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_has_default_call */
extern void T65f993(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_generic_parameter_count_call */
extern void T65f992(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_default_call */
extern void T65f991(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_query_call */
extern void T65f744(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_object_comparison_call */
extern void T65f990(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_count_call */
extern void T65f989(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f989ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_query_call */
extern void T65f743(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_item_call */
extern void T65f988(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_item_access */
extern void T65f617(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_item_name */
extern void T65f630(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_element_size_call */
extern void T65f987(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f987ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_count_call */
extern void T65f986(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_count_access */
extern void T65f606(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_count_name */
extern void T65f647(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.first */
extern T0* T220f4(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_special_capacity_call */
extern void T65f985(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_capacity_access */
extern void T65f605(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_capacity_name */
extern void T65f646(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_base_address_call */
extern void T65f984(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_query_call */
extern void T65f742(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_positive_infinity_call */
extern void T65f983(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_negative_infinity_call */
extern void T65f982(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_nan_call */
extern void T65f981(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_item_call */
extern void T65f980(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_unboxed_expression */
extern void T65f670(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_query_call */
extern void T65f741(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_real_call */
extern void T65f979(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_integer_64_call */
extern void T65f978(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_integer_call */
extern void T65f977(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_to_double_call */
extern void T65f976(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_quotient_call */
extern void T65f975(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_product_call */
extern void T65f974(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_power_call */
extern void T65f973(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_plus_call */
extern void T65f972(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_opposite_call */
extern void T65f971(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_minus_call */
extern void T65f970(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_positive_infinity_call */
extern void T65f969(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_negative_infinity_call */
extern void T65f968(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_nan_call */
extern void T65f967(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_less_call */
extern void T65f966(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_query_c_call */
extern void T65f1002(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3, T0* a4, T1 a5);
/* [detachable] ET_C_GENERATOR.print_declaration_type_cast */
extern void T65f665(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.total_order_on_reals_mode */
extern T1 T65f276(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_minimum_number_call */
extern void T65f965(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_maximum_number_call */
extern void T65f964(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_less_equal_call */
extern void T65f963(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_less_call */
extern void T65f962(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_greater_equal_call */
extern void T65f961(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_greater_call */
extern void T65f960(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_equal_call */
extern void T65f959(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_identity_call */
extern void T65f958(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_floor_real_64_call */
extern void T65f957(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_floor_real_32_call */
extern void T65f956(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ceiling_real_64_call */
extern void T65f955(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ceiling_real_32_call */
extern void T65f954(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_query_call */
extern void T65f740(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_item_call */
extern void T65f953(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_query_call */
extern void T65f739(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_to_integer_32_call */
extern void T65f952(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_plus_call */
extern void T65f951(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_is_default_pointer_call */
extern void T65f950(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.current_universe_impl */
extern T0* T65f280(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.universe */
extern T0* T73f82(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_hash_code_call */
extern void T65f949(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_query_call */
extern void T65f738(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_wide_character_bytes_call */
extern void T65f948(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_real_bytes_call */
extern void T65f947(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_pointer_bytes_call */
extern void T65f946(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_windows_call */
extern void T65f945(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_vxworks_call */
extern void T65f944(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_vms_call */
extern void T65f943(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_unix_call */
extern void T65f942(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_thread_capable_call */
extern void T65f941(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_scoop_capable_call */
extern void T65f940(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_mac_call */
extern void T65f939(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_dotnet_call */
extern void T65f938(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_64_bits_call */
extern void T65f937(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_integer_bytes_call */
extern void T65f936(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_double_bytes_call */
extern void T65f935(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_character_bytes_call */
extern void T65f934(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_boolean_bytes_call */
extern void T65f933(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_query_call */
extern void T65f737(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_find_referers_call */
extern void T65f932(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_query_call */
extern void T65f736(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_type_id_from_name_call */
extern void T65f931(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_type_conforms_to_call */
extern void T65f930(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_storable_version_of_type_call */
extern void T65f929(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_at_offset_call */
extern void T65f928(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_at_call */
extern void T65f927(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_call */
extern void T65f926(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_64_field_at_call */
extern void T65f925(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_64_field_call */
extern void T65f924(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_32_field_at_call */
extern void T65f923(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_32_field_call */
extern void T65f922(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_raw_reference_field_at_offset_call */
extern void T65f921(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_raw_reference_field_at_call */
extern void T65f920(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pointer_field_at_call */
extern void T65f919(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pointer_field_call */
extern void T65f918(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_persistent_field_count_of_type_call */
extern void T65f917(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_once_objects_call */
extern void T65f916(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_object_size_call */
extern void T65f915(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_type_instance_of_call */
extern void T65f914(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_tuple_instance_of_call */
extern void T65f913(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_special_of_reference_instance_of_call */
extern void T65f912(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_instance_of_call */
extern void T65f911(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_64_field_at_call */
extern void T65f910(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_64_field_call */
extern void T65f909(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_32_field_at_call */
extern void T65f908(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_32_field_call */
extern void T65f907(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_16_field_at_call */
extern void T65f906(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_16_field_call */
extern void T65f905(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_8_field_at_call */
extern void T65f904(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_8_field_call */
extern void T65f903(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_tuple_type_call */
extern void T65f902(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_tuple_call */
extern void T65f901(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_type_call */
extern void T65f900(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_or_basic_type_call */
extern void T65f899(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_call */
extern void T65f898(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_expanded_call */
extern void T65f897(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_copy_semantics_item_call */
extern void T65f896(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_call */
extern void T65f895(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_object_marked_call */
extern void T65f894(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_field_transient_of_type_call */
extern void T65f893(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_field_expanded_of_type_call */
extern void T65f892(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_expanded_call */
extern void T65f891(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_copy_semantics_field_call */
extern void T65f890(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_attached_type_call */
extern void T65f889(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_64_field_at_call */
extern void T65f888(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_64_field_call */
extern void T65f887(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_32_field_at_call */
extern void T65f886(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_32_field_call */
extern void T65f885(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_16_field_at_call */
extern void T65f884(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_16_field_call */
extern void T65f883(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_8_field_at_call */
extern void T65f882(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_8_field_call */
extern void T65f881(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generic_parameter_count_call */
extern void T65f880(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generator_of_type_call */
extern void T65f879(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generating_type_of_type_call */
extern void T65f878(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_type_of_type_call */
extern void T65f877(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_static_type_of_type_call */
extern void T65f876(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_offset_of_type_call */
extern void T65f875(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_name_of_type_call */
extern void T65f874(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_count_of_type_call */
extern void T65f873(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_eif_gen_param_id_call */
extern void T65f872(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_dynamic_type_at_offset_call */
extern void T65f871(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_dynamic_type_call */
extern void T65f870(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_detachable_type_call */
extern void T65f869(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_compiler_version_call */
extern void T65f868(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_check_assert_call */
extern void T65f867(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_32_field_at_call */
extern void T65f866(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_32_field_call */
extern void T65f865(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_8_field_at_call */
extern void T65f864(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_8_field_call */
extern void T65f863(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_boolean_field_at_call */
extern void T65f862(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_boolean_field_call */
extern void T65f861(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_attached_type_call */
extern void T65f860(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_query_call */
extern void T65f735(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_item_call */
extern void T65f859(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_query_call */
extern void T65f734(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_64_call */
extern void T65f858(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_32_call */
extern void T65f857(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_call */
extern void T65f856(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_double_call */
extern void T65f855(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_character_32_call */
extern void T65f854(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_character_8_call */
extern void T65f853(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_quotient_call */
extern void T65f852(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_product_call */
extern void T65f851(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_power_call */
extern void T65f850(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_plus_call */
extern void T65f849(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_opposite_call */
extern void T65f848(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_minus_call */
extern void T65f847(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_is_less_call */
extern void T65f846(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_integer_remainder_call */
extern void T65f845(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_integer_quotient_call */
extern void T65f844(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_identity_call */
extern void T65f843(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_xor_call */
extern void T65f842(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_shift_right_call */
extern void T65f841(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_shift_left_call */
extern void T65f840(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_or_call */
extern void T65f839(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_not_call */
extern void T65f838(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_and_call */
extern void T65f837(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_64_call */
extern void T65f836(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_32_call */
extern void T65f835(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_16_call */
extern void T65f834(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_8_call */
extern void T65f833(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_64_call */
extern void T65f832(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_32_call */
extern void T65f831(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_16_call */
extern void T65f830(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_8_call */
extern void T65f829(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_query_call */
extern void T65f733(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_object_id_call */
extern void T65f828(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_id_object_call */
extern void T65f827(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_function_query_call */
extern void T65f732(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_function_item_call */
extern void T65f826(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_routine_call_call */
extern void T65f1001(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f1001ot1(T0* a1, T0** a2);
extern T1 T65f1001ot2(T0* a1, T0** a2);
extern T1 T65f1001ot3(T0* a1, T0** a2);
extern T1 T65f1001ot4(T0* a1, T0** a2);
extern T1 T65f1001ot5(T0* a1, T0** a2);
extern T1 T65f1001ot6(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.wipe_out */
extern void T260f20(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T412f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.mark_call_operands_unfrozen */
extern void T65f1004(GE_context* ac, T0* C);
extern T1 T65f1004ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.make */
extern T0* T302c12(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.make */
extern T0* T261c18(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_type */
extern void T261f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_target */
extern void T278f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.new_agent_tuple_item_expression */
extern T0* T65f453(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].replace */
extern void T280f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.make */
extern T0* T278c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.make_qualified_call */
extern void T278f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.set_tuple_label */
extern void T180f118(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].item */
extern T0* T280f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].put_last */
extern void T280f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].force */
extern void T647f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].fill_with */
extern void T646f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].extend */
extern void T646f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].resize */
extern void T280f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T647f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T646f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_types */
extern void T261f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.make_empty */
extern T0* T261c13(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.mark_call_operands_frozen */
extern void T65f1003(GE_context* ac, T0* C);
extern T1 T65f1003ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_frozen */
extern void T65f795(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.reset_with_types */
extern void T261f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.result_type_set */
extern T0* T443f40(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.is_ise */
extern T1 T62f42(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.is_ise */
extern T1 T371f44(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_procedure */
extern T1 T298f24(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_procedure */
extern T1 T232f22(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_function_fast_item_call */
extern void T65f825(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_query_call */
extern void T65f731(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_cwin_error_text_call */
extern void T65f824(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_character_size_call */
extern void T65f823(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_facility_call */
extern void T65f822(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_code_call */
extern void T65f821(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_call */
extern void T65f820(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_c_strlen_call */
extern void T65f819(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_query_call */
extern void T65f730(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_item_call */
extern void T65f818(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_query_call */
extern void T65f729(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_to_character_32_call */
extern void T65f817(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_to_character_8_call */
extern void T65f816(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_natural_32_code_call */
extern void T65f815(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_code_call */
extern void T65f814(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_query_call */
extern void T65f728(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_item_call */
extern void T65f813(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_query_call */
extern void T65f727(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_xor_call */
extern void T65f812(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_or_else_call */
extern void T65f811(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_or_call */
extern void T65f810(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_not_call */
extern void T65f809(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_implies_call */
extern void T65f808(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_and_then_call */
extern void T65f807(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_and_call */
extern void T65f806(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_query_call */
extern void T65f726(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_i_th_argument_string_call */
extern void T65f805(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_i_th_argument_pointer_call */
extern void T65f804(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_argument_count_call */
extern void T65f803(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_query_call */
extern void T65f725(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_call */
extern void T65f802(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f802ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_same_type_call */
extern void T65f801(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_is_equal_call */
extern void T65f800(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_generator_call */
extern void T65f799(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_generating_type_call */
extern void T65f798(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_deep_twin_call */
extern void T65f797(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin */
extern T1 T298f21(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin */
extern T1 T232f17(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_attribute_access */
extern void T65f703(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_access */
extern void T65f724(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_once_value_name */
extern void T65f621(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].force_last */
extern void T285f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_put */
extern void T285f58(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].force */
extern void T655f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE].fill_with */
extern void T654f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE].extend */
extern void T654f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_put */
extern void T285f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_put */
extern void T285f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_item */
extern T6 T285f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].hash_position */
extern T6 T285f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].resize */
extern void T285f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_resize */
extern void T285f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].special_integer_ */
extern T0* T285f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_resize */
extern void T285f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].aliased_resized_area */
extern T0* T655f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FEATURE].aliased_resized_area */
extern T0* T654f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_resize */
extern void T285f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].aliased_resized_area */
extern T0* T659f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].aliased_resized_area */
extern T0* T657f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_item */
extern T0* T285f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_item */
extern T6 T285f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_resize */
extern void T285f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_modulus */
extern T6 T285f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_capacity */
extern T6 T285f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_put */
extern void T285f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].force */
extern void T659f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].fill_with */
extern void T657f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].extend */
extern void T657f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].search_position */
extern void T285f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_adapted_expression */
extern void T65f684(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_typed_default_entity_value */
extern void T65f722(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_default_name */
extern void T65f604(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_target_expression */
extern void T65f671(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].put_last */
extern void T263f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].force */
extern void T623f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].fill_with */
extern void T622f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].extend */
extern void T622f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.mark_expressions_frozen */
extern void T65f721(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.put_first */
extern void T281f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].force */
extern void T645f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].fill_with */
extern void T643f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].extend */
extern void T643f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.resize */
extern void T281f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T645f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T643f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.capacity */
extern T6 T281f11(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.expression */
extern T0* T281f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.item */
extern T0* T281f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_target_operand */
extern void T65f690(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_used */
extern void T65f720(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].first */
extern T0* T263f10(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.has_item_cursor */
extern T1 T986f29(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_is */
extern T1 T409f39(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.has_item_cursor */
extern T1 T989f17(GE_context* ac, T0* C);
/* detachable ET_ACROSS_COMPONENT_LIST.across_component */
extern T0* T361f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_temporary_variable */
extern void T65f694(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_temporary */
extern T1 T180f12(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_formal_argument */
extern void T65f693(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_argument_name */
extern void T65f652(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].force */
extern void T262f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].resize */
extern void T262f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].aliased_resized_area */
extern T0* T623f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].aliased_resized_area */
extern T0* T622f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].new_capacity */
extern T6 T262f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].extendible */
extern T1 T262f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.fill_call_operands */
extern void T65f679(GE_context* ac, T0* C, T6 a1);
extern T1 T65f679ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].prune */
extern void T262f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_free */
extern void T65f608(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].i_th */
extern T0* T262f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].resize */
extern void T263f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_operand */
extern void T65f678(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CREATE_EXPRESSION.arguments */
extern T0* T320f14(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.name */
extern T0* T320f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.dynamic_type_set */
extern T0* T65f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.item */
extern T0* T265f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.valid_index */
extern T1 T265f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.valid_index */
extern T1 T255f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.force_last */
extern void T255f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.resize */
extern void T255f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].aliased_resized_area */
extern T0* T615f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].aliased_resized_area */
extern T0* T614f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.new_capacity */
extern T6 T255f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.capacity */
extern T6 T255f6(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.make */
extern T0* T320c28(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.create_keyword */
extern T0* T72f215(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_create */
extern T0* T409c50(GE_context* ac);
/* [detachable] ET_CREATE_EXPRESSION.tokens */
extern T0* T320f8(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.make */
extern T0* T321c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_show_console_call */
extern void T65f631(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].item */
extern T6 T284f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_routine_name */
extern void T65f624(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T65f624ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_escaped_string */
extern void T65f625(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_INTEGER_FORMATTER.put_octal_integer */
extern void T349f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_C_GENERATOR.integer_formatter_ */
extern T0* T65f263(GE_context* ac, T0* C);
/* [detachable] UT_INTEGER_FORMATTER.default_create */
extern T0* T349c1(GE_context* ac);
/* [detachable] ET_C_GENERATOR.print_temp_name */
extern void T65f602(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.temp_variable */
extern T0* T65f148(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_type_declaration */
extern void T65f598(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_indentation */
extern void T65f600(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.indent */
extern void T65f599(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_types_array */
extern void T65f558(GE_context* ac, T0* C);
extern T1 T65f558ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SYSTEM.is_new_instance_type */
extern T1 T61f60(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].has */
extern T1 T69f25(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].search_position */
extern void T69f47(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].hash_position */
extern T6 T69f32(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_item */
extern T0* T69f28(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_equality_tester */
extern T0* T69f26(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_item */
extern T6 T69f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.replace_substring_all */
extern void T17f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] SPECIAL [CHARACTER_8].resized_area_with_default */
extern T0* T15f8(GE_context* ac, T0* C, T2 a1, T6 a2);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].area */
extern T0* T756f3(GE_context* ac, T0* C);
/* [detachable] STRING_8_SEARCHER.substring_index_list_with_deltas */
extern T0* T755f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].extend */
extern void T756f9(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].additional_space */
extern T6 T756f5(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].capacity */
extern T6 T756f6(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].count */
extern T6 T756f4(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].make */
extern T0* T756c8(GE_context* ac, T6 a1);
/* [detachable] STRING_8_SEARCHER.initialize_deltas */
extern void T755f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_deferred */
extern T1 T73f188(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_deferred_mark */
extern T1 T73f79(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_deferred */
extern T1 T409f31(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.item_feature_name */
extern T0* T72f26(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T75f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T75f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T413f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T413f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T75f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T314f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort_with_comparator */
extern void T314f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].subsort_with_comparator */
extern void T314f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].swap */
extern void T75f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].replace */
extern void T75f21(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID.less_than */
extern T1 T316f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T75f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].append_last */
extern void T75f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].extend_last */
extern void T75f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T328f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_forth */
extern void T266f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].add_traversing_cursor */
extern void T266f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T328f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_position */
extern void T328f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].remove_traversing_cursor */
extern void T266f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T328f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T328f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_item */
extern T0* T266f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T328f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T328f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_start */
extern void T266f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_off */
extern T1 T266f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].off */
extern T1 T328f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T266f21(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T314c2(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.dynamic_primary_type_comparator */
extern T0* T65f156(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID.make */
extern T0* T316c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T75c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T75f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T75f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T413c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T0* T287f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T287f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T287f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T287f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T287f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T287f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T287f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T287f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T287f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_dynamic_type_id_set_constants */
extern void T65f557(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].forth */
extern void T212f54(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.occurrences */
extern T6 T743f27(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.code_occurrences */
extern T6 T743f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.occurrences */
extern T6 T17f34(GE_context* ac, T0* C, T2 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_for_iteration */
extern T0* T212f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_for_iteration */
extern T0* T212f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].after */
extern T1 T212f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_after */
extern T1 T212f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].start */
extern void T212f53(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_init_const_function */
extern void T65f556(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].forth */
extern void T286f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_forth */
extern void T286f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].add_traversing_cursor */
extern void T286f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_position */
extern void T660f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].remove_traversing_cursor */
extern void T286f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_item */
extern T6 T286f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_for_iteration */
extern T0* T286f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_item */
extern T0* T286f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_item */
extern T0* T286f36(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.constant */
extern T0* T1502f12(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_inline_constant_name */
extern void T65f622(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_for_iteration */
extern T0* T286f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_key */
extern T0* T286f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_item */
extern T0* T286f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].after */
extern T1 T286f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_after */
extern T1 T286f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].after */
extern T1 T660f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].start */
extern void T286f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_start */
extern void T286f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_off */
extern T1 T286f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].off */
extern T1 T660f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].is_empty */
extern T1 T286f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].forth */
extern void T285f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_forth */
extern void T285f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].add_traversing_cursor */
extern void T285f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_position */
extern void T656f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].remove_traversing_cursor */
extern void T285f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_for_iteration */
extern T0* T285f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_item */
extern T0* T285f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_item */
extern T0* T285f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_once_status_name */
extern void T65f623(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].has */
extern T1 T283f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].search_position */
extern void T283f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].hash_position */
extern T6 T283f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_item */
extern T0* T283f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_item */
extern T6 T283f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_item */
extern T6 T283f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_for_iteration */
extern T0* T285f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_key */
extern T0* T285f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].after */
extern T1 T285f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_after */
extern T1 T285f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].after */
extern T1 T656f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].start */
extern void T285f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_start */
extern void T285f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_off */
extern T1 T285f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].off */
extern T1 T656f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].is_empty */
extern T1 T285f28(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_constants_declaration */
extern void T65f555(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_functions */
extern void T65f554(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T267f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T267f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T267f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T267f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T267f50(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].keep_head */
extern void T627f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].set_count */
extern void T627f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T267f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T267f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T629f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T629f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T267f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_forth */
extern void T267f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].add_traversing_cursor */
extern void T267f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_position */
extern void T629f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].remove_traversing_cursor */
extern void T267f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T267f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_deep_twin_polymorphic_call_function */
extern void T65f620(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_default_entity_value */
extern void T65f657(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].wipe_out */
extern void T264f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_wipe_out */
extern void T264f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_wipe_out */
extern void T264f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T264f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T264f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].unset_found_item */
extern void T264f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].move_all_cursors_after */
extern void T264f45(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_next_cursor */
extern void T626f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_after */
extern void T626f7(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_binary_search_polymorphic_call */
extern void T65f677(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T0* a4, T0* a5);
/* [detachable] ET_C_GENERATOR.print_adapted_deep_twin_call */
extern void T65f676(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_deep_twin_call */
extern T0* T65f676ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_deep_twin_call */
extern void T65f692(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item */
extern T0* T264f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_item */
extern T0* T264f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].search_position */
extern void T264f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].hash_position */
extern T6 T264f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_item */
extern T6 T264f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_item */
extern T6 T264f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_item */
extern T6 T264f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].force_last */
extern void T264f36(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_put */
extern void T264f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_put */
extern void T264f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_put */
extern void T264f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].resize */
extern void T264f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_resize */
extern void T264f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].special_integer_ */
extern T0* T264f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_resize */
extern void T264f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_resize */
extern void T264f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_resize */
extern void T264f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_modulus */
extern T6 T264f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_capacity */
extern T6 T264f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_put */
extern void T264f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_C_GENERATOR.print_current_name */
extern void T65f651(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_for_iteration */
extern T0* T267f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_item */
extern T0* T267f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T267f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T267f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_after */
extern T1 T267f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T629f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T267f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_start */
extern void T267f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_off */
extern T1 T267f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].off */
extern T1 T629f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T267f22(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern void T65f619(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac1(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #2 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #3 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac3(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #4 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac4(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #5 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac5(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #6 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #7 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac7(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #8 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac8(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #9 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac9(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #10 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac10(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #11 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac11(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #12 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T65f619ac12(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T65f619ot1(T0* a1, T0** a2);
extern T1 T65f619ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_attribute_special_indexed_item_access */
extern void T65f674(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.reset_temp_variables */
extern void T65f610(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T258f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T617f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].set_count */
extern void T617f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T258f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T619f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T619f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_set_deep_twined_attribute */
extern void T65f675(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_deep_twined_attribute */
extern void T65f691(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].force_last_new */
extern void T267f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T267f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T267f58(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].force */
extern void T628f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].fill_with */
extern void T627f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].extend */
extern void T627f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T267f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T267f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T267f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T267f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T267f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T267f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T267f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T267f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T267f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].aliased_resized_area */
extern T0* T628f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].aliased_resized_area */
extern T0* T627f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T267f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T267f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T267f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T267f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.force_last */
extern void T265f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.fixed_array */
extern T0* T265f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].default_create */
extern T0* T628c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.resize */
extern void T265f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.new_capacity */
extern T6 T265f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.capacity */
extern T6 T265f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.reset */
extern void T261f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T0* T267f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T267f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T267f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T267f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.result_type_set_in_feature */
extern T0* T65f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.result_keyword */
extern T0* T72f25(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.make */
extern T0* T301c31(GE_context* ac);
/* [detachable] ET_RESULT.make_token */
extern void T301f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RESULT.make_leaf */
extern void T301f34(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.tokens */
extern T0* T301f5(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_result_name */
extern void T65f601(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].wipe_out */
extern void T282f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_wipe_out */
extern void T282f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_wipe_out */
extern void T282f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_wipe_out */
extern void T282f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_wipe_out */
extern void T282f48(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].keep_head */
extern void T650f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].set_count */
extern void T650f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].unset_found_item */
extern void T282f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].move_all_cursors_after */
extern void T282f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_next_cursor */
extern void T649f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_after */
extern void T649f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].forth */
extern void T282f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_forth */
extern void T282f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].add_traversing_cursor */
extern void T282f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_position */
extern void T649f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].remove_traversing_cursor */
extern void T282f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_item */
extern T6 T282f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_manifest_tuple_function */
extern void T65f553(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_for_iteration */
extern T0* T282f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_item */
extern T0* T282f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_item */
extern T0* T282f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].after */
extern T1 T282f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_after */
extern T1 T282f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].after */
extern T1 T649f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].start */
extern void T282f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_start */
extern void T282f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_off */
extern T1 T282f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].off */
extern T1 T649f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].is_empty */
extern T1 T282f21(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_big_manifest_array_function */
extern void T65f552(GE_context* ac, T0* C, T0* a1);
extern T1 T65f552ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T266f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T266f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T266f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T266f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T266f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T266f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T266f39(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_manifest_array_function */
extern void T65f551(GE_context* ac, T0* C, T0* a1);
extern T1 T65f551ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_for_iteration */
extern T0* T266f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T266f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_after */
extern T1 T266f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T266f38(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_polymorphic_tuple_label_call_functions */
extern void T65f550(GE_context* ac, T0* C);
extern T1 T65f550ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_polymorphic_call_function */
extern void T65f611(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T65f611ot1(T0* a1, T0** a2);
extern T1 T65f611ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_binary_search_polymorphic_calls */
extern void T65f658(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_C_GENERATOR.print_named_procedure_call */
extern void T65f656(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_procedure_call */
extern void T65f687(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_non_inlined_procedure_call */
extern void T65f699(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_call */
extern void T65f698(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_procedure_call */
extern void T65f719(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_set_object_comparison_call */
extern void T65f794(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_procedure_call */
extern void T65f718(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_put_call */
extern void T65f793(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_procedure_call */
extern void T65f717(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_set_item_call */
extern void T65f792(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_procedure_call */
extern void T65f716(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_fast_call_call */
extern void T65f791(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_call_call */
extern void T65f790(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_procedure_call */
extern void T65f715(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_set_item_call */
extern void T65f789(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_procedure_call */
extern void T65f714(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_free_call */
extern void T65f788(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_c_call */
extern void T65f1000(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3, T0* a4, T1 a5);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_procedure_call */
extern void T65f713(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_unmark_object_call */
extern void T65f787(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_unlock_marking_call */
extern void T65f786(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_reference_field_at_call */
extern void T65f785(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_reference_field_call */
extern void T65f784(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_64_field_at_call */
extern void T65f783(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_64_field_call */
extern void T65f782(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_32_field_at_call */
extern void T65f781(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_32_field_call */
extern void T65f780(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pointer_field_at_call */
extern void T65f779(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pointer_field_call */
extern void T65f778(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_64_field_at_call */
extern void T65f777(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_64_field_call */
extern void T65f776(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_32_field_at_call */
extern void T65f775(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_32_field_call */
extern void T65f774(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_16_field_at_call */
extern void T65f773(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_16_field_call */
extern void T65f772(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_8_field_at_call */
extern void T65f771(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_8_field_call */
extern void T65f770(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_64_field_at_call */
extern void T65f769(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_64_field_call */
extern void T65f768(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_32_field_at_call */
extern void T65f767(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_32_field_call */
extern void T65f766(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_16_field_at_call */
extern void T65f765(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_16_field_call */
extern void T65f764(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_8_field_at_call */
extern void T65f763(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_8_field_call */
extern void T65f762(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_32_field_at_call */
extern void T65f761(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_32_field_call */
extern void T65f760(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_8_field_at_call */
extern void T65f759(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_8_field_call */
extern void T65f758(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_boolean_field_at_call */
extern void T65f757(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_boolean_field_call */
extern void T65f756(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_mark_object_call */
extern void T65f755(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_lock_marking_call */
extern void T65f754(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_exception_manager_procedure_call */
extern void T65f712(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_exception_manager_developer_raise_call */
extern void T65f753(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_procedure_call */
extern void T65f711(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_set_item_call */
extern void T65f752(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_procedure_call */
extern void T65f710(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_object_id_free_call */
extern void T65f751(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_procedure_call */
extern void T65f709(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_cwin_local_free_call */
extern void T65f750(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_procedure_call */
extern void T65f708(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_set_item_call */
extern void T65f749(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_procedure_call */
extern void T65f707(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_set_item_call */
extern void T65f748(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_procedure_call */
extern void T65f706(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_call */
extern void T65f747(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f747ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_info_message_call */
extern void T65f999(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_copy_call */
extern void T65f746(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_tuple_label_setter_call */
extern void T65f686(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T65f686ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.set_polymorphic_call_argument_type_sets */
extern void T65f654(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T65f654ot1(T0* a1, T0** a2);
extern T1 T65f654ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.set_static_type */
extern void T261f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.put_last */
extern void T265f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.is_call_agent */
extern T1 T306f6(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.is_call_agent */
extern T1 T302f8(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.is_tuple_label */
extern T1 T306f8(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_call_name_comment */
extern void T65f649(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T65f649ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].force_last_new */
extern void T264f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].has */
extern T1 T264f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].replace */
extern void T263f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_C_GENERATOR.fill_call_formal_arguments */
extern void T65f648(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.static_feature */
extern T0* T306f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.seeded_procedure */
extern T0* T73f180(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.seeded_feature */
extern T0* T254f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.static_feature */
extern T0* T302f9(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.formal_argument */
extern T0* T65f232(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].put_last */
extern void T257f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.set_argument */
extern void T180f117(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].force_last */
extern void T263f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].new_capacity */
extern T6 T263f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].extendible */
extern T1 T263f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.set_next */
extern void T302f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.same_declared_signature */
extern T1 T65f146(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T65f146ot1(T0* a1);
extern T1 T65f146ot2(T0* a1);
extern T1 T65f146ot3(T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.is_tuple_label */
extern T1 T302f7(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_feature */
extern void T65f549(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].wipe_out */
extern void T270f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].keep_head */
extern void T637f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].set_count */
extern void T637f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].move_all_cursors_after */
extern void T270f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].set_next_cursor */
extern void T639f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].set_after */
extern void T639f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_equality_function */
extern void T65f616(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_BRACKET_SYMBOL.is_not_equal */
extern T1 T610f43(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.tokens */
extern T0* T610f8(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_not_equal */
extern T1 T343f36(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.tokens */
extern T0* T343f3(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_equality_expression */
extern void T65f673(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].force_last */
extern void T270f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].force */
extern void T638f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].fill_with */
extern void T637f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].extend */
extern void T637f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].resize */
extern void T270f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T638f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T637f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].new_capacity */
extern T6 T270f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].extendible */
extern T1 T270f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_equality_function_name */
extern void T65f672(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.print_static_routine_name */
extern void T65f688(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T65f688ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].item */
extern T0* T270f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].wipe_out */
extern void T269f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].keep_head */
extern void T634f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_count */
extern void T634f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].move_all_cursors_after */
extern void T269f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_next_cursor */
extern void T636f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_after */
extern void T636f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_equality_function */
extern void T65f615(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_BRACKET_SYMBOL.is_not_tilde */
extern T1 T610f42(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_not_tilde */
extern T1 T343f35(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_equality_expression */
extern void T65f669(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].force_last */
extern void T269f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].force */
extern void T635f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].fill_with */
extern void T634f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].extend */
extern void T634f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].resize */
extern void T269f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T635f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T634f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].new_capacity */
extern T6 T269f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].extendible */
extern T1 T269f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_object_equality_function_name */
extern void T65f668(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].item */
extern T0* T269f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].wipe_out */
extern void T268f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].keep_head */
extern void T631f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].set_count */
extern void T631f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].move_all_cursors_after */
extern void T268f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].set_next_cursor */
extern void T633f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].set_after */
extern void T633f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_test_function */
extern void T65f614(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.is_empty */
extern T1 T260f9(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_test_function_name */
extern void T65f666(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].item */
extern T0* T268f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].wipe_out */
extern void T271f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_AGENT].keep_head */
extern void T640f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT].set_count */
extern void T640f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].move_all_cursors_after */
extern void T271f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].set_next_cursor */
extern void T642f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].set_after */
extern void T642f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_declaration */
extern void T65f613(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T65f613ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_agent_closed_operand_access */
extern void T65f664(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_agent_creation_name */
extern void T65f663(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.print_agent_body_declaration */
extern void T65f662(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.put_first */
extern void T274f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.fixed_array */
extern T0* T274f9(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.resize */
extern void T274f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.capacity */
extern T6 T274f8(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.wipe_out */
extern void T274f14(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_open_operand_access */
extern void T65f661(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_function_name */
extern void T65f660(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.make */
extern T0* T248c116(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.tuple_keyword */
extern T0* T72f89(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.tokens */
extern T0* T248f23(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.put_first */
extern void T236f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].force */
extern void T612f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T609f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].extend */
extern void T609f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.fixed_array */
extern T0* T236f11(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T612c3(GE_context* ac);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern T0* T236c26(GE_context* ac, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern void T236f26p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].make */
extern T0* T612f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T609c5(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_bracket_symbol */
extern T0* T72f208(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_bracket */
extern T0* T343c78(GE_context* ac);
/* [detachable] ET_SYMBOL.make_leaf */
extern void T343f81(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.left_bracket_symbol */
extern T0* T72f207(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.make */
extern T0* T610c108(GE_context* ac);
/* [detachable] ET_BRACKET_SYMBOL.make_leaf */
extern void T610f109(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.tokens */
extern T0* T236f9(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_agent_closed_operand */
extern void T180f113(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IDENTIFIER.set_agent_open_operand */
extern void T180f110(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].item */
extern T0* T271f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.is_semistrict */
extern T1 T298f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.is_semistrict */
extern T1 T232f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].remove_last */
extern void T288f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].keep_head */
extern void T595f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].set_count */
extern void T595f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].move_last_cursors_after */
extern void T288f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].set_next_cursor */
extern void T668f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].set_after */
extern void T668f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].last */
extern T0* T288f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].is_empty */
extern T1 T288f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].force_last */
extern void T288f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].force */
extern void T596f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].fill_with */
extern void T595f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].extend */
extern void T595f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].resize */
extern void T288f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T596f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T595f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].new_capacity */
extern T6 T288f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].extendible */
extern T1 T288f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_generated */
extern void T298f38(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_generated */
extern void T232f32(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_object_allocation_functions */
extern void T65f548(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_allocation_function */
extern void T65f612(GE_context* ac, T0* C, T0* a1);
extern T1 T65f612ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_dispose_registration */
extern void T65f659(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T65f659ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].force_last_new */
extern void T287f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T287f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T287f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].force */
extern void T667f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].fill_with */
extern void T664f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].extend */
extern void T664f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T287f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T287f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T287f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T287f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T287f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T287f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T287f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T667f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T664f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T287f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T287f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T287f32(GE_context* ac, T0* C, T6 a1);
/* detachable ET_COMPOUND.is_empty */
extern T1 T330f4(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_polymorphic_procedure_call_functions */
extern void T65f547(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.set_next */
extern void T306f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_polymorphic_query_call_functions */
extern void T65f546(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_immutable_string_32_function */
extern void T65f545(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_immutable_string_8_function */
extern void T65f544(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_string_32_function */
extern void T65f543(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_string_8_function */
extern void T65f542(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_default_declarations */
extern void T65f541(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_default_declaration */
extern void T65f597(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_default_object_value */
extern void T65f643(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T65f643ot1(T0* a1, T0** a2);
extern T1 T65f643ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_default_attribute_value */
extern void T65f683(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.flush_to_c_file */
extern void T65f540(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_append */
extern void T293f35(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_open_append */
extern void T293f44(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_append */
extern void T293f44p1(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_types */
extern void T65f539(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_boxed_type_struct */
extern void T65f596(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_attribute_flags_name */
extern void T65f640(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_struct_name */
extern void T65f638(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_type_struct */
extern void T65f595(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T65f595ot1(T0* a1, T0** a2);
extern T1 T65f595ot2(T0* a1, T0** a2);
extern T1 T65f595ot3(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print */
extern void T65f641(GE_context* ac, T0* C, T0* a1);
/* [detachable] STD_FILES.put_string */
extern void T352f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] STD_FILES.standard_default */
extern T0* T352f2(GE_context* ac, T0* C);
/* [detachable] STRING_8.out */
extern T0* T17f16(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.io */
extern T0* T65f269(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_struct_name */
extern void T65f634(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].has_cycle */
extern T1 T300f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T300f15(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] BOOLEAN].item */
extern T1 T677f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] BOOLEAN].put */
extern void T677f6(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] ARRAY [[attached] BOOLEAN].make_filled */
extern T0* T677c5(GE_context* ac, T1 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] BOOLEAN].make_filled_area */
extern void T677f7(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_last */
extern void T75f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].replace */
extern void T676f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].item */
extern T0* T676f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].cloned_object */
extern T0* T676f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].twin */
extern T0* T676f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].copy */
extern void T676f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].resized_area */
extern T0* T1360f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].copy_data */
extern void T1360f10(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].force */
extern void T1360f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].extend */
extern void T1360f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].move_data */
extern void T1360f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].overlapping_move */
extern void T1360f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].fill_with */
extern void T1360f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].non_overlapping_move */
extern void T1360f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].make_empty */
extern T0* T1360c6(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].new_cursor */
extern T0* T676f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T1362c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].set_internal_cursor */
extern void T676f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].valid_cursor */
extern T1 T676f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].move_all_cursors_after */
extern void T676f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].set_next_cursor */
extern void T1362f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].set_after */
extern void T1362f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].cloned_object */
extern T0* T259f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].twin */
extern T0* T259f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].copy */
extern void T259f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].valid_cursor */
extern T1 T259f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].reset */
extern void T300f20(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_boxed_type_definition */
extern void T65f594(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_type_type_definition */
extern void T65f593(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_function */
extern void T65f592(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_type_id_access */
extern void T65f636(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_boxed_type_declaration */
extern void T65f635(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_relation */
extern void T300f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_indexed_relation */
extern void T300f19(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_LINKABLE [[attached] INTEGER_32].put_right */
extern void T493f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] INTEGER_32].make */
extern T0* T493c3(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].count */
extern T6 T300f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].index_of */
extern T6 T300f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T6 T675f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T6 T675f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T675f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T675f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T675f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T675f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T675f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T675f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T675f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_type_definition */
extern void T65f591(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_pointer_type_definition */
extern void T65f590(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_real_64_type_definition */
extern void T65f589(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_real_32_type_definition */
extern void T65f588(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_64_type_definition */
extern void T65f587(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_32_type_definition */
extern void T65f586(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_16_type_definition */
extern void T65f585(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_8_type_definition */
extern void T65f584(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_64_type_definition */
extern void T65f583(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_32_type_definition */
extern void T65f582(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_16_type_definition */
extern void T65f581(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_8_type_definition */
extern void T65f580(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_character_32_type_definition */
extern void T65f579(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_character_8_type_definition */
extern void T65f578(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boolean_type_definition */
extern void T65f577(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T300f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put */
extern void T300f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].put_last */
extern void T676f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].force */
extern void T1361f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].put_new */
extern void T675f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T675f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T675f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T675f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T675f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T675f40(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].resize */
extern void T676f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].aliased_resized_area */
extern T0* T1361f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].aliased_resized_area */
extern T0* T1360f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T675f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T675f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T675f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T675f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T675f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T675f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T675f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].capacity */
extern T6 T300f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T300f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T675f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_default */
extern T0* T300c12(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern void T300f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T300c16(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T676c11(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T1361f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].default_create */
extern T0* T1361c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T675c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T675f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T675f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T675f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T675f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T1359c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T675f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T675f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T675f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T675f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_aliased_double_type_definition */
extern void T65f576(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.upper_name */
extern T0* T180f11(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_upper */
extern T0* T743f61(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_upper */
extern void T743f104(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code */
extern T6 T423f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.upper_codes */
extern T0* T423f31(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_upper_code_plane */
extern T0* T423f38(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_upper_code_segment */
extern T0* T423f45(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_1 */
extern T0* T423f37(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_1_segment_4 */
extern T0* T423f56(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0 */
extern T0* T423f36(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_255 */
extern T0* T423f55(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_167 */
extern T0* T423f54(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_166 */
extern T0* T423f53(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_45 */
extern T0* T423f52(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_44 */
extern T0* T423f51(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_36 */
extern T0* T423f50(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_33 */
extern T0* T423f49(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_31 */
extern T0* T423f48(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_30 */
extern T0* T423f47(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_29 */
extern T0* T423f46(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_5 */
extern T0* T423f44(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_4 */
extern T0* T423f43(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_3 */
extern T0* T423f42(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_2 */
extern T0* T423f41(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_1 */
extern T0* T423f40(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_0 */
extern T0* T423f39(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.unicode */
extern T0* T743f57(GE_context* ac, T0* C);
/* [detachable] STRING_8.as_upper */
extern T0* T17f32(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_upper */
extern void T17f72(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_upper_area */
extern void T17f75(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_TOKEN_CONSTANTS.double_class_name */
extern T0* T72f32(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_aliased_real_type_definition */
extern void T65f575(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.real_class_name */
extern T0* T72f31(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_aliased_natural_type_definition */
extern void T65f574(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.natural_class_name */
extern T0* T72f30(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_aliased_integer_type_definition */
extern void T65f573(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.integer_class_name */
extern T0* T72f29(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_aliased_wide_character_type_definition */
extern void T65f572(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.wide_character_class_name */
extern T0* T72f28(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_aliased_character_type_definition */
extern void T65f571(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.character_class_name */
extern T0* T72f27(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_start_extern_c */
extern void T65f538(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_runtime_header_file */
extern void T65f537(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_C_GENERATOR.include_runtime_c_file */
extern void T65f569(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_SET [[attached] STRING_8].force_last */
extern void T69f44(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_put */
extern void T69f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_put */
extern void T69f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].resize */
extern void T69f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_resize */
extern void T69f69(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].special_integer_ */
extern T0* T69f35(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_resize */
extern void T69f68(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_resize */
extern void T69f67(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_resize */
extern void T69f66(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_modulus */
extern T6 T69f23(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_capacity */
extern T6 T69f31(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_put */
extern void T69f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].force_last */
extern void T289f48(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_put */
extern void T289f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_put */
extern void T289f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_put */
extern void T289f61(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_item */
extern T6 T289f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].hash_position */
extern T6 T289f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].resize */
extern void T289f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_resize */
extern void T289f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].special_integer_ */
extern T0* T289f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_resize */
extern void T289f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_resize */
extern void T289f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_resize */
extern void T289f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_modulus */
extern T6 T289f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_capacity */
extern T6 T289f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_put */
extern void T289f59(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].search_position */
extern void T289f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].replace */
extern void T289f47(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item */
extern T1 T289f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].has */
extern T1 T289f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.exception_trace_mode */
extern T1 T65f192(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.trace_mode */
extern T1 T65f115(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_ids */
extern void T65f536(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_id */
extern void T298f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_FEATURE.set_id */
extern void T232f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.sort_types */
extern void T65f568(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_relation */
extern void T300f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_iterator */
extern T0* T266f34(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.open_cpp_file */
extern void T65f535(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.open_c_file */
extern void T65f534(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_split_threshold */
extern void T65f526(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.set_split_mode */
extern void T65f525(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.no_split */
extern T1 T25f47(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.parameter */
extern T1 T46f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].last */
extern T1 T153f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].is_empty */
extern T1 T153f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_finalize_mode */
extern void T65f524(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_finalize */
extern T1 T25f44(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_use_boehm_gc */
extern void T65f523(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.use_boehm_gc */
extern T1 T25f46(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.make */
extern T0* T65c522(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.make_external_regexps */
extern void T65f529(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.make_rescue_data */
extern void T65f528(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].set_equality_tester */
extern void T290f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.identifier_tester */
extern T0* T65f112(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TESTER.default_create */
extern T0* T291c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make */
extern T0* T290c31(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].set_internal_cursor */
extern void T290f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_cursor */
extern T0* T290f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].make */
extern T0* T672c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_slots */
extern void T290f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_clashes */
extern void T290f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_key_storage */
extern void T290f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_item_storage */
extern void T290f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].make */
extern T0* T589f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].make_empty */
extern T0* T590c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].default_create */
extern T0* T589c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].set_key_equality_tester */
extern void T289f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make */
extern T0* T289c42(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_with_equality_testers */
extern void T289f50(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_sparse_container */
extern void T289f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].set_internal_cursor */
extern void T289f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_cursor */
extern T0* T289f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].make */
extern T0* T669c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_slots */
extern void T289f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_clashes */
extern void T289f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_key_storage */
extern void T289f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_item_storage */
extern void T289f71(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] STRING_8].default_create */
extern T0* T150c2(GE_context* ac);
/* detachable DS_HASH_SET [[attached] STRING_8].set_equality_tester */
extern void T69f41(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make */
extern void T69f40(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make */
extern T0* T69c40(GE_context* ac, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].set_internal_cursor */
extern void T69f57(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_cursor */
extern T0* T69f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].make */
extern T0* T372c9(GE_context* ac, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_slots */
extern void T69f56(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_clashes */
extern void T69f55(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_key_storage */
extern void T69f54(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_item_storage */
extern void T69f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T288c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].set_internal_cursor */
extern void T288f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].new_cursor */
extern T0* T288f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T668c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T287c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T287f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T287f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T287f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T287f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T666c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T287f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T287f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T287f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T287f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].make */
extern T0* T667f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].make_empty */
extern T0* T664c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].default_create */
extern T0* T667c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_map */
extern T0* T286c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_with_equality_testers */
extern void T286f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_sparse_container */
extern void T286f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_internal_cursor */
extern void T286f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_cursor */
extern T0* T286f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make */
extern T0* T660c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_slots */
extern void T286f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_modulus */
extern T6 T286f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_clashes */
extern void T286f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_key_storage */
extern void T286f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].make */
extern T0* T663f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].make_empty */
extern T0* T661c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].default_create */
extern T0* T663c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_item_storage */
extern void T286f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_map */
extern T0* T285c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_with_equality_testers */
extern void T285f44(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_sparse_container */
extern void T285f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_internal_cursor */
extern void T285f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_cursor */
extern T0* T285f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].make */
extern T0* T656c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_slots */
extern void T285f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_clashes */
extern void T285f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_key_storage */
extern void T285f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].make */
extern T0* T655f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE].make_empty */
extern T0* T654c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].default_create */
extern T0* T655c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_item_storage */
extern void T285f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].make */
extern T0* T659f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].make_empty */
extern T0* T657c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].default_create */
extern T0* T659c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_map */
extern T0* T283c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_with_equality_testers */
extern void T283f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_sparse_container */
extern void T283f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].set_internal_cursor */
extern void T283f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_cursor */
extern T0* T283f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].make */
extern T0* T653c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_slots */
extern void T283f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_modulus */
extern T6 T283f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_clashes */
extern void T283f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_key_storage */
extern void T283f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_item_storage */
extern void T283f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T282c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].set_internal_cursor */
extern void T282f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_cursor */
extern T0* T282f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T649c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_slots */
extern void T282f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_modulus */
extern T6 T282f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_clashes */
extern void T282f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_key_storage */
extern void T282f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_item_storage */
extern void T282f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T651f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].make_empty */
extern T0* T650c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].default_create */
extern T0* T651c3(GE_context* ac);
/* [detachable] ET_MANIFEST_TUPLE.make_with_capacity */
extern T0* T281c28(GE_context* ac, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.make_with_capacity */
extern void T281f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].make */
extern T0* T645f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].make_empty */
extern T0* T643c5(GE_context* ac, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.tokens */
extern T0* T281f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T280c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].set_internal_cursor */
extern void T280f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_QUALIFIED_CALL_EXPRESSION].new_cursor */
extern T0* T280f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T648c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T647f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].make_empty */
extern T0* T646c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].default_create */
extern T0* T647c3(GE_context* ac);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.make */
extern T0* T279c28(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.make */
extern T0* T275c9(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.make_qualified_call */
extern void T275f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T274f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern T0* T274c13(GE_context* ac, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T274f13p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_parenthesis_symbol */
extern T0* T72f212(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_parenthesis */
extern T0* T343c80(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_parenthesis_symbol */
extern T0* T72f211(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_parenthesis */
extern T0* T343c79(GE_context* ac);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.tokens */
extern T0* T274f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].make */
extern T0* T271c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].set_internal_cursor */
extern void T271f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].new_cursor */
extern T0* T271f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].make */
extern T0* T642c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].make */
extern T0* T641f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT].make_empty */
extern T0* T640c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].default_create */
extern T0* T641c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T270c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].set_internal_cursor */
extern void T270f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].new_cursor */
extern T0* T270f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T639c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T638f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].make_empty */
extern T0* T637c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].default_create */
extern T0* T638c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T269c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_internal_cursor */
extern void T269f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].new_cursor */
extern T0* T269f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T636c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T635f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make_empty */
extern T0* T634c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].default_create */
extern T0* T635c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].make */
extern T0* T268c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].set_internal_cursor */
extern void T268f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].new_cursor */
extern T0* T268f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].make */
extern T0* T633c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].make */
extern T0* T632f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].make_empty */
extern T0* T631c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].default_create */
extern T0* T632c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T267c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T267f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T267f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T267f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T267f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T629c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T267f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T267f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T267f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T267f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make */
extern T0* T628f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make_empty */
extern T0* T627c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.make_with_capacity */
extern T0* T265c8(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_map */
extern T0* T264c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_with_equality_testers */
extern void T264f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_sparse_container */
extern void T264f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_internal_cursor */
extern void T264f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_cursor */
extern T0* T264f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make */
extern T0* T626c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_slots */
extern void T264f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_clashes */
extern void T264f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_key_storage */
extern void T264f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_item_storage */
extern void T264f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].make */
extern T0* T263c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].set_internal_cursor */
extern void T263f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].new_cursor */
extern T0* T263f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].make */
extern T0* T624c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].make */
extern T0* T623f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].make_empty */
extern T0* T622c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].default_create */
extern T0* T623c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].make */
extern T0* T262c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T258c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T258f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T258f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T619c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T618f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].make_empty */
extern T0* T617c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].default_create */
extern T0* T618c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].make */
extern T0* T257c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].set_internal_cursor */
extern void T257f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].new_cursor */
extern T0* T257f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER].make */
extern T0* T616c3(GE_context* ac, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.make */
extern T0* T256c4(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.null_output_stream */
extern T0* T65f282(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.lower_name */
extern T0* T73f160(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class */
extern T0* T72f1(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_implementation_error */
extern void T73f259(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_interface_error */
extern void T73f258(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_flattening_error */
extern void T73f257(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_ancestors_error */
extern void T73f256(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_syntax_error */
extern void T73f255(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_group */
extern void T73f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_group */
extern T0* T72f2(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.make */
extern T0* T406c31(GE_context* ac);
/* [detachable] ET_CLASS.make_unknown */
extern T0* T73c253(GE_context* ac);
/* [detachable] ET_CLASS.make */
extern void T73f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.make */
extern T0* T73c260(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_CODES.class_code */
extern T8 T410f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].found_item */
extern T8 T742f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_item */
extern T8 T742f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].found */
extern T1 T742f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].search */
extern void T742f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].search_position */
extern void T742f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].hash_position */
extern T6 T742f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_NAME_TESTER.test */
extern T1 T555f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.same_class_name */
extern T1 T180f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_item */
extern T0* T742f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_item */
extern T6 T742f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_item */
extern T6 T742f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_CODES.codes_by_name */
extern T0* T410f2(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.iteration_cursor_class_name */
extern T0* T72f122(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.iterable_class_name */
extern T0* T72f121(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disposable_class_name */
extern T0* T72f120(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.system_object_class_name */
extern T0* T72f119(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.any_class_name */
extern T0* T72f118(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.array_class_name */
extern T0* T72f117(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.typed_pointer_class_name */
extern T0* T72f116(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_class_name */
extern T0* T72f115(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.special_class_name */
extern T0* T72f114(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.routine_class_name */
extern T0* T72f113(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.procedure_class_name */
extern T0* T72f112(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.predicate_class_name */
extern T0* T72f111(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.function_class_name */
extern T0* T72f110(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.tuple_class_name */
extern T0* T72f109(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_class_name */
extern T0* T72f108(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_class_name */
extern T0* T72f107(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_class_name */
extern T0* T72f106(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_class_name */
extern T0* T72f105(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_class_name */
extern T0* T72f104(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_class_name */
extern T0* T72f103(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_class_name */
extern T0* T72f102(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_class_name */
extern T0* T72f101(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_class_name */
extern T0* T72f100(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_class_name */
extern T0* T72f99(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_class_name */
extern T0* T72f98(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_class_name */
extern T0* T72f97(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_class_name */
extern T0* T72f96(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].force_last */
extern void T742f37(GE_context* ac, T0* C, T8 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_put */
extern void T742f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].force */
extern void T602f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].fill_with */
extern void T603f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].extend */
extern void T603f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_put */
extern void T742f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_put */
extern void T742f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].resize */
extern void T742f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_resize */
extern void T742f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].special_integer_ */
extern T0* T742f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_resize */
extern void T742f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].aliased_resized_area */
extern T0* T602f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].aliased_resized_area */
extern T0* T603f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_resize */
extern void T742f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].aliased_resized_area */
extern T0* T1382f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_8].aliased_resized_area */
extern T0* T477f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_resize */
extern void T742f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_modulus */
extern T6 T742f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_capacity */
extern T6 T742f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_put */
extern void T742f41(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].force */
extern void T1382f4(GE_context* ac, T0* C, T0* a1, T8 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].unset_found_item */
extern void T742f40(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_class_name */
extern T0* T72f95(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.tokens */
extern T0* T410f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].set_key_equality_tester */
extern void T742f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_CODES.class_name_tester */
extern T0* T410f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_NAME_TESTER.default_create */
extern T0* T555c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_map */
extern T0* T742c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_with_equality_testers */
extern void T742f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_sparse_container */
extern void T742f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].set_internal_cursor */
extern void T742f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_cursor */
extern T0* T742f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make */
extern T0* T1383c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_slots */
extern void T742f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_clashes */
extern void T742f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_key_storage */
extern void T742f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].make */
extern T0* T602f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].make_empty */
extern T0* T603c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].default_create */
extern T0* T602c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_item_storage */
extern void T742f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].make */
extern T0* T1382f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].default_create */
extern T0* T1382c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.class_keyword */
extern T0* T72f87(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_class */
extern T0* T409c46(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_procedures */
extern T0* T72f86(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.make_with_capacity */
extern T0* T254c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].make */
extern T0* T889f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].make_empty */
extern T0* T888c5(GE_context* ac, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.fixed_array */
extern T0* T254f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].default_create */
extern T0* T889c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_queries */
extern T0* T72f85(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.make_with_capacity */
extern T0* T253c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].make */
extern T0* T887f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUERY].make_empty */
extern T0* T885c5(GE_context* ac, T6 a1);
/* [detachable] ET_QUERY_LIST.fixed_array */
extern T0* T253f7(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].default_create */
extern T0* T887c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_ancestors */
extern T0* T72f84(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_LIST.make_with_capacity */
extern T0* T408c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].make */
extern T0* T900f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].make_empty */
extern T0* T899c5(GE_context* ac, T6 a1);
/* [detachable] ET_BASE_TYPE_LIST.fixed_array */
extern T0* T408f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].default_create */
extern T0* T900c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class_name */
extern T0* T72f83(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.tokens */
extern T0* T73f178(GE_context* ac, T0* C);
/* [detachable] GEC.tokens */
extern T0* T25f45(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.record_end_time */
extern void T62f57(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.print_time */
extern void T62f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STDOUT_FILE.put_line */
extern void T137f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.out */
extern T0* T461f9(GE_context* ac, T0* C);
/* [detachable] DT_DATE_TIME_DURATION.append_to_string */
extern void T461f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.append_time_to_string */
extern void T461f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_INTEGER_ROUTINES.append_decimal_integer */
extern void T753f6(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DT_DATE_TIME_DURATION.integer_ */
extern T0* T461f14(GE_context* ac, T0* C);
/* [detachable] DT_DATE_TIME_DURATION.append_date_to_string */
extern void T461f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.set_canonical */
extern void T461f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.make_canonical_from_dates */
extern void T461f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DT_DATE_TIME_DURATION.make_precise */
extern void T461f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6, T6 a7);
/* [detachable] DT_DATE_TIME_DURATION.make_precise */
extern T0* T461c16(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6, T6 a7);
/* [detachable] DT_DATE_TIME_DURATION.make_precise_time_duration */
extern void T461f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DT_DATE_TIME_DURATION.make_date_duration */
extern void T461f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.days_in_current_month */
extern T6 T67f43(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.days_in_month */
extern T6 T67f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.leap_year */
extern T1 T67f38(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.days_in_previous_month */
extern T6 T67f42(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.is_less */
extern T1 T67f41(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.millisecond_count */
extern T6 T67f40(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.day */
extern T6 T67f10(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.month */
extern T6 T67f9(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.year */
extern T6 T67f8(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.add_duration */
extern void T67f62(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.add_milliseconds */
extern void T67f66(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_normalized_time_storage */
extern void T67f69(GE_context* ac, T0* C, T7 a1);
/* detachable DT_DATE_TIME.add_days */
extern void T67f70(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_from_epoch_days */
extern void T67f68(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_year_month_day */
extern void T67f57(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.epoch_to_y2k_days */
extern T6 T67f46(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.epoch_days */
extern T6 T67f11(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.epoch_days_at_month */
extern T6 T67f24(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.days_at_month */
extern T6 T67f23(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.epoch_leap_years */
extern T6 T67f22(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.div */
extern T6 T753f4(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.integer_ */
extern T0* T67f21(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.set_day */
extern void T67f67(GE_context* ac, T0* C, T6 a1);
/* INTEGER_64.to_integer */
extern T6 T7f18(GE_context* ac, T7* C);
/* [detachable] DT_DATE_TIME_DURATION.millisecond_count */
extern T6 T461f15(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.add_years_months_days */
extern void T67f65(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.set_time */
extern void T67f61(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.set_date */
extern void T67f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.tmp_date_time */
extern T0* T461f8(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.make */
extern T0* T67c59(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6);
/* detachable DT_DATE_TIME.make_time */
extern void T67f64(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.make_date */
extern void T67f63(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.minus */
extern T0* T67f12(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.duration */
extern T0* T67f16(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.millisecond */
extern T6 T67f7(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.second */
extern T6 T67f6(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.minute */
extern T6 T67f5(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.hour */
extern T6 T67f4(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.day_count */
extern T6 T67f3(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.date_time_now */
extern T0* T446f10(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.set_date_time_to_now */
extern void T446f12(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.set_precise_hour_minute_second */
extern void T67f58(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DT_SYSTEM_CLOCK.set_local_time */
extern void T446f13(GE_context* ac, T0* C);
/* detachable C_DATE.second_now */
extern T6 T907f6(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_sec */
extern T6 T907f13(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.minute_now */
extern T6 T907f5(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_min */
extern T6 T907f12(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.hour_now */
extern T6 T907f4(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_hour */
extern T6 T907f11(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.day_now */
extern T6 T907f15(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_mday */
extern T6 T907f10(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.month_now */
extern T6 T907f8(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_mon */
extern T6 T907f9(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.year_now */
extern T6 T907f7(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_year */
extern T6 T907f14(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.update */
extern void T907f23(GE_context* ac, T0* C);
/* POINTER.memory_free */
extern void T14f11(GE_context* ac, T14* C);
/* POINTER.default_pointer */
extern T14 T14f8(GE_context* ac, T14* C);
/* POINTER.c_free */
extern void T14f13(GE_context* ac, T14* C, T14 a1);
/* detachable C_DATE.get_millitm */
extern T6 T907f21(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.tm_structure_size */
extern T6 T907f20(GE_context* ac, T0* C);
/* detachable C_DATE.localtime */
extern T14 T907f19(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.gmtime */
extern T14 T907f18(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.get_time */
extern void T907f25(GE_context* ac, T0* C, T14 a1, T14 a2);
/* detachable C_DATE.ftime */
extern void T907f24(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.time_t_structure_size */
extern T6 T907f17(GE_context* ac, T0* C);
/* detachable C_DATE.timeb_structure_size */
extern T6 T907f16(GE_context* ac, T0* C);
/* detachable C_DATE.default_create */
extern T0* T907c22(GE_context* ac);
/* detachable DT_DATE_TIME.make_from_storage */
extern T0* T67c56(GE_context* ac, T6 a1, T6 a2);
/* [detachable] DT_SHARED_SYSTEM_CLOCK.system_clock */
extern T0* T445f1(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.make */
extern T0* T446c11(GE_context* ac);
/* [detachable] DT_SHARED_SYSTEM_CLOCK.default_create */
extern T0* T445c2(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.stop_requested */
extern T1 T62f39(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.record_end_time */
extern void T371f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.print_time */
extern void T371f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.stop_requested */
extern T1 T371f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.compile */
extern void T61f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_system */
extern void T61f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.build_dynamic_type_sets */
extern void T61f90(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_dynamic_type_sets */
extern void T226f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_dynamic_type_sets */
extern void T70f105(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_validity */
extern void T70f109(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_call_validity */
extern void T70f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_target_validity */
extern void T70f116(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f116ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.seeded_dynamic_feature */
extern T0* T306f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.seeded_dynamic_feature */
extern T0* T302f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_catcall_error */
extern void T70f120(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.report_catcall_error */
extern void T51f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_info_message */
extern void T51f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_eiffel_error */
extern void T51f201(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.report_catcall_error */
extern void T35f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_info_message */
extern void T35f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_eiffel_error */
extern void T35f201(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.append_catcall_error_message */
extern void T70f121(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.position */
extern T0* T306f9(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.position */
extern T0* T302f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.string_ */
extern T0* T70f65(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.shared_error_message */
extern T0* T70f64(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_feature_dynamic_type_sets */
extern void T70f108(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.wipe_out */
extern void T255f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].wipe_out */
extern void T377f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_wipe_out */
extern void T377f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_wipe_out */
extern void T377f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_wipe_out */
extern void T377f39(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].keep_head */
extern void T687f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].set_count */
extern void T687f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_wipe_out */
extern void T377f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].unset_found_item */
extern void T377f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].move_all_cursors_after */
extern void T377f36(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_next_cursor */
extern void T686f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_after */
extern void T686f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_built */
extern void T298f36(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_built */
extern void T232f37(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_feature_validity */
extern void T70f112(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE.wipe_out */
extern void T393f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].wipe_out */
extern void T724f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_wipe_out */
extern void T724f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_wipe_out */
extern void T724f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_wipe_out */
extern void T724f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_wipe_out */
extern void T724f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].unset_found_item */
extern void T724f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_all_cursors_after */
extern void T724f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_next_cursor */
extern void T1378f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_after */
extern void T1378f10(GE_context* ac, T0* C);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.wipe_out */
extern void T392f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].wipe_out */
extern void T723f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].keep_head */
extern void T721f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].set_count */
extern void T721f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].move_all_cursors_after */
extern void T723f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT].set_next_cursor */
extern void T1377f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT].set_after */
extern void T1377f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].wipe_out */
extern void T391f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].slots_wipe_out */
extern void T391f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].clashes_wipe_out */
extern void T391f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].key_storage_wipe_out */
extern void T391f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_storage_wipe_out */
extern void T391f50(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].keep_head */
extern void T709f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].set_count */
extern void T709f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].unset_found_item */
extern void T391f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].move_all_cursors_after */
extern void T391f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].set_next_cursor */
extern void T719f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].set_after */
extern void T719f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].forth */
extern void T391f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].cursor_forth */
extern void T391f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].add_traversing_cursor */
extern void T391f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].set_position */
extern void T719f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].remove_traversing_cursor */
extern void T391f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].clashes_item */
extern T6 T391f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_for_iteration */
extern T0* T391f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].cursor_item */
extern T0* T391f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_storage_item */
extern T0* T391f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].after */
extern T1 T391f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].cursor_after */
extern T1 T391f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].after */
extern T1 T719f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].start */
extern void T391f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].cursor_start */
extern void T391f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].cursor_off */
extern T1 T391f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].off */
extern T1 T719f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].is_empty */
extern T1 T391f27(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.wipe_out */
extern void T389f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].wipe_out */
extern void T718f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].keep_head */
extern void T716f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].set_count */
extern void T716f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].move_all_cursors_after */
extern void T718f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].set_next_cursor */
extern void T1376f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].set_after */
extern void T1376f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].wipe_out */
extern void T388f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_wipe_out */
extern void T388f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_wipe_out */
extern void T388f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_wipe_out */
extern void T388f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_wipe_out */
extern void T388f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].unset_found_item */
extern void T388f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].move_all_cursors_after */
extern void T388f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_next_cursor */
extern void T712f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_after */
extern void T712f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].forth */
extern void T388f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_forth */
extern void T388f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].add_traversing_cursor */
extern void T388f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_position */
extern void T712f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].remove_traversing_cursor */
extern void T388f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_item */
extern T6 T388f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_context */
extern void T70f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.reset */
extern void T237f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.wipe_out */
extern void T237f55(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].keep_head */
extern void T613f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].keep_head */
extern void T251f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE].set_count */
extern void T251f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.fixed_array */
extern T0* T237f4(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].default_create */
extern T0* T613c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.tokens */
extern T0* T70f71(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].force_last */
extern void T387f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].force */
extern void T710f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].fill_with */
extern void T709f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].extend */
extern void T709f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].resize */
extern void T387f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].aliased_resized_area */
extern T0* T710f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].aliased_resized_area */
extern T0* T709f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].new_capacity */
extern T6 T387f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].extendible */
extern T1 T387f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_for_iteration */
extern T0* T388f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_item */
extern T0* T388f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_item */
extern T0* T388f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].after */
extern T1 T388f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_after */
extern T1 T388f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].after */
extern T1 T712f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].start */
extern void T388f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_start */
extern void T388f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_off */
extern T1 T388f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].off */
extern T1 T712f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].is_empty */
extern T1 T388f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_preconditions_attachment_scope */
extern void T70f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_attachment_scope */
extern void T70f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].force_last */
extern void T395f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].force */
extern void T727f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].fill_with */
extern void T726f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].extend */
extern void T726f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].resize */
extern void T395f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].aliased_resized_area */
extern T0* T727f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].aliased_resized_area */
extern T0* T726f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].new_capacity */
extern T6 T395f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].extendible */
extern T1 T395f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].forth */
extern void T402f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_forth */
extern void T402f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].add_traversing_cursor */
extern void T402f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_next_cursor */
extern void T738f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_position */
extern void T738f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].remove_traversing_cursor */
extern void T402f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_after */
extern void T738f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_item */
extern T6 T402f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE.merge_scope */
extern void T393f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].forth */
extern void T724f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_forth */
extern void T724f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].add_traversing_cursor */
extern void T724f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_position */
extern void T1378f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove_traversing_cursor */
extern void T724f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_item */
extern T6 T724f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove */
extern void T724f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove_position */
extern void T724f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_put */
extern void T724f80(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_put */
extern void T724f79(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_keep_head */
extern void T724f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_keep_head */
extern void T724f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_put */
extern void T724f68(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_put */
extern void T724f67(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_cursors_forth */
extern void T724f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].internal_set_key_equality_tester */
extern void T724f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_equality_tester */
extern T0* T724f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_item */
extern T6 T724f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].hash_position */
extern T6 T724f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_item */
extern T6 T724f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_item */
extern T6 T724f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].search_position */
extern void T724f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].has */
extern T1 T724f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_for_iteration */
extern T6 T724f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_item */
extern T6 T724f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].after */
extern T1 T724f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_after */
extern T1 T724f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].after */
extern T1 T1378f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].start */
extern void T724f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_start */
extern void T724f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_off */
extern T1 T724f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].off */
extern T1 T1378f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].is_empty */
extern T1 T724f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].is_last */
extern T1 T402f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_is_last */
extern T1 T402f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].is_empty */
extern T1 T402f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_for_iteration */
extern T0* T402f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_item */
extern T0* T402f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_item */
extern T0* T402f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].start */
extern void T402f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_start */
extern void T402f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_off */
extern T1 T402f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].off */
extern T1 T738f5(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE.copy_scope */
extern void T393f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].append_last */
extern void T724f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].extend_last */
extern void T724f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].forth */
extern void T1378f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].put_last */
extern void T724f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].compress */
extern void T724f81(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_all_cursors */
extern void T724f82(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].item */
extern T6 T1378f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].start */
extern void T1378f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_cursor */
extern T0* T724f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].make */
extern T0* T1378c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].resize */
extern void T724f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_resize */
extern void T724f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].special_integer_ */
extern T0* T724f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_resize */
extern void T724f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_resize */
extern void T724f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_resize */
extern void T724f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_modulus */
extern T6 T724f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_capacity */
extern T6 T724f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].extendible */
extern T1 T724f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_attachment_scope */
extern T0* T70f95(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].remove_last */
extern void T395f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].keep_head */
extern void T726f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].set_count */
extern void T726f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].move_last_cursors_after */
extern void T395f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].set_next_cursor */
extern void T728f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].set_after */
extern void T728f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].last */
extern T0* T395f6(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE.make */
extern T0* T393c10(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make */
extern T0* T724c43(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].set_internal_cursor */
extern void T724f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_slots */
extern void T724f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_clashes */
extern void T724f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_key_storage */
extern void T724f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_item_storage */
extern void T724f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].is_empty */
extern T1 T395f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_assertions_attachment_scope */
extern void T70f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.build_scope */
extern void T394f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PRECONDITIONS.assertion */
extern T0* T401f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECONDITIONS.item */
extern T0* T401f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].first */
extern T0* T402f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].force_last */
extern void T402f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_put */
extern void T402f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_put */
extern void T402f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_put */
extern void T402f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_item */
extern T6 T402f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].hash_position */
extern T6 T402f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].resize */
extern void T402f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_resize */
extern void T402f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].special_integer_ */
extern T0* T402f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_resize */
extern void T402f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_resize */
extern void T402f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].aliased_resized_area */
extern T0* T741f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].aliased_resized_area */
extern T0* T739f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_item */
extern T0* T402f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_resize */
extern void T402f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_modulus */
extern T6 T402f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_capacity */
extern T6 T402f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_put */
extern void T402f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].force */
extern void T741f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].fill_with */
extern void T739f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].extend */
extern void T739f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].search_position */
extern void T402f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].unset_found_item */
extern void T402f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.add_precursors_with_preconditions_recursive */
extern void T70f117(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_LIST.item */
extern T0* T404f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_map */
extern T0* T402c38(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_with_equality_testers */
extern void T402f42(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_sparse_container */
extern void T402f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_internal_cursor */
extern void T402f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_cursor */
extern T0* T402f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make */
extern T0* T738c7(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_slots */
extern void T402f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_clashes */
extern void T402f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_key_storage */
extern void T402f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_item_storage */
extern void T402f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].make */
extern T0* T741f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].make_empty */
extern T0* T739c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].default_create */
extern T0* T741c3(GE_context* ac);
/* [detachable] ET_FEATURE_CHECKER.check_preconditions_validity */
extern void T376f74(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECONDITIONS.set_validity_error */
extern void T401f21(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.set_validity_checked */
extern void T401f20(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.free_context */
extern void T376f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.tokens */
extern T0* T376f52(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.keep_object_tests */
extern void T389f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].keep_first */
extern void T718f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.current_system */
extern T0* T376f54(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.current_system */
extern T0* T73f83(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.build_scope */
extern void T390f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwbe0a_error */
extern void T51f203(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_validity_error */
extern void T51f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_info */
extern void T51f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_vwbe0a */
extern T0* T759c830(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.set_compilers */
extern void T759f831(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_VALIDITY_ERROR.filename */
extern T0* T759f21(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.default_message_template */
extern T0* T759f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.null_position */
extern T0* T759f12(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.make_default */
extern T0* T189c12(GE_context* ac);
/* [detachable] ET_VALIDITY_ERROR.template_code */
extern T0* T759f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwbe_error */
extern T1 T51f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwbe0a_error */
extern void T35f203(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_validity_error */
extern void T35f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_info */
extern void T35f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vwbe_error */
extern T1 T35f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_type */
extern T1 T237f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_expression_validity */
extern void T376f77(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_expression_supplier */
extern void T376f78(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_expression_supplier */
extern void T1299f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.add_supplier */
extern void T1299f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].force_last */
extern void T926f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_put */
extern void T926f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_put */
extern void T926f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_item */
extern T6 T926f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].hash_position */
extern T6 T926f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.hash_code */
extern T6 T560f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].resize */
extern void T926f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_resize */
extern void T926f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].special_integer_ */
extern T0* T926f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_resize */
extern void T926f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_resize */
extern void T926f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].aliased_resized_area */
extern T0* T1600f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].aliased_resized_area */
extern T0* T1598f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_item */
extern T0* T926f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_item */
extern T0* T926f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_item */
extern T6 T926f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_resize */
extern void T926f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_modulus */
extern T6 T926f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_capacity */
extern T6 T926f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_put */
extern void T926f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].force */
extern void T1600f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].fill_with */
extern void T1598f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].extend */
extern void T1598f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].search_position */
extern void T926f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_equality_tester */
extern T0* T926f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].unset_found_item */
extern void T926f46(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.actual_class */
extern T0* T560f21(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_class */
extern T0* T73f78(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.count */
extern T6 T389f3(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.new_context */
extern T0* T376f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].remove_last */
extern void T387f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_last_cursors_after */
extern void T387f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_next_cursor */
extern void T711f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_after */
extern void T711f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].last */
extern T0* T387f6(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.make_with_capacity */
extern T0* T237c51(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.make_ast_list_with_capacity */
extern void T237f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].make */
extern T0* T613f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE].make_empty */
extern T0* T251c5(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].is_empty */
extern T1 T387f10(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.current_universe_impl */
extern T0* T376f49(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.error_handler */
extern T0* T376f60(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_fatal_error */
extern void T376f76(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_feature_validity */
extern void T70f111(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.is_precursor */
extern T1 T232f23(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_dynamic_type_sets */
extern void T298f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.set_dynamic_type_sets */
extern void T232f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_catcall_warning_mode */
extern void T226f15(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_catcall_warning_mode */
extern void T70f101(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_catcall_error_mode */
extern void T226f14(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_catcall_error_mode */
extern void T70f100(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_no_assertion */
extern void T226f18(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_no_assertion */
extern void T70f104(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_no_debug */
extern void T226f17(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_no_debug */
extern void T70f103(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_creation */
extern void T298f39(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_creation */
extern void T232f33(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.is_creation_directly_exported_to */
extern T1 T73f197(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR_LIST.is_directly_exported_to */
extern T1 T1080f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR.is_directly_exported_to */
extern T1 T1077f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR.has_feature_name */
extern T1 T1077f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.has_class */
extern T1 T1038f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT.base_class */
extern T0* T398f3(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.has_class */
extern T1 T397f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.is_empty */
extern T1 T1077f11(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc5a_error */
extern void T51f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_system_error */
extern void T51f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc5a */
extern T0* T757c32(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc5_error */
extern T1 T51f16(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc5a_error */
extern void T35f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_system_error */
extern void T35f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc5_error */
extern T1 T35f16(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vsrc1a_error */
extern void T51f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_vsrc1a */
extern T0* T757c31(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vsrc1_error */
extern T1 T51f15(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vsrc1a_error */
extern void T35f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vsrc1_error */
extern T1 T35f15(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_syntax_error */
extern T1 T73f101(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_parsed */
extern T1 T73f100(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc4a_error */
extern void T51f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc4a */
extern T0* T757c30(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc4_error */
extern T1 T51f14(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc4a_error */
extern void T35f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc4_error */
extern T1 T35f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_preparsed */
extern T1 T73f86(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc6a_error */
extern void T51f186(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc6a */
extern T0* T757c29(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc6_error */
extern T1 T51f13(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc6a_error */
extern void T35f186(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc6_error */
extern T1 T35f13(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.default_create_feature_name */
extern T0* T72f3(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc3a_error */
extern void T51f185(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc3a */
extern T0* T757c28(GE_context* ac);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc3_error */
extern T1 T51f12(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc3a_error */
extern void T35f185(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc3_error */
extern T1 T35f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.error_handler */
extern T0* T61f58(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_kernel */
extern void T61f88(GE_context* ac, T0* C, T0* a1);
extern T1 T61f88ot1(T0* a1, T0** a2);
/* [detachable] ET_TOKEN_CONSTANTS.set_exception_data_feature_name */
extern T0* T72f19(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.once_raise_feature_name */
extern T0* T72f18(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.last_exception_feature_name */
extern T0* T72f17(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.init_exception_manager_feature_name */
extern T0* T72f16(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe6a_error */
extern void T51f196(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe6a */
extern T0* T759c829(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ARRAY [[attached] ET_TYPE].item */
extern T0* T250f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe6_error */
extern T1 T51f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe6a_error */
extern void T35f196(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe6_error */
extern T1 T35f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe4a_error */
extern void T51f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe4a */
extern T0* T759c828(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe4_error */
extern T1 T51f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe4a_error */
extern void T35f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe4_error */
extern T1 T35f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_rout_disp_final_feature_name */
extern T0* T72f15(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.rout_disp_feature_name */
extern T0* T72f14(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.closed_operands_feature_name */
extern T0* T72f13(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_pointer_feature_name */
extern T0* T72f10(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_item_feature_name */
extern T0* T72f9(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.older_ise_version */
extern T1 T62f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.is_less */
extern T1 T44f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.build */
extern T6 T44f9(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_build */
extern T1 T44f8(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.revision */
extern T6 T44f7(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_revision */
extern T1 T44f6(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.minor */
extern T6 T44f5(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_minor */
extern T1 T44f13(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.major */
extern T6 T44f12(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_major */
extern T1 T44f11(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.older_ise_version */
extern T1 T371f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.ise_6_0_6_7057 */
extern T0* T61f76(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.make */
extern T0* T44c19(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ET_TOKEN_CONSTANTS.upper_feature_name */
extern T0* T72f8(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.lower_feature_name */
extern T0* T72f7(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.put_last */
extern void T237f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].force */
extern void T613f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].fill_with */
extern void T251f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].extend */
extern void T251f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.formal_parameter_type */
extern T0* T73f89(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].replace */
extern void T866f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.make */
extern T0* T246c75(GE_context* ac, T0* a1, T0* a2, T6 a3, T0* a4);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].item */
extern T0* T866f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].put_last */
extern void T866f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].force */
extern void T1570f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].fill_with */
extern void T1569f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].extend */
extern void T1569f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].resize */
extern void T866f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].aliased_resized_area */
extern T0* T1570f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].aliased_resized_area */
extern T0* T1569f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T866c8(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].set_internal_cursor */
extern void T866f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].new_cursor */
extern T0* T866f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T1571c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T1570f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].make_empty */
extern T0* T1569c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].default_create */
extern T0* T1570c3(GE_context* ac);
/* detachable ET_FORMAL_PARAMETER_LIST.formal_parameter */
extern T0* T593f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.formal_parameter */
extern T0* T1450f10(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.formal_parameter */
extern T0* T867f8(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.item */
extern T0* T593f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.attached_type */
extern T0* T61f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.attached_like_current */
extern T0* T72f20(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.make */
extern T0* T442c77(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.like_keyword */
extern T0* T72f269(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_like */
extern T0* T409c52(GE_context* ac);
/* [detachable] ET_LIKE_CURRENT.tokens */
extern T0* T442f11(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.attached_keyword */
extern T0* T72f60(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_attached */
extern T0* T409c43(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.area_feature_name */
extern T0* T72f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.make_generic */
extern T0* T71c123(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TOKEN_CONSTANTS.capacity_feature_name */
extern T0* T72f5(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe3a_error */
extern void T51f194(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe3a */
extern T0* T759c827(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe3_error */
extern T1 T51f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe3a_error */
extern void T35f194(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe3_error */
extern T1 T35f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe1a_error */
extern void T51f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe1a */
extern T0* T759c826(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe1_error */
extern T1 T51f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe1a_error */
extern void T35f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe1_error */
extern T1 T35f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe2a_error */
extern void T51f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe2a */
extern T0* T759c825(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe2_error */
extern T1 T51f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe2a_error */
extern void T35f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe2_error */
extern T1 T35f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.count_feature_name */
extern T0* T72f4(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvknl1a_error */
extern void T51f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvknl1a */
extern T0* T757c33(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvknl1_error */
extern T1 T51f17(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvknl1a_error */
extern void T35f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvknl1_error */
extern T1 T35f17(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_6 */
extern void T62f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_6_metrics */
extern void T62f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T6 T56f147(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T0* T56f147ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T0* T56f147ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]].set_rout_disp_final */
extern void T224f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS].call */
extern void T808f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_COUNTER.increment */
extern void T807f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_do_recursive */
extern void T56f168(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive */
extern T0* T56f168ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]].set_rout_disp_final */
extern void T812f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM.universes_do_recursive */
extern void T56f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].do_all */
extern void T829f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_item */
extern T0* T829f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_item */
extern T6 T829f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern void T56f186(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern T0* T56f186ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern T0* T56f186ac2(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T819f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.add_universe_recursive */
extern void T207f170(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.add_universe_recursive */
extern T0* T207f170ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.do_all */
extern void T585f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].item */
extern T0* T1356f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.add_universe_recursive */
extern void T586f170(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.add_universe_recursive */
extern T0* T586f170ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]].set_rout_disp_final */
extern void T828f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern void T183f195(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern T0* T183f195ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern T0* T183f195ac2(GE_context* ac, T0* a1);
/* [detachable] ET_LIBRARY.add_universe_recursive */
extern void T826f170(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_universe_recursive */
extern T0* T826f170ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.add_universe_recursive */
extern T0* T826f170ac2(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_all */
extern void T206f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].item */
extern T0* T584f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_all */
extern void T184f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].item */
extern T0* T568f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].force_last */
extern void T829f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_put */
extern void T829f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_put */
extern void T829f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_item */
extern T6 T829f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].hash_position */
extern T6 T829f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].resize */
extern void T829f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_resize */
extern void T829f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].special_integer_ */
extern T0* T829f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_storage_resize */
extern void T829f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_resize */
extern void T829f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].aliased_resized_area */
extern T0* T1559f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].aliased_resized_area */
extern T0* T1557f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_storage_item */
extern T0* T829f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_resize */
extern void T829f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_modulus */
extern T6 T829f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_capacity */
extern T6 T829f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_put */
extern void T829f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].force */
extern void T1559f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].fill_with */
extern void T1557f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].extend */
extern void T1557f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].search_position */
extern void T829f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_equality_tester */
extern T0* T829f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].unset_found_item */
extern void T829f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].has */
extern T1 T829f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make */
extern T0* T829c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].set_internal_cursor */
extern void T829f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_cursor */
extern T0* T829f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].make */
extern T0* T1556c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_slots */
extern void T829f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_clashes */
extern void T829f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_key_storage */
extern void T829f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_item_storage */
extern void T829f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].make */
extern T0* T1559f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].make_empty */
extern T0* T1557c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].default_create */
extern T0* T1559c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM.initial_universes_capacity */
extern T6 T56f137(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.class_actions */
extern T0* T56f142(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS].default_create */
extern T0* T808c1(GE_context* ac);
/* [detachable] UT_COUNTER.make */
extern T0* T807c2(GE_context* ac, T6 a1);
/* [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern void T56f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.build_scm_write_mappings */
extern void T56f175(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.build_scm_write_mappings */
extern T0* T56f175ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]].set_rout_disp_final */
extern void T832f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM.build_scm_write_mapping */
extern void T56f189(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_scm_write_mapping */
extern void T776f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.clusters_do_explicit_recursive */
extern void T56f188(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.clusters_do_explicit_recursive */
extern T0* T56f188ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]].set_rout_disp_final */
extern void T822f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM.internal_universes_do_recursive */
extern void T56f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].do_all */
extern void T834f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_item */
extern T0* T834f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_item */
extern T6 T834f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.add_internal_universe_recursive */
extern void T56f190(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.add_internal_universe_recursive */
extern T0* T56f190ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_LIBRARY.add_internal_universe_recursive */
extern void T183f196(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_internal_universe_recursive */
extern T0* T183f196ac1(GE_context* ac, T0* a1);
/* [detachable] ET_LIBRARY.add_internal_universe_recursive */
extern void T826f171(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_internal_universe_recursive */
extern T0* T826f171ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].force_last */
extern void T834f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_put */
extern void T834f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_put */
extern void T834f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_item */
extern T6 T834f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].hash_position */
extern T6 T834f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].resize */
extern void T834f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_resize */
extern void T834f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].special_integer_ */
extern T0* T834f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_storage_resize */
extern void T834f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_resize */
extern void T834f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].aliased_resized_area */
extern T0* T1562f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].aliased_resized_area */
extern T0* T1563f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_storage_item */
extern T0* T834f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_resize */
extern void T834f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_modulus */
extern T6 T834f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_capacity */
extern T6 T834f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_put */
extern void T834f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].force */
extern void T1562f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].fill_with */
extern void T1563f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].extend */
extern void T1563f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].search_position */
extern void T834f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_equality_tester */
extern T0* T834f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].unset_found_item */
extern void T834f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].has */
extern T1 T834f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T834c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].set_internal_cursor */
extern void T834f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_cursor */
extern T0* T834f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T1560c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_slots */
extern void T834f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_clashes */
extern void T834f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_key_storage */
extern void T834f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_item_storage */
extern void T834f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T1562f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].make_empty */
extern T0* T1563c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].default_create */
extern T0* T1562c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM.build_scm_read_mappings */
extern void T56f174(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.build_scm_read_mappings */
extern T0* T56f174ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.build_scm_read_mapping */
extern void T56f187(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_scm_read_mapping */
extern void T776f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_master_class_validity */
extern void T371f110(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.check_master_class_validity */
extern T0* T371f110ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]].set_rout_disp_final */
extern void T837f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_MASTER_CLASS.process */
extern void T560f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_master_class */
extern void T457f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.process_master_class */
extern void T448f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_master_class_validity */
extern void T448f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_overriding_classes_validity */
extern void T448f12(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0d_error */
extern void T51f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_universe_error */
extern void T51f397(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0d */
extern T0* T1945c21(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_MASTER_CLASS.upper_name */
extern T0* T560f49(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vscn_error */
extern T1 T51f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vscn0d_error */
extern void T35f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_universe_error */
extern void T35f397(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vscn_error */
extern T1 T35f108(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].item */
extern T0* T1347f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.current_universe */
extern T0* T448f6(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.error_handler */
extern T0* T448f5(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.set_fatal_error */
extern void T448f13(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].is_empty */
extern T1 T1347f12(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_intrinsic_classes_validity */
extern void T448f11(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_overridden_class_validity */
extern void T448f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0c_error */
extern void T51f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0c */
extern T0* T1945c23(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vscn0c_error */
extern void T35f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.is_in_dotnet_assembly */
extern T1 T73f95(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0b_error */
extern void T51f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0b */
extern T0* T1945c22(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vscn0b_error */
extern void T35f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MASTER_CLASS.actual_intrinsic_class */
extern T0* T560f26(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_intrinsic_class */
extern T0* T73f192(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0a_error */
extern void T51f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0a */
extern T0* T1945c20(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vscn0a_error */
extern void T35f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].item */
extern T0* T455f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.reset_fatal_error */
extern void T448f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern void T56f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern T0* T56f251ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern T0* T56f251ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.universes_do_recursive_until */
extern void T56f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].forth */
extern void T829f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_forth */
extern void T829f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].add_traversing_cursor */
extern void T829f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_next_cursor */
extern void T1556f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_position */
extern void T1556f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].remove_traversing_cursor */
extern void T829f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_after */
extern void T1556f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_for_iteration */
extern T0* T829f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_item */
extern T0* T829f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].go_after */
extern void T829f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_go_after */
extern void T829f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_off */
extern T1 T829f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].off */
extern T1 T1556f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].after */
extern T1 T829f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_after */
extern T1 T829f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].after */
extern T1 T1556f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].start */
extern void T829f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_start */
extern void T829f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].is_empty */
extern T1 T829f31(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.check_master_class_validity */
extern void T62f106(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.check_master_class_validity */
extern T0* T62f106ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern void T56f170p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac2(GE_context* ac);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac4(GE_context* ac, T0* a1);
/* Creation of agent #5 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac5(GE_context* ac, T0* a1);
/* Creation of agent #6 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac6(GE_context* ac, T0* a1);
/* Creation of agent #7 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac7(GE_context* ac);
/* Creation of agent #8 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T56f170p1ac8(GE_context* ac);
/* [detachable] ET_ECF_LIBRARY.import_classes */
extern void T183f194(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T183f194ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T183f194ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T183f194ac3(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T839f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.export_classes */
extern void T209f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].forth */
extern void T599f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_forth */
extern void T221f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].add_traversing_cursor */
extern void T221f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_next_cursor */
extern void T599f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_position */
extern void T599f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].remove_traversing_cursor */
extern void T221f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_after */
extern void T599f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_item */
extern T6 T221f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS.add_last_imported_class */
extern void T560f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.update_intrinsic_class */
extern void T560f55(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern void T560f56(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T560f56ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T560f56ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T560f56ac3(GE_context* ac, T0* a1);
extern T1 T560f56ot1(T0* a1, T0** a2);
/* [detachable] ET_MASTER_CLASS.set_marked */
extern void T560f51(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.set_marked */
extern void T73f284(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_MASTER_CLASS.is_marked */
extern T1 T560f36(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].do_all */
extern void T1347f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.unmark_overridden */
extern void T560f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_overriding_class */
extern void T560f64(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove_first */
extern void T1347f19(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove */
extern void T1347f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].keep_head */
extern void T601f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].set_count */
extern void T601f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_left */
extern void T1347f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_cursors_left */
extern void T1347f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_position */
extern void T1678f6(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove_last */
extern void T1347f24(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_last_cursors_after */
extern void T1347f31(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_next_cursor */
extern void T1678f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_after */
extern void T1678f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].first */
extern T0* T1347f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].delete */
extern void T1347f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_all_cursors_after */
extern void T1347f27(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_unknown_class */
extern T1 T73f169(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_unknown */
extern T1 T73f85(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.tokens */
extern T0* T560f22(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.mark_overridden */
extern void T560f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_overriding_class */
extern void T560f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.is_override */
extern T1 T560f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_override */
extern T1 T73f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_in_override_group */
extern T1 T73f171(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].force_last */
extern void T1347f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].force */
extern void T600f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].fill_with */
extern void T601f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].extend */
extern void T601f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].resize */
extern void T1347f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].aliased_resized_area */
extern T0* T600f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].aliased_resized_area */
extern T0* T601f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].new_capacity */
extern T6 T1347f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].extendible */
extern T1 T1347f9(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].make */
extern T0* T1347c14(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].set_internal_cursor */
extern void T1347f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].new_cursor */
extern T0* T1347f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].make */
extern T0* T1678c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].make */
extern T0* T600f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].make_empty */
extern T0* T601c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].default_create */
extern T0* T600c3(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.has_imported_class */
extern T1 T560f47(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].has */
extern T1 T1347f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item */
extern T0* T212f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key */
extern T0* T599f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_key */
extern T0* T221f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_item */
extern T0* T221f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS.is_mapped */
extern T1 T560f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item */
extern T0* T599f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_item */
extern T0* T221f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_item */
extern T0* T221f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].after */
extern T1 T599f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].start */
extern void T599f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_start */
extern void T221f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_off */
extern T1 T221f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].off */
extern T1 T599f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].is_empty */
extern T1 T221f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_cursor */
extern T0* T221f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make */
extern T0* T599c10(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]].set_rout_disp_final */
extern void T187f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.export_classes */
extern void T203f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_unknown_imported_classes */
extern void T560f60(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.remove_imported_class */
extern void T560f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.is_declared_locally */
extern T1 T560f24(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_adapted */
extern void T206f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].do_all */
extern void T584f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_adapted */
extern void T184f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].do_all */
extern void T568f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.import_classes */
extern void T826f169(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T826f169ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T826f169ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T826f169ac3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_classes */
extern void T207f169(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_kernel_class */
extern void T207f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_class */
extern T0* T207f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].force_last_new */
extern void T221f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_put */
extern void T221f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_put */
extern void T221f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_put */
extern void T221f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_put */
extern void T221f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_item */
extern T6 T221f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].hash_position */
extern T6 T221f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].resize */
extern void T221f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_resize */
extern void T221f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].special_integer_ */
extern T0* T221f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_resize */
extern void T221f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_resize */
extern void T221f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_resize */
extern void T221f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_modulus */
extern T6 T221f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_capacity */
extern T6 T221f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].unset_found_item */
extern void T221f46(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.make */
extern T0* T560c50(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].found_item */
extern T0* T221f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].found */
extern T1 T221f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].search */
extern void T221f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].search_position */
extern void T221f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.tokens */
extern T0* T207f88(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_kernel_classes */
extern void T207f180(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.native_array_class_name */
extern T0* T72f1520(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_classes */
extern void T586f169(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_kernel_class */
extern void T586f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.master_class */
extern T0* T586f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.tokens */
extern T0* T586f88(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_kernel_classes */
extern void T586f180(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.preparse */
extern void T183f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.preparse */
extern void T183f193p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T183f193p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T183f193p1ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T183f193p1ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T183f193p1ac4(GE_context* ac, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_unknown_local_classes */
extern void T560f59(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.remove_local_ignored_class */
extern void T560f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_first */
extern void T455f26(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].first */
extern T0* T455f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].is_empty */
extern T1 T455f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].delete */
extern void T455f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] ET_CLASS].keep_head */
extern void T1304f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS].set_count */
extern void T1304f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_all_cursors_after */
extern void T455f21(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_next_cursor */
extern void T1306f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_after */
extern void T1306f10(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.remove_local_non_override_class */
extern void T560f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_local_override_class */
extern void T560f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove */
extern void T455f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_left */
extern void T455f32(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_cursors_left */
extern void T455f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_position */
extern void T1306f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_last */
extern void T455f30(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_last_cursors_after */
extern void T455f34(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern void T560f58(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T560f58ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T560f58ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T560f58ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T560f58ac4(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #5 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T560f58ac5(GE_context* ac);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]].set_rout_disp_final */
extern void T844f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_MASTER_CLASS.reset_local_modified_class */
extern void T560f67(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.reset_parsed */
extern void T73f291(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_preparsed */
extern void T73f275(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_ancestors_built */
extern void T73f299(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_features_flattened */
extern void T73f298(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_interface_checked */
extern void T73f297(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_implementation_checked */
extern void T73f296(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.file_time_stamp */
extern T6 T124f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.file_time_stamp */
extern T6 T123f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.file_system */
extern T0* T560f23(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.unix_file_system */
extern T0* T560f31(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.windows_file_system */
extern T0* T560f30(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.operating_system */
extern T0* T560f29(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset */
extern void T73f290(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_preparsed */
extern void T73f277(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_in_cluster */
extern T1 T73f158(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_parsed_and_errors */
extern void T73f285(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_parsed */
extern void T73f286(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FEATURE_LIST.reset */
extern void T1069f10(GE_context* ac, T0* C);
/* detachable ET_CONVERT_PROCEDURE.reset */
extern void T1458f10(GE_context* ac, T0* C);
/* [detachable] ET_BRACED_TYPE_LIST.reset */
extern void T1346f12(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.reset */
extern void T1457f9(GE_context* ac, T0* C);
/* detachable ET_CONVERT_PROCEDURE.convert_feature */
extern T0* T1458f6(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.convert_feature */
extern T0* T1457f5(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.reset */
extern void T1080f11(GE_context* ac, T0* C);
/* detachable ET_CREATOR.reset */
extern void T1077f19(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset */
extern void T1168f34(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset_validity_checked */
extern void T1168f31(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset */
extern void T1168f34p1(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.reset */
extern void T1206f8(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.reset */
extern void T558f14(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.reset */
extern void T557f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.reset */
extern void T564f10(GE_context* ac, T0* C);
/* detachable ET_EXPORT_LIST.reset */
extern void T563f9(GE_context* ac, T0* C);
/* detachable ET_FEATURE_EXPORT.reset */
extern void T1118f13(GE_context* ac, T0* C);
/* detachable ET_ALL_EXPORT.reset */
extern void T1455f8(GE_context* ac, T0* C);
/* detachable ET_RENAME_LIST.reset */
extern void T562f10(GE_context* ac, T0* C);
/* [detachable] ET_RENAME.reset */
extern void T1295f7(GE_context* ac, T0* C);
/* [detachable] ET_RENAME.rename_pair */
extern T0* T1295f5(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.parent */
extern T0* T557f10(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset */
extern void T254f11(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset */
extern void T253f11(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.local_ignored_classes_do_if */
extern void T560f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].do_if */
extern void T455f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_ignored_classes_do_all */
extern void T560f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].do_all */
extern void T455f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_all */
extern void T560f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_all */
extern void T560f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.process */
extern void T776f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_cluster */
extern void T457f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.process_cluster */
extern void T447f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern void T447f182(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern T0* T447f182ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern T0* T447f182ac2(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS.is_in_group */
extern T1 T73f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_clusters */
extern void T447f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].remove */
extern void T1527f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].keep_head */
extern void T1804f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].set_count */
extern void T1804f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_left */
extern void T1527f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_cursors_left */
extern void T1527f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_position */
extern void T1806f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].remove_last */
extern void T1527f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_last_cursors_after */
extern void T1527f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_next_cursor */
extern void T1806f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_after */
extern void T1806f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].item */
extern T0* T1527f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.build_dependant_constraint */
extern void T447f185(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_dependant_constraint */
extern void T776f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_dependant_constraint */
extern void T761f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.make */
extern T0* T1706c17(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].set_key_equality_tester */
extern void T1986f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.string_equality_tester */
extern T0* T1706f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_map */
extern T0* T1986c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_with_equality_testers */
extern void T1986f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_sparse_container */
extern void T1986f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].set_internal_cursor */
extern void T1986f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].new_cursor */
extern T0* T1986f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].make */
extern T0* T1987c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].unset_found_item */
extern void T1986f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_slots */
extern void T1986f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].new_modulus */
extern T6 T1986f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_clashes */
extern void T1986f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_key_storage */
extern void T1986f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_item_storage */
extern void T1986f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].make */
extern T0* T2028f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].make_empty */
extern T0* T2025c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].default_create */
extern T0* T2028c2(GE_context* ac);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.make_with_pathnames */
extern T0* T1706c16(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].put_last */
extern void T1986f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_put */
extern void T1986f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_put */
extern void T1986f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_put */
extern void T1986f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_item */
extern T6 T1986f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].compress */
extern void T1986f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].hash_position */
extern T6 T1986f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_wipe_out */
extern void T1986f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_wipe_out */
extern void T1986f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_keep_head */
extern void T1986f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_keep_head */
extern void T1986f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].keep_head */
extern void T2025f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].set_count */
extern void T2025f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].move_all_cursors */
extern void T1986f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_position */
extern void T1987f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_item */
extern T0* T1986f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_item */
extern T0* T1986f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_item */
extern T6 T1986f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_put */
extern void T1986f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].force */
extern void T2028f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].fill_with */
extern void T2025f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].extend */
extern void T2025f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].search_position */
extern void T1986f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.compile */
extern void T2014f20(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_DFA_WILDCARD.compile */
extern T0* T2014c20(GE_context* ac, T0* a1, T1 a2);
/* [detachable] LX_FULL_DFA.make */
extern T0* T2042c38(GE_context* ac, T0* a1);
/* [detachable] LX_FULL_DFA.build */
extern void T2042f41(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_accept_table */
extern void T2042f47(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].first */
extern T0* T2054f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_nxt_table */
extern void T2042f46(GE_context* ac, T0* C);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].target */
extern T0* T2093f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].value */
extern T0* T2143f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_item */
extern T0* T2143f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].position_of_key */
extern T6 T2143f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_item */
extern T6 T2143f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_item */
extern T6 T2143f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].hash_position */
extern T6 T2143f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_item */
extern T6 T2143f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.is_accepting */
extern T1 T2085f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].is_empty */
extern T1 T2054f11(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_transitions */
extern void T2042f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].set_target */
extern void T2093f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].force_new */
extern void T2143f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_put */
extern void T2143f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_put */
extern void T2143f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].force */
extern void T2140f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].fill_with */
extern void T2139f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].extend */
extern void T2139f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_put */
extern void T2143f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_put */
extern void T2143f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].resize */
extern void T2143f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_resize */
extern void T2143f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].special_integer_ */
extern T0* T2143f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_resize */
extern void T2143f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_resize */
extern void T2143f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].aliased_resized_area */
extern T0* T2140f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].aliased_resized_area */
extern T0* T2139f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_resize */
extern void T2143f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_modulus */
extern T6 T2143f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_capacity */
extern T6 T2143f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].unset_found_item */
extern void T2143f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].replace_found_item */
extern void T2143f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].found */
extern T1 T2143f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].search */
extern void T2143f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].search_position */
extern void T2143f47(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.previous_symbol */
extern T6 T2088f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].item */
extern T0* T2127f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.new_state */
extern T0* T2042f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.set_id */
extern void T2085f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].put_last */
extern void T2087f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.is_equal */
extern T1 T2085f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].is_equal */
extern T1 T2086f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].any_ */
extern T0* T2086f17(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.any_ */
extern T0* T2085f14(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.new_state */
extern T0* T2085f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.make */
extern T0* T2085c16(GE_context* ac, T0* a1, T6 a2, T6 a3);
extern T1 T2085f16ot1(T0* a1, T0** a2);
/* detachable LX_RULE.set_useful */
extern void T2071f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].sort */
extern void T2054f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].sort */
extern void T2132f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].sort_with_comparator */
extern void T2132f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].subsort_with_comparator */
extern void T2132f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].replace */
extern void T2054f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].less_than */
extern T1 T2150f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].detachable_less_than */
extern T1 T2150f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].attached_less_than */
extern T1 T2150f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable LX_RULE.is_less */
extern T1 T2071f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].item */
extern T0* T2054f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.rule_sorter */
extern T0* T2085f7(GE_context* ac, T0* C);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].make */
extern T0* T2132c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].make */
extern T0* T2150c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].sort */
extern void T2086f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].sort */
extern void T2130f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].sort_with_comparator */
extern void T2130f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].subsort_with_comparator */
extern void T2130f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].replace */
extern void T2086f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].less_than */
extern T1 T2153f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].detachable_less_than */
extern T1 T2153f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].attached_less_than */
extern T1 T2153f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] LX_NFA_STATE.is_less */
extern T1 T2096f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.bubble_sorter */
extern T0* T2085f8(GE_context* ac, T0* C);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].make */
extern T0* T2130c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].make */
extern T0* T2153c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].last */
extern T0* T2086f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].is_empty */
extern T1 T2086f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].force_last */
extern void T2054f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].force */
extern void T2104f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_RULE].fill_with */
extern void T2084f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_RULE].extend */
extern void T2084f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].resize */
extern void T2054f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].aliased_resized_area */
extern T0* T2104f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_RULE].aliased_resized_area */
extern T0* T2084f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].new_capacity */
extern T6 T2054f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].extendible */
extern T1 T2054f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_NFA_STATE.is_accepting_head */
extern T1 T2096f6(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.has_transition */
extern T1 T2096f8(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.is_accepting */
extern T1 T2096f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].has */
extern T1 T2086f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].force_last */
extern void T2086f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].force */
extern void T2136f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].fill_with */
extern void T2134f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].extend */
extern void T2134f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].resize */
extern void T2086f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].aliased_resized_area */
extern T0* T2136f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].aliased_resized_area */
extern T0* T2134f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].new_capacity */
extern T6 T2086f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].extendible */
extern T1 T2086f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].make */
extern T0* T2054c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].set_internal_cursor */
extern void T2054f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].new_cursor */
extern T0* T2054f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE].make */
extern T0* T2105c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].make */
extern T0* T2104f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_RULE].make_empty */
extern void T2084f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_RULE].make_empty */
extern T0* T2084c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].default_create */
extern T0* T2104c3(GE_context* ac);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].make */
extern T0* T2093c8(GE_context* ac, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_map */
extern T0* T2143c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_with_equality_testers */
extern void T2143f36(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_sparse_container */
extern void T2143f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].set_internal_cursor */
extern void T2143f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_cursor */
extern T0* T2143f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32].make */
extern T0* T2162c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_slots */
extern void T2143f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_clashes */
extern void T2143f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_key_storage */
extern void T2143f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_item_storage */
extern void T2143f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].make */
extern T0* T2140f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].make_empty */
extern T0* T2139c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].default_create */
extern T0* T2140c3(GE_context* ac);
/* [detachable] LX_DFA_STATE.maximum_symbol */
extern T6 T2085f13(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.minimum_symbol */
extern T6 T2085f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put_last */
extern void T2086f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].item */
extern T0* T2086f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].make */
extern T0* T2086c19(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].set_internal_cursor */
extern void T2086f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].new_cursor */
extern T0* T2086f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].make */
extern T0* T2137c9(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].make */
extern T0* T2136f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].make_empty */
extern T0* T2134c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].default_create */
extern T0* T2136c3(GE_context* ac);
/* detachable LX_SYMBOL_PARTITIONS.is_representative */
extern T1 T2088f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.partition */
extern void T2085f18(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.initialize */
extern void T2088f11(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.initialize */
extern void T2088f11p1(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.new_lower */
extern T6 T2088f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put_left */
extern void T2128f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].attach_right */
extern void T2128f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put */
extern void T2128f8(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.lower */
extern T6 T2088f7(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.upper */
extern T6 T2088f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.resize */
extern void T2042f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].resize */
extern void T2087f10(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.capacity */
extern T6 T2088f5(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].count */
extern T6 T2127f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].item */
extern T0* T2087f7(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.make */
extern T0* T2088c10(GE_context* ac, T6 a1, T6 a2);
/* detachable LX_SYMBOL_PARTITIONS.make */
extern void T2088f10p1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].put */
extern void T2127f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled */
extern T0* T2127c6(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled_area */
extern void T2127f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled */
extern T0* T2129c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].fill_with */
extern void T2129f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].extend */
extern void T2129f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_empty */
extern void T2129f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].make */
extern T0* T2128c4(GE_context* ac, T6 a1);
/* [detachable] LX_FULL_DFA.put_eob_state */
extern void T2042f40(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].force_first */
extern void T2054f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].put */
extern void T2054f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].move_cursors_right */
extern void T2054f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE].set_position */
extern void T2105f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].move_right */
extern void T2054f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].put_last */
extern void T2054f19(GE_context* ac, T0* C, T0* a1);
/* detachable LX_RULE.make_default */
extern T0* T2071c14(GE_context* ac, T6 a1);
/* detachable LX_RULE.dummy_action */
extern T0* T2071f11(GE_context* ac, T0* C);
/* [detachable] LX_ACTION.make */
extern T0* T2126c2(GE_context* ac, T0* a1);
/* detachable LX_RULE.dummy_pattern */
extern T0* T2071f12(GE_context* ac, T0* C);
/* [detachable] LX_NFA.make_epsilon */
extern T0* T2069c6(GE_context* ac, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put_first */
extern void T2086f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put */
extern void T2086f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].move_cursors_right */
extern void T2086f31(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_position */
extern void T2137f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].move_right */
extern void T2086f30(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] LX_NFA_STATE.set_transition */
extern void T2096f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2120c3(GE_context* ac, T0* a1);
/* [detachable] LX_NFA_STATE.make */
extern T0* T2096c10(GE_context* ac, T1 a1);
/* [detachable] LX_FULL_DFA.initialize */
extern void T2042f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_DFA.initialize_dfa */
extern void T2042f44(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] LX_FULL_DFA.put_start_condition */
extern void T2042f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.start_state */
extern T0* T2069f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].first */
extern T0* T2086f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].item */
extern T0* T2094f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.item */
extern T0* T2055f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.set_nfa_state_ids */
extern void T2042f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA_STATE.set_id */
extern void T2096f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].force */
extern void T2095f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_put */
extern void T2095f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_put */
extern void T2095f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_put */
extern void T2095f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_item */
extern T6 T2095f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_item */
extern T6 T2095f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].hash_position */
extern T6 T2095f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].resize */
extern void T2095f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_resize */
extern void T2095f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].special_integer_ */
extern T0* T2095f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_resize */
extern void T2095f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_resize */
extern void T2095f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].aliased_resized_area */
extern T0* T2068f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_NFA].aliased_resized_area */
extern T0* T2070f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_item */
extern T6 T2095f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_resize */
extern void T2095f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_modulus */
extern T6 T2095f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_capacity */
extern T6 T2095f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_put */
extern void T2095f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].force */
extern void T2068f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA].fill_with */
extern void T2070f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA].extend */
extern void T2070f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].search_position */
extern void T2095f36(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32].test */
extern T1 T625f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].unset_found_item */
extern void T2095f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item */
extern T0* T2095f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_item */
extern T0* T2095f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].has */
extern T1 T2095f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make */
extern T0* T2095c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_with_equality_testers */
extern void T2095f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_sparse_container */
extern void T2095f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].set_internal_cursor */
extern void T2095f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_cursor */
extern T0* T2095f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32].make */
extern T0* T2146c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_slots */
extern void T2095f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_clashes */
extern void T2095f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_key_storage */
extern void T2095f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_item_storage */
extern void T2095f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].make */
extern T0* T2068f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA].make_empty */
extern T0* T2070c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].default_create */
extern T0* T2068c3(GE_context* ac);
/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32].default_create */
extern T0* T625c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].make */
extern T0* T2087c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].set_internal_cursor */
extern void T2087f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].new_cursor */
extern T0* T2087f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE].make */
extern T0* T2141c3(GE_context* ac, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.to_array */
extern T0* T2073f6(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] LX_EQUIVALENCE_CLASSES.lower */
extern T6 T2073f5(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.upper */
extern T6 T2073f4(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.built */
extern T1 T2073f3(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.new_lower */
extern T6 T2073f7(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_eof_rules */
extern void T2042f43(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable LX_RULE].put */
extern void T2090f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [detachable LX_RULE].make_filled */
extern T0* T2090c4(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable LX_RULE].make_filled_area */
extern void T2090f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_RULE].make_filled */
extern T0* T2091c3(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_RULE].fill_with */
extern void T2091f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable LX_RULE].extend */
extern void T2091f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable LX_RULE].make_empty */
extern void T2091f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.build_rules */
extern void T2042f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] LX_RULE].put */
extern void T2083f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] LX_RULE].make_filled */
extern T0* T2083c4(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] LX_RULE].make_filled_area */
extern void T2083f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_RULE].make_filled */
extern T0* T2084c5(GE_context* ac, T0* a1, T6 a2);
/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE].make_empty_with_lower */
extern T0* T2089f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] LX_RULE].make_from_array */
extern T0* T2083c8(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] LX_RULE].set_area */
extern void T2083f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY [[attached] LX_RULE].new_rebase */
extern void T2142f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY [[attached] LX_RULE].make_empty */
extern T0* T2142c4(GE_context* ac);
/* [detachable] KL_ARRAY [[attached] LX_RULE].make_empty_area */
extern void T2142f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] LX_RULE].make_empty */
extern T0* T2083c7(GE_context* ac);
/* [detachable] ARRAY [[attached] LX_RULE].make_empty_area */
extern void T2083f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE].default_create */
extern T0* T2089c2(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.names */
extern T0* T2055f7(GE_context* ac, T0* C);
/* [detachable] LX_START_CONDITIONS.first */
extern T0* T2055f6(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8].make_empty_with_lower */
extern T0* T2106f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].make_from_array */
extern T0* T324c11(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].set_area */
extern void T324f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY [[attached] STRING_8].new_rebase */
extern void T2160f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY [[attached] STRING_8].make_empty */
extern T0* T2160c4(GE_context* ac);
/* [detachable] KL_ARRAY [[attached] STRING_8].make_empty_area */
extern void T2160f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].make_empty */
extern T0* T324c10(GE_context* ac);
/* [detachable] ARRAY [[attached] STRING_8].make_empty_area */
extern void T324f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.string_array_routines */
extern T0* T2055f11(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8].default_create */
extern T0* T2106c2(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.parse_string */
extern void T2040f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.parse */
extern void T2040f251(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_pop_last_value */
extern void T2040f264(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_push_error_value */
extern void T2040f263(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_do_action */
extern void T2040f262(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.twin */
extern T0* T2058f21(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.copy */
extern void T2058f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].twin */
extern T0* T748f7(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.chunk_of_ones */
extern T0* T2058f20(GE_context* ac, T0* C);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].make_filled */
extern T0* T2115c5(GE_context* ac, T0* a1, T6 a2);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].fill_with */
extern void T2115f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].extend */
extern void T2115f9(GE_context* ac, T0* C, T0* a1);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].make_empty */
extern void T2115f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.chunk_index */
extern T6 T2058f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.remove_symbol_class */
extern void T2058f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.add_symbol_class */
extern void T2058f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].filled_with */
extern T1 T748f6(GE_context* ac, T0* C, T11 a1, T6 a2, T6 a3);
/* [detachable] LX_SYMBOL_CLASS.attached_other_sets */
extern T0* T2058f19(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.append_character_set_to_character_class */
extern T0* T2040f188(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] LX_SYMBOL_CLASS.add_symbol */
extern void T2058f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.index_in_chunk */
extern T6 T2058f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.attached_chunk */
extern T0* T2058f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.masks */
extern T0* T2058f10(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_negative_range_in_character_class_error */
extern void T2040f304(GE_context* ac, T0* C);
/* [detachable] UT_ERROR_HANDLER.report_error */
extern void T134f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.report_error_message */
extern void T134f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.message */
extern T0* T134f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.make */
extern T0* T2103c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.filename */
extern T0* T2040f209(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].aliased_resized_area */
extern T0* T2066f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].aliased_resized_area */
extern T0* T2067f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_character_to_character_class */
extern T0* T2040f187(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] LX_WILDCARD_PARSER.new_character_class */
extern T0* T2040f186(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.make */
extern T0* T2058c25(GE_context* ac, T6 a1, T6 a2);
/* [detachable] LX_SYMBOL_CLASS.set_negated */
extern void T2058f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].force */
extern void T2061f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_put */
extern void T2061f47(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].force */
extern void T2064f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_32].fill_with */
extern void T2065f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_32].extend */
extern void T2065f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_put */
extern void T2061f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_put */
extern void T2061f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_item */
extern T6 T2061f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_item */
extern T6 T2061f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].hash_position */
extern T6 T2061f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.hash_code */
extern T6 T18f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].resize */
extern void T2061f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_resize */
extern void T2061f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].special_integer_ */
extern T0* T2061f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_resize */
extern void T2061f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].aliased_resized_area */
extern T0* T2064f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_32].aliased_resized_area */
extern T0* T2065f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_resize */
extern void T2061f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_item */
extern T0* T2061f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_resize */
extern void T2061f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_modulus */
extern T6 T2061f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_capacity */
extern T6 T2061f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_put */
extern void T2061f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].search_position */
extern void T2061f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32].test */
extern T1 T2062f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] STRING_32.is_equal */
extern T1 T18f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_32].same_items */
extern T1 T16f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].unset_found_item */
extern void T2061f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].force_new */
extern void T2060f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_put */
extern void T2060f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_put */
extern void T2060f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_put */
extern void T2060f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_item */
extern T6 T2060f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].hash_position */
extern T6 T2060f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.hash_code */
extern T6 T2058f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_item */
extern T6 T2060f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].resize */
extern void T2060f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_resize */
extern void T2060f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].special_integer_ */
extern T0* T2060f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_storage_resize */
extern void T2060f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_resize */
extern void T2060f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_storage_item */
extern T0* T2060f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_item */
extern T0* T2060f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_resize */
extern void T2060f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_modulus */
extern T6 T2060f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_capacity */
extern T6 T2060f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].unset_found_item */
extern void T2060f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].found_item */
extern T0* T2060f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].found */
extern T1 T2060f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].search */
extern void T2060f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].search_position */
extern void T2060f46(GE_context* ac, T0* C, T0* a1);
/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS].test */
extern T1 T2111f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] LX_SYMBOL_CLASS.is_equal */
extern T1 T2058f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.same_symbol_class */
extern T1 T2058f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].same_items */
extern T1 T748f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_equality_tester */
extern T0* T2060f27(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].force */
extern void T2066f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].fill_with */
extern void T2067f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].extend */
extern void T2067f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_character_class */
extern T0* T2040f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_symbol_class_nfa */
extern T0* T2040f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.make_symbol_class */
extern T0* T2069c12(GE_context* ac, T0* a1, T1 a2);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2123c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].force */
extern void T2060f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_utf8_character_class */
extern T0* T2040f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_unions_of_concatenations_of_symbol_classes */
extern T0* T2040f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].item */
extern T0* T2102f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].first */
extern T0* T2102f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].item */
extern T0* T2101f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].force_new */
extern void T2057f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_put */
extern void T2057f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_put */
extern void T2057f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].force */
extern void T2113f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].fill_with */
extern void T2110f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].extend */
extern void T2110f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_put */
extern void T2057f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_put */
extern void T2057f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_item */
extern T6 T2057f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].hash_position */
extern T6 T2057f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_item */
extern T6 T2057f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].resize */
extern void T2057f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_resize */
extern void T2057f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].special_integer_ */
extern T0* T2057f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_resize */
extern void T2057f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_resize */
extern void T2057f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].aliased_resized_area */
extern T0* T2113f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].aliased_resized_area */
extern T0* T2110f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_item */
extern T0* T2057f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_resize */
extern void T2057f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_modulus */
extern T6 T2057f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_capacity */
extern T6 T2057f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].unset_found_item */
extern void T2057f39(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].is_empty */
extern T1 T2101f6(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.append_concatenations_of_symbol_classes_from_utf8_n_byte_character_class */
extern void T2040f313(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_iterator */
extern T0* T724f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].is_equal */
extern T1 T724f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].same_equality_tester */
extern T1 T724f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].any_ */
extern T0* T724f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].force */
extern void T724f87(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].put */
extern void T2099f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].item */
extern T0* T2099f4(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled */
extern T0* T2099c5(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled_area */
extern void T2099f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled */
extern T0* T2100c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].fill_with */
extern void T2100f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].extend */
extern void T2100f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].make_empty */
extern void T2100f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_concatenations_of_symbol_classes_from_utf8_2_byte_character_class */
extern void T2040f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].put */
extern void T2097f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].item */
extern T0* T2097f4(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].make_filled */
extern T0* T2097c5(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].make_filled_area */
extern void T2097f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].make_filled */
extern T0* T2098c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].fill_with */
extern void T2098f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].extend */
extern void T2098f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].make_empty */
extern void T2098f6(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].force_last */
extern void T2101f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].force */
extern void T2157f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].fill_with */
extern void T2156f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].extend */
extern void T2156f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].resize */
extern void T2101f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].aliased_resized_area */
extern T0* T2157f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].aliased_resized_area */
extern T0* T2156f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].new_capacity */
extern T6 T2101f9(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].extendible */
extern T1 T2101f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].put_last */
extern void T2102f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2102c9(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2102f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2102f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2159c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2066f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].make_empty */
extern T0* T2067c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].default_create */
extern T0* T2066c3(GE_context* ac);
/* [detachable] LX_SYMBOL_CLASS.has */
extern T1 T2058f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.added */
extern T1 T2058f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.chunk */
extern T0* T2058f11(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2101c11(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].set_internal_cursor */
extern void T2101f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].new_cursor */
extern T0* T2101f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2158c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2157f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make_empty */
extern T0* T2156c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].default_create */
extern T0* T2157c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].found_item */
extern T0* T2057f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_item */
extern T0* T2057f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].found */
extern T1 T2057f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].search */
extern void T2057f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].search_position */
extern void T2057f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.question_character_class */
extern T0* T2040f183(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].force_new */
extern void T2061f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].found_item */
extern T0* T2061f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_item */
extern T0* T2061f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].found */
extern T1 T2061f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].search */
extern void T2061f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.build_optional */
extern void T2069f11(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.set_epsilon_transition */
extern void T2096f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.final_state */
extern T0* T2069f5(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_positive_closure */
extern void T2069f10(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_closure */
extern void T2069f9(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_character */
extern T0* T2040f182(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.new_symbol_nfa */
extern T0* T2040f239(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_NFA.make_symbol */
extern T0* T2069c13(GE_context* ac, T6 a1, T1 a2);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2124c4(GE_context* ac, T6 a1, T0* a2);
/* [detachable] STRING_32.append_integer */
extern void T18f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_character_to_string */
extern T0* T2040f181(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] LX_WILDCARD_PARSER.new_epsilon_nfa */
extern T0* T2040f179(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].item */
extern T1 T2063f5(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_concatenation */
extern void T2069f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].append_last */
extern void T2086f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].extend_last */
extern void T2086f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].forth */
extern void T2137f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_forth */
extern void T2086f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].add_traversing_cursor */
extern void T2086f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_next_cursor */
extern void T2137f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].remove_traversing_cursor */
extern void T2086f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].item */
extern T0* T2137f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_item */
extern T0* T2086f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].after */
extern T1 T2137f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].start */
extern void T2137f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_start */
extern void T2086f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_after */
extern void T2137f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].off */
extern T1 T2137f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].before */
extern T1 T2137f8(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_union */
extern void T2069f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.report_unrecognized_rule_error */
extern void T2040f279(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.make */
extern T0* T2074c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_rule */
extern void T2040f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.add_nfa_to_all */
extern void T2055f13(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITION.add_nfa */
extern void T2092f7(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITION.put_nfa */
extern void T2092f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].force_last */
extern void T2094f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].resize */
extern void T2094f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].new_capacity */
extern T6 T2094f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].extendible */
extern T1 T2094f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITION.put_bol_nfa */
extern void T2092f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.add_nfa_to_non_exclusive */
extern void T2055f15(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITIONS.is_empty */
extern T1 T2055f8(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_bol_needed */
extern void T2041f40(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_variable_trail_context */
extern void T2041f39(GE_context* ac, T0* C, T1 a1);
/* detachable LX_RULE.set_column_count */
extern void T2071f21(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_line_count */
extern void T2071f20(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_trail_count */
extern void T2071f19(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_head_count */
extern void T2071f18(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_trail_context */
extern void T2071f17(GE_context* ac, T0* C, T1 a1);
/* detachable LX_RULE.set_line_nb */
extern void T2071f16(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_pattern */
extern void T2071f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.set_accepted_rule */
extern void T2069f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA_STATE.set_accepted_rule */
extern void T2096f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.check_options */
extern void T2040f277(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_variable_trailing_context_error */
extern void T2040f296(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.make */
extern T0* T2081c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_reject_error */
extern void T2040f295(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.make */
extern T0* T2080c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_meta_equiv_classes_error */
extern void T2040f294(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.make */
extern T0* T2079c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.build_equiv_classes */
extern void T2040f276(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_equiv_classes */
extern void T2041f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.convert_to_equivalence */
extern void T2058f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.equivalence_class */
extern T6 T2073f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.is_representative */
extern T1 T2073f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.build */
extern void T2073f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].forth */
extern void T2072f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_forth */
extern void T2060f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].add_traversing_cursor */
extern void T2060f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_next_cursor */
extern void T2072f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_position */
extern void T2072f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].remove_traversing_cursor */
extern void T2060f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_after */
extern void T2072f12(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.add */
extern void T2073f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].forget_right */
extern void T2128f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].forget_left */
extern void T2128f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put_right */
extern void T2128f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].attach_left */
extern void T2128f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].item */
extern T0* T2072f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_item */
extern T0* T2060f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].after */
extern T1 T2072f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].start */
extern void T2072f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_start */
extern void T2060f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_off */
extern T1 T2060f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].off */
extern T1 T2072f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].is_empty */
extern T1 T2060f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2060f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2072c9(GE_context* ac, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.make */
extern T0* T2073c10(GE_context* ac, T6 a1, T6 a2);
/* [detachable] LX_EQUIVALENCE_CLASSES.initialize */
extern void T2073f13(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.set_maximum_symbol_equivalence_class */
extern void T2040f275(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.set_maximum_symbol_equivalence_class */
extern void T2058f32(GE_context* ac, T0* C, T6 a1);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].upper */
extern T6 T2115f4(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_maximum_symbol */
extern void T2041f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_push_last_value */
extern void T2040f261(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.accept */
extern void T2040f260(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_do_error_action */
extern void T2040f258(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.report_error */
extern void T2040f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.report_eof_expected_error */
extern void T2040f273(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.read_token */
extern void T2040f257(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_execute_action */
extern void T2040f272(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.character_text */
extern T0* T2040f208(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.unicode_text_item */
extern T3 T2040f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.unicode_item */
extern T3 T1429f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_CHARACTER_BUFFER.unicode_item */
extern T3 T1396f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.text */
extern T0* T2040f241(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.substring */
extern T0* T1429f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.append_substring_to_string */
extern void T1429f21(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.item */
extern T2 T1429f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.text_count */
extern T6 T2040f227(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_bad_character_error */
extern void T2040f293(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_ERROR.make */
extern T0* T2078c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.report_bad_character_class_error */
extern void T2040f292(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.make */
extern T0* T2077c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_unicode_character */
extern void T2040f291(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_character */
extern void T2040f299(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.report_invalid_character_error */
extern void T2040f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.make */
extern T0* T2075c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.report_character_out_of_range_error */
extern void T2040f301(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.make */
extern T0* T2082c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.process_hexadecimal_character */
extern void T2040f290(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_octal_character */
extern void T2040f289(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_escaped_character */
extern void T2040f288(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_missing_quote_error */
extern void T2040f287(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.make */
extern T0* T2076c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_single_character */
extern void T2040f286(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.text_item */
extern T2 T2040f201(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.less */
extern void T2040f285(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_set_line_column */
extern void T2040f298(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.character_class_with_name */
extern T0* T2040f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.unicode_text */
extern T0* T2040f194(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.unicode_substring */
extern T0* T1429f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.unicode_substring */
extern T0* T1396f10(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.append_unicode_substring_to_string */
extern void T1396f19(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] STRING_32.append_substring_general */
extern void T18f40(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] LX_WILDCARD_PARSER.set_start_condition */
extern void T2040f284(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_execute_eof_action */
extern void T2040f271(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.terminate */
extern void T2040f283(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.wrap */
extern T1 T2040f165(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_refill_input_buffer */
extern void T2040f270(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_set_content */
extern void T2040f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.count */
extern T6 T1429f6(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.as_special */
extern T0* T1429f12(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_BUFFER.fill */
extern void T1585f15(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_BUFFER.set_index */
extern void T1585f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_null_trans_state */
extern T6 T2040f164(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_previous_state */
extern T6 T2040f163(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.fatal_error */
extern void T2040f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_character */
extern void T136f9(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_STANDARD_FILES.error */
extern T0* T135f1(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.make */
extern T0* T136c5(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.std */
extern T0* T2040f189(GE_context* ac, T0* C);
/* [detachable] KL_STANDARD_FILES.default_create */
extern T0* T135c4(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.report_invalid_unicode_character_error */
extern void T2040f268(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.item_code */
extern T10 T1429f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_at_beginning_of_line */
extern T6 T2040f132(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.special_integer_ */
extern T0* T2040f129(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_init_value_stacks */
extern void T2040f256(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_clear_all */
extern void T2040f265(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.clear_all */
extern void T2040f280(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.clear_stacks */
extern void T2040f297(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_clear_value_stacks */
extern void T2040f300(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] LX_NFA].keep_head */
extern void T2070f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA].set_count */
extern void T2070f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].keep_head */
extern void T2067f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].set_count */
extern void T2067f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].keep_head */
extern void T2065f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].set_count */
extern void T2065f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.abort */
extern void T2040f259(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.set_input_buffer */
extern void T2040f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_load_input_buffer */
extern void T2040f255(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_BUFFER.set_position */
extern void T1585f14(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] LX_WILDCARD_PARSER.new_unicode_string_buffer */
extern T0* T2040f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_BUFFER.make */
extern T0* T1585c12(GE_context* ac, T0* a1);
/* [detachable] YY_UNICODE_BUFFER.make_from_buffer */
extern void T1585f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.put */
extern void T1429f17(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_string */
extern void T1429f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_substring */
extern void T1429f23(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] STRING_32.put */
extern void T18f36(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] STRING_32.subcopy */
extern void T18f37(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] YY_UNICODE_BUFFER.new_default_buffer */
extern T0* T1585f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.make */
extern T0* T1429c15(GE_context* ac, T6 a1);
/* [detachable] STRING_32.make_filled */
extern T0* T18c35(GE_context* ac, T3 a1, T6 a2);
/* [detachable] STRING_32.fill_character */
extern void T18f38(GE_context* ac, T0* C, T3 a1);
/* [detachable] LX_WILDCARD_PARSER.make_from_description */
extern T0* T2040c246(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make */
extern T0* T2057c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_with_equality_testers */
extern void T2057f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_sparse_container */
extern void T2057f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2057f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2057f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make */
extern T0* T2114c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_slots */
extern void T2057f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_clashes */
extern void T2057f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_key_storage */
extern void T2057f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_item_storage */
extern void T2057f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].make */
extern T0* T2113f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].make_empty */
extern T0* T2110c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].default_create */
extern T0* T2113c3(GE_context* ac);
/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS].default_create */
extern T0* T2111c2(GE_context* ac);
/* [detachable] LX_ACTION_FACTORY.make */
extern T0* T2056c1(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.make */
extern void T2055f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.make */
extern T0* T2055c12(GE_context* ac, T6 a1);
/* [detachable] LX_START_CONDITIONS.set_internal_cursor */
extern void T2055f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.new_cursor */
extern T0* T2055f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION].make */
extern T0* T2109c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].make */
extern T0* T2108f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].make_empty */
extern T0* T2107c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].default_create */
extern T0* T2108c2(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.make_parser_skeleton */
extern void T2040f249(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_build_parser_tables */
extern void T2040f254(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yycheck_template */
extern T0* T2040f178(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yyfixed_array */
extern T0* T2040f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yytable_template */
extern T0* T2040f177(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yypgoto_template */
extern T0* T2040f176(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yypact_template */
extern T0* T2040f174(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yydefgoto_template */
extern T0* T2040f173(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yydefact_template */
extern T0* T2040f172(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytypes2_template */
extern T0* T2040f171(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytypes1_template */
extern T0* T2040f169(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yyr1_template */
extern T0* T2040f160(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template */
extern T0* T2040f159(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template_2 */
extern void T2040f303(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yyarray_subcopy */
extern void T2040f314(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] LX_WILDCARD_PARSER.integer_array_ */
extern T0* T2040f245(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template_1 */
extern void T2040f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_create_value_stacks */
extern void T2040f253(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].make */
extern T0* T2064f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].make_empty */
extern T0* T2065c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].default_create */
extern T0* T2064c3(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.make_lex_scanner_from_description */
extern void T2040f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].force */
extern void T2063f9(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].resize */
extern void T2063f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].new_capacity */
extern T6 T2063f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].extendible */
extern T1 T2063f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].make */
extern T0* T2063c8(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make */
extern void T2060f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2060c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2060f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_slots */
extern void T2060f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_clashes */
extern void T2060f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_key_storage */
extern void T2060f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_item_storage */
extern void T2060f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].set_key_equality_tester */
extern void T2061f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.string_32_equality_tester */
extern T0* T2040f143(GE_context* ac, T0* C);
/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32].default_create */
extern T0* T2062c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_map */
extern T0* T2061c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_with_equality_testers */
extern void T2061f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_sparse_container */
extern void T2061f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].set_internal_cursor */
extern void T2061f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_cursor */
extern T0* T2061f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make */
extern T0* T2119c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_slots */
extern void T2061f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_clashes */
extern void T2061f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_key_storage */
extern void T2061f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_item_storage */
extern void T2061f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_equal */
extern T0* T2060c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].set_key_equality_tester */
extern void T2059f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.string_equality_tester */
extern T0* T2040f126(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_map */
extern T0* T2059c20(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_with_equality_testers */
extern void T2059f22(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_sparse_container */
extern void T2059f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].set_internal_cursor */
extern void T2059f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].new_cursor */
extern T0* T2059f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8].make */
extern T0* T2117c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].unset_found_item */
extern void T2059f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_slots */
extern void T2059f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].new_modulus */
extern T6 T2059f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_clashes */
extern void T2059f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_key_storage */
extern void T2059f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_item_storage */
extern void T2059f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.make_with_buffer */
extern void T2040f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_initialize */
extern void T2040f266(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_initialize */
extern void T2040f266p1(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_build_tables */
extern void T2040f281(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_accept_template */
extern T0* T2040f218(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_fixed_array */
extern T0* T2040f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_meta_template */
extern T0* T2040f217(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_ec_template */
extern T0* T2040f216(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_ec_template_1 */
extern void T2040f311(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_array_subcopy */
extern void T2040f315(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] LX_WILDCARD_PARSER.yy_def_template */
extern T0* T2040f215(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_base_template */
extern T0* T2040f214(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template */
extern T0* T2040f213(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_3 */
extern void T2040f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_2 */
extern void T2040f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_1 */
extern void T2040f308(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template */
extern T0* T2040f212(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_3 */
extern void T2040f307(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_2 */
extern void T2040f306(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_1 */
extern void T2040f305(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.empty_buffer */
extern T0* T2040f139(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_case_insensitive */
extern void T2041f36(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_full_table */
extern void T2041f35(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_meta_equiv_classes_used */
extern void T2041f34(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_equiv_classes_used */
extern void T2041f33(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.make */
extern T0* T2041c32(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.make_with_initial */
extern T0* T2055c14(GE_context* ac, T6 a1);
/* [detachable] LX_START_CONDITIONS.put_first */
extern void T2055f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.put */
extern void T2055f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] LX_START_CONDITIONS.move_cursors_right */
extern void T2055f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION].set_position */
extern void T2109f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.move_right */
extern void T2055f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].force */
extern void T2108f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].fill_with */
extern void T2107f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].extend */
extern void T2107f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.put_last */
extern void T2055f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITION.make */
extern T0* T2092c6(GE_context* ac, T0* a1, T6 a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].make */
extern T0* T2094c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].set_internal_cursor */
extern void T2094f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].new_cursor */
extern T0* T2094f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA].make */
extern T0* T2144c3(GE_context* ac, T0* a1);
/* [detachable] UT_ERROR_HANDLER.make_null */
extern T0* T134c6(GE_context* ac);
/* [detachable] UT_ERROR_HANDLER.null_output_stream */
extern T0* T134f4(GE_context* ac, T0* C);
/* [detachable] LX_DFA_WILDCARD.wipe_out */
extern void T2014f22(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.replace_backslashes */
extern void T1706f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.put */
extern void T743f79(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_relative_pathname */
extern T1 T124f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.file_system */
extern T0* T1706f13(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.unix_file_system */
extern T0* T1706f15(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.windows_file_system */
extern T0* T1706f14(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.execution_environment */
extern T0* T1706f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make */
extern void T143f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make */
extern T0* T143c21(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T143f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].new_cursor */
extern T0* T143f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T486c8(GE_context* ac, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.operating_system */
extern T0* T1706f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove */
extern void T143f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_left */
extern void T143f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_cursors_left */
extern void T143f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_position */
extern void T486f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove_last */
extern void T143f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_last_cursors_after */
extern void T143f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T486f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_after */
extern void T486f9(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_lower */
extern T0* T743f9(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_lower */
extern void T743f84(GE_context* ac, T0* C);
/* [detachable] STRING_8.as_lower */
extern T0* T17f30(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_lower */
extern void T17f71(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_lower_area */
extern void T17f74(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_EIFFEL_PREPARSER.current_system */
extern T0* T447f62(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make_from_linear */
extern T0* T143c37(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].forth */
extern void T489f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_forth */
extern void T148f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T148f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T489f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T148f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].set */
extern void T489f9(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].item */
extern T0* T489f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].start */
extern void T489f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_start */
extern void T148f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_off */
extern T1 T148f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T489f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].new_cursor */
extern T0* T148f9(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T489c8(GE_context* ac, T0* a1);
/* [detachable] ST_SPLITTER.split */
extern T0* T1898f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_SPLITTER.do_split */
extern T0* T1898f7(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].force_last */
extern void T148f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] STRING_8].put_right */
extern void T490f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] STRING_8].make */
extern T0* T490c3(GE_context* ac, T0* a1);
/* [detachable] ST_SPLITTER.string_ */
extern T0* T1898f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].make */
extern T0* T148c17(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T148f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_SPLITTER.set_separators */
extern void T1898f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].put */
extern void T724f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_SPLITTER.make */
extern T0* T1898c11(GE_context* ac);
/* [detachable] KL_TEXT_INPUT_FILE.read_string */
extern void T32f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_read_to_string */
extern T6 T32f36(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_TEXT_INPUT_FILE.file_gss */
extern T6 T32f44(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].item_address */
extern T14 T15f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_end_of_file */
extern T1 T32f25(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_feof */
extern T1 T32f31(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_EIFFEL_PREPARSER.build_provider_constraint */
extern void T447f184(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_EIFFEL_PREPARSER.build_provider_constraint */
extern T0* T447f184ac1(GE_context* ac, T0* a1, T0* a2);
extern T1 T447f184ot1(T0* a1);
/* detachable ET_ECF_CLUSTER.set_provider_constraint */
extern void T776f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_provider_constraint */
extern void T761f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.universes_do_all */
extern void T183f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.universes_do_all */
extern void T206f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.universes_do_all */
extern void T184f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.universes_do_all */
extern void T56f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.add_universe_full_name */
extern void T447f190(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].force_last */
extern void T143f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].resize */
extern void T143f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].new_capacity */
extern T6 T143f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].extendible */
extern T1 T143f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcaaa_error */
extern void T51f551(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_cluster_error */
extern void T51f353(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTER_ERROR.make_gcaaa */
extern T0* T1887c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcaaa_error */
extern T1 T51f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcaaa_error */
extern void T35f551(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_cluster_error */
extern void T35f353(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gcaaa_error */
extern T1 T35f175(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.add_recursive_cluster */
extern void T776f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.add_subcluster */
extern void T776f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_parent */
extern void T776f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.put_last */
extern void T761f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].force_last */
extern void T1527f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].force */
extern void T1805f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].fill_with */
extern void T1804f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].extend */
extern void T1804f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].resize */
extern void T1527f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].aliased_resized_area */
extern T0* T1805f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].aliased_resized_area */
extern T0* T1804f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].new_capacity */
extern T6 T1527f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].extendible */
extern T1 T1527f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.set_subclusters */
extern void T776f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_parent */
extern void T761f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.make_empty */
extern T0* T761c9(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].make */
extern T0* T1527c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].set_internal_cursor */
extern void T1527f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].new_cursor */
extern T0* T1527f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].make */
extern T0* T1806c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].make */
extern T0* T1805f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].make_empty */
extern T0* T1804c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].default_create */
extern T0* T1805c3(GE_context* ac);
/* detachable ET_ECF_CLUSTER.new_recursive_cluster */
extern T0* T776f61(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_overridden_group */
extern void T776f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_provider_groups */
extern void T776f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_visible_classes */
extern void T776f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_options */
extern void T776f78(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_options */
extern void T776f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_mappings */
extern void T776f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_renamings */
extern void T776f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_classname_prefix */
extern void T776f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_use_obsolete_syntax */
extern void T776f88(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_hidden */
extern void T776f71(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_read_only */
extern void T776f73(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_override */
extern void T776f85(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_implicit */
extern void T776f103(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_recursive */
extern void T776f74(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_file_rules */
extern void T776f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.make_relative */
extern T0* T776c101(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] UT_VERSION.is_less_equal */
extern T1 T44f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.ecf_1_4_0 */
extern T0* T776f37(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.set_scm_mapping_constraint_enabled */
extern void T776f87(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTERS.set_scm_mapping_constraint_enabled */
extern void T761f10(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.string_ */
extern T0* T776f54(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.directory_exists */
extern T1 T124f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.exists */
extern T1 T327f29(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.tmp_directory */
extern T0* T124f30(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.directory_exists */
extern T1 T123f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.tmp_directory */
extern T0* T123f33(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_valid_directory_name */
extern T1 T776f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.is_included */
extern T1 T764f3(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULES.is_included */
extern T0* T764f3ac1(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]].set_rout_disp_final */
extern void T2006f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_FILE_RULE.is_included */
extern T1 T784f10(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULE.is_included */
extern T0* T784f10ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_FILE_RULE.is_included */
extern T0* T784f10ac2(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]].set_rout_disp_final */
extern void T2048f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.matches */
extern T1 T83f306(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].there_exists */
extern T1 T1541f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].item */
extern T0* T1541f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].for_all */
extern T1 T1530f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].item */
extern T0* T1530f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.implicit_relative_name */
extern T0* T776f52(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.unix_file_system */
extern T0* T776f51(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_valid_directory_name */
extern T1 T776f40p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_file */
extern void T447f183(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcaab_error */
extern void T51f245(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLUSTER_ERROR.make_gcaab */
extern T0* T1887c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcaab_error */
extern T1 T51f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcaab_error */
extern void T35f245(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gcaab_error */
extern T1 T35f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T447f189(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_end_of_file */
extern void T908f28(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T447f189p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T447f189p0(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T447f189p2(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_syntax_error */
extern void T51f260(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYNTAX_ERROR.make */
extern T0* T1888c18(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYNTAX_ERROR.set_compilers */
extern void T1888f19(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.report_syntax_error */
extern void T35f260(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.current_position */
extern T0* T447f61(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.make */
extern T0* T1515c12(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_FILE_POSITION.make_compressed_position */
extern void T1515f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_FILE_POSITION.set_position */
extern void T1515f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PREPARSER.read_token */
extern void T447f188(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_execute_action */
extern void T447f195(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_silu_error */
extern void T51f259(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_silu_error */
extern void T35f259(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sifu_error */
extern void T51f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sifu_error */
extern void T35f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssnp_error */
extern void T51f247(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssnp_error */
extern void T35f247(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssnl_error */
extern void T51f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssnl_error */
extern void T35f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sssc_error */
extern void T51f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sssc_error */
extern void T35f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.more */
extern void T447f201(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.is_verbatim_string_closer */
extern T1 T447f116(GE_context* ac, T0* C, T6 a1, T6 a2);
/* CHARACTER_32.plus */
extern T3 T3f5(GE_context* ac, T3* C, T10 a1);
/* [detachable] ET_EIFFEL_PREPARSER.report_syntax_error */
extern void T447f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.set_start_condition */
extern void T447f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.unicode_text_item */
extern T3 T447f113(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.utf8_text_substring */
extern T0* T447f111(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].force_new */
extern void T1692f41(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_put */
extern void T1692f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_put */
extern void T1692f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_put */
extern void T1692f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_put */
extern void T1692f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_item */
extern T6 T1692f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].hash_position */
extern T6 T1692f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_item */
extern T6 T1692f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].resize */
extern void T1692f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_resize */
extern void T1692f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].special_integer_ */
extern T0* T1692f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_resize */
extern void T1692f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_resize */
extern void T1692f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_item */
extern T0* T1692f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_resize */
extern void T1692f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_modulus */
extern T6 T1692f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_capacity */
extern T6 T1692f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].unset_found_item */
extern void T1692f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found_key */
extern T0* T1692f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found */
extern T1 T1692f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].search */
extern void T1692f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].search_position */
extern void T1692f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.strings */
extern T0* T447f134(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_key_equality_tester */
extern void T1692f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.string_equality_tester */
extern T0* T447f135(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_map */
extern T0* T1692c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_with_equality_testers */
extern void T1692f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_sparse_container */
extern void T1692f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_internal_cursor */
extern void T1692f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_cursor */
extern T0* T1692f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8].make */
extern T0* T1930c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_slots */
extern void T1692f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_clashes */
extern void T1692f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_key_storage */
extern void T1692f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_item_storage */
extern void T1692f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.append_utf8_text_substring_to_string */
extern void T447f202(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_CHARACTER_BUFFER.append_utf8_substring_to_string */
extern void T1396f16(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PREPARSER.string_ */
extern T0* T447f133(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.string_buffer */
extern T0* T447f132(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sceq_error */
extern void T51f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sceq_error */
extern void T35f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scqq_error */
extern void T51f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scqq_error */
extern void T35f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scac_error */
extern void T51f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scac_error */
extern void T35f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scas_error */
extern void T51f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scas_error */
extern void T35f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.set_syntax_error */
extern void T447f198(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.text_count */
extern T6 T447f109(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_reference_keyword */
extern T1 T447f108(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_note_keyword */
extern T1 T447f107(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_detachable_keyword */
extern T1 T447f106(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_attribute_keyword */
extern T1 T447f105(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_identifier */
extern T0* T458f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.set_position */
extern void T180f114(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PREPARSER.last_identifier */
extern T0* T447f131(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].replace_found_item */
extern void T1692f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found_item */
extern T6 T1692f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_item */
extern T6 T1692f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.append_text_substring_to_string */
extern void T447f203(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_CHARACTER_BUFFER.append_substring_to_string */
extern void T1396f18(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] STRING_8.append_substring */
extern void T17f80(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_EIFFEL_PARSER.last_identifier */
extern T0* T449f987(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.strings */
extern T0* T449f964(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.tokens */
extern T0* T449f656(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.string_equality_tester */
extern T0* T449f973(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.append_text_substring_to_string */
extern void T449f2206(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.string_ */
extern T0* T449f963(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.string_buffer */
extern T0* T449f962(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.ast_factory */
extern T0* T447f104(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_attached_keyword */
extern T1 T447f100(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_set_line_column */
extern void T447f200(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_execute_eof_action */
extern void T447f194(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sseq_error */
extern void T51f246(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sseq_error */
extern void T35f246(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.terminate */
extern void T447f197(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.wrap */
extern T1 T447f92(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_refill_input_buffer */
extern void T447f193(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_set_content */
extern void T447f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_null_trans_state */
extern T6 T447f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_previous_state */
extern T6 T447f90(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.fatal_error */
extern void T447f192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.report_invalid_unicode_character_error */
extern void T447f191(GE_context* ac, T0* C, T10 a1);
/* [detachable] ET_EIFFEL_PREPARSER.special_integer_ */
extern T0* T447f60(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_at_beginning_of_line */
extern T6 T447f72(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_load_input_buffer */
extern void T447f140(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.tmp_file */
extern T0* T447f58(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_closed */
extern T1 T32f17(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.shared_file */
extern T0* T447f129(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_class */
extern void T447f187(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_MASTER_CLASS.add_last_local_class */
extern void T560f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_non_override_class */
extern void T560f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].force_last */
extern void T455f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].force */
extern void T1305f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS].fill_with */
extern void T1304f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS].extend */
extern void T1304f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].resize */
extern void T455f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].aliased_resized_area */
extern T0* T1305f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLASS].aliased_resized_area */
extern T0* T1304f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].new_capacity */
extern T6 T455f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].extendible */
extern T1 T455f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].make */
extern T0* T455c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].set_internal_cursor */
extern void T455f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].new_cursor */
extern T0* T455f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].make */
extern T0* T1306c9(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].make */
extern T0* T1305f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS].make_empty */
extern T0* T1304c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].default_create */
extern T0* T1305c3(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.add_last_local_override_class */
extern void T560f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_ignored_class */
extern void T560f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_time_stamp */
extern void T73f318(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.set_filename */
extern void T73f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.register_class */
extern void T56f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_id */
extern void T73f263(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYSTEM.register_class */
extern void T36f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.basename */
extern T0* T124f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_root_directory */
extern T1 T124f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.basename */
extern T0* T123f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_root_directory */
extern T1 T123f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.tokens */
extern T0* T447f93(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.same_pathnames */
extern T1 T124f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_PATHNAME.same_pathname */
extern T1 T475f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.is_subpathname */
extern T1 T475f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.string_ */
extern T0* T475f18(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.string_equality_tester */
extern T0* T475f19(GE_context* ac, T0* C);
/* [detachable] UC_STRING_EQUALITY_TESTER.default_create */
extern T0* T871c3(GE_context* ac);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.same_pathnames */
extern T1 T123f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_PATHNAME.same_case_insensitive */
extern T1 T475f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.is_case_insensitive_subpathname */
extern T1 T475f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.case_insensitive_string_equality_tester */
extern T0* T475f17(GE_context* ac, T0* C);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.default_create */
extern T0* T171c3(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.local_classes_do_if */
extern void T560f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_if */
extern void T560f88(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_if */
extern void T560f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.current_universe */
extern T0* T447f57(GE_context* ac, T0* C);
/* detachable DS_CELL [detachable ET_CLASS].put */
extern void T1997f3(GE_context* ac, T0* C, T0* a1);
/* detachable DS_CELL [detachable ET_CLASS].make */
extern T0* T1997c2(GE_context* ac, T0* a1);
/* detachable ET_ECF_CLUSTER.is_valid_eiffel_filename */
extern T1 T776f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.is_valid_eiffel_filename */
extern T1 T776f39p1(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.has_eiffel_extension */
extern T1 T776f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.operating_system */
extern T0* T776f59(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.reset */
extern void T327f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.tmp_directory */
extern T0* T447f56(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.shared_directory */
extern T0* T447f127(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.file_system */
extern T0* T447f55(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.unix_file_system */
extern T0* T447f126(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.windows_file_system */
extern T0* T447f125(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.operating_system */
extern T0* T447f124(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.full_pathname */
extern T0* T776f38(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.dirname */
extern T0* T124f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.dirname */
extern T0* T123f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.value */
extern T0* T55f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname_from_file_system */
extern T0* T124f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.any_ */
extern T0* T124f17(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname_from_file_system */
extern T0* T123f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.any_ */
extern T0* T123f19(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.windows_file_system */
extern T0* T776f50(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.file_system */
extern T0* T776f49(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.interpreted_string */
extern T0* T60f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VARIABLES.expanded_string */
extern T0* T60f9(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_VARIABLES.value */
extern T0* T60f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VARIABLES.string_ */
extern T0* T60f5(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.has */
extern T1 T743f66(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.index_of */
extern T6 T743f21(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.index_of_item_code */
extern T6 T743f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PREPARSER.execution_environment */
extern T0* T447f112(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.error_handler */
extern T0* T447f103(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.set_preparsed */
extern void T776f99(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLUSTERS.do_all */
extern void T554f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].item */
extern T0* T1342f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_preparsing_status */
extern void T51f516(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_preparsing_status */
extern void T35f516(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.preparse */
extern void T826f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.preparse */
extern void T826f168p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T826f168p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T826f168p1ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T826f168p1ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T826f168p1ac4(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.preparse */
extern void T207f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.preparse */
extern void T586f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.put_last */
extern void T585f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].force_last */
extern void T1356f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].force */
extern void T1553f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].fill_with */
extern void T1554f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].extend */
extern void T1554f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].resize */
extern void T1356f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1553f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1554f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].new_capacity */
extern T6 T1356f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].extendible */
extern T1 T1356f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T818f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.is_consumable */
extern T1 T207f92(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.is_consumable */
extern T1 T586f92(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].force_last */
extern void T816f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_put */
extern void T816f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_put */
extern void T816f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_item */
extern T6 T816f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].hash_position */
extern T6 T816f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].resize */
extern void T816f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_resize */
extern void T816f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].special_integer_ */
extern T0* T816f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_storage_resize */
extern void T816f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_resize */
extern void T816f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_storage_item */
extern T0* T816f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_item */
extern T0* T816f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_item */
extern T6 T816f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_resize */
extern void T816f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_modulus */
extern T6 T816f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_capacity */
extern T6 T816f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_put */
extern void T816f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].search_position */
extern void T816f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_equality_tester */
extern T0* T816f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].unset_found_item */
extern void T816f37(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern void T56f179(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac4(GE_context* ac);
/* Creation of agent #5 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac5(GE_context* ac);
/* Creation of agent #6 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac6(GE_context* ac);
/* Creation of agent #7 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac7(GE_context* ac);
/* Creation of agent #8 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac8(GE_context* ac);
/* Creation of agent #9 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac9(GE_context* ac);
/* Creation of agent #10 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac10(GE_context* ac);
/* Creation of agent #11 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac11(GE_context* ac);
/* Creation of agent #12 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac12(GE_context* ac);
/* Creation of agent #13 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac13(GE_context* ac);
/* Creation of agent #14 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac14(GE_context* ac);
/* Creation of agent #15 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac15(GE_context* ac);
/* Creation of agent #16 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac16(GE_context* ac);
/* Creation of agent #17 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac17(GE_context* ac);
/* Creation of agent #18 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac18(GE_context* ac);
/* Creation of agent #19 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac19(GE_context* ac);
/* Creation of agent #20 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac20(GE_context* ac);
/* Creation of agent #21 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac21(GE_context* ac);
/* Creation of agent #22 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac22(GE_context* ac);
/* Creation of agent #23 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac23(GE_context* ac, T0* a1);
/* Creation of agent #24 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac24(GE_context* ac);
/* Creation of agent #25 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac25(GE_context* ac, T0* a1);
/* Creation of agent #26 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac26(GE_context* ac);
/* Creation of agent #27 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac27(GE_context* ac, T0* a1);
/* Creation of agent #28 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac28(GE_context* ac);
/* Creation of agent #29 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac29(GE_context* ac, T0* a1);
/* Creation of agent #30 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac30(GE_context* ac);
/* Creation of agent #31 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T56f179ac31(GE_context* ac, T1 a1);
/* [detachable] ET_MASTER_CLASS.set_modified */
extern void T560f62(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.process_class */
extern void T843f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.internal_process_class */
extern void T843f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.check_providers_validity */
extern void T843f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].forth */
extern void T1568f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_forth */
extern void T926f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].add_traversing_cursor */
extern void T926f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_next_cursor */
extern void T1568f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_position */
extern void T1568f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].remove_traversing_cursor */
extern void T926f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_after */
extern void T1568f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].go_after */
extern void T1568f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_go_after */
extern void T926f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_off */
extern T1 T926f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].off */
extern T1 T1568f6(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_interface_error */
extern T1 T560f41(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.interface_checked */
extern T1 T560f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].item */
extern T0* T1568f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_item */
extern T0* T926f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].after */
extern T1 T1568f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].start */
extern void T1568f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_start */
extern void T926f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].is_empty */
extern T1 T926f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_cursor */
extern T0* T926f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].make */
extern T0* T1568c9(GE_context* ac, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.check_suppliers_validity */
extern void T843f9(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.parent */
extern T0* T558f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENT_LIST.item */
extern T0* T558f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.parents */
extern T0* T73f103(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.implicit_parent_clause */
extern T0* T73f131(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_system_object_class */
extern T1 T73f91(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_dotnet */
extern T1 T73f92(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_any_class */
extern T1 T73f104(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.item */
extern T0* T1206f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.parents_count */
extern T6 T73f102(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_explicit_conforming_parent */
extern T1 T73f130(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.has_conforming_parent */
extern T1 T1206f5(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.has_conforming_parent */
extern T1 T558f10(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.set_fatal_error */
extern void T843f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.error_handler */
extern T0* T843f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_implementation_error */
extern void T73f295(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.process_class */
extern void T842f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.internal_process_class */
extern void T842f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_parents_validity */
extern void T842f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_type_validity */
extern void T1567f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_formal_parameters_validity */
extern void T842f12(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.constraint */
extern T0* T867f26(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_qualified_anchored_signatures_validity */
extern void T842f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_signatures_validity */
extern void T1566f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_feature_signatures_validity */
extern void T1566f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROCEDURE_LIST.item */
extern T0* T254f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUERY_LIST.item */
extern T0* T253f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.set_fatal_error */
extern void T842f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.error_handler */
extern T0* T842f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_interface_error */
extern void T73f294(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.process_class */
extern void T841f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.internal_process_class */
extern void T841f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_signatures_validity */
extern void T841f12(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_feature_signatures_validity */
extern void T841f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_type_validity */
extern void T1565f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_parents_validity */
extern void T841f11(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_formal_parameters_validity */
extern void T841f10(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.set_fatal_error */
extern void T841f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.error_handler */
extern T0* T841f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_flattening_error */
extern void T73f293(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.process_class */
extern void T840f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.internal_process_class */
extern void T840f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.check_parents_validity */
extern void T840f11(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_type_validity */
extern void T1564f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.check_formal_parameters_validity */
extern void T840f10(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.set_fatal_error */
extern void T840f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.error_handler */
extern T0* T840f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_ancestors_error */
extern void T73f292(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.not_implementation_checked */
extern T1 T73f172(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_interface_checked */
extern void T73f289(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset_after_interface_checked */
extern void T254f13(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset_after_interface_checked */
extern void T253f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_features_flattened */
extern void T73f288(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_constraint_creation_procedures */
extern void T593f36(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_creation_procedures */
extern void T1450f85(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.reset */
extern void T1054f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_constraint_creation_procedures */
extern void T867f82(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.creation_procedures */
extern T0* T867f27(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_constraint_renames */
extern void T593f35(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_renames */
extern void T1450f84(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_constraint_renames */
extern void T867f81(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset_after_features_flattened */
extern void T254f12(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset_after_features_flattened */
extern void T253f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.features_flattened */
extern T1 T73f107(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_flattening_error */
extern T1 T73f108(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_ancestors_built */
extern void T73f287(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_type */
extern void T1278f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.make */
extern T0* T1278c6(GE_context* ac, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.make */
extern void T1278f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.tokens */
extern T0* T1278f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.ancestors_built */
extern T1 T73f99(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_ancestors_error */
extern T1 T73f106(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.local_classes_do_unless_actual */
extern void T560f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_unless_actual */
extern void T560f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_unless_actual */
extern void T560f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.make */
extern T0* T843c5(GE_context* ac, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.tokens */
extern T0* T843f3(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.make */
extern T0* T842c7(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.make */
extern T0* T1567c2(GE_context* ac);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.make */
extern T0* T1566c5(GE_context* ac, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.tokens */
extern T0* T1566f4(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.make */
extern void T842f7p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.tokens */
extern T0* T842f5(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.make */
extern T0* T841c6(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.make */
extern T0* T1565c2(GE_context* ac);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.make */
extern void T841f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.tokens */
extern T0* T841f4(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.make */
extern T0* T840c6(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.make */
extern T0* T1564c2(GE_context* ac);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.make */
extern void T840f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.tokens */
extern T0* T840f4(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_do_if_recursive */
extern void T56f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_if_recursive */
extern T0* T56f193ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_recursive */
extern void T56f192(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive */
extern T0* T56f192ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T1 T56f140(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T56f140ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T56f140ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T56f140ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T56f140ac4(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE].set_rout_disp_final */
extern void T1403f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]].set_rout_disp_final */
extern void T1402f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] KL_AGENT_ROUTINES [[attached] ANY].call */
extern void T1399f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_TRISTATE.set_true */
extern void T1398f4(GE_context* ac, T0* C);
/* [detachable] UT_TRISTATE.is_true */
extern T1 T1398f2(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern void T56f252(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern T0* T56f252ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern T0* T56f252ac2(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM.any_actions */
extern T0* T56f135(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ANY].default_create */
extern T0* T1399c1(GE_context* ac);
/* [detachable] UT_TRISTATE.make_false */
extern T0* T1398c3(GE_context* ac);
/* [detachable] UT_TRISTATE.set_false */
extern void T1398f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.import_classes */
extern void T56f178(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T56f178ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T56f178ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T56f178ac3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.preparse */
extern void T56f177(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T56f177ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T56f177ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T56f177ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T56f177ac4(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_recursive */
extern void T184f8(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ADAPTED_LIBRARIES.do_recursive */
extern T0* T184f8ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_LIBRARY.add_library_recursive */
extern void T183f197(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_library_recursive */
extern T0* T183f197ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].force_last */
extern void T878f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_put */
extern void T878f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_put */
extern void T878f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_item */
extern T6 T878f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].hash_position */
extern T6 T878f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].resize */
extern void T878f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_resize */
extern void T878f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].special_integer_ */
extern T0* T878f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_storage_resize */
extern void T878f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_resize */
extern void T878f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].aliased_resized_area */
extern T0* T1576f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].aliased_resized_area */
extern T0* T1577f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_storage_item */
extern T0* T878f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_item */
extern T0* T878f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_item */
extern T6 T878f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_resize */
extern void T878f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_modulus */
extern T6 T878f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_capacity */
extern T6 T878f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_put */
extern void T878f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].force */
extern void T1576f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].fill_with */
extern void T1577f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].extend */
extern void T1577f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].search_position */
extern void T878f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_equality_tester */
extern T0* T878f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].unset_found_item */
extern void T878f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].has */
extern T1 T878f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.add_library_recursive */
extern void T826f172(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_library_recursive */
extern T0* T826f172ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].do_all */
extern void T878f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make */
extern T0* T878c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].set_internal_cursor */
extern void T878f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_cursor */
extern T0* T878f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY].make */
extern T0* T1575c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_slots */
extern void T878f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_clashes */
extern void T878f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_key_storage */
extern void T878f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_item_storage */
extern void T878f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].make */
extern T0* T1576f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].make_empty */
extern T0* T1577c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].default_create */
extern T0* T1576c3(GE_context* ac);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_recursive */
extern void T206f7(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_recursive */
extern T0* T206f7ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern void T207f171(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern T0* T207f171ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].has */
extern T1 T816f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern void T586f171(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern T0* T586f171ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_assemblies */
extern void T460f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.consume */
extern void T207f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_assembly */
extern void T460f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gaaaa_error */
extern void T51f327(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_assembly_error */
extern void T51f354(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.make_gaaaa */
extern T0* T1889c14(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gaaaa_error */
extern T1 T51f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gaaaa_error */
extern void T35f327(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_assembly_error */
extern void T35f354(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gaaaa_error */
extern T1 T35f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.error_handler */
extern T0* T460f2(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.consume */
extern void T586f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T585f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.count */
extern T6 T585f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].do_all */
extern void T816f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.make_empty */
extern T0* T585c5(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1356c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T1356f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].new_cursor */
extern T0* T1356f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1679c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1553f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].make_empty */
extern T0* T1554c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].default_create */
extern T0* T1553c3(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T816c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T816f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_cursor */
extern T0* T816f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1552c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_slots */
extern void T816f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_clashes */
extern void T816f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_key_storage */
extern void T816f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_item_storage */
extern void T816f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.count */
extern T6 T206f3(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_6 */
extern void T371f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_6_metrics */
extern void T371f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.activate_dynamic_type_set_builder */
extern void T61f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_all */
extern void T61f85(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DYNAMIC_SYSTEM.compile_all */
extern T0* T61f85ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_all_features */
extern void T61f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern void T56f169(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern T0* T56f169ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern T0* T56f169ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_all */
extern void T62f56(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.compile_all */
extern T0* T62f56ac1(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_classes */
extern void T62f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_3 */
extern void T62f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_3_metrics */
extern void T62f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_metrics */
extern void T62f103(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.report_processor_metrics */
extern void T62f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_processed_class_count_stack */
extern void T62f105(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.total_processed_class_count */
extern T6 T62f40(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_implementation_internal_error */
extern void T62f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.check_implementation_validity */
extern void T62f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.reset_processed_class_count */
extern void T62f102(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.reset_total_processed_class_count */
extern void T62f101(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_4 */
extern void T62f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_4_metrics */
extern void T62f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_interface_internal_error */
extern void T62f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.check_interface_validity */
extern void T62f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flattening_internal_error */
extern void T62f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.flatten_features */
extern void T62f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ancestors_internal_error */
extern void T62f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.build_ancestors */
extern void T62f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_5 */
extern void T62f81(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_PROCESSOR.check_provider_validity */
extern void T62f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_5_2 */
extern void T62f84(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_5_2_metrics */
extern void T62f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_syntax_internal_error */
extern void T62f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.parse_classes */
extern void T62f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.remove_unmarked_classes */
extern void T62f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].keep_first */
extern void T455f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].replace */
extern void T455f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_CLASS.is_marked */
extern T1 T73f175(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.parse_marked_classes */
extern void T62f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.nested_benchmark_start_time */
extern T0* T62f43(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_all */
extern void T371f58(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_all */
extern T0* T371f58ac1(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_classes */
extern void T371f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_3 */
extern void T371f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_3_metrics */
extern void T371f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_metrics */
extern void T371f106(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processor_metrics_recursive */
extern void T371f107(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processor_metrics */
extern void T371f108(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processed_class_count_stack */
extern void T371f109(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.total_processed_class_count */
extern T6 T371f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].item */
extern T0* T680f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.total_processed_class_count_recursive */
extern T6 T371f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_implementation_internal_error */
extern void T371f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern void T371f96(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern T0* T371f96ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern void T371f96p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.reset_processed_class_count */
extern void T371f105(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.reset_total_processed_class_count */
extern void T371f104(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.tokens */
extern T0* T371f47(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.launch */
extern void T681f10(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.launch_with_attributes */
extern void T681f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] WORKER_THREAD.last_created_thread */
extern T14 T681f6(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.thr_set_terminated */
extern void T681f14(GE_context* ac, T0* C, T1 a1);
/* [detachable] WORKER_THREAD.thr_set_terminated */
extern void T681a14(T0* C, T1 a1);
/* [detachable] WORKER_THREAD.thr_main */
extern void T681f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] WORKER_THREAD.thr_main */
extern void T681a13(T0* C, T6 a1);
/* [detachable] WORKER_THREAD.execute_procedure */
extern void T681f16(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.default_pointer */
extern T14 T681f8(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.create_thread_with_attr */
extern void T681f15(GE_context* ac, T0* C, T0* a1, T14 a2, T14 a3, T14 a4);
/* [detachable] THREAD_ATTRIBUTES.item */
extern T14 T1367f4(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.is_launchable */
extern T1 T681f5(GE_context* ac, T0* C);
/* [detachable] MUTEX.is_set */
extern T1 T133f5(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] BOOLEAN].put */
extern void T1368f2(GE_context* ac, T0* C, T1 a1);
/* [detachable] CELL [[attached] BOOLEAN].put */
extern T0* T1368c2(GE_context* ac, T1 a1);
/* [detachable] WORKER_THREAD.is_last_launch_successful_cell */
extern T0* T681f7(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.make */
extern T0* T1367c7(GE_context* ac);
/* [detachable] THREAD_ATTRIBUTES.set_priority */
extern void T1367f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] THREAD_ATTRIBUTES.c_set_priority */
extern void T1367f10(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] THREAD_ATTRIBUTES.default_priority */
extern T6 T1367f5(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.mem_make */
extern void T1367f8(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.structure_size */
extern T6 T1367f6(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.make */
extern T0* T681c9(GE_context* ac, T0* a1);
/* [detachable] WORKER_THREAD.thread_make */
extern void T681f11(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_4 */
extern void T371f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_4_metrics */
extern void T371f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_interface_internal_error */
extern void T371f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_interface_validity */
extern void T371f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flattening_internal_error */
extern void T371f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.flatten_features */
extern void T371f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ancestors_internal_error */
extern void T371f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.build_ancestors */
extern void T371f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_5 */
extern void T371f81(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_provider_validity */
extern void T371f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_5_2 */
extern void T371f87(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_5_2_metrics */
extern void T371f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_syntax_internal_error */
extern void T371f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern void T371f101(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern T0* T371f101ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern void T371f101p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.remove_unmarked_classes */
extern void T371f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern void T371f99(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern T0* T371f99ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern void T371f99p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.nested_benchmark_start_time */
extern T0* T371f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_dynamic_type_set_builder */
extern void T61f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make */
extern T0* T70c99(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_map */
extern T0* T377c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_with_equality_testers */
extern void T377f35(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_sparse_container */
extern void T377f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_internal_cursor */
extern void T377f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_cursor */
extern T0* T377f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make */
extern T0* T686c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_slots */
extern void T377f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_modulus */
extern T6 T377f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_clashes */
extern void T377f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_key_storage */
extern void T377f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_item_storage */
extern void T377f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_dynamic_feature */
extern T0* T70f92(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_feature */
extern T0* T70f69(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.make */
extern T0* T468c77(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DEFERRED_PROCEDURE.make */
extern void T468f77p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.deferred_keyword */
extern T0* T72f277(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_deferred */
extern T0* T409c53(GE_context* ac);
/* [detachable] ET_DEFERRED_PROCEDURE.tokens */
extern T0* T468f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_dynamic_type */
extern T0* T70f79(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.make */
extern T0* T376c73(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.make */
extern T0* T400c9(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.tokens */
extern T0* T400f7(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.make */
extern void T400f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T399c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_internal_cursor */
extern void T399f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].new_cursor */
extern T0* T399f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T737c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T735f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make_empty */
extern T0* T736c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].default_create */
extern T0* T735c3(GE_context* ac);
/* [detachable] ET_CLIENT.make */
extern T0* T398c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_CLIENT_LIST.make_with_capacity */
extern T0* T397c13(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].make */
extern T0* T396c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].set_internal_cursor */
extern void T396f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].new_cursor */
extern T0* T396f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].make */
extern T0* T732c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].make */
extern T0* T729f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].make_empty */
extern T0* T731c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].default_create */
extern T0* T729c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T395c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].set_internal_cursor */
extern void T395f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].new_cursor */
extern T0* T395f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T728c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T727f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].make_empty */
extern T0* T726c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].default_create */
extern T0* T727c3(GE_context* ac);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.make */
extern T0* T394c3(GE_context* ac);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.make */
extern T0* T392c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].make */
extern T0* T723c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].set_internal_cursor */
extern void T723f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].new_cursor */
extern T0* T723f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACROSS_COMPONENT].make */
extern T0* T1377c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT].make */
extern T0* T722f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].make_empty */
extern T0* T721c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT].default_create */
extern T0* T722c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_map */
extern T0* T391c41(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_with_equality_testers */
extern void T391f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_sparse_container */
extern void T391f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].set_internal_cursor */
extern void T391f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].new_cursor */
extern T0* T391f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make */
extern T0* T719c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_slots */
extern void T391f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].new_modulus */
extern T6 T391f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_clashes */
extern void T391f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_key_storage */
extern void T391f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].make_item_storage */
extern void T391f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T710f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].make_empty */
extern T0* T709c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].default_create */
extern T0* T710c3(GE_context* ac);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.make */
extern T0* T390c8(GE_context* ac, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.tokens */
extern T0* T390f6(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.make */
extern T0* T389c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T718c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].set_internal_cursor */
extern void T718f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].new_cursor */
extern T0* T718f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T1376c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T717f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].make_empty */
extern T0* T716c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].default_create */
extern T0* T717c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_map */
extern T0* T388c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_with_equality_testers */
extern void T388f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_sparse_container */
extern void T388f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_internal_cursor */
extern void T388f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_cursor */
extern T0* T388f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T712c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_slots */
extern void T388f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_modulus */
extern T6 T388f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_clashes */
extern void T388f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_key_storage */
extern void T388f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_item_storage */
extern void T388f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T387c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].set_internal_cursor */
extern void T387f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].new_cursor */
extern T0* T387f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T711c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T386c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_internal_cursor */
extern void T386f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].new_cursor */
extern T0* T386f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T708c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T705f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make_empty */
extern T0* T707c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].default_create */
extern T0* T705c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T385c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_internal_cursor */
extern void T385f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].new_cursor */
extern T0* T385f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T704c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T701f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make_empty */
extern T0* T703c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].default_create */
extern T0* T701c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T384c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_internal_cursor */
extern void T384f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].new_cursor */
extern T0* T384f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T700c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T697f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make_empty */
extern T0* T699c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].default_create */
extern T0* T697c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T383c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_internal_cursor */
extern void T383f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].new_cursor */
extern T0* T383f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T696c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T693f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make_empty */
extern T0* T695c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].default_create */
extern T0* T693c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].make */
extern T0* T382c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].set_internal_cursor */
extern void T382f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].new_cursor */
extern T0* T382f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].make */
extern T0* T692c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].make */
extern T0* T690f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].make_empty */
extern T0* T691c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].default_create */
extern T0* T690c3(GE_context* ac);
/* [detachable] ET_FEATURE_CHECKER.dummy_feature */
extern T0* T376f48(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.make */
extern T0* T380c22(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].make */
extern T0* T689c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].set_internal_cursor */
extern void T689f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].new_cursor */
extern T0* T689f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].make */
extern T0* T1373c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].make */
extern T0* T1370f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].make_empty */
extern T0* T1372c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].default_create */
extern T0* T1370c3(GE_context* ac);
/* [detachable] ET_TYPE_CHECKER.dummy_feature */
extern T0* T380f16(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.tokens */
extern T0* T380f19(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.make */
extern void T380f22p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.make */
extern void T376f73p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make_feature_checker */
extern void T70f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_context */
extern T0* T70f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make */
extern void T70f106p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_new_instance_types */
extern void T61f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_catcall_warning_mode */
extern void T61f80(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.catcall_warning_mode */
extern T1 T25f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_catcall_error_mode */
extern void T61f79(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.catcall_error_mode */
extern T1 T25f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_full_class_checking */
extern void T61f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.make */
extern T0* T61c77(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_DYNAMIC_SYSTEM.make */
extern T0* T61f77ac1(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.make */
extern T0* T226c13(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make */
extern T0* T222c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_with_equality_testers */
extern void T222f36(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_sparse_container */
extern void T222f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].set_internal_cursor */
extern void T222f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_cursor */
extern T0* T222f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make */
extern T0* T605c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_slots */
extern void T222f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_clashes */
extern void T222f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_key_storage */
extern void T222f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_item_storage */
extern void T222f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.make_basic_types */
extern void T61f84(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile */
extern void T62f55(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.compile */
extern T0* T62f55ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_SYSTEM_PROCESSOR.compile */
extern T0* T62f55ac2(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_marked_classes */
extern void T62f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern void T371f57(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern T0* T371f57ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern T0* T371f57ac2(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_marked_classes */
extern void T371f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.benchmark_start_time */
extern T0* T62f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.benchmark_start_time */
extern T0* T371f43(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_dbc_mode */
extern void T62f54(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_dbc_mode_only */
extern void T62f66(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_dbc_mode */
extern void T371f56(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_dbc_mode_only */
extern void T371f68(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_mode */
extern void T62f53(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_mode_only */
extern void T62f65(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_mode */
extern void T371f55(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_mode_only */
extern void T371f67(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_gelint */
extern T1 T25f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_metrics_shown */
extern void T62f52(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_metrics_shown_only */
extern void T62f64(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_metrics_shown */
extern void T371f54(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_metrics_shown_only */
extern void T371f66(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_metrics */
extern T1 T25f40(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_nested_benchmark_shown */
extern void T62f51(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_nested_benchmark_shown_only */
extern void T62f63(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_nested_benchmark_shown */
extern void T371f53(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_nested_benchmark_shown_only */
extern void T371f65(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_nested_benchmark */
extern T1 T25f39(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_benchmark_shown */
extern void T62f50(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_benchmark_shown_only */
extern void T62f62(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_benchmark_shown */
extern void T371f52(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_benchmark_shown_only */
extern void T371f64(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_silent */
extern T1 T25f37(GE_context* ac, T0* C);
/* [detachable] GEC.is_no_benchmark */
extern T1 T25f38(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ise_version */
extern void T62f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ise_version_only */
extern void T62f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ecma_version_only */
extern void T62f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ise_version */
extern void T371f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ise_version_only */
extern void T371f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_default_keyword_usage_only */
extern void T371f80(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ecma_version_only */
extern void T371f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_detachable_keyword_only */
extern void T371f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_attached_keyword_only */
extern void T371f77(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_reference_keyword_only */
extern void T371f76(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_note_keyword_only */
extern void T371f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_attribute_keyword_only */
extern void T371f74(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_error_handler */
extern void T62f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_error_handler_only */
extern void T62f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_error_handler */
extern void T371f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_error_handler_only */
extern void T371f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.thread_count */
extern T6 T25f36(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.parameter */
extern T6 T47f14(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].last */
extern T6 T156f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].is_empty */
extern T1 T156f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_verbose */
extern void T51f179(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.set_verbose */
extern void T35f179(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_verbose */
extern T1 T25f35(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_ise */
extern void T51f178(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.set_ise */
extern void T35f178(GE_context* ac, T0* C);
/* [detachable] GEC.parse_ecf_file */
extern void T25f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.to_boolean */
extern T1 T743f64(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_boolean */
extern T1 T17f43(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_boolean */
extern T1 T743f62(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_boolean */
extern T1 T17f42(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.has_error */
extern T1 T54f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern void T53f210(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern T0* T53f210ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern T0* T53f210ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]].set_rout_disp_final */
extern void T167f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_system */
extern void T53f213(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.build_system */
extern T0* T53f213ac1(GE_context* ac);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.propagate_read_only */
extern void T203f35(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_read_only */
extern void T826f181(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM].put */
extern void T159f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].wipe_out */
extern void T172f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_wipe_out */
extern void T172f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_wipe_out */
extern void T172f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_wipe_out */
extern void T172f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_wipe_out */
extern void T172f42(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].keep_head */
extern void T501f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].set_count */
extern void T501f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].unset_found_item */
extern void T172f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].move_all_cursors_after */
extern void T172f40(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_next_cursor */
extern void T504f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_after */
extern void T504f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].wipe_out */
extern void T170f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_wipe_out */
extern void T170f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_wipe_out */
extern void T170f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_wipe_out */
extern void T170f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_wipe_out */
extern void T170f52(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].keep_head */
extern void T497f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].set_count */
extern void T497f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].unset_found_item */
extern void T170f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].move_all_cursors_after */
extern void T170f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_next_cursor */
extern void T496f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_after */
extern void T496f9(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_dotnet_assemblies */
extern void T53f217(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].force_last_new */
extern void T172f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_put */
extern void T172f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_put */
extern void T172f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].force */
extern void T505f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].fill_with */
extern void T501f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].extend */
extern void T501f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_put */
extern void T172f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_put */
extern void T172f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_item */
extern T6 T172f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].hash_position */
extern T6 T172f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].resize */
extern void T172f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_resize */
extern void T172f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].special_integer_ */
extern T0* T172f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_resize */
extern void T172f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_resize */
extern void T172f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T505f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T501f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_item */
extern T0* T172f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_item */
extern T6 T172f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_resize */
extern void T172f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_modulus */
extern T6 T172f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_capacity */
extern T6 T172f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_dotnet_assembly */
extern T0* T158f2(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make */
extern T0* T207c104(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make_from_system */
extern void T207f105(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_default_class_mapping */
extern void T207f110(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_wide_character_ref_class_mapping */
extern void T207f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_class_mapping */
extern void T207f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.set_mapped_class */
extern void T560f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.wide_character_ref_class_name */
extern T0* T72f192(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_ref_class_name */
extern T0* T72f173(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_wide_character_class_mapping */
extern void T207f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_class_mapping */
extern void T207f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.string_class_name */
extern T0* T72f184(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.string_8_class_name */
extern T0* T72f172(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_ref_class_mapping */
extern void T207f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.real_ref_class_name */
extern T0* T72f191(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_ref_class_name */
extern T0* T72f171(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_class_mapping */
extern void T207f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_ref_class_mapping */
extern void T207f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.natural_ref_class_name */
extern T0* T72f190(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_ref_class_name */
extern T0* T72f170(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_class_mapping */
extern void T207f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_ref_class_mapping */
extern void T207f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.integer_ref_class_name */
extern T0* T72f189(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_ref_class_name */
extern T0* T72f169(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_class_mapping */
extern void T207f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_double_ref_class_mapping */
extern void T207f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.double_ref_class_name */
extern T0* T72f188(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_ref_class_name */
extern T0* T72f168(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_double_class_mapping */
extern void T207f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_ref_class_mapping */
extern void T207f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.character_ref_class_name */
extern T0* T72f187(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_ref_class_name */
extern T0* T72f167(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_class_mapping */
extern void T207f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_kernel_types */
extern void T207f109(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_wide_character_type */
extern void T207f152(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.make */
extern T0* T71c125(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_typed_pointer_type */
extern void T207f151(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identity_type */
extern T0* T72f177(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_type_type */
extern void T207f150(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_like_current */
extern T0* T72f186(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_keyword */
extern T0* T72f174(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_detachable */
extern T0* T409c48(GE_context* ac);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_system_string_type */
extern void T207f149(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.system_string_class_name */
extern T0* T72f185(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_system_object_type */
extern void T207f148(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.put_first */
extern void T558f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].force */
extern void T1345f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].fill_with */
extern void T1344f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].extend */
extern void T1344f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_LIST.fixed_array */
extern T0* T558f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].default_create */
extern T0* T1345c2(GE_context* ac);
/* [detachable] ET_PARENT_LIST.make_with_capacity */
extern T0* T558c11(GE_context* ac, T6 a1);
/* [detachable] ET_PARENT_LIST.make_with_capacity */
extern void T558f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].make */
extern T0* T1345f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].make_empty */
extern T0* T1344c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.inherit_keyword */
extern T0* T72f278(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_inherit */
extern T0* T409c54(GE_context* ac);
/* [detachable] ET_PARENT_LIST.tokens */
extern T0* T558f9(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.make */
extern T0* T557c11(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_PARENT.tokens */
extern T0* T557f9(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.has_feature_adaptation */
extern T1 T557f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_type */
extern void T207f147(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_32_type */
extern void T207f146(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.make */
extern T0* T559c6(GE_context* ac, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.make */
extern T0* T1346c8(GE_context* ac);
/* [detachable] ET_BRACED_TYPE_LIST.make */
extern void T1346f8p1(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].make */
extern T0* T1677f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].make_empty */
extern T0* T1676c4(GE_context* ac, T6 a1);
/* [detachable] ET_BRACED_TYPE_LIST.fixed_array */
extern T0* T1346f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].default_create */
extern T0* T1677c2(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.right_brace_symbol */
extern T0* T72f702(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_brace */
extern T0* T343c95(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_brace_symbol */
extern T0* T72f701(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_brace */
extern T0* T343c94(GE_context* ac);
/* [detachable] ET_BRACED_TYPE_LIST.tokens */
extern T0* T1346f5(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.string_32_class_name */
extern T0* T72f183(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_8_type */
extern void T207f145(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_special_type */
extern void T207f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_routine_type */
extern void T207f143(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_64_type */
extern void T207f142(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_32_type */
extern void T207f141(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_type */
extern void T207f140(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_procedure_type */
extern void T207f139(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_predicate_type */
extern void T207f138(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_pointer_type */
extern void T207f137(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_64_type */
extern void T207f136(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_32_type */
extern void T207f135(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_16_type */
extern void T207f134(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_8_type */
extern void T207f133(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_type */
extern void T207f132(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_iterable_type */
extern void T207f131(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_ise_exception_manager_type */
extern void T207f130(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ise_exception_manager_class_name */
extern T0* T72f182(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_64_type */
extern void T207f129(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_32_type */
extern void T207f128(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_16_type */
extern void T207f127(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_8_type */
extern void T207f126(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_type */
extern void T207f125(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_immutable_string_32_type */
extern void T207f124(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.immutable_string_32_class_name */
extern T0* T72f181(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_immutable_string_8_type */
extern void T207f123(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.immutable_string_8_class_name */
extern T0* T72f180(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_function_type */
extern void T207f122(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_exception_manager_type */
extern void T207f121(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_class_name */
extern T0* T72f179(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_exception_type */
extern void T207f120(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_class_name */
extern T0* T72f178(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_double_type */
extern void T207f119(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_32_type */
extern void T207f118(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_8_type */
extern void T207f117(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_type */
extern void T207f116(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_boolean_type */
extern void T207f115(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_array_type */
extern void T207f114(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_tuple_type */
extern void T207f113(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.make */
extern T0* T567c17(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_none_type */
extern void T207f112(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.add_first_imported_class */
extern void T560f52(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].force_first */
extern void T1347f15(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].put */
extern void T1347f23(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_cursors_right */
extern void T1347f30(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_right */
extern void T1347f29(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].put_last */
extern void T1347f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.none_class_name */
extern T0* T72f176(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_any_type */
extern void T207f111(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.put_first */
extern void T397f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].force */
extern void T734f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].fill_with */
extern void T733f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].extend */
extern void T733f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_separate_type_mark */
extern T0* T72f175(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.make */
extern T0* T561c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.separate_keyword */
extern T0* T72f244(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_separate */
extern T0* T409c51(GE_context* ac);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make_adapted */
extern void T207f108(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T207f89(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_unknown_kernel_types */
extern void T207f107(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_convert_feature */
extern T0* T72f166(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_generic_class_type */
extern T0* T72f165(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make */
extern T0* T236c31(GE_context* ac);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make */
extern void T236f31p1(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.empty_actual_parameters */
extern T0* T72f164(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_tuple_type */
extern T0* T72f163(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_parents */
extern T0* T72f161(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_parent */
extern T0* T72f160(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class_type */
extern T0* T72f159(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.initialize */
extern void T207f106(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_key_equality_tester */
extern void T221f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.class_name_tester */
extern T0* T207f95(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_map */
extern T0* T221c38(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_with_equality_testers */
extern void T221f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_sparse_container */
extern void T221f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_internal_cursor */
extern void T221f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_slots */
extern void T221f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_clashes */
extern void T221f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_key_storage */
extern void T221f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_item_storage */
extern void T221f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_dotnet_assembly */
extern void T209f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].found_item */
extern T0* T172f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_item */
extern T0* T172f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].found */
extern T1 T172f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].search */
extern void T172f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].search_position */
extern void T172f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.operating_system */
extern T0* T53f26(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.full_filename */
extern T0* T53f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.windows_file_system */
extern T0* T53f19(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.file_system */
extern T0* T53f18(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.unix_file_system */
extern T0* T53f32(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.pathname */
extern T0* T209f22(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T206f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_TARGET.fill_options */
extern void T57f55(GE_context* ac, T0* C, T0* a1);
extern T1 T57f55ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM.set_trace_mode */
extern void T56f257(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_obsolete_routine_type_mode */
extern void T56f256(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_LIBRARY.set_obsolete_routine_type_mode */
extern void T183f192(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_implicit_attachment_type_mark */
extern void T56f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_implicit_attachment_type_mark */
extern void T183f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.tokens */
extern T0* T57f29(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.value */
extern T0* T437f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.fill_capabilities */
extern void T57f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_target_type_attachment_mode */
extern void T56f263(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_attachment_type_conformance_mode */
extern void T56f262(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_multithreaded_mode */
extern void T56f261(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.string_ */
extern T0* T57f30(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.support_value */
extern T0* T59f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.use_value */
extern T0* T59f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].found_item */
extern T0* T217f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_item */
extern T0* T217f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].found */
extern T1 T217f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].search */
extern void T217f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].search_position */
extern void T217f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].hash_position */
extern T6 T217f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_item */
extern T0* T217f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_item */
extern T6 T217f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_item */
extern T6 T217f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_TARGET.fill_settings */
extern void T57f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_total_order_on_reals_mode */
extern void T56f260(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_exception_trace_mode */
extern void T56f259(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_console_application_mode */
extern void T56f258(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.fill_root */
extern void T57f59(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.fill_root */
extern void T790f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_root_creation */
extern void T56f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_root_type */
extern void T56f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ROOT_ALL_CLASSES.fill_root */
extern void T791f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.unset_root_type */
extern void T56f276(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].forth */
extern void T170f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_forth */
extern void T170f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].add_traversing_cursor */
extern void T170f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_position */
extern void T496f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].remove_traversing_cursor */
extern void T170f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_item */
extern T6 T170f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_for_iteration */
extern T0* T170f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_item */
extern T0* T170f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_item */
extern T0* T170f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].after */
extern T1 T170f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_after */
extern T1 T170f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].after */
extern T1 T496f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].start */
extern void T170f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_start */
extern void T170f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_off */
extern T1 T170f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].off */
extern T1 T496f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].is_empty */
extern T1 T170f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_libraries */
extern void T53f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_libraries */
extern T0* T53f216ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_library */
extern void T53f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY].put */
extern void T202f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_library */
extern T0* T53f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].force_last_new */
extern void T170f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_put */
extern void T170f61(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_put */
extern void T170f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].force */
extern void T500f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].fill_with */
extern void T497f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].extend */
extern void T497f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_put */
extern void T170f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_put */
extern void T170f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_item */
extern T6 T170f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].hash_position */
extern T6 T170f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].resize */
extern void T170f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_resize */
extern void T170f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].special_integer_ */
extern T0* T170f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_resize */
extern void T170f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_resize */
extern void T170f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].aliased_resized_area */
extern T0* T500f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].aliased_resized_area */
extern T0* T497f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_item */
extern T0* T170f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_resize */
extern void T170f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_modulus */
extern T6 T170f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_capacity */
extern T6 T170f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.fill_system_config */
extern void T53f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f220ot1(T0* a1, T0** a2);
extern T1 T53f220ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eltm_error */
extern void T54f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error */
extern void T54f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error_message */
extern void T54f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error_message */
extern void T54f32p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.message */
extern T0* T54f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.default_message */
extern T0* T210f9(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.message */
extern T0* T210f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.arguments */
extern T0* T210f6(GE_context* ac, T0* C);
/* [detachable] KL_ARGUMENTS.make */
extern T0* T33c5(GE_context* ac);
/* [detachable] KL_ARGUMENTS.argument */
extern T0* T33f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.as_string_8 */
extern T0* T20f9(GE_context* ac, T0* C);
/* [detachable] ARGUMENTS_32.argument */
extern T0* T431f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].item */
extern T0* T902f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARGUMENTS_32.internal_argument_array */
extern T0* T431f3(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].put */
extern void T902f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].compare_objects */
extern void T902f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].make_filled */
extern T0* T902c6(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].make_filled_area */
extern void T902f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_filled */
extern T0* T903c6(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].fill_with */
extern void T903f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].extend */
extern void T903f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_empty */
extern void T903f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_empty */
extern T0* T903c4(GE_context* ac, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.make_empty */
extern T0* T20c16(GE_context* ac);
/* [detachable] IMMUTABLE_STRING_32.make */
extern void T20f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARGUMENTS.internal_arguments */
extern T0* T33f4(GE_context* ac, T0* C);
/* [detachable] ARGUMENTS_32.default_create */
extern T0* T431c5(GE_context* ac);
/* [detachable] ARRAY [[attached] STRING_8].valid_index */
extern T1 T324f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.to_integer */
extern T6 T743f29(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parsed_integer */
extern T6 T428f13(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parse_string_with_type */
extern void T428f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parse_character */
extern void T428f23(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.overflowed */
extern T1 T428f15(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.overflow_checker */
extern T0* T428f14(GE_context* ac, T0* C);
/* CHARACTER_8.is_digit */
extern T1 T2f26(GE_context* ac, T2* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.reset */
extern void T428f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.ctoi_convertor */
extern T0* T743f50(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_trailing_separators_acceptable */
extern void T428f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_leading_separators_acceptable */
extern void T428f20(GE_context* ac, T0* C, T1 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_trailing_separators */
extern void T428f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_leading_separators */
extern void T428f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.make */
extern T0* T428c16(GE_context* ac);
/* [detachable] STRING_8.to_integer */
extern T6 T17f22(GE_context* ac, T0* C);
/* [detachable] STRING_8.ctoi_convertor */
extern T0* T17f23(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_integer */
extern T1 T743f28(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_valid_integer_or_natural */
extern T1 T743f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.is_integral_integer */
extern T1 T428f12(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_integer */
extern T1 T17f26(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_valid_integer_or_natural */
extern T1 T17f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR.string_ */
extern T0* T210f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.make_eltm */
extern T0* T210c105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.filename */
extern T0* T210f14(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.default_message_template */
extern T0* T210f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.add_dependent_universes */
extern void T210f116(GE_context* ac, T0* C, T0* a1);
extern T1 T210f116ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_TARGET.parent_path */
extern T0* T57f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].put_last */
extern void T884f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].force */
extern void T1583f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].fill_with */
extern void T1582f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].extend */
extern void T1582f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make */
extern void T884f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T884c13(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].set_internal_cursor */
extern void T884f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].new_cursor */
extern T0* T884f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T1584c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T1583f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].make_empty */
extern T0* T1582c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].default_create */
extern T0* T1583c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].item */
extern T0* T882f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR.add_parent_path */
extern void T210f117(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].item */
extern T0* T884f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].force_last */
extern void T884f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].resize */
extern void T884f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].aliased_resized_area */
extern T0* T1583f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].aliased_resized_area */
extern T0* T1582f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].new_capacity */
extern T6 T884f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].extendible */
extern T1 T884f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make_default */
extern T0* T884c11(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].default_capacity */
extern T6 T884f8(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGETS.target_by_name */
extern T0* T199f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.string_ */
extern T0* T199f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].item */
extern T0* T881f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_etam_error */
extern void T54f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_etam */
extern T0* T210c104(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_TARGETS.is_empty */
extern T1 T199f5(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGETS.count */
extern T6 T199f3(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].forth */
extern void T440f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_forth */
extern void T869f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].add_traversing_cursor */
extern void T869f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_next_cursor */
extern void T440f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_traversing_cursor */
extern void T869f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set */
extern void T440f15(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatb_error */
extern void T54f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatb */
extern T0* T210c103(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.attribute_value */
extern T0* T53f51(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable XM_POSITION_TABLE.item */
extern T0* T165f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].forth */
extern void T873f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_forth */
extern void T874f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].add_traversing_cursor */
extern void T874f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_next_cursor */
extern void T873f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].remove_traversing_cursor */
extern void T874f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set */
extern void T873f12(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].go_after */
extern void T873f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_go_after */
extern void T874f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_after */
extern void T873f13(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_off */
extern T1 T874f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].off */
extern T1 T873f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].item */
extern T0* T873f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].start */
extern void T873f9(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_start */
extern void T874f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].new_cursor */
extern T0* T874f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T873c8(GE_context* ac, T0* a1);
/* detachable XM_POSITION_TABLE.has */
extern T1 T165f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.is_false */
extern T1 T53f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.is_true */
extern T1 T53f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_targets */
extern T0* T53f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_ECF_TARGETS.put_last */
extern void T199f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].force_last */
extern void T881f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].force */
extern void T579f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].fill_with */
extern void T577f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].extend */
extern void T577f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].resize */
extern void T881f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].aliased_resized_area */
extern T0* T579f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].aliased_resized_area */
extern T0* T577f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].new_capacity */
extern T6 T881f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].extendible */
extern T1 T881f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_targets */
extern T0* T158f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.make */
extern T0* T199c8(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].put_last */
extern void T881f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].make */
extern T0* T881c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].set_internal_cursor */
extern void T881f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].new_cursor */
extern T0* T881f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET].make */
extern T0* T1578c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].make */
extern T0* T579f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].make_empty */
extern T0* T577c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].default_create */
extern T0* T579c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_target */
extern T0* T53f73(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
extern T1 T53f73ot1(T0* a1, T0** a2);
extern T1 T53f73ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_TARGET.set_notes */
extern void T57f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_post_compile_actions */
extern void T57f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_pre_compile_actions */
extern void T57f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_resources */
extern void T57f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_objects */
extern void T57f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_makes */
extern void T57f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_linker_flags */
extern void T57f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_libraries */
extern void T57f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_includes */
extern void T57f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_cflags */
extern void T57f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_class_mappings */
extern void T57f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_file_rules */
extern void T57f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_dotnet_assemblies */
extern void T57f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_libraries */
extern void T57f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_clusters */
extern void T57f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_target_options */
extern void T53f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_options */
extern void T53f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_OPTIONS.set_primary_value */
extern void T437f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_value */
extern T0* T437f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.primary_value */
extern T0* T55f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_capabilities */
extern void T53f237(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_5_0 */
extern T0* T53f69(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.set_primary_use_value */
extern void T59f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_CAPABILITIES.set_primary_use_value_id */
extern void T59f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].force_last */
extern void T217f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_put */
extern void T217f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_put */
extern void T217f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_put */
extern void T217f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].resize */
extern void T217f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_resize */
extern void T217f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].special_integer_ */
extern T0* T217f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_resize */
extern void T217f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_resize */
extern void T217f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_resize */
extern void T217f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_modulus */
extern T6 T217f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_capacity */
extern T6 T217f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_put */
extern void T217f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].unset_found_item */
extern void T217f44(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.set_primary_support_value */
extern void T59f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_11_0 */
extern T0* T53f59(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.primary_use_value */
extern T0* T59f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.primary_use_value_id */
extern T0* T59f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.primary_support_value */
extern T0* T59f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_abstract */
extern void T57f38(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.set_version */
extern void T57f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_version */
extern T0* T53f116(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VERSION.set_copyright */
extern void T775f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.set_company */
extern void T775f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.set_product */
extern void T775f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.make_major */
extern T0* T775c11(GE_context* ac, T6 a1);
/* detachable ET_ECF_VERSION.make_major_minor */
extern T0* T775c10(GE_context* ac, T6 a1, T6 a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eats_error */
extern void T54f39(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eats */
extern T0* T210c107(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_VERSION.make_major_minor_release */
extern T0* T775c9(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* detachable ET_ECF_VERSION.make */
extern T0* T775c8(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eati_error */
extern void T54f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eati */
extern T0* T210c106(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_variable */
extern void T53f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_VARIABLES.set_primary_value */
extern void T60f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_VARIABLES.primary_value */
extern T0* T60f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_setting */
extern void T53f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_TARGET.set_root */
extern void T57f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_root */
extern T0* T53f112(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ROOT_CLASS.set_cluster_name */
extern void T790f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.set_creation_procedure_name */
extern void T790f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_root_class */
extern T0* T158f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.make */
extern T0* T790c4(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_root_all_classes */
extern T0* T158f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ROOT_ALL_CLASSES.make */
extern T0* T791c1(GE_context* ac);
/* [detachable] ET_ECF_TARGET.set_precompiled_library */
extern void T57f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_precompiled_library */
extern T0* T53f110(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_precompiled_library */
extern T0* T53f110ac1(GE_context* ac, T0* a1);
extern T1 T53f110ot1(T0* a1, T0** a2);
extern T1 T53f110ot2(T0* a1, T0** a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]].set_rout_disp_final */
extern void T789f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_conditions */
extern void T773f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_visible_classes */
extern void T773f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_class_renamings */
extern void T773f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_class_options */
extern void T773f26(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].do_all */
extern void T786f35(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_item */
extern T0* T786f33(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_item */
extern T6 T786f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_use_application_options */
extern void T773f24(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_read_only */
extern void T773f23(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_classname_prefix */
extern void T773f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_eifgens_location */
extern void T773f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_visible_classes */
extern T0* T53f139(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].force_last */
extern void T787f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].force */
extern void T1547f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].fill_with */
extern void T1548f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].extend */
extern void T1548f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].resize */
extern void T787f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].aliased_resized_area */
extern T0* T1547f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].aliased_resized_area */
extern T0* T1548f3(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].new_capacity */
extern T6 T787f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].extendible */
extern T1 T787f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_visible_classes */
extern T0* T158f41(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T787c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].set_internal_cursor */
extern void T787f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].new_cursor */
extern T0* T787f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T1549c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T1547f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].make_empty */
extern T0* T1548c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].default_create */
extern T0* T1547c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_visible_class */
extern T0* T53f171(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VISIBLE_CLASS.set_new_feature_name */
extern void T794f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.set_new_class_name */
extern void T794f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.set_feature_name */
extern void T794f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_visible_class */
extern T0* T158f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.make */
extern T0* T794c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_renamings */
extern T0* T53f137(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_renaming */
extern void T53f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatd_error */
extern void T54f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatd */
extern T0* T210c115(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_class_renamings */
extern T0* T158f40(GE_context* ac, T0* C);
/* [detachable] ET_ECF_AST_FACTORY.case_insensitive_string_equality_tester */
extern T0* T158f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_options */
extern T0* T53f135(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_AST_FACTORY.new_options */
extern T0* T158f39(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.make */
extern T0* T437c14(GE_context* ac);
/* Creation of agent #1 in feature [detachable] ET_ECF_OPTIONS.make */
extern T0* T437f14ac1(GE_context* ac, T0* a1);
/* [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32].set_rout_disp_final */
extern void T214f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] KL_STRING_ROUTINES.case_insensitive_hash_code */
extern T6 T84f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_hash_function */
extern void T212f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8].make */
extern T0* T211c3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTIONS.string_ */
extern T0* T437f10(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.case_insensitive_string_equality_tester */
extern T0* T437f9(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_description */
extern void T773f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_class_options */
extern T0* T53f134(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].force_last */
extern void T786f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_put */
extern void T786f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_put */
extern void T786f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_put */
extern void T786f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_item */
extern T6 T786f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].hash_position */
extern T6 T786f23(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].resize */
extern void T786f43(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_resize */
extern void T786f51(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].special_integer_ */
extern T0* T786f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_resize */
extern void T786f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_resize */
extern void T786f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].aliased_resized_area */
extern T0* T1545f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].aliased_resized_area */
extern T0* T1542f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_item */
extern T0* T786f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_resize */
extern void T786f48(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_modulus */
extern T6 T786f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_capacity */
extern T6 T786f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_put */
extern void T786f42(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].force */
extern void T1545f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].fill_with */
extern void T1542f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].extend */
extern void T1542f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].search_position */
extern void T786f40(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].unset_found_item */
extern void T786f41(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].found_item */
extern T0* T786f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].found */
extern T1 T786f27(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].search */
extern void T786f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_class_options */
extern T0* T158f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].set_key_equality_tester */
extern void T786f38(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_map */
extern T0* T786c34(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_with_equality_testers */
extern void T786f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_sparse_container */
extern void T786f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].set_internal_cursor */
extern void T786f56(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_cursor */
extern T0* T786f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make */
extern T0* T1546c4(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_slots */
extern void T786f55(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_clashes */
extern void T786f54(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_key_storage */
extern void T786f53(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_item_storage */
extern void T786f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].make */
extern T0* T1545f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].make_empty */
extern T0* T1542c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].default_create */
extern T0* T1545c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_conditions */
extern T0* T53f132(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T53f132ot1(T0* a1, T0** a2);
extern T1 T53f132ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ORED_CONDITIONS.put_last */
extern void T792f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].force_last */
extern void T1550f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].force */
extern void T1840f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].fill_with */
extern void T1841f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].extend */
extern void T1841f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].resize */
extern void T1550f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].aliased_resized_area */
extern T0* T1840f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].aliased_resized_area */
extern T0* T1841f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].new_capacity */
extern T6 T1550f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].extendible */
extern T1 T1550f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_ored_conditions */
extern T0* T158f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.make */
extern T0* T792c6(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].put_last */
extern void T1550f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1550c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].set_internal_cursor */
extern void T1550f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].new_cursor */
extern T0* T1550f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1842c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1840f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].make_empty */
extern T0* T1841c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].default_create */
extern T0* T1840c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_condition */
extern T0* T53f170(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f170ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_version_condition */
extern T0* T53f188(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabj_error */
extern void T54f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabj */
extern T0* T210c114(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.make */
extern T0* T805c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.make */
extern T0* T804c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabi_error */
extern void T54f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eabi */
extern T0* T210c113(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabh_error */
extern void T54f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR.make_eabh */
extern T0* T210c112(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabg_error */
extern void T54f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabg */
extern T0* T210c111(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabf_error */
extern void T54f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabf */
extern T0* T210c110(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] UT_VERSION.make_major_minor_revision */
extern T0* T44c21(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* [detachable] UT_VERSION.make_major_minor */
extern T0* T44c18(GE_context* ac, T6 a1, T6 a2);
/* [detachable] UT_VERSION.make_major */
extern T0* T44c17(GE_context* ac, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_platform_condition */
extern T0* T53f187(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatn_error */
extern void T54f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatn */
extern T0* T210c109(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_platform_condition */
extern T0* T158f51(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_PLATFORM_CONDITION.make */
extern T0* T802c9(GE_context* ac, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.make_excluded */
extern T0* T802c8(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatx_error */
extern void T54f40(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatx */
extern T0* T210c108(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_multithreaded_condition */
extern T0* T53f185(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_concurrency_condition */
extern T0* T158f46(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_CONCURRENCY_CONDITION.make */
extern T0* T797c18(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.make_excluded */
extern T0* T797c17(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_dynamic_runtime_condition */
extern T0* T53f183(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_dynamic_runtime_condition */
extern T0* T158f50(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.make */
extern T0* T801c5(GE_context* ac, T1 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_dotnet_condition */
extern T0* T53f181(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_dotnet_condition */
extern T0* T158f49(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_DOTNET_CONDITION.make */
extern T0* T800c5(GE_context* ac, T1 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_custom_condition */
extern T0* T53f179(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_custom_condition */
extern T0* T158f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* detachable ET_ECF_CUSTOM_CONDITION.make */
extern T0* T799c13(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CUSTOM_CONDITION.make_excluded */
extern T0* T799c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_void_safety_condition */
extern T0* T53f177(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_void_safety_condition */
extern T0* T158f47(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.make */
extern T0* T798c18(GE_context* ac, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.make_excluded */
extern T0* T798c17(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_concurrency_condition */
extern T0* T53f175(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ANDED_CONDITIONS.put_last */
extern void T793f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].force_last */
extern void T1551f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].force */
extern void T1843f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].fill_with */
extern void T1844f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].extend */
extern void T1844f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].resize */
extern void T1551f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].aliased_resized_area */
extern T0* T1843f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].aliased_resized_area */
extern T0* T1844f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].new_capacity */
extern T6 T1551f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].extendible */
extern T1 T1551f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_build_condition */
extern T0* T53f173(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_build_condition */
extern T0* T158f45(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_BUILD_CONDITION.make */
extern T0* T796c8(GE_context* ac, T0* a1);
/* detachable ET_ECF_BUILD_CONDITION.make_excluded */
extern T0* T796c7(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_anded_conditions */
extern T0* T158f43(GE_context* ac, T0* C);
/* detachable ET_ECF_ANDED_CONDITIONS.make_empty */
extern T0* T793c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1551c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].set_internal_cursor */
extern void T1551f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].new_cursor */
extern T0* T1551f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1845c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1843f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].make_empty */
extern T0* T1844c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].default_create */
extern T0* T1843c3(GE_context* ac);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_precompiled_library */
extern T0* T158f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.make */
extern T0* T773c19(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.make_adapted_library */
extern void T773f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.default_read_only_value */
extern T1 T773f18(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_library */
extern T0* T72f260(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.make */
extern T0* T826c105(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.make_from_system */
extern void T826f106(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.set_default_class_mapping */
extern void T826f111(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_wide_character_ref_class_mapping */
extern void T826f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_class_mapping */
extern void T826f167(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.set_wide_character_class_mapping */
extern void T826f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_string_class_mapping */
extern void T826f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_real_ref_class_mapping */
extern void T826f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_real_class_mapping */
extern void T826f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_natural_ref_class_mapping */
extern void T826f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_natural_class_mapping */
extern void T826f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_integer_ref_class_mapping */
extern void T826f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_integer_class_mapping */
extern void T826f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_double_ref_class_mapping */
extern void T826f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_double_class_mapping */
extern void T826f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_character_ref_class_mapping */
extern void T826f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.set_character_class_mapping */
extern void T826f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.tokens */
extern T0* T826f89(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_kernel_types */
extern void T826f110(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_wide_character_type */
extern void T826f153(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_typed_pointer_type */
extern void T826f152(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_type_type */
extern void T826f151(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_system_string_type */
extern void T826f150(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_system_object_type */
extern void T826f149(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_type */
extern void T826f148(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_32_type */
extern void T826f147(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_8_type */
extern void T826f146(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_special_type */
extern void T826f145(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_routine_type */
extern void T826f144(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_64_type */
extern void T826f143(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_32_type */
extern void T826f142(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_type */
extern void T826f141(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_procedure_type */
extern void T826f140(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_predicate_type */
extern void T826f139(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_pointer_type */
extern void T826f138(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_64_type */
extern void T826f137(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_32_type */
extern void T826f136(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_16_type */
extern void T826f135(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_8_type */
extern void T826f134(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_type */
extern void T826f133(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_iterable_type */
extern void T826f132(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_ise_exception_manager_type */
extern void T826f131(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_64_type */
extern void T826f130(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_32_type */
extern void T826f129(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_16_type */
extern void T826f128(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_8_type */
extern void T826f127(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_type */
extern void T826f126(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_immutable_string_32_type */
extern void T826f125(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_immutable_string_8_type */
extern void T826f124(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_function_type */
extern void T826f123(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_exception_manager_type */
extern void T826f122(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_exception_type */
extern void T826f121(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_double_type */
extern void T826f120(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_32_type */
extern void T826f119(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_8_type */
extern void T826f118(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_type */
extern void T826f117(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_boolean_type */
extern void T826f116(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_array_type */
extern void T826f115(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_tuple_type */
extern void T826f114(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_none_type */
extern void T826f113(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_any_type */
extern void T826f112(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.make_adapted */
extern void T826f109(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.default_read_only_value */
extern T1 T826f90(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_unknown_kernel_types */
extern void T826f108(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.initialize */
extern void T826f107(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.make_empty */
extern T0* T206c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T584c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T584f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].new_cursor */
extern T0* T584f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1355c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1354f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make_empty */
extern T0* T1353c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].default_create */
extern T0* T1354c3(GE_context* ac);
/* [detachable] ET_CLUSTERS.make_empty */
extern T0* T554c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].make */
extern T0* T1342c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].set_internal_cursor */
extern void T1342f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].new_cursor */
extern T0* T1342f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER].make */
extern T0* T1675c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].make */
extern T0* T1674f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].make_empty */
extern T0* T1673c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].default_create */
extern T0* T1674c3(GE_context* ac);
/* [detachable] ET_ADAPTED_LIBRARIES.make_empty */
extern T0* T184c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T568c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].set_internal_cursor */
extern void T568f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].new_cursor */
extern T0* T568f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T1350c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T1349f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].make_empty */
extern T0* T1348c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].default_create */
extern T0* T1349c3(GE_context* ac);
/* [detachable] ET_LIBRARY.initialize */
extern void T826f107p1(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.class_name_tester */
extern T0* T826f94(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_system */
extern T0* T72f227(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.make */
extern T0* T36c134(GE_context* ac, T0* a1);
/* detachable ET_SYSTEM.set_default_class_mapping */
extern void T36f140(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_wide_character_ref_class_mapping */
extern void T36f195(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_class_mapping */
extern void T36f197(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.set_wide_character_class_mapping */
extern void T36f194(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_string_class_mapping */
extern void T36f193(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_real_ref_class_mapping */
extern void T36f192(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_real_class_mapping */
extern void T36f191(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_natural_ref_class_mapping */
extern void T36f190(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_natural_class_mapping */
extern void T36f189(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_integer_ref_class_mapping */
extern void T36f188(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_integer_class_mapping */
extern void T36f187(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_double_ref_class_mapping */
extern void T36f186(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_double_class_mapping */
extern void T36f185(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_character_ref_class_mapping */
extern void T36f184(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_character_class_mapping */
extern void T36f183(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.tokens */
extern T0* T36f117(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_kernel_types */
extern void T36f139(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_wide_character_type */
extern void T36f182(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_typed_pointer_type */
extern void T36f181(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_type_type */
extern void T36f180(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_string_type */
extern void T36f179(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_object_type */
extern void T36f178(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_type */
extern void T36f177(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_32_type */
extern void T36f176(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_8_type */
extern void T36f175(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_special_type */
extern void T36f174(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_routine_type */
extern void T36f173(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_64_type */
extern void T36f172(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_32_type */
extern void T36f171(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_type */
extern void T36f170(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_procedure_type */
extern void T36f169(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_predicate_type */
extern void T36f168(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_pointer_type */
extern void T36f167(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_64_type */
extern void T36f166(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_32_type */
extern void T36f165(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_16_type */
extern void T36f164(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_8_type */
extern void T36f163(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_type */
extern void T36f162(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iterable_type */
extern void T36f161(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_ise_exception_manager_type */
extern void T36f160(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_64_type */
extern void T36f159(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_32_type */
extern void T36f158(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_16_type */
extern void T36f157(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_8_type */
extern void T36f156(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_type */
extern void T36f155(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_immutable_string_32_type */
extern void T36f154(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_immutable_string_8_type */
extern void T36f153(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_function_type */
extern void T36f152(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_exception_manager_type */
extern void T36f151(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_exception_type */
extern void T36f150(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_double_type */
extern void T36f149(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_32_type */
extern void T36f148(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_8_type */
extern void T36f147(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_type */
extern void T36f146(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_boolean_type */
extern void T36f145(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_array_type */
extern void T36f144(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_tuple_type */
extern void T36f143(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_none_type */
extern void T36f142(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.add_first_local_class */
extern void T560f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_first_local_non_override_class */
extern void T560f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].force_first */
extern void T455f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].put */
extern void T455f33(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_cursors_right */
extern void T455f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_right */
extern void T455f35(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].put_last */
extern void T455f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_first_local_override_class */
extern void T560f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_first_local_ignored_class */
extern void T560f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NONE_GROUP.make */
extern T0* T861c30(GE_context* ac, T0* a1);
/* detachable ET_SYSTEM.set_any_type */
extern void T36f141(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.make_adapted */
extern void T36f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.default_read_only_value */
extern T1 T36f131(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_unknown_kernel_types */
extern void T36f137(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_name */
extern void T36f136(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.initialize */
extern void T36f135(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.initialize */
extern void T36f135p1(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.class_name_tester */
extern T0* T36f126(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.tokens */
extern T0* T773f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_actions */
extern T0* T53f107(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].force_last */
extern void T772f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].force */
extern void T1538f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].fill_with */
extern void T1539f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].extend */
extern void T1539f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].resize */
extern void T772f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].aliased_resized_area */
extern T0* T1538f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].aliased_resized_area */
extern T0* T1539f3(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].new_capacity */
extern T6 T772f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].extendible */
extern T1 T772f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_actions */
extern T0* T158f21(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].make */
extern T0* T772c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].set_internal_cursor */
extern void T772f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].new_cursor */
extern T0* T772f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION].make */
extern T0* T1540c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].make */
extern T0* T1538f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].make_empty */
extern T0* T1539c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].default_create */
extern T0* T1538c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_action */
extern T0* T53f130(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f130ot1(T0* a1, T0** a2);
extern T1 T53f130ot2(T0* a1, T0** a2);
/* detachable ET_ECF_ACTION.set_conditions */
extern void T785f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.set_working_directory */
extern void T785f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.set_must_succeed */
extern void T785f8(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ACTION.set_description */
extern void T785f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_action */
extern T0* T158f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.make */
extern T0* T785c6(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_options */
extern void T53f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T53f234ot1(T0* a1, T0** a2);
extern T1 T53f234ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_OPTIONS.set_description */
extern void T437f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_warning */
extern void T53f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_OPTIONS.set_primary_warning_value */
extern void T437f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_warning_value */
extern T0* T437f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_debug */
extern void T53f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_OPTIONS.set_primary_debug_value */
extern void T437f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_debug_value */
extern T0* T437f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_assertions */
extern void T53f246(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T53f246ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_iterator */
extern T0* T212f43(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.set_primary_assertion_value */
extern void T437f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_assertion_value */
extern T0* T437f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_mappings */
extern T0* T53f104(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_mapping */
extern void T53f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_class_mappings */
extern T0* T158f20(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_libraries */
extern T0* T53f102(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ECF_ADAPTED_LIBRARIES.put_last */
extern void T762f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].force_last */
extern void T1528f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].force */
extern void T1808f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].fill_with */
extern void T1807f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].extend */
extern void T1807f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].resize */
extern void T1528f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1808f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1807f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].new_capacity */
extern T6 T1528f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].extendible */
extern T1 T1528f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_libraries */
extern T0* T158f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.make */
extern T0* T762c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].put_last */
extern void T1528f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1528c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].set_internal_cursor */
extern void T1528f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].new_cursor */
extern T0* T1528f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1809c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1808f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].make_empty */
extern T0* T1807c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].default_create */
extern T0* T1808c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_library */
extern T0* T53f129(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_library */
extern T0* T53f129ac1(GE_context* ac, T0* a1);
extern T1 T53f129ot1(T0* a1, T0** a2);
extern T1 T53f129ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_notes */
extern void T203f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_conditions */
extern void T203f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_visible_classes */
extern void T203f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_class_renamings */
extern void T203f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_class_options */
extern void T203f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_use_application_options */
extern void T203f26(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_read_only */
extern void T203f25(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_classname_prefix */
extern void T203f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_description */
extern void T203f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_library */
extern T0* T158f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.make */
extern T0* T203c22(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.make_adapted_library */
extern void T203f33(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.default_read_only_value */
extern T1 T203f18(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.tokens */
extern T0* T203f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_override_clusters */
extern T0* T53f100(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_AST_FACTORY.new_clusters */
extern T0* T158f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.make */
extern T0* T761c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].put_last */
extern void T1527f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_override_cluster */
extern T0* T53f128(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_override_cluster */
extern T0* T53f128ac1(GE_context* ac, T0* a1);
extern T1 T53f128ot1(T0* a1, T0** a2);
extern T1 T53f128ot2(T0* a1, T0** a2);
/* detachable ET_ECF_CLUSTER.set_conditions */
extern void T776f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_conditioned_file_rules */
extern void T776f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_conditioned_subclusters */
extern void T776f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_provider_groups */
extern T0* T53f159(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_provider_group */
extern void T53f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_provider_groups */
extern T0* T158f42(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_overridden_group */
extern T0* T53f166(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CLUSTER.set_description */
extern void T776f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_cluster */
extern T0* T158f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_CLUSTER.make */
extern T0* T776c69(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_file_rules */
extern T0* T53f98(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_FILE_RULES.put_last */
extern void T764f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].force_last */
extern void T1530f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].force */
extern void T1813f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].fill_with */
extern void T1814f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].extend */
extern void T1814f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].resize */
extern void T1530f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].aliased_resized_area */
extern T0* T1813f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].aliased_resized_area */
extern T0* T1814f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].new_capacity */
extern T6 T1530f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].extendible */
extern T1 T1530f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_file_rules */
extern T0* T158f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.make */
extern T0* T764c4(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].put_last */
extern void T1530f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1530c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].set_internal_cursor */
extern void T1530f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].new_cursor */
extern T0* T1530f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1815c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1813f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].make_empty */
extern T0* T1814c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].default_create */
extern T0* T1813c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_file_rule */
extern T0* T53f127(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f127ot1(T0* a1, T0** a2);
/* detachable ET_ECF_FILE_RULE.set_description */
extern void T784f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULE.set_conditions */
extern void T784f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_file_rule */
extern T0* T158f34(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULE.make */
extern T0* T784c11(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULE.compiled_regexp */
extern T0* T784f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].forth */
extern void T372f14(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].put_last */
extern void T1541f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].force */
extern void T1837f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].fill_with */
extern void T1838f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].extend */
extern void T1838f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_compiled */
extern T1 T83f307(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].item */
extern T0* T372f6(GE_context* ac, T0* C);
/* detachable ET_ECF_FILE_RULE.operating_system */
extern T0* T784f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].start */
extern void T372f13(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1541c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].set_internal_cursor */
extern void T1541f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].new_cursor */
extern T0* T1541f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1839c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1837f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].make_empty */
extern T0* T1838c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].default_create */
extern T0* T1837c2(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_resources */
extern T0* T53f96(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_RESOURCES.put_last */
extern void T771f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].force_last */
extern void T1537f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].force */
extern void T1834f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].fill_with */
extern void T1835f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].extend */
extern void T1835f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].resize */
extern void T1537f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].aliased_resized_area */
extern T0* T1834f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].aliased_resized_area */
extern T0* T1835f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].new_capacity */
extern T6 T1537f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].extendible */
extern T1 T1537f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_resources */
extern T0* T158f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCES.make */
extern T0* T771c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].put_last */
extern void T1537f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1537c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].set_internal_cursor */
extern void T1537f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].new_cursor */
extern T0* T1537f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1836c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1834f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].make_empty */
extern T0* T1835c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].default_create */
extern T0* T1834c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_resource */
extern T0* T53f126(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f126ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.set_conditions */
extern void T783f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCE.set_description */
extern void T783f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_resource */
extern T0* T158f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCE.make */
extern T0* T783c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_objects */
extern T0* T53f94(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_OBJECTS.put_last */
extern void T770f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].force_last */
extern void T1536f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].force */
extern void T1831f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].fill_with */
extern void T1832f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].extend */
extern void T1832f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].resize */
extern void T1536f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].aliased_resized_area */
extern T0* T1831f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].aliased_resized_area */
extern T0* T1832f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].new_capacity */
extern T6 T1536f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].extendible */
extern T1 T1536f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_objects */
extern T0* T158f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECTS.make */
extern T0* T770c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].put_last */
extern void T1536f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1536c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].set_internal_cursor */
extern void T1536f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].new_cursor */
extern T0* T1536f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1833c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1831f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].make_empty */
extern T0* T1832c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].default_create */
extern T0* T1831c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_object */
extern T0* T53f125(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f125ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.set_conditions */
extern void T782f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECT.set_description */
extern void T782f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_object */
extern T0* T158f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECT.make */
extern T0* T782c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_makes */
extern T0* T53f92(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_MAKES.put_last */
extern void T769f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].force_last */
extern void T1535f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].force */
extern void T1828f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].fill_with */
extern void T1829f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].extend */
extern void T1829f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].resize */
extern void T1535f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].aliased_resized_area */
extern T0* T1828f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].aliased_resized_area */
extern T0* T1829f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].new_capacity */
extern T6 T1535f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].extendible */
extern T1 T1535f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_makes */
extern T0* T158f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKES.make */
extern T0* T769c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].put_last */
extern void T1535f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1535c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].set_internal_cursor */
extern void T1535f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].new_cursor */
extern T0* T1535f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1830c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1828f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].make_empty */
extern T0* T1829c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].default_create */
extern T0* T1828c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_make */
extern T0* T53f124(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f124ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_MAKE.set_conditions */
extern void T781f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKE.set_description */
extern void T781f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_make */
extern T0* T158f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKE.make */
extern T0* T781c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_linker_flags */
extern T0* T53f90(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.put_last */
extern void T768f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].force_last */
extern void T1534f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].force */
extern void T1825f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].fill_with */
extern void T1826f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].extend */
extern void T1826f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].resize */
extern void T1534f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].aliased_resized_area */
extern T0* T1825f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].aliased_resized_area */
extern T0* T1826f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].new_capacity */
extern T6 T1534f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].extendible */
extern T1 T1534f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_linker_flags */
extern T0* T158f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.make */
extern T0* T768c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].put_last */
extern void T1534f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1534c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].set_internal_cursor */
extern void T1534f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].new_cursor */
extern T0* T1534f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1827c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1825f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make_empty */
extern T0* T1826c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].default_create */
extern T0* T1825c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_linker_flag */
extern T0* T53f123(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f123ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.set_conditions */
extern void T780f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.set_description */
extern void T780f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_linker_flag */
extern T0* T158f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.make */
extern T0* T780c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_libraries */
extern T0* T53f88(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.put_last */
extern void T767f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].force_last */
extern void T1533f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].force */
extern void T1822f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].fill_with */
extern void T1823f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].extend */
extern void T1823f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].resize */
extern void T1533f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].aliased_resized_area */
extern T0* T1822f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].aliased_resized_area */
extern T0* T1823f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].new_capacity */
extern T6 T1533f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].extendible */
extern T1 T1533f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_libraries */
extern T0* T158f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.make */
extern T0* T767c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].put_last */
extern void T1533f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1533c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].set_internal_cursor */
extern void T1533f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].new_cursor */
extern T0* T1533f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1824c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1822f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].make_empty */
extern T0* T1823c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].default_create */
extern T0* T1822c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_library */
extern T0* T53f122(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f122ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.set_conditions */
extern void T779f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARY.set_description */
extern void T779f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_library */
extern T0* T158f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARY.make */
extern T0* T779c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_includes */
extern T0* T53f86(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_INCLUDES.put_last */
extern void T766f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].force_last */
extern void T1532f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].force */
extern void T1819f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].fill_with */
extern void T1820f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].extend */
extern void T1820f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].resize */
extern void T1532f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].aliased_resized_area */
extern T0* T1819f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].aliased_resized_area */
extern T0* T1820f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].new_capacity */
extern T6 T1532f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].extendible */
extern T1 T1532f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_includes */
extern T0* T158f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDES.make */
extern T0* T766c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].put_last */
extern void T1532f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1532c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].set_internal_cursor */
extern void T1532f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].new_cursor */
extern T0* T1532f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1821c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1819f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].make_empty */
extern T0* T1820c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].default_create */
extern T0* T1819c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_include */
extern T0* T53f121(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f121ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.set_conditions */
extern void T778f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDE.set_description */
extern void T778f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_include */
extern T0* T158f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDE.make */
extern T0* T778c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_cflags */
extern T0* T53f84(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_CFLAGS.put_last */
extern void T765f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].force_last */
extern void T1531f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].force */
extern void T1816f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].fill_with */
extern void T1817f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].extend */
extern void T1817f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].resize */
extern void T1531f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].aliased_resized_area */
extern T0* T1816f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].aliased_resized_area */
extern T0* T1817f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].new_capacity */
extern T6 T1531f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].extendible */
extern T1 T1531f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_cflags */
extern T0* T158f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAGS.make */
extern T0* T765c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].put_last */
extern void T1531f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1531c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].set_internal_cursor */
extern void T1531f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].new_cursor */
extern T0* T1531f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1818c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1816f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].make_empty */
extern T0* T1817c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].default_create */
extern T0* T1816c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_cflag */
extern T0* T53f120(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f120ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.set_conditions */
extern void T777f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAG.set_description */
extern void T777f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_cflag */
extern T0* T158f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAG.make */
extern T0* T777c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_TARGET.set_description */
extern void T57f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_clusters */
extern T0* T53f82(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_cluster */
extern T0* T53f119(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_cluster */
extern T0* T53f119ac1(GE_context* ac, T0* a1);
extern T1 T53f119ot1(T0* a1, T0** a2);
extern T1 T53f119ot2(T0* a1, T0** a2);
/* detachable ET_ECF_CLUSTER.set_notes */
extern void T776f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_capabilities */
extern void T53f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T53f233ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assemblies */
extern T0* T53f79(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.put_last */
extern void T763f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].force_last */
extern void T1529f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].force */
extern void T1811f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].fill_with */
extern void T1810f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].extend */
extern void T1810f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].resize */
extern void T1529f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1811f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1810f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].new_capacity */
extern T6 T1529f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].extendible */
extern T1 T1529f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_dotnet_assemblies */
extern T0* T158f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.make */
extern T0* T763c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].put_last */
extern void T1529f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1529c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T1529f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].new_cursor */
extern T0* T1529f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1812c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1811f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make_empty */
extern T0* T1810c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].default_create */
extern T0* T1811c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assembly */
extern T0* T53f118(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assembly */
extern T0* T53f118ac1(GE_context* ac, T0* a1);
extern T1 T53f118ot1(T0* a1, T0** a2);
extern T1 T53f118ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_notes */
extern void T209f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_conditions */
extern void T209f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_class_renamings */
extern void T209f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_class_options */
extern void T209f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_read_only */
extern void T209f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_classname_prefix */
extern void T209f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_version */
extern void T209f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_name */
extern void T209f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_key */
extern void T209f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_culture */
extern void T209f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_description */
extern void T209f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_dotnet_assembly */
extern T0* T158f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.make */
extern T0* T209c24(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.make_adapted_dotnet_assembly */
extern void T209f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T209f20(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_dotnet_assembly */
extern T0* T72f261(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.make */
extern T0* T586c104(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.make_from_system */
extern void T586f105(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.set_default_class_mapping */
extern void T586f110(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_wide_character_ref_class_mapping */
extern void T586f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_class_mapping */
extern void T586f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.set_wide_character_class_mapping */
extern void T586f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_class_mapping */
extern void T586f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_ref_class_mapping */
extern void T586f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_class_mapping */
extern void T586f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_ref_class_mapping */
extern void T586f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_class_mapping */
extern void T586f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_ref_class_mapping */
extern void T586f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_class_mapping */
extern void T586f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_double_ref_class_mapping */
extern void T586f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_double_class_mapping */
extern void T586f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_ref_class_mapping */
extern void T586f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_class_mapping */
extern void T586f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.set_kernel_types */
extern void T586f109(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_wide_character_type */
extern void T586f152(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_typed_pointer_type */
extern void T586f151(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_type_type */
extern void T586f150(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_system_string_type */
extern void T586f149(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_system_object_type */
extern void T586f148(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_type */
extern void T586f147(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_32_type */
extern void T586f146(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_8_type */
extern void T586f145(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_special_type */
extern void T586f144(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_routine_type */
extern void T586f143(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_64_type */
extern void T586f142(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_32_type */
extern void T586f141(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_type */
extern void T586f140(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_procedure_type */
extern void T586f139(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_predicate_type */
extern void T586f138(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_pointer_type */
extern void T586f137(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_64_type */
extern void T586f136(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_32_type */
extern void T586f135(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_16_type */
extern void T586f134(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_8_type */
extern void T586f133(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_type */
extern void T586f132(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_iterable_type */
extern void T586f131(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_ise_exception_manager_type */
extern void T586f130(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_64_type */
extern void T586f129(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_32_type */
extern void T586f128(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_16_type */
extern void T586f127(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_8_type */
extern void T586f126(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_type */
extern void T586f125(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_immutable_string_32_type */
extern void T586f124(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_immutable_string_8_type */
extern void T586f123(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_function_type */
extern void T586f122(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_exception_manager_type */
extern void T586f121(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_exception_type */
extern void T586f120(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_double_type */
extern void T586f119(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_32_type */
extern void T586f118(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_8_type */
extern void T586f117(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_type */
extern void T586f116(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_boolean_type */
extern void T586f115(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_array_type */
extern void T586f114(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_tuple_type */
extern void T586f113(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_none_type */
extern void T586f112(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_any_type */
extern void T586f111(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.make_adapted */
extern void T586f108(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T586f89(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_unknown_kernel_types */
extern void T586f107(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.initialize */
extern void T586f106(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.class_name_tester */
extern T0* T586f94(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.tokens */
extern T0* T209f19(GE_context* ac, T0* C);
/* [detachable] ET_ECF_TARGET.set_parent */
extern void T57f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.execution_environment */
extern T0* T57f28(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.set_secondary_variables */
extern void T60f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGET_PARENT.make */
extern T0* T196c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.set_secondary_options */
extern void T437f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.set_secondary_capabilities */
extern void T59f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.set_secondary_settings */
extern void T55f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_target */
extern T0* T158f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.make */
extern T0* T57c32(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_CAPABILITIES.make */
extern T0* T59c13(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_CAPABILITIES.make */
extern T0* T59f13ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_hash_function */
extern void T217f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.string_ */
extern T0* T59f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_key_equality_tester */
extern void T217f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.case_insensitive_string_equality_tester */
extern T0* T59f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_map */
extern T0* T217c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_with_equality_testers */
extern void T217f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_sparse_container */
extern void T217f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_internal_cursor */
extern void T217f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_cursor */
extern T0* T217f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].make */
extern T0* T591c10(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_slots */
extern void T217f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_clashes */
extern void T217f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_key_storage */
extern void T217f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_item_storage */
extern void T217f56(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_VARIABLES.make */
extern T0* T60c10(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_VARIABLES.make */
extern T0* T60f10ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_VARIABLES.case_insensitive_string_equality_tester */
extern T0* T60f4(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.execution_environment */
extern T0* T60f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_notes */
extern T0* T53f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].force_last */
extern void T438f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].force */
extern void T904f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].fill_with */
extern void T905f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].extend */
extern void T905f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].resize */
extern void T438f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].aliased_resized_area */
extern T0* T904f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].aliased_resized_area */
extern T0* T905f3(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].new_capacity */
extern T6 T438f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].extendible */
extern T1 T438f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_notes */
extern T0* T158f5(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make */
extern void T438f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T438c10(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].set_internal_cursor */
extern void T438f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].new_cursor */
extern T0* T438f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T906c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T904f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].make_empty */
extern T0* T905c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].default_create */
extern T0* T904c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_note_element */
extern T0* T53f72(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T53f72ot1(T0* a1, T0** a2);
extern T1 T53f72ot2(T0* a1, T0** a2);
/* detachable ET_ECF_NOTE_ELEMENT.set_content */
extern void T760f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.add_element */
extern void T760f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.add_attribute */
extern void T760f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_AST_FACTORY.new_note */
extern T0* T158f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.make */
extern T0* T760c7(GE_context* ac, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_NOTE_ELEMENT.make */
extern T0* T760f7ac1(GE_context* ac, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make_default */
extern T0* T438c14(GE_context* ac);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].default_capacity */
extern T6 T438f9(GE_context* ac, T0* C);
/* detachable ET_ECF_NOTE_ELEMENT.string_ */
extern T0* T760f6(GE_context* ac, T0* C);
/* detachable ET_ECF_NOTE_ELEMENT.case_insensitive_string_equality_tester */
extern T0* T760f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map_default */
extern T0* T212c93(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].default_capacity */
extern T6 T212f46(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.text */
extern T0* T164f14(GE_context* ac, T0* C);
extern T1 T164f14ot1(T0* a1, T0** a2);
/* [detachable] XM_ELEMENT.string_ */
extern T0* T164f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.string_ */
extern T0* T53f20(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].item */
extern T0* T440f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].start */
extern void T440f11(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_start */
extern void T869f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_off */
extern T1 T869f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].off */
extern T1 T440f9(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.new_cursor */
extern T0* T164f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].new_cursor */
extern T0* T869f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].make */
extern T0* T440c10(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options */
extern T0* T53f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_0_0 */
extern T0* T53f71(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_0_0 */
extern void T53f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_5_0 */
extern void T53f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_9_0 */
extern void T53f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_11_0 */
extern void T53f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_15_0 */
extern void T53f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_18_0 */
extern void T53f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_latest */
extern void T53f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_5_0 */
extern T0* T53f70(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_9_0 */
extern T0* T53f68(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_9_0 */
extern T0* T53f67(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_11_0 */
extern T0* T53f66(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_15_0 */
extern T0* T53f65(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_15_0 */
extern T0* T53f64(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_18_0 */
extern T0* T53f63(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.is_greater_equal */
extern T1 T44f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_18_0 */
extern T0* T53f62(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities */
extern T0* T53f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_0_0 */
extern T0* T53f61(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_0_0 */
extern void T53f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_11_0 */
extern void T53f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_13_0 */
extern void T53f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_16_0 */
extern void T53f223(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_latest */
extern void T53f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_11_0 */
extern T0* T53f60(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_13_0 */
extern T0* T53f58(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_13_0 */
extern T0* T53f57(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_16_0 */
extern T0* T53f56(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_16_0 */
extern T0* T53f55(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings */
extern T0* T53f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings_1_0_0 */
extern T0* T53f54(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_1_0_0 */
extern void T53f222(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_1_20_0 */
extern void T53f221(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_latest */
extern void T53f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings_1_20_0 */
extern T0* T53f53(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_20_0 */
extern T0* T53f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_latest */
extern T0* T53f38(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.make_latest */
extern T0* T44c20(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_version */
extern T0* T53f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eate_error */
extern void T54f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eate */
extern T0* T210c102(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.attribute_name */
extern T0* T53f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].found_item */
extern T0* T170f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].found */
extern T1 T170f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].search */
extern void T170f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].search_position */
extern void T170f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatm_error */
extern void T54f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eatm */
extern T0* T210c101(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_library */
extern T0* T158f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.make */
extern T0* T183c118(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.make_library */
extern void T183f127(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.make_from_system */
extern void T183f128(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.set_default_class_mapping */
extern void T183f133(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_wide_character_ref_class_mapping */
extern void T183f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_class_mapping */
extern void T183f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.set_wide_character_class_mapping */
extern void T183f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_string_class_mapping */
extern void T183f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_real_ref_class_mapping */
extern void T183f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_real_class_mapping */
extern void T183f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_natural_ref_class_mapping */
extern void T183f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_natural_class_mapping */
extern void T183f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_integer_ref_class_mapping */
extern void T183f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_integer_class_mapping */
extern void T183f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_double_ref_class_mapping */
extern void T183f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_double_class_mapping */
extern void T183f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_character_ref_class_mapping */
extern void T183f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_character_class_mapping */
extern void T183f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.tokens */
extern T0* T183f100(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_kernel_types */
extern void T183f132(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_wide_character_type */
extern void T183f175(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_typed_pointer_type */
extern void T183f174(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_type_type */
extern void T183f173(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_system_string_type */
extern void T183f172(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_system_object_type */
extern void T183f171(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_type */
extern void T183f170(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_32_type */
extern void T183f169(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_8_type */
extern void T183f168(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_special_type */
extern void T183f167(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_routine_type */
extern void T183f166(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_64_type */
extern void T183f165(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_32_type */
extern void T183f164(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_type */
extern void T183f163(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_procedure_type */
extern void T183f162(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_predicate_type */
extern void T183f161(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_pointer_type */
extern void T183f160(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_64_type */
extern void T183f159(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_32_type */
extern void T183f158(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_16_type */
extern void T183f157(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_8_type */
extern void T183f156(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_type */
extern void T183f155(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_iterable_type */
extern void T183f154(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_ise_exception_manager_type */
extern void T183f153(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_64_type */
extern void T183f152(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_32_type */
extern void T183f151(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_16_type */
extern void T183f150(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_8_type */
extern void T183f149(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_type */
extern void T183f148(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_immutable_string_32_type */
extern void T183f147(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_immutable_string_8_type */
extern void T183f146(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_function_type */
extern void T183f145(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_exception_manager_type */
extern void T183f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_exception_type */
extern void T183f143(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_double_type */
extern void T183f142(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_32_type */
extern void T183f141(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_8_type */
extern void T183f140(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_type */
extern void T183f139(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_boolean_type */
extern void T183f138(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_array_type */
extern void T183f137(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_tuple_type */
extern void T183f136(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_none_type */
extern void T183f135(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_any_type */
extern void T183f134(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.make_adapted */
extern void T183f131(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.default_read_only_value */
extern T1 T183f106(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_unknown_kernel_types */
extern void T183f130(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.initialize */
extern void T183f129(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.initialize */
extern void T183f129p1(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.class_name_tester */
extern T0* T183f101(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.attribute_by_name */
extern T0* T164f5(GE_context* ac, T0* C, T0* a1);
extern T1 T164f5ot1(T0* a1, T0** a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].go_after */
extern void T440f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_go_after */
extern void T869f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_after */
extern void T440f16(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.attribute_same_name */
extern T1 T164f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.same_string */
extern T1 T164f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.string_equality_tester */
extern T0* T164f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].forth */
extern void T201f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_forth */
extern void T200f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].add_traversing_cursor */
extern void T200f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_next_cursor */
extern void T201f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_position */
extern void T201f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove_traversing_cursor */
extern void T200f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_after */
extern void T201f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_item */
extern T6 T200f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadp_error */
extern void T54f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eadp */
extern T0* T210c100(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ANDED_CONDITIONS.is_capability_supported */
extern T1 T793f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ANDED_CONDITIONS.is_capability_supported */
extern T1 T793f3p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].item */
extern T0* T1551f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_capability_supported */
extern T1 T792f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].item */
extern T0* T1550f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_capability_supported */
extern T1 T792f3p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CAPABILITIES.is_capability_supported */
extern T1 T59f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].found_item */
extern T0* T1693f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_item */
extern T0* T1693f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].found */
extern T1 T1693f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].search */
extern void T1693f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].search_position */
extern void T1693f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].hash_position */
extern T6 T1693f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_item */
extern T0* T1693f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_item */
extern T6 T1693f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_item */
extern T6 T1693f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CAPABILITIES.capability_order */
extern T0* T59f12(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_CAPABILITIES.capability_order */
extern T0* T59f12ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].put_last */
extern void T1693f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_put */
extern void T1693f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_put */
extern void T1693f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_put */
extern void T1693f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].compress */
extern void T1693f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_wipe_out */
extern void T1693f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_wipe_out */
extern void T1693f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_keep_head */
extern void T1693f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_keep_head */
extern void T1693f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].keep_head */
extern void T1931f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].set_count */
extern void T1931f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].move_all_cursors */
extern void T1693f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_position */
extern void T1935f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_put */
extern void T1693f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].force */
extern void T1934f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].fill_with */
extern void T1931f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].extend */
extern void T1931f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].unset_found_item */
extern void T1693f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].put_last */
extern void T1692f39(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].compress */
extern void T1692f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_wipe_out */
extern void T1692f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_wipe_out */
extern void T1692f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_keep_head */
extern void T1692f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_keep_head */
extern void T1692f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].move_all_cursors */
extern void T1692f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8].set_position */
extern void T1930f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_hash_function */
extern void T1692f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_hash_function */
extern void T1693f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_key_equality_tester */
extern void T1693f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_map */
extern T0* T1693c32(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_with_equality_testers */
extern void T1693f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_sparse_container */
extern void T1693f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_internal_cursor */
extern void T1693f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].new_cursor */
extern T0* T1693f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make */
extern T0* T1935c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_slots */
extern void T1693f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].new_modulus */
extern T6 T1693f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_clashes */
extern void T1693f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_key_storage */
extern void T1693f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_item_storage */
extern void T1693f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].make */
extern T0* T1934f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].make_empty */
extern T0* T1931c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].default_create */
extern T0* T1934c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key */
extern T0* T201f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_key */
extern T0* T200f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_item */
extern T0* T200f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].after */
extern T1 T201f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_iterator */
extern T0* T200f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].start */
extern void T201f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_start */
extern void T200f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_off */
extern T1 T200f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].off */
extern T1 T201f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].is_empty */
extern T1 T200f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor */
extern T0* T200f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T201c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.valid_capabilities_latest */
extern T0* T53f28(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.valid_capabilities_latest */
extern T0* T53f28ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.void_safety_detection_capability_value_regexp */
extern T0* T53f35(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.concurrency_capability_value_regexp */
extern T0* T53f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].force_last */
extern void T200f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_put */
extern void T200f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_put */
extern void T200f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_put */
extern void T200f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_item */
extern T6 T200f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].hash_position */
extern T6 T200f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].resize */
extern void T200f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_resize */
extern void T200f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].special_integer_ */
extern T0* T200f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_resize */
extern void T200f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_resize */
extern void T200f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].aliased_resized_area */
extern T0* T582f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].aliased_resized_area */
extern T0* T583f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_resize */
extern void T200f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_modulus */
extern T6 T200f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_capacity */
extern T6 T200f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_put */
extern void T200f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].force */
extern void T582f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].fill_with */
extern void T583f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].extend */
extern void T583f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].search_position */
extern void T200f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].unset_found_item */
extern void T200f44(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.catcall_detection_capability_value_regexp */
extern T0* T53f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_hash_function */
extern void T200f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_key_equality_tester */
extern void T200f40(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].internal_set_equality_tester */
extern void T581f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.case_insensitive_string_equality_tester */
extern T0* T53f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_map */
extern T0* T200c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_with_equality_testers */
extern void T200f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_sparse_container */
extern void T200f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_internal_cursor */
extern void T200f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_slots */
extern void T200f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_clashes */
extern void T200f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_key_storage */
extern void T200f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_item_storage */
extern void T200f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].make */
extern T0* T582f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].make_empty */
extern T0* T583c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].default_create */
extern T0* T582c3(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabw_error */
extern void T54f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabw */
extern T0* T210c99(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_library */
extern void T203f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabv_error */
extern void T54f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabv */
extern T0* T210c98(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.pathname */
extern T0* T203f20(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_LIBRARIES.library */
extern T0* T184f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_LIBRARIES.count */
extern T6 T184f3(GE_context* ac, T0* C);
/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY].make */
extern T0* T202c2(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.select_target */
extern void T53f215(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.select_target */
extern T0* T53f215ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_system_config */
extern void T53f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG].put */
extern void T197f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_system_config */
extern T0* T53f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_system_config */
extern T0* T158f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_CONFIG.make */
extern T0* T181c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_TARGET.fill_universe */
extern void T57f54(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T57f54ot1(T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.fill_external_linker_flags */
extern void T768f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_LINKER_FLAGS.fill_external_linker_flags */
extern T0* T768f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]].set_rout_disp_final */
extern void T1744f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.fill_external_linker_flags */
extern void T780f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.add_external_value */
extern void T780f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.is_enabled */
extern T1 T780f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ANDED_CONDITIONS.is_enabled */
extern T1 T793f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_enabled */
extern T1 T792f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].do_all */
extern void T1534f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].item */
extern T0* T1534f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_CFLAGS.fill_external_cflags */
extern void T765f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_CFLAGS.fill_external_cflags */
extern T0* T765f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]].set_rout_disp_final */
extern void T1737f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_CFLAG.fill_external_cflags */
extern void T777f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.add_external_value */
extern void T777f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.is_enabled */
extern T1 T777f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].do_all */
extern void T1531f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].item */
extern T0* T1531f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_MAKES.fill_external_makes */
extern void T769f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_MAKES.fill_external_makes */
extern T0* T769f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]].set_rout_disp_final */
extern void T1746f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_MAKE.fill_external_makes */
extern void T781f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_MAKE.add_external_value */
extern void T781f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_MAKE.file_system */
extern T0* T781f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.unix_file_system */
extern T0* T781f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.windows_file_system */
extern T0* T781f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.operating_system */
extern T0* T781f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.is_enabled */
extern T1 T781f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].do_all */
extern void T1535f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].item */
extern T0* T1535f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_RESOURCES.fill_external_resources */
extern void T771f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_RESOURCES.fill_external_resources */
extern T0* T771f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]].set_rout_disp_final */
extern void T1750f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_RESOURCE.fill_external_resources */
extern void T783f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.add_external_value */
extern void T783f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.file_system */
extern T0* T783f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.unix_file_system */
extern T0* T783f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.windows_file_system */
extern T0* T783f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.operating_system */
extern T0* T783f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.is_enabled */
extern T1 T783f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].do_all */
extern void T1537f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].item */
extern T0* T1537f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.fill_external_libraries */
extern void T767f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_LIBRARIES.fill_external_libraries */
extern T0* T767f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]].set_rout_disp_final */
extern void T1742f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_LIBRARY.fill_external_libraries */
extern void T779f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.add_external_value */
extern void T779f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.file_system */
extern T0* T779f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.unix_file_system */
extern T0* T779f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.windows_file_system */
extern T0* T779f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.operating_system */
extern T0* T779f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.is_enabled */
extern T1 T779f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].do_all */
extern void T1533f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].item */
extern T0* T1533f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_OBJECTS.fill_external_objects */
extern void T770f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_OBJECTS.fill_external_objects */
extern T0* T770f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]].set_rout_disp_final */
extern void T1748f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_OBJECT.fill_external_objects */
extern void T782f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.add_external_value */
extern void T782f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.file_system */
extern T0* T782f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.unix_file_system */
extern T0* T782f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.windows_file_system */
extern T0* T782f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.operating_system */
extern T0* T782f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.is_enabled */
extern T1 T782f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].do_all */
extern void T1536f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].item */
extern T0* T1536f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_INCLUDES.fill_external_includes */
extern void T766f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_INCLUDES.fill_external_includes */
extern T0* T766f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]].set_rout_disp_final */
extern void T1740f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_INCLUDE.fill_external_includes */
extern void T778f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.add_external_value */
extern void T778f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.file_system */
extern T0* T778f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.unix_file_system */
extern T0* T778f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.windows_file_system */
extern T0* T778f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.operating_system */
extern T0* T778f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.is_enabled */
extern T1 T778f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].do_all */
extern void T1532f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].item */
extern T0* T1532f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_dotnet */
extern void T56f254(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.put_last */
extern void T206f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].force_last */
extern void T584f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].force */
extern void T1354f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].fill_with */
extern void T1353f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].extend */
extern void T1353f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].resize */
extern void T584f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1354f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1353f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].new_capacity */
extern T6 T584f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].extendible */
extern T1 T584f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.is_enabled */
extern T1 T209f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T763f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].item */
extern T0* T1529f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.count */
extern T6 T763f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_STATE.is_dotnet */
extern T1 T179f3(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_LIBRARIES.put_last */
extern void T184f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].force_last */
extern void T568f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].force */
extern void T1349f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].fill_with */
extern void T1348f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].extend */
extern void T1348f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].resize */
extern void T568f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1349f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1348f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].new_capacity */
extern T6 T568f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].extendible */
extern T1 T568f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.is_enabled */
extern T1 T203f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.library */
extern T0* T762f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].item */
extern T0* T1528f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.count */
extern T6 T762f3(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.fill_file_rules */
extern void T776f93(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_CLUSTER.fill_file_rules */
extern T0* T776f93ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]].set_rout_disp_final */
extern void T1681f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_CLUSTERS.do_all */
extern void T761f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.fill_file_rules */
extern void T764f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULES.fill_file_rules */
extern T0* T764f7ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_FILE_RULES.fill_file_rules */
extern T0* T764f7ac2(GE_context* ac, T0* a1);
/* detachable ET_ECF_FILE_RULE.is_enabled */
extern T1 T784f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]].set_rout_disp_final */
extern void T2004f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].do_if */
extern void T1530f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.fill_file_rules */
extern void T57f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULES.make_empty */
extern T0* T764c6(GE_context* ac);
/* detachable ET_ECF_CLUSTER.fill_subclusters */
extern void T776f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTERS.put_last */
extern void T554f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].force_last */
extern void T1342f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].force */
extern void T1674f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].fill_with */
extern void T1673f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].extend */
extern void T1673f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].resize */
extern void T1342f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].aliased_resized_area */
extern T0* T1674f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].aliased_resized_area */
extern T0* T1673f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].new_capacity */
extern T6 T1342f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].extendible */
extern T1 T1342f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.is_enabled */
extern T1 T776f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.cluster */
extern T0* T761f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTERS.count */
extern T6 T761f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcd_error */
extern void T54f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ECF_ERROR.make_epcd */
extern T0* T210c97(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcl_error */
extern void T54f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_epcl */
extern T0* T210c96(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcr_error */
extern void T54f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR.make_epcr */
extern T0* T210c95(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadq_error */
extern void T54f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadq */
extern T0* T210c94(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].before */
extern T1 T195f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_before */
extern T1 T195f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].before */
extern T1 T576f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].back */
extern void T195f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_back */
extern void T195f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].add_traversing_cursor */
extern void T195f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_next_cursor */
extern void T576f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_position */
extern void T576f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].remove_traversing_cursor */
extern void T195f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_before */
extern void T576f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_item */
extern T6 T195f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].finish */
extern void T195f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_finish */
extern void T195f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_off */
extern T1 T195f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].off */
extern T1 T576f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].is_empty */
extern T1 T195f36(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGET_PARENT.set_target */
extern void T196f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcy_error */
extern void T54f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epcy */
extern T0* T210c93(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].put_last */
extern void T143f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_for_iteration */
extern T0* T195f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_item */
extern T0* T195f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_item */
extern T0* T195f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].forth */
extern void T195f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_forth */
extern void T195f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_after */
extern void T576f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_for_iteration */
extern T0* T195f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_key */
extern T0* T195f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_item */
extern T0* T195f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].after */
extern T1 T195f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_after */
extern T1 T195f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].after */
extern T1 T576f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].start */
extern void T195f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_start */
extern void T195f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].has */
extern T1 T195f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].search_position */
extern void T195f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].hash_position */
extern T6 T195f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_item */
extern T6 T195f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_internal_error */
extern void T54f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_internal */
extern T0* T210c92(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epud_error */
extern void T54f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epud */
extern T0* T210c91(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epul_error */
extern void T54f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_epul */
extern T0* T210c90(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epur_error */
extern void T54f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epur */
extern T0* T210c89(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epfe_error */
extern void T54f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epfe */
extern T0* T210c88(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].found_item */
extern T0* T194f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_item */
extern T0* T194f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].found */
extern T1 T194f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].search */
extern void T194f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].search_position */
extern void T194f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].hash_position */
extern T6 T194f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_item */
extern T0* T194f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_item */
extern T6 T194f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_item */
extern T6 T194f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG].make */
extern T0* T197c2(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].force_last */
extern void T195f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_put */
extern void T195f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_put */
extern void T195f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_put */
extern void T195f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].resize */
extern void T195f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_resize */
extern void T195f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].special_integer_ */
extern T0* T195f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_resize */
extern void T195f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_resize */
extern void T195f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_resize */
extern void T195f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_modulus */
extern T6 T195f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_capacity */
extern T6 T195f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_put */
extern void T195f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].unset_found_item */
extern void T195f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_default */
extern T0* T195c43(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make */
extern void T195f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_with_equality_testers */
extern void T195f61(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_sparse_container */
extern void T195f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].set_internal_cursor */
extern void T195f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_cursor */
extern T0* T195f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].make */
extern T0* T576c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_slots */
extern void T195f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_clashes */
extern void T195f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_key_storage */
extern void T195f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_item_storage */
extern void T195f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].default_capacity */
extern T6 T195f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].force_last */
extern void T194f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_put */
extern void T194f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_put */
extern void T194f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_put */
extern void T194f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].resize */
extern void T194f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_resize */
extern void T194f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].special_integer_ */
extern T0* T194f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_resize */
extern void T194f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_resize */
extern void T194f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].aliased_resized_area */
extern T0* T574f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].aliased_resized_area */
extern T0* T572f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_resize */
extern void T194f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_modulus */
extern T6 T194f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_capacity */
extern T6 T194f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_put */
extern void T194f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].force */
extern void T574f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].fill_with */
extern void T572f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].extend */
extern void T572f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].unset_found_item */
extern void T194f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_default */
extern T0* T194c34(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make */
extern void T194f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_with_equality_testers */
extern void T194f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_sparse_container */
extern void T194f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].set_internal_cursor */
extern void T194f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_cursor */
extern T0* T194f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make */
extern T0* T575c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_slots */
extern void T194f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_clashes */
extern void T194f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_key_storage */
extern void T194f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_item_storage */
extern void T194f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].make */
extern T0* T574f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].make_empty */
extern T0* T572c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].default_create */
extern T0* T574c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].default_capacity */
extern T6 T194f27(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.select_target */
extern void T183f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.select_target */
extern void T56f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_STATE.make */
extern T0* T179c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_etnu_error */
extern void T54f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_etnu */
extern T0* T210c81(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.element_name */
extern T0* T53f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.target_with_name */
extern T0* T56f141(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.target */
extern T0* T199f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_system */
extern T0* T53f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_system */
extern T0* T158f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.make */
extern T0* T56c158(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.make_system */
extern void T56f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_default_class_mapping */
extern void T56f185(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_wide_character_ref_class_mapping */
extern void T56f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_class_mapping */
extern void T56f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.set_wide_character_class_mapping */
extern void T56f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_string_class_mapping */
extern void T56f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_real_ref_class_mapping */
extern void T56f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_real_class_mapping */
extern void T56f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_natural_ref_class_mapping */
extern void T56f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_natural_class_mapping */
extern void T56f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_integer_ref_class_mapping */
extern void T56f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_integer_class_mapping */
extern void T56f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_double_ref_class_mapping */
extern void T56f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_double_class_mapping */
extern void T56f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_character_ref_class_mapping */
extern void T56f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_character_class_mapping */
extern void T56f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.tokens */
extern T0* T56f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_kernel_types */
extern void T56f184(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_wide_character_type */
extern void T56f235(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_typed_pointer_type */
extern void T56f234(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_type_type */
extern void T56f233(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_string_type */
extern void T56f232(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_object_type */
extern void T56f231(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_type */
extern void T56f230(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_32_type */
extern void T56f229(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_8_type */
extern void T56f228(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_special_type */
extern void T56f227(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_routine_type */
extern void T56f226(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_64_type */
extern void T56f225(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_32_type */
extern void T56f224(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_type */
extern void T56f223(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_procedure_type */
extern void T56f222(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_predicate_type */
extern void T56f221(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_pointer_type */
extern void T56f220(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_64_type */
extern void T56f219(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_32_type */
extern void T56f218(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_16_type */
extern void T56f217(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_8_type */
extern void T56f216(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_type */
extern void T56f215(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_iterable_type */
extern void T56f214(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_ise_exception_manager_type */
extern void T56f213(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_64_type */
extern void T56f212(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_32_type */
extern void T56f211(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_16_type */
extern void T56f210(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_8_type */
extern void T56f209(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_type */
extern void T56f208(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_immutable_string_32_type */
extern void T56f207(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_immutable_string_8_type */
extern void T56f206(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_function_type */
extern void T56f205(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_exception_manager_type */
extern void T56f204(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_exception_type */
extern void T56f203(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_double_type */
extern void T56f202(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_32_type */
extern void T56f201(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_8_type */
extern void T56f200(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_type */
extern void T56f199(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_boolean_type */
extern void T56f198(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_array_type */
extern void T56f197(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_tuple_type */
extern void T56f196(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_none_type */
extern void T56f195(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_any_type */
extern void T56f194(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.make_adapted */
extern void T56f183(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.default_read_only_value */
extern T1 T56f143(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_unknown_kernel_types */
extern void T56f182(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_name */
extern void T56f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.initialize */
extern void T56f180(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.initialize */
extern void T56f180p1(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.class_name_tester */
extern T0* T56f145(GE_context* ac, T0* C);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]].set_rout_disp_final */
extern void T161f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM_PARSER.override_target */
extern void T53f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_variables */
extern void T57f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_capabilities */
extern void T57f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].forth */
extern void T591f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_forth */
extern void T217f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].add_traversing_cursor */
extern void T217f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_next_cursor */
extern void T591f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_position */
extern void T591f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].remove_traversing_cursor */
extern void T217f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_after */
extern void T591f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].item */
extern T0* T591f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_item */
extern T0* T217f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].key */
extern T0* T591f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_key */
extern T0* T217f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].after */
extern T1 T591f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_iterator */
extern T0* T217f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].start */
extern void T591f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_start */
extern void T217f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_off */
extern T1 T217f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].off */
extern T1 T591f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].is_empty */
extern T1 T217f37(GE_context* ac, T0* C);
/* [detachable] ET_ECF_TARGET.override_settings */
extern void T57f56(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_settings */
extern T0* T57f56ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_TARGET.override_all_assertions */
extern void T57f62(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T57f62ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T57f62ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T57f62ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T1687f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]].set_rout_disp_final */
extern void T1685f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_cluster */
extern void T57f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_cluster */
extern T0* T57f64ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_adapted_group */
extern void T57f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_adapted_group */
extern T0* T57f65ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.do_adapted */
extern void T763f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.do_adapted */
extern void T762f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.do_explicit */
extern void T761f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.do_recursive */
extern void T761f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_option */
extern void T57f63(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_option */
extern T0* T57f63ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]].set_rout_disp_final */
extern void T1690f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].do_all */
extern void T581f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].forth */
extern void T1981f7(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_forth */
extern void T581f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item */
extern T0* T1981f4(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_item */
extern T0* T581f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].after */
extern T1 T1981f3(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].start */
extern void T1981f6(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_start */
extern void T581f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor */
extern T0* T581f4(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T1981c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].keys */
extern T0* T200f35(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T581c7(GE_context* ac, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_internal_cursor */
extern void T581f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor_with_table_cursor */
extern T0* T581f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_with_table_cursor */
extern T0* T1981c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_TARGETS.do_all */
extern void T199f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].do_all */
extern void T881f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_file_with_action */
extern void T53f214(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.last_error_description */
extern T0* T174f120(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.position */
extern T0* T174f114(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].first */
extern T0* T509f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.positions */
extern T0* T174f118(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_positions */
extern T0* T174f88(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].remove */
extern void T510f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].item */
extern T0* T510f4(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].is_empty */
extern T1 T510f3(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].force_last */
extern void T509f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].put_right */
extern void T1316f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].attach_left */
extern void T1316f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].make */
extern T0* T1316c4(GE_context* ac, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].make */
extern T0* T509c7(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].set_internal_cursor */
extern void T509f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].new_cursor */
extern T0* T509f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION].make */
extern T0* T1317c3(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].copy */
extern void T510f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER].put_right */
extern void T1318f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER].make */
extern T0* T1318c3(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].make */
extern T0* T510c5(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.report_syntax_error */
extern void T54f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_syntax */
extern T0* T210c87(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_TREE_CALLBACKS_PIPE.last_error */
extern T0* T175f5(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.make */
extern T0* T189c11(GE_context* ac, T6 a1, T6 a2);
/* [detachable] ET_COMPRESSED_POSITION.set_position */
extern void T189f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabx_error */
extern void T54f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eabx */
extern T0* T210c86(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadf_error */
extern void T54f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadf */
extern T0* T210c85(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadi_error */
extern void T54f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadi */
extern T0* T210c84(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable DS_HASH_SET [[attached] STRING_8].remove */
extern void T69f43(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].remove_position */
extern void T69f59(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_put */
extern void T69f76(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_keep_head */
extern void T69f75(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_keep_head */
extern void T69f74(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].move_cursors_forth */
extern void T69f73(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].internal_set_key_equality_tester */
extern void T69f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.string_equality_tester */
extern T0* T53f13(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.execution_environment */
extern T0* T53f25(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadh_error */
extern void T54f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eadh */
extern T0* T210c83(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadg_error */
extern void T54f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eadg */
extern T0* T210c82(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_TREE_CALLBACKS_PIPE.document */
extern T0* T175f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.is_correct */
extern T1 T174f90(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.parse_from_stream */
extern void T174f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.parse_with_events */
extern void T174f213(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_finish */
extern void T174f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.parse */
extern void T174f218(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_pop_last_value */
extern void T174f228(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.yy_push_error_value */
extern void T174f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_do_action */
extern void T174f226(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.on_notation_declaration */
extern void T174f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_notation_declaration */
extern void T516f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_EXTERNAL_ID.set_public */
extern void T526f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].aliased_resized_area */
extern T0* T525f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].aliased_resized_area */
extern T0* T527f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_EXTERNAL_ID.set_system */
extern void T526f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_external_id */
extern T0* T174f168(GE_context* ac, T0* C);
/* [detachable] XM_DTD_EXTERNAL_ID.make */
extern T0* T526c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.when_pe_entity_declared */
extern void T174f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].force_new */
extern void T512f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_put */
extern void T512f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_put */
extern void T512f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].force */
extern void T1323f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].fill_with */
extern void T1319f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].extend */
extern void T1319f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_put */
extern void T512f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_put */
extern void T512f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_item */
extern T6 T512f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].hash_position */
extern T6 T512f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_item */
extern T6 T512f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].resize */
extern void T512f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_resize */
extern void T512f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].special_integer_ */
extern T0* T512f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_resize */
extern void T512f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_resize */
extern void T512f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].aliased_resized_area */
extern T0* T1323f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].aliased_resized_area */
extern T0* T1319f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_item */
extern T0* T512f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_resize */
extern void T512f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_modulus */
extern T6 T512f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_capacity */
extern T6 T512f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].unset_found_item */
extern void T512f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].has */
extern T1 T512f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].search_position */
extern void T512f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_def */
extern T0* T549c214(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_literal */
extern void T549f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.default_create */
extern T0* T514c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_scanner */
extern void T549f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T549f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset_sent */
extern void T549f222(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T549f219p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T549f219p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T549f219p2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T549f219p3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_with_buffer */
extern void T549f218(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_load_input_buffer */
extern void T549f221(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_set_content */
extern void T549f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_initialize */
extern void T549f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_initialize */
extern void T549f220p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_build_tables */
extern void T549f223(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template */
extern T0* T549f54(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_fixed_array */
extern T0* T549f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_5 */
extern void T549f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_array_subcopy */
extern void T549f282(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.integer_array_ */
extern T0* T549f60(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_4 */
extern void T549f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_3 */
extern void T549f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_2 */
extern void T549f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_1 */
extern void T549f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_meta_template */
extern T0* T549f52(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template */
extern T0* T549f73(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template_2 */
extern void T549f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template_1 */
extern void T549f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template */
extern T0* T549f68(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_5 */
extern void T549f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_4 */
extern void T549f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_3 */
extern void T549f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_2 */
extern void T549f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_1 */
extern void T549f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template */
extern T0* T549f57(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_5 */
extern void T549f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_4 */
extern void T549f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_3 */
extern void T549f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_2 */
extern void T549f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_1 */
extern void T549f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template */
extern T0* T549f55(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_20 */
extern void T549f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_19 */
extern void T549f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_18 */
extern void T549f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_17 */
extern void T549f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_16 */
extern void T549f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_15 */
extern void T549f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_14 */
extern void T549f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_13 */
extern void T549f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_12 */
extern void T549f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_11 */
extern void T549f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_10 */
extern void T549f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_9 */
extern void T549f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_8 */
extern void T549f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_7 */
extern void T549f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_6 */
extern void T549f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_5 */
extern void T549f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_4 */
extern void T549f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_3 */
extern void T549f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_2 */
extern void T549f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_1 */
extern void T549f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template */
extern T0* T549f97(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_20 */
extern void T549f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_19 */
extern void T549f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_18 */
extern void T549f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_17 */
extern void T549f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_16 */
extern void T549f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_15 */
extern void T549f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_14 */
extern void T549f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_13 */
extern void T549f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_12 */
extern void T549f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_11 */
extern void T549f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_10 */
extern void T549f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_9 */
extern void T549f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_8 */
extern void T549f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_7 */
extern void T549f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_6 */
extern void T549f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_5 */
extern void T549f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_4 */
extern void T549f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_3 */
extern void T549f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_2 */
extern void T549f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_1 */
extern void T549f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.special_integer_ */
extern T0* T549f56(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.empty_buffer */
extern T0* T549f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.make */
extern T0* T1314c8(GE_context* ac);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_external */
extern void T549f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.new_external_entity */
extern T0* T174f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_external */
extern T0* T546c213(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_scanner */
extern void T546f214(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T546f216(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T546f216p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T546f216p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T546f216p2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_with_buffer */
extern void T546f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_load_input_buffer */
extern void T546f218(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_set_content */
extern void T546f220(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_initialize */
extern void T546f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_initialize */
extern void T546f217p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_build_tables */
extern void T546f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template */
extern T0* T546f52(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_fixed_array */
extern T0* T546f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_5 */
extern void T546f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_array_subcopy */
extern void T546f278(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_ENTITY_DEF.integer_array_ */
extern T0* T546f58(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_4 */
extern void T546f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_3 */
extern void T546f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_2 */
extern void T546f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_1 */
extern void T546f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_meta_template */
extern T0* T546f50(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template */
extern T0* T546f71(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template_2 */
extern void T546f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template_1 */
extern void T546f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template */
extern T0* T546f66(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_5 */
extern void T546f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_4 */
extern void T546f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_3 */
extern void T546f223(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_2 */
extern void T546f222(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_1 */
extern void T546f221(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template */
extern T0* T546f55(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_5 */
extern void T546f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_4 */
extern void T546f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_3 */
extern void T546f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_2 */
extern void T546f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_1 */
extern void T546f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template */
extern T0* T546f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_20 */
extern void T546f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_19 */
extern void T546f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_18 */
extern void T546f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_17 */
extern void T546f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_16 */
extern void T546f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_15 */
extern void T546f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_14 */
extern void T546f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_13 */
extern void T546f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_12 */
extern void T546f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_11 */
extern void T546f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_10 */
extern void T546f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_9 */
extern void T546f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_8 */
extern void T546f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_7 */
extern void T546f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_6 */
extern void T546f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_5 */
extern void T546f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_4 */
extern void T546f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_3 */
extern void T546f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_2 */
extern void T546f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_1 */
extern void T546f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template */
extern T0* T546f94(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_20 */
extern void T546f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_19 */
extern void T546f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_18 */
extern void T546f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_17 */
extern void T546f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_16 */
extern void T546f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_15 */
extern void T546f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_14 */
extern void T546f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_13 */
extern void T546f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_12 */
extern void T546f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_11 */
extern void T546f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_10 */
extern void T546f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_9 */
extern void T546f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_8 */
extern void T546f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_7 */
extern void T546f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_6 */
extern void T546f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_5 */
extern void T546f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_4 */
extern void T546f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_3 */
extern void T546f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_2 */
extern void T546f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_1 */
extern void T546f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.special_integer_ */
extern T0* T546f70(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.empty_buffer */
extern T0* T546f60(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_entity_declaration */
extern void T174f254(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3, T0* a4, T0* a5);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_entity_declaration */
extern void T516f7(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3, T0* a4, T0* a5);
/* [detachable] XM_EIFFEL_PARSER.when_entity_declared */
extern void T174f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.new_literal_entity */
extern T0* T174f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_literal */
extern T0* T546c212(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_default_value */
extern void T535f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_fixed */
extern void T535f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_implied */
extern void T535f24(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_required */
extern void T535f23(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].aliased_resized_area */
extern T0* T537f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].aliased_resized_area */
extern T0* T539f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].force_last */
extern void T538f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].put_right */
extern void T1332f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].attach_left */
extern void T1332f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].make */
extern T0* T1332c4(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_string_bilinked_list */
extern T0* T174f163(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].set_equality_tester */
extern void T538f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.string_equality_tester */
extern T0* T174f177(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].make */
extern T0* T538c10(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T538f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].new_cursor */
extern T0* T538f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T1331c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].force */
extern void T537f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].fill_with */
extern void T539f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].extend */
extern void T539f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_enumeration_list */
extern void T535f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_enumeration */
extern void T535f27(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.default_enumeration_list */
extern T0* T535f11(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_notation */
extern void T535f21(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_token */
extern void T535f20(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_entity */
extern void T535f19(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_list_type */
extern void T535f18(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_id_ref */
extern void T535f17(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_id */
extern void T535f16(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_data */
extern void T535f15(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_attribute_content */
extern T0* T174f162(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].aliased_resized_area */
extern T0* T534f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].aliased_resized_area */
extern T0* T536f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.make */
extern T0* T535c12(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].force */
extern void T534f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].fill_with */
extern void T536f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].extend */
extern void T536f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.copy_default */
extern void T535f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_implied */
extern T1 T535f10(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_required */
extern T1 T535f9(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.has_default_value */
extern T1 T535f8(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_fixed */
extern T1 T535f7(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_name */
extern void T535f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].force */
extern void T531f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].fill_with */
extern void T533f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].extend */
extern void T533f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].aliased_resized_area */
extern T0* T531f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].aliased_resized_area */
extern T0* T533f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].force_last */
extern void T532f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].put_right */
extern void T1330f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].attach_left */
extern void T1330f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T1330c4(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_attribute_content_list */
extern T0* T174f160(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T532c7(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_internal_cursor */
extern void T532f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].new_cursor */
extern T0* T532f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T1329c8(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_attribute_declarations */
extern void T174f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].forth */
extern void T1329f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_forth */
extern void T532f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].add_traversing_cursor */
extern void T532f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_next_cursor */
extern void T1329f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].remove_traversing_cursor */
extern void T532f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].set */
extern void T1329f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] XM_EIFFEL_PARSER.on_attribute_declaration */
extern void T174f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_attribute_declaration */
extern void T516f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].item */
extern T0* T1329f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].start */
extern void T1329f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_start */
extern void T532f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_off */
extern T1 T532f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].off */
extern T1 T1329f7(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_zero_or_more */
extern void T529f11(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_mixed */
extern T0* T529c10(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_mixed */
extern void T529f21(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_choice */
extern void T529f19(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_list */
extern void T529f16(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_default */
extern void T529f22(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_one */
extern void T529f12(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_sequence */
extern void T529f20(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1328c6(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].set_internal_cursor */
extern void T1328f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].new_cursor */
extern T0* T1328f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1667c3(GE_context* ac, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_sequence */
extern T0* T529c9(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].force_last */
extern void T1328f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].put_right */
extern void T1668f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].attach_left */
extern void T1668f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1668c4(GE_context* ac, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_choice */
extern T0* T529c8(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].force_first */
extern void T1328f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.element_name */
extern T0* T174f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_name */
extern T0* T529c15(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.set_element_repetition */
extern void T174f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_zero_or_one */
extern void T529f14(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_one_or_more */
extern void T529f13(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_any */
extern T0* T529c7(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_any */
extern void T529f18(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].force */
extern void T528f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].fill_with */
extern void T530f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].extend */
extern void T530f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].aliased_resized_area */
extern T0* T528f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].aliased_resized_area */
extern T0* T530f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_empty */
extern T0* T529c6(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_empty */
extern void T529f17(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_element_declaration */
extern void T174f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_element_declaration */
extern void T516f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_attribute */
extern void T174f249(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].has */
extern T1 T523f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].search_position */
extern void T523f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].hash_position */
extern T6 T523f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.hash_code */
extern T6 T520f16(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.item */
extern T0* T520f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].item */
extern T0* T538f8(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME].test */
extern T1 T1326f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_storage_item */
extern T0* T523f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_item */
extern T0* T523f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_item */
extern T6 T523f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_equality_tester */
extern T0* T523f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_item */
extern T6 T523f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].force */
extern void T522f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].fill_with */
extern void T524f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].extend */
extern void T524f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].aliased_resized_area */
extern T0* T522f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].aliased_resized_area */
extern T0* T524f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].force_new */
extern void T523f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_put */
extern void T523f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_put */
extern void T523f38(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_put */
extern void T523f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].resize */
extern void T523f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_resize */
extern void T523f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].special_integer_ */
extern T0* T523f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_storage_resize */
extern void T523f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_resize */
extern void T523f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_resize */
extern void T523f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_modulus */
extern T6 T523f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_capacity */
extern T6 T523f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].unset_found_item */
extern void T523f35(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_name_set */
extern T0* T174f148(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_equal */
extern T0* T523c31(GE_context* ac, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME].default_create */
extern T0* T1326c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make */
extern void T523f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].set_internal_cursor */
extern void T523f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_cursor */
extern T0* T523f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME].make */
extern T0* T1327c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_slots */
extern void T523f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_clashes */
extern void T523f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_key_storage */
extern void T523f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_item_storage */
extern void T523f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].make */
extern T0* T519f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].make_empty */
extern T0* T521c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].default_create */
extern T0* T519c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.on_start_tag */
extern void T174f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_EIFFEL_PARSER.on_end_tag */
extern void T174f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_EIFFEL_PARSER_NAME.local_part */
extern T0* T520f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].forth */
extern void T1331f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_forth */
extern void T538f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T538f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T1331f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T538f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].set */
extern void T1331f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].item */
extern T0* T1331f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].start */
extern void T1331f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_start */
extern void T538f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_off */
extern T1 T538f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T1331f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.string_ */
extern T0* T520f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.last */
extern T0* T520f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].last */
extern T0* T538f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.ns_prefix */
extern T0* T520f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.is_namespace_name */
extern T1 T520f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_no_namespaces */
extern void T520f18(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_no_namespaces */
extern T0* T520c18(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER_NAME.new_string_bilinked_list */
extern T0* T520f15(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.string_equality_tester */
extern T0* T520f14(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_start_tag_finish */
extern void T174f246(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.is_equal */
extern T1 T520f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.same_string */
extern T1 T520f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.when_external_dtd */
extern void T174f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_scanner */
extern T0* T547c205(GE_context* ac);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_start_condition */
extern void T547f212(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_scanner */
extern void T547f205p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T547f214(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T547f214p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T547f214p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_with_buffer */
extern void T547f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_load_input_buffer */
extern void T547f225(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_set_content */
extern void T547f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_initialize */
extern void T547f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_initialize */
extern void T547f224p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_build_tables */
extern void T547f229(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template */
extern T0* T547f199(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_fixed_array */
extern T0* T547f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_5 */
extern void T547f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_array_subcopy */
extern void T547f287(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_SCANNER_DTD.integer_array_ */
extern T0* T547f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_4 */
extern void T547f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_3 */
extern void T547f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_2 */
extern void T547f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_1 */
extern void T547f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_meta_template */
extern T0* T547f198(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template */
extern T0* T547f197(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template_2 */
extern void T547f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template_1 */
extern void T547f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template */
extern T0* T547f196(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_5 */
extern void T547f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_4 */
extern void T547f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_3 */
extern void T547f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_2 */
extern void T547f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_1 */
extern void T547f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template */
extern T0* T547f195(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_5 */
extern void T547f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_4 */
extern void T547f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_3 */
extern void T547f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_2 */
extern void T547f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_1 */
extern void T547f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template */
extern T0* T547f194(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_20 */
extern void T547f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_19 */
extern void T547f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_18 */
extern void T547f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_17 */
extern void T547f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_16 */
extern void T547f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_15 */
extern void T547f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_14 */
extern void T547f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_13 */
extern void T547f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_12 */
extern void T547f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_11 */
extern void T547f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_10 */
extern void T547f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_9 */
extern void T547f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_8 */
extern void T547f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_7 */
extern void T547f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_6 */
extern void T547f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_5 */
extern void T547f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_4 */
extern void T547f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_3 */
extern void T547f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_2 */
extern void T547f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_1 */
extern void T547f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template */
extern T0* T547f193(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_20 */
extern void T547f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_19 */
extern void T547f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_18 */
extern void T547f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_17 */
extern void T547f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_16 */
extern void T547f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_15 */
extern void T547f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_14 */
extern void T547f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_13 */
extern void T547f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_12 */
extern void T547f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_11 */
extern void T547f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_10 */
extern void T547f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_9 */
extern void T547f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_8 */
extern void T547f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_7 */
extern void T547f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_6 */
extern void T547f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_5 */
extern void T547f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_4 */
extern void T547f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_3 */
extern void T547f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_2 */
extern void T547f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_1 */
extern void T547f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.special_integer_ */
extern T0* T547f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.empty_buffer */
extern T0* T547f52(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].force */
extern void T510f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.null_resolver */
extern T0* T174f112(GE_context* ac, T0* C);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.last_error */
extern T0* T514f1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.resolve_external_id */
extern void T174f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve */
extern void T514f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve_public */
extern void T514f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].force */
extern void T525f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].fill_with */
extern void T527f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].extend */
extern void T527f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_end */
extern void T174f244(GE_context* ac, T0* C);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_end */
extern void T516f5(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_doctype */
extern void T174f243(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_doctype */
extern void T516f4(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_EIFFEL_DECLARATION.set_encoding */
extern void T542f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_DECLARATION.set_stand_alone */
extern void T542f8(GE_context* ac, T0* C, T1 a1);
/* [detachable] XM_EIFFEL_DECLARATION.set_version */
extern void T542f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].force */
extern void T541f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].fill_with */
extern void T543f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].extend */
extern void T543f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].aliased_resized_area */
extern T0* T541f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].aliased_resized_area */
extern T0* T543f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_DECLARATION.make */
extern T0* T542c7(GE_context* ac);
/* [detachable] XM_EIFFEL_DECLARATION.process */
extern void T542f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_xml_declaration */
extern void T174f287(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_EIFFEL_PARSER.apply_encoding */
extern void T174f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_content */
extern void T174f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_processing_instruction */
extern void T174f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_processing_instruction */
extern void T516f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_processing_instruction */
extern void T174f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_comment */
extern void T174f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_comment */
extern void T516f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_comment */
extern void T174f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.entity_referenced_in_entity_value */
extern T0* T174f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.defined_entity_referenced */
extern T0* T174f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.external_entity_to_string */
extern T0* T174f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.remove */
extern void T17f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve_finish */
extern void T514f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_string */
extern void T876f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.string_ */
extern T0* T876f17(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.platform */
extern T0* T753f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.integer_ */
extern T0* T174f203(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.make_from_stream */
extern T0* T876c29(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].make */
extern T0* T1573c5(GE_context* ac);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.last_stream */
extern T0* T514f2(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item */
extern T0* T512f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_item */
extern T0* T512f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.string_ */
extern T0* T174f137(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.force_error */
extern void T174f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.report_error */
extern void T174f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_error */
extern void T174f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.last_error_extended_description */
extern T0* T174f173(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.safe_error_component */
extern T0* T174f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_safe_error_character */
extern T1 T174f201(GE_context* ac, T0* C, T2 a1);
/* [detachable] XM_EIFFEL_PARSER.setup_error_state */
extern void T174f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.onstring_utf8 */
extern T0* T174f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.maximum_item_code */
extern T6 T174f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_latin1 */
extern T1 T174f185(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_unicode_string_from_utf8 */
extern T0* T174f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.make_from_utf8 */
extern T0* T743c67(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_STRING.append_utf8 */
extern void T743f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.valid_utf8 */
extern T1 T416f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.valid_utf8_substring */
extern T1 T416f16(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_next_byte */
extern T1 T416f25(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_second_byte */
extern T1 T416f20(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_first_byte */
extern T1 T416f17(GE_context* ac, T0* C, T2 a1);
/* [detachable] XM_EIFFEL_PARSER.utf8 */
extern T0* T174f184(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_ascii */
extern T1 T174f182(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.onstring_ascii */
extern T0* T174f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_unicode */
extern T1 T174f180(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string */
extern T0* T174f135(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string_string */
extern T0* T174f190(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string_uc */
extern T0* T174f189(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_unicode_string_empty */
extern T0* T174f197(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.make_empty */
extern T0* T743c72(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].force */
extern void T519f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].fill_with */
extern void T521f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].extend */
extern void T521f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].aliased_resized_area */
extern T0* T519f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].aliased_resized_area */
extern T0* T521f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.namespace_force_last */
extern void T174f235(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER_NAME.force_last */
extern void T520f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.can_force_last */
extern T1 T520f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_namespace_name */
extern T0* T174f134(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_namespaces */
extern T0* T520c20(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.yy_push_last_value */
extern void T174f225(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.accept */
extern void T174f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_do_error_action */
extern void T174f222(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.report_eof_expected_error */
extern void T174f233(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.read_token */
extern void T174f221(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.process_attribute_entity */
extern void T174f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.process_entity_scanner */
extern void T174f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_start_condition */
extern void T546f280(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_start_condition */
extern void T549f284(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.apply_input_buffer */
extern void T546f279(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.make */
extern T0* T1427c9(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.to_utf8 */
extern T0* T416f49(GE_context* ac, T0* C, T0* a1);
extern T1 T416f49ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_ROUTINES.append_code_to_utf8 */
extern void T416f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] UC_UTF8_ROUTINES.append_natural_32_code_to_utf8 */
extern void T416f54(GE_context* ac, T0* C, T0* a1, T10 a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.utf8 */
extern T0* T546f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.fatal_error */
extern void T546f287(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.apply_input_buffer */
extern void T549f283(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.utf8 */
extern T0* T549f71(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.fatal_error */
extern void T549f291(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_literal */
extern T1 T546f64(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_external */
extern T1 T546f67(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_literal */
extern T1 T549f66(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_external */
extern T1 T549f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.process_entity */
extern void T174f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.process_pe_entity */
extern void T174f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.special_integer_ */
extern T0* T174f97(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_init_value_stacks */
extern void T174f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_clear_all */
extern void T174f229(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.clear_all */
extern void T174f257(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.clear_stacks */
extern void T174f263(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_clear_value_stacks */
extern void T174f264(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].keep_head */
extern void T543f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].set_count */
extern void T543f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].keep_head */
extern void T539f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].set_count */
extern void T539f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].keep_head */
extern void T536f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_count */
extern void T536f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].keep_head */
extern void T533f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].set_count */
extern void T533f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].keep_head */
extern void T530f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].set_count */
extern void T530f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].keep_head */
extern void T527f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].set_count */
extern void T527f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].keep_head */
extern void T524f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].set_count */
extern void T524f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].keep_head */
extern void T521f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].set_count */
extern void T521f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.abort */
extern void T174f223(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_start */
extern void T174f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.reset_error_state */
extern void T174f216(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.reset */
extern void T174f212(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].wipe_out */
extern void T512f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_wipe_out */
extern void T512f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_wipe_out */
extern void T512f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_wipe_out */
extern void T512f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_wipe_out */
extern void T512f44(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].keep_head */
extern void T1319f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].set_count */
extern void T1319f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].move_all_cursors_after */
extern void T512f42(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_next_cursor */
extern void T1322f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_after */
extern void T1322f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make_scanner */
extern void T174f210(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.make_scanner */
extern T0* T506c203(GE_context* ac);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T506f211(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T506f211p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T506f211p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.make_with_buffer */
extern void T506f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_load_input_buffer */
extern void T506f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_set_content */
extern void T506f221(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_initialize */
extern void T506f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_initialize */
extern void T506f219p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_build_tables */
extern void T506f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template */
extern T0* T506f197(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_fixed_array */
extern T0* T506f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_5 */
extern void T506f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_array_subcopy */
extern void T506f285(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_SCANNER.integer_array_ */
extern T0* T506f202(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_4 */
extern void T506f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_3 */
extern void T506f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_2 */
extern void T506f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_1 */
extern void T506f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_meta_template */
extern T0* T506f196(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template */
extern T0* T506f195(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template_2 */
extern void T506f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template_1 */
extern void T506f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template */
extern T0* T506f194(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_5 */
extern void T506f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_4 */
extern void T506f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_3 */
extern void T506f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_2 */
extern void T506f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_1 */
extern void T506f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template */
extern T0* T506f193(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_5 */
extern void T506f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_4 */
extern void T506f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_3 */
extern void T506f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_2 */
extern void T506f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_1 */
extern void T506f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template */
extern T0* T506f192(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_20 */
extern void T506f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_19 */
extern void T506f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_18 */
extern void T506f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_17 */
extern void T506f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_16 */
extern void T506f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_15 */
extern void T506f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_14 */
extern void T506f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_13 */
extern void T506f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_12 */
extern void T506f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_11 */
extern void T506f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_10 */
extern void T506f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_9 */
extern void T506f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_8 */
extern void T506f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_7 */
extern void T506f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_6 */
extern void T506f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_5 */
extern void T506f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_4 */
extern void T506f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_3 */
extern void T506f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_2 */
extern void T506f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_1 */
extern void T506f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template */
extern T0* T506f191(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_20 */
extern void T506f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_19 */
extern void T506f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_18 */
extern void T506f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_17 */
extern void T506f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_16 */
extern void T506f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_15 */
extern void T506f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_14 */
extern void T506f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_13 */
extern void T506f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_12 */
extern void T506f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_11 */
extern void T506f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_10 */
extern void T506f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_9 */
extern void T506f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_8 */
extern void T506f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_7 */
extern void T506f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_6 */
extern void T506f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_5 */
extern void T506f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_4 */
extern void T506f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_3 */
extern void T506f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_2 */
extern void T506f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_1 */
extern void T506f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.special_integer_ */
extern T0* T506f54(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.empty_buffer */
extern T0* T506f45(GE_context* ac, T0* C);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM].make */
extern T0* T159c2(GE_context* ac, T0* a1);
/* [detachable] GEC.target_name */
extern T0* T25f33(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.parameter */
extern T0* T41f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_variables */
extern void T53f209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_capabilities */
extern void T53f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_settings */
extern void T53f207(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.set_primary_value */
extern void T55f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_SETTINGS.make */
extern T0* T55c7(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_SETTINGS.make */
extern T0* T55f7ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_SETTINGS.string_ */
extern T0* T55f3(GE_context* ac, T0* C);
/* detachable ET_ECF_SETTINGS.case_insensitive_string_equality_tester */
extern T0* T55f6(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.make */
extern T0* T53c206(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.make_with_factory */
extern void T53f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.enable_position_table */
extern void T178f10(GE_context* ac, T0* C, T0* a1);
/* detachable XM_POSITION_TABLE.make */
extern T0* T165c4(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T874c7(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_internal_cursor */
extern void T874f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.set_callbacks */
extern void T174f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_TREE_CALLBACKS_PIPE.make */
extern T0* T175c10(GE_context* ac);
/* [detachable] XM_TREE_CALLBACKS_PIPE.callbacks_pipe */
extern T0* T175f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER].item */
extern T0* T551f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.make_null */
extern T0* T178c9(GE_context* ac);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.make_next */
extern void T178f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.initialize */
extern void T178f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_equal */
extern T0* T553c31(GE_context* ac, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE].default_create */
extern T0* T1338c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make */
extern void T553f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].set_internal_cursor */
extern void T553f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_cursor */
extern T0* T553f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE].make */
extern T0* T1339c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].unset_found_item */
extern void T553f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_slots */
extern void T553f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_modulus */
extern T6 T553f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_clashes */
extern void T553f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_key_storage */
extern void T553f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_item_storage */
extern void T553f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].make */
extern T0* T1340f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].make_empty */
extern T0* T1341c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].default_create */
extern T0* T1340c3(GE_context* ac);
/* [detachable] XM_DOCUMENT.make */
extern T0* T188c10(GE_context* ac);
/* [detachable] XM_DOCUMENT.make_with_root_named */
extern void T188f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DOCUMENT.force_last */
extern void T188f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].force_last */
extern void T569f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].force_last */
extern void T569f12p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].put_right */
extern void T1351f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].make */
extern T0* T1351c3(GE_context* ac, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].before_addition */
extern void T569f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.before_addition */
extern void T188f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.equality_delete */
extern void T164f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].remove */
extern void T440f14(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_at_cursor */
extern void T869f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_all_cursors */
extern void T869f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_current_cell */
extern void T440f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].put */
extern void T1698f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].put_right */
extern void T1698f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_last_cell */
extern void T869f23(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].forget_right */
extern void T1698f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_last */
extern void T869f22(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_last_cursors_after */
extern void T869f27(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].wipe_out */
extern void T869f25(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_all_cursors_after */
extern void T869f28(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].is_last */
extern T1 T440f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_is_last */
extern T1 T869f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_first */
extern void T869f21(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_first_cell */
extern void T869f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].is_first */
extern T1 T440f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_is_first */
extern T1 T869f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.equality_delete */
extern void T188f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].forth */
extern void T880f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_forth */
extern void T569f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].add_traversing_cursor */
extern void T569f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_next_cursor */
extern void T880f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_traversing_cursor */
extern void T569f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set */
extern void T880f15(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].remove */
extern void T880f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_at_cursor */
extern void T569f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_all_cursors */
extern void T569f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_current_cell */
extern void T880f17(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].put */
extern void T1351f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_last_cell */
extern void T569f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].forget_right */
extern void T1351f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_last */
extern void T569f20(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_last_cursors_after */
extern void T569f26(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_after */
extern void T880f18(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].wipe_out */
extern void T569f24(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_all_cursors_after */
extern void T569f27(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].is_last */
extern T1 T880f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_is_last */
extern T1 T569f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_first */
extern void T569f19(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_first_cell */
extern void T569f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].is_first */
extern T1 T880f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_is_first */
extern T1 T569f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].item */
extern T0* T880f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].start */
extern void T880f11(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_start */
extern void T569f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_off */
extern T1 T569f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].off */
extern T1 T880f9(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.new_cursor */
extern T0* T188f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].new_cursor */
extern T0* T569f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].make */
extern T0* T880c10(GE_context* ac, T0* a1);
/* [detachable] XM_DOCUMENT.last */
extern T0* T188f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].last */
extern T0* T569f7(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.is_empty */
extern T1 T188f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].is_empty */
extern T1 T569f6(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.make */
extern T0* T164c16(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].make */
extern T0* T869c11(GE_context* ac);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_internal_cursor */
extern void T869f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].make */
extern T0* T569c11(GE_context* ac);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_internal_cursor */
extern void T569f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.default_ns */
extern T0* T188f3(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE.make_default */
extern T0* T439c7(GE_context* ac);
/* [detachable] XM_NAMESPACE.make */
extern void T439f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE.make */
extern T0* T439c8(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.null_callbacks */
extern T0* T178f7(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.make */
extern T0* T511c1(GE_context* ac);
/* [detachable] XM_TREE_CALLBACKS_PIPE.new_stop_on_error */
extern T0* T175f7(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.make_null */
extern T0* T191c5(GE_context* ac);
/* [detachable] XM_STOP_ON_ERROR_FILTER.make_next */
extern void T191f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.initialize */
extern void T191f7(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.null_callbacks */
extern T0* T191f4(GE_context* ac, T0* C);
/* [detachable] XM_TREE_CALLBACKS_PIPE.new_namespace_resolver */
extern T0* T175f6(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.make_null */
extern T0* T550c27(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER.make_next */
extern void T550f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.initialize */
extern void T550f39(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_make */
extern void T550f43(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.new_string_queue */
extern T0* T550f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].make */
extern T0* T1337c5(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER.new_detachable_string_queue */
extern T0* T550f20(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].make */
extern T0* T1336c6(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.make */
extern T0* T1333c10(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1670c8(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_internal_cursor */
extern void T1670f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].new_cursor */
extern T0* T1670f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1671c8(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.null_callbacks */
extern T0* T550f13(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.set_string_mode_mixed */
extern void T174f206(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make */
extern T0* T174c205(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.new_entities_table */
extern T0* T174f85(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_key_equality_tester */
extern void T512f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_map_default */
extern T0* T512c36(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_map */
extern void T512f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_with_equality_testers */
extern void T512f53(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_sparse_container */
extern void T512f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_internal_cursor */
extern void T512f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_cursor */
extern T0* T512f34(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make */
extern T0* T1322c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_slots */
extern void T512f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_clashes */
extern void T512f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_key_storage */
extern void T512f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_item_storage */
extern void T512f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].make */
extern T0* T1323f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].make_empty */
extern T0* T1319c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].default_create */
extern T0* T1323c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].default_capacity */
extern T6 T512f30(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make_parser */
extern void T174f211(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_build_parser_tables */
extern void T174f215(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template */
extern T0* T174f121(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yyfixed_array */
extern T0* T174f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_4 */
extern void T174f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyarray_subcopy */
extern void T174f286(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_PARSER.integer_array_ */
extern T0* T174f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_3 */
extern void T174f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_2 */
extern void T174f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_1 */
extern void T174f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template */
extern T0* T174f119(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_4 */
extern void T174f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_3 */
extern void T174f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_2 */
extern void T174f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_1 */
extern void T174f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypgoto_template */
extern T0* T174f117(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yypact_template */
extern T0* T174f115(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_3 */
extern void T174f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_2 */
extern void T174f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_1 */
extern void T174f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefgoto_template */
extern T0* T174f113(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template */
extern T0* T174f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_3 */
extern void T174f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_2 */
extern void T174f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_1 */
extern void T174f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes2_template */
extern T0* T174f98(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template */
extern T0* T174f93(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_3 */
extern void T174f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_2 */
extern void T174f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_1 */
extern void T174f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template */
extern T0* T174f89(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template_2 */
extern void T174f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template_1 */
extern void T174f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template */
extern T0* T174f87(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template_2 */
extern void T174f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template_1 */
extern void T174f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yy_create_value_stacks */
extern void T174f214(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].make */
extern T0* T541f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].make_empty */
extern T0* T543c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].default_create */
extern T0* T541c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].make */
extern T0* T537f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].make_empty */
extern T0* T539c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].default_create */
extern T0* T537c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T534f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].make_empty */
extern T0* T536c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].default_create */
extern T0* T534c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].make */
extern T0* T531f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].make_empty */
extern T0* T533c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].default_create */
extern T0* T531c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T528f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].make_empty */
extern T0* T530c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].default_create */
extern T0* T528c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].make */
extern T0* T525f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].make_empty */
extern T0* T527c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].default_create */
extern T0* T525c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].make */
extern T0* T522f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].make_empty */
extern T0* T524c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].default_create */
extern T0* T522c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.initialize */
extern void T174f209(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.null_dtd_callbacks */
extern T0* T174f86(GE_context* ac, T0* C);
/* [detachable] XM_DTD_CALLBACKS_NULL.make */
extern T0* T516c1(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.null_callbacks */
extern T0* T174f84(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_key_equality_tester */
extern void T172f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_map */
extern T0* T172c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_with_equality_testers */
extern void T172f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_sparse_container */
extern void T172f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_internal_cursor */
extern void T172f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_cursor */
extern T0* T172f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make */
extern T0* T504c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_slots */
extern void T172f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_clashes */
extern void T172f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_key_storage */
extern void T172f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_item_storage */
extern void T172f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].make */
extern T0* T505f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].make_empty */
extern T0* T501c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].default_create */
extern T0* T505c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_key_equality_tester */
extern void T170f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_map */
extern T0* T170c40(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_with_equality_testers */
extern void T170f47(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_sparse_container */
extern void T170f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_internal_cursor */
extern void T170f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_cursor */
extern T0* T170f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make */
extern T0* T496c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_slots */
extern void T170f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_clashes */
extern void T170f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_key_storage */
extern void T170f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_item_storage */
extern void T170f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].make */
extern T0* T500f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].make_empty */
extern T0* T497c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].default_create */
extern T0* T500c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.make_with_factory */
extern void T53f211p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ise_latest */
extern T0* T53f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_AST_FACTORY.make */
extern T0* T158c53(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.make_standard */
extern T0* T54c10(GE_context* ac);
/* [detachable] KL_STANDARD_FILES.output */
extern T0* T135f2(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.make */
extern T0* T137c5(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.std */
extern T0* T54f7(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.make_null */
extern T0* T54c9(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.null_output_stream */
extern T0* T54f6(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_open_read */
extern T1 T32f14(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_open_read */
extern T1 T32f28(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.open_read */
extern void T32f61(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_open_read */
extern void T32f66(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.default_pointer */
extern T14 T32f21(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.open_read */
extern void T32f66p1(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_open */
extern T14 T32f24(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_readable */
extern T1 T32f16(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.is_readable */
extern T1 T132f8(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.file_eaccess */
extern T1 T132f6(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern void T32f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern T0* T32c60(GE_context* ac, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern void T32f60p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_make */
extern void T32f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make_with_name */
extern void T32f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.parse_arguments */
extern void T25f60(GE_context* ac, T0* C);
/* [detachable] GEC.set_new_instance_types */
extern void T25f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_SET [[attached] STRING_8].force */
extern void T69f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.adjust */
extern void T17f60(GE_context* ac, T0* C);
/* [detachable] STRING_8.right_adjust */
extern void T17f65(GE_context* ac, T0* C);
/* CHARACTER_8.is_space */
extern T1 T2f4(GE_context* ac, T2* C);
/* [detachable] STRING_8.left_adjust */
extern void T17f64(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].make_equal */
extern T0* T69c38(GE_context* ac, T6 a1);
/* [detachable] GEC.set_split_size */
extern void T25f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] AP_ERROR.make_invalid_parameter_error */
extern T0* T88c21(GE_context* ac, T0* a1, T0* a2);
/* [detachable] GEC.set_override_variables */
extern void T25f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].forth */
extern void T147f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_forth */
extern void T146f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_position */
extern void T147f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].add_traversing_cursor */
extern void T146f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_next_cursor */
extern void T147f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].remove_traversing_cursor */
extern void T146f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.index_of */
extern T6 T17f33(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].item */
extern T0* T147f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_item */
extern T0* T146f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].after */
extern T1 T147f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_iterator */
extern T0* T146f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].start */
extern void T147f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_start */
extern void T146f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_after */
extern void T147f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].off */
extern T1 T147f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].before */
extern T1 T147f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_cursor */
extern T0* T146f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].make */
extern T0* T147c9(GE_context* ac, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.set_variable_value */
extern void T87f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.put */
extern void T122f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.eif_putenv */
extern T6 T122f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].force */
extern void T473f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].force */
extern void T1312f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].extend */
extern void T1312f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].position */
extern T6 T473f22(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].force */
extern void T903f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].force */
extern void T127f6(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].deleted_position */
extern T6 T473f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].add_space */
extern void T473f37(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].accommodate */
extern void T473f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_indexes_map */
extern void T473f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_deleted_marks */
extern void T473f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_keys */
extern void T473f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_content */
extern void T473f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].put */
extern void T473f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_conflict */
extern void T473f45(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].found */
extern T1 T473f31(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].occupied */
extern T1 T473f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].empty_duplicate */
extern T0* T473f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].compare_objects */
extern void T473f44(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].make */
extern T0* T473c34(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].make_empty */
extern T0* T1312c4(GE_context* ac, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].soon_full */
extern T1 T473f20(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].not_found */
extern T1 T473f19(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].internal_search */
extern void T473f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].same_keys */
extern T1 T473f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] IMMUTABLE_STRING_32.is_equal */
extern T1 T20f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].hash_code_of */
extern T6 T473f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] IMMUTABLE_STRING_32.hash_code */
extern T6 T20f7(GE_context* ac, T0* C);
/* [detachable] IMMUTABLE_STRING_32.item */
extern T3 T20f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_general */
extern T0* T20c11(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_STRING.to_string_32 */
extern T0* T743f18(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_string_32 */
extern T0* T17f40(GE_context* ac, T0* C);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_32 */
extern void T20f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_8 */
extern void T20f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] IMMUTABLE_STRING_32.make_from_area_and_bounds */
extern void T20f15(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] EXECUTION_ENVIRONMENT.environ */
extern T0* T122f6(GE_context* ac, T0* C);
/* [detachable] GEC.execution_environment */
extern T0* T25f56(GE_context* ac, T0* C);
/* [detachable] GEC.set_override_capabilities */
extern void T25f68(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.set_override_settings */
extern void T25f67(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.set_ise_version */
extern void T25f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].first */
extern T0* T143f19(GE_context* ac, T0* C);
/* [detachable] GEC.report_usage_message */
extern void T25f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.full_usage_instruction */
extern T0* T39f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].forth */
extern void T142f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_forth */
extern void T142f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_position */
extern void T483f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].add_traversing_cursor */
extern void T142f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_next_cursor */
extern void T483f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].remove_traversing_cursor */
extern void T142f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.alternative_usage_instruction */
extern T0* T39f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.wrapped_string */
extern T0* T138f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.canonify_whitespace */
extern void T138f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.is_space */
extern T1 T138f8(GE_context* ac, T0* C, T2 a1);
/* [detachable] ST_WORD_WRAPPER.string_ */
extern T0* T138f4(GE_context* ac, T0* C);
/* [detachable] ST_WORD_WRAPPER.set_new_line_indentation */
extern void T138f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_PARSER.wrapper */
extern T0* T39f12(GE_context* ac, T0* C);
/* [detachable] ST_WORD_WRAPPER.set_maximum_text_width */
extern void T138f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_WORD_WRAPPER.make */
extern T0* T138c9(GE_context* ac);
/* [detachable] AP_PARSER.arguments */
extern T0* T39f27(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.file_system */
extern T0* T39f26(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.unix_file_system */
extern T0* T39f15(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.windows_file_system */
extern T0* T39f14(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.operating_system */
extern T0* T39f13(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.forth */
extern void T40f17(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_forth */
extern void T40f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.add_traversing_cursor */
extern void T40f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].set_next_cursor */
extern void T145f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.remove_traversing_cursor */
extern void T40f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].set */
extern void T145f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.item_for_iteration */
extern T0* T40f6(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_item */
extern T0* T40f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.after */
extern T1 T40f5(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_after */
extern T1 T40f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.start */
extern void T40f16(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_start */
extern void T40f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_off */
extern T1 T40f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].off */
extern T1 T145f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].item_for_iteration */
extern T0* T142f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_item */
extern T0* T142f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].after */
extern T1 T142f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_after */
extern T1 T142f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].after */
extern T1 T483f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].start */
extern void T142f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_start */
extern void T142f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_after */
extern void T483f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].is_empty */
extern T1 T142f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].off */
extern T1 T483f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].before */
extern T1 T483f6(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.usage_instruction */
extern T0* T39f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].forth */
extern void T141f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_forth */
extern void T141f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_position */
extern void T479f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].add_traversing_cursor */
extern void T141f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_next_cursor */
extern void T479f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].remove_traversing_cursor */
extern void T141f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].item_for_iteration */
extern T0* T141f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_item */
extern T0* T141f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].after */
extern T1 T141f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_after */
extern T1 T141f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].after */
extern T1 T479f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].start */
extern void T141f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_start */
extern void T141f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_after */
extern void T479f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].is_empty */
extern T1 T141f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].off */
extern T1 T479f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].before */
extern T1 T479f6(GE_context* ac, T0* C);
/* [detachable] GEC.report_version_number */
extern void T25f64(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.make */
extern T0* T81c8(GE_context* ac, T0* a1);
/* [detachable] AP_PARSER.parse_arguments */
extern void T39f44(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.parse_list */
extern void T39f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.final_error_action */
extern void T39f49(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.exceptions */
extern T0* T39f31(GE_context* ac, T0* C);
/* [detachable] KL_EXCEPTIONS.default_create */
extern T0* T38c1(GE_context* ac);
/* [detachable] AP_ERROR_HANDLER.report_error_message */
extern void T144f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR_HANDLER.report_error_message */
extern void T144f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.check_options_after_parsing */
extern void T39f48(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.make_surplus_option_error */
extern T0* T88c23(GE_context* ac, T0* a1);
/* [detachable] AP_ERROR_HANDLER.report_error */
extern void T144f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR_HANDLER.message */
extern T0* T144f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_missing_option_error */
extern T0* T88c22(GE_context* ac, T0* a1);
/* [detachable] AP_PARSER.parse_argument */
extern void T39f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.parse_short */
extern void T39f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_missing_parameter_error */
extern T0* T88c25(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].off */
extern T1 T143f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_off */
extern T1 T143f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T486f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].before */
extern T1 T486f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].after */
extern T1 T486f5(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.make_unknown_option_error */
extern T0* T88c24(GE_context* ac, T0* a1);
/* CHARACTER_8.out */
extern T0* T2f3(GE_context* ac, T2* C);
/* [detachable] AP_PARSER.parse_long */
extern void T39f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_unnecessary_parameter_error */
extern T0* T88c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] AP_PARSER.string_ */
extern T0* T39f33(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].forth */
extern void T143f24(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_forth */
extern void T143f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T143f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T143f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].item_for_iteration */
extern T0* T143f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_item */
extern T0* T143f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].after */
extern T1 T143f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_after */
extern T1 T143f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].start */
extern void T143f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_start */
extern void T143f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].is_empty */
extern T1 T143f16(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.reset_parser */
extern void T39f46(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.all_options */
extern T0* T39f25(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].has */
extern T1 T141f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].make */
extern T0* T141c17(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].set_internal_cursor */
extern void T141f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].new_cursor */
extern T0* T141f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].make */
extern T0* T479c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].make */
extern T0* T482f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] AP_OPTION].make_empty */
extern T0* T481c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].default_create */
extern T0* T482c3(GE_context* ac);
/* [detachable] AP_ERROR_HANDLER.reset */
extern void T144f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].force */
extern void T143f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].put */
extern void T143f29(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_cursors_right */
extern void T143f33(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_right */
extern void T143f32(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_ARGUMENTS.argument_count */
extern T6 T33f3(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].force_last */
extern void T142f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].force */
extern void T485f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].fill_with */
extern void T484f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].extend */
extern void T484f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].resize */
extern void T142f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].aliased_resized_area */
extern T0* T485f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].aliased_resized_area */
extern T0* T484f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].new_capacity */
extern T6 T142f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].extendible */
extern T1 T142f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.make */
extern T0* T40c11(GE_context* ac, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.old_make */
extern void T40f14(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_internal_cursor */
extern void T40f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.new_cursor */
extern T0* T40f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].make */
extern T0* T145c7(GE_context* ac, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_parameters_description */
extern void T40f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_introduction_option */
extern void T40f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.make */
extern T0* T43c19(GE_context* ac, T2 a1, T0* a2);
/* [detachable] AP_FLAG.set_long_form */
extern void T43f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.set_short_form */
extern void T43f23(GE_context* ac, T0* C, T2 a1);
/* [detachable] AP_FLAG.initialize */
extern void T43f20(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.set_parameter_description */
extern void T47f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.set_description */
extern void T47f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.make_with_long_form */
extern T0* T47c23(GE_context* ac, T0* a1);
/* [detachable] AP_INTEGER_OPTION.set_long_form */
extern void T47f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T47f26(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T47f26p1(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T47f26p0(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.set_parameter_description */
extern void T46f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.set_description */
extern void T46f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.make_with_long_form */
extern T0* T46c26(GE_context* ac, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.set_long_form */
extern void T46f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T46f29(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T46f29p1(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T46f29p0(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.set_parameter_description */
extern void T45f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.extend */
extern void T45f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.update_parameter_description */
extern void T45f32(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].forth */
extern void T148f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].is_last */
extern T1 T148f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_is_last */
extern T1 T148f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].item_for_iteration */
extern T0* T148f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_item */
extern T0* T148f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].after */
extern T1 T148f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_after */
extern T1 T148f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].start */
extern void T148f20(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.set_description */
extern void T45f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.make_with_long_form */
extern T0* T45c25(GE_context* ac, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.set_long_form */
extern void T45f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T45f29(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].set_equality_tester */
extern void T148f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.string_equality_tester */
extern T0* T45f14(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T45f29p1(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T45f29p0(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.set_default_parameter */
extern void T41f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.out */
extern T0* T44f10(GE_context* ac, T0* C);
/* [detachable] GEC.ise_latest */
extern T0* T25f32(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.set_description */
extern void T43f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.make_with_long_form */
extern T0* T43c17(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].force_last */
extern void T141f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].force */
extern void T482f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_OPTION].fill_with */
extern void T481f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_OPTION].extend */
extern void T481f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].resize */
extern void T141f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].aliased_resized_area */
extern T0* T482f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] AP_OPTION].aliased_resized_area */
extern T0* T481f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].new_capacity */
extern T6 T141f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].extendible */
extern T1 T141f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_STRING_OPTION.set_parameter_description */
extern void T41f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.set_description */
extern void T41f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.make_with_long_form */
extern T0* T41c24(GE_context* ac, T0* a1);
/* [detachable] AP_STRING_OPTION.set_long_form */
extern void T41f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.initialize */
extern void T41f28(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.initialize */
extern void T41f28p1(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.set_parameters_description */
extern void T39f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.set_application_description */
extern void T39f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.make */
extern T0* T39c41(GE_context* ac);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_description */
extern void T140f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.make */
extern T0* T140c19(GE_context* ac, T2 a1, T0* a2);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_long_form */
extern void T140f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_short_form */
extern void T140f22(GE_context* ac, T0* C, T2 a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.initialize */
extern void T140f21(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.make */
extern void T39f41p1(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.make_standard */
extern T0* T144c7(GE_context* ac);
/* [detachable] AP_ERROR_HANDLER.std */
extern T0* T144f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T142c14(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_internal_cursor */
extern void T142f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].new_cursor */
extern T0* T142f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T483c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T485f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make_empty */
extern T0* T484c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].default_create */
extern T0* T485c3(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.make_standard */
extern T0* T35c177(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.make_standard */
extern void T35f177p1(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.std */
extern T0* T35f23(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_ise_library_variable */
extern void T34f7(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.execution_environment */
extern T0* T34f3(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.ise_eiffel_value */
extern T0* T34f2(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.ise_library_value */
extern T0* T34f1(GE_context* ac, T0* C);
/* [detachable] GEC.ise_variables */
extern T0* T25f31(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.default_create */
extern T0* T34c6(GE_context* ac);
/* [detachable] KL_ARGUMENTS.set_program_name */
extern void T33f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.arguments */
extern T0* T25f34(GE_context* ac, T0* C);
/* [detachable] KL_EXCEPTIONS.die */
extern void T38f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_EXCEPTIONS.exceptions_die */
extern void T38f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] GEC.exceptions */
extern T0* T25f48(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.set_exception_data */
extern void T21f16(GE_context* ac, T0* C, T6 a1, T1 a2, T6 a3, T6 a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T6 a11, T1 a12);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_from_data */
extern T0* T21f10(GE_context* ac, T0* C);
extern T1 T21f10ot1(T0* a1, T0** a2);
extern T1 T21f10ot2(T0* a1, T0** a2);
extern T1 T21f10ot3(T0* a1, T0** a2);
extern T1 T21f10ot4(T0* a1, T0** a2);
extern T1 T21f10ot5(T0* a1, T0** a2);
extern T1 T21f10ot6(T0* a1, T0** a2);
extern T1 T21f10ot7(T0* a1, T0** a2);
/* [detachable] COM_FAILURE.set_exception_information */
extern void T102f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.ccom_hresult_to_string */
extern T0* T102f11(GE_context* ac, T0* C, T6 a1);
/* INTEGER_32.to_hex_string */
extern T0* T6f26(GE_context* ac, T6* C);
/* [detachable] COM_FAILURE.set_hresult_code */
extern void T102f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_error_code */
extern void T101f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] IO_FAILURE.set_error_code */
extern void T100f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_signal_code */
extern void T99f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] INVARIANT_VIOLATION.set_is_entry */
extern void T98f9(GE_context* ac, T0* C, T1 a1);
/* [detachable] ROUTINE_FAILURE.set_class_name */
extern void T93f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_routine_name */
extern void T93f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_from_code */
extern T0* T21f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SERIALIZATION_FAILURE.default_create */
extern T0* T121c6(GE_context* ac);
/* [detachable] OLD_VIOLATION.default_create */
extern T0* T95c6(GE_context* ac);
/* [detachable] COM_FAILURE.default_create */
extern T0* T102c12(GE_context* ac);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.default_create */
extern T0* T120c6(GE_context* ac);
/* [detachable] MISMATCH_FAILURE.default_create */
extern T0* T118c6(GE_context* ac);
/* [detachable] OPERATING_SYSTEM_FAILURE.default_create */
extern T0* T101c7(GE_context* ac);
/* [detachable] IO_FAILURE.set_code */
extern void T100f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] IO_FAILURE.default_create */
extern T0* T100c8(GE_context* ac);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.default_create */
extern T0* T117c6(GE_context* ac);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.default_create */
extern T0* T116c6(GE_context* ac);
/* [detachable] EXTERNAL_FAILURE.default_create */
extern T0* T115c6(GE_context* ac);
/* [detachable] CREATE_ON_DEFERRED.default_create */
extern T0* T114c6(GE_context* ac);
/* [detachable] RESUMPTION_FAILURE.default_create */
extern T0* T113c6(GE_context* ac);
/* [detachable] RESCUE_FAILURE.default_create */
extern T0* T112c6(GE_context* ac);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_code */
extern void T103f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.code */
extern T6 T103f7(GE_context* ac, T0* C);
/* [detachable] EIFFEL_RUNTIME_PANIC.default_create */
extern T0* T103c8(GE_context* ac);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.default_create */
extern T0* T99c7(GE_context* ac);
/* [detachable] LOOP_INVARIANT_VIOLATION.default_create */
extern T0* T111c6(GE_context* ac);
/* [detachable] VARIANT_VIOLATION.default_create */
extern T0* T110c6(GE_context* ac);
/* [detachable] BAD_INSPECT_VALUE.default_create */
extern T0* T109c6(GE_context* ac);
/* [detachable] ROUTINE_FAILURE.default_create */
extern T0* T93c8(GE_context* ac);
/* [detachable] CHECK_VIOLATION.default_create */
extern T0* T108c6(GE_context* ac);
/* [detachable] INVARIANT_VIOLATION.default_create */
extern T0* T98c7(GE_context* ac);
/* [detachable] FLOATING_POINT_FAILURE.default_create */
extern T0* T107c6(GE_context* ac);
/* [detachable] POSTCONDITION_VIOLATION.default_create */
extern T0* T106c6(GE_context* ac);
/* [detachable] PRECONDITION_VIOLATION.default_create */
extern T0* T105c6(GE_context* ac);
/* [detachable] NO_MORE_MEMORY.set_code */
extern void T97f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] NO_MORE_MEMORY.code */
extern T6 T97f7(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.no_memory_exception_object_cell */
extern T0* T21f7(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] NO_MORE_MEMORY].put */
extern T0* T29c2(GE_context* ac, T0* a1);
/* [detachable] NO_MORE_MEMORY.default_create */
extern T0* T97c8(GE_context* ac);
/* [detachable] VOID_TARGET.default_create */
extern T0* T91c6(GE_context* ac);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_data */
extern T0* T21f11(GE_context* ac, T0* C);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].put */
extern void T28f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].put */
extern T0* T28c2(GE_context* ac, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_data_cell */
extern T0* T21f6(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.once_raise */
extern void T21f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.last_exception */
extern T0* T21f1(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.init_exception_manager */
extern void T21f14(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] NO_MORE_MEMORY].do_nothing */
extern void T29f3(GE_context* ac, T0* C);
/* [detachable] CELL [detachable EXCEPTION].do_nothing */
extern void T26f3(GE_context* ac, T0* C);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].do_nothing */
extern void T28f3(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.unraisable_exceptions */
extern T0* T21f5(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].force */
extern void T27f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].position */
extern T6 T27f22(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].deleted_position */
extern T6 T27f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].add_space */
extern void T27f39(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].accommodate */
extern void T27f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_indexes_map */
extern void T27f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_deleted_marks */
extern void T27f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_keys */
extern void T27f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_content */
extern void T27f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].put */
extern void T27f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_conflict */
extern void T27f47(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].found */
extern T1 T27f31(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].occupied */
extern T1 T27f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].empty_duplicate */
extern T0* T27f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].compare_objects */
extern void T27f46(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].soon_full */
extern T1 T27f20(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].not_found */
extern T1 T27f19(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].internal_search */
extern void T27f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ISE_EXCEPTION_MANAGER.unignorable_exceptions */
extern T0* T21f4(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].do_nothing */
extern void T27f36(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dispose */
extern void T327f30(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.dispose */
extern void T293f27(GE_context* ac, T0* C);
/* [detachable] MUTEX.dispose */
extern void T133f9(GE_context* ac, T0* C);
/* [detachable] MUTEX.destroy */
extern void T133f13(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_destroy */
extern void T133f16(GE_context* ac, T0* C, T14 a1);
/* [detachable] MANAGED_POINTER.dispose */
extern void T131f9(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.dispose */
extern void T32f59(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_close */
extern void T32f64(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_close */
extern void T32f69(GE_context* ac, T0* C, T14 a1);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2124f5(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.put */
extern void T2088f13(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.put */
extern void T2088f13p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2123f5(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.add */
extern void T2088f12(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.add */
extern void T2088f12p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2120f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIS_INSTRUCTION.set_arguments */
extern void T1947f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.set_arguments */
extern void T1704f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_arguments */
extern void T1490f34(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_arguments */
extern void T1489f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_INSTRUCTION.set_arguments */
extern void T1487f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_EXPRESSION.set_arguments */
extern void T1230f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BRACKET_EXPRESSION.set_arguments */
extern void T1013f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIS_EXPRESSION.set_arguments */
extern void T862f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL.set_arguments */
extern void T321f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_arguments */
extern void T279f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_arguments */
extern void T278f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_arguments */
extern void T275f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.set_parenthesis_call */
extern void T1704f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_INSTRUCTION.make */
extern T0* T1947c8(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_parenthesis_call */
extern void T279f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_EXPRESSION.make */
extern T0* T862c11(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_parenthesis_call */
extern void T278f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_parenthesis_call */
extern void T275f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.reset */
extern void T1497f36(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.reset */
extern void T1496f36(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.reset */
extern void T972f100(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.reset */
extern void T959f55(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.reset */
extern void T951f52(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_options */
extern void T773f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_options */
extern void T209f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_options */
extern void T203f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CHARACTER_DATA.node_set_parent */
extern void T1421f4(GE_context* ac, T0* C, T0* a1);
/* detachable XM_ATTRIBUTE.node_set_parent */
extern void T192f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.add_overloaded_queries */
extern void T1794f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove */
extern void T698f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUERY].keep_head */
extern void T885f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUERY].set_count */
extern void T885f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_left */
extern void T698f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_cursors_left */
extern void T698f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_position */
extern void T1375f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove_last */
extern void T698f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_last_cursors_after */
extern void T698f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_next_cursor */
extern void T1375f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_after */
extern void T1375f10(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.index_of_old_name */
extern T6 T1057f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].item */
extern T0* T698f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.rename_pair */
extern T0* T1057f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.item */
extern T0* T1057f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.index_of_new_name */
extern T6 T1057f8(GE_context* ac, T0* C, T0* a1);
extern T1 T1057f8ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.add_overloaded_queries */
extern void T248f128(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_overloaded_queries */
extern void T73f329(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_QUERY_LIST.add_overloaded_features */
extern void T253f14(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T253f14ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_last */
extern void T698f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].force */
extern void T887f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUERY].fill_with */
extern void T885f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUERY].extend */
extern void T885f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].resize */
extern void T698f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].aliased_resized_area */
extern T0* T887f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_QUERY].aliased_resized_area */
extern T0* T885f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].new_capacity */
extern T6 T698f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].extendible */
extern T1 T698f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.add_overloaded_queries */
extern void T71f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.add_overloaded_procedures */
extern void T1794f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove */
extern void T694f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].keep_head */
extern void T888f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].set_count */
extern void T888f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_left */
extern void T694f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_cursors_left */
extern void T694f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_position */
extern void T1374f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove_last */
extern void T694f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_last_cursors_after */
extern void T694f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_next_cursor */
extern void T1374f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_after */
extern void T1374f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].item */
extern T0* T694f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.add_overloaded_procedures */
extern void T248f129(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_overloaded_procedures */
extern void T73f330(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PROCEDURE_LIST.add_overloaded_features */
extern void T254f14(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T254f14ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_last */
extern void T694f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].force */
extern void T889f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].fill_with */
extern void T888f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].extend */
extern void T888f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].resize */
extern void T694f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].aliased_resized_area */
extern T0* T889f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].aliased_resized_area */
extern T0* T888f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].new_capacity */
extern T6 T694f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].extendible */
extern T1 T694f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.add_overloaded_procedures */
extern void T71f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TYPE_CONSTRAINT_LIST.reset_renames */
extern void T1244f17(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.reset */
extern void T1244f18(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset_renames */
extern void T1478f89(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset_renames */
extern void T1477f93(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset_renames */
extern void T1476f94(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.reset_renames */
extern void T1451f10(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.reset */
extern void T1057f14(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset_renames */
extern void T442f87(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset_renames */
extern void T248f126(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset_renames */
extern void T246f83(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_renames */
extern void T73f327(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset_renames */
extern void T71f136(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.reset */
extern void T1451f11(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset */
extern void T1704f8(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset */
extern void T1704f8p1(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset_arguments */
extern void T1704f9(GE_context* ac, T0* C);
extern T1 T1704f9ot1(T0* a1, T0** a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.reset */
extern void T1924f14(GE_context* ac, T0* C);
extern T1 T1924f14ot1(T0* a1, T0** a2);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.reset */
extern void T274f18(GE_context* ac, T0* C);
extern T1 T274f18ot1(T0* a1, T0** a2);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset */
extern void T1489f12(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.type */
extern T0* T1489f6(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset */
extern void T1489f12p1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset_arguments */
extern void T1489f13(GE_context* ac, T0* C);
extern T1 T1489f13ot1(T0* a1, T0** a2);
/* detachable ET_PRECURSOR_INSTRUCTION.reset */
extern void T1487f13(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_INSTRUCTION.reset_arguments */
extern void T1487f14(GE_context* ac, T0* C);
extern T1 T1487f14ot1(T0* a1, T0** a2);
/* detachable ET_CREATE_INSTRUCTION.reset */
extern void T1485f16(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.reset */
extern void T321f6(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.reset_arguments */
extern void T321f7(GE_context* ac, T0* C);
extern T1 T321f7ot1(T0* a1, T0** a2);
/* detachable ET_CREATE_INSTRUCTION.type */
extern T0* T1485f7(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.reset */
extern void T1484f14(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.reset */
extern void T1482f15(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.until_expression */
extern T0* T1482f9(GE_context* ac, T0* C);
/* detachable ET_VARIANT.reset */
extern void T1253f7(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.reset */
extern void T1184f12(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.reset */
extern void T330f13(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT_ATTEMPT.reset */
extern void T1481f7(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT.reset */
extern void T1480f7(GE_context* ac, T0* C);
extern T1 T1480f7ot1(T0* a1, T0** a2);
/* detachable ET_ASSIGNER_INSTRUCTION.reset */
extern void T1479f17(GE_context* ac, T0* C);
extern T1 T1479f17ot1(T0* a1, T0** a2);
/* detachable ET_ASSIGNER_INSTRUCTION.dummy_name */
extern T0* T1479f6(GE_context* ac, T0* C);
/* detachable ET_INSPECT_INSTRUCTION.reset */
extern void T1162f10(GE_context* ac, T0* C);
/* detachable ET_WHEN_PART_LIST.reset */
extern void T1259f7(GE_context* ac, T0* C);
/* detachable ET_WHEN_PART.reset */
extern void T1256f4(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.reset */
extern void T1031f11(GE_context* ac, T0* C);
/* detachable ET_INSPECT_INSTRUCTION.expression */
extern T0* T1162f6(GE_context* ac, T0* C);
/* detachable ET_IF_INSTRUCTION.reset */
extern void T1143f12(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART_LIST.reset */
extern void T1095f7(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART.reset */
extern void T1092f6(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART.expression */
extern T0* T1092f3(GE_context* ac, T0* C);
/* detachable ET_IF_INSTRUCTION.expression */
extern T0* T1143f7(GE_context* ac, T0* C);
/* detachable ET_DEBUG_INSTRUCTION.reset */
extern void T1083f7(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.reset */
extern void T1019f15(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.reset */
extern void T1019f15p1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.reset */
extern void T989f29(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.reset_unfolded_form */
extern void T989f30(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.reset */
extern void T969f18(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T275f16(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T275f16p1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T275f16p0(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset_arguments */
extern void T275f17(GE_context* ac, T0* C);
extern T1 T275f17ot1(T0* a1, T0** a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.process */
extern void T1704f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.process */
extern void T1489f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_INSTRUCTION.process */
extern void T1487f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.process */
extern void T1485f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BANG_INSTRUCTION.process */
extern void T1484f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.process */
extern void T1482f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNMENT_ATTEMPT.process */
extern void T1481f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNMENT.process */
extern void T1480f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.process */
extern void T1479f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INSPECT_INSTRUCTION.process */
extern void T1162f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_INSTRUCTION.process */
extern void T1143f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEBUG_INSTRUCTION.process */
extern void T1083f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.process */
extern void T1019f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.process */
extern void T989f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.process */
extern void T969f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.process */
extern void T275f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_RANGE.reset */
extern void T1486f8(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_sign */
extern void T1438f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_sign */
extern void T1437f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_sign */
extern void T1436f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_sign */
extern void T1435f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_sign */
extern void T366f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_type */
extern void T1438f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_type */
extern void T1437f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_type */
extern void T1436f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_type */
extern void T1435f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_type */
extern void T366f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_cast_type */
extern void T1438f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_cast_type */
extern void T1437f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_cast_type */
extern void T1436f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_cast_type */
extern void T1435f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_cast_type */
extern void T366f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.set_type */
extern void T1445f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_type */
extern void T1444f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_type */
extern void T1432f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.set_cast_type */
extern void T1445f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_cast_type */
extern void T1444f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_cast_type */
extern void T1432f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.set_infix */
extern void T1470f52(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.tokens */
extern T0* T1470f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.set_infix */
extern void T1468f106(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.set_infix */
extern void T1467f74(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.tokens */
extern T0* T1467f33(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.set_infix */
extern void T1466f92(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.tokens */
extern T0* T1466f27(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.set_infix */
extern void T886f91(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.tokens */
extern T0* T886f66(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.set_infix */
extern void T864f74(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.set_prefix */
extern void T1470f51(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.set_prefix */
extern void T1468f105(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.tokens */
extern T0* T1468f41(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.set_prefix */
extern void T1467f73(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.set_prefix */
extern void T1466f91(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.set_prefix */
extern void T886f90(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.set_prefix */
extern void T864f73(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.tokens */
extern T0* T864f17(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.reset */
extern void T1470f53(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.reset */
extern void T886f92(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.dotnet_assemblies_do_if */
extern void T826f182(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_if */
extern void T206f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.dotnet_assemblies_do_if */
extern void T183f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.dotnet_assemblies_do_if */
extern void T56f283(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.clusters_do_explicit */
extern void T826f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTERS.do_explicit */
extern void T554f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.clusters_do_explicit */
extern void T183f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.clusters_do_explicit */
extern void T56f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.put_type_from_type_set */
extern void T1989f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.propagate_type */
extern void T1989f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_is_equal_argument_type */
extern void T226f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_is_equal_argument_type */
extern void T70f545(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_actual_argument_dynamic_types */
extern void T70f168(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.error_handler */
extern T0* T226f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_fatal_error */
extern void T226f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.put_type_from_type_set */
extern void T1988f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.propagate_type */
extern void T1988f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_from_type_set */
extern void T306f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_with_tuple_label */
extern void T306f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_tuple_label_setter_dynamic_types */
extern void T226f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_setter_dynamic_types */
extern void T70f485(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f485ot1(T0* a1, T0** a2);
extern T1 T70f485ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_argument_dynamic_types */
extern void T70f503(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_with_feature */
extern void T306f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T306f15ot1(T0* a1, T0** a2);
extern T1 T306f15ot2(T0* a1, T0** a2);
extern T1 T306f15ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type */
extern void T1416f20(GE_context* ac, T0* C, T0* a1);
extern T1 T1416f20ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_function_item */
extern T1 T298f30(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_function_item */
extern T1 T232f28(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_procedure_call */
extern T1 T298f31(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_procedure_call */
extern T1 T232f29(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_from_type_set */
extern void T302f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_tuple_label */
extern void T302f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_tuple_label_expression_dynamic_types */
extern void T226f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_expression_dynamic_types */
extern void T70f484(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f484ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_result_dynamic_types */
extern void T70f502(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_feature */
extern void T302f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_feature */
extern void T302f16p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T302f16p1ot1(T0* a1, T0** a2);
extern T1 T302f16p1ot2(T0* a1, T0** a2);
extern T1 T302f16p1ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.propagate_can_be_void */
extern void T1989f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.propagate_can_be_void */
extern void T1988f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.propagate_can_be_void */
extern void T306f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.propagate_can_be_void */
extern void T302f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_PROCESSING_INSTRUCTION.node_set_parent */
extern void T1420f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_COMMENT.node_set_parent */
extern void T1419f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.node_set_parent */
extern void T164f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_stream */
extern void T549f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_buffer */
extern void T549f292(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.name */
extern T0* T908f14(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_buffer */
extern void T549f292p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.new_file_buffer */
extern T0* T549f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_stream */
extern void T547f206(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_buffer */
extern void T547f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_buffer */
extern void T547f215p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.new_file_buffer */
extern T0* T547f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_stream */
extern void T546f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_buffer */
extern void T546f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_buffer */
extern void T546f288p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.new_file_buffer */
extern T0* T546f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_stream */
extern void T506f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_buffer */
extern void T506f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_buffer */
extern void T506f212p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.new_file_buffer */
extern T0* T506f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T549f286(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T549f286p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T549f286p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_execute_action */
extern void T549f297(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.from_hexadecimal */
extern void T1314f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.hexadecimal_to_integer */
extern T6 T84f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.string_ */
extern T0* T1314f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.to_utf8 */
extern T0* T1314f3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.utf8 */
extern T0* T1314f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.is_ascii */
extern T1 T1314f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.is_valid */
extern T1 T1314f5(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.valid_non_surrogate_code */
extern T1 T423f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.unicode */
extern T0* T1314f4(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.from_decimal */
extern void T1314f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.has_normalized_newline */
extern T1 T549f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_substring */
extern T0* T549f187(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_count */
extern T6 T549f186(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.system_literal_text */
extern T0* T549f168(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_item */
extern T2 T549f213(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.normalized_newline */
extern T0* T549f127(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.normalized_newline */
extern T0* T549f127p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_set_line_column */
extern void T549f301(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text */
extern T0* T549f111(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.pop_start_condition */
extern void T549f300(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.push_start_condition */
extern void T549f298(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_last_token */
extern void T549f299(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_execute_eof_action */
extern void T549f296(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.wrap */
extern T1 T549f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_refill_input_buffer */
extern void T549f295(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_null_trans_state */
extern T6 T549f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_previous_state */
extern T6 T549f103(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.report_invalid_unicode_character_error */
extern void T549f294(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_at_beginning_of_line */
extern T6 T549f92(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.terminate */
extern void T549f293(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.read_token */
extern void T547f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.read_token */
extern void T547f207p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_execute_action */
extern void T547f222(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.has_normalized_newline */
extern T1 T547f187(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_substring */
extern T0* T547f167(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_count */
extern T6 T547f166(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.system_literal_text */
extern T0* T547f148(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_item */
extern T2 T547f201(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.normalized_newline */
extern T0* T547f110(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_set_line_column */
extern void T547f228(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text */
extern T0* T547f84(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.pop_start_condition */
extern void T547f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.push_start_condition */
extern void T547f223(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_execute_eof_action */
extern void T547f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.wrap */
extern T1 T547f72(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_refill_input_buffer */
extern void T547f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_null_trans_state */
extern T6 T547f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_previous_state */
extern T6 T547f70(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.fatal_error */
extern void T547f219(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.report_invalid_unicode_character_error */
extern void T547f218(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_at_beginning_of_line */
extern T6 T547f62(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.terminate */
extern void T547f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_last_token */
extern void T547f216(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.read_token */
extern void T546f282(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.read_token */
extern void T546f282p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_execute_action */
extern void T546f292(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.has_normalized_newline */
extern T1 T546f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_substring */
extern T0* T546f184(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_count */
extern T6 T546f183(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.system_literal_text */
extern T0* T546f165(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_item */
extern T2 T546f211(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.normalized_newline */
extern T0* T546f124(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.normalized_newline */
extern T0* T546f124p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_set_line_column */
extern void T546f297(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text */
extern T0* T546f108(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.pop_start_condition */
extern void T546f296(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.push_start_condition */
extern void T546f293(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_last_token */
extern void T546f295(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_execute_eof_action */
extern void T546f291(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.terminate */
extern void T546f294(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.wrap */
extern T1 T546f102(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_refill_input_buffer */
extern void T546f290(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_null_trans_state */
extern T6 T546f101(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_previous_state */
extern T6 T546f100(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.report_invalid_unicode_character_error */
extern void T546f289(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_at_beginning_of_line */
extern T6 T546f89(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.read_token */
extern void T506f205(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_execute_action */
extern void T506f217(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.has_normalized_newline */
extern T1 T506f185(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text_substring */
extern T0* T506f165(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_SCANNER.text_count */
extern T6 T506f164(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.system_literal_text */
extern T0* T506f146(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text_item */
extern T2 T506f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.normalized_newline */
extern T0* T506f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_set_line_column */
extern void T506f225(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text */
extern T0* T506f79(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.pop_start_condition */
extern void T506f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.set_start_condition */
extern void T506f226(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.push_start_condition */
extern void T506f218(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.set_last_token */
extern void T506f223(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_execute_eof_action */
extern void T506f216(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.terminate */
extern void T506f222(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.wrap */
extern T1 T506f67(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_refill_input_buffer */
extern void T506f215(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_null_trans_state */
extern T6 T506f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_previous_state */
extern T6 T506f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.fatal_error */
extern void T506f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.report_invalid_unicode_character_error */
extern void T506f213(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_at_beginning_of_line */
extern T6 T506f50(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.push_start_condition_dtd_ignore */
extern void T549f288(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.push_start_condition_dtd_ignore */
extern void T547f209(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.push_start_condition_dtd_ignore */
extern void T546f284(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.push_start_condition_dtd_ignore */
extern void T506f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_encoding */
extern void T549f289(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.set_encoding */
extern void T876f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_encoding */
extern void T547f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_encoding */
extern void T546f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_encoding */
extern void T506f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_from_resolver */
extern void T549f290(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_from_resolver */
extern void T547f211(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_from_resolver */
extern void T546f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_from_resolver */
extern void T506f209(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.close_input */
extern void T549f287(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.close */
extern void T1427f11(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.is_closable */
extern T1 T1427f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.close_input */
extern void T547f208(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.close_input */
extern void T546f283(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.close_input */
extern void T506f206(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.reset */
extern void T1469f8(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.reset */
extern void T1513f9(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.reset */
extern void T1473f9(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.process */
extern void T1513f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_tagged_assertion */
extern void T1284f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.process */
extern void T1473f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_class_assertion */
extern void T1284f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.set_cast_type */
extern void T1434f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_cast_type */
extern void T1433f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.set_cast_type */
extern void T863f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.set_type */
extern void T1434f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_type */
extern void T1433f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.set_type */
extern void T863f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1508f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1507f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1506f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1505f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1504f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1503f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_across_components */
extern void T1508f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_across_components */
extern void T1507f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_across_components */
extern void T1506f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_across_components */
extern void T1505f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_across_components */
extern void T1504f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_across_components */
extern void T1503f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1508f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1507f59(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1506f54(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1505f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1504f60(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1503f55(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1508f50(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1507f58(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1506f53(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1505f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1504f59(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1503f54(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.reset */
extern void T1510f12(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.reset */
extern void T983f46(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.process */
extern void T1510f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_typed_open_argument */
extern void T1284f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.process */
extern void T983f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_question_mark_symbol */
extern void T1284f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.reset */
extern void T1760f7(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.reset */
extern void T1509f10(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.process */
extern void T1760f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_implicit_current_target */
extern void T1284f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_OPEN_TARGET.process */
extern void T1509f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_open_target */
extern void T1284f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.set_index */
extern void T1994f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.set_index */
extern void T1993f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.set_index */
extern void T1992f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.set_index */
extern void T1974f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.set_index */
extern void T1926f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.set_index */
extern void T1760f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.set_index */
extern void T1510f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_OPEN_TARGET.set_index */
extern void T1509f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_index */
extern void T1508f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_index */
extern void T1507f60(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_index */
extern void T1506f55(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_index */
extern void T1505f53(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_index */
extern void T1504f61(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_index */
extern void T1503f56(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_EXPRESSION.set_index */
extern void T1500f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_EXPRESSION.set_index */
extern void T1499f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_EXPRESSION.set_index */
extern void T1495f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_index */
extern void T1490f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STATIC_CALL_EXPRESSION.set_index */
extern void T1230f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.set_index */
extern void T1201f26(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.set_index */
extern void T1187f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_EXPRESSION.set_index */
extern void T1140f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CALL_AGENT.set_index */
extern void T1016f37(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BRACKET_EXPRESSION.set_index */
extern void T1013f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUESTION_MARK_SYMBOL.set_index */
extern void T983f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIS_EXPRESSION.set_index */
extern void T862f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CREATE_EXPRESSION.set_index */
extern void T320f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_index */
extern void T278f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1508f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1507f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1506f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1505f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1504f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1503f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CALL_AGENT.set_arguments */
extern void T1016f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENTHESIS_SYMBOL.set_seed */
extern void T1967f42(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.set_seed */
extern void T1497f37(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_AND_THEN_OPERATOR.set_seed */
extern void T1496f37(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYMBOL_OPERATOR.set_seed */
extern void T972f101(GE_context* ac, T0* C, T6 a1);
/* detachable ET_KEYWORD_OPERATOR.set_seed */
extern void T959f56(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FREE_OPERATOR.set_seed */
extern void T951f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BRACKET_SYMBOL.set_seed */
extern void T610f112(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_validity_checked */
extern void T1465f89(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.set_validity_checked */
extern void T1464f84(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.set_validity_checked */
extern void T1462f82(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.set_validity_checked */
extern void T1461f85(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.set_validity_checked */
extern void T1460f83(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.set_validity_checked */
extern void T468f79(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.set_validity_checked */
extern void T466f80(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.set_validity_checked */
extern void T367f79(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_validity_checked */
extern void T365f83(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_validity_checked */
extern void T363f82(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_validity_checked */
extern void T329f82(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.set_validity_checked */
extern void T247f91(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_validity_error */
extern void T1465f90(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.set_validity_error */
extern void T1464f85(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.set_validity_error */
extern void T1462f83(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.set_validity_error */
extern void T1461f86(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.set_validity_error */
extern void T1460f84(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.set_validity_error */
extern void T468f80(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.set_validity_error */
extern void T466f81(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.set_validity_error */
extern void T367f80(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_validity_error */
extern void T365f84(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_validity_error */
extern void T363f83(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_validity_error */
extern void T329f83(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.set_validity_error */
extern void T247f92(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_implementation_feature */
extern void T1465f115(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_implementation_feature */
extern void T1464f112(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_implementation_feature */
extern void T1462f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_implementation_feature */
extern void T1461f114(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_implementation_feature */
extern void T1460f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_implementation_feature */
extern void T468f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_implementation_feature */
extern void T466f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_implementation_feature */
extern void T367f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_implementation_feature */
extern void T365f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_implementation_feature */
extern void T363f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_implementation_feature */
extern void T329f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_implementation_feature */
extern void T247f118(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_implementation_class */
extern void T1465f116(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_implementation_class */
extern void T1464f113(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_implementation_class */
extern void T1462f109(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_implementation_class */
extern void T1461f115(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_implementation_class */
extern void T1460f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_implementation_class */
extern void T468f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_implementation_class */
extern void T466f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_implementation_class */
extern void T367f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_implementation_class */
extern void T365f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_implementation_class */
extern void T363f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_implementation_class */
extern void T329f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_implementation_class */
extern void T247f119(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.reset_preconditions */
extern void T1465f113(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_preconditions */
extern void T1464f110(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_preconditions */
extern void T1462f106(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_preconditions */
extern void T1461f112(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_preconditions */
extern void T1460f109(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_preconditions */
extern void T468f103(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_preconditions */
extern void T466f106(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_preconditions */
extern void T367f97(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_preconditions */
extern void T365f104(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_preconditions */
extern void T363f101(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_preconditions */
extern void T329f107(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_preconditions */
extern void T247f116(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_postconditions */
extern void T1465f114(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_postconditions */
extern void T1464f111(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_postconditions */
extern void T1462f107(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_postconditions */
extern void T1461f113(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_postconditions */
extern void T1460f110(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_postconditions */
extern void T468f104(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_postconditions */
extern void T466f107(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_postconditions */
extern void T367f98(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_postconditions */
extern void T365f105(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_postconditions */
extern void T363f102(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_postconditions */
extern void T329f108(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_postconditions */
extern void T247f117(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_precursor */
extern void T1465f117(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_first_precursor */
extern void T1464f114(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_first_precursor */
extern void T1462f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_first_precursor */
extern void T1461f116(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_first_precursor */
extern void T1460f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_precursor */
extern void T468f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_first_precursor */
extern void T466f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_first_precursor */
extern void T367f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_precursor */
extern void T365f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_precursor */
extern void T363f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_precursor */
extern void T329f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_precursor */
extern void T247f120(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_other_precursors */
extern void T1465f118(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_other_precursors */
extern void T1464f115(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_other_precursors */
extern void T1462f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_other_precursors */
extern void T1461f117(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_other_precursors */
extern void T1460f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_other_precursors */
extern void T468f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_other_precursors */
extern void T466f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_other_precursors */
extern void T367f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_other_precursors */
extern void T365f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_other_precursors */
extern void T363f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_other_precursors */
extern void T329f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_other_precursors */
extern void T247f121(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_version */
extern void T1465f111(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_version */
extern void T1464f108(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_version */
extern void T1462f104(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_version */
extern void T1461f110(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_version */
extern void T1460f107(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_version */
extern void T468f101(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_version */
extern void T466f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_version */
extern void T367f95(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_version */
extern void T365f102(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_version */
extern void T363f99(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_version */
extern void T329f105(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_version */
extern void T247f114(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_clients */
extern void T1465f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_clients */
extern void T1464f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_clients */
extern void T1462f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_clients */
extern void T1461f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_clients */
extern void T1460f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_clients */
extern void T468f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_clients */
extern void T466f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_clients */
extern void T367f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_clients */
extern void T365f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_clients */
extern void T363f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_clients */
extern void T329f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_clients */
extern void T247f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_seed */
extern void T1465f110(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_first_seed */
extern void T1464f107(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_first_seed */
extern void T1462f103(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_first_seed */
extern void T1461f109(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_first_seed */
extern void T1460f106(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_seed */
extern void T468f100(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_first_seed */
extern void T466f103(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_first_seed */
extern void T367f94(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_seed */
extern void T365f101(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_seed */
extern void T363f98(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_seed */
extern void T329f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_seed */
extern void T247f113(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_other_seeds */
extern void T1465f119(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_other_seeds */
extern void T1464f116(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_other_seeds */
extern void T1462f112(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_other_seeds */
extern void T1461f118(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_other_seeds */
extern void T1460f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_other_seeds */
extern void T468f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_other_seeds */
extern void T466f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_other_seeds */
extern void T367f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_other_seeds */
extern void T365f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_other_seeds */
extern void T363f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_other_seeds */
extern void T329f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_other_seeds */
extern void T247f122(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.process */
extern void T1465f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.process */
extern void T1464f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.process */
extern void T1462f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.process */
extern void T1461f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.process */
extern void T1460f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.process */
extern void T468f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.process */
extern void T466f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.process */
extern void T367f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.process */
extern void T365f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.process */
extern void T363f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.process */
extern void T329f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.process */
extern void T247f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_id */
extern void T1465f87(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_id */
extern void T1464f82(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_id */
extern void T1462f80(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_id */
extern void T1461f83(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_id */
extern void T1460f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_id */
extern void T468f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_id */
extern void T466f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_id */
extern void T367f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_id */
extern void T365f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_id */
extern void T363f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_id */
extern void T329f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_id */
extern void T247f89(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.set_index */
extern void T1502f27(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_OBJECT_TEST.set_index */
extern void T1501f33(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_TYPE.set_index */
extern void T1498f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXPRESSION_ADDRESS.set_index */
extern void T1494f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RESULT_ADDRESS.set_index */
extern void T1493f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CURRENT_ADDRESS.set_index */
extern void T1492f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_ADDRESS.set_index */
extern void T1491f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.set_index */
extern void T1445f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_index */
extern void T1444f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FALSE_CONSTANT.set_index */
extern void T1443f44(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TRUE_CONSTANT.set_index */
extern void T1442f44(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_index */
extern void T1440f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_index */
extern void T1439f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_index */
extern void T1438f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_index */
extern void T1437f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_index */
extern void T1436f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_index */
extern void T1435f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VERBATIM_STRING.set_index */
extern void T1434f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_index */
extern void T1433f48(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_index */
extern void T1432f49(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.set_index */
extern void T1233f32(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_EXPRESSION.set_index */
extern void T986f49(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VOID.set_index */
extern void T975f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.set_index */
extern void T863f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NAMED_OBJECT_TEST.set_index */
extern void T715f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_index */
extern void T366f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EQUALITY_EXPRESSION.set_index */
extern void T310f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_index */
extern void T309f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_TEST.set_index */
extern void T308f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_RESULT.set_index */
extern void T301f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.set_index */
extern void T281f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_index */
extern void T279f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CURRENT.set_index */
extern void T273f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_index */
extern void T180f112(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.reset */
extern void T1994f25(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.reset */
extern void T1993f25(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.reset */
extern void T1992f20(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.reset */
extern void T1926f20(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.reset */
extern void T1508f49(GE_context* ac, T0* C);
extern T1 T1508f49ot1(T0* a1, T0** a2);
/* detachable ET_POSTCONDITIONS.reset */
extern void T1213f24(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.reset_validity_checked */
extern void T1213f21(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.reset */
extern void T1213f24p1(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset */
extern void T401f23(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset_validity_checked */
extern void T401f22(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset */
extern void T401f23p1(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.reset */
extern void T249f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.reset */
extern void T1715f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.reset */
extern void T252f13(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.reset */
extern void T1003f14(GE_context* ac, T0* C);
extern T1 T1003f14ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.reset */
extern void T1507f57(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.reset */
extern void T331f14(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.reset */
extern void T1716f14(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.type */
extern T0* T1716f6(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.name */
extern T0* T1716f8(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.reset */
extern void T1176f13(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.type */
extern T0* T1176f6(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.name */
extern T0* T1176f8(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.local_variable */
extern T0* T1176f7(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.local_variable */
extern T0* T1716f7(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.reset */
extern void T1507f57p1(GE_context* ac, T0* C);
extern T1 T1507f57p1ot1(T0* a1, T0** a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.reset */
extern void T1506f52(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.reset */
extern void T1506f52p1(GE_context* ac, T0* C);
extern T1 T1506f52p1ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.reset */
extern void T1505f50(GE_context* ac, T0* C);
extern T1 T1505f50ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.reset */
extern void T1504f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.reset */
extern void T1504f58p1(GE_context* ac, T0* C);
extern T1 T1504f58p1ot1(T0* a1, T0** a2);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.reset */
extern void T1503f53(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.reset */
extern void T1503f53p1(GE_context* ac, T0* C);
extern T1 T1503f53p1ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_MANIFEST_STRING.reset */
extern void T1502f29(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.reset */
extern void T1501f32(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.reset */
extern void T1500f24(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.reset */
extern void T1499f28(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.reset */
extern void T1498f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.reset */
extern void T1495f28(GE_context* ac, T0* C);
extern T1 T1495f28ot1(T0* a1, T0** a2);
extern T1 T1495f28ot2(T0* a1, T0** a2);
extern T1 T1495f28ot3(T0* a1, T0** a2);
/* detachable ET_EXPRESSION_ADDRESS.reset */
extern void T1494f24(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.reset */
extern void T1493f25(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.reset */
extern void T1492f25(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.reset */
extern void T1491f25(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.reset */
extern void T1490f29(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.reset_arguments */
extern void T1490f30(GE_context* ac, T0* C);
extern T1 T1490f30ot1(T0* a1, T0** a2);
/* detachable ET_C2_CHARACTER_CONSTANT.reset */
extern void T1445f54(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.reset */
extern void T1444f53(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.reset */
extern void T1443f45(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.reset */
extern void T1442f45(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.reset */
extern void T1440f49(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.reset */
extern void T1439f49(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.reset */
extern void T1438f70(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.reset */
extern void T1437f70(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.reset */
extern void T1436f70(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.reset */
extern void T1435f70(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.reset */
extern void T1434f56(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.reset */
extern void T1433f52(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.reset */
extern void T1432f53(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.reset */
extern void T1233f33(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset */
extern void T1230f33(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.type */
extern T0* T1230f11(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset */
extern void T1230f33p1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset_arguments */
extern void T1230f34(GE_context* ac, T0* C);
extern T1 T1230f34ot1(T0* a1, T0** a2);
/* detachable ET_PARENTHESIZED_EXPRESSION.reset */
extern void T1201f25(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.reset */
extern void T1187f33(GE_context* ac, T0* C);
extern T1 T1187f33ot1(T0* a1, T0** a2);
/* detachable ET_IF_EXPRESSION.reset */
extern void T1140f30(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION_LIST.reset */
extern void T1089f8(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.reset */
extern void T1086f8(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.conditional_expression */
extern T0* T1086f5(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.conditional_expression */
extern T0* T1140f14(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.reset */
extern void T1016f34(GE_context* ac, T0* C);
extern T1 T1016f34ot1(T0* a1, T0** a2);
/* detachable ET_BRACKET_EXPRESSION.reset */
extern void T1013f27(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.reset_arguments */
extern void T1013f30(GE_context* ac, T0* C);
extern T1 T1013f30ot1(T0* a1, T0** a2);
/* [detachable] ET_BRACKET_SYMBOL.reset */
extern void T610f113(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.reset */
extern void T986f47(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.reset_unfolded_form */
extern void T986f48(GE_context* ac, T0* C);
/* detachable ET_VOID.reset */
extern void T975f36(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.reset */
extern void T863f52(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.reset */
extern void T715f30(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.reset */
extern void T366f72(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.reset */
extern void T320f33(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.type */
extern T0* T320f20(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.reset */
extern void T310f23(GE_context* ac, T0* C);
extern T1 T310f23ot1(T0* a1, T0** a2);
extern T1 T310f23ot2(T0* a1, T0** a2);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.reset */
extern void T309f25(GE_context* ac, T0* C);
extern T1 T309f25ot1(T0* a1, T0** a2);
extern T1 T309f25ot2(T0* a1, T0** a2);
/* [detachable] ET_OBJECT_TEST.reset */
extern void T308f27(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.reset */
extern void T301f37(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.reset */
extern void T281f35(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset */
extern void T279f30(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset */
extern void T279f30p1(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset_arguments */
extern void T279f34(GE_context* ac, T0* C);
extern T1 T279f34ot1(T0* a1, T0** a2);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T278f33(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T278f33p1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T278f33p0(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset_arguments */
extern void T278f34(GE_context* ac, T0* C);
extern T1 T278f34ot1(T0* a1, T0** a2);
/* [detachable] ET_CURRENT.reset */
extern void T273f38(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.process */
extern void T1994f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.process */
extern void T1993f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.process */
extern void T1992f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.process */
extern void T1926f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.process */
extern void T1508f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.process */
extern void T1507f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.process */
extern void T1506f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.process */
extern void T1505f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.process */
extern void T1504f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.process */
extern void T1503f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.process */
extern void T1502f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_OBJECT_TEST.process */
extern void T1501f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_EXPRESSION.process */
extern void T1500f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_EXPRESSION.process */
extern void T1499f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_TYPE.process */
extern void T1498f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_EXPRESSION.process */
extern void T1495f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPRESSION_ADDRESS.process */
extern void T1494f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RESULT_ADDRESS.process */
extern void T1493f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CURRENT_ADDRESS.process */
extern void T1492f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_ADDRESS.process */
extern void T1491f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.process */
extern void T1490f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.process */
extern void T1445f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.process */
extern void T1444f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.process */
extern void T1443f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.process */
extern void T1442f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.process */
extern void T1440f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.process */
extern void T1439f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.process */
extern void T1438f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.process */
extern void T1437f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.process */
extern void T1436f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.process */
extern void T1435f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.process */
extern void T1434f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.process */
extern void T1433f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.process */
extern void T1432f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.process */
extern void T1233f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_EXPRESSION.process */
extern void T1230f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.process */
extern void T1201f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.process */
extern void T1187f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_EXPRESSION.process */
extern void T1140f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CALL_AGENT.process */
extern void T1016f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BRACKET_EXPRESSION.process */
extern void T1013f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.process */
extern void T986f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.process */
extern void T975f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.process */
extern void T863f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.process */
extern void T715f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.process */
extern void T366f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CREATE_EXPRESSION.process */
extern void T320f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EQUALITY_EXPRESSION.process */
extern void T310f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.process */
extern void T309f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST.process */
extern void T308f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RESULT.process */
extern void T301f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.process */
extern void T281f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.process */
extern void T279f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.process */
extern void T278f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.process */
extern void T273f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.process */
extern void T180f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_meta_type */
extern void T444f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_meta_type */
extern void T443f59(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.set_meta_type */
extern void T315f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_meta_type */
extern void T311f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_meta_type */
extern void T294f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_meta_type */
extern void T219f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type_from_type_set */
extern void T1416f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T1416f17ot1(T0* a1, T0** a2);
extern T1 T1416f17ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type_from_type_set */
extern void T1416f17p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T1416f17p1ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type_from_type_set */
extern void T469f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_TARGET_LIST.item */
extern T0* T1311f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type_from_type_set */
extern void T469f17p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T469f17p1ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_type_from_type_set */
extern void T444f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_type_from_type_set */
extern void T443f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.put_type_from_type_set */
extern void T315f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_type_from_type_set */
extern void T311f70(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_type_from_type_set */
extern void T294f70(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_type_from_type_set */
extern void T261f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_type_from_type_set */
extern void T219f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.propagate_can_be_void */
extern void T1416f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.propagate_can_be_void */
extern void T469f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.propagate_can_be_void */
extern void T444f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.propagate_can_be_void */
extern void T443f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.propagate_can_be_void */
extern void T315f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.propagate_can_be_void */
extern void T311f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.propagate_can_be_void */
extern void T294f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.propagate_can_be_void */
extern void T261f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.propagate_can_be_void */
extern void T219f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.set_never_void */
extern void T1416f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.set_never_void */
extern void T469f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_never_void */
extern void T444f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_never_void */
extern void T443f63(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.set_never_void */
extern void T315f26(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_never_void */
extern void T311f61(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_never_void */
extern void T294f61(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.set_never_void */
extern void T261f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_never_void */
extern void T219f60(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_target */
extern void T1416f19(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_target */
extern void T469f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_TARGET_LIST.force_last */
extern void T1311f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].force */
extern void T1665f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].fill_with */
extern void T1664f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].extend */
extern void T1664f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.fixed_array */
extern T0* T1311f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].default_create */
extern T0* T1665c3(GE_context* ac);
/* detachable ET_DYNAMIC_TARGET_LIST.resize */
extern void T1311f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].aliased_resized_area */
extern T0* T1665f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].aliased_resized_area */
extern T0* T1664f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_TARGET_LIST.new_capacity */
extern T6 T1311f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_TARGET_LIST.capacity */
extern T6 T1311f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_TARGET_LIST.has */
extern T1 T1311f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.index_of */
extern T6 T1311f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.put_last */
extern void T1311f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.make_with_capacity */
extern T0* T1311c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].make */
extern T0* T1665f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].make_empty */
extern T0* T1664c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_target */
extern void T444f68(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_target */
extern void T443f68(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.put_target */
extern void T315f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_target */
extern void T311f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_target */
extern void T294f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_target */
extern void T261f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_target */
extern void T219f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_runtime_name_to_string */
extern void T1902f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T1901f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_runtime_name_to_string */
extern void T593f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T567f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_runtime_name_to_string */
extern void T236f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_unaliased_to_string */
extern void T1902f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T1901f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_unaliased_to_string */
extern void T593f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T567f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_unaliased_to_string */
extern void T236f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_to_string */
extern void T1902f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_to_string */
extern void T1901f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_to_string */
extern void T593f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_to_string */
extern void T567f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_to_string */
extern void T236f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.reset */
extern void T1902f26(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.reset */
extern void T1901f23(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset */
extern void T593f33(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.reset */
extern void T567f21(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.reset */
extern void T236f33(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.reset_qualified_anchored_types */
extern void T1902f27(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T1901f24(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_qualified_anchored_types */
extern void T593f34(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T567f22(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.reset_qualified_anchored_types */
extern void T236f34(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.append_to_string_with_space */
extern void T983f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.expandedness_text */
extern T0* T983f27(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.tokens */
extern T0* T983f15(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_expandedness_mark */
extern T1 T983f31(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_reference */
extern T1 T983f17(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expanded */
extern T1 T983f16(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.separateness_text */
extern T0* T983f26(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_separateness_mark */
extern T1 T983f30(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separate */
extern T1 T983f18(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.attachment_text */
extern T0* T983f25(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.append_to_string_with_space */
extern void T610f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.expandedness_text */
extern T0* T610f31(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_expandedness_mark */
extern T1 T610f41(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_reference */
extern T1 T610f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expanded */
extern T1 T610f23(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.separateness_text */
extern T0* T610f30(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_separateness_mark */
extern T1 T610f40(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separate */
extern T1 T610f25(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.attachment_text */
extern T0* T610f29(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.append_to_string_with_space */
extern void T561f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.expandedness_text */
extern T0* T561f15(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.tokens */
extern T0* T561f18(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_expandedness_mark */
extern T1 T561f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_reference */
extern T1 T561f11(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expanded */
extern T1 T561f10(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.separateness_text */
extern T0* T561f14(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_separateness_mark */
extern T1 T561f19(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_separateness_mark */
extern T1 T409f25(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separate */
extern T1 T409f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.attachment_text */
extern T0* T561f13(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_attachment_mark */
extern T1 T561f17(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.append_to_string_with_space */
extern void T409f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.expandedness_text */
extern T0* T409f23(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_expandedness_mark */
extern T1 T409f26(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_reference */
extern T1 T409f19(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expanded */
extern T1 T409f18(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.separateness_text */
extern T0* T409f22(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.attachment_text */
extern T0* T409f21(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.append_to_string_with_space */
extern void T343f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.expandedness_text */
extern T0* T343f34(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_expandedness_mark */
extern T1 T343f26(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_reference */
extern T1 T343f24(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expanded */
extern T1 T343f23(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.separateness_text */
extern T0* T343f33(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_separateness_mark */
extern T1 T343f27(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separate */
extern T1 T343f25(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.attachment_text */
extern T0* T343f32(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.append_to_string_with_space */
extern void T342f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.expandedness_text */
extern T0* T342f11(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.tokens */
extern T0* T342f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_expandedness_mark */
extern T1 T342f15(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_reference */
extern T1 T342f22(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_expanded */
extern T1 T342f21(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.separateness_text */
extern T0* T342f10(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_separateness_mark */
extern T1 T342f14(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_separate */
extern T1 T342f20(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.attachment_text */
extern T0* T342f9(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.classes_do_all */
extern void T826f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_all */
extern void T586f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_all */
extern void T586f172p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_all */
extern void T207f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_all */
extern void T207f172p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.classes_do_all */
extern void T183f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_do_all */
extern void T56f277(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.classes_do_all */
extern void T36f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.classes_do_until */
extern void T826f174(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].go_after */
extern void T599f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_go_after */
extern void T221f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_until */
extern void T586f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_until */
extern void T207f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.classes_do_until */
extern void T183f199(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.classes_do_until */
extern void T56f278(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.classes_do_until */
extern void T36f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.classes_do_if */
extern void T826f176(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_if */
extern void T586f175(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_if */
extern void T586f175p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_if */
extern void T207f175(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_if */
extern void T207f175p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.classes_do_if */
extern void T183f201(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.classes_do_if */
extern void T56f279(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.classes_do_if */
extern void T36f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.master_classes_do_all */
extern void T826f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].do_all */
extern void T221f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_all */
extern void T586f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_all */
extern void T207f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_all */
extern void T183f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_all */
extern void T56f191(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.master_classes_do_all */
extern void T36f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.master_classes_do_until */
extern void T826f178(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_until */
extern void T586f177(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_until */
extern void T207f177(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_until */
extern void T183f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_until */
extern void T56f280(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.master_classes_do_until */
extern void T36f212(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.master_classes_do_if */
extern void T826f179(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].do_if */
extern void T221f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_if */
extern void T586f178(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_if */
extern void T207f178(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_if */
extern void T183f204(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if */
extern void T56f281(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.master_classes_do_if */
extern void T36f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.master_classes_do_if_until */
extern void T826f180(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_if_until */
extern void T586f179(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_if_until */
extern void T207f179(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_if_until */
extern void T183f205(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if_until */
extern void T56f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_SYSTEM.master_classes_do_if_until */
extern void T36f214(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIBRARY.universes_do_if */
extern void T826f177(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.universes_do_if */
extern void T206f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_LIBRARIES.universes_do_if */
extern void T184f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.universes_do_if */
extern void T586f176(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLIES.universes_do_if */
extern void T585f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.universes_do_if */
extern void T207f176(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.universes_do_if */
extern void T183f202(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.universes_do_if */
extern void T56f285(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.universes_do_if */
extern void T36f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_runtime_name_to_string */
extern void T1478f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_runtime_name_to_string */
extern void T1477f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.append_runtime_name_to_string */
extern void T1476f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_runtime_name_to_string */
extern void T1450f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.append_runtime_name_to_string */
extern void T867f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.append_runtime_name_to_string */
extern void T442f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_runtime_name_to_string */
extern void T246f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_unaliased_to_string */
extern void T1478f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_unaliased_to_string */
extern void T1477f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.append_unaliased_to_string */
extern void T1476f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_unaliased_to_string */
extern void T1450f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.append_unaliased_to_string */
extern void T867f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.append_unaliased_to_string */
extern void T442f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.append_unaliased_to_string */
extern void T248f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_unaliased_to_string */
extern void T246f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.append_unaliased_to_string */
extern void T73f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.append_unaliased_to_string */
extern void T71f126(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_to_string */
extern void T1478f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_to_string */
extern void T1477f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.append_to_string */
extern void T1476f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_to_string */
extern void T1450f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.upper_name */
extern T0* T1450f28(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.append_to_string */
extern void T867f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.upper_name */
extern T0* T867f29(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.append_to_string */
extern void T442f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_to_string */
extern void T246f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.upper_name */
extern T0* T246f46(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset_qualified_anchored_types */
extern void T1478f83(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset_qualified_anchored_types */
extern void T1477f90(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset_qualified_anchored_types */
extern void T1476f90(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_qualified_anchored_types */
extern void T1450f92(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_qualified_anchored_types */
extern void T867f89(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset_qualified_anchored_types */
extern void T442f84(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset_qualified_anchored_types */
extern void T248f123(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset_qualified_anchored_types */
extern void T246f78(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_qualified_anchored_types */
extern void T73f321(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset_qualified_anchored_types */
extern void T71f130(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.process */
extern void T1478f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.process */
extern void T1477f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.process */
extern void T1476f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.process */
extern void T1450f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.process */
extern void T867f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.process */
extern void T442f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.process */
extern void T246f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset */
extern void T1478f82(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset */
extern void T1477f89(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset */
extern void T1476f89(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_feature_name */
extern void T180f123(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LIKE_FEATURE.is_like_argument */
extern T1 T1476f30(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset */
extern void T1450f91(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_base_types */
extern void T1450f93(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset */
extern void T867f88(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset */
extern void T442f83(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset */
extern void T248f122(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset */
extern void T246f77(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_type */
extern void T73f320(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset */
extern void T71f129(GE_context* ac, T0* C);
extern T1 T71f129ot1(T0* a1, T0** a2);
extern T1 T71f129ot2(T0* a1, T0** a2);
extern T1 T71f129ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.add_adapted_base_classes_to_list */
extern void T1478f88(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f88ot1(T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.keep_first */
extern void T237f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.force_last */
extern void T237f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.resize */
extern void T237f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].aliased_resized_area */
extern T0* T613f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_TYPE].aliased_resized_area */
extern T0* T251f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.new_capacity */
extern T6 T237f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.capacity */
extern T6 T237f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].force_last */
extern void T689f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].force */
extern void T1370f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].fill_with */
extern void T1372f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].extend */
extern void T1372f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].resize */
extern void T689f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].aliased_resized_area */
extern T0* T1370f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].aliased_resized_area */
extern T0* T1372f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].new_capacity */
extern T6 T689f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].extendible */
extern T1 T689f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.tokens */
extern T0* T1478f37(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.seed */
extern T6 T1478f36(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.add_adapted_base_classes_to_list */
extern void T1477f92(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f92ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.tokens */
extern T0* T1477f33(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.seed */
extern T6 T1477f32(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.add_adapted_base_classes_to_list */
extern void T1476f93(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.seeded_feature */
extern T0* T73f196(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.index */
extern T6 T1476f31(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.tokens */
extern T0* T1476f29(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.add_adapted_base_classes_to_list */
extern void T1450f86(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f86ot1(T0* a1, T0** a2);
extern T1 T1450f86ot2(T0* a1, T0** a2);
extern T1 T1450f86ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.constraint_base_types */
extern T0* T867f28(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.tokens */
extern T0* T1450f30(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.ancestor */
extern T0* T73f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.base_type */
extern T0* T408f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.add_adapted_base_classes_to_list */
extern void T1424f41(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.remove_last */
extern void T237f57(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.item */
extern T0* T237f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.valid_index */
extern T1 T237f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER.add_adapted_base_classes_to_list */
extern void T867f83(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f83ot1(T0* a1, T0** a2);
extern T1 T867f83ot2(T0* a1, T0** a2);
extern T1 T867f83ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.tokens */
extern T0* T867f40(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.add_adapted_base_classes_to_list */
extern void T442f85(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.add_adapted_base_classes_to_list */
extern void T237f62(GE_context* ac, T0* C, T0* a1);
extern T1 T237f62ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.put */
extern void T237f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.last */
extern T0* T237f17(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.add_adapted_base_classes_to_list */
extern void T246f82(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f82ot1(T0* a1, T0** a2);
extern T1 T246f82ot2(T0* a1, T0** a2);
extern T1 T246f82ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.tokens */
extern T0* T246f8(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_class */
extern void T1567f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_class */
extern void T1566f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_class */
extern void T1565f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_class */
extern void T1564f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_class */
extern void T1296f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_class */
extern void T1288f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_class */
extern void T1287f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_class */
extern void T1282f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_class */
extern void T1281f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_class */
extern void T1280f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_class */
extern void T1279f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_class */
extern void T1278f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_class */
extern void T1277f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_class */
extern void T1266f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_class */
extern void T1265f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_class */
extern void T457f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_class */
extern void T454f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.internal_process_class */
extern void T454f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_class_processed */
extern void T62f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_class_processed */
extern void T371f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_implementation_checked */
extern void T73f272(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].wipe_out */
extern void T926f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_wipe_out */
extern void T926f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_wipe_out */
extern void T926f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_wipe_out */
extern void T926f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_wipe_out */
extern void T926f50(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].keep_head */
extern void T1598f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].set_count */
extern void T1598f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].move_all_cursors_after */
extern void T926f49(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_suppliers */
extern void T73f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].extend */
extern void T926f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].put */
extern void T926f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make */
extern T0* T926c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].set_internal_cursor */
extern void T926f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_slots */
extern void T926f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_clashes */
extern void T926f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_key_storage */
extern void T926f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_item_storage */
extern void T926f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].make */
extern T0* T1600f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].make_empty */
extern T0* T1598c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].default_create */
extern T0* T1600c3(GE_context* ac);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_invariants_validity */
extern void T454f30(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.flat_mode */
extern T1 T454f17(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_invariants_validity */
extern void T376f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INVARIANTS.set_validity_error */
extern void T1168f33(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.set_validity_checked */
extern void T1168f32(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.assertion */
extern T0* T1168f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INVARIANTS.item */
extern T0* T1168f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.set_supplier_handler */
extern void T376f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.set_supplier_handler */
extern void T380f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_features_validity */
extern void T454f29(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_procedure_validity */
extern void T454f34(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].wipe_out */
extern void T1297f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_wipe_out */
extern void T1297f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_wipe_out */
extern void T1297f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_wipe_out */
extern void T1297f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_wipe_out */
extern void T1297f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].unset_found_item */
extern void T1297f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].move_all_cursors_after */
extern void T1297f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_next_cursor */
extern void T1660f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_after */
extern void T1660f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].forth */
extern void T1297f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_forth */
extern void T1297f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].add_traversing_cursor */
extern void T1297f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_position */
extern void T1660f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].remove_traversing_cursor */
extern void T1297f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_item */
extern T6 T1297f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_precursor_procedure_validity */
extern void T454f38(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_feature_validity */
extern void T376f85(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_for_iteration */
extern T0* T1297f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_item */
extern T0* T1297f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_item */
extern T0* T1297f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].after */
extern T1 T1297f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_after */
extern T1 T1297f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].after */
extern T1 T1660f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].start */
extern void T1297f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_start */
extern void T1297f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_off */
extern T1 T1297f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].off */
extern T1 T1660f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].is_empty */
extern T1 T1297f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_feature_validity */
extern void T376f82(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.build_preconditions_attachment_scope */
extern void T376f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.free_attachment_scope */
extern void T376f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.new_attachment_scope */
extern T0* T376f50(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.build_assertions_attachment_scope */
extern void T376f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.add_precursors_with_preconditions_recursive */
extern void T376f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.feature_checker */
extern T0* T376f46(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_assertions_validity */
extern void T454f36(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_IMPLEMENTATION_CHECKER.flat_dbc_mode */
extern T1 T454f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_postconditions_validity */
extern void T376f84(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_POSTCONDITIONS.set_validity_error */
extern void T1213f23(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.set_validity_checked */
extern void T1213f22(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.assertion */
extern T0* T1213f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_POSTCONDITIONS.item */
extern T0* T1213f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.set_precursor_procedures */
extern void T376f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_query_validity */
extern void T454f33(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].wipe_out */
extern void T1298f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_wipe_out */
extern void T1298f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_wipe_out */
extern void T1298f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_wipe_out */
extern void T1298f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_wipe_out */
extern void T1298f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].unset_found_item */
extern void T1298f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].move_all_cursors_after */
extern void T1298f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_next_cursor */
extern void T1661f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_after */
extern void T1661f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].forth */
extern void T1298f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_forth */
extern void T1298f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].add_traversing_cursor */
extern void T1298f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_position */
extern void T1661f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].remove_traversing_cursor */
extern void T1298f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_item */
extern T6 T1298f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_precursor_query_validity */
extern void T454f37(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_for_iteration */
extern T0* T1298f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_item */
extern T0* T1298f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_item */
extern T0* T1298f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].after */
extern T1 T1298f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_after */
extern T1 T1298f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].after */
extern T1 T1661f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].start */
extern void T1298f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_start */
extern void T1298f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_off */
extern T1 T1298f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].off */
extern T1 T1661f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].is_empty */
extern T1 T1298f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_precursor_queries */
extern void T376f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SUPPLIER_BUILDER.set */
extern void T1299f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IMPLEMENTATION_CHECKER.suppliers_enabled */
extern T1 T454f15(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_signatures_validity */
extern void T454f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].wipe_out */
extern void T1267f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_wipe_out */
extern void T1267f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_wipe_out */
extern void T1267f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_wipe_out */
extern void T1267f60(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].keep_head */
extern void T1613f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].set_count */
extern void T1613f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_wipe_out */
extern void T1267f59(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].keep_head */
extern void T1610f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].set_count */
extern void T1610f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1267f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].move_all_cursors_after */
extern void T1267f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_next_cursor */
extern void T1609f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_after */
extern void T1609f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].forth */
extern void T1267f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_forth */
extern void T1267f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].add_traversing_cursor */
extern void T1267f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_position */
extern void T1609f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].remove_traversing_cursor */
extern void T1267f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1267f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_signature_validity */
extern void T454f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_signature_validity */
extern void T1280f11(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].forth */
extern void T1643f10(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_forth */
extern void T1644f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].add_traversing_cursor */
extern void T1644f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_next_cursor */
extern void T1643f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].remove_traversing_cursor */
extern void T1644f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set */
extern void T1643f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_SIGNATURE_CHECKER.check_selected_signature_validity */
extern void T1280f15(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T1280f15ot1(T0* a1);
extern T1 T1280f15ot2(T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.is_empty */
extern T1 T249f8(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_actual_parameters_unfolded */
extern void T1280f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SIGNATURE_CHECKER.tokens */
extern T0* T1280f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2c_error */
extern void T51f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2c */
extern T0* T759c962(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd2_error */
extern T1 T51f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2c_error */
extern void T35f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd2_error */
extern T1 T35f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2d_error */
extern void T51f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2d */
extern T0* T759c961(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENT_FEATURE.name */
extern T0* T1302f13(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2d_error */
extern void T35f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SIGNATURE_CHECKER.error_handler */
extern T0* T1280f8(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.set_fatal_error */
extern void T1280f16(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.set */
extern void T237f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].item */
extern T0* T1643f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].start */
extern void T1643f9(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_start */
extern void T1644f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_off */
extern T1 T1644f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].off */
extern T1 T1643f7(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].new_cursor */
extern T0* T1644f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].make */
extern T0* T1643c8(GE_context* ac, T0* a1);
/* detachable ET_PARENT_FEATURE.is_deferred */
extern T1 T1302f14(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_undefine */
extern T1 T1302f19(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.check_joined_signature_validity */
extern void T1280f14(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0b_error */
extern void T51f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0b */
extern T0* T759c960(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdjr_error */
extern T1 T51f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0b_error */
extern void T35f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdjr_error */
extern T1 T35f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0a_error */
extern void T51f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0a */
extern T0* T759c959(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0a_error */
extern void T35f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0c_error */
extern void T51f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0c */
extern T0* T759c958(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0c_error */
extern void T35f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SIGNATURE_CHECKER.check_redeclared_signature_validity */
extern void T1280f13(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T1280f13ot1(T0* a1);
extern T1 T1280f13ot2(T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd6b_error */
extern void T51f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd6b */
extern T0* T759c957(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd6_error */
extern T1 T51f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd6b_error */
extern void T35f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd6_error */
extern T1 T35f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2a_error */
extern void T51f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2a */
extern T0* T759c956(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2a_error */
extern void T35f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2b_error */
extern void T51f345(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2b */
extern T0* T759c955(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2b_error */
extern void T35f345(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_for_iteration */
extern T0* T1267f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_item */
extern T0* T1267f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1267f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].after */
extern T1 T1267f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_after */
extern T1 T1267f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].after */
extern T1 T1609f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].start */
extern void T1267f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_start */
extern void T1267f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_off */
extern T1 T1267f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].off */
extern T1 T1609f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].is_empty */
extern T1 T1267f31(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.resolve_feature_adaptations */
extern void T454f31(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.resolve_inherited_features */
extern void T454f35(GE_context* ac, T0* C, T0* a1);
extern T1 T454f35ot1(T0* a1, T0** a2);
/* [detachable] ET_INHERITED_FEATURE.set_flattened_parent */
extern void T1303f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_flattened_feature */
extern void T1303f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].found_item */
extern T0* T1267f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].found */
extern T1 T1267f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].search */
extern void T1267f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].search_position */
extern void T1267f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1267f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_NAME_TESTER.test */
extern T1 T1268f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1267f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1267f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.resolve_feature_adaptations */
extern void T1275f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].wipe_out */
extern void T1629f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].slots_wipe_out */
extern void T1629f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].clashes_wipe_out */
extern void T1629f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].key_storage_wipe_out */
extern void T1629f46(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].keep_head */
extern void T1618f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].set_count */
extern void T1618f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].item_storage_wipe_out */
extern void T1629f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].unset_found_item */
extern void T1629f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].move_all_cursors_after */
extern void T1629f43(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].set_next_cursor */
extern void T1857f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].set_after */
extern void T1857f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.process_replication */
extern void T1275f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].wipe_out */
extern void T1630f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_wipe_out */
extern void T1630f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_wipe_out */
extern void T1630f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T1630f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T1630f54(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].keep_head */
extern void T1859f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].set_count */
extern void T1859f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T1630f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].move_all_cursors_after */
extern void T1630f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_next_cursor */
extern void T1858f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_after */
extern void T1858f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].forth */
extern void T1630f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_forth */
extern void T1630f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].add_traversing_cursor */
extern void T1630f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_position */
extern void T1858f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].remove_traversing_cursor */
extern void T1630f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T1630f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.process_replicated_feature */
extern void T1275f37(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmrc2b_error */
extern void T51f343(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmrc2b */
extern T0* T759c953(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].after */
extern T1 T1866f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].forth */
extern void T1866f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_forth */
extern void T1634f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_position */
extern void T1866f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].add_traversing_cursor */
extern void T1634f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_next_cursor */
extern void T1866f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].remove_traversing_cursor */
extern void T1634f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].item */
extern T0* T1866f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_item */
extern T0* T1634f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].start */
extern void T1866f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_start */
extern void T1634f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_after */
extern void T1866f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].is_empty */
extern T1 T1634f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].off */
extern T1 T1866f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].before */
extern T1 T1866f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].new_cursor */
extern T0* T1634f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].make */
extern T0* T1866c9(GE_context* ac, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.ast_position */
extern T0* T759f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].first */
extern T0* T1634f7(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmrc2_error */
extern T1 T51f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmrc2b_error */
extern void T35f343(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vmrc2_error */
extern T1 T35f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.selected_feature */
extern T0* T1520f17(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_select */
extern T1 T1302f22(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.selected_feature */
extern T0* T1303f18(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.add_replicated_feature */
extern void T1520f33(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].force_last */
extern void T1644f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE].put_right */
extern void T2017f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE].make */
extern T0* T2017c3(GE_context* ac, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].make */
extern T0* T1644c7(GE_context* ac);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].set_internal_cursor */
extern void T1644f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.add_replicated_feature */
extern void T1303f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.has_seed */
extern T1 T1302f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.set_selected */
extern void T1520f35(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.set_selected */
extern void T1303f39(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.set_replicated */
extern void T1520f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_IDS.put */
extern void T594f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].resize_with_default */
extern void T877f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ARRAY [[attached] INTEGER_32].conservative_resize_with_default */
extern void T284f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable ET_FEATURE_IDS.integer_array_ */
extern T0* T594f5(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.has */
extern T1 T594f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_IDS.make */
extern T0* T594c9(GE_context* ac, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.set_replicated */
extern void T1303f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.has_selected_feature */
extern T1 T1520f18(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.has_selected_feature */
extern T1 T1303f19(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmrc2a_error */
extern void T51f342(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmrc2a */
extern T0* T759c952(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vmrc2a_error */
extern void T35f342(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.error_handler */
extern T0* T1275f20(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.set_fatal_error */
extern void T1275f30(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].forth */
extern void T1633f16(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_forth */
extern void T1633f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].add_traversing_cursor */
extern void T1633f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].set_next_cursor */
extern void T1862f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].remove_traversing_cursor */
extern void T1633f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].set */
extern void T1862f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].put_last */
extern void T1634f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].force */
extern void T1864f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].fill_with */
extern void T1865f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].extend */
extern void T1865f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.seeded_feature */
extern T0* T1520f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.seeded_feature */
extern T0* T1303f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].item_for_iteration */
extern T0* T1633f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_item */
extern T0* T1633f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].after */
extern T1 T1633f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_after */
extern T1 T1633f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].start */
extern void T1633f15(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_start */
extern void T1633f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_off */
extern T1 T1633f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].off */
extern T1 T1862f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].make */
extern T0* T1634c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].set_internal_cursor */
extern void T1634f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].make */
extern T0* T1864f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].make_empty */
extern T0* T1865c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].default_create */
extern T0* T1864c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_for_iteration */
extern T6 T1630f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_key */
extern T6 T1630f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T1630f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_for_iteration */
extern T0* T1630f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_item */
extern T0* T1630f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T1630f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].after */
extern T1 T1630f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_after */
extern T1 T1630f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].after */
extern T1 T1858f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].start */
extern void T1630f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_start */
extern void T1630f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_off */
extern T1 T1630f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].off */
extern T1 T1858f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].is_empty */
extern T1 T1630f28(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.item */
extern T6 T594f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.record_replicable_feature */
extern void T1275f36(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].force_new */
extern void T1630f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T1630f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T1630f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].force */
extern void T1861f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].fill_with */
extern void T1859f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].extend */
extern void T1859f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_put */
extern void T1630f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T1630f61(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T1630f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T1630f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].resize */
extern void T1630f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T1630f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T1630f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T1630f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T1630f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].aliased_resized_area */
extern T0* T1861f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].aliased_resized_area */
extern T0* T1859f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T1630f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T1630f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T1630f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].replace_found_item */
extern void T1630f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REPLICATED_FEATURE.make */
extern T0* T1632c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1633c12(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].set_internal_cursor */
extern void T1633f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].new_cursor */
extern T0* T1633f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1862c7(GE_context* ac, T0* a1);
/* [detachable] ET_REPLICATED_FEATURE.put_feature */
extern void T1632f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].force_last */
extern void T1633f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE].put_right */
extern void T1863f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1863c3(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].found_item */
extern T0* T1630f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].found */
extern T1 T1630f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].search */
extern void T1630f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].search_position */
extern void T1630f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features */
extern void T1275f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss1a_error */
extern void T51f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmss1a */
extern T0* T759c945(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss1_error */
extern T1 T51f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss1a_error */
extern void T35f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss1_error */
extern T1 T35f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs1a_error */
extern void T51f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs1a */
extern T0* T759c944(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs1_error */
extern T1 T51f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs1a_error */
extern void T35f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs1_error */
extern T1 T35f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].wipe_out */
extern void T1628f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].slots_wipe_out */
extern void T1628f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].clashes_wipe_out */
extern void T1628f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].key_storage_wipe_out */
extern void T1628f60(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].keep_head */
extern void T1653f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].set_count */
extern void T1653f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].item_storage_wipe_out */
extern void T1628f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].unset_found_item */
extern void T1628f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].move_all_cursors_after */
extern void T1628f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].set_next_cursor */
extern void T1854f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].set_after */
extern void T1854f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].forth */
extern void T1628f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_forth */
extern void T1628f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].add_traversing_cursor */
extern void T1628f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].set_position */
extern void T1854f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].remove_traversing_cursor */
extern void T1628f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].clashes_item */
extern T6 T1628f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus1a_error */
extern void T51f333(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus1a */
extern T0* T759c943(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus1_error */
extern T1 T51f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus1a_error */
extern void T35f333(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus1_error */
extern T1 T35f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].key_for_iteration */
extern T0* T1628f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_key */
extern T0* T1628f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].key_storage_item */
extern T0* T1628f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].item_for_iteration */
extern T1 T1628f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_item */
extern T1 T1628f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].item_storage_item */
extern T1 T1628f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].after */
extern T1 T1628f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_after */
extern T1 T1628f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].after */
extern T1 T1854f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].start */
extern void T1628f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_start */
extern void T1628f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].cursor_off */
extern T1 T1628f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].off */
extern T1 T1854f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].is_empty */
extern T1 T1628f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].wipe_out */
extern void T1627f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].slots_wipe_out */
extern void T1627f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].clashes_wipe_out */
extern void T1627f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_storage_wipe_out */
extern void T1627f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_storage_wipe_out */
extern void T1627f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].unset_found_item */
extern void T1627f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].move_all_cursors_after */
extern void T1627f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].set_next_cursor */
extern void T1853f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].set_after */
extern void T1853f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].forth */
extern void T1627f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].cursor_forth */
extern void T1627f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].add_traversing_cursor */
extern void T1627f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].set_position */
extern void T1853f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].remove_traversing_cursor */
extern void T1627f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].clashes_item */
extern T6 T1627f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel2a_error */
extern void T51f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vlel2a */
extern T0* T759c942(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel2_error */
extern T1 T51f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel2a_error */
extern void T35f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel2_error */
extern T1 T35f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_for_iteration */
extern T0* T1627f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].cursor_item */
extern T0* T1627f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_storage_item */
extern T0* T1627f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].after */
extern T1 T1627f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].cursor_after */
extern T1 T1627f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].after */
extern T1 T1853f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].start */
extern void T1627f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].cursor_start */
extern void T1627f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].cursor_off */
extern T1 T1627f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].off */
extern T1 T1853f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].wipe_out */
extern void T1289f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].slots_wipe_out */
extern void T1289f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].clashes_wipe_out */
extern void T1289f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].key_storage_wipe_out */
extern void T1289f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_storage_wipe_out */
extern void T1289f55(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_RENAME].keep_head */
extern void T1652f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME].set_count */
extern void T1652f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].unset_found_item */
extern void T1289f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].move_all_cursors_after */
extern void T1289f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].set_next_cursor */
extern void T1651f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].set_after */
extern void T1651f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].forth */
extern void T1289f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].cursor_forth */
extern void T1289f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].add_traversing_cursor */
extern void T1289f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].set_position */
extern void T1651f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].remove_traversing_cursor */
extern void T1289f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].clashes_item */
extern T6 T1289f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhrc1a_error */
extern void T51f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vhrc1a */
extern T0* T759c941(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhrc1_error */
extern T1 T51f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhrc1a_error */
extern void T35f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vhrc1_error */
extern T1 T35f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_for_iteration */
extern T0* T1289f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].cursor_item */
extern T0* T1289f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_storage_item */
extern T0* T1289f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].after */
extern T1 T1289f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].cursor_after */
extern T1 T1289f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].after */
extern T1 T1651f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].start */
extern void T1289f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].cursor_start */
extern void T1289f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].cursor_off */
extern T1 T1289f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].off */
extern T1 T1651f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].force_last_new */
extern void T1629f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].key_storage_put */
extern void T1629f53(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].force */
extern void T1617f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].fill_with */
extern void T1618f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].extend */
extern void T1618f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].item_storage_put */
extern void T1629f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].force */
extern void T1615f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].fill_with */
extern void T1613f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].extend */
extern void T1613f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].slots_put */
extern void T1629f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].clashes_put */
extern void T1629f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].slots_item */
extern T6 T1629f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].hash_position */
extern T6 T1629f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].resize */
extern void T1629f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].clashes_resize */
extern void T1629f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].special_integer_ */
extern T0* T1629f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].key_storage_resize */
extern void T1629f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1617f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1618f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].item_storage_resize */
extern void T1629f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1615f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1613f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].key_storage_item */
extern T0* T1629f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].clashes_item */
extern T6 T1629f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].slots_resize */
extern void T1629f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].new_modulus */
extern T6 T1629f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].new_capacity */
extern T6 T1629f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].has */
extern T1 T1629f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].search_position */
extern void T1629f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ALIAS_NAME_TESTER.test */
extern T1 T1271f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].put_last_new */
extern void T1267f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1267f86(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1267f85(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].force */
extern void T1614f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].fill_with */
extern void T1610f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].extend */
extern void T1610f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_put */
extern void T1267f79(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_put */
extern void T1267f80(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].compress */
extern void T1267f84(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_keep_head */
extern void T1267f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_keep_head */
extern void T1267f88(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].move_all_cursors */
extern void T1267f87(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_inherited_feature */
extern T0* T1275f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_next */
extern void T1303f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.make */
extern T0* T1303c29(GE_context* ac, T0* a1);
/* detachable ET_PARENT_FEATURE.other_seeds */
extern T0* T1302f12(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.first_seed */
extern T6 T1302f11(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.reset */
extern void T1303f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.put_parent_feature */
extern void T1520f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_IDS.cloned_object */
extern T0* T594f4(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.twin */
extern T0* T594f3(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.copy */
extern void T594f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].cloned_array */
extern T0* T877f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] INTEGER_32].twin */
extern T0* T284f9(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].copy */
extern void T284f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].twin */
extern T0* T126f7(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.has_seed */
extern T1 T1520f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.is_other_seeds_shared */
extern T1 T1520f20(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.set_merged_feature */
extern void T1302f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_has_other_effective */
extern void T1302f32(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PARENT_FEATURE.set_has_other_deferred */
extern void T1302f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_INHERITED_FEATURE.put_parent_feature */
extern void T1303f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.has_seed */
extern T1 T1303f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.is_other_seeds_shared */
extern T1 T1303f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].replace_found_item */
extern void T1267f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_redeclared_feature */
extern T0* T1275f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_REDECLARED_FEATURE.set_next */
extern void T1520f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.make */
extern T0* T1520c27(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_REDECLARED_FEATURE.reset */
extern void T1520f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].found_item */
extern T0* T1629f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].item_storage_item */
extern T0* T1629f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].found_key */
extern T0* T1629f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].found */
extern T1 T1629f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].search */
extern void T1629f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_select_name */
extern void T1302f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_redefine_name */
extern void T1302f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].replace_found_item */
extern void T1628f46(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].item_storage_put */
extern void T1628f54(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].found_item */
extern T1 T1628f25(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.set_undefine_name */
extern void T1302f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].found_key */
extern T0* T1628f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].found */
extern T1 T1628f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].search */
extern void T1628f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].search_position */
extern void T1628f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].hash_position */
extern T6 T1628f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CALL_NAME_TESTER.test */
extern T1 T1290f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].slots_item */
extern T6 T1628f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].remove_found_item */
extern void T1627f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].remove_position */
extern void T1627f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_storage_put */
extern void T1627f72(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_storage_put */
extern void T1627f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].force */
extern void T1657f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].fill_with */
extern void T1653f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].extend */
extern void T1653f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_storage_keep_head */
extern void T1627f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_storage_keep_head */
extern void T1627f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].clashes_put */
extern void T1627f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].slots_put */
extern void T1627f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].move_cursors_forth */
extern void T1627f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].search_position */
extern void T1627f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].internal_set_key_equality_tester */
extern void T1627f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_equality_tester */
extern T0* T1627f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].slots_item */
extern T6 T1627f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].hash_position */
extern T6 T1627f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_storage_item */
extern T0* T1627f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].found_item */
extern T0* T1627f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].found */
extern T1 T1627f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].search */
extern void T1627f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_new_name */
extern void T1302f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].remove_found_item */
extern void T1289f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].remove_position */
extern void T1289f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].key_storage_put */
extern void T1289f66(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_storage_put */
extern void T1289f65(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].force */
extern void T1656f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME].fill_with */
extern void T1652f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME].extend */
extern void T1652f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].key_storage_keep_head */
extern void T1289f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_storage_keep_head */
extern void T1289f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].clashes_put */
extern void T1289f61(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].slots_put */
extern void T1289f60(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].move_cursors_forth */
extern void T1289f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].search_position */
extern void T1289f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].internal_set_key_equality_tester */
extern void T1289f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].slots_item */
extern T6 T1289f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].hash_position */
extern T6 T1289f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].key_storage_item */
extern T0* T1289f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].found_item */
extern T0* T1289f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].found */
extern T1 T1289f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.current_system */
extern T0* T1275f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].search */
extern void T1289f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_parent_feature */
extern T0* T1275f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PARENT_FEATURE.set_next */
extern void T1302f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.make */
extern T0* T1302c24(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_PARENT_FEATURE.reset */
extern void T1302f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].resize */
extern void T1267f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1267f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1267f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1267f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1267f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].aliased_resized_area */
extern T0* T1614f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].aliased_resized_area */
extern T0* T1610f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_resize */
extern void T1267f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1267f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_select_table */
extern void T1275f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss2a_error */
extern void T51f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vmss2a */
extern T0* T759c951(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss2_error */
extern T1 T51f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss2a_error */
extern void T35f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss2_error */
extern T1 T35f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].put_new */
extern void T1628f51(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].key_storage_put */
extern void T1628f69(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].slots_put */
extern void T1628f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].clashes_put */
extern void T1628f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.feature_name */
extern T0* T564f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.item */
extern T0* T564f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].resize */
extern void T1628f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].clashes_resize */
extern void T1628f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].special_integer_ */
extern T0* T1628f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].key_storage_resize */
extern void T1628f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].aliased_resized_area */
extern T0* T1657f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].aliased_resized_area */
extern T0* T1653f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].item_storage_resize */
extern void T1628f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].slots_resize */
extern void T1628f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].new_modulus */
extern T6 T1628f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_redefine_table */
extern void T1275f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs3a_error */
extern void T51f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs3a */
extern T0* T759c950(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs3_error */
extern T1 T51f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs3a_error */
extern void T35f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs3_error */
extern T1 T35f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_undefine_table */
extern void T1275f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus4a_error */
extern void T51f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdus4a */
extern T0* T759c949(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus4_error */
extern T1 T51f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus4a_error */
extern void T35f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus4_error */
extern T1 T35f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].is_empty */
extern T1 T1627f32(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_export_table */
extern void T1275f32(GE_context* ac, T0* C, T0* a1);
extern T1 T1275f32ot1(T0* a1, T0** a2);
extern T1 T1275f32ot2(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel3a_error */
extern void T51f338(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vlel3a */
extern T0* T759c948(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel3_error */
extern T1 T51f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel3a_error */
extern void T35f338(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel3_error */
extern T1 T35f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].put_new */
extern void T1627f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.feature_name */
extern T0* T1118f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_EXPORT.item */
extern T0* T1118f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].resize */
extern void T1627f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].clashes_resize */
extern void T1627f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].special_integer_ */
extern T0* T1627f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].key_storage_resize */
extern void T1627f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].item_storage_resize */
extern void T1627f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].slots_resize */
extern void T1627f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].new_modulus */
extern T6 T1627f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel1a_error */
extern void T51f337(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vlel1a */
extern T0* T759c947(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel1_error */
extern T1 T51f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel1a_error */
extern void T35f337(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel1_error */
extern T1 T35f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPORT_LIST.item */
extern T0* T563f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].is_empty */
extern T1 T1289f37(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_rename_table */
extern void T1275f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhrc2a_error */
extern void T51f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vhrc2a */
extern T0* T759c946(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhrc2_error */
extern T1 T51f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhrc2a_error */
extern void T35f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vhrc2_error */
extern T1 T35f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].put_new */
extern void T1289f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_RENAME_LIST.rename_pair */
extern T0* T562f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RENAME_LIST.item */
extern T0* T562f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].resize */
extern void T1289f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].clashes_resize */
extern void T1289f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].special_integer_ */
extern T0* T1289f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].key_storage_resize */
extern void T1289f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].item_storage_resize */
extern void T1289f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].aliased_resized_area */
extern T0* T1656f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_RENAME].aliased_resized_area */
extern T0* T1652f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].slots_resize */
extern void T1289f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].new_modulus */
extern T6 T1289f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.add_current_features */
extern void T1275f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0a_error */
extern void T51f330(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0a */
extern T0* T759c940(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmfn_error */
extern T1 T51f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0a_error */
extern void T35f330(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmfn_error */
extern T1 T35f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.resolve_feature_adaptations */
extern void T1276f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].wipe_out */
extern void T1638f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_wipe_out */
extern void T1638f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_wipe_out */
extern void T1638f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_wipe_out */
extern void T1638f43(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].keep_head */
extern void T1867f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].set_count */
extern void T1867f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_wipe_out */
extern void T1638f42(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].keep_head */
extern void T1870f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].set_count */
extern void T1870f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].unset_found_item */
extern void T1638f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].move_all_cursors_after */
extern void T1638f40(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1873f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_after */
extern void T1873f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].wipe_out */
extern void T1636f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_wipe_out */
extern void T1636f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_wipe_out */
extern void T1636f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_wipe_out */
extern void T1636f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_wipe_out */
extern void T1636f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].unset_found_item */
extern void T1636f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].move_all_cursors_after */
extern void T1636f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1868f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].set_after */
extern void T1868f7(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_any_features */
extern void T1276f23(GE_context* ac, T0* C, T0* a1);
extern T1 T1276f23ot1(T0* a1, T0** a2);
extern T1 T1276f23ot2(T0* a1, T0** a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_inherited_feature */
extern T0* T1276f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RENAME.make */
extern T0* T1295c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.as_keyword */
extern T0* T72f708(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_as */
extern T0* T409c68(GE_context* ac);
/* [detachable] ET_RENAME.tokens */
extern T0* T1295f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].has */
extern T1 T1267f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_parent_feature */
extern T0* T1276f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.current_system */
extern T0* T1276f15(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.current_universe */
extern T0* T1276f13(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features_from_parents */
extern void T1276f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features */
extern void T1276f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.error_handler */
extern T0* T1276f14(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.set_fatal_error */
extern void T1276f25(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].forth */
extern void T1640f15(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_forth */
extern void T1640f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].add_traversing_cursor */
extern void T1640f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1875f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].remove_traversing_cursor */
extern void T1640f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].set */
extern void T1875f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].item_for_iteration */
extern T0* T1640f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_item */
extern T0* T1640f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].after */
extern T1 T1640f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_after */
extern T1 T1640f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].start */
extern void T1640f14(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_start */
extern void T1640f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_off */
extern T1 T1640f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].off */
extern T1 T1875f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].found_item */
extern T0* T1638f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_item */
extern T0* T1638f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].found */
extern T1 T1638f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].search */
extern void T1638f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].search_position */
extern void T1638f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].hash_position */
extern T6 T1638f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_SIGNATURE_TESTER.test */
extern T1 T1635f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_item */
extern T0* T1638f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_item */
extern T6 T1638f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_item */
extern T6 T1638f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_feature */
extern void T1276f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_redeclared_feature */
extern T0* T1276f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].found_item */
extern T0* T1636f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_item */
extern T0* T1636f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].found */
extern T1 T1636f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].search */
extern void T1636f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].search_position */
extern void T1636f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].hash_position */
extern T6 T1636f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_item */
extern T0* T1636f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_item */
extern T6 T1636f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_equality_tester */
extern T0* T1636f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_item */
extern T6 T1636f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_current_features */
extern void T1276f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_current_feature */
extern void T1276f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].force_last */
extern void T1636f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_put */
extern void T1636f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_put */
extern void T1636f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_capacity */
extern T6 T1636f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_put */
extern void T1636f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].force */
extern void T1869f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].fill_with */
extern void T1867f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].extend */
extern void T1867f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].force_last_new */
extern void T1638f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_put */
extern void T1638f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_put */
extern void T1638f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].force */
extern void T1874f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].fill_with */
extern void T1870f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].extend */
extern void T1870f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_put */
extern void T1638f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_put */
extern void T1638f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].resize */
extern void T1638f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_resize */
extern void T1638f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].special_integer_ */
extern T0* T1638f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_resize */
extern void T1638f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].aliased_resized_area */
extern T0* T1869f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].aliased_resized_area */
extern T0* T1867f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_resize */
extern void T1638f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].aliased_resized_area */
extern T0* T1874f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].aliased_resized_area */
extern T0* T1870f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_resize */
extern void T1638f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_modulus */
extern T6 T1638f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_capacity */
extern T6 T1638f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1640c11(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1640f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1640f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1875c7(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].force_last */
extern void T1640f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE].put_right */
extern void T1876f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1876c3(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].force_last_new */
extern void T1267f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_capacity */
extern T6 T1267f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].resize */
extern void T1636f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_resize */
extern void T1636f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].special_integer_ */
extern T0* T1636f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_resize */
extern void T1636f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_resize */
extern void T1636f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_resize */
extern void T1636f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_modulus */
extern T6 T1636f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_parents_validity */
extern void T454f27(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.check_parents_validity */
extern void T1296f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_compilation_status */
extern void T51f204(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_compilation_status */
extern void T35f204(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SYSTEM_PROCESSOR.report_class_postponed */
extern void T62f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_class_postponed */
extern void T371f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_parent_class */
extern void T454f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_checking_implementation */
extern void T73f273(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.interface_checked_successfully */
extern T1 T73f110(GE_context* ac, T0* C);
/* [detachable] MUTEX.try_lock */
extern T1 T133f7(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_trylock */
extern T1 T133f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.set_fatal_error */
extern void T454f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make */
extern T0* T454c19(GE_context* ac, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make_with_feature_checker */
extern void T454f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SUPPLIER_BUILDER.make */
extern T0* T1299c5(GE_context* ac);
/* [detachable] ET_SUPPLIER_BUILDER.dummy_suppliers */
extern T0* T1299f4(GE_context* ac, T0* C);
/* [detachable] ET_SUPPLIER_BUILDER.tokens */
extern T0* T1299f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make */
extern T0* T1298c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].set_internal_cursor */
extern void T1298f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_cursor */
extern T0* T1298f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].make */
extern T0* T1661c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_slots */
extern void T1298f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_modulus */
extern T6 T1298f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_clashes */
extern void T1298f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_key_storage */
extern void T1298f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_item_storage */
extern void T1298f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make */
extern T0* T1297c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].set_internal_cursor */
extern void T1297f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_cursor */
extern T0* T1297f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].make */
extern T0* T1660c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_slots */
extern void T1297f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_modulus */
extern T6 T1297f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_clashes */
extern void T1297f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_key_storage */
extern void T1297f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_item_storage */
extern void T1297f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENT_CHECKER3.make */
extern T0* T1296c6(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.make */
extern void T1296f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.tokens */
extern T0* T1296f4(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.make */
extern T0* T1280c10(GE_context* ac, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.make */
extern void T1280f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.make */
extern T0* T1276c19(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_key_equality_tester */
extern void T1638f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_map */
extern T0* T1638c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_with_equality_testers */
extern void T1638f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_sparse_container */
extern void T1638f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1638f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1638f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make */
extern T0* T1873c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_slots */
extern void T1638f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_clashes */
extern void T1638f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_key_storage */
extern void T1638f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1869f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].make_empty */
extern T0* T1867c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].default_create */
extern T0* T1869c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_item_storage */
extern void T1638f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].make */
extern T0* T1874f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].make_empty */
extern T0* T1870c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].default_create */
extern T0* T1874c3(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].set_equality_tester */
extern void T1636f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_SIGNATURE_TESTER.make */
extern T0* T1635c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1636c31(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1636f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1636f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1868c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_slots */
extern void T1636f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_clashes */
extern void T1636f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_key_storage */
extern void T1636f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_item_storage */
extern void T1636f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.make */
extern void T1276f19p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.tokens */
extern T0* T1276f12(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.make */
extern T0* T1275c25(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T1630c42(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T1630f49(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T1630f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T1630f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T1630f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make */
extern T0* T1858c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_slots */
extern void T1630f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T1630f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T1630f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T1630f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].make */
extern T0* T1861f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].make_empty */
extern T0* T1859c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].default_create */
extern T0* T1861c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].set_key_equality_tester */
extern void T1629f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.alias_name_tester */
extern T0* T1275f17(GE_context* ac, T0* C);
/* [detachable] ET_ALIAS_NAME_TESTER.default_create */
extern T0* T1271c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_map */
extern T0* T1629c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_with_equality_testers */
extern void T1629f42(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_sparse_container */
extern void T1629f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].set_internal_cursor */
extern void T1629f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].new_cursor */
extern T0* T1629f33(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make */
extern T0* T1857c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_slots */
extern void T1629f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_clashes */
extern void T1629f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_key_storage */
extern void T1629f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].make */
extern T0* T1617f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].make_empty */
extern T0* T1618c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].default_create */
extern T0* T1617c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE_NAME, [attached] ET_ALIAS_NAME].make_item_storage */
extern void T1629f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].make */
extern T0* T1615f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].make_empty */
extern T0* T1613c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].default_create */
extern T0* T1615c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].set_key_equality_tester */
extern void T1628f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_map */
extern T0* T1628c43(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_with_equality_testers */
extern void T1628f52(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_sparse_container */
extern void T1628f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].set_internal_cursor */
extern void T1628f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].new_cursor */
extern T0* T1628f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_CALL_NAME].make */
extern T0* T1854c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_slots */
extern void T1628f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_clashes */
extern void T1628f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_key_storage */
extern void T1628f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].make */
extern T0* T1657f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].make_empty */
extern T0* T1653c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].default_create */
extern T0* T1657c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_CALL_NAME].make_item_storage */
extern void T1628f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].set_equality_tester */
extern void T1627f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].make */
extern T0* T1627c37(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].set_internal_cursor */
extern void T1627f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].new_cursor */
extern T0* T1627f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CALL_NAME].make */
extern T0* T1853c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].make_slots */
extern void T1627f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].make_clashes */
extern void T1627f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].make_key_storage */
extern void T1627f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CALL_NAME].make_item_storage */
extern void T1627f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].set_key_equality_tester */
extern void T1289f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.call_name_tester */
extern T0* T1275f19(GE_context* ac, T0* C);
/* [detachable] ET_CALL_NAME_TESTER.default_create */
extern T0* T1290c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_map */
extern T0* T1289c42(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_with_equality_testers */
extern void T1289f52(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_sparse_container */
extern void T1289f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].set_internal_cursor */
extern void T1289f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].new_cursor */
extern T0* T1289f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_CALL_NAME].make */
extern T0* T1651c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_slots */
extern void T1289f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_clashes */
extern void T1289f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_key_storage */
extern void T1289f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].make_item_storage */
extern void T1289f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].make */
extern T0* T1656f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME].make_empty */
extern T0* T1652c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].default_create */
extern T0* T1656c3(GE_context* ac);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.make */
extern void T1275f25p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.tokens */
extern T0* T1275f18(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1267f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.feature_name_tester */
extern T0* T454f14(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_NAME_TESTER.default_create */
extern T0* T1268c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_map */
extern T0* T1267c44(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1267f54(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1267f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1267f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1267f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make */
extern T0* T1609c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_slots */
extern void T1267f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_clashes */
extern void T1267f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1267f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1267f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].make */
extern T0* T1614f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].make_empty */
extern T0* T1610c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].default_create */
extern T0* T1614c3(GE_context* ac);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make_class_processor */
extern void T454f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.tokens */
extern T0* T454f16(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.error_handler */
extern T0* T454f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_none_class */
extern void T454f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.process_class */
extern void T453f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].remove */
extern void T1286f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].remove_position */
extern void T1286f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_put */
extern void T1286f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_put */
extern void T1286f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_keep_head */
extern void T1286f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_keep_head */
extern void T1286f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_put */
extern void T1286f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_put */
extern void T1286f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_item */
extern T6 T1286f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].move_cursors_forth */
extern void T1286f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_position */
extern void T1649f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_next_cursor */
extern void T1649f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_after */
extern void T1649f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].internal_set_key_equality_tester */
extern void T1286f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_equality_tester */
extern T0* T1286f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_item */
extern T6 T1286f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].hash_position */
extern T6 T1286f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_item */
extern T0* T1286f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_item */
extern T0* T1286f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].search_position */
extern void T1286f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].unset_found_item */
extern void T1286f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].last */
extern T0* T1286f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].is_empty */
extern T1 T1286f28(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.internal_process_class */
extern void T453f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_creations_validity */
extern void T453f18(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_creation_validity */
extern void T453f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3b_error */
extern void T51f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3b */
extern T0* T759c846(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc3_error */
extern T1 T51f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc3b_error */
extern void T35f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc3_error */
extern T1 T35f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3a_error */
extern void T51f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3a */
extern T0* T759c845(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vggc3a_error */
extern void T35f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3c_error */
extern void T51f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3c */
extern T0* T759c844(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc3c_error */
extern void T35f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3d_error */
extern void T51f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3d */
extern T0* T759c843(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vggc3d_error */
extern void T35f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3c_error */
extern void T51f215(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3c */
extern T0* T759c842(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp3_error */
extern T1 T51f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3c_error */
extern void T35f215(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp3_error */
extern T1 T35f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.feature_name */
extern T0* T1054f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.item */
extern T0* T1054f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.features_flattened_successfully */
extern T1 T73f219(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renamings_validity */
extern void T453f17(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renaming_validity */
extern void T453f20(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T453f20ot1(T0* a1, T0** a2);
extern T1 T453f20ot2(T0* a1, T0** a2);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_not_renamed_names_validity */
extern void T453f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T453f23ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3g_error */
extern void T51f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3g */
extern T0* T759c866(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav3_error */
extern T1 T51f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav3g_error */
extern void T35f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav3_error */
extern T1 T35f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2g_error */
extern void T51f238(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2g */
extern T0* T759c865(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav2_error */
extern T1 T51f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav2g_error */
extern void T35f238(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav2_error */
extern T1 T35f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2f_error */
extern void T51f237(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2f */
extern T0* T759c864(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc2_error */
extern T1 T51f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc2f_error */
extern void T35f237(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc2_error */
extern T1 T35f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1t_error */
extern void T51f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1t */
extern T0* T759c863(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav1_error */
extern T1 T51f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav1t_error */
extern void T35f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav1_error */
extern T1 T35f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1s_error */
extern void T51f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1s */
extern T0* T759c862(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1s_error */
extern void T35f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].found_key */
extern T0* T1289f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_CALL_NAME].has */
extern T1 T1289f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renamed_names_validity */
extern void T453f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T453f22ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3f_error */
extern void T51f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3f */
extern T0* T759c861(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav3f_error */
extern void T35f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2f_error */
extern void T51f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2f */
extern T0* T759c860(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav2f_error */
extern void T35f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2e_error */
extern void T51f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2e */
extern T0* T759c859(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vggc2e_error */
extern void T35f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1r_error */
extern void T51f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1r */
extern T0* T759c858(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1r_error */
extern void T35f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1q_error */
extern void T51f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1q */
extern T0* T759c857(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1q_error */
extern void T35f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4a_error */
extern void T51f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4a */
extern T0* T759c856(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav4_error */
extern T1 T51f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav4a_error */
extern void T35f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav4_error */
extern T1 T35f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1p_error */
extern void T51f228(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1p */
extern T0* T759c855(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.alias_lower_name */
extern T0* T759f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav1p_error */
extern void T35f228(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1n_error */
extern void T51f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1n */
extern T0* T759c854(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav1n_error */
extern void T35f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1m_error */
extern void T51f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1m */
extern T0* T759c853(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav1m_error */
extern void T35f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3e_error */
extern void T51f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3e */
extern T0* T759c852(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav3e_error */
extern void T35f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2e_error */
extern void T51f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2e */
extern T0* T759c851(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav2e_error */
extern void T35f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1l_error */
extern void T51f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1l */
extern T0* T759c850(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav1l_error */
extern void T35f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1o_error */
extern void T51f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1o */
extern T0* T759c849(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav1o_error */
extern void T35f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2a_error */
extern void T51f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2a */
extern T0* T759c848(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc2a_error */
extern void T35f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2d_error */
extern void T51f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2d */
extern T0* T759c847(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vggc2d_error */
extern void T35f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2c_error */
extern void T51f214(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2c */
extern T0* T759c841(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc2c_error */
extern void T35f214(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2b_error */
extern void T51f213(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2b */
extern T0* T759c840(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc2b_error */
extern void T35f213(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_INTERFACE_CHECKER.resolve_signatures_unfolded_tuple_actual_parameters */
extern void T453f16(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.resolve_signature_unfolded_tuple_actual_parameters */
extern void T453f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_type */
extern void T1288f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_INTERFACE_CHECKER.check_qualified_anchored_signatures_validity */
extern void T453f15(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_signatures_validity */
extern void T1287f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.reset_fatal_error */
extern void T1287f16(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_INTERFACE_CHECKER.set_fatal_error */
extern void T453f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.make */
extern T0* T453c11(GE_context* ac, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.call_name_tester */
extern T0* T453f9(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.make */
extern T0* T1288c6(GE_context* ac, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.make */
extern void T1288f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.tokens */
extern T0* T1288f4(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.set_classes_to_be_processed */
extern void T1287f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.make */
extern T0* T1287c13(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_feature_flattening_error_only */
extern void T400f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.make */
extern void T1287f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.tokens */
extern T0* T1287f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make */
extern T0* T1286c32(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].set_internal_cursor */
extern void T1286f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_cursor */
extern T0* T1286f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].make */
extern T0* T1649c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_slots */
extern void T1286f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_modulus */
extern T6 T1286f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_clashes */
extern void T1286f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_key_storage */
extern void T1286f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_item_storage */
extern void T1286f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INTERFACE_CHECKER.make */
extern void T453f11p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.tokens */
extern T0* T453f8(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.error_handler */
extern T0* T453f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_interface_checked */
extern void T73f270(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.process_class */
extern void T452f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.internal_process_class */
extern void T452f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_features */
extern void T452f36(GE_context* ac, T0* C);
extern T1 T452f36ot1(T0* a1, T0** a2);
extern T1 T452f36ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_FLATTENER.check_kernel_features_validity */
extern void T452f47(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iteration_cursor_forth_seed */
extern void T36f207(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_forth_seed */
extern void T56f273(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.forth_feature_name */
extern T0* T72f276(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iteration_cursor_after_seed */
extern void T36f206(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_after_seed */
extern void T56f272(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.after_feature_name */
extern T0* T72f275(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iteration_cursor_item_seed */
extern void T36f205(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_item_seed */
extern void T56f271(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.is_iteration_cursor_class */
extern T1 T73f167(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iterable_new_cursor_seed */
extern void T36f204(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iterable_new_cursor_seed */
extern void T56f270(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.new_cursor_feature_name */
extern T0* T72f274(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_iterable_class */
extern T1 T73f166(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_dispose_seed */
extern void T36f203(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_dispose_seed */
extern void T56f269(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.dispose_feature_name */
extern T0* T72f273(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_disposable_class */
extern T1 T73f165(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_function_item_seed */
extern void T36f202(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_function_item_seed */
extern void T56f268(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYSTEM.set_routine_call_seed */
extern void T36f201(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_routine_call_seed */
extern void T56f267(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.call_feature_name */
extern T0* T72f272(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_routine_class */
extern T1 T73f164(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe5a_error */
extern void T51f303(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe5a */
extern T0* T759c913(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe5_error */
extern T1 T51f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe5a_error */
extern void T35f303(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe5_error */
extern T1 T35f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_is_equal_seed */
extern void T36f200(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_is_equal_seed */
extern void T56f266(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.is_equal_feature_name */
extern T0* T72f271(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_copy_seed */
extern void T36f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_copy_seed */
extern void T56f265(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.copy_feature_name */
extern T0* T72f270(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_default_create_seed */
extern void T36f198(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_default_create_seed */
extern void T56f264(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.current_system */
extern T0* T452f27(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.tokens */
extern T0* T452f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_convert_validity */
extern void T452f46(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FEATURE_LIST.convert_feature */
extern T0* T1069f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONVERT_FEATURE_LIST.item */
extern T0* T1069f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_creators_validity */
extern void T452f45(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp2a_error */
extern void T51f302(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp2a */
extern T0* T759c912(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp2_error */
extern T1 T51f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp2a_error */
extern void T35f302(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp2_error */
extern T1 T35f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp2b_error */
extern void T51f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp2b */
extern T0* T759c911(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp2b_error */
extern void T35f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6a_error */
extern void T51f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6a */
extern T0* T759c910(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc6_error */
extern T1 T51f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6a_error */
extern void T35f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc6_error */
extern T1 T35f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3b_error */
extern void T51f299(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3b */
extern T0* T759c909(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3b_error */
extern void T35f299(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3a_error */
extern void T51f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3a */
extern T0* T759c908(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3a_error */
extern void T35f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR.feature_name */
extern T0* T1077f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR.item */
extern T0* T1077f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR_LIST.item */
extern T0* T1080f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp1a_error */
extern void T51f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp1a */
extern T0* T759c907(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CREATOR.position */
extern T0* T1077f7(GE_context* ac, T0* C);
/* detachable ET_CREATOR.first */
extern T0* T1077f12(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.position */
extern T0* T1038f13(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT.position */
extern T0* T398f5(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.first */
extern T0* T1038f16(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.is_empty */
extern T1 T1038f15(GE_context* ac, T0* C);
/* detachable ET_CREATOR.clients_clause */
extern T0* T1077f10(GE_context* ac, T0* C);
extern T1 T1077f10ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp1_error */
extern T1 T51f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp1a_error */
extern void T35f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp1_error */
extern T1 T35f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR_LIST.first */
extern T0* T1080f5(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.is_empty */
extern T1 T1080f4(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_assigners_validity */
extern void T452f44(GE_context* ac, T0* C);
extern T1 T452f44ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac4a_error */
extern void T51f296(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfac4a */
extern T0* T759c906(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac4_error */
extern T1 T51f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac4a_error */
extern void T35f296(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac4_error */
extern T1 T35f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac3a_error */
extern void T51f295(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfac3a */
extern T0* T759c905(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac3_error */
extern T1 T51f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac3a_error */
extern void T35f295(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac3_error */
extern T1 T35f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac2a_error */
extern void T51f294(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfac2a */
extern T0* T759c904(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac2_error */
extern T1 T51f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac2a_error */
extern void T35f294(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac2_error */
extern T1 T35f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac1b_error */
extern void T51f293(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfac1b */
extern T0* T759c903(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac1_error */
extern T1 T51f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac1b_error */
extern void T35f293(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac1_error */
extern T1 T35f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac1a_error */
extern void T51f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfac1a */
extern T0* T759c902(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfac1a_error */
extern void T35f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch2a_error */
extern void T51f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.set_ise_reported */
extern void T759f898(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_VALIDITY_ERROR.make_vcch2a */
extern T0* T759c897(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcch2_error */
extern T1 T51f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcch2a_error */
extern void T35f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vcch2_error */
extern T1 T35f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch1a_error */
extern void T51f287(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vcch1a */
extern T0* T759c896(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcch1_error */
extern T1 T51f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcch1a_error */
extern void T35f287(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vcch1_error */
extern T1 T35f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch1b_error */
extern void T51f286(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vcch1b */
extern T0* T759c895(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_INHERITED_FEATURE.name */
extern T0* T1303f22(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vcch1b_error */
extern void T35f286(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].go_after */
extern void T1267f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_go_after */
extern void T1267f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_has_deferred_features */
extern void T73f282(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].wipe_out */
extern void T1270f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_wipe_out */
extern void T1270f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_wipe_out */
extern void T1270f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_wipe_out */
extern void T1270f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_wipe_out */
extern void T1270f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].unset_found_item */
extern void T1270f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].move_all_cursors_after */
extern void T1270f47(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_next_cursor */
extern void T1619f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_after */
extern void T1619f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_signature_validity */
extern void T452f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].force_last_new */
extern void T1270f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_put */
extern void T1270f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_put */
extern void T1270f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_put */
extern void T1270f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_put */
extern void T1270f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_item */
extern T6 T1270f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].hash_position */
extern T6 T1270f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].resize */
extern void T1270f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_resize */
extern void T1270f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].special_integer_ */
extern T0* T1270f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_resize */
extern void T1270f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_resize */
extern void T1270f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_item */
extern T0* T1270f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_item */
extern T6 T1270f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_resize */
extern void T1270f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_modulus */
extern T6 T1270f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_capacity */
extern T6 T1270f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1f_error */
extern void T51f285(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1f */
extern T0* T759c894(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1f_error */
extern void T35f285(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1c_error */
extern void T51f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1c */
extern T0* T759c893(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1c_error */
extern void T35f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3b_error */
extern void T51f283(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3b */
extern T0* T759c892(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav3b_error */
extern void T35f283(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2b_error */
extern void T51f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2b */
extern T0* T759c891(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav2b_error */
extern void T35f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1g_error */
extern void T51f281(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1g */
extern T0* T759c890(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENT_FEATURE.alias_name */
extern T0* T1302f20(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vfav1g_error */
extern void T35f281(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1d_error */
extern void T51f280(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1d */
extern T0* T759c889(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1d_error */
extern void T35f280(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3c_error */
extern void T51f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3c */
extern T0* T759c888(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav3c_error */
extern void T35f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2c_error */
extern void T51f278(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2c */
extern T0* T759c887(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav2c_error */
extern void T35f278(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1h_error */
extern void T51f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1h */
extern T0* T759c886(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1h_error */
extern void T35f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1e_error */
extern void T51f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1e */
extern T0* T759c885(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1e_error */
extern void T35f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3d_error */
extern void T51f275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3d */
extern T0* T759c884(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav3d_error */
extern void T35f275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2d_error */
extern void T51f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2d */
extern T0* T759c883(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav2d_error */
extern void T35f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].found_item */
extern T0* T1270f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_item */
extern T0* T1270f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].found */
extern T1 T1270f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].search */
extern void T1270f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].search_position */
extern void T1270f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_redeclared_signatures_checked */
extern void T73f281(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_anchored_signatures */
extern void T452f42(GE_context* ac, T0* C);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.check_signatures */
extern void T1279f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].wipe_out */
extern void T1641f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].wipe_out */
extern void T676f19(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].keep_head */
extern void T1360f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].set_count */
extern void T1360f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].wipe_out */
extern void T1877f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_wipe_out */
extern void T1877f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_wipe_out */
extern void T1877f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_wipe_out */
extern void T1877f41(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].keep_head */
extern void T1878f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].set_count */
extern void T1878f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_wipe_out */
extern void T1877f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].unset_found_item */
extern void T1877f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].move_all_cursors_after */
extern void T1877f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_next_cursor */
extern void T1939f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_after */
extern void T1939f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].wipe_out */
extern void T1642f12(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].move_all_cursors_after */
extern void T1642f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_next_cursor */
extern void T1880f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_after */
extern void T1880f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].reset */
extern void T1641f14(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat2a_error */
extern void T51f344(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtat2a */
extern T0* T759c954(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].forth */
extern void T1880f14(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_forth */
extern void T1642f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_position */
extern void T1880f15(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].add_traversing_cursor */
extern void T1642f19(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].remove_traversing_cursor */
extern void T1642f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].go_after */
extern void T1880f13(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_go_after */
extern void T1642f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].off */
extern T1 T1880f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].before */
extern T1 T1880f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].item */
extern T0* T1880f6(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_item */
extern T0* T1642f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].after */
extern T1 T1880f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].start */
extern void T1880f12(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_start */
extern void T1642f16(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].new_cursor */
extern T0* T1642f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].make */
extern T0* T1880c9(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtat2_error */
extern T1 T51f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtat2a_error */
extern void T35f344(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtat2_error */
extern T1 T35f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.error_handler */
extern T0* T1279f6(GE_context* ac, T0* C);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.set_fatal_error */
extern void T1279f10(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].is_empty */
extern T1 T1642f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].sort */
extern void T1641f11(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].put_last */
extern void T1642f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].force */
extern void T1879f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].fill_with */
extern void T1878f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].extend */
extern void T1878f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].item */
extern T0* T1642f8(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].make */
extern T0* T1642c10(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].set_internal_cursor */
extern void T1642f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].make */
extern T0* T1879f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].make_empty */
extern T0* T1878c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].default_create */
extern T0* T1879c3(GE_context* ac);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd7b_error */
extern void T51f273(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd7b */
extern T0* T759c882(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vffd7_error */
extern T1 T51f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vffd7b_error */
extern void T35f273(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vffd7_error */
extern T1 T35f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd7a_error */
extern void T51f272(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd7a */
extern T0* T759c881(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vffd7a_error */
extern void T35f272(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1k_error */
extern void T51f271(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1k */
extern T0* T759c880(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav1k_error */
extern void T35f271(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1b_error */
extern void T51f270(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1b */
extern T0* T759c879(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav1b_error */
extern void T35f270(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1a_error */
extern void T51f269(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1a */
extern T0* T759c878(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav1a_error */
extern void T35f269(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3a_error */
extern void T51f268(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3a */
extern T0* T759c877(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav3a_error */
extern void T35f268(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2a_error */
extern void T51f267(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2a */
extern T0* T759c876(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav2a_error */
extern void T35f267(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1j_error */
extern void T51f266(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1j */
extern T0* T759c875(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav1j_error */
extern void T35f266(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1i_error */
extern void T51f265(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1i */
extern T0* T759c874(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vfav1i_error */
extern void T35f265(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd4a_error */
extern void T51f264(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd4a */
extern T0* T759c873(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vffd4_error */
extern T1 T51f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vffd4a_error */
extern void T35f264(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vffd4_error */
extern T1 T35f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.resolve_signature_unfolded_tuple_actual_parameters */
extern void T452f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_signature_vtct_validity */
extern void T452f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_signature_vtct_validity */
extern void T1280f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.resolve_identifier_signature */
extern void T452f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrfa0a_error */
extern void T51f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrfa0a */
extern T0* T759c901(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrfa_error */
extern T1 T51f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrfa0a_error */
extern void T35f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrfa_error */
extern T1 T35f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item */
extern T0* T1267f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vreg0a_error */
extern void T51f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vreg0a */
extern T0* T759c900(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vreg_error */
extern T1 T51f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vreg0a_error */
extern void T35f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vreg_error */
extern T1 T35f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_type */
extern void T1277f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].wipe_out */
extern void T694f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_all_cursors_after */
extern void T694f23(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_procedures */
extern void T73f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROCEDURE_LIST.set_declared_count */
extern void T254f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.put_first */
extern void T254f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].subsort */
extern void T1518f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].subsort_with_comparator */
extern void T1518f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].swap */
extern void T694f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].replace */
extern void T694f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR.less_than */
extern T1 T1710f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.procedure_sorter */
extern T0* T452f26(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].make */
extern T0* T1518c2(GE_context* ac, T0* a1);
/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR.make */
extern T0* T1710c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].wipe_out */
extern void T698f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_all_cursors_after */
extern void T698f23(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_queries */
extern void T73f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUERY_LIST.set_declared_count */
extern void T253f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUERY_LIST.put_first */
extern void T253f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].subsort */
extern void T1517f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].subsort_with_comparator */
extern void T1517f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].swap */
extern void T698f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].replace */
extern void T698f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_SEEDED_QUERY_COMPARATOR.less_than */
extern T1 T1707f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.query_sorter */
extern T0* T452f25(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].make */
extern T0* T1517c2(GE_context* ac, T0* a1);
/* [detachable] ET_SEEDED_QUERY_COMPARATOR.make */
extern T0* T1707c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].back */
extern void T1267f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_back */
extern void T1267f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_before */
extern void T1609f13(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.flatten_feature */
extern void T452f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_inherited_feature */
extern void T452f50(GE_context* ac, T0* C, T0* a1);
extern T1 T452f50ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr2b_error */
extern void T51f310(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr2b */
extern T0* T759c920(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdjr2_error */
extern T1 T51f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdjr2b_error */
extern void T35f310(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdjr2_error */
extern T1 T35f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn2b_error */
extern void T51f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn2b */
extern T0* T759c919(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmfn2_error */
extern T1 T51f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmfn2b_error */
extern void T35f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmfn2_error */
extern T1 T35f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr2a_error */
extern void T51f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr2a */
extern T0* T759c918(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr2a_error */
extern void T35f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn2a_error */
extern void T51f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn2a */
extern T0* T759c917(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn2a_error */
extern void T35f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_FLATTENER.process_replicated_seeds */
extern void T452f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_REDECLARED_FEATURE.set_other_seeds */
extern void T1520f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_other_seeds */
extern void T1303f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.set_first_seed */
extern void T1520f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.set_first_seed */
extern void T1303f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_IDS.replace */
extern void T594f10(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FEATURE_IDS.first */
extern T6 T594f8(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.remove */
extern void T594f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.register_feature */
extern void T73f283(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.extended_name */
extern T0* T1302f21(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.make */
extern T0* T1469c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].wipe_out */
extern void T1285f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_wipe_out */
extern void T1285f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_wipe_out */
extern void T1285f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T1285f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T1285f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T1285f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].move_all_cursors_after */
extern void T1285f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_next_cursor */
extern void T1646f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_after */
extern void T1646f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].forth */
extern void T1285f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_forth */
extern void T1285f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].add_traversing_cursor */
extern void T1285f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_position */
extern void T1646f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove_traversing_cursor */
extern void T1285f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T1285f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_LIST.put_first */
extern void T404f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_LIST.fixed_array */
extern T0* T404f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_for_iteration */
extern T0* T1285f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_item */
extern T0* T1285f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T1285f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].after */
extern T1 T1285f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_after */
extern T1 T1285f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].after */
extern T1 T1646f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].start */
extern void T1285f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_start */
extern void T1285f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_off */
extern T1 T1285f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].off */
extern T1 T1646f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].is_empty */
extern T1 T1285f33(GE_context* ac, T0* C);
/* detachable ET_FEATURE_LIST.make_with_capacity */
extern T0* T404c5(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove */
extern void T1285f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove_position */
extern void T1285f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T1285f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T1285f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_keep_head */
extern void T1285f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_keep_head */
extern void T1285f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T1285f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_put */
extern void T1285f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].move_cursors_forth */
extern void T1285f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].internal_set_key_equality_tester */
extern void T1285f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T1285f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T1285f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T1285f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].search_position */
extern void T1285f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.same_clients */
extern T1 T1038f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.same_clients */
extern T1 T397f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.clients */
extern T0* T1302f18(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_rename */
extern T1 T1302f17(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0c_error */
extern void T51f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0c */
extern T0* T759c916(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0c_error */
extern void T35f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENT_FEATURE.same_version */
extern T1 T1302f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].force_last */
extern void T1285f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].resize */
extern void T1285f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T1285f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T1285f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T1285f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T1285f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T1285f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T1285f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T1285f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_no_redeclaration_validity */
extern void T452f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs4a_error */
extern void T51f321(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs4a */
extern T0* T759c932(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.feature_lower_name */
extern T0* T759f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs4_error */
extern T1 T51f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs4a_error */
extern void T35f321(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs4_error */
extern T1 T35f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.has_redefine */
extern T1 T1302f15(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_redefine_clause_validity */
extern void T452f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs2b_error */
extern void T51f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs2b */
extern T0* T759c937(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs2_error */
extern T1 T51f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs2b_error */
extern void T35f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs2_error */
extern T1 T35f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs2a_error */
extern void T51f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs2a */
extern T0* T759c936(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrs2a_error */
extern void T35f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_FLATTENER.check_undefine_clause_validity */
extern void T452f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus2b_error */
extern void T51f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus2b */
extern T0* T759c935(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus2_error */
extern T1 T51f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus2b_error */
extern void T35f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus2_error */
extern T1 T35f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus2a_error */
extern void T51f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus2a */
extern T0* T759c934(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdus2a_error */
extern void T35f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus3a_error */
extern void T51f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus3a */
extern T0* T759c933(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus3_error */
extern T1 T51f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus3a_error */
extern void T35f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus3_error */
extern T1 T35f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_rename_clause_validity */
extern void T452f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.is_replicated */
extern T1 T1303f20(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.inherited_clients */
extern T0* T452f28(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_FEATURE_FLATTENER.inherited_clients */
extern T0* T452f28ac1(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].wipe_out */
extern void T1273f16(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].keep_head */
extern void T1621f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].set_count */
extern void T1621f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].move_all_cursors_after */
extern void T1273f21(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].set_next_cursor */
extern void T1622f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].set_after */
extern void T1622f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].wipe_out */
extern void T1274f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_wipe_out */
extern void T1274f66(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_wipe_out */
extern void T1274f65(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1274f64(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1274f63(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CLIENT].keep_head */
extern void T1626f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT].set_count */
extern void T1626f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].unset_found_item */
extern void T1274f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].move_all_cursors_after */
extern void T1274f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_next_cursor */
extern void T1624f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_after */
extern void T1624f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].back */
extern void T1274f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_back */
extern void T1274f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].add_traversing_cursor */
extern void T1274f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_position */
extern void T1624f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].remove_traversing_cursor */
extern void T1274f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_before */
extern void T1624f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_item */
extern T6 T1274f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.put_first */
extern void T1038f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.fixed_array */
extern T0* T1038f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_for_iteration */
extern T0* T1274f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_item */
extern T0* T1274f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_item */
extern T0* T1274f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].before */
extern T1 T1274f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_before */
extern T1 T1274f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].before */
extern T1 T1624f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].finish */
extern void T1274f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_finish */
extern void T1274f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_off */
extern T1 T1274f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].off */
extern T1 T1624f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].is_empty */
extern T1 T1274f36(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].for_all */
extern T1 T1797f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].forth */
extern void T1936f8(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_forth */
extern void T1797f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].forth */
extern void T1624f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_forth */
extern void T1274f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].go_after */
extern void T1936f7(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_go_after */
extern void T1797f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].go_after */
extern void T1624f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_go_after */
extern void T1274f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].item */
extern T0* T1936f4(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_item */
extern T0* T1797f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].key */
extern T0* T1624f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_key */
extern T0* T1274f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_item */
extern T0* T1274f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].after */
extern T1 T1936f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].after */
extern T1 T1624f8(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].start */
extern void T1936f6(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_start */
extern void T1797f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].start */
extern void T1624f15(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_start */
extern void T1274f73(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor */
extern T0* T1797f4(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1936c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor */
extern T0* T1274f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1624c10(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].keys */
extern T0* T1274f39(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1797c8(GE_context* ac, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].set_internal_cursor */
extern void T1797f9(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor_with_table_cursor */
extern T0* T1797f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make_with_table_cursor */
extern T0* T1936c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].is_empty */
extern T1 T1273f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].force */
extern void T1274f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_put */
extern void T1274f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_put */
extern void T1274f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_put */
extern void T1274f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_item */
extern T6 T1274f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].hash_position */
extern T6 T1274f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].resize */
extern void T1274f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_resize */
extern void T1274f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].special_integer_ */
extern T0* T1274f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_resize */
extern void T1274f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_resize */
extern void T1274f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].aliased_resized_area */
extern T0* T1625f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT].aliased_resized_area */
extern T0* T1626f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_resize */
extern void T1274f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_modulus */
extern T6 T1274f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_capacity */
extern T6 T1274f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_put */
extern void T1274f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].force */
extern void T1625f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT].fill_with */
extern void T1626f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT].extend */
extern void T1626f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].search_position */
extern void T1274f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.client */
extern T0* T1038f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT.client */
extern T0* T398f4(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.item */
extern T0* T1038f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.client */
extern T0* T397f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.item */
extern T0* T397f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].remove */
extern void T1273f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].move_left */
extern void T1273f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].move_cursors_left */
extern void T1273f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].set_position */
extern void T1622f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].remove_last */
extern void T1273f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].move_last_cursors_after */
extern void T1273f22(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.is_none */
extern T1 T1038f7(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.is_none */
extern T1 T397f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].item */
extern T0* T1273f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].first */
extern T0* T1273f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].force_last */
extern void T1273f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].force */
extern void T1620f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].fill_with */
extern void T1621f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].extend */
extern void T1621f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].resize */
extern void T1273f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].aliased_resized_area */
extern T0* T1620f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].aliased_resized_area */
extern T0* T1621f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].new_capacity */
extern T6 T1273f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].extendible */
extern T1 T1273f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_EXPORT.clients */
extern T0* T1118f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALL_EXPORT.clients */
extern T0* T1455f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.has_feature_name */
extern T1 T1118f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALL_EXPORT.has_feature_name */
extern T1 T1455f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.is_all */
extern T1 T1118f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.flatten_redeclared_feature */
extern void T452f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_precursor_validity */
extern void T452f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.check_feature_validity */
extern void T1284f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd3b_error */
extern void T51f305(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd3b */
extern T0* T759c915(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_POSTCONDITIONS.position */
extern T0* T1213f14(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.first */
extern T0* T1213f16(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.is_empty */
extern T1 T1213f15(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd3_error */
extern T1 T51f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd3b_error */
extern void T35f305(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd3_error */
extern T1 T35f59(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.is_ensure_then */
extern T1 T1213f13(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd3a_error */
extern void T51f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd3a */
extern T0* T759c914(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECONDITIONS.position */
extern T0* T401f13(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.first */
extern T0* T401f15(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.is_empty */
extern T1 T401f14(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vdrd3a_error */
extern void T35f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECONDITIONS.is_require_else */
extern T1 T401f12(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_redeclaration_validity */
extern void T452f53(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd6a_error */
extern void T51f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd6a */
extern T0* T759c931(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd6a_error */
extern void T35f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4b_error */
extern void T51f319(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4b */
extern T0* T759c930(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd4_error */
extern T1 T51f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4b_error */
extern void T35f319(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd4_error */
extern T1 T35f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4c_error */
extern void T51f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4c */
extern T0* T759c929(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4c_error */
extern void T35f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0b_error */
extern void T51f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.set_ge_reported */
extern void T759f928(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0b */
extern T0* T759c927(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0b_error */
extern void T35f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4a_error */
extern void T51f316(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4a */
extern T0* T759c926(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4a_error */
extern void T35f316(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs4b_error */
extern void T51f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs4b */
extern T0* T759c925(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrs4b_error */
extern void T35f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_FLATTENER.ise_19_01_0 */
extern T0* T452f29(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd5a_error */
extern void T51f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd5a */
extern T0* T759c924(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd5_error */
extern T1 T51f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd5a_error */
extern void T35f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd5_error */
extern T1 T35f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7c_error */
extern void T51f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7c */
extern T0* T759c923(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd7_error */
extern T1 T51f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7c_error */
extern void T35f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd7_error */
extern T1 T35f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7a_error */
extern void T51f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7a */
extern T0* T759c922(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7a_error */
extern void T35f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7b_error */
extern void T51f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7b */
extern T0* T759c921(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7b_error */
extern void T35f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REDECLARED_FEATURE.is_replicated */
extern T1 T1520f19(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_builtin_feature_validity */
extern void T452f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_feature_validity */
extern void T1283f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss3a_error */
extern void T51f289(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmss3a */
extern T0* T759c899(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss3_error */
extern T1 T51f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss3a_error */
extern void T35f289(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss3_error */
extern T1 T35f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_immediate_feature */
extern void T452f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].before */
extern T1 T1267f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_before */
extern T1 T1267f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].before */
extern T1 T1609f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].finish */
extern void T1267f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_finish */
extern void T1267f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.resolve_feature_adaptations */
extern void T452f37(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_parents_validity */
extern void T452f35(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.check_parents_validity */
extern void T1281f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_formal_parameters_validity */
extern void T452f34(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_formal_parameters_validity */
extern void T1282f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_constraint_validity */
extern void T1282f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.ancestors_built_successfully */
extern T1 T73f161(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.set_fatal_error */
extern void T452f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.make */
extern T0* T452c30(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T1285c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T1285f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T1285f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T1285f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T1285f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].make */
extern T0* T1646c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_slots */
extern void T1285f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T1285f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T1285f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T1285f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PRECURSOR_CHECKER.make */
extern T0* T1284c10(GE_context* ac, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.dummy_feature */
extern T0* T1284f6(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.current_system */
extern T0* T1284f7(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.make */
extern void T1284f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.tokens */
extern T0* T1284f5(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.make */
extern T0* T1283c12(GE_context* ac, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.make */
extern void T1283f12p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.tokens */
extern T0* T1283f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_map */
extern T0* T1645c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_with_equality_testers */
extern void T1645f34(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_sparse_container */
extern void T1645f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].set_internal_cursor */
extern void T1645f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_cursor */
extern T0* T1645f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make */
extern T0* T1882c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].unset_found_item */
extern void T1645f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_slots */
extern void T1645f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_modulus */
extern T6 T1645f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_clashes */
extern void T1645f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_key_storage */
extern void T1645f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_item_storage */
extern void T1645f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].make */
extern T0* T1883f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].make_empty */
extern T0* T1885c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].default_create */
extern T0* T1883c3(GE_context* ac);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.make */
extern T0* T1282c8(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.make */
extern void T1282f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.tokens */
extern T0* T1282f5(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.make */
extern T0* T1281c9(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.make */
extern void T1281f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.tokens */
extern T0* T1281f5(GE_context* ac, T0* C);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.make */
extern T0* T1279c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].make_default */
extern T0* T1641c10(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].make */
extern void T1641f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_map */
extern T0* T1877c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_with_equality_testers */
extern void T1877f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_sparse_container */
extern void T1877f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_internal_cursor */
extern void T1877f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].new_cursor */
extern T0* T1877f26(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make */
extern T0* T1939c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_slots */
extern void T1877f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].new_modulus */
extern T6 T1877f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_clashes */
extern void T1877f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_key_storage */
extern void T1877f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_item_storage */
extern void T1877f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.make */
extern void T1279f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.tokens */
extern T0* T1279f7(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.make */
extern T0* T1277c7(GE_context* ac, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.tokens */
extern T0* T1277f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_map */
extern T0* T1274c41(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_with_equality_testers */
extern void T1274f42(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_sparse_container */
extern void T1274f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].set_internal_cursor */
extern void T1274f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_slots */
extern void T1274f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_clashes */
extern void T1274f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_key_storage */
extern void T1274f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_item_storage */
extern void T1274f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].make */
extern T0* T1625f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT].make_empty */
extern T0* T1626c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].default_create */
extern T0* T1625c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].make */
extern T0* T1273c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].set_internal_cursor */
extern void T1273f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].new_cursor */
extern T0* T1273f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].make */
extern T0* T1622c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].make */
extern T0* T1620f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].make_empty */
extern T0* T1621c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].default_create */
extern T0* T1620c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_key_equality_tester */
extern void T1270f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.alias_name_tester */
extern T0* T452f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_map */
extern T0* T1270c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_with_equality_testers */
extern void T1270f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_sparse_container */
extern void T1270f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_internal_cursor */
extern void T1270f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_cursor */
extern T0* T1270f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make */
extern T0* T1619c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_slots */
extern void T1270f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_clashes */
extern void T1270f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_key_storage */
extern void T1270f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_item_storage */
extern void T1270f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].make */
extern T0* T694c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].set_internal_cursor */
extern void T694f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].new_cursor */
extern T0* T694f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].make */
extern T0* T1374c9(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].make */
extern T0* T698c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].set_internal_cursor */
extern void T698f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].new_cursor */
extern T0* T698f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].make */
extern T0* T1375c9(GE_context* ac, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.feature_name_tester */
extern T0* T452f22(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.make */
extern void T452f30p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.error_handler */
extern T0* T452f23(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_features_flattened */
extern void T73f278(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.process_class */
extern void T451f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.internal_process_class */
extern void T451f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr1a_error */
extern void T51f205(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr1a */
extern T0* T759c832(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].after */
extern T1 T1306f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].forth */
extern void T1306f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_forth */
extern void T455f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].add_traversing_cursor */
extern void T455f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_traversing_cursor */
extern void T455f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].item */
extern T0* T1306f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_item */
extern T0* T455f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].start */
extern void T1306f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_start */
extern void T455f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].off */
extern T1 T1306f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].before */
extern T1 T1306f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr1_error */
extern T1 T51f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr1a_error */
extern void T35f205(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr1_error */
extern T1 T35f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_parents_inheritance_error */
extern void T451f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].wipe_out */
extern void T1263f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].wipe_out */
extern void T1602f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_wipe_out */
extern void T1602f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_wipe_out */
extern void T1602f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1602f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1602f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].unset_found_item */
extern void T1602f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].move_all_cursors_after */
extern void T1602f41(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].set_next_cursor */
extern void T1848f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].set_after */
extern void T1848f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].wipe_out */
extern void T455f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].reset */
extern void T1263f17(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.check_parents_validity */
extern void T451f20(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER1.check_parents_validity */
extern void T1265f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.check_formal_parameters_validity */
extern void T451f19(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_formal_parameters_validity */
extern void T1266f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].wipe_out */
extern void T1606f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].keep_head */
extern void T1849f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].set_count */
extern void T1849f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].move_all_cursors_after */
extern void T1606f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].set_next_cursor */
extern void T1851f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].set_after */
extern void T1851f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_constraint_base_types */
extern void T1266f28(GE_context* ac, T0* C, T0* a1);
extern T1 T1266f28ot1(T0* a1, T0** a2);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.put_first */
extern void T1608f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].force */
extern void T1850f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].fill_with */
extern void T1849f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].extend */
extern void T1849f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.fixed_array */
extern T0* T1608f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].default_create */
extern T0* T1850c3(GE_context* ac);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern T0* T1608c11(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.make_empty_with_capacity */
extern void T1608f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1850f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].make_empty */
extern T0* T1849c5(GE_context* ac, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.tokens */
extern T0* T1608f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].item */
extern T0* T1606f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_constraint_base_types */
extern void T1450f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].first */
extern T0* T1606f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].force_last */
extern void T1606f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].resize */
extern void T1606f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1850f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1849f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].new_capacity */
extern T6 T1606f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].extendible */
extern T1 T1606f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.base_type_constraint */
extern T0* T1266f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.make */
extern T0* T1794c12(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.tokens */
extern T0* T1266f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_recursive_formal_constraints */
extern void T1266f27(GE_context* ac, T0* C, T0* a1);
extern T1 T1266f27ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_recursive_formal_constraints */
extern void T1450f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].item */
extern T10 T1607f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].replace */
extern void T1607f12(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_n_flags_to_list */
extern void T1266f30(GE_context* ac, T0* C, T10 a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].force_last */
extern void T1607f13(GE_context* ac, T0* C, T10 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].resize */
extern void T1607f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].new_capacity */
extern T6 T1607f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].extendible */
extern T1 T1607f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].wipe_out */
extern void T1607f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].move_all_cursors_after */
extern void T1607f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].set_next_cursor */
extern void T1852f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].set_after */
extern void T1852f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.propagate_recursive_formal_dependencies */
extern void T1266f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_constraint_validity */
extern void T1266f25(GE_context* ac, T0* C, T0* a1);
extern T1 T1266f25ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcfg2a_error */
extern void T51f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vcfg2a */
extern T0* T759c939(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcfg2_error */
extern T1 T51f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcfg2a_error */
extern void T35f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vcfg2_error */
extern T1 T35f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcfg1a_error */
extern void T51f328(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vcfg1a */
extern T0* T759c938(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcfg1_error */
extern T1 T51f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcfg1a_error */
extern void T35f328(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vcfg1_error */
extern T1 T35f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.error_handler */
extern T0* T1266f12(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_fatal_error */
extern void T1266f24(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.current_universe */
extern T0* T1266f15(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.initalize_base_type_dependencies */
extern void T1266f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_n_items_to_list */
extern void T1266f29(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.initalize_formal_dependencies */
extern void T1266f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_ancestors */
extern void T451f18(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_conforming_ancestors */
extern void T73f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_ancestors */
extern void T73f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].wipe_out */
extern void T1264f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_wipe_out */
extern void T1264f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_wipe_out */
extern void T1264f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1264f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1264f51(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].keep_head */
extern void T899f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].set_count */
extern void T899f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].unset_found_item */
extern void T1264f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].move_all_cursors_after */
extern void T1264f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_next_cursor */
extern void T1603f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_after */
extern void T1603f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].forth */
extern void T1264f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_forth */
extern void T1264f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].add_traversing_cursor */
extern void T1264f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_position */
extern void T1603f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].remove_traversing_cursor */
extern void T1264f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_item */
extern T6 T1264f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_LIST.put_last */
extern void T408f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].force */
extern void T900f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].fill_with */
extern void T899f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].extend */
extern void T899f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_for_iteration */
extern T0* T1264f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_item */
extern T0* T1264f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_item */
extern T0* T1264f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].after */
extern T1 T1264f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_after */
extern T1 T1264f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].after */
extern T1 T1603f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].start */
extern void T1264f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_start */
extern void T1264f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_off */
extern T1 T1264f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].off */
extern T1 T1603f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].is_empty */
extern T1 T1264f24(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_parent_to_ancestors */
extern void T451f23(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].force_new */
extern void T1264f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_put */
extern void T1264f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_put */
extern void T1264f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_put */
extern void T1264f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_put */
extern void T1264f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_item */
extern T6 T1264f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].hash_position */
extern T6 T1264f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].resize */
extern void T1264f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_resize */
extern void T1264f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].special_integer_ */
extern T0* T1264f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_resize */
extern void T1264f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_resize */
extern void T1264f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].aliased_resized_area */
extern T0* T900f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].aliased_resized_area */
extern T0* T899f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_item */
extern T0* T1264f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_resize */
extern void T1264f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_modulus */
extern T6 T1264f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_capacity */
extern T6 T1264f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvagp0a_error */
extern void T51f212(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvagp0a */
extern T0* T759c839(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvagp_error */
extern T1 T51f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvagp0a_error */
extern void T35f212(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvagp_error */
extern T1 T35f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].found_item */
extern T0* T1264f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].found */
extern T1 T1264f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].search */
extern void T1264f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].search_position */
extern void T1264f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr2a_error */
extern void T51f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr2a */
extern T0* T759c838(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr2_error */
extern T1 T51f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr2a_error */
extern void T35f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr2_error */
extern T1 T35f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_frozen */
extern T1 T73f217(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_frozen_mark */
extern T1 T73f224(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhso2a_error */
extern void T51f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_gvhso2a */
extern T0* T759c837(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhso2_error */
extern T1 T51f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhso2a_error */
extern void T35f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhso2_error */
extern T1 T35f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtct0a_error */
extern void T51f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtct0a */
extern T0* T759c836(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtct_error */
extern T1 T51f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtct0a_error */
extern void T35f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtct_error */
extern T1 T35f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhso1a_error */
extern void T51f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_gvhso1a */
extern T0* T759c835(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhso1_error */
extern T1 T51f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhso1a_error */
extern void T35f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhso1_error */
extern T1 T35f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhay0a_error */
extern void T51f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_vhay0a */
extern T0* T759c834(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhay_error */
extern T1 T51f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhay0a_error */
extern void T35f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vhay_error */
extern T1 T35f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr1b_error */
extern void T51f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr1b */
extern T0* T759c833(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vhpr1b_error */
extern void T35f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENT_LIST.is_conforming */
extern T1 T558f7(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.current_system */
extern T0* T451f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].sort */
extern void T1263f12(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_class_to_sorter */
extern void T451f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.current_universe */
extern T0* T451f11(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_parents_to_sorter */
extern void T451f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].force_relation */
extern void T1263f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].put_indexed_relation */
extern void T1263f19(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].count */
extern T6 T1263f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].index_of */
extern T6 T1263f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].found_item */
extern T6 T1602f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_item */
extern T6 T1602f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].found */
extern T1 T1602f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].search */
extern void T1602f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].search_position */
extern void T1602f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].hash_position */
extern T6 T1602f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_item */
extern T0* T1602f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_item */
extern T6 T1602f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_item */
extern T6 T1602f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].force */
extern void T1263f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].put */
extern void T1263f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].put_new */
extern void T1602f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_put */
extern void T1602f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_put */
extern void T1602f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_put */
extern void T1602f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_put */
extern void T1602f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].resize */
extern void T1602f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_resize */
extern void T1602f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].special_integer_ */
extern T0* T1602f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_resize */
extern void T1602f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_resize */
extern void T1602f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_resize */
extern void T1602f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].new_modulus */
extern T6 T1602f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].capacity */
extern T6 T1263f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].has */
extern T1 T1263f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].has */
extern T1 T1602f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_fatal_error */
extern void T451f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.make */
extern T0* T451c13(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.make */
extern T0* T1266c20(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].make */
extern T0* T1607c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].set_internal_cursor */
extern void T1607f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].new_cursor */
extern T0* T1607f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].make */
extern T0* T1852c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1606c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].set_internal_cursor */
extern void T1606f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].new_cursor */
extern T0* T1606f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1851c5(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.make */
extern void T1266f20p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.make */
extern T0* T1265c8(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.tokens */
extern T0* T1265f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_map */
extern T0* T1264c40(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_with_equality_testers */
extern void T1264f46(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_sparse_container */
extern void T1264f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_internal_cursor */
extern void T1264f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_cursor */
extern T0* T1264f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make */
extern T0* T1603c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_slots */
extern void T1264f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_clashes */
extern void T1264f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_key_storage */
extern void T1264f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_item_storage */
extern void T1264f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].make_default */
extern T0* T1263c11(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].make */
extern void T1263f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_map */
extern T0* T1602c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_with_equality_testers */
extern void T1602f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_sparse_container */
extern void T1602f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].set_internal_cursor */
extern void T1602f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].new_cursor */
extern T0* T1602f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].make */
extern T0* T1848c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_slots */
extern void T1602f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_clashes */
extern void T1602f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_key_storage */
extern void T1602f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_item_storage */
extern void T1602f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCESTOR_BUILDER.make */
extern void T451f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.tokens */
extern T0* T451f9(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.error_handler */
extern T0* T451f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_ancestors_built */
extern void T73f267(GE_context* ac, T0* C);
/* [detachable] ET_PROVIDER_CHECKER.process_class */
extern void T450f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.internal_process_class */
extern void T450f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.check_cluster_dependence_constraints */
extern void T450f10(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcdep_error */
extern void T51f396(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLUSTER_ERROR.make_gcdep */
extern T0* T1887c25(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].forth */
extern void T1987f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_forth */
extern void T1986f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].add_traversing_cursor */
extern void T1986f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_next_cursor */
extern void T1987f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].remove_traversing_cursor */
extern void T1986f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_after */
extern void T1987f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].key */
extern T0* T1987f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_key */
extern T0* T1986f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].after */
extern T1 T1987f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].start */
extern void T1987f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_start */
extern void T1986f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_off */
extern T1 T1986f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].off */
extern T1 T1987f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].is_empty */
extern T1 T1986f29(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcdep_error */
extern T1 T51f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcdep_error */
extern void T35f396(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gcdep_error */
extern T1 T35f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcpro_error */
extern void T51f395(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLUSTER_ERROR.make_gcpro */
extern T0* T1887c24(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcpro_error */
extern T1 T51f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcpro_error */
extern void T35f395(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gcpro_error */
extern T1 T35f109(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T1 T1706f8(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1706f8ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* Creation of agent #2 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1706f8ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1706f8ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1706f8ac4(GE_context* ac, T0* a1, T0* a2);
extern T1 T1706f8ot1(T0* a1, T0** a2);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]].set_rout_disp_final */
extern void T2009f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]].set_rout_disp_final */
extern void T2020f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.matches_pathname */
extern T1 T1706f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] LX_DFA_WILDCARD.recognizes */
extern T1 T2014f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.longest_end_position */
extern T6 T2014f16(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] LX_DFA_WILDCARD.set_subject */
extern void T2014f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.is_compiled */
extern T1 T2014f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].there_exists */
extern T1 T143f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.string_ */
extern T0* T1706f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].there_exists */
extern T1 T1986f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.group_pathname */
extern T0* T1706f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.first_non_override_overridden_class */
extern T0* T73f195(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.master_class_in_universe */
extern T0* T73f201(GE_context* ac, T0* C);
/* [detachable] ET_PROVIDER_CHECKER.scm_mapped_cluster */
extern T0* T450f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.scm_write_mapping_recursive */
extern T0* T776f57(GE_context* ac, T0* C);
/* [detachable] ET_PROVIDER_CHECKER.set_fatal_error */
extern void T450f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.make */
extern T0* T450c6(GE_context* ac, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.tokens */
extern T0* T450f3(GE_context* ac, T0* C);
/* [detachable] ET_PROVIDER_CHECKER.error_handler */
extern T0* T450f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_class */
extern void T449f1006(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.internal_process_class */
extern void T449f1168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_class */
extern void T460f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvscn1b_error */
extern void T51f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvscn1b */
extern T0* T759c871(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvscn1_error */
extern T1 T51f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvscn1b_error */
extern void T35f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvscn1_error */
extern T1 T35f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.syntax_error */
extern T1 T449f717(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.parse_file */
extern void T449f1171(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ET_CLASS.set_has_utf8_bom */
extern void T73f303(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_EIFFEL_PARSER.yyparse */
extern void T449f1172(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_pop_last_value */
extern void T449f1182(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_push_error_value */
extern void T449f1181(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action */
extern void T449f1180(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_801_1000 */
extern void T449f1207(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_944 */
extern void T449f2167(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_counter */
extern void T449f2175(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].force */
extern void T915f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].resize */
extern void T915f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].new_capacity */
extern T6 T915f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].extendible */
extern T1 T915f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_943 */
extern void T449f2166(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].force */
extern void T953f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].fill_with */
extern void T954f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].extend */
extern void T954f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_942 */
extern void T449f2165(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_941 */
extern void T449f2164(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_940 */
extern void T449f2163(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].force */
extern void T963f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].fill_with */
extern void T965f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].extend */
extern void T965f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_cast_type */
extern void T1440f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_cast_type */
extern void T1439f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_target_type */
extern T0* T458f267(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.ast_factory */
extern T0* T449f731(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_939 */
extern void T449f2162(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_sign */
extern void T1440f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_sign */
extern void T1439f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_938 */
extern void T449f2161(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_937 */
extern void T449f2160(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_936 */
extern void T449f2159(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_935 */
extern void T449f2158(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_934 */
extern void T449f2157(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_933 */
extern void T449f2156(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].force */
extern void T955f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].fill_with */
extern void T957f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].extend */
extern void T957f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_932 */
extern void T449f2155(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_931 */
extern void T449f2154(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_930 */
extern void T449f2153(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_929 */
extern void T449f2152(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_928 */
extern void T449f2151(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_927 */
extern void T449f2150(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_926 */
extern void T449f2149(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].force */
extern void T939f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].fill_with */
extern void T941f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].extend */
extern void T941f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_925 */
extern void T449f2148(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_924 */
extern void T449f2147(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].force */
extern void T945f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].fill_with */
extern void T947f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].extend */
extern void T947f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_923 */
extern void T449f2146(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].aliased_resized_area */
extern T0* T945f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].aliased_resized_area */
extern T0* T947f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_922 */
extern void T449f2145(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_921 */
extern void T449f2144(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_920 */
extern void T449f2143(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_919 */
extern void T449f2142(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].force */
extern void T961f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].fill_with */
extern void T962f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].extend */
extern void T962f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_918 */
extern void T449f2141(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].aliased_resized_area */
extern T0* T961f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].aliased_resized_area */
extern T0* T962f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_917 */
extern void T449f2140(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_916 */
extern void T449f2139(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_915 */
extern void T449f2138(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_914 */
extern void T449f2137(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_913 */
extern void T449f2136(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_912 */
extern void T449f2135(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_911 */
extern void T449f2134(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_910 */
extern void T449f2133(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_909 */
extern void T449f2132(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_908 */
extern void T449f2131(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_907 */
extern void T449f2130(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_906 */
extern void T449f2129(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_905 */
extern void T449f2128(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_904 */
extern void T449f2127(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_903 */
extern void T449f2126(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_902 */
extern void T449f2125(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_901 */
extern void T449f2124(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_900 */
extern void T449f2123(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_899 */
extern void T449f2122(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_898 */
extern void T449f2121(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_897 */
extern void T449f2120(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_896 */
extern void T449f2119(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_895 */
extern void T449f2118(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_894 */
extern void T449f2117(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_893 */
extern void T449f2116(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_892 */
extern void T449f2115(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].force */
extern void T997f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].fill_with */
extern void T998f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].extend */
extern void T998f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].aliased_resized_area */
extern T0* T997f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].aliased_resized_area */
extern T0* T998f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_typed_open_argument */
extern T0* T458f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.make */
extern T0* T1510c11(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.question_mark_symbol */
extern T0* T72f1517(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.make */
extern T0* T983c42(GE_context* ac);
/* detachable ET_QUESTION_MARK_SYMBOL.make_leaf */
extern void T983f45(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.tokens */
extern T0* T1510f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_891 */
extern void T449f2114(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_890 */
extern void T449f2113(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_889 */
extern void T449f2112(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].force */
extern void T999f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].fill_with */
extern void T1001f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].extend */
extern void T1001f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].aliased_resized_area */
extern T0* T999f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].aliased_resized_area */
extern T0* T1001f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.increment_counter */
extern void T449f2176(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].replace */
extern void T915f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_argument_operand_comma */
extern T0* T458f328(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_888 */
extern void T449f2111(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].force */
extern void T1002f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].fill_with */
extern void T1004f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].extend */
extern void T1004f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.put_first */
extern void T1003f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].force */
extern void T1759f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].fill_with */
extern void T1758f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].extend */
extern void T1758f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.fixed_array */
extern T0* T1003f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].default_create */
extern T0* T1759c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_887 */
extern void T449f2110(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].aliased_resized_area */
extern T0* T1002f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].aliased_resized_area */
extern T0* T1004f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_argument_operands */
extern T0* T458f327(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.make_with_capacity */
extern T0* T1003c12(GE_context* ac, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.make_with_capacity */
extern void T1003f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].make */
extern T0* T1759f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].make_empty */
extern T0* T1758c4(GE_context* ac, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.tokens */
extern T0* T1003f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.counter_value */
extern T6 T449f905(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_symbol */
extern T0* T449f907(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].item */
extern T0* T919f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_886 */
extern void T449f2109(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_885 */
extern void T449f2108(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_symbol */
extern void T449f2179(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].force */
extern void T919f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].force */
extern void T935f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL].fill_with */
extern void T936f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL].extend */
extern void T936f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].resize */
extern void T919f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].aliased_resized_area */
extern T0* T935f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SYMBOL].aliased_resized_area */
extern T0* T936f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].new_capacity */
extern T6 T919f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].extendible */
extern T1 T919f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_884 */
extern void T449f2107(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_counter */
extern void T449f2173(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_symbol */
extern void T449f2178(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].remove */
extern void T919f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_SYMBOL].keep_head */
extern void T936f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL].set_count */
extern void T936f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_883 */
extern void T449f2106(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_882 */
extern void T449f2105(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_881 */
extern void T449f2104(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].force */
extern void T1005f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].fill_with */
extern void T1006f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].extend */
extern void T1006f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].aliased_resized_area */
extern T0* T1005f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].aliased_resized_area */
extern T0* T1006f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_open_target */
extern T0* T458f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_AGENT_OPEN_TARGET.make */
extern T0* T1509c9(GE_context* ac, T0* a1);
/* detachable ET_AGENT_OPEN_TARGET.tokens */
extern T0* T1509f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_880 */
extern void T449f2103(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_879 */
extern void T449f2102(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_878 */
extern void T449f2101(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_877 */
extern void T449f2100(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_agent_identifier_target */
extern T0* T449f953(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_unqualified_call_expression */
extern T0* T458f351(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.set_object_test_local */
extern void T180f125(GE_context* ac, T0* C, T1 a1);
/* detachable ET_OBJECT_TEST_LIST.index_of_name */
extern T6 T1431f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.set_across_cursor */
extern void T180f122(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ACROSS_COMPONENT_LIST.index_of_name */
extern T6 T361f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_used */
extern void T1716f11(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOCAL_VARIABLE.set_used */
extern void T1176f10(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.local_variable */
extern T0* T331f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.item */
extern T0* T331f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_local */
extern void T180f124(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.index_of */
extern T6 T331f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_used */
extern void T1715f13(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_used */
extern void T252f12(GE_context* ac, T0* C, T1 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.index_of */
extern T6 T249f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_876 */
extern void T449f2099(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_start_closure */
extern void T449f2187(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].force */
extern void T922f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST].force */
extern void T1591f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].fill_with */
extern void T1590f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].extend */
extern void T1590f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].resize */
extern void T922f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST].aliased_resized_area */
extern T0* T1591f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].aliased_resized_area */
extern T0* T1590f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].new_capacity */
extern T6 T922f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].extendible */
extern T1 T922f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].force */
extern void T920f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].force */
extern void T1587f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].fill_with */
extern void T1586f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].extend */
extern void T1586f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].resize */
extern void T920f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1587f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1586f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].new_capacity */
extern T6 T920f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].extendible */
extern T1 T920f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].force */
extern void T917f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].force */
extern void T1181f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].fill_with */
extern void T1182f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].extend */
extern void T1182f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].resize */
extern void T917f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].aliased_resized_area */
extern T0* T1181f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].aliased_resized_area */
extern T0* T1182f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].new_capacity */
extern T6 T917f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].extendible */
extern T1 T917f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].force */
extern void T916f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].force */
extern void T1130f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].fill_with */
extern void T1131f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].extend */
extern void T1131f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].resize */
extern void T916f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1130f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1131f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].new_capacity */
extern T6 T916f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].extendible */
extern T1 T916f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_875 */
extern void T449f2098(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_874 */
extern void T449f2097(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].force */
extern void T1159f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].fill_with */
extern void T1160f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].extend */
extern void T1160f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].aliased_resized_area */
extern T0* T1159f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].aliased_resized_area */
extern T0* T1160f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.register_inline_agent */
extern void T449f2198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.set_end_closure */
extern void T449f2202(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].remove */
extern void T922f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].keep_head */
extern void T1590f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].set_count */
extern void T1590f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].item */
extern T0* T922f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].is_empty */
extern T1 T922f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].force */
extern void T923f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST].force */
extern void T1592f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].fill_with */
extern void T1593f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].extend */
extern void T1593f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].resize */
extern void T923f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST].aliased_resized_area */
extern T0* T1592f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].aliased_resized_area */
extern T0* T1593f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].new_capacity */
extern T6 T923f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].extendible */
extern T1 T923f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_COMPONENT_LIST.wipe_out */
extern void T361f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT].keep_head */
extern void T722f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_ACROSS_COMPONENT_LIST.fixed_array */
extern T0* T361f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].remove */
extern void T920f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].keep_head */
extern void T1586f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].set_count */
extern void T1586f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].item */
extern T0* T920f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].is_empty */
extern T1 T920f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].force */
extern void T921f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].force */
extern void T1588f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].fill_with */
extern void T1589f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].extend */
extern void T1589f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].resize */
extern void T921f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1588f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1589f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].new_capacity */
extern T6 T921f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].extendible */
extern T1 T921f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.wipe_out */
extern void T1431f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].keep_head */
extern void T717f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_OBJECT_TEST_LIST.fixed_array */
extern T0* T1431f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].remove */
extern void T917f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].keep_head */
extern void T1182f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].set_count */
extern void T1182f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].item */
extern T0* T917f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].is_empty */
extern T1 T917f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].remove */
extern void T916f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].keep_head */
extern void T1131f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].set_count */
extern void T1131f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].item */
extern T0* T916f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].is_empty */
extern T1 T916f5(GE_context* ac, T0* C);
/* detachable ET_ACROSS_COMPONENT_LIST.cloned_across_component_list */
extern T0* T361f3(GE_context* ac, T0* C);
/* detachable ET_ACROSS_COMPONENT_LIST.put_last */
extern void T361f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT].force */
extern void T722f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].fill_with */
extern void T721f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].extend */
extern void T721f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_COMPONENT_LIST.make_with_capacity */
extern T0* T361c9(GE_context* ac, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.cloned_object_test_list */
extern T0* T1431f3(GE_context* ac, T0* C);
/* detachable ET_OBJECT_TEST_LIST.put_last */
extern void T1431f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].force */
extern void T717f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].fill_with */
extern void T716f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].extend */
extern void T716f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OBJECT_TEST_LIST.make_with_capacity */
extern T0* T1431c9(GE_context* ac, T6 a1);
/* [detachable] ET_AST_FACTORY.new_external_procedure_inline_agent */
extern T0* T458f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1508f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_alias_clause */
extern void T1508f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1508f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1508f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make */
extern T0* T1508c38(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make_procedure */
extern void T1508f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1508f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.make */
extern T0* T1760c6(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.agent_keyword */
extern T0* T72f889(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.make */
extern T0* T930c13(GE_context* ac);
/* detachable ET_AGENT_KEYWORD.make_token */
extern void T930f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_KEYWORD.make_leaf */
extern void T930f16(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.tokens */
extern T0* T930f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1508f16(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_external_language */
extern T0* T458f179(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_873 */
extern void T449f2096(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_872 */
extern void T449f2095(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_procedure_inline_agent */
extern T0* T458f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_rescue_clause */
extern void T1507f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1507f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_compound */
extern void T1507f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_keys */
extern void T1507f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_locals */
extern void T1507f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1507f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1507f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.make */
extern T0* T1507c44(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1507f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1507f19(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_compound */
extern T0* T458f176(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_COMPOUND.set_keyword */
extern void T330f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.once_keyword */
extern T0* T72f743(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_once */
extern T0* T409c55(GE_context* ac);
/* [detachable] ET_AST_FACTORY.tokens */
extern T0* T458f361(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_871 */
extern void T449f2094(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_870 */
extern void T449f2093(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_procedure_inline_agent */
extern T0* T458f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_rescue_clause */
extern void T1506f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1506f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_compound */
extern void T1506f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_locals */
extern void T1506f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1506f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1506f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.make */
extern T0* T1506c40(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1506f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1506f35(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_869 */
extern void T449f2092(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_868 */
extern void T449f2091(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_external_function_inline_agent */
extern T0* T458f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1505f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_alias_clause */
extern void T1505f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1505f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1505f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make */
extern T0* T1505c39(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make_function */
extern void T1505f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1505f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RESULT.set_position */
extern void T301f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1505f17(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_colon_type */
extern T0* T458f170(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_867 */
extern void T449f2090(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_866 */
extern void T449f2089(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_function_inline_agent */
extern T0* T458f321(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_rescue_clause */
extern void T1504f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1504f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_compound */
extern void T1504f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_keys */
extern void T1504f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_locals */
extern void T1504f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1504f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1504f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.make */
extern T0* T1504c45(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1504f57(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1504f20(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_865 */
extern void T449f2088(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_864 */
extern void T449f2087(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_function_inline_agent */
extern T0* T458f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_rescue_clause */
extern void T1503f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1503f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_compound */
extern void T1503f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_locals */
extern void T1503f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1503f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1503f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.make */
extern T0* T1503c41(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1503f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1503f20(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_863 */
extern void T449f2086(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_862 */
extern void T449f2085(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_inline_agent_actual_arguments */
extern void T449f2197(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_861 */
extern void T449f2084(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].force */
extern void T1015f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].fill_with */
extern void T1017f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].extend */
extern void T1017f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].aliased_resized_area */
extern T0* T1015f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].aliased_resized_area */
extern T0* T1017f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_call_agent */
extern T0* T458f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CALL_AGENT.set_agent_keyword */
extern void T1016f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CALL_AGENT.make */
extern T0* T1016c32(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CALL_AGENT.tokens */
extern T0* T1016f10(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.make_unqualified */
extern T0* T1016c30(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_dot_feature_name */
extern T0* T458f319(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_860 */
extern void T449f2083(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_859 */
extern void T449f2082(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].force */
extern void T985f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].fill_with */
extern void T987f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].extend */
extern void T987f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].aliased_resized_area */
extern T0* T985f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].aliased_resized_area */
extern T0* T987f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_across_expression_header */
extern T0* T449f952(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ACROSS_COMPONENT_LIST.force_last */
extern void T361f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_COMPONENT_LIST.resize */
extern void T361f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT].aliased_resized_area */
extern T0* T722f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT].aliased_resized_area */
extern T0* T721f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_COMPONENT_LIST.new_capacity */
extern T6 T361f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_COMPONENT_LIST.capacity */
extern T6 T361f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_component_list */
extern T0* T449f971(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].remove */
extern void T923f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].keep_head */
extern void T1593f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].set_count */
extern void T1593f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].item */
extern T0* T923f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].is_empty */
extern T1 T923f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_across_all_expression */
extern T0* T458f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_ACROSS_EXPRESSION.set_as_keyword */
extern void T986f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_variant_part */
extern void T986f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_invariant_part */
extern void T986f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.make_all */
extern T0* T986c36(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ACROSS_EXPRESSION.create_unfolded_form */
extern void T986f46(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.across_keyword */
extern T0* T72f888(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_across */
extern T0* T409c61(GE_context* ac);
/* detachable ET_ACROSS_EXPRESSION.tokens */
extern T0* T986f20(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.true_keyword */
extern T0* T72f294(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.make */
extern T0* T1442c39(GE_context* ac);
/* detachable ET_TRUE_CONSTANT.make_token */
extern void T1442f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.make_leaf */
extern void T1442f43(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.tokens */
extern T0* T1442f12(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_858 */
extern void T449f2081(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_857 */
extern void T449f2080(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_all_expression */
extern T0* T449f951(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_ACROSS_EXPRESSION.set_end_keyword */
extern void T986f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_all */
extern void T986f44(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ACROSS_EXPRESSION.set_iteration_conditional */
extern void T986f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_until_conditional */
extern void T986f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_conditional */
extern T0* T458f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_KEYWORD_EXPRESSION.make */
extern T0* T1719c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_KEYWORD.is_inspect */
extern T1 T409f29(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_if */
extern T1 T409f28(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_856 */
extern void T449f2079(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_some_expression */
extern T0* T449f950(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_ACROSS_EXPRESSION.set_some */
extern void T986f40(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_855 */
extern void T449f2078(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].force */
extern void T1045f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTANT].fill_with */
extern void T1046f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTANT].extend */
extern void T1046f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].aliased_resized_area */
extern T0* T1045f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTANT].aliased_resized_area */
extern T0* T1046f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_854 */
extern void T449f2077(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_853 */
extern void T449f2076(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_852 */
extern void T449f2075(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_851 */
extern void T449f2074(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_850 */
extern void T449f2073(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].force */
extern void T1232f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].fill_with */
extern void T1234f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].extend */
extern void T1234f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.put_first */
extern void T1233f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].force */
extern void T1731f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].fill_with */
extern void T1730f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].extend */
extern void T1730f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.fixed_array */
extern T0* T1233f14(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].default_create */
extern T0* T1731c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_849 */
extern void T449f2072(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].aliased_resized_area */
extern T0* T1232f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].aliased_resized_area */
extern T0* T1234f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_strip_expression */
extern T0* T458f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* detachable ET_STRIP_EXPRESSION.set_strip_keyword */
extern void T1233f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.make_with_capacity */
extern T0* T1233c28(GE_context* ac, T6 a1);
/* detachable ET_STRIP_EXPRESSION.make_with_capacity */
extern void T1233f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].make */
extern T0* T1731f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].make_empty */
extern T0* T1730c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.strip_keyword */
extern T0* T72f894(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_strip */
extern T0* T409c60(GE_context* ac);
/* detachable ET_STRIP_EXPRESSION.tokens */
extern T0* T1233f12(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_keyword */
extern T0* T449f904(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].item */
extern T0* T918f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_848 */
extern void T449f2071(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_847 */
extern void T449f2070(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_keyword */
extern void T449f2174(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].force */
extern void T918f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].force */
extern void T927f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD].fill_with */
extern void T928f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD].extend */
extern void T928f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].resize */
extern void T918f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].aliased_resized_area */
extern T0* T927f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD].aliased_resized_area */
extern T0* T928f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].new_capacity */
extern T6 T918f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].extendible */
extern T1 T918f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_846 */
extern void T449f2069(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_keyword */
extern void T449f2172(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].remove */
extern void T918f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_KEYWORD].keep_head */
extern void T928f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD].set_count */
extern void T928f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_845 */
extern void T449f2068(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_844 */
extern void T449f2067(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].force */
extern void T1195f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].fill_with */
extern void T1196f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].extend */
extern void T1196f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_843 */
extern void T449f2066(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].aliased_resized_area */
extern T0* T1195f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].aliased_resized_area */
extern T0* T1196f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_tuple */
extern T0* T458f316(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_MANIFEST_TUPLE.set_left_symbol */
extern void T281f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_842 */
extern void T449f2065(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_841 */
extern void T449f2064(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_840 */
extern void T449f2063(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_839 */
extern void T449f2062(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_838 */
extern void T449f2061(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].force */
extern void T1186f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].fill_with */
extern void T1188f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].extend */
extern void T1188f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.put_first */
extern void T1187f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.fixed_array */
extern T0* T1187f14(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_837 */
extern void T449f2060(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].aliased_resized_area */
extern T0* T1186f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].aliased_resized_area */
extern T0* T1188f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_array */
extern T0* T458f315(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_MANIFEST_ARRAY.set_left_symbol */
extern void T1187f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.make_with_capacity */
extern T0* T1187c28(GE_context* ac, T6 a1);
/* detachable ET_MANIFEST_ARRAY.make_with_capacity */
extern void T1187f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_array_symbol */
extern T0* T72f703(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_array */
extern T0* T343c96(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_array_symbol */
extern T0* T72f892(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_array */
extern T0* T343c89(GE_context* ac);
/* detachable ET_MANIFEST_ARRAY.tokens */
extern T0* T1187f12(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_836 */
extern void T449f2059(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_835 */
extern void T449f2058(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.set_cast_type */
extern void T1187f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_834 */
extern void T449f2057(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_833 */
extern void T449f2056(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_832 */
extern void T449f2055(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_831 */
extern void T449f2054(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].force */
extern void T1200f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].fill_with */
extern void T1202f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].extend */
extern void T1202f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].aliased_resized_area */
extern T0* T1200f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].aliased_resized_area */
extern T0* T1202f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parenthesized_expression */
extern T0* T458f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIZED_EXPRESSION.set_left_parenthesis */
extern void T1201f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.make */
extern T0* T1201c22(GE_context* ac, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.tokens */
extern T0* T1201f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_830 */
extern void T449f2053(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].force */
extern void T991f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].fill_with */
extern void T992f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].extend */
extern void T992f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_829 */
extern void T449f2052(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T991f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T992f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_bracket_arguments */
extern T0* T458f313(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_bracketed_with_capacity */
extern T0* T274c17(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_828 */
extern void T449f2051(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_827 */
extern void T449f2050(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].force */
extern void T1012f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].fill_with */
extern void T1014f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].extend */
extern void T1014f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].aliased_resized_area */
extern T0* T1012f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].aliased_resized_area */
extern T0* T1014f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_826 */
extern void T449f2049(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_bracket_expression */
extern T0* T458f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BRACKET_EXPRESSION.make */
extern T0* T1013c25(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_825 */
extern void T449f2048(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_824 */
extern void T449f2047(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_823 */
extern void T449f2046(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_822 */
extern void T449f2045(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_821 */
extern void T449f2044(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_820 */
extern void T449f2043(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_819 */
extern void T449f2042(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_818 */
extern void T449f2041(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_817 */
extern void T449f2040(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].force */
extern void T1102f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].fill_with */
extern void T1103f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].extend */
extern void T1103f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].aliased_resized_area */
extern T0* T1102f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].aliased_resized_area */
extern T0* T1103f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_816 */
extern void T449f2039(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.raise_error */
extern void T449f2169(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_17_11_0 */
extern T0* T449f949(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_815 */
extern void T449f2038(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_814 */
extern void T449f2037(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_813 */
extern void T449f2036(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_812 */
extern void T449f2035(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_811 */
extern void T449f2034(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_810 */
extern void T449f2033(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_once_manifest_string */
extern T0* T449f948(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.register_inline_constant */
extern void T73f319(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.set_id */
extern void T1502f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_once_manifest_string */
extern T0* T458f359(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_MANIFEST_STRING.set_once_keyword */
extern void T1502f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.make */
extern T0* T1502c24(GE_context* ac, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.tokens */
extern T0* T1502f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_809 */
extern void T449f2032(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_808 */
extern void T449f2031(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_807 */
extern void T449f2030(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_806 */
extern void T449f2029(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_805 */
extern void T449f2028(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_804 */
extern void T449f2027(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_803 */
extern void T449f2026(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_802 */
extern void T449f2025(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_801 */
extern void T449f2024(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_601_800 */
extern void T449f1206(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_800 */
extern void T449f2023(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_799 */
extern void T449f2022(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_798 */
extern void T449f2021(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_797 */
extern void T449f2020(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_named_object_test */
extern T0* T449f947(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_OBJECT_TEST_LIST.force_last */
extern void T1431f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OBJECT_TEST_LIST.resize */
extern void T1431f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].aliased_resized_area */
extern T0* T717f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].aliased_resized_area */
extern T0* T716f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.new_capacity */
extern T6 T1431f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.capacity */
extern T6 T1431f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_object_test_list */
extern T0* T449f972(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].remove */
extern void T921f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].keep_head */
extern void T1589f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].set_count */
extern void T1589f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].item */
extern T0* T921f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].is_empty */
extern T1 T921f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_named_object_test */
extern T0* T458f358(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NAMED_OBJECT_TEST.set_attached_keyword */
extern void T715f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.make */
extern T0* T715c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NAMED_OBJECT_TEST.make_object_test */
extern void T715f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NAMED_OBJECT_TEST.tokens */
extern T0* T715f11(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_796 */
extern void T449f2019(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_795 */
extern void T449f2018(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_object_test */
extern T0* T458f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST.set_attached_keyword */
extern void T308f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST.make */
extern T0* T308c24(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_OBJECT_TEST.tokens */
extern T0* T308f11(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_794 */
extern void T449f2017(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_793 */
extern void T449f2016(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_old_object_test */
extern T0* T449f946(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_AST_FACTORY.new_old_object_test */
extern T0* T458f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_OLD_OBJECT_TEST.set_left_brace */
extern void T1501f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_OBJECT_TEST.make */
extern T0* T1501c29(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.colon_symbol */
extern T0* T72f700(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_colon */
extern T0* T343c93(GE_context* ac);
/* detachable ET_OLD_OBJECT_TEST.tokens */
extern T0* T1501f14(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_6_1_0 */
extern T0* T449f934(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_792 */
extern void T449f2015(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_old_expression */
extern T0* T458f310(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_OLD_EXPRESSION.set_old_keyword */
extern void T1500f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_EXPRESSION.make */
extern T0* T1500c21(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.old_keyword */
extern T0* T72f1516(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_old */
extern T0* T409c58(GE_context* ac);
/* detachable ET_OLD_EXPRESSION.tokens */
extern T0* T1500f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_791 */
extern void T449f2014(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_expression */
extern T0* T458f308(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_EXPRESSION.make */
extern T0* T1499c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_prefix_free_operator */
extern T0* T458f309(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_prefix */
extern void T951f49(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.tokens */
extern T0* T951f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_790 */
extern void T449f2013(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_789 */
extern void T449f2012(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_prefix_minus_expression */
extern T0* T449f945(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T449f945ot1(T0* a1, T0** a2);
extern T1 T449f945ot2(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_prefix_minus_operator */
extern T0* T458f356(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_prefix_minus */
extern void T972f98(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.tokens */
extern T0* T972f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_788 */
extern void T449f2011(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_prefix_plus_expression */
extern T0* T449f944(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T449f944ot1(T0* a1, T0** a2);
extern T1 T449f944ot2(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_prefix_plus_operator */
extern T0* T458f355(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_prefix_plus */
extern void T972f97(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_787 */
extern void T449f2010(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_786 */
extern void T449f2009(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_785 */
extern void T449f2008(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_784 */
extern void T449f2007(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_783 */
extern void T449f2006(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_782 */
extern void T449f2005(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_781 */
extern void T449f2004(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_780 */
extern void T449f2003(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_779 */
extern void T449f2002(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_778 */
extern void T449f2001(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_manifest_type */
extern T0* T458f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_TYPE.make */
extern T0* T1498c22(GE_context* ac, T0* a1);
/* detachable ET_MANIFEST_TYPE.tokens */
extern T0* T1498f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_777 */
extern void T449f2000(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_776 */
extern void T449f1999(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_775 */
extern void T449f1998(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_774 */
extern void T449f1997(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_773 */
extern void T449f1996(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_772 */
extern void T449f1995(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_object_equality_expression */
extern T0* T458f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.make */
extern T0* T309c23(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_771 */
extern void T449f1994(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_770 */
extern void T449f1993(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_equality_expression */
extern T0* T458f305(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EQUALITY_EXPRESSION.make */
extern T0* T310c21(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_769 */
extern void T449f1992(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_768 */
extern void T449f1991(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_expression */
extern T0* T458f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_EXPRESSION.make */
extern T0* T1495c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_767 */
extern void T449f1990(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_or_else_operator */
extern T0* T458f304(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.set_or_keyword */
extern void T1497f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.make */
extern T0* T1497c34(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.else_keyword */
extern T0* T72f712(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_else */
extern T0* T409c72(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.or_keyword */
extern T0* T72f1515(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.make_or */
extern T0* T959c47(GE_context* ac);
/* detachable ET_KEYWORD_OPERATOR.make_token */
extern void T959f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_leaf */
extern void T959f54(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.tokens */
extern T0* T959f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.tokens */
extern T0* T1497f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_766 */
extern void T449f1989(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_and_then_operator */
extern T0* T458f303(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_AND_THEN_OPERATOR.set_and_keyword */
extern void T1496f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_AND_THEN_OPERATOR.make */
extern T0* T1496c34(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.then_keyword */
extern T0* T72f716(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_then */
extern T0* T409c76(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.and_keyword */
extern T0* T72f1514(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.make_and */
extern T0* T959c49(GE_context* ac);
/* detachable ET_INFIX_AND_THEN_OPERATOR.tokens */
extern T0* T1496f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_765 */
extern void T449f1988(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_764 */
extern void T449f1987(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_763 */
extern void T449f1986(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_762 */
extern void T449f1985(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_761 */
extern void T449f1984(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_760 */
extern void T449f1983(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_759 */
extern void T449f1982(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_758 */
extern void T449f1981(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_757 */
extern void T449f1980(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_756 */
extern void T449f1979(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_755 */
extern void T449f1978(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_754 */
extern void T449f1977(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_753 */
extern void T449f1976(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_minus_operator */
extern T0* T458f302(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_infix_minus */
extern void T972f96(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_752 */
extern void T449f1975(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_plus_operator */
extern T0* T458f301(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_infix_plus */
extern void T972f95(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_751 */
extern void T449f1974(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_free_operator */
extern T0* T458f299(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_infix */
extern void T951f48(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_750 */
extern void T449f1973(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_749 */
extern void T449f1972(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_748 */
extern void T449f1971(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].force */
extern void T1261f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WRITABLE].fill_with */
extern void T1262f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WRITABLE].extend */
extern void T1262f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].aliased_resized_area */
extern T0* T1261f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WRITABLE].aliased_resized_area */
extern T0* T1262f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_747 */
extern void T449f1970(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_writable */
extern T0* T449f943(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_746 */
extern void T449f1969(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_expression_address */
extern T0* T458f298(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXPRESSION_ADDRESS.set_dollar */
extern void T1494f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPRESSION_ADDRESS.make */
extern T0* T1494c21(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.dollar_symbol */
extern T0* T72f1513(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dollar */
extern T0* T343c85(GE_context* ac);
/* detachable ET_EXPRESSION_ADDRESS.tokens */
extern T0* T1494f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_745 */
extern void T449f1968(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_result_address */
extern T0* T458f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_RESULT_ADDRESS.set_dollar */
extern void T1493f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RESULT_ADDRESS.make */
extern T0* T1493c21(GE_context* ac);
/* detachable ET_RESULT_ADDRESS.tokens */
extern T0* T1493f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_744 */
extern void T449f1967(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_current_address */
extern T0* T458f296(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CURRENT_ADDRESS.set_dollar */
extern void T1492f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CURRENT_ADDRESS.make */
extern T0* T1492c21(GE_context* ac);
/* detachable ET_CURRENT_ADDRESS.tokens */
extern T0* T1492f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_743 */
extern void T449f1966(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_feature_address */
extern T0* T449f942(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T449f942ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_feature_address */
extern T0* T458f354(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_ADDRESS.set_dollar */
extern void T1491f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_ADDRESS.make */
extern T0* T1491c21(GE_context* ac, T0* a1);
/* detachable ET_FEATURE_ADDRESS.tokens */
extern T0* T1491f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_742 */
extern void T449f1965(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].force */
extern void T1104f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].fill_with */
extern void T1105f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].extend */
extern void T1105f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T1104f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T1105f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_expression_comma */
extern T0* T458f295(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_741 */
extern void T449f1964(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_740 */
extern void T449f1963(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_actual_arguments */
extern T0* T458f294(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_739 */
extern void T449f1962(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_738 */
extern void T449f1961(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_737 */
extern void T449f1960(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_736 */
extern void T449f1959(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_735 */
extern void T449f1958(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_734 */
extern void T449f1957(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_call_expression */
extern T0* T458f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.new_dot_feature_name */
extern T0* T449f933(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_733 */
extern void T449f1956(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_5_7_59914 */
extern T0* T449f916(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_732 */
extern void T449f1955(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_731 */
extern void T449f1954(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_730 */
extern void T449f1953(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_729 */
extern void T449f1952(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_728 */
extern void T449f1951(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_727 */
extern void T449f1950(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_726 */
extern void T449f1949(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_725 */
extern void T449f1948(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_unqualified_call_expression */
extern T0* T449f941(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_724 */
extern void T449f1947(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_expression */
extern T0* T458f293(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_EXPRESSION.set_precursor_keyword */
extern void T1490f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_parent_prefixed */
extern void T1490f27(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PRECURSOR_EXPRESSION.make */
extern T0* T1490c25(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.precursor_keyword */
extern T0* T72f1512(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.make */
extern T0* T933c21(GE_context* ac);
/* detachable ET_PRECURSOR_KEYWORD.make_token */
extern void T933f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.make_leaf */
extern void T933f24(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.tokens */
extern T0* T933f7(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.tokens */
extern T0* T1490f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_class_name */
extern T0* T458f289(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_723 */
extern void T449f1946(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_722 */
extern void T449f1945(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].force */
extern void T1229f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].fill_with */
extern void T1231f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].extend */
extern void T1231f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1229f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1231f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_static_call_expression */
extern T0* T458f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_STATIC_CALL_EXPRESSION.set_feature_keyword */
extern void T1230f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_EXPRESSION.make */
extern T0* T1230c29(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.make_qualified_call */
extern void T1230f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_721 */
extern void T449f1944(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_720 */
extern void T449f1943(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_719 */
extern void T449f1942(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_718 */
extern void T449f1941(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_717 */
extern void T449f1940(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_716 */
extern void T449f1939(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_715 */
extern void T449f1938(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].force */
extern void T1164f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].fill_with */
extern void T1166f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].extend */
extern void T1166f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].aliased_resized_area */
extern T0* T1164f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].aliased_resized_area */
extern T0* T1166f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_static_call_instruction */
extern T0* T458f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_feature_keyword */
extern void T1489f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.make */
extern T0* T1489c8(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_INSTRUCTION.make_qualified_call */
extern void T1489f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_714 */
extern void T449f1937(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_713 */
extern void T449f1936(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_instruction */
extern T0* T458f288(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.set_precursor_keyword */
extern void T1487f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parent_prefixed */
extern void T1487f10(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PRECURSOR_INSTRUCTION.make */
extern T0* T1487c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_PRECURSOR_INSTRUCTION.tokens */
extern T0* T1487f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_712 */
extern void T449f1935(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_711 */
extern void T449f1934(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_call_instruction */
extern T0* T458f287(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_710 */
extern void T449f1933(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_709 */
extern void T449f1932(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_unqualified_call_instruction */
extern T0* T449f940(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_unqualified_call_instruction */
extern T0* T458f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.make */
extern T0* T1704c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_708 */
extern void T449f1931(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].force */
extern void T1189f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].fill_with */
extern void T1191f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].extend */
extern void T1191f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].aliased_resized_area */
extern T0* T1189f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].aliased_resized_area */
extern T0* T1191f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_string_comma */
extern T0* T458f286(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_707 */
extern void T449f1930(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].force */
extern void T1192f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].fill_with */
extern void T1194f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].extend */
extern void T1194f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_STRING_LIST.put_first */
extern void T1193f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].force */
extern void T1785f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].fill_with */
extern void T1784f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].extend */
extern void T1784f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_STRING_LIST.fixed_array */
extern T0* T1193f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].default_create */
extern T0* T1785c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_706 */
extern void T449f1929(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].aliased_resized_area */
extern T0* T1192f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].aliased_resized_area */
extern T0* T1194f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_string_list */
extern T0* T458f285(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_MANIFEST_STRING_LIST.make_with_capacity */
extern T0* T1193c9(GE_context* ac, T6 a1);
/* detachable ET_MANIFEST_STRING_LIST.tokens */
extern T0* T1193f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.make_with_capacity */
extern void T1193f9p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].make */
extern T0* T1785f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].make_empty */
extern T0* T1784c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_705 */
extern void T449f1928(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_704 */
extern void T449f1927(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_703 */
extern void T449f1926(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_702 */
extern void T449f1925(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_701 */
extern void T449f1924(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].force */
extern void T1082f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].fill_with */
extern void T1084f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].extend */
extern void T1084f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].aliased_resized_area */
extern T0* T1082f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].aliased_resized_area */
extern T0* T1084f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_debug_instruction */
extern T0* T458f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DEBUG_INSTRUCTION.make */
extern T0* T1083c5(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_DEBUG_INSTRUCTION.tokens */
extern T0* T1083f4(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_debug_compound */
extern T0* T458f283(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.debug_keyword */
extern T0* T72f720(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_debug */
extern T0* T409c80(GE_context* ac);
/* detachable ET_COMPOUND.make */
extern T0* T330c9(GE_context* ac);
/* detachable ET_COMPOUND.make */
extern void T330f9p1(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].make */
extern T0* T1913f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].make_empty */
extern T0* T1911c4(GE_context* ac, T6 a1);
/* detachable ET_COMPOUND.fixed_array */
extern T0* T330f7(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].default_create */
extern T0* T1913c2(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.do_keyword */
extern T0* T72f709(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_do */
extern T0* T409c69(GE_context* ac);
/* detachable ET_COMPOUND.tokens */
extern T0* T330f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_700 */
extern void T449f1923(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].force */
extern void T1042f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].fill_with */
extern void T1044f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].extend */
extern void T1044f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].aliased_resized_area */
extern T0* T1042f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].aliased_resized_area */
extern T0* T1044f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_699 */
extern void T449f1922(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_698 */
extern void T449f1921(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].force */
extern void T988f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].fill_with */
extern void T990f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].extend */
extern void T990f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].aliased_resized_area */
extern T0* T988f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].aliased_resized_area */
extern T0* T990f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_across_instruction_header */
extern T0* T449f939(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_AST_FACTORY.new_across_instruction */
extern T0* T458f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_ACROSS_INSTRUCTION.set_as_keyword */
extern void T989f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_variant_part */
extern void T989f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_invariant_part */
extern void T989f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.make */
extern T0* T989c19(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ACROSS_INSTRUCTION.create_unfolded_form */
extern void T989f27(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.tokens */
extern T0* T989f16(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_697 */
extern void T449f1920(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_696 */
extern void T449f1919(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].force */
extern void T1024f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].fill_with */
extern void T1026f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].extend */
extern void T1026f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].aliased_resized_area */
extern T0* T1024f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].aliased_resized_area */
extern T0* T1026f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_695 */
extern void T449f1918(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_choice_attribute_constant */
extern T0* T449f938(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_694 */
extern void T449f1917(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_693 */
extern void T449f1916(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_692 */
extern void T449f1915(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].force */
extern void T1021f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE].fill_with */
extern void T1023f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE].extend */
extern void T1023f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].aliased_resized_area */
extern T0* T1021f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE].aliased_resized_area */
extern T0* T1023f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_range */
extern T0* T458f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CHOICE_RANGE.make */
extern T0* T1486c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.dotdot_symbol */
extern T0* T72f705(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dotdot */
extern T0* T343c98(GE_context* ac);
/* detachable ET_CHOICE_RANGE.tokens */
extern T0* T1486f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_691 */
extern void T449f1914(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_690 */
extern void T449f1913(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].force */
extern void T1027f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].fill_with */
extern void T1029f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].extend */
extern void T1029f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].aliased_resized_area */
extern T0* T1027f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].aliased_resized_area */
extern T0* T1029f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_comma */
extern T0* T458f281(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_689 */
extern void T449f1912(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].force */
extern void T1030f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].fill_with */
extern void T1032f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].extend */
extern void T1032f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.put_first */
extern void T1031f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].force */
extern void T1763f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].fill_with */
extern void T1762f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].extend */
extern void T1762f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.fixed_array */
extern T0* T1031f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].default_create */
extern T0* T1763c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_688 */
extern void T449f1911(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].aliased_resized_area */
extern T0* T1030f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].aliased_resized_area */
extern T0* T1032f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_list */
extern T0* T458f280(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_CHOICE_LIST.set_when_keyword */
extern void T1031f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.make_with_capacity */
extern T0* T1031c8(GE_context* ac, T6 a1);
/* detachable ET_CHOICE_LIST.make_with_capacity */
extern void T1031f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].make */
extern T0* T1763f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].make_empty */
extern T0* T1762c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.when_keyword */
extern T0* T72f717(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_when */
extern T0* T409c77(GE_context* ac);
/* detachable ET_CHOICE_LIST.tokens */
extern T0* T1031f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_687 */
extern void T449f1910(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_686 */
extern void T449f1909(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_685 */
extern void T449f1908(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_684 */
extern void T449f1907(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_683 */
extern void T449f1906(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].force */
extern void T1255f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].fill_with */
extern void T1257f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].extend */
extern void T1257f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].aliased_resized_area */
extern T0* T1255f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].aliased_resized_area */
extern T0* T1257f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_part */
extern T0* T458f279(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_WHEN_PART.make */
extern T0* T1256c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_682 */
extern void T449f1905(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].force */
extern void T1258f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].fill_with */
extern void T1260f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].extend */
extern void T1260f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.put_first */
extern void T1259f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].force */
extern void T1792f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].fill_with */
extern void T1791f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].extend */
extern void T1791f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.fixed_array */
extern T0* T1259f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].default_create */
extern T0* T1792c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_681 */
extern void T449f1904(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].aliased_resized_area */
extern T0* T1258f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].aliased_resized_area */
extern T0* T1260f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_part_list */
extern T0* T458f278(GE_context* ac, T0* C, T6 a1);
/* detachable ET_WHEN_PART_LIST.make_with_capacity */
extern T0* T1259c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].make */
extern T0* T1792f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].make_empty */
extern T0* T1791c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_680 */
extern void T449f1903(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_679 */
extern void T449f1902(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_678 */
extern void T449f1901(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].force */
extern void T1161f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].fill_with */
extern void T1163f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].extend */
extern void T1163f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].aliased_resized_area */
extern T0* T1161f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].aliased_resized_area */
extern T0* T1163f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_inspect_instruction */
extern T0* T458f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_INSPECT_INSTRUCTION.set_else_compound */
extern void T1162f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INSPECT_INSTRUCTION.make */
extern T0* T1162c7(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_INSPECT_INSTRUCTION.tokens */
extern T0* T1162f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_677 */
extern void T449f1900(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_676 */
extern void T449f1899(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].force */
extern void T1085f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].fill_with */
extern void T1087f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].extend */
extern void T1087f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].aliased_resized_area */
extern T0* T1085f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].aliased_resized_area */
extern T0* T1087f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_expression */
extern T0* T458f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ELSEIF_EXPRESSION.make */
extern T0* T1086c7(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ELSEIF_EXPRESSION.tokens */
extern T0* T1086f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_675 */
extern void T449f1898(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].force */
extern void T1088f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].fill_with */
extern void T1090f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].extend */
extern void T1090f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.put_first */
extern void T1089f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].force */
extern void T1773f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].fill_with */
extern void T1772f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].extend */
extern void T1772f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.fixed_array */
extern T0* T1089f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].default_create */
extern T0* T1773c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_674 */
extern void T449f1897(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1088f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1090f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_expression_list */
extern T0* T458f275(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.make_with_capacity */
extern T0* T1089c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].make */
extern T0* T1773f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].make_empty */
extern T0* T1772c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_673 */
extern void T449f1896(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_672 */
extern void T449f1895(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].force */
extern void T1139f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].fill_with */
extern void T1141f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].extend */
extern void T1141f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].aliased_resized_area */
extern T0* T1139f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].aliased_resized_area */
extern T0* T1141f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_if_expression */
extern T0* T458f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_IF_EXPRESSION.set_elseif_parts */
extern void T1140f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_EXPRESSION.make */
extern T0* T1140c27(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_IF_EXPRESSION.tokens */
extern T0* T1140f13(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_671 */
extern void T449f1894(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_670 */
extern void T449f1893(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].force */
extern void T1091f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].fill_with */
extern void T1093f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].extend */
extern void T1093f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].aliased_resized_area */
extern T0* T1091f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].aliased_resized_area */
extern T0* T1093f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_part */
extern T0* T458f273(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ELSEIF_PART.make */
extern T0* T1092c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_669 */
extern void T449f1892(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].force */
extern void T1094f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].fill_with */
extern void T1096f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].extend */
extern void T1096f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART_LIST.put_first */
extern void T1095f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].force */
extern void T1775f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].fill_with */
extern void T1774f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].extend */
extern void T1774f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART_LIST.fixed_array */
extern T0* T1095f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].default_create */
extern T0* T1775c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_668 */
extern void T449f1891(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].aliased_resized_area */
extern T0* T1094f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].aliased_resized_area */
extern T0* T1096f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_part_list */
extern T0* T458f272(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ELSEIF_PART_LIST.make_with_capacity */
extern T0* T1095c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].make */
extern T0* T1775f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].make_empty */
extern T0* T1774c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_667 */
extern void T449f1890(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_666 */
extern void T449f1889(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].force */
extern void T1142f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].fill_with */
extern void T1144f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].extend */
extern void T1144f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].aliased_resized_area */
extern T0* T1142f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].aliased_resized_area */
extern T0* T1144f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_if_instruction */
extern T0* T458f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_IF_INSTRUCTION.set_else_compound */
extern void T1143f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_INSTRUCTION.set_elseif_parts */
extern void T1143f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_INSTRUCTION.make */
extern T0* T1143c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_IF_INSTRUCTION.tokens */
extern T0* T1143f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_665 */
extern void T449f1888(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_664 */
extern void T449f1887(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_663 */
extern void T449f1886(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_662 */
extern void T449f1885(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].force */
extern void T1073f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].fill_with */
extern void T1075f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].extend */
extern void T1075f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].aliased_resized_area */
extern T0* T1073f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].aliased_resized_area */
extern T0* T1075f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_creation_region */
extern T0* T458f270(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATION_REGION.set_right_symbol */
extern void T1074f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATION_REGION.set_left_symbol */
extern void T1074f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATION_REGION.make */
extern T0* T1074c5(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.greater_than_symbol */
extern T0* T72f891(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.make_gt */
extern T0* T972c89(GE_context* ac);
/* detachable ET_SYMBOL_OPERATOR.make_leaf */
extern void T972f99(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.less_than_symbol */
extern T0* T72f890(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.make_lt */
extern T0* T972c90(GE_context* ac);
/* detachable ET_CREATION_REGION.tokens */
extern T0* T1074f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_661 */
extern void T449f1884(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_660 */
extern void T449f1883(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].force */
extern void T1071f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].fill_with */
extern void T1072f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].extend */
extern void T1072f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].aliased_resized_area */
extern T0* T1071f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].aliased_resized_area */
extern T0* T1072f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_create_expression */
extern T0* T458f269(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CREATE_EXPRESSION.set_creation_region */
extern void T320f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CREATE_EXPRESSION.set_create_keyword */
extern void T320f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_qualified_call */
extern T0* T458f266(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_659 */
extern void T449f1882(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_658 */
extern void T449f1881(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_create_instruction */
extern T0* T458f268(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CREATE_INSTRUCTION.set_creation_region */
extern void T1485f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.set_create_keyword */
extern void T1485f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.make */
extern T0* T1485c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATE_INSTRUCTION.tokens */
extern T0* T1485f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_657 */
extern void T449f1880(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_656 */
extern void T449f1879(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_655 */
extern void T449f1878(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_654 */
extern void T449f1877(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_bang_instruction */
extern T0* T458f265(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_BANG_INSTRUCTION.set_left_bang */
extern void T1484f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BANG_INSTRUCTION.make */
extern T0* T1484c11(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.bang_symbol */
extern T0* T72f1511(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_bang */
extern T0* T343c84(GE_context* ac);
/* detachable ET_BANG_INSTRUCTION.tokens */
extern T0* T1484f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_653 */
extern void T449f1876(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_652 */
extern void T449f1875(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_651 */
extern void T449f1874(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_650 */
extern void T449f1873(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].force */
extern void T1018f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].fill_with */
extern void T1020f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].extend */
extern void T1020f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].aliased_resized_area */
extern T0* T1018f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].aliased_resized_area */
extern T0* T1020f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_check_instruction */
extern T0* T449f937(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].remove_last */
extern void T924f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].keep_head */
extern void T1594f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].set_count */
extern void T1594f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].move_last_cursors_after */
extern void T924f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].set_next_cursor */
extern void T1596f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].set_after */
extern void T1596f6(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.put_first */
extern void T1019f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].force */
extern void T1595f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].fill_with */
extern void T1594f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].extend */
extern void T1594f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.fixed_array */
extern T0* T1019f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].default_create */
extern T0* T1595c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].item */
extern T0* T924f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_check_instruction */
extern T0* T458f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* detachable ET_CHECK_INSTRUCTION.set_check_keyword */
extern void T1019f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.set_then_compound */
extern void T1019f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.make_with_capacity */
extern T0* T1019c10(GE_context* ac, T6 a1);
/* detachable ET_CHECK_INSTRUCTION.make_with_capacity */
extern void T1019f10p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].make */
extern T0* T1595f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].make_empty */
extern T0* T1594c5(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.check_keyword */
extern T0* T72f719(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_check */
extern T0* T409c79(GE_context* ac);
/* detachable ET_CHECK_INSTRUCTION.tokens */
extern T0* T1019f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].remove_last */
extern void T259f29(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].move_last_cursors_after */
extern void T259f30(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].last */
extern T6 T259f15(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_649 */
extern void T449f1872(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_648 */
extern void T449f1871(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_647 */
extern void T449f1870(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_646 */
extern void T449f1869(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_check_instruction */
extern void T449f2196(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_assertions */
extern void T449f2201(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_645 */
extern void T449f1868(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_null_instruction */
extern T0* T458f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_644 */
extern void T449f1867(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_643 */
extern void T449f1866(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_642 */
extern void T449f1865(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_641 */
extern void T449f1864(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_instruction */
extern T0* T449f936(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_ACROSS_INSTRUCTION.set_end_keyword */
extern void T989f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_loop_compound */
extern void T989f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_until_conditional */
extern void T989f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_from_compound */
extern void T989f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_640 */
extern void T449f1863(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_instruction */
extern T0* T458f263(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_LOOP_INSTRUCTION.set_has_old_variant_syntax */
extern void T1482f13(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOOP_INSTRUCTION.set_variant_part */
extern void T1482f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.set_invariant_part */
extern void T1482f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.make */
extern T0* T1482c10(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_LOOP_INSTRUCTION.tokens */
extern T0* T1482f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_6_3_7_4554 */
extern T0* T449f935(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_639 */
extern void T449f1862(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_instruction_old_syntax */
extern T0* T458f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_638 */
extern void T449f1861(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_637 */
extern void T449f1860(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_636 */
extern void T449f1859(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_635 */
extern void T449f1858(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assignment_attempt */
extern T0* T458f260(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNMENT_ATTEMPT.make */
extern T0* T1481c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.assign_attempt_symbol */
extern T0* T72f707(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_assign_attempt */
extern T0* T343c100(GE_context* ac);
/* detachable ET_ASSIGNMENT_ATTEMPT.tokens */
extern T0* T1481f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_634 */
extern void T449f1857(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assignment */
extern T0* T458f259(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNMENT.make */
extern T0* T1480c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.assign_symbol */
extern T0* T72f706(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_assign */
extern T0* T343c99(GE_context* ac);
/* detachable ET_ASSIGNMENT.tokens */
extern T0* T1480f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_633 */
extern void T449f1856(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assigner_instruction */
extern T0* T458f258(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNER_INSTRUCTION.make */
extern T0* T1479c15(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ASSIGNER_INSTRUCTION.tokens */
extern T0* T1479f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_632 */
extern void T449f1855(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_631 */
extern void T449f1854(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_630 */
extern void T449f1853(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_629 */
extern void T449f1852(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_628 */
extern void T449f1851(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].force */
extern void T1040f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_COMPOUND].fill_with */
extern void T1041f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_COMPOUND].extend */
extern void T1041f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].aliased_resized_area */
extern T0* T1040f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_COMPOUND].aliased_resized_area */
extern T0* T1041f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_627 */
extern void T449f1850(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.put_first */
extern void T330f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].force */
extern void T1913f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].fill_with */
extern void T1911f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].extend */
extern void T1911f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_626 */
extern void T449f1849(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_compound */
extern T0* T458f257(GE_context* ac, T0* C, T6 a1);
/* detachable ET_COMPOUND.make_with_capacity */
extern T0* T330c12(GE_context* ac, T6 a1);
/* detachable ET_COMPOUND.make_with_capacity */
extern void T330f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_625 */
extern void T449f1848(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_624 */
extern void T449f1847(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_623 */
extern void T449f1846(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_622 */
extern void T449f1845(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_empty_compound */
extern T0* T458f256(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_621 */
extern void T449f1844(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_620 */
extern void T449f1843(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_compound */
extern T0* T458f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.loop_keyword */
extern T0* T72f714(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_loop */
extern T0* T409c74(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_619 */
extern void T449f1842(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_618 */
extern void T449f1841(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_617 */
extern void T449f1840(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_from_compound */
extern T0* T458f254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.from_keyword */
extern T0* T72f713(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_from */
extern T0* T409c73(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_616 */
extern void T449f1839(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_rescue_compound */
extern T0* T458f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.rescue_keyword */
extern T0* T72f729(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_rescue */
extern T0* T409c89(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_615 */
extern void T449f1838(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_else_compound */
extern T0* T458f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_614 */
extern void T449f1837(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_613 */
extern void T449f1836(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_then_compound */
extern T0* T458f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_612 */
extern void T449f1835(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_611 */
extern void T449f1834(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_attribute_compound */
extern T0* T458f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.attribute_keyword */
extern T0* T72f741(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_attribute */
extern T0* T409c101(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_610 */
extern void T449f1833(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_compound */
extern T0* T458f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_609 */
extern void T449f1832(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].force */
extern void T1219f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].fill_with */
extern void T1221f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].extend */
extern void T1221f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].aliased_resized_area */
extern T0* T1219f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].aliased_resized_area */
extern T0* T1221f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_qualified_like_type */
extern T0* T458f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.make */
extern T0* T1478c80(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_AST_FACTORY.new_attachment_mark_separate_keyword */
extern T0* T458f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_608 */
extern void T449f1831(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_607 */
extern void T449f1830(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_606 */
extern void T449f1829(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_605 */
extern void T449f1828(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_604 */
extern void T449f1827(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_603 */
extern void T449f1826(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_602 */
extern void T449f1825(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_601 */
extern void T449f1824(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_401_600 */
extern void T449f1205(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_600 */
extern void T449f1823(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_like_braced_type */
extern T0* T458f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_like_keyword */
extern void T1477f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.make */
extern T0* T1477c83(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_599 */
extern void T449f1822(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_598 */
extern void T449f1821(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_597 */
extern void T449f1820(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_596 */
extern void T449f1819(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_595 */
extern void T449f1818(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_594 */
extern void T449f1817(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_593 */
extern void T449f1816(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_592 */
extern void T449f1815(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_591 */
extern void T449f1814(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_590 */
extern void T449f1813(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_589 */
extern void T449f1812(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_588 */
extern void T449f1811(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].force */
extern void T1172f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].fill_with */
extern void T1174f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].extend */
extern void T1174f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].aliased_resized_area */
extern T0* T1172f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].aliased_resized_area */
extern T0* T1174f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_587 */
extern void T449f1810(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_like_current */
extern T0* T458f246(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.set_like_keyword */
extern void T442f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_586 */
extern void T449f1809(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_585 */
extern void T449f1808(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_584 */
extern void T449f1807(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_583 */
extern void T449f1806(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_582 */
extern void T449f1805(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_581 */
extern void T449f1804(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_580 */
extern void T449f1803(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_579 */
extern void T449f1802(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_578 */
extern void T449f1801(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_577 */
extern void T449f1800(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_like_feature */
extern T0* T449f932(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_like_feature */
extern T0* T458f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.set_like_keyword */
extern void T1476f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.make */
extern T0* T1476c83(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_576 */
extern void T449f1799(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_575 */
extern void T449f1798(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_574 */
extern void T449f1797(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_573 */
extern void T449f1796(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_572 */
extern void T449f1795(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_571 */
extern void T449f1794(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_570 */
extern void T449f1793(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_569 */
extern void T449f1792(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_568 */
extern void T449f1791(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_567 */
extern void T449f1790(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_566 */
extern void T449f1789(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_565 */
extern void T449f1788(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_564 */
extern void T449f1787(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].force */
extern void T993f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T994f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].extend */
extern void T994f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T993f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T994f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_labeled_actual_parameter_semicolon */
extern T0* T458f245(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_labeled_actual_parameter */
extern T0* T458f242(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LABELED_ACTUAL_PARAMETER.make */
extern T0* T1474c10(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_563 */
extern void T449f1786(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_562 */
extern void T449f1785(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].force */
extern void T995f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].fill_with */
extern void T996f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].extend */
extern void T996f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.add_to_actual_parameter_list */
extern void T449f2195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_labeled_comma_actual_parameter */
extern T0* T458f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.make */
extern T0* T1717c10(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.first */
extern T0* T236f15(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_label_comma */
extern T0* T458f243(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_561 */
extern void T449f1784(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_560 */
extern void T449f1783(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_559 */
extern void T449f1782(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_558 */
extern void T449f1781(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_557 */
extern void T449f1780(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T995f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T996f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_actual_parameters */
extern T0* T458f240(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.set_left_bracket */
extern void T236f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_556 */
extern void T449f1779(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_555 */
extern void T449f1778(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_554 */
extern void T449f1777(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_553 */
extern void T449f1776(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_552 */
extern void T449f1775(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_551 */
extern void T449f1774(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_actual_parameter_comma */
extern T0* T458f241(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_550 */
extern void T449f1773(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_549 */
extern void T449f1772(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_tuple_type */
extern T0* T449f931(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T449f931ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_tuple_type */
extern T0* T458f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.set_tuple_keyword */
extern void T248f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.providers_enabled */
extern T1 T449f954(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.current_universe */
extern T0* T449f965(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_548 */
extern void T449f1771(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_named_type */
extern T0* T449f930(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T449f930ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_class_type */
extern T0* T458f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_generic_class_type */
extern T0* T458f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_AST_FACTORY.new_formal_parameter_type */
extern T0* T458f347(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ET_CLASS.formal_parameter */
extern T0* T73f194(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.formal_parameter_by_name */
extern T0* T593f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_547 */
extern void T449f1770(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_546 */
extern void T449f1769(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_545 */
extern void T449f1768(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_544 */
extern void T449f1767(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_543 */
extern void T449f1766(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_542 */
extern void T449f1765(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_541 */
extern void T449f1764(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_540 */
extern void T449f1763(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_539 */
extern void T449f1762(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_538 */
extern void T449f1761(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_537 */
extern void T449f1760(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].force */
extern void T1235f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE].fill_with */
extern void T1236f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE].extend */
extern void T1236f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].aliased_resized_area */
extern T0* T1235f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE].aliased_resized_area */
extern T0* T1236f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_536 */
extern void T449f1759(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_535 */
extern void T449f1758(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_534 */
extern void T449f1757(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_533 */
extern void T449f1756(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_532 */
extern void T449f1755(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_531 */
extern void T449f1754(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_530 */
extern void T449f1753(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_529 */
extern void T449f1752(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_528 */
extern void T449f1751(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_527 */
extern void T449f1750(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_526 */
extern void T449f1749(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_525 */
extern void T449f1748(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_524 */
extern void T449f1747(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_523 */
extern void T449f1746(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_522 */
extern void T449f1745(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_521 */
extern void T449f1744(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_520 */
extern void T449f1743(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_519 */
extern void T449f1742(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_518 */
extern void T449f1741(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_517 */
extern void T449f1740(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_516 */
extern void T449f1739(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_515 */
extern void T449f1738(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_514 */
extern void T449f1737(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_513 */
extern void T449f1736(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_512 */
extern void T449f1735(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_511 */
extern void T449f1734(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_510 */
extern void T449f1733(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_509 */
extern void T449f1732(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_508 */
extern void T449f1731(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_507 */
extern void T449f1730(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_506 */
extern void T449f1729(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_505 */
extern void T449f1728(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_504 */
extern void T449f1727(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_503 */
extern void T449f1726(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_502 */
extern void T449f1725(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_501 */
extern void T449f1724(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_500 */
extern void T449f1723(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_499 */
extern void T449f1722(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_498 */
extern void T449f1721(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_497 */
extern void T449f1720(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_496 */
extern void T449f1719(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_495 */
extern void T449f1718(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_494 */
extern void T449f1717(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_493 */
extern void T449f1716(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_492 */
extern void T449f1715(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_491 */
extern void T449f1714(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_490 */
extern void T449f1713(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_489 */
extern void T449f1712(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_488 */
extern void T449f1711(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_487 */
extern void T449f1710(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_486 */
extern void T449f1709(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_485 */
extern void T449f1708(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_484 */
extern void T449f1707(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_483 */
extern void T449f1706(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_482 */
extern void T449f1705(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_481 */
extern void T449f1704(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_480 */
extern void T449f1703(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_479 */
extern void T449f1702(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_478 */
extern void T449f1701(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_477 */
extern void T449f1700(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_476 */
extern void T449f1699(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_475 */
extern void T449f1698(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_474 */
extern void T449f1697(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_473 */
extern void T449f1696(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_472 */
extern void T449f1695(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_471 */
extern void T449f1694(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_470 */
extern void T449f1693(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_469 */
extern void T449f1692(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_468 */
extern void T449f1691(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_467 */
extern void T449f1690(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_466 */
extern void T449f1689(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].force */
extern void T1252f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VARIANT].fill_with */
extern void T1254f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VARIANT].extend */
extern void T1254f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_465 */
extern void T449f1688(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].aliased_resized_area */
extern T0* T1252f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_VARIANT].aliased_resized_area */
extern T0* T1254f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_464 */
extern void T449f1687(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_variant */
extern T0* T458f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VARIANT.set_variant_keyword */
extern void T1253f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VARIANT.make */
extern T0* T1253c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.variant_keyword */
extern T0* T72f734(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_variant */
extern T0* T409c94(GE_context* ac);
/* detachable ET_VARIANT.tokens */
extern T0* T1253f4(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_tag */
extern T0* T458f122(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_463 */
extern void T449f1686(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_462 */
extern void T449f1685(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].force */
extern void T1183f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].fill_with */
extern void T1185f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].extend */
extern void T1185f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].aliased_resized_area */
extern T0* T1183f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].aliased_resized_area */
extern T0* T1185f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_loop_invariants */
extern T0* T449f929(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.put_first */
extern void T1184f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.fixed_array */
extern T0* T1184f5(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_invariants */
extern T0* T458f346(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_LOOP_INVARIANTS.set_invariant_keyword */
extern void T1184f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.make_with_capacity */
extern T0* T1184c9(GE_context* ac, T6 a1);
/* detachable ET_LOOP_INVARIANTS.make_with_capacity */
extern void T1184f9p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.invariant_keyword */
extern T0* T72f742(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_invariant */
extern T0* T409c102(GE_context* ac);
/* detachable ET_LOOP_INVARIANTS.tokens */
extern T0* T1184f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_461 */
extern void T449f1684(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_460 */
extern void T449f1683(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_459 */
extern void T449f1682(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_458 */
extern void T449f1681(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_loop_invariant */
extern void T449f2194(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_457 */
extern void T449f1680(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_456 */
extern void T449f1679(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].force */
extern void T1167f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].fill_with */
extern void T1169f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].extend */
extern void T1169f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].aliased_resized_area */
extern T0* T1167f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].aliased_resized_area */
extern T0* T1169f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_invariants */
extern T0* T449f928(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_across_components_stack */
extern void T449f1187(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].wipe_out */
extern void T922f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].i_th */
extern T0* T922f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_object_tests_stack */
extern void T449f1186(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].wipe_out */
extern void T920f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].i_th */
extern T0* T920f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_local_variables_stack */
extern void T449f1185(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].wipe_out */
extern void T917f10(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_formal_arguments_stack */
extern void T449f1184(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].wipe_out */
extern void T916f10(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.set_across_components */
extern void T1168f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.set_object_tests */
extern void T1168f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.put_first */
extern void T1168f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.fixed_array */
extern T0* T1168f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_invariants */
extern T0* T458f345(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_INVARIANTS.set_invariant_keyword */
extern void T1168f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.make_with_capacity */
extern T0* T1168c25(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_INVARIANTS.make_assertions_with_capacity */
extern void T1168f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INVARIANTS.tokens */
extern T0* T1168f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_455 */
extern void T449f1678(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_454 */
extern void T449f1677(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_453 */
extern void T449f1676(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_452 */
extern void T449f1675(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_invariant */
extern void T449f2193(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_451 */
extern void T449f1674(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].force */
extern void T1212f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].fill_with */
extern void T1214f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].extend */
extern void T1214f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].aliased_resized_area */
extern T0* T1212f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].aliased_resized_area */
extern T0* T1214f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_postconditions */
extern T0* T449f927(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_POSTCONDITIONS.put_first */
extern void T1213f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.fixed_array */
extern T0* T1213f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_postconditions */
extern T0* T458f344(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_POSTCONDITIONS.set_then_keyword */
extern void T1213f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.set_ensure_keyword */
extern void T1213f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.make_with_capacity */
extern T0* T1213c17(GE_context* ac, T6 a1);
/* detachable ET_POSTCONDITIONS.make_with_capacity */
extern void T1213f17p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.ensure_keyword */
extern T0* T72f893(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_ensure */
extern T0* T409c62(GE_context* ac);
/* detachable ET_POSTCONDITIONS.tokens */
extern T0* T1213f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_450 */
extern void T449f1673(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_449 */
extern void T449f1672(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_448 */
extern void T449f1671(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_447 */
extern void T449f1670(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_446 */
extern void T449f1669(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_postcondition */
extern void T449f2192(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_445 */
extern void T449f1668(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].force */
extern void T1215f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].fill_with */
extern void T1216f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].extend */
extern void T1216f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].aliased_resized_area */
extern T0* T1215f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].aliased_resized_area */
extern T0* T1216f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_preconditions */
extern T0* T449f926(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PRECONDITIONS.put_first */
extern void T401f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.fixed_array */
extern T0* T401f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_preconditions */
extern T0* T458f343(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_PRECONDITIONS.set_else_keyword */
extern void T401f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.set_require_keyword */
extern void T401f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.make_with_capacity */
extern T0* T401c16(GE_context* ac, T6 a1);
/* detachable ET_PRECONDITIONS.make_with_capacity */
extern void T401f16p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.require_keyword */
extern T0* T72f1519(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_require */
extern T0* T409c65(GE_context* ac);
/* detachable ET_PRECONDITIONS.tokens */
extern T0* T401f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_444 */
extern void T449f1667(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_443 */
extern void T449f1666(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_442 */
extern void T449f1665(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_441 */
extern void T449f1664(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_440 */
extern void T449f1663(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_precondition */
extern void T449f2191(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_439 */
extern void T449f1662(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_class_assertion */
extern void T449f2190(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.add_untagged_assertion */
extern void T449f2200(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T449f2200ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].force_last */
extern void T924f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].resize */
extern void T924f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].aliased_resized_area */
extern T0* T1595f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].aliased_resized_area */
extern T0* T1594f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].new_capacity */
extern T6 T924f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].extendible */
extern T1 T924f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].replace */
extern void T924f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_assertion_semicolon */
extern T0* T458f331(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TAGGED_ASSERTION.set_untagged_assertion */
extern void T1513f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].last */
extern T0* T924f9(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_class_assertion */
extern T0* T458f238(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.set_class_keyword */
extern void T1473f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.make */
extern T0* T1473c7(GE_context* ac);
/* detachable ET_CLASS_ASSERTION.tokens */
extern T0* T1473f2(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_438 */
extern void T449f1661(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_437 */
extern void T449f1660(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_tagged_assertion */
extern void T449f2189(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T449f2189ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_tagged_assertion */
extern T0* T458f330(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TAGGED_ASSERTION.make */
extern T0* T1513c7(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.report_syntax_error */
extern void T449f1209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.set_syntax_error */
extern void T449f1213(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.column */
extern T6 T1515f3(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.line */
extern T6 T1515f6(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.is_null */
extern T1 T1515f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.current_position */
extern T0* T449f721(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_436 */
extern void T449f1659(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_435 */
extern void T449f1658(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_expression_assertion */
extern void T449f2188(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_434 */
extern void T449f1657(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_433 */
extern void T449f1656(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_432 */
extern void T449f1655(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_431 */
extern void T449f1654(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_430 */
extern void T449f1653(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_429 */
extern void T449f1652(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_428 */
extern void T449f1651(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_427 */
extern void T449f1650(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].force */
extern void T1178f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].fill_with */
extern void T1180f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].extend */
extern void T1180f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].aliased_resized_area */
extern T0* T1178f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].aliased_resized_area */
extern T0* T1180f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_local_variable_semicolon */
extern T0* T458f237(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_local_variable */
extern T0* T458f236(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LOCAL_VARIABLE.make */
extern T0* T1176c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_426 */
extern void T449f1649(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_425 */
extern void T449f1648(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].force */
extern void T1175f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].fill_with */
extern void T1177f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].extend */
extern void T1177f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].aliased_resized_area */
extern T0* T1175f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].aliased_resized_area */
extern T0* T1177f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_local_comma_variable */
extern T0* T458f235(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.make */
extern T0* T1716c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.dummy_type */
extern T0* T449f923(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_424 */
extern void T449f1647(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_local_name_comma */
extern T0* T458f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_423 */
extern void T449f1646(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.put_first */
extern void T331f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].force */
extern void T1916f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].fill_with */
extern void T1914f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].extend */
extern void T1914f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.fixed_array */
extern T0* T331f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].default_create */
extern T0* T1916c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_422 */
extern void T449f1645(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_421 */
extern void T449f1644(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_declared_type */
extern void T1716f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.first */
extern T0* T331f6(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.is_empty */
extern T1 T331f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_420 */
extern void T449f1643(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_419 */
extern void T449f1642(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_local_variables */
extern T0* T449f924(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_local_variables */
extern T0* T458f342(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_LOCAL_VARIABLE_LIST.set_local_keyword */
extern void T331f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.make_with_capacity */
extern T0* T331c11(GE_context* ac, T6 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.make_with_capacity */
extern void T331f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].make */
extern T0* T1916f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].make_empty */
extern T0* T1914c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.local_keyword */
extern T0* T72f722(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_local */
extern T0* T409c82(GE_context* ac);
/* detachable ET_LOCAL_VARIABLE_LIST.tokens */
extern T0* T331f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_418 */
extern void T449f1641(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_417 */
extern void T449f1640(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_416 */
extern void T449f1639(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_415 */
extern void T449f1638(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_414 */
extern void T449f1637(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_413 */
extern void T449f1636(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].force */
extern void T1127f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].fill_with */
extern void T1129f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].extend */
extern void T1129f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].aliased_resized_area */
extern T0* T1127f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].aliased_resized_area */
extern T0* T1129f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_argument_semicolon */
extern T0* T458f233(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_formal_argument */
extern T0* T458f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_ARGUMENT.make */
extern T0* T252c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_412 */
extern void T449f1635(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_411 */
extern void T449f1634(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].force */
extern void T1125f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].fill_with */
extern void T1126f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].extend */
extern void T1126f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].aliased_resized_area */
extern T0* T1125f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].aliased_resized_area */
extern T0* T1126f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_comma_argument */
extern T0* T458f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.make */
extern T0* T1715c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_410 */
extern void T449f1633(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_argument_name_comma */
extern T0* T458f230(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_409 */
extern void T449f1632(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.put_first */
extern void T249f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].force */
extern void T1907f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].fill_with */
extern void T1905f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].extend */
extern void T1905f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.fixed_array */
extern T0* T249f11(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].default_create */
extern T0* T1907c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_408 */
extern void T449f1631(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_407 */
extern void T449f1630(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_declared_type */
extern void T1715f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.first */
extern T0* T249f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_406 */
extern void T449f1629(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_405 */
extern void T449f1628(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_formal_arguments */
extern T0* T449f922(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_AST_FACTORY.new_formal_arguments */
extern T0* T458f341(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_FORMAL_ARGUMENT_LIST.make_with_capacity */
extern T0* T249c12(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.make_with_capacity */
extern void T249f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].make */
extern T0* T1907f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].make_empty */
extern T0* T1905c4(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.tokens */
extern T0* T249f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_404 */
extern void T449f1627(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_403 */
extern void T449f1626(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_402 */
extern void T449f1625(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_401 */
extern void T449f1624(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_201_400 */
extern void T449f1204(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_400 */
extern void T449f1623(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_399 */
extern void T449f1622(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_398 */
extern void T449f1621(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_397 */
extern void T449f1620(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].force */
extern void T1007f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].fill_with */
extern void T1008f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].extend */
extern void T1008f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1007f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1008f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_invalid_alias_name */
extern T0* T449f921(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.new_alias_free_name */
extern T0* T449f970(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_alias_free_name */
extern T0* T458f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_FREE_NAME.set_convert_keyword */
extern void T1470f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.set_alias_keyword */
extern void T1470f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.make_infix */
extern T0* T1470c48(GE_context* ac, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.string_ */
extern T0* T1470f15(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.default_keyword */
extern T0* T1470f21(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_keyword */
extern T0* T72f718(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_alias */
extern T0* T409c78(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_396 */
extern void T449f1619(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_parenthesis_name */
extern T0* T458f229(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.set_convert_keyword */
extern void T886f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.set_alias_keyword */
extern void T886f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.make_parenthesis */
extern T0* T886c89(GE_context* ac, T0* a1);
/* detachable ET_ALIAS_NAME.default_keyword */
extern T0* T886f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_395 */
extern void T449f1618(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_bracket_name */
extern T0* T458f228(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_bracket */
extern T0* T886c88(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_394 */
extern void T449f1617(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_393 */
extern void T449f1616(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_dotdot_name */
extern T0* T458f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_dotdot */
extern T0* T886c87(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_392 */
extern void T449f1615(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_xor_name */
extern T0* T458f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_xor */
extern T0* T886c86(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_391 */
extern void T449f1614(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_implies_name */
extern T0* T458f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_implies */
extern T0* T886c85(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_390 */
extern void T449f1613(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_else_name */
extern T0* T458f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or_else */
extern T0* T886c84(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_389 */
extern void T449f1612(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_name */
extern T0* T458f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or */
extern T0* T886c83(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_388 */
extern void T449f1611(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_then_name */
extern T0* T458f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and_then */
extern T0* T886c82(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_387 */
extern void T449f1610(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_name */
extern T0* T458f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and */
extern T0* T886c81(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_386 */
extern void T449f1609(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_ge_name */
extern T0* T458f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_ge */
extern T0* T886c80(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_385 */
extern void T449f1608(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_gt_name */
extern T0* T458f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_gt */
extern T0* T886c79(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_384 */
extern void T449f1607(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_le_name */
extern T0* T458f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_le */
extern T0* T886c78(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_383 */
extern void T449f1606(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_lt_name */
extern T0* T458f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_lt */
extern T0* T886c77(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_382 */
extern void T449f1605(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_power_name */
extern T0* T458f215(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_power */
extern T0* T886c76(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_381 */
extern void T449f1604(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_mod_name */
extern T0* T458f214(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_mod */
extern T0* T886c75(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_380 */
extern void T449f1603(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_div_name */
extern T0* T458f213(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_div */
extern T0* T886c74(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_379 */
extern void T449f1602(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_divide_name */
extern T0* T458f212(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_divide */
extern T0* T886c73(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_378 */
extern void T449f1601(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_times_name */
extern T0* T458f211(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_times */
extern T0* T886c72(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_377 */
extern void T449f1600(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_minus_name */
extern T0* T458f210(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_minus */
extern T0* T886c71(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_376 */
extern void T449f1599(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_plus_name */
extern T0* T458f209(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_plus */
extern T0* T886c70(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_375 */
extern void T449f1598(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_not_name */
extern T0* T458f208(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_not */
extern T0* T886c67(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_374 */
extern void T449f1597(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].force */
extern void T1106f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].fill_with */
extern void T1107f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].extend */
extern void T1107f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].aliased_resized_area */
extern T0* T1106f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].aliased_resized_area */
extern T0* T1107f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_aliased_feature_name */
extern T0* T458f207(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_373 */
extern void T449f1596(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_372 */
extern void T449f1595(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].force */
extern void T1120f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].fill_with */
extern void T1121f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].extend */
extern void T1121f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1120f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1121f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_invalid_infix_name */
extern T0* T449f920(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.new_infix_free_name */
extern T0* T449f969(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_infix_free_name */
extern T0* T458f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_INFIX_FREE_NAME.set_infix_keyword */
extern void T864f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.make */
extern T0* T864c69(GE_context* ac, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.string_ */
extern T0* T864f7(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.default_keyword */
extern T0* T864f29(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.infix_keyword */
extern T0* T72f721(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_infix */
extern T0* T409c81(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_371 */
extern void T449f1594(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_370 */
extern void T449f1593(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_invalid_prefix_name */
extern T0* T449f919(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.new_prefix_free_name */
extern T0* T449f968(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_prefix_free_name */
extern T0* T458f188(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_FREE_NAME.set_prefix_keyword */
extern void T1467f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.make */
extern T0* T1467c69(GE_context* ac, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.string_ */
extern T0* T1467f30(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.default_keyword */
extern T0* T1467f31(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.prefix_keyword */
extern T0* T72f727(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_prefix */
extern T0* T409c87(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_369 */
extern void T449f1592(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_368 */
extern void T449f1591(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_367 */
extern void T449f1590(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_366 */
extern void T449f1589(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_365 */
extern void T449f1588(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_364 */
extern void T449f1587(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_363 */
extern void T449f1586(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_362 */
extern void T449f1585(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_361 */
extern void T449f1584(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_360 */
extern void T449f1583(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_359 */
extern void T449f1582(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_358 */
extern void T449f1581(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_357 */
extern void T449f1580(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_356 */
extern void T449f1579(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_355 */
extern void T449f1578(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_354 */
extern void T449f1577(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_353 */
extern void T449f1576(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_xor_name */
extern T0* T458f205(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.set_infix_keyword */
extern void T1468f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_NAME.make_xor */
extern T0* T1468c102(GE_context* ac, T0* a1);
/* detachable ET_INFIX_NAME.default_keyword */
extern T0* T1468f46(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_352 */
extern void T449f1575(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_implies_name */
extern T0* T458f204(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_implies */
extern T0* T1468c101(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_351 */
extern void T449f1574(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_or_else_name */
extern T0* T458f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_or_else */
extern T0* T1468c100(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_350 */
extern void T449f1573(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_or_name */
extern T0* T458f202(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_or */
extern T0* T1468c99(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_349 */
extern void T449f1572(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_and_then_name */
extern T0* T458f201(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_and_then */
extern T0* T1468c98(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_348 */
extern void T449f1571(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_and_name */
extern T0* T458f200(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_and */
extern T0* T1468c97(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_347 */
extern void T449f1570(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_ge_name */
extern T0* T458f199(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_ge */
extern T0* T1468c96(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_346 */
extern void T449f1569(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_gt_name */
extern T0* T458f198(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_gt */
extern T0* T1468c95(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_345 */
extern void T449f1568(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_le_name */
extern T0* T458f197(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_le */
extern T0* T1468c94(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_344 */
extern void T449f1567(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_lt_name */
extern T0* T458f196(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_lt */
extern T0* T1468c93(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_343 */
extern void T449f1566(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_power_name */
extern T0* T458f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_power */
extern T0* T1468c92(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_342 */
extern void T449f1565(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_mod_name */
extern T0* T458f194(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_mod */
extern T0* T1468c91(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_341 */
extern void T449f1564(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_div_name */
extern T0* T458f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_div */
extern T0* T1468c90(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_340 */
extern void T449f1563(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_divide_name */
extern T0* T458f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_divide */
extern T0* T1468c89(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_339 */
extern void T449f1562(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_times_name */
extern T0* T458f191(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_times */
extern T0* T1468c88(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_338 */
extern void T449f1561(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_minus_name */
extern T0* T458f190(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_minus */
extern T0* T1468c87(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_337 */
extern void T449f1560(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_plus_name */
extern T0* T458f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_NAME.make_plus */
extern T0* T1468c85(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_336 */
extern void T449f1559(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_335 */
extern void T449f1558(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_minus_name */
extern T0* T458f187(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_NAME.set_prefix_keyword */
extern void T1466f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_NAME.make_minus */
extern T0* T1466c88(GE_context* ac, T0* a1);
/* detachable ET_PREFIX_NAME.default_keyword */
extern T0* T1466f32(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_334 */
extern void T449f1557(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_plus_name */
extern T0* T458f186(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_NAME.make_plus */
extern T0* T1466c87(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_333 */
extern void T449f1556(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_not_name */
extern T0* T458f185(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_NAME.make_not */
extern T0* T1466c85(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_332 */
extern void T449f1555(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_331 */
extern void T449f1554(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].force */
extern void T1009f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].fill_with */
extern void T1011f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].extend */
extern void T1011f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].aliased_resized_area */
extern T0* T1009f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].aliased_resized_area */
extern T0* T1011f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_assigner */
extern T0* T458f184(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_330 */
extern void T449f1553(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_329 */
extern void T449f1552(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].force */
extern void T1108f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].fill_with */
extern void T1110f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].extend */
extern void T1110f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].aliased_resized_area */
extern T0* T1108f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].aliased_resized_area */
extern T0* T1110f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_external_alias */
extern T0* T458f183(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_328 */
extern void T449f1551(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_327 */
extern void T449f1550(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].force */
extern void T977f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].fill_with */
extern void T979f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].extend */
extern void T979f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_326 */
extern void T449f1549(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].aliased_resized_area */
extern T0* T977f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].aliased_resized_area */
extern T0* T979f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_325 */
extern void T449f1548(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_324 */
extern void T449f1547(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_323 */
extern void T449f1546(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].force */
extern void T1217f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].fill_with */
extern void T1218f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].extend */
extern void T1218f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].aliased_resized_area */
extern T0* T1217f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].aliased_resized_area */
extern T0* T1218f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_external_procedure */
extern T0* T449f918(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14);
/* [detachable] ET_AST_FACTORY.new_external_procedure */
extern T0* T458f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_indexing */
extern void T1465f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_alias_clause */
extern void T1465f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_postconditions */
extern void T1465f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_preconditions */
extern void T1465f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_obsolete_message */
extern void T1465f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_is_keyword */
extern void T1465f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.make */
extern T0* T1465c86(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_PROCEDURE.tokens */
extern T0* T1465f44(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_322 */
extern void T449f1545(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_321 */
extern void T449f1544(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_deferred_procedure */
extern T0* T458f182(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_indexing */
extern void T468f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_postconditions */
extern void T468f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_preconditions */
extern void T468f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_obsolete_message */
extern void T468f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_is_keyword */
extern void T468f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_320 */
extern void T449f1543(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_319 */
extern void T449f1542(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_procedure */
extern T0* T458f181(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* detachable ET_ONCE_PROCEDURE.set_first_indexing */
extern void T1464f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_rescue_clause */
extern void T1464f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_postconditions */
extern void T1464f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_compound */
extern void T1464f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_keys */
extern void T1464f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_locals */
extern void T1464f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_preconditions */
extern void T1464f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_obsolete_message */
extern void T1464f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_is_keyword */
extern void T1464f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.make */
extern T0* T1464c81(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE.tokens */
extern T0* T1464f32(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_318 */
extern void T449f1541(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_317 */
extern void T449f1540(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_procedure */
extern T0* T458f180(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* [detachable] ET_DO_PROCEDURE.set_first_indexing */
extern void T466f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_rescue_clause */
extern void T466f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_postconditions */
extern void T466f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_compound */
extern void T466f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_locals */
extern void T466f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_preconditions */
extern void T466f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_obsolete_message */
extern void T466f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_is_keyword */
extern void T466f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_316 */
extern void T449f1539(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_315 */
extern void T449f1538(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].force */
extern void T1222f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUERY].fill_with */
extern void T1223f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUERY].extend */
extern void T1223f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].aliased_resized_area */
extern T0* T1222f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUERY].aliased_resized_area */
extern T0* T1223f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_external_function */
extern T0* T449f917(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* [detachable] ET_AST_FACTORY.new_external_function */
extern T0* T458f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_indexing */
extern void T247f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_alias_clause */
extern void T247f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_postconditions */
extern void T247f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_preconditions */
extern void T247f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_obsolete_message */
extern void T247f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_assigner */
extern void T247f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_is_keyword */
extern void T247f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.make */
extern T0* T247c88(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_EXTERNAL_FUNCTION.tokens */
extern T0* T247f51(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_314 */
extern void T449f1537(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_313 */
extern void T449f1536(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_312 */
extern void T449f1535(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_311 */
extern void T449f1534(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_deferred_function */
extern T0* T458f178(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* detachable ET_DEFERRED_FUNCTION.set_first_indexing */
extern void T1462f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_postconditions */
extern void T1462f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_preconditions */
extern void T1462f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_obsolete_message */
extern void T1462f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_assigner */
extern void T1462f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_is_keyword */
extern void T1462f91(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.make */
extern T0* T1462c79(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DEFERRED_FUNCTION.make */
extern void T1462f79p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DEFERRED_FUNCTION.tokens */
extern T0* T1462f47(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_310 */
extern void T449f1533(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_309 */
extern void T449f1532(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_308 */
extern void T449f1531(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_307 */
extern void T449f1530(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_function */
extern T0* T458f177(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16, T0* a17, T0* a18);
/* detachable ET_ONCE_FUNCTION.set_first_indexing */
extern void T1461f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_rescue_clause */
extern void T1461f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_postconditions */
extern void T1461f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_compound */
extern void T1461f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_keys */
extern void T1461f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_locals */
extern void T1461f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_preconditions */
extern void T1461f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_obsolete_message */
extern void T1461f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_assigner */
extern void T1461f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_is_keyword */
extern void T1461f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.make */
extern T0* T1461c82(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ONCE_FUNCTION.tokens */
extern T0* T1461f38(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_306 */
extern void T449f1529(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_305 */
extern void T449f1528(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_304 */
extern void T449f1527(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_303 */
extern void T449f1526(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_function */
extern T0* T458f175(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16, T0* a17);
/* detachable ET_DO_FUNCTION.set_first_indexing */
extern void T1460f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_rescue_clause */
extern void T1460f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_postconditions */
extern void T1460f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_compound */
extern void T1460f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_locals */
extern void T1460f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_preconditions */
extern void T1460f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_obsolete_message */
extern void T1460f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_assigner */
extern void T1460f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_is_keyword */
extern void T1460f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.make */
extern T0* T1460c80(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DO_FUNCTION.tokens */
extern T0* T1460f49(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_302 */
extern void T449f1525(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_301 */
extern void T449f1524(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_300 */
extern void T449f1523(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_299 */
extern void T449f1522(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_unique_attribute */
extern T0* T458f174(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_assigner */
extern void T365f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.make */
extern T0* T365c80(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.make */
extern T0* T366c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.make_leaf */
extern void T366f65(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unique_keyword */
extern T0* T72f1518(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_unique */
extern T0* T409c64(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.is_keyword */
extern T0* T72f711(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_is */
extern T0* T409c71(GE_context* ac);
/* [detachable] ET_UNIQUE_ATTRIBUTE.tokens */
extern T0* T365f61(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_298 */
extern void T449f1521(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_297 */
extern void T449f1520(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_constant_attribute */
extern T0* T458f173(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_assigner */
extern void T363f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.make */
extern T0* T363c79(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CONSTANT_ATTRIBUTE.tokens */
extern T0* T363f59(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_296 */
extern void T449f1519(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_295 */
extern void T449f1518(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_extended_attribute */
extern T0* T458f172(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_indexing */
extern void T329f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_rescue_clause */
extern void T329f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_postconditions */
extern void T329f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_compound */
extern void T329f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_locals */
extern void T329f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_preconditions */
extern void T329f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_obsolete_message */
extern void T329f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_assigner */
extern void T329f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.make */
extern T0* T329c79(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTENDED_ATTRIBUTE.make */
extern void T329f79p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTENDED_ATTRIBUTE.tokens */
extern T0* T329f59(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_294 */
extern void T449f1517(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_attribute */
extern T0* T458f171(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ATTRIBUTE.set_assigner */
extern void T367f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.make */
extern T0* T367c76(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTRIBUTE.tokens */
extern T0* T367f54(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_293 */
extern void T449f1516(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_292 */
extern void T449f1515(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_procedure_synonym */
extern void T449f2186(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].back */
extern void T694f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_back */
extern void T694f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].add_traversing_cursor */
extern void T694f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove_traversing_cursor */
extern void T694f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_left */
extern void T694f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_left_cursor */
extern void T694f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].put */
extern void T694f31(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_cursors_right */
extern void T694f34(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_right */
extern void T694f33(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].put_last */
extern void T694f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].index */
extern T6 T1374f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_index */
extern T6 T694f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].forth */
extern void T694f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_forth */
extern void T694f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].before */
extern T1 T694f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_before */
extern T1 T694f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].before */
extern T1 T1374f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_procedure_synonym */
extern T0* T449f915(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_291 */
extern void T449f1514(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_extended_feature_name_comma */
extern T0* T458f169(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_290 */
extern void T449f1513(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_289 */
extern void T449f1512(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_288 */
extern void T449f1511(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_procedure */
extern void T449f2185(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].finish */
extern void T694f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_finish */
extern void T694f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].off */
extern T1 T1374f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].after */
extern T1 T1374f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_287 */
extern void T449f1510(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_286 */
extern void T449f1509(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_query_synonym */
extern void T449f2184(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].back */
extern void T698f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_back */
extern void T698f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].add_traversing_cursor */
extern void T698f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove_traversing_cursor */
extern void T698f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_left */
extern void T698f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_left_cursor */
extern void T698f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].put */
extern void T698f31(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_cursors_right */
extern void T698f34(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_right */
extern void T698f33(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].put_last */
extern void T698f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].index */
extern T6 T1375f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_index */
extern T6 T698f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].forth */
extern void T698f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_forth */
extern void T698f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].before */
extern T1 T698f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_before */
extern T1 T698f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].before */
extern T1 T1375f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_query_synonym */
extern T0* T449f914(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_285 */
extern void T449f1508(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_284 */
extern void T449f1507(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_283 */
extern void T449f1506(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_282 */
extern void T449f1505(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_query */
extern void T449f2183(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].finish */
extern void T698f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_finish */
extern void T698f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].off */
extern T1 T1375f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].after */
extern T1 T1375f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_281 */
extern void T449f1504(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_280 */
extern void T449f1503(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_279 */
extern void T449f1502(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_278 */
extern void T449f1501(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_277 */
extern void T449f1500(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_276 */
extern void T449f1499(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].force */
extern void T1111f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].fill_with */
extern void T1113f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].extend */
extern void T1113f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].aliased_resized_area */
extern T0* T1111f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].aliased_resized_area */
extern T0* T1113f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_clause */
extern T0* T458f168(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_CLAUSE.make */
extern T0* T1112c4(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.feature_keyword */
extern T0* T72f732(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_feature */
extern T0* T409c92(GE_context* ac);
/* detachable ET_FEATURE_CLAUSE.tokens */
extern T0* T1112f3(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_any_clients */
extern T0* T449f913(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.new_client */
extern T0* T449f911(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_client */
extern T0* T458f337(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.current_system */
extern T0* T449f967(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_275 */
extern void T449f1498(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_274 */
extern void T449f1497(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_273 */
extern void T449f1496(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_272 */
extern void T449f1495(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].force */
extern void T1114f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].fill_with */
extern void T1116f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].extend */
extern void T1116f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_CLAUSE_LIST.put_first */
extern void T1115f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].force */
extern void T1777f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].fill_with */
extern void T1776f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].extend */
extern void T1776f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_CLAUSE_LIST.fixed_array */
extern T0* T1115f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].default_create */
extern T0* T1777c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_271 */
extern void T449f1494(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].aliased_resized_area */
extern T0* T1114f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].aliased_resized_area */
extern T0* T1116f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_clauses */
extern T0* T458f167(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_CLAUSE_LIST.make_with_capacity */
extern T0* T1115c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].make */
extern T0* T1777f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].make_empty */
extern T0* T1776c3(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_270 */
extern void T449f1493(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_features */
extern void T449f2182(GE_context* ac, T0* C);
extern T1 T449f2182ot1(T0* a1, T0** a2);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_value */
extern void T366f64(GE_context* ac, T0* C, T11 a1);
/* NATURAL_64.out */
extern T0* T11f18(GE_context* ac, T11* C);
/* [detachable] STRING_8.append_natural_64 */
extern void T17f81(GE_context* ac, T0* C, T11 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_269 */
extern void T449f1492(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_268 */
extern void T449f1491(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_267 */
extern void T449f1490(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].force */
extern void T1246f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].fill_with */
extern void T1248f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].extend */
extern void T1248f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].aliased_resized_area */
extern T0* T1246f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].aliased_resized_area */
extern T0* T1248f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_comma */
extern T0* T458f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_266 */
extern void T449f1489(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].force */
extern void T1249f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].fill_with */
extern void T1251f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].extend */
extern void T1251f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.put_first */
extern void T1346f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].force */
extern void T1677f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].fill_with */
extern void T1676f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].extend */
extern void T1676f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_265 */
extern void T449f1488(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].aliased_resized_area */
extern T0* T1249f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].aliased_resized_area */
extern T0* T1251f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_types */
extern T0* T458f165(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_BRACED_TYPE_LIST.set_left_brace */
extern void T1346f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.make_with_capacity */
extern T0* T1346c9(GE_context* ac, T6 a1);
/* [detachable] ET_BRACED_TYPE_LIST.make_with_capacity */
extern void T1346f9p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_264 */
extern void T449f1487(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_263 */
extern void T449f1486(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_262 */
extern void T449f1485(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_261 */
extern void T449f1484(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].force */
extern void T1062f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].fill_with */
extern void T1064f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].extend */
extern void T1064f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].aliased_resized_area */
extern T0* T1062f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].aliased_resized_area */
extern T0* T1064f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_procedure */
extern T0* T458f164(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONVERT_PROCEDURE.make */
extern T0* T1458c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CONVERT_PROCEDURE.tokens */
extern T0* T1458f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_260 */
extern void T449f1483(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_convert_function */
extern T0* T458f163(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONVERT_FUNCTION.make */
extern T0* T1457c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CONVERT_FUNCTION.tokens */
extern T0* T1457f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_259 */
extern void T449f1482(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].force */
extern void T1065f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].fill_with */
extern void T1067f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].extend */
extern void T1067f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].aliased_resized_area */
extern T0* T1065f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].aliased_resized_area */
extern T0* T1067f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_feature_comma */
extern T0* T458f162(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_258 */
extern void T449f1481(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].force */
extern void T1068f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].fill_with */
extern void T1070f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].extend */
extern void T1070f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONVERT_FEATURE_LIST.put_first */
extern void T1069f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].force */
extern void T1769f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].fill_with */
extern void T1768f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].extend */
extern void T1768f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONVERT_FEATURE_LIST.fixed_array */
extern T0* T1069f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].default_create */
extern T0* T1769c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_257 */
extern void T449f1480(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].aliased_resized_area */
extern T0* T1068f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].aliased_resized_area */
extern T0* T1070f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_features */
extern T0* T458f161(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_CONVERT_FEATURE_LIST.make_with_capacity */
extern T0* T1069c8(GE_context* ac, T6 a1);
/* detachable ET_CONVERT_FEATURE_LIST.make_with_capacity */
extern void T1069f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].make */
extern T0* T1769f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].make_empty */
extern T0* T1768c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.convert_keyword */
extern T0* T72f731(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_convert */
extern T0* T409c91(GE_context* ac);
/* detachable ET_CONVERT_FEATURE_LIST.tokens */
extern T0* T1069f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_256 */
extern void T449f1479(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_255 */
extern void T449f1478(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_254 */
extern void T449f1477(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_253 */
extern void T449f1476(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_252 */
extern void T449f1475(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_251 */
extern void T449f1474(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].force */
extern void T1122f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].fill_with */
extern void T1124f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].extend */
extern void T1124f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].aliased_resized_area */
extern T0* T1122f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].aliased_resized_area */
extern T0* T1124f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_name_comma */
extern T0* T458f158(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_250 */
extern void T449f1473(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].force */
extern void T1076f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR].fill_with */
extern void T1078f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR].extend */
extern void T1078f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.put_first */
extern void T1077f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.fixed_array */
extern T0* T1077f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_249 */
extern void T449f1472(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].aliased_resized_area */
extern T0* T1076f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATOR].aliased_resized_area */
extern T0* T1078f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_creator */
extern T0* T458f160(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CREATOR.make_with_capacity */
extern T0* T1077c16(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_CREATOR.make_feature_name_list_with_capacity */
extern void T1077f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR.tokens */
extern T0* T1077f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_248 */
extern void T449f1471(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_247 */
extern void T449f1470(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_246 */
extern void T449f1469(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_245 */
extern void T449f1468(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_244 */
extern void T449f1467(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_243 */
extern void T449f1466(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_242 */
extern void T449f1465(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_241 */
extern void T449f1464(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_240 */
extern void T449f1463(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_239 */
extern void T449f1462(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_238 */
extern void T449f1461(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_237 */
extern void T449f1460(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_236 */
extern void T449f1459(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_235 */
extern void T449f1458(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].force */
extern void T1079f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].fill_with */
extern void T1081f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].extend */
extern void T1081f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].aliased_resized_area */
extern T0* T1079f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].aliased_resized_area */
extern T0* T1081f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_234 */
extern void T449f1457(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.put_first */
extern void T1080f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].force */
extern void T1771f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CREATOR].fill_with */
extern void T1770f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CREATOR].extend */
extern void T1770f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR_LIST.fixed_array */
extern T0* T1080f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].default_create */
extern T0* T1771c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_233 */
extern void T449f1456(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_creators */
extern T0* T458f159(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR_LIST.make_with_capacity */
extern T0* T1080c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].make */
extern T0* T1771f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CREATOR].make_empty */
extern T0* T1770c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_232 */
extern void T449f1455(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_231 */
extern void T449f1454(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_230 */
extern void T449f1453(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_229 */
extern void T449f1452(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].force */
extern void T1170f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].fill_with */
extern void T1171f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].extend */
extern void T1171f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.put_first */
extern void T564f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.fixed_array */
extern T0* T564f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_228 */
extern void T449f1451(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].aliased_resized_area */
extern T0* T1170f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].aliased_resized_area */
extern T0* T1171f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_keyword_feature_name_list */
extern T0* T458f157(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.make_with_capacity */
extern T0* T564c8(GE_context* ac, T6 a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.make_with_capacity */
extern void T564f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.redefine_keyword */
extern T0* T72f739(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_redefine */
extern T0* T409c99(GE_context* ac);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.tokens */
extern T0* T564f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_227 */
extern void T449f1450(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_226 */
extern void T449f1449(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_225 */
extern void T449f1448(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_224 */
extern void T449f1447(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_223 */
extern void T449f1446(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_222 */
extern void T449f1445(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_221 */
extern void T449f1444(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_220 */
extern void T449f1443(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_219 */
extern void T449f1442(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_218 */
extern void T449f1441(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_217 */
extern void T449f1440(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_216 */
extern void T449f1439(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_215 */
extern void T449f1438(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_214 */
extern void T449f1437(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_213 */
extern void T449f1436(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_212 */
extern void T449f1435(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_211 */
extern void T449f1434(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].force */
extern void T1037f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENTS].fill_with */
extern void T1039f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENTS].extend */
extern void T1039f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].aliased_resized_area */
extern T0* T1037f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLIENTS].aliased_resized_area */
extern T0* T1039f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_clients */
extern T0* T458f156(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CLIENTS.set_left_brace */
extern void T1038f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.make_with_capacity */
extern T0* T1038c20(GE_context* ac, T6 a1);
/* detachable ET_CLIENTS.make_with_capacity */
extern void T1038f20p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.tokens */
extern T0* T1038f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_210 */
extern void T449f1433(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_209 */
extern void T449f1432(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_208 */
extern void T449f1431(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].force */
extern void T1035f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].fill_with */
extern void T1036f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].extend */
extern void T1036f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T1035f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T1036f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_client_comma */
extern T0* T449f912(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_client_comma */
extern T0* T458f338(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_207 */
extern void T449f1430(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_206 */
extern void T449f1429(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_205 */
extern void T449f1428(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_204 */
extern void T449f1427(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_203 */
extern void T449f1426(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_202 */
extern void T449f1425(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_none_clients */
extern T0* T458f155(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CLIENTS.make */
extern T0* T1038c17(GE_context* ac);
/* detachable ET_CLIENTS.make */
extern void T1038f17p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_201 */
extern void T449f1424(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_1_200 */
extern void T449f1203(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_200 */
extern void T449f1423(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_199 */
extern void T449f1422(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].force */
extern void T1117f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].fill_with */
extern void T1119f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].extend */
extern void T1119f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.put_first */
extern void T1118f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.fixed_array */
extern T0* T1118f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_198 */
extern void T449f1421(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].aliased_resized_area */
extern T0* T1117f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].aliased_resized_area */
extern T0* T1119f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_export */
extern T0* T458f153(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_FEATURE_EXPORT.make_with_capacity */
extern T0* T1118c10(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_FEATURE_EXPORT.make_feature_name_list_with_capacity */
extern void T1118f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_197 */
extern void T449f1420(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_196 */
extern void T449f1419(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].force */
extern void T1097f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT].fill_with */
extern void T1099f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT].extend */
extern void T1099f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].aliased_resized_area */
extern T0* T1097f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPORT].aliased_resized_area */
extern T0* T1099f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_null_export */
extern T0* T458f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_195 */
extern void T449f1418(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_194 */
extern void T449f1417(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_193 */
extern void T449f1416(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_192 */
extern void T449f1415(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_all_export */
extern T0* T458f152(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ALL_EXPORT.make */
extern T0* T1455c7(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.all_keyword */
extern T0* T72f1510(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_all */
extern T0* T409c57(GE_context* ac);
/* detachable ET_ALL_EXPORT.tokens */
extern T0* T1455f3(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_191 */
extern void T449f1414(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].force */
extern void T1100f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].fill_with */
extern void T1101f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].extend */
extern void T1101f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].aliased_resized_area */
extern T0* T1100f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].aliased_resized_area */
extern T0* T1101f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_190 */
extern void T449f1413(GE_context* ac, T0* C);
/* detachable ET_EXPORT_LIST.put_first */
extern void T563f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].force */
extern void T1728f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPORT].fill_with */
extern void T1727f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPORT].extend */
extern void T1727f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPORT_LIST.fixed_array */
extern T0* T563f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].default_create */
extern T0* T1728c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_189 */
extern void T449f1412(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_exports */
extern T0* T458f151(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_EXPORT_LIST.make_with_capacity */
extern T0* T563c7(GE_context* ac, T6 a1);
/* detachable ET_EXPORT_LIST.make_with_capacity */
extern void T563f7p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].make */
extern T0* T1728f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPORT].make_empty */
extern T0* T1727c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.export_keyword */
extern T0* T72f726(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_export */
extern T0* T409c86(GE_context* ac);
/* detachable ET_EXPORT_LIST.tokens */
extern T0* T563f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_188 */
extern void T449f1411(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_187 */
extern void T449f1410(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_186 */
extern void T449f1409(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_185 */
extern void T449f1408(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_184 */
extern void T449f1407(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_183 */
extern void T449f1406(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].force */
extern void T1224f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].fill_with */
extern void T1226f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].extend */
extern void T1226f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].aliased_resized_area */
extern T0* T1224f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].aliased_resized_area */
extern T0* T1226f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_rename_comma */
extern T0* T458f150(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_182 */
extern void T449f1405(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_rename */
extern T0* T458f149(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_181 */
extern void T449f1404(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].force */
extern void T1227f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].fill_with */
extern void T1228f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].extend */
extern void T1228f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RENAME_LIST.put_first */
extern void T562f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].force */
extern void T1725f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].fill_with */
extern void T1724f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].extend */
extern void T1724f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RENAME_LIST.fixed_array */
extern T0* T562f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].default_create */
extern T0* T1725c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_180 */
extern void T449f1403(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].aliased_resized_area */
extern T0* T1227f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].aliased_resized_area */
extern T0* T1228f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_renames */
extern T0* T458f148(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_RENAME_LIST.make_with_capacity */
extern T0* T562c8(GE_context* ac, T6 a1);
/* detachable ET_RENAME_LIST.make_with_capacity */
extern void T562f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].make */
extern T0* T1725f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].make_empty */
extern T0* T1724c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.rename_keyword */
extern T0* T72f728(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_rename */
extern T0* T409c88(GE_context* ac);
/* detachable ET_RENAME_LIST.tokens */
extern T0* T562f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_179 */
extern void T449f1402(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_178 */
extern void T449f1401(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_177 */
extern void T449f1400(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_176 */
extern void T449f1399(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_175 */
extern void T449f1398(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].force */
extern void T1208f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].fill_with */
extern void T1209f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].extend */
extern void T1209f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].aliased_resized_area */
extern T0* T1208f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].aliased_resized_area */
extern T0* T1209f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parent_semicolon */
extern T0* T458f147(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_174 */
extern void T449f1397(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].force */
extern void T1210f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].fill_with */
extern void T1211f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].extend */
extern void T1211f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_173 */
extern void T449f1396(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_172 */
extern void T449f1395(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].aliased_resized_area */
extern T0* T1210f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].aliased_resized_area */
extern T0* T1211f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parents */
extern T0* T458f146(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_171 */
extern void T449f1394(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_170 */
extern void T449f1393(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].force */
extern void T1203f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT].fill_with */
extern void T1204f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT].extend */
extern void T1204f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].aliased_resized_area */
extern T0* T1203f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT].aliased_resized_area */
extern T0* T1204f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_parent */
extern T0* T449f910(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* [detachable] ET_AST_FACTORY.new_parent */
extern T0* T458f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_CLASS.has_formal_parameter */
extern T1 T73f193(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.has_formal_parameter */
extern T1 T593f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_169 */
extern void T449f1392(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_168 */
extern void T449f1391(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_167 */
extern void T449f1390(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_166 */
extern void T449f1389(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_165 */
extern void T449f1388(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_164 */
extern void T449f1387(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_163 */
extern void T449f1386(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.set_clients_clause */
extern void T558f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_162 */
extern void T449f1385(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_161 */
extern void T449f1384(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].force */
extern void T1205f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].fill_with */
extern void T1207f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].extend */
extern void T1207f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].aliased_resized_area */
extern T0* T1205f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].aliased_resized_area */
extern T0* T1207f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_160 */
extern void T449f1383(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.put_first */
extern void T1206f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].force */
extern void T1787f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].fill_with */
extern void T1786f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].extend */
extern void T1786f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_CLAUSE_LIST.fixed_array */
extern T0* T1206f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].default_create */
extern T0* T1787c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_159 */
extern void T449f1382(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_parent_clauses */
extern T0* T458f145(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENT_CLAUSE_LIST.make_with_capacity */
extern T0* T1206c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].make */
extern T0* T1787f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].make_empty */
extern T0* T1786c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_158 */
extern void T449f1381(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_157 */
extern void T449f1380(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_156 */
extern void T449f1379(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].force */
extern void T1197f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].fill_with */
extern void T1199f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].extend */
extern void T1199f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].aliased_resized_area */
extern T0* T1197f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].aliased_resized_area */
extern T0* T1199f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_obsolete_message */
extern T0* T458f144(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_155 */
extern void T449f1378(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_154 */
extern void T449f1377(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].force */
extern void T1047f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T1049f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].extend */
extern void T1049f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1047f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1049f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_actual_parameter_semicolon */
extern T0* T458f143(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_actual_parameter */
extern T0* T458f141(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.make */
extern T0* T1453c6(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_153 */
extern void T449f1376(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_152 */
extern void T449f1375(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].force */
extern void T1050f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].fill_with */
extern void T1052f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].extend */
extern void T1052f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.add_to_constraint_actual_parameter_list */
extern void T449f2181(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.put_first */
extern void T1051f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].force */
extern void T1766f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T1764f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].extend */
extern void T1764f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.fixed_array */
extern T0* T1051f8(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T1766c2(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_comma_actual_parameter */
extern T0* T458f142(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.make */
extern T0* T1454c6(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.first */
extern T0* T1051f7(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.is_empty */
extern T1 T1051f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_151 */
extern void T449f1374(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_150 */
extern void T449f1373(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_149 */
extern void T449f1372(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1050f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1052f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_actual_parameters */
extern T0* T458f139(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.set_left_bracket */
extern void T1051f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern T0* T1051c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern void T1051f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make */
extern T0* T1766f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T1764c4(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.tokens */
extern T0* T1051f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_148 */
extern void T449f1371(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_147 */
extern void T449f1370(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_146 */
extern void T449f1369(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_145 */
extern void T449f1368(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_144 */
extern void T449f1367(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_143 */
extern void T449f1366(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_constraint_actual_parameter_comma */
extern T0* T458f140(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_142 */
extern void T449f1365(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_constraint_named_type */
extern T0* T449f909(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_constraint_named_type */
extern T0* T458f333(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_NAMED_TYPE.make */
extern T0* T1452c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_constraint_generic_named_type */
extern T0* T458f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.make */
extern T0* T1703c8(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_141 */
extern void T449f1364(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_140 */
extern void T449f1363(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_139 */
extern void T449f1362(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_138 */
extern void T449f1361(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_137 */
extern void T449f1360(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_136 */
extern void T449f1359(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_135 */
extern void T449f1358(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].force */
extern void T1059f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].fill_with */
extern void T1061f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].extend */
extern void T1061f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].aliased_resized_area */
extern T0* T1059f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].aliased_resized_area */
extern T0* T1061f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_134 */
extern void T449f1357(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_133 */
extern void T449f1356(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_132 */
extern void T449f1355(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_131 */
extern void T449f1354(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_130 */
extern void T449f1353(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_129 */
extern void T449f1352(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_128 */
extern void T449f1351(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_127 */
extern void T449f1350(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_126 */
extern void T449f1349(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_125 */
extern void T449f1348(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_124 */
extern void T449f1347(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_123 */
extern void T449f1346(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_122 */
extern void T449f1345(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_121 */
extern void T449f1344(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_120 */
extern void T449f1343(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_119 */
extern void T449f1342(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_118 */
extern void T449f1341(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_117 */
extern void T449f1340(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_116 */
extern void T449f1339(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_115 */
extern void T449f1338(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_114 */
extern void T449f1337(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_113 */
extern void T449f1336(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_112 */
extern void T449f1335(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_111 */
extern void T449f1334(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_110 */
extern void T449f1333(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_109 */
extern void T449f1332(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_108 */
extern void T449f1331(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_107 */
extern void T449f1330(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_106 */
extern void T449f1329(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_105 */
extern void T449f1328(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_104 */
extern void T449f1327(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_103 */
extern void T449f1326(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_102 */
extern void T449f1325(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_101 */
extern void T449f1324(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_100 */
extern void T449f1323(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_99 */
extern void T449f1322(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_98 */
extern void T449f1321(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_97 */
extern void T449f1320(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_96 */
extern void T449f1319(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_95 */
extern void T449f1318(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_94 */
extern void T449f1317(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_93 */
extern void T449f1316(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_92 */
extern void T449f1315(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_91 */
extern void T449f1314(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].force */
extern void T1053f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].fill_with */
extern void T1055f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].extend */
extern void T1055f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.put_first */
extern void T1054f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.fixed_array */
extern T0* T1054f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_90 */
extern void T449f1313(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].aliased_resized_area */
extern T0* T1053f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].aliased_resized_area */
extern T0* T1055f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_creator */
extern T0* T458f137(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_CREATOR.make_with_capacity */
extern T0* T1054c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.make_with_capacity */
extern void T1054f12p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.tokens */
extern T0* T1054f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_89 */
extern void T449f1312(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_88 */
extern void T449f1311(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_87 */
extern void T449f1310(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_86 */
extern void T449f1309(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_85 */
extern void T449f1308(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_84 */
extern void T449f1307(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_83 */
extern void T449f1306(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].force */
extern void T1056f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].fill_with */
extern void T1058f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].extend */
extern void T1058f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.put_first */
extern void T1057f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.fixed_array */
extern T0* T1057f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_82 */
extern void T449f1305(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].aliased_resized_area */
extern T0* T1056f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].aliased_resized_area */
extern T0* T1058f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_renames */
extern T0* T458f136(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern T0* T1057c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern void T1057f12p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern void T1057f12p0(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.tokens */
extern T0* T1057f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_81 */
extern void T449f1304(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_80 */
extern void T449f1303(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_79 */
extern void T449f1302(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_78 */
extern void T449f1301(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_77 */
extern void T449f1300(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].force */
extern void T1240f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].fill_with */
extern void T1242f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].extend */
extern void T1242f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].aliased_resized_area */
extern T0* T1240f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].aliased_resized_area */
extern T0* T1242f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_constraint_comma */
extern T0* T458f135(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_76 */
extern void T449f1299(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].force */
extern void T1243f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].fill_with */
extern void T1245f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].extend */
extern void T1245f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.put_first */
extern void T1244f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].force */
extern void T1790f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].fill_with */
extern void T1789f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].extend */
extern void T1789f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.fixed_array */
extern T0* T1244f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].default_create */
extern T0* T1790c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_75 */
extern void T449f1298(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].aliased_resized_area */
extern T0* T1243f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].aliased_resized_area */
extern T0* T1245f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_constraint_list */
extern T0* T458f134(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_TYPE_CONSTRAINT_LIST.set_left_brace */
extern void T1244f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern T0* T1244c11(GE_context* ac, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern void T1244f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].make */
extern T0* T1790f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].make_empty */
extern T0* T1789c4(GE_context* ac, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.tokens */
extern T0* T1244f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_74 */
extern void T449f1297(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_73 */
extern void T449f1296(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_72 */
extern void T449f1295(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].force */
extern void T1237f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].fill_with */
extern void T1239f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].extend */
extern void T1239f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1237f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1239f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.register_constraint */
extern void T449f2180(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].force_last */
extern void T925f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].resize */
extern void T925f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].new_capacity */
extern T6 T925f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].extendible */
extern T1 T925f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_rename_constraint */
extern T0* T458f133(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TYPE_RENAME_CONSTRAINT.make */
extern T0* T1451c8(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.dummy_constraint */
extern T0* T449f908(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_71 */
extern void T449f1294(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_70 */
extern void T449f1293(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].force */
extern void T1132f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].fill_with */
extern void T1133f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].extend */
extern void T1133f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].aliased_resized_area */
extern T0* T1132f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].aliased_resized_area */
extern T0* T1133f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constrained_formal_parameter */
extern T0* T458f132(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_type_mark */
extern void T1450f78(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.make */
extern T0* T1450c77(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.make_unconstrained */
extern void T1450f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.arrow_symbol */
extern T0* T72f704(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_arrow */
extern T0* T343c97(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_69 */
extern void T449f1292(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_68 */
extern void T449f1291(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_67 */
extern void T449f1290(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_66 */
extern void T449f1289(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_65 */
extern void T449f1288(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_64 */
extern void T449f1287(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_formal_parameter */
extern T0* T458f131(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER.set_type_mark */
extern void T867f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.make */
extern T0* T867c78(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_63 */
extern void T449f1286(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_62 */
extern void T449f1285(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_61 */
extern void T449f1284(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].force */
extern void T1134f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].fill_with */
extern void T1136f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].extend */
extern void T1136f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1134f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1136f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_parameter_comma */
extern T0* T458f130(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_60 */
extern void T449f1283(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].force */
extern void T1137f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].fill_with */
extern void T1138f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].extend */
extern void T1138f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.put_first */
extern void T593f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_index */
extern void T1450f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER.set_index */
extern void T867f80(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.capacity */
extern T6 T593f9(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.put_first */
extern void T593f28p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].force */
extern void T1733f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].fill_with */
extern void T1732f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].extend */
extern void T1732f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.fixed_array */
extern T0* T593f8(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].default_create */
extern T0* T1733c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_59 */
extern void T449f1282(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1137f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1138f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_parameters */
extern T0* T458f129(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_FORMAL_PARAMETER_LIST.set_left_bracket */
extern void T593f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.make_with_capacity */
extern T0* T593c26(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.make_with_capacity */
extern void T593f26p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].make */
extern T0* T1733f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].make_empty */
extern T0* T1732c4(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.tokens */
extern T0* T593f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_58 */
extern void T449f1281(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_57 */
extern void T449f1280(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_formal_parameters */
extern void T449f2177(GE_context* ac, T0* C, T0* a1);
extern T1 T449f2177ot1(T0* a1, T0** a2);
extern T1 T449f2177ot2(T0* a1, T0** a2);
extern T1 T449f2177ot3(T0* a1, T0** a2);
extern T1 T449f2177ot4(T0* a1, T0** a2);
extern T1 T449f2177ot5(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].wipe_out */
extern void T925f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].keep_head */
extern void T1061f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].set_count */
extern void T1061f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].move_all_cursors_after */
extern void T925f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].set_next_cursor */
extern void T1597f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].set_after */
extern void T1597f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_formal_parameters */
extern void T73f305(GE_context* ac, T0* C, T0* a1);
extern T1 T73f305ot1(T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.put */
extern void T1244f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_TYPE_CONSTRAINT_LIST.remove */
extern void T1244f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1790f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1789f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].set_count */
extern void T1789f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.item */
extern T0* T1244f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_constraint */
extern void T1450f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.remove */
extern void T593f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1733f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1732f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].set_count */
extern void T1732f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.set_type */
extern void T1451f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].item */
extern T0* T925f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_56 */
extern void T449f1279(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_55 */
extern void T449f1278(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_54 */
extern void T449f1277(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_53 */
extern void T449f1276(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_52 */
extern void T449f1275(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_51 */
extern void T449f1274(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_50 */
extern void T449f1273(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].force */
extern void T1033f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLASS].fill_with */
extern void T1034f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLASS].extend */
extern void T1034f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].aliased_resized_area */
extern T0* T1033f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLASS].aliased_resized_area */
extern T0* T1034f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.set_external_keyword */
extern void T73f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_frozen_keyword */
extern void T73f301(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_class_mark */
extern void T73f316(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_class_keyword */
extern void T73f300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.new_class */
extern T0* T449f906(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_name */
extern void T73f317(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.file_system */
extern T0* T449f966(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.unix_file_system */
extern T0* T449f976(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.windows_file_system */
extern T0* T449f975(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.operating_system */
extern T0* T449f974(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvscn1a_error */
extern void T51f263(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvscn1a */
extern T0* T759c872(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvscn1a_error */
extern void T35f263(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MASTER_CLASS.first_local_class */
extern T0* T560f28(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_local_class */
extern T1 T560f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_hidden_class */
extern T1 T560f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_ignored_class */
extern T1 T560f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].has */
extern T1 T455f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_non_override_class */
extern T1 T560f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_override_class */
extern T1 T560f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_49 */
extern void T449f1272(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_48 */
extern void T449f1271(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_47 */
extern void T449f1270(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_46 */
extern void T449f1269(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].force */
extern void T1153f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].fill_with */
extern void T1155f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].extend */
extern void T1155f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].aliased_resized_area */
extern T0* T1153f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].aliased_resized_area */
extern T0* T1155f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_indexing_term_comma */
extern T0* T458f128(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_45 */
extern void T449f1268(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].force */
extern void T1151f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].fill_with */
extern void T1152f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].extend */
extern void T1152f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].aliased_resized_area */
extern T0* T1151f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].aliased_resized_area */
extern T0* T1152f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_custom_attribute */
extern T0* T458f127(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CUSTOM_ATTRIBUTE.make */
extern T0* T1449c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CUSTOM_ATTRIBUTE.tokens */
extern T0* T1449f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_44 */
extern void T449f1267(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_43 */
extern void T449f1266(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_42 */
extern void T449f1265(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_41 */
extern void T449f1264(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_40 */
extern void T449f1263(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_39 */
extern void T449f1262(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_38 */
extern void T449f1261(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_37 */
extern void T449f1260(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].force */
extern void T1156f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].fill_with */
extern void T1158f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].extend */
extern void T1158f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_TERM_LIST.put_first */
extern void T1157f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].force */
extern void T1782f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].fill_with */
extern void T1781f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].extend */
extern void T1781f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_TERM_LIST.fixed_array */
extern T0* T1157f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].default_create */
extern T0* T1782c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_36 */
extern void T449f1259(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].aliased_resized_area */
extern T0* T1156f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].aliased_resized_area */
extern T0* T1158f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_indexing_terms */
extern T0* T458f126(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INDEXING_TERM_LIST.make_with_capacity */
extern T0* T1157c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].make */
extern T0* T1782f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].make_empty */
extern T0* T1781c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_35 */
extern void T449f1258(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].force */
extern void T1148f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].fill_with */
extern void T1150f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].extend */
extern void T1150f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_indexing_semicolon */
extern T0* T458f124(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_34 */
extern void T449f1257(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_33 */
extern void T449f1256(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].aliased_resized_area */
extern T0* T1148f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].aliased_resized_area */
extern T0* T1150f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_tagged_indexing */
extern T0* T458f123(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TAGGED_INDEXING.make */
extern T0* T1447c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_32 */
extern void T449f1255(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_indexing */
extern T0* T458f125(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING.make */
extern T0* T1448c5(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_31 */
extern void T449f1254(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_30 */
extern void T449f1253(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].force */
extern void T1145f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].fill_with */
extern void T1147f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].extend */
extern void T1147f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].aliased_resized_area */
extern T0* T1145f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].aliased_resized_area */
extern T0* T1147f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_29 */
extern void T449f1252(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.put_first */
extern void T1146f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].force */
extern void T1780f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].fill_with */
extern void T1779f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].extend */
extern void T1779f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_LIST.fixed_array */
extern T0* T1146f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].default_create */
extern T0* T1780c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_28 */
extern void T449f1251(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_27 */
extern void T449f1250(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_26 */
extern void T449f1249(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_indexings */
extern T0* T458f121(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_INDEXING_LIST.set_indexing_keyword */
extern void T1146f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_LIST.make_with_capacity */
extern T0* T1146c11(GE_context* ac, T6 a1);
/* detachable ET_INDEXING_LIST.make_with_capacity */
extern void T1146f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].make */
extern T0* T1780f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].make_empty */
extern T0* T1779c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.indexing_keyword */
extern T0* T72f737(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_indexing */
extern T0* T409c97(GE_context* ac);
/* detachable ET_INDEXING_LIST.tokens */
extern T0* T1146f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_25 */
extern void T449f1248(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_24 */
extern void T449f1247(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_23 */
extern void T449f1246(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_22 */
extern void T449f1245(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_21 */
extern void T449f1244(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_20 */
extern void T449f1243(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_19 */
extern void T449f1242(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_18 */
extern void T449f1241(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_17 */
extern void T449f1240(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_16 */
extern void T449f1239(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_15 */
extern void T449f1238(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_14 */
extern void T449f1237(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_13 */
extern void T449f1236(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_12 */
extern void T449f1235(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_11 */
extern void T449f1234(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_10 */
extern void T449f1233(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_9 */
extern void T449f1232(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_8 */
extern void T449f1231(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_providers */
extern void T449f2171(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].forth */
extern void T926f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].put_last */
extern void T926f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].compress */
extern void T926f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_keep_head */
extern void T926f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_keep_head */
extern void T926f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].move_all_cursors */
extern void T926f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_put */
extern void T926f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_for_iteration */
extern T0* T926f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].after */
extern T1 T926f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_after */
extern T1 T926f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].start */
extern void T926f39(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_providers */
extern void T73f304(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_7 */
extern void T449f1230(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_to_end */
extern void T449f2170(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_CLASS.set_parsed */
extern void T73f274(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_end_keyword */
extern void T73f315(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_second_indexing */
extern void T73f314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_invariants */
extern void T73f313(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_feature_clauses */
extern void T73f312(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_convert_features */
extern void T73f311(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_creators */
extern void T73f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_parent_clauses */
extern void T73f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_obsolete_message */
extern void T73f308(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_6 */
extern void T449f1229(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_5 */
extern void T449f1228(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_4 */
extern void T449f1227(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_3 */
extern void T449f1226(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_first_indexing */
extern void T73f307(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_2 */
extern void T449f1225(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_leading_break */
extern void T73f306(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_1 */
extern void T449f1224(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_push_last_value */
extern void T449f1179(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].force */
extern void T982f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].fill_with */
extern void T984f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].extend */
extern void T984f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].aliased_resized_area */
extern T0* T982f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].aliased_resized_area */
extern T0* T984f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].force */
extern void T980f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].fill_with */
extern void T981f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].extend */
extern void T981f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].aliased_resized_area */
extern T0* T980f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].aliased_resized_area */
extern T0* T981f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].force */
extern void T974f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VOID].fill_with */
extern void T976f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VOID].extend */
extern void T976f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].aliased_resized_area */
extern T0* T974f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_VOID].aliased_resized_area */
extern T0* T976f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].force */
extern void T971f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].fill_with */
extern void T973f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].extend */
extern void T973f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].aliased_resized_area */
extern T0* T971f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].aliased_resized_area */
extern T0* T973f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].force */
extern void T968f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].fill_with */
extern void T970f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].extend */
extern void T970f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].aliased_resized_area */
extern T0* T968f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].aliased_resized_area */
extern T0* T970f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].force */
extern void T966f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RESULT].fill_with */
extern void T967f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RESULT].extend */
extern void T967f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].aliased_resized_area */
extern T0* T966f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RESULT].aliased_resized_area */
extern T0* T967f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].aliased_resized_area */
extern T0* T963f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].aliased_resized_area */
extern T0* T965f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].force */
extern void T958f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].fill_with */
extern void T960f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].extend */
extern void T960f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].aliased_resized_area */
extern T0* T958f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].aliased_resized_area */
extern T0* T960f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].aliased_resized_area */
extern T0* T955f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].aliased_resized_area */
extern T0* T957f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].aliased_resized_area */
extern T0* T953f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].aliased_resized_area */
extern T0* T954f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].force */
extern void T950f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].fill_with */
extern void T952f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].extend */
extern void T952f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].aliased_resized_area */
extern T0* T950f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].aliased_resized_area */
extern T0* T952f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].force */
extern void T948f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CURRENT].fill_with */
extern void T949f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CURRENT].extend */
extern void T949f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].aliased_resized_area */
extern T0* T948f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CURRENT].aliased_resized_area */
extern T0* T949f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].force */
extern void T942f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BREAK].fill_with */
extern void T944f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BREAK].extend */
extern void T944f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].aliased_resized_area */
extern T0* T942f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BREAK].aliased_resized_area */
extern T0* T944f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].aliased_resized_area */
extern T0* T939f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].aliased_resized_area */
extern T0* T941f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].force */
extern void T937f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSITION].fill_with */
extern void T938f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSITION].extend */
extern void T938f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].aliased_resized_area */
extern T0* T937f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_POSITION].aliased_resized_area */
extern T0* T938f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].force */
extern void T932f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].fill_with */
extern void T934f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].extend */
extern void T934f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].aliased_resized_area */
extern T0* T932f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].aliased_resized_area */
extern T0* T934f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].force */
extern void T929f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].fill_with */
extern void T931f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].extend */
extern void T931f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].aliased_resized_area */
extern T0* T929f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].aliased_resized_area */
extern T0* T931f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.accept */
extern void T449f1178(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action */
extern void T449f1176(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.report_error */
extern void T449f1202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1600_1799 */
extern void T449f1201(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.report_eof_expected_error */
extern void T449f1223(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1400_1599 */
extern void T449f1200(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1200_1399 */
extern void T449f1199(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1000_1199 */
extern void T449f1198(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_800_999 */
extern void T449f1197(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_600_799 */
extern void T449f1196(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_400_599 */
extern void T449f1195(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_200_399 */
extern void T449f1194(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_0_199 */
extern void T449f1193(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.read_token */
extern void T449f1175(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_execute_action */
extern void T449f1192(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.less */
extern void T449f1222(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.process_break */
extern void T449f1211(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_comment */
extern T0* T458f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_break */
extern T0* T458f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_underscored_real_constant */
extern T0* T458f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_position */
extern void T1440f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.make */
extern T0* T1440c41(GE_context* ac, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.make_leaf */
extern void T1440f46(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_literal */
extern T0* T449f986(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.text_substring */
extern T0* T449f991(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_regular_real_constant */
extern T0* T458f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_position */
extern void T1439f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_REGULAR_REAL_CONSTANT.make */
extern T0* T1439c41(GE_context* ac, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.make_leaf */
extern void T1439f46(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_binary_integer_constant */
extern T0* T458f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_position */
extern void T1438f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_binary_integer_constant */
extern T0* T449f985(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.make */
extern T0* T1438c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_BINARY_INTEGER_CONSTANT.make_leaf */
extern void T1438f71(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.compute_natural_64_value */
extern void T449f2205(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_AST_FACTORY.new_octal_integer_constant */
extern T0* T458f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_position */
extern void T1437f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_octal_integer_constant */
extern T0* T449f984(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.make */
extern T0* T1437c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_OCTAL_INTEGER_CONSTANT.make_leaf */
extern void T1437f71(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_hexadecimal_integer_constant */
extern T0* T458f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_position */
extern void T1436f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_hexadecimal_integer_constant */
extern T0* T449f983(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.make */
extern T0* T1436c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.make_leaf */
extern void T1436f71(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_underscored_integer_constant */
extern T0* T458f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_position */
extern void T1435f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_underscored_integer_constant */
extern T0* T449f982(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.make */
extern T0* T1435c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.make_leaf */
extern void T1435f71(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_regular_integer_constant */
extern T0* T458f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_position */
extern void T366f71(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_regular_integer_constant */
extern T0* T449f981(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssel_error */
extern void T51f257(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssel_error */
extern void T35f257(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sscu_error */
extern void T51f254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sscu_error */
extern void T35f254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssac_error */
extern void T51f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssac_error */
extern void T35f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssas_error */
extern void T51f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssas_error */
extern void T35f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_verbatim_string */
extern T0* T458f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* detachable ET_VERBATIM_STRING.set_position */
extern void T1434f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_verbatim_string */
extern T0* T449f980(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* detachable ET_VERBATIM_STRING.make */
extern T0* T1434c50(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T1 a6);
/* detachable ET_VERBATIM_STRING.make_leaf */
extern void T1434f57(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_utf8_literal */
extern T0* T449f977(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.is_verbatim_string_closer */
extern T1 T449f750(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.utf8_text */
extern T0* T449f748(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.append_utf8_text_to_string */
extern void T449f2204(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.append_utf8_substring_to_string */
extern void T1429f16(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.utf8_text_substring */
extern T0* T449f745(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.append_utf8_text_substring_to_string */
extern void T449f2203(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_special_manifest_string */
extern T0* T458f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_position */
extern void T1433f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_special_manifest_string */
extern T0* T449f979(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.make */
extern T0* T1433c46(GE_context* ac, T0* a1, T0* a2, T1 a3);
/* detachable ET_SPECIAL_MANIFEST_STRING.make_leaf */
extern void T1433f53(GE_context* ac, T0* C);
/* NATURAL_64.to_natural_32 */
extern T10 T11f19(GE_context* ac, T11* C);
/* [detachable] ET_AST_FACTORY.new_regular_manifest_string */
extern T0* T458f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.set_position */
extern void T863f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_REGULAR_MANIFEST_STRING.make */
extern T0* T863c45(GE_context* ac, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.make_leaf */
extern void T863f46(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_regular_manifest_string */
extern void T449f1221(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_c3_character_constant */
extern T0* T458f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_position */
extern void T1432f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_c3_character_constant */
extern T0* T449f978(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.make */
extern T0* T1432c47(GE_context* ac, T0* a1, T3 a2, T1 a3);
/* detachable ET_C3_CHARACTER_CONSTANT.make_leaf */
extern void T1432f54(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_c2_character_constant */
extern void T449f1220(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_AST_FACTORY.new_c2_character_constant */
extern T0* T458f120(GE_context* ac, T0* C, T2 a1, T0* a2);
/* detachable ET_C2_CHARACTER_CONSTANT.set_position */
extern void T1445f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_C2_CHARACTER_CONSTANT.make */
extern T0* T1445c47(GE_context* ac, T2 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.make_leaf */
extern void T1445f51(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scsc_error */
extern void T51f262(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scsc_error */
extern void T35f262(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sccu_error */
extern void T51f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sccu_error */
extern void T35f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.process_c1_character_constant */
extern void T449f1219(GE_context* ac, T0* C, T3 a1);
/* [detachable] ET_AST_FACTORY.new_c1_character_constant */
extern T0* T458f119(GE_context* ac, T0* C, T3 a1, T0* a2);
/* detachable ET_C1_CHARACTER_CONSTANT.set_position */
extern void T1444f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_C1_CHARACTER_CONSTANT.make */
extern T0* T1444c46(GE_context* ac, T3 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.make_leaf */
extern void T1444f50(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.unicode_text_item */
extern T3 T449f737(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_free_operator */
extern T0* T458f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_position */
extern void T951f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FREE_OPERATOR.make_prefix */
extern T0* T951c46(GE_context* ac, T0* a1);
/* detachable ET_FREE_OPERATOR.string_ */
extern T0* T951f10(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.make_token */
extern void T951f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.make_leaf */
extern void T951f51(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_start_condition */
extern void T449f1212(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.more */
extern void T449f1218(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_identifier */
extern void T449f1217(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_detachable_keyword */
extern T0* T458f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.set_position */
extern void T409f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.use_detachable_keyword */
extern T1 T449f880(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_reference_keyword */
extern T0* T458f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_reference */
extern T0* T409c67(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.use_reference_keyword */
extern T1 T449f878(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_keyword */
extern T0* T458f116(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.set_position */
extern void T933f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_invariant_keyword */
extern T0* T458f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_attribute_keyword */
extern T0* T458f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.use_attribute_keyword */
extern T1 T449f874(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_undefine_keyword */
extern T0* T458f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.undefine_keyword */
extern T0* T72f740(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_undefine */
extern T0* T409c100(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_separate_keyword */
extern T0* T458f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_redefine_keyword */
extern T0* T458f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_obsolete_keyword */
extern T0* T458f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.obsolete_keyword */
extern T0* T72f738(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_obsolete */
extern T0* T409c98(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_indexing_keyword */
extern T0* T458f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_external_keyword */
extern T0* T458f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.external_keyword */
extern T0* T72f736(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_external */
extern T0* T409c96(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_expanded_keyword */
extern T0* T458f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_expanded */
extern T0* T409c66(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_deferred_keyword */
extern T0* T458f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_creation_keyword */
extern T0* T458f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.creation_keyword */
extern T0* T72f735(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_creation */
extern T0* T409c95(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_attached_keyword */
extern T0* T458f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.use_attached_keyword */
extern T1 T449f863(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_variant_keyword */
extern T0* T458f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_require_keyword */
extern T0* T458f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_inspect_keyword */
extern T0* T458f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.inspect_keyword */
extern T0* T72f733(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_inspect */
extern T0* T409c93(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_inherit_keyword */
extern T0* T458f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_implies_keyword */
extern T0* T458f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.set_position */
extern void T959f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_KEYWORD_OPERATOR.make_implies */
extern T0* T959c52(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_feature_keyword */
extern T0* T458f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_convert_keyword */
extern T0* T458f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_current_keyword */
extern T0* T458f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.set_position */
extern void T273f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_unique_keyword */
extern T0* T458f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_select_keyword */
extern T0* T458f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.select_keyword */
extern T0* T72f730(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_select */
extern T0* T409c90(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_result_keyword */
extern T0* T458f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_rescue_keyword */
extern T0* T458f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_rename_keyword */
extern T0* T458f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_prefix_keyword */
extern T0* T458f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_frozen_keyword */
extern T0* T458f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_frozen */
extern T0* T409c63(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_export_keyword */
extern T0* T458f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_ensure_keyword */
extern T0* T458f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_elseif_keyword */
extern T0* T458f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.elseif_keyword */
extern T0* T72f725(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_elseif */
extern T0* T409c85(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_create_keyword */
extern T0* T458f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_assign_keyword */
extern T0* T458f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.assign_keyword */
extern T0* T72f724(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_assign */
extern T0* T409c84(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_across_keyword */
extern T0* T458f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_until_keyword */
extern T0* T458f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.until_keyword */
extern T0* T72f723(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_until */
extern T0* T409c83(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_strip_keyword */
extern T0* T458f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_retry_keyword */
extern T0* T458f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.set_position */
extern void T969f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_RETRY_INSTRUCTION.make */
extern T0* T969c13(GE_context* ac);
/* detachable ET_RETRY_INSTRUCTION.make_token */
extern void T969f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.make_leaf */
extern void T969f16(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.tokens */
extern T0* T969f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_local_keyword */
extern T0* T458f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_infix_keyword */
extern T0* T458f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_false_keyword */
extern T0* T458f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.set_position */
extern void T1443f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FALSE_CONSTANT.make */
extern T0* T1443c39(GE_context* ac);
/* detachable ET_FALSE_CONSTANT.make_token */
extern void T1443f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.make_leaf */
extern void T1443f43(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.tokens */
extern T0* T1443f12(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_debug_keyword */
extern T0* T458f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_class_keyword */
extern T0* T458f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_check_keyword */
extern T0* T458f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_alias_keyword */
extern T0* T458f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_agent_keyword */
extern T0* T458f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_KEYWORD.set_position */
extern void T930f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_when_keyword */
extern T0* T458f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_void_keyword */
extern T0* T458f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.set_position */
extern void T975f31(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_VOID.make */
extern T0* T975c30(GE_context* ac);
/* detachable ET_VOID.make_token */
extern void T975f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.make_leaf */
extern void T975f34(GE_context* ac, T0* C);
/* detachable ET_VOID.tokens */
extern T0* T975f12(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_true_keyword */
extern T0* T458f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.set_position */
extern void T1442f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_then_keyword */
extern T0* T458f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_some_keyword */
extern T0* T458f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_some */
extern T0* T409c59(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_note_keyword */
extern T0* T458f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.note_keyword */
extern T0* T72f715(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_note */
extern T0* T409c75(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.use_note_keyword */
extern T1 T449f822(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_keyword */
extern T0* T458f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_like_keyword */
extern T0* T458f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_from_keyword */
extern T0* T458f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_else_keyword */
extern T0* T458f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_xor_keyword */
extern T0* T458f60(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_xor */
extern T0* T959c51(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_old_keyword */
extern T0* T458f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_not_keyword */
extern T0* T458f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_not */
extern T0* T959c50(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_end_keyword */
extern T0* T458f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_all_keyword */
extern T0* T458f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_and_keyword */
extern T0* T458f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_or_keyword */
extern T0* T458f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_is_keyword */
extern T0* T458f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_if_keyword */
extern T0* T458f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.if_keyword */
extern T0* T72f710(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_if */
extern T0* T409c70(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_do_keyword */
extern T0* T458f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_as_keyword */
extern T0* T458f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_once_keyword */
extern T0* T458f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.process_two_char_symbol */
extern void T449f1216(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] ET_AST_FACTORY.new_assign_attempt_symbol */
extern T0* T458f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_assign_symbol */
extern T0* T458f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_dotdot_symbol */
extern T0* T458f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_arrow_symbol */
extern T0* T458f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_array_symbol */
extern T0* T458f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.set_position */
extern void T343f83(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_le_symbol */
extern T0* T458f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_position */
extern void T972f84(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_SYMBOL_OPERATOR.make_le */
extern T0* T972c94(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_right_array_symbol */
extern T0* T458f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_ge_symbol */
extern T0* T458f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_ge */
extern T0* T972c93(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_mod_symbol */
extern T0* T458f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_mod */
extern T0* T972c92(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_not_tilde_symbol */
extern T0* T458f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_not_tilde */
extern T0* T343c88(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_not_equal_symbol */
extern T0* T458f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_not_equal */
extern T0* T343c87(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_div_symbol */
extern T0* T458f38(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_div */
extern T0* T972c91(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.text_count */
extern T6 T449f729(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_set_line_column */
extern void T449f1215(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_one_char_symbol */
extern void T449f1214(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_AST_FACTORY.new_tilde_symbol */
extern T0* T458f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_tilde */
extern T0* T343c86(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_dollar_symbol */
extern T0* T458f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_right_bracket_symbol */
extern T0* T458f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_bracket_symbol */
extern T0* T458f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.set_position */
extern void T610f110(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_right_brace_symbol */
extern T0* T458f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_brace_symbol */
extern T0* T458f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_right_parenthesis_symbol */
extern T0* T458f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_parenthesis_symbol */
extern T0* T458f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_question_mark_symbol */
extern T0* T458f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.set_position */
extern void T983f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_bang_symbol */
extern T0* T458f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_colon_symbol */
extern T0* T458f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_comma_symbol */
extern T0* T458f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.comma_symbol */
extern T0* T72f699(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_comma */
extern T0* T343c92(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_semicolon_symbol */
extern T0* T458f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.semicolon_symbol */
extern T0* T72f698(GE_context* ac, T0* C);
/* detachable ET_SEMICOLON_SYMBOL.make */
extern T0* T978c4(GE_context* ac);
/* detachable ET_SEMICOLON_SYMBOL.make_leaf */
extern void T978f5(GE_context* ac, T0* C);
/* detachable ET_SEMICOLON_SYMBOL.tokens */
extern T0* T978f3(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_dot_symbol */
extern T0* T458f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.dot_symbol */
extern T0* T72f697(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dot */
extern T0* T343c91(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_lt_symbol */
extern T0* T458f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_gt_symbol */
extern T0* T458f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_equal_symbol */
extern T0* T458f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_equal */
extern T0* T343c82(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_power_symbol */
extern T0* T458f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_power */
extern T0* T972c88(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_divide_symbol */
extern T0* T458f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_divide */
extern T0* T972c87(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_times_symbol */
extern T0* T458f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_times */
extern T0* T972c86(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_plus_symbol */
extern T0* T458f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_plus */
extern T0* T972c85(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_minus_symbol */
extern T0* T458f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_minus */
extern T0* T972c83(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.text_item */
extern T2 T449f728(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_execute_eof_action */
extern void T449f1191(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.terminate */
extern void T449f1210(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.wrap */
extern T1 T449f691(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_refill_input_buffer */
extern void T449f1190(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_set_content */
extern void T449f1005(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.fill */
extern void T914f26(GE_context* ac, T0* C);
/* CHARACTER_8.to_character_8 */
extern T2 T2f29(GE_context* ac, T2* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_iso_8859_1_stream */
extern T6 T1429f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_iso_8859_1_substring */
extern void T1429f20(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_utf8_stream */
extern T6 T1429f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] YY_UNICODE_FILE_BUFFER.compact_left */
extern void T914f28(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.move_left */
extern void T1429f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.resize */
extern void T1429f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.default_capacity */
extern T6 T914f19(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_index */
extern void T914f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_null_trans_state */
extern T6 T449f690(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_previous_state */
extern T6 T449f689(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.fatal_error */
extern void T449f1189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.report_invalid_unicode_character_error */
extern void T449f1188(GE_context* ac, T0* C, T10 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_at_beginning_of_line */
extern T6 T449f679(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.special_integer_ */
extern T0* T449f702(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_init_value_stacks */
extern void T449f1174(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_clear_all */
extern void T449f1183(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.clear_all */
extern void T449f1208(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.clear_stacks */
extern void T449f2168(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_clear_value_stacks */
extern void T449f2199(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_WRITABLE].keep_head */
extern void T1262f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WRITABLE].set_count */
extern void T1262f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].keep_head */
extern void T1260f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].set_count */
extern void T1260f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].keep_head */
extern void T1257f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].set_count */
extern void T1257f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].keep_head */
extern void T1254f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].set_count */
extern void T1254f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].keep_head */
extern void T1251f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].set_count */
extern void T1251f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].keep_head */
extern void T1248f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].set_count */
extern void T1248f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].keep_head */
extern void T1245f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].set_count */
extern void T1245f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1242f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].set_count */
extern void T1242f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].keep_head */
extern void T1239f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].set_count */
extern void T1239f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].keep_head */
extern void T1236f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].set_count */
extern void T1236f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].keep_head */
extern void T1234f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].set_count */
extern void T1234f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].keep_head */
extern void T1231f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].set_count */
extern void T1231f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].keep_head */
extern void T1228f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].set_count */
extern void T1228f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].keep_head */
extern void T1226f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].set_count */
extern void T1226f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].keep_head */
extern void T1223f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].set_count */
extern void T1223f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].keep_head */
extern void T1221f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].set_count */
extern void T1221f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].keep_head */
extern void T1218f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].set_count */
extern void T1218f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].keep_head */
extern void T1216f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].set_count */
extern void T1216f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].keep_head */
extern void T1214f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].set_count */
extern void T1214f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].keep_head */
extern void T1211f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].set_count */
extern void T1211f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].keep_head */
extern void T1209f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].set_count */
extern void T1209f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].keep_head */
extern void T1207f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].set_count */
extern void T1207f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].keep_head */
extern void T1204f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].set_count */
extern void T1204f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].keep_head */
extern void T1202f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].set_count */
extern void T1202f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].keep_head */
extern void T1199f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].set_count */
extern void T1199f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].keep_head */
extern void T1196f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].set_count */
extern void T1196f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].keep_head */
extern void T1194f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].set_count */
extern void T1194f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].keep_head */
extern void T1191f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].set_count */
extern void T1191f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].keep_head */
extern void T1188f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].set_count */
extern void T1188f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].keep_head */
extern void T1185f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].set_count */
extern void T1185f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].keep_head */
extern void T1180f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].set_count */
extern void T1180f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].keep_head */
extern void T1177f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].set_count */
extern void T1177f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].keep_head */
extern void T1174f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].set_count */
extern void T1174f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].keep_head */
extern void T1171f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].set_count */
extern void T1171f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].keep_head */
extern void T1169f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].set_count */
extern void T1169f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].keep_head */
extern void T1166f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].set_count */
extern void T1166f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].keep_head */
extern void T1163f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].set_count */
extern void T1163f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].keep_head */
extern void T1160f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].set_count */
extern void T1160f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].keep_head */
extern void T1158f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].set_count */
extern void T1158f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].keep_head */
extern void T1155f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].set_count */
extern void T1155f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].keep_head */
extern void T1152f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].set_count */
extern void T1152f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].keep_head */
extern void T1150f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].set_count */
extern void T1150f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].keep_head */
extern void T1147f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].set_count */
extern void T1147f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].keep_head */
extern void T1144f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].set_count */
extern void T1144f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].keep_head */
extern void T1141f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].set_count */
extern void T1141f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].keep_head */
extern void T1138f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].set_count */
extern void T1138f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1136f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].set_count */
extern void T1136f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].keep_head */
extern void T1133f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].set_count */
extern void T1133f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].keep_head */
extern void T1129f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].set_count */
extern void T1129f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].keep_head */
extern void T1126f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].set_count */
extern void T1126f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].keep_head */
extern void T1124f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].set_count */
extern void T1124f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].keep_head */
extern void T1121f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].set_count */
extern void T1121f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].keep_head */
extern void T1119f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].set_count */
extern void T1119f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].keep_head */
extern void T1116f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].set_count */
extern void T1116f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].keep_head */
extern void T1113f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].set_count */
extern void T1113f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].keep_head */
extern void T1110f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].set_count */
extern void T1110f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].keep_head */
extern void T1107f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].set_count */
extern void T1107f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].keep_head */
extern void T1105f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].set_count */
extern void T1105f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].keep_head */
extern void T1103f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].set_count */
extern void T1103f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].keep_head */
extern void T1101f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].set_count */
extern void T1101f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].keep_head */
extern void T1099f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].set_count */
extern void T1099f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].keep_head */
extern void T1096f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].set_count */
extern void T1096f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].keep_head */
extern void T1093f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].set_count */
extern void T1093f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].keep_head */
extern void T1090f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].set_count */
extern void T1090f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].keep_head */
extern void T1087f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].set_count */
extern void T1087f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].keep_head */
extern void T1084f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].set_count */
extern void T1084f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].keep_head */
extern void T1081f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].set_count */
extern void T1081f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].keep_head */
extern void T1078f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].set_count */
extern void T1078f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].keep_head */
extern void T1075f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].set_count */
extern void T1075f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].keep_head */
extern void T1072f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].set_count */
extern void T1072f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].keep_head */
extern void T1070f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].set_count */
extern void T1070f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].keep_head */
extern void T1067f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].set_count */
extern void T1067f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].keep_head */
extern void T1064f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].set_count */
extern void T1064f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].keep_head */
extern void T1058f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].set_count */
extern void T1058f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].keep_head */
extern void T1055f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].set_count */
extern void T1055f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].keep_head */
extern void T1052f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].set_count */
extern void T1052f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].keep_head */
extern void T1049f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].set_count */
extern void T1049f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].keep_head */
extern void T1046f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].set_count */
extern void T1046f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].keep_head */
extern void T1044f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].set_count */
extern void T1044f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].keep_head */
extern void T1041f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].set_count */
extern void T1041f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].keep_head */
extern void T1039f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].set_count */
extern void T1039f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].keep_head */
extern void T1036f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].set_count */
extern void T1036f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].keep_head */
extern void T1034f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].set_count */
extern void T1034f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].keep_head */
extern void T1032f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].set_count */
extern void T1032f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].keep_head */
extern void T1029f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].set_count */
extern void T1029f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].keep_head */
extern void T1026f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].set_count */
extern void T1026f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].keep_head */
extern void T1023f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].set_count */
extern void T1023f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].keep_head */
extern void T1020f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].set_count */
extern void T1020f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].keep_head */
extern void T646f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].set_count */
extern void T646f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].keep_head */
extern void T1017f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].set_count */
extern void T1017f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].keep_head */
extern void T1014f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].set_count */
extern void T1014f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].keep_head */
extern void T1011f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].set_count */
extern void T1011f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].keep_head */
extern void T1008f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].set_count */
extern void T1008f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].keep_head */
extern void T1006f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].set_count */
extern void T1006f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].keep_head */
extern void T1004f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].set_count */
extern void T1004f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].keep_head */
extern void T1001f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].set_count */
extern void T1001f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].keep_head */
extern void T998f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].set_count */
extern void T998f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].keep_head */
extern void T996f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].set_count */
extern void T996f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].keep_head */
extern void T994f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].set_count */
extern void T994f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].keep_head */
extern void T992f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].set_count */
extern void T992f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].keep_head */
extern void T990f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].set_count */
extern void T990f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].keep_head */
extern void T987f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].set_count */
extern void T987f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].keep_head */
extern void T984f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].set_count */
extern void T984f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].keep_head */
extern void T981f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].set_count */
extern void T981f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].keep_head */
extern void T979f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].set_count */
extern void T979f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].keep_head */
extern void T976f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].set_count */
extern void T976f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].keep_head */
extern void T973f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].set_count */
extern void T973f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].keep_head */
extern void T970f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].set_count */
extern void T970f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].keep_head */
extern void T967f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].set_count */
extern void T967f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].keep_head */
extern void T965f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].set_count */
extern void T965f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].keep_head */
extern void T962f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].set_count */
extern void T962f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].keep_head */
extern void T960f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].set_count */
extern void T960f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].keep_head */
extern void T957f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].set_count */
extern void T957f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].keep_head */
extern void T954f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].set_count */
extern void T954f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].keep_head */
extern void T952f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].set_count */
extern void T952f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].keep_head */
extern void T949f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].set_count */
extern void T949f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].keep_head */
extern void T947f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].set_count */
extern void T947f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].keep_head */
extern void T944f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].set_count */
extern void T944f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].keep_head */
extern void T941f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].set_count */
extern void T941f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].keep_head */
extern void T938f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].set_count */
extern void T938f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].keep_head */
extern void T934f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].set_count */
extern void T934f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].keep_head */
extern void T931f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].set_count */
extern void T931f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.abort */
extern void T449f1177(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_load_input_buffer */
extern void T449f1003(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_file */
extern void T914f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_file_with_size */
extern void T914f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_UNICODE_FILE_BUFFER.flush */
extern void T914f29(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.flush */
extern void T914f29p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T449f1173(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].wipe_out */
extern void T924f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].move_all_cursors_after */
extern void T924f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].wipe_out */
extern void T919f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].wipe_out */
extern void T918f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].wipe_out */
extern void T915f10(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_end_of_file */
extern void T914f24(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T449f1173p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T449f1173p0(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T449f1173p2(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.tmp_file */
extern T0* T449f716(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.shared_file */
extern T0* T449f960(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.error_handler */
extern T0* T449f715(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_fatal_error */
extern void T449f1167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.process_none_class */
extern void T449f1166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_class */
extern void T380f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_identifier */
extern void T1284f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_identifier */
extern void T394f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_identifier */
extern void T390f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_identifier */
extern void T376f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_cursor_validity */
extern void T376f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.current_closure_impl */
extern T0* T376f55(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_across_cursor */
extern void T376f275(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.copy_type_context */
extern void T237f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.set_root_context */
extern void T237f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].found_item */
extern T0* T391f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].found */
extern T1 T391f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].search */
extern void T391f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].search_position */
extern void T391f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].hash_position */
extern T6 T391f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.hash_code */
extern T6 T989f18(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.hash_code */
extern T6 T986f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].key_storage_item */
extern T0* T391f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].slots_item */
extern T6 T391f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen9b_error */
extern void T51f411(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen9b */
extern T0* T759c1011(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen9_error */
extern T1 T51f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen9b_error */
extern void T35f411(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_veen9_error */
extern T1 T35f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen9a_error */
extern void T51f410(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen9a */
extern T0* T759c1010(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen9a_error */
extern void T35f410(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.across_component */
extern T0* T392f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].item */
extern T0* T723f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.has_implementation_error */
extern T1 T376f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_local_validity */
extern void T376f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_OBJECT_TEST_LIST.object_test */
extern T0* T1431f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.report_object_test_local */
extern void T376f272(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].found_item */
extern T0* T388f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].found */
extern T1 T388f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].search */
extern void T388f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].search_position */
extern void T388f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].hash_position */
extern T6 T388f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.hash_code */
extern T6 T715f23(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.hash_code */
extern T6 T1501f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_item */
extern T0* T388f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_item */
extern T6 T388f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen8b_error */
extern void T51f409(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen8b */
extern T0* T759c1009(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen8_error */
extern T1 T51f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen8b_error */
extern void T35f409(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_veen8_error */
extern T1 T35f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen8a_error */
extern void T51f408(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen8a */
extern T0* T759c1008(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen8a_error */
extern void T35f408(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST_SCOPE.object_test */
extern T0* T389f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].item */
extern T0* T718f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_local_variable_validity */
extern void T376f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_local_variable */
extern void T376f281(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0a_error */
extern void T51f440(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0a */
extern T0* T759c1040(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vevi_error */
extern T1 T51f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vevi0a_error */
extern void T35f440(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vevi_error */
extern T1 T35f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_expanded */
extern T1 T237f43(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_detachable */
extern T1 T237f37(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE.has_local_variable */
extern T1 T393f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_attached */
extern T1 T237f36(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2c_error */
extern void T51f415(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2c */
extern T0* T759c1015(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen2_error */
extern T1 T51f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen2c_error */
extern void T35f415(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_veen2_error */
extern T1 T35f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2e_error */
extern void T51f414(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2e */
extern T0* T759c1014(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2e_error */
extern void T35f414(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_formal_argument_validity */
extern void T376f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_formal_argument */
extern void T376f280(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE.has_formal_argument */
extern T1 T393f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_identifier */
extern void T70f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_cursor_validity */
extern void T70f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_closure_impl */
extern T0* T70f72(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_across_cursor */
extern void T70f246(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.has_implementation_error */
extern T1 T70f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_local_validity */
extern void T70f233(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test_local */
extern void T70f245(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_variable_validity */
extern void T70f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_variable */
extern void T70f244(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_cap_dynamic_types */
extern void T70f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_dynamic_type_set */
extern void T70f146(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.put */
extern void T255f16(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dynamic_type_set */
extern T0* T70f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_argument_validity */
extern void T70f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_formal_argument */
extern void T70f243(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_identifier */
extern void T65f1005(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_closed_operand */
extern void T65f1011(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_agent_closed_operand */
extern T1 T180f33(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_open_operand */
extern void T65f1010(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_agent_open_operand */
extern T1 T180f32(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_test_local */
extern void T65f1009(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_test_local_name */
extern void T65f1014(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_current */
extern void T1284f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_current */
extern void T394f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_current */
extern void T390f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_current */
extern void T376f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_current_validity */
extern void T376f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_current */
extern void T376f295(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0f_error */
extern void T51f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0f */
extern T0* T759c1035(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vucr_error */
extern T1 T51f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vucr0f_error */
extern void T35f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vucr_error */
extern T1 T35f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_current */
extern void T70f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_current_validity */
extern void T70f310(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_current */
extern void T70f370(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_current */
extern void T65f1006(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_current */
extern void T65f1012(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_qualified_call_expression */
extern void T1284f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_actual_argument_list */
extern void T1284f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_qualified_call_expression */
extern void T394f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_qualified_call_expression */
extern void T390f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_qualified_call_expression */
extern void T376f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_expression_validity */
extern void T376f202(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f202ot1(T0* a1, T0** a2);
extern T1 T376f202ot2(T0* a1, T0** a2);
extern T1 T376f202ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.free_adapted_base_classes */
extern void T376f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].force_last */
extern void T399f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].force */
extern void T735f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].fill_with */
extern void T736f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].extend */
extern void T736f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].resize */
extern void T399f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].aliased_resized_area */
extern T0* T735f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].aliased_resized_area */
extern T0* T736f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].new_capacity */
extern T6 T399f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].extendible */
extern T1 T399f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].wipe_out */
extern void T689f14(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].keep_head */
extern void T1372f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].set_count */
extern void T1372f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_all_cursors_after */
extern void T689f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_next_cursor */
extern void T1373f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_after */
extern void T1373f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0f_error */
extern void T51f401(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0f */
extern T0* T759c1001(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgmc_error */
extern T1 T51f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0f_error */
extern void T35f401(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vgmc_error */
extern T1 T35f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0e_error */
extern void T51f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0e */
extern T0* T759c993(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0e_error */
extern void T35f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_context */
extern T1 T237f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_context_with_type_marks */
extern T1 T237f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NESTED_TYPE_CONTEXT.tokens */
extern T0* T237f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].item */
extern T0* T689f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].is_empty */
extern T1 T689f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex2a_error */
extern void T51f400(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuex2a */
extern T0* T759c1000(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuex2_error */
extern T1 T51f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuex2a_error */
extern void T35f400(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vuex2_error */
extern T1 T35f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_tuple_label_call_expression_validity */
extern void T376f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_expression */
extern void T376f327(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1c_error */
extern void T51f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1c */
extern T0* T759c1044(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar1_error */
extern T1 T51f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar1c_error */
extern void T35f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar1_error */
extern T1 T35f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuta2b_error */
extern void T51f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuta2b */
extern T0* T759c1043(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuta2_error */
extern T1 T51f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuta2b_error */
extern void T35f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuta2_error */
extern T1 T35f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_tuple_label_parenthesis_call_validity */
extern void T376f261(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T376f261ot1(T0* a1, T0** a2);
extern T1 T376f261ot2(T0* a1, T0** a2);
extern T1 T376f261ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_expression_validity */
extern void T376f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f322ot1(T0* a1, T0** a2);
extern T1 T376f322ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_instruction_validity */
extern void T376f321(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T376f321ot1(T0* a1, T0** a2);
extern T1 T376f321ot2(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1a_error */
extern void T51f486(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1a */
extern T0* T759c1086(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vkcn1_error */
extern T1 T51f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1a_error */
extern void T35f486(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vkcn1_error */
extern T1 T35f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_procedure_call_instruction_validity */
extern void T376f363(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_call_instruction */
extern void T376f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_validity */
extern void T376f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.reset_fatal_error */
extern void T376f326(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FEATURE_CHECKER.check_actual_arguments_validity */
extern void T376f330(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f330ot1(T0* a1, T0** a2);
extern T1 T376f330ot2(T0* a1);
extern T1 T376f330ot3(T0* a1, T0** a2);
extern T1 T376f330ot4(T0* a1);
extern T1 T376f330ot5(T0* a1, T0** a2);
extern T1 T376f330ot6(T0* a1, T0** a2);
extern T1 T376f330ot7(T0* a1, T0** a2);
extern T1 T376f330ot8(T0* a1, T0** a2);
extern T1 T376f330ot9(T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar2b_error */
extern void T51f494(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vuar2b */
extern T0* T759c1094(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar2_error */
extern T1 T51f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar2b_error */
extern void T35f494(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar2_error */
extern T1 T35f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar2a_error */
extern void T51f493(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vuar2a */
extern T0* T759c1093(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vuar2a_error */
extern void T35f493(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr4b_error */
extern void T51f492(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr4b */
extern T0* T759c1092(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr4_error */
extern T1 T51f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr4b_error */
extern void T35f492(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr4_error */
extern T1 T35f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_converted_target_infix_expression_validity */
extern void T376f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_INFIX_EXPRESSION.set_left */
extern void T1495f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_type */
extern T1 T237f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INFIX_CAST_EXPRESSION.make */
extern T0* T1926c17(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_INFIX_EXPRESSION.set_right */
extern void T1495f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.put */
extern void T1924f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.put */
extern void T274f19(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.item */
extern T0* T1924f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.item */
extern T0* T274f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.convert_expression */
extern T0* T376f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_builtin_conversion */
extern void T376f385(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.make */
extern T0* T1992c18(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_call_expression */
extern void T376f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CONVERT_TO_EXPRESSION.make */
extern T0* T1994c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_creation_expression */
extern void T376f371(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_creation_vape_validity */
extern void T376f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape2a_error */
extern void T51f511(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape2a */
extern T0* T759c1111(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vape2_error */
extern T1 T51f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vape2a_error */
extern void T35f511(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vape2_error */
extern T1 T35f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.add_non_descendant_caller_clients_to */
extern void T376f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLIENT_LIST.force_first */
extern void T397f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.resize */
extern void T397f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T734f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T733f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.new_capacity */
extern T6 T397f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.capacity */
extern T6 T397f10(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.has_descendant */
extern T1 T1038f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.has_ancestor */
extern T1 T73f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.has_descendant */
extern T1 T397f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CLIENTS.is_none_or_unknown */
extern T1 T1038f10(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.is_none_or_unknown */
extern T1 T397f7(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.wipe_out */
extern void T397f16(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].keep_head */
extern void T734f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].keep_head */
extern void T733f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].set_count */
extern void T733f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.make */
extern T0* T1993c23(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.current_universe */
extern T0* T376f64(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.convert_feature */
extern T0* T380f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.convert_from_feature */
extern T0* T73f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BRACED_TYPE_LIST.has_named_type_with_type_marks */
extern T1 T1346f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TOKEN_CONSTANTS.attached_separate_type_mark */
extern T0* T72f1522(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.convert_to_feature */
extern T0* T73f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_context */
extern T1 T237f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_context_with_type_marks */
extern T1 T237f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_actual_argument_validity */
extern void T376f364(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f364ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.build_negated_scope */
extern void T394f57(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.build_negated_scope */
extern void T390f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1b_error */
extern void T51f491(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1b */
extern T0* T759c1091(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuar1b_error */
extern void T35f491(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1a_error */
extern void T51f490(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1a */
extern T0* T759c1090(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vuar1a_error */
extern void T35f490(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr4a_error */
extern void T51f489(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr4a */
extern T0* T759c1089(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vdpr4a_error */
extern void T35f489(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make */
extern T0* T1924c13(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_MANIFEST_TUPLE.make */
extern T0* T281c36(GE_context* ac);
/* [detachable] ET_MANIFEST_TUPLE.make */
extern void T281f36p1(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.put_first */
extern void T1924f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.fixed_array */
extern T0* T1924f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T1924f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T1924f16p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.tokens */
extern T0* T1924f10(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.tuple_argument_position */
extern T6 T376f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_vape_validity */
extern void T376f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape1b_error */
extern void T51f488(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape1b */
extern T0* T759c1088(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vape1_error */
extern T1 T51f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vape1b_error */
extern void T35f488(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vape1_error */
extern T1 T35f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex2b_error */
extern void T51f445(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuex2b */
extern T0* T759c1045(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vuex2b_error */
extern void T35f445(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuta2a_error */
extern void T51f487(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuta2a */
extern T0* T759c1087(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vuta2a_error */
extern void T35f487(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.set_parenthesis_call_position */
extern void T376f320(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENTHESIS_SYMBOL.set_position */
extern void T1967f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_PARENTHESIS_SYMBOL.make */
extern T0* T1967c41(GE_context* ac);
/* [detachable] ET_PARENTHESIS_SYMBOL.make_leaf */
extern void T1967f44(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.tokens */
extern T0* T1967f36(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn2a_error */
extern void T51f399(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn2a */
extern T0* T759c999(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vkcn2_error */
extern T1 T51f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vkcn2a_error */
extern void T35f399(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vkcn2_error */
extern T1 T35f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_query_call_expression_validity */
extern void T376f260(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.check_query_call_type_validity */
extern void T376f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f325ot1(T0* a1, T0** a2);
extern T1 T376f325ot2(T0* a1, T0** a2);
extern T1 T376f325ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_query_parenthesis_call_validity */
extern void T376f259(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f259ot1(T0* a1, T0** a2);
extern T1 T376f259ot2(T0* a1, T0** a2);
extern T1 T376f259ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_query_call_expression_validity */
extern void T376f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_call_expression */
extern void T376f359(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_vape_validity */
extern void T376f276(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape1a_error */
extern void T51f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vape1a */
extern T0* T759c1048(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vape1a_error */
extern void T35f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0d_error */
extern void T51f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0d */
extern T0* T759c1078(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0d_error */
extern void T35f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_dotnet_query_call_expression_validity */
extern void T376f258(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_queries */
extern void T376f319(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].force_last */
extern void T384f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].force */
extern void T697f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].fill_with */
extern void T699f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].extend */
extern void T699f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].resize */
extern void T384f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].aliased_resized_area */
extern T0* T697f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].aliased_resized_area */
extern T0* T699f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].new_capacity */
extern T6 T384f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].extendible */
extern T1 T384f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].first */
extern T0* T698f14(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.keep_best_overloaded_features */
extern void T376f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T1 a6);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].keep_first */
extern void T694f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].keep_first */
extern void T698f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_features */
extern void T376f362(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].force_last */
extern void T385f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].force */
extern void T701f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].fill_with */
extern void T703f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].extend */
extern void T703f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].resize */
extern void T385f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].aliased_resized_area */
extern T0* T701f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].aliased_resized_area */
extern T0* T703f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].new_capacity */
extern T6 T385f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].extendible */
extern T1 T385f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].has */
extern T1 T702f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].remove */
extern void T702f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_left */
extern void T702f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_cursors_left */
extern void T702f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_position */
extern void T2002f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].remove_last */
extern void T702f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_last_cursors_after */
extern void T702f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_next_cursor */
extern void T2002f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_after */
extern void T2002f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].item */
extern T0* T702f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].wipe_out */
extern void T702f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_all_cursors_after */
extern void T702f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].force_last */
extern void T702f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].resize */
extern void T702f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].new_capacity */
extern T6 T702f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].extendible */
extern T1 T702f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_features */
extern T0* T376f65(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].remove_last */
extern void T385f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].keep_head */
extern void T703f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_count */
extern void T703f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].move_last_cursors_after */
extern void T385f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_next_cursor */
extern void T704f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_after */
extern void T704f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].last */
extern T0* T385f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].make */
extern T0* T702c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].set_internal_cursor */
extern void T702f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].new_cursor */
extern T0* T702f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].make */
extern T0* T2002c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].is_empty */
extern T1 T385f9(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.adapted_name */
extern T0* T376f63(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].is_empty */
extern T1 T698f13(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_queries */
extern T0* T376f62(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].remove_last */
extern void T384f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].keep_head */
extern void T699f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_count */
extern void T699f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].move_last_cursors_after */
extern void T384f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_next_cursor */
extern void T700f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_after */
extern void T700f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].last */
extern T0* T384f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].is_empty */
extern T1 T384f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.set_boolean_operator */
extern void T1495f29(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PREFIX_EXPRESSION.set_boolean_operator */
extern void T1499f29(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.is_boolean_class */
extern T1 T73f115(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.reset_context_if_multiple_constraints */
extern void T400f11(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.like_0 */
extern T0* T72f262(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.make_zero */
extern T0* T1424c39(GE_context* ac, T0* a1);
/* [detachable] ET_LIKE_N.tokens */
extern T0* T1424f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].first */
extern T0* T689f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_orphan_actual_arguments_validity */
extern void T376f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_adapted_base_classes_validity */
extern void T376f256(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.check_adapted_base_classes_validity */
extern void T400f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T400f13ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0a_error */
extern void T51f515(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0a */
extern T0* T759c1115(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0a_error */
extern void T35f515(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0d_error */
extern void T51f514(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0d */
extern T0* T759c1114(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0d_error */
extern void T35f514(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0c_error */
extern void T51f513(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0c */
extern T0* T759c1113(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0c_error */
extern void T35f513(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0b_error */
extern void T51f512(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0b */
extern T0* T759c1112(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0b_error */
extern void T35f512(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.error_handler */
extern T0* T400f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].put_last */
extern void T689f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].remove */
extern void T689f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_left */
extern void T689f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_cursors_left */
extern void T689f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_position */
extern void T1373f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].remove_last */
extern void T689f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_last_cursors_after */
extern void T689f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].keep_first */
extern void T689f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_fatal_error */
extern void T400f14(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.new_adapted_base_classes */
extern T0* T376f56(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].remove_last */
extern void T399f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].keep_head */
extern void T736f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_count */
extern void T736f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].move_last_cursors_after */
extern void T399f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_next_cursor */
extern void T737f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_after */
extern void T737f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].last */
extern T0* T399f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].is_empty */
extern T1 T399f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_qualified_call_expression */
extern void T70f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_expression_validity */
extern void T70f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f322ot1(T0* a1, T0** a2);
extern T1 T70f322ot2(T0* a1, T0** a2);
extern T1 T70f322ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_adapted_base_classes */
extern void T70f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_tuple_label_call_expression_validity */
extern void T70f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_expression */
extern void T70f369(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f369ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_qualified_call_target_dynamic_types */
extern void T70f364(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_tuple_label_parenthesis_call_validity */
extern void T70f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f241ot1(T0* a1, T0** a2);
extern T1 T70f241ot2(T0* a1, T0** a2);
extern T1 T70f241ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_expression_validity */
extern void T70f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f307ot1(T0* a1, T0** a2);
extern T1 T70f307ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_instruction_validity */
extern void T70f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f306ot1(T0* a1, T0** a2);
extern T1 T70f306ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_procedure_call_instruction_validity */
extern void T70f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_instruction */
extern void T70f302(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.make */
extern T0* T306c11(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_validity */
extern void T70f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.reset_fatal_error */
extern void T70f363(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_actual_arguments_validity */
extern void T70f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f362ot1(T0* a1, T0** a2);
extern T1 T70f362ot2(T0* a1);
extern T1 T70f362ot3(T0* a1, T0** a2);
extern T1 T70f362ot4(T0* a1);
extern T1 T70f362ot5(T0* a1, T0** a2);
extern T1 T70f362ot6(T0* a1, T0** a2);
extern T1 T70f362ot7(T0* a1, T0** a2);
extern T1 T70f362ot8(T0* a1, T0** a2);
extern T1 T70f362ot9(T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_converted_target_infix_expression_validity */
extern void T70f423(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.convert_expression */
extern T0* T70f87(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_conversion */
extern void T70f501(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_expression */
extern void T70f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f365ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_creation_expression */
extern void T70f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_argument_operand_dynamic_types */
extern void T70f425(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_vape_validity */
extern void T70f477(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.add_non_descendant_caller_clients_to */
extern void T70f421(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_universe */
extern T0* T70f84(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_actual_argument_validity */
extern void T70f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f422ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.tuple_argument_position */
extern T6 T70f90(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_vape_validity */
extern void T70f361(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_parenthesis_call_position */
extern void T70f303(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_query_call_expression_validity */
extern void T70f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_call_type_validity */
extern void T70f366(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f366ot1(T0* a1, T0** a2);
extern T1 T70f366ot2(T0* a1, T0** a2);
extern T1 T70f366ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_parenthesis_call_validity */
extern void T70f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f240ot1(T0* a1, T0** a2);
extern T1 T70f240ot2(T0* a1, T0** a2);
extern T1 T70f240ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_query_call_expression_validity */
extern void T70f305(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_expression */
extern void T70f368(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f368ot1(T0* a1, T0** a2);
extern T1 T70f368ot2(T0* a1, T0** a2);
extern T1 T70f368ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_like_argument_dynamic_types */
extern void T70f426(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_argument_dynamic_types */
extern void T70f424(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_vape_validity */
extern void T70f367(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_dotnet_query_call_expression_validity */
extern void T70f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_queries */
extern void T70f432(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.keep_best_overloaded_features */
extern void T70f299(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T1 a6);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_features */
extern void T70f360(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_features */
extern T0* T70f85(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.adapted_name */
extern T0* T70f83(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_queries */
extern T0* T70f91(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_orphan_actual_arguments_validity */
extern void T70f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_adapted_base_classes_validity */
extern void T70f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_adapted_base_classes */
extern T0* T70f77(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_expression_validity */
extern void T70f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_expression_supplier */
extern void T70f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_qualified_call_expression */
extern void T65f1007(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_result */
extern void T1284f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_result */
extern void T394f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_result */
extern void T390f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_result */
extern void T376f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_result_validity */
extern void T376f237(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_result */
extern void T376f294(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0b_error */
extern void T51f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0b */
extern T0* T759c1034(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0b_error */
extern void T35f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE.has_result */
extern T1 T393f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2a_error */
extern void T51f420(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2a */
extern T0* T759c1020(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2a_error */
extern void T35f420(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2g_error */
extern void T51f433(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2g */
extern T0* T759c1033(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2g_error */
extern void T35f433(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2d_error */
extern void T51f418(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen2d */
extern T0* T759c1018(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_veen2d_error */
extern void T35f418(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2b_error */
extern void T51f417(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2b */
extern T0* T759c1017(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2b_error */
extern void T35f417(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_result */
extern void T70f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_result_validity */
extern void T70f317(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result */
extern void T70f374(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.process_result */
extern void T65f1008(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_result */
extern void T65f1013(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_value */
extern void T65f1015(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_DYNAMIC_PRECURSOR.is_once_per_process */
extern T1 T298f19(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_once_per_process */
extern T1 T232f20(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.once_prefixes */
extern T0* T65f461(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.once_kind */
extern T6 T65f460(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_real_64_class */
extern T1 T73f128(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_real_32_class */
extern T1 T73f127(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_pointer_class */
extern T1 T73f126(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_64_class */
extern T1 T73f125(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_32_class */
extern T1 T73f124(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_16_class */
extern T1 T73f123(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_8_class */
extern T1 T73f122(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_64_class */
extern T1 T73f121(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_32_class */
extern T1 T73f120(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_16_class */
extern T1 T73f119(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_8_class */
extern T1 T73f118(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_character_32_class */
extern T1 T73f117(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_character_8_class */
extern T1 T73f116(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].found_item */
extern T6 T283f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_item */
extern T6 T283f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].found */
extern T1 T283f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].search */
extern void T283f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_tuple */
extern void T1284f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_tuple */
extern void T394f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_tuple */
extern void T390f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_tuple */
extern void T376f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_tuple_validity */
extern void T376f245(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f245ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_tuple */
extern void T376f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_tuple */
extern void T70f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_tuple_validity */
extern void T70f338(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f338ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_tuple */
extern void T70f396(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f396ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_universe_impl */
extern T0* T70f75(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_manifest_tuple */
extern void T65f1016(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_tuple */
extern void T65f1019(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1019ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].force_last */
extern void T282f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_put */
extern void T282f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_put */
extern void T282f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_item */
extern T6 T282f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].hash_position */
extern T6 T282f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].resize */
extern void T282f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_resize */
extern void T282f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].special_integer_ */
extern T0* T282f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_resize */
extern void T282f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_resize */
extern void T282f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].aliased_resized_area */
extern T0* T651f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].aliased_resized_area */
extern T0* T650f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_item */
extern T0* T282f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_resize */
extern void T282f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_capacity */
extern T6 T282f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_put */
extern void T282f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].force */
extern void T651f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].fill_with */
extern void T650f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].extend */
extern void T650f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].search_position */
extern void T282f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_equality_tester */
extern T0* T282f30(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_procedure */
extern void T1284f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_compound */
extern void T1284f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_COMPOUND.item */
extern T0* T330f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_do_procedure */
extern void T1283f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_procedure */
extern void T376f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_procedure_validity */
extern void T376f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_rescue_validity */
extern void T376f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_instructions_validity */
extern void T376f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_locals_validity */
extern void T376f117(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_local_supplier */
extern void T376f132(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_local_supplier */
extern void T1299f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_local_variable_declaration */
extern void T376f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_local_type_validity */
extern void T376f130(GE_context* ac, T0* C, T0* a1);
extern T1 T376f130ot1(T0* a1, T0** a2);
/* [detachable] ET_TYPE_CHECKER.check_creation_type_validity */
extern void T380f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T380f25ot1(T0* a1, T0** a2);
extern T1 T380f25ot2(T0* a1, T0** a2);
extern T1 T380f25ot3(T0* a1, T0** a2);
extern T1 T380f25ot4(T0* a1, T0** a2);
extern T1 T380f25ot5(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg4a_error */
extern void T51f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg4a */
extern T0* T759c988(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtcg4_error */
extern T1 T51f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtcg4a_error */
extern void T35f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vtcg4_error */
extern T1 T35f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.current_system */
extern T0* T380f18(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.ise_6_0_6_7358 */
extern T0* T380f17(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg4b_error */
extern void T51f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg4b */
extern T0* T759c987(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vtcg4b_error */
extern void T35f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_CONSTRAINT_CREATOR.has_feature */
extern T1 T1054f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.is_empty */
extern T1 T1054f9(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.error_handler */
extern T0* T380f15(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.set_fatal_error */
extern void T380f26(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.check_type_validity */
extern void T380f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrle1a_error */
extern void T51f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrle1a */
extern T0* T759c870(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrle1_error */
extern T1 T51f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrle1a_error */
extern void T35f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrle1_error */
extern T1 T35f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrle2a_error */
extern void T51f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrle2a */
extern T0* T759c869(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrle2_error */
extern T1 T51f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrle2a_error */
extern void T35f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrle2_error */
extern T1 T35f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vreg0b_error */
extern void T51f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vreg0b */
extern T0* T759c868(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vreg0b_error */
extern void T35f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_across_components_validity */
extern void T376f116(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_object_tests_validity */
extern void T376f115(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot3a_error */
extern void T51f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuot3a */
extern T0* T759c867(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot3_error */
extern T1 T51f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot3a_error */
extern void T35f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot3_error */
extern T1 T35f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.ise_6_4_7_7252 */
extern T0* T376f51(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_formal_arguments_validity */
extern void T376f114(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_argument_supplier */
extern void T376f129(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_argument_supplier */
extern void T1299f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_formal_argument_declaration */
extern void T376f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_signature_type_validity */
extern void T376f127(GE_context* ac, T0* C, T0* a1);
extern T1 T376f127ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_procedure */
extern void T70f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_procedure_validity */
extern void T70f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_rescue_validity */
extern void T70f139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_instructions_validity */
extern void T70f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_locals_validity */
extern void T70f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_supplier */
extern void T70f145(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_variable_declaration */
extern void T70f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_attached_index */
extern void T1716f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE.set_attached_index */
extern void T1176f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_index */
extern void T1716f12(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE.set_index */
extern void T1176f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_type_validity */
extern void T70f143(GE_context* ac, T0* C, T0* a1);
extern T1 T70f143ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_components_validity */
extern void T70f136(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_tests_validity */
extern void T70f135(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_6_4_7_7252 */
extern T0* T70f67(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_arguments_validity */
extern void T70f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_argument_supplier */
extern void T70f142(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_formal_argument_declaration */
extern void T70f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_attached_index */
extern void T1715f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_attached_index */
extern void T252f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_signature_type_validity */
extern void T70f140(GE_context* ac, T0* C, T0* a1);
extern T1 T70f140ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_do_procedure */
extern void T65f1017(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure */
extern void T65f1020(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_procedure */
extern void T65f1022(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_address_routine */
extern void T65f1024(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.argument_type_set */
extern T0* T65f465(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_address_routine_name */
extern void T65f1046(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_feature_name_comment */
extern void T65f1025(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_internal_routine */
extern void T65f1023(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.print_assign_completed_to_once_status */
extern void T65f1045(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_internal_routine_body_declaration */
extern void T65f1044(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_assign_last_exception_to_once_exception */
extern void T65f1054(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_exception */
extern void T65f1050(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_compound */
extern void T65f1053(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_instruction */
extern void T65f1055(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.reset_rescue_data */
extern void T65f1052(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].wipe_out */
extern void T290f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_wipe_out */
extern void T290f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_wipe_out */
extern void T290f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_wipe_out */
extern void T290f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_wipe_out */
extern void T290f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].move_all_cursors_after */
extern void T290f57(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_malloc_current */
extern void T65f1043(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1043ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.current_universe */
extern T0* T65f474(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_feature_trace_message_call */
extern void T65f1042(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_unindented_feature_info_message_call */
extern void T65f1051(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assign_result_to_once_value */
extern void T65f1041(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_called_to_once_status */
extern void T65f1040(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_once_value */
extern void T65f1039(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_status */
extern void T65f1038(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_status_is_completed */
extern void T65f1037(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_statement */
extern void T65f1036(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assign_called_to_thread_safe_once_status */
extern void T65f1035(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_once_exception_to_thread_safe_once_exception */
extern void T65f1034(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_exception */
extern void T65f1047(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_result_to_thread_safe_once_value */
extern void T65f1033(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_value */
extern void T65f1048(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_once_value_to_result */
extern void T65f1032(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex_unlock */
extern void T65f1031(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex */
extern void T65f1049(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex_lock */
extern void T65f1030(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex_try_lock */
extern void T65f1029(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_thread_safe_once_value */
extern void T65f1028(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_raise_thread_safe_once_exception_if_any */
extern void T65f1027(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_status */
extern void T65f1026(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].force_last */
extern void T283f51(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_put */
extern void T283f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_put */
extern void T283f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_put */
extern void T283f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].resize */
extern void T283f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_resize */
extern void T283f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].special_integer_ */
extern T0* T283f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_resize */
extern void T283f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_resize */
extern void T283f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_resize */
extern void T283f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_capacity */
extern T6 T283f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_put */
extern void T283f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_deferred_procedure */
extern void T1284f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_deferred_procedure */
extern void T1283f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_deferred_procedure */
extern void T376f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_deferred_procedure_validity */
extern void T376f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_deferred_procedure */
extern void T70f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_deferred_procedure_validity */
extern void T70f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_deferred_procedure */
extern void T65f1018(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_deferred_procedure */
extern void T65f1021(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_named_object_test */
extern void T1284f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_named_object_test */
extern void T394f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_named_object_test */
extern void T390f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE.add_object_test */
extern void T389f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].force_last */
extern void T718f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].resize */
extern void T718f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].new_capacity */
extern T6 T718f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].extendible */
extern T1 T718f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1f_error */
extern void T51f519(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1f */
extern T0* T759c1118(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot1_error */
extern T1 T51f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot1f_error */
extern void T35f519(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot1_error */
extern T1 T35f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.error_handler */
extern T0* T390f7(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.set_fatal_error */
extern void T390f63(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_named_object_test */
extern void T376f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_named_object_test_validity */
extern void T376f227(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_named_object_test */
extern void T376f284(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot4b_error */
extern void T51f428(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuot4b */
extern T0* T759c1028(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot4_error */
extern T1 T51f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot4b_error */
extern void T35f428(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot4_error */
extern T1 T35f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot4a_error */
extern void T51f427(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuot4a */
extern T0* T759c1027(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vuot4a_error */
extern void T35f427(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.ise_6_3_7_5660 */
extern T0* T376f59(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1e_error */
extern void T51f426(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1e */
extern T0* T759c1026(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1e_error */
extern void T35f426(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1d_error */
extern void T51f425(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1d */
extern T0* T759c1025(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1d_error */
extern void T35f425(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].remove_last */
extern void T382f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].keep_head */
extern void T691f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].set_count */
extern void T691f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].move_last_cursors_after */
extern void T382f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].set_next_cursor */
extern void T692f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].set_after */
extern void T692f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1c_error */
extern void T51f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1c */
extern T0* T759c1024(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1c_error */
extern void T35f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1b_error */
extern void T51f423(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1b */
extern T0* T759c1023(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1b_error */
extern void T35f423(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].item */
extern T0* T382f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].force_last */
extern void T382f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].force */
extern void T690f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].fill_with */
extern void T691f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].extend */
extern void T691f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].resize */
extern void T382f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].aliased_resized_area */
extern T0* T690f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].aliased_resized_area */
extern T0* T691f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].new_capacity */
extern T6 T382f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].extendible */
extern T1 T382f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1a_error */
extern void T51f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1a */
extern T0* T759c1022(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1a_error */
extern void T35f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].force_last */
extern void T388f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_put */
extern void T388f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_put */
extern void T388f66(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_put */
extern void T388f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].resize */
extern void T388f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_resize */
extern void T388f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].special_integer_ */
extern T0* T388f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_resize */
extern void T388f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_resize */
extern void T388f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_resize */
extern void T388f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_capacity */
extern T6 T388f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_put */
extern void T388f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FEATURE_CHECKER.check_type_validity */
extern void T376f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_named_object_test */
extern void T70f287(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_named_object_test_validity */
extern void T70f349(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_named_object_test */
extern void T70f416(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_named_object_test_dynamic_types */
extern void T70f461(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_constant_expression */
extern void T70f225(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].force_last */
extern void T377f49(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_put */
extern void T377f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].force */
extern void T688f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].fill_with */
extern void T687f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].extend */
extern void T687f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_put */
extern void T377f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_put */
extern void T377f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_item */
extern T6 T377f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].hash_position */
extern T6 T377f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].resize */
extern void T377f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_resize */
extern void T377f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].special_integer_ */
extern T0* T377f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_resize */
extern void T377f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].aliased_resized_area */
extern T0* T688f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].aliased_resized_area */
extern T0* T687f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_resize */
extern void T377f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_item */
extern T0* T377f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_item */
extern T6 T377f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_resize */
extern void T377f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_capacity */
extern T6 T377f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_put */
extern void T377f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].search_position */
extern void T377f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].found_item */
extern T6 T377f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_item */
extern T6 T377f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].found */
extern T1 T377f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].search */
extern void T377f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_6_3_7_5660 */
extern T0* T70f88(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_type_validity */
extern void T70f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_named_object_test */
extern void T65f1056(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_test */
extern void T65f1095(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].force_last */
extern void T268f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].force */
extern void T632f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].fill_with */
extern void T631f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].extend */
extern void T631f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].resize */
extern void T268f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].aliased_resized_area */
extern T0* T632f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].aliased_resized_area */
extern T0* T631f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].new_capacity */
extern T6 T268f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].extendible */
extern T1 T268f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_manifest_string */
extern void T1284f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_manifest_string */
extern void T394f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_manifest_string */
extern void T390f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_manifest_string */
extern void T376f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_manifest_string_validity */
extern void T376f235(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_string_validity */
extern void T376f126(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_immutable_string_32_constant */
extern void T376f147(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_string_32_constant */
extern void T376f146(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_immutable_string_8_constant */
extern void T376f145(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2c_error */
extern void T51f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2c */
extern T0* T759c986(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvwmc2_error */
extern T1 T51f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2c_error */
extern void T35f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvwmc2_error */
extern T1 T35f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_string_8_constant */
extern void T376f144(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0d_error */
extern void T51f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0d */
extern T0* T759c985(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.type_name */
extern T0* T759f409(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwmq_error */
extern T1 T51f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0d_error */
extern void T35f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwmq_error */
extern T1 T35f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_manifest_string */
extern void T70f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_manifest_string_validity */
extern void T70f311(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_string_validity */
extern void T70f183(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_immutable_string_32_constant */
extern void T70f224(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_constant */
extern void T70f226(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_manifest_string_area_dynamic_type */
extern void T70f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_32_constant */
extern void T70f223(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_immutable_string_8_constant */
extern void T70f222(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_8_constant */
extern void T70f221(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_manifest_string */
extern void T65f1057(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_manifest_string */
extern void T65f1096(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_string */
extern void T65f1139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_utf8_as_escaped_string_8 */
extern void T65f1148(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_INTEGER_FORMATTER.put_octal_natural_32 */
extern void T349f3(GE_context* ac, T0* C, T0* a1, T10 a2);
/* [detachable] ET_C_GENERATOR.print_utf8_as_escaped_string_32 */
extern void T65f1147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_void */
extern void T1284f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_void */
extern void T394f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_void */
extern void T390f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_void */
extern void T376f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_void_validity */
extern void T376f236(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_void_constant */
extern void T376f293(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_void */
extern void T70f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_void_validity */
extern void T70f315(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_void_constant */
extern void T70f373(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_void */
extern void T65f1058(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_void */
extern void T65f1097(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_across_expression */
extern void T1284f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_loop_invariants */
extern void T1284f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.assertion */
extern T0* T1184f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOOP_INVARIANTS.item */
extern T0* T1184f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_across_expression */
extern void T394f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_across_expression */
extern void T390f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_across_expression */
extern void T376f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_expression_validity */
extern void T376f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_across_expression */
extern void T376f316(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].remove_found_item */
extern void T391f63(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].remove_position */
extern void T391f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].key_storage_put */
extern void T391f71(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_storage_put */
extern void T391f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].key_storage_keep_head */
extern void T391f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_storage_keep_head */
extern void T391f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].clashes_put */
extern void T391f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].slots_put */
extern void T391f70(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].move_cursors_forth */
extern void T391f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].internal_set_key_equality_tester */
extern void T391f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.remove_across_components */
extern void T392f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].prune_last */
extern void T723f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_loop_variant_validity */
extern void T376f315(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vave0a_error */
extern void T51f485(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vave0a */
extern T0* T759c1085(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vave_error */
extern T1 T51f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vave0a_error */
extern void T35f485(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vave_error */
extern T1 T35f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_loop_invariant_validity */
extern void T376f314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.add_across_component */
extern void T392f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].force_last */
extern void T723f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].resize */
extern void T723f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].new_capacity */
extern T6 T723f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACROSS_COMPONENT].extendible */
extern T1 T723f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_component_header_validity */
extern void T376f313(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_instruction_validity */
extern void T376f361(GE_context* ac, T0* C, T0* a1);
extern T1 T376f361ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1b_error */
extern void T51f505(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1b */
extern T0* T759c1105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1b_error */
extern void T35f505(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_INSTRUCTION.parenthesis_call */
extern T0* T1947f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_dotnet_procedure_call_instruction_validity */
extern void T376f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_procedures */
extern void T376f372(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].force_last */
extern void T383f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].force */
extern void T693f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].fill_with */
extern void T695f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].extend */
extern void T695f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].resize */
extern void T383f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].aliased_resized_area */
extern T0* T693f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].aliased_resized_area */
extern T0* T695f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].new_capacity */
extern T6 T383f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].extendible */
extern T1 T383f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].first */
extern T0* T694f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].is_empty */
extern T1 T694f13(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_procedures */
extern T0* T376f67(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].remove_last */
extern void T383f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].keep_head */
extern void T695f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_count */
extern void T695f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].move_last_cursors_after */
extern void T383f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_next_cursor */
extern void T696f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_after */
extern void T696f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].last */
extern T0* T383f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].is_empty */
extern T1 T383f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].force_last */
extern void T391f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].resize */
extern void T391f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].clashes_resize */
extern void T391f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].special_integer_ */
extern T0* T391f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].key_storage_resize */
extern void T391f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].item_storage_resize */
extern void T391f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].slots_resize */
extern void T391f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ACROSS_COMPONENT].new_capacity */
extern T6 T391f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.report_across_cursor_declaration */
extern void T376f360(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.hidden_across_component */
extern T0* T392f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2e_error */
extern void T51f484(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2e */
extern T0* T759c1084(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_voit2_error */
extern T1 T51f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_voit2e_error */
extern void T35f484(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_voit2_error */
extern T1 T35f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE.hidden_object_test */
extern T0* T389f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2d_error */
extern void T51f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2d */
extern T0* T759c1083(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2d_error */
extern void T35f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2c_error */
extern void T51f482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2c */
extern T0* T759c1082(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2c_error */
extern void T35f482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2b_error */
extern void T51f481(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2b */
extern T0* T759c1081(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2b_error */
extern void T35f481(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2a_error */
extern void T51f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2a */
extern T0* T759c1080(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2a_error */
extern void T35f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit1a_error */
extern void T51f479(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_voit1a */
extern T0* T759c1079(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_voit1_error */
extern T1 T51f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_voit1a_error */
extern void T35f479(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_voit1_error */
extern T1 T35f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_across_expression */
extern void T70f289(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_expression_validity */
extern void T70f351(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_across_expression */
extern void T70f420(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_variant_validity */
extern void T70f419(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_variant_validity */
extern void T70f419p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_invariant_validity */
extern void T70f418(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_invariant_validity */
extern void T70f418p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_component_header_validity */
extern void T70f417(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_instruction_validity */
extern void T70f230(GE_context* ac, T0* C, T0* a1);
extern T1 T70f230ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_dotnet_procedure_call_instruction_validity */
extern void T70f238(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_procedures */
extern void T70f300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_procedures */
extern T0* T70f82(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_across_cursor_declaration */
extern void T70f462(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_across_expression */
extern void T65f1059(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_across_expression */
extern void T65f1098(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_bracket_expression */
extern void T1284f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.process */
extern void T1924f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unfolded_tuple_actual_argument_list */
extern void T1284f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.process */
extern void T274f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_bracket_expression */
extern void T394f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_bracket_expression */
extern void T390f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_bracket_expression */
extern void T376f192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_bracket_expression_validity */
extern void T376f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_bracket_expression */
extern void T70f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_bracket_expression_validity */
extern void T70f335(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_bracket_expression */
extern void T65f1060(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_bracket_expression */
extern void T65f1099(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_call_agent */
extern void T1284f63(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f63ot1(T0* a1, T0** a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_argument_operand_list */
extern void T1284f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_call_agent */
extern void T394f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_call_agent */
extern void T390f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_call_agent */
extern void T376f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_call_agent_validity */
extern void T376f246(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f246ot1(T0* a1, T0** a2);
extern T1 T376f246ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_typed_call_agent_validity */
extern void T376f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f307ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0g_error */
extern void T51f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0g */
extern T0* T759c1067(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0g_error */
extern void T35f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_CALL_AGENT.is_procedure */
extern T1 T1016f22(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca1b_error */
extern void T51f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vpca1b */
extern T0* T759c1066(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca1_error */
extern T1 T51f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca1b_error */
extern void T35f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca1_error */
extern T1 T35f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_typed_tuple_label_call_agent_validity */
extern void T376f354(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_call_agent */
extern void T376f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_typed_query_call_agent_validity */
extern void T376f353(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_query_call_agent */
extern void T376f376(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_agent_arguments_validity */
extern void T376f342(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f342ot1(T0* a1, T0** a2);
extern T1 T376f342ot2(T0* a1, T0** a2);
extern T1 T376f342ot3(T0* a1, T0** a2);
extern T1 T376f342ot4(T0* a1, T0** a2);
extern T1 T376f342ot5(T0* a1, T0** a2);
extern T1 T376f342ot6(T0* a1, T0** a2);
extern T1 T376f342ot7(T0* a1, T0** a2);
extern T1 T376f342ot8(T0* a1, T0** a2);
extern T1 T376f342ot9(T0* a1, T0** a2);
extern T1 T376f342ot10(T0* a1, T0** a2);
extern T1 T376f342ot11(T0* a1, T0** a2);
extern T1 T376f342ot12(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca5b_error */
extern void T51f503(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vpca5b */
extern T0* T759c1103(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca5_error */
extern T1 T51f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca5b_error */
extern void T35f503(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca5_error */
extern T1 T35f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca5a_error */
extern void T51f502(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vpca5a */
extern T0* T759c1102(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vpca5a_error */
extern void T35f502(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca4b_error */
extern void T51f501(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vpca4b */
extern T0* T759c1101(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca4_error */
extern T1 T51f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca4b_error */
extern void T35f501(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca4_error */
extern T1 T35f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca4a_error */
extern void T51f500(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vpca4a */
extern T0* T759c1100(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vpca4a_error */
extern void T35f500(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.put */
extern void T1003f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.item */
extern T0* T1003f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca3b_error */
extern void T51f499(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vpca3b */
extern T0* T759c1099(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca3_error */
extern T1 T51f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca3b_error */
extern void T35f499(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca3_error */
extern T1 T35f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca3a_error */
extern void T51f498(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpca3a */
extern T0* T759c1098(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpca3a_error */
extern void T35f498(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.force_first */
extern void T236f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.resize */
extern void T236f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T612f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T609f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.new_capacity */
extern T6 T236f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.capacity */
extern T6 T236f23(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_implicit_agent_open_arguments */
extern void T376f341(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.put_last */
extern void T1973f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].force */
extern void T2024f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].fill_with */
extern void T2023f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].extend */
extern void T2023f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.fixed_array */
extern T0* T1973f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].default_create */
extern T0* T2024c2(GE_context* ac);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.make */
extern T0* T1974c6(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.make_with_capacity */
extern T0* T1973c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].make */
extern T0* T2024f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].make_empty */
extern T0* T2023c4(GE_context* ac, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca2a_error */
extern void T51f504(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vpca2a */
extern T0* T759c1104(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca2_error */
extern T1 T51f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca2a_error */
extern void T35f504(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca2_error */
extern T1 T35f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_typed_procedure_call_agent_validity */
extern void T376f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_procedure_call_agent */
extern void T376f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CALL_AGENT.set_procedure */
extern void T1016f35(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_agent_validity */
extern void T376f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f306ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_tuple_label_call_agent_validity */
extern void T376f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_query_call_agent_validity */
extern void T376f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_procedure_call_agent_validity */
extern void T376f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_agent_validity */
extern void T376f305(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca1a_error */
extern void T51f465(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpca1a */
extern T0* T759c1065(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vpca1a_error */
extern void T35f465(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_query_call_agent_validity */
extern void T376f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_query_call_agent */
extern void T376f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_procedure_call_agent_validity */
extern void T376f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_procedure_call_agent */
extern void T376f373(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0j_error */
extern void T51f464(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0j */
extern T0* T759c1064(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0j_error */
extern void T35f464(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_call_agent */
extern void T70f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_call_agent_validity */
extern void T70f340(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f340ot1(T0* a1, T0** a2);
extern T1 T70f340ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_call_agent_validity */
extern void T70f400(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f400ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_tuple_label_call_agent_validity */
extern void T70f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_call_agent */
extern void T70f473(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f473ot1(T0* a1, T0** a2);
extern T1 T70f473ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_agent_qualified_query_call */
extern void T70f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.is_empty */
extern T1 T255f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_query_call_agent_validity */
extern void T70f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_query_call_agent */
extern void T70f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_agent */
extern void T70f479(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f479ot1(T0* a1, T0** a2);
extern T1 T70f479ot2(T0* a1, T0** a2);
extern T1 T70f479ot3(T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_agent_closed_operands_dynamic_types */
extern void T70f482(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f482ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_agent_arguments_validity */
extern void T70f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f454ot1(T0* a1, T0** a2);
extern T1 T70f454ot2(T0* a1, T0** a2);
extern T1 T70f454ot3(T0* a1, T0** a2);
extern T1 T70f454ot4(T0* a1, T0** a2);
extern T1 T70f454ot5(T0* a1, T0** a2);
extern T1 T70f454ot6(T0* a1, T0** a2);
extern T1 T70f454ot7(T0* a1, T0** a2);
extern T1 T70f454ot8(T0* a1, T0** a2);
extern T1 T70f454ot9(T0* a1, T0** a2);
extern T1 T70f454ot10(T0* a1, T0** a2);
extern T1 T70f454ot11(T0* a1, T0** a2);
extern T1 T70f454ot12(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_implicit_agent_open_arguments */
extern void T70f453(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_procedure_call_agent_validity */
extern void T70f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_procedure_call_agent */
extern void T70f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_agent_validity */
extern void T70f399(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f399ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_tuple_label_call_agent_validity */
extern void T70f445(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_query_call_agent_validity */
extern void T70f444(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_procedure_call_agent_validity */
extern void T70f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_agent_validity */
extern void T70f398(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_query_call_agent_validity */
extern void T70f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_query_call_agent */
extern void T70f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_agent */
extern void T70f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f478ot1(T0* a1, T0** a2);
extern T1 T70f478ot2(T0* a1, T0** a2);
extern T1 T70f478ot3(T0* a1, T0** a2);
extern T1 T70f478ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_call_agent_result_dynamic_types */
extern void T70f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_procedure_call_agent_validity */
extern void T70f441(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_procedure_call_agent */
extern void T70f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.process_call_agent */
extern void T65f1061(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_call_agent */
extern void T65f1101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent */
extern void T65f1141(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1141ot1(T0* a1, T0** a2);
extern T1 T65f1141ot2(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].force_last */
extern void T271f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].force */
extern void T641f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT].fill_with */
extern void T640f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT].extend */
extern void T640f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].resize */
extern void T271f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].aliased_resized_area */
extern T0* T641f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_AGENT].aliased_resized_area */
extern T0* T640f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].new_capacity */
extern T6 T271f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].extendible */
extern T1 T271f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_call_agent_body_declaration */
extern void T65f1100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_qualified_call_instruction */
extern void T65f1140(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1140ot1(T0* a1, T0** a2);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_name */
extern void T275f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_target */
extern void T275f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_name */
extern void T278f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_if_expression */
extern void T1284f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_expression_list */
extern void T1284f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION.process */
extern void T1086f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_expression */
extern void T1284f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.item */
extern T0* T1089f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_if_expression */
extern void T394f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_if_expression */
extern void T390f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_if_expression */
extern void T376f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_if_expression_validity */
extern void T376f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.free_common_ancestor_types */
extern void T376f299(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].keep_first */
extern void T387f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_all_cursors_after */
extern void T387f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].item */
extern T0* T387f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.report_if_expression */
extern void T376f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.update_common_ancestor_type_list */
extern void T376f297(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].remove */
extern void T387f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_left */
extern void T387f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_cursors_left */
extern void T387f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_position */
extern void T711f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_if_expression */
extern void T70f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_if_expression_validity */
extern void T70f321(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_common_ancestor_types */
extern void T70f378(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_if_expression */
extern void T70f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_if_expression_dynamic_types */
extern void T70f431(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.update_common_ancestor_type_list */
extern void T70f376(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_C_GENERATOR.process_if_expression */
extern void T65f1062(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_if_expression */
extern void T65f1102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_array */
extern void T1284f60(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.expression */
extern T0* T1187f15(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.item */
extern T0* T1187f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_array */
extern void T394f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_array */
extern void T390f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_array */
extern void T376f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_array_validity */
extern void T376f200(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_untyped_manifest_array_validity */
extern void T376f255(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f255ot1(T0* a1, T0** a2);
extern T1 T376f255ot2(T0* a1, T0** a2);
/* detachable ET_MANIFEST_ARRAY.put */
extern void T1187f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_array */
extern void T376f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.conforms_to_context */
extern T1 T71f109(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.conforms_to_context_with_type_marks */
extern T1 T71f119(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.tokens */
extern T0* T71f17(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_typed_manifest_array_validity */
extern void T376f254(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f254ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwma2a_error */
extern void T51f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vwma2a */
extern T0* T759c1042(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwma2_error */
extern T1 T51f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwma2a_error */
extern void T35f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vwma2_error */
extern T1 T35f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwma1a_error */
extern void T51f441(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwma1a */
extern T0* T759c1041(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwma1_error */
extern T1 T51f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwma1a_error */
extern void T35f441(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwma1_error */
extern T1 T35f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_array */
extern void T70f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_array_validity */
extern void T70f337(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_untyped_manifest_array_validity */
extern void T70f395(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f395ot1(T0* a1, T0** a2);
extern T1 T70f395ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_array */
extern void T70f440(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f440ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_manifest_array_validity */
extern void T70f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f394ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_manifest_array */
extern void T65f1063(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_array */
extern void T65f1103(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1103ot1(T0* a1, T0** a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_parenthesized_expression */
extern void T1284f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_parenthesized_expression */
extern void T394f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_parenthesized_expression */
extern void T390f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_parenthesized_expression */
extern void T376f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_parenthesized_expression_validity */
extern void T376f201(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_parenthesized_expression */
extern void T70f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_parenthesized_expression_validity */
extern void T70f336(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_parenthesized_expression */
extern void T65f1064(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_parenthesized_expression */
extern void T65f1104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_static_call_expression */
extern void T1284f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_static_call_expression */
extern void T394f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_static_call_expression */
extern void T390f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_static_call_expression */
extern void T376f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_call_expression_validity */
extern void T376f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_static_query_call_expression_validity */
extern void T376f266(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_static_call_expression */
extern void T376f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_feature_validity */
extern void T376f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T376f331ot1(T0* a1, T0** a2);
extern T1 T376f331ot2(T0* a1, T0** a2);
extern T1 T376f331ot3(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno5a_error */
extern void T51f497(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuno5a */
extern T0* T759c1097(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuno5_error */
extern T1 T51f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuno5a_error */
extern void T35f497(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuno5_error */
extern T1 T35f147(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_implicitly_static */
extern T1 T1465f53(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin_static */
extern T1 T1465f45(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.string_ */
extern T0* T1465f49(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin */
extern T1 T1465f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_implicitly_static */
extern T1 T247f48(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin_static */
extern T1 T247f63(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.string_ */
extern T0* T247f67(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin */
extern T1 T247f58(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_implicitly_static */
extern T1 T365f48(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_implicitly_static */
extern T1 T363f47(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno5b_error */
extern void T51f496(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuno5b */
extern T0* T759c1096(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuno5b_error */
extern void T35f496(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno3a_error */
extern void T51f495(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuno3a */
extern T0* T759c1095(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuno3_error */
extern T1 T51f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuno3a_error */
extern void T35f495(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vuno3_error */
extern T1 T35f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_static_supplier */
extern void T376f328(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_static_supplier */
extern void T1299f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_parenthesis_call_validity */
extern void T376f265(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T376f265ot1(T0* a1, T0** a2);
extern T1 T376f265ot2(T0* a1, T0** a2);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_parenthesis_call */
extern void T1489f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.set_parenthesis_call */
extern void T1230f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_dotnet_query_call_expression_validity */
extern void T376f264(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_static_call_expression */
extern void T70f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_call_expression_validity */
extern void T70f323(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_query_call_expression_validity */
extern void T70f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_call_expression */
extern void T70f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_feature_validity */
extern void T70f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f434ot1(T0* a1, T0** a2);
extern T1 T70f434ot2(T0* a1, T0** a2);
extern T1 T70f434ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_supplier */
extern void T70f433(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_parenthesis_call_validity */
extern void T70f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f381ot1(T0* a1, T0** a2);
extern T1 T70f381ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_dotnet_query_call_expression_validity */
extern void T70f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_static_call_expression */
extern void T65f1065(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_static_call_expression */
extern void T65f1105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_expression_static_call */
extern void T65f1142(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T65f1142ot1(T0* a1, T0** a2);
extern T1 T65f1142ot2(T0* a1, T0** a2);
extern T1 T65f1142ot3(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.set_static_generated */
extern void T298f33(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_static_generated */
extern void T232f38(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_strip_expression */
extern void T1284f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_strip_expression */
extern void T394f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_strip_expression */
extern void T390f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_strip_expression */
extern void T376f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_strip_expression_validity */
extern void T376f204(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_strip_expression */
extern void T376f267(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst2a_error */
extern void T51f404(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwst2a */
extern T0* T759c1004(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwst2_error */
extern T1 T51f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwst2a_error */
extern void T35f404(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwst2_error */
extern T1 T35f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst1a_error */
extern void T51f403(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vwst1a */
extern T0* T759c1003(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwst1_error */
extern T1 T51f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwst1a_error */
extern void T35f403(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vwst1_error */
extern T1 T35f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst1b_error */
extern void T51f402(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwst1b */
extern T0* T759c1002(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwst1b_error */
extern void T35f402(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STRIP_EXPRESSION.feature_name */
extern T0* T1233f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.item */
extern T0* T1233f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_strip_expression */
extern void T70f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_strip_expression_validity */
extern void T70f339(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_strip_expression */
extern void T70f397(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_strip_expression */
extern void T65f1066(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_strip_expression */
extern void T65f1106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_real_constant */
extern void T1284f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_real_constant */
extern void T394f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_real_constant */
extern void T390f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_real_constant */
extern void T376f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_real_constant_validity */
extern void T376f212(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_real_constant_validity */
extern void T376f268(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_type */
extern void T1440f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_type */
extern void T1439f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_real_64_constant */
extern void T376f334(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_real_32_constant */
extern void T376f333(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0b_error */
extern void T51f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0b */
extern T0* T759c1046(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0b_error */
extern void T35f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_real_constant */
extern void T70f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_real_constant_validity */
extern void T70f312(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_real_constant_validity */
extern void T70f371(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_real_64_constant */
extern void T70f428(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_real_32_constant */
extern void T70f427(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_real_constant */
extern void T65f1067(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_real_constant */
extern void T65f1107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.is_negative */
extern T1 T1439f17(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_minus */
extern T1 T972f19(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.in_target */
extern T1 T65f492(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_underscored_real_constant */
extern void T1284f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_underscored_real_constant */
extern void T394f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_underscored_real_constant */
extern void T390f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_underscored_real_constant */
extern void T376f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_underscored_real_constant_validity */
extern void T376f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_underscored_real_constant */
extern void T70f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_underscored_real_constant_validity */
extern void T70f313(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_underscored_real_constant */
extern void T65f1068(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_underscored_real_constant */
extern void T65f1108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_negative */
extern T1 T1440f17(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_true_constant */
extern void T1284f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_true_constant */
extern void T394f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_true_constant */
extern void T390f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_true_constant */
extern void T376f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_true_constant_validity */
extern void T376f214(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_boolean_constant */
extern void T376f269(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_true_constant */
extern void T70f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_true_constant_validity */
extern void T70f314(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_boolean_constant */
extern void T70f372(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_true_constant */
extern void T65f1069(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_true_constant */
extern void T65f1109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_false_constant */
extern void T1284f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_false_constant */
extern void T394f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_false_constant */
extern void T390f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_false_constant */
extern void T376f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_false_constant_validity */
extern void T376f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_false_constant */
extern void T70f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_false_constant_validity */
extern void T70f316(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_false_constant */
extern void T65f1070(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_false_constant */
extern void T65f1110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c1_character_constant */
extern void T1284f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c1_character_constant */
extern void T394f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c1_character_constant */
extern void T390f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c1_character_constant */
extern void T376f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c1_character_constant_validity */
extern void T376f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_character_constant_validity */
extern void T376f124(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_character_32_constant */
extern void T376f135(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2b_error */
extern void T51f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2b */
extern T0* T759c982(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T759f982ot1(T0* a1, T0** a2);
extern T1 T759f982ot2(T0* a1, T0** a2);
/* detachable ET_C2_CHARACTER_CONSTANT.literal */
extern T2 T1445f29(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2b_error */
extern void T35f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_character_8_constant */
extern void T376f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0c_error */
extern void T51f373(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0c */
extern T0* T759c981(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0c_error */
extern void T35f373(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c1_character_constant */
extern void T70f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c1_character_constant_validity */
extern void T70f318(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_character_constant_validity */
extern void T70f181(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_character_32_constant */
extern void T70f212(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_character_8_constant */
extern void T70f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c1_character_constant */
extern void T65f1071(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_character_constant */
extern void T65f1111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_escaped_character_8 */
extern void T65f1143(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_natural_32 */
extern void T157f11(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_natural_64 */
extern void T157f12(GE_context* ac, T0* C, T11 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_natural_32 */
extern void T256f12(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_natural_64 */
extern void T256f13(GE_context* ac, T0* C, T11 a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c2_character_constant */
extern void T1284f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c2_character_constant */
extern void T394f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c2_character_constant */
extern void T390f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c2_character_constant */
extern void T376f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c2_character_constant_validity */
extern void T376f217(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c2_character_constant */
extern void T70f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c2_character_constant_validity */
extern void T70f319(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c2_character_constant */
extern void T65f1072(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_function */
extern void T1284f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_do_function */
extern void T1283f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_function */
extern void T376f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_function_validity */
extern void T376f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0c_error */
extern void T51f355(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0c */
extern T0* T759c963(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0c_error */
extern void T35f355(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_result_supplier */
extern void T376f122(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_result_supplier */
extern void T1299f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_result_declaration */
extern void T376f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_query_type_validity */
extern void T376f120(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_function */
extern void T70f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_function_validity */
extern void T70f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_supplier */
extern void T70f159(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_declaration */
extern void T70f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_type_validity */
extern void T70f157(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_do_function */
extern void T65f1073(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function */
extern void T65f1112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_function */
extern void T65f1144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_function */
extern void T1284f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_once_function */
extern void T1283f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_function */
extern void T376f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_function_validity */
extern void T376f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0b_error */
extern void T51f356(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0b */
extern T0* T759c964(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vucr0b_error */
extern void T35f356(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_once_keys_validity */
extern void T376f123(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok2a_error */
extern void T51f361(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vvok2a */
extern T0* T759c969(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vvok2_error */
extern T1 T51f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vvok2a_error */
extern void T35f361(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vvok2_error */
extern T1 T35f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_process_key */
extern T1 T1802f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_expected_key */
extern T1 T1802f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_STANDARD_ONCE_KEYS.string_ */
extern T0* T1802f7(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_thread_key */
extern T1 T1802f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1b_error */
extern void T51f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1b */
extern T0* T759c968(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vvok1_error */
extern T1 T51f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vvok1b_error */
extern void T35f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vvok1_error */
extern T1 T35f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1a_error */
extern void T51f359(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1a */
extern T0* T759c967(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vvok1a_error */
extern void T35f359(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_object_key */
extern T1 T1802f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.standard_once_keys */
extern T0* T376f53(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.default_create */
extern T0* T1802c12(GE_context* ac);
/* detachable ET_MANIFEST_STRING_LIST.manifest_string */
extern T0* T1193f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_STRING_LIST.item */
extern T0* T1193f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].wipe_out */
extern void T396f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].keep_head */
extern void T731f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].set_count */
extern void T731f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].move_all_cursors_after */
extern void T396f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].set_next_cursor */
extern void T732f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].set_after */
extern void T732f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok2b_error */
extern void T51f358(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vvok2b */
extern T0* T759c966(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vvok2b_error */
extern void T35f358(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1c_error */
extern void T51f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1c */
extern T0* T759c965(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vvok1c_error */
extern void T35f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].item */
extern T0* T396f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INDEXING_LIST.append_tagged_indexing_terms_to_list */
extern void T1146f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INDEXING_TERM_LIST.append_indexing_terms_to_list */
extern void T1157f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].force_last */
extern void T396f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].force */
extern void T729f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].fill_with */
extern void T731f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].extend */
extern void T731f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].resize */
extern void T396f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].aliased_resized_area */
extern T0* T729f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].aliased_resized_area */
extern T0* T731f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].new_capacity */
extern T6 T396f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].extendible */
extern T1 T396f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INDEXING_LIST.string_ */
extern T0* T1146f7(GE_context* ac, T0* C);
/* detachable ET_INDEXING.tag */
extern T0* T1448f4(GE_context* ac, T0* C);
/* detachable ET_INDEXING.indexing_clause */
extern T0* T1448f2(GE_context* ac, T0* C);
/* detachable ET_TAGGED_INDEXING.indexing_clause */
extern T0* T1447f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_function */
extern void T70f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_function_validity */
extern void T70f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_keys_validity */
extern void T70f160(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.standard_once_keys */
extern T0* T70f73(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_once_function */
extern void T65f1074(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_function */
extern void T65f1113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_deferred_function */
extern void T1284f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_deferred_function */
extern void T1283f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_deferred_function */
extern void T376f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_deferred_function_validity */
extern void T376f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_deferred_function */
extern void T70f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_deferred_function_validity */
extern void T70f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_deferred_function */
extern void T65f1075(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_deferred_function */
extern void T65f1114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_procedure */
extern void T1284f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_once_procedure */
extern void T1283f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_procedure */
extern void T376f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_procedure_validity */
extern void T376f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.is_once_per_object */
extern T1 T1464f62(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_object_key */
extern T1 T1802f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_expected_key */
extern T1 T1802f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE.standard_once_keys */
extern T0* T1464f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_procedure */
extern void T70f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_procedure_validity */
extern void T70f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_once_procedure */
extern void T65f1076(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_procedure */
extern void T65f1115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_procedure */
extern void T1284f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_external_procedure */
extern void T1283f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_external_feature_validity */
extern void T1283f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_feature_validity */
extern void T1283f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkbu1a_error */
extern void T51f517(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkbu1a */
extern T0* T759c1116(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkbu1_error */
extern T1 T51f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkbu1a_error */
extern void T35f517(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkbu1_error */
extern T1 T35f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.error_handler */
extern T0* T1283f9(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.set_fatal_error */
extern void T1283f48(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.unknown_builtin_reported */
extern T1 T1283f8(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_builtin_code */
extern void T1465f109(GE_context* ac, T0* C, T8 a1, T8 a2);
/* [detachable] ET_EXTERNAL_FUNCTION.set_builtin_code */
extern void T247f112(GE_context* ac, T0* C, T8 a1, T8 a2);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_real_n_feature_validity */
extern void T1283f47(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_expected_builtin_feature_validity */
extern void T1283f50(GE_context* ac, T0* C, T0* a1, T8 a2, T0* a3);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_signature_validity */
extern void T1283f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkbs0a_error */
extern void T51f518(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvkbs0a */
extern T0* T759c1117(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkbs_error */
extern T1 T51f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkbs0a_error */
extern void T35f518(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkbs_error */
extern T1 T35f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] ET_TYPE].count */
extern T6 T250f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].found_item */
extern T0* T1884f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1884f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].found */
extern T1 T1884f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].search */
extern void T1884f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].search_position */
extern void T1884f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1884f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1884f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1884f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1884f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_real_feature_name */
extern T0* T72f1506(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_integer_64_feature_name */
extern T0* T72f1504(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_integer_feature_name */
extern T0* T72f1502(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_double_feature_name */
extern T0* T72f1449(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.quotient_feature_name */
extern T0* T72f1445(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.product_feature_name */
extern T0* T72f1443(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.power_feature_name */
extern T0* T72f1441(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.plus_feature_name */
extern T0* T72f1029(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.out_feature_name */
extern T0* T72f1027(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.opposite_feature_name */
extern T0* T72f1438(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.minus_feature_name */
extern T0* T72f1436(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_positive_infinity_feature_name */
extern T0* T72f1492(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_negative_infinity_feature_name */
extern T0* T72f1490(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_nan_feature_name */
extern T0* T72f1488(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_less_feature_name */
extern T0* T72f1434(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_minimum_number_feature_name */
extern T0* T72f1485(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_maximum_number_feature_name */
extern T0* T72f1483(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_less_equal_feature_name */
extern T0* T72f1481(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_less_feature_name */
extern T0* T72f1479(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_greater_equal_feature_name */
extern T0* T72f1477(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_greater_feature_name */
extern T0* T72f1475(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_equal_feature_name */
extern T0* T72f1473(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identity_feature_name */
extern T0* T72f1428(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.floor_real_64_feature_name */
extern T0* T72f1470(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.floor_real_32_feature_name */
extern T0* T72f1468(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ceiling_real_64_feature_name */
extern T0* T72f1466(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.register_builtin_feature */
extern void T1283f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T8 a4, T0* a5);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].force_last */
extern void T1884f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1884f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_put */
extern void T1884f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_put */
extern void T1884f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].resize */
extern void T1884f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1884f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1884f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1884f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1884f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].aliased_resized_area */
extern T0* T1943f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].aliased_resized_area */
extern T0* T1940f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_resize */
extern void T1884f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1884f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_capacity */
extern T6 T1884f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1884f41(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].force */
extern void T1943f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].fill_with */
extern void T1940f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].extend */
extern void T1940f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1884f39(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ceiling_real_32_feature_name */
extern T0* T72f1464(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].force_last */
extern void T1645f43(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_put */
extern void T1645f49(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_put */
extern void T1645f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_put */
extern void T1645f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_item */
extern T6 T1645f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].hash_position */
extern T6 T1645f23(GE_context* ac, T0* C, T8 a1);
/* NATURAL_8.hash_code */
extern T6 T8f20(GE_context* ac, T8* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].resize */
extern void T1645f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_resize */
extern void T1645f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].special_integer_ */
extern T0* T1645f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_resize */
extern void T1645f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_resize */
extern void T1645f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].aliased_resized_area */
extern T0* T1883f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].aliased_resized_area */
extern T0* T1885f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_item */
extern T8 T1645f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_item */
extern T6 T1645f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_resize */
extern void T1645f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_capacity */
extern T6 T1645f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_put */
extern void T1645f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].force */
extern void T1883f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].fill_with */
extern void T1885f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].extend */
extern void T1885f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].search_position */
extern void T1645f44(GE_context* ac, T0* C, T8 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1884f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.feature_name_tester */
extern T0* T1283f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_map */
extern T0* T1884c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1884f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1884f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1884f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1884f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make */
extern T0* T1944c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_slots */
extern void T1884f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_clashes */
extern void T1884f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1884f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1884f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].make */
extern T0* T1943f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].make_empty */
extern T0* T1940c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].default_create */
extern T0* T1943c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].found_item */
extern T0* T1645f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_item */
extern T0* T1645f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].found */
extern T1 T1645f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].search */
extern void T1645f42(GE_context* ac, T0* C, T8 a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_real_n_ref_feature_validity */
extern void T1283f46(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_TOKEN_CONSTANTS.set_item_feature_name */
extern T0* T72f995(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.positive_infinity_feature_name */
extern T0* T72f1461(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.negative_infinity_feature_name */
extern T0* T72f1459(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.nan_feature_name */
extern T0* T72f1457(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_ref_class_name */
extern T0* T72f927(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_ref_class_name */
extern T0* T72f922(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_ref_class_name */
extern T0* T72f919(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_ref_class_name */
extern T0* T72f916(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_ref_class_name */
extern T0* T72f911(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_integer_n_feature_validity */
extern void T1283f45(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_64_feature_name */
extern T0* T72f1455(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_32_feature_name */
extern T0* T72f1453(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_feature_name */
extern T0* T72f1451(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_character_32_feature_name */
extern T0* T72f1003(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_character_8_feature_name */
extern T0* T72f1001(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_remainder_feature_name */
extern T0* T72f1432(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_quotient_feature_name */
extern T0* T72f1430(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_xor_feature_name */
extern T0* T72f1426(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_shift_right_feature_name */
extern T0* T72f1424(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_shift_left_feature_name */
extern T0* T72f1422(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_or_feature_name */
extern T0* T72f1420(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_not_feature_name */
extern T0* T72f1418(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_and_feature_name */
extern T0* T72f1416(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_64_feature_name */
extern T0* T72f1414(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_32_feature_name */
extern T0* T72f1412(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_16_feature_name */
extern T0* T72f1410(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_8_feature_name */
extern T0* T72f1408(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_64_feature_name */
extern T0* T72f1406(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_32_feature_name */
extern T0* T72f1404(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_16_feature_name */
extern T0* T72f1402(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_8_feature_name */
extern T0* T72f1400(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_integer_n_ref_feature_validity */
extern void T1283f44(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_ref_class_name */
extern T0* T72f908(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_tuple_feature_validity */
extern void T1283f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_object_comparison_feature_name */
extern T0* T72f1397(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_reference_feature_name */
extern T0* T72f91(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_real_64_feature_name */
extern T0* T72f1394(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_real_32_feature_name */
extern T0* T72f1392(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_pointer_feature_name */
extern T0* T72f1390(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_64_feature_name */
extern T0* T72f1388(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_32_feature_name */
extern T0* T72f1386(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_16_feature_name */
extern T0* T72f1384(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_8_feature_name */
extern T0* T72f1382(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_64_feature_name */
extern T0* T72f1380(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_32_feature_name */
extern T0* T72f1378(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_16_feature_name */
extern T0* T72f1376(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_8_feature_name */
extern T0* T72f1374(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_character_32_feature_name */
extern T0* T72f1372(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_character_8_feature_name */
extern T0* T72f1370(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_boolean_feature_name */
extern T0* T72f1368(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_item_feature_name */
extern T0* T72f93(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_item_feature_name */
extern T0* T72f1365(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_item_feature_name */
extern T0* T72f1363(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.object_comparison_feature_name */
extern T0* T72f1361(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_item_feature_name */
extern T0* T72f1359(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_item_feature_name */
extern T0* T72f1357(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_item_feature_name */
extern T0* T72f1355(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_item_feature_name */
extern T0* T72f1353(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.item_code_feature_name */
extern T0* T72f1351(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_item_feature_name */
extern T0* T72f1349(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_item_feature_name */
extern T0* T72f1347(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_item_feature_name */
extern T0* T72f1345(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_item_feature_name */
extern T0* T72f1343(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_item_feature_name */
extern T0* T72f1340(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_item_feature_name */
extern T0* T72f1338(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_item_feature_name */
extern T0* T72f1336(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_function_feature_validity */
extern void T1283f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.detachable_separate_formal_parameter_type */
extern T0* T1283f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.fast_item_feature_name */
extern T0* T72f1334(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_procedure_feature_validity */
extern void T1283f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.fast_call_feature_name */
extern T0* T72f1332(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_platform_feature_validity */
extern void T1283f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.wide_character_bytes_feature_name */
extern T0* T72f1330(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_bytes_feature_name */
extern T0* T72f1328(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_bytes_feature_name */
extern T0* T72f1326(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_windows_feature_name */
extern T0* T72f1324(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_vxworks_feature_name */
extern T0* T72f1322(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_vms_feature_name */
extern T0* T72f1320(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_unix_feature_name */
extern T0* T72f1318(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_thread_capable_feature_name */
extern T0* T72f1316(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_scoop_capable_feature_name */
extern T0* T72f1314(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_mac_feature_name */
extern T0* T72f1312(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_dotnet_feature_name */
extern T0* T72f1310(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_64_bits_feature_name */
extern T0* T72f1308(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_bytes_feature_name */
extern T0* T72f1306(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.double_bytes_feature_name */
extern T0* T72f1304(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_bytes_feature_name */
extern T0* T72f1302(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_bytes_feature_name */
extern T0* T72f1300(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.platform_class_name */
extern T0* T72f907(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_ise_runtime_feature_validity */
extern void T1283f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.unmark_object_feature_name */
extern T0* T72f1297(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unlock_marking_feature_name */
extern T0* T72f1295(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_reference_field_at_feature_name */
extern T0* T72f1294(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_reference_field_feature_name */
extern T0* T72f1293(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_64_field_at_feature_name */
extern T0* T72f1291(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_64_field_feature_name */
extern T0* T72f1289(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_32_field_at_feature_name */
extern T0* T72f1287(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_32_field_feature_name */
extern T0* T72f1285(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pre_ecma_mapping_feature_name */
extern T0* T72f1283(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pointer_field_at_feature_name */
extern T0* T72f1281(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pointer_field_feature_name */
extern T0* T72f1279(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_64_field_at_feature_name */
extern T0* T72f1277(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_64_field_feature_name */
extern T0* T72f1275(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_32_field_at_feature_name */
extern T0* T72f1273(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_32_field_feature_name */
extern T0* T72f1271(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_16_field_at_feature_name */
extern T0* T72f1269(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_16_field_feature_name */
extern T0* T72f1267(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_8_field_at_feature_name */
extern T0* T72f1265(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_8_field_feature_name */
extern T0* T72f1263(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_64_field_at_feature_name */
extern T0* T72f1261(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_64_field_feature_name */
extern T0* T72f1259(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_32_field_at_feature_name */
extern T0* T72f1257(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_32_field_feature_name */
extern T0* T72f1255(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_16_field_at_feature_name */
extern T0* T72f1253(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_16_field_feature_name */
extern T0* T72f1251(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_8_field_at_feature_name */
extern T0* T72f1249(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_8_field_feature_name */
extern T0* T72f1247(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_32_field_at_feature_name */
extern T0* T72f1245(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_32_field_feature_name */
extern T0* T72f1243(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_8_field_at_feature_name */
extern T0* T72f1241(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_8_field_feature_name */
extern T0* T72f1239(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_boolean_field_at_feature_name */
extern T0* T72f1237(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_boolean_field_feature_name */
extern T0* T72f1235(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.mark_object_feature_name */
extern T0* T72f1233(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.lock_marking_feature_name */
extern T0* T72f1231(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_id_from_name_feature_name */
extern T0* T72f1229(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_conforms_to_feature_name */
extern T0* T72f1227(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.storable_version_of_type_feature_name */
extern T0* T72f1225(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_at_offset_feature_name */
extern T0* T72f1224(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_at_feature_name */
extern T0* T72f1223(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_feature_name */
extern T0* T72f1222(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_field_at_feature_name */
extern T0* T72f1220(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_field_feature_name */
extern T0* T72f1218(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_field_at_feature_name */
extern T0* T72f1216(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_field_feature_name */
extern T0* T72f1214(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raw_reference_field_at_offset_feature_name */
extern T0* T72f1212(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raw_reference_field_at_feature_name */
extern T0* T72f1210(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pre_ecma_mapping_status_feature_name */
extern T0* T72f1208(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_field_at_feature_name */
extern T0* T72f1206(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_field_feature_name */
extern T0* T72f1204(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.persistent_field_count_of_type_feature_name */
extern T0* T72f1202(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.once_objects_feature_name */
extern T0* T72f1200(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.object_size_feature_name */
extern T0* T72f1198(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_type_instance_of_feature_name */
extern T0* T72f1197(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_tuple_instance_of_feature_name */
extern T0* T72f1196(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_special_of_reference_instance_of_feature_name */
extern T0* T72f1195(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_instance_of_feature_name */
extern T0* T72f1194(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_field_at_feature_name */
extern T0* T72f1192(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_field_feature_name */
extern T0* T72f1190(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_field_at_feature_name */
extern T0* T72f1188(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_field_feature_name */
extern T0* T72f1186(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_field_at_feature_name */
extern T0* T72f1184(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_field_feature_name */
extern T0* T72f1182(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_field_at_feature_name */
extern T0* T72f1180(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_field_feature_name */
extern T0* T72f1178(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_tuple_type_feature_name */
extern T0* T72f1176(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_tuple_feature_name */
extern T0* T72f1174(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_type_feature_name */
extern T0* T72f1172(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_or_basic_type_feature_name */
extern T0* T72f1170(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_feature_name */
extern T0* T72f1168(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_expanded_feature_name */
extern T0* T72f1166(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_copy_semantics_item_feature_name */
extern T0* T72f1164(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_feature_name */
extern T0* T72f1162(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_object_marked_feature_name */
extern T0* T72f1160(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_field_transient_of_type_feature_name */
extern T0* T72f1158(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_field_expanded_of_type_feature_name */
extern T0* T72f1156(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_expanded_feature_name */
extern T0* T72f973(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_copy_semantics_field_feature_name */
extern T0* T72f1153(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_attached_type_feature_name */
extern T0* T72f1151(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_field_at_feature_name */
extern T0* T72f1149(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_field_feature_name */
extern T0* T72f1147(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_field_at_feature_name */
extern T0* T72f1145(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_field_feature_name */
extern T0* T72f1143(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_field_at_feature_name */
extern T0* T72f1141(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_field_feature_name */
extern T0* T72f1139(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_field_at_feature_name */
extern T0* T72f1137(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_field_feature_name */
extern T0* T72f1135(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.in_assertion_feature_name */
extern T0* T72f1133(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generic_parameter_count_feature_name */
extern T0* T72f965(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generator_of_type_feature_name */
extern T0* T72f1130(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generating_type_of_type_feature_name */
extern T0* T72f1128(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_type_of_type_feature_name */
extern T0* T72f1126(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_static_type_of_type_feature_name */
extern T0* T72f1124(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_offset_of_type_feature_name */
extern T0* T72f1122(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_name_of_type_feature_name */
extern T0* T72f1120(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_count_of_type_feature_name */
extern T0* T72f1118(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_gen_param_id_feature_name */
extern T0* T72f1116(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.dynamic_type_at_offset_feature_name */
extern T0* T72f1114(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.dynamic_type_feature_name */
extern T0* T72f1112(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_type_feature_name */
extern T0* T72f1110(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.compiler_version_feature_name */
extern T0* T72f1108(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.check_assert_feature_name */
extern T0* T72f1106(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_field_at_feature_name */
extern T0* T72f1104(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_field_feature_name */
extern T0* T72f1102(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_field_at_feature_name */
extern T0* T72f1100(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_field_feature_name */
extern T0* T72f1098(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_field_at_feature_name */
extern T0* T72f1096(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_field_feature_name */
extern T0* T72f1094(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.attached_type_feature_name */
extern T0* T72f1092(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ise_runtime_class_name */
extern T0* T72f906(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_identified_routines_feature_validity */
extern void T1283f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.eif_object_id_free_feature_name */
extern T0* T72f1090(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_object_id_feature_name */
extern T0* T72f1089(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_id_object_feature_name */
extern T0* T72f1088(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identified_routines_class_name */
extern T0* T72f905(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_ise_exception_manager_feature_validity */
extern void T1283f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.developer_raise_feature_name */
extern T0* T72f1086(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_exception_manager_factory_feature_validity */
extern void T1283f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_feature_name */
extern T0* T72f1083(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_factory_class_name */
extern T0* T72f904(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_exception_manager_feature_validity */
extern void T1283f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_is_ignored_feature_name */
extern T0* T72f1080(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raise_feature_name */
extern T0* T72f1078(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ignore_feature_name */
extern T0* T72f1076(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.catch_feature_name */
extern T0* T72f1074(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_of_code_feature_name */
extern T0* T72f1072(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_raisable_feature_name */
extern T0* T72f1069(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_ignored_feature_name */
extern T0* T72f1067(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_ignorable_feature_name */
extern T0* T72f1065(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_caught_feature_name */
extern T0* T72f1063(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_from_code_feature_name */
extern T0* T72f1061(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_memory_feature_validity */
extern void T1283f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.free_feature_name */
extern T0* T72f1058(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.find_referers_feature_name */
extern T0* T72f1056(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.memory_class_name */
extern T0* T72f903(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_com_failure_feature_validity */
extern void T1283f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.cwin_local_free_feature_name */
extern T0* T72f1053(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.cwin_error_text_feature_name */
extern T0* T72f1051(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_size_feature_name */
extern T0* T72f1049(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_facility_feature_name */
extern T0* T72f1047(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_code_feature_name */
extern T0* T72f1045(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_feature_name */
extern T0* T72f1043(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.c_strlen_feature_name */
extern T0* T72f1041(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.com_failure_class_name */
extern T0* T72f902(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_arguments_32_feature_validity */
extern void T1283f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.i_th_argument_string_feature_name */
extern T0* T72f1038(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.i_th_argument_pointer_feature_name */
extern T0* T72f1036(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.argument_count_feature_name */
extern T0* T72f1034(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.arguments_32_class_name */
extern T0* T72f901(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_pointer_feature_validity */
extern void T1283f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.to_integer_32_feature_name */
extern T0* T72f1031(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_default_pointer_feature_name */
extern T0* T72f1025(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.hash_code_feature_name */
extern T0* T72f1023(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_pointer_ref_feature_validity */
extern void T1283f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_ref_class_name */
extern T0* T72f900(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_boolean_feature_validity */
extern void T1283f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.negated_feature_name */
extern T0* T72f1019(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implication_feature_name */
extern T0* T72f1017(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_semistrict_feature_name */
extern T0* T72f1015(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_exclusive_feature_name */
extern T0* T72f1013(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_feature_name */
extern T0* T72f1011(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conjuncted_semistrict_feature_name */
extern T0* T72f1009(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conjuncted_feature_name */
extern T0* T72f1007(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_boolean_ref_feature_validity */
extern void T1283f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_ref_class_name */
extern T0* T72f899(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_character_n_feature_validity */
extern void T1283f27(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_code_feature_name */
extern T0* T72f999(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.code_feature_name */
extern T0* T72f997(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_character_n_ref_feature_validity */
extern void T1283f26(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.current_universe */
extern T0* T1283f7(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_special_feature_validity */
extern void T1283f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_count_feature_name */
extern T0* T72f993(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_feature_name */
extern T0* T72f90(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.make_empty_feature_name */
extern T0* T72f990(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.extend_feature_name */
extern T0* T72f94(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.element_size_feature_name */
extern T0* T72f986(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.base_address_feature_name */
extern T0* T72f982(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.like_current */
extern T0* T72f938(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.aliased_resized_area_feature_name */
extern T0* T72f980(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_type_feature_validity */
extern void T1283f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.type_id_feature_name */
extern T0* T72f977(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.runtime_name_feature_name */
extern T0* T72f975(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_deferred_feature_name */
extern T0* T72f971(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_attached_feature_name */
extern T0* T72f969(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.has_default_feature_name */
extern T0* T72f967(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generic_parameter_type_feature_name */
extern T0* T72f963(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.default_feature_name */
extern T0* T72f961(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_any_feature_validity */
extern void T1283f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.standard_copy_feature_name */
extern T0* T72f958(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.twin_feature_name */
extern T0* T72f955(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.tagged_out_feature_name */
extern T0* T72f953(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.standard_twin_feature_name */
extern T0* T72f951(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.standard_is_equal_feature_name */
extern T0* T72f949(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.same_type_feature_name */
extern T0* T72f947(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_deep_equal_feature_name */
extern T0* T72f944(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generator_feature_name */
extern T0* T72f942(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generating_type_feature_name */
extern T0* T72f940(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.deep_twin_feature_name */
extern T0* T72f937(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conforms_to_feature_name */
extern T0* T72f935(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.string_ */
extern T0* T1283f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_external_procedure */
extern void T376f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_procedure_validity */
extern void T376f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_procedure */
extern void T70f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_validity */
extern void T70f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_builtin_procedure_validity */
extern void T70f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_tuple_put_reference */
extern void T70f167(GE_context* ac, T0* C, T0* a1);
extern T1 T70f167ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_procedure */
extern void T70f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_set_is_ignored */
extern void T70f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.argument_type_set */
extern T0* T70f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_raise */
extern void T70f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_ignore */
extern void T70f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_catch */
extern void T70f162(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin_unknown */
extern T1 T1465f68(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_validity */
extern void T70f156p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_external_procedure */
extern void T65f1077(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_procedure */
extern void T65f1116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_routine */
extern void T65f1145(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_C_GENERATOR.flush_to_cpp_file */
extern void T65f1159(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_dllwin_body */
extern void T65f1158(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_C_GENERATOR.print_external_cpp_body */
extern void T65f1157(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].item */
extern T0* T148f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_SPLITTER.make_with_separators */
extern T0* T1898c9(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_c_struct_body */
extern void T65f1156(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_C_GENERATOR.print_external_c_body */
extern void T65f1155(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T1 a7);
/* [detachable] ET_C_GENERATOR.print_external_c_prototype */
extern void T65f1154(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_C_GENERATOR.print_external_builtin_body */
extern void T65f1153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_procedure_body */
extern void T65f1163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_tuple_procedure_body */
extern void T65f1201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_reference_body */
extern void T65f1260(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1260ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_real_64_body */
extern void T65f1259(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_basic_expanded_item_body */
extern void T65f1262(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1262ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_real_32_body */
extern void T65f1258(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_pointer_body */
extern void T65f1257(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_64_body */
extern void T65f1256(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_32_body */
extern void T65f1255(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_16_body */
extern void T65f1254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_8_body */
extern void T65f1253(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_64_body */
extern void T65f1252(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_32_body */
extern void T65f1251(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_16_body */
extern void T65f1250(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_8_body */
extern void T65f1249(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_character_32_body */
extern void T65f1248(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_character_8_body */
extern void T65f1247(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_boolean_body */
extern void T65f1246(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_special_procedure_body */
extern void T65f1200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_special_set_count_body */
extern void T65f1245(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1245ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_extend_body */
extern void T65f1244(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1244ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_ref_procedure_body */
extern void T65f1199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_procedure_procedure_body */
extern void T65f1198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_ref_procedure_body */
extern void T65f1197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_memory_procedure_body */
extern void T65f1196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_runtime_procedure_body */
extern void T65f1195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pre_ecma_mapping_body */
extern void T65f1243(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_exception_manager_procedure_body */
extern void T65f1194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_ref_procedure_body */
extern void T65f1193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_identified_routines_procedure_body */
extern void T65f1192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_procedure_body */
extern void T65f1191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_set_is_ignored_body */
extern void T65f1242(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_raise_body */
extern void T65f1241(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_ignore_body */
extern void T65f1240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_catch_body */
extern void T65f1239(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_com_failure_procedure_body */
extern void T65f1190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_ref_procedure_body */
extern void T65f1189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_ref_procedure_body */
extern void T65f1188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_any_procedure_body */
extern void T65f1187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_function_body */
extern void T65f1162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_type_function_body */
extern void T65f1186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_type_generic_parameter_type_body */
extern void T65f1238(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_semicolon_newline */
extern void T65f1204(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_indentation_assign_to_result */
extern void T65f1203(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_builtin_tuple_function_body */
extern void T65f1185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_reference_item_body */
extern void T65f1237(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_builtin_tuple_reference_item_body */
extern T0* T65f1237ac1(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T65f1237ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_real_64_item_body */
extern void T65f1236(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_basic_expanded_item_body */
extern void T65f1261(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1261ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_real_32_item_body */
extern void T65f1235(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_pointer_item_body */
extern void T65f1234(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_64_item_body */
extern void T65f1233(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_32_item_body */
extern void T65f1232(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_16_item_body */
extern void T65f1231(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_8_item_body */
extern void T65f1230(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_item_code_body */
extern void T65f1229(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1229ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_64_item_body */
extern void T65f1228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_32_item_body */
extern void T65f1227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_16_item_body */
extern void T65f1226(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_8_item_body */
extern void T65f1225(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_character_32_item_body */
extern void T65f1224(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_character_8_item_body */
extern void T65f1223(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_boolean_item_body */
extern void T65f1222(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_special_function_body */
extern void T65f1184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_special_aliased_resized_area_body */
extern void T65f1221(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1221ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_ref_function_body */
extern void T65f1183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_function_body */
extern void T65f1182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_out_body */
extern void T65f1220(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_ref_function_body */
extern void T65f1181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_function_body */
extern void T65f1180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_out_body */
extern void T65f1219(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_platform_function_body */
extern void T65f1179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_memory_function_body */
extern void T65f1178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_runtime_function_body */
extern void T65f1177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pre_ecma_mapping_status_body */
extern void T65f1218(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_in_assertion_body */
extern void T65f1217(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_ref_function_body */
extern void T65f1176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_function_body */
extern void T65f1175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_identified_routines_function_body */
extern void T65f1174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_function_function_body */
extern void T65f1173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_factory_function_body */
extern void T65f1172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_factory_exception_manager_body */
extern void T65f1216(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_function_body */
extern void T65f1171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_type_of_code_body */
extern void T65f1215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_last_exception_body */
extern void T65f1214(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_raisable_body */
extern void T65f1213(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_ignored_body */
extern void T65f1212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_ignorable_body */
extern void T65f1211(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_caught_body */
extern void T65f1210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_exception_from_code_body */
extern void T65f1209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_com_failure_function_body */
extern void T65f1170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_ref_function_body */
extern void T65f1169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_function_body */
extern void T65f1168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_ref_function_body */
extern void T65f1167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_function_body */
extern void T65f1166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_arguments_32_function_body */
extern void T65f1165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_any_function_body */
extern void T65f1164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_body */
extern void T65f1208(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1208ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_tagged_out_body */
extern void T65f1207(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1207ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_twin_body */
extern void T65f1206(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1206ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_is_deep_equal_body */
extern void T65f1205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_conforms_to_body */
extern void T65f1202(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_function */
extern T1 T1465f67(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_c_inline_body */
extern void T65f1152(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_ROUTINES.as_lower */
extern T2 T1899f1(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_C_GENERATOR.character_ */
extern T0* T65f499(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_ROUTINES.default_create */
extern T0* T1899c2(GE_context* ac);
/* [detachable] UC_UTF8_STRING.has_substring */
extern T1 T743f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.has_substring */
extern T1 T17f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_c_includes */
extern void T65f1151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_header_filename */
extern void T65f1161(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_substring_count */
extern T6 T83f308(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_feature_info_message_call */
extern void T65f1150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_inline_routine_name */
extern void T65f1149(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T65f1149ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_like_feature */
extern void T1567f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_like_feature */
extern void T1566f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_like_feature */
extern void T1565f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_like_feature */
extern void T1564f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_like_feature */
extern void T1296f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_like_feature */
extern void T1288f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_like_feature */
extern void T1287f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_like_feature_validity */
extern void T1287f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat2b_error */
extern void T51f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat2b */
extern T0* T759c989(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vtat2b_error */
extern void T35f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.error_handler */
extern T0* T1287f12(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.set_fatal_error */
extern void T1287f27(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_like_feature */
extern void T1282f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_like_feature */
extern void T1281f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_like_feature */
extern void T1280f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_like_feature */
extern void T1279f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_like_feature_to_sorter */
extern void T1279f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].force_relation */
extern void T1641f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].put_indexed_relation */
extern void T1641f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].force */
extern void T1641f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].put */
extern void T1641f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].put_new */
extern void T1877f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_put */
extern void T1877f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_put */
extern void T1877f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_put */
extern void T1877f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_put */
extern void T1877f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_item */
extern T6 T1877f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].hash_position */
extern T6 T1877f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.hash_code */
extern T6 T1476f78(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_item */
extern T6 T1877f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].resize */
extern void T1877f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_resize */
extern void T1877f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].special_integer_ */
extern T0* T1877f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_resize */
extern void T1877f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].aliased_resized_area */
extern T0* T1879f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].aliased_resized_area */
extern T0* T1878f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_resize */
extern void T1877f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_item */
extern T0* T1877f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_resize */
extern void T1877f44(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].resize */
extern void T1642f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].capacity */
extern T6 T1641f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].count */
extern T6 T1641f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].index_of */
extern T6 T1641f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].found_item */
extern T6 T1877f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_item */
extern T6 T1877f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].found */
extern T1 T1877f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].search */
extern void T1877f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].search_position */
extern void T1877f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_like_feature */
extern void T1278f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_like_feature */
extern void T1277f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_like_feature */
extern void T1277f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1a_error */
extern void T51f383(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1a */
extern T0* T759c991(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtat1_error */
extern T1 T51f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtat1a_error */
extern void T35f383(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtat1_error */
extern T1 T35f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1b_error */
extern void T51f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1b */
extern T0* T759c990(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vtat1b_error */
extern void T35f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.error_handler */
extern T0* T1277f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.set_fatal_error */
extern void T1277f19(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.resolve_like_argument */
extern void T1476f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolve_like_feature */
extern void T1476f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_feature */
extern void T1266f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_type */
extern void T1266f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_like_type_constraint */
extern void T1266f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc1a_error */
extern void T51f522(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vggc1a */
extern T0* T759c1121(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc1_error */
extern T1 T51f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc1a_error */
extern void T35f522(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc1_error */
extern T1 T35f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_feature */
extern void T1265f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_type */
extern void T1265f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_like_type_validity */
extern void T1265f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr3a_error */
extern void T51f521(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr3a */
extern T0* T759c1120(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr3_error */
extern T1 T51f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr3a_error */
extern void T35f521(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr3_error */
extern T1 T35f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.error_handler */
extern T0* T1265f6(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER1.set_fatal_error */
extern void T1265f12(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.process_like_feature */
extern void T380f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_like_feature_validity */
extern void T380f33(GE_context* ac, T0* C, T0* a1);
extern T1 T380f33ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_qualified_like_braced_type */
extern void T1567f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_qualified_like_identifier_validity */
extern void T1567f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_qualified_like_braced_type */
extern void T1566f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_qualified_like_identifier_validity */
extern void T1566f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.set_fatal_error */
extern void T1566f12(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_flattening_error */
extern T1 T560f39(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.features_flattened */
extern T1 T560f38(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_qualified_like_braced_type */
extern void T1565f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_qualified_like_identifier_validity */
extern void T1565f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_qualified_like_braced_type */
extern void T1564f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_qualified_like_identifier_validity */
extern void T1564f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_qualified_like_braced_type */
extern void T1296f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_qualified_like_braced_type */
extern void T1288f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_qualified_like_identifier */
extern void T1288f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T1287f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_qualified_like_identifier_validity */
extern void T1287f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1c_error */
extern void T51f384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1c */
extern T0* T759c992(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vtat1c_error */
extern void T35f384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolve_identifier_type */
extern void T1478f87(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolve_identifier_type */
extern void T1477f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].force_last */
extern void T1286f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].resize */
extern void T1286f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_resize */
extern void T1286f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].special_integer_ */
extern T0* T1286f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_resize */
extern void T1286f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_resize */
extern void T1286f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_resize */
extern void T1286f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_capacity */
extern T6 T1286f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.add_adapted_base_classes_to_list */
extern void T1287f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T1287f28ot1(T0* a1, T0** a2);
extern T1 T1287f28ot2(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.copy_to_type_context */
extern void T237f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.copy_to_type_context */
extern void T73f324(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_qualified_like_braced_type */
extern void T1282f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_qualified_like_braced_type */
extern void T1281f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_qualified_like_braced_type */
extern void T1280f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_qualified_like_identifier_tuple_actual_parameters_unfolded */
extern void T1280f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_qualified_like_identifier_vtct_validity */
extern void T1280f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T1279f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_qualified_like_identifier_to_sorter */
extern void T1279f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_qualified_like_braced_type */
extern void T1278f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_qualified_like_identifier */
extern void T1278f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_qualified_like_braced_type */
extern void T1277f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_qualified_like_identifier */
extern void T1277f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_qualified_like_braced_type */
extern void T1266f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_qualified_like_braced_type */
extern void T1265f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T380f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_qualified_like_identifier_validity */
extern void T380f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_adapted_base_classes_validity */
extern void T380f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_class_interface_error_only */
extern void T400f12(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TYPE_CHECKER.report_qualified_anchored_type_supplier */
extern void T380f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SUPPLIER_BUILDER.report_qualified_anchored_type_supplier */
extern void T1299f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_qualified_like_type */
extern void T1567f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_qualified_like_type */
extern void T1566f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_qualified_like_type */
extern void T1565f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_qualified_like_type */
extern void T1564f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_qualified_like_type */
extern void T1296f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_qualified_like_type */
extern void T1288f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_qualified_like_type */
extern void T1287f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_qualified_like_type */
extern void T1282f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_qualified_like_type */
extern void T1281f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_qualified_like_type */
extern void T1280f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_qualified_like_type */
extern void T1279f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_qualified_like_type */
extern void T1278f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_qualified_like_type */
extern void T1277f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_qualified_like_type */
extern void T1266f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_qualified_like_type */
extern void T1265f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_qualified_like_type */
extern void T380f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_precursor_expression */
extern void T1284f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.check_precursor_validity */
extern void T1284f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3c_error */
extern void T51f555(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3c */
extern T0* T759c1153(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.tokens */
extern T0* T759f817(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr3_error */
extern T1 T51f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3c_error */
extern void T35f555(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr3_error */
extern T1 T35f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3b_error */
extern void T51f554(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3b */
extern T0* T759c1152(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.position */
extern T0* T1487f8(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3b_error */
extern void T35f554(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parent_type */
extern void T1487f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_parent_type */
extern void T1490f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.set_seed */
extern void T933f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3a_error */
extern void T51f553(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3a */
extern T0* T759c1151(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3a_error */
extern void T35f553(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr2a_error */
extern void T51f552(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr2a */
extern T0* T759c1150(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr2_error */
extern T1 T51f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr2a_error */
extern void T35f552(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr2_error */
extern T1 T35f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.error_handler */
extern T0* T1284f9(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.set_fatal_error */
extern void T1284f93(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.current_universe */
extern T0* T1284f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.class_name */
extern T0* T180f98(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_precursor_expression */
extern void T394f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_precursor_expression */
extern void T390f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_precursor_expression */
extern void T376f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_expression_validity */
extern void T376f218(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f218ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_query_expression_validity */
extern void T376f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].force_last */
extern void T1298f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_put */
extern void T1298f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_put */
extern void T1298f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_item */
extern T6 T1298f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].hash_position */
extern T6 T1298f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].resize */
extern void T1298f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_resize */
extern void T1298f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].special_integer_ */
extern T0* T1298f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_resize */
extern void T1298f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_resize */
extern void T1298f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_item */
extern T0* T1298f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_resize */
extern void T1298f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_capacity */
extern T6 T1298f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_put */
extern void T1298f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].search_position */
extern void T1298f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_equality_tester */
extern T0* T1298f30(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_precursor_expression */
extern void T376f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0h_error */
extern void T51f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0h */
extern T0* T759c1047(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0h_error */
extern void T35f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_parenthesis_call_validity */
extern void T376f270(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T376f270ot1(T0* a1, T0** a2);
extern T1 T376f270ot2(T0* a1, T0** a2);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parenthesis_call */
extern void T1487f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_EXPRESSION.set_parenthesis_call */
extern void T1490f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.is_generic */
extern T1 T71f18(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3d_error */
extern void T51f407(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3d */
extern T0* T759c1007(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3d_error */
extern void T35f407(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3e_error */
extern void T51f406(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3e */
extern T0* T759c1006(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3e_error */
extern void T35f406(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr1b_error */
extern void T51f405(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr1b */
extern T0* T759c1005(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr1_error */
extern T1 T51f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr1b_error */
extern void T35f405(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr1_error */
extern T1 T35f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_precursor_expression */
extern void T70f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_expression_validity */
extern void T70f324(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_query_expression_validity */
extern void T70f384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_precursor_expression */
extern void T70f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_PRECURSOR.dynamic_precursor */
extern T0* T298f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_FEATURE.dynamic_precursor */
extern T0* T232f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.force_last */
extern void T299f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].force */
extern void T1910f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].fill_with */
extern void T1909f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].extend */
extern void T1909f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.fixed_array */
extern T0* T299f4(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].default_create */
extern T0* T1910c3(GE_context* ac);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.resize */
extern void T299f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].aliased_resized_area */
extern T0* T1910f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].aliased_resized_area */
extern T0* T1909f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.new_capacity */
extern T6 T299f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.capacity */
extern T6 T299f5(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.put_last */
extern void T299f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.make_with_capacity */
extern T0* T299c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].make */
extern T0* T1910f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].make_empty */
extern T0* T1909c5(GE_context* ac, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.make */
extern T0* T298c32(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DYNAMIC_PRECURSOR.make_feature */
extern void T298f40(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T298f40ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.tokens */
extern T0* T298f28(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.empty_dynamic_type_sets */
extern T0* T298f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_parenthesis_call_validity */
extern void T70f383(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T70f383ot1(T0* a1, T0** a2);
extern T1 T70f383ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_precursor_expression */
extern void T65f1078(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_precursor_expression */
extern void T65f1117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_feature_address */
extern void T1284f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_feature_address */
extern void T394f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_feature_address */
extern void T390f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_feature_address */
extern void T376f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_feature_address_validity */
extern void T376f219(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar4a_error */
extern void T51f413(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuar4a */
extern T0* T759c1013(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar4_error */
extern T1 T51f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar4a_error */
extern void T35f413(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar4_error */
extern T1 T35f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_function_address */
extern void T376f279(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_attribute_address */
extern void T376f278(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0e_error */
extern void T51f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0e */
extern T0* T759c1012(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0e_error */
extern void T35f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_procedure_address */
extern void T376f277(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_pointer_expression */
extern void T376f274(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_typed_pointer_expression */
extern void T376f273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_feature_address */
extern void T70f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_feature_address_validity */
extern void T70f348(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_function_address */
extern void T70f415(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_PRECURSOR.set_address */
extern void T298f41(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_address */
extern void T232f39(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_attribute_address */
extern void T70f414(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_procedure_address */
extern void T70f413(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_pointer_expression */
extern void T70f386(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_typed_pointer_expression */
extern void T70f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_feature_address */
extern void T65f1079(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_feature_address */
extern void T65f1118(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1118ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.is_attribute */
extern T1 T298f25(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_attribute */
extern T1 T232f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.is_empty */
extern T1 T220f9(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_current_address */
extern void T1284f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_current_address */
extern void T394f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_current_address */
extern void T390f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_current_address */
extern void T376f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_current_address_validity */
extern void T376f220(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0g_error */
extern void T51f416(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0g */
extern T0* T759c1016(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0g_error */
extern void T35f416(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_current_address */
extern void T70f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_current_address_validity */
extern void T70f325(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_current_address */
extern void T65f1080(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_current_address */
extern void T65f1119(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1119ot1(T0* a1, T0** a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_result_address */
extern void T1284f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_result_address */
extern void T394f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_result_address */
extern void T390f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_result_address */
extern void T376f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_result_address_validity */
extern void T376f221(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2f_error */
extern void T51f419(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2f */
extern T0* T759c1019(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2f_error */
extern void T35f419(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_result_address */
extern void T70f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_result_address_validity */
extern void T70f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_result_address */
extern void T65f1081(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_result_address */
extern void T65f1120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_expression_address */
extern void T1284f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_expression_address */
extern void T394f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_expression_address */
extern void T390f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_expression_address */
extern void T376f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_expression_address_validity */
extern void T376f222(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_expression_address */
extern void T70f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_expression_address_validity */
extern void T70f327(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_expression_address */
extern void T65f1082(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_expression_address */
extern void T65f1121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_infix_expression */
extern void T1284f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_infix_expression */
extern void T394f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_infix_expression */
extern void T390f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_infix_expression */
extern void T376f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_infix_expression_validity */
extern void T376f223(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.free_call_info */
extern void T376f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].force_last */
extern void T386f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].force */
extern void T705f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].fill_with */
extern void T707f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].extend */
extern void T707f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].resize */
extern void T386f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].aliased_resized_area */
extern T0* T705f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].aliased_resized_area */
extern T0* T707f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].new_capacity */
extern T6 T386f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].extendible */
extern T1 T386f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.new_call_info */
extern T0* T376f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].remove_last */
extern void T386f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].keep_head */
extern void T707f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_count */
extern void T707f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].move_last_cursors_after */
extern void T386f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_next_cursor */
extern void T708f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_after */
extern void T708f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].last */
extern T0* T386f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].is_empty */
extern T1 T386f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_infix_expression */
extern void T70f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_infix_expression_validity */
extern void T70f328(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_call_info */
extern void T70f388(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_call_info */
extern T0* T70f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_infix_expression */
extern void T65f1083(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_infix_expression */
extern void T65f1122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_type */
extern void T1284f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_type */
extern void T394f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_type */
extern void T390f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_type */
extern void T376f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_type_validity */
extern void T376f224(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_type */
extern void T376f283(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_type */
extern void T70f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_type_validity */
extern void T70f331(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_type */
extern void T70f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_manifest_type */
extern void T65f1084(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_type */
extern void T65f1123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_prefix_expression */
extern void T1284f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_prefix_expression */
extern void T394f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_prefix_expression */
extern void T390f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_prefix_expression */
extern void T376f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_prefix_expression_validity */
extern void T376f225(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_prefix_expression */
extern void T70f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_prefix_expression_validity */
extern void T70f332(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_prefix_expression */
extern void T65f1085(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_prefix_expression */
extern void T65f1124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_old_expression */
extern void T1284f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_old_expression */
extern void T394f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_old_expression */
extern void T390f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_old_expression */
extern void T376f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_old_expression_validity */
extern void T376f226(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vaol1a_error */
extern void T51f421(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vaol1a */
extern T0* T759c1021(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vaol1_error */
extern T1 T51f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vaol1a_error */
extern void T35f421(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vaol1_error */
extern T1 T35f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_old_expression */
extern void T70f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_old_expression_validity */
extern void T70f333(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_old_expression */
extern void T65f1086(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_old_expression */
extern void T65f1125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_old_object_test */
extern void T1284f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_old_object_test */
extern void T394f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_old_object_test */
extern void T390f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_old_object_test */
extern void T376f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_old_object_test */
extern void T70f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_old_object_test */
extern void T65f1087(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_manifest_string */
extern void T1284f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_manifest_string */
extern void T394f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_manifest_string */
extern void T390f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_manifest_string */
extern void T376f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_manifest_string_validity */
extern void T376f228(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_manifest_string */
extern void T70f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_manifest_string_validity */
extern void T70f350(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_once_manifest_string */
extern void T65f1088(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_manifest_string */
extern void T65f1126(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].force_last */
extern void T286f60(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_put */
extern void T286f66(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].force */
extern void T663f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].fill_with */
extern void T661f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].extend */
extern void T661f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_put */
extern void T286f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_put */
extern void T286f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_item */
extern T6 T286f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].hash_position */
extern T6 T286f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.hash_code */
extern T6 T1502f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].resize */
extern void T286f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_resize */
extern void T286f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].special_integer_ */
extern T0* T286f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_resize */
extern void T286f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].aliased_resized_area */
extern T0* T663f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].aliased_resized_area */
extern T0* T661f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_resize */
extern void T286f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_resize */
extern void T286f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_capacity */
extern T6 T286f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_put */
extern void T286f62(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].search_position */
extern void T286f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_function_inline_agent */
extern void T1284f64(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f64ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_do_function_inline_agent */
extern void T394f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_do_function_inline_agent */
extern void T390f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_function_inline_agent */
extern void T376f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_function_inline_agent_validity */
extern void T376f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_query_inline_agent_validity */
extern void T376f291(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_query_inline_agent */
extern void T376f343(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].last */
extern T0* T382f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].is_empty */
extern T1 T382f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0d_error */
extern void T51f430(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0d */
extern T0* T759c1030(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0d_error */
extern void T35f430(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_locals_validity */
extern void T376f290(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_local_supplier */
extern void T376f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_local_supplier */
extern void T1299f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_local_variable_declaration */
extern void T376f339(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1h_error */
extern void T51f462(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1h */
extern T0* T759c1062(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpir1_error */
extern T1 T51f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpir1h_error */
extern void T35f462(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vpir1_error */
extern T1 T35f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1f_error */
extern void T51f461(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1f */
extern T0* T759c1061(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1f_error */
extern void T35f461(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1d_error */
extern void T51f460(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1d */
extern T0* T759c1060(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1d_error */
extern void T35f460(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1c_error */
extern void T51f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1c */
extern T0* T759c1059(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1c_error */
extern void T35f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrlv1b_error */
extern void T51f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrlv1b */
extern T0* T759c1058(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrlv1_error */
extern T1 T51f134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrlv1b_error */
extern void T35f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vrlv1_error */
extern T1 T35f134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrlv2b_error */
extern void T51f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrlv2b */
extern T0* T759c1057(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrlv2_error */
extern T1 T51f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrlv2b_error */
extern void T35f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vrlv2_error */
extern T1 T35f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vred0d_error */
extern void T51f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vred0d */
extern T0* T759c1056(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vred0d_error */
extern void T35f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_across_components_validity */
extern void T376f289(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_object_tests_validity */
extern void T376f288(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot3b_error */
extern void T51f455(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot3b */
extern T0* T759c1055(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot3b_error */
extern void T35f455(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_result_supplier */
extern void T376f287(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_result_supplier */
extern void T1299f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_result_declaration */
extern void T376f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_formal_arguments_validity */
extern void T376f285(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_argument_supplier */
extern void T376f338(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_argument_supplier */
extern void T1299f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_formal_argument_declaration */
extern void T376f337(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1g_error */
extern void T51f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1g */
extern T0* T759c1054(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1g_error */
extern void T35f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1e_error */
extern void T51f453(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1e */
extern T0* T759c1053(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1e_error */
extern void T35f453(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1b_error */
extern void T51f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1b */
extern T0* T759c1052(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1b_error */
extern void T35f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1a_error */
extern void T51f451(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1a */
extern T0* T759c1051(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1a_error */
extern void T35f451(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrfa0b_error */
extern void T51f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrfa0b */
extern T0* T759c1050(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vrfa0b_error */
extern void T35f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vred0c_error */
extern void T51f449(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vred0c */
extern T0* T759c1049(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vred0c_error */
extern void T35f449(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0i_error */
extern void T51f429(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0i */
extern T0* T759c1029(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0i_error */
extern void T35f429(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.hide_across_components */
extern void T392f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACROSS_CURSOR_SCOPE.count */
extern T6 T392f4(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.hide_object_tests */
extern void T389f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_function_inline_agent */
extern void T70f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_function_inline_agent_validity */
extern void T70f341(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_inline_agent_validity */
extern void T70f407(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_query_inline_agent */
extern void T70f455(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent */
extern void T70f474(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T70f474ot1(T0* a1, T0** a2);
extern T1 T70f474ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_inline_agent_result_dynamic_types */
extern void T70f481(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_locals_validity */
extern void T70f406(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_local_supplier */
extern void T70f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_local_variable_declaration */
extern void T70f451(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_across_components_validity */
extern void T70f405(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_object_tests_validity */
extern void T70f404(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_result_supplier */
extern void T70f403(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_result_declaration */
extern void T70f402(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_formal_arguments_validity */
extern void T70f401(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_argument_supplier */
extern void T70f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_formal_argument_declaration */
extern void T70f449(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_do_function_inline_agent */
extern void T65f1089(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function_inline_agent */
extern void T65f1128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function_inline_agent_body_declaration */
extern void T65f1127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_routine_inline_agent_body_declaration */
extern void T65f1146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_trace_message_call */
extern void T65f1160(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_function_inline_agent */
extern void T1284f65(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f65ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_function_inline_agent */
extern void T394f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_function_inline_agent */
extern void T390f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_function_inline_agent */
extern void T376f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_function_inline_agent_validity */
extern void T376f230(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir3a_error */
extern void T51f431(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpir3a */
extern T0* T759c1031(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpir3_error */
extern T1 T51f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpir3a_error */
extern void T35f431(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vpir3_error */
extern T1 T35f124(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.first_indexing */
extern T0* T1504f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_function_inline_agent */
extern void T70f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_function_inline_agent_validity */
extern void T70f342(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_once_function_inline_agent */
extern void T65f1090(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_function_inline_agent */
extern void T65f1130(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_once_per_object */
extern T1 T1504f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.standard_once_keys */
extern T0* T1504f30(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_function_inline_agent_body_declaration */
extern void T65f1129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_function_inline_agent */
extern void T1284f66(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f66ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_external_function_inline_agent */
extern void T394f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_external_function_inline_agent */
extern void T390f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_function_inline_agent */
extern void T376f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_function_inline_agent_validity */
extern void T376f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir3b_error */
extern void T51f432(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpir3b */
extern T0* T759c1032(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vpir3b_error */
extern void T35f432(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_function_inline_agent */
extern void T70f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_inline_agent_validity */
extern void T70f343(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_external_function_inline_agent */
extern void T65f1091(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function_inline_agent */
extern void T65f1132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function_inline_agent_body_declaration */
extern void T65f1131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_procedure_inline_agent */
extern void T1284f67(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f67ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_do_procedure_inline_agent */
extern void T394f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_do_procedure_inline_agent */
extern void T390f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_procedure_inline_agent */
extern void T376f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_procedure_inline_agent_validity */
extern void T376f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_procedure_inline_agent_validity */
extern void T376f292(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_procedure_inline_agent */
extern void T376f344(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_procedure_inline_agent */
extern void T70f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_procedure_inline_agent_validity */
extern void T70f344(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_procedure_inline_agent_validity */
extern void T70f408(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_procedure_inline_agent */
extern void T70f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_do_procedure_inline_agent */
extern void T65f1092(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure_inline_agent */
extern void T65f1134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure_inline_agent_body_declaration */
extern void T65f1133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_procedure_inline_agent */
extern void T1284f68(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f68ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_procedure_inline_agent */
extern void T394f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_procedure_inline_agent */
extern void T390f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_procedure_inline_agent */
extern void T376f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_procedure_inline_agent_validity */
extern void T376f233(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.first_indexing */
extern T0* T1507f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_procedure_inline_agent */
extern void T70f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_procedure_inline_agent_validity */
extern void T70f345(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_once_procedure_inline_agent */
extern void T65f1093(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_procedure_inline_agent */
extern void T65f1136(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_once_per_object */
extern T1 T1507f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.standard_once_keys */
extern T0* T1507f29(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_procedure_inline_agent_body_declaration */
extern void T65f1135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_procedure_inline_agent */
extern void T1284f69(GE_context* ac, T0* C, T0* a1);
extern T1 T1284f69ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_external_procedure_inline_agent */
extern void T394f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_external_procedure_inline_agent */
extern void T390f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_procedure_inline_agent */
extern void T376f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_procedure_inline_agent_validity */
extern void T376f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_procedure_inline_agent */
extern void T70f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_inline_agent_validity */
extern void T70f346(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_external_procedure_inline_agent */
extern void T65f1094(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_procedure_inline_agent */
extern void T65f1138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_procedure_inline_agent_body_declaration */
extern void T65f1137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_class_type */
extern void T1567f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_class_type_validity */
extern void T1567f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.set_fatal_error */
extern void T1567f6(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_class_type */
extern void T1566f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_actual_parameters_validity */
extern void T1566f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_class_type */
extern void T1565f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_class_type_validity */
extern void T1565f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.set_fatal_error */
extern void T1565f14(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_ancestors_error */
extern T1 T560f46(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.ancestors_built */
extern T1 T560f45(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_class_type */
extern void T1564f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_class_type_validity */
extern void T1564f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_syntax_error */
extern T1 T560f44(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.is_parsed */
extern T1 T560f43(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.set_fatal_error */
extern void T1564f6(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.is_preparsed */
extern T1 T560f42(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.process_class_type */
extern void T1296f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.check_class_type_validity */
extern void T1296f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1296f9ot1(T0* a1, T0** a2);
extern T1 T1296f9ot2(T0* a1, T0** a2);
/* [detachable] ET_PARENT_CHECKER3.error_handler */
extern T0* T1296f5(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.set_fatal_error */
extern void T1296f10(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_class_type */
extern void T1288f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_class_type */
extern void T1288f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolve_unfolded_tuple_actual_parameters_2 */
extern void T71f132(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.make */
extern T0* T1901c19(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.make */
extern T0* T1902c22(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_CLASS_TYPE.resolve_unfolded_tuple_actual_parameters_1 */
extern void T71f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.position */
extern T0* T1902f15(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.null_position */
extern T0* T72f794(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.tokens */
extern T0* T1902f19(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.current_universe */
extern T0* T1288f5(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_class_type */
extern void T1287f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_actual_parameters_validity */
extern void T1287f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_class_type */
extern void T1282f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_class_type_constraint */
extern void T1282f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg3a_error */
extern void T51f390(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg3a */
extern T0* T759c998(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtcg3_error */
extern T1 T51f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtcg3a_error */
extern void T35f390(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vtcg3_error */
extern T1 T35f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.error_handler */
extern T0* T1282f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.resolve_unfolded_tuple_actual_parameters_2 */
extern void T73f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.set_fatal_error */
extern void T1282f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.resolve_unfolded_tuple_actual_parameters_1 */
extern void T73f322(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.current_universe */
extern T0* T1282f6(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.process_class_type */
extern void T1281f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.check_class_type_validity */
extern void T1281f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.error_handler */
extern T0* T1281f8(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.current_system */
extern T0* T1281f7(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.set_fatal_error */
extern void T1281f13(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.current_universe */
extern T0* T1281f6(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.process_class_type */
extern void T1280f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_class_type_tuple_actual_parameters_unfolded */
extern void T1280f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_class_type_vtct_validity */
extern void T1280f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_class_type */
extern void T1279f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_actual_parameters_to_sorter */
extern void T1279f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_class_type */
extern void T1278f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_class_type */
extern void T1278f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.current_universe */
extern T0* T1278f5(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_class_type */
extern void T1277f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_actual_parameters */
extern void T1277f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.reset_fatal_error */
extern void T1277f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_class_type */
extern void T1266f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_class_type_constraint */
extern void T1266f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_base_type_dependency */
extern void T1266f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvtcg5a_error */
extern void T51f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_gvtcg5a */
extern T0* T759c997(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvtcg5_error */
extern T1 T51f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvtcg5a_error */
extern void T35f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_gvtcg5_error */
extern T1 T35f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_reference */
extern T1 T1450f32(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_reference */
extern T1 T867f33(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvtcg5b_error */
extern void T51f388(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_gvtcg5b */
extern T0* T759c996(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_gvtcg5b_error */
extern void T35f388(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_expanded */
extern T1 T1450f29(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_expanded */
extern T1 T867f30(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtug2a_error */
extern void T51f387(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtug2a */
extern T0* T759c995(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtug2_error */
extern T1 T51f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtug2a_error */
extern void T35f387(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtug2_error */
extern T1 T35f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtug1a_error */
extern void T51f386(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtug1a */
extern T0* T759c994(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtug1_error */
extern T1 T51f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtug1a_error */
extern void T35f386(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtug1_error */
extern T1 T35f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_class_type */
extern void T1265f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_class_type_validity */
extern void T1265f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENT_CHECKER1.current_universe */
extern T0* T1265f7(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.process_class_type */
extern void T380f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_class_type_validity */
extern void T380f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.reset_fatal_error */
extern void T380f40(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TYPE_CHECKER.current_universe */
extern T0* T380f20(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_formal_parameter_type */
extern void T1567f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_formal_parameter_type */
extern void T1566f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_formal_parameter_type */
extern void T1565f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_formal_parameter_type */
extern void T1564f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_formal_parameter_type */
extern void T1296f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_formal_parameter_type */
extern void T1288f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_formal_parameter_type */
extern void T1287f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_formal_parameter_type */
extern void T1282f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_formal_parameter_type */
extern void T1281f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_formal_parameter_type */
extern void T1280f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_formal_parameter_type */
extern void T1279f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_formal_parameter_type */
extern void T1278f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_formal_parameter_type */
extern void T1277f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_formal_parameter_type */
extern void T1266f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_formal_parameter_type_constraint */
extern void T1266f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_PARENT_CHECKER1.process_formal_parameter_type */
extern void T1265f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_formal_parameter_type */
extern void T380f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_function */
extern void T1284f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_external_function */
extern void T1283f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_function */
extern void T376f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_function_validity */
extern void T376f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_function */
extern void T70f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_validity */
extern void T70f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_warning_message */
extern void T51f398(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_warning_message */
extern void T35f398(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.mark_string_type_alive */
extern void T70f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.result_type_set */
extern T0* T70f76(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_builtin_function_validity */
extern void T70f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_type_generic_parameter_type */
extern void T70f209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_type_default */
extern void T70f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_tuple_reference_item */
extern void T70f207(GE_context* ac, T0* C, T0* a1);
extern T1 T70f207ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_special_item */
extern void T70f206(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_type_conforms_to */
extern void T70f205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_type_conforms_to_feature */
extern void T61f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_storable_version_of_type */
extern void T70f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field_at_offset */
extern void T70f203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field_at */
extern void T70f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field */
extern void T70f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_type_instance_of */
extern void T70f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_tuple_instance_of */
extern void T70f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_special_of_reference_instance_of */
extern void T70f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_instance_of */
extern void T70f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_object_id */
extern void T70f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_id_object */
extern void T70f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_function_item */
extern void T70f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_factory_exception_manager */
extern void T70f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_type_of_code */
extern void T70f192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_last_exception */
extern void T70f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_raisable */
extern void T70f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_ignored */
extern void T70f189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_ignorable */
extern void T70f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_caught */
extern void T70f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_exception_from_code */
extern void T70f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_function */
extern void T70f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_any_twin */
extern void T70f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin_unknown */
extern T1 T247f78(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_validity */
extern void T70f174p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_external_function */
extern void T65f1263(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function */
extern void T65f1281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_tuple_type */
extern void T1567f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_tuple_type_validity */
extern void T1567f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_tuple_type */
extern void T1566f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_tuple_type */
extern void T1565f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_tuple_type_validity */
extern void T1565f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_tuple_type */
extern void T1564f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_tuple_type_validity */
extern void T1564f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_tuple_type */
extern void T1296f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.check_tuple_type_validity */
extern void T1296f19(GE_context* ac, T0* C, T0* a1);
extern T1 T1296f19ot1(T0* a1, T0** a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_tuple_type */
extern void T1288f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_tuple_type */
extern void T1288f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_tuple_type */
extern void T1287f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_tuple_type */
extern void T1282f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_tuple_type_constraint */
extern void T1282f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_tuple_type */
extern void T1281f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.check_tuple_type_validity */
extern void T1281f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_tuple_type */
extern void T1280f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_type_tuple_actual_parameters_unfolded */
extern void T1280f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_type_vtct_validity */
extern void T1280f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_tuple_type */
extern void T1279f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_tuple_type */
extern void T1278f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_tuple_type */
extern void T1278f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_tuple_type */
extern void T1277f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_tuple_type */
extern void T1266f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_tuple_type_constraint */
extern void T1266f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_PARENT_CHECKER1.process_tuple_type */
extern void T1265f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_tuple_type_validity */
extern void T1265f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhpr5a_error */
extern void T51f520(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvhpr5a */
extern T0* T759c1119(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhpr5_error */
extern T1 T51f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhpr5a_error */
extern void T35f520(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhpr5_error */
extern T1 T35f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_tuple_type */
extern void T380f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_tuple_type_validity */
extern void T380f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_qualified_call_instruction */
extern void T1284f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_qualified_call_instruction */
extern void T376f383(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_qualified_call_instruction */
extern void T70f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_qualified_call_instruction */
extern void T65f1264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unqualified_call_expression */
extern void T1284f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_unqualified_call_expression */
extern void T394f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_unqualified_call_expression */
extern void T390f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unqualified_call_expression */
extern void T376f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_expression_validity */
extern void T376f247(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f247ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex1a_error */
extern void T51f439(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuex1a */
extern T0* T759c1039(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuex1_error */
extern T1 T51f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuex1a_error */
extern void T35f439(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuex1_error */
extern T1 T35f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn2c_error */
extern void T51f438(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn2c */
extern T0* T759c1038(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn2c_error */
extern void T35f438(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_object_test_local_call_expression_validity */
extern void T376f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0c_error */
extern void T51f477(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0c */
extern T0* T759c1077(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuao_error */
extern T1 T51f139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0c_error */
extern void T35f477(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuao_error */
extern T1 T35f139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0a_error */
extern void T51f476(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0a */
extern T0* T759c1076(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0a_error */
extern void T35f476(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0b_error */
extern void T51f475(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0b */
extern T0* T759c1075(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0b_error */
extern void T35f475(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_local_parenthesis_call_validity */
extern void T376f358(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_identifier_parenthesis_call_validity */
extern void T376f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T376f378ot1(T0* a1, T0** a2);
extern T1 T376f378ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_across_cursor_call_expression_validity */
extern void T376f310(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0c_error */
extern void T51f474(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0c */
extern T0* T759c1074(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuac_error */
extern T1 T51f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0c_error */
extern void T35f474(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuac_error */
extern T1 T35f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0a_error */
extern void T51f473(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0a */
extern T0* T759c1073(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0a_error */
extern void T35f473(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0b_error */
extern void T51f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0b */
extern T0* T759c1072(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0b_error */
extern void T35f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_across_cursor_parenthesis_call_validity */
extern void T376f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_local_variable_call_expression_validity */
extern void T376f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvual0a_error */
extern void T51f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvual0a */
extern T0* T759c1071(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvual_error */
extern T1 T51f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvual0a_error */
extern void T35f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvual_error */
extern T1 T35f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvual0b_error */
extern void T51f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvual0b */
extern T0* T759c1070(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvual0b_error */
extern void T35f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_local_variable_parenthesis_call_validity */
extern void T376f356(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_formal_argument_call_expression_validity */
extern void T376f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuaa0a_error */
extern void T51f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuaa0a */
extern T0* T759c1069(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuaa_error */
extern T1 T51f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuaa0a_error */
extern void T35f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuaa_error */
extern T1 T35f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuaa0b_error */
extern void T51f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuaa0b */
extern T0* T759c1068(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuaa0b_error */
extern void T35f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_formal_argument_parenthesis_call_validity */
extern void T376f355(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unqualified_call_expression */
extern void T70f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_expression_validity */
extern void T70f347(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f347ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_object_test_local_call_expression_validity */
extern void T70f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_local_parenthesis_call_validity */
extern void T70f460(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_identifier_parenthesis_call_validity */
extern void T70f475(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T70f475ot1(T0* a1, T0** a2);
extern T1 T70f475ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_across_cursor_call_expression_validity */
extern void T70f411(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_cursor_parenthesis_call_validity */
extern void T70f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_local_variable_call_expression_validity */
extern void T70f410(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_variable_parenthesis_call_validity */
extern void T70f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_formal_argument_call_expression_validity */
extern void T70f409(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_argument_parenthesis_call_validity */
extern void T70f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_unqualified_call_expression */
extern void T65f1265(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unqualified_call_expression */
extern void T65f1282(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1282ot1(T0* a1, T0** a2);
extern T1 T65f1282ot2(T0* a1, T0** a2);
extern T1 T65f1282ot3(T0* a1, T0** a2);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_call_agent */
extern T1 T279f16(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_constant_attribute */
extern T1 T298f29(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_constant_attribute */
extern T1 T232f27(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].is_empty */
extern T1 T263f11(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_object_test */
extern void T1284f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_object_test */
extern void T394f48(GE_context* ac, T0* C, T0* a1);
extern T1 T394f48ot1(T0* a1);
extern T1 T394f48ot2(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE.add_name */
extern void T393f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_attribute */
extern void T393f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].force_last */
extern void T724f84(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_formal_argument */
extern void T393f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_local_variable */
extern void T393f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_result */
extern void T393f14(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_object_test */
extern void T390f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_object_test */
extern void T376f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_validity */
extern void T376f243(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_object_test */
extern void T376f303(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_object_test_type */
extern void T376f302(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_object_test */
extern void T70f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_validity */
extern void T70f334(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test */
extern void T70f393(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test_type */
extern void T70f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_object_test */
extern void T65f1266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_object_equality_expression */
extern void T1284f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_object_equality_expression */
extern void T394f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_object_equality_expression */
extern void T390f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_object_equality_expression */
extern void T376f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_equality_expression_validity */
extern void T376f242(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_object_equality_expression */
extern void T376f301(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vweq0b_error */
extern void T51f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vweq0b */
extern T0* T759c1037(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_BRACKET_SYMBOL.text */
extern T0* T610f69(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.text */
extern T0* T343f40(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vweq_error */
extern T1 T51f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vweq0b_error */
extern void T35f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vweq_error */
extern T1 T35f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_left */
extern void T309f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_right */
extern void T309f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.name */
extern T0* T309f16(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.twin */
extern T0* T180f93(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.copy */
extern void T180f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.tokens */
extern T0* T309f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_object_equality_expression */
extern void T70f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_equality_expression_validity */
extern void T70f330(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_equality_expression */
extern void T70f390(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_object_equality_expression_target_dynamic_types */
extern void T70f439(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.make */
extern T0* T1989c6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.process_object_equality_expression */
extern void T65f1267(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_equality_expression */
extern void T1284f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_equality_expression */
extern void T394f46(GE_context* ac, T0* C, T0* a1);
extern T1 T394f46ot1(T0* a1);
extern T1 T394f46ot2(T0* a1);
extern T1 T394f46ot3(T0* a1, T0** a2);
extern T1 T394f46ot4(T0* a1);
extern T1 T394f46ot5(T0* a1);
extern T1 T394f46ot6(T0* a1, T0** a2);
/* [detachable] ET_BRACKET_SYMBOL.is_equal_symbol */
extern T1 T610f46(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_equal_symbol */
extern T1 T343f37(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_equality_expression */
extern void T390f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_equality_expression */
extern void T376f189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_equality_expression_validity */
extern void T376f241(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_equality_expression */
extern void T376f300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vweq0a_error */
extern void T51f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vweq0a */
extern T0* T759c1036(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vweq0a_error */
extern void T35f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_EQUALITY_EXPRESSION.set_left */
extern void T310f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EQUALITY_EXPRESSION.set_right */
extern void T310f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_equality_expression */
extern void T70f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_equality_expression_validity */
extern void T70f329(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_equality_expression */
extern void T70f389(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_equality_expression_target_dynamic_types */
extern void T70f438(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.make */
extern T0* T1988c6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.process_equality_expression */
extern void T65f1268(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_create_expression */
extern void T1284f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_create_expression */
extern void T394f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_create_expression */
extern void T390f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_create_expression */
extern void T376f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_create_expression_validity */
extern void T376f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_creation_expression_validity */
extern void T376f296(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6b_error */
extern void T51f463(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6b */
extern T0* T759c1063(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6b_error */
extern void T35f463(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_dotnet_creation_procedure_call_expression_validity */
extern void T376f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_expression_validity */
extern void T376f345(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6);
extern T1 T376f345ot1(T0* a1, T0** a2);
extern T1 T376f345ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_validity */
extern void T376f370(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6c_error */
extern void T51f509(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6c */
extern T0* T759c1109(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6c_error */
extern void T35f509(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.first */
extern T0* T237f47(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_formal_creation_procedure_call_validity */
extern void T376f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FEATURE_CHECKER.check_formal_parameter_creation_vape_validity */
extern void T376f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape2b_error */
extern void T51f510(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape2b */
extern T0* T759c1110(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vape2b_error */
extern void T35f510(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_CLIENT.reset */
extern void T398f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc8a_error */
extern void T51f508(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc8a */
extern T0* T759c1108(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc8_error */
extern T1 T51f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc8a_error */
extern void T35f508(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc8_error */
extern T1 T35f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_no_creation_procedure_call_validity */
extern void T376f368(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc1a_error */
extern void T51f507(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc1a */
extern T0* T759c1107(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc1_error */
extern T1 T51f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc1a_error */
extern void T35f507(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc1_error */
extern T1 T35f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc5a_error */
extern void T51f506(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc5a */
extern T0* T759c1106(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc5_error */
extern T1 T51f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc5a_error */
extern void T35f506(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc5_error */
extern T1 T35f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_creation_type_validity */
extern void T376f367(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.type_position */
extern T0* T1993f21(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.type_position */
extern T0* T320f23(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_create_supplier */
extern void T376f366(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_create_supplier */
extern void T1299f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_NULL_LEAF.column */
extern T6 T2039f3(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.line */
extern T6 T2039f2(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.last_position */
extern T0* T1993f20(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.last_position */
extern T0* T320f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.creation_call */
extern T0* T1993f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_create_expression */
extern void T70f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_create_expression_validity */
extern void T70f320(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_expression_validity */
extern void T70f375(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_dotnet_creation_procedure_call_expression_validity */
extern void T70f430(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_expression_validity */
extern void T70f429(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6);
extern T1 T70f429ot1(T0* a1, T0** a2);
extern T1 T70f429ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_validity */
extern void T70f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_creation_procedure_call_validity */
extern void T70f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_parameter_creation_vape_validity */
extern void T70f476(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_no_creation_procedure_call_validity */
extern void T70f465(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_type_validity */
extern void T70f464(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_create_supplier */
extern void T70f463(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_create_expression */
extern void T65f1269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_create_expression */
extern void T65f1283(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_extended_attribute */
extern void T1284f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_extended_attribute */
extern void T1283f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_extended_attribute */
extern void T376f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_extended_attribute_validity */
extern void T376f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0a_error */
extern void T51f363(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0a */
extern T0* T759c971(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vucr0a_error */
extern void T35f363(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0e_error */
extern void T51f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0e */
extern T0* T759c970(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0e_error */
extern void T35f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_COMPOUND.has_non_null_instruction */
extern T1 T330f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_extended_attribute */
extern void T70f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_extended_attribute_validity */
extern void T70f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_extended_attribute */
extern void T65f1270(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_extended_attribute */
extern void T65f1284(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_attribute */
extern void T65f1287(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_attribute_wrapper */
extern void T65f1294(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_name */
extern void T279f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_constant_attribute */
extern void T1284f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_constant_attribute */
extern void T1283f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_constant_attribute */
extern void T376f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_constant_attribute_validity */
extern void T376f111(GE_context* ac, T0* C, T0* a1);
extern T1 T376f111ot1(T0* a1, T0** a2);
extern T1 T376f111ot2(T0* a1, T0** a2);
extern T1 T376f111ot3(T0* a1, T0** a2);
extern T1 T376f111ot4(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc5a_error */
extern void T51f371(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc5a */
extern T0* T759c979(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc5_error */
extern T1 T51f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc5a_error */
extern void T35f371(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc5_error */
extern T1 T35f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc5b_error */
extern void T51f370(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc5b */
extern T0* T759c978(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vqmc5b_error */
extern void T35f370(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc4a_error */
extern void T51f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc4a */
extern T0* T759c977(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc4_error */
extern T1 T51f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc4a_error */
extern void T35f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc4_error */
extern T1 T35f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc3a_error */
extern void T51f368(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc3a */
extern T0* T759c976(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc3_error */
extern T1 T51f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc3a_error */
extern void T35f368(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc3_error */
extern T1 T35f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc3b_error */
extern void T51f367(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc3b */
extern T0* T759c975(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vqmc3b_error */
extern void T35f367(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_integer_constant_validity */
extern void T376f125(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_64_constant */
extern void T376f143(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_32_constant */
extern void T376f142(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_16_constant */
extern void T376f141(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_8_constant */
extern void T376f140(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_64_constant */
extern void T376f139(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_32_constant */
extern void T376f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_16_constant */
extern void T376f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2a_error */
extern void T51f376(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2a */
extern T0* T759c984(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2a_error */
extern void T35f376(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_integer_8_constant */
extern void T376f136(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0a_error */
extern void T51f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0a */
extern T0* T759c983(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0a_error */
extern void T35f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc2a_error */
extern void T51f366(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc2a */
extern T0* T759c974(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc2_error */
extern T1 T51f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc2a_error */
extern void T35f366(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc2_error */
extern T1 T35f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc2b_error */
extern void T51f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc2b */
extern T0* T759c973(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T759f973ot1(T0* a1, T0** a2);
extern T1 T759f973ot2(T0* a1, T0** a2);
/* [detachable] ET_ERROR_HANDLER.report_vqmc2b_error */
extern void T35f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc1a_error */
extern void T51f364(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc1a */
extern T0* T759c972(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc1_error */
extern T1 T51f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc1a_error */
extern void T35f364(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc1_error */
extern T1 T35f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_constant_attribute */
extern void T70f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_constant_attribute_validity */
extern void T70f176(GE_context* ac, T0* C, T0* a1);
extern T1 T70f176ot1(T0* a1, T0** a2);
extern T1 T70f176ot2(T0* a1, T0** a2);
extern T1 T70f176ot3(T0* a1, T0** a2);
extern T1 T70f176ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_integer_constant_validity */
extern void T70f182(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_64_constant */
extern void T70f220(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_32_constant */
extern void T70f219(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_16_constant */
extern void T70f218(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_8_constant */
extern void T70f217(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_64_constant */
extern void T70f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_32_constant */
extern void T70f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_16_constant */
extern void T70f214(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_8_constant */
extern void T70f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_constant_attribute */
extern void T65f1271(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_constant_attribute */
extern void T65f1285(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unique_attribute */
extern void T1284f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_unique_attribute */
extern void T1283f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unique_attribute */
extern void T376f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unique_attribute_validity */
extern void T376f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqui0a_error */
extern void T51f372(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqui0a */
extern T0* T759c980(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqui_error */
extern T1 T51f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqui0a_error */
extern void T35f372(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqui_error */
extern T1 T35f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unique_attribute */
extern void T70f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unique_attribute_validity */
extern void T70f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_unique_attribute */
extern void T65f1272(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unique_attribute */
extern void T65f1286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_attribute */
extern void T1284f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_attribute */
extern void T1283f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_attribute */
extern void T376f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_attribute_validity */
extern void T376f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_attribute */
extern void T70f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_attribute_validity */
extern void T70f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_attribute */
extern void T65f1273(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_like_current */
extern void T1567f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_like_current */
extern void T1566f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_like_current */
extern void T1565f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_like_current */
extern void T1564f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_like_current */
extern void T1296f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_like_current */
extern void T1288f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_like_current */
extern void T1287f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_like_current */
extern void T1282f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_like_current */
extern void T1281f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_like_current */
extern void T1280f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_like_current */
extern void T1279f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_like_current */
extern void T1278f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_like_current */
extern void T1277f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_current */
extern void T1266f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_current */
extern void T1265f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_like_current */
extern void T380f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_like_current_validity */
extern void T380f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c3_character_constant */
extern void T1284f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c3_character_constant */
extern void T394f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c3_character_constant */
extern void T390f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c3_character_constant */
extern void T376f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c3_character_constant_validity */
extern void T376f205(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c3_character_constant */
extern void T70f290(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c3_character_constant_validity */
extern void T70f352(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c3_character_constant */
extern void T65f1274(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_special_manifest_string */
extern void T1284f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_special_manifest_string */
extern void T394f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_special_manifest_string */
extern void T390f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_special_manifest_string */
extern void T376f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_special_manifest_string_validity */
extern void T376f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_special_manifest_string */
extern void T70f291(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_special_manifest_string_validity */
extern void T70f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_special_manifest_string */
extern void T65f1275(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_special_manifest_string */
extern void T65f1288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_verbatim_string */
extern void T1284f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_verbatim_string */
extern void T394f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_verbatim_string */
extern void T390f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_verbatim_string */
extern void T376f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_verbatim_string_validity */
extern void T376f207(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_verbatim_string */
extern void T70f292(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_verbatim_string_validity */
extern void T70f354(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_verbatim_string */
extern void T65f1276(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_verbatim_string */
extern void T65f1289(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_underscored_integer_constant */
extern void T1284f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_underscored_integer_constant */
extern void T394f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_underscored_integer_constant */
extern void T390f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_underscored_integer_constant */
extern void T376f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_underscored_integer_constant_validity */
extern void T376f208(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_underscored_integer_constant */
extern void T70f294(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_underscored_integer_constant_validity */
extern void T70f356(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_underscored_integer_constant */
extern void T65f1277(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_underscored_integer_constant */
extern void T65f1290(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_integer_constant */
extern void T65f1295(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_integer_value */
extern void T65f1296(GE_context* ac, T0* C, T11 a1, T1 a2, T0* a3, T1 a4);
/* [detachable] ET_PRECURSOR_CHECKER.process_hexadecimal_integer_constant */
extern void T1284f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_hexadecimal_integer_constant */
extern void T394f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_hexadecimal_integer_constant */
extern void T390f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_hexadecimal_integer_constant */
extern void T376f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_hexadecimal_integer_constant_validity */
extern void T376f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_hexadecimal_integer_constant */
extern void T70f295(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_hexadecimal_integer_constant_validity */
extern void T70f357(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_hexadecimal_integer_constant */
extern void T65f1278(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_hexadecimal_integer_constant */
extern void T65f1291(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_octal_integer_constant */
extern void T1284f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_octal_integer_constant */
extern void T394f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_octal_integer_constant */
extern void T390f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_octal_integer_constant */
extern void T376f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_octal_integer_constant_validity */
extern void T376f210(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_octal_integer_constant */
extern void T70f296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_octal_integer_constant_validity */
extern void T70f358(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_octal_integer_constant */
extern void T65f1279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_octal_integer_constant */
extern void T65f1292(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_binary_integer_constant */
extern void T1284f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_binary_integer_constant */
extern void T394f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_binary_integer_constant */
extern void T390f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_binary_integer_constant */
extern void T376f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_binary_integer_constant_validity */
extern void T376f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_binary_integer_constant */
extern void T70f297(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_binary_integer_constant_validity */
extern void T70f359(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_binary_integer_constant */
extern void T65f1280(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_binary_integer_constant */
extern void T65f1293(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_integer_constant */
extern void T1284f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_integer_constant */
extern void T394f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_integer_constant */
extern void T390f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_integer_constant */
extern void T376f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_integer_constant_validity */
extern void T376f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_integer_constant */
extern void T70f293(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_integer_constant_validity */
extern void T70f355(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_integer_constant */
extern void T65f1297(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_integer_constant */
extern void T65f1299(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_formal_parameter */
extern void T1567f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_formal_parameter */
extern void T1566f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_formal_parameter */
extern void T1565f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_formal_parameter */
extern void T1564f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_formal_parameter */
extern void T1296f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_formal_parameter */
extern void T1288f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_formal_parameter */
extern void T1287f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_formal_parameter */
extern void T1282f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_formal_parameter */
extern void T1281f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_formal_parameter */
extern void T1280f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_formal_parameter */
extern void T1279f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_formal_parameter */
extern void T1278f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_formal_parameter */
extern void T1277f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_formal_parameter */
extern void T1266f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_formal_parameter */
extern void T1265f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_formal_parameter */
extern void T380f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_constrained_formal_parameter */
extern void T1567f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_constrained_formal_parameter */
extern void T1566f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_constrained_formal_parameter */
extern void T1565f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_constrained_formal_parameter */
extern void T1564f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_constrained_formal_parameter */
extern void T1296f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_constrained_formal_parameter */
extern void T1288f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T1287f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_constrained_formal_parameter */
extern void T1282f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_constrained_formal_parameter */
extern void T1281f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_constrained_formal_parameter */
extern void T1280f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T1279f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_constrained_formal_parameter */
extern void T1278f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_constrained_formal_parameter */
extern void T1277f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_constrained_formal_parameter */
extern void T1266f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_constrained_formal_parameter */
extern void T1265f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T380f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_infix_cast_expression */
extern void T1284f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_infix_cast_expression */
extern void T394f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_infix_cast_expression */
extern void T390f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_infix_cast_expression */
extern void T376f384(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_infix_cast_expression_validity */
extern void T376f386(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_infix_cast_expression */
extern void T70f387(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_infix_cast_expression_validity */
extern void T70f437(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_infix_cast_expression */
extern void T65f1298(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_infix_cast_expression */
extern void T65f1300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_retry_instruction */
extern void T1284f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_retry_instruction */
extern void T376f387(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_retry_instruction_validity */
extern void T376f402(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vxrt0a_error */
extern void T51f523(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vxrt0a */
extern T0* T759c1122(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_RETRY_INSTRUCTION.position */
extern T0* T969f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vxrt_error */
extern T1 T51f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vxrt0a_error */
extern void T35f523(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vxrt_error */
extern T1 T35f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_retry_instruction */
extern void T70f486(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_retry_instruction_validity */
extern void T70f504(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_retry_instruction */
extern void T65f1301(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_retry_instruction */
extern void T65f1316(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_across_instruction */
extern void T1284f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_across_instruction */
extern void T376f388(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_instruction_validity */
extern void T376f403(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_loop_component_no_from_validity */
extern void T376f417(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.is_subset */
extern T1 T393f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].go_after */
extern void T724f85(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_go_after */
extern void T724f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_across_instruction */
extern void T70f487(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_instruction_validity */
extern void T70f505(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_component_no_from_validity */
extern void T70f519(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_across_instruction */
extern void T65f1302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_across_instruction */
extern void T65f1317(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_check_instruction */
extern void T1284f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.assertion */
extern T0* T1019f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CHECK_INSTRUCTION.item */
extern T0* T1019f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.process_check_instruction */
extern void T376f389(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_check_instruction_validity */
extern void T376f404(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.set_code_unreachable */
extern void T393f19(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_check_instruction */
extern void T70f488(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_check_instruction_validity */
extern void T70f506(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_check_instruction_validity */
extern void T70f506p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_check_instruction */
extern void T65f1303(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_check_instruction */
extern void T65f1318(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_debug_instruction */
extern void T1284f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_debug_instruction */
extern void T376f390(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_debug_instruction_validity */
extern void T376f405(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_debug_instruction */
extern void T70f489(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_debug_instruction_validity */
extern void T70f507(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_debug_instruction_validity */
extern void T70f507p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_debug_instruction */
extern void T65f1304(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_debug_instruction */
extern void T65f1319(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_if_instruction */
extern void T1284f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_part_list */
extern void T1284f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART.process */
extern void T1092f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_part */
extern void T1284f114(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART_LIST.item */
extern T0* T1095f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.process_if_instruction */
extern void T376f391(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_if_instruction_validity */
extern void T376f406(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_if_instruction */
extern void T70f490(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_if_instruction_validity */
extern void T70f508(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_if_instruction */
extern void T65f1305(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_if_instruction */
extern void T65f1320(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_inspect_instruction */
extern void T1284f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_part_list */
extern void T1284f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_part */
extern void T1284f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.item */
extern T0* T1259f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.process_inspect_instruction */
extern void T376f392(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_inspect_instruction_validity */
extern void T376f407(GE_context* ac, T0* C, T0* a1);
extern T1 T376f407ot1(T0* a1, T0** a2);
extern T1 T376f407ot2(T0* a1, T0** a2);
extern T1 T376f407ot3(T0* a1, T0** a2);
extern T1 T376f407ot4(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb2a_error */
extern void T51f526(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vomb2a */
extern T0* T759c1125(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vomb2_error */
extern T1 T51f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vomb2a_error */
extern void T35f526(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vomb2_error */
extern T1 T35f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb2b_error */
extern void T51f525(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vomb2b */
extern T0* T759c1124(GE_context* ac, T0* a1, T0* a2, T0* a3);
extern T1 T759f1124ot1(T0* a1, T0** a2);
extern T1 T759f1124ot2(T0* a1, T0** a2);
/* [detachable] ET_ERROR_HANDLER.report_vomb2b_error */
extern void T35f525(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.choice_constant */
extern T0* T376f68(GE_context* ac, T0* C, T0* a1);
extern T1 T376f68ot1(T0* a1, T0** a2);
extern T1 T376f68ot2(T0* a1, T0** a2);
extern T1 T376f68ot3(T0* a1, T0** a2);
extern T1 T376f68ot4(T0* a1, T0** a2);
extern T1 T376f68ot5(T0* a1, T0** a2);
extern T1 T376f68ot6(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.integer_choice_constant */
extern T0* T376f72(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.character_choice_constant */
extern T0* T376f71(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.arguments_count */
extern T6 T180f104(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.arguments */
extern T0* T180f108(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.arguments_count */
extern T6 T279f23(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.choice */
extern T0* T1031f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CHOICE_LIST.item */
extern T0* T1031f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb1a_error */
extern void T51f524(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vomb1a */
extern T0* T759c1123(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vomb1_error */
extern T1 T51f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vomb1a_error */
extern void T35f524(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vomb1_error */
extern T1 T35f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_inspect_instruction */
extern void T70f491(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inspect_instruction_validity */
extern void T70f509(GE_context* ac, T0* C, T0* a1);
extern T1 T70f509ot1(T0* a1, T0** a2);
extern T1 T70f509ot2(T0* a1, T0** a2);
extern T1 T70f509ot3(T0* a1, T0** a2);
extern T1 T70f509ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.choice_constant */
extern T0* T70f93(GE_context* ac, T0* C, T0* a1);
extern T1 T70f93ot1(T0* a1, T0** a2);
extern T1 T70f93ot2(T0* a1, T0** a2);
extern T1 T70f93ot3(T0* a1, T0** a2);
extern T1 T70f93ot4(T0* a1, T0** a2);
extern T1 T70f93ot5(T0* a1, T0** a2);
extern T1 T70f93ot6(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.integer_choice_constant */
extern T0* T70f98(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.character_choice_constant */
extern T0* T70f97(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_inspect_instruction */
extern void T65f1306(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_inspect_instruction */
extern void T65f1321(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1321ot1(T0* a1, T0** a2);
extern T1 T65f1321ot2(T0* a1, T0** a2);
extern T1 T65f1321ot3(T0* a1, T0** a2);
extern T1 T65f1321ot4(T0* a1, T0** a2);
extern T1 T65f1321ot5(T0* a1, T0** a2);
extern T1 T65f1321ot6(T0* a1, T0** a2);
/* INTEGER_32.abs */
extern T6 T6f32(GE_context* ac, T6* C);
/* INTEGER_32.abs_ref */
extern T6 T6f33(GE_context* ac, T6* C);
/* [detachable] ET_C_GENERATOR.choice_constant */
extern T0* T65f520(GE_context* ac, T0* C, T0* a1);
extern T1 T65f520ot1(T0* a1, T0** a2);
extern T1 T65f520ot2(T0* a1, T0** a2);
extern T1 T65f520ot3(T0* a1, T0** a2);
extern T1 T65f520ot4(T0* a1, T0** a2);
extern T1 T65f520ot5(T0* a1, T0** a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_assigner_instruction */
extern void T1284f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assigner_instruction */
extern void T376f393(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assigner_instruction_validity */
extern void T376f408(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_setter */
extern void T376f418(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vbac1a_error */
extern void T51f528(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vbac1a */
extern T0* T759c1127(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vbac1_error */
extern T1 T51f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vbac1a_error */
extern void T35f528(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vbac1_error */
extern T1 T35f165(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.set_source */
extern void T1479f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vbac2a_error */
extern void T51f527(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vbac2a */
extern T0* T759c1126(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vbac2_error */
extern T1 T51f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vbac2a_error */
extern void T35f527(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vbac2_error */
extern T1 T35f164(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.set_name */
extern void T1479f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assigner_instruction */
extern void T70f492(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assigner_instruction_validity */
extern void T70f510(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_setter */
extern void T70f520(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_assigner_instruction */
extern void T65f1307(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assigner_instruction */
extern void T65f1322(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_assignment */
extern void T1284f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assignment */
extern void T376f394(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assignment_validity */
extern void T376f409(GE_context* ac, T0* C, T0* a1);
extern T1 T376f409ot1(T0* a1);
extern T1 T376f409ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_assignment */
extern void T376f420(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_name */
extern void T393f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_attribute */
extern void T393f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_formal_argument */
extern void T393f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_local_variable */
extern void T393f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_result */
extern void T393f20(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjar0a_error */
extern void T51f529(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vjar0a */
extern T0* T759c1128(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjar_error */
extern T1 T51f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjar0a_error */
extern void T35f529(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vjar_error */
extern T1 T35f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_basic */
extern T1 T73f181(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.is_basic */
extern T1 T410f34(GE_context* ac, T0* C, T8 a1);
/* detachable ET_ASSIGNMENT.set_source */
extern void T1480f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_writable_validity */
extern void T376f419(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T376f419ot1(T0* a1, T0** a2);
extern T1 T376f419ot2(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen0a_error */
extern void T51f538(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen0a */
extern T0* T759c1137(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen_error */
extern T1 T51f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen0a_error */
extern void T35f538(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_veen_error */
extern T1 T35f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen0b_error */
extern void T51f537(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen0b */
extern T0* T759c1136(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen0b_error */
extern void T35f537(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0b_error */
extern void T51f536(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0b */
extern T0* T759c1135(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjaw_error */
extern T1 T51f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0b_error */
extern void T35f536(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vjaw_error */
extern T1 T35f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0c_error */
extern void T51f535(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0c */
extern T0* T759c1134(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0c_error */
extern void T35f535(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_attribute_assignment_target */
extern void T376f433(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0c_error */
extern void T51f534(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0c */
extern T0* T759c1133(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0c_error */
extern void T35f534(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0a_error */
extern void T51f533(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0a */
extern T0* T759c1132(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0a_error */
extern void T35f533(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_local_assignment_target */
extern void T376f432(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_result_assignment_target */
extern void T376f431(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assignment */
extern void T70f493(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assignment_validity */
extern void T70f511(GE_context* ac, T0* C, T0* a1);
extern T1 T70f511ot1(T0* a1);
extern T1 T70f511ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_assignment */
extern void T70f522(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_assignment_dynamic_types */
extern void T70f536(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_writable_validity */
extern void T70f521(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f521ot1(T0* a1, T0** a2);
extern T1 T70f521ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_attribute_assignment_target */
extern void T70f535(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_assignment_target */
extern void T70f534(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_assignment_target */
extern void T70f533(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_assignment */
extern void T65f1308(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assignment */
extern void T65f1323(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_result */
extern T1 T180f99(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_assignment_attempt */
extern void T1284f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assignment_attempt */
extern void T376f395(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assignment_attempt_validity */
extern void T376f410(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_assignment_attempt */
extern void T376f421(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjrv0a_error */
extern void T51f530(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vjrv0a */
extern T0* T759c1129(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjrv_error */
extern T1 T51f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjrv0a_error */
extern void T35f530(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vjrv_error */
extern T1 T35f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_reference */
extern T1 T237f50(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.newer_or_same_ise_version */
extern T1 T62f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.newer_or_same_ise_version */
extern T1 T371f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.ise_5_7_0 */
extern T0* T376f69(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assignment_attempt */
extern void T70f494(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assignment_attempt_validity */
extern void T70f512(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_assignment_attempt */
extern void T70f523(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_assignment_attempt_dynamic_types */
extern void T70f537(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_5_7_0 */
extern T0* T70f94(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_assignment_attempt */
extern void T65f1309(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assignment_attempt */
extern void T65f1324(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_loop_instruction */
extern void T1284f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_loop_instruction */
extern void T376f396(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_loop_instruction_validity */
extern void T376f411(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_loop_instruction */
extern void T70f495(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_instruction_validity */
extern void T70f513(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_loop_instruction */
extern void T65f1310(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_loop_instruction */
extern void T65f1325(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_bang_instruction */
extern void T1284f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_bang_instruction */
extern void T376f397(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_bang_instruction_validity */
extern void T376f412(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_creation_instruction_validity */
extern void T376f422(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6d_error */
extern void T51f539(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6d */
extern T0* T759c1138(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6d_error */
extern void T35f539(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_dotnet_creation_procedure_call_instruction_validity */
extern void T376f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5, T0* a6);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_instruction_validity */
extern void T376f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6, T0* a7);
extern T1 T376f434ot1(T0* a1, T0** a2);
extern T1 T376f434ot2(T0* a1, T0** a2);
extern T1 T376f434ot3(T0* a1);
extern T1 T376f434ot4(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_creation_instruction */
extern void T376f439(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.is_type_valid */
extern T1 T376f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.set_class_interface_error_only */
extern void T380f45(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc3a_error */
extern void T51f550(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc3a */
extern T0* T759c1149(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc3_error */
extern T1 T51f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc3a_error */
extern void T35f550(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc3_error */
extern T1 T35f174(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.last_position */
extern T0* T1485f8(GE_context* ac, T0* C);
/* detachable ET_CREATE_INSTRUCTION.last_leaf */
extern T0* T1485f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.last_leaf */
extern T0* T321f4(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.last_leaf */
extern T0* T1924f12(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.last_leaf */
extern T0* T274f12(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.last_position */
extern T0* T1484f7(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.last_leaf */
extern T0* T1484f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_bang_instruction */
extern void T70f496(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_bang_instruction_validity */
extern void T70f514(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_instruction_validity */
extern void T70f524(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_dotnet_creation_procedure_call_instruction_validity */
extern void T70f539(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5, T0* a6);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_instruction_validity */
extern void T70f538(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6, T0* a7);
extern T1 T70f538ot1(T0* a1, T0** a2);
extern T1 T70f538ot2(T0* a1, T0** a2);
extern T1 T70f538ot3(T0* a1);
extern T1 T70f538ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_creation_instruction */
extern void T70f543(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_creation_dynamic_type */
extern void T70f544(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATE_INSTRUCTION.arguments */
extern T0* T1485f10(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.arguments */
extern T0* T1484f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.is_type_valid */
extern T1 T70f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_bang_instruction */
extern void T65f1311(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_bang_instruction */
extern void T65f1326(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_creation_instruction */
extern void T65f1331(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_create_instruction */
extern void T1284f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_create_instruction */
extern void T376f398(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_create_instruction_validity */
extern void T376f413(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_create_instruction */
extern void T70f497(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_create_instruction_validity */
extern void T70f515(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_create_instruction */
extern void T65f1312(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_create_instruction */
extern void T65f1327(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_precursor_instruction */
extern void T1284f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_precursor_instruction */
extern void T376f399(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_instruction_validity */
extern void T376f414(GE_context* ac, T0* C, T0* a1);
extern T1 T376f414ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_procedure_instruction_validity */
extern void T376f423(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].force_last */
extern void T1297f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_put */
extern void T1297f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_put */
extern void T1297f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_item */
extern T6 T1297f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].hash_position */
extern T6 T1297f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].resize */
extern void T1297f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_resize */
extern void T1297f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].special_integer_ */
extern T0* T1297f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_resize */
extern void T1297f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_resize */
extern void T1297f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_item */
extern T0* T1297f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_resize */
extern void T1297f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_capacity */
extern T6 T1297f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_put */
extern void T1297f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].search_position */
extern void T1297f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_equality_tester */
extern T0* T1297f30(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_precursor_instruction */
extern void T376f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr1a_error */
extern void T51f531(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr1a */
extern T0* T759c1130(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vdpr1a_error */
extern void T35f531(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_precursor_instruction */
extern void T70f498(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_instruction_validity */
extern void T70f516(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_procedure_instruction_validity */
extern void T70f525(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_precursor_instruction */
extern void T70f540(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_precursor_instruction */
extern void T65f1313(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_precursor_instruction */
extern void T65f1328(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_static_call_instruction */
extern void T1284f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_static_call_instruction */
extern void T376f400(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_call_instruction_validity */
extern void T376f415(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_procedure_call_instruction_validity */
extern void T376f425(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_static_call_instruction */
extern void T376f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_dotnet_procedure_call_instruction_validity */
extern void T376f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_static_call_instruction */
extern void T70f499(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_call_instruction_validity */
extern void T70f517(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_procedure_call_instruction_validity */
extern void T70f527(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_call_instruction */
extern void T70f541(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_dotnet_procedure_call_instruction_validity */
extern void T70f526(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_static_call_instruction */
extern void T65f1314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_static_call_instruction */
extern void T65f1329(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_instruction_static_call */
extern void T65f1332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_PRECURSOR_CHECKER.process_unqualified_call_instruction */
extern void T1284f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unqualified_call_instruction */
extern void T376f401(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_instruction_validity */
extern void T376f416(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1c_error */
extern void T51f532(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1c */
extern T0* T759c1131(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1c_error */
extern void T35f532(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_procedure_call_instruction_validity */
extern void T376f430(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_call_instruction */
extern void T376f438(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_object_test_local_call_instruction_validity */
extern void T376f429(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0c_error */
extern void T51f549(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0c */
extern T0* T759c1148(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuio_error */
extern T1 T51f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0c_error */
extern void T35f549(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuio_error */
extern T1 T35f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0a_error */
extern void T51f548(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0a */
extern T0* T759c1147(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0a_error */
extern void T35f548(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0b_error */
extern void T51f547(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0b */
extern T0* T759c1146(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0b_error */
extern void T35f547(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_across_cursor_call_instruction_validity */
extern void T376f428(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0c_error */
extern void T51f546(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0c */
extern T0* T759c1145(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuic_error */
extern T1 T51f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0c_error */
extern void T35f546(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuic_error */
extern T1 T35f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0a_error */
extern void T51f545(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0a */
extern T0* T759c1144(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0a_error */
extern void T35f545(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0b_error */
extern void T51f544(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0b */
extern T0* T759c1143(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0b_error */
extern void T35f544(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_local_variable_call_instruction_validity */
extern void T376f427(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuil0a_error */
extern void T51f543(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuil0a */
extern T0* T759c1142(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuil_error */
extern T1 T51f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuil0a_error */
extern void T35f543(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuil_error */
extern T1 T35f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuil0b_error */
extern void T51f542(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuil0b */
extern T0* T759c1141(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuil0b_error */
extern void T35f542(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_formal_argument_call_instruction_validity */
extern void T376f426(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuia0a_error */
extern void T51f541(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuia0a */
extern T0* T759c1140(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuia_error */
extern T1 T51f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuia0a_error */
extern void T35f541(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuia_error */
extern T1 T35f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuia0b_error */
extern void T51f540(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuia0b */
extern T0* T759c1139(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuia0b_error */
extern void T35f540(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unqualified_call_instruction */
extern void T70f500(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_instruction_validity */
extern void T70f518(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_procedure_call_instruction_validity */
extern void T70f532(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_instruction */
extern void T70f542(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T70f542ot1(T0* a1, T0** a2);
extern T1 T70f542ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_object_test_local_call_instruction_validity */
extern void T70f531(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_across_cursor_call_instruction_validity */
extern void T70f530(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_local_variable_call_instruction_validity */
extern void T70f529(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_formal_argument_call_instruction_validity */
extern void T70f528(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_unqualified_call_instruction */
extern void T65f1315(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unqualified_call_instruction */
extern void T65f1330(GE_context* ac, T0* C, T0* a1);
extern T1 T65f1330ot1(T0* a1, T0** a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.is_call_agent */
extern T1 T1704f4(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_builtin_expression */
extern void T1284f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_builtin_expression */
extern void T394f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_builtin_expression */
extern void T390f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_convert_builtin_expression */
extern void T376f440(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_builtin_expression_validity */
extern void T376f443(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_builtin_expression */
extern void T70f546(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_builtin_expression_validity */
extern void T70f549(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_convert_builtin_expression */
extern void T65f1333(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_builtin_expression */
extern void T65f1336(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_from_expression */
extern void T1284f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_from_expression */
extern void T394f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_from_expression */
extern void T390f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_convert_from_expression */
extern void T376f441(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_from_expression_validity */
extern void T376f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_from_expression */
extern void T70f547(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_from_expression_validity */
extern void T70f550(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_convert_from_expression */
extern void T65f1334(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_from_expression */
extern void T65f1337(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_to_expression */
extern void T1284f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_to_expression */
extern void T394f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_to_expression */
extern void T390f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_convert_to_expression */
extern void T376f442(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_to_expression_validity */
extern void T376f445(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_to_expression */
extern void T70f548(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_to_expression_validity */
extern void T70f551(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_convert_to_expression */
extern void T65f1335(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_to_expression */
extern void T65f1338(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.reset_validity_checked */
extern void T1462f84(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_validity_checked */
extern void T1461f87(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_validity_checked */
extern void T1460f85(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_validity_checked */
extern void T367f81(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_validity_checked */
extern void T365f85(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_validity_checked */
extern void T363f84(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_validity_checked */
extern void T329f84(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_validity_checked */
extern void T247f93(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.set_frozen_keyword */
extern void T1462f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_frozen_keyword */
extern void T1461f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_frozen_keyword */
extern void T1460f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_frozen_keyword */
extern void T367f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_frozen_keyword */
extern void T365f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_frozen_keyword */
extern void T363f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_frozen_keyword */
extern void T329f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_frozen_keyword */
extern void T247f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_object_tests */
extern void T1462f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_object_tests */
extern void T1461f91(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_object_tests */
extern void T1460f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_object_tests */
extern void T367f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_object_tests */
extern void T365f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_object_tests */
extern void T363f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_object_tests */
extern void T329f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_object_tests */
extern void T247f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_across_components */
extern void T1462f89(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_across_components */
extern void T1461f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_across_components */
extern void T1460f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_across_components */
extern void T367f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_across_components */
extern void T365f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_across_components */
extern void T363f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_across_components */
extern void T329f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_across_components */
extern void T247f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.reset */
extern void T1462f85(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset */
extern void T1462f85p1(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset */
extern void T1461f88(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset */
extern void T1461f88p1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset */
extern void T1460f86(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset */
extern void T1460f86p1(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset */
extern void T367f82(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset */
extern void T367f82p1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset */
extern void T365f86(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset */
extern void T365f86p1(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset */
extern void T363f85(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset */
extern void T363f85p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset */
extern void T329f85(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset */
extern void T329f85p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset */
extern void T247f94(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset */
extern void T247f94p1(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_after_features_flattened */
extern void T1462f86(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1462f105(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.reset_qualified_anchored_types */
extern void T249f15(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_features_flattened */
extern void T1461f89(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1461f111(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_features_flattened */
extern void T1460f87(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1460f108(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_after_features_flattened */
extern void T367f83(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T367f96(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_after_features_flattened */
extern void T365f87(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T365f103(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_after_features_flattened */
extern void T363f86(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T363f100(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_features_flattened */
extern void T329f86(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T329f106(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_after_features_flattened */
extern void T247f95(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_signature_qualified_anchored_types */
extern void T247f115(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_after_interface_checked */
extern void T1462f87(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_interface_checked */
extern void T1461f90(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_interface_checked */
extern void T1461f90p1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_interface_checked */
extern void T1460f88(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_interface_checked */
extern void T1460f88p1(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_after_interface_checked */
extern void T367f84(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_after_interface_checked */
extern void T365f88(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_after_interface_checked */
extern void T363f87(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_interface_checked */
extern void T329f87(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_interface_checked */
extern void T329f87p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_after_interface_checked */
extern void T247f96(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_validity_checked */
extern void T1465f91(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_validity_checked */
extern void T1464f86(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_validity_checked */
extern void T468f82(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_validity_checked */
extern void T466f83(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_frozen_keyword */
extern void T1465f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_frozen_keyword */
extern void T1464f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_frozen_keyword */
extern void T468f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_frozen_keyword */
extern void T466f89(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_object_tests */
extern void T1465f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_object_tests */
extern void T1464f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_object_tests */
extern void T468f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_object_tests */
extern void T466f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_across_components */
extern void T1465f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_across_components */
extern void T1464f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_across_components */
extern void T468f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_across_components */
extern void T466f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.reset */
extern void T1465f92(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset */
extern void T1464f87(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset */
extern void T468f83(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset */
extern void T466f84(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_after_features_flattened */
extern void T1465f93(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T1465f112(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_features_flattened */
extern void T1464f88(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T1464f109(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_after_features_flattened */
extern void T468f84(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T468f102(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_features_flattened */
extern void T466f85(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T466f105(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_after_interface_checked */
extern void T1465f94(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_interface_checked */
extern void T1464f89(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_interface_checked */
extern void T1464f89p1(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_after_interface_checked */
extern void T468f85(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_interface_checked */
extern void T466f86(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_interface_checked */
extern void T466f86p1(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.add_creation_clients_to */
extern void T1465f120(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR_LIST.add_creation_clients_to */
extern void T1080f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR.add_creation_clients_to */
extern void T1077f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLIENT_LIST.append_first */
extern void T397f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.add_creation_clients_to */
extern void T1464f117(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DEFERRED_PROCEDURE.add_creation_clients_to */
extern void T468f110(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DO_PROCEDURE.add_creation_clients_to */
extern void T466f113(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_NAME.reset */
extern void T1468f103(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.reset */
extern void T1467f71(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.reset */
extern void T1466f89(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.reset */
extern void T864f71(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.reset */
extern void T180f121(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.set_seed */
extern void T1468f104(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_FREE_NAME.set_seed */
extern void T1467f72(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_NAME.set_seed */
extern void T1466f90(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_FREE_NAME.set_seed */
extern void T864f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_seed */
extern void T180f111(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_id */
extern void T444f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_id */
extern void T443f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_id */
extern void T311f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_id */
extern void T294f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_id */
extern void T219f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_attached_type */
extern void T444f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_attached_type */
extern void T443f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_attached_type */
extern void T311f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_attached_type */
extern void T294f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attached_type */
extern void T219f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_hash_code */
extern void T444f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_hash_code */
extern void T443f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_hash_code */
extern void T311f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_hash_code */
extern void T294f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_hash_code */
extern void T219f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_next_type */
extern void T444f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_next_type */
extern void T443f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_next_type */
extern void T311f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_next_type */
extern void T294f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_next_type */
extern void T219f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_attribute_position */
extern void T444f65(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.put */
extern void T220f12(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.index_of */
extern T6 T220f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_attribute_position */
extern void T443f65(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_attribute_position */
extern void T311f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_attribute_position */
extern void T294f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attribute_position */
extern void T219f66(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_alive */
extern void T444f58(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_alive */
extern void T443f58(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_alive */
extern void T311f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_alive */
extern void T294f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_alive */
extern void T219f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.use_all_attributes */
extern void T444f67(GE_context* ac, T0* C, T0* a1);
extern T1 T444f67ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.use_all_attributes */
extern void T443f67(GE_context* ac, T0* C, T0* a1);
extern T1 T443f67ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.use_all_attributes */
extern void T311f68(GE_context* ac, T0* C, T0* a1);
extern T1 T311f68ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.use_all_attributes */
extern void T294f68(GE_context* ac, T0* C, T0* a1);
extern T1 T294f68ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.use_all_attributes */
extern void T219f67(GE_context* ac, T0* C, T0* a1);
extern T1 T219f67ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_equality_expression */
extern void T444f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.set_next */
extern void T1988f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_equality_expression */
extern void T443f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_equality_expression */
extern void T311f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_equality_expression */
extern void T294f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_equality_expression */
extern void T219f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_object_equality_expression */
extern void T444f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.set_next */
extern void T1989f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_object_equality_expression */
extern void T443f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_object_equality_expression */
extern void T311f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_object_equality_expression */
extern void T294f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_object_equality_expression */
extern void T219f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_query_call */
extern void T444f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_query_call */
extern void T443f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_query_call */
extern void T311f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_query_call */
extern void T294f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_query_call */
extern void T219f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_procedure_call */
extern void T444f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_procedure_call */
extern void T443f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_procedure_call */
extern void T311f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_procedure_call */
extern void T294f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_procedure_call */
extern void T219f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_static */
extern void T444f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_static */
extern void T443f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_static */
extern void T311f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_static */
extern void T294f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_static */
extern void T219f73(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_LIBRARY.set_ecf_namespace */
extern void T183f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_namespace */
extern void T181f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_ecf_namespace */
extern void T56f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_ecf_version */
extern void T183f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_version */
extern void T181f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_ecf_version */
extern void T56f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_description */
extern void T183f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_description */
extern void T181f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_description */
extern void T56f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_notes */
extern void T183f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_notes */
extern void T181f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_notes */
extern void T56f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_targets */
extern void T183f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_targets */
extern void T181f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_targets */
extern void T56f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_read_only */
extern void T183f124(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_read_only */
extern void T181f18(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_read_only */
extern void T56f165(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_LIBRARY.set_uuid */
extern void T183f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_uuid */
extern void T181f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_uuid */
extern void T56f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_library_target */
extern void T183f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_library_target */
extern void T181f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_library_target */
extern void T56f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start */
extern void T550f28(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_start */
extern void T511f2(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start */
extern void T191f9(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start */
extern void T191f9p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start */
extern void T178f14(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_finish */
extern void T550f29(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_finish */
extern void T511f3(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_finish */
extern void T191f10(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_finish */
extern void T191f10p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_finish */
extern void T178f15(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_comment */
extern void T550f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_comment */
extern void T511f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_comment */
extern void T191f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_comment */
extern void T191f11p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_comment */
extern void T178f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.handle_position */
extern void T178f24(GE_context* ac, T0* C, T0* a1);
/* detachable XM_POSITION_TABLE.put */
extern void T165f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].put_last */
extern void T874f9(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].put_right */
extern void T1755f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T1755c3(GE_context* ac, T0* a1);
/* [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE].make */
extern T0* T875c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_COMMENT.make_last_in_document */
extern T0* T1419c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_COMMENT.make_last */
extern T0* T1419c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.force_last */
extern void T164f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].force_last */
extern void T869f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].force_last */
extern void T869f13p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].make */
extern T0* T1698c3(GE_context* ac, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].before_addition */
extern void T869f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.before_addition */
extern void T164f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.last */
extern T0* T164f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].last */
extern T0* T869f8(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.is_empty */
extern T1 T164f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].is_empty */
extern T1 T869f7(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_processing_instruction */
extern void T550f31(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_NULL.on_processing_instruction */
extern void T511f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_processing_instruction */
extern void T191f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_processing_instruction */
extern void T191f12p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_processing_instruction */
extern void T178f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_PROCESSING_INSTRUCTION.make_last_in_document */
extern T0* T1420c5(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_PROCESSING_INSTRUCTION.make_last */
extern T0* T1420c4(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER.on_content */
extern void T550f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_content */
extern void T511f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_content */
extern void T191f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_content */
extern void T191f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_content */
extern void T178f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CHARACTER_DATA.make_last */
extern T0* T1421c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag */
extern void T550f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.push */
extern void T1333f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].force_last */
extern void T1670f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].put_right */
extern void T1886f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].attach_left */
extern void T1886f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1886c4(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.new_string_string_table */
extern T0* T1333f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_equality_tester */
extern void T212f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.string_equality_tester */
extern T0* T1333f9(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_start_tag */
extern void T511f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag */
extern void T191f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag */
extern void T191f14p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start_tag */
extern void T178f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_ELEMENT.make_root */
extern T0* T164c19(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_DOCUMENT.set_root_element */
extern void T188f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.remove_previous_root_element */
extern void T188f16(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].go_after */
extern void T880f14(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_go_after */
extern void T569f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.make_last */
extern T0* T164c18(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.new_namespace */
extern T0* T178f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].force_last */
extern void T553f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_put */
extern void T553f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_put */
extern void T553f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_item */
extern T6 T553f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].hash_position */
extern T6 T553f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE.hash_code */
extern T6 T439f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].resize */
extern void T553f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_resize */
extern void T553f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].special_integer_ */
extern T0* T553f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_storage_resize */
extern void T553f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_resize */
extern void T553f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].aliased_resized_area */
extern T0* T1340f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].aliased_resized_area */
extern T0* T1341f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_storage_item */
extern T0* T553f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_item */
extern T0* T553f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_item */
extern T6 T553f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_resize */
extern void T553f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_capacity */
extern T6 T553f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_put */
extern void T553f41(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].force */
extern void T1340f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].fill_with */
extern void T1341f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].extend */
extern void T1341f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].search_position */
extern void T553f40(GE_context* ac, T0* C, T0* a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE].test */
extern T1 T1338f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE.is_equal */
extern T1 T439f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE.string_ */
extern T0* T439f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_equality_tester */
extern T0* T553f25(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE.same_prefix */
extern T1 T439f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item */
extern T0* T553f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].has */
extern T1 T553f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag_finish */
extern void T550f33(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag_finish */
extern void T550f33p1(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.resolve_default */
extern T0* T1333f3(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.default_pseudo_prefix */
extern T0* T1333f7(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.string_ */
extern T0* T550f14(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_delayed_attributes */
extern void T550f40(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_remove */
extern void T550f44(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].remove */
extern void T1337f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].wipe_out */
extern void T1337f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].remove */
extern void T1336f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].wipe_out */
extern void T1336f9(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.unprefixed_attribute_namespace */
extern T0* T550f19(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.default_namespace */
extern T0* T550f25(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xmlns_namespace */
extern T0* T550f18(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.is_xmlns */
extern T1 T550f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.same_string */
extern T1 T550f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE_RESOLVER.string_equality_tester */
extern T0* T550f26(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xmlns */
extern T0* T550f22(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xml_prefix_namespace */
extern T0* T550f17(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.is_xml */
extern T1 T550f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.xml_prefix */
extern T0* T550f24(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].item */
extern T0* T1337f4(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].item */
extern T0* T1336f5(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_is_empty */
extern T1 T550f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].is_empty */
extern T1 T1336f4(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.resolve */
extern T0* T1333f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].back */
extern void T1671f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_back */
extern void T1670f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].add_traversing_cursor */
extern void T1670f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_next_cursor */
extern void T1671f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].remove_traversing_cursor */
extern void T1670f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set */
extern void T1671f14(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].go_before */
extern void T1671f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_go_before */
extern void T1670f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_before */
extern void T1671f15(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_off */
extern T1 T1670f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].off */
extern T1 T1671f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].item */
extern T0* T1671f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].finish */
extern void T1671f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_finish */
extern void T1670f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.default_namespace */
extern T0* T1333f6(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.has */
extern T1 T1333f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.has_prefix */
extern T1 T550f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_start_tag_finish */
extern void T511f7(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag_finish */
extern void T191f15(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag_finish */
extern void T191f15p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start_tag_finish */
extern void T178f20(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_end_tag */
extern void T550f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.pop */
extern void T1333f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].remove_last */
extern void T1670f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_last_cell */
extern void T1670f14(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].forget_right */
extern void T1886f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].move_last_cursors_after */
extern void T1670f13(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_after */
extern void T1671f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].wipe_out */
extern void T1670f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].move_all_cursors_after */
extern void T1670f15(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_end_tag */
extern void T550f34p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_NULL.on_end_tag */
extern void T511f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_end_tag */
extern void T191f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_end_tag */
extern void T191f16p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_end_tag */
extern void T178f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_ELEMENT.parent_element */
extern T0* T164f10(GE_context* ac, T0* C);
extern T1 T164f10ot1(T0* a1, T0** a2);
/* [detachable] XM_ELEMENT.is_root_node */
extern T1 T164f11(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.is_root_node */
extern T1 T188f8(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.parent */
extern T0* T188f9(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_error */
extern void T550f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_error */
extern void T511f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_error */
extern void T191f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_error */
extern void T191f17p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_error */
extern void T178f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.on_attribute */
extern void T550f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.add */
extern void T1333f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_new */
extern void T212f95(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].last */
extern T0* T1670f5(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.shallow_has */
extern T1 T1333f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_force */
extern void T550f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].force */
extern void T1337f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].force */
extern void T1336f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [detachable STRING_8].put_right */
extern void T1672f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [detachable STRING_8].make */
extern T0* T1672c3(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.add_default */
extern void T1333f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_attribute */
extern void T511f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_attribute */
extern void T191f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_attribute */
extern void T191f18p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_attribute */
extern void T178f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable XM_ATTRIBUTE.make_last */
extern T0* T192c5(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_NAMESPACE_RESOLVER.on_xml_declaration */
extern void T550f45(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_CALLBACKS_NULL.on_xml_declaration */
extern void T511f12(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_xml_declaration */
extern void T191f19(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_xml_declaration */
extern void T178f25(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_NAMESPACE_RESOLVER.set_next */
extern void T550f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.set_next */
extern void T191f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.set_next */
extern void T178f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_line */
extern void T293f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_line */
extern void T256f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_line */
extern void T157f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.append */
extern void T293f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.append */
extern void T293f37p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.append */
extern void T256f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.append */
extern void T256f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.append */
extern void T157f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.append */
extern void T157f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_string */
extern void T293f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_put_string */
extern void T293f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_ps */
extern void T293f49(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_string */
extern void T256f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.string_ */
extern T0* T256f3(GE_context* ac, T0* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_string */
extern void T157f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_new_line */
extern void T293f33(GE_context* ac, T0* C);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_new_line */
extern void T256f5(GE_context* ac, T0* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_new_line */
extern void T157f7(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_character */
extern void T293f31(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_put_character */
extern void T293f43(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_pc */
extern void T293f50(GE_context* ac, T0* C, T14 a1, T2 a2);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_character */
extern void T256f7(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_character */
extern void T157f8(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_integer */
extern void T293f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_integer_64 */
extern void T293f45(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_integer */
extern void T256f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_integer_64 */
extern void T256f11(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_integer */
extern void T157f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_integer_64 */
extern void T157f10(GE_context* ac, T0* C, T7 a1);
/* [detachable] ET_TUPLE_TYPE.append_runtime_name_to_string */
extern void T248f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.append_runtime_name_to_string */
extern void T73f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.append_runtime_name_to_string */
extern void T71f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.process */
extern void T248f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.process */
extern void T73f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.process */
extern void T71f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_add_adapted_base_classes_to_list */
extern void T248f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_add_adapted_base_classes_to_list */
extern void T73f325(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_add_adapted_base_classes_to_list */
extern void T71f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.add_adapted_base_classes_to_list */
extern void T248f124(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_adapted_base_classes_to_list */
extern void T73f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.add_adapted_base_classes_to_list */
extern void T71f133(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.append_to_string */
extern void T248f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.append_to_string */
extern void T73f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.append_to_string */
extern void T71f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.reset */
extern void T140f24(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.reset */
extern void T47f28(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].make */
extern T0* T156c8(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].set_internal_cursor */
extern void T156f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].new_cursor */
extern T0* T156f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32].make */
extern T0* T494c3(GE_context* ac, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.reset */
extern void T46f31(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].make */
extern T0* T153c8(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].set_internal_cursor */
extern void T153f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].new_cursor */
extern T0* T153f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN].make */
extern T0* T492c3(GE_context* ac, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.reset */
extern void T45f31(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].make */
extern T0* T146c13(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].set_internal_cursor */
extern void T146f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].make */
extern T0* T488f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].default_create */
extern T0* T488c3(GE_context* ac);
/* [detachable] AP_FLAG.reset */
extern void T43f22(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.reset */
extern void T41f30(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.record_occurrence */
extern void T140f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.display_help */
extern void T140f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.exceptions */
extern T0* T140f17(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.report_info_message */
extern void T144f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.full_help_text */
extern T0* T140f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.full_help_text */
extern T0* T39f35(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.option_help_text */
extern T0* T39f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].sort */
extern void T1891f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].sort_with_comparator */
extern void T1891f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].subsort_with_comparator */
extern void T1891f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].swap */
extern void T141f28(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].replace */
extern void T141f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] AP_OPTION_COMPARATOR.less_than */
extern T1 T1890f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STRING_ROUTINES.three_way_comparison */
extern T6 T84f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] AP_OPTION_COMPARATOR.string_ */
extern T0* T1890f2(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].item */
extern T0* T141f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].make */
extern T0* T1891c2(GE_context* ac, T0* a1);
/* [detachable] AP_OPTION_COMPARATOR.default_create */
extern T0* T1890c3(GE_context* ac);
/* [detachable] AP_INTEGER_OPTION.record_occurrence */
extern void T47f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].force_last */
extern void T156f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_BOOLEAN_OPTION.record_occurrence */
extern void T46f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.false_strings */
extern T0* T46f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].set_equality_tester */
extern void T143f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.case_insensitive_string_equality_tester */
extern T0* T46f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].has */
extern T1 T143f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.true_strings */
extern T0* T46f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].force_last */
extern void T153f10(GE_context* ac, T0* C, T1 a1);
/* detachable DS_LINKABLE [[attached] BOOLEAN].put_right */
extern void T491f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] BOOLEAN].make */
extern T0* T491c3(GE_context* ac, T1 a1);
/* [detachable] AP_ENUMERATION_OPTION.record_occurrence */
extern void T45f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].has */
extern T1 T148f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].force_last */
extern void T146f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].force */
extern void T488f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].resize */
extern void T146f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].aliased_resized_area */
extern T0* T488f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_capacity */
extern T6 T146f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].extendible */
extern T1 T146f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_FLAG.record_occurrence */
extern void T43f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.record_occurrence */
extern void T41f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDIN_FILE.close */
extern void T909f13(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.close */
extern void T876f38(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.close */
extern void T32f62(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.read_character */
extern void T909f14(GE_context* ac, T0* C);
/* [detachable] CONSOLE.read_character */
extern void T896f30(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_readchar */
extern T2 T896f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] CONSOLE.end_of_file */
extern T1 T896f11(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_eof */
extern T1 T896f13(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_STDIN_FILE.console */
extern T0* T909f7(GE_context* ac, T0* C);
/* [detachable] STD_FILES.input */
extern T0* T352f5(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stdin */
extern T0* T896c29(GE_context* ac, T0* a1);
/* [detachable] CONSOLE.set_read_mode */
extern void T896f32(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.io */
extern T0* T909f9(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_character */
extern void T876f32(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].remove */
extern void T1573f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].wipe_out */
extern void T1573f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.noqueue_read_character */
extern void T876f34(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16_read_character */
extern void T876f36(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.append_character */
extern void T876f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].force */
extern void T1573f7(GE_context* ac, T0* C, T2 a1);
/* detachable DS_LINKABLE [[attached] CHARACTER_8].put_right */
extern void T1847f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] CHARACTER_8].make */
extern T0* T1847c3(GE_context* ac, T2 a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf8 */
extern T0* T876f22(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf8_buffer */
extern T0* T876f25(GE_context* ac, T0* C);
/* [detachable] UC_UTF16_ROUTINES.surrogate */
extern T6 T1574f7(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_low_surrogate */
extern T1 T1574f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF16_ROUTINES.least_10_bits */
extern T6 T1574f5(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_high_surrogate */
extern T1 T1574f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF16_ROUTINES.is_surrogate */
extern T1 T1574f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16 */
extern T0* T876f20(GE_context* ac, T0* C);
/* [detachable] UC_UTF16_ROUTINES.default_create */
extern T0* T1574c16(GE_context* ac);
/* [detachable] XM_EIFFEL_INPUT_STREAM.least_significant */
extern T6 T876f24(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] XM_EIFFEL_INPUT_STREAM.most_significant */
extern T6 T876f23(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] XM_EIFFEL_INPUT_STREAM.latin1_read_character */
extern void T876f35(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.read_character */
extern void T1427f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16_detect_read_character */
extern void T876f33(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character */
extern T1 T416f39(GE_context* ac, T0* C, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character_start */
extern T1 T416f40(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] UC_UTF16_ROUTINES.is_endian_detection_character_least_first */
extern T1 T1574f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_endian_detection_character_most_first */
extern T1 T1574f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.read_character */
extern void T32f67(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_read_character */
extern void T32f71(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_gc */
extern T2 T32f26(GE_context* ac, T0* C, T14 a1);
/* [detachable] SERIALIZATION_FAILURE.set_throwing_exception */
extern void T121f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_throwing_exception */
extern void T120f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_throwing_exception */
extern void T119f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_throwing_exception */
extern void T118f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_throwing_exception */
extern void T117f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_throwing_exception */
extern void T116f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_throwing_exception */
extern void T115f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_throwing_exception */
extern void T114f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_throwing_exception */
extern void T113f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_throwing_exception */
extern void T112f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_throwing_exception */
extern void T111f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_throwing_exception */
extern void T110f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_throwing_exception */
extern void T109f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_throwing_exception */
extern void T108f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_throwing_exception */
extern void T107f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_throwing_exception */
extern void T106f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_throwing_exception */
extern void T105f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_throwing_exception */
extern void T103f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_throwing_exception */
extern void T102f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_throwing_exception */
extern void T101f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_throwing_exception */
extern void T100f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_throwing_exception */
extern void T99f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_throwing_exception */
extern void T98f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_throwing_exception */
extern void T97f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_throwing_exception */
extern void T95f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_throwing_exception */
extern void T93f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_throwing_exception */
extern void T91f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_exception_trace */
extern void T121f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_exception_trace */
extern void T120f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_exception_trace */
extern void T119f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_exception_trace */
extern void T118f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_exception_trace */
extern void T117f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_exception_trace */
extern void T116f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_exception_trace */
extern void T115f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_exception_trace */
extern void T114f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_exception_trace */
extern void T113f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_exception_trace */
extern void T112f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_exception_trace */
extern void T111f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_exception_trace */
extern void T110f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_exception_trace */
extern void T109f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_exception_trace */
extern void T108f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_exception_trace */
extern void T107f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_exception_trace */
extern void T106f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_exception_trace */
extern void T105f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_exception_trace */
extern void T103f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_exception_trace */
extern void T102f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_exception_trace */
extern void T101f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_exception_trace */
extern void T100f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_exception_trace */
extern void T99f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_exception_trace */
extern void T98f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_exception_trace */
extern void T97f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_exception_trace */
extern void T95f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_exception_trace */
extern void T93f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_exception_trace */
extern void T91f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_description */
extern void T121f9(GE_context* ac, T0* C, T0* a1);
/* detachable C_STRING.set_count */
extern void T30f8(GE_context* ac, T0* C, T6 a1);
/* UTF_CONVERTER.utf_32_string_into_utf_8_0_pointer */
extern void T471f12(GE_context* ac, T471* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_description */
extern void T120f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_description */
extern void T119f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_description */
extern void T118f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_description */
extern void T117f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_description */
extern void T116f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_description */
extern void T115f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_description */
extern void T114f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_description */
extern void T113f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_description */
extern void T112f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_description */
extern void T111f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_description */
extern void T110f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_description */
extern void T109f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_description */
extern void T108f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_description */
extern void T107f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_description */
extern void T106f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_description */
extern void T105f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_description */
extern void T103f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_description */
extern void T102f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_description */
extern void T101f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_description */
extern void T100f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_description */
extern void T99f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_description */
extern void T98f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_description */
extern void T97f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_description */
extern void T95f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_description */
extern void T93f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_description */
extern void T91f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_recipient_name */
extern void T121f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_recipient_name */
extern void T120f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_recipient_name */
extern void T119f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_recipient_name */
extern void T118f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_recipient_name */
extern void T117f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_recipient_name */
extern void T116f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_recipient_name */
extern void T115f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_recipient_name */
extern void T114f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_recipient_name */
extern void T113f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_recipient_name */
extern void T112f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_recipient_name */
extern void T111f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_recipient_name */
extern void T110f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_recipient_name */
extern void T109f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_recipient_name */
extern void T108f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_recipient_name */
extern void T107f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_recipient_name */
extern void T106f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_recipient_name */
extern void T105f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_recipient_name */
extern void T103f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_recipient_name */
extern void T102f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_recipient_name */
extern void T101f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_recipient_name */
extern void T100f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_recipient_name */
extern void T99f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_recipient_name */
extern void T98f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_recipient_name */
extern void T97f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_recipient_name */
extern void T95f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_recipient_name */
extern void T93f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_recipient_name */
extern void T91f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_type_name */
extern void T121f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_type_name */
extern void T120f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_type_name */
extern void T119f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_type_name */
extern void T118f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_type_name */
extern void T117f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_type_name */
extern void T116f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_type_name */
extern void T115f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_type_name */
extern void T114f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_type_name */
extern void T113f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_type_name */
extern void T112f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_type_name */
extern void T111f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_type_name */
extern void T110f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_type_name */
extern void T109f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_type_name */
extern void T108f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_type_name */
extern void T107f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_type_name */
extern void T106f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_type_name */
extern void T105f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_type_name */
extern void T103f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_type_name */
extern void T102f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_type_name */
extern void T101f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_type_name */
extern void T100f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_type_name */
extern void T99f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_type_name */
extern void T98f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_type_name */
extern void T97f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_type_name */
extern void T95f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_type_name */
extern void T93f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_type_name */
extern void T91f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2124f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2123f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2120f2(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATE_INSTRUCTION.position */
extern T0* T1485f9(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.position */
extern T0* T1484f8(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.target */
extern T0* T1704f5(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.target */
extern T0* T279f17(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.arguments */
extern T0* T1993f15(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.name */
extern T0* T1489f7(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.name */
extern T0* T1230f12(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.name */
extern T0* T321f3(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.name */
extern T0* T1478f39(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.name */
extern T0* T1477f35(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.target */
extern T0* T1479f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_freeop */
extern T1 T1470f13(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_freeop */
extern T1 T1467f29(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_freeop */
extern T1 T951f25(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.operator_name */
extern T0* T1470f14(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.operator_name */
extern T0* T1467f17(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.operator_name */
extern T0* T864f8(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.position */
extern T0* T930f9(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.kind_lower_name */
extern T0* T826f95(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.kind_lower_name */
extern T0* T776f65(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.kind_lower_name */
extern T0* T183f104(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.full_name */
extern T0* T861f17(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_NONE_GROUP.relative_name */
extern T0* T861f21(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_NONE_GROUP.string_ */
extern T0* T861f15(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.relative_name */
extern T0* T36f130(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_SYSTEM.string_ */
extern T0* T36f123(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].is_empty */
extern T1 T882f10(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.relative_name */
extern T0* T56f155(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_SYSTEM.string_ */
extern T0* T56f151(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.full_name */
extern T0* T776f62(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.relative_name */
extern T0* T776f66(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_ECF_CLUSTER.relative_name */
extern T0* T776f66p1(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.relative_name */
extern T0* T183f112(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.string_ */
extern T0* T183f115(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.full_name */
extern T0* T406f18(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_UNKNOWN_GROUP.relative_name */
extern T0* T406f22(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_UNKNOWN_GROUP.string_ */
extern T0* T406f17(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.parent */
extern T0* T861f18(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.parent */
extern T0* T406f19(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.absolute_pathname */
extern T0* T861f19(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.file_system */
extern T0* T861f25(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.unix_file_system */
extern T0* T861f29(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.windows_file_system */
extern T0* T861f28(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.operating_system */
extern T0* T861f27(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.full_pathname */
extern T0* T861f24(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.pathname */
extern T0* T861f26(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.execution_environment */
extern T0* T861f23(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.absolute_pathname */
extern T0* T776f63(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.execution_environment */
extern T0* T776f68(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.absolute_pathname */
extern T0* T406f20(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.file_system */
extern T0* T406f26(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.unix_file_system */
extern T0* T406f30(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.windows_file_system */
extern T0* T406f29(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.operating_system */
extern T0* T406f28(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.full_pathname */
extern T0* T406f25(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.pathname */
extern T0* T406f27(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.execution_environment */
extern T0* T406f24(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.last_position */
extern T0* T2039f4(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.last_position */
extern T0* T1445f46(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.last_position */
extern T0* T1444f45(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.last_position */
extern T0* T1443f38(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.last_position */
extern T0* T1442f38(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.last_position */
extern T0* T1440f40(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.last_position */
extern T0* T1439f40(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.last_position */
extern T0* T1438f62(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.last_position */
extern T0* T1437f62(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.last_position */
extern T0* T1436f62(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.last_position */
extern T0* T1435f62(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.last_position */
extern T0* T1434f38(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.last_position */
extern T0* T1433f34(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.last_position */
extern T0* T1432f46(GE_context* ac, T0* C);
/* detachable ET_VOID.last_position */
extern T0* T975f29(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.last_position */
extern T0* T933f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.last_position */
extern T0* T863f34(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.last_position */
extern T0* T610f107(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.last_position */
extern T0* T409f42(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.last_position */
extern T0* T366f62(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.last_position */
extern T0* T343f41(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.last_position */
extern T0* T301f30(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.last_position */
extern T0* T273f29(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.last_position */
extern T0* T180f100(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_once_per_process */
extern T1 T1507f25(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_process_key */
extern T1 T1802f1(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_once_per_process */
extern T1 T1506f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_once_per_process */
extern T1 T1504f26(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_once_per_process */
extern T1 T1503f24(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.selected_count */
extern T6 T1520f25(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.selected_count */
extern T6 T1303f27(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.has_replication */
extern T1 T1520f22(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.has_replication */
extern T1 T1303f24(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.replicated_feature */
extern T0* T1632f4(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.replicated_feature */
extern T0* T1520f23(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.replicated_feature */
extern T0* T1303f25(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.first_feature */
extern T0* T1632f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].first */
extern T0* T1633f8(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.first_feature */
extern T0* T1520f24(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.first_feature */
extern T0* T1303f26(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.position */
extern T0* T1513f6(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.position */
extern T0* T1473f6(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.assertion */
extern T0* T1513f3(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1508f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1507f27(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1506f25(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1505f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1504f28(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1503f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.assertion */
extern T0* T1502f13(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.assertion */
extern T0* T1501f18(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.assertion */
extern T0* T1500f11(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.assertion */
extern T0* T1499f13(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.assertion */
extern T0* T1498f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.assertion */
extern T0* T1495f13(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.assertion */
extern T0* T1494f11(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.assertion */
extern T0* T1493f11(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.assertion */
extern T0* T1492f11(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.assertion */
extern T0* T1491f11(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.assertion */
extern T0* T1490f15(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.assertion */
extern T0* T1473f3(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.assertion */
extern T0* T1445f21(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.assertion */
extern T0* T1444f21(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.assertion */
extern T0* T1443f20(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.assertion */
extern T0* T1442f20(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.assertion */
extern T0* T1440f21(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.assertion */
extern T0* T1439f21(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.assertion */
extern T0* T1438f21(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.assertion */
extern T0* T1437f21(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.assertion */
extern T0* T1436f21(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.assertion */
extern T0* T1435f21(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.assertion */
extern T0* T1434f13(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.assertion */
extern T0* T1433f13(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.assertion */
extern T0* T1432f11(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.assertion */
extern T0* T1233f17(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.assertion */
extern T0* T1230f15(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.assertion */
extern T0* T1201f13(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.assertion */
extern T0* T1187f18(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.assertion */
extern T0* T1140f17(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.assertion */
extern T0* T1016f16(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.assertion */
extern T0* T1013f13(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.assertion */
extern T0* T986f24(GE_context* ac, T0* C);
/* detachable ET_VOID.assertion */
extern T0* T975f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.assertion */
extern T0* T863f24(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.assertion */
extern T0* T715f15(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.assertion */
extern T0* T366f27(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.assertion */
extern T0* T320f15(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.assertion */
extern T0* T310f11(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.assertion */
extern T0* T309f11(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.assertion */
extern T0* T308f12(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.assertion */
extern T0* T301f20(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.assertion */
extern T0* T281f18(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.assertion */
extern T0* T279f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.assertion */
extern T0* T278f15(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.assertion */
extern T0* T273f20(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.assertion */
extern T0* T180f79(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.base_type_index_of_label */
extern T6 T1794f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.adapted_class_base_type_index_of_label */
extern T6 T248f113(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.adapted_class_base_type_index_of_label */
extern T6 T73f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_index_of_label */
extern T6 T237f49(GE_context* ac, T0* C, T0* a1);
extern T1 T237f49ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.adapted_class_base_type_index_of_label */
extern T6 T71f113(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_procedure */
extern T0* T1794f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_procedure */
extern T0* T248f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_procedure */
extern T0* T73f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROCEDURE_LIST.named_feature */
extern T0* T254f5(GE_context* ac, T0* C, T0* a1);
extern T1 T254f5ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.named_procedure */
extern T0* T71f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_query */
extern T0* T1794f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_query */
extern T0* T248f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_query */
extern T0* T73f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUERY_LIST.named_feature */
extern T0* T253f6(GE_context* ac, T0* C, T0* a1);
extern T1 T253f6ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.named_query */
extern T0* T71f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.immediate_feature */
extern T0* T1520f10(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_immediate_feature */
extern T0* T1465f57(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_immediate_feature */
extern T0* T1464f52(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_immediate_feature */
extern T0* T1462f55(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_immediate_feature */
extern T0* T1461f59(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_immediate_feature */
extern T0* T1460f57(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.immediate_feature */
extern T0* T1303f11(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_immediate_feature */
extern T0* T468f48(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_immediate_feature */
extern T0* T466f51(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_immediate_feature */
extern T0* T367f35(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_immediate_feature */
extern T0* T365f37(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_immediate_feature */
extern T0* T363f36(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_immediate_feature */
extern T0* T329f38(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_immediate_feature */
extern T0* T247f38(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_feature */
extern T0* T1465f55(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_feature */
extern T0* T1464f50(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_feature */
extern T0* T1462f53(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_feature */
extern T0* T1461f57(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_feature */
extern T0* T1460f55(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_feature */
extern T0* T468f46(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_feature */
extern T0* T466f49(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_feature */
extern T0* T367f33(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_feature */
extern T0* T365f35(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_feature */
extern T0* T363f34(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_feature */
extern T0* T329f36(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_feature */
extern T0* T247f36(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_redeclared */
extern T1 T1465f59(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_redeclared */
extern T1 T1464f54(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_redeclared */
extern T1 T1462f57(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_redeclared */
extern T1 T1461f61(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_redeclared */
extern T1 T1460f59(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_redeclared */
extern T1 T1303f13(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_redeclared */
extern T1 T468f50(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_redeclared */
extern T1 T466f53(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_redeclared */
extern T1 T367f37(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_redeclared */
extern T1 T365f39(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_redeclared */
extern T1 T363f38(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_redeclared */
extern T1 T329f40(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_redeclared */
extern T1 T247f40(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.redeclared_feature */
extern T0* T1520f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_redeclared_feature */
extern T0* T1465f60(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_redeclared_feature */
extern T0* T1464f55(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_redeclared_feature */
extern T0* T1462f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_redeclared_feature */
extern T0* T1461f62(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_redeclared_feature */
extern T0* T1460f60(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.redeclared_feature */
extern T0* T1303f14(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_redeclared_feature */
extern T0* T468f51(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_redeclared_feature */
extern T0* T466f54(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T367f38(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T365f40(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T363f39(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T329f41(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_redeclared_feature */
extern T0* T247f41(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_inherited */
extern T1 T1520f14(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_inherited */
extern T1 T1465f61(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_inherited */
extern T1 T1464f56(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_inherited */
extern T1 T1462f59(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_inherited */
extern T1 T1461f63(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_inherited */
extern T1 T1460f61(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_inherited */
extern T1 T468f52(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_inherited */
extern T1 T466f55(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_inherited */
extern T1 T367f39(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_inherited */
extern T1 T365f41(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_inherited */
extern T1 T363f40(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_inherited */
extern T1 T329f42(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_inherited */
extern T1 T247f42(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.inherited_feature */
extern T0* T1520f15(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_inherited_feature */
extern T0* T1465f62(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_inherited_feature */
extern T0* T1464f57(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_inherited_feature */
extern T0* T1462f60(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_inherited_feature */
extern T0* T1461f64(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_inherited_feature */
extern T0* T1460f62(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.inherited_feature */
extern T0* T1303f16(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_inherited_feature */
extern T0* T468f53(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_inherited_feature */
extern T0* T466f56(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_inherited_feature */
extern T0* T367f40(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_inherited_feature */
extern T0* T365f42(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_inherited_feature */
extern T0* T363f41(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_inherited_feature */
extern T0* T329f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_inherited_feature */
extern T0* T247f43(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_adapted */
extern T1 T1465f63(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_adapted */
extern T1 T1464f58(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_adapted */
extern T1 T1462f61(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_adapted */
extern T1 T1461f65(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_adapted */
extern T1 T1460f63(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_adapted */
extern T1 T468f54(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_adapted */
extern T1 T466f57(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_adapted */
extern T1 T367f41(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_adapted */
extern T1 T365f43(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_adapted */
extern T1 T363f42(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_adapted */
extern T1 T329f44(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_adapted */
extern T1 T247f44(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.adapted_feature */
extern T0* T1520f11(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_adapted_feature */
extern T0* T1465f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_adapted_feature */
extern T0* T1464f53(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_adapted_feature */
extern T0* T1462f56(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_adapted_feature */
extern T0* T1461f60(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_adapted_feature */
extern T0* T1460f58(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.adapted_feature */
extern T0* T1303f12(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_adapted_feature */
extern T0* T468f49(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_adapted_feature */
extern T0* T466f52(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_adapted_feature */
extern T0* T367f36(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_adapted_feature */
extern T0* T365f38(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_adapted_feature */
extern T0* T363f37(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_adapted_feature */
extern T0* T329f39(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_adapted_feature */
extern T0* T247f39(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_feature */
extern T0* T1794f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_feature */
extern T0* T248f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_feature */
extern T0* T73f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_feature */
extern T0* T71f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.base_class */
extern T0* T1794f7(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.types_to_text */
extern T0* T1794f5(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.append_types_to_string */
extern void T1794f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.types_to_text */
extern T0* T1608f8(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.append_types_to_string */
extern void T1608f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.conforms_to_type_with_type_marks */
extern T1 T1794f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.conforms_to_type_with_type_marks */
extern T1 T1608f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.type_constraint */
extern T0* T1794f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.type_constraint */
extern T0* T1608f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.item */
extern T0* T1608f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.types_to_text */
extern T0* T1478f73(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_types_to_string */
extern void T1478f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.types_to_text */
extern T0* T1477f76(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_types_to_string */
extern void T1477f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.types_to_text */
extern T0* T1476f74(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.append_types_to_string */
extern void T1476f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.types_to_text */
extern T0* T1451f5(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.append_types_to_string */
extern void T1451f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.types_to_text */
extern T0* T1244f9(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.append_types_to_string */
extern void T1244f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.types_to_text */
extern T0* T442f71(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.append_types_to_string */
extern void T442f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.types_to_text */
extern T0* T248f105(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.append_types_to_string */
extern void T248f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.types_to_text */
extern T0* T246f70(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_types_to_string */
extern void T246f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.types_to_text */
extern T0* T73f245(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.append_types_to_string */
extern void T73f328(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.types_to_text */
extern T0* T71f94(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.append_types_to_string */
extern void T71f134(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_constraint */
extern T0* T1478f72(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_constraint */
extern T0* T1477f75(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.type_constraint */
extern T0* T1476f73(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.type_constraint */
extern T0* T1451f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.type_constraint */
extern T0* T1244f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LIKE_CURRENT.type_constraint */
extern T0* T442f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.type_constraint */
extern T0* T248f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_constraint */
extern T0* T246f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.type_constraint */
extern T0* T73f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.type_constraint */
extern T0* T71f86(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.last_leaf */
extern T0* T1451f7(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.last_leaf */
extern T0* T1057f11(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.last_leaf */
extern T0* T1244f10(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.current_type_constraint */
extern T0* T1478f69(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.current_type_constraint */
extern T0* T1477f72(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.current_type_constraint */
extern T0* T1476f70(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.current_type_constraint */
extern T0* T1451f6(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.current_type_constraint */
extern T0* T442f67(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.current_type_constraint */
extern T0* T248f101(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.current_type_constraint */
extern T0* T246f66(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.current_type_constraint */
extern T0* T73f246(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.current_type_constraint */
extern T0* T71f114(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.renames */
extern T0* T1478f75(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.renames */
extern T0* T1477f78(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.renames */
extern T0* T1476f77(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.renames */
extern T0* T442f74(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.renames */
extern T0* T248f109(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.renames */
extern T0* T246f72(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.renames */
extern T0* T73f225(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.renames */
extern T0* T71f96(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_constraint_type */
extern T0* T1478f70(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_constraint_type */
extern T0* T1477f73(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.type_constraint_type */
extern T0* T1476f71(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.type_constraint_type */
extern T0* T442f68(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type_constraint_type */
extern T0* T248f102(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_constraint_type */
extern T0* T246f67(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.type_constraint_type */
extern T0* T73f212(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type_constraint_type */
extern T0* T71f95(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.like_keyword */
extern T0* T1478f34(GE_context* ac, T0* C);
/* detachable ET_CUSTOM_ATTRIBUTE.indexing_term */
extern T0* T1449f7(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.indexing_term */
extern T0* T1445f35(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.indexing_term */
extern T0* T1444f34(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.indexing_term */
extern T0* T1443f30(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.indexing_term */
extern T0* T1442f30(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.indexing_term */
extern T0* T1440f32(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.indexing_term */
extern T0* T1439f32(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.indexing_term */
extern T0* T1438f49(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.indexing_term */
extern T0* T1437f49(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.indexing_term */
extern T0* T1436f49(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.indexing_term */
extern T0* T1435f49(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.indexing_term */
extern T0* T1434f41(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.indexing_term */
extern T0* T1433f37(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.indexing_term */
extern T0* T1432f35(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.indexing_term */
extern T0* T863f37(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.indexing_term */
extern T0* T366f50(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.indexing_term */
extern T0* T180f97(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.manifest_string */
extern T0* T1434f28(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.manifest_string */
extern T0* T1433f10(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.manifest_string */
extern T0* T863f16(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.is_convert_from */
extern T1 T1457f7(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.is_convert_from */
extern T1 T559f4(GE_context* ac, T0* C);
/* detachable ET_CONVERT_PROCEDURE.is_convert_to */
extern T1 T1458f7(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.is_convert_to */
extern T1 T559f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.resolved_syntactical_constraint */
extern T0* T1703f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_generic_named_type */
extern T0* T449f992(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T449f992ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.resolved_syntactical_constraint */
extern T0* T1051f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_actual_parameter_list */
extern T0* T449f993(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.item */
extern T0* T1051f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.position */
extern T0* T1051f9(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_syntactical_constraint */
extern T0* T1478f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_syntactical_constraint */
extern T0* T1477f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.resolved_syntactical_constraint */
extern T0* T1476f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_NAMED_TYPE.resolved_syntactical_constraint */
extern T0* T1452f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_named_type */
extern T0* T449f988(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T449f988ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.resolved_syntactical_constraint */
extern T0* T442f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.position */
extern T0* T1703f6(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.position */
extern T0* T1454f4(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.position */
extern T0* T1453f4(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_NAMED_TYPE.position */
extern T0* T1452f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.type */
extern T0* T1703f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_NAMED_TYPE.type */
extern T0* T1452f4(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1703f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1478f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1477f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LIKE_FEATURE.resolved_syntactical_constraint_with_type */
extern T0* T1476f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.resolved_syntactical_constraint_with_type */
extern T0* T1454f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_labeled_comma_actual_parameter */
extern T0* T449f995(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.resolved_syntactical_constraint_with_type */
extern T0* T1453f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_labeled_actual_parameter */
extern T0* T449f994(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_NAMED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1452f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.resolved_syntactical_constraint_with_type */
extern T0* T442f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_KEYWORD_EXPRESSION.position */
extern T0* T1719f3(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.choice */
extern T0* T1486f5(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.choice */
extern T0* T1445f41(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.choice */
extern T0* T1444f40(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.choice */
extern T0* T1438f57(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.choice */
extern T0* T1437f57(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.choice */
extern T0* T1436f57(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.choice */
extern T0* T1435f57(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.choice */
extern T0* T1432f41(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.choice */
extern T0* T1230f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.choice */
extern T0* T366f57(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.choice */
extern T0* T279f22(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.choice */
extern T0* T180f103(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.lower */
extern T0* T1445f43(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.lower */
extern T0* T1444f42(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.lower */
extern T0* T1438f59(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.lower */
extern T0* T1437f59(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.lower */
extern T0* T1436f59(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.lower */
extern T0* T1435f59(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.lower */
extern T0* T1432f43(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.lower */
extern T0* T1230f26(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.lower */
extern T0* T366f60(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.lower */
extern T0* T279f26(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.lower */
extern T0* T180f106(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_range */
extern T1 T1445f42(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_range */
extern T1 T1444f41(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_range */
extern T1 T1438f58(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_range */
extern T1 T1437f58(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_range */
extern T1 T1436f58(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_range */
extern T1 T1435f58(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_range */
extern T1 T1432f42(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_range */
extern T1 T1230f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_range */
extern T1 T366f59(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_range */
extern T1 T279f25(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_range */
extern T1 T180f105(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.upper */
extern T0* T1445f44(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.upper */
extern T0* T1444f43(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.upper */
extern T0* T1438f60(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.upper */
extern T0* T1437f60(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.upper */
extern T0* T1436f60(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.upper */
extern T0* T1435f60(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.upper */
extern T0* T1432f44(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.upper */
extern T0* T1230f27(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.upper */
extern T0* T366f61(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.upper */
extern T0* T279f27(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.upper */
extern T0* T180f107(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.agent_actual_argument */
extern T0* T1994f14(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.agent_actual_argument */
extern T0* T1993f14(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.agent_actual_argument */
extern T0* T1992f13(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.agent_actual_argument */
extern T0* T1926f14(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.agent_actual_argument */
extern T0* T1510f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1508f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1507f33(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1506f29(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1505f27(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1504f34(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1503f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.agent_actual_argument */
extern T0* T1502f16(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.agent_actual_argument */
extern T0* T1501f21(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.agent_actual_argument */
extern T0* T1500f14(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.agent_actual_argument */
extern T0* T1499f17(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.agent_actual_argument */
extern T0* T1498f15(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.agent_actual_argument */
extern T0* T1495f17(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.agent_actual_argument */
extern T0* T1494f14(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.agent_actual_argument */
extern T0* T1493f14(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.agent_actual_argument */
extern T0* T1492f14(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.agent_actual_argument */
extern T0* T1491f14(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.agent_actual_argument */
extern T0* T1490f18(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1445f26(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1444f26(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.agent_actual_argument */
extern T0* T1443f23(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.agent_actual_argument */
extern T0* T1442f23(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.agent_actual_argument */
extern T0* T1440f25(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.agent_actual_argument */
extern T0* T1439f25(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1438f41(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1437f41(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1436f41(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1435f41(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.agent_actual_argument */
extern T0* T1434f32(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.agent_actual_argument */
extern T0* T1433f28(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1432f27(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.agent_actual_argument */
extern T0* T1233f20(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.agent_actual_argument */
extern T0* T1230f18(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.agent_actual_argument */
extern T0* T1201f16(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.agent_actual_argument */
extern T0* T1187f21(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.agent_actual_argument */
extern T0* T1140f20(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.agent_actual_argument */
extern T0* T1016f19(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.agent_actual_argument */
extern T0* T1013f16(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.agent_actual_argument */
extern T0* T986f27(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.agent_actual_argument */
extern T0* T983f35(GE_context* ac, T0* C);
/* detachable ET_VOID.agent_actual_argument */
extern T0* T975f22(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.agent_actual_argument */
extern T0* T863f27(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.agent_actual_argument */
extern T0* T715f18(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T366f30(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.agent_actual_argument */
extern T0* T320f18(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.agent_actual_argument */
extern T0* T310f14(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.agent_actual_argument */
extern T0* T309f14(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.agent_actual_argument */
extern T0* T308f15(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.agent_actual_argument */
extern T0* T301f23(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.agent_actual_argument */
extern T0* T281f21(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.agent_actual_argument */
extern T0* T279f14(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.agent_actual_argument */
extern T0* T278f18(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.agent_actual_argument */
extern T0* T273f23(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.agent_actual_argument */
extern T0* T180f82(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1438f10(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1438f24(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1437f10(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1437f25(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1436f10(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1436f24(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1435f10(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1435f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_8 */
extern T1 T366f17(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T366f35(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1438f11(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1438f25(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1437f11(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1437f26(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1436f11(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1436f25(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1435f11(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1435f26(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_16 */
extern T1 T366f18(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T366f36(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1438f14(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1437f14(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1436f14(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1435f14(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_8 */
extern T1 T366f21(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1438f15(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1437f15(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1436f15(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1435f15(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_16 */
extern T1 T366f22(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1438f16(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1437f16(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1436f16(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1435f16(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_32 */
extern T1 T366f23(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1438f12(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1438f26(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1437f12(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1437f27(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1436f12(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1436f26(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1435f12(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1435f27(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_32 */
extern T1 T366f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T366f16(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1438f13(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1438f27(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1437f13(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1437f28(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1436f13(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1436f27(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1435f13(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1435f28(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_64 */
extern T1 T366f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T366f37(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1438f18(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1437f18(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1435f18(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T366f25(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_binary */
extern T1 T1437f19(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_binary */
extern T1 T1436f19(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_binary */
extern T1 T1435f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_binary */
extern T1 T366f26(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1438f17(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1437f17(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1436f17(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1435f17(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_64 */
extern T1 T366f24(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.value_position */
extern T0* T1438f44(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.value_position */
extern T0* T1437f44(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.value_position */
extern T0* T1436f44(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.value_position */
extern T0* T1435f44(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.value_position */
extern T0* T366f46(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1438f54(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1437f54(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1436f54(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1435f54(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.to_integer_32 */
extern T6 T366f12(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_negative */
extern T1 T1438f28(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_negative */
extern T1 T1437f24(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_negative */
extern T1 T1436f28(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_negative */
extern T1 T1435f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_negative */
extern T1 T366f40(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1445f19(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.has_invalid_code */
extern T1 T1445f22(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1444f19(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.has_invalid_code */
extern T1 T1444f22(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1432f9(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1445f20(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1444f20(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1432f10(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.value_position */
extern T0* T1445f30(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.value_position */
extern T0* T1444f29(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.value_position */
extern T0* T1432f30(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infixable */
extern T1 T1470f20(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infixable */
extern T1 T886f11(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefixable */
extern T1 T1470f17(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefixable */
extern T1 T886f10(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.alias_lower_name */
extern T0* T1470f43(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.alias_name */
extern T0* T1470f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.alias_lower_name */
extern T0* T1468f70(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.alias_name */
extern T0* T1468f72(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.alias_lower_name */
extern T0* T1467f59(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.alias_name */
extern T0* T1467f61(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.alias_lower_name */
extern T0* T1466f56(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.alias_name */
extern T0* T1466f58(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.alias_lower_name */
extern T0* T886f37(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.alias_name */
extern T0* T886f63(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.alias_lower_name */
extern T0* T864f59(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.alias_name */
extern T0* T864f60(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.same_alias_name */
extern T1 T1470f45(GE_context* ac, T0* C, T0* a1);
extern T1 T1470f45ot1(T0* a1, T0** a2);
/* detachable ET_INFIX_NAME.same_alias_name */
extern T1 T1468f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.same_alias_name */
extern T1 T1467f63(GE_context* ac, T0* C, T0* a1);
extern T1 T1467f63ot1(T0* a1, T0** a2);
/* detachable ET_PREFIX_NAME.same_alias_name */
extern T1 T1466f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.same_alias_name */
extern T1 T886f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.same_alias_name */
extern T1 T864f63(GE_context* ac, T0* C, T0* a1);
extern T1 T864f63ot1(T0* a1, T0** a2);
/* [detachable] ET_DOTNET_ASSEMBLY.lower_name */
extern T0* T586f95(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.lower_name */
extern T0* T209f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.lower_name */
extern T0* T207f94(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.lower_name */
extern T0* T203f21(GE_context* ac, T0* C);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_capability_aware */
extern T1 T805f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_capability_aware */
extern T1 T804f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.is_capability_aware */
extern T1 T802f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_capability_aware */
extern T1 T801f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DOTNET_CONDITION.is_capability_aware */
extern T1 T800f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.is_capability_aware */
extern T1 T799f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_aware */
extern T1 T798f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.string_ */
extern T0* T798f6(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_aware */
extern T1 T797f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.string_ */
extern T0* T797f6(GE_context* ac, T0* C);
/* detachable ET_ECF_BUILD_CONDITION.is_capability_aware */
extern T1 T796f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_capability_supported */
extern T1 T805f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_capability_supported */
extern T1 T804f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_PLATFORM_CONDITION.is_capability_supported */
extern T1 T802f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_capability_supported */
extern T1 T801f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_DOTNET_CONDITION.is_capability_supported */
extern T1 T800f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CUSTOM_CONDITION.is_capability_supported */
extern T1 T799f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T1 T798f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T0* T798f4ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T0* T798f4ac2(GE_context* ac, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].go_after */
extern void T69f80(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_go_after */
extern void T69f83(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].put_last */
extern void T69f79(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].compress */
extern void T69f82(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].move_all_cursors */
extern void T69f84(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].do_all */
extern void T148f27(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].append_last */
extern void T69f78(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].extend_last */
extern void T69f81(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].extendible */
extern T1 T69f37(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.supported_void_safety_capability_values */
extern T0* T798f8(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.supported_void_safety_capability_values */
extern T0* T798f8ac1(GE_context* ac, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].set_hash_function */
extern void T69f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.case_insensitive_string_equality_tester */
extern T0* T798f7(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T1 T797f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T0* T797f4ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T0* T797f4ac2(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.supported_concurrency_capability_values */
extern T0* T797f8(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.supported_concurrency_capability_values */
extern T0* T797f8ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.case_insensitive_string_equality_tester */
extern T0* T797f7(GE_context* ac, T0* C);
/* detachable ET_ECF_BUILD_CONDITION.is_capability_supported */
extern T1 T796f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_enabled */
extern T1 T805f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_included */
extern T1 T805f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_enabled */
extern T1 T804f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_included */
extern T1 T804f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.is_enabled */
extern T1 T802f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_PLATFORM_CONDITION.is_enabled */
extern T0* T802f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].there_exists */
extern T1 T148f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.string_ */
extern T0* T802f7(GE_context* ac, T0* C);
/* detachable ET_ECF_PLATFORM_CONDITION.operating_system */
extern T0* T802f6(GE_context* ac, T0* C);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_enabled */
extern T1 T801f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DOTNET_CONDITION.is_enabled */
extern T1 T800f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.is_enabled */
extern T1 T799f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.compile_case_sensitive */
extern T0* T2014c19(GE_context* ac, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.string_ */
extern T0* T799f8(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_enabled */
extern T1 T798f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_enabled */
extern T0* T798f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities */
extern T0* T798f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_0_0 */
extern T0* T798f16(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_0_0 */
extern void T798f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_11_0 */
extern void T798f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_13_0 */
extern void T798f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_16_0 */
extern void T798f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_latest */
extern void T798f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_11_0 */
extern T0* T798f15(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_11_0 */
extern T0* T798f14(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_13_0 */
extern T0* T798f13(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_13_0 */
extern T0* T798f12(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_16_0 */
extern T0* T798f11(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_16_0 */
extern T0* T798f10(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_enabled */
extern T1 T797f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_enabled */
extern T0* T797f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities */
extern T0* T797f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_0_0 */
extern T0* T797f16(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_0_0 */
extern void T797f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_11_0 */
extern void T797f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_13_0 */
extern void T797f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_16_0 */
extern void T797f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_latest */
extern void T797f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_11_0 */
extern T0* T797f15(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_11_0 */
extern T0* T797f14(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_13_0 */
extern T0* T797f13(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_13_0 */
extern T0* T797f12(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_16_0 */
extern T0* T797f11(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_16_0 */
extern T0* T797f10(GE_context* ac, T0* C);
/* detachable ET_ECF_BUILD_CONDITION.is_enabled */
extern T1 T796f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_BUILD_CONDITION.is_enabled */
extern T0* T796f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_BUILD_CONDITION.string_ */
extern T0* T796f6(GE_context* ac, T0* C);
/* detachable ET_CUSTOM_ATTRIBUTE.has_indexing_term_value */
extern T1 T1449f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1445f33(GE_context* ac, T0* C, T0* a1);
/* CHARACTER_32.as_lower */
extern T3 T3f8(GE_context* ac, T3* C);
/* [detachable] CHARACTER_PROPERTY.to_lower */
extern T3 T2029f1(GE_context* ac, T0* C, T3 a1);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_15 */
extern T0* T2029f17(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_14 */
extern T0* T2029f16(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_13 */
extern T0* T2029f15(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_12 */
extern T0* T2029f14(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_11 */
extern T0* T2029f13(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_10 */
extern T0* T2029f12(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_9 */
extern T0* T2029f11(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_8 */
extern T0* T2029f10(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_7 */
extern T0* T2029f9(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_6 */
extern T0* T2029f8(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_5 */
extern T0* T2029f7(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_4 */
extern T0* T2029f6(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_3 */
extern T0* T2029f5(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_2 */
extern T0* T2029f4(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_1 */
extern T0* T2029f3(GE_context* ac, T0* C);
/* CHARACTER_32.is_less_equal */
extern T1 T3f11(GE_context* ac, T3* C, T3 a1);
/* CHARACTER_32.is_less */
extern T1 T3f12(GE_context* ac, T3* C, T3 a1);
/* CHARACTER_32.properties */
extern T0* T3f9(GE_context* ac, T3* C);
/* [detachable] CHARACTER_PROPERTY.make */
extern T0* T2029c18(GE_context* ac);
/* detachable ET_C1_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1444f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.has_indexing_term_value */
extern T1 T1443f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.string_ */
extern T0* T1443f31(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.has_indexing_term_value */
extern T1 T1442f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.string_ */
extern T0* T1442f31(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.has_indexing_term_value */
extern T1 T1440f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.is_plus */
extern T1 T972f57(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.string_ */
extern T0* T1440f33(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.has_indexing_term_value */
extern T1 T1439f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.string_ */
extern T0* T1439f33(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1438f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.string_ */
extern T0* T1438f50(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1437f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.string_ */
extern T0* T1437f50(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1436f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.string_ */
extern T0* T1436f50(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1435f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.string_ */
extern T0* T1435f50(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.has_indexing_term_value */
extern T1 T1434f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.string_ */
extern T0* T1434f43(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.has_indexing_term_value */
extern T1 T1433f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.string_ */
extern T0* T1433f39(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1432f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.has_indexing_term_value */
extern T1 T863f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.string_ */
extern T0* T863f38(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T366f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.string_ */
extern T0* T366f51(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.has_indexing_term_value */
extern T1 T180f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CUSTOM_ATTRIBUTE.position */
extern T0* T1449f6(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1445f37(GE_context* ac, T0* C);
/* CHARACTER_32.out */
extern T0* T3f10(GE_context* ac, T3* C);
/* detachable ET_C1_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1444f36(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.indexing_term_value */
extern T0* T1443f33(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.indexing_term_value */
extern T0* T1442f33(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.indexing_term_value */
extern T0* T1440f35(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.text */
extern T0* T972f59(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.indexing_term_value */
extern T0* T1439f35(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1438f52(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1437f52(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1436f52(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1435f52(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.indexing_term_value */
extern T0* T1434f45(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.indexing_term_value */
extern T0* T1433f41(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1432f37(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.indexing_term_value */
extern T0* T863f39(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.indexing_term_value */
extern T0* T366f53(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.indexing_term_value */
extern T0* T180f101(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T1717f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_parameter_with_type */
extern T0* T1478f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_parameter_with_type */
extern T0* T1477f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.named_parameter_with_type */
extern T0* T1476f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LABELED_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T1474f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_parameter_with_type */
extern T0* T1450f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.named_parameter_with_type */
extern T0* T867f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.named_parameter_with_type */
extern T0* T442f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_parameter_with_type */
extern T0* T248f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_parameter_with_type */
extern T0* T246f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_parameter_with_type */
extern T0* T73f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_parameter_with_type */
extern T0* T71f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.position */
extern T0* T1717f8(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.position */
extern T0* T1474f8(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.label */
extern T0* T1717f5(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.label */
extern T0* T1478f32(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.label */
extern T0* T1477f28(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.label */
extern T0* T1476f22(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.label */
extern T0* T1474f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.label */
extern T0* T1450f68(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.label */
extern T0* T867f70(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.label */
extern T0* T442f24(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.label */
extern T0* T248f53(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.label */
extern T0* T246f42(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.label */
extern T0* T73f142(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.label */
extern T0* T71f49(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.last_leaf */
extern T0* T1717f9(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.last_leaf */
extern T0* T1474f9(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.last_leaf */
extern T0* T1450f75(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.renames */
extern T0* T1450f76(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.last_leaf */
extern T0* T1054f11(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.last_leaf */
extern T0* T867f77(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.error_position */
extern T0* T549f61(GE_context* ac, T0* C);
/* [detachable] XM_DEFAULT_POSITION.make */
extern T0* T507c5(GE_context* ac, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] XM_EIFFEL_SCANNER_DTD.error_position */
extern T0* T547f59(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.error_position */
extern T0* T546f59(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.error_position */
extern T0* T506f70(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.end_of_file */
extern T1 T549f63(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.end_of_file */
extern T1 T547f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.end_of_file */
extern T1 T546f61(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.end_of_file */
extern T1 T506f72(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_applicable_encoding */
extern T1 T549f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_applicable_encoding */
extern T1 T876f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_valid_encoding */
extern T1 T876f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.is_applicable_encoding */
extern T1 T547f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_applicable_encoding */
extern T1 T546f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.is_applicable_encoding */
extern T1 T506f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.start_condition */
extern T6 T549f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.start_condition */
extern T6 T547f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.start_condition */
extern T6 T546f63(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.start_condition */
extern T6 T506f46(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.same_extended_feature_name */
extern T1 T1469f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ANY_ROUTINES.same_objects */
extern T1 T427f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ALIASED_FEATURE_NAME.any_ */
extern T0* T1469f5(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.same_extended_feature_name */
extern T1 T1468f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_NAME.any_ */
extern T0* T1468f42(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.same_extended_feature_name */
extern T1 T1467f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.any_ */
extern T0* T1467f19(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.same_extended_feature_name */
extern T1 T1466f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_NAME.any_ */
extern T0* T1466f28(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.same_extended_feature_name */
extern T1 T864f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.any_ */
extern T0* T864f31(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.same_extended_feature_name */
extern T1 T180f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.any_ */
extern T0* T180f91(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.position */
extern T0* T1469f3(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.same_call_name */
extern T1 T1469f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_NAME.feature_alias_name */
extern T0* T1468f22(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.feature_alias_name */
extern T0* T1467f10(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.feature_alias_name */
extern T0* T1466f22(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.feature_alias_name */
extern T0* T864f10(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.alias_name */
extern T0* T180f50(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_none */
extern T1 T776f41(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_none */
extern T1 T406f11(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_dotnet_assembly */
extern T1 T861f22(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_dotnet_assembly */
extern T1 T776f43(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_dotnet_assembly */
extern T1 T406f5(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.dotnet_assembly */
extern T0* T861f6(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.dotnet_assembly */
extern T0* T776f47(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.dotnet_assembly */
extern T0* T406f8(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.cluster */
extern T0* T861f5(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.cluster */
extern T0* T776f46(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.cluster */
extern T0* T406f7(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_unknown */
extern T1 T861f10(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_unknown */
extern T1 T776f42(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.universe */
extern T0* T406f23(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.tokens */
extern T0* T406f6(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.current_system */
extern T0* T861f11(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.current_system */
extern T0* T776f45(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.current_system */
extern T0* T406f12(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_cluster */
extern T1 T861f8(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_cluster */
extern T1 T406f9(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_hidden */
extern T1 T861f16(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_hidden */
extern T1 T406f14(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.full_lower_name */
extern T0* T861f12(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.full_lower_name */
extern T0* T776f48(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_UNKNOWN_GROUP.full_lower_name */
extern T0* T406f13(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_NONE_GROUP.kind_name */
extern T0* T861f20(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.kind_name */
extern T0* T776f64(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.kind_name */
extern T0* T406f21(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.relative_lower_name */
extern T0* T861f13(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_SYSTEM.relative_lower_name */
extern T0* T36f118(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_SYSTEM.lower_name */
extern T0* T36f119(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.relative_lower_name */
extern T0* T56f148(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_SYSTEM.lower_name */
extern T0* T56f150(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.lower_name */
extern T0* T861f14(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.relative_lower_name */
extern T0* T776f58(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_ECF_CLUSTER.relative_lower_name */
extern T0* T776f58p1(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.relative_lower_name */
extern T0* T183f110(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.lower_name */
extern T0* T183f114(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.lower_name */
extern T0* T776f60(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.relative_lower_name */
extern T0* T406f15(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_UNKNOWN_GROUP.lower_name */
extern T0* T406f16(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.is_class_assertion */
extern T1 T1513f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1508f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1507f31(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1506f27(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1505f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1504f32(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1503f28(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_class_assertion */
extern T1 T1502f15(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_class_assertion */
extern T1 T1501f20(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_class_assertion */
extern T1 T1500f13(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_class_assertion */
extern T1 T1499f15(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_class_assertion */
extern T1 T1498f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_class_assertion */
extern T1 T1495f14(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_class_assertion */
extern T1 T1494f12(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_class_assertion */
extern T1 T1493f12(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_class_assertion */
extern T1 T1492f12(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_class_assertion */
extern T1 T1491f12(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_class_assertion */
extern T1 T1490f16(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1445f23(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1444f23(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_class_assertion */
extern T1 T1443f21(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_class_assertion */
extern T1 T1442f21(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_class_assertion */
extern T1 T1440f22(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_class_assertion */
extern T1 T1439f22(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1438f32(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1437f32(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1436f32(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1435f32(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_class_assertion */
extern T1 T1434f19(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_class_assertion */
extern T1 T1433f19(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1432f18(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_class_assertion */
extern T1 T1233f18(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_class_assertion */
extern T1 T1230f16(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_class_assertion */
extern T1 T1201f14(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_class_assertion */
extern T1 T1187f19(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_class_assertion */
extern T1 T1140f19(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_class_assertion */
extern T1 T1016f18(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_class_assertion */
extern T1 T1013f15(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_class_assertion */
extern T1 T986f26(GE_context* ac, T0* C);
/* detachable ET_VOID.is_class_assertion */
extern T1 T975f21(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_class_assertion */
extern T1 T863f26(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_class_assertion */
extern T1 T715f17(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_class_assertion */
extern T1 T366f29(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_class_assertion */
extern T1 T320f17(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_class_assertion */
extern T1 T310f13(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_class_assertion */
extern T1 T309f13(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_class_assertion */
extern T1 T308f14(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_class_assertion */
extern T1 T301f22(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_class_assertion */
extern T1 T281f20(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_class_assertion */
extern T1 T279f13(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_class_assertion */
extern T1 T278f17(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_class_assertion */
extern T1 T273f22(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_class_assertion */
extern T1 T180f81(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.expression */
extern T0* T1513f4(GE_context* ac, T0* C);
extern T1 T1513f4ot1(T0* a1, T0** a2);
/* detachable ET_CLASS_ASSERTION.expression */
extern T0* T1473f4(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.clients */
extern T0* T1168f19(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.preconditions */
extern T0* T1168f23(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.postconditions */
extern T0* T1168f24(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.implementation_feature */
extern T0* T1168f22(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.first_precursor */
extern T0* T1168f13(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.is_feature */
extern T1 T1168f14(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.as_feature */
extern T0* T1465f78(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.as_feature */
extern T0* T1464f74(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.as_feature */
extern T0* T1462f71(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.as_feature */
extern T0* T1461f75(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.as_feature */
extern T0* T1460f73(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.as_feature */
extern T0* T1168f15(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.as_feature */
extern T0* T468f69(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.as_feature */
extern T0* T466f71(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.as_feature */
extern T0* T367f68(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.as_feature */
extern T0* T365f72(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.as_feature */
extern T0* T363f71(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.as_feature */
extern T0* T329f72(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.as_feature */
extern T0* T247f80(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_invariants */
extern T1 T1465f80(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_invariants */
extern T1 T1464f75(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_invariants */
extern T1 T1462f73(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_invariants */
extern T1 T1461f76(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_invariants */
extern T1 T1460f74(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_invariants */
extern T1 T468f71(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_invariants */
extern T1 T466f72(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_invariants */
extern T1 T367f70(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_invariants */
extern T1 T365f74(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_invariants */
extern T1 T363f73(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_invariants */
extern T1 T329f73(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_invariants */
extern T1 T247f82(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.as_invariants */
extern T0* T1465f81(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.as_invariants */
extern T0* T1464f76(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.as_invariants */
extern T0* T1462f74(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.as_invariants */
extern T0* T1461f77(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.as_invariants */
extern T0* T1460f75(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.as_invariants */
extern T0* T1168f21(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.as_invariants */
extern T0* T468f72(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.as_invariants */
extern T0* T466f73(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.as_invariants */
extern T0* T367f71(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.as_invariants */
extern T0* T365f75(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.as_invariants */
extern T0* T363f74(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.as_invariants */
extern T0* T329f74(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.as_invariants */
extern T0* T247f83(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_string_8 */
extern T1 T1434f11(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.has_invalid_code */
extern T1 T1434f18(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_string_8 */
extern T1 T1433f11(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_string_8 */
extern T1 T863f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.has_invalid_code */
extern T1 T863f22(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_string_32 */
extern T1 T1434f12(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_string_32 */
extern T1 T1433f12(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_string_32 */
extern T1 T863f21(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.value_position */
extern T0* T1434f35(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.value_position */
extern T0* T1433f31(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.value_position */
extern T0* T863f30(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.arguments_count */
extern T6 T1994f21(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.arguments_count */
extern T6 T1499f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.arguments_count */
extern T6 T1495f20(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.arguments_count */
extern T6 T1013f19(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.arguments_count */
extern T6 T862f9(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.arguments_count */
extern T6 T278f21(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.parenthesis_call */
extern T0* T1994f20(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.parenthesis_call */
extern T0* T1499f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.parenthesis_call */
extern T0* T1495f18(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.parenthesis_call */
extern T0* T1013f17(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.parenthesis_call */
extern T0* T862f8(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.locals */
extern T0* T1508f28(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.locals */
extern T0* T1505f29(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.locals */
extern T0* T1465f79(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.locals */
extern T0* T1462f72(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.locals */
extern T0* T1168f17(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.locals */
extern T0* T468f70(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.locals */
extern T0* T367f69(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.locals */
extern T0* T365f73(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.locals */
extern T0* T363f72(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.locals */
extern T0* T247f81(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.arguments */
extern T0* T1168f16(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.type */
extern T0* T1508f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.type */
extern T0* T1507f32(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.type */
extern T0* T1506f28(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.type */
extern T0* T1505f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.type */
extern T0* T1504f33(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.type */
extern T0* T1503f29(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.type */
extern T0* T1168f18(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.has_type */
extern T1 T260f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.index_of */
extern T6 T444f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.index_of */
extern T6 T443f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.index_of */
extern T6 T315f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.index_of */
extern T6 T311f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.index_of */
extern T6 T294f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.index_of */
extern T6 T260f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.index_of */
extern T6 T219f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.dynamic_type */
extern T0* T260f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.is_instance_free */
extern T1 T1510f10(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_instance_free */
extern T1 T983f41(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.position */
extern T0* T1760f3(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.position */
extern T0* T1509f5(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.is_instance_free */
extern T1 T1760f5(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.is_instance_free */
extern T1 T1509f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_current */
extern T1 T1994f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_current */
extern T1 T1993f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_current */
extern T1 T1992f5(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.is_current */
extern T1 T1974f5(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_current */
extern T1 T1926f5(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.is_current */
extern T1 T1510f9(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_current */
extern T1 T983f39(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_current */
extern T1 T862f5(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_open_operand */
extern T1 T1994f19(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_open_operand */
extern T1 T1993f17(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_open_operand */
extern T1 T1992f15(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_open_operand */
extern T1 T1926f15(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.is_open_operand */
extern T1 T1760f4(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1508f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1507f37(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1506f33(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1505f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1504f38(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1503f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_open_operand */
extern T1 T1502f20(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_open_operand */
extern T1 T1501f24(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_open_operand */
extern T1 T1500f17(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_open_operand */
extern T1 T1499f22(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_open_operand */
extern T1 T1498f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_open_operand */
extern T1 T1495f22(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_open_operand */
extern T1 T1494f17(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_open_operand */
extern T1 T1493f17(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_open_operand */
extern T1 T1492f17(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_open_operand */
extern T1 T1491f17(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_open_operand */
extern T1 T1490f21(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1445f31(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1444f30(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_open_operand */
extern T1 T1443f26(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_open_operand */
extern T1 T1442f26(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_open_operand */
extern T1 T1440f28(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_open_operand */
extern T1 T1439f28(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_open_operand */
extern T1 T1438f45(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_open_operand */
extern T1 T1437f45(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_open_operand */
extern T1 T1436f45(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_open_operand */
extern T1 T1435f45(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_open_operand */
extern T1 T1434f36(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_open_operand */
extern T1 T1433f32(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1432f31(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_open_operand */
extern T1 T1233f24(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_open_operand */
extern T1 T1230f21(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_open_operand */
extern T1 T1201f12(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_open_operand */
extern T1 T1187f24(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_open_operand */
extern T1 T1140f23(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_open_operand */
extern T1 T1016f24(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_open_operand */
extern T1 T1013f21(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_open_operand */
extern T1 T986f31(GE_context* ac, T0* C);
/* detachable ET_VOID.is_open_operand */
extern T1 T975f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_open_operand */
extern T1 T863f31(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_open_operand */
extern T1 T715f21(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_open_operand */
extern T1 T366f38(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_open_operand */
extern T1 T320f19(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_open_operand */
extern T1 T310f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_open_operand */
extern T1 T309f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_open_operand */
extern T1 T308f18(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_open_operand */
extern T1 T301f19(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_open_operand */
extern T1 T281f22(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_open_operand */
extern T1 T279f10(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_open_operand */
extern T1 T278f19(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_open_operand */
extern T1 T273f19(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_open_operand */
extern T1 T180f75(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.expression_item */
extern T0* T1994f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.expression_item */
extern T0* T1993f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.expression_item */
extern T0* T1992f9(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.expression_item */
extern T0* T1926f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1508f17(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1507f18(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1506f18(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1505f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1504f41(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1503f18(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.expression_item */
extern T0* T1502f11(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.expression_item */
extern T0* T1501f13(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.expression_item */
extern T0* T1500f4(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.expression_item */
extern T0* T1499f16(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.expression_item */
extern T0* T1498f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.expression_item */
extern T0* T1495f8(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.expression_item */
extern T0* T1494f4(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.expression_item */
extern T0* T1493f4(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.expression_item */
extern T0* T1492f4(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.expression_item */
extern T0* T1491f4(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.expression_item */
extern T0* T1490f12(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.expression_item */
extern T0* T1445f12(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.expression_item */
extern T0* T1444f12(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.expression_item */
extern T0* T1443f16(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.expression_item */
extern T0* T1442f16(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.expression_item */
extern T0* T1440f13(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.expression_item */
extern T0* T1439f13(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.expression_item */
extern T0* T1438f36(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.expression_item */
extern T0* T1437f36(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.expression_item */
extern T0* T1436f36(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.expression_item */
extern T0* T1435f36(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.expression_item */
extern T0* T1434f21(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.expression_item */
extern T0* T1433f18(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.expression_item */
extern T0* T1432f8(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.expression_item */
extern T0* T1233f10(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.expression_item */
extern T0* T1230f8(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.expression_item */
extern T0* T1201f9(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.expression_item */
extern T0* T1187f10(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.expression_item */
extern T0* T1140f9(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.expression_item */
extern T0* T1016f9(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.expression_item */
extern T0* T1013f8(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.expression_item */
extern T0* T986f18(GE_context* ac, T0* C);
/* detachable ET_VOID.expression_item */
extern T0* T975f16(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.expression_item */
extern T0* T863f6(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.expression_item */
extern T0* T715f7(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.expression_item */
extern T0* T366f10(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.expression_item */
extern T0* T320f10(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.expression_item */
extern T0* T310f10(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.expression_item */
extern T0* T309f10(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.expression_item */
extern T0* T308f9(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.expression_item */
extern T0* T301f8(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.expression_item */
extern T0* T281f15(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.expression_item */
extern T0* T279f6(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.expression_item */
extern T0* T278f12(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.expression_item */
extern T0* T273f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.expression_item */
extern T0* T180f45(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_empty_actual_argument */
extern T1 T1994f13(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_empty_actual_argument */
extern T1 T1993f13(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_empty_actual_argument */
extern T1 T1992f12(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.is_empty */
extern T1 T1973f3(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_empty_actual_argument */
extern T1 T1926f12(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.is_empty */
extern T1 T1924f9(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1508f29(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1507f35(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1506f31(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1505f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1504f36(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1503f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T1502f18(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_empty_actual_argument */
extern T1 T1501f22(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_empty_actual_argument */
extern T1 T1500f15(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_empty_actual_argument */
extern T1 T1499f19(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_empty_actual_argument */
extern T1 T1498f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_empty_actual_argument */
extern T1 T1495f19(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_empty_actual_argument */
extern T1 T1494f15(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_empty_actual_argument */
extern T1 T1493f15(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_empty_actual_argument */
extern T1 T1492f15(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_empty_actual_argument */
extern T1 T1491f15(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_empty_actual_argument */
extern T1 T1490f19(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_empty */
extern T1 T1479f14(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1445f27(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1444f27(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_empty_actual_argument */
extern T1 T1443f24(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_empty_actual_argument */
extern T1 T1442f24(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_empty_actual_argument */
extern T1 T1440f26(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_empty_actual_argument */
extern T1 T1439f26(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1438f42(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1437f42(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1436f42(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1435f42(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_empty_actual_argument */
extern T1 T1434f33(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T1433f29(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1432f28(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_empty_actual_argument */
extern T1 T1233f22(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T1230f19(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_empty_actual_argument */
extern T1 T1201f17(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_empty_actual_argument */
extern T1 T1187f22(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_empty_actual_argument */
extern T1 T1140f21(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_empty_actual_argument */
extern T1 T1016f21(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_empty_actual_argument */
extern T1 T1013f18(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.is_empty */
extern T1 T1003f9(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_empty_actual_argument */
extern T1 T986f28(GE_context* ac, T0* C);
/* detachable ET_VOID.is_empty_actual_argument */
extern T1 T975f23(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T863f28(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_empty_actual_argument */
extern T1 T715f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T366f44(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_empty_actual_argument */
extern T1 T320f21(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_empty_actual_argument */
extern T1 T310f16(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_empty_actual_argument */
extern T1 T309f17(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_empty_actual_argument */
extern T1 T308f19(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_empty_actual_argument */
extern T1 T301f24(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_empty_actual_argument */
extern T1 T281f23(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T279f18(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T278f20(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.is_empty */
extern T1 T274f10(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_empty_actual_argument */
extern T1 T273f24(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_empty_actual_argument */
extern T1 T180f84(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.count */
extern T6 T1479f12(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.actual_argument */
extern T0* T1994f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.actual_argument */
extern T0* T1993f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.actual_argument */
extern T0* T1992f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.actual_argument */
extern T0* T1973f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.actual_argument */
extern T0* T1926f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.actual_argument */
extern T0* T1924f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1508f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1507f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1506f20(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1505f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1504f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1503f22(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.actual_argument */
extern T0* T1502f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_OBJECT_TEST.actual_argument */
extern T0* T1501f17(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_EXPRESSION.actual_argument */
extern T0* T1500f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_EXPRESSION.actual_argument */
extern T0* T1499f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_TYPE.actual_argument */
extern T0* T1498f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_EXPRESSION.actual_argument */
extern T0* T1495f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXPRESSION_ADDRESS.actual_argument */
extern T0* T1494f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RESULT_ADDRESS.actual_argument */
extern T0* T1493f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CURRENT_ADDRESS.actual_argument */
extern T0* T1492f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_ADDRESS.actual_argument */
extern T0* T1491f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECURSOR_EXPRESSION.actual_argument */
extern T0* T1490f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ASSIGNER_INSTRUCTION.actual_argument */
extern T0* T1479f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.actual_argument */
extern T0* T1445f17(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.actual_argument */
extern T0* T1444f17(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FALSE_CONSTANT.actual_argument */
extern T0* T1443f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TRUE_CONSTANT.actual_argument */
extern T0* T1442f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.actual_argument */
extern T0* T1440f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_REAL_CONSTANT.actual_argument */
extern T0* T1439f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.actual_argument */
extern T0* T1438f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.actual_argument */
extern T0* T1437f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.actual_argument */
extern T0* T1436f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.actual_argument */
extern T0* T1435f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VERBATIM_STRING.actual_argument */
extern T0* T1434f26(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.actual_argument */
extern T0* T1433f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C3_CHARACTER_CONSTANT.actual_argument */
extern T0* T1432f22(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.actual_argument */
extern T0* T1233f16(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STATIC_CALL_EXPRESSION.actual_argument */
extern T0* T1230f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.actual_argument */
extern T0* T1201f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.actual_argument */
extern T0* T1187f17(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_EXPRESSION.actual_argument */
extern T0* T1140f16(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CALL_AGENT.actual_argument */
extern T0* T1016f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BRACKET_EXPRESSION.actual_argument */
extern T0* T1013f12(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.actual_argument */
extern T0* T1003f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_EXPRESSION.actual_argument */
extern T0* T986f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VOID.actual_argument */
extern T0* T975f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.actual_argument */
extern T0* T863f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NAMED_OBJECT_TEST.actual_argument */
extern T0* T715f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.actual_argument */
extern T0* T366f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CREATE_EXPRESSION.actual_argument */
extern T0* T320f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EQUALITY_EXPRESSION.actual_argument */
extern T0* T310f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.actual_argument */
extern T0* T309f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_TEST.actual_argument */
extern T0* T308f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_RESULT.actual_argument */
extern T0* T301f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.actual_argument */
extern T0* T281f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.actual_argument */
extern T0* T279f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.actual_argument */
extern T0* T278f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.actual_argument */
extern T0* T274f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CURRENT.actual_argument */
extern T0* T273f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.actual_argument */
extern T0* T180f73(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.last_leaf */
extern T0* T1478f77(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.last_leaf */
extern T0* T1477f80(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.last_leaf */
extern T0* T1476f80(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.last_leaf */
extern T0* T442f73(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.last_leaf */
extern T0* T248f110(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.last_leaf */
extern T0* T246f73(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.last_leaf */
extern T0* T73f251(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.last_leaf */
extern T0* T71f121(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1445f18(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1444f18(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_boolean_constant */
extern T1 T1440f20(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_boolean_constant */
extern T1 T1439f20(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1438f37(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1437f37(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1436f37(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1435f37(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_boolean_constant */
extern T1 T1434f27(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_boolean_constant */
extern T1 T1433f9(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1432f23(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_boolean_constant */
extern T1 T863f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T366f15(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_character_constant */
extern T1 T1443f35(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_character_constant */
extern T1 T1442f35(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_character_constant */
extern T1 T1440f37(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_character_constant */
extern T1 T1439f37(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_character_constant */
extern T1 T1438f55(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_character_constant */
extern T1 T1437f55(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_character_constant */
extern T1 T1436f55(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_character_constant */
extern T1 T1435f55(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_character_constant */
extern T1 T1434f47(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_character_constant */
extern T1 T1433f43(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_character_constant */
extern T1 T863f42(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_character_constant */
extern T1 T366f55(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1445f40(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1444f39(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_integer_constant */
extern T1 T1443f36(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_integer_constant */
extern T1 T1442f36(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_integer_constant */
extern T1 T1440f38(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_integer_constant */
extern T1 T1439f38(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_integer_constant */
extern T1 T1434f48(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_integer_constant */
extern T1 T1433f44(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1432f40(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_integer_constant */
extern T1 T863f43(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.name */
extern T0* T308f16(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.type */
extern T0* T1501f15(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.type */
extern T0* T715f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.type */
extern T0* T308f17(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1508f21(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1507f24(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1506f22(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_qualified_call */
extern T1 T1016f15(GE_context* ac, T0* C);
extern T1 T1016f15ot1(T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1508f27(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1508f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1507f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1507f38(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1506f30(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1506f34(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1505f28(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1505f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1504f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1504f39(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1503f31(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1503f35(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.implicit_target_position */
extern T0* T1016f20(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix */
extern T1 T1470f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix */
extern T1 T886f13(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix */
extern T1 T1470f47(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix */
extern T1 T886f12(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.feature_name */
extern T0* T1470f46(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.feature_name */
extern T0* T886f65(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.same_call_name */
extern T1 T1470f8(GE_context* ac, T0* C, T0* a1);
extern T1 T1470f8ot1(T0* a1, T0** a2);
/* detachable ET_INFIX_NAME.same_call_name */
extern T1 T1468f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.same_call_name */
extern T1 T1467f20(GE_context* ac, T0* C, T0* a1);
extern T1 T1467f20ot1(T0* a1, T0** a2);
/* detachable ET_PREFIX_NAME.same_call_name */
extern T1 T1466f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.same_call_name */
extern T1 T886f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.same_call_name */
extern T1 T864f24(GE_context* ac, T0* C, T0* a1);
extern T1 T864f24ot1(T0* a1, T0** a2);
/* [detachable] ET_IDENTIFIER.same_call_name */
extern T1 T180f74(GE_context* ac, T0* C, T0* a1);
extern T1 T180f74ot1(T0* a1, T0** a2);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_bracket */
extern T1 T1967f23(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_bracket */
extern T1 T1497f28(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_bracket */
extern T1 T1496f28(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_bracket */
extern T1 T1470f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_bracket */
extern T1 T1468f18(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_bracket */
extern T1 T1467f23(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_bracket */
extern T1 T1466f18(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_bracket */
extern T1 T972f37(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_bracket */
extern T1 T959f37(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_bracket */
extern T1 T951f40(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_bracket */
extern T1 T886f8(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_bracket */
extern T1 T864f25(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_bracket */
extern T1 T180f76(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_parenthesis */
extern T1 T1497f29(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_parenthesis */
extern T1 T1496f29(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_parenthesis */
extern T1 T1470f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_parenthesis */
extern T1 T1468f19(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_parenthesis */
extern T1 T1467f24(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_parenthesis */
extern T1 T1466f19(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_parenthesis */
extern T1 T972f38(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_parenthesis */
extern T1 T959f38(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_parenthesis */
extern T1 T951f41(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_parenthesis */
extern T1 T886f9(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_parenthesis */
extern T1 T864f26(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_parenthesis */
extern T1 T610f63(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_parenthesis */
extern T1 T180f77(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and */
extern T1 T1967f6(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and */
extern T1 T1497f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_and */
extern T1 T1496f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and */
extern T1 T1470f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_and */
extern T1 T1468f15(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_and */
extern T1 T1467f15(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_and */
extern T1 T1466f15(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and */
extern T1 T972f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and */
extern T1 T959f16(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and */
extern T1 T951f15(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and */
extern T1 T886f19(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_and */
extern T1 T864f22(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and */
extern T1 T610f36(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and */
extern T1 T180f41(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_div */
extern T1 T1967f8(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_div */
extern T1 T1497f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_div */
extern T1 T1496f16(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_div */
extern T1 T1470f27(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_div */
extern T1 T1468f50(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_div */
extern T1 T1467f39(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_div */
extern T1 T1466f36(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_div */
extern T1 T972f25(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_div */
extern T1 T959f25(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_div */
extern T1 T951f28(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_div */
extern T1 T886f21(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_div */
extern T1 T864f39(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_div */
extern T1 T610f50(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_div */
extern T1 T180f54(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_divide */
extern T1 T1967f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_divide */
extern T1 T1497f17(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_divide */
extern T1 T1496f17(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_divide */
extern T1 T1470f28(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_divide */
extern T1 T1468f51(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_divide */
extern T1 T1467f40(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_divide */
extern T1 T1466f37(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_divide */
extern T1 T972f26(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_divide */
extern T1 T959f26(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_divide */
extern T1 T951f29(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_divide */
extern T1 T886f22(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_divide */
extern T1 T864f40(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_divide */
extern T1 T610f51(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_divide */
extern T1 T180f55(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_ge */
extern T1 T1967f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_ge */
extern T1 T1497f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_ge */
extern T1 T1496f18(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_ge */
extern T1 T1470f29(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_ge */
extern T1 T1468f52(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_ge */
extern T1 T1467f41(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_ge */
extern T1 T1466f38(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_ge */
extern T1 T972f27(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_ge */
extern T1 T959f27(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_ge */
extern T1 T951f30(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_ge */
extern T1 T886f23(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_ge */
extern T1 T864f41(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_ge */
extern T1 T610f52(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_ge */
extern T1 T180f56(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_gt */
extern T1 T1967f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_gt */
extern T1 T1497f19(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_gt */
extern T1 T1496f19(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_gt */
extern T1 T1470f30(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_gt */
extern T1 T1468f53(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_gt */
extern T1 T1467f42(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_gt */
extern T1 T1466f39(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_gt */
extern T1 T972f28(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_gt */
extern T1 T959f28(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_gt */
extern T1 T951f31(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_gt */
extern T1 T886f24(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_gt */
extern T1 T864f42(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_gt */
extern T1 T610f53(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_gt */
extern T1 T180f57(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_le */
extern T1 T1967f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_le */
extern T1 T1497f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_le */
extern T1 T1496f20(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_le */
extern T1 T1470f32(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_le */
extern T1 T1468f54(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_le */
extern T1 T1467f43(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_le */
extern T1 T1466f40(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_le */
extern T1 T972f29(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_le */
extern T1 T959f29(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_le */
extern T1 T951f32(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_le */
extern T1 T886f26(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_le */
extern T1 T864f43(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_le */
extern T1 T610f54(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_le */
extern T1 T180f58(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_lt */
extern T1 T1967f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_lt */
extern T1 T1497f21(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_lt */
extern T1 T1496f21(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_lt */
extern T1 T1470f33(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_lt */
extern T1 T1468f55(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_lt */
extern T1 T1467f44(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_lt */
extern T1 T1466f41(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_lt */
extern T1 T972f30(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_lt */
extern T1 T959f30(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_lt */
extern T1 T951f33(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_lt */
extern T1 T886f27(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_lt */
extern T1 T864f44(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_lt */
extern T1 T610f55(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_lt */
extern T1 T180f59(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_minus */
extern T1 T1967f15(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_minus */
extern T1 T1497f22(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_minus */
extern T1 T1496f22(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_minus */
extern T1 T1470f34(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_minus */
extern T1 T1468f56(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_minus */
extern T1 T1467f45(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_minus */
extern T1 T1466f42(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_minus */
extern T1 T972f31(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_minus */
extern T1 T959f31(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_minus */
extern T1 T951f34(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_minus */
extern T1 T886f28(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_minus */
extern T1 T864f45(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_minus */
extern T1 T610f56(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_minus */
extern T1 T180f60(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_mod */
extern T1 T1967f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_mod */
extern T1 T1497f23(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_mod */
extern T1 T1496f23(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_mod */
extern T1 T1470f35(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_mod */
extern T1 T1468f57(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_mod */
extern T1 T1467f46(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_mod */
extern T1 T1466f43(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_mod */
extern T1 T972f32(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_mod */
extern T1 T959f32(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_mod */
extern T1 T951f35(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_mod */
extern T1 T886f29(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_mod */
extern T1 T864f46(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_mod */
extern T1 T610f57(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_mod */
extern T1 T180f61(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or */
extern T1 T1967f17(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_or */
extern T1 T1497f8(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or */
extern T1 T1496f8(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or */
extern T1 T1470f36(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_or */
extern T1 T1468f16(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_or */
extern T1 T1467f16(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_or */
extern T1 T1466f16(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or */
extern T1 T972f13(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or */
extern T1 T959f17(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or */
extern T1 T951f16(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or */
extern T1 T886f30(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_or */
extern T1 T864f23(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or */
extern T1 T610f37(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or */
extern T1 T180f42(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_plus */
extern T1 T1967f19(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_plus */
extern T1 T1497f24(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_plus */
extern T1 T1496f24(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_plus */
extern T1 T1470f38(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_plus */
extern T1 T1468f58(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_plus */
extern T1 T1467f47(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_plus */
extern T1 T1466f44(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_plus */
extern T1 T972f33(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_plus */
extern T1 T959f33(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_plus */
extern T1 T951f36(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_plus */
extern T1 T886f32(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_plus */
extern T1 T864f47(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_plus */
extern T1 T610f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_plus */
extern T1 T180f62(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_power */
extern T1 T1967f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_power */
extern T1 T1497f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_power */
extern T1 T1496f25(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_power */
extern T1 T1470f39(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_power */
extern T1 T1468f59(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_power */
extern T1 T1467f48(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_power */
extern T1 T1466f45(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_power */
extern T1 T972f34(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_power */
extern T1 T959f34(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_power */
extern T1 T951f37(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_power */
extern T1 T886f33(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_power */
extern T1 T864f48(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_power */
extern T1 T610f59(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_power */
extern T1 T180f63(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_times */
extern T1 T1967f21(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_times */
extern T1 T1497f26(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_times */
extern T1 T1496f26(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_times */
extern T1 T1470f40(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_times */
extern T1 T1468f60(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_times */
extern T1 T1467f49(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_times */
extern T1 T1466f46(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_times */
extern T1 T972f35(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_times */
extern T1 T959f35(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_times */
extern T1 T951f38(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_times */
extern T1 T886f34(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_times */
extern T1 T864f49(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_times */
extern T1 T610f60(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_times */
extern T1 T180f64(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_xor */
extern T1 T1967f22(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_xor */
extern T1 T1497f27(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_xor */
extern T1 T1496f27(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_xor */
extern T1 T1470f41(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_xor */
extern T1 T1468f61(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_xor */
extern T1 T1467f50(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_xor */
extern T1 T1466f47(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_xor */
extern T1 T972f36(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_xor */
extern T1 T959f36(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_xor */
extern T1 T951f39(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_xor */
extern T1 T886f35(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_xor */
extern T1 T864f50(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_xor */
extern T1 T610f61(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_xor */
extern T1 T180f65(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_dotdot */
extern T1 T1967f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_dotdot */
extern T1 T1497f30(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_dotdot */
extern T1 T1496f30(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_dotdot */
extern T1 T1470f44(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_dotdot */
extern T1 T1468f71(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_dotdot */
extern T1 T1467f60(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_dotdot */
extern T1 T1466f57(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_dotdot */
extern T1 T972f39(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_dotdot */
extern T1 T959f39(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_dotdot */
extern T1 T951f42(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_dotdot */
extern T1 T886f61(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_dotdot */
extern T1 T864f62(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_dotdot */
extern T1 T610f64(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_dotdot */
extern T1 T180f78(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_minus */
extern T1 T1967f37(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_minus */
extern T1 T1497f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_minus */
extern T1 T1496f13(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_minus */
extern T1 T1470f22(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_prefix_minus */
extern T1 T1468f47(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_prefix_minus */
extern T1 T1467f36(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_prefix_minus */
extern T1 T1466f33(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_minus */
extern T1 T972f23(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_minus */
extern T1 T959f23(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_minus */
extern T1 T951f26(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_minus */
extern T1 T886f16(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_prefix_minus */
extern T1 T864f36(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_minus */
extern T1 T610f47(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_minus */
extern T1 T180f51(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_plus */
extern T1 T1967f4(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_plus */
extern T1 T1497f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_plus */
extern T1 T1496f14(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_plus */
extern T1 T1470f23(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_prefix_plus */
extern T1 T1468f48(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_prefix_plus */
extern T1 T1467f37(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_prefix_plus */
extern T1 T1466f34(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_plus */
extern T1 T972f24(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_plus */
extern T1 T959f24(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_plus */
extern T1 T951f27(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_plus */
extern T1 T886f17(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_prefix_plus */
extern T1 T864f37(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_plus */
extern T1 T610f48(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_plus */
extern T1 T180f52(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_not */
extern T1 T1967f5(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_not */
extern T1 T1497f15(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_not */
extern T1 T1496f15(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_not */
extern T1 T1470f24(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_prefix_not */
extern T1 T1468f49(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_prefix_not */
extern T1 T1467f38(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_prefix_not */
extern T1 T1466f35(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_not */
extern T1 T972f22(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_not */
extern T1 T959f22(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_not */
extern T1 T951f24(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_not */
extern T1 T886f18(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_prefix_not */
extern T1 T864f38(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_not */
extern T1 T610f49(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_not */
extern T1 T180f53(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.hash_code */
extern T6 T886f14(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_precursor */
extern T1 T1967f28(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_precursor */
extern T1 T1497f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_precursor */
extern T1 T1496f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_precursor */
extern T1 T1468f80(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_precursor */
extern T1 T1467f64(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_precursor */
extern T1 T1466f80(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_precursor */
extern T1 T972f20(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_precursor */
extern T1 T959f20(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_precursor */
extern T1 T951f22(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_precursor */
extern T1 T933f14(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_precursor */
extern T1 T864f64(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_precursor */
extern T1 T610f44(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_precursor */
extern T1 T180f85(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.precursor_keyword */
extern T0* T1967f29(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.precursor_keyword */
extern T0* T1497f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.precursor_keyword */
extern T0* T1496f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.precursor_keyword */
extern T0* T1468f81(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.precursor_keyword */
extern T0* T1467f65(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.precursor_keyword */
extern T0* T1466f81(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.precursor_keyword */
extern T0* T972f21(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.precursor_keyword */
extern T0* T959f21(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.precursor_keyword */
extern T0* T951f23(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.precursor_keyword */
extern T0* T933f15(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.precursor_keyword */
extern T0* T864f65(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.precursor_keyword */
extern T0* T610f45(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.precursor_keyword */
extern T0* T180f86(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and_then */
extern T1 T1967f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and_then */
extern T1 T1497f4(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and_then */
extern T1 T1470f26(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_and_then */
extern T1 T1468f12(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_and_then */
extern T1 T1467f12(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_and_then */
extern T1 T1466f12(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and_then */
extern T1 T972f9(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and_then */
extern T1 T959f13(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and_then */
extern T1 T951f12(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_and_then */
extern T1 T933f16(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and_then */
extern T1 T886f20(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_and_then */
extern T1 T864f19(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and_then */
extern T1 T610f33(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and_then */
extern T1 T180f38(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_implies */
extern T1 T1967f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_implies */
extern T1 T1497f6(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_implies */
extern T1 T1496f6(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_implies */
extern T1 T1470f31(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_implies */
extern T1 T1468f14(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_implies */
extern T1 T1467f14(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_implies */
extern T1 T1466f14(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_implies */
extern T1 T972f11(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_implies */
extern T1 T959f15(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_implies */
extern T1 T951f14(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_implies */
extern T1 T933f17(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_implies */
extern T1 T886f25(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_implies */
extern T1 T864f21(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_implies */
extern T1 T610f35(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_implies */
extern T1 T180f40(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or_else */
extern T1 T1967f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or_else */
extern T1 T1496f5(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or_else */
extern T1 T1470f37(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_infix_or_else */
extern T1 T1468f13(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix_or_else */
extern T1 T1467f13(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix_or_else */
extern T1 T1466f13(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or_else */
extern T1 T972f10(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or_else */
extern T1 T959f14(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or_else */
extern T1 T951f13(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_or_else */
extern T1 T933f18(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or_else */
extern T1 T886f31(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_infix_or_else */
extern T1 T864f20(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or_else */
extern T1 T610f34(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or_else */
extern T1 T180f39(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.lower_name */
extern T0* T1967f27(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.name */
extern T0* T1967f33(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.lower_name */
extern T0* T1497f32(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.name */
extern T0* T1497f33(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.lower_name */
extern T0* T1496f32(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.name */
extern T0* T1496f33(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.lower_name */
extern T0* T1470f10(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.name */
extern T0* T1470f16(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.lower_name */
extern T0* T972f40(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.name */
extern T0* T972f41(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.lower_name */
extern T0* T959f40(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.name */
extern T0* T959f41(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.lower_name */
extern T0* T951f43(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.name */
extern T0* T951f44(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.lower_name */
extern T0* T933f19(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.lower_name */
extern T0* T886f15(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.name */
extern T0* T886f62(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.lower_name */
extern T0* T610f65(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.name */
extern T0* T610f67(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_tuple_label */
extern T1 T1967f34(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_tuple_label */
extern T1 T1497f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_tuple_label */
extern T1 T1496f10(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_tuple_label */
extern T1 T972f15(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_tuple_label */
extern T1 T959f19(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_tuple_label */
extern T1 T951f18(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_tuple_label */
extern T1 T610f39(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.position */
extern T0* T1967f26(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.position */
extern T0* T1497f31(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.position */
extern T0* T1496f31(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.position */
extern T0* T1470f42(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.position */
extern T0* T972f8(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.position */
extern T0* T959f5(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.position */
extern T0* T951f11(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.position */
extern T0* T933f9(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.position */
extern T0* T886f36(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_call_agent */
extern T1 T1994f17(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.is_call_agent */
extern T1 T1947f5(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_call_agent */
extern T1 T1499f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_call_agent */
extern T1 T1495f12(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_call_agent */
extern T1 T1479f10(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_call_agent */
extern T1 T1013f10(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_call_agent */
extern T1 T862f7(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_call_agent */
extern T1 T278f9(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.is_call_agent */
extern T1 T275f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_tuple_label */
extern T1 T1994f16(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.is_tuple_label */
extern T1 T1947f4(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_tuple_label */
extern T1 T1499f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_tuple_label */
extern T1 T1495f11(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_tuple_label */
extern T1 T1479f9(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_tuple_label */
extern T1 T1016f25(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_tuple_label */
extern T1 T1013f9(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_tuple_label */
extern T1 T862f6(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_tuple_label */
extern T1 T278f8(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.is_tuple_label */
extern T1 T275f5(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.position */
extern T0* T1947f7(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.position */
extern T0* T1479f11(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.position */
extern T0* T862f10(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.position */
extern T0* T275f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.arguments */
extern T0* T1994f15(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.arguments */
extern T0* T1499f10(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.arguments */
extern T0* T1479f8(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.name */
extern T0* T1016f7(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.name */
extern T0* T278f7(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.name */
extern T0* T275f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.feature_name */
extern T0* T1468f10(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.feature_name */
extern T0* T1467f27(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.feature_name */
extern T0* T1466f10(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.feature_name */
extern T0* T864f9(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.feature_name */
extern T0* T180f31(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_once */
extern T1 T1465f65(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once */
extern T1 T468f56(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_once */
extern T1 T466f31(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_query */
extern T1 T1465f66(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_query */
extern T1 T1464f61(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_query */
extern T1 T1462f63(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_query */
extern T1 T1461f67(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_query */
extern T1 T1460f65(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_query */
extern T1 T468f57(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_query */
extern T1 T466f59(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_query */
extern T1 T367f43(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_query */
extern T1 T365f45(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_query */
extern T1 T363f44(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_query */
extern T1 T329f46(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_query */
extern T1 T247f46(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_dotnet */
extern T1 T1465f75(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_dotnet */
extern T1 T1464f71(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_dotnet */
extern T1 T1462f68(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_dotnet */
extern T1 T1461f72(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_dotnet */
extern T1 T1460f70(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_dotnet */
extern T1 T468f66(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_dotnet */
extern T1 T466f68(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_dotnet */
extern T1 T367f64(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_dotnet */
extern T1 T365f68(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_dotnet */
extern T1 T363f67(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_dotnet */
extern T1 T329f69(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_dotnet */
extern T1 T247f75(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.same_version */
extern T1 T1465f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.same_version */
extern T1 T1464f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.same_version */
extern T1 T1462f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.same_version */
extern T1 T1461f73(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.same_version */
extern T1 T1460f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.same_version */
extern T1 T468f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.same_version */
extern T1 T466f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.same_version */
extern T1 T367f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.same_version */
extern T1 T365f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.same_version */
extern T1 T363f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.same_version */
extern T1 T329f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.same_version */
extern T1 T247f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.undefined_feature */
extern T0* T1465f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_feature_clause */
extern void T468f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_semicolon */
extern void T468f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_end_keyword */
extern void T468f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.undefined_feature */
extern T0* T1464f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.undefined_feature */
extern T0* T1462f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_feature_clause */
extern void T1462f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_semicolon */
extern void T1462f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_end_keyword */
extern void T1462f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.undefined_feature */
extern T0* T1461f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.undefined_feature */
extern T0* T1460f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.undefined_feature */
extern T0* T468f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.undefined_feature */
extern T0* T466f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.undefined_feature */
extern T0* T367f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.obsolete_message */
extern T0* T367f66(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.undefined_feature */
extern T0* T365f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.obsolete_message */
extern T0* T365f70(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.undefined_feature */
extern T0* T363f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.obsolete_message */
extern T0* T363f69(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.undefined_feature */
extern T0* T329f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.undefined_feature */
extern T0* T247f73(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.renamed_feature */
extern T0* T1465f74(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_feature_clause */
extern void T1465f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_semicolon */
extern void T1465f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_end_keyword */
extern void T1465f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.renamed_feature */
extern T0* T1464f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_feature_clause */
extern void T1464f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_semicolon */
extern void T1464f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_end_keyword */
extern void T1464f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.renamed_feature */
extern T0* T1462f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_deferred_keyword */
extern void T1462f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.renamed_feature */
extern T0* T1461f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_feature_clause */
extern void T1461f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_semicolon */
extern void T1461f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_end_keyword */
extern void T1461f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.renamed_feature */
extern T0* T1460f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_feature_clause */
extern void T1460f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_semicolon */
extern void T1460f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_end_keyword */
extern void T1460f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.renamed_feature */
extern T0* T468f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_deferred_keyword */
extern void T468f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.renamed_feature */
extern T0* T466f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_feature_clause */
extern void T466f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_semicolon */
extern void T466f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_end_keyword */
extern void T466f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.renamed_feature */
extern T0* T367f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_first_indexing */
extern void T367f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_feature_clause */
extern void T367f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_semicolon */
extern void T367f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.renamed_feature */
extern T0* T365f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_constant */
extern void T365f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_indexing */
extern void T365f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_feature_clause */
extern void T365f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_semicolon */
extern void T365f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_unique_keyword */
extern void T365f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_is_keyword */
extern void T365f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.renamed_feature */
extern T0* T363f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_indexing */
extern void T363f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_feature_clause */
extern void T363f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_semicolon */
extern void T363f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_is_keyword */
extern void T363f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.renamed_feature */
extern T0* T329f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_feature_clause */
extern void T329f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_semicolon */
extern void T329f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_end_keyword */
extern void T329f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.renamed_feature */
extern T0* T247f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_feature_clause */
extern void T247f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_semicolon */
extern void T247f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_end_keyword */
extern void T247f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_bracketable */
extern T1 T1465f72(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_bracketable */
extern T1 T1464f66(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_bracketable */
extern T1 T468f61(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_bracketable */
extern T1 T466f63(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_prefixable */
extern T1 T1465f71(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_prefixable */
extern T1 T1464f65(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_prefixable */
extern T1 T468f60(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_prefixable */
extern T1 T466f62(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_infixable */
extern T1 T1465f70(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_infixable */
extern T1 T1464f64(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_infixable */
extern T1 T468f59(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_infixable */
extern T1 T466f61(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_once_per_process */
extern T1 T1465f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_once_per_process */
extern T1 T1464f38(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.has_tagged_indexing_term_value */
extern T1 T1146f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INDEXING_TERM_LIST.has_indexing_term_value */
extern T1 T1157f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.is_once_per_process */
extern T1 T1462f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once_per_process */
extern T1 T1461f42(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.standard_once_keys */
extern T0* T1461f36(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once_per_process */
extern T1 T1460f40(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once_per_process */
extern T1 T468f41(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_once_per_process */
extern T1 T466f44(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once_per_process */
extern T1 T367f44(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once_per_process */
extern T1 T365f46(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once_per_process */
extern T1 T363f45(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once_per_process */
extern T1 T329f47(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once_per_process */
extern T1 T247f52(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_attribute */
extern T1 T1465f41(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_attribute */
extern T1 T1464f48(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_attribute */
extern T1 T468f30(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_attribute */
extern T1 T466f45(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_unique_attribute */
extern T1 T1465f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_unique_attribute */
extern T1 T1464f37(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_unique_attribute */
extern T1 T468f39(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_unique_attribute */
extern T1 T466f33(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.has_class_postcondition_recursive */
extern T1 T1465f48(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.has_class_assertion */
extern T1 T1213f7(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.has_class_postcondition_recursive */
extern T1 T1464f44(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_class_postcondition_recursive */
extern T1 T1462f46(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.has_class_postcondition_recursive */
extern T1 T1461f51(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.has_class_postcondition_recursive */
extern T1 T1460f48(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.has_class_postcondition_recursive */
extern T1 T468f28(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.has_class_postcondition_recursive */
extern T1 T466f47(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T367f53(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T365f60(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T363f58(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T329f58(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.has_class_postcondition_recursive */
extern T1 T247f66(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.type */
extern T0* T1465f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.type */
extern T0* T1464f34(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.type */
extern T0* T468f37(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.type */
extern T0* T466f48(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_procedure */
extern T1 T1465f64(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_procedure */
extern T1 T1464f59(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_procedure */
extern T1 T1462f62(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_procedure */
extern T1 T1461f66(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_procedure */
extern T1 T1460f64(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_procedure */
extern T1 T468f55(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_procedure */
extern T1 T466f58(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_procedure */
extern T1 T367f42(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_procedure */
extern T1 T365f44(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_procedure */
extern T1 T363f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_procedure */
extern T1 T329f45(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_procedure */
extern T1 T247f45(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_constant_attribute */
extern T1 T1465f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_constant_attribute */
extern T1 T1464f36(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_constant_attribute */
extern T1 T468f38(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_constant_attribute */
extern T1 T466f30(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T1465f46(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.is_instance_free */
extern T1 T401f8(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T1464f69(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1462f66(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1461f70(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1460f68(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T468f64(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T466f66(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T367f62(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T365f58(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T363f56(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T329f67(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T247f64(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T1465f47(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.is_instance_free */
extern T1 T1213f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T1464f70(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1462f67(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1461f71(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1460f69(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T468f65(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T466f67(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T367f63(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T365f59(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T363f57(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T329f68(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T247f65(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_never_void */
extern T1 T1994f18(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_never_void */
extern T1 T1992f14(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_never_void */
extern T1 T1926f8(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_never_void */
extern T1 T1501f12(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_never_void */
extern T1 T1500f7(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_never_void */
extern T1 T1499f6(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_never_void */
extern T1 T1495f6(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_never_void */
extern T1 T1490f8(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_never_void */
extern T1 T1230f7(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_never_void */
extern T1 T1201f7(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_never_void */
extern T1 T1140f11(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_never_void */
extern T1 T1013f6(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_never_void */
extern T1 T986f17(GE_context* ac, T0* C);
/* detachable ET_VOID.is_never_void */
extern T1 T975f8(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_never_void */
extern T1 T715f9(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_never_void */
extern T1 T310f8(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_never_void */
extern T1 T309f8(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_never_void */
extern T1 T308f10(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_never_void */
extern T1 T301f6(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_never_void */
extern T1 T279f9(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_never_void */
extern T1 T278f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_never_void */
extern T1 T180f8(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1508f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1507f39(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1506f36(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1505f18(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1504f19(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1503f19(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_current */
extern T1 T1502f5(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_current */
extern T1 T1501f10(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_current */
extern T1 T1500f5(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_current */
extern T1 T1499f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_current */
extern T1 T1498f5(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_current */
extern T1 T1495f16(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_current */
extern T1 T1494f5(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_current */
extern T1 T1493f5(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_current */
extern T1 T1492f5(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_current */
extern T1 T1491f5(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_current */
extern T1 T1490f9(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_current */
extern T1 T1445f25(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_current */
extern T1 T1444f25(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_current */
extern T1 T1443f9(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_current */
extern T1 T1442f9(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_current */
extern T1 T1440f24(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_current */
extern T1 T1439f24(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_current */
extern T1 T1438f9(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_current */
extern T1 T1437f9(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_current */
extern T1 T1436f9(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_current */
extern T1 T1435f9(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_current */
extern T1 T1434f23(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_current */
extern T1 T1433f21(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_current */
extern T1 T1432f16(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_current */
extern T1 T1233f11(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_current */
extern T1 T1230f9(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_current */
extern T1 T1201f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_current */
extern T1 T1187f11(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_current */
extern T1 T1140f10(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_current */
extern T1 T1016f11(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_current */
extern T1 T1013f5(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_current */
extern T1 T986f19(GE_context* ac, T0* C);
/* detachable ET_VOID.is_current */
extern T1 T975f9(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_current */
extern T1 T863f7(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_current */
extern T1 T715f8(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_current */
extern T1 T366f11(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_current */
extern T1 T320f7(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_current */
extern T1 T310f7(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_current */
extern T1 T309f7(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_current */
extern T1 T308f7(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_current */
extern T1 T301f7(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_current */
extern T1 T281f13(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_current */
extern T1 T279f15(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_current */
extern T1 T278f10(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_current */
extern T1 T180f7(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.unparenthesized_expression */
extern T0* T1994f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.unparenthesized_expression */
extern T0* T1993f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.unparenthesized_expression */
extern T0* T1992f7(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.unparenthesized_expression */
extern T0* T1926f9(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1508f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1507f36(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1506f32(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1505f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1504f37(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1503f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.unparenthesized_expression */
extern T0* T1502f19(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.unparenthesized_expression */
extern T0* T1501f23(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.unparenthesized_expression */
extern T0* T1500f16(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.unparenthesized_expression */
extern T0* T1499f21(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.unparenthesized_expression */
extern T0* T1498f17(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.unparenthesized_expression */
extern T0* T1495f21(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.unparenthesized_expression */
extern T0* T1494f16(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.unparenthesized_expression */
extern T0* T1493f16(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.unparenthesized_expression */
extern T0* T1492f16(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.unparenthesized_expression */
extern T0* T1491f16(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.unparenthesized_expression */
extern T0* T1490f20(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1445f28(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1444f28(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.unparenthesized_expression */
extern T0* T1443f25(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.unparenthesized_expression */
extern T0* T1442f25(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.unparenthesized_expression */
extern T0* T1440f27(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.unparenthesized_expression */
extern T0* T1439f27(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1438f43(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1437f43(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1436f43(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1435f43(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.unparenthesized_expression */
extern T0* T1434f34(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.unparenthesized_expression */
extern T0* T1433f30(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1432f29(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.unparenthesized_expression */
extern T0* T1233f23(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T1230f20(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.unparenthesized_expression */
extern T0* T1201f18(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.unparenthesized_expression */
extern T0* T1187f23(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.unparenthesized_expression */
extern T0* T1140f22(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.unparenthesized_expression */
extern T0* T1016f23(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.unparenthesized_expression */
extern T0* T1013f20(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.unparenthesized_expression */
extern T0* T986f30(GE_context* ac, T0* C);
/* detachable ET_VOID.unparenthesized_expression */
extern T0* T975f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.unparenthesized_expression */
extern T0* T863f29(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.unparenthesized_expression */
extern T0* T715f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T366f45(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.unparenthesized_expression */
extern T0* T320f24(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.unparenthesized_expression */
extern T0* T310f17(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.unparenthesized_expression */
extern T0* T309f18(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.unparenthesized_expression */
extern T0* T308f20(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.unparenthesized_expression */
extern T0* T301f25(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.unparenthesized_expression */
extern T0* T281f24(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T279f19(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T278f22(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.unparenthesized_expression */
extern T0* T273f25(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.unparenthesized_expression */
extern T0* T180f87(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1994f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1993f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1992f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INFIX_CAST_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1926f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1508f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1507f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1506f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1505f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1504f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1503f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T1502f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OLD_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T1501f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OLD_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1500f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PREFIX_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1499f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_TYPE.manifest_constant_convert_feature */
extern T0* T1498f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1495f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXPRESSION_ADDRESS.manifest_constant_convert_feature */
extern T0* T1494f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_RESULT_ADDRESS.manifest_constant_convert_feature */
extern T0* T1493f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CURRENT_ADDRESS.manifest_constant_convert_feature */
extern T0* T1492f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FEATURE_ADDRESS.manifest_constant_convert_feature */
extern T0* T1491f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1490f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C2_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1445f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C1_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1444f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FALSE_CONSTANT.manifest_constant_convert_feature */
extern T0* T1443f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_TRUE_CONSTANT.manifest_constant_convert_feature */
extern T0* T1442f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.manifest_constant_convert_feature */
extern T0* T1440f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_REGULAR_REAL_CONSTANT.manifest_constant_convert_feature */
extern T0* T1439f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BINARY_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1438f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OCTAL_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1437f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1436f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1435f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VERBATIM_STRING.manifest_constant_convert_feature */
extern T0* T1434f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VERBATIM_STRING.tokens */
extern T0* T1434f42(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T1433f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_SPECIAL_MANIFEST_STRING.tokens */
extern T0* T1433f38(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1432f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STRIP_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1233f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1230f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIZED_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1201f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_ARRAY.manifest_constant_convert_feature */
extern T0* T1187f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_IF_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1140f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CALL_AGENT.manifest_constant_convert_feature */
extern T0* T1016f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BRACKET_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1013f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ACROSS_EXPRESSION.manifest_constant_convert_feature */
extern T0* T986f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VOID.manifest_constant_convert_feature */
extern T0* T975f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T863f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_MANIFEST_STRING.tokens */
extern T0* T863f33(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T715f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T366f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CREATE_EXPRESSION.manifest_constant_convert_feature */
extern T0* T320f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EQUALITY_EXPRESSION.manifest_constant_convert_feature */
extern T0* T310f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.manifest_constant_convert_feature */
extern T0* T309f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T308f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_RESULT.manifest_constant_convert_feature */
extern T0* T301f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MANIFEST_TUPLE.manifest_constant_convert_feature */
extern T0* T281f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T279f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T278f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CURRENT.manifest_constant_convert_feature */
extern T0* T273f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_IDENTIFIER.manifest_constant_convert_feature */
extern T0* T180f95(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1508f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1507f42(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1506f38(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1505f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1504f43(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1503f39(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_false */
extern T1 T1502f22(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_false */
extern T1 T1501f27(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_false */
extern T1 T1500f19(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_false */
extern T1 T1499f24(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_false */
extern T1 T1498f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_false */
extern T1 T1495f24(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_false */
extern T1 T1494f19(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_false */
extern T1 T1493f19(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_false */
extern T1 T1492f19(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_false */
extern T1 T1491f19(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_false */
extern T1 T1490f23(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_false */
extern T1 T1445f38(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_false */
extern T1 T1444f37(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_false */
extern T1 T1443f34(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_false */
extern T1 T1442f34(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_false */
extern T1 T1440f36(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_false */
extern T1 T1439f36(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_false */
extern T1 T1438f53(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_false */
extern T1 T1437f53(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_false */
extern T1 T1436f53(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_false */
extern T1 T1435f53(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_false */
extern T1 T1434f46(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_false */
extern T1 T1433f42(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_false */
extern T1 T1432f38(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_false */
extern T1 T1233f26(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_false */
extern T1 T1230f23(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_false */
extern T1 T1201f20(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_false */
extern T1 T1187f26(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_false */
extern T1 T1140f25(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_false */
extern T1 T1016f28(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_false */
extern T1 T1013f23(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_false */
extern T1 T986f34(GE_context* ac, T0* C);
/* detachable ET_VOID.is_false */
extern T1 T975f27(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_false */
extern T1 T863f41(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_false */
extern T1 T715f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_false */
extern T1 T366f54(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_false */
extern T1 T320f27(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_false */
extern T1 T310f19(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_false */
extern T1 T309f21(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_false */
extern T1 T308f22(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_false */
extern T1 T301f28(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_false */
extern T1 T281f26(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_false */
extern T1 T279f21(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_false */
extern T1 T278f24(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_false */
extern T1 T273f30(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_false */
extern T1 T180f102(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_instance_free */
extern T1 T1994f5(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_instance_free */
extern T1 T1993f18(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_instance_free */
extern T1 T1992f16(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_instance_free */
extern T1 T1926f4(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1508f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1507f28(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1506f26(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1505f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1504f29(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1503f27(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_instance_free */
extern T1 T1502f14(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_instance_free */
extern T1 T1501f19(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_instance_free */
extern T1 T1500f12(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_instance_free */
extern T1 T1499f14(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_instance_free */
extern T1 T1498f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_instance_free */
extern T1 T1495f15(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_instance_free */
extern T1 T1494f13(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_instance_free */
extern T1 T1493f13(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_instance_free */
extern T1 T1492f13(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_instance_free */
extern T1 T1491f13(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_instance_free */
extern T1 T1490f17(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1445f24(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1444f24(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_instance_free */
extern T1 T1443f22(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_instance_free */
extern T1 T1442f22(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_instance_free */
extern T1 T1440f23(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_instance_free */
extern T1 T1439f23(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_instance_free */
extern T1 T1438f33(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_instance_free */
extern T1 T1437f33(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_instance_free */
extern T1 T1436f33(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_instance_free */
extern T1 T1435f33(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_instance_free */
extern T1 T1434f20(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_instance_free */
extern T1 T1433f20(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1432f19(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_instance_free */
extern T1 T1233f19(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_instance_free */
extern T1 T1230f17(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_instance_free */
extern T1 T1201f15(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_instance_free */
extern T1 T1187f20(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_instance_free */
extern T1 T1140f18(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION_LIST.is_instance_free */
extern T1 T1089f5(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.is_instance_free */
extern T1 T1086f6(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_instance_free */
extern T1 T1016f17(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.is_instance_free */
extern T1 T1973f4(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.is_instance_free */
extern T1 T1003f10(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_instance_free */
extern T1 T1013f14(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.is_instance_free */
extern T1 T1924f6(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.is_instance_free */
extern T1 T274f11(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_instance_free */
extern T1 T986f25(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.is_instance_free */
extern T1 T1184f7(GE_context* ac, T0* C);
/* detachable ET_VOID.is_instance_free */
extern T1 T975f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_instance_free */
extern T1 T863f25(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_instance_free */
extern T1 T715f16(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_instance_free */
extern T1 T366f28(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_instance_free */
extern T1 T320f16(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_instance_free */
extern T1 T310f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_instance_free */
extern T1 T309f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_instance_free */
extern T1 T308f13(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_instance_free */
extern T1 T301f21(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_instance_free */
extern T1 T281f19(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_instance_free */
extern T1 T279f12(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_instance_free */
extern T1 T278f16(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_instance_free */
extern T1 T273f21(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_instance_free */
extern T1 T180f80(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_feature_name */
extern T1 T180f83(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.position */
extern T0* T1994f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.position */
extern T0* T1993f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.position */
extern T0* T1992f8(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.position */
extern T0* T1926f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.position */
extern T0* T1508f15(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.position */
extern T0* T1507f17(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.position */
extern T0* T1506f17(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.position */
extern T0* T1505f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.position */
extern T0* T1504f40(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.position */
extern T0* T1503f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.position */
extern T0* T1502f6(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.position */
extern T0* T1501f11(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.position */
extern T0* T1500f6(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.position */
extern T0* T1499f7(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.position */
extern T0* T1498f6(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.position */
extern T0* T1495f7(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.position */
extern T0* T1494f8(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.position */
extern T0* T1493f8(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.position */
extern T0* T1492f8(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.position */
extern T0* T1491f8(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.position */
extern T0* T1490f11(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.position */
extern T0* T1445f10(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.position */
extern T0* T1444f10(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.position */
extern T0* T1443f11(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.position */
extern T0* T1442f11(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.position */
extern T0* T1440f11(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.position */
extern T0* T1439f11(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.position */
extern T0* T1438f20(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.position */
extern T0* T1437f20(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.position */
extern T0* T1436f20(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.position */
extern T0* T1435f20(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.position */
extern T0* T1434f24(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.position */
extern T0* T1433f23(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.position */
extern T0* T1432f17(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.position */
extern T0* T1233f13(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_empty */
extern T1 T1233f7(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.position */
extern T0* T1230f10(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.position */
extern T0* T1201f6(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.position */
extern T0* T1187f13(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.position */
extern T0* T1187f13p1(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_empty */
extern T1 T1187f7(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.position */
extern T0* T1140f12(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.position */
extern T0* T1016f12(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.position */
extern T0* T1013f7(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.position */
extern T0* T986f21(GE_context* ac, T0* C);
/* detachable ET_VOID.position */
extern T0* T975f11(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.position */
extern T0* T863f23(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.position */
extern T0* T715f10(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.position */
extern T0* T366f9(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.position */
extern T0* T320f9(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.position */
extern T0* T310f9(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.position */
extern T0* T309f9(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.position */
extern T0* T308f8(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.position */
extern T0* T301f9(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.position */
extern T0* T281f16(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_empty */
extern T1 T281f17(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.position */
extern T0* T279f5(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.position */
extern T0* T278f11(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.position */
extern T0* T273f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.last_leaf */
extern T0* T1994f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.last_leaf */
extern T0* T1993f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.last_leaf */
extern T0* T1992f17(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.last_leaf */
extern T0* T1926f16(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1508f37(GE_context* ac, T0* C);
extern T1 T1508f37ot1(T0* a1, T0** a2);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.last_leaf */
extern T0* T1003f11(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1507f43(GE_context* ac, T0* C);
extern T1 T1507f43ot1(T0* a1, T0** a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1506f39(GE_context* ac, T0* C);
extern T1 T1506f39ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1505f38(GE_context* ac, T0* C);
extern T1 T1505f38ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1504f44(GE_context* ac, T0* C);
extern T1 T1504f44ot1(T0* a1, T0** a2);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1503f40(GE_context* ac, T0* C);
extern T1 T1503f40ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_MANIFEST_STRING.last_leaf */
extern T0* T1502f23(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.last_leaf */
extern T0* T1501f28(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.last_leaf */
extern T0* T1500f20(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.last_leaf */
extern T0* T1499f25(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.last_leaf */
extern T0* T1498f21(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.last_leaf */
extern T0* T1495f25(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.last_leaf */
extern T0* T1494f20(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.last_leaf */
extern T0* T1493f20(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.last_leaf */
extern T0* T1492f20(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.last_leaf */
extern T0* T1491f20(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.last_leaf */
extern T0* T1490f24(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.last_leaf */
extern T0* T1445f45(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.last_leaf */
extern T0* T1444f44(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.last_leaf */
extern T0* T1443f37(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.last_leaf */
extern T0* T1442f37(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.last_leaf */
extern T0* T1440f39(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.last_leaf */
extern T0* T1439f39(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.last_leaf */
extern T0* T1438f61(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.last_leaf */
extern T0* T1437f61(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.last_leaf */
extern T0* T1436f61(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.last_leaf */
extern T0* T1435f61(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.last_leaf */
extern T0* T1434f49(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.last_leaf */
extern T0* T1433f45(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.last_leaf */
extern T0* T1432f45(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.last_leaf */
extern T0* T1233f27(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.last_leaf */
extern T0* T1230f28(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.last_leaf */
extern T0* T1201f21(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.last_leaf */
extern T0* T1187f27(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.last_leaf */
extern T0* T1140f26(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.last_leaf */
extern T0* T1016f29(GE_context* ac, T0* C);
extern T1 T1016f29ot1(T0* a1, T0** a2);
/* detachable ET_BRACKET_EXPRESSION.last_leaf */
extern T0* T1013f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.last_leaf */
extern T0* T610f106(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.last_leaf */
extern T0* T986f35(GE_context* ac, T0* C);
/* detachable ET_VOID.last_leaf */
extern T0* T975f28(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.last_leaf */
extern T0* T863f44(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.last_leaf */
extern T0* T715f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.last_leaf */
extern T0* T366f58(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.last_leaf */
extern T0* T320f25(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.last_leaf */
extern T0* T310f20(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.last_leaf */
extern T0* T309f22(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.last_leaf */
extern T0* T308f23(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.last_leaf */
extern T0* T301f29(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.last_leaf */
extern T0* T281f27(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.last_leaf */
extern T0* T279f24(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.last_leaf */
extern T0* T278f25(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.last_leaf */
extern T0* T273f31(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.overridden_type_mark */
extern T0* T1450f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.overridden_type_mark */
extern T0* T867f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.overridden_type_mark */
extern T0* T246f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.type_id */
extern T6 T444f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.type_id */
extern T6 T443f29(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.type_id */
extern T6 T315f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.type_id */
extern T6 T311f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.type_id */
extern T6 T294f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_id */
extern T6 T219f38(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_generic */
extern T1 T315f6(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.base_type */
extern T0* T315f9(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.base_class */
extern T0* T315f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.conforms_to_type */
extern T1 T444f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.conforms_to_type */
extern T1 T443f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.conforms_to_type */
extern T1 T315f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.conforms_to_type */
extern T1 T311f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.conforms_to_type */
extern T1 T294f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_type */
extern T1 T219f34(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.hash_code */
extern T6 T315f21(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_basic */
extern T1 T315f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_type */
extern T1 T1416f7(GE_context* ac, T0* C, T0* a1);
extern T1 T1416f7ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_type */
extern T1 T1416f7p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.index_of */
extern T6 T1416f11(GE_context* ac, T0* C, T0* a1);
extern T1 T1416f11ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.index_of */
extern T6 T1416f11p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_type */
extern T1 T469f11(GE_context* ac, T0* C, T0* a1);
extern T1 T469f11ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_type */
extern T1 T469f11p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.index_of */
extern T6 T469f13(GE_context* ac, T0* C, T0* a1);
extern T1 T469f13ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.index_of */
extern T6 T469f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_type */
extern T1 T444f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_type */
extern T1 T443f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.has_type */
extern T1 T315f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_type */
extern T1 T311f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_type */
extern T1 T294f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_type */
extern T1 T261f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_type */
extern T1 T261f7p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.index_of */
extern T6 T261f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.index_of */
extern T6 T261f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_type */
extern T1 T219f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_types */
extern T0* T444f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_types */
extern T0* T443f51(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.dynamic_types */
extern T0* T315f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_types */
extern T0* T311f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_types */
extern T0* T294f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_types */
extern T0* T219f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.special_type */
extern T0* T1416f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.special_type */
extern T0* T469f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.special_type */
extern T0* T444f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.special_type */
extern T0* T443f53(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.special_type */
extern T0* T315f23(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.special_type */
extern T0* T311f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.special_type */
extern T0* T294f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.special_type */
extern T0* T261f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.special_type */
extern T0* T219f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.is_empty */
extern T1 T1416f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.is_empty */
extern T1 T469f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_empty */
extern T1 T444f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_empty */
extern T1 T443f24(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_empty */
extern T1 T315f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_empty */
extern T1 T311f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_empty */
extern T1 T294f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.is_empty */
extern T1 T261f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_empty */
extern T1 T219f26(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.is_expanded */
extern T1 T1416f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.is_expanded */
extern T1 T469f10(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_expanded */
extern T1 T315f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.is_expanded */
extern T1 T261f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_expanded */
extern T1 T1416f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_expanded */
extern T1 T469f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_expanded */
extern T1 T444f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_expanded */
extern T1 T443f31(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.has_expanded */
extern T1 T315f18(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_expanded */
extern T1 T311f35(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_expanded */
extern T1 T294f34(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_expanded */
extern T1 T261f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_expanded */
extern T1 T219f24(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.count */
extern T6 T444f25(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.count */
extern T6 T443f25(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.count */
extern T6 T315f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.count */
extern T6 T311f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.count */
extern T6 T294f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.count */
extern T6 T219f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.dynamic_type */
extern T0* T1416f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.dynamic_type */
extern T0* T469f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_type */
extern T0* T444f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_type */
extern T0* T443f26(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.dynamic_type */
extern T0* T315f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_type */
extern T0* T311f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_type */
extern T0* T294f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.dynamic_type */
extern T0* T261f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_type */
extern T0* T219f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.can_be_void */
extern T1 T1416f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.can_be_void */
extern T1 T469f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.can_be_void */
extern T1 T444f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.can_be_void */
extern T1 T443f48(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.can_be_void */
extern T1 T315f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.can_be_void */
extern T1 T311f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.can_be_void */
extern T1 T294f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.can_be_void */
extern T1 T261f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.can_be_void */
extern T1 T219f23(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_never_void */
extern T1 T444f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_never_void */
extern T1 T443f32(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_never_void */
extern T1 T315f13(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_never_void */
extern T1 T311f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_never_void */
extern T1 T294f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_never_void */
extern T1 T219f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.static_type */
extern T0* T444f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.static_type */
extern T0* T443f46(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.static_type */
extern T0* T315f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.static_type */
extern T0* T311f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.static_type */
extern T0* T294f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.static_type */
extern T0* T219f45(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.named_types */
extern T0* T1902f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.named_types */
extern T0* T1901f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.named_types */
extern T0* T593f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.named_types */
extern T0* T567f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.named_types */
extern T0* T236f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.conforms_to_types */
extern T1 T1902f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T1901f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_FORMAL_PARAMETER_LIST.conforms_to_types */
extern T1 T593f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T567f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.conforms_to_types */
extern T1 T236f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_unqualified_anchored_type */
extern T1 T1902f7(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T1901f7(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.has_unqualified_anchored_type */
extern T1 T593f15(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T567f8(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_unqualified_anchored_type */
extern T1 T236f16(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_formal_types */
extern T1 T1902f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T1901f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.has_formal_types */
extern T1 T593f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T567f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_formal_types */
extern T1 T236f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.depends_on_qualified_anchored_type */
extern T1 T1902f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T1901f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.depends_on_qualified_anchored_type */
extern T1 T593f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T567f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.depends_on_qualified_anchored_type */
extern T1 T236f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_identifier_anchored_type */
extern T1 T1902f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T1901f11(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.has_identifier_anchored_type */
extern T1 T593f21(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T567f11(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_identifier_anchored_type */
extern T1 T236f19(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.resolved_formal_parameters */
extern T0* T1902f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T1901f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.resolved_formal_parameters */
extern T0* T593f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T567f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.resolved_formal_parameters */
extern T0* T236f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.same_syntactical_types */
extern T1 T1902f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T1901f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FORMAL_PARAMETER_LIST.same_syntactical_types */
extern T1 T593f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T567f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.same_syntactical_types */
extern T1 T236f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.actual_parameter */
extern T0* T1902f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T1901f16(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.actual_parameter */
extern T0* T593f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T567f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.actual_parameter */
extern T0* T236f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.item */
extern T0* T236f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.index_of_label */
extern T6 T1902f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.index_of_label */
extern T6 T1901f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.index_of_label */
extern T6 T593f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.index_of_label */
extern T6 T567f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.index_of_label */
extern T6 T236f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.same_named_types */
extern T1 T1902f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.same_named_types */
extern T1 T1901f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FORMAL_PARAMETER_LIST.same_named_types */
extern T1 T593f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.same_named_types */
extern T1 T567f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.same_named_types */
extern T1 T236f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.is_empty */
extern T1 T1902f4(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.is_empty */
extern T1 T1901f15(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.is_empty */
extern T1 T593f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.is_empty */
extern T1 T567f5(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.is_empty */
extern T1 T236f10(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.count */
extern T6 T1902f17(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.type */
extern T0* T1902f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.type */
extern T0* T1901f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.type */
extern T0* T593f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.type */
extern T0* T567f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.type */
extern T0* T236f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.last_leaf */
extern T0* T1902f21(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.null_leaf */
extern T0* T72f1524(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.make */
extern T0* T2039c5(GE_context* ac);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.last_leaf */
extern T0* T1901f18(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.last_leaf */
extern T0* T593f25(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.last_leaf */
extern T0* T567f16(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.last_leaf */
extern T0* T236f25(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_reference_mark */
extern T1 T983f6(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_reference_mark */
extern T1 T610f14(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_reference_mark */
extern T1 T561f4(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_reference_mark */
extern T1 T409f9(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_reference_mark */
extern T1 T343f10(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separate_mark */
extern T1 T983f7(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separate_mark */
extern T1 T610f15(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_separate_mark */
extern T1 T561f5(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separate_mark */
extern T1 T409f10(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separate_mark */
extern T1 T343f11(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_attachment_mark */
extern T1 T983f10(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_question_mark */
extern T1 T983f14(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_bang */
extern T1 T983f12(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_detachable */
extern T1 T983f13(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attached */
extern T1 T983f11(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_attachment_mark */
extern T1 T610f18(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_question_mark */
extern T1 T610f22(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_bang */
extern T1 T610f20(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_detachable */
extern T1 T610f21(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attached */
extern T1 T610f19(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_attachment_mark */
extern T1 T409f24(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_question_mark */
extern T1 T409f17(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_bang */
extern T1 T409f15(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_detachable */
extern T1 T409f16(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attached */
extern T1 T409f14(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_attachment_mark */
extern T1 T343f14(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_question_mark */
extern T1 T343f19(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_bang */
extern T1 T343f17(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_detachable */
extern T1 T343f18(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attached */
extern T1 T343f16(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_attachment_mark */
extern T1 T342f12(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_question_mark */
extern T1 T342f19(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_bang */
extern T1 T342f18(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_detachable */
extern T1 T342f17(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_attached */
extern T1 T342f16(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.overridden_type_mark */
extern T0* T983f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_type_mark */
extern T0* T72f44(GE_context* ac, T0* C, T1 a1, T1 a2, T1 a3, T1 a4, T1 a5);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_no_type_mark */
extern T0* T72f82(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_reference_type_mark */
extern T0* T72f81(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_reference_mark */
extern void T342f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_expanded_type_mark */
extern T0* T72f80(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_expanded_mark */
extern void T342f30(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_type_mark */
extern T0* T72f79(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_separate_mark */
extern void T342f29(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_reference_type_mark */
extern T0* T72f78(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_expanded_type_mark */
extern T0* T72f77(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_reference_type_mark */
extern T0* T72f76(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_expanded_type_mark */
extern T0* T72f75(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_type_mark */
extern T0* T72f74(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_reference_type_mark */
extern T0* T72f73(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_expanded_type_mark */
extern T0* T72f72(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_reference_type_mark */
extern T0* T72f71(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_expanded_type_mark */
extern T0* T72f70(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_type_mark */
extern T0* T72f69(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_reference_type_mark */
extern T0* T72f68(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_expanded_type_mark */
extern T0* T72f67(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.overridden_type_mark */
extern T0* T610f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.overridden_type_mark */
extern T0* T561f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.overridden_type_mark */
extern T0* T409f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.overridden_type_mark */
extern T0* T343f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.overridden_type_mark */
extern T0* T342f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expanded_mark */
extern T1 T983f5(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expanded_mark */
extern T1 T610f13(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expanded_mark */
extern T1 T561f3(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expanded_mark */
extern T1 T409f8(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expanded_mark */
extern T1 T343f9(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separateness_mark */
extern T1 T983f9(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separateness_mark */
extern T1 T610f17(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_separateness_mark */
extern T1 T561f7(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separateness_mark */
extern T1 T409f12(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separateness_mark */
extern T1 T343f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_separateness_mark */
extern T1 T342f6(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_implicit_mark */
extern T1 T983f28(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_implicit_mark */
extern T1 T610f32(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_implicit_mark */
extern T1 T561f21(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_implicit_mark */
extern T1 T409f30(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_implicit_mark */
extern T1 T343f15(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_implicit_mark */
extern T1 T342f24(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attachment_mark */
extern T1 T983f19(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attachment_mark */
extern T1 T610f9(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_attachment_mark */
extern T1 T561f8(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attachment_mark */
extern T1 T409f13(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attachment_mark */
extern T1 T343f4(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_attachment_mark */
extern T1 T342f7(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expandedness_mark */
extern T1 T983f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expandedness_mark */
extern T1 T610f16(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expandedness_mark */
extern T1 T561f6(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expandedness_mark */
extern T1 T409f11(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expandedness_mark */
extern T1 T343f12(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_expandedness_mark */
extern T1 T342f8(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.position */
extern T0* T983f29(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.position */
extern T0* T610f5(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.position */
extern T0* T561f22(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.position */
extern T0* T409f27(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.position */
extern T0* T343f7(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.position */
extern T0* T342f25(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attached_mark */
extern T1 T983f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attached_mark */
extern T1 T610f10(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_attached_mark */
extern T1 T561f9(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attached_mark */
extern T1 T409f6(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attached_mark */
extern T1 T343f5(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_detachable_mark */
extern T1 T983f37(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_detachable_mark */
extern T1 T610f11(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_detachable_mark */
extern T1 T561f12(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_detachable_mark */
extern T1 T409f7(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_detachable_mark */
extern T1 T343f6(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.actual_parameter */
extern T0* T1717f6(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.actual_parameter */
extern T0* T1478f38(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.actual_parameter */
extern T0* T1477f31(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.actual_parameter */
extern T0* T1476f25(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.actual_parameter */
extern T0* T1474f6(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.actual_parameter */
extern T0* T1450f37(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.actual_parameter */
extern T0* T867f38(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.actual_parameter */
extern T0* T442f14(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.actual_parameter */
extern T0* T248f44(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.actual_parameter */
extern T0* T246f20(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_parameter */
extern T0* T73f114(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.actual_parameter */
extern T0* T71f43(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.type */
extern T0* T1717f3(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.type */
extern T0* T1474f3(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.type */
extern T0* T1450f27(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.type */
extern T0* T867f25(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.type */
extern T0* T73f113(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1717f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters_with_type */
extern T0* T1478f63(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters_with_type */
extern T0* T1477f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters_with_type */
extern T0* T1476f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LABELED_ACTUAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1474f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1450f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T867f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters_with_type */
extern T0* T442f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters_with_type */
extern T0* T248f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters_with_type */
extern T0* T246f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.resolved_formal_parameters_with_type */
extern T0* T73f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters_with_type */
extern T0* T71f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.hash_code */
extern T6 T826f92(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.hash_code */
extern T6 T586f91(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.hash_code */
extern T6 T207f91(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.hash_code */
extern T6 T183f102(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.hash_code */
extern T6 T56f138(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.hash_code */
extern T6 T36f121(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T183f116(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T183f116ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T183f116ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T183f116ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.initial_classes_in_group_capacity */
extern T6 T183f117(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T56f156(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T56f156ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T56f156ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T56f156ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.initial_classes_in_group_capacity */
extern T6 T56f157(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.classes_in_group */
extern T0* T36f132(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T36f132ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T36f132ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T36f132ac3(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_SYSTEM.initial_classes_in_group_capacity */
extern T6 T36f133(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.full_name */
extern T0* T826f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.relative_name */
extern T0* T826f100(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_LIBRARY.string_ */
extern T0* T826f102(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.full_name */
extern T0* T586f96(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.relative_name */
extern T0* T586f98(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_DOTNET_ASSEMBLY.string_ */
extern T0* T586f100(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.full_name */
extern T0* T207f96(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.relative_name */
extern T0* T207f100(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.string_ */
extern T0* T207f103(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.full_name */
extern T0* T183f108(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_SYSTEM.full_name */
extern T0* T56f152(GE_context* ac, T0* C, T2 a1);
/* detachable ET_SYSTEM.full_name */
extern T0* T36f127(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_LIBRARY.has_master_class */
extern T1 T183f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.has_master_class */
extern T1 T56f139(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.has_master_class */
extern T1 T36f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.master_class */
extern T0* T826f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.master_class */
extern T0* T183f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.master_class */
extern T0* T56f146(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.master_class */
extern T0* T36f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.kind_name */
extern T0* T826f98(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.kind_name */
extern T0* T586f101(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.kind_name */
extern T0* T207f98(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.kind_name */
extern T0* T183f111(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.kind_name */
extern T0* T56f153(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.kind_name */
extern T0* T36f128(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.full_lower_name */
extern T0* T826f96(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.relative_lower_name */
extern T0* T826f99(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_LIBRARY.lower_name */
extern T0* T826f101(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.full_lower_name */
extern T0* T586f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.relative_lower_name */
extern T0* T586f99(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.full_lower_name */
extern T0* T207f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.relative_lower_name */
extern T0* T207f101(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.full_lower_name */
extern T0* T183f105(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_SYSTEM.full_lower_name */
extern T0* T56f154(GE_context* ac, T0* C, T2 a1);
/* detachable ET_SYSTEM.full_lower_name */
extern T0* T36f129(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.shortest_path */
extern T0* T826f103(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T826f103ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T826f103ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T826f103ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE].negated */
extern T1 T1408f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]].set_rout_disp_final */
extern void T1409f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_LIBRARY.insert_in_hash_table */
extern void T826f184(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].force_first */
extern void T882f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].put */
extern void T882f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].move_cursors_right */
extern void T882f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE].set_position */
extern void T1581f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].move_right */
extern void T882f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].force */
extern void T1580f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].fill_with */
extern void T1579f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].extend */
extern void T1579f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].put_last */
extern void T882f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].resize */
extern void T882f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].aliased_resized_area */
extern T0* T1580f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].aliased_resized_area */
extern T0* T1579f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].new_capacity */
extern T6 T882f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].extendible */
extern T1 T882f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].before */
extern T1 T1406f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].back */
extern void T1406f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_back */
extern void T1405f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].add_traversing_cursor */
extern void T1405f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_next_cursor */
extern void T1406f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_position */
extern void T1406f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].remove_traversing_cursor */
extern void T1405f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_before */
extern void T1406f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_item */
extern T6 T1405f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].finish */
extern void T1406f15(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_finish */
extern void T1405f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_off */
extern T1 T1405f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].off */
extern T1 T1406f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].is_empty */
extern T1 T1405f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].forth */
extern void T1406f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_forth */
extern void T1405f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_after */
extern void T1406f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].go_after */
extern void T1406f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_go_after */
extern void T1405f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].has */
extern T1 T1405f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].search_position */
extern void T1405f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].hash_position */
extern T6 T1405f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_item */
extern T0* T1405f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_item */
extern T6 T1405f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LIBRARY.universe_actions */
extern T0* T826f104(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE].default_create */
extern T0* T1408c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].item */
extern T6 T1406f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_item */
extern T6 T1405f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_item */
extern T6 T1405f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].key */
extern T0* T1406f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_key */
extern T0* T1405f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].after */
extern T1 T1406f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].start */
extern void T1406f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_start */
extern void T1405f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_cursor */
extern T0* T1405f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].make */
extern T0* T1406c11(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].force_last */
extern void T1405f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_put */
extern void T1405f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_put */
extern void T1405f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_put */
extern void T1405f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].resize */
extern void T1405f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_resize */
extern void T1405f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].special_integer_ */
extern T0* T1405f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_resize */
extern void T1405f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_resize */
extern void T1405f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_resize */
extern void T1405f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_modulus */
extern T6 T1405f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_capacity */
extern T6 T1405f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_put */
extern void T1405f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].unset_found_item */
extern void T1405f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_map */
extern T0* T1405c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_with_equality_testers */
extern void T1405f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_sparse_container */
extern void T1405f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_internal_cursor */
extern void T1405f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_slots */
extern void T1405f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_clashes */
extern void T1405f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_key_storage */
extern void T1405f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_item_storage */
extern void T1405f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T882c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].set_internal_cursor */
extern void T882f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].new_cursor */
extern T0* T882f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T1581c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T1580f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].make_empty */
extern T0* T1579c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].default_create */
extern T0* T1580c3(GE_context* ac);
/* [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T586f102(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T586f102ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T586f102ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T586f102ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.insert_in_hash_table */
extern void T586f182(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.universe_actions */
extern T0* T586f103(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T207f99(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T207f99ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T207f99ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T207f99ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.insert_in_hash_table */
extern void T207f182(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.universe_actions */
extern T0* T207f102(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T183f109(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T183f109ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T183f109ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T183f109ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.insert_in_hash_table */
extern void T183f209(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.universe_actions */
extern T0* T183f113(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T56f134(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T56f134ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T56f134ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T56f134ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.insert_in_hash_table */
extern void T56f253(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM.universe_actions */
extern T0* T56f136(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.shortest_path */
extern T0* T36f122(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T36f122ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T36f122ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T36f122ac3(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_SYSTEM.insert_in_hash_table */
extern void T36f216(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* detachable ET_SYSTEM.universe_actions */
extern T0* T36f124(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.adapted_universe */
extern T0* T826f93(GE_context* ac, T0* C, T0* a1);
extern T1 T826f93ot1(T0* a1, T0** a2);
extern T1 T826f93ot2(T0* a1, T0** a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.adapted_dotnet_assembly */
extern T0* T206f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.adapted_library */
extern T0* T184f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.adapted_universe */
extern T0* T586f93(GE_context* ac, T0* C, T0* a1);
extern T1 T586f93ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].has */
extern T1 T1356f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.adapted_universe */
extern T0* T207f93(GE_context* ac, T0* C, T0* a1);
extern T1 T207f93ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_LIBRARY.adapted_universe */
extern T0* T183f103(GE_context* ac, T0* C, T0* a1);
extern T1 T183f103ot1(T0* a1, T0** a2);
extern T1 T183f103ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM.adapted_universe */
extern T0* T56f149(GE_context* ac, T0* C, T0* a1);
extern T1 T56f149ot1(T0* a1, T0** a2);
extern T1 T56f149ot2(T0* a1, T0** a2);
/* detachable ET_SYSTEM.adapted_universe */
extern T0* T36f120(GE_context* ac, T0* C, T0* a1);
extern T1 T36f120ot1(T0* a1, T0** a2);
extern T1 T36f120ot2(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.context_base_type_with_type_mark */
extern T0* T248f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_with_type_mark */
extern T0* T237f29(GE_context* ac, T0* C, T0* a1);
extern T1 T237f29ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_base_type_with_type_mark */
extern T0* T73f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_base_type_with_type_mark */
extern T0* T71f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_base_class */
extern T0* T237f27(GE_context* ac, T0* C);
extern T1 T237f27ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.context_is_type_expanded_with_type_mark */
extern T1 T248f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_expanded_with_type_mark */
extern T1 T237f33(GE_context* ac, T0* C, T0* a1);
extern T1 T237f33ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_is_type_expanded_with_type_mark */
extern T1 T73f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_expanded_with_type_mark */
extern T1 T71f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_is_type_detachable_with_type_mark */
extern T1 T248f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_detachable_with_type_mark */
extern T1 T237f34(GE_context* ac, T0* C, T0* a1);
extern T1 T237f34ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_is_type_detachable_with_type_mark */
extern T1 T73f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_detachable_with_type_mark */
extern T1 T71f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_class_type_with_type_marks */
extern T1 T248f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_class_type_with_type_marks */
extern T1 T237f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T237f30ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.cloned_type_context */
extern T0* T237f9(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.context_conforms_from_class_type_with_type_marks */
extern T1 T73f149(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_class_type_with_type_marks */
extern T1 T71f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_tuple_type_with_type_marks */
extern T1 T248f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_tuple_type_with_type_marks */
extern T1 T237f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T237f32ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_conforms_from_tuple_type_with_type_marks */
extern T1 T73f150(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_tuple_type_with_type_marks */
extern T1 T71f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_class_type_with_type_marks */
extern T1 T248f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_class_type_with_type_marks */
extern T1 T237f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T237f23ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_same_named_class_type_with_type_marks */
extern T1 T73f151(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_class_type_with_type_marks */
extern T1 T71f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.context_same_named_tuple_type_with_type_marks */
extern T1 T248f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_tuple_type_with_type_marks */
extern T1 T237f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T237f24ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_same_named_tuple_type_with_type_marks */
extern T1 T73f152(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_tuple_type_with_type_marks */
extern T1 T71f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.context_conforms_to_type_with_type_marks */
extern T1 T248f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_type_with_type_marks */
extern T1 T237f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T237f25ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_conforms_to_type_with_type_marks */
extern T1 T73f153(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_to_type_with_type_marks */
extern T1 T71f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_type_with_type_marks */
extern T1 T248f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_type_with_type_marks */
extern T1 T237f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T237f15ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_same_named_type_with_type_marks */
extern T1 T73f154(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_type_with_type_marks */
extern T1 T71f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.context_named_type_with_type_mark */
extern T0* T248f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_type_with_type_mark */
extern T0* T237f16(GE_context* ac, T0* C, T0* a1);
extern T1 T237f16ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_named_type_with_type_mark */
extern T0* T73f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_named_type_with_type_mark */
extern T0* T71f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T248f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_formal_parameter_type_with_type_marks */
extern T1 T237f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T237f26ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T73f157(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T71f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T248f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T237f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T237f31ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T73f191(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T71f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.adapted_base_class_with_named_feature */
extern T0* T237f38(GE_context* ac, T0* C, T0* a1);
extern T1 T237f38ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.adapted_base_class_with_seeded_feature */
extern T0* T237f39(GE_context* ac, T0* C, T6 a1);
extern T1 T237f39ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.context_is_type_reference_with_type_mark */
extern T1 T248f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_reference_with_type_mark */
extern T1 T237f40(GE_context* ac, T0* C, T0* a1);
extern T1 T237f40ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_is_type_reference_with_type_mark */
extern T1 T73f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_reference_with_type_mark */
extern T1 T71f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_is_type_attached_with_type_mark */
extern T1 T248f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_attached_with_type_mark */
extern T1 T237f42(GE_context* ac, T0* C, T0* a1);
extern T1 T237f42ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_is_type_attached_with_type_mark */
extern T1 T73f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_attached_with_type_mark */
extern T1 T71f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_same_base_type_with_type_marks */
extern T1 T248f100(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_base_type_with_type_marks */
extern T1 T237f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T237f48ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_same_base_type_with_type_marks */
extern T1 T73f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_base_type_with_type_marks */
extern T1 T71f112(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_type */
extern T0* T237f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.context_named_type */
extern T0* T73f237(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.context_named_type */
extern T0* T71f108(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.attachment_type_conformance_mode */
extern T1 T248f32(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.attachment_type_conformance_mode */
extern T1 T237f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.attachment_type_conformance_mode */
extern T1 T73f67(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.attachment_type_conformance_mode */
extern T1 T71f31(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.scoop_mode */
extern T1 T248f33(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.scoop_mode */
extern T1 T237f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.scoop_mode */
extern T1 T73f68(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.scoop_mode */
extern T1 T71f32(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_class */
extern T0* T237f20(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.context_base_type_actual_count */
extern T6 T248f46(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_actual_count */
extern T6 T237f21(GE_context* ac, T0* C);
extern T1 T237f21ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_base_type_actual_count */
extern T6 T73f147(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.context_base_type_actual_count */
extern T6 T71f51(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.context_base_type_actual */
extern T0* T248f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_actual */
extern T0* T237f22(GE_context* ac, T0* C, T6 a1);
extern T1 T237f22ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_base_type_actual */
extern T0* T73f148(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.context_base_type_actual */
extern T0* T71f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.is_root_context */
extern T1 T248f14(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_root_context */
extern T1 T237f19(GE_context* ac, T0* C);
extern T1 T237f19ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.is_root_context */
extern T1 T73f112(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_root_context */
extern T1 T71f20(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.root_context */
extern T0* T248f15(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.root_context */
extern T0* T73f159(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.root_context */
extern T0* T71f21(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.as_nested_type_context */
extern T0* T248f34(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.as_nested_type_context */
extern T0* T237f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.as_nested_type_context */
extern T0* T73f69(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.as_nested_type_context */
extern T0* T71f34(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.to_nested_type_context */
extern T0* T248f35(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.to_nested_type_context */
extern T0* T237f8(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.to_nested_type_context */
extern T0* T73f70(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.to_nested_type_context */
extern T0* T71f35(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.type */
extern T0* T1478f35(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type */
extern T0* T1477f30(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.type */
extern T0* T1476f24(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.type */
extern T0* T442f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type */
extern T0* T248f30(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type */
extern T0* T246f19(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type */
extern T0* T71f28(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_base_type */
extern T0* T1478f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_base_type */
extern T0* T1477f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_base_type */
extern T0* T1476f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_base_type */
extern T0* T1450f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.shallow_base_type */
extern T0* T867f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.shallow_base_type */
extern T0* T442f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.shallow_base_type */
extern T0* T248f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_base_type */
extern T0* T246f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.shallow_base_type */
extern T0* T73f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.shallow_base_type */
extern T0* T71f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_reference */
extern T1 T1478f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_reference */
extern T1 T1477f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_reference */
extern T1 T1476f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_reference */
extern T1 T1450f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_reference */
extern T1 T867f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_reference */
extern T1 T442f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_reference */
extern T1 T248f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_reference */
extern T1 T246f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_reference */
extern T1 T73f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_reference */
extern T1 T71f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_constraint */
extern T1 T1478f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T1478f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_constraint */
extern T1 T1477f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T1477f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_LIKE_FEATURE.conforms_to_constraint */
extern T1 T1476f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_LIKE_FEATURE.conforms_to_constraint_with_type_marks */
extern T1 T1476f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_constraint */
extern T1 T1450f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_constraint_with_type_marks */
extern T1 T1450f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_constraint */
extern T1 T867f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_constraint_with_type_marks */
extern T1 T867f73(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_LIKE_CURRENT.conforms_to_constraint */
extern T1 T442f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_CURRENT.conforms_to_constraint_with_type_marks */
extern T1 T442f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_to_constraint */
extern T1 T248f80(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T248f94(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_constraint */
extern T1 T246f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T246f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_to_constraint */
extern T1 T73f211(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_to_constraint_with_type_marks */
extern T1 T73f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_to_constraint */
extern T1 T71f82(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T71f98(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_class */
extern T0* T1478f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_class */
extern T0* T1477f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.base_class */
extern T0* T1476f38(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_class */
extern T0* T1450f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.base_class */
extern T0* T867f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_class */
extern T0* T442f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.base_class_in_context */
extern T0* T248f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_class */
extern T0* T246f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.base_class_in_context */
extern T0* T73f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.base_class_in_context */
extern T0* T71f75(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_type_with_type_mark */
extern T0* T1478f8(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f8ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.overridden_type_mark */
extern T0* T1478f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_type_with_type_mark */
extern T0* T1477f26(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f26ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.overridden_type_mark */
extern T0* T1477f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.named_type_with_type_mark */
extern T0* T1476f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LIKE_FEATURE.overridden_type_mark */
extern T0* T1476f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_type_with_type_mark */
extern T0* T1450f12(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f12ot1(T0* a1, T0** a2);
extern T1 T1450f12ot2(T0* a1, T0** a2);
extern T1 T1450f12ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.named_type_with_type_mark */
extern T0* T1424f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_N.overridden_type_mark */
extern T0* T1424f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.named_type_with_type_mark */
extern T0* T867f10(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f10ot1(T0* a1, T0** a2);
extern T1 T867f10ot2(T0* a1, T0** a2);
extern T1 T867f10ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.named_type_with_type_mark */
extern T0* T442f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.overridden_type_mark */
extern T0* T442f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_type_with_type_mark */
extern T0* T246f6(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f6ot1(T0* a1, T0** a2);
extern T1 T246f6ot2(T0* a1, T0** a2);
extern T1 T246f6ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_actual_count */
extern T6 T1478f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1478f9ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_actual_count */
extern T6 T1477f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1477f9ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.base_type_actual_count */
extern T6 T1476f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_actual_count */
extern T6 T1450f13(GE_context* ac, T0* C, T0* a1);
extern T1 T1450f13ot1(T0* a1, T0** a2);
extern T1 T1450f13ot2(T0* a1, T0** a2);
extern T1 T1450f13ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.base_type_actual_count */
extern T6 T1424f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.base_type_actual_count */
extern T6 T867f11(GE_context* ac, T0* C, T0* a1);
extern T1 T867f11ot1(T0* a1, T0** a2);
extern T1 T867f11ot2(T0* a1, T0** a2);
extern T1 T867f11ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.base_type_actual_count */
extern T6 T442f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_actual_count */
extern T6 T246f9(GE_context* ac, T0* C, T0* a1);
extern T1 T246f9ot1(T0* a1, T0** a2);
extern T1 T246f9ot2(T0* a1, T0** a2);
extern T1 T246f9ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_actual */
extern T0* T1478f10(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1478f10ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_actual */
extern T0* T1477f40(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1477f40ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.base_type_actual */
extern T0* T1476f41(GE_context* ac, T0* C, T6 a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_actual */
extern T0* T1450f14(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1450f14ot1(T0* a1, T0** a2);
extern T1 T1450f14ot2(T0* a1, T0** a2);
extern T1 T1450f14ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.base_type_actual */
extern T0* T1424f6(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_actual */
extern T0* T867f12(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T867f12ot1(T0* a1, T0** a2);
extern T1 T867f12ot2(T0* a1, T0** a2);
extern T1 T867f12ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.base_type_actual */
extern T0* T442f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_actual */
extern T0* T246f10(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T246f10ot1(T0* a1, T0** a2);
extern T1 T246f10ot2(T0* a1, T0** a2);
extern T1 T246f10ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_class_type_with_type_marks */
extern T1 T1478f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1478f11ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_class_type_with_type_marks */
extern T1 T1477f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1477f14ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_class_type_with_type_marks */
extern T1 T1476f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_class_type_with_type_marks */
extern T1 T1450f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f15ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_named_class_type_with_type_marks */
extern T1 T1424f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_named_class_type_with_type_marks */
extern T1 T867f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f13ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_named_class_type_with_type_marks */
extern T1 T442f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_class_type_with_type_marks */
extern T1 T246f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f11ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T1478f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1478f12ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T1477f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1477f15ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_tuple_type_with_type_marks */
extern T1 T1476f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_tuple_type_with_type_marks */
extern T1 T1450f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f16ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_named_tuple_type_with_type_marks */
extern T1 T1424f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_named_tuple_type_with_type_marks */
extern T1 T867f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f14ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_named_tuple_type_with_type_marks */
extern T1 T442f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T246f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f12ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_type_with_type_marks */
extern T1 T1478f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1478f13ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_type_with_type_marks */
extern T1 T1477f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1477f24ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_to_type_with_type_marks */
extern T1 T1476f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_type_with_type_marks */
extern T1 T1450f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1450f17ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.conforms_to_type_with_type_marks */
extern T1 T1424f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_type_with_type_marks */
extern T1 T867f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T867f15ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.conforms_to_type_with_type_marks */
extern T1 T442f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_type_with_type_marks */
extern T1 T246f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T246f13ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1478f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1478f14ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1477f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1477f16ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1476f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_formal_parameter_type_with_type_marks */
extern T1 T1450f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f18ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_named_formal_parameter_type_with_type_marks */
extern T1 T1424f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_named_formal_parameter_type_with_type_marks */
extern T1 T867f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f16ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_named_formal_parameter_type_with_type_marks */
extern T1 T442f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T246f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f14ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_base_class */
extern T0* T1478f15(GE_context* ac, T0* C, T0* a1);
extern T1 T1478f15ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_base_class */
extern T0* T1477f21(GE_context* ac, T0* C, T0* a1);
extern T1 T1477f21ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.named_base_class */
extern T0* T1476f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_base_class */
extern T0* T1450f19(GE_context* ac, T0* C, T0* a1);
extern T1 T1450f19ot1(T0* a1, T0** a2);
extern T1 T1450f19ot2(T0* a1, T0** a2);
extern T1 T1450f19ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.named_base_class */
extern T0* T1424f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.named_base_class */
extern T0* T867f17(GE_context* ac, T0* C, T0* a1);
extern T1 T867f17ot1(T0* a1, T0** a2);
extern T1 T867f17ot2(T0* a1, T0** a2);
extern T1 T867f17ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.named_base_class */
extern T0* T442f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_base_class_in_context */
extern T0* T248f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_base_class */
extern T0* T246f15(GE_context* ac, T0* C, T0* a1);
extern T1 T246f15ot1(T0* a1, T0** a2);
extern T1 T246f15ot2(T0* a1, T0** a2);
extern T1 T246f15ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS.named_base_class_in_context */
extern T0* T73f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_base_class_in_context */
extern T0* T71f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_with_type_mark */
extern T0* T1478f21(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f21ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_with_type_mark */
extern T0* T1477f43(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f43ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.base_type_with_type_mark */
extern T0* T1476f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_with_type_mark */
extern T0* T1450f20(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f20ot1(T0* a1, T0** a2);
extern T1 T1450f20ot2(T0* a1, T0** a2);
extern T1 T1450f20ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.base_type_with_type_mark */
extern T0* T1424f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_with_type_mark */
extern T0* T867f18(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f18ot1(T0* a1, T0** a2);
extern T1 T867f18ot2(T0* a1, T0** a2);
extern T1 T867f18ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.base_type_with_type_mark */
extern T0* T442f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_with_type_mark */
extern T0* T246f28(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f28ot1(T0* a1, T0** a2);
extern T1 T246f28ot2(T0* a1, T0** a2);
extern T1 T246f28ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T1478f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1478f23ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T1477f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1477f10ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_class_type_with_type_marks */
extern T1 T1476f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_class_type_with_type_marks */
extern T1 T1450f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1450f21ot1(T0* a1);
/* [detachable] ET_LIKE_N.conforms_from_class_type_with_type_marks */
extern T1 T1424f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_class_type_with_type_marks */
extern T1 T867f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T867f19ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.conforms_from_class_type_with_type_marks */
extern T1 T442f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T246f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T246f30ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1478f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1478f25ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1477f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1477f12ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1476f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1450f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1450f22ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.recursive_formal_constraints */
extern T0* T867f31(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1424f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T867f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T867f20ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T442f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T246f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T246f31ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T1478f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1478f24ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T1477f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1477f11ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_tuple_type_with_type_marks */
extern T1 T1476f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_tuple_type_with_type_marks */
extern T1 T1450f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1450f23ot1(T0* a1);
/* [detachable] ET_LIKE_N.conforms_from_tuple_type_with_type_marks */
extern T1 T1424f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_tuple_type_with_type_marks */
extern T1 T867f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T867f21ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.conforms_from_tuple_type_with_type_marks */
extern T1 T442f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T246f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T246f32ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_expanded_with_type_mark */
extern T1 T1478f29(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f29ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_expanded_with_type_mark */
extern T1 T1477f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f22ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_expanded_with_type_mark */
extern T1 T1476f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_expanded_with_type_mark */
extern T1 T1450f24(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f24ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_expanded_with_type_mark */
extern T1 T1424f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_expanded_with_type_mark */
extern T1 T867f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f22ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.is_type_expanded_with_type_mark */
extern T1 T442f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_expanded_with_type_mark */
extern T1 T246f33(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f33ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_detachable_with_type_mark */
extern T1 T1478f30(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f30ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_detachable_with_type_mark */
extern T1 T1477f23(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f23ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_detachable_with_type_mark */
extern T1 T1476f33(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_detachable_with_type_mark */
extern T1 T1450f25(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f25ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_detachable_with_type_mark */
extern T1 T1424f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_detachable_with_type_mark */
extern T1 T867f23(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f23ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.is_type_detachable_with_type_mark */
extern T1 T442f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable_with_type_mark */
extern T1 T246f17(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f17ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_attached_with_type_mark */
extern T1 T1478f51(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f51ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_attached_with_type_mark */
extern T1 T1477f54(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f54ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_attached_with_type_mark */
extern T1 T1476f52(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_attached_with_type_mark */
extern T1 T1450f31(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f31ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_attached_with_type_mark */
extern T1 T1424f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_attached_with_type_mark */
extern T1 T867f32(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f32ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.is_type_attached_with_type_mark */
extern T1 T442f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached_with_type_mark */
extern T1 T246f16(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f16ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_reference_with_type_mark */
extern T1 T1478f52(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f52ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_reference_with_type_mark */
extern T1 T1477f55(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f55ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_reference_with_type_mark */
extern T1 T1476f53(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_reference_with_type_mark */
extern T1 T1450f35(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f35ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_reference_with_type_mark */
extern T1 T1424f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_reference_with_type_mark */
extern T1 T867f36(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f36ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.is_type_reference_with_type_mark */
extern T1 T442f48(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_reference_with_type_mark */
extern T1 T246f43(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f43ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_named_type_with_type_mark */
extern T0* T1478f46(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f46ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_named_type_with_type_mark */
extern T0* T1477f49(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f49ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_named_type_with_type_mark */
extern T0* T1476f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_named_type_with_type_mark */
extern T0* T1450f38(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f38ot1(T0* a1, T0** a2);
extern T1 T1450f38ot2(T0* a1, T0** a2);
extern T1 T1450f38ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.shallow_named_type_with_type_mark */
extern T0* T1424f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_named_type_with_type_mark */
extern T0* T867f56(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f56ot1(T0* a1, T0** a2);
extern T1 T867f56ot2(T0* a1, T0** a2);
extern T1 T867f56ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.shallow_named_type_with_type_mark */
extern T0* T442f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_named_type_with_type_mark */
extern T0* T246f45(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f45ot1(T0* a1, T0** a2);
extern T1 T246f45ot2(T0* a1, T0** a2);
extern T1 T246f45ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_type_with_type_marks */
extern T1 T1478f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_type_with_type_marks */
extern T1 T1477f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_type_with_type_marks */
extern T1 T1476f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_type_with_type_marks */
extern T1 T1450f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f44ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_type_with_type_marks */
extern T1 T867f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f45ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_type_with_type_marks */
extern T1 T442f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_type_with_type_marks */
extern T1 T248f86(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.is_generic */
extern T1 T248f24(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_type_with_type_marks */
extern T1 T246f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f53ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.same_syntactical_type_with_type_marks */
extern T1 T73f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_type_with_type_marks */
extern T1 T71f91(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T1478f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T1477f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_like_feature_with_type_marks */
extern T1 T1476f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_attachment_marks_with_default */
extern T1 T1476f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_like_feature_with_type_marks */
extern T1 T1450f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f45ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_syntactical_like_feature_with_type_marks */
extern T1 T1424f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_like_feature_with_type_marks */
extern T1 T867f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f46ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_like_feature_with_type_marks */
extern T1 T442f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T248f87(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T246f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f54ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_like_feature_with_type_marks */
extern T1 T73f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T71f92(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1478f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_attachment_marks_with_default */
extern T1 T1478f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1477f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_attachment_marks_with_default */
extern T1 T1477f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1476f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1450f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f46ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1424f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T867f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f47ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T442f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T248f88(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T246f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f55ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T73f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T71f93(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1478f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1477f67(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1476f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1450f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f49ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1424f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T867f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f50ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T442f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T248f93(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T246f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f61ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T73f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T71f105(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_base_type_with_type_mark */
extern T0* T1478f45(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f45ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_base_type_with_type_mark */
extern T0* T1477f48(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f48ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_base_type_with_type_mark */
extern T0* T1476f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_base_type_with_type_mark */
extern T0* T1450f56(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f56ot1(T0* a1, T0** a2);
extern T1 T1450f56ot2(T0* a1, T0** a2);
extern T1 T1450f56ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_base_type_with_type_mark */
extern T0* T867f55(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f55ot1(T0* a1, T0** a2);
extern T1 T867f55ot2(T0* a1, T0** a2);
extern T1 T867f55ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.shallow_base_type_with_type_mark */
extern T0* T442f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_base_type_with_type_mark */
extern T0* T246f44(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f44ot1(T0* a1, T0** a2);
extern T1 T246f44ot2(T0* a1, T0** a2);
extern T1 T246f44ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_with_type_mark */
extern T0* T1478f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_with_type_mark */
extern T0* T1477f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_right_brace */
extern void T1477f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_left_brace */
extern void T1477f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.type_with_type_mark */
extern T0* T1476f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.set_type_mark */
extern void T1476f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.twin */
extern T0* T1476f76(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.copy */
extern void T1476f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.type_with_type_mark */
extern T0* T1450f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.type_with_type_mark */
extern T0* T867f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.type_with_type_mark */
extern T0* T442f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.set_current_keyword */
extern void T442f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_with_type_mark */
extern T0* T246f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_unqualified_anchored_type */
extern T1 T1478f17(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_unqualified_anchored_type */
extern T1 T1477f19(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.has_unqualified_anchored_type */
extern T1 T1476f19(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_unqualified_anchored_type */
extern T1 T1450f62(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.has_unqualified_anchored_type */
extern T1 T867f62(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.has_unqualified_anchored_type */
extern T1 T442f33(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.has_unqualified_anchored_type */
extern T1 T248f67(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_unqualified_anchored_type */
extern T1 T246f24(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_unqualified_anchored_type */
extern T1 T73f204(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.has_unqualified_anchored_type */
extern T1 T71f66(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_formal_types */
extern T1 T1478f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_formal_types */
extern T1 T1477f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.has_formal_types */
extern T1 T1476f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_formal_types */
extern T1 T1450f52(GE_context* ac, T0* C, T0* a1);
extern T1 T1450f52ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.has_formal_types */
extern T1 T867f54(GE_context* ac, T0* C, T0* a1);
extern T1 T867f54ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.has_formal_types */
extern T1 T442f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.has_formal_types */
extern T1 T248f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_formal_types */
extern T1 T246f25(GE_context* ac, T0* C, T0* a1);
extern T1 T246f25ot1(T0* a1);
/* [detachable] ET_CLASS.has_formal_types */
extern T1 T73f205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.has_formal_types */
extern T1 T71f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.depends_on_qualified_anchored_type */
extern T1 T1478f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.depends_on_qualified_anchored_type */
extern T1 T1477f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.depends_on_qualified_anchored_type */
extern T1 T1476f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.depends_on_qualified_anchored_type */
extern T1 T1450f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.depends_on_qualified_anchored_type */
extern T1 T867f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.depends_on_qualified_anchored_type */
extern T1 T442f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.depends_on_qualified_anchored_type */
extern T1 T248f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.depends_on_qualified_anchored_type */
extern T1 T246f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.depends_on_qualified_anchored_type */
extern T1 T73f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.depends_on_qualified_anchored_type */
extern T1 T71f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_identifier_anchored_type */
extern T1 T1478f53(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_identifier_anchored_type */
extern T1 T1477f56(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.has_identifier_anchored_type */
extern T1 T1476f54(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_identifier_anchored_type */
extern T1 T1450f63(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.has_identifier_anchored_type */
extern T1 T867f63(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.has_identifier_anchored_type */
extern T1 T442f51(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.has_identifier_anchored_type */
extern T1 T248f84(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_identifier_anchored_type */
extern T1 T246f51(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_identifier_anchored_type */
extern T1 T73f218(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.has_identifier_anchored_type */
extern T1 T71f90(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters */
extern T0* T1478f62(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T1478f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters */
extern T0* T1477f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T1477f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters */
extern T0* T1476f63(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters_with_type_mark */
extern T0* T1476f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters */
extern T0* T1450f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters_with_type_mark */
extern T0* T1450f72(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f72ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters */
extern T0* T867f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters_with_type_mark */
extern T0* T867f74(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f74ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters */
extern T0* T442f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters_with_type_mark */
extern T0* T442f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters */
extern T0* T248f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T248f96(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.overridden_type_mark */
extern T0* T248f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_separate */
extern T1 T73f81(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_separate_mark */
extern T1 T73f93(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_separate */
extern T1 T248f21(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters */
extern T0* T246f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T246f62(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f62ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.resolved_formal_parameters */
extern T0* T73f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.resolved_formal_parameters_with_type_mark */
extern T0* T73f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.overridden_type_mark */
extern T0* T73f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters */
extern T0* T71f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T71f101(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.overridden_type_mark */
extern T0* T71f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_separate */
extern T1 T71f15(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_type */
extern T1 T1478f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_type */
extern T1 T1477f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_syntactical_type */
extern T1 T1476f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_type */
extern T1 T1450f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_type */
extern T1 T867f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_type */
extern T1 T442f52(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_type */
extern T1 T248f85(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_type */
extern T1 T246f52(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.same_syntactical_type */
extern T1 T73f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.same_syntactical_type */
extern T1 T71f87(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_expanded */
extern T1 T1478f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_expanded */
extern T1 T1477f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_expanded */
extern T1 T1476f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_expanded */
extern T1 T1450f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_expanded */
extern T1 T867f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_expanded */
extern T1 T442f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_expanded */
extern T1 T248f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_expanded */
extern T1 T246f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_expanded */
extern T1 T73f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_expanded */
extern T1 T71f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_named_type */
extern T1 T1478f47(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_named_type */
extern T1 T1477f50(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_named_type */
extern T1 T1476f48(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_named_type */
extern T1 T1450f64(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_named_type */
extern T1 T867f64(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.is_named_type */
extern T1 T442f44(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_named_type */
extern T1 T248f77(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_named_type */
extern T1 T246f47(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_named_type */
extern T1 T71f78(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_type */
extern T1 T1478f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_type */
extern T1 T1477f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_named_type */
extern T1 T1476f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_type */
extern T1 T1450f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER.same_named_type */
extern T1 T867f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.same_named_type */
extern T1 T442f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_type */
extern T1 T246f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_type */
extern T1 T1478f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_type */
extern T1 T1477f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_LIKE_FEATURE.conforms_to_type */
extern T1 T1476f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_type */
extern T1 T1450f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_type */
extern T1 T867f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_CURRENT.conforms_to_type */
extern T1 T442f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.conforms_to_type */
extern T1 T248f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_type */
extern T1 T246f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_to_type */
extern T1 T73f137(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.conforms_to_type */
extern T1 T71f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_type */
extern T0* T1478f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_type */
extern T0* T1477f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.named_type */
extern T0* T1476f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_type */
extern T0* T1450f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.named_type */
extern T0* T867f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.named_type */
extern T0* T442f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_type */
extern T0* T248f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_type */
extern T0* T246f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_type */
extern T0* T73f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_type */
extern T0* T71f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_like_current */
extern T1 T1478f44(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_like_current */
extern T1 T1477f47(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_like_current */
extern T1 T1476f45(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_like_current */
extern T1 T1450f55(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.is_like_current */
extern T1 T1424f30(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_like_current */
extern T1 T867f53(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_like_current */
extern T1 T248f73(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_like_current */
extern T1 T246f40(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_like_current */
extern T1 T73f200(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_like_current */
extern T1 T71f74(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T1478f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T1477f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_class_type_with_type_marks */
extern T1 T1476f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_class_type_with_type_marks */
extern T1 T1450f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f48ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_syntactical_class_type_with_type_marks */
extern T1 T1424f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_class_type_with_type_marks */
extern T1 T867f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f49ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_class_type_with_type_marks */
extern T1 T442f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T248f90(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T246f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f57ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_class_type_with_type_marks */
extern T1 T73f229(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.is_type_separate_with_type_mark */
extern T1 T71f33(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_separate_with_type_mark */
extern T1 T73f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T71f102(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1478f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1477f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1476f67(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_tuple_type_with_type_marks */
extern T1 T1450f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f50ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_syntactical_tuple_type_with_type_marks */
extern T1 T1424f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_tuple_type_with_type_marks */
extern T1 T867f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f65ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_tuple_type_with_type_marks */
extern T1 T442f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T248f95(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.is_type_separate_with_type_mark */
extern T1 T248f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T246f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f63ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_tuple_type_with_type_marks */
extern T1 T73f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T71f106(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T1478f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T1477f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_like_current_with_type_marks */
extern T1 T1476f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_like_current_with_type_marks */
extern T1 T1450f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f51ot1(T0* a1);
/* [detachable] ET_LIKE_N.same_syntactical_like_current_with_type_marks */
extern T1 T1424f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_like_current_with_type_marks */
extern T1 T867f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f66ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_like_current_with_type_marks */
extern T1 T442f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_attachment_marks_with_default */
extern T1 T442f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T248f99(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T246f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f65ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_like_current_with_type_marks */
extern T1 T73f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T71f111(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_base_type */
extern T1 T1478f67(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_base_type */
extern T1 T1477f70(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_base_type */
extern T1 T1476f68(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_base_type */
extern T1 T1450f73(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_base_type */
extern T1 T867f75(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.is_base_type */
extern T1 T442f63(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_base_type */
extern T1 T248f97(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_base_type */
extern T1 T246f64(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_base_type */
extern T1 T73f248(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_base_type */
extern T1 T71f107(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.to_text */
extern T0* T1478f16(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.to_text */
extern T0* T1477f41(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.to_text */
extern T0* T1476f42(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.to_text */
extern T0* T1450f26(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.to_text */
extern T0* T1424f12(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.append_to_string */
extern void T1424f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.to_text */
extern T0* T867f24(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.to_text */
extern T0* T442f38(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.to_text */
extern T0* T246f18(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_type_with_type_marks */
extern T1 T1478f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1478f7ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_type_with_type_marks */
extern T1 T1477f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1477f25ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_type_with_type_marks */
extern T1 T1476f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_type_with_type_marks */
extern T1 T1450f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f11ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_named_type_with_type_marks */
extern T1 T1424f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_named_type_with_type_marks */
extern T1 T867f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f9ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_named_type_with_type_marks */
extern T1 T442f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_type_with_type_marks */
extern T1 T246f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f5ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_base_type_with_type_marks */
extern T1 T1478f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1478f60ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_base_type_with_type_marks */
extern T1 T1477f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1477f63ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_base_type_with_type_marks */
extern T1 T1476f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_base_type_with_type_marks */
extern T1 T1450f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1450f47ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.same_base_formal_parameter_type_with_type_marks */
extern T1 T71f118(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_base_type_with_type_marks */
extern T1 T1424f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_base_type_with_type_marks */
extern T1 T867f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T867f48ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_base_type_with_type_marks */
extern T1 T442f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_base_type_with_type_marks */
extern T1 T246f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T246f56ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.adapted_base_class_with_named_feature */
extern T0* T1478f5(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f5ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.adapted_base_class_with_named_feature */
extern T0* T1477f38(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f38ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.adapted_base_class_with_named_feature */
extern T0* T1476f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.adapted_base_class_with_named_feature */
extern T0* T1450f33(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f33ot1(T0* a1, T0** a2);
extern T1 T1450f33ot2(T0* a1, T0** a2);
extern T1 T1450f33ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.adapted_base_class_with_named_feature */
extern T0* T1424f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.adapted_base_class_with_named_feature */
extern T0* T867f34(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f34ot1(T0* a1, T0** a2);
extern T1 T867f34ot2(T0* a1, T0** a2);
extern T1 T867f34ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.adapted_base_class_with_named_feature */
extern T0* T442f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.adapted_base_class_with_named_feature */
extern T0* T246f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f22ot1(T0* a1, T0** a2);
extern T1 T246f22ot2(T0* a1, T0** a2);
extern T1 T246f22ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T1478f6(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1478f6ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T1477f39(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1477f39ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.adapted_base_class_with_seeded_feature */
extern T0* T1476f40(GE_context* ac, T0* C, T6 a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.adapted_base_class_with_seeded_feature */
extern T0* T1450f34(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1450f34ot1(T0* a1, T0** a2);
extern T1 T1450f34ot2(T0* a1, T0** a2);
extern T1 T1450f34ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.adapted_base_class_with_seeded_feature */
extern T0* T1424f25(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.adapted_base_class_with_seeded_feature */
extern T0* T867f35(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T867f35ot1(T0* a1, T0** a2);
extern T1 T867f35ot2(T0* a1, T0** a2);
extern T1 T867f35ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.adapted_base_class_with_seeded_feature */
extern T0* T442f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T246f23(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T246f23ot1(T0* a1, T0** a2);
extern T1 T246f23ot2(T0* a1, T0** a2);
extern T1 T246f23ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_index_of_label */
extern T6 T1478f76(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1478f76ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_index_of_label */
extern T6 T1477f79(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1477f79ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.base_type_index_of_label */
extern T6 T1476f79(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_index_of_label */
extern T6 T1450f74(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1450f74ot1(T0* a1, T0** a2);
extern T1 T1450f74ot2(T0* a1, T0** a2);
extern T1 T1450f74ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.base_type_index_of_label */
extern T6 T1424f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_index_of_label */
extern T6 T867f76(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T867f76ot1(T0* a1, T0** a2);
extern T1 T867f76ot2(T0* a1, T0** a2);
extern T1 T867f76ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.base_type_index_of_label */
extern T6 T442f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_index_of_label */
extern T6 T246f71(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T246f71ot1(T0* a1, T0** a2);
extern T1 T246f71ot2(T0* a1, T0** a2);
extern T1 T246f71ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_attached */
extern T1 T1478f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_attached */
extern T1 T1477f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_attached */
extern T1 T1476f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_attached */
extern T1 T442f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_attached */
extern T1 T248f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached */
extern T1 T246f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_attached */
extern T1 T71f122(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_named_type */
extern T0* T1478f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_named_type */
extern T0* T1477f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_named_type */
extern T0* T1476f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_named_type */
extern T0* T1450f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.shallow_named_type */
extern T0* T1424f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.shallow_named_type */
extern T0* T867f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.shallow_named_type */
extern T0* T442f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.shallow_named_type */
extern T0* T248f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_named_type */
extern T0* T246f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.shallow_named_type */
extern T0* T73f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.shallow_named_type */
extern T0* T71f71(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_detachable */
extern T1 T1478f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_detachable */
extern T1 T1477f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_detachable */
extern T1 T1476f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_detachable */
extern T1 T1450f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_detachable */
extern T1 T867f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_detachable */
extern T1 T442f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_detachable */
extern T1 T248f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable */
extern T1 T246f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_detachable */
extern T1 T73f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_detachable */
extern T1 T71f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.position */
extern T0* T1478f41(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.position */
extern T0* T1477f34(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.position */
extern T0* T1476f28(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.position */
extern T0* T1450f57(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.position */
extern T0* T867f57(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.position */
extern T0* T442f32(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.position */
extern T0* T248f66(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.position */
extern T0* T246f21(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.position */
extern T0* T73f209(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.position */
extern T0* T1146f8(GE_context* ac, T0* C);
/* detachable ET_INDEXING.position */
extern T0* T1448f3(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.position */
extern T0* T1157f5(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.tokens */
extern T0* T1157f8(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.first */
extern T0* T1157f7(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.is_empty */
extern T1 T1157f6(GE_context* ac, T0* C);
/* detachable ET_TAGGED_INDEXING.position */
extern T0* T1447f4(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.first */
extern T0* T1146f10(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.is_empty */
extern T1 T1146f9(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.position */
extern T0* T71f65(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.preconditions */
extern T0* T367f48(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.preconditions */
extern T0* T365f51(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.preconditions */
extern T0* T363f50(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.postconditions */
extern T0* T367f58(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.postconditions */
extern T0* T365f63(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.postconditions */
extern T0* T363f62(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_deferred */
extern T1 T1461f45(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_deferred */
extern T1 T1460f43(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_deferred */
extern T1 T367f59(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_deferred */
extern T1 T365f64(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_deferred */
extern T1 T363f63(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_deferred */
extern T1 T329f60(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_deferred */
extern T1 T247f60(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_frozen */
extern T1 T1462f44(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_frozen */
extern T1 T1461f46(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_frozen */
extern T1 T1460f44(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_frozen */
extern T1 T367f23(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_frozen */
extern T1 T365f65(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_frozen */
extern T1 T363f64(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_frozen */
extern T1 T329f61(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_frozen */
extern T1 T247f61(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_bracketable */
extern T1 T1462f50(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_bracketable */
extern T1 T1461f50(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_bracketable */
extern T1 T1460f50(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_bracketable */
extern T1 T367f27(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_bracketable */
extern T1 T365f29(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_bracketable */
extern T1 T363f28(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_bracketable */
extern T1 T329f30(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_bracketable */
extern T1 T247f70(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_parenthesisable */
extern T1 T1462f51(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_parenthesisable */
extern T1 T1461f52(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_parenthesisable */
extern T1 T1460f51(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_parenthesisable */
extern T1 T367f28(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_parenthesisable */
extern T1 T365f30(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_parenthesisable */
extern T1 T363f29(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_parenthesisable */
extern T1 T329f31(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_parenthesisable */
extern T1 T247f71(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once */
extern T1 T1462f52(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once */
extern T1 T1460f52(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once */
extern T1 T367f29(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once */
extern T1 T365f31(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once */
extern T1 T363f30(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once */
extern T1 T329f32(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once */
extern T1 T247f72(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once_per_object */
extern T1 T1462f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once_per_object */
extern T1 T1461f54(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once_per_object */
extern T1 T1460f53(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once_per_object */
extern T1 T367f30(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once_per_object */
extern T1 T365f32(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once_per_object */
extern T1 T363f31(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once_per_object */
extern T1 T329f33(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once_per_object */
extern T1 T247f33(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_infixable */
extern T1 T1462f49(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_infixable */
extern T1 T1461f49(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_infixable */
extern T1 T1460f47(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_infixable */
extern T1 T367f26(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_infixable */
extern T1 T365f28(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_infixable */
extern T1 T363f27(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_infixable */
extern T1 T329f64(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_infixable */
extern T1 T247f69(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_prefixable */
extern T1 T1462f48(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_prefixable */
extern T1 T1461f48(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_prefixable */
extern T1 T1460f46(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_prefixable */
extern T1 T247f68(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.new_synonym */
extern T0* T1462f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_synonym */
extern void T1462f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.new_synonym */
extern T0* T1461f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_synonym */
extern void T1461f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.new_synonym */
extern T0* T1460f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_synonym */
extern void T1460f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.new_synonym */
extern T0* T367f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_synonym */
extern void T367f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.new_synonym */
extern T0* T365f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_synonym */
extern void T365f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.new_synonym */
extern T0* T363f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_synonym */
extern void T363f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.new_synonym */
extern T0* T329f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_synonym */
extern void T329f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.new_synonym */
extern T0* T247f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_synonym */
extern void T247f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.is_constant_attribute */
extern T1 T1462f38(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_constant_attribute */
extern T1 T1461f40(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_constant_attribute */
extern T1 T1460f38(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_constant_attribute */
extern T1 T367f52(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_constant_attribute */
extern T1 T365f55(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_constant_attribute */
extern T1 T329f53(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_constant_attribute */
extern T1 T247f54(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_unique_attribute */
extern T1 T1462f39(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_unique_attribute */
extern T1 T1461f41(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_unique_attribute */
extern T1 T1460f39(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_unique_attribute */
extern T1 T367f55(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_unique_attribute */
extern T1 T363f55(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_unique_attribute */
extern T1 T329f54(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_unique_attribute */
extern T1 T247f55(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_seed */
extern T1 T1462f41(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.has_seed */
extern T1 T1461f43(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.has_seed */
extern T1 T1460f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.has_seed */
extern T1 T367f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_seed */
extern T1 T365f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_seed */
extern T1 T363f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_seed */
extern T1 T329f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.has_seed */
extern T1 T247f57(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.arguments_count */
extern T6 T1462f42(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.arguments_count */
extern T6 T1461f44(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.arguments_count */
extern T6 T1460f42(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.arguments_count */
extern T6 T367f45(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.arguments_count */
extern T6 T365f47(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.arguments_count */
extern T6 T363f49(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.arguments_count */
extern T6 T329f48(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.arguments_count */
extern T6 T247f53(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_static */
extern T1 T1462f37(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_static_mark */
extern T1 T1462f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_static */
extern T1 T1461f39(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.has_static_mark */
extern T1 T1461f34(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_static */
extern T1 T1460f37(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.has_static_mark */
extern T1 T1460f34(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_static */
extern T1 T367f51(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.has_static_mark */
extern T1 T367f46(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_static */
extern T1 T365f54(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_static */
extern T1 T365f54p1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_static_mark */
extern T1 T365f49(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_static */
extern T1 T363f53(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_static */
extern T1 T363f53p1(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_static_mark */
extern T1 T363f48(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_static */
extern T1 T329f52(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_static_mark */
extern T1 T329f49(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_static */
extern T1 T247f50(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_static */
extern T1 T247f50p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.has_static_mark */
extern T1 T247f49(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_exported_to */
extern T1 T1462f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_FUNCTION.is_exported_to */
extern T1 T1461f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DO_FUNCTION.is_exported_to */
extern T1 T1460f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTRIBUTE.is_exported_to */
extern T1 T367f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_exported_to */
extern T1 T365f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_exported_to */
extern T1 T363f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_exported_to */
extern T1 T329f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTERNAL_FUNCTION.is_exported_to */
extern T1 T247f84(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTRIBUTE.arguments */
extern T0* T367f56(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.arguments */
extern T0* T365f57(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.arguments */
extern T0* T363f60(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.arguments */
extern T0* T329f55(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.name */
extern T0* T1462f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.name */
extern T0* T1461f55(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.name */
extern T0* T1460f54(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.name */
extern T0* T367f47(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.name */
extern T0* T365f50(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.name */
extern T0* T363f51(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.name */
extern T0* T329f50(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.name */
extern T0* T247f56(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_name */
extern T0* T1462f76(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_extended_name */
extern T0* T1462f78(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_name */
extern T0* T1461f79(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_extended_name */
extern T0* T1461f81(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_name */
extern T0* T1460f77(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_extended_name */
extern T0* T1460f79(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_name */
extern T0* T367f73(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_extended_name */
extern T0* T367f75(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_name */
extern T0* T365f77(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_extended_name */
extern T0* T365f79(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_name */
extern T0* T363f76(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_extended_name */
extern T0* T363f78(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_name */
extern T0* T329f76(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_extended_name */
extern T0* T329f78(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_name */
extern T0* T247f85(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_extended_name */
extern T0* T247f87(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.alias_name */
extern T0* T1462f45(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.alias_name */
extern T0* T1461f47(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.alias_name */
extern T0* T1460f45(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.alias_name */
extern T0* T367f24(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.alias_name */
extern T0* T365f26(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.alias_name */
extern T0* T363f25(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.alias_name */
extern T0* T329f62(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.alias_name */
extern T0* T247f62(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_alias_name */
extern T0* T1462f77(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_alias_name */
extern T0* T1461f80(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_alias_name */
extern T0* T1460f78(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_alias_name */
extern T0* T367f74(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_alias_name */
extern T0* T365f78(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_alias_name */
extern T0* T363f77(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_alias_name */
extern T0* T329f77(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_alias_name */
extern T0* T247f86(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_attribute */
extern T1 T1462f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_attribute */
extern T1 T1461f33(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_attribute */
extern T1 T1460f33(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_attribute */
extern T1 T365f52(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_attribute */
extern T1 T363f32(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_attribute */
extern T1 T247f59(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.type */
extern T0* T1462f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.type */
extern T0* T1461f37(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.type */
extern T0* T1460f36(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.type */
extern T0* T367f50(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.type */
extern T0* T365f53(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.type */
extern T0* T363f52(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.type */
extern T0* T329f51(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.type */
extern T0* T247f47(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.lower_name */
extern T0* T1462f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.lower_name */
extern T0* T1461f35(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.lower_name */
extern T0* T1460f35(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.lower_name */
extern T0* T367f31(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.lower_name */
extern T0* T365f33(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.lower_name */
extern T0* T363f46(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.lower_name */
extern T0* T329f34(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.lower_name */
extern T0* T247f34(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_deferred */
extern T1 T1465f42(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_deferred */
extern T1 T1464f42(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_deferred */
extern T1 T466f36(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_frozen */
extern T1 T1465f50(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_frozen */
extern T1 T1464f43(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_frozen */
extern T1 T468f34(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_frozen */
extern T1 T466f37(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_parenthesisable */
extern T1 T1465f51(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_parenthesisable */
extern T1 T1464f45(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_parenthesisable */
extern T1 T468f35(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_parenthesisable */
extern T1 T466f38(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.has_seed */
extern T1 T1465f39(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.has_seed */
extern T1 T1464f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.has_seed */
extern T1 T468f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.has_seed */
extern T1 T466f34(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.new_synonym */
extern T0* T1465f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_synonym */
extern void T1465f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.new_synonym */
extern T0* T1464f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_synonym */
extern void T1464f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.new_synonym */
extern T0* T468f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_synonym */
extern void T468f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.new_synonym */
extern T0* T466f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_synonym */
extern void T466f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_creation_exported_to */
extern T1 T1465f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.is_creation_exported_to */
extern T1 T73f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR_LIST.is_exported_to */
extern T1 T1080f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR.is_exported_to */
extern T1 T1077f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE.is_creation_exported_to */
extern T1 T1464f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DEFERRED_PROCEDURE.is_creation_exported_to */
extern T1 T468f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DO_PROCEDURE.is_creation_exported_to */
extern T1 T466f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE.lower_name */
extern T0* T1465f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.lower_name */
extern T0* T1464f31(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.lower_name */
extern T0* T468f44(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.lower_name */
extern T0* T466f42(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.name */
extern T0* T1465f38(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.name */
extern T0* T1464f41(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.name */
extern T0* T468f40(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.name */
extern T0* T466f40(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_name */
extern T0* T1465f83(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_extended_name */
extern T0* T1465f85(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_name */
extern T0* T1464f78(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_extended_name */
extern T0* T1464f80(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_name */
extern T0* T468f74(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_extended_name */
extern T0* T468f76(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_name */
extern T0* T466f75(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_extended_name */
extern T0* T466f77(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.alias_name */
extern T0* T1465f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.alias_name */
extern T0* T1464f40(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.alias_name */
extern T0* T468f32(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.alias_name */
extern T0* T466f35(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_alias_name */
extern T0* T1465f84(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_alias_name */
extern T0* T1464f79(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_alias_name */
extern T0* T468f75(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_alias_name */
extern T0* T466f76(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.arguments_count */
extern T6 T1465f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.arguments_count */
extern T6 T1464f33(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.arguments_count */
extern T6 T468f43(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.arguments_count */
extern T6 T466f41(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_exported_to */
extern T1 T1465f82(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE.is_exported_to */
extern T1 T1464f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DEFERRED_PROCEDURE.is_exported_to */
extern T1 T468f73(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DO_PROCEDURE.is_exported_to */
extern T1 T466f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.is_static */
extern T1 T1465f33(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_static */
extern T1 T1465f33p1(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.has_static_mark */
extern T1 T1465f54(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_static */
extern T1 T1464f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.has_static_mark */
extern T1 T1464f47(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_static */
extern T1 T468f42(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.has_static_mark */
extern T1 T468f45(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_static */
extern T1 T466f43(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.has_static_mark */
extern T1 T466f46(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_infix */
extern T1 T1467f9(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_infix */
extern T1 T1466f9(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix */
extern T1 T180f49(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_prefix */
extern T1 T1468f8(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_prefix */
extern T1 T864f15(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix */
extern T1 T180f48(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.first_leaf */
extern T0* T1468f82(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.first_leaf */
extern T0* T1467f66(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.first_leaf */
extern T0* T1466f82(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.first_leaf */
extern T0* T864f66(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.first_leaf */
extern T0* T180f88(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.hash_code */
extern T6 T1468f44(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.hash_code */
extern T6 T1466f30(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_argument */
extern T1 T1468f62(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_argument */
extern T1 T1467f51(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_argument */
extern T1 T1466f48(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_argument */
extern T1 T864f51(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_argument */
extern T1 T180f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.argument_name */
extern T0* T1468f63(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.argument_name */
extern T0* T1467f52(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.argument_name */
extern T0* T1466f49(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.argument_name */
extern T0* T864f52(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.argument_name */
extern T0* T180f66(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_local */
extern T1 T1468f64(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_local */
extern T1 T1467f53(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_local */
extern T1 T1466f50(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_local */
extern T1 T864f53(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_local */
extern T1 T180f29(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.local_name */
extern T0* T1468f65(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.local_name */
extern T0* T1467f54(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.local_name */
extern T0* T1466f51(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.local_name */
extern T0* T864f54(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.local_name */
extern T0* T180f67(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_object_test_local */
extern T1 T1468f66(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_object_test_local */
extern T1 T1467f55(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_object_test_local */
extern T1 T1466f52(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_object_test_local */
extern T1 T864f55(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_object_test_local */
extern T1 T180f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.object_test_local_name */
extern T0* T1468f67(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.object_test_local_name */
extern T0* T1467f56(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.object_test_local_name */
extern T0* T1466f53(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.object_test_local_name */
extern T0* T864f56(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.object_test_local_name */
extern T0* T180f68(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_across_cursor */
extern T1 T1468f68(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_across_cursor */
extern T1 T1467f57(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_across_cursor */
extern T1 T1466f54(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_across_cursor */
extern T1 T864f57(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_across_cursor */
extern T1 T180f27(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.across_cursor_name */
extern T0* T1468f69(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.across_cursor_name */
extern T0* T1467f58(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.across_cursor_name */
extern T0* T1466f55(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.across_cursor_name */
extern T0* T864f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.across_cursor_name */
extern T0* T180f69(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.is_tuple_label */
extern T1 T1468f11(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.is_tuple_label */
extern T1 T1467f11(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.is_tuple_label */
extern T1 T1466f11(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.is_tuple_label */
extern T1 T864f18(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_tuple_label */
extern T1 T180f37(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.same_feature_name */
extern T1 T1468f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_FREE_NAME.same_feature_name */
extern T1 T1467f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_NAME.same_feature_name */
extern T1 T1466f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_FREE_NAME.same_feature_name */
extern T1 T864f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.same_feature_name */
extern T1 T180f36(GE_context* ac, T0* C, T0* a1);
extern T1 T180f36ot1(T0* a1, T0** a2);
/* detachable ET_INFIX_NAME.position */
extern T0* T1468f7(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.position */
extern T0* T1467f7(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.position */
extern T0* T1466f7(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.position */
extern T0* T864f14(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.position */
extern T0* T180f30(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.lower_name */
extern T0* T1468f43(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.lower_name */
extern T0* T1467f22(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.lower_name */
extern T0* T1466f29(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.lower_name */
extern T0* T864f12(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.lower_name */
extern T0* T180f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.name */
extern T0* T1468f23(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.name */
extern T0* T1467f21(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.name */
extern T0* T1466f23(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.name */
extern T0* T864f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_NAME.last_leaf */
extern T0* T1468f84(GE_context* ac, T0* C);
/* detachable ET_PREFIX_FREE_NAME.last_leaf */
extern T0* T1467f68(GE_context* ac, T0* C);
/* detachable ET_PREFIX_NAME.last_leaf */
extern T0* T1466f84(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_FREE_NAME.last_leaf */
extern T0* T864f68(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.last_leaf */
extern T0* T180f94(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_tuple */
extern T1 T444f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_tuple */
extern T1 T443f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_tuple */
extern T1 T311f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_tuple */
extern T1 T219f47(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.type_mark */
extern T0* T444f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.tokens */
extern T0* T444f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.type_mark */
extern T0* T443f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.tokens */
extern T0* T443f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.type_mark */
extern T0* T311f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.tokens */
extern T0* T311f26(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.type_mark */
extern T0* T294f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.tokens */
extern T0* T294f26(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_mark */
extern T0* T219f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.tokens */
extern T0* T219f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_nested_reference_attributes */
extern T1 T444f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_nested_reference_attributes */
extern T1 T443f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_reference_attributes */
extern T1 T311f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_reference_attributes */
extern T1 T294f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_attributes */
extern T1 T219f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_generic */
extern T1 T444f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_generic */
extern T1 T443f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_generic */
extern T1 T311f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_generic */
extern T1 T294f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_generic */
extern T1 T219f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.primary_type */
extern T0* T444f35(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.primary_type */
extern T0* T443f35(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.primary_type */
extern T0* T311f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.primary_type */
extern T0* T294f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.primary_type */
extern T0* T219f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_builtin_attribute */
extern T1 T444f34(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_builtin_attribute */
extern T1 T443f34(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_builtin_attribute */
extern T1 T311f37(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_builtin_attribute */
extern T1 T294f37(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_builtin_attribute */
extern T1 T219f42(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_special */
extern T1 T444f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_special */
extern T1 T443f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_special */
extern T1 T294f47(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_special */
extern T1 T219f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.seeded_dynamic_procedure */
extern T0* T444f49(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.force_last */
extern void T220f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.resize */
extern void T220f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.new_capacity */
extern T6 T220f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.capacity */
extern T6 T220f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.register_feature_seeds */
extern void T444f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].force_last */
extern void T592f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T592f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_put */
extern void T592f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T592f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T592f23(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T592f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].resize */
extern void T592f39(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T592f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T592f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T592f46(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T592f45(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T592f28(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T592f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T592f44(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T592f27(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T592f32(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T592f38(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].search_position */
extern void T592f37(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T592f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_procedure */
extern T0* T444f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].value */
extern T0* T592f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T592f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].position_of_key */
extern T6 T592f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T592c33(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T592f35(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T592f43(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T592f52(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T592f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make */
extern T0* T1358c4(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_slots */
extern void T592f51(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T592f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T592f49(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T592f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.make_with_capacity */
extern T0* T220c10(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.seeded_dynamic_procedure */
extern T0* T443f55(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.register_feature_seeds */
extern void T443f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_procedure */
extern T0* T443f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_agent_call */
extern void T226f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_agent_call */
extern void T70f129(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.make */
extern T0* T1416c15(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_procedure */
extern T0* T443f44p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.seeded_dynamic_procedure */
extern T0* T311f54(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.register_feature_seeds */
extern void T311f62(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_procedure */
extern T0* T311f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_procedure */
extern T0* T311f24p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.seeded_dynamic_procedure */
extern T0* T294f54(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.register_feature_seeds */
extern void T294f62(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_procedure */
extern T0* T294f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_tuple_put */
extern void T226f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_tuple_put */
extern void T70f126(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_procedure */
extern T0* T294f24p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_procedure */
extern T0* T219f28(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.register_feature_seeds */
extern void T219f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.new_dynamic_procedure */
extern T0* T219f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_query */
extern T0* T444f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_attribute */
extern void T444f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_query */
extern T0* T444f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.set_result_type_set */
extern void T232f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_query */
extern T0* T444f46p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_query */
extern T0* T443f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_attribute */
extern void T443f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_query */
extern T0* T443f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_query */
extern T0* T311f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_attribute */
extern void T311f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_query */
extern T0* T311f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_query */
extern T0* T311f30p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_query */
extern T0* T294f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_attribute */
extern void T294f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_query */
extern T0* T294f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_tuple_item */
extern void T226f25(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_tuple_item */
extern void T70f127(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.infix_at_feature_name */
extern T0* T72f92(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_query */
extern T0* T294f30p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_query */
extern T0* T219f40(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_attribute */
extern void T219f62(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.new_dynamic_query */
extern T0* T219f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_agent_type */
extern T1 T311f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_agent_type */
extern T1 T294f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_agent_type */
extern T1 T219f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_procedure */
extern T0* T444f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_procedure */
extern T0* T443f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_procedure */
extern T0* T311f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_procedure */
extern T0* T294f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_procedure */
extern T0* T219f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_expanded */
extern T1 T444f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_expanded */
extern T1 T443f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_expanded */
extern T1 T311f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_expanded */
extern T1 T294f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_expanded */
extern T1 T219f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_basic */
extern T1 T444f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_basic */
extern T1 T443f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_basic */
extern T1 T311f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_basic */
extern T1 T294f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_basic */
extern T1 T219f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.seeded_dynamic_query */
extern T0* T444f30(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.seeded_dynamic_query */
extern T0* T443f30(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.seeded_dynamic_query */
extern T0* T311f34(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.seeded_dynamic_query */
extern T0* T294f33(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_query */
extern T0* T219f41(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.conforms_to_primary_type */
extern T1 T444f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T266f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.conforms_to_primary_type */
extern T1 T443f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.conforms_to_primary_type */
extern T1 T311f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.conforms_to_primary_type */
extern T1 T294f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_primary_type */
extern T1 T219f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_null_position */
extern T1 T1967f32(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_null_position */
extern T1 T1445f8(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_null_position */
extern T1 T1444f8(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_null_position */
extern T1 T1443f7(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_null_position */
extern T1 T1442f7(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_null_position */
extern T1 T1440f9(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_null_position */
extern T1 T1439f9(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_null_position */
extern T1 T1438f31(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_null_position */
extern T1 T1437f31(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_null_position */
extern T1 T1436f31(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_null_position */
extern T1 T1435f31(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_null_position */
extern T1 T1434f17(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_null_position */
extern T1 T1433f17(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_null_position */
extern T1 T1432f14(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_null_position */
extern T1 T983f32(GE_context* ac, T0* C);
/* detachable ET_VOID.is_null_position */
extern T1 T975f7(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_null_position */
extern T1 T972f18(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.is_null_position */
extern T1 T969f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_null_position */
extern T1 T959f6(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_null_position */
extern T1 T951f21(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_null_position */
extern T1 T933f13(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.is_null_position */
extern T1 T930f12(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.is_null_position */
extern T1 T863f10(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_null_position */
extern T1 T610f4(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_null_position */
extern T1 T409f32(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_null_position */
extern T1 T366f34(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_null_position */
extern T1 T343f20(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_null_position */
extern T1 T301f12(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_null_position */
extern T1 T273f12(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.is_null */
extern T1 T189f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_null_position */
extern T1 T180f47(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.to_text */
extern T0* T1515f10(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.append_to_string */
extern void T1515f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FILE_POSITION.append_to_string_with_filename */
extern void T1515f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.append_context_to_string */
extern void T1515f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.append_line_to_string */
extern void T1515f19(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.skip_lines */
extern void T1515f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FILE_POSITION.integer_ */
extern T0* T1515f11(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.position_to_text */
extern T0* T1445f34(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.append_position_to_string */
extern void T1445f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.integer_ */
extern T0* T1445f36(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.position_to_text */
extern T0* T1444f33(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.append_position_to_string */
extern void T1444f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.integer_ */
extern T0* T1444f35(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.position_to_text */
extern T0* T1443f29(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.append_position_to_string */
extern void T1443f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.integer_ */
extern T0* T1443f32(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.position_to_text */
extern T0* T1442f29(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.append_position_to_string */
extern void T1442f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.integer_ */
extern T0* T1442f32(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.position_to_text */
extern T0* T1440f31(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.append_position_to_string */
extern void T1440f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.integer_ */
extern T0* T1440f34(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.position_to_text */
extern T0* T1439f31(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.append_position_to_string */
extern void T1439f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.integer_ */
extern T0* T1439f34(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.position_to_text */
extern T0* T1438f48(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.append_position_to_string */
extern void T1438f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_ */
extern T0* T1438f51(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.position_to_text */
extern T0* T1437f48(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.append_position_to_string */
extern void T1437f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_ */
extern T0* T1437f51(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.position_to_text */
extern T0* T1436f48(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.append_position_to_string */
extern void T1436f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_ */
extern T0* T1436f51(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.position_to_text */
extern T0* T1435f48(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.append_position_to_string */
extern void T1435f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_ */
extern T0* T1435f51(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.position_to_text */
extern T0* T1434f40(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.append_position_to_string */
extern void T1434f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.integer_ */
extern T0* T1434f44(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.position_to_text */
extern T0* T1433f36(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.append_position_to_string */
extern void T1433f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.integer_ */
extern T0* T1433f40(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.position_to_text */
extern T0* T1432f34(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.append_position_to_string */
extern void T1432f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.integer_ */
extern T0* T1432f36(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.position_to_text */
extern T0* T983f38(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.append_position_to_string */
extern void T983f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.integer_ */
extern T0* T983f40(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.position_to_text */
extern T0* T972f56(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.append_position_to_string */
extern void T972f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.integer_ */
extern T0* T972f58(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.position_to_text */
extern T0* T863f36(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.append_position_to_string */
extern void T863f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_MANIFEST_STRING.integer_ */
extern T0* T863f40(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.position_to_text */
extern T0* T610f66(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.append_position_to_string */
extern void T610f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.integer_ */
extern T0* T610f68(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.position_to_text */
extern T0* T409f40(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.append_position_to_string */
extern void T409f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.integer_ */
extern T0* T409f41(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.position_to_text */
extern T0* T366f49(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.append_position_to_string */
extern void T366f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_ */
extern T0* T366f52(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.position_to_text */
extern T0* T343f38(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.append_position_to_string */
extern void T343f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.integer_ */
extern T0* T343f39(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.position_to_text */
extern T0* T273f26(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.append_position_to_string */
extern void T273f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.integer_ */
extern T0* T273f27(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.to_text */
extern T0* T189f9(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.append_to_string */
extern void T189f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_COMPRESSED_POSITION.integer_ */
extern T0* T189f10(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.position_to_text */
extern T0* T180f90(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.append_position_to_string */
extern void T180f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.integer_ */
extern T0* T180f92(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.line */
extern T6 T1967f30(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.line */
extern T6 T1445f6(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.line */
extern T6 T1444f6(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.line */
extern T6 T1443f5(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.line */
extern T6 T1442f5(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.line */
extern T6 T1440f7(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.line */
extern T6 T1439f7(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.line */
extern T6 T1438f29(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.line */
extern T6 T1437f29(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.line */
extern T6 T1436f29(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.line */
extern T6 T1435f29(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.line */
extern T6 T1434f15(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.line */
extern T6 T1433f15(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.line */
extern T6 T1432f12(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.line */
extern T6 T983f33(GE_context* ac, T0* C);
/* detachable ET_VOID.line */
extern T6 T975f5(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.line */
extern T6 T972f16(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.line */
extern T6 T969f10(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.line */
extern T6 T959f11(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.line */
extern T6 T951f19(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.line */
extern T6 T933f11(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.line */
extern T6 T930f10(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.line */
extern T6 T863f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.line */
extern T6 T610f7(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.line */
extern T6 T409f33(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.line */
extern T6 T366f32(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.line */
extern T6 T343f21(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.line */
extern T6 T301f10(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.line */
extern T6 T273f10(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.line */
extern T6 T189f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.line */
extern T6 T180f34(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.column */
extern T6 T1967f31(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.column */
extern T6 T1445f7(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.column */
extern T6 T1444f7(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.column */
extern T6 T1443f6(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.column */
extern T6 T1442f6(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.column */
extern T6 T1440f8(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.column */
extern T6 T1439f8(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.column */
extern T6 T1438f30(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.column */
extern T6 T1437f30(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.column */
extern T6 T1436f30(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.column */
extern T6 T1435f30(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.column */
extern T6 T1434f16(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.column */
extern T6 T1433f16(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.column */
extern T6 T1432f13(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.column */
extern T6 T983f34(GE_context* ac, T0* C);
/* detachable ET_VOID.column */
extern T6 T975f6(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.column */
extern T6 T972f17(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.column */
extern T6 T969f11(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.column */
extern T6 T959f12(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.column */
extern T6 T951f20(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.column */
extern T6 T933f12(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.column */
extern T6 T930f11(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_MANIFEST_STRING.column */
extern T6 T863f9(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.column */
extern T6 T610f38(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.column */
extern T6 T409f34(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.column */
extern T6 T366f33(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.column */
extern T6 T343f22(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.column */
extern T6 T301f11(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.column */
extern T6 T273f11(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.column */
extern T6 T189f7(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.column */
extern T6 T180f35(GE_context* ac, T0* C);
/* [detachable] UC_STRING_EQUALITY_TESTER.test */
extern T1 T871f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UC_STRING_EQUALITY_TESTER.string_ */
extern T0* T871f2(GE_context* ac, T0* C);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.test */
extern T1 T171f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.string_ */
extern T0* T171f2(GE_context* ac, T0* C);
/* detachable KL_EQUALITY_TESTER [[attached] STRING_8].test */
extern T1 T150f1(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T150f1oe1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] UC_UTF8_STRING.is_equal */
extern T1 T743f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_EQUALITY_TESTER.test */
extern T1 T149f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.default_message */
extern T0* T2103f5(GE_context* ac, T0* C);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.message */
extern T0* T2103f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.arguments */
extern T0* T2103f6(GE_context* ac, T0* C);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.string_ */
extern T0* T2103f4(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.default_message */
extern T0* T2082f3(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.message */
extern T0* T2082f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.arguments */
extern T0* T2082f7(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.string_ */
extern T0* T2082f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.default_message */
extern T0* T2081f3(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.message */
extern T0* T2081f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.arguments */
extern T0* T2081f7(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.string_ */
extern T0* T2081f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.default_message */
extern T0* T2080f3(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.message */
extern T0* T2080f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_REJECT_ERROR.arguments */
extern T0* T2080f7(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.string_ */
extern T0* T2080f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.default_message */
extern T0* T2079f3(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.message */
extern T0* T2079f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_META_ERROR.arguments */
extern T0* T2079f7(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.string_ */
extern T0* T2079f6(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.default_message */
extern T0* T2078f3(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.message */
extern T0* T2078f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_ERROR.arguments */
extern T0* T2078f7(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.string_ */
extern T0* T2078f6(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.default_message */
extern T0* T2077f3(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.message */
extern T0* T2077f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.arguments */
extern T0* T2077f7(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.string_ */
extern T0* T2077f6(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.default_message */
extern T0* T2076f3(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.message */
extern T0* T2076f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_MISSING_QUOTE_ERROR.arguments */
extern T0* T2076f7(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.string_ */
extern T0* T2076f6(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.default_message */
extern T0* T2075f3(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.message */
extern T0* T2075f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.arguments */
extern T0* T2075f7(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.string_ */
extern T0* T2075f6(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.default_message */
extern T0* T2074f3(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.message */
extern T0* T2074f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.arguments */
extern T0* T2074f7(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.string_ */
extern T0* T2074f6(GE_context* ac, T0* C);
/* [detachable] ET_UNIVERSE_ERROR.default_message */
extern T0* T1945f12(GE_context* ac, T0* C);
/* [detachable] ET_UNIVERSE_ERROR.message */
extern T0* T1945f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.arguments */
extern T0* T1945f6(GE_context* ac, T0* C);
/* [detachable] ET_UNIVERSE_ERROR.string_ */
extern T0* T1945f9(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.default_message */
extern T0* T1889f12(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.message */
extern T0* T1889f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.arguments */
extern T0* T1889f6(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.string_ */
extern T0* T1889f9(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.default_message */
extern T0* T1888f11(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.message */
extern T0* T1888f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYNTAX_ERROR.arguments */
extern T0* T1888f12(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.string_ */
extern T0* T1888f16(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.default_message */
extern T0* T1887f12(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.message */
extern T0* T1887f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTER_ERROR.arguments */
extern T0* T1887f6(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.string_ */
extern T0* T1887f9(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.default_message */
extern T0* T1309f5(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.message */
extern T0* T1309f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.arguments */
extern T0* T1309f6(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.string_ */
extern T0* T1309f4(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.default_message */
extern T0* T1308f5(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.message */
extern T0* T1308f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_SYNTAX_ERROR.arguments */
extern T0* T1308f6(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.string_ */
extern T0* T1308f4(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.default_message */
extern T0* T759f38(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.message */
extern T0* T759f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.arguments */
extern T0* T759f13(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.string_ */
extern T0* T759f37(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.default_message */
extern T0* T758f5(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.message */
extern T0* T758f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERNAL_ERROR.arguments */
extern T0* T758f6(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.string_ */
extern T0* T758f9(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.default_message */
extern T0* T757f5(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.message */
extern T0* T757f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.arguments */
extern T0* T757f6(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.string_ */
extern T0* T757f9(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.default_message */
extern T0* T297f5(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.message */
extern T0* T297f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.arguments */
extern T0* T297f6(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.string_ */
extern T0* T297f4(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.default_message */
extern T0* T88f17(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.message */
extern T0* T88f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.arguments */
extern T0* T88f20(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.string_ */
extern T0* T88f19(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.default_message */
extern T0* T82f3(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.message */
extern T0* T82f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_MESSAGE.arguments */
extern T0* T82f7(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.string_ */
extern T0* T82f6(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.default_message */
extern T0* T81f3(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.message */
extern T0* T81f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION_NUMBER.arguments */
extern T0* T81f7(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.string_ */
extern T0* T81f6(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.default_message */
extern T0* T79f3(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.message */
extern T0* T79f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.arguments */
extern T0* T79f7(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.string_ */
extern T0* T79f6(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.character_32_item */
extern T3 T743f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.item */
extern T3 T18f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.character_32_item */
extern T3 T17f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.substring */
extern T0* T743f12(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.make_from_substring */
extern T0* T743c74(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.make_from_substring_general */
extern void T743f93(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_32.substring */
extern T0* T18f15(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] STRING_32.new_string */
extern T0* T18f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.substring */
extern T0* T17f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.code */
extern T10 T743f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.code */
extern T10 T18f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.code */
extern T10 T17f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.runtime_name_to_text */
extern T0* T248f20(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.runtime_name_to_text */
extern T0* T73f189(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.runtime_name_to_text */
extern T0* T71f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_as_base_class */
extern T1 T248f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.same_as_base_class */
extern T1 T73f57(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.same_as_base_class */
extern T1 T71f12(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.upper_name */
extern T0* T248f106(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.upper_name */
extern T0* T73f174(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.upper_name */
extern T0* T71f80(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_named_type_with_type_marks */
extern T1 T248f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_named_type_with_type_marks */
extern T1 T73f76(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_type_with_type_marks */
extern T1 T71f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.named_type_with_type_mark */
extern T0* T248f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.named_type_with_type_mark */
extern T0* T73f145(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.named_type_with_type_mark */
extern T0* T71f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.base_type_actual_count */
extern T6 T248f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.base_type_actual_count */
extern T6 T73f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.base_type_actual_count */
extern T6 T71f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.same_named_class_type_with_type_marks */
extern T1 T248f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_named_class_type_with_type_marks */
extern T1 T73f138(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_class_type_with_type_marks */
extern T1 T71f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T248f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_named_tuple_type_with_type_marks */
extern T1 T73f139(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T71f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T248f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_named_formal_parameter_type_with_type_marks */
extern T1 T73f155(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T71f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.unaliased_to_text */
extern T0* T248f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unaliased_to_text */
extern T0* T73f111(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.unaliased_to_text */
extern T0* T71f19(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_named_type */
extern T1 T248f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.same_named_type */
extern T1 T73f185(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.same_named_type */
extern T1 T71f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.shallow_base_type_with_type_mark */
extern T0* T248f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.shallow_base_type_with_type_mark */
extern T0* T73f202(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.shallow_base_type_with_type_mark */
extern T0* T71f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.base_type_with_type_mark */
extern T0* T248f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.base_type_with_type_mark */
extern T0* T73f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.base_type_with_type_mark */
extern T0* T71f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T248f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_from_class_type_with_type_marks */
extern T1 T73f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforming_ancestor */
extern T0* T73f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T71f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T248f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T73f190(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T71f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T248f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.tuple_conforms_to_types */
extern T1 T1902f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.tuple_conforms_to_types */
extern T1 T236f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_from_tuple_type_with_type_marks */
extern T1 T73f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T71f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.is_type_expanded_with_type_mark */
extern T1 T248f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_expanded_with_type_mark */
extern T1 T73f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_expanded_with_type_mark */
extern T1 T71f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_type_detachable_with_type_mark */
extern T1 T248f48(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_detachable_with_type_mark */
extern T1 T73f129(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_detachable_with_type_mark */
extern T1 T71f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_attached */
extern T1 T248f22(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_attached */
extern T1 T73f75(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_attached */
extern T1 T71f16(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_expanded */
extern T1 T73f170(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_expanded_mark */
extern T1 T73f74(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_expanded */
extern T1 T71f14(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T248f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_adapted_base_class_with_named_feature */
extern T0* T73f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T71f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T248f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.context_adapted_base_class_with_seeded_feature */
extern T0* T73f214(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T71f84(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.is_type_reference_with_type_mark */
extern T1 T248f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_reference_with_type_mark */
extern T1 T73f208(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_reference_with_type_mark */
extern T1 T71f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.shallow_named_type_with_type_mark */
extern T0* T248f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.shallow_named_type_with_type_mark */
extern T0* T73f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.shallow_named_type_with_type_mark */
extern T0* T71f73(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.actual_parameter_count */
extern T6 T248f49(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.formal_parameter_count */
extern T6 T73f144(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.actual_parameter_count */
extern T6 T71f46(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.base_type_actual */
extern T0* T248f54(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T248f54ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.base_type_actual */
extern T0* T73f177(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.base_type_actual */
extern T0* T71f62(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T71f62ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.is_type_attached_with_type_mark */
extern T1 T248f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_attached_with_type_mark */
extern T1 T73f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_attached_with_type_mark */
extern T1 T71f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.adapted_base_class_with_named_feature */
extern T0* T248f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.adapted_base_class_with_named_feature */
extern T0* T73f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.adapted_base_class_with_named_feature */
extern T0* T71f68(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T248f70(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_CLASS.adapted_base_class_with_seeded_feature */
extern T0* T73f207(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T71f69(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.type_with_type_mark */
extern T0* T248f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.type_with_type_mark */
extern T0* T73f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.type_with_type_mark */
extern T0* T71f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.to_text */
extern T0* T248f28(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.to_text */
extern T0* T73f58(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.to_text */
extern T0* T71f26(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_base_type_with_type_marks */
extern T1 T248f89(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_base_tuple_type_with_type_marks */
extern T1 T71f117(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_base_type_with_type_marks */
extern T1 T73f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_base_class_type_with_type_marks */
extern T1 T71f116(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_base_type_with_type_marks */
extern T1 T71f97(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.base_type_index_of_label */
extern T6 T248f108(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.base_type_index_of_label */
extern T6 T73f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.base_type_index_of_label */
extern T6 T71f120(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.index_of_label */
extern T6 T248f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.index_of_label */
extern T6 T73f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.index_of_label */
extern T6 T71f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.base_class */
extern T0* T248f11(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.base_class */
extern T0* T73f56(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.base_class */
extern T0* T71f11(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.conforms_to_type_with_type_marks */
extern T1 T248f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_to_type_with_type_marks */
extern T1 T73f163(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_to_type_with_type_marks */
extern T1 T71f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] AP_DISPLAY_HELP_FLAG.example */
extern T0* T140f11(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.example */
extern T0* T47f15(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.example */
extern T0* T46f17(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.example */
extern T0* T45f13(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.example */
extern T0* T43f11(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.example */
extern T0* T41f18(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.was_found */
extern T1 T140f12(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.was_found */
extern T1 T47f13(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.was_found */
extern T1 T46f16(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.was_found */
extern T1 T45f18(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.was_found */
extern T1 T43f9(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.was_found */
extern T1 T41f13(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.occurrences */
extern T6 T47f16(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.occurrences */
extern T6 T46f14(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.occurrences */
extern T6 T45f16(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.occurrences */
extern T6 T41f17(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.allows_parameter */
extern T1 T140f10(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.allows_parameter */
extern T1 T47f18(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.allows_parameter */
extern T1 T46f15(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.allows_parameter */
extern T1 T45f17(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.allows_parameter */
extern T1 T43f13(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.allows_parameter */
extern T1 T41f16(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.needs_parameter */
extern T1 T140f13(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.needs_parameter */
extern T1 T43f12(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.name */
extern T0* T140f9(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.name */
extern T0* T47f17(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.name */
extern T0* T46f13(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.name */
extern T0* T45f15(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.name */
extern T0* T43f10(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.name */
extern T0* T41f14(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.names */
extern T0* T140f18(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.names */
extern T0* T47f21(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.has_long_form */
extern T1 T47f22(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.names */
extern T0* T47f21p1(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.names */
extern T0* T46f24(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.has_long_form */
extern T1 T46f25(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.names */
extern T0* T46f24p1(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.names */
extern T0* T45f23(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.has_long_form */
extern T1 T45f24(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.names */
extern T0* T45f23p1(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.names */
extern T0* T43f16(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.names */
extern T0* T41f22(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.has_long_form */
extern T1 T41f23(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.names */
extern T0* T41f22p1(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.is_closable */
extern T1 T909f5(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_closable */
extern T1 T876f27(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_closable */
extern T1 T32f38(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.name */
extern T0* T876f26(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.name */
extern T0* T1427f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.last_character */
extern T2 T876f19(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].item */
extern T2 T1573f4(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.end_of_input */
extern T1 T876f18(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.end_of_input */
extern T1 T32f33(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.read_to_string */
extern T6 T1427f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_STDIN_FILE.read_to_string */
extern T6 T909f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] CONSOLE.read_stream */
extern void T896f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] CONSOLE.read_stream_thread_aware */
extern void T896f33(GE_context* ac, T0* C, T6 a1);
/* detachable C_STRING.read_string_into */
extern void T30f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.console_readstream */
extern T6 T896f17(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] CONSOLE.read_data_buffer */
extern T0* T896f16(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.any_ */
extern T0* T909f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_to_string */
extern T6 T876f28(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_to_string */
extern T6 T876f28p1(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_TEXT_INPUT_FILE.read_to_string */
extern T6 T32f37(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_TEXT_INPUT_FILE.dummy_kl_character_buffer */
extern T0* T32f47(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.any_ */
extern T0* T32f45(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.is_string_32 */
extern T1 T416s51(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.is_string_8 */
extern T1 T416s50(GE_context* ac, T0* a1);
/* [detachable] UT_SHARED_ECF_VERSIONS.ecf_last_known */
extern T0* T2013s1(GE_context* ac);
/* [detachable] UT_SHARED_ECF_VERSIONS.ecf_1_18_0 */
extern T0* T2013s2(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.unicode_substring_character_count */
extern T6 T416s48(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.unicode_character_count */
extern T6 T416s47(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_next_byte */
extern T1 T416s25(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_second_byte */
extern T1 T416s20(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_first_byte */
extern T1 T416s17(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character */
extern T1 T416s39(GE_context* ac, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character_start */
extern T1 T416s40(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UNICODE_CONSTANTS.bom_character */
extern T3 T744s6(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.append_natural_32_code_to_utf8 */
extern void T416s54(GE_context* ac, T0* a1, T10 a2);
/* [detachable] KL_ANY_ROUTINES.same_objects */
extern T1 T427s2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.reset_context_if_multiple_constraints */
extern void T400s11(GE_context* ac, T1 a1, T0* a2, T0* a3);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.tokens */
extern T0* T400s7(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_latest */
extern T0* T1688s2(GE_context* ac);
/* Creation of agent #1 in feature [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_latest */
extern T0* T1688s2ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTION_DEFAULTS.boolean_option_value_regexp */
extern T0* T1688s8(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.string_ */
extern T0* T1688s7(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.case_insensitive_string_equality_tester */
extern T0* T1688s6(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions */
extern T0* T1688s1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_1_0_0 */
extern T0* T1688s5(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_1_2_0 */
extern T0* T1688s4(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.ecf_1_2_0 */
extern T0* T1688s3(GE_context* ac);
/* [detachable] EXECUTION_ENVIRONMENT.available_cpu_count */
extern T10 T122s7(GE_context* ac);
/* [detachable] PLATFORM.is_thread_capable */
extern T1 T48s1(GE_context* ac);
/* [detachable] THREAD_CONTROL.join_all */
extern void T683s1(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.append_substring_to_utf8 */
extern void T416s53(GE_context* ac, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] UC_UTF8_ROUTINES.four_byte_character_code */
extern T10 T416s46(GE_context* ac, T2 a1, T2 a2, T2 a3, T2 a4);
/* [detachable] UC_UTF8_ROUTINES.three_byte_character_code */
extern T10 T416s45(GE_context* ac, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.two_byte_character_code */
extern T10 T416s44(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.encoded_byte_count */
extern T6 T416s18(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.append_code_to_utf8 */
extern void T416s56(GE_context* ac, T0* a1, T6 a2);
/* [detachable] UC_UNICODE_ROUTINES.valid_non_surrogate_natural_32_code */
extern T1 T423s1(GE_context* ac, T10 a1);
/* [detachable] UC_UTF8_ROUTINES.string_to_utf8 */
extern T0* T416s1(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.substring_to_utf8 */
extern T0* T416s2(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.substring_byte_count */
extern T6 T416s3(GE_context* ac, T0* a1, T6 a2, T6 a3);
extern T1 T416s3ot1(T0* a1, T0** a2);
extern T1 T416s3ot2(T0* a1, T0** a2);
extern T1 T416s3ot3(T0* a1, T0** a2);
/* [detachable] UC_UTF8_ROUTINES.character_32_byte_count */
extern T6 T416s8(GE_context* ac, T3 a1);
/* [detachable] UC_UTF8_ROUTINES.natural_32_code_byte_count */
extern T6 T416s10(GE_context* ac, T10 a1);
/* [detachable] UC_UTF8_ROUTINES.dummy_uc_string */
extern T0* T416s7(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.character_8_byte_count */
extern T6 T416s6(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.any_ */
extern T0* T416s4(GE_context* ac);
/* [detachable] ISE_RUNTIME.check_assert */
extern T1 T1384s1(GE_context* ac, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR_FACTORY.new_system_processor */
extern T0* T68s1(GE_context* ac, T6 a1);
/* [detachable] PLATFORM.is_windows */
extern T1 T48s2(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.make */
extern T0* T62c47(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.standard_error_handler */
extern T0* T72f123(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.make */
extern T0* T449c996(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.make_parser_skeleton */
extern void T449f998(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_build_parser_tables */
extern void T449f1001(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template */
extern T0* T449f648(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yyfixed_array */
extern T0* T449f711(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_37 */
extern void T449f1102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyarray_subcopy */
extern void T449f1169(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PARSER.integer_array_ */
extern T0* T449f718(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_36 */
extern void T449f1101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_35 */
extern void T449f1100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_34 */
extern void T449f1099(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_33 */
extern void T449f1098(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_32 */
extern void T449f1097(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_31 */
extern void T449f1096(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_30 */
extern void T449f1095(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_29 */
extern void T449f1094(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_28 */
extern void T449f1093(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_27 */
extern void T449f1092(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_26 */
extern void T449f1091(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_25 */
extern void T449f1090(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_24 */
extern void T449f1089(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_23 */
extern void T449f1088(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_22 */
extern void T449f1087(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_21 */
extern void T449f1086(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_20 */
extern void T449f1085(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_19 */
extern void T449f1084(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_18 */
extern void T449f1083(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_17 */
extern void T449f1082(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_16 */
extern void T449f1081(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_15 */
extern void T449f1080(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_14 */
extern void T449f1079(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_13 */
extern void T449f1078(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_12 */
extern void T449f1077(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_11 */
extern void T449f1076(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_10 */
extern void T449f1075(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_9 */
extern void T449f1074(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_8 */
extern void T449f1073(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_7 */
extern void T449f1072(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_6 */
extern void T449f1071(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_5 */
extern void T449f1070(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_4 */
extern void T449f1069(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_3 */
extern void T449f1068(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_2 */
extern void T449f1067(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_1 */
extern void T449f1066(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template */
extern T0* T449f641(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_37 */
extern void T449f1065(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_36 */
extern void T449f1064(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_35 */
extern void T449f1063(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_34 */
extern void T449f1062(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_33 */
extern void T449f1061(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_32 */
extern void T449f1060(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_31 */
extern void T449f1059(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_30 */
extern void T449f1058(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_29 */
extern void T449f1057(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_28 */
extern void T449f1056(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_27 */
extern void T449f1055(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_26 */
extern void T449f1054(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_25 */
extern void T449f1053(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_24 */
extern void T449f1052(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_23 */
extern void T449f1051(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_22 */
extern void T449f1050(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_21 */
extern void T449f1049(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_20 */
extern void T449f1048(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_19 */
extern void T449f1047(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_18 */
extern void T449f1046(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_17 */
extern void T449f1045(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_16 */
extern void T449f1044(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_15 */
extern void T449f1043(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_14 */
extern void T449f1042(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_13 */
extern void T449f1041(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_12 */
extern void T449f1040(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_11 */
extern void T449f1039(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_10 */
extern void T449f1038(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_9 */
extern void T449f1037(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_8 */
extern void T449f1036(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_7 */
extern void T449f1035(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_6 */
extern void T449f1034(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_5 */
extern void T449f1033(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_4 */
extern void T449f1032(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_3 */
extern void T449f1031(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_2 */
extern void T449f1030(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_1 */
extern void T449f1029(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template */
extern T0* T449f989(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template_2 */
extern void T449f1028(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template_1 */
extern void T449f1027(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template */
extern T0* T449f732(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_9 */
extern void T449f1026(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_8 */
extern void T449f1025(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_7 */
extern void T449f1024(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_6 */
extern void T449f1023(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_5 */
extern void T449f1022(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_4 */
extern void T449f1021(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_3 */
extern void T449f1020(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_2 */
extern void T449f1019(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_1 */
extern void T449f1018(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template */
extern T0* T449f725(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template_2 */
extern void T449f1017(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template_1 */
extern void T449f1016(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template */
extern T0* T449f720(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_9 */
extern void T449f1015(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_8 */
extern void T449f1014(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_7 */
extern void T449f1013(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_6 */
extern void T449f1012(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_5 */
extern void T449f1011(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_4 */
extern void T449f1010(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_3 */
extern void T449f1009(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_2 */
extern void T449f1008(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_1 */
extern void T449f1007(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes2_template */
extern T0* T449f710(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template */
extern T0* T449f708(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_9 */
extern void T449f1165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_8 */
extern void T449f1164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_7 */
extern void T449f1163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_6 */
extern void T449f1162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_5 */
extern void T449f1161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_4 */
extern void T449f1160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_3 */
extern void T449f1159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_2 */
extern void T449f1158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_1 */
extern void T449f1157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template */
extern T0* T449f704(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_5 */
extern void T449f1149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_4 */
extern void T449f1148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_3 */
extern void T449f1147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_2 */
extern void T449f1146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_1 */
extern void T449f1145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template */
extern T0* T449f700(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template_2 */
extern void T449f1140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template_1 */
extern void T449f1139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_create_value_stacks */
extern void T449f1000(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].make */
extern T0* T1261f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WRITABLE].make_empty */
extern T0* T1262c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].default_create */
extern T0* T1261c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].make */
extern T0* T1258f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].make_empty */
extern T0* T1260c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].default_create */
extern T0* T1258c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].make */
extern T0* T1255f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].make_empty */
extern T0* T1257c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].default_create */
extern T0* T1255c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].make */
extern T0* T1252f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].make_empty */
extern T0* T1254c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].default_create */
extern T0* T1252c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].make */
extern T0* T1249f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].make_empty */
extern T0* T1251c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].default_create */
extern T0* T1249c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].make */
extern T0* T1246f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].make_empty */
extern T0* T1248c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].default_create */
extern T0* T1246c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].make */
extern T0* T1243f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].make_empty */
extern T0* T1245c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].default_create */
extern T0* T1243c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].make */
extern T0* T1240f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].make_empty */
extern T0* T1242c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].default_create */
extern T0* T1240c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].make */
extern T0* T1237f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].make_empty */
extern T0* T1239c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].default_create */
extern T0* T1237c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].make */
extern T0* T1235f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].make_empty */
extern T0* T1236c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].default_create */
extern T0* T1235c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].make */
extern T0* T1232f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].make_empty */
extern T0* T1234c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].default_create */
extern T0* T1232c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].make */
extern T0* T1229f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].make_empty */
extern T0* T1231c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].default_create */
extern T0* T1229c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].make */
extern T0* T1227f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].make_empty */
extern T0* T1228c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].default_create */
extern T0* T1227c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].make */
extern T0* T1224f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].make_empty */
extern T0* T1226c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].default_create */
extern T0* T1224c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].make */
extern T0* T1222f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].make_empty */
extern T0* T1223c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].default_create */
extern T0* T1222c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].make */
extern T0* T1219f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].make_empty */
extern T0* T1221c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].default_create */
extern T0* T1219c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].make */
extern T0* T1217f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].make_empty */
extern T0* T1218c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].default_create */
extern T0* T1217c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].make */
extern T0* T1215f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].make_empty */
extern T0* T1216c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].default_create */
extern T0* T1215c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].make */
extern T0* T1212f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].make_empty */
extern T0* T1214c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].default_create */
extern T0* T1212c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].make */
extern T0* T1210f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].make_empty */
extern T0* T1211c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].default_create */
extern T0* T1210c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].make */
extern T0* T1208f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].make_empty */
extern T0* T1209c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].default_create */
extern T0* T1208c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].make */
extern T0* T1205f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].make_empty */
extern T0* T1207c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].default_create */
extern T0* T1205c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].make */
extern T0* T1203f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].make_empty */
extern T0* T1204c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].default_create */
extern T0* T1203c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].make */
extern T0* T1200f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].make_empty */
extern T0* T1202c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].default_create */
extern T0* T1200c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].make */
extern T0* T1197f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].make_empty */
extern T0* T1199c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].default_create */
extern T0* T1197c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].make */
extern T0* T1195f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].make_empty */
extern T0* T1196c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].default_create */
extern T0* T1195c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].make */
extern T0* T1192f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].make_empty */
extern T0* T1194c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].default_create */
extern T0* T1192c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].make */
extern T0* T1189f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].make_empty */
extern T0* T1191c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].default_create */
extern T0* T1189c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].make */
extern T0* T1186f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].make_empty */
extern T0* T1188c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].default_create */
extern T0* T1186c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].make */
extern T0* T1183f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].make_empty */
extern T0* T1185c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].default_create */
extern T0* T1183c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].make */
extern T0* T1181f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].make_empty */
extern T0* T1182c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].default_create */
extern T0* T1181c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].make */
extern T0* T1178f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].make_empty */
extern T0* T1180c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].default_create */
extern T0* T1178c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].make */
extern T0* T1175f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].make_empty */
extern T0* T1177c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].default_create */
extern T0* T1175c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].make */
extern T0* T1172f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].make_empty */
extern T0* T1174c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].default_create */
extern T0* T1172c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].make */
extern T0* T1170f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].make_empty */
extern T0* T1171c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].default_create */
extern T0* T1170c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].make */
extern T0* T1167f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].make_empty */
extern T0* T1169c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].default_create */
extern T0* T1167c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].make */
extern T0* T1164f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].make_empty */
extern T0* T1166c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].default_create */
extern T0* T1164c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].make */
extern T0* T1161f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].make_empty */
extern T0* T1163c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].default_create */
extern T0* T1161c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].make */
extern T0* T1159f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].make_empty */
extern T0* T1160c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].default_create */
extern T0* T1159c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].make */
extern T0* T1156f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].make_empty */
extern T0* T1158c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].default_create */
extern T0* T1156c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].make */
extern T0* T1153f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].make_empty */
extern T0* T1155c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].default_create */
extern T0* T1153c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].make */
extern T0* T1151f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].make_empty */
extern T0* T1152c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].default_create */
extern T0* T1151c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].make */
extern T0* T1148f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].make_empty */
extern T0* T1150c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].default_create */
extern T0* T1148c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].make */
extern T0* T1145f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].make_empty */
extern T0* T1147c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].default_create */
extern T0* T1145c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].make */
extern T0* T1142f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].make_empty */
extern T0* T1144c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].default_create */
extern T0* T1142c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].make */
extern T0* T1139f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].make_empty */
extern T0* T1141c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].default_create */
extern T0* T1139c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].make */
extern T0* T1137f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].make_empty */
extern T0* T1138c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].default_create */
extern T0* T1137c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].make */
extern T0* T1134f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].make_empty */
extern T0* T1136c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].default_create */
extern T0* T1134c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].make */
extern T0* T1132f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].make_empty */
extern T0* T1133c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].default_create */
extern T0* T1132c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].make */
extern T0* T1130f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].make_empty */
extern T0* T1131c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].default_create */
extern T0* T1130c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].make */
extern T0* T1127f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].make_empty */
extern T0* T1129c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].default_create */
extern T0* T1127c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].make */
extern T0* T1125f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].make_empty */
extern T0* T1126c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].default_create */
extern T0* T1125c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].make */
extern T0* T1122f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].make_empty */
extern T0* T1124c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].default_create */
extern T0* T1122c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].make */
extern T0* T1120f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].make_empty */
extern T0* T1121c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].default_create */
extern T0* T1120c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].make */
extern T0* T1117f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].make_empty */
extern T0* T1119c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].default_create */
extern T0* T1117c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].make */
extern T0* T1114f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].make_empty */
extern T0* T1116c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].default_create */
extern T0* T1114c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].make */
extern T0* T1111f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].make_empty */
extern T0* T1113c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].default_create */
extern T0* T1111c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].make */
extern T0* T1108f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].make_empty */
extern T0* T1110c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].default_create */
extern T0* T1108c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].make */
extern T0* T1106f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].make_empty */
extern T0* T1107c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].default_create */
extern T0* T1106c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].make */
extern T0* T1104f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].make_empty */
extern T0* T1105c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].default_create */
extern T0* T1104c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].make */
extern T0* T1102f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].make_empty */
extern T0* T1103c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].default_create */
extern T0* T1102c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].make */
extern T0* T1100f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].make_empty */
extern T0* T1101c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].default_create */
extern T0* T1100c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].make */
extern T0* T1097f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].make_empty */
extern T0* T1099c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].default_create */
extern T0* T1097c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].make */
extern T0* T1094f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].make_empty */
extern T0* T1096c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].default_create */
extern T0* T1094c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].make */
extern T0* T1091f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].make_empty */
extern T0* T1093c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].default_create */
extern T0* T1091c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].make */
extern T0* T1088f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].make_empty */
extern T0* T1090c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].default_create */
extern T0* T1088c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].make */
extern T0* T1085f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].make_empty */
extern T0* T1087c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].default_create */
extern T0* T1085c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].make */
extern T0* T1082f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].make_empty */
extern T0* T1084c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].default_create */
extern T0* T1082c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].make */
extern T0* T1079f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].make_empty */
extern T0* T1081c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].default_create */
extern T0* T1079c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].make */
extern T0* T1076f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].make_empty */
extern T0* T1078c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].default_create */
extern T0* T1076c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].make */
extern T0* T1073f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].make_empty */
extern T0* T1075c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].default_create */
extern T0* T1073c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].make */
extern T0* T1071f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].make_empty */
extern T0* T1072c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].default_create */
extern T0* T1071c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].make */
extern T0* T1068f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].make_empty */
extern T0* T1070c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].default_create */
extern T0* T1068c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].make */
extern T0* T1065f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].make_empty */
extern T0* T1067c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].default_create */
extern T0* T1065c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].make */
extern T0* T1062f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].make_empty */
extern T0* T1064c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].default_create */
extern T0* T1062c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T1059f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].make_empty */
extern T0* T1061c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].default_create */
extern T0* T1059c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].make */
extern T0* T1056f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].make_empty */
extern T0* T1058c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].default_create */
extern T0* T1056c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].make */
extern T0* T1053f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].make_empty */
extern T0* T1055c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].default_create */
extern T0* T1053c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].make */
extern T0* T1050f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].make_empty */
extern T0* T1052c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].default_create */
extern T0* T1050c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make */
extern T0* T1047f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T1049c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T1047c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].make */
extern T0* T1045f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].make_empty */
extern T0* T1046c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].default_create */
extern T0* T1045c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].make */
extern T0* T1042f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].make_empty */
extern T0* T1044c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].default_create */
extern T0* T1042c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].make */
extern T0* T1040f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].make_empty */
extern T0* T1041c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].default_create */
extern T0* T1040c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].make */
extern T0* T1037f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].make_empty */
extern T0* T1039c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].default_create */
extern T0* T1037c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].make */
extern T0* T1035f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].make_empty */
extern T0* T1036c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].default_create */
extern T0* T1035c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].make */
extern T0* T1033f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].make_empty */
extern T0* T1034c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].default_create */
extern T0* T1033c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].make */
extern T0* T1030f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].make_empty */
extern T0* T1032c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].default_create */
extern T0* T1030c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].make */
extern T0* T1027f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].make_empty */
extern T0* T1029c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].default_create */
extern T0* T1027c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].make */
extern T0* T1024f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].make_empty */
extern T0* T1026c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].default_create */
extern T0* T1024c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].make */
extern T0* T1021f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].make_empty */
extern T0* T1023c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].default_create */
extern T0* T1021c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].make */
extern T0* T1018f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].make_empty */
extern T0* T1020c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].default_create */
extern T0* T1018c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].make */
extern T0* T1015f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].make_empty */
extern T0* T1017c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].default_create */
extern T0* T1015c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].make */
extern T0* T1012f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].make_empty */
extern T0* T1014c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].default_create */
extern T0* T1012c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].make */
extern T0* T1009f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].make_empty */
extern T0* T1011c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].default_create */
extern T0* T1009c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].make */
extern T0* T1007f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].make_empty */
extern T0* T1008c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].default_create */
extern T0* T1007c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].make */
extern T0* T1005f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].make_empty */
extern T0* T1006c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].default_create */
extern T0* T1005c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].make */
extern T0* T1002f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].make_empty */
extern T0* T1004c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].default_create */
extern T0* T1002c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].make */
extern T0* T999f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].make_empty */
extern T0* T1001c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].default_create */
extern T0* T999c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].make */
extern T0* T997f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].make_empty */
extern T0* T998c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].default_create */
extern T0* T997c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].make */
extern T0* T995f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].make_empty */
extern T0* T996c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].default_create */
extern T0* T995c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].make */
extern T0* T993f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T994c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T993c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].make */
extern T0* T991f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].make_empty */
extern T0* T992c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].default_create */
extern T0* T991c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].make */
extern T0* T988f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].make_empty */
extern T0* T990c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].default_create */
extern T0* T988c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].make */
extern T0* T985f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].make_empty */
extern T0* T987c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].default_create */
extern T0* T985c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].make */
extern T0* T982f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].make_empty */
extern T0* T984c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].default_create */
extern T0* T982c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].make */
extern T0* T980f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].make_empty */
extern T0* T981c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].default_create */
extern T0* T980c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].make */
extern T0* T977f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].make_empty */
extern T0* T979c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].default_create */
extern T0* T977c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].make */
extern T0* T974f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].make_empty */
extern T0* T976c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].default_create */
extern T0* T974c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].make */
extern T0* T971f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].make_empty */
extern T0* T973c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].default_create */
extern T0* T971c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].make */
extern T0* T968f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].make_empty */
extern T0* T970c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].default_create */
extern T0* T968c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].make */
extern T0* T966f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].make_empty */
extern T0* T967c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].default_create */
extern T0* T966c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].make */
extern T0* T963f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].make_empty */
extern T0* T965c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].default_create */
extern T0* T963c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].make */
extern T0* T961f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].make_empty */
extern T0* T962c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].default_create */
extern T0* T961c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].make */
extern T0* T958f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].make_empty */
extern T0* T960c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].default_create */
extern T0* T958c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].make */
extern T0* T955f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].make_empty */
extern T0* T957c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].default_create */
extern T0* T955c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].make */
extern T0* T953f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].make_empty */
extern T0* T954c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].default_create */
extern T0* T953c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].make */
extern T0* T950f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].make_empty */
extern T0* T952c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].default_create */
extern T0* T950c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].make */
extern T0* T948f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].make_empty */
extern T0* T949c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].default_create */
extern T0* T948c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].make */
extern T0* T945f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].make_empty */
extern T0* T947c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].default_create */
extern T0* T945c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].make */
extern T0* T942f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].make_empty */
extern T0* T944c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].default_create */
extern T0* T942c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].make */
extern T0* T939f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].make_empty */
extern T0* T941c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].default_create */
extern T0* T939c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].make */
extern T0* T937f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].make_empty */
extern T0* T938c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].default_create */
extern T0* T937c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].make */
extern T0* T935f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL].make_empty */
extern T0* T936c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].default_create */
extern T0* T935c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].make */
extern T0* T932f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].make_empty */
extern T0* T934c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].default_create */
extern T0* T932c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].make */
extern T0* T929f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].make_empty */
extern T0* T931c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].default_create */
extern T0* T929c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].make */
extern T0* T927f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD].make_empty */
extern T0* T928c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].default_create */
extern T0* T927c3(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.make_eiffel_scanner */
extern void T449f997(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.make_with_buffer */
extern void T449f999(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_initialize */
extern void T449f1002(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_initialize */
extern void T449f1002p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_build_tables */
extern void T449f1004(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template */
extern T0* T449f707(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_fixed_array */
extern T0* T449f712(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_4 */
extern void T449f1156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_array_subcopy */
extern void T449f1170(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_3 */
extern void T449f1155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_2 */
extern void T449f1154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_1 */
extern void T449f1153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template */
extern T0* T449f705(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_3 */
extern void T449f1152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_2 */
extern void T449f1151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_1 */
extern void T449f1150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_meta_template */
extern T0* T449f703(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template */
extern T0* T449f701(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_4 */
extern void T449f1144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_3 */
extern void T449f1143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_2 */
extern void T449f1142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_1 */
extern void T449f1141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template */
extern T0* T449f699(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_3 */
extern void T449f1138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_2 */
extern void T449f1137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_1 */
extern void T449f1136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template */
extern T0* T449f697(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_3 */
extern void T449f1135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_2 */
extern void T449f1134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_1 */
extern void T449f1133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template */
extern T0* T449f695(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_15 */
extern void T449f1132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_14 */
extern void T449f1131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_13 */
extern void T449f1130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_12 */
extern void T449f1129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_11 */
extern void T449f1128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_10 */
extern void T449f1127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_9 */
extern void T449f1126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_8 */
extern void T449f1125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_7 */
extern void T449f1124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_6 */
extern void T449f1123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_5 */
extern void T449f1122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_4 */
extern void T449f1121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_3 */
extern void T449f1120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_2 */
extern void T449f1119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_1 */
extern void T449f1118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template */
extern T0* T449f672(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_15 */
extern void T449f1117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_14 */
extern void T449f1116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_13 */
extern void T449f1115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_12 */
extern void T449f1114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_11 */
extern void T449f1113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_10 */
extern void T449f1112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_9 */
extern void T449f1111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_8 */
extern void T449f1110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_7 */
extern void T449f1109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_6 */
extern void T449f1108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_5 */
extern void T449f1107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_4 */
extern void T449f1106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_3 */
extern void T449f1105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_2 */
extern void T449f1104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_1 */
extern void T449f1103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.empty_buffer */
extern T0* T449f664(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T925c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].set_internal_cursor */
extern void T925f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].new_cursor */
extern T0* T925f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T1597c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].make */
extern T0* T924c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].set_internal_cursor */
extern void T924f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].new_cursor */
extern T0* T924f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].make */
extern T0* T1596c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ACROSS_COMPONENT_LIST].make */
extern T0* T923c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST].make */
extern T0* T1592f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACROSS_COMPONENT_LIST].make_empty */
extern T0* T1593c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACROSS_COMPONENT_LIST].default_create */
extern T0* T1592c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ACROSS_COMPONENT_LIST].make */
extern T0* T922c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST].make */
extern T0* T1591f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_COMPONENT_LIST].make_empty */
extern T0* T1590c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_COMPONENT_LIST].default_create */
extern T0* T1591c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].make */
extern T0* T921c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].make */
extern T0* T1588f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].make_empty */
extern T0* T1589c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].default_create */
extern T0* T1588c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].make */
extern T0* T920c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].make */
extern T0* T1587f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].make_empty */
extern T0* T1586c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].default_create */
extern T0* T1587c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].make */
extern T0* T919c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].make */
extern T0* T918c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].make */
extern T0* T917c9(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].make */
extern T0* T916c9(GE_context* ac, T6 a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.make_with_size */
extern T0* T914c22(GE_context* ac, T0* a1, T6 a2);
/* [detachable] YY_UNICODE_FILE_BUFFER.new_default_buffer */
extern T0* T914f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STANDARD_FILES.input */
extern T0* T135f3(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.make */
extern T0* T909c12(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.std */
extern T0* T449f698(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.make */
extern void T449f996p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.make */
extern T0* T448c7(GE_context* ac, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.tokens */
extern T0* T448f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.make */
extern T0* T447c136(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.make_eiffel_scanner */
extern void T447f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.make_with_buffer */
extern void T447f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_initialize */
extern void T447f139(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_initialize */
extern void T447f139p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_build_tables */
extern void T447f141(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template */
extern T0* T447f87(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_fixed_array */
extern T0* T447f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_4 */
extern void T447f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_array_subcopy */
extern void T447f180(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PREPARSER.integer_array_ */
extern T0* T447f102(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_3 */
extern void T447f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_2 */
extern void T447f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_1 */
extern void T447f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template */
extern T0* T447f85(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_3 */
extern void T447f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_2 */
extern void T447f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_1 */
extern void T447f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_meta_template */
extern T0* T447f80(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template */
extern T0* T447f74(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_4 */
extern void T447f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_3 */
extern void T447f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_2 */
extern void T447f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_1 */
extern void T447f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template */
extern T0* T447f71(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_3 */
extern void T447f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_2 */
extern void T447f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_1 */
extern void T447f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template */
extern T0* T447f63(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_3 */
extern void T447f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_2 */
extern void T447f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_1 */
extern void T447f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template */
extern T0* T447f59(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_10 */
extern void T447f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_9 */
extern void T447f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_8 */
extern void T447f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_7 */
extern void T447f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_6 */
extern void T447f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_5 */
extern void T447f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_4 */
extern void T447f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_3 */
extern void T447f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_2 */
extern void T447f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_1 */
extern void T447f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template */
extern T0* T447f94(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_10 */
extern void T447f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_9 */
extern void T447f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_8 */
extern void T447f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_7 */
extern void T447f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_6 */
extern void T447f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_5 */
extern void T447f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_4 */
extern void T447f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_3 */
extern void T447f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_2 */
extern void T447f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_1 */
extern void T447f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.empty_buffer */
extern T0* T447f83(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.std */
extern T0* T447f95(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make */
extern T0* T371c49(GE_context* ac, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make_single */
extern void T371f61(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make_null_single */
extern void T371f73(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_qualified_anchored_types_cycle_detection_enabled_only */
extern void T371f86(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_unknown_builtin_reported_only */
extern void T371f85(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_preparse_shallow_mode_only */
extern void T371f84(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].put_last */
extern void T680f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].force */
extern void T1365f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].fill_with */
extern void T1364f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].extend */
extern void T1364f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T680c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].set_internal_cursor */
extern void T680f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].new_cursor */
extern T0* T680f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T1366c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T1365f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].make_empty */
extern T0* T1364c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].default_create */
extern T0* T1365c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove */
extern void T200f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove_position */
extern void T200f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_keep_head */
extern void T200f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_keep_head */
extern void T200f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].keep_head */
extern void T583f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].set_count */
extern void T583f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].move_cursors_forth */
extern void T200f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].internal_set_key_equality_tester */
extern void T200f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].twin */
extern T0* T200f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].copy */
extern void T200f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].copy */
extern void T200f66p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_clashes */
extern void T200f73(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_slots */
extern void T200f72(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_key_storage */
extern void T200f71(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] STRING_8].resized_area */
extern T0* T325f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].copy_data */
extern void T325f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] STRING_8].force */
extern void T325f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].move_data */
extern void T325f14(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].overlapping_move */
extern void T325f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].non_overlapping_move */
extern void T325f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_item_storage */
extern void T200f70(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].resized_area */
extern T0* T583f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].copy_data */
extern void T583f11(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].force */
extern void T583f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].move_data */
extern void T583f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].overlapping_move */
extern void T583f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].non_overlapping_move */
extern void T583f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].valid_cursor */
extern T1 T200f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].move_all_cursors_after */
extern void T200f68(GE_context* ac, T0* C);
extern T0* GE_ma424(T6 c, T6 n, ...);
extern T0* GE_ma324(T6 c, T6 n, ...);
extern T0* GE_ma284(T6 c, T6 n, ...);
extern T0* GE_ma250(T6 c, T6 n, ...);
extern T0* GE_ma551(T6 c, T6 n, ...);
extern T0* GE_ma2053(T6 c, T6 n, ...);
extern T0* GE_ma2052(T6 c, T6 n, ...);
extern void GE_bma2052(T0* C, T6 s, T6 n, ...);
extern T0* GE_mt162(void);
extern T0* GE_mt831(T0* a1);
extern T0* GE_mt31(T6 a1, T6 a2, T6 a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T6 a10, T1 a11);
extern T0* GE_mt1925(T0* a1, T0* a2, T8 a3);
extern T0* GE_mt706(T0* a1, T0* a2, T0* a3);
extern T0* ge392ov5881863;
extern T0* ge923ov15384578;
extern T0* ge391ov5996545;
extern T0* ge347ov5685281;
extern T0* ge703ov11517958;
extern T0* ge478ov7831607;
extern T0* ge468ov7667772;
extern T0* ge393ov5996545;
extern T0* ge1458ov23888393;
extern T0* ge1458ov23888397;
extern T0* ge377ov6176776;
extern T0* ge1458ov23888389;
extern T0* ge1458ov23888396;
extern T0* ge1458ov23888394;
extern T0* ge1458ov23888395;
extern T0* ge378ov6193168;
extern T0* ge1458ov23888390;
extern T0* ge1458ov23888392;
extern T0* ge1458ov23888028;
extern T0* ge1458ov23888331;
extern T0* ge1458ov23888344;
extern T0* ge376ov6160403;
extern T0* ge376ov6160404;
extern T0* ge763ov12501023;
extern T0* ge763ov12501009;
extern T0* ge763ov12501007;
extern T0* ge763ov12500993;
extern T0* ge763ov12501011;
extern T0* ge763ov12501008;
extern T0* ge763ov12501012;
extern T0* ge763ov12501017;
extern T0* ge763ov12501014;
extern T0* ge763ov12501010;
extern T0* ge763ov12501004;
extern T0* ge763ov12501005;
extern T0* ge763ov12501013;
extern T0* ge763ov12501006;
extern T0* ge763ov12501018;
extern T0* ge763ov12500994;
extern T0* ge763ov12501020;
extern T0* ge763ov12501021;
extern T0* ge763ov12500996;
extern T0* ge763ov12500995;
extern T0* ge763ov12501019;
extern T0* ge763ov12501022;
extern T0* ge763ov12501002;
extern T0* ge763ov12501003;
extern T0* ge763ov12500999;
extern T0* ge763ov12501016;
extern T0* ge763ov12501015;
extern T0* ge763ov12501000;
extern T0* ge763ov12501001;
extern T0* ge763ov12500998;
extern T0* ge763ov12500997;
extern T0* ge759ov12435474;
extern T0* ge1458ov23888398;
extern T0* ge1458ov23888391;
extern T0* ge1458ov23888329;
extern T0* ge1458ov23888062;
extern T0* ge1458ov23888036;
extern T0* ge1458ov23888324;
extern T0* ge1458ov23888328;
extern T0* ge1458ov23888325;
extern T0* ge1458ov23888327;
extern T0* ge1458ov23888326;
extern T0* ge1458ov23888332;
extern T0* ge1458ov23888334;
extern T0* ge1458ov23888333;
extern T0* ge1458ov23888335;
extern T0* ge1452ov23789571;
extern T0* ge1452ov23789570;
extern T0* ge1452ov23789569;
extern T0* ge1233ov20202398;
extern T0* ge399ov5832735;
extern T0* ge399ov5832736;
extern T0* ge384ov6291468;
extern T0* ge384ov6291467;
extern T0* ge402ov5832735;
extern T0* ge402ov5832736;
extern T0* ge377ov6176778;
extern T0* ge1458ov23888107;
extern T0* ge1458ov23888374;
extern T0* ge1458ov23888113;
extern T0* ge1458ov23888017;
extern T0* ge1458ov23888375;
extern T0* ge1458ov23888200;
extern T0* ge1458ov23888199;
extern T0* ge1458ov23888138;
extern T0* ge1458ov23888169;
extern T0* ge1458ov23888216;
extern T0* ge1458ov23888273;
extern T0* ge1458ov23888300;
extern T0* ge1458ov23888340;
extern T0* ge1458ov23888070;
extern T0* ge1458ov23888368;
extern T0* ge1458ov23888115;
extern T0* ge1458ov23888092;
extern T0* ge1458ov23888019;
extern T0* ge1458ov23888336;
extern T0* ge1458ov23888088;
extern T0* ge1233ov20202369;
extern T0* ge1458ov23888339;
extern T0* ge1458ov23888069;
extern T0* ge1458ov23888063;
extern T0* ge1458ov23888037;
extern T0* ge1458ov23888338;
extern T0* ge1458ov23888130;
extern T0* ge1458ov23888064;
extern T0* ge1458ov23888378;
extern T0* ge1458ov23888155;
extern T0* ge1458ov23888152;
extern T0* ge1458ov23888369;
extern T0* ge1458ov23888244;
extern T0* ge1458ov23888261;
extern T0* ge1458ov23888243;
extern T0* ge1458ov23888260;
extern T0* ge1458ov23888242;
extern T0* ge1458ov23888259;
extern T0* ge1458ov23888220;
extern T0* ge1458ov23888029;
extern T0* ge1458ov23888241;
extern T0* ge1458ov23888258;
extern T0* ge1458ov23888240;
extern T0* ge1458ov23888257;
extern T0* ge1458ov23888239;
extern T0* ge1458ov23888256;
extern T0* ge1458ov23888238;
extern T0* ge1458ov23888255;
extern T0* ge1458ov23888236;
extern T0* ge1458ov23888253;
extern T0* ge1458ov23888235;
extern T0* ge1458ov23888252;
extern T0* ge1458ov23888234;
extern T0* ge1458ov23888251;
extern T0* ge1458ov23888233;
extern T0* ge1458ov23888250;
extern T0* ge1458ov23888232;
extern T0* ge1458ov23888249;
extern T0* ge1458ov23888231;
extern T0* ge1458ov23888248;
extern T0* ge1458ov23888230;
extern T0* ge1458ov23888247;
extern T0* ge1458ov23888129;
extern T0* ge1458ov23888100;
extern T0* ge1458ov23888025;
extern T0* ge1458ov23888034;
extern T0* ge1458ov23888058;
extern T0* ge1458ov23888056;
extern T0* ge1458ov23888047;
extern T0* ge1458ov23888046;
extern T0* ge1458ov23888045;
extern T0* ge1458ov23888044;
extern T0* ge1458ov23888043;
extern T0* ge1458ov23888151;
extern T0* ge1458ov23888041;
extern T0* ge1458ov23888059;
extern T0* ge1458ov23888033;
extern T0* ge1458ov23888031;
extern T0* ge1458ov23888077;
extern T0* ge1458ov23888117;
extern T0* ge1458ov23888116;
extern T0* ge1458ov23888299;
extern T0* ge1458ov23888210;
extern T0* ge1458ov23888263;
extern T0* ge1458ov23888262;
extern T0* ge1458ov23888246;
extern T0* ge1458ov23888245;
extern T0* ge1458ov23888229;
extern T0* ge1458ov23888228;
extern T0* ge1458ov23888225;
extern T0* ge1458ov23888226;
extern T0* ge1458ov23888219;
extern T0* ge1458ov23888218;
extern T0* ge1458ov23888217;
extern T0* ge1458ov23888363;
extern T0* ge1458ov23888215;
extern T0* ge1458ov23888208;
extern T0* ge1458ov23888207;
extern T0* ge1458ov23888204;
extern T0* ge1458ov23888201;
extern T0* ge1458ov23888198;
extern T0* ge1458ov23888197;
extern T0* ge1458ov23888196;
extern T0* ge1458ov23888195;
extern T0* ge1458ov23888194;
extern T0* ge1458ov23888193;
extern T0* ge1458ov23888192;
extern T0* ge1458ov23888191;
extern T0* ge1458ov23888165;
extern T0* ge1458ov23888166;
extern T0* ge1458ov23888162;
extern T0* ge1458ov23888164;
extern T0* ge1458ov23888163;
extern T0* ge1458ov23888161;
extern T0* ge1458ov23888159;
extern T0* ge1458ov23888160;
extern T0* ge1458ov23888158;
extern T0* ge1458ov23888157;
extern T0* ge1458ov23888156;
extern T0* ge1458ov23888154;
extern T0* ge1458ov23888153;
extern T0* ge1458ov23888150;
extern T0* ge1458ov23888149;
extern T0* ge1458ov23888148;
extern T0* ge1458ov23888147;
extern T0* ge1458ov23888146;
extern T0* ge1458ov23888145;
extern T0* ge1458ov23888144;
extern T0* ge1458ov23888143;
extern T0* ge1458ov23888133;
extern T0* ge1458ov23888131;
extern T0* ge1458ov23888128;
extern T0* ge1458ov23888127;
extern T0* ge1458ov23888126;
extern T0* ge1458ov23888125;
extern T0* ge1458ov23888124;
extern T0* ge1458ov23888132;
extern T0* ge1458ov23888212;
extern T0* ge1458ov23888211;
extern T0* ge1458ov23888209;
extern T0* ge1458ov23888106;
extern T0* ge1458ov23888105;
extern T0* ge1458ov23888104;
extern T0* ge1458ov23888103;
extern T0* ge1458ov23888102;
extern T0* ge1458ov23888101;
extern T0* ge1458ov23888087;
extern T0* ge1458ov23888086;
extern T0* ge1458ov23888084;
extern T0* ge1458ov23888213;
extern T0* ge1458ov23888134;
extern T0* ge1458ov23888385;
extern T0* ge1458ov23888223;
extern T0* ge1458ov23888122;
extern T0* ge1458ov23888094;
extern T0* ge1458ov23888093;
extern T0* ge1458ov23888097;
extern T0* ge1458ov23888096;
extern T0* ge1458ov23888095;
extern T0* ge1458ov23888099;
extern T0* ge1458ov23888167;
extern T0* ge1458ov23888083;
extern T0* ge1458ov23888082;
extern T0* ge1458ov23888237;
extern T0* ge1458ov23888254;
extern T0* ge1458ov23888353;
extern T0* ge1458ov23888180;
extern T0* ge1458ov23888321;
extern T0* ge1458ov23888111;
extern T0* ge1458ov23888112;
extern T0* ge1233ov20202389;
extern T0* ge1458ov23888298;
extern T0* ge1458ov23888362;
extern T0* ge1458ov23888320;
extern T0* ge1458ov23888318;
extern T0* ge1458ov23888319;
extern T0* ge1458ov23888314;
extern T0* ge1458ov23888301;
extern T0* ge1458ov23888302;
extern T0* ge1458ov23888303;
extern T0* ge1458ov23888306;
extern T0* ge1458ov23888307;
extern T0* ge1458ov23888308;
extern T0* ge1458ov23888309;
extern T0* ge1458ov23888310;
extern T0* ge1458ov23888311;
extern T0* ge1458ov23888312;
extern T0* ge1458ov23888313;
extern T0* ge1458ov23888315;
extern T0* ge1458ov23888316;
extern T0* ge1458ov23888317;
extern T0* ge1458ov23888305;
extern T0* ge1458ov23888304;
extern T0* ge1233ov20202219;
extern T0* ge1458ov23888076;
extern T0* ge1458ov23888065;
extern T0* ge1458ov23888137;
extern T0* ge1458ov23888110;
extern T0* ge1458ov23888373;
extern T0* ge1458ov23888024;
extern T0* ge1458ov23888354;
extern T0* ge1458ov23888341;
extern T0* ge1458ov23888175;
extern T0* ge1458ov23888176;
extern T0* ge1458ov23888355;
extern T0* ge1458ov23888080;
extern T0* ge1233ov20202372;
extern T0* ge1458ov23888183;
extern T0* ge1458ov23888085;
extern T0* ge1458ov23888172;
extern T0* ge1458ov23888021;
extern T0* ge1458ov23888027;
extern T0* ge1458ov23888048;
extern T0* ge1458ov23888323;
extern T0* ge1458ov23888322;
extern T0* ge1458ov23888297;
extern T0* ge1458ov23888296;
extern T0* ge1458ov23888295;
extern T0* ge1458ov23888294;
extern T0* ge1458ov23888293;
extern T0* ge1458ov23888292;
extern T0* ge1458ov23888291;
extern T0* ge1458ov23888290;
extern T0* ge1458ov23888289;
extern T0* ge1458ov23888288;
extern T0* ge1458ov23888287;
extern T0* ge1458ov23888286;
extern T0* ge1458ov23888285;
extern T0* ge1458ov23888284;
extern T0* ge1458ov23888283;
extern T0* ge1458ov23888282;
extern T0* ge1458ov23888281;
extern T0* ge1458ov23888280;
extern T0* ge1458ov23888279;
extern T0* ge1458ov23888278;
extern T0* ge1458ov23888277;
extern T0* ge1458ov23888276;
extern T0* ge1458ov23888275;
extern T0* ge1458ov23888274;
extern T0* ge1458ov23888272;
extern T0* ge1458ov23888271;
extern T0* ge1458ov23888270;
extern T0* ge1458ov23888269;
extern T0* ge1458ov23888268;
extern T0* ge1458ov23888267;
extern T0* ge1458ov23888177;
extern T0* ge1458ov23888171;
extern T0* ge1458ov23888114;
extern T0* ge1458ov23888214;
extern T0* ge1458ov23888098;
extern T0* ge1458ov23888081;
extern T0* ge1458ov23888371;
extern T0* ge1458ov23888364;
extern T0* ge1458ov23888357;
extern T0* ge1458ov23888358;
extern T0* ge1458ov23888018;
extern T0* ge1458ov23888359;
extern T0* ge1458ov23888071;
extern T0* ge1458ov23888066;
extern T0* ge1458ov23888089;
extern T0* ge1233ov20202015;
extern T0* ge1233ov20202040;
extern T0* ge1233ov20202082;
extern T0* ge1458ov23888022;
extern T0* ge1458ov23888023;
extern T0* ge1233ov20202051;
extern T0* ge1458ov23888345;
extern T0* ge1458ov23888173;
extern T0* ge1458ov23888174;
extern T0* ge1458ov23888227;
extern T0* ge1458ov23888203;
extern T0* ge1458ov23888202;
extern T0* ge1458ov23888206;
extern T0* ge1458ov23888205;
extern T0* ge1458ov23888030;
extern T0* ge1458ov23888372;
extern T0* ge1458ov23888380;
extern T0* ge1458ov23888078;
extern T0* ge1458ov23888381;
extern T0* ge1458ov23888361;
extern T0* ge1458ov23888379;
extern T0* ge1458ov23888054;
extern T0* ge1458ov23888053;
extern T0* ge1458ov23888052;
extern T0* ge1458ov23888051;
extern T0* ge1458ov23888042;
extern T0* ge1458ov23888040;
extern T0* ge1458ov23888039;
extern T0* ge1458ov23888035;
extern T0* ge1233ov20202011;
extern T0* ge1458ov23888057;
extern T0* ge1233ov20202041;
extern T0* ge1458ov23888050;
extern T0* ge1233ov20202031;
extern T0* ge1458ov23888038;
extern T0* ge1233ov20202019;
extern T0* ge1458ov23888067;
extern T0* ge1233ov20202054;
extern T0* ge1458ov23888032;
extern T0* ge1233ov20202006;
extern T0* ge1233ov20202028;
extern T0* ge1233ov20202027;
extern T0* ge1233ov20202010;
extern T0* ge1233ov20202049;
extern T0* ge1233ov20202002;
extern T0* ge1233ov20202004;
extern T0* ge1233ov20202053;
extern T0* ge1233ov20202052;
extern T0* ge1233ov20202045;
extern T0* ge1233ov20202044;
extern T0* ge1233ov20202039;
extern T0* ge1233ov20202038;
extern T0* ge1233ov20202043;
extern T0* ge1233ov20202042;
extern T0* ge1233ov20202035;
extern T0* ge1233ov20202034;
extern T0* ge1233ov20202033;
extern T0* ge1233ov20202032;
extern T0* ge1233ov20202023;
extern T0* ge1233ov20202022;
extern T0* ge1233ov20202021;
extern T0* ge1233ov20202020;
extern T0* ge1233ov20202008;
extern T0* ge1233ov20202007;
extern T0* ge1233ov20202005;
extern T0* ge1233ov20202387;
extern T0* ge1233ov20202055;
extern T0* ge1457ov23872269;
extern T0* ge1457ov23871950;
extern T0* ge1457ov23871772;
extern T0* ge1455ov23838737;
extern T0* ge1455ov23838731;
extern T0* ge1455ov23838725;
extern T0* ge1455ov23838733;
extern T0* ge1455ov23838727;
extern T0* ge1455ov23838721;
extern T0* ge1455ov23838736;
extern T0* ge1455ov23838730;
extern T0* ge1455ov23838724;
extern T0* ge1455ov23838738;
extern T0* ge1455ov23838732;
extern T0* ge1455ov23838726;
extern T0* ge1233ov20202111;
extern T0* ge1455ov23838735;
extern T0* ge1455ov23838729;
extern T0* ge1455ov23838723;
extern T0* ge1233ov20202311;
extern T0* ge1233ov20202257;
extern T0* ge1233ov20202221;
extern T0* ge1233ov20202167;
extern T0* ge1457ov23872294;
extern T0* ge1457ov23871971;
extern T0* ge1457ov23871797;
extern T0* ge1457ov23872292;
extern T0* ge1457ov23871969;
extern T0* ge1457ov23871795;
extern T0* ge1233ov20202341;
extern T0* ge1233ov20202301;
extern T0* ge1233ov20202099;
extern T0* ge1233ov20202352;
extern T0* ge1233ov20202265;
extern T0* ge1233ov20202367;
extern T0* ge1233ov20202224;
extern T0* ge1233ov20202415;
extern T0* ge1233ov20202384;
extern T0* ge1233ov20202058;
extern T0* ge1233ov20202083;
extern T0* ge1457ov23872291;
extern T0* ge1457ov23871968;
extern T0* ge1457ov23871794;
extern T0* ge1457ov23872289;
extern T0* ge1457ov23871966;
extern T0* ge1457ov23871792;
extern T0* ge1457ov23872290;
extern T0* ge1457ov23871967;
extern T0* ge1457ov23871793;
extern T0* ge1233ov20202106;
extern T0* ge1455ov23838734;
extern T0* ge1455ov23838728;
extern T0* ge1455ov23838722;
extern T0* ge1456ov23855114;
extern T0* ge1456ov23855110;
extern T0* ge1456ov23855109;
extern T0* ge1456ov23855113;
extern T0* ge1456ov23855108;
extern T0* ge1456ov23855112;
extern T0* ge1456ov23855107;
extern T0* ge1456ov23855111;
extern T0* ge1456ov23855106;
extern T0* ge1233ov20202030;
extern T0* ge1621ov26558481;
extern T0* ge1629ov26689559;
extern T0* ge522ov8552452;
extern T0* ge1447ov23707660;
extern T0* ge1447ov23707655;
extern T0* ge1447ov23707650;
extern T0* ge1133ov18563123;
extern T0* ge1133ov18563124;
extern T0* ge1447ov23707661;
extern T0* ge1447ov23707656;
extern T0* ge1447ov23707651;
extern T0* ge1469ov24068147;
extern T0* ge1454ov23822339;
extern T0* ge1454ov23822338;
extern T0* ge1233ov20202491;
extern T0* ge1233ov20202492;
extern T0* ge1233ov20202493;
extern T0* ge1233ov20202494;
extern T0* ge1233ov20202495;
extern T0* ge1233ov20202496;
extern T0* ge1233ov20202497;
extern T0* ge1233ov20202498;
extern T0* ge1233ov20202499;
extern T0* ge1233ov20202500;
extern T0* ge1233ov20202501;
extern T0* ge1233ov20202502;
extern T0* ge1233ov20202503;
extern T0* ge1233ov20202504;
extern T0* ge1233ov20202505;
extern T0* ge1233ov20202506;
extern T0* ge1233ov20202507;
extern T0* ge1233ov20202508;
extern T0* ge1233ov20202509;
extern T0* ge1233ov20202510;
extern T0* ge1233ov20202553;
extern T0* ge1233ov20202554;
extern T0* ge1233ov20202555;
extern T0* ge1233ov20202003;
extern T0* ge1233ov20202009;
extern T0* ge1233ov20202012;
extern T0* ge1233ov20202013;
extern T0* ge1233ov20202014;
extern T0* ge1233ov20202016;
extern T0* ge1233ov20202017;
extern T0* ge1233ov20202018;
extern T0* ge1233ov20202024;
extern T0* ge1233ov20202025;
extern T0* ge1233ov20202026;
extern T0* ge1233ov20202029;
extern T0* ge1233ov20202036;
extern T0* ge1233ov20202037;
extern T0* ge1233ov20202046;
extern T0* ge1233ov20202047;
extern T0* ge1233ov20202048;
extern T0* ge1233ov20202050;
extern T0* ge1233ov20202056;
extern T0* ge1233ov20202057;
extern T0* ge1233ov20202059;
extern T0* ge1233ov20202060;
extern T0* ge1233ov20202061;
extern T0* ge1233ov20202062;
extern T0* ge1233ov20202063;
extern T0* ge1233ov20202064;
extern T0* ge1233ov20202065;
extern T0* ge1233ov20202066;
extern T0* ge1233ov20202067;
extern T0* ge1233ov20202068;
extern T0* ge1233ov20202069;
extern T0* ge1233ov20202070;
extern T0* ge1233ov20202071;
extern T0* ge1233ov20202072;
extern T0* ge1233ov20202073;
extern T0* ge1233ov20202074;
extern T0* ge1233ov20202075;
extern T0* ge1233ov20202076;
extern T0* ge1233ov20202077;
extern T0* ge1233ov20202078;
extern T0* ge1233ov20202079;
extern T0* ge1233ov20202080;
extern T0* ge1233ov20202081;
extern T0* ge1233ov20202084;
extern T0* ge1233ov20202085;
extern T0* ge1233ov20202086;
extern T0* ge1233ov20202087;
extern T0* ge1233ov20202088;
extern T0* ge1233ov20202089;
extern T0* ge1233ov20202090;
extern T0* ge1233ov20202091;
extern T0* ge1233ov20202092;
extern T0* ge1233ov20202093;
extern T0* ge1233ov20202094;
extern T0* ge1233ov20202095;
extern T0* ge1233ov20202096;
extern T0* ge1233ov20202097;
extern T0* ge1233ov20202098;
extern T0* ge1233ov20202100;
extern T0* ge1233ov20202101;
extern T0* ge1233ov20202102;
extern T0* ge1233ov20202103;
extern T0* ge1233ov20202104;
extern T0* ge1233ov20202105;
extern T0* ge1233ov20202107;
extern T0* ge1233ov20202108;
extern T0* ge1233ov20202109;
extern T0* ge1233ov20202110;
extern T0* ge1233ov20202112;
extern T0* ge1233ov20202113;
extern T0* ge1233ov20202114;
extern T0* ge1233ov20202115;
extern T0* ge1233ov20202116;
extern T0* ge1233ov20202117;
extern T0* ge1233ov20202118;
extern T0* ge1233ov20202119;
extern T0* ge1233ov20202120;
extern T0* ge1233ov20202121;
extern T0* ge1233ov20202122;
extern T0* ge1233ov20202123;
extern T0* ge1233ov20202124;
extern T0* ge1233ov20202125;
extern T0* ge1233ov20202126;
extern T0* ge1233ov20202127;
extern T0* ge1233ov20202128;
extern T0* ge1233ov20202129;
extern T0* ge1233ov20202130;
extern T0* ge1233ov20202131;
extern T0* ge1233ov20202132;
extern T0* ge1233ov20202133;
extern T0* ge1233ov20202134;
extern T0* ge1233ov20202135;
extern T0* ge1233ov20202136;
extern T0* ge1233ov20202137;
extern T0* ge1233ov20202138;
extern T0* ge1233ov20202139;
extern T0* ge1233ov20202140;
extern T0* ge1233ov20202141;
extern T0* ge1233ov20202142;
extern T0* ge1233ov20202143;
extern T0* ge1233ov20202144;
extern T0* ge1233ov20202145;
extern T0* ge1233ov20202146;
extern T0* ge1233ov20202147;
extern T0* ge1233ov20202148;
extern T0* ge1233ov20202149;
extern T0* ge1233ov20202150;
extern T0* ge1233ov20202151;
extern T0* ge1233ov20202152;
extern T0* ge1233ov20202153;
extern T0* ge1233ov20202154;
extern T0* ge1233ov20202155;
extern T0* ge1233ov20202156;
extern T0* ge1233ov20202157;
extern T0* ge1233ov20202158;
extern T0* ge1233ov20202159;
extern T0* ge1233ov20202160;
extern T0* ge1233ov20202161;
extern T0* ge1233ov20202162;
extern T0* ge1233ov20202163;
extern T0* ge1233ov20202164;
extern T0* ge1233ov20202165;
extern T0* ge1233ov20202166;
extern T0* ge1233ov20202168;
extern T0* ge1233ov20202169;
extern T0* ge1233ov20202170;
extern T0* ge1233ov20202171;
extern T0* ge1233ov20202172;
extern T0* ge1233ov20202173;
extern T0* ge1233ov20202174;
extern T0* ge1233ov20202175;
extern T0* ge1233ov20202176;
extern T0* ge1233ov20202177;
extern T0* ge1233ov20202178;
extern T0* ge1233ov20202179;
extern T0* ge1233ov20202180;
extern T0* ge1233ov20202181;
extern T0* ge1233ov20202182;
extern T0* ge1233ov20202183;
extern T0* ge1233ov20202184;
extern T0* ge1233ov20202185;
extern T0* ge1233ov20202186;
extern T0* ge1233ov20202187;
extern T0* ge1233ov20202188;
extern T0* ge1233ov20202189;
extern T0* ge1233ov20202190;
extern T0* ge1233ov20202191;
extern T0* ge1233ov20202192;
extern T0* ge1233ov20202193;
extern T0* ge1233ov20202194;
extern T0* ge1233ov20202195;
extern T0* ge1233ov20202196;
extern T0* ge1233ov20202197;
extern T0* ge1233ov20202198;
extern T0* ge1233ov20202199;
extern T0* ge1233ov20202200;
extern T0* ge1233ov20202201;
extern T0* ge1233ov20202202;
extern T0* ge1233ov20202203;
extern T0* ge1233ov20202204;
extern T0* ge1233ov20202205;
extern T0* ge1233ov20202206;
extern T0* ge1233ov20202207;
extern T0* ge1233ov20202208;
extern T0* ge1233ov20202209;
extern T0* ge1233ov20202210;
extern T0* ge1233ov20202211;
extern T0* ge1233ov20202212;
extern T0* ge1233ov20202213;
extern T0* ge1233ov20202214;
extern T0* ge1233ov20202215;
extern T0* ge1233ov20202216;
extern T0* ge1233ov20202217;
extern T0* ge1233ov20202218;
extern T0* ge1233ov20202220;
extern T0* ge1233ov20202222;
extern T0* ge1233ov20202223;
extern T0* ge1233ov20202225;
extern T0* ge1233ov20202226;
extern T0* ge1233ov20202227;
extern T0* ge1233ov20202228;
extern T0* ge1233ov20202229;
extern T0* ge1233ov20202230;
extern T0* ge1233ov20202231;
extern T0* ge1233ov20202232;
extern T0* ge1233ov20202233;
extern T0* ge1233ov20202234;
extern T0* ge1233ov20202235;
extern T0* ge1233ov20202236;
extern T0* ge1233ov20202237;
extern T0* ge1233ov20202238;
extern T0* ge1233ov20202239;
extern T0* ge1233ov20202240;
extern T0* ge1233ov20202241;
extern T0* ge1233ov20202242;
extern T0* ge1233ov20202243;
extern T0* ge1233ov20202244;
extern T0* ge1233ov20202245;
extern T0* ge1233ov20202246;
extern T0* ge1233ov20202247;
extern T0* ge1233ov20202248;
extern T0* ge1233ov20202249;
extern T0* ge1233ov20202250;
extern T0* ge1233ov20202251;
extern T0* ge1233ov20202252;
extern T0* ge1233ov20202253;
extern T0* ge1233ov20202254;
extern T0* ge1233ov20202255;
extern T0* ge1233ov20202256;
extern T0* ge1233ov20202258;
extern T0* ge1233ov20202259;
extern T0* ge1233ov20202260;
extern T0* ge1233ov20202261;
extern T0* ge1233ov20202262;
extern T0* ge1233ov20202263;
extern T0* ge1233ov20202264;
extern T0* ge1233ov20202266;
extern T0* ge1233ov20202267;
extern T0* ge1233ov20202268;
extern T0* ge1233ov20202269;
extern T0* ge1233ov20202270;
extern T0* ge1233ov20202271;
extern T0* ge1233ov20202272;
extern T0* ge1233ov20202273;
extern T0* ge1233ov20202274;
extern T0* ge1233ov20202275;
extern T0* ge1233ov20202276;
extern T0* ge1233ov20202277;
extern T0* ge1233ov20202278;
extern T0* ge1233ov20202279;
extern T0* ge1233ov20202280;
extern T0* ge1233ov20202281;
extern T0* ge1233ov20202282;
extern T0* ge1233ov20202283;
extern T0* ge1233ov20202284;
extern T0* ge1233ov20202285;
extern T0* ge1233ov20202286;
extern T0* ge1233ov20202287;
extern T0* ge1233ov20202288;
extern T0* ge1233ov20202289;
extern T0* ge1233ov20202290;
extern T0* ge1233ov20202291;
extern T0* ge1233ov20202292;
extern T0* ge1233ov20202293;
extern T0* ge1233ov20202294;
extern T0* ge1233ov20202295;
extern T0* ge1233ov20202296;
extern T0* ge1233ov20202297;
extern T0* ge1233ov20202298;
extern T0* ge1233ov20202299;
extern T0* ge1233ov20202300;
extern T0* ge1233ov20202302;
extern T0* ge1233ov20202303;
extern T0* ge1233ov20202304;
extern T0* ge1233ov20202305;
extern T0* ge1233ov20202306;
extern T0* ge1233ov20202307;
extern T0* ge1233ov20202308;
extern T0* ge1233ov20202309;
extern T0* ge1233ov20202310;
extern T0* ge1233ov20202312;
extern T0* ge1233ov20202313;
extern T0* ge1233ov20202314;
extern T0* ge1233ov20202315;
extern T0* ge1233ov20202316;
extern T0* ge1233ov20202317;
extern T0* ge1233ov20202318;
extern T0* ge1233ov20202319;
extern T0* ge1233ov20202320;
extern T0* ge1233ov20202321;
extern T0* ge1233ov20202322;
extern T0* ge1233ov20202323;
extern T0* ge1233ov20202324;
extern T0* ge1233ov20202325;
extern T0* ge1233ov20202326;
extern T0* ge1233ov20202327;
extern T0* ge1233ov20202328;
extern T0* ge1233ov20202329;
extern T0* ge1233ov20202330;
extern T0* ge1233ov20202331;
extern T0* ge1233ov20202332;
extern T0* ge1233ov20202333;
extern T0* ge1233ov20202334;
extern T0* ge1233ov20202335;
extern T0* ge1233ov20202336;
extern T0* ge1233ov20202337;
extern T0* ge1233ov20202338;
extern T0* ge1233ov20202339;
extern T0* ge1233ov20202340;
extern T0* ge1233ov20202342;
extern T0* ge1233ov20202343;
extern T0* ge1233ov20202344;
extern T0* ge1233ov20202345;
extern T0* ge1233ov20202346;
extern T0* ge1233ov20202347;
extern T0* ge1233ov20202348;
extern T0* ge1233ov20202349;
extern T0* ge1233ov20202350;
extern T0* ge1233ov20202351;
extern T0* ge1233ov20202353;
extern T0* ge1233ov20202354;
extern T0* ge1233ov20202355;
extern T0* ge1233ov20202356;
extern T0* ge1233ov20202357;
extern T0* ge1233ov20202358;
extern T0* ge1233ov20202359;
extern T0* ge1233ov20202360;
extern T0* ge1233ov20202361;
extern T0* ge1233ov20202362;
extern T0* ge1233ov20202363;
extern T0* ge1233ov20202364;
extern T0* ge1233ov20202365;
extern T0* ge1233ov20202366;
extern T0* ge1233ov20202368;
extern T0* ge1233ov20202370;
extern T0* ge1233ov20202371;
extern T0* ge1233ov20202373;
extern T0* ge1233ov20202374;
extern T0* ge1233ov20202375;
extern T0* ge1233ov20202376;
extern T0* ge1233ov20202377;
extern T0* ge1233ov20202378;
extern T0* ge1233ov20202379;
extern T0* ge1233ov20202380;
extern T0* ge1233ov20202382;
extern T0* ge1233ov20202383;
extern T0* ge1233ov20202385;
extern T0* ge1233ov20202386;
extern T0* ge1233ov20202388;
extern T0* ge1233ov20202390;
extern T0* ge1233ov20202391;
extern T0* ge1233ov20202392;
extern T0* ge1233ov20202393;
extern T0* ge1233ov20202394;
extern T0* ge1233ov20202395;
extern T0* ge1233ov20202396;
extern T0* ge1233ov20202397;
extern T0* ge1233ov20202399;
extern T0* ge1233ov20202400;
extern T0* ge1233ov20202401;
extern T0* ge1233ov20202402;
extern T0* ge1233ov20202403;
extern T0* ge1233ov20202404;
extern T0* ge1233ov20202405;
extern T0* ge1233ov20202406;
extern T0* ge1233ov20202407;
extern T0* ge1233ov20202408;
extern T0* ge1233ov20202409;
extern T0* ge1233ov20202410;
extern T0* ge1233ov20202411;
extern T0* ge1233ov20202412;
extern T0* ge1233ov20202413;
extern T0* ge1233ov20202414;
extern T0* ge1233ov20202416;
extern T0* ge1233ov20202417;
extern T0* ge1233ov20202418;
extern T0* ge1233ov20202419;
extern T0* ge1233ov20202420;
extern T0* ge1233ov20202421;
extern T0* ge1233ov20202422;
extern T0* ge1233ov20202423;
extern T0* ge1233ov20202425;
extern T0* ge1233ov20202426;
extern T0* ge1233ov20202427;
extern T0* ge1233ov20202428;
extern T0* ge1233ov20202429;
extern T0* ge1233ov20202430;
extern T0* ge1233ov20202431;
extern T0* ge1233ov20202432;
extern T0* ge1233ov20202433;
extern T0* ge1233ov20202434;
extern T0* ge1233ov20202435;
extern T0* ge1233ov20202436;
extern T0* ge1233ov20202437;
extern T0* ge1233ov20202438;
extern T0* ge1233ov20202439;
extern T0* ge1233ov20202440;
extern T0* ge1233ov20202441;
extern T0* ge1233ov20202442;
extern T0* ge1233ov20202443;
extern T0* ge1233ov20202444;
extern T0* ge1233ov20202445;
extern T0* ge1233ov20202446;
extern T0* ge1233ov20202453;
extern T0* ge1233ov20202454;
extern T0* ge1233ov20202455;
extern T0* ge1233ov20202456;
extern T0* ge1233ov20202457;
extern T0* ge1233ov20202459;
extern T0* ge1233ov20202460;
extern T0* ge1233ov20202461;
extern T0* ge1233ov20202462;
extern T0* ge1233ov20202463;
extern T0* ge1233ov20202464;
extern T0* ge1233ov20202465;
extern T0* ge1233ov20202466;
extern T0* ge1233ov20202467;
extern T0* ge1233ov20202468;
extern T0* ge1233ov20202469;
extern T0* ge1233ov20202470;
extern T0* ge1233ov20202471;
extern T0* ge1233ov20202472;
extern T0* ge1233ov20202473;
extern T0* ge1233ov20202474;
extern T0* ge1233ov20202475;
extern T0* ge1233ov20202476;
extern T0* ge1233ov20202477;
extern T0* ge1233ov20202480;
extern T0* ge1233ov20202481;
extern T0* ge1233ov20202482;
extern T0* ge1233ov20202483;
extern T0* ge1233ov20202484;
extern T0* ge1233ov20202485;
extern T0* ge1233ov20202486;
extern T0* ge1233ov20202487;
extern T0* ge1233ov20202488;
extern T0* ge1233ov20202489;
extern T0* ge1469ov24068168;
extern T0* ge834ov13664281;
extern T0* ge1448ov23724038;
extern T0* ge1448ov23724036;
extern T0* ge1448ov23724034;
extern T0* ge834ov13664274;
extern T0* ge834ov13664315;
extern T0* ge826ov13533194;
extern T0* ge826ov13533195;
extern T0* ge826ov13533204;
extern T0* ge783ov12828674;
extern T0* ge783ov12828681;
extern T0* ge783ov12828680;
extern T0* ge783ov12828675;
extern T0* ge783ov12828679;
extern T0* ge783ov12828676;
extern T0* ge783ov12828682;
extern T0* ge834ov13664266;
extern T0* ge834ov13664271;
extern T0* ge834ov13664298;
extern T0* ge850ov13926485;
extern T0* ge850ov13926455;
extern T0* ge850ov13926423;
extern T0* ge850ov13926457;
extern T0* ge850ov13926479;
extern T0* ge850ov13926446;
extern T0* ge850ov13926469;
extern T0* ge846ov13860930;
extern T0* ge846ov13860895;
extern T0* ge846ov13860939;
extern T0* ge846ov13860904;
extern T0* ge846ov13860922;
extern T0* ge846ov13860887;
extern T0* ge853ov13975620;
extern T0* ge853ov13975619;
extern T0* ge850ov13926430;
extern T0* ge850ov13926431;
extern T0* ge850ov13926403;
extern T0* ge850ov13926410;
extern T0* ge850ov13926414;
extern T0* ge850ov13926432;
extern T0* ge850ov13926433;
extern T0* ge850ov13926434;
extern T0* ge850ov13926435;
extern T0* ge850ov13926436;
extern T0* ge850ov13926437;
extern T0* ge850ov13926438;
extern T0* ge850ov13926441;
extern T0* ge850ov13926461;
extern T0* ge850ov13926445;
extern T0* ge850ov13926448;
extern T0* ge850ov13926460;
extern T0* ge850ov13926464;
extern T0* ge850ov13926465;
extern T0* ge850ov13926466;
extern T0* ge850ov13926474;
extern T0* ge850ov13926475;
extern T0* ge850ov13926487;
extern T0* ge850ov13926488;
extern T0* ge850ov13926401;
extern T0* ge834ov13664276;
extern T0* ge826ov13533193;
extern T0* ge826ov13533203;
extern T0* ge826ov13533202;
extern T0* ge826ov13533228;
extern T0* ge826ov13533231;
extern T0* ge826ov13533230;
extern T0* ge826ov13533205;
extern T0* ge826ov13533221;
extern T0* ge826ov13533196;
extern T0* ge826ov13533191;
extern T0* ge783ov12828673;
extern T0* ge834ov13664265;
extern T0* ge834ov13664293;
extern T0* ge850ov13926447;
extern T0* ge850ov13926452;
extern T0* ge850ov13926472;
extern T0* ge850ov13926409;
extern T0* ge850ov13926468;
extern T0* ge850ov13926416;
extern T0* ge850ov13926420;
extern T0* ge850ov13926480;
extern T0* ge846ov13860928;
extern T0* ge846ov13860893;
extern T0* ge846ov13860925;
extern T0* ge846ov13860890;
extern T0* ge850ov13926486;
extern T0* ge834ov13664267;
extern T0* ge834ov13664305;
extern T0* ge834ov13664311;
extern T0* ge834ov13664306;
extern T0* ge850ov13926402;
extern T0* ge850ov13926411;
extern T0* ge850ov13926439;
extern T0* ge850ov13926458;
extern T0* ge850ov13926419;
extern T0* ge850ov13926412;
extern T0* ge850ov13926473;
extern T0* ge850ov13926489;
extern T0* ge850ov13926426;
extern T0* ge850ov13926467;
extern T0* ge850ov13926483;
extern T0* ge850ov13926413;
extern T0* ge850ov13926440;
extern T0* ge850ov13926459;
extern T0* ge850ov13926456;
extern T0* ge846ov13860923;
extern T0* ge846ov13860888;
extern T0* ge850ov13926418;
extern T0* ge850ov13926490;
extern T0* ge850ov13926421;
extern T0* ge850ov13926424;
extern T0* ge850ov13926425;
extern T0* ge850ov13926454;
extern T0* ge850ov13926463;
extern T0* ge850ov13926451;
extern T0* ge850ov13926450;
extern T0* ge850ov13926481;
extern T0* ge850ov13926417;
extern T0* ge850ov13926453;
extern T0* ge846ov13860912;
extern T0* ge846ov13860877;
extern T0* ge846ov13860911;
extern T0* ge846ov13860876;
extern T0* ge846ov13860910;
extern T0* ge846ov13860875;
extern T0* ge846ov13860909;
extern T0* ge846ov13860874;
extern T0* ge846ov13860908;
extern T0* ge846ov13860873;
extern T0* ge850ov13926429;
extern T0* ge846ov13860927;
extern T0* ge846ov13860892;
extern T0* ge846ov13860929;
extern T0* ge846ov13860894;
extern T0* ge850ov13926449;
extern T0* ge850ov13926415;
extern T0* ge850ov13926476;
extern T0* ge850ov13926492;
extern T0* ge850ov13926408;
extern T0* ge850ov13926422;
extern T0* ge850ov13926491;
extern T0* ge850ov13926427;
extern T0* ge850ov13926462;
extern T0* ge850ov13926484;
extern T0* ge850ov13926443;
extern T0* ge850ov13926470;
extern T0* ge850ov13926442;
extern T0* ge850ov13926428;
extern T0* ge850ov13926444;
extern T0* ge850ov13926477;
extern T0* ge850ov13926482;
extern T0* ge850ov13926404;
extern T0* ge850ov13926405;
extern T0* ge850ov13926406;
extern T0* ge850ov13926407;
extern T0* ge826ov13533223;
extern T0* ge826ov13533232;
extern T0* ge826ov13533197;
extern T0* ge826ov13533226;
extern T0* ge826ov13533208;
extern T0* ge826ov13533192;
extern T0* ge826ov13533198;
extern T0* ge826ov13533200;
extern T0* ge826ov13533201;
extern T0* ge826ov13533206;
extern T0* ge826ov13533185;
extern T0* ge826ov13533186;
extern T0* ge826ov13533187;
extern T0* ge826ov13533188;
extern T0* ge826ov13533189;
extern T0* ge826ov13533190;
extern T0* ge826ov13533207;
extern T0* ge826ov13533209;
extern T0* ge826ov13533210;
extern T0* ge826ov13533211;
extern T0* ge826ov13533212;
extern T0* ge826ov13533213;
extern T0* ge826ov13533214;
extern T0* ge826ov13533215;
extern T0* ge826ov13533216;
extern T0* ge826ov13533217;
extern T0* ge826ov13533218;
extern T0* ge826ov13533219;
extern T0* ge826ov13533220;
extern T0* ge834ov13664257;
extern T0* ge834ov13664258;
extern T0* ge834ov13664259;
extern T0* ge834ov13664260;
extern T0* ge834ov13664261;
extern T0* ge834ov13664262;
extern T0* ge834ov13664263;
extern T0* ge834ov13664264;
extern T0* ge834ov13664301;
extern T0* ge834ov13664268;
extern T0* ge834ov13664269;
extern T0* ge834ov13664270;
extern T0* ge834ov13664275;
extern T0* ge834ov13664277;
extern T0* ge834ov13664278;
extern T0* ge834ov13664279;
extern T0* ge834ov13664280;
extern T0* ge834ov13664282;
extern T0* ge834ov13664283;
extern T0* ge834ov13664302;
extern T0* ge834ov13664289;
extern T0* ge834ov13664290;
extern T0* ge834ov13664304;
extern T0* ge834ov13664292;
extern T0* ge834ov13664294;
extern T0* ge834ov13664295;
extern T0* ge834ov13664299;
extern T0* ge834ov13664300;
extern T0* ge846ov13860924;
extern T0* ge846ov13860889;
extern T0* ge846ov13860926;
extern T0* ge846ov13860891;
extern T0* ge846ov13860920;
extern T0* ge846ov13860885;
extern T0* ge783ov12828677;
extern T0* ge783ov12828678;
extern T0* ge846ov13860914;
extern T0* ge846ov13860879;
extern T0* ge846ov13860913;
extern T0* ge846ov13860878;
extern T0* ge846ov13860931;
extern T0* ge846ov13860896;
extern T0* ge846ov13860932;
extern T0* ge846ov13860897;
extern T0* ge846ov13860933;
extern T0* ge846ov13860898;
extern T0* ge846ov13860921;
extern T0* ge846ov13860886;
extern T0* ge846ov13860934;
extern T0* ge846ov13860899;
extern T0* ge846ov13860907;
extern T0* ge846ov13860872;
extern T0* ge846ov13860936;
extern T0* ge846ov13860901;
extern T0* ge846ov13860937;
extern T0* ge846ov13860902;
extern T0* ge846ov13860938;
extern T0* ge846ov13860903;
extern T0* ge846ov13860935;
extern T0* ge846ov13860900;
extern T0* ge850ov13926478;
extern T0* ge850ov13926471;
extern T0* ge846ov13860906;
extern T0* ge846ov13860871;
extern T0* ge846ov13860915;
extern T0* ge846ov13860880;
extern T0* ge846ov13860916;
extern T0* ge846ov13860881;
extern T0* ge846ov13860919;
extern T0* ge846ov13860884;
extern T0* ge846ov13860918;
extern T0* ge846ov13860883;
extern T0* ge846ov13860917;
extern T0* ge846ov13860882;
extern T0* ge1725ov28262442;
extern T0* ge1725ov28262404;
extern T0* ge1725ov28262402;
extern T0* ge1725ov28262443;
extern T0* ge1725ov28262415;
extern T0* ge1725ov28262414;
extern T0* ge1725ov28262425;
extern T0* ge1725ov28262419;
extern T0* ge1725ov28262418;
extern T0* ge1725ov28262417;
extern T0* ge1725ov28262423;
extern T0* ge1725ov28262422;
extern T0* ge1725ov28262424;
extern T0* ge1725ov28262401;
extern T0* ge1725ov28262427;
extern T0* ge1725ov28262436;
extern T0* ge1728ov28311579;
extern T0* ge1728ov28311577;
extern T0* ge1728ov28311578;
extern T0* ge1725ov28262437;
extern T0* ge1725ov28262438;
extern T0* ge1725ov28262441;
extern T0* ge1725ov28262439;
extern T0* ge1725ov28262440;
extern T0* ge1725ov28262434;
extern T0* ge1721ov28196865;
extern T0* ge1721ov28196866;
extern T0* ge1725ov28262407;
extern T0* ge1725ov28262430;
extern T0* ge1736ov28442641;
extern T0* ge1736ov28442640;
extern T0* ge1725ov28262450;
extern T0* ge1725ov28262451;
extern T0* ge1725ov28262452;
extern T0* ge1725ov28262445;
extern T0* ge1725ov28262428;
extern T0* ge1725ov28262429;
extern T0* ge1725ov28262431;
extern T0* ge1725ov28262435;
extern T0* ge1780ov29163521;
extern T0* ge318ov5210114;
extern T0* ge318ov5210121;
extern T0* ge313ov5128214;
extern T0* ge420ov6881281;
extern T0* ge315ov5160967;
extern T0* ge318ov5210117;
extern T0* ge318ov5210124;
extern T0* ge318ov5210115;
extern T0* ge318ov5210122;
extern T0* ge318ov5210116;
extern T0* ge318ov5210123;
extern T0* ge318ov5210118;
extern T0* ge318ov5210125;
extern T0* ge318ov5210119;
extern T0* ge318ov5210126;
extern T0* ge315ov5160962;
extern T0* ge315ov5160961;
extern T0* ge315ov5160969;
extern T0* ge315ov5160968;
extern T0* ge1509ov24723459;
extern T0* ge1509ov24723457;
extern T0* ge1732ov28377107;
extern T0* ge1732ov28377108;
extern T0* ge1723ov28229636;
extern T0* ge1723ov28229638;
extern T0* ge1233ov20202449;
extern T0* ge1233ov20202450;
extern T0* ge1233ov20202452;
extern T0* ge1233ov20202451;
extern T0* ge1233ov20202447;
extern T0* ge1233ov20202448;
extern T0* ge1331ov21807131;
extern T0* ge1324ov21692424;
extern T0* ge1320ov21626920;
extern T0* ge1457ov23872015;
extern T0* ge1457ov23871843;
extern T0* ge1457ov23871523;
extern T0* ge1457ov23872016;
extern T0* ge1457ov23871524;
extern T0* ge1457ov23871999;
extern T0* ge1457ov23871836;
extern T0* ge1457ov23871507;
extern T0* ge1457ov23871998;
extern T0* ge1457ov23871506;
extern T0* ge1457ov23872000;
extern T0* ge1457ov23871508;
extern T0* ge1457ov23872024;
extern T0* ge1457ov23871847;
extern T0* ge1457ov23871532;
extern T0* ge1457ov23872013;
extern T0* ge1457ov23871521;
extern T0* ge1457ov23872014;
extern T0* ge1457ov23871522;
extern T0* ge1457ov23872161;
extern T0* ge1457ov23871900;
extern T0* ge1457ov23871666;
extern T0* ge1457ov23872160;
extern T0* ge1457ov23871665;
extern T0* ge1457ov23872162;
extern T0* ge1457ov23871901;
extern T0* ge1457ov23871667;
extern T0* ge1457ov23872028;
extern T0* ge1457ov23871849;
extern T0* ge1457ov23871536;
extern T0* ge1457ov23872034;
extern T0* ge1457ov23871853;
extern T0* ge1457ov23871542;
extern T0* ge1457ov23872153;
extern T0* ge1457ov23871896;
extern T0* ge1457ov23871658;
extern T0* ge1457ov23872140;
extern T0* ge1457ov23871888;
extern T0* ge1457ov23871645;
extern T0* ge1457ov23872163;
extern T0* ge1457ov23871902;
extern T0* ge1457ov23871668;
extern T0* ge1457ov23872031;
extern T0* ge1457ov23871851;
extern T0* ge1457ov23871539;
extern T0* ge1457ov23872038;
extern T0* ge1457ov23871856;
extern T0* ge1457ov23871546;
extern T0* ge1457ov23872154;
extern T0* ge1457ov23871897;
extern T0* ge1457ov23871659;
extern T0* ge1457ov23872152;
extern T0* ge1457ov23871895;
extern T0* ge1457ov23871657;
extern T0* ge1457ov23872141;
extern T0* ge1457ov23871889;
extern T0* ge1457ov23871646;
extern T0* ge1457ov23872155;
extern T0* ge1457ov23871898;
extern T0* ge1457ov23871660;
extern T0* ge1457ov23872125;
extern T0* ge1457ov23871884;
extern T0* ge1457ov23871630;
extern T0* ge1457ov23872124;
extern T0* ge1457ov23871629;
extern T0* ge1457ov23872126;
extern T0* ge1457ov23871631;
extern T0* ge1457ov23872127;
extern T0* ge1457ov23871632;
extern T0* ge1457ov23872115;
extern T0* ge1457ov23871881;
extern T0* ge1457ov23871621;
extern T0* ge1457ov23872095;
extern T0* ge1457ov23871868;
extern T0* ge1457ov23871603;
extern T0* ge1457ov23872088;
extern T0* ge1457ov23871867;
extern T0* ge1457ov23871596;
extern T0* ge1457ov23872122;
extern T0* ge1457ov23871883;
extern T0* ge1457ov23871628;
extern T0* ge1457ov23872081;
extern T0* ge1457ov23871866;
extern T0* ge1457ov23871589;
extern T0* ge1457ov23872080;
extern T0* ge1457ov23871588;
extern T0* ge1457ov23872094;
extern T0* ge1457ov23871602;
extern T0* ge1457ov23872087;
extern T0* ge1457ov23871595;
extern T0* ge1457ov23872121;
extern T0* ge1457ov23871627;
extern T0* ge1457ov23872079;
extern T0* ge1457ov23871587;
extern T0* ge1457ov23872078;
extern T0* ge1457ov23871586;
extern T0* ge1457ov23872096;
extern T0* ge1457ov23871869;
extern T0* ge1457ov23871604;
extern T0* ge1457ov23872077;
extern T0* ge1457ov23871585;
extern T0* ge1457ov23872075;
extern T0* ge1457ov23871583;
extern T0* ge1457ov23872074;
extern T0* ge1457ov23871582;
extern T0* ge1457ov23872093;
extern T0* ge1457ov23871601;
extern T0* ge1457ov23872086;
extern T0* ge1457ov23871594;
extern T0* ge1457ov23872073;
extern T0* ge1457ov23871581;
extern T0* ge1457ov23872076;
extern T0* ge1457ov23871584;
extern T0* ge1457ov23872117;
extern T0* ge1457ov23871623;
extern T0* ge1457ov23872120;
extern T0* ge1457ov23871626;
extern T0* ge1457ov23872119;
extern T0* ge1457ov23871625;
extern T0* ge1457ov23872118;
extern T0* ge1457ov23871624;
extern T0* ge1457ov23872293;
extern T0* ge1457ov23871970;
extern T0* ge1457ov23871796;
extern T0* ge1457ov23872111;
extern T0* ge1457ov23871880;
extern T0* ge1457ov23871617;
extern T0* ge1457ov23872112;
extern T0* ge1457ov23871618;
extern T0* ge1457ov23872105;
extern T0* ge1457ov23871877;
extern T0* ge1457ov23871611;
extern T0* ge1457ov23872114;
extern T0* ge1457ov23871620;
extern T0* ge1457ov23872113;
extern T0* ge1457ov23871619;
extern T0* ge1457ov23872110;
extern T0* ge1457ov23871879;
extern T0* ge1457ov23871616;
extern T0* ge1457ov23872061;
extern T0* ge1457ov23871865;
extern T0* ge1457ov23871569;
extern T0* ge1457ov23872060;
extern T0* ge1457ov23871864;
extern T0* ge1457ov23871568;
extern T0* ge1457ov23872059;
extern T0* ge1457ov23871863;
extern T0* ge1457ov23871567;
extern T0* ge1457ov23872058;
extern T0* ge1457ov23871862;
extern T0* ge1457ov23871566;
extern T0* ge1457ov23872057;
extern T0* ge1457ov23871565;
extern T0* ge1457ov23871995;
extern T0* ge1457ov23871833;
extern T0* ge1457ov23871503;
extern T0* ge1457ov23871993;
extern T0* ge1457ov23871832;
extern T0* ge1457ov23871501;
extern T0* ge1457ov23872067;
extern T0* ge1457ov23871575;
extern T0* ge1457ov23872064;
extern T0* ge1457ov23871572;
extern T0* ge1457ov23872090;
extern T0* ge1457ov23871598;
extern T0* ge1457ov23872083;
extern T0* ge1457ov23871591;
extern T0* ge1457ov23872068;
extern T0* ge1457ov23871576;
extern T0* ge1457ov23872065;
extern T0* ge1457ov23871573;
extern T0* ge1457ov23872091;
extern T0* ge1457ov23871599;
extern T0* ge1457ov23872084;
extern T0* ge1457ov23871592;
extern T0* ge1457ov23872069;
extern T0* ge1457ov23871577;
extern T0* ge1457ov23872066;
extern T0* ge1457ov23871574;
extern T0* ge1457ov23872092;
extern T0* ge1457ov23871600;
extern T0* ge1457ov23872085;
extern T0* ge1457ov23871593;
extern T0* ge1457ov23872219;
extern T0* ge1457ov23871930;
extern T0* ge1457ov23871724;
extern T0* ge1457ov23872101;
extern T0* ge1457ov23871871;
extern T0* ge1457ov23871607;
extern T0* ge1457ov23872100;
extern T0* ge1457ov23871606;
extern T0* ge1457ov23872072;
extern T0* ge1457ov23871580;
extern T0* ge1457ov23872063;
extern T0* ge1457ov23871571;
extern T0* ge1457ov23872062;
extern T0* ge1457ov23871570;
extern T0* ge1457ov23872089;
extern T0* ge1457ov23871597;
extern T0* ge1457ov23872082;
extern T0* ge1457ov23871590;
extern T0* ge1457ov23872071;
extern T0* ge1457ov23871579;
extern T0* ge1457ov23872070;
extern T0* ge1457ov23871578;
extern T0* ge1457ov23872097;
extern T0* ge1457ov23871870;
extern T0* ge1457ov23871605;
extern T0* ge1457ov23872208;
extern T0* ge1457ov23871924;
extern T0* ge1457ov23871713;
extern T0* ge1457ov23872206;
extern T0* ge1457ov23871923;
extern T0* ge1457ov23871711;
extern T0* ge1457ov23872002;
extern T0* ge1457ov23871837;
extern T0* ge1457ov23871510;
extern T0* ge1457ov23872159;
extern T0* ge1457ov23871899;
extern T0* ge1457ov23871664;
extern T0* ge1457ov23872001;
extern T0* ge1457ov23871509;
extern T0* ge1457ov23872158;
extern T0* ge1457ov23871663;
extern T0* ge1457ov23872157;
extern T0* ge1457ov23871662;
extern T0* ge1457ov23872032;
extern T0* ge1457ov23871852;
extern T0* ge1457ov23871540;
extern T0* ge1457ov23872030;
extern T0* ge1457ov23871850;
extern T0* ge1457ov23871538;
extern T0* ge1457ov23872029;
extern T0* ge1457ov23871537;
extern T0* ge1457ov23872036;
extern T0* ge1457ov23871854;
extern T0* ge1457ov23871544;
extern T0* ge1457ov23872035;
extern T0* ge1457ov23871543;
extern T0* ge1457ov23872037;
extern T0* ge1457ov23871855;
extern T0* ge1457ov23871545;
extern T0* ge1457ov23872018;
extern T0* ge1457ov23871844;
extern T0* ge1457ov23871526;
extern T0* ge1457ov23872017;
extern T0* ge1457ov23871525;
extern T0* ge1457ov23872023;
extern T0* ge1457ov23871531;
extern T0* ge1457ov23872020;
extern T0* ge1457ov23871845;
extern T0* ge1457ov23871528;
extern T0* ge1457ov23872021;
extern T0* ge1457ov23871529;
extern T0* ge1457ov23872156;
extern T0* ge1457ov23871661;
extern T0* ge1457ov23872019;
extern T0* ge1457ov23871527;
extern T0* ge1457ov23872033;
extern T0* ge1457ov23871541;
extern T0* ge1457ov23872022;
extern T0* ge1457ov23871846;
extern T0* ge1457ov23871530;
extern T0* ge1457ov23872027;
extern T0* ge1457ov23871848;
extern T0* ge1457ov23871535;
extern T0* ge1457ov23872025;
extern T0* ge1457ov23871533;
extern T0* ge1457ov23872026;
extern T0* ge1457ov23871534;
extern T0* ge1457ov23872164;
extern T0* ge1457ov23871903;
extern T0* ge1457ov23871669;
extern T0* ge1457ov23872136;
extern T0* ge1457ov23871872;
extern T0* ge1457ov23871641;
extern T0* ge1457ov23871997;
extern T0* ge1457ov23871835;
extern T0* ge1457ov23871505;
extern T0* ge1457ov23871996;
extern T0* ge1457ov23871834;
extern T0* ge1457ov23871504;
extern T0* ge1457ov23872283;
extern T0* ge1457ov23871959;
extern T0* ge1457ov23871786;
extern T0* ge1457ov23872138;
extern T0* ge1457ov23871873;
extern T0* ge1457ov23871643;
extern T0* ge1457ov23872286;
extern T0* ge1457ov23871963;
extern T0* ge1457ov23871789;
extern T0* ge1457ov23872224;
extern T0* ge1457ov23871933;
extern T0* ge1457ov23871729;
extern T0* ge1457ov23872285;
extern T0* ge1457ov23871962;
extern T0* ge1457ov23871788;
extern T0* ge1457ov23872135;
extern T0* ge1457ov23871886;
extern T0* ge1457ov23871640;
extern T0* ge1457ov23872137;
extern T0* ge1457ov23871642;
extern T0* ge1447ov23707662;
extern T0* ge1447ov23707657;
extern T0* ge1447ov23707652;
extern T0* ge1447ov23707663;
extern T0* ge1447ov23707658;
extern T0* ge1447ov23707653;
extern T0* ge1457ov23872296;
extern T0* ge1457ov23871972;
extern T0* ge1457ov23871799;
extern T0* ge1457ov23872295;
extern T0* ge1457ov23871798;
extern T0* ge478ov7831564;
extern T0* ge1457ov23872051;
extern T0* ge1457ov23871860;
extern T0* ge1457ov23871559;
extern T0* ge1457ov23872050;
extern T0* ge1457ov23871558;
extern T0* ge1457ov23872049;
extern T0* ge1457ov23871859;
extern T0* ge1457ov23871557;
extern T0* ge1457ov23872048;
extern T0* ge1457ov23871556;
extern T0* ge1457ov23872052;
extern T0* ge1457ov23871861;
extern T0* ge1457ov23871560;
extern T0* ge1457ov23872043;
extern T0* ge1457ov23871858;
extern T0* ge1457ov23871551;
extern T0* ge1457ov23872045;
extern T0* ge1457ov23871553;
extern T0* ge1457ov23872241;
extern T0* ge1457ov23871939;
extern T0* ge1457ov23871744;
extern T0* ge1457ov23872133;
extern T0* ge1457ov23871885;
extern T0* ge1457ov23871638;
extern T0* ge1457ov23872132;
extern T0* ge1457ov23871637;
extern T0* ge1457ov23872247;
extern T0* ge1457ov23871941;
extern T0* ge1457ov23871750;
extern T0* ge1457ov23872232;
extern T0* ge1457ov23871936;
extern T0* ge1457ov23871735;
extern T0* ge1457ov23872263;
extern T0* ge1457ov23871947;
extern T0* ge1457ov23871766;
extern T0* ge1457ov23872147;
extern T0* ge1457ov23871893;
extern T0* ge1457ov23871652;
extern T0* ge1457ov23872234;
extern T0* ge1457ov23871937;
extern T0* ge1457ov23871737;
extern T0* ge1457ov23872233;
extern T0* ge1457ov23871736;
extern T0* ge1457ov23872012;
extern T0* ge1457ov23871842;
extern T0* ge1457ov23871520;
extern T0* ge1457ov23871988;
extern T0* ge1457ov23871828;
extern T0* ge1457ov23871496;
extern T0* ge1457ov23872231;
extern T0* ge1457ov23871734;
extern T0* ge1457ov23872230;
extern T0* ge1457ov23871733;
extern T0* ge1457ov23872011;
extern T0* ge1457ov23871841;
extern T0* ge1457ov23871519;
extern T0* ge1457ov23871987;
extern T0* ge1457ov23871827;
extern T0* ge1457ov23871495;
extern T0* ge1457ov23872248;
extern T0* ge1457ov23871751;
extern T0* ge1457ov23872262;
extern T0* ge1457ov23871765;
extern T0* ge1457ov23872150;
extern T0* ge1457ov23871894;
extern T0* ge1457ov23871655;
extern T0* ge1457ov23871986;
extern T0* ge1457ov23871494;
extern T0* ge1457ov23872239;
extern T0* ge1457ov23871742;
extern T0* ge1457ov23872128;
extern T0* ge1457ov23871633;
extern T0* ge1457ov23872131;
extern T0* ge1457ov23871636;
extern T0* ge1457ov23872130;
extern T0* ge1457ov23871635;
extern T0* ge1457ov23872129;
extern T0* ge1457ov23871634;
extern T0* ge1457ov23872053;
extern T0* ge1457ov23871561;
extern T0* ge1457ov23872041;
extern T0* ge1457ov23871549;
extern T0* ge1457ov23872047;
extern T0* ge1457ov23871555;
extern T0* ge1457ov23872044;
extern T0* ge1457ov23871552;
extern T0* ge1457ov23872042;
extern T0* ge1457ov23871550;
extern T0* ge1458ov23888222;
extern T0* ge1458ov23888376;
extern T0* ge1458ov23888384;
extern T0* ge1457ov23872222;
extern T0* ge1457ov23871932;
extern T0* ge1457ov23871727;
extern T0* ge1457ov23872223;
extern T0* ge1457ov23871728;
extern T0* ge1457ov23872210;
extern T0* ge1457ov23871925;
extern T0* ge1457ov23871715;
extern T0* ge1457ov23872211;
extern T0* ge1457ov23871926;
extern T0* ge1457ov23871716;
extern T0* ge1457ov23872207;
extern T0* ge1457ov23871712;
extern T0* ge1457ov23872258;
extern T0* ge1457ov23871945;
extern T0* ge1457ov23871761;
extern T0* ge1458ov23888108;
extern T0* ge1458ov23888360;
extern T0* ge1458ov23888264;
extern T0* ge1458ov23888377;
extern T0* ge1458ov23888343;
extern T0* ge1458ov23888266;
extern T0* ge1458ov23888120;
extern T0* ge1458ov23888168;
extern T0* ge1458ov23888265;
extern T0* ge1458ov23888366;
extern T0* ge1458ov23888352;
extern T0* ge1458ov23888386;
extern T0* ge1458ov23888170;
extern T0* ge1458ov23888074;
extern T0* ge1458ov23888370;
extern T0* ge1458ov23888367;
extern T0* ge1458ov23888186;
extern T0* ge1458ov23888221;
extern T0* ge1458ov23888356;
extern T0* ge1458ov23888184;
extern T0* ge1458ov23888185;
extern T0* ge1458ov23888224;
extern T0* ge1457ov23872257;
extern T0* ge1457ov23871944;
extern T0* ge1457ov23871760;
extern T0* ge1457ov23872261;
extern T0* ge1457ov23871946;
extern T0* ge1457ov23871764;
extern T0* ge1457ov23872260;
extern T0* ge1457ov23871763;
extern T0* ge1457ov23872256;
extern T0* ge1457ov23871759;
extern T0* ge1457ov23872255;
extern T0* ge1457ov23871758;
extern T0* ge1457ov23872254;
extern T0* ge1457ov23871757;
extern T0* ge1457ov23872253;
extern T0* ge1457ov23871756;
extern T0* ge1457ov23872252;
extern T0* ge1457ov23871755;
extern T0* ge1457ov23872321;
extern T0* ge1457ov23871982;
extern T0* ge1457ov23871824;
extern T0* ge1457ov23872278;
extern T0* ge1457ov23871955;
extern T0* ge1457ov23871781;
extern T0* ge1458ov23888182;
extern T0* ge1458ov23888136;
extern T0* ge1458ov23888135;
extern T0* ge1458ov23888181;
extern T0* ge1457ov23871990;
extern T0* ge1457ov23871829;
extern T0* ge1457ov23871498;
extern T0* ge1457ov23872148;
extern T0* ge1457ov23871653;
extern T0* ge1457ov23872170;
extern T0* ge1457ov23871905;
extern T0* ge1457ov23871675;
extern T0* ge1457ov23872169;
extern T0* ge1457ov23871674;
extern T0* ge1457ov23872168;
extern T0* ge1457ov23871673;
extern T0* ge1457ov23872167;
extern T0* ge1457ov23871672;
extern T0* ge1457ov23872166;
extern T0* ge1457ov23871671;
extern T0* ge1457ov23872165;
extern T0* ge1457ov23871904;
extern T0* ge1457ov23871670;
extern T0* ge1458ov23888387;
extern T0* ge1457ov23872134;
extern T0* ge1457ov23871639;
extern T0* ge1457ov23872175;
extern T0* ge1457ov23871908;
extern T0* ge1457ov23871680;
extern T0* ge1457ov23872182;
extern T0* ge1457ov23871912;
extern T0* ge1457ov23871687;
extern T0* ge1457ov23872181;
extern T0* ge1457ov23871686;
extern T0* ge1457ov23872180;
extern T0* ge1457ov23871911;
extern T0* ge1457ov23871685;
extern T0* ge1457ov23872179;
extern T0* ge1457ov23871684;
extern T0* ge1457ov23872178;
extern T0* ge1457ov23871910;
extern T0* ge1457ov23871683;
extern T0* ge1457ov23872177;
extern T0* ge1457ov23871682;
extern T0* ge1457ov23872176;
extern T0* ge1457ov23871909;
extern T0* ge1457ov23871681;
extern T0* ge1457ov23872174;
extern T0* ge1457ov23871679;
extern T0* ge1457ov23872245;
extern T0* ge1457ov23871748;
extern T0* ge1457ov23872274;
extern T0* ge1457ov23871954;
extern T0* ge1457ov23871777;
extern T0* ge1457ov23872273;
extern T0* ge1457ov23871953;
extern T0* ge1457ov23871776;
extern T0* ge1458ov23888346;
extern T0* ge1457ov23872250;
extern T0* ge1457ov23871943;
extern T0* ge1457ov23871753;
extern T0* ge1457ov23872251;
extern T0* ge1457ov23871754;
extern T0* ge1457ov23872249;
extern T0* ge1457ov23871942;
extern T0* ge1457ov23871752;
extern T0* ge1457ov23872281;
extern T0* ge1457ov23871957;
extern T0* ge1457ov23871784;
extern T0* ge1457ov23872279;
extern T0* ge1457ov23871956;
extern T0* ge1457ov23871782;
extern T0* ge1457ov23872276;
extern T0* ge1457ov23871779;
extern T0* ge1457ov23872320;
extern T0* ge1457ov23871823;
extern T0* ge1457ov23872277;
extern T0* ge1457ov23871780;
extern T0* ge1458ov23888189;
extern T0* ge1457ov23872054;
extern T0* ge1457ov23871562;
extern T0* ge1457ov23872237;
extern T0* ge1457ov23871740;
extern T0* ge1233ov20202557;
extern T0* ge1233ov20202559;
extern T0* ge1233ov20202560;
extern T0* ge1457ov23872267;
extern T0* ge1457ov23871949;
extern T0* ge1457ov23871770;
extern T0* ge1227ov20103178;
extern T0* ge1227ov20103177;
extern T0* ge1457ov23872265;
extern T0* ge1457ov23871948;
extern T0* ge1457ov23871768;
extern T0* ge1457ov23872264;
extern T0* ge1457ov23871767;
extern T0* ge1227ov20103179;
extern T0* ge1457ov23872268;
extern T0* ge1457ov23871771;
extern T0* ge1457ov23872266;
extern T0* ge1457ov23871769;
extern T0* ge1457ov23872288;
extern T0* ge1457ov23871965;
extern T0* ge1457ov23871791;
extern T0* ge1457ov23872287;
extern T0* ge1457ov23871964;
extern T0* ge1457ov23871790;
extern T0* ge1458ov23888061;
extern T0* ge1458ov23888049;
extern T0* ge1458ov23888365;
extern T0* ge1458ov23888073;
extern T0* ge1458ov23888342;
extern T0* ge1457ov23872220;
extern T0* ge1457ov23871725;
extern T0* ge1457ov23872216;
extern T0* ge1457ov23871929;
extern T0* ge1457ov23871721;
extern T0* ge1457ov23872217;
extern T0* ge1457ov23871722;
extern T0* ge1457ov23872116;
extern T0* ge1457ov23871882;
extern T0* ge1457ov23871622;
extern T0* ge1457ov23872139;
extern T0* ge1457ov23871887;
extern T0* ge1457ov23871644;
extern T0* ge1457ov23872218;
extern T0* ge1457ov23871723;
extern T0* ge1457ov23872007;
extern T0* ge1457ov23871840;
extern T0* ge1457ov23871515;
extern T0* ge1457ov23872006;
extern T0* ge1457ov23871514;
extern T0* ge1457ov23872005;
extern T0* ge1457ov23871839;
extern T0* ge1457ov23871513;
extern T0* ge1457ov23872243;
extern T0* ge1457ov23871746;
extern T0* ge1457ov23872009;
extern T0* ge1457ov23871517;
extern T0* ge1457ov23872010;
extern T0* ge1457ov23871518;
extern T0* ge1457ov23872004;
extern T0* ge1457ov23871838;
extern T0* ge1457ov23871512;
extern T0* ge1457ov23872235;
extern T0* ge1457ov23871938;
extern T0* ge1457ov23871738;
extern T0* ge1457ov23872240;
extern T0* ge1457ov23871743;
extern T0* ge1458ov23888351;
extern T0* ge1457ov23872242;
extern T0* ge1457ov23871745;
extern T0* ge1457ov23871554;
extern T0* ge1457ov23871985;
extern T0* ge1457ov23871826;
extern T0* ge1457ov23871493;
extern T0* ge1457ov23872055;
extern T0* ge1457ov23871563;
extern T0* ge1457ov23872190;
extern T0* ge1457ov23871913;
extern T0* ge1457ov23871695;
extern T0* ge1457ov23872188;
extern T0* ge1457ov23871693;
extern T0* ge1457ov23872186;
extern T0* ge1457ov23871691;
extern T0* ge1457ov23872185;
extern T0* ge1457ov23871690;
extern T0* ge1457ov23872213;
extern T0* ge1457ov23871927;
extern T0* ge1457ov23871718;
extern T0* ge1457ov23872215;
extern T0* ge1457ov23871928;
extern T0* ge1457ov23871720;
extern T0* ge1457ov23872205;
extern T0* ge1457ov23871922;
extern T0* ge1457ov23871710;
extern T0* ge1457ov23872259;
extern T0* ge1457ov23871762;
extern T0* ge1457ov23872189;
extern T0* ge1457ov23871694;
extern T0* ge1457ov23872187;
extern T0* ge1457ov23871692;
extern T0* ge1457ov23872184;
extern T0* ge1457ov23871689;
extern T0* ge1457ov23872183;
extern T0* ge1457ov23871688;
extern T0* ge1457ov23872209;
extern T0* ge1457ov23871714;
extern T0* ge1457ov23872204;
extern T0* ge1457ov23871709;
extern T0* ge1457ov23872244;
extern T0* ge1457ov23871747;
extern T0* ge1457ov23872191;
extern T0* ge1457ov23871914;
extern T0* ge1457ov23871696;
extern T0* ge1457ov23872192;
extern T0* ge1457ov23871697;
extern T0* ge1457ov23872221;
extern T0* ge1457ov23871931;
extern T0* ge1457ov23871726;
extern T0* ge1457ov23872297;
extern T0* ge1457ov23871973;
extern T0* ge1457ov23871800;
extern T0* ge1457ov23872298;
extern T0* ge1457ov23871801;
extern T0* ge1457ov23872229;
extern T0* ge1457ov23871935;
extern T0* ge1457ov23871732;
extern T0* ge1457ov23872228;
extern T0* ge1457ov23871934;
extern T0* ge1457ov23871731;
extern T0* ge1457ov23872284;
extern T0* ge1457ov23871961;
extern T0* ge1457ov23871787;
extern T0* ge1457ov23872246;
extern T0* ge1457ov23871940;
extern T0* ge1457ov23871749;
extern T0* ge1457ov23872151;
extern T0* ge1457ov23871656;
extern T0* ge1457ov23872308;
extern T0* ge1457ov23871977;
extern T0* ge1457ov23871811;
extern T0* ge1457ov23872306;
extern T0* ge1457ov23871809;
extern T0* ge1457ov23872307;
extern T0* ge1457ov23871810;
extern T0* ge1457ov23872303;
extern T0* ge1457ov23871975;
extern T0* ge1457ov23871806;
extern T0* ge1457ov23872301;
extern T0* ge1457ov23871804;
extern T0* ge1457ov23872302;
extern T0* ge1457ov23871805;
extern T0* ge1457ov23872304;
extern T0* ge1457ov23871976;
extern T0* ge1457ov23871807;
extern T0* ge1457ov23872305;
extern T0* ge1457ov23871808;
extern T0* ge1457ov23872299;
extern T0* ge1457ov23871974;
extern T0* ge1457ov23871802;
extern T0* ge1457ov23872300;
extern T0* ge1457ov23871803;
extern T0* ge1457ov23872272;
extern T0* ge1457ov23871952;
extern T0* ge1457ov23871775;
extern T0* ge1233ov20202478;
extern T0* ge1233ov20202490;
extern T0* ge1457ov23872271;
extern T0* ge1457ov23871774;
extern T0* ge1457ov23872106;
extern T0* ge1457ov23871612;
extern T0* ge1457ov23872107;
extern T0* ge1457ov23871613;
extern T0* ge1457ov23871989;
extern T0* ge1457ov23871497;
extern T0* ge1457ov23872109;
extern T0* ge1457ov23871878;
extern T0* ge1457ov23871615;
extern T0* ge1457ov23872102;
extern T0* ge1457ov23871874;
extern T0* ge1457ov23871608;
extern T0* ge1457ov23872104;
extern T0* ge1457ov23871876;
extern T0* ge1457ov23871610;
extern T0* ge1457ov23872236;
extern T0* ge1457ov23871739;
extern T0* ge1457ov23872056;
extern T0* ge1457ov23871564;
extern T0* ge1457ov23872199;
extern T0* ge1457ov23871919;
extern T0* ge1457ov23871704;
extern T0* ge1457ov23872200;
extern T0* ge1457ov23871705;
extern T0* ge1457ov23872198;
extern T0* ge1457ov23871918;
extern T0* ge1457ov23871703;
extern T0* ge1457ov23872196;
extern T0* ge1457ov23871917;
extern T0* ge1457ov23871701;
extern T0* ge1457ov23872197;
extern T0* ge1457ov23871702;
extern T0* ge1457ov23872319;
extern T0* ge1457ov23871822;
extern T0* ge1457ov23872275;
extern T0* ge1457ov23871778;
extern T0* ge1457ov23872194;
extern T0* ge1457ov23871916;
extern T0* ge1457ov23871699;
extern T0* ge1457ov23872195;
extern T0* ge1457ov23871700;
extern T0* ge1457ov23872193;
extern T0* ge1457ov23871915;
extern T0* ge1457ov23871698;
extern T0* ge1457ov23872201;
extern T0* ge1457ov23871921;
extern T0* ge1457ov23871706;
extern T0* ge1458ov23888139;
extern T0* ge1458ov23888140;
extern T0* ge1458ov23888178;
extern T0* ge1458ov23888141;
extern T0* ge1458ov23888090;
extern T0* ge1458ov23888179;
extern T0* ge1458ov23888142;
extern T0* ge1458ov23888187;
extern T0* ge1458ov23888188;
extern T0* ge1458ov23888091;
extern T0* ge1458ov23888190;
extern T0* ge1457ov23872282;
extern T0* ge1457ov23871958;
extern T0* ge1457ov23871785;
extern T0* ge1458ov23888072;
extern T0* ge1458ov23888075;
extern T0* ge1458ov23888337;
extern T0* ge1458ov23888119;
extern T0* ge1457ov23872172;
extern T0* ge1457ov23871907;
extern T0* ge1457ov23871677;
extern T0* ge1457ov23872173;
extern T0* ge1457ov23871678;
extern T0* ge1457ov23872171;
extern T0* ge1457ov23871906;
extern T0* ge1457ov23871676;
extern T0* ge1458ov23888123;
extern T0* ge1457ov23871991;
extern T0* ge1457ov23871830;
extern T0* ge1457ov23871499;
extern T0* ge1457ov23871992;
extern T0* ge1457ov23871831;
extern T0* ge1457ov23871500;
extern T0* ge1457ov23872142;
extern T0* ge1457ov23871890;
extern T0* ge1457ov23871647;
extern T0* ge1457ov23872039;
extern T0* ge1457ov23871857;
extern T0* ge1457ov23871547;
extern T0* ge1457ov23872040;
extern T0* ge1457ov23871548;
extern T0* ge1457ov23872144;
extern T0* ge1457ov23871891;
extern T0* ge1457ov23871649;
extern T0* ge1457ov23872145;
extern T0* ge1457ov23871650;
extern T0* ge1457ov23872238;
extern T0* ge1457ov23871741;
extern T0* ge1457ov23872143;
extern T0* ge1457ov23871648;
extern T0* ge1457ov23872146;
extern T0* ge1457ov23871892;
extern T0* ge1457ov23871651;
extern T0* ge1457ov23872108;
extern T0* ge1457ov23871614;
extern T0* ge1457ov23872103;
extern T0* ge1457ov23871875;
extern T0* ge1457ov23871609;
extern T0* ge1457ov23872003;
extern T0* ge1457ov23871511;
extern T0* ge1457ov23872149;
extern T0* ge1457ov23871654;
extern T0* ge1457ov23872318;
extern T0* ge1457ov23871981;
extern T0* ge1457ov23871821;
extern T0* ge1457ov23872316;
extern T0* ge1457ov23871819;
extern T0* ge1457ov23872317;
extern T0* ge1457ov23871820;
extern T0* ge1457ov23872313;
extern T0* ge1457ov23871979;
extern T0* ge1457ov23871816;
extern T0* ge1457ov23872311;
extern T0* ge1457ov23871814;
extern T0* ge1457ov23872312;
extern T0* ge1457ov23871815;
extern T0* ge1457ov23872314;
extern T0* ge1457ov23871980;
extern T0* ge1457ov23871817;
extern T0* ge1457ov23872315;
extern T0* ge1457ov23871818;
extern T0* ge1457ov23872309;
extern T0* ge1457ov23871978;
extern T0* ge1457ov23871812;
extern T0* ge1457ov23872310;
extern T0* ge1457ov23871813;
extern T0* ge1692ov27721740;
extern T0* ge1692ov27721741;
extern T0* ge397ov5996545;
extern T0* ge395ov5996545;
extern T0* ge382ov5996545;
extern T0* ge313ov5128215;
extern T0* ge313ov5128216;
extern T0* ge313ov5128217;
extern T0* ge1237ov20267019;
extern T0* ge1233ov20202551;
extern T0* ge1233ov20202552;
extern T0* ge1233ov20202532;
extern T0* ge1233ov20202548;
extern T0* ge1233ov20202536;
extern T0* ge1233ov20202537;
extern T0* ge1233ov20202538;
extern T0* ge1233ov20202539;
extern T0* ge1233ov20202533;
extern T0* ge1233ov20202540;
extern T0* ge1233ov20202541;
extern T0* ge1233ov20202542;
extern T0* ge1233ov20202543;
extern T0* ge1233ov20202534;
extern T0* ge1233ov20202549;
extern T0* ge1233ov20202544;
extern T0* ge1233ov20202545;
extern T0* ge1233ov20202546;
extern T0* ge1233ov20202535;
extern T0* ge1233ov20202547;
extern T0* ge1233ov20202550;
extern T0* ge783ov12828684;
extern T0* ge834ov13664318;
extern T0* ge834ov13664316;
extern T0* ge792ov12976133;
extern T0* ge792ov12976134;
extern T0* ge792ov12976135;
extern T0* ge834ov13664307;
extern T0* ge834ov13664319;
extern T0* ge1723ov28229635;
extern T0* ge1723ov28229637;
extern T0* ge1259ov20627463;
extern T0* ge1233ov20202511;
extern T0* ge1233ov20202527;
extern T0* ge1233ov20202515;
extern T0* ge1233ov20202516;
extern T0* ge1233ov20202517;
extern T0* ge1233ov20202518;
extern T0* ge1233ov20202512;
extern T0* ge1233ov20202519;
extern T0* ge1233ov20202520;
extern T0* ge1233ov20202521;
extern T0* ge1233ov20202522;
extern T0* ge1233ov20202513;
extern T0* ge1233ov20202528;
extern T0* ge1233ov20202523;
extern T0* ge1233ov20202524;
extern T0* ge1233ov20202525;
extern T0* ge1233ov20202514;
extern T0* ge1278ov20938755;
extern T0* ge1233ov20202530;
extern T0* ge1233ov20202531;
extern T0* ge1233ov20202529;
extern T0* ge1266ov20742147;
extern T0* ge1606ov11517956;
extern T0* ge1591ov11517956;
extern T0* ge1597ov11517956;
extern T0* ge1596ov11517956;
extern T0* ge1595ov11517956;
extern T0* ge1586ov11517956;
extern T0* ge1585ov11517956;
extern T0* ge1603ov11517956;
extern T0* ge1600ov11517956;
extern T0* ge1616ov11517956;
extern T0* ge709ov11517956;
extern T0* ge708ov11517956;
extern T0* ge702ov11517956;
extern T0* ge706ov11517956;
extern T0* ge713ov11517956;
extern T0* ge701ov11517956;
extern T0* ge392ov6422538;
extern T0* ge379ov6209546;
extern T0* geic23887873;
extern T0* geic23887874;
extern T0* geic23887875;
extern T0* geic12845057;
extern void GE_init_const1(void);
extern void GE_init_const2(void);
extern void GE_init_const3(void);
extern void GE_init_const4(void);
extern void GE_init_const5(void);
extern void GE_init_const6(void);
extern void GE_init_const7(void);
extern void GE_init_const8(void);
extern void GE_init_const9(void);
extern void GE_init_const10(void);
extern void GE_init_const11(void);
extern void GE_init_const12(void);
extern void GE_init_const13(void);
extern void GE_init_const14(void);
extern void GE_init_const15(void);
extern void GE_init_const16(void);
extern void GE_init_const17(void);
extern void GE_init_const18(void);
extern void GE_init_const19(void);
extern void GE_init_const20(void);
extern void GE_init_const21(void);
extern void GE_init_const22(void);
extern void GE_init_const(void);

#ifdef __cplusplus
}
#endif


/*
	description:

		"C functions used to implement class C_DATE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_TIME_H
#define GE_TIME_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#if defined _XOPEN_UNIX || defined __USE_XOPEN || defined __USE_BSD
#define HAS_GETTIMEOFDAY
#else
#ifndef WIN32
#define HAS_GETTIMEOFDAY
#endif
#endif

#ifdef HAS_GETTIMEOFDAY
/*
	OpenBSD does not support 'ftime' unless -lcompat is specified.
	Use the recommended XSI (X/Open System Interface) extension instead.
*/
#include <sys/time.h>
#include <sys/types.h>
#define GE_ftime(p) gettimeofday((struct timeval*)(p),((void*)0))
#define GE_timebsz sizeof(struct timeval)
#define GE_timebmillitm(p) (((struct timeval*)(p))->tv_usec/1000)
#define GE_timebtime(p) ((struct timeval*)(p))->tv_sec
#else
#include <sys/timeb.h>
extern void ftime(struct timeb *); /* Needed for lcc-win32 */
#define GE_ftime(p) ftime((struct timeb*)(p))
#define GE_timebsz sizeof(struct timeb)
#define GE_timebmillitm(p) ((struct timeb*)(p))->millitm
#define GE_timebtime(p) ((struct timeb*)(p))->time
#endif

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_CECIL_H
#define EIF_CECIL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef void (*EIF_PROCEDURE)(EIF_REFERENCE, ...);
#define EIF_PROC EIF_PROCEDURE
typedef EIF_INTEGER_32 (*EIF_INTEGER_32_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Integer 32 bits */
#define EIF_INTEGER_FUNCTION EIF_INTEGER_32_FUNCTION
typedef EIF_REFERENCE (*EIF_REFERENCE_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Reference */
typedef EIF_POINTER (*EIF_POINTER_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Pointer */
typedef EIF_BOOLEAN (*EIF_BOOLEAN_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Boolean */
typedef EIF_CHARACTER_8 (*EIF_CHARACTER_8_FUNCTION)(EIF_REFERENCE, ...); /* Returns char */
#define EIF_CHARACTER_FUNCTION EIF_CHARACTER_8_FUNCTION
typedef EIF_REAL_64 (*EIF_REAL_64_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Double */
#define EIF_DOUBLE_FUNCTION EIF_REAL_64_FUNCTION
typedef EIF_REAL_32 (*EIF_REAL_32_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Real */
#define EIF_REAL_FUNCTION EIF_REAL_32_FUNCTION

extern EIF_PROCEDURE eif_procedure(char* rout, EIF_TYPE_ID cid);
#define eif_proc eif_procedure /* Use `eif_procedure' instead */
extern EIF_INTEGER_FUNCTION eif_integer_function(char* rout, EIF_TYPE_ID cid);
extern EIF_REFERENCE_FUNCTION eif_reference_function(char* rout, EIF_TYPE_ID cid);
extern EIF_POINTER_FUNCTION eif_pointer_function(char* rout, EIF_TYPE_ID cid);
extern EIF_BOOLEAN_FUNCTION eif_boolean_function(char* rout, EIF_TYPE_ID cid);
extern EIF_REAL_64_FUNCTION eif_real_64_function(char* rout, EIF_TYPE_ID cid);
#define eif_double_function(rout,cid) eif_real_64_function((rout),(cid))
extern EIF_REAL_32_FUNCTION eif_real_32_function(char* rout, EIF_TYPE_ID cid);
#define eif_real_function(rout,cid) eif_real_32_function((rout),(cid))
extern EIF_CHARACTER_8_FUNCTION eif_character_8_function(char* rout, EIF_TYPE_ID cid);
#define eif_character_function(rout,cid) eif_character_8_function((rout),(cid))
extern EIF_OBJECT eif_create(EIF_TYPE_ID cid);
extern EIF_TYPE_ID eif_type_id(char* type_string);
extern EIF_INTEGER eifaddr_offset(EIF_REFERENCE object, char *name, int * const ret);
#define eif_string(s) RTMS(s)
#define makestr(s,c) RTMS_EX((s),(c))
#define eifaddr(object,name,ret) ((void *) ((char *) object + eifaddr_offset (object, name, ret)))
#define eif_field(object,name,type) *(type *)(eifaddr(object,name, NULL)) /* Obsolete. Use "eif_attribute" instead. */
#define eif_attribute(object,name,type,ret) *(type *)(eifaddr(object,name,ret)) /* Returns the attribute of an object. Return status in "ret".*/

/* Miscellaneous useful functions. */

/* Make an Eiffel array from a C array:
 * `eif_array' is the direct reference to the Eiffel array.
 * `c_array' is the C array.
 * `nelts' the number of elements to copy in the eiffel array, it has to
 * be equal to `eif_array.count'.
 * type is an Eiffel type.
 */
#define eif_make_from_c(eif_array, c_array, nelts, type) \
	{ \
		EIF_REFERENCE area = eif_field(eif_array, "area", EIF_REFERENCE); \
		memcpy((type *)area, c_array, nelts * sizeof(type));\
	}

#define eif_type eiftype /* Dynamic type ID */
extern EIF_TYPE_ID eiftype(EIF_OBJECT object); /* Give dynamic type of EIF_OBJECT. Obsolete, use "eif_type_by_object". */
extern EIF_TYPE_ID eif_type_by_reference(EIF_REFERENCE object);
extern EIF_TYPE_INDEX eif_gen_param_id(EIF_TYPE_INDEX dftype, uint32_t pos);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_PLUG_H
#define EIF_PLUG_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int nstcall;	/* Nested call global variable: signals a nested call and
					 * trigger an invariant check in generated C routines  */

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement Thread support"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_THREADS_H
#define EIF_THREADS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef EIF_CECIL_H
#include "eif_cecil.h"
#endif
#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifdef GE_USE_THREADS
#ifndef GE_THREAD_H
#include "ge_thread.h"
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifndef GE_USE_THREADS

/*
 * Empty stubs for EiffelThread library so that it may be compiled against a non-multithreaded run-time.
 */

/* THREAD_ENVIRONMENT */
#define eif_thr_thread_id() NULL

/* THREAD_CONTROL */
#define eif_thr_yield()
#define eif_thr_join_all()

/* THREAD */
#define eif_thr_create_with_attr(current_obj, init_func, set_terminated_func, attr)
#define eif_thr_wait(obj,get_terminated)
#define eif_thr_wait_with_timeout(obj,get_terminated,tms) EIF_TRUE
#define eif_thr_last_thread() NULL
#define eif_thr_exit()

/* THREAD_ATTRIBUTES */
#define eif_thr_default_priority() 0
#define eif_thr_min_priority() 0
#define eif_thr_max_priority() 0

/* MUTEX */
#define eif_thr_mutex_create() NULL
#define eif_thr_mutex_lock(a_mutex_pointer)
#define eif_thr_mutex_trylock(a_mutex_pointer) EIF_TRUE
#define eif_thr_mutex_unlock(a_mutex_pointer)
#define eif_thr_mutex_destroy(a_mutex_pointer)

/* CONDITION_VARIABLE */
#define eif_thr_cond_create() NULL
#define eif_thr_cond_broadcast(a_cond_ptr)
#define eif_thr_cond_signal(a_cond_ptr)
#define eif_thr_cond_wait(a_cond_ptr,a_mutex_ptr)
#define eif_thr_cond_wait_with_timeout(a_cond_ptr,a_mutex_ptr,a_timeout) 1
#define eif_thr_cond_destroy(a_cond_ptr)

/* SEMAPHORE */
#define eif_thr_sem_create(a_count) NULL
#define eif_thr_sem_wait(a_sem_pointer)
#define eif_thr_sem_post(a_sem_pointer)
#define eif_thr_sem_trywait(a_sem_pointer) EIF_TRUE
#define eif_thr_sem_destroy(a_sem_pointer)

/* READ_WRITE_LOCK */
#define eif_thr_rwl_create() NULL
#define eif_thr_rwl_rdlock(a_read_write_lock_pointer)
#define eif_thr_rwl_unlock(a_read_write_lock_pointer)
#define eif_thr_rwl_wrlock(a_read_write_lock_pointer)
#define eif_thr_rwl_destroy(a_read_write_lock_pointer)

#else

/* THREAD_ENVIRONMENT */
#define eif_thr_thread_id() GE_thread_id()

/* THREAD_CONTROL */
#define eif_thr_yield() GE_thread_yield()
#define eif_thr_join_all() GE_thread_join_all()

/* THREAD */
#define eif_thr_create_with_attr(current_obj,init_func,set_terminated_func,attr) GE_thread_create_with_attr((current_obj),(void (*)(EIF_REFERENCE))(init_func),(void (*)(EIF_REFERENCE,EIF_BOOLEAN))(set_terminated_func),(EIF_THR_ATTR_TYPE*)(attr))
#define eif_thr_wait(obj,get_terminated) GE_thread_wait((obj),(EIF_BOOLEAN (*)(EIF_REFERENCE))(get_terminated))
#define eif_thr_wait_with_timeout(obj,get_terminated,tms) GE_thread_wait_with_timeout((obj),(EIF_BOOLEAN (*)(EIF_REFERENCE))(get_terminated),(tms))
#define eif_thr_last_thread() GE_last_thread_created()
#define eif_thr_exit() GE_thread_exit()

/* THREAD_ATTRIBUTES */
#define eif_thr_default_priority() GE_thread_default_priority()
#define eif_thr_min_priority() GE_thread_min_priority()
#define eif_thr_max_priority() GE_thread_max_priority()

/* MUTEX */
#define eif_thr_mutex_create() GE_mutex_create()
#define eif_thr_mutex_lock(a_mutex_pointer) GE_mutex_lock(a_mutex_pointer)
#define eif_thr_mutex_trylock(a_mutex_pointer) GE_mutex_try_lock(a_mutex_pointer)
#define eif_thr_mutex_unlock(a_mutex_pointer) GE_mutex_unlock(a_mutex_pointer)
#define eif_thr_mutex_destroy(a_mutex_pointer) GE_mutex_destroy(a_mutex_pointer)

/* CONDITION_VARIABLE */
#define eif_thr_cond_create() GE_condition_variable_create()
#define eif_thr_cond_broadcast(a_cond_ptr) GE_condition_variable_broadcast(a_cond_ptr)
#define eif_thr_cond_signal(a_cond_ptr) GE_condition_variable_signal(a_cond_ptr)
#define eif_thr_cond_wait(a_cond_ptr,a_mutex_ptr) GE_condition_variable_wait((a_cond_ptr),(a_mutex_ptr))
#define eif_thr_cond_wait_with_timeout(a_cond_ptr,a_mutex_ptr,a_timeout) GE_condition_variable_wait_with_timeout((a_cond_ptr),(a_mutex_ptr),(a_timeout))
#define eif_thr_cond_destroy(a_cond_ptr) GE_condition_variable_destroy(a_cond_ptr)

/* SEMAPHORE */
#define eif_thr_sem_create(a_count) GE_semaphore_create(a_count)
#define eif_thr_sem_wait(a_sem_pointer) GE_semaphore_wait(a_sem_pointer)
#define eif_thr_sem_post(a_sem_pointer) GE_semaphore_post(a_sem_pointer)
#define eif_thr_sem_trywait(a_sem_pointer) GE_semaphore_try_wait(a_sem_pointer)
#define eif_thr_sem_destroy(a_sem_pointer) GE_semaphore_destroy(a_sem_pointer)

/* READ_WRITE_LOCK */
#define eif_thr_rwl_create() GE_read_write_lock_create()
#define eif_thr_rwl_rdlock(a_read_write_lock_pointer) GE_read_write_lock_read_lock(a_read_write_lock_pointer)
#define eif_thr_rwl_unlock(a_read_write_lock_pointer) GE_read_write_lock_unlock(a_read_write_lock_pointer)
#define eif_thr_rwl_wrlock(a_read_write_lock_pointer) GE_read_write_lock_write_lock(a_read_write_lock_pointer)
#define eif_thr_rwl_destroy(a_read_write_lock_pointer) GE_read_write_lock_destroy(a_read_write_lock_pointer)

/* WEL */
#ifdef EIF_WINDOWS
#define eif_thr_create_wel_per_thread_data(a_size) GE_thread_create_wel_per_thread_data(a_size)
#endif

#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class CONSOLE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_CONSOLE_H
#define EIF_CONSOLE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_POINTER console_def(EIF_INTEGER file);
extern EIF_BOOLEAN console_eof(FILE* fp);
extern EIF_CHARACTER_8 console_separator(FILE* f);
extern void console_ps(FILE* f, char* str, EIF_INTEGER len);
extern void console_pr(FILE* f, EIF_REAL_32 number);
extern void console_pc(FILE* f, EIF_CHARACTER_8 c);
extern void console_pd(FILE* f, EIF_REAL_64 val);
extern void console_pi(FILE* f, EIF_INTEGER number);
extern void console_tnwl(FILE* f);
extern EIF_CHARACTER_8 console_readchar(FILE* f);
extern EIF_REAL_32 console_readreal(FILE* f);
extern EIF_INTEGER console_readint(FILE* f);
extern EIF_REAL_64 console_readdouble(FILE* f);
extern EIF_INTEGER console_readword(FILE* f, char* s, EIF_INTEGER bound, EIF_INTEGER start);
extern EIF_INTEGER console_readline(FILE* f, char* s, EIF_INTEGER bound, EIF_INTEGER start);
extern void console_next_line(FILE* f);
extern EIF_INTEGER console_readstream(FILE* f, char* s, EIF_INTEGER bound);
extern void console_file_close (FILE* f);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class FILE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_FILE_H
#define EIF_FILE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <time.h>
#include <sys/stat.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Let's define the stat structure for our platforms. */
/* The definition is the same for both ANSI and Unicode versions on Windows. */
#ifdef EIF_WINDOWS
#	ifdef __LCC__
#		define rt_stat_buf	struct stat
#	elif defined EIF_64_BITS
#		define rt_stat_buf	struct _stat64
#	else
#		define rt_stat_buf	struct _stat64i32
#	endif
#else
#	define rt_stat_buf		struct stat
#endif

/*
 * Create directory `dirname'.
 */
extern void eif_file_mkdir(EIF_FILENAME dirname);

/*
 * Rename file `from' into `to'.
 */
extern void eif_file_rename(EIF_FILENAME from, EIF_FILENAME  to);

/*
 * Link file `from' into `to'.
 */
extern void eif_file_link(EIF_FILENAME from, EIF_FILENAME to);

/*
 * Delete file or directory `name'.
 */
extern void eif_file_unlink(EIF_FILENAME name);

/*
 * Open file `name' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_open(EIF_FILENAME name, int how);

/*
 * Open file `fd' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_dopen(int fd, int how);

/*
 * Reopen file `name' with the corresponding type `how' and substitute that
 * to the old stream described by `old'. This is useful to redirect 'stdout'
 * to another place, for instance.
 */
extern EIF_POINTER eif_file_reopen(EIF_FILENAME name, int how, FILE *old);

/*
 * Close the file.
 */
extern void eif_file_close(FILE *fp);

/*
 * Flush data held in stdio buffer.
 */
extern void eif_file_flush(FILE *fp);

/*
 * Return the associated file descriptor.
 */
extern EIF_INTEGER eif_file_fd(FILE *f);

/*
 * Get a character from `f'.
 */
extern EIF_CHARACTER_8 eif_file_gc(FILE *f);

/*
 * Get a string from `f' and fill it into `s' (at most `bound' characters),
 * with `start' being the amount of bytes already stored within s. This
 * means we really have to read (bound - start) characters.
 */
extern EIF_INTEGER eif_file_gs(FILE *f, char *s, EIF_INTEGER bound, EIF_INTEGER start);

/*
 * Read min (bound, remaining bytes in file) characters into `s' and
 * return the number of characters read.
 */
extern EIF_INTEGER eif_file_gss(FILE *f, char *s, EIF_INTEGER bound);

/*
 * Get a word from `f' and fill it into `s' (at most `bound' characters),
 * with `start' being the amount of bytes already stored within s. This
 * means we really have to read (bound - start) characters. Any leading
 * spaces are skipped.
 */
extern EIF_INTEGER eif_file_gw(FILE *f, char *s, EIF_INTEGER bound, EIF_INTEGER start);

/*
 * Look ahead one character. If EOF, return 0.
 */
extern EIF_CHARACTER_8 eif_file_lh(FILE *f);

/*
 * Size of file `fp'.
 */
extern EIF_INTEGER eif_file_size(FILE *fp);

/*
 * Read upto next input line.
 */
extern void eif_file_tnil(FILE *f);

/*
 * Current position within file.
 */
extern EIF_INTEGER eif_file_tell(FILE *f);

/*
 * Touch file `name' by setting both access and modification time to the
 * current time stamp. This external function exists only because there
 * is no way within UNIX_FILE to get the current time stamp. Otherwise,
 * we could simply call file_utime.
 */
extern void eif_file_touch(EIF_FILENAME name);

/*
 * Modify the modification and/or the access time stored in the file's
 * inode. The 'how' parameter tells which attributes should be set.
 */
extern void eif_file_utime(EIF_FILENAME name, time_t stamp, int how);

/*
 * This is an encapsulation of the stat() system call. The routine either
 * succeeds and returns or fails and raises the appropriate exception.
 */
extern int eif_file_stat(EIF_FILENAME path, rt_stat_buf *buf, int follow);

/*
 * Change permissions of file `name', using an interface like chmod(1).
 * The flag is true if permissions are to be added, 0 to remove them.
 */
extern void eif_file_perm(EIF_FILENAME name, char *who, char *what, int flag);

/*
 * Change permission mode on file `path'.
 */
extern void eif_file_chmod(EIF_FILENAME path, int mode);

/*
 * Change the owner of the file to `uid'.
 */
extern void eif_file_chown(EIF_FILENAME name, int uid);

/*
 * Change the group of the file to `gid'.
 */
extern void eif_file_chgrp(EIF_FILENAME name, int gid);

/*
 * Put new_line onto `f'.
 */
extern void eif_file_tnwl(FILE *f);

/*
 * Append a copy of `other' to `f'.
 */
extern void eif_file_append(FILE *f, FILE *other, EIF_INTEGER l);

/*
 * Write string `str' on `f'.
 */
extern void eif_file_ps(FILE *f, char *str, EIF_INTEGER len);

/*
 * Write character `c' on `f'.
 */
extern void eif_file_pc(FILE *f, char c);

/*
 * Go to absolute position `pos' counted from start.
 */
extern void eif_file_go(FILE *f, EIF_INTEGER pos);

/*
 * Go to absolute position `pos' counted from end.
 */
extern void eif_file_recede(FILE *f, EIF_INTEGER pos);

/*
 * Go to absolute position `pos' counted from current position.
 */
extern void eif_file_move(FILE *f, EIF_INTEGER pos);

/*
 * End of file.
 */
extern EIF_BOOLEAN eif_file_feof(FILE *fp);

/*
 * Test whether file exists or not. If `name' represents a symbolic link,
 * it will check that pointed file does exist.
 */
extern EIF_BOOLEAN eif_file_exists(EIF_FILENAME name);

/*
 * Test whether file exists or not without following the symbolic link
 * if `name' represents one.
 */
extern EIF_BOOLEAN eif_file_path_exists(EIF_FILENAME name);

/*
 * Check whether access permission `op' are possible on file `name' using
 * real UID and real GID. This is probably only useful to setuid or setgid
 * programs.
 */
extern EIF_BOOLEAN eif_file_access(EIF_FILENAME name, EIF_INTEGER op);

/*
 * Check whether the file `path' may be created: we need write permissions
 * in the parent directory and there must not be any file bearing that name
 * with no write permissions...
 */
extern EIF_BOOLEAN eif_file_creatable(EIF_FILENAME path, EIF_INTEGER nbytes);

/*
 * Get an integer from `f'.
 */
extern EIF_INTEGER eif_file_gi(FILE *f);

/*
 * Get a real from `f'.
 */
extern EIF_REAL_32 eif_file_gr(FILE *f);

/*
 * Get a double from `f'.
 */
extern EIF_REAL_64 eif_file_gd(FILE *f);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pi(FILE *f, EIF_INTEGER number);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pr(FILE *f, EIF_REAL_32 number);

/*
 * Write double `val' onto `f'.
 */
extern void eif_file_pd(FILE *f, EIF_REAL_64 val);

/*
 * Size of the stat structure. This is used by the Eiffel side to create
 * the area (special object) which will play the role of a stat buffer
 * structure.
 */
extern EIF_INTEGER stat_size(void);

/*
 * Check file permissions using effective UID and effective GID. The
 * current permission mode is held in the st_mode field of the stat()
 * buffer structure `buf'.
 */
extern EIF_BOOLEAN eif_file_eaccess(rt_stat_buf *buf, int op);

/*
 * Perform the field dereferencing from the appropriate stat structure,
 * which Eiffel cannot do directly.
 */
extern EIF_INTEGER eif_file_info(rt_stat_buf *buf, int op);

/*
 * Return the Eiffel string filled in with the name associated with `uid'
 * if found in /etc/passwd. Otherwise, return fill it in with the numeric
 * value.
 */
extern EIF_REFERENCE eif_file_owner(int uid);

/*
 * Return the Eiffel string filled in with the name associated with `gid'
 * if found in /etc/group. Otherwise, return fill it in with the numeric
 * value.
 */
extern EIF_REFERENCE eif_file_group(int gid);

/*
 * Get an integer from `f'.
 */
extern EIF_INTEGER eif_file_gib(FILE* f);

/*
 * Get a real from `f'.
 */
extern EIF_REAL_32 eif_file_grb(FILE* f);

/*
 * Get a double from `f'.
 */
extern EIF_REAL_64 eif_file_gdb(FILE* f);

/*
 * Open file `name' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_binary_open(EIF_FILENAME name, int how);

/*
 * Open file `fd' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_binary_dopen(int fd, int how);

/*
 * Reopen file `name' with the corresponding type `how' and substitute that
 * to the old stream described by `old'. This is useful to redirect 'stdout'
 * to another place, for instance.
 */
extern EIF_POINTER eif_file_binary_reopen(EIF_FILENAME name, int how, FILE* old);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pib(FILE* f, EIF_INTEGER number);

/*
 * Write `number' on `f'.
 */
extern void eif_file_prb(FILE* f, EIF_REAL_32 number);

/*
 * Write double `val' onto `f'.
 */
extern void eif_file_pdb(FILE* f, EIF_REAL_64 val);

/*
 * Modification time of a file.
 * Seconds since epoch (01 January 1970) in UTC or 0 if time cannot be retrieved.
 */
extern EIF_INTEGER eif_file_date(EIF_FILENAME  name);

/*
 * Access time of a file.
 * Seconds since epoch (01 January 1970) in UTC or 0 if time cannot be retrieved.
 */
extern EIF_INTEGER eif_file_access_date(EIF_FILENAME name);

/*
 * Generate a temporary file with a name based on a template and
 * return a file descriptor to the file.
 * The template is overwritten with the name of the new file.
 * Return a non-zero file descriptor for the temporary file upon success, -1 otherwise.
 * `template'is the template to use for creating the temporary file.
 * It must match the rules for mk[s]temp (i.e. end in "XXXXXX").
 * On exit, it is updated with the new name.
 * `is_text_mode', if non-zero, then the temporary file is created
 * in text mode, otherwise in binary mode.<param>
 * Note: Code was inspired from https://github.com/mirror/mingw-w64/blob/master/mingw-w64-crt/misc/mkstemp.c
 */
extern EIF_INTEGER eif_file_mkstemp(EIF_FILENAME a_template, EIF_BOOLEAN is_text_mode);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class EXCEPTIONS"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_EXCEPT_H
#define EIF_EXCEPT_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Predefined exception codes.
 */
#define EN_VOID		GE_EX_VOID			/* Feature applied to void reference */
#define EN_MEM		GE_EX_MEM			/* No more memory */
#define EN_PRE		GE_EX_PRE			/* Pre-condition violated */
#define EN_POST		GE_EX_POST			/* Post-condition violated */
#define EN_FLOAT	GE_EX_FLOAT			/* Floating point exception (signal SIGFPE) */
#define EN_CINV		GE_EX_CINV			/* Class invariant violated */
#define EN_CHECK	GE_EX_CHECK			/* Assertion violated */
#define EN_FAIL		GE_EX_FAIL			/* Routine failure */
#define EN_WHEN		GE_EX_WHEN			/* Unmatched inspect value */
#define EN_VAR		GE_EX_VAR			/* Non-decreasing loop variant */
#define EN_LINV		GE_EX_LINV			/* Loop invariant violated */
#define EN_SIG		GE_EX_SIG			/* Operating system signal */
#define EN_BYE		GE_EX_BYE			/* Eiffel run-time panic */
#define EN_RESC		GE_EX_RESC			/* Exception in rescue clause */
#define EN_OMEM		GE_EX_OMEM			/* Out of memory (cannot be ignored) */
#define EN_RES		GE_EX_RES			/* Resumption failed (retry did not succeed) */
#define EN_CDEF		GE_EX_CDEF			/* Create on deferred */
#define EN_EXT		GE_EX_EXT			/* External event */
#define EN_VEXP		GE_EX_VEXP			/* Void assigned to expanded */
#define EN_HDLR		GE_EX_HDLR			/* Exception in signal handler */
#define EN_IO		GE_EX_IO			/* I/O error */
#define EN_SYS		GE_EX_SYS			/* Operating system error */
#define EN_RETR		GE_EX_RETR			/* Retrieval error */
#define EN_PROG		GE_EX_PROG			/* Developer exception */
#define EN_FATAL	GE_EX_FATAL			/* Eiffel run-time fatal error */
#define EN_DOL		GE_EX_DOL			/* $ applied to melted feature */
#define EN_ISE_IO	GE_EX_ISE_IO		/* I/O error raised by the ISE Eiffel runtime */
#define EN_COM		GE_EX_COM			/* COM error raised by EiffelCOM runtime */
#define EN_RT_CHECK	GE_EX_RT_CHECK		/* Runtime check error such as out-of-bound array access */
#define EN_OLD		GE_EX_OLD			/* Old violation */
#define EN_SEL		GE_EX_SEL			/* Serialization failure */
#define EN_DIRTY	GE_EX_DIRTY			/* SCOOP processor dirty exception. */
#define EN_NEX		GE_EX_NEX			/* Number of internal exceptions */

/*
 * Raise an Eiffel exception.
 */
extern void eraise(const char* name, long code);

/*
 * Raise an Eiffel exception of the given code with no associated tag.
 */
extern void xraise(int code);

/*
 * Raise an "Operating system error" exception.
 */
extern void esys(void);

/*
 * As a special case, an I/O error is raised when a system call which is I/O bound fails.
 */
extern void eise_io(const char* tag);

/*
 * Raise a "No more memory" exception.
 */
extern void enomem(void);

/*
 * Raise EiffelCOM exception.
 */
extern void com_eraise(const char* tag, long num);

/*
 * Terminate execution with exit status `code',
 * without triggering an exception.
 */
extern void esdie(int code);

/*
 * Exception tag associated with `code'.
 * This is a duplication from Eiffel classes, but still used for trace printing and in EiffelCom.
 */
extern EIF_REFERENCE eename(long except);

/*
 * Is exception `ex' defined?
 * Used in EiffelCOM.
 */
extern char eedefined(long ex);

/*
 * Enable/diable printing of the exception trace.
 * Per thead information.
 */
extern void eetrace(char b);

/*
 * Is current execution during rescue?
 */
extern EIF_BOOLEAN eif_is_in_rescue(void);

#ifdef EIF_WINDOWS
/*
 * Set default exception handler.
 */
extern void set_windows_exception_filter(void);
#endif

/*
 * Used in EiffelCOM.
 */
#define echval GE_current_context()->exception_code
#define echtg GE_current_context()->exception_tag

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class REAL"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_REAL_H
#define GE_REAL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <math.h>

#ifndef GE_power
#define GE_power(x,y) pow((x),(y))
#endif
#define GE_ceiling(x) ceil(x)
#define GE_floor(x) floor(x)

extern EIF_NATURAL_32 GE_real_32_to_bits (EIF_REAL_32 v);
extern EIF_NATURAL_64 GE_real_64_to_bits (EIF_REAL_64 v);
extern EIF_REAL_32 GE_real_32_from_bits (EIF_NATURAL_32 v);
extern EIF_REAL_64 GE_real_64_from_bits (EIF_NATURAL_64 v);

#define GE_real_32_is_nan(v) ((v) != (v) ? EIF_TRUE : ((GE_real_32_to_bits(v) & ~GE_nat32(0x80000000)) > GE_nat32(0x7FF00000)))
#define GE_real_64_is_nan(v) ((v) != (v) ? EIF_TRUE : ((GE_real_64_to_bits(v) & ~GE_nat64(0x8000000000000000)) > GE_nat64(0x7FF0000000000000)))
#define GE_real_32_is_negative_infinity(v) (GE_real_32_to_bits(v) == GE_nat32(0xFF800000))
#define GE_real_64_is_negative_infinity(v) (GE_real_64_to_bits(v) == GE_nat64(0xFFF0000000000000))
#define GE_real_32_is_positive_infinity(v) (GE_real_32_to_bits(v) == GE_nat32(0x7F800000))
#define GE_real_64_is_positive_infinity(v) (GE_real_64_to_bits(v) == GE_nat64(0x7FF0000000000000))
#define GE_real_32_nan GE_real_32_from_bits(GE_nat32(0x7FC00000))
#define GE_real_64_nan GE_real_64_from_bits(GE_nat64(0x7FF8000000000000))
#define GE_real_32_negative_infinity GE_real_32_from_bits(GE_nat32(0xFF800000))
#define GE_real_64_negative_infinity GE_real_64_from_bits(GE_nat64(0xFFF0000000000000))
#define GE_real_32_positive_infinity GE_real_32_from_bits(GE_nat32(0x7F800000))
#define GE_real_64_positive_infinity GE_real_64_from_bits(GE_nat64(0x7FF0000000000000))
#define GE_real_32_is_equal(v1,v2) ((v1) == (v1) ? (v1) == (v2) : (v2) != (v2))
#define GE_real_64_is_equal(v1,v2) ((v1) == (v1) ? (v1) == (v2) : (v2) != (v2))
#define GE_real_32_is_less(v1,v2) ((v1) == (v1) ? (v1) < (v2) : (v2) == (v2))
#define GE_real_64_is_less(v1,v2) ((v1) == (v1) ? (v1) < (v2) : (v2) == (v2))
#define GE_real_32_ieee_is_equal(v1,v2) (v1) == (v2)
#define GE_real_64_ieee_is_equal(v1,v2) (v1) == (v2)
#define GE_real_32_ieee_is_greater(v1,v2) isgreater(v1,v2)
#define GE_real_64_ieee_is_greater(v1,v2) isgreater(v1,v2)
#define GE_real_32_ieee_is_greater_equal(v1,v2) isgreaterequal(v1,v2)
#define GE_real_64_ieee_is_greater_equal(v1,v2) isgreaterequal(v1,v2)
#define GE_real_32_ieee_is_less(v1,v2) isless(v1,v2)
#define GE_real_64_ieee_is_less(v1,v2) isless(v1,v2)
#define GE_real_32_ieee_is_less_equal(v1,v2) islessequal(v1,v2)
#define GE_real_64_ieee_is_less_equal(v1,v2) islessequal(v1,v2)
#define GE_real_32_ieee_maximum_number(v1,v2) fmaxf(v1,v2)
#define GE_real_64_ieee_maximum_number(v1,v2) fmax(v1,v2)
#define GE_real_32_ieee_minimum_number(v1,v2) fminf(v1,v2)
#define GE_real_64_ieee_minimum_number(v1,v2) fmin(v1,v2)

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2016, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_GLOBALS_H
#define EIF_GLOBALS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif
#ifndef EIF_THREADS_H
#include "eif_threads.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_THREADS
#define EIF_GET_CONTEXT \
	GE_context* eif_globals = GE_current_context();
#else
#define EIF_GET_CONTEXT
#endif

#define GTCX EIF_GET_CONTEXT

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C declarations for the ISE Eiffel runtime."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2016, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_EIFFEL_H
#define EIF_EIFFEL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_STRING_H
#include "ge_string.h"
#endif
#ifndef EIF_GLOBALS_H
#include "eif_globals.h"
#endif
#ifndef EIF_EXCEPT_H
#include "eif_except.h"
#endif

#endif
/*
	description:

		"C functions used to access DLLs"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_DLL_H
#define GE_DLL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef EIF_WINDOWS

#include <windows.h>

/* DLL declarations */
#define GE_load_dll(name) LoadLibraryA((LPCSTR)(name))

#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class EXECUTION_ENVIRONMENT"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_MISC_H
#define EIF_MISC_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_DLL_H
#include "ge_dll.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_INTEGER eif_system(EIF_NATIVE_CHAR* s);
extern void eif_system_asynchronous(EIF_NATIVE_CHAR* cmd);
extern void eif_sleep(EIF_INTEGER_64);

#ifdef EIF_WINDOWS
/* DLL declarations */
#define eif_load_dll(name) GE_load_dll(name)
#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class DIRECTORY"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_DIR_H
#define EIF_DIR_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_POINTER eif_dir_open(EIF_FILENAME dirname);
extern EIF_POINTER eif_dir_next(EIF_POINTER dir);
extern EIF_POINTER eif_dir_rewind(EIF_POINTER d, EIF_FILENAME dir_name);
extern void eif_dir_close(EIF_POINTER dir);
extern EIF_BOOLEAN eif_dir_exists(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_readable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_writable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_executable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_deletable(EIF_FILENAME name);
extern EIF_CHARACTER_8 eif_dir_separator(void);
extern EIF_INTEGER eif_dir_current(EIF_FILENAME a_buffer, EIF_INTEGER a_count);
extern EIF_INTEGER eif_chdir(EIF_FILENAME path);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class COM_FAILURE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef GE_COM_FAILURE_H
#define GE_COM_FAILURE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_INTEGER GE_ccom_hresult(EIF_POINTER an_exception_code);

extern EIF_INTEGER GE_ccom_hresult_code(EIF_INTEGER an_hresult);

extern EIF_INTEGER GE_ccom_hresult_facility(EIF_INTEGER an_hresult);

/*
 * Get text from error `a_code'. It is up to the caller to free
 * the returned buffer using `GE_ccom_local_free'.
 */
extern EIF_POINTER GE_ccom_error_text(EIF_INTEGER a_code);

/*
 * Number of characters in `ptr'.
 */
extern EIF_INTEGER_32 GE_ccom_strlen(EIF_POINTER ptr);

/*
 * Number of bytes occupied by a TCHAR.
 */
extern EIF_INTEGER_32 GE_ccom_character_size(void);

/*
 * Free `a_ptr' using LocalFree.
 */
extern void GE_ccom_local_free(EIF_POINTER a_ptr);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		""Part of ISE Eiffel runtime. Needed to compile class EXECUTION_ENVIRONMENT"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2020-01-08 10:24:11 +0000 (Wed, 08 Jan 2020) $"
	revision: "$Revision: 103831 $"
*/

#ifndef EIF_SYSTEM_H
#define EIF_SYSTEM_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef EIF_WINDOWS
#include "Winbase.h"
#elif defined(EIF_VMS) || defined(EIF_MACOSX)
#include "sysctl.conf"
#else
#include <unistd.h>
#endif

#endif
#include <string.h>
#include <stdlib.h>
#include <time.h>
